module CADGen where

import LMath

import Control.Exception
import Control.Monad
import Data.Char
import Data.Maybe
import System
import System.IO

data CParams = CParams { cpVars :: Integer,
                         cpParams :: Integer,
                         cpCtx :: QfFormula (Polynomial Rational),
                         cpDoms :: [QfFormula (Polynomial Rational)]
                       }
             deriving Show

data Method = CAD | FM deriving (Eq, Show)

getNextLine :: IO String
getNextLine = do
    l <- getLine
    case dropWhile isSpace l of
       '#':_ -> getNextLine
       ""    -> getNextLine
       line  -> return line

getInt :: IO Integer
getInt = do
    l <- getNextLine
    return (read l)

getFormula :: Integer -> IO (QfFormula (Polynomial Rational))
getFormula nIndets = do
    l <- getNextLine
    case parseFormula (Exactly nIndets) l of
        Right f   -> return f
        Left  err -> error ("Could not parse '" ++ l ++ "':\n" ++ err)

parseInput :: IO CParams
parseInput = do
    nV <- getInt
    nP <- getInt
    ctx <- getFormula (nV+nP)
    nF <- getInt
    doms <- mapM (\_ -> getFormula (nV+nP)) [1..nF]
    return $ CParams { cpVars = nV, cpParams = nP, cpCtx = ctx, cpDoms = doms }

handleToolIOError :: IO a -> IO ()
handleToolIOError action =
    catchDyn (action >> return ())
             (\(tio :: ToolIOError) ->
                 do hPutStrLn stderr ("Tool IO error: " ++ show tio)
             )

revVarsAndParams :: Num a => Integer -> Polynomial a -> Polynomial a
revVarsAndParams nV p =
    createPolyFromList $ [(co,f es) | (co,es) <- polyToList p]
    where
    f es = reverse es1 ++ reverse es2
        where
        (es1,es2) = splitAt (fromIntegral nV) es

bycad :: CParams -> String
bycad cp =
    code2text (cpParams cp) (paramNames ++ varNames) $
    codeSimpl $
    cadgen fullctx fulltispcs
    where
    paramNames = ['x':show i | i <- [cpVars cp .. cpVars cp+cpParams cp-1]]
    varNames = ['x':show i | i <- [0..cpVars cp - 1]]
    fullctx = fmap (revVarsAndParams nV) $ cpCtx cp
    fulltispcs = map (fmap (revVarsAndParams nV)) $ cpDoms cp
    nV = cpVars cp

byfm :: CParams -> GIneqSystem (Polynomial Rational) -> String
byfm cp sys =
    dtreeSorted2code varNames paramNames "S1;" $
    makeDTree (fracPolyDTreeContext () [ctx]) $
    fourierMotzkin (fmap (%%1) sys) Nothing Nothing
    where
    paramNames = ['x':show i | i <- [cpVars cp .. cpVars cp+cpParams cp-1]]
    varNames = ['x':show i | i <- [0..cpVars cp - 1]]
    ctx = fmap (deleteIndeterminates (cpVars cp)) (cpCtx cp)

extractFMinput :: CParams -> Maybe (GIneqSystem (Polynomial Rational))
extractFMinput cp
    | length (cpDoms cp) /= 1  =  Nothing
    | otherwise = case formulaToIneqSystems (cpVars cp) (cpDoms cp !! 0) of
                    Just [sys] -> Just sys
                    _          -> Nothing
    

output :: String -> IO ()
output str =
    handleToolIOError $ putStrLn str

main :: IO ()
main = do
    args <- getArgs
    when (length args /= 2) $
         usage
    let reduceCmd = args!!0
    method <- case map toLower (args!!1) of
                     "cad" -> return CAD
                     "fm"  -> return FM
                     _     -> usage >> return undefined
    setReduceCommand reduceCmd
    input <- parseInput
    let mbFmInput = extractFMinput input
        Just fmInput = mbFmInput
    when (method == FM && not (isJust mbFmInput)) $
         do hPutStrLn stderr "For FM, there must be exactly 1 conjuction of inequalities."
            exitFailure
    hPutStrLn stderr ("Input parses as: " ++ show input ++ "\n")
    putStrLn ("/* Code generated with method " ++ show method ++ " */")
    output $ case method of
               CAD -> bycad input
               FM  -> byfm  input fmInput
    where
    usage :: IO () 
    usage =
        do pn <- getProgName
           hPutStrLn stderr ("USAGE: " ++ pn ++ " REDUCE-PATH  fm|cad")
           exitFailure

