module CfFADA
    ( run, writeNewDeps,
      module CfFADA.GraphToDot,
      module LoopoDS
    )
where

import LoopoDS
import LMath

import CfFADA.FilledCfGraph
import CfFADA.CfFADA
import CfFADA.GraphToList
import CfFADA.GraphToDot

-- |
-- Run the CfFADA analysis.
--
run :: FilePath   -- ^ The path to the '.src.out' file
    -> AccessType -- ^ The desired destination access type
    -> AccessType -- ^ The desired source access type
    -> Bool       -- ^ Whether or not to optimize the resulting dependences
    -> IO ()      -- ^ The complete dependence data
run basename dat sat optimize =
    do deps <- runCfFADA basename dat sat optimize
       let p []     = putStr ""
           p (x:xs) = do print x
                         putStrLn ""
                         putStrLn ""
                         p xs
       p deps

-- |
-- Run the CfFADA analysis and write the output to the '.src.out.dep_new' file
--
writeNewDeps :: FilePath   -- ^ The path to the '.src.out' file
             -> AccessType -- ^ The desired destination access type
             -> AccessType -- ^ The desired source access type
             -> Bool       -- ^ Whether or not to optimize the resulting dependences
             -> IO ()      -- ^ The complete dependence data
writeNewDeps basename dat sat optimize =
    do deps <- runCfFADA basename dat sat optimize
       let getDepData (aid, sid, (ed1, ed2, sd, depd)) = depAccesses depd
           addParInfo (a1,a2) (b1,b2) = (a1+b1,a2)
           relDeps   = map getDepData deps

           -- Shifting is needed because different CfFADA runs can
           -- introduce different new parameters that have the same index,
           -- so we shift them behind the new parameters from previous runs
           (np,fidx) = foldl addParInfo (0,0) $ map (getParamInfo True) relDeps
           shifted   = if (np,fidx) == (0,0) then relDeps
                       else shiftDependences relDeps np fidx 0

           ldeps     = depAccessToLoopoDep 1 (dat, sat) np (foldl (++) [] shifted)
       print ldeps
       writeDependences (basename ++ ".dep_new") ldeps
       putStrLn ("Wrote file: " ++ basename ++ ".dep_new")

runCfFADA :: FilePath
          -> AccessType
          -> AccessType
          -> Bool
          -> IO [(ArrayID,
                  StatementID,
                  ([GraphID FilledCfgEdgeData],
                   [GraphID FilledCfgEdgeData],
                   [SrcStatementData],
                   DependenceData))]
runCfFADA basename dat sat optimize =
    do (stmt,accs) <- loadStatementsAndAccesses basename
       slp         <- loadSurroundingLoops basename
       sif         <- loadSurroundingIfs basename
       cfg         <- loadCfGraph basename
       let graphTraversal destAcc srcAccType =
               traverseCfGraph cfg stmt accs slp destAcc srcAccType
           computeDeps destAccType srcAccType opt =
               [(accArrayID destAcc,
                 accStmtID destAcc,
                 cffada destAcc (graphTraversal destAcc srcAccType) sif srcAccType opt
                ) | destAcc <- accs, accType destAcc == destAccType
               ]
--            trueDeps   = computeDeps Read  Write True
--            outputDeps = computeDeps Write Write True
--            antiDeps   = computeDeps Write Read  False
--            inputDeps  = computeDeps Read  Read  True
       return (computeDeps dat sat optimize)

----------------
-- Test cases --
----------------

ex = run "CfFADA/testdata/piter/piter.src.out" Read Write
exw = run "CfFADA/testdata/piter/piter.src.out" Read Write
ext = run "CfFADA/testdata/trivial/trivial.src.out"
exe = run "CfFADA/testdata/extreme/extreme.src.out"
exs = run "CfFADA/testdata/strange/strange.src.out"
exi = run "CfFADA/testdata/script/script.src.out"
exm = run "CfFADA/testdata/matmult-systol/matmult-systol.src.out"
exu = run "CfFADA/testdata/mult/mult.src.out"
exu2 = run "CfFADA/testdata/mult2/mult2.src.out"
ex3 = run "CfFADA/testdata/mult3/mult3.src.out" Read Write
exw3 = writeNewDeps "CfFADA/testdata/mult3/mult3.src.out" Read Write
ex4 = run "CfFADA/testdata/mult4/mult4.src.out" Read Write
exw4 = writeNewDeps "CfFADA/testdata/mult4/mult4.src.out" Read Write
exa = run "CfFADA/testdata/across/across.src.out"
exd = showDot "CfFADA/testdata/piter/piter.src.out"
exf = writeDot "CfFADA/testdata/piter/piter.src.out"

