-- |
-- Control-flow based Fuzzy Array Data Analysis.
-- Based on the paper 'A Precise Fixpoint Reaching Definition for Arrays'
-- by Jean-Francois Collard and Martin Griebl.
--
module CfFADA.CfFADA
    (cffada,
     depAccessToLoopoDep,
     getParamInfo,
     shiftDependences,
     StatementData(..),
     DependenceAccess(..),
     DependenceData(..),
     SrcStatementData
    )
where

import List (nub)

import LoopoDS
import LMath

import CfFADA.FilledCfGraph
import CfFADA.GraphToList
import CfFADA.GraphToDot
import CfFADA.IterationVector as IV

-- Holds the data of the source statements needed for the analysis later.
--
-- Fields:
-- reqs:    The accIndexFct of the source access, the accIndexFct of the
--          destination access and the resulting conflict equation as an
--          IneqSystem
-- reqsis:  An IneqSystem consisting of the above conflict equation and the
--          loop constraints of both the source and the destination access
-- predics: The surrounding ifs of the statement, with the StatementIDs of
--          the if-cases in the first list and the else-cases in the second
data StatementData = StatData {
                         reqs    :: (AffineFunction, AffineFunction, IneqSystem),
                         reqsis  :: IneqSystem,
                         predics :: ([StatementID],[StatementID])
                     }
instance Show StatementData where
    show (StatData r ris p) = "(" ++ show(r) ++ ", " ++ show(ris) ++ ", " ++ show(p) ++ ")\n\n"

-- The resulting dependence data.
--
-- Fields:
-- sid:     The StatementID of the access/statement
-- ivec:    The IterationVector where this dependence belongs
-- dQuast:  The IneqSystem joined together from the reqsis of the source
--          StatementData and the IterationVector constraints, and the Quast
--          generated from it
-- dRange:  A list with all the result data from the Quast. First the
--          h-transformation as an AffineFunction, then a list with the
--          new parameters together with their index, and then the
--          constraints posed upon the Quast result as a list of
--          IneqSystems.
-- predifs: The if-else cases needed for the access to be the real source
--          of a dependence, with the tuple just like in predics of the
--          StatementData and the h-transformations indicating the
--          corresponding loop instance
data DependenceAccess = DepAccess {
                            ssid    :: StatementID,
                            dsid    :: StatementID,
                            ivec    :: IterationVector,
                            dQuast  :: (IneqSystem, GQuast QVector Rational Integer),
                            dRange  :: (AffineFunction, [(Integer, GVector Integer, Integer)], ZDomain),
                            predifs :: [(([StatementID], [StatementID]), AffineFunction)]
                        }
instance Show DependenceAccess where
    show (DepAccess ssid dsid i q r p) = "(SourceSID: " ++ show(ssid) ++ ", DestSID: " ++ show(dsid) ++ ",\n"
                                         ++ "IterationVector: " ++ show(i) ++ ",\n"
                                         ++ show(q) ++ ",\n"
                                         ++ show(r) ++ ",\n"
                                         ++ show(p) ++ ")\n\n"

-- A list of the resulting DependenceAccesses.
data DependenceData = DepData {
                          depAccesses :: [DependenceAccess]
                      }
instance Show DependenceData where
    show (DepData a) = "\n\n(Accesses: " ++ show(a) ++ "\n\n,)"

-- The StatementData together with its StatementID for easy use in the
-- 'lookup' function.
type SrcStatementData = (StatementID, StatementData)

-- |
-- The main function that delegates the separate steps of the algorithm to
-- specialized subfunctions.
--
cffada :: Access                      -- ^ The destination access
       -> [GraphID FilledCfgEdgeData] -- ^ The graph traversal as a list of nodes
       -> SurroundingIfs              -- ^ A list with the surrounding ifs
       -> AccessType                  -- ^ The access type that is searched for in the source accesses
       -> Bool                        -- ^ Whether to optimize the dependences or not
       -> ([GraphID FilledCfgEdgeData],[GraphID FilledCfgEdgeData],[SrcStatementData],DependenceData)
                                      -- ^ The graph traversal in reverse order,
                                      --   the graph traversal after filtering for the relevant edges,
                                      --   a list with the StatementDatas,
                                      --   the actual dependences
cffada destAcc traversal sif acct opt =
    (reverse traversal,
     relEdgeList,
     srcStatements,
     DepData {
         depAccesses = if opt then split
                       else adjusted
     }
    )
    where
    relEdgeList = filterEdges (reverse traversal) acct (accArrayID destAcc)
    srcStatements = getSrcStatements destAcc [] relEdgeList sif
    merged = mergeDependences srcStatements (accStmtID destAcc) relEdgeList [] (IV [])

    (newParameters,firstPIdx) = getParamInfo False merged

    adjusted = adjustDependences merged newParameters firstPIdx 0 sif []
    optimized = optimizeDependences adjusted sif []
    split = splitDependences optimized
        where
        splitDependences []     = []
        splitDependences (x:xs) = map split' zd ++ splitDependences xs
            where
            (a,p,zd) = dRange x
            split' zp = x { dRange = (a,p,[zp]) }

depAccessToLoopoDep :: Integer
                    -> (AccessType, AccessType)
                    -> Integer
                    -> [DependenceAccess]
                    -> [Dependence]
depAccessToLoopoDep _ _ _ []                           = []
depAccessToLoopoDep curDepID (dat, sat) numPars (x:xs) =
    Dependence {
        depID         = DependenceID curDepID,
        depType       = deptype,

                        -- this is just a (more or less) temporary workaround
                        -- since it doesn't guarantee an integral result.
        depSrcDomain  = image af ((zpRealImage (head zdom)) `multijoin` parIneqs),

        depDestDomain = zpRealImage (head zdom) `multijoin` parIneqs,
        depSrcStmtID  = ssid x,
        depDestStmtID = dsid x,
        depHTrafo     = af,
        depContext    = reverse $ genContext numPars
    } : depAccessToLoopoDep (curDepID + 1) (dat, sat) numPars xs
    where
    deptype = case v == um of
                  True  -> case (dat, sat) of
                               (Read,  Write) -> UniformTrue
                               (Write, Read)  -> UniformAnti
                               (Read,  Read)  -> UniformInput
                               (Write, Write) -> UniformOutput
                  False -> case (dat, sat) of
                               (Read,  Write) -> NonUniformTrue
                               (Write, Read)  -> NonUniformAnti
                               (Read,  Read)  -> NonUniformInput
                               (Write, Write) -> NonUniformOutput
        where
        um      = unitMatrix (nRows v) (nColumns v)
        (v,_,_) = partsVPCih af
    (af, pars, zdom) = dRange x
    parIneqs = parsToIneqs pars
    parsToIneqs :: [(Integer, GVector Integer, Integer)] -> [IneqSystem]
    parsToIneqs []           = []
    parsToIneqs ((i,v,d):xs) =
        ineqSystem ((nVars af, nParams af), m) : parsToIneqs xs
        where
        nv = setVectorEntry v (nVars af + i) (-d)
        l  = vectorLength nv
        c  = vectorEntry nv (l - 1)
        iv = - setVectorEntry nv (l - 1) (c - d + 1)
        m  = rowVectorsToMatrix1 [fmap toRational nv, fmap toRational iv]
    genContext 0 = []
    genContext n = Context {
                       parN = n,
                       stride = 1%1,
                       parDescription =
                           ineqSystem ((1, (nParams af) - numPars),
                                       matrix (0, 1 + (nParams af) - numPars + 1, [])
                                      )
                   } : genContext (n - 1)

-- Get the StatementID of a statement belonging to an edge
getStatementID :: FilledCfgEdgeData -> StatementID
getStatementID (FilledCfgEdgeData a (Just (b, c, Just d))) = b

getParamInfo :: Bool
             -> [DependenceAccess]
             -> (Integer, Integer)
getParamInfo filter deps =
    if length parlist == 0 then (0,0)
    else (fromIntegral $ l, ((\(a,b,c) -> a) $ head parlist))
    where
    parlist = foldl (++) [] (getPars deps)
    l = if filter then length $ nub parlist
        else length parlist
    getPars []     = []
    getPars (x:xs) = p : getPars xs
        where
        (_,p,_) = dRange x

-- Generate the complete data of the dependences for every IterationVector
-- instance and put them in a list. This data is not optimized yet.
mergeDependences :: [SrcStatementData]          -- ^ A list with the data of the single source statements
                 -> StatementID                 -- ^ The statement id of the destination access
                 -> [GraphID FilledCfgEdgeData] -- ^ The (filtered) graph traversal
                 -> [DependenceAccess]          -- ^ The resulting list of the dependence accesses, initialized with an empty list
                 -> IterationVector             -- ^ The current IterationVector, also initialized with an empty one
                 -> [DependenceAccess]          -- ^ A list with all possible dependences before optimization
mergeDependences stmts destsid edges depaccs iv =
    case edges of
        []                   -> depaccs
        ((GEdge x):xs)       -> case quast of
                                    QNoResult -> mergeDependences stmts destsid xs depaccs iv
                                    _         -> mergeDependences
                                                     stmts
                                                     destsid
                                                     xs
                                                     (depaccs ++ newaccs)
                                                     iv
                                where
                                (is,quast) = genQuast stmts (getStatementID (edgeData x)) iv
                                ranges     = quastToRanges quast (nVars is) [] []
                                newaccs    = map buildNewAcc ranges
                                buildNewAcc range = DepAccess {
                                                        ssid    = getStatementID (edgeData x),
                                                        dsid    = destsid,
                                                        ivec    = iv,
                                                        dQuast  = (is,quast),
                                                        dRange  = range,
                                                        predifs = []
                                                    }
        ((GIterVector x):xs) -> mergeDependences stmts destsid xs depaccs x

-- Adjust the data of the dependences for new parameters introduces by the
-- mathematical analysis in the Quast.
adjustDependences :: [DependenceAccess] -- ^ The list of the dependence accesses
                  -> Integer            -- ^ The number of the new parameters
                  -> Integer            -- ^ The first index value of all parameters
                  -> Integer            -- ^ The current parameter offset, initialized with 0
                  -> SurroundingIfs     -- ^ The surrounding ifs of the accesses
                  -> [DependenceAccess] -- ^ An empty list that will hold the adjusted dependences
                  -> [DependenceAccess] -- ^ A list with the adjusted dependences
adjustDependences deps np fidx curpoffs sif adj =
    case deps of
        []     -> adj
        (x:xs) -> adjustDependences
                      xs
                      np
                      fidx
                      (curpoffs + (fromIntegral $ length p))
                      sif
                      (adj ++ [newacc])
                  where
                  (_,p,_) = dRange x
                  newacc = x {
                               dRange = newrange,
                               predifs = newpreds
                           }
                  newrange = if np == 0 then (dRange x)
                             else adjustRange (dRange x) np fidx curpoffs (fromIntegral $ length p)
                  (na,_,_) = newrange
                  newpreds = case (ifs,elses) of
                                 ([],[]) -> []
                                 _       -> [((ifs,elses), na)]
                  (ifs,elses) = lookupSurroundingIfs sif (ssid x)

-- Part of the adjustDependences function; adjust the dRange data field of
-- a dependence.
adjustRange :: (AffineFunction,[(Integer,GVector Integer,Integer)],ZDomain) -- ^ The dRange field of a dependence
            -> Integer                                                               -- ^ The number of new parameters
            -> Integer                                                               -- ^ The first index value of all parameters
            -> Integer                                                               -- ^ The current parameter offset
            -> Integer                                                               -- ^ The length of the item to offset to
            -> (AffineFunction,[(Integer,GVector Integer,Integer)],ZDomain) -- ^ The adjusted dRange field
adjustRange (a,p,z) np fidx co len =
    (adjustAF a, map adjustP p, map adjustZP z)
    where
    adjustP (n,v,d) = (n + co,
                       insertEntries
                           (insertEntries
                                v
                                (nVars a + fidx)
                                (take (fromIntegral co) $ repeat 0))
                           (nVars a + fidx + co + len)
                           (take (fromIntegral (np - co - len))
                            $ repeat 0),
                       d
                      )
    adjustZP zp = zpInsertZeroParams
                      (zpInsertZeroParams zp fidx co)
                      (fidx + co + len)
                      (np - co - len)
    adjustAF af = extendAF
                      (partsVPCih (extendAF (partsVPCih af) fidx co))
                      (fidx + co + len)
                      (np - co - len)
    extendAF (v,p,c) idx o = affineFunctionVPC
                                 (v, insertZeroColumns p idx o, c)
                                 Inhomogeneous

shiftDependences :: [[DependenceAccess]]
                 -> Integer
                 -> Integer
                 -> Integer
                 -> [[DependenceAccess]]
shiftDependences [] _ _ _          = []
shiftDependences (x:xs) np fidx co =
    shiftDependences' x np fidx co (fst $ getParamInfo True x)
    : shiftDependences xs np fidx (co + (fst $ getParamInfo True x))
    where
    shiftDependences' [] _ _ _ _     = []
    shiftDependences' (x:xs) n i c l = x {
                                           dRange = adjustRange oldrange n i c l
                                       } : shiftDependences' xs n i c l
                                       where
                                       oldrange = dRange x

-- Optimize the dependence accesses. This means eliminating accesses that
-- will definitely get overwritten by later accesses, adjusting the range
-- fields of accesses that will get only partially overwritten and add
-- if-else data if they will get only overwritten under certain
-- circumstances.
optimizeDependences :: [DependenceAccess] -- ^ The list of dependence accesses to optimize
                    -> SurroundingIfs     -- ^ The list with the surrounding ifs
                    -> [DependenceAccess] -- ^ The list with the optimized dependences, initialized with an empty list
                    -> [DependenceAccess] -- ^ A list with the optimized dependences
optimizeDependences deps sif opt =
    case deps of
        []     -> opt
        (x:xs) -> optimizeDependences
                      xs
                      sif
                      (opt ++ [optresult])
                  where
                  optresult = foldl (optimize sif) x opt

-- Get all the data of the source statements/accesses for use later
-- in the analysis.
getSrcStatements :: Access                      -- ^ The destination access
                 -> [SrcStatementData]          -- ^ A list with the resulting statements, initialized with an empty list
                 -> [GraphID FilledCfgEdgeData] -- ^ The (filtered) graph traversal
                 -> SurroundingIfs              -- ^ The surrounding ifs
                 -> [SrcStatementData]          -- ^ A list with the resulting statements
getSrcStatements destAcc stmts list sif =
    case list of
        []                   -> stmts
        ((GEdge x):xs)       -> case lookup (getStatementID (edgeData x)) stmts of
                                    Nothing -> getSrcStatements
                                                   destAcc
                                                   (stmts ++ [genStatementData x])
                                                   xs
                                                   sif
                                    Just s  -> getSrcStatements destAcc stmts xs sif
        ((GIterVector x):xs) -> getSrcStatements destAcc stmts xs sif
    where
    getAccess (FilledCfgEdgeData a (Just (b, c, Just d))) = d
    genStatementData edge =
        (getStatementID (edgeData edge),
         StatData {
             reqs    = (saif,
                        daif,
                        genStatementEquation saif daif
                       ),
             reqsis  = genReqData (genStatementEquation saif daif)
                                  (accDomain $ getAccess $ edgeData edge)
                                  (accDomain destAcc),
             predics = lookupSurroundingIfs sif (getStatementID (edgeData edge))
         }
        )
        where
        saif = accIndexFct $ getAccess $ edgeData edge
        daif = accIndexFct destAcc
        -- Converts an equation of a source accIndexFct
        -- and a destination accIndexFct like
        --  i' + j' + 1 = i + j
        -- to an IneqSystem
        --  i' + j' - i - j + 1 (>= 0)
        -- -i' - j' + i + j - 1 (>= 0)
        -- with i and j as parameters.
        -- Already existing parameters are preserved.
        genStatementEquation saif daif =
            ineqSystem ((nVars saif, nVars daif + nParams saif), m)
            where
            (vs,ps,cs) = partsVPCih saif
            (vd,pd,cd) = partsVPCih daif
            m = multijoin1 $ map multiconc1
                    [[ vs,-vd,  ps-pd ,  cs-cd ],
                     [-vs, vd,-(ps-pd),-(cs-cd)]
                    ]

-- Filters out the edges accessing the same array as the destination access
-- and having the requested accType.
filterEdges :: [GraphID FilledCfgEdgeData] -- ^ The graph traversal
            -> AccessType                  -- ^ The requested access type of the source statements
            -> ArrayID                     -- ^ The requested array id
            -> [GraphID FilledCfgEdgeData] -- ^ The filtered graph traversal\/edge list
filterEdges list acct' aid =
    case list of
        []                                     -> []
        ((GIterVector x):xs)                   -> (GIterVector x) : filterEdges xs acct' aid
        ((GEdge x):xs)
            | hasAccess (edgeData x) acct' aid -> (GEdge x)       : filterEdges xs acct' aid
            | otherwise                        ->                   filterEdges xs acct' aid
    where
    hasAccess (FilledCfgEdgeData a (Just (b, c, Just d))) acct' aid = (accType d == acct')
                                                                      && (accArrayID d == aid)
    hasAccess (FilledCfgEdgeData a _) acct' aid                     = False

-- This funcion takes a list of the source statement data and uses it to
-- generate a Quast for the current one with the data from the reqsis field
-- and the current iteration vector. This Quast holds most of the
-- information we are looking for in a dependence.
genQuast :: [SrcStatementData]                            -- ^ The list with the source statement data
         -> StatementID                                   -- ^ The statement id of the current statement
         -> IterationVector                               -- ^ The current iteration vector
         -> (IneqSystem, GQuast QVector Rational Integer) -- ^ The complete IneqSystem used for the generation of the Quast
                                                          --   and the actual Quast
genQuast stmts sid iv =
    case (lookup sid stmts) of
        Just a  -> (is a, pipLexmax (is a) Nothing True)
        Nothing -> error "genQuast: Source statement not found, this shouldn't happen"
    where
    is stmt = foldl join (reqsis stmt) (genPredIS iv nv)
        where
        nv = nVars (reqsis stmt)
        np = nParams (reqsis stmt)
        genPredIS (IV [])     0 = []
        genPredIS (IV (x:xs)) n =
            case x of
                IV.Equal -> ineqSystem ((nv,np), join l (-l))
                            : genPredIS (IV xs) (n - 1)
                            where
                            l = m `conc` (-m) `conc` (zeroMatrix 1 (np - nv)) `conc` (zeroMatrix 1 1)
                            m = setEntry (zeroMatrix 1 nv) 0 (nv - n) 1
                IV.Lower -> ineqSystemVPC ((-m), conc m (zeroMatrix 1 (np - nv)), matrix [[-1]])
                            : genPredIS (IV xs) (n - 1)
                            where
                            m = setEntry (zeroMatrix 1 nv) 0 (nv - n) 1
                IV.Star  -> genPredIS (IV xs) (n - 1)

-- Parse the Quast to get the actual result, the constraints for it and
-- the new parameters as separate structures.
quastToRanges :: GQuast QVector Rational Integer     -- ^ The Quast to parse
              -> Integer                             -- ^ The original number of variables
              -> [GVector Integer]                   -- ^ A temporary list with the Quast results as vectors,
                                                     --   initialized with an empty list
              -> [(Integer,GVector Integer,Integer)] -- ^ A temporary list with the new parameters,
                                                     --   initialized with an empty list
              -> [(AffineFunction, [(Integer, GVector Integer, Integer)], ZDomain)]
                                                     -- ^ The dRange field for the current dependence data
quastToRanges quast nv vecs newpars =
    case quast of
        QNoResult         -> []
        QIf v q1 q2       -> (quastToRanges q1 nv (vecs ++ [v]) newpars) ++ (quastToRanges q2 nv (vecs ++ [v]) newpars)
        QList l           -> [(affineFunction ((nv, numparsa - nv), lm) Inhomogeneous,
                              map changePars newpars,
                              [zpolyhedron
                                  (ineqSystem ((nv, numparsi - nv), vm))
                                  (affineFunction ((nv, numparsa - nv), lm) Inhomogeneous)
                              ]
                             )]
                             where
                             lvecs = parseQList l
                             lm = rowVectorsToMatrix (length l) lvecs
                             numparsa = if length lvecs == 0 then 0
                                        else (nColumns lm) - 1
                             vm = fmap fromIntegral (rowVectorsToMatrix (length vecs) vecs)
                             numparsi = if length vecs == 0 then 0
                                        else (nColumns vm) - 1
                             changePars (i, v, d) = (i - nv, v, d)
        QNewParam n v d q -> quastToRanges q nv (map extendVec vecs) ((map extendPar newpars) ++ [(fromIntegral n, extendVec v, fromIntegral d)])
                             -- extend all vectors collected up to now with a column for the
                             -- new parameter, even the vector of the new parameter itself
                             where
                             extendVec vec = insertEntries vec (fromIntegral n) [0]
                             extendPar (pn,pv,pd) = (pn, extendVec pv, pd)
    where
    parseQList []                    = []
    parseQList ((QFinite x):xs)      = x : parseQList xs
    parseQList ((QNegInfinity x):xs) =     parseQList xs
    parseQList ((QPosInfinity x):xs) =     parseQList xs

-- This function takes the conflict equation of a source statement data and
-- the loop constraints of the source statement and the destination
-- statement, and joins them together to form the reqsis field of the
-- StatementData structure.
genReqData :: IneqSystem -- ^ The conflict equation
           -> IneqSystem -- ^ The loop constraints of the source access
           -> IneqSystem -- ^ The loop constraints of the destination access
           -> IneqSystem -- ^ The complete inequality system
genReqData seq sad dad = join seq (extDomain (partsVPC sad) (partsVPC dad))
    where
    extDomain (sv,sp,sc) (dv,dp,dc) =
        ineqSystemVPC (join sv dfill,
                       join (conc sfill spr) (conc dv dpr),
                       join sc dc
                      )
        where
        sfill = zeroMatrix (nRows sv) (nColumns sv)
        dfill = zeroMatrix (nRows dv) (nColumns dv)
        -- necessary because of superfluous zero columns generated by old cffada
        spr = deleteZeroColumns sp
        dpr = deleteZeroColumns dp

-- Part of the optimizeDependences function; the actual optimization step
-- that adjusts the dRange field for later accesses to the same indexes and
-- adds the if-else cases for accesses that will not definitely get
-- overwritten.
optimize :: SurroundingIfs   -- ^ The list of surrounding ifs
         -> DependenceAccess -- ^ The new dependence access to optimize
         -> DependenceAccess -- ^ The current old dependence access to compare to
         -> DependenceAccess -- ^ The optimized new dependence access
optimize sif n o =
    case (oif,oelse) of
        ([],[]) -> n {
                       dRange = newrange range
                   }
                   where
                   range = dRange n
                   newrange (na,np,nz) = (na, np, zdDisjointUnionDifference nz oz)
                   (_,_,oz) = dRange o
        _       -> n {
                       predifs = npreds ++ [((oelse,oif), oa)]
                   }
                   where
                   npreds = predifs n
                   (oa,_,_) = dRange o
    where
    (oif,oelse) = lookupSurroundingIfs sif (ssid o)

-- Delete columns that only consist of zeros.
deleteZeroColumns :: Num a
                  => GMatrix a -- ^ The matrix to adjust
                  -> GMatrix a -- ^ The adjusted matrix
deleteZeroColumns = mtranspose . matrix . filter isNotZero . columns'
    where
    isNotZero = not . and . map (== 0)

