-- |
-- Generates a '.dot' file of the given graph
-- for use with the 'dot' program from Graphviz.
--
module CfFADA.GraphToDot(showDot, writeDot)
where

import LoopoDS

-- | Print dot markup to stdout
showDot :: FilePath -> IO ()
showDot basename =
    do dotList <- generateDot basename
       putStrLn "digraph CfGraph {"
       printDotList dotList
       putStrLn "}"

-- | Write file 'FilePath ++ \".dot\"'
writeDot :: FilePath -> IO ()
writeDot basename =
    do dotList <- generateDot basename
       writeFile (basename ++ ".dot") (generateFileString dotList)
       putStrLn ("Wrote file: " ++ (basename ++ ".dot"))

generateDot :: FilePath -> IO [String]
generateDot basename =
    do cfg <- loadCfGraph basename
       let edgeInfoList = getEdgeInfoList cfg
           lineList     = [list | list <- map getEdgeLine edgeInfoList]
           dataList     = [list | list <- map getEdgeData edgeInfoList]
           combinedList = sortEdgeList $ combineLists lineList dataList
           dotList      = [list | list <- map getDotString combinedList]
       return dotList

getEdgeInfoList :: CfGraph -> [EdgeInfo CfgEdgeData]
getEdgeInfoList (AGraph a b) = b

getEdgeLine :: EdgeInfo a -> (EdgeID,NodeID,NodeID)
getEdgeLine (EdgeInfo id d n1 n2) = (id,n1,n2)

getEdgeData :: EdgeInfo CfgEdgeData -> String
getEdgeData (EdgeInfo id (CfgEdgeData d1 Nothing d3) n1 n2)   = d3
getEdgeData (EdgeInfo id (CfgEdgeData d1 (Just d2) d3) n1 n2) = "[" ++ show(d2) ++ "] " ++ d3

combineLists :: [(EdgeID,NodeID,NodeID)] -> [String] -> [(NodeID,NodeID,String)]
combineLists [] []                 = []
combineLists ((e,n1,n2):ns) (s:ss) = (n1,n2,"(" ++ show(e) ++ ") " ++ s)
                                     : combineLists ns ss

sortEdgeList :: [(NodeID,NodeID,String)] -> [(NodeID,NodeID,String)]
sortEdgeList []              = []
sortEdgeList ((x1,x2,x3):xs) = sortEdgeList [(y1,y2,y3) | (y1,y2,y3) <- xs, y1 <= x1]
                               ++ [(x1,x2,x3)]
                               ++ sortEdgeList [(y1,y2,y3) | (y1,y2,y3) <- xs, y1 > x1]

getDotString :: (NodeID,NodeID,String) -> String
getDotString (n1,n2,s) = show(n1) ++ " -> " ++ show(n2)
                         ++ " [fontsize=12, label = " ++ show(s) ++ "]"

printDotList :: [String] -> IO ()
printDotList []     = return ()
printDotList (x:xs) = do putStrLn ("    " ++ x)
                         printDotList xs

generateFileString :: [String] -> String
generateFileString [] = ""
generateFileString l  = "digraph CfGraph {\n" ++ recString l ++ "}\n"
                        where recString []     = ""
                              recString (x:xs) = "    " ++ x ++ "\n"
                                                 ++ recString xs

