-- -*- Mode:Haskell -*-
---
-- Module for generating communication code
-- 
-- @author Tobias Langhammer
-- @version $Id: CommCode.hs 3389 2006-11-19 13:05:22Z groessli $
--

module CommCode (main, mcSenderCode) where
import LMath
import DepAcc (srcAcc,acpsToAccs)
import Tiling(defaultTiledLoops,TileInformation(..),
              defaultTile,
              test_l -- just for testing
             )
import LMath.LinAlg.Matrix(showPlainMatrix,getMatrix, 
			   matrix, rows') 
import LMath.LinAlg.AffineFunction(AffineFunction)
import LMath.LinAlg.LClasses(Applicable(apply))
import LMath.LinAlg.Vector(Vector)
import LoopoDS
import IOExts(unsafePerformIO,trace) -- just for testing
import List
import LMath.LinAlg.AffineExp(nVars,nParams,toAffineExp
		,substParam)
import CodeGen(loopNest, getBoundText)
import LMath.LinAlg.IneqSystem(IneqSystem, ineqSystem, simplifyTrivially, 
		  disjUnionAll, nondisjointUnion, sortIneqs, projectToVars,
		  ToolMethod(PipMethod, DefaultMethod))
import LMath.Logic.FM(projectToVarsFM)
import Ratio -- MC 2003-07-10

import IO(openFile, hGetContents, IOMode(..))
import System(getArgs)
import System.Cmd(system)
import GHC.IOBase(ExitCode(..))
import CppInterface
import Foreign.C.String
import Maybe
import Ix

---
-- Data structure containing labeled fields holding information
-- for generating input for CLooG directly for each statement.
data CloogStatement = 		    
    -- Used for "preparing" a send action, i.e. for filling
    -- send buffers:
    PrepareSend { cloogIS :: IneqSystem, 
		  globalTimeDim :: Integer, 
		  realProcDim :: Integer,   
		  coordsDim :: Integer, 
		  depNr :: DependenceID,
		  tiling :: Bool } |

    -- Used for "preparing" a receive action, i.e. for reading
    -- from receive buffers:
    PrepareReceive { cloogIS :: IneqSystem, 
		     globalTimeDim :: Integer, 
		     realProcDim :: Integer,   
		     coordsDim :: Integer, 
		     depNr :: DependenceID,
		     tiling :: Bool } |

    -- Used for a computation statement in the target program:
    Compute { cloogIS :: IneqSystem,
	      ispcDim :: Integer,
	      globalTimeDim :: Integer, 
	      realProcDim :: Integer,   
	      coordsDim :: Integer, 
	      trStatementNr :: TStatementID,
	      includeCommunicationDimensions :: Bool,
	      tiling :: Bool } |

    -- Used for the communication statement (currently all-to-all):
    Communicate { cloogIS :: IneqSystem,
		  globalTimeDim :: Integer,
		  realProcDim :: Integer,
		  coordsDim :: Integer,
		  tiling :: Bool }
    deriving Show


---
-- Data structure containing data needed for generating 
-- an abstract representation of an actual statement.
data CloogStmtData = 
    CloogStmtData {
		   trDepData :: TrDependence, 
		   trStmtNrData :: TStatementID,
		   origStmtData :: Statement,
		   invTrafoData :: AffineFunction,
		   linCombEqsData :: IneqSystem,
		   trcConstrData :: IneqSystem,
		   latticeData :: [Vector],
		   timeDimData :: Integer,
		   useTilingData :: Bool,
		   useCommunicationData :: Bool
		  }
    deriving Show

---
-- Data structure which holds all cloog-statements, some
-- additional information concerning tiling and/or communication
-- and information about the loop "quality vectors"
data CloogStmts = CloogStmts { comps :: [CloogStatement],
			       sends :: [CloogStatement],
			       recvs :: [CloogStatement],
			       comms :: [CloogStatement],
			       computeStmtsOnly :: Bool,
			       indexNames :: [String],
			       blobNames :: [String],
			       qualityVectors :: [(TStatementID, [QualEntry])]
			     }

---
-- Data structure containing "raw" data structures from
-- which the data for all cloog statements can be extracted from.
data CloogDS = 
    CloogDS { ispcsDS :: [Statement],
	      tispcsDS :: [TStatement],
              trDepsDS :: [TrDependence],
	      invTrafosDS :: [(TStatementID, AffineFunction)],
	      linCombEqsDS :: [(TStatementID, IneqSystem)],
	      trcsDS :: [TRC],
	      latticeDS :: [Vector],
	      timeDimDS :: Integer,
	      useTilingDS :: Bool,
	      useCommunicationDS :: Bool
	    }
    deriving Show


simpleTest :: Bool -> IO ()
simpleTest onScreen = 
    let 
      lattice = [vector [1,0], vector[0,1]]
    in do
       mctestmain "exs/simple/simple.src.out" 1 lattice onScreen


lattice_2dim = [vector [3,0], vector [0,3%1]]

get_lattice_2dim size = [vector [size%1, 0%1], vector [0%1, size%1]]

large_2dim = [vector [b, 0], vector [0%1, b]]
    where
    b = 1000%1

getNonTilingLattice = []

myCode tempDir projFile timeDimension latticeVec = 
    generateBackendCodeOnly tempDir 
			    projFile 
			    -- timeDimension 
			    latticeVec 
			    False 
			    "../cloog/cloog-0.12.2/bin/cloog"
			    "../loopo/targout/targout"




main :: IO ()
main = 
    do [tempDir, projFile, latString, cloogExec, targoutExec] <- getArgs
       let latticeIntLists = (read latString) :: [[Integer]]
       let latticeRatLists = map (map (\x -> x % 1)) latticeIntLists
       let lattice = 
	       trace (latString ++ (show latticeIntLists)) $
	       if length latticeIntLists > 0
	       then Just (map vector latticeRatLists)
	       else Nothing
       putStrLn $ latString
       generateBackendCodeOnly tempDir projFile lattice 
			       False -- onScreen
			       cloogExec targoutExec
       
---
-- Generates only the sequentially runnable code from a given
-- project file. Tiling technique can be used optionally.
-- @param tempDir LooPo's tempDir
-- @param projFile a string containing the name of the project file
--                 e.g.: "example10_2.src.out"
-- @param latticeVec a list of lattice vectors for tiling. If tiling is
--                   not used, use "Nothing" instead of "Just ..."
-- @param onScreen True for output on screen only. False for output into 
--                 an output-file.
-- @param cloogExec the cloog executable
-- @param targoutExec LooPo's targout executable
generateBackendCodeOnly :: String -> String -> Maybe [Vector] -> 
			   Bool -> String -> String -> IO ()
generateBackendCodeOnly tempDir projFile latticeVec 
			onScreen cloogExec targoutExec =
    do
    if onScreen then putStr cloogCode 
                else do
		     putStrLn $ "writing cloog input into: " ++ 
				  (loopoFile ++".cloog")
		     writeFile (loopoFile ++".cloog") cloogCode
		     putStrLn $ "cloog output: " ++ 
				  (loopoFile ++".cloogOutput")
		     runCloog cloogExec (loopoFile ++ ".cloog") 
				  (loopoFile ++ ".cloogOutput")
		     createDefinesFile targoutExec tempDir projFile
    where

    timeDimension = getTimeDim (loopoQUAL loopoDS)
    loopoFile = tempDir ++ "/" ++ projFile
    cloogCode = generateCloogInput cloogStmts
    useTiling = isJust latticeVec

    -- "cloogIndices" are names for loop indices used by cloog for
    -- the source index space (the original names cannot be used, because
    -- after scanning one generated loop can represent different 
    -- source indices)
    cloogIndices = ["cloog" ++ (show i) | i <- [1 .. (length indices)]]

    (indices, blobs) = getIndicesAndBlobs tempDir projFile
    lat = maybe (getNonTilingLattice) id latticeVec

    loopoDS = 
	trace "reading loopoDS..." $
	unsafeReadLoopoDS loopoFile
    cloogDS' = 
	   CloogDS { 
		    ispcsDS = error "not yet defined", 
		    tispcsDS = error "not yet defined", 
		    trDepsDS = error "not yet defined",
		    invTrafosDS = error "not yet defined", 
		    linCombEqsDS = error "not yet defined", 
		    trcsDS = error "not yet defined",
                    latticeDS = lat,
		    timeDimDS = timeDimension,
		    useTilingDS = useTiling,
		    useCommunicationDS = False
		   }

    cloogDS = 
	completeCloogDS loopoDS cloogDS'

    cloogStmts = 
	CloogStmts { comps = computeStmts,
		     comms = [],
		     sends = [],
		     recvs = [],
		     computeStmtsOnly = True,
		     indexNames = cloogIndices,
		     blobNames = blobs,
		     qualityVectors = loopoQUAL loopoDS
		   }

    computeStmts = 
	--trace "generating compute stmts..." $
	map createComputeStatement (createComputeStmtData cloogDS)
	
---
-- Completes a given CloogDS with values from a given LoopoDS.
-- Some preparations are applied like reducing to surrounding loops.
completeCloogDS :: LoopoDS -> CloogDS -> CloogDS
completeCloogDS loopoDS cloogDS = 
    cloogDS {
	     ispcsDS = ispcs, 
	     tispcsDS = tispcs,
	     trDepsDS = trdeps, 
	     invTrafosDS = invtrafos,
	     linCombEqsDS = lincombs,
	     trcsDS = trc
	    }			     
    where     
    ispcs = map ({-reduceStmt-}trimIndexSpace slp) (loopoISPC loopoDS)
    tispcs = loopoTISPC loopoDS
    trdeps = transformDeps tispcs trafos deps 
    invtrafos = makeProperInvTrafos (loopoTrafo loopoDS)
    lincombs = makeProperLinCombEquations (loopoTrafo loopoDS)
    trc = loopoTRC loopoDS

    deps = map ({-reduceDep-}trimDependence slp) (loopoDEP loopoDS)
    slp  = loopoSLP loopoDS
    trafos = makeProperTrafos (loopoTrafo loopoDS)

---
-- Counts the number of time dimensions in the target program from the
-- 
-- @param qualVects the quality vectors for each target statement
-- @return timeDim, the number of schedule dimensions in the target program
getTimeDim :: [(TStatementID, [QualEntry])] -> Integer
getTimeDim qualVects = timeDim
    where
    (_, qualEntries) = unzip qualVects
    schedEntries = map (filter (\e -> e == Sched)) qualEntries
    timeDimNumbers = map genericLength schedEntries
    timeDim = 
	if ((length $ group timeDimNumbers) == 1) 
	then head timeDimNumbers
	else error ("number of schedule dimensions not equal for " ++
                    "all target statements!")

---
-- Creates a file containing "#define" expressions for the target 
-- program. These are used for mapping the generated index variables 
-- of CLooG to the used variables in the array accesses and substituting
-- the source statements in the placeholder generated by CLooG.
-- @param targoutExec the executable of LooPo's targout module used for
--                    generating the original statment code.
-- @param tempDir the temporary directory containing the intermediate 
--                files used by LooPo/HS-LooPo
-- @param projFile the filename of the project file
createDefinesFile :: String -> String -> String -> IO ()
createDefinesFile targoutExec tempDir projFile =
    do stmtsNrAndText <-
	   getTextForStatements targoutExec 
				tempDir 
				projFile 
				langFile
       let defines = (generateAllStmtDefines 
		      origToTrStmts stmtsNrAndText indexMappings)    
       writeFile (loopoFile ++ suffix) defines	      
    where
    loopoFile = tempDir ++ "/" ++ projFile
    suffix = ".defines"
    langFile = "c_stmtsOnly.lang"
    loopoDS = unsafeReadLoopoDS loopoFile
    slp  = loopoSLP loopoDS
    (indices, blobs) = getIndicesAndBlobs tempDir projFile
    cloogIndices = ["cloog" ++ (show i) | i <- [1 .. (length indices)]]
    indexMappings = generateAllSurrLoopsIndices indices cloogIndices slp
    (trStmtNrs, origStmtNrs, _) = unzip3 (loopoTISPC loopoDS)
    origToTrStmts = zip origStmtNrs trStmtNrs

---
-- Returns a list of identifiers used for the scatter-dimensions in the
-- target program. These identifiers contain information about the 
-- kind of each loop (parallel or sequential) and about time/processor
-- loops or virtuall processor loops.
getScatterNames :: CloogStmts -> [String]
getScatterNames allStmts = names
    where

    commonQualVects = 
	map ((genericTake $ getCommonCoordsDim allStmts).snd) 
		(qualityVectors allStmts)
    qualVect = 
	if ((length $ group commonQualVects) == 1) 
	then head commonQualVects
	else error "quality vectors have different prefixes!"

    numPar = length $ filter (== Place) qualVect

    parPrefix = "par"
    seqPrefix = "seq"

    tileTimePrefix = "globalTime"
    tileRealProcPrefix = "realProc"

    coordTimePrefix = 
	if useTiling then "vTime" else "time"
    coordProcPrefix =
	if useTiling then "vProc" else "proc"
    bl = "bl"
    stmtType = "stmtType" 

    useComm = not $ computeStmtsOnly allStmts
    useTiling = tiling $ head $ comps allStmts

    tile_tD = globalTimeDim $ head $ comps allStmts
    tile_rpD = realProcDim $ head $ comps allStmts
    tileD = tile_tD + tile_rpD

    coord_tD = globalTimeDim $ head $ comps allStmts
    coord_pD = coordD - coord_tD
    coordD = getCommonCoordsDim allStmts

    isPlace i = 
	if inRange (0, genericLength qualVect-1) i
	then genericIndex qualVect i == Place
	else error $ "getScatterNames: index (i = " ++ show i ++
	             ") out of range for quality Vector"

    -- the part for global time/processors
    -- (tileNr, if tiling is used, otherwise target coordinates):
    globalPart =
	if useTiling 
	then [if isPlace (i-1) 
	      then parPrefix ++ tileRealProcPrefix ++ show i
	      else seqPrefix ++ tileTimePrefix     ++ show i 
	      | i <- [1..tileD]]
	else [if isPlace (i-1) 
	      then parPrefix ++ coordProcPrefix ++ show i
	      else seqPrefix ++ coordTimePrefix ++ show i 
	      | i <- [1..coordD]]


    globalPartComm = 	
	if useComm
	then insertElemAt bl tile_tD globalPart
	else globalPart

--     globalPart_old =
-- 	if useTiling
-- 	then [(if (not useComm && isPlace (i-1)) 
-- 	       then parPrefix
-- 	       else seqPrefix) ++
-- 	      tileTimePrefix ++ show i | i <- [1..tile_tD]] ++ 
-- 	     if useComm then [bl] else [] ++
-- 	     [(if isPlace (i+tile_tD-1)
-- 	       then parPrefix 
-- 	       else seqPrefix) ++
-- 	      tileRealProcPrefix ++ show i | i <- [1..tile_rpD]]
-- 	else [(if (not useComm && isPlace (i-1))
-- 	       then parPrefix 
-- 	       else seqPrefix) ++
-- 	      coordTimePrefix ++ show i | i <- [1..coord_tD]] ++
-- 	    if useComm then [bl] else [] ++
-- 	     [(if isPlace (i+coord_tD-1)
-- 	       then parPrefix 
-- 	       else seqPrefix) ++
-- 	      coordProcPrefix ++ show i | i <- [1..coord_pD]]

    -- part for virtual time/processors (only used with tiling):
    virtualTimePart =
	if useTiling
	then [(if (not useComm && isPlace (i-1) && i > tileD) 
	       then parPrefix else seqPrefix) ++
	      (if isPlace (i-1) then coordProcPrefix else coordTimePrefix) ++
	      show i | i <- [1..coord_tD]]
	else []

--     virtualTimePart_old =
-- 	if useTiling 
-- 	then [seqPrefix ++ coordTimePrefix ++ show i | i <- [1..tile_tD]] ++
-- 	     [(if (not useComm && isPlace (i-1)) 
-- 	       then parPrefix
-- 	       else seqPrefix) ++
-- 	      coordTimePrefix ++ show i | i <- [tile_tD+1 .. coord_tD]]
-- 	else []

    -- part for the innermost dimensions which can differ from 
    -- stmt to stmt if communication is used:
    -- (if communication is not used, then the dimensions can be 
    --  named uniquely)
    innerPart =
	if useTiling
	then if useComm 
	     then ["stmtType"] ++		  
		  ["scatter" ++ show i | i <- [1 .. 1+coord_pD]]
	     else [(if (i > tileD && isPlace (i-1))
		    then parPrefix
		    else seqPrefix) ++
		   (if isPlace (i-1) then coordProcPrefix else coordTimePrefix) ++ 
		   show i | i <- [coord_tD+1 .. coord_tD+coord_pD]] ++
		   ["stmtNr"]
	else if useComm 
	     then ["stmtType"] ++ ["stmtOrDepNum"] 
	     else ["stmtNr"]
	   
    names = 
	(trace "{global}" globalPartComm) ++ 
	(trace "{virtualTime}" virtualTimePart) ++ 
	(trace "{inner}" innerPart)

---
-- inserts a given element at the position in a list given by n 
-- (counting from 0).
-- example: insertAt 'b' 1 ['a', 'c', 'd'] --> ['a', 'b', 'c', 'd']
insertElemAt :: a -> Integer -> [a] -> [a]
insertElemAt elem n xs = before ++ [elem] ++ after
    where
    (before, after) = genericSplitAt n xs
		

checkResult :: ExitCode -> (Bool, Int)
checkResult ExitSuccess = (True, 0)
checkResult (ExitFailure x) = (False, x)

---
-- Executes the given CLooG executable using the given input file
-- and directs the output into a given output file.
runCloog :: String -> String -> String -> IO ()
runCloog cloogExec cloogInputFile outputFile = 
    let      
      cloogFlags = "-block 1 -esp 0 -csp 0 -strides 1"
      cloogCommand = 
	  cloogExec ++ " " ++ cloogFlags ++ 
          " -o " ++ outputFile ++ " " ++ cloogInputFile
    in do 
       exitCode <- do system cloogCommand
       let (ok, errCode) = checkResult exitCode
       if ok then do putStrLn "cloog completed."
             else do putStrLn ("cloog exited with error code: " ++ 
			       (show errCode))       

---
-- Gets the text for the original computation statements in the 
-- untransformed program.
-- Note: LooPo's "targout" tool is used
getTextForStatements :: String -> String -> String -> String -> IO [(StatementID, String)]
getTextForStatements targoutExec tempDir projFile langFile =
    let
    tmpFile = "stmtsText.tmp"
    targoutFlags = 
	" -temp " ++ tempDir ++
	" -proj " ++ tempDir ++ "/" ++ projFile ++
	" -pp " ++
	" -count " ++
	" -lang " ++ langFile ++
	" -f " ++ tmpFile
    in do
       exitCode <- do 
		   putStrLn ("calling targout for statement text...\n" ++
			     targoutExec ++ targoutFlags)
		   system (targoutExec ++ targoutFlags)
       let (ok, errCode) = checkResult exitCode
       if ok then do putStrLn "targout completed."
             else do error ("targout exited with error code: " ++
			    (show errCode))
       stmtsTextFile <- openFile (tempDir ++ "/" ++ tmpFile) ReadMode
       rawStmtsText <- hGetContents stmtsTextFile
       let stmtsLines = filter (isPrefixOf stmtPrefix) $
			lines rawStmtsText
       return $ map getStmtNumberAndText stmtsLines

stmtPrefix :: String
stmtPrefix = "StatementNr: "
       
---
-- Reads the origStmtNr and the corresponding code for a statement from a 
-- given line of the temporary file generated by "targout".
getStmtNumberAndText :: String -> (StatementID, String)
getStmtNumberAndText line =
    (integerToID (read stmtNr::Integer), unwords stmtText)
    where
    (stmtNr:stmtText) = words $ drop (length stmtPrefix) line

---
-- Generates all defines for all target statements.
-- @param origToTrStmts a list containing mapping from origStmtNr to trStmtNr
-- @param stmtsNrAndText a list of origStmtNrs and corresponding statement code
-- @param indexMappings a list of origStmtNrs and correspondig code for mapping
--                      cloog indices to the original indices in the src-program.
generateAllStmtDefines :: [(StatementID, TStatementID)] -> 
			  [(StatementID, String)] -> 
			  [(StatementID, String)] -> String
generateAllStmtDefines origToTrStmts stmtsNrAndText indexMappings = 
    if (length stmtsNrAndText /= length indexMappings) 
    then error "generateAllStmtDefines: #stmts doesn't match!"
    else unlines $ zipWith (generateStmtDefine origToTrStmts) stmtsNrAndText indexMappings

---
-- Helper function for generateAllStmtDefines.
generateStmtDefine :: [(StatementID, TStatementID)] -> 
		      (StatementID, String) -> 
		      (StatementID, String) -> String
generateStmtDefine origToTrStmts (stmtNr1, stmtText) (stmtNr2, indexMapping) =
    if (stmtNr1 /= stmtNr2)
    then error "generateStmtDefine: stmtNr doesn't match!"
    else unlines $
	 map (\trStmtNr -> 
	      defineText ++ cloogStmtDesc ++ (show trStmtNr) ++ " " ++ 
	      indexMapping ++ stmtText) trStmtNrs
    where
    trStmtNrs = foldr (\(orig, tr) trNums -> if (orig == stmtNr1) 
		                         then tr:trNums 
		                         else trNums) [] origToTrStmts
    defineText = "#define "
    cloogStmtDesc = "S"

---
-- Generates for each origStmtNr code for mapping cloog indices to the 
-- original indices in the src-program.
-- @param loopoIndices a list of all index identifiers from the src-program
-- @param cloogIndices a list of index identifiers used instead of 
--        the original identifiers in cloog
generateAllSurrLoopsIndices :: [String] -> [String] -> SurroundingLoops -> 
			       [(StatementID, String)]    
generateAllSurrLoopsIndices loopoIndices cloogIndices (SLP surrLoopsList) =
    map (\(stmtNr, surrLoops) -> 
	   generateSurrLoopsIndices loopoIndices cloogIndices surrLoops stmtNr) 
	surrLoopsList

---
-- Helper function for generateAllSurrLoopsIndices.
generateSurrLoopsIndices :: [String] -> [String] -> [Bool] -> StatementID -> 
			    (StatementID, String)
generateSurrLoopsIndices loopoIndices cloogIndices surrLoops stmtNr =
    (stmtNr, 
     concat $ map (\(l, c) -> l ++ " = " ++ c ++ "; ") indexPairs)
    where
    usedLoops = elemIndices True surrLoops
    indexPairs = 
	zipWith (\u i -> (loopoIndices!!u, cloogIndices!!i))
		usedLoops [0..(length usedLoops)-1]

mctestmain :: String -> Integer -> [Vector] -> Bool -> IO ()
mctestmain test_file test_timeDim test_lat onScreen = 
    do
    if onScreen then putStr code 
                else do
		     putStrLn $ "writing cloog input into: " ++ 
				  (test_file ++".cloog")
		     writeFile (test_file ++".cloog") code
    where
    code = cloogInput ++
           --generateProcMapping test_tispcs test_lat test_trafo test_timeDim ++ 
	   --initBufferCode ++
	   ""
	   {- MC 2004-01-14: code wird ab jetzt mit Cloog generiert
           (concat $ niceLoops globalTimeLoopCode [globalTimeLoopBody])
	   -}
    globalTimeLoopCode = ["loop1", "loop2"]
    globalTimeLoopBody =
        mcSenderCode test_accs test_acps test_invtrafos test_lat 
	{- MC TODO: alle Abhaengigkeiten betrachten! -}
			 (head test_trdeps) 
    
    cloogDS = CloogDS { ispcsDS = test_ispcs, 
			tispcsDS = test_tispcs,
			trDepsDS = test_trdeps, 
		        invTrafosDS = test_invtrafos,
			linCombEqsDS = test_lincombs,
			trcsDS = test_trc,
			latticeDS = test_lat,
			timeDimDS =  test_timeDim,
			useTilingDS = True,
			useCommunicationDS = True
		      }

    allStmts = 
	CloogStmts { comps = map createComputeStatement 
                                 (createComputeStmtData cloogDS),
		     comms = map createCommunicationStatement 
                                 (createCommunicationStmtData cloogDS),
		     sends = map createPrepareSendStatement 
                                 (createPrepareSendStmtData cloogDS),
		     recvs = map createPrepareReceiveStatement 
                                 (createPrepareReceiveStmtData cloogDS),
		     computeStmtsOnly = False,
		     indexNames = undefined,
		     blobNames = undefined,
		     qualityVectors = undefined
		   }

    cloogInput = generateCloogInput allStmts 

    loopoDS = unsafeReadLoopoDS test_file
    test_raccs = loopoACC loopoDS
    test_acps = loopoACP loopoDS
    test_slp  = loopoSLP loopoDS
    test_deps'= loopoDEP loopoDS
    test_ispcs' = loopoISPC loopoDS
    test_tispcs'= loopoTISPC loopoDS
    test_trafo = loopoTrafo loopoDS
    test_trc = loopoTRC loopoDS

    test_ispcs = map ({-reduceStmt-}trimIndexSpace test_slp) test_ispcs'
    test_tispcs = test_tispcs' -- MC 2003-06-12
    test_deps = map ({-reduceDep-}trimDependence test_slp) test_deps'
    test_trafos = makeProperTrafos test_trafo
    test_invtrafos = makeProperInvTrafos test_trafo
    test_lincombs = makeProperLinCombEquations test_trafo
    test_trdeps = transformDeps test_tispcs test_trafos test_deps 
    test_accs = acpsToAccs test_slp test_acps test_ispcs test_trdeps test_raccs 



---
-- Creates a list of data structures which each entry holding information for 
-- one computation statement. 
createComputeStmtData :: CloogDS -> [CloogStmtData]
createComputeStmtData CloogDS { ispcsDS = ispcs',
				invTrafosDS = invTrafos', 
				linCombEqsDS = linCombs,
				trcsDS = trcs', 
				latticeDS = l,
				timeDimDS = tD,
			        useTilingDS = useTiling,
				useCommunicationDS = useCommunication 
			      } 
    = compStmtDataList
  where (trStmtNrs, invTs) = unzip invTrafos'
	linCs = map snd linCombs
	(stmtNrs, trcConstrs) = unzip $ map (findTRC trcs') trStmtNrs        

        compStmtData5Tupel = 
	    zip5 trStmtNrs stmtNrs invTs linCs trcConstrs        
        compStmtDataList = 
	    map (\(trNr, nr, oneInvT, oneLinC, oneTRCConstr) ->
	    --trace ("creating compute/commStmtData: (trNr, nr) = " ++ show (trNr, nr)) $
      	
		    CloogStmtData { trDepData = 
				      error ("trying to use trDepData" ++
					     " in a compute/comm-stmt data"),
				    trStmtNrData = trNr,
				    origStmtData = fromJust $ find ((==nr).fst) ispcs', --genericIndex ispcs' (nr-1),
				    invTrafoData = oneInvT,
				    linCombEqsData = oneLinC,
				    trcConstrData = oneTRCConstr,
				    latticeData = l,
				    timeDimData = tD,
				    useTilingData = useTiling,
				    useCommunicationData = useCommunication
				  }
		    ) compStmtData5Tupel


createCommunicationStmtData :: CloogDS -> [CloogStmtData]
createCommunicationStmtData = createComputeStmtData
    

createPrepareSendStmtData :: CloogDS -> [CloogStmtData]
createPrepareSendStmtData cD = 
  map (\dep -> CloogStmtData { trDepData = dep,
			       trStmtNrData = 
			         error ("trying to use trStmtNrData " ++
					"in a send/recv-stmt data"),
			       origStmtData = 
			         error ("trying to use origStmtData " ++
					"in a send/recv-stmt data"),
			       invTrafoData = 
			         error ("trying to use invTrafoData " ++
					"in a send/recv-stmt data"),
			       linCombEqsData = 
			         error ("trying to use linCombEqsData " ++
					"in a send/recv-stmt data"),
			       trcConstrData = 
			       error ("trying to use trcConstrData " ++
				      "in a send/recv-stmt data"),
			       latticeData = latticeDS cD,
			       timeDimData = timeDimDS cD,
			       useTilingData = useTilingDS cD,
			       useCommunicationData = useCommunicationDS cD
			     }) (trDepsDS cD)


createPrepareReceiveStmtData :: CloogDS -> [CloogStmtData]
createPrepareReceiveStmtData = createPrepareSendStmtData


-- Creates the big IneqSystem for use in createComputeStatement and
-- createCommunicationStatement. 
-- Also returns the dimension of tileCoords and tileNr.
-- (dimension of tileNr is set to 0, if tiling is not used!)
createCloogIneqSystem :: CloogStmtData -> 
			 (IneqSystem, (Integer, Integer))
createCloogIneqSystem CloogStmtData { trStmtNrData = trStNr,
				      origStmtData = (stNr, is),
				      invTrafoData = iT,
				      linCombEqsData = linCombs,
				      trcConstrData = trcIS,
				      latticeData = l,
				      timeDimData = tD,
				      useTilingData = useTiling } 
    = trace (if (coordVars /= trCoordsDim) 
	     then "Error: coordVars /= trCoordsDim!" else "") $
      (cloogConstraintsIS, (coordVars, tileNrVars))
    where 
    trCoordsDim = nColumns iT_v

    -- tiling auf 0-zeiligem (leerem!) IneqSystem (mit dim(iT), iT inhomogen)
    emptyIneqSystem = ineqSystem ((trCoordsDim, nColumns iT_p), 
				 zeroMatrix 0 (nColumns iT))

    emptyStmt = (integerToID (-1), integerToID (-1), emptyIneqSystem)

    -- in order for using the space-time-mapping equations combined
    -- with the ISPC to get equations for the TISPC, we have to use
    -- tiling on an empty IneqSystem in order to generate an 
    -- "infinite tiled index space" where the relation between 
    -- tile number and absolute coordinates is represented
    (coordTileIS, [coordVars,tileNrVars]) = 
	if useTiling
	then
	  -- scrCoord / scrTile Reihenfolge hier so, 
	  -- damit IS und coords ueberlappen
	  getTiles l [SrcCoord,SrcTile] emptyStmt
	else
	-- if tiling is not used, we don't need the relation
	-- between tile numbers and coordinates at all,
	-- so we just use a dummy IneqSystem for the coordinates 
	-- and set the tileNr dimension to 0.
	  (emptyIneqSystem, 
	   [trCoordsDim, 0])

    

    (is_v, is_p, is_c) = partsVPC is
    (iT_v, iT_p, iT_c) = partsVPCih iT -- hier sicher inhomogen!
    (linC_v, linC_p, linC_c) = partsVPC linCombs
    (coordTile_v, coordTile_p, coordTile_c) = partsVPC coordTileIS
    (trc_v, trc_p, trc_c) = partsVPC trcIS

    -- the space-time mapping matrix: [ -E | iT ]
    stMapping_v = negUnitMatrix `conc` iT_v
    negUnitMatrix = -(unitMatrix (nRows iT_v) (nColumns is_v))
    
    is_vars = nColumns is_v   
    iT_vars = nColumns iT_v
    linC_vars = if (nColumns linC_v == iT_vars) 
		then nColumns linC_v
		else error "linear combinations don't match trafo matrix!"
    linC_rows = nRows linC_v
    coordTile_vars = nColumns coordTile_v    
    trc_vars = if (nColumns trc_v == is_vars) 
	       then nColumns trc_v
	       else error ("trc_vars != is_vars: (trc_vars, is_vars)=" ++ 
			   show (nColumns trc_v, is_vars))
    stMapping_vars = is_vars + iT_vars
    
    total_vars = is_vars + coordTile_vars
    total_params = nColumns is_p
    
    cloogConstraintsMatrix =
	trace "cloogConstraintsMatrix" $
	if (nColumns iT_p /= nColumns trc_p) then error ("trafo params != trc params") else 
	myMultijoin1 $
	[ 
	 trace "non-tiling part of compute/comm-IneqSystem..." $
	 is_M,
	 stMapping_M,
	 stMappingNeg_M,
	 linCombEqs_M,
	 trc_M
	] ++ 
        if useTiling 
	then [ 
	       trace "tiling part of IneqSystem" $
	       leadZerol (total_vars - coordTile_vars) coordTile_v `conc` 
	       coordTile_p `conc` coordTile_c 
	     ]
	else trace "not using tiling for Compute stmt IneqSystem..." $
		 []

    -- all inequations for the source index space:
    is_M = trailZerol (total_vars - is_vars) is_v `conc` is_p `conc` is_c

    -- space-time mapping:
    stMapping_M = trailZerol (total_vars - stMapping_vars) stMapping_v `conc` 
		  iT_p `conc` iT_c

    stMappingNeg_M = trailZerol (total_vars - stMapping_vars) (-stMapping_v) `conc` 
		     (-iT_p) `conc` (-iT_c)

    -- all equations (represented as 2 inequations) for restoring
    -- lines of the trafo-matrix which have been removed:
    linCombEqs_M = 
	zeroMatrix linC_rows is_vars `conc`
	linC_v `conc` 
	zeroMatrix linC_rows (total_vars - is_vars - linC_vars) `conc`
	linC_p `conc` linC_c
		   

    -- additional restrictions to the source index-space (see trc-file)
    trc_M = trailZerol (total_vars - trc_vars) trc_v `conc` 
	    trc_p `conc` trc_c	  


    -- When tiling is used, the order of variables has to be changed 
    -- for the resulting IS. If tiling is not used, then tileNr is empty, so
    -- this doesn't do any harm anyway.
    -- it should be: IS, tileNr, tileCoords
    projectVars = 
	[0 .. is_vars-1] ++ 
	[stMapping_vars .. stMapping_vars + tileNrVars -1] ++
	[is_vars .. is_vars + coordVars-1]

    constraintsIS = 
	--simplifyTrivially $ 
        -- trace ("trCoordsDim=" ++ (show trCoordsDim)) $
	ineqSystem ((total_vars, total_params), cloogConstraintsMatrix)

    cloogConstraintsIS = 	
	trace ("projectVars: " ++ show projectVars ++ "\n") $
	projectToVarsMC constraintsIS projectVars


---
-- Creates an abstract representation of a "compute statement" 
-- (i.e. a representation for each statement from which
-- the input for Cloog can be generated).
-- That means e.g. that the domain for this statement has to be computed
-- from the original IS, the s-t-mapping and some other constraints.
createComputeStatement :: CloogStmtData -> CloogStatement
createComputeStatement clData@CloogStmtData { trStmtNrData = trStNr,
					      origStmtData = (stNr, is),
					      timeDimData = tD,
					      useTilingData = useTiling,
					      useCommunicationData = 
					        useCommunication } 
    = trace "creating compute stmt...\n" $
      compStmt
      where 
      (cloogConstraintsIS, (coordsD, tileNrD)) = 
	  createCloogIneqSystem clData
      isD = nVars is
      compStmt = Compute { cloogIS = cloogConstraintsIS,
			   ispcDim = isD,
			   globalTimeDim = tD,
			   realProcDim = tileNrD - tD,
			   coordsDim = coordsD,
			   trStatementNr = trStNr,
			   includeCommunicationDimensions = 
			         useCommunication,
			   tiling = useTiling
			 }

---
-- Creates an abstract representation of a "prepare send statement" 
-- (i.e. a representation for each statement from which
-- the input for Cloog can be generated).
-- Here, the statement's domain is computed from the [tiled]
-- IS of the according transformed relation.
createPrepareSendStatement CloogStmtData { trDepData = dep,
					   latticeData = l,
					   timeDimData = tD,
					   useTilingData = useTiling
					 } = prepareSendStmt
    where
    -- Die Ungleichungssysteme fuer Schreiben in ("send") bzw.
    -- Lesen aus dem Sendepuffer 
    (tileIS, [srcTileNrVars, srcTileCoordVars, destTileNrVars]) =
                  defaultTile l (trDepRelation dep) 
				  [SrcTile, SrcCoord, DestTile]
    prepareSendStmt = PrepareSend { cloogIS = tileIS, 
				    globalTimeDim = tD, 
				    realProcDim = srcTileNrVars - tD,   
				    coordsDim = srcTileCoordVars, 
				    depNr =  trDepOrigID dep,
				    tiling = useTiling
				  }


---
-- Creates an abstract representation of a "prepare recieve statement" 
-- (i.e. a representation for each statement from which
-- the input for Cloog can be generated).
-- Here, the statement's domain is computed from the [tiled]
-- IS of the according transformed relation.
-- (Note: Currently, all fields should be the same as the according
--        fields of the prepare send statement. Refactoring could
--        be wise here...)
createPrepareReceiveStatement CloogStmtData { trDepData = dep,
					      latticeData = l,
					      timeDimData = tD,
					      useTilingData = useTiling
					    } = prepareReceiveStmt
    where
    (tileIS, [srcTileNrVars, srcTileCoordVars, destTileNrVars]) =
           defaultTile l (trDepRelation dep) 
			   [SrcTile, SrcCoord, DestTile]
    prepareReceiveStmt = PrepareReceive { cloogIS = tileIS, 
					  globalTimeDim = tD, 
					  realProcDim = srcTileNrVars - tD,   
					  coordsDim = srcTileCoordVars, 
					  depNr =  trDepOrigID dep,
					  tiling = useTiling
					}


---
-- Creates an abstract representation of THE "communication statement".
-- Note: Currently, MPI all-to-all communication is inteded to be used.
-- Here, the domain is computed similar to the "compute statement", but
-- because *every* real processor has to communicate at *every* 
-- global time step (from the start of computation on the first 
-- real processor to the end of computation of the last real processor), 
-- we have to first skew the index system here, then later use 
-- CLooG's scatter function to project to the global time dimension.
createCommunicationStatement clData@CloogStmtData { trStmtNrData = trStNr,
						    origStmtData = (stNr, is),
						    timeDimData = tD,
						    useTilingData = useTiling }  
    = commStmt
      where 
      (cloogConstraintsIS, (coordsD, tileNrD)) = 
	  trace "creating IneqSystem for Comm-Stmt..." $
	  createCloogIneqSystem clData
      isD = nVars is
      rpD = tileNrD - tD

      tileDims = [isD .. isD+tileNrD-1]
      tileIS = 
	trace ("(tD, rpD): " ++ show (tD, rpD) ++ "\n") $
	trace ("projecting to: " ++ show tileDims ++ "\n") $
	projectToVarsMC cloogConstraintsIS tileDims

      skewedIS = 
	  trace ("WARNING: skewing-fct should be applied here...") $
	  --applySkewingFct tD rpD tileIS
	  tileIS
		
      commStmt = Communicate { cloogIS = skewedIS,
			       globalTimeDim = tD,
			       realProcDim = rpD,   
			       coordsDim = coordsD,
			       tiling = useTiling
			     }

{---
-- kein guter Name... ist eigentlich nur fuer Compute/Communicate zu gebrauchen
applySkewingFct :: Integer -> Integer -> IneqSystem -> IneqSystem
applySkewingFct tD rpD is = 
    image skewingFct is
    where
    params = nParams is
    rows = nRows is
    skewingMatrix = 
	myMultijoin1 
	[	 
	 -- ACHTUNG: Wohl noch voellig falsch!
	  unitMatrix tD tD `conc` zeroMatrix tD (params+1),
	  matrix (rpD, tD, genericReplicate (rpD * tD) (1%1)) `conc` 
	    zeroMatrix rpD (params+1),
	  zeroMatrix params tD `conc` unitMatrix params params `conc`
	    zeroMatrix params 1,
	  zeroMatrix 1 (tD+params) `conc` unitMatrix 1 1
	]
    skewingFct = affineFunction ((tD, params), skewingMatrix) Homogeneous
---}


---
-- takes all compute-, send-, receive-statements and 
-- communication statements and generates the input for cloog
generateCloogInput ::  CloogStmts -> String
generateCloogInput allStmts@CloogStmts { comps = compStmts,
					 sends = sendStmts,
					 recvs = recvStmts, 
					 comms = commStmts,
				         computeStmtsOnly = computeOnly,
					 indexNames = indices,
					 blobNames = blobs,
					 qualityVectors = qualVects
				       } 
    = header ++
      descriptions ++
      scatterFcts
    where params = nParams $ cloogIS (head compStmts)


          numTotalVars = 
	      maximum $ map (nVars.cloogIS) (sendStmts ++ compStmts)
	  additionalVars = 
	      ["loopo" ++ show i |
	       i <- [(length indices)+1 .. numTotalVars]]

	  numTotalStmts = length compStmts + 
			  length recvStmts + length sendStmts + 
			  -- only one communication statement! 
			  -- (which can have multiple domains though...)
			  if computeOnly then 0 else 1
          

          header = 
	      "# ---------------------- CONTEXT ----------------------\n" ++
	      "c # language is C\n\n" ++
	      (if (params > 0) 
	       then "1 # Context\n" ++
	            (showIS2cloog $ ineqSystem ((0, params), 
						unitMatrix params params `conc` 
						zeroMatrix params 1)) ++ 
	            "\n\n" 
	       else "1 # Context (no parameter, so always true)\n" ++
	            "1 2       # 1 line and 2 columns  \n" ++
	            "#  1 \n" ++
		    "1  0   # 0 >= 0, always true\n\n") ++
	      "1 # set parameter names\n" ++
	      (concat $ intersperse " " blobs) ++ "\n\n" ++

	      "# --------------------- STATEMENTS --------------------\n" ++
	      show numTotalStmts ++ "\n\n"
          descriptions = stmtsDescriptions allStmts ++
			 "\n1 # set the iterator names\n" ++
			 (concat $ intersperse " " 
			  (indices ++ additionalVars)) ++ 
			 "\n\n"
	  scatterFcts = 
              "# --------------------- SCATTERING --------------------\n" ++
	      show numTotalStmts ++ " # Scattering functions\n" ++
	      scatterFunctions allStmts ++
	      "\n1 # we set the scattering dimension names\n" ++
	      (concat $ intersperse " " $ getScatterNames allStmts) ++ "\n\n"



foreign import ccall getIndex_c :: CString -> CString -> CInt ->
				   IO CString
getIndex :: String -> String -> Integer -> Maybe String
getIndex tempDir projFile indexNum = 
    unsafeCppiConvert3 getIndex_c tempDir projFile indexNum

foreign import ccall getBlob_c :: CString -> CString -> CInt ->
				   IO CString
getBlob :: String -> String -> Integer -> Maybe String
getBlob tempDir projFile blobNum = 
    unsafeCppiConvert3 getBlob_c tempDir projFile blobNum

---
-- Returns lists of lexems (names) of indices and blobs in the original 
-- program. 
-- Note: at this time, the loopo parser is used for retrieving the lexems
--       from the symbol table. Therefore, a C++ interface function has
--       to be called.
getIndicesAndBlobs :: String -> String -> ([String], [String])
getIndicesAndBlobs tempDir projFile =
    (indices, blobs)
    where
    indices = 
	map fromJust $ 
	takeWhile isJust $ map (getIndex tempDir projFile) [1..]
    blobs   = 
	map fromJust $ 
	takeWhile isJust $ map (getBlob tempDir projFile) [1..]

---
-- Generates the description of each statement as CLooG expects it.
stmtsDescriptions :: CloogStmts -> String
stmtsDescriptions cS 
    = compStmtsDescription ++
      if (computeStmtsOnly cS) 
      then []
      else sendStmtsDescription ++ 
	   recvStmtsDescription ++ 
	   commStmtsDescription
    where
    compStmtsDescription =
	"# computation stmts:\n" ++
	concat (map (\stmt -> 
		     "1 # one domain per statement\n" ++
		     showIS2cloog (cloogIS stmt) ++ 
		     "0 0 0 # for future options...\n" ++
		     "\n\n") (comps cS))

    sendStmtsDescription =
	"# prepare-send stmts:\n" ++
	concat (map (\stmt -> 
		     "1 # one domain per statement\n" ++
		     showIS2cloog (cloogIS stmt) ++ 
		     "0 0 0 # for future options...\n" ++
		     "\n\n") (sends cS))

    recvStmtsDescription =
	"# prepare-receive stmts:\n" ++
	concat (map (\stmt -> 
		     "1 # one domain per statement\n" ++
		     showIS2cloog (cloogIS stmt) ++ 
		     "0 0 0 # for future options...\n" ++
		     "\n\n") (recvs cS))

    commStmtsDescription =
	"# communication statement:\n" ++
	(show $ length (comms cS)) ++ 
	" # number of domains for communication statement\n" ++
	concat (map (\stmt -> showIS2cloog (cloogIS stmt) ++ "\n\n") 
		(comms cS)) ++
	"0 0 0 # for future options...\n\n"
	

---
-- Generates the input for CLooG's scatter functions from 
-- all "cloog statements".
scatterFunctions  ::  CloogStmts -> String
scatterFunctions allStmts@CloogStmts { comps = compStmts,
				       sends = sendStmts,
				       recvs = recvStmts, 
				       comms = commStmts,
				       computeStmtsOnly = computeOnly
				     } 
    = trace "generating scatter fcts...\n" $
      if computeOnly
      then computeStmtsScatterCode
      else computeStmtsScatterCode ++ communicationScatterCode
    where 
    computeStmtsScatterCode = "# compute stmts: \n" ++ compFcts ++ "\n\n"
    communicationScatterCode = 
	"# prepare send stmts: \n" ++ sendFcts ++ "\n\n" ++
        "# prepare receive stmts: \n" ++ recvFcts ++ "\n\n" ++
        "# communication stmt: \n" ++ commFct ++ "\n"
    commonCoordsDim = getCommonCoordsDim allStmts	
    compFcts = 
	trace "compFcts..." $
	concat $ map (genScatterFct commonCoordsDim) compStmts
    sendFcts = 	concat $ map (genScatterFct commonCoordsDim) sendStmts
    recvFcts = concat $ map (genScatterFct commonCoordsDim) recvStmts
    commFct = 
	(genScatterFct commonCoordsDim) (head commStmts)

-- get the dimension of the common loop-nest of all target statements
-- (Cloog needs same number of scatter dimensions for all statements).
-- Note: currently, only the minimum dimension of coordinates over
--       all statements is used, but it should rather be:
--       dim(Schedule) + dim(Allocation)
getCommonCoordsDim :: CloogStmts -> Integer
getCommonCoordsDim clStmts = minimum $ map coordsDim allStmts
    where
    allStmts = 
	comps clStmts ++ 
	sends clStmts ++ 
	recvs clStmts ++ 
	comms clStmts

-- Generates the Cloog input code for a given target coordinate dimension
-- (needed for scatter dimensions) and a given statement.
genScatterFct :: Integer -> CloogStatement -> String
genScatterFct commonCoordsDim Compute { cloogIS = clIS,
					ispcDim = isD,
					globalTimeDim = tD, 
					realProcDim = rpD,   
					coordsDim = coordsD, 
					trStatementNr = stmtNr,
					includeCommunicationDimensions = 
					    useCommunication,
					tiling = useTiling
				      } 
    = trace ("generating scatter-fct for compute-stmt...") $
      showScatterMatrix2cloog fctM
    where
    stmtType = 0%1 -- first compute, then do other stuff
    blockNr = 0%1

    tile_tD = if useTiling then tD else 0
    tile_rpD = if useTiling then rpD else 0
    auxDim = if useCommunication then 1 else 0
    blockNrD = auxDim
    stmtTypeD = auxDim
    stmtNrD = 1 -- even without communication!

    c_tD = trace ("comp: (tD, rpD, params, coordsD)=" ++ show (tD, rpD, params, coordsD)) $
	   tD
    c_vpD = coordsD - c_tD
    commonCs_vpD =  commonCoordsDim - c_tD

    -- remaining target coordinate dimension:
    -- (to achieve uniform scatter dimension for all statements)
    leftOverCs = 
	coordsD - commonCoordsDim

    -- send/recv have additional src/destTileRP:
    -- (and depNr instead of stmtNr)
    fillVars = if useCommunication 
	       then rpD -- ACHTUNG: stimmt das noch mit Tiling an/aus?
	       else 0

    params = nParams clIS
    fctM = 
	--trace ("tile_tD=" ++ show tile_tD ++ ", tile_rpD=" ++ show tile_rpD ++ ", c_tD=" ++ show c_tD ++ ", c_vpD=" ++ show c_vpD ++ ", fillVars=" ++ show fillVars ++ "\n") $
	myMultijoin1
	[ tileTime_M, blockNr_M, tileRP_M, coordsTime_M, stmtType_M, coordsVP_M,
	  stmtNr_M, fillVars_M ]

	 -- tileTime:
	 -- Ziel-Ausfuehrzeitpunkte werden aus dem Tiling-Ungleichungssystem
	 -- fuer die globale Zeit berechnet
    tileTime_M = 
	trace "tileTime." $
        --trace (show $ tile_tD + blockNrD + tile_rpD + commonCoordsDim +
	--       stmtTypeD + stmtNrD + fillVars + isD + tile_tD +
	--       if useTiling then tile_rpD else 0 +
	--       coordsD + params + 1) $
        (unitMatrix tile_tD tile_tD `conc` 
	 zeroMatrix tile_tD (blockNrD + tile_rpD + commonCoordsDim +
			     stmtTypeD + stmtNrD + fillVars + isD) `conc`
	 (-(unitMatrix tile_tD tile_tD)) `conc`
	 (if useTiling 
	  then (rowMatrix (genericReplicate tile_rpD (-1%1)) `join` 
		zeroMatrix (tile_tD-1) tile_rpD) 
	  else zeroMatrix tile_tD tile_rpD) `conc`
	 zeroMatrix tile_tD (coordsD + params + 1)
	)

    -- blockNr: 
    -- zur Trennung von Schreiben/Lesen in/aus Puffer sowie Berechnungs-
    -- Statements von dem abschliessenden All-to-all Kommunikationsstmt.
    blockNr_M = 
	trace "blockNr." $
	--trace (show $ tD + blockNrD + rpD + commonCoordsDim + stmtTypeD + stmtNrD + 
	--       fillVars + isD + tD + rpD + coordsD + params + 1) $
        (zeroMatrix blockNrD tD `conc` unitMatrix blockNrD blockNrD `conc` 
	 zeroMatrix blockNrD (rpD + commonCoordsDim + stmtTypeD + stmtNrD + 
			      fillVars + isD + tD + rpD + coordsD + 
			      params) `conc`
	 if useCommunication 	  
	 then matrix [[-blockNr]]
	 else zeroMatrix blockNrD 1)
	
    -- tileRP: 
    -- Realprozessoren werden direkt aus dem entsprechenden 
    -- Ungleichungssystem uebernommen
    tileRP_M =
	trace "tileRP." $
        --trace (show $ tile_tD + blockNrD + tile_rpD + commonCoordsDim + 
	--       stmtTypeD + stmtNrD + fillVars + isD + tile_tD + tile_rpD + 
	--       coordsD + params + 1) $
        (zeroMatrix tile_rpD (tile_tD + blockNrD) `conc` 
	 unitMatrix tile_rpD tile_rpD `conc`
	 zeroMatrix tile_rpD (commonCoordsDim + stmtTypeD + stmtNrD + 
			      fillVars + isD + tile_tD) `conc`
	 (-(unitMatrix tile_rpD tile_rpD)) `conc` 
	 zeroMatrix tile_rpD (coordsD + params + 1))
	
    -- coordsTime:
    -- der Anteil der absoluten Zielkoordinaten, der fuer die 
    -- virtuelle Zeit steht (feine Zeit). Wird direkt aus dem entspr.
    -- Ungl. Sys. uebernommen.
    coordsTime_M = 
	trace "coordsTime." $
        (zeroMatrix c_tD (tile_tD + blockNrD + tile_rpD) `conc` 
	 unitMatrix c_tD c_tD `conc` 
	 zeroMatrix c_tD (stmtTypeD + commonCs_vpD + stmtNrD + fillVars + 
			  isD + tile_tD + tile_rpD) `conc`
	 (-(unitMatrix c_tD c_tD)) `conc` zeroMatrix c_tD (c_vpD + params + 1))

    -- stmtType:
    -- zum Sicherstellen der Reihenfolge (nur bei Kommunikation):
    -- 1. Berechnen
    -- 2. Schreiben in Sendepuffer
    -- 3. Lesen aus Sendepuffer, schreiben in lokales Array
    stmtType_M = 
	trace "stmtType." $
        (zeroMatrix stmtTypeD (tile_tD + blockNrD + tile_rpD + c_tD) `conc` 
	 unitMatrix stmtTypeD stmtTypeD `conc` 
	 zeroMatrix stmtTypeD (commonCs_vpD + stmtNrD + fillVars + 
			       isD + tile_tD + tile_rpD + coordsD + 
			       params) `conc`
	 if useCommunication 
	 then matrix [[-stmtType]]
	 else zeroMatrix stmtTypeD 1)

    -- coordsVP: 
    -- Anteil der virtuellen Prozessoren an den absoluten Zielkoordinaten.
    coordsVP_M = 
	trace "coordsVP." $
	--trace ("\n<coordsD = " ++ show coordsD ++ "> " ++
	--       "<c_tD = " ++ show c_tD ++ "> " ++
	--       "<commonCs_vpD = " ++ show commonCs_vpD ++ "> " ++
	--       "<leftOverCs = " ++ show leftOverCs ++ ">\n") $
	       
	(zeroMatrix commonCs_vpD (tile_tD + blockNrD + tile_rpD + c_tD + 
				  stmtTypeD) `conc` 
	 unitMatrix commonCs_vpD commonCs_vpD `conc`
	 zeroMatrix commonCs_vpD (stmtNrD + fillVars + isD + tile_tD + 
				  tile_rpD + c_tD) `conc`
	 (-(unitMatrix commonCs_vpD commonCs_vpD)) `conc` 
	 zeroMatrix commonCs_vpD  (leftOverCs + params + 1))
	
    -- stmtNr:
    -- Nummer des jeweiligen Berechnungsstatements, um sequentielle
    -- Ausfuehrungsreihenfolge zu gewaehren.
    stmtNr_M = 
	trace "stmtNr." $
        (zeroMatrix stmtNrD (tile_tD + blockNrD + tile_rpD + c_tD + 
			     stmtTypeD + commonCs_vpD) `conc` 
	 unitMatrix stmtNrD stmtNrD `conc` 
	 zeroMatrix stmtNrD (fillVars + isD + tile_tD + tile_rpD + coordsD + 
			     params) `conc` 
	 matrix [[-(integerFromID stmtNr % 1)]])

    -- fillVars:
    -- zusaetzliche Dimensionen, um eine einheitliche Anzahl der 
    -- Zieldimensionen fuer alle Cloog-Statements sicherzustellen.
    fillVars_M = 
	trace "fillVars." $
        (zeroMatrix fillVars (tile_tD + blockNrD + tile_rpD + commonCoordsDim + 
			      stmtTypeD + stmtNrD) `conc`
	 unitMatrix fillVars fillVars `conc`
	 zeroMatrix fillVars (isD + tile_tD + tile_rpD + coordsD + 
			      params + 1))	 

genScatterFct commonCoordsDim stmtData@PrepareSend { cloogIS = clIS, 
						     globalTimeDim = tD, 
						     realProcDim = rpD,   
						     coordsDim = coordsD, 
						     depNr = depN,
						     tiling = useTiling
						   }
    = showScatterMatrix2cloog sendM
      where 

      helperMs = getSendRecvSubMatrices stmtData commonCoordsDim True

      stmtType = 1%1
      blockNr = 0%1

      tile_tD = if useTiling then tD else 0
      tile_rpD = if useTiling then rpD else 0

      tileD = tile_tD + tile_rpD
      c_tD = trace ("send: " ++ show (tD, rpD, params, coordsD)) $
	   tD
      c_vpD = coordsD - c_tD
      commonCs_vpD = commonCoordsDim - c_tD

      leftOverCs = coordsD - commonCoordsDim

      params = nParams clIS
      sendM = 	  
        myMultijoin1
	[
	 -- srcTileTime
	 (unitMatrix tile_tD tile_tD `conc` 
	  zeroMatrix tile_tD (1 + tile_rpD + c_tD + 1 + 1 + 
			      tile_rpD + commonCs_vpD) `conc` 
	  (-(unitMatrix tile_tD tile_tD)) `conc`
	  (rowMatrix (genericReplicate tile_rpD (-1%1)) `join` 
	   zeroMatrix (max (tile_tD-1) 0) tile_rpD) `conc`
	  zeroMatrix tile_tD (coordsD + tileD + params + 1)
	 ),
	 -- blockNr
	 blockNrM helperMs,
	 -- srcTileRP
	 (zeroMatrix tile_rpD (tile_tD + 1) `conc` 
	  unitMatrix tile_rpD tile_rpD `conc`
	  zeroMatrix tile_rpD (c_tD + 1 + 1 + tile_rpD + commonCs_vpD + tile_tD) `conc`
	  (-(unitMatrix tile_rpD tile_rpD)) `conc` 
	  zeroMatrix tile_rpD (coordsD + tileD + params + 1)),
         -- srcCoordsTime
	 srcCoordsTimeM helperMs,
         -- stmtType
	 stmtTypeM helperMs,
         -- depNr
	 depNrM helperMs,
	 -- destTileRP
	 (zeroMatrix tile_rpD (tile_tD + 1 + tile_rpD + c_tD + 1 + 1) `conc`
	  unitMatrix tile_rpD tile_rpD `conc` 
	  zeroMatrix tile_rpD (commonCs_vpD + tileD + coordsD + tile_tD) `conc`
	  (-(unitMatrix tile_rpD tile_rpD)) `conc` 
	  zeroMatrix tile_rpD (params + 1)
	 ),
	 -- srcCoordsVP
         srcCoordsVPM helperMs
	]

genScatterFct commonCoordsDim stmtData@PrepareReceive { cloogIS = clIS, 
							globalTimeDim = tD, 
							realProcDim = rpD,   
							coordsDim = coordsD, 
							depNr = depN,
							tiling = useTiling
						      }
    = showScatterMatrix2cloog recvM
    where 

      helperMs = getSendRecvSubMatrices stmtData commonCoordsDim False

      stmtType = 2%1
      blockNr = 0%1

      tile_tD = if useTiling then tD else 0
      tile_rpD = if useTiling then rpD else 0

      tileD = tile_tD + tile_rpD
      c_tD = trace ("recv: " ++ show (tD, rpD, params, coordsD)) $
	     tD
      c_vpD = coordsD - c_tD
      commonCs_vpD = commonCoordsDim - c_tD

      leftOverCs = coordsD - commonCoordsDim

      params = nParams clIS
      recvM =
        myMultijoin1
	[
	 -- srcTileTime
	 (unitMatrix tile_tD tile_tD `conc` 
	  zeroMatrix tile_tD (1 + tile_rpD + c_tD + 1 + 1 + 
			      tile_rpD + commonCs_vpD) `conc` 
	  (-(unitMatrix tile_tD tile_tD)) `conc`
	  (rowMatrix (genericReplicate tile_rpD (-1%1)) `join` 
	   zeroMatrix (max (tile_tD-1) 0) tile_rpD) `conc`
	  zeroMatrix tile_tD (coordsD + tileD + params + 1)
	 ),
	 -- blockNr
	 blockNrM helperMs,
	 -- destTileRP
	 (zeroMatrix tile_rpD (tile_tD + 1) `conc` 
	  unitMatrix tile_rpD tile_rpD `conc`
	  zeroMatrix tile_rpD (c_tD + 1 + 1 + tile_rpD + commonCs_vpD + 
			  tileD + coordsD + tile_tD) `conc`
	  (-(unitMatrix tile_rpD tile_rpD)) `conc` 
	  zeroMatrix tile_rpD (params + 1)),
         -- srcCoordsTime
	 srcCoordsTimeM helperMs,
         -- stmtType
	 stmtTypeM helperMs,
         -- depNr
	 depNrM helperMs,
	 -- srcTileRP
	 (zeroMatrix tile_rpD (tile_tD + 1 + tile_rpD + c_tD + 1 + 1) `conc`
	  unitMatrix tile_rpD tile_rpD `conc` 
	  zeroMatrix tile_rpD (commonCs_vpD + tile_tD) `conc`
	  (-(unitMatrix tile_rpD tile_rpD)) `conc` 
	  zeroMatrix tile_rpD  (coordsD + tileD + params + 1)
	 ),
	 -- srcCoordsVP
	 srcCoordsVPM helperMs
	]


genScatterFct commonCoordsDim Communicate { cloogIS = is,
					    globalTimeDim = tD,
					    realProcDim = rpD,   
					    coordsDim = coordsD
					  }
    = showScatterMatrix2cloog fctM
    where
    myRpD = nVars is - tD
    blockNr = 1%1

    fillVars = rpD + 1 + 1 + rpD + commonCoordsDim
    params = nParams is
    fctM = 
	myMultijoin1
	[
	 -- tileTime

         -- ATTENTION: no skewing is used here! It should already be done
	 --            completely at the time when the index system is
	 --            generated.
	 (unitMatrix tD tD `conc` 
	  zeroMatrix tD (1 + fillVars) `conc` 
	  (-(unitMatrix tD tD)) `conc`
	  --(rowMatrix (genericReplicate rpD (-1%1)) `join` zeroMatrix (tD-1) rpD) 
	  zeroMatrix tD (params + 1)),
	 -- blockNr
	 (zeroMatrix 1 tD `conc` 
	  unitMatrix 1 1 `conc`
	  --zeroMatrix 1 (fillVars + tD + rpD + params) `conc` 
	  zeroMatrix 1 (fillVars + tD + params) `conc` 
	  matrix[[-blockNr]]),
	 -- fillVars
         (zeroMatrix fillVars (tD + 1) `conc`
	  unitMatrix fillVars fillVars `conc`
	  zeroMatrix fillVars (tD + params + 1))
	]
	  

---
-- Data structure for some common sub matrices of the monstrous matrices
-- used for generating the scatter functions.
data HelperSubMatrices =
    HelperSubMatrices {
		       blockNrM :: Matrix,
		       srcCoordsTimeM :: Matrix,
		       stmtTypeM :: Matrix,
		       depNrM :: Matrix,
		       srcCoordsVPM :: Matrix
		      }
    deriving Show

---
-- Generates some common sub matrices of the monstrous matrices
-- used for generating the scatter functions.
-- @param stmt The CloogStatment to which the various sub matrices
--             should be computed (only prepare-send/receive stmts).
-- @param commonCoordsDim The common dimension of absolute target coordinats
--                        for all statements in the TISPC
-- @param isSendStmt A flag indicating whether the sub matrices should
--                   be created for prepare-send or -receive stmts
-- @return a HelperSubMatrices labeled field containing the various
--         sub matrices.
getSendRecvSubMatrices :: CloogStatement -> Integer -> Bool -> 
			  HelperSubMatrices
getSendRecvSubMatrices stmt commonCoordsDim isSendStmt = 
    HelperSubMatrices {
		       blockNrM = blockNr_M,
		       srcCoordsTimeM = srcCoordsTime_M,
		       stmtTypeM = stmtType_M,
		       depNrM = depNr_M,
		       srcCoordsVPM = srcCoordsVP_M
		      }
    where
    tD = globalTimeDim stmt
    rpD = realProcDim stmt
    coordsD = coordsDim stmt
    depN = integerFromID $ depNr stmt
    useTiling = tiling stmt

    commonCs_vpD = commonCoordsDim - c_tD
    leftOverCs = coordsD - commonCoordsDim

    stmtType = 
	if isSendStmt
	then 1%1
	else 2%1
    blockNr = 0%1
	      
    tile_tD = if useTiling then tD else 0
    tile_rpD = if useTiling then rpD else 0

    tileD = tile_tD + tile_rpD
    c_tD = tD
    c_vpD = coordsD - c_tD
    params = nParams (cloogIS stmt)

    blockNr_M = 
    	(zeroMatrix 1 tile_tD `conc` unitMatrix 1 1 `conc` 
	 zeroMatrix 1 (tile_rpD + c_tD + 1 + 1 + tile_rpD + commonCs_vpD + 
		       tileD + coordsD + tileD + params) `conc`
	 matrix [[-blockNr]])
	 
    srcCoordsTime_M = 
	(zeroMatrix c_tD (tile_tD + 1 + tile_rpD) `conc` 
	 unitMatrix c_tD c_tD `conc` 
	 zeroMatrix c_tD (1 + 1 + tile_rpD + commonCs_vpD + tileD) `conc`
	 (-(unitMatrix c_tD c_tD)) `conc` 
	 zeroMatrix c_tD (c_vpD + tileD + params + 1))

    stmtType_M =
	(zeroMatrix 1 (tile_tD + 1 + tile_rpD + c_tD) `conc` 
	 unitMatrix 1 1 `conc` 
	 zeroMatrix 1 (1 + tile_rpD + commonCs_vpD + tileD + coordsD + 
		       tileD + params) `conc`
	 matrix [[-stmtType]])
	
    depNr_M =
	(zeroMatrix 1 (tile_tD + 1 + tile_rpD + c_tD + 1) `conc` 
	 unitMatrix 1 1 `conc` 
	 zeroMatrix 1 (tile_rpD + commonCs_vpD + tileD + coordsD + 
		       tileD + params) `conc`
	 matrix [[-depN%1]])

    srcCoordsVP_M =
	(zeroMatrix commonCs_vpD (tile_tD + 1 + tile_rpD + c_tD + 
				  1 + 1 + tile_rpD) `conc` 
	 unitMatrix commonCs_vpD commonCs_vpD `conc` 
	 zeroMatrix commonCs_vpD (tileD + c_tD) `conc`
	 (-(unitMatrix commonCs_vpD commonCs_vpD)) `conc` 
	 zeroMatrix commonCs_vpD (leftOverCs + tileD + params + 1))

--     spareCs_M =
-- 	(zeroMatrix spareCs (tile_tD + 1 + tile_rpD + c_tD +
-- 			     1 + 1 + tile_rpD + c_vpD) `conc`
-- 	 unitMatrix spareCs spareCs `conc`
-- 	 zeroMatrix spareCs (tileD + c_tD + c_vpD + tileD + params + 1))



--- 
-- shows a "LooPo style" IneqSystem in Cloog-conform syntax
showIS2cloog :: IneqSystem -> String
showIS2cloog is = output
    where
    rawM = fmap fromIntegral $ getMatrix $ integralIneqSystem is
    oneColumn = columnVectorsToMatrix1 [vector (replicate (nRows rawM) (1%1))]
    finalM = oneColumn `conc` rawM
    output = show (nRows finalM) ++ " " ++ show (nColumns finalM) ++ "\n" ++
	     intShowMat finalM False

--- 
-- shows a scatter function matrix in Cloog-conform syntax
showScatterMatrix2cloog :: Matrix -> String
showScatterMatrix2cloog m = output
    where
    zeroColumn = columnVectorsToMatrix1 [vector (replicate (nRows m) (0%1))]
    finalM = zeroColumn `conc` m
    output = show (nRows finalM) ++ " " ++ show (nColumns finalM) ++ "\n" ++
	     intShowMat finalM False ++ "\n"

---
-- finds the matching TRC for a given target statement number
-- and returns its original statement number and constraints
findTRC :: [TRC] -> TStatementID -> (StatementID, IneqSystem)
findTRC trcs trStmtNr = (trcStmtID trc, trcConstraints trc)
    where
    Just trc = find (\aTRC -> trcTargetStmtID aTRC == trStmtNr) trcs


-- ispc <-> tispc: #stmnts gleich? nein: #tispc >= #ispc!

-- creates an inequality system for one statement providing all information
-- which "Cloog" expects
-- @param is the (untransformed) index space
-- @param tispc the transformed index space
-- @param trafo the space time transformation function
-- @param lattice the lattice vector used for tiling
-- @param timeDim the number of time dimensions in a tile
-- @return resultIS containing all necessary information for cloog
createCloogStatementIneqs :: Statement -> (TStatementID, StatementID, IneqSystem) -> 
			     (Integer,AffineFunction) -> [Vector] -> Integer -> 
							 IneqSystem
createCloogStatementIneqs stmt@(stmtNum, is) tispc (_,trafo) lattice timeDim =
    let (tilingIS, [coordVars,tileNrVars]) = 
	    getTiles lattice [SrcCoord,SrcTile] (integerToID (-1),stmtNum,is)

	(isV, isP, isC) = partsVPC is
	(trafoV, trafoP, trafoC) = partsVPC trafo
        (tileV, tileP, tileC) = partsVPC tilingIS
	isVs = nColumns isV
	trafoVs = nColumns trafoV
	tileVs = nColumns tileV

	totalPs = nColumns isP -- same number of parameters for all

	totalVs = isVs + tileVs

        numCols = isVs + tileVs + totalPs + nColumns isC

	resultMatrix = 
	    myMultijoin1 
	    [trailZerol (totalVs - isVs) isV `conc` isP `conc` isC,
	     trailZerol (totalVs - trafoVs) trafoV `conc` trafoP `conc` trafoC,
	     trailZerol (totalVs - trafoVs) (-trafoV) `conc` (-trafoP) `conc` (-trafoC),
	     leadZerol  (totalVs - tileVs) tileV `conc` tileP `conc` tileC
	    ]	    

        assertParams = case nColumns tileP == nColumns isP of
		       True -> ""
		       False -> "number of tiling parameters differs from ISPC parameters"
        assertConst1 = case (nColumns tileC == 1) && (nColumns isC == 1) of
		       True -> ""
		       False -> "number of constants should be 1"
	assertCols = case nColumns (resultMatrix :: Matrix) == numCols of
		     True -> ""
		     False -> "wrong number of columns in the resulting matrix! "
        failedAssertions = 
	    concat [assertParams, assertConst1, assertCols]

	resultIS = 
	    case failedAssertions of
	      "" -> ineqSystem ((totalVs, totalPs), resultMatrix)
	      s  -> error s
    in resultIS

	   
leadZerol :: Integer -> Matrix -> Matrix
leadZerol cols m = zeroMatrix (nRows m) cols `conc` m

trailZerol :: Integer -> Matrix -> Matrix
trailZerol cols m = m `conc` zeroMatrix (nRows m) cols


---
-- returns the tiled IneqSystem and dimensions
-- for all requested tiling information for one statement.
-- @param lattice  a given lattice vectors
-- @param infos    a given list of TileInfromation 
-- @param trStmnt  a given target-statement
-- @return res     the tiled IS with all dimensions for each info
getTiles :: [Vector] -> [TileInformation] -> TStatement -> (IneqSystem, [Integer])
getTiles lattice infos (tsID,sID,ispc) = res
    where
    res = defaultTile lattice (selfDepRelation (sID,ispc)) infos

projectToVarsMC is vars = fromMaybe (error "problem in projectToVarsFM")
			            projected
    where
    projected = projectToVarsFM is Nothing vars

myMultijoin1 ms = 
    if ok 
    then multijoin1 ms
    else error ("multijoin1 failed with different columns in submatrices: " ++ 
		show cols)
    where
    ok = all (== head cols) cols
    cols = map nColumns ms


--- ------------------------------------------------------------------------
-- old code:

trafoAccessFkt :: AffineFunction -> Access -> Access
trafoAccessFkt f a = a { accIndexFct = accIndexFct a `apply` f }


---
-- Generates code for mapping all tiles (from all statements!) to real processors
-- @param tispcs the statements used for collecting all ISs
-- @param lattice the lattice used for tiling
-- @param trafo the used transformation
-- @param timeDim the number of time dimensions in the ISs
generateProcMapping :: [(TStatementID, StatementID, IneqSystem)] -> [Vector] -> Trafo -> Integer 
		    -> String
generateProcMapping tispcs lattice trafo timeDim
    = header ++ concat code
    where

      code = --trace (show allTiles) $ trace (show (disjUnionAll allTiles)) $
             map (oneMapping.(prepareIS tileDim)) (disjUnionAll allTiles)

      -- all tiles (-> real processors!) for all statements,
      -- represented as a list of IneqSystems
      -- also returns the dimension of the tile enumerating loops
      (allTiles, tileDim) = --trace (show (map fst procInfo, snd $ head procInfo)) $
	                    (map fst procInfo,
			     snd $ head procInfo)
                          where
--                          trStmts = 
--			      map ignoreOrigStmt tispcs
--			  paddedStmts = map (padStmnt trafo) trStmts -- Aufblasen oder nicht aufblasen?
			  info = --trace (show paddedStmts) $
				 map (getTiles lattice [SrcTile]) {-paddedStmts-}tispcs
			  (timeInfo, procInfo) = splitTimeProcDims info timeDim
      header = "#include <vector>\n" ++
               "#include <map>\n" ++
               "#include \"NTupel.h\"\n" ++
               "\n" ++
               "::std::map<NTupel,int> " ++ tileMapName ++ ";\n" ++
	       "int LOOPO_procNum = 0;\n\n"


ignoreOrigStmt :: (TStatementID, StatementID, IneqSystem) -> (TStatementID,IneqSystem)
ignoreOrigStmt (stmnt,origstmnt,is) = (stmnt,is)

-- MC 200401-15: WICHTIG!! es wird davon ausgegangen, dass nur 
--               Ungleichungssysteme fuer die SrcTileNr 
--               benutzt werden, d.h. all (==1) (map length dims)
splitTimeProcDims :: [(IneqSystem,[Integer])] -> Integer -> 
		     ([(IneqSystem,Integer)], [(IneqSystem,Integer)]) 
splitTimeProcDims tiles timeDim = (timeTilesAndDims, procTilesAndDims)
    where
    (iss, dims) = unzip tiles
    vars = nVars (head iss)
    procDim = vars - timeDim
    timeDims = [0 .. timeDim-1]
    procDims = [timeDim .. vars-1]
    timeIss' = map (\is -> projectToVarsMC is timeDims) iss
    procIss' = map (\is -> projectToVarsMC is procDims) iss
    timeDims' = 
	if all (==1) (map length dims)
	then map (\x -> x-procDim) (head dims)
	else error ("dims == " ++ (show dims) ++ 
		    ",  splitTimeProcDims has to be used on SrcTiles only!")
    procDims' = 
	if all (==1) (map length dims)
	then map (\x -> x-timeDim) (head dims)
	else error ("dims == " ++ (show dims) ++ 
		    ",  splitTimeProcDims has to be used on SrcTiles only!")
    timeTilesAndDims = --trace (show (timeIss', timeDims')) $ 
		       zip timeIss' timeDims'
    procTilesAndDims = --trace (show (procIss', procDims')) $ 
		       zip procIss' procDims'


-- MC 2004-01-04: noch nicht korrekt!
showIneqs :: IneqSystem -> Integer -> Integer -> [Rational] -> String
showIneqs is fstVar sndVar parVals = case failedAsserts of
				    "" -> output
				    s  -> error s
    where
    vars = nVars is
    assertCorrectVars = if ((fstVar>=0 && fstVar < vars) && 
			    (sndVar>=0 && sndVar < vars)) 
			then ""
			else "wrong variables to project to! "
    assertCorrectParVals = if (length parVals == fromInteger (nParams is))
			  then ""
			  else "wrong number of parameters! "
    failedAsserts = concat [assertCorrectVars, assertCorrectParVals]
    projectedIS = projectToVarsMC is [fstVar, sndVar]
    outputIS = head (reverse $ 
		     zipWith (\nr val -> substParam projectedIS nr val) 
		             [0..] parVals)
    output = show outputIS

tileMapName :: String
tileMapName = "LOOPO_tileMap"

sendBufferName :: String
sendBufferName = "LOOPO_buffer"

tmpValueName :: String
tmpValueName = "LOOPO_tmpValue"

initBufferCode :: String
initBufferCode = "\n" ++
		 "::std::vector< ::std::vector<int> > " ++ sendBufferName 
							    ++";\n" ++
		 "for (int i = 0; i < MPI_NP; i++) {\n" ++
		 "  ::std::vector<int> LOOPO_tmpVec;\n" ++
		 "  " ++ sendBufferName ++ ".push_back(LOOPO_tmpVec);\n" ++
		 "}\n\n"

---
-- prepares one IneqSystem by restoring its variable/parameter-dimensions
-- and sorting it (to insert some needed inequalities...)
prepareIS :: Integer -> IneqSystem -> IneqSystem
prepareIS tileDim is = 
    unMaybe (sortIneqs (restoreDims tileDim is) DefaultMethod)
	where unMaybe (Just x) = x
	      unMaybe Nothing = error ("prepareIS: should not happen!" ++ 
				       " Perhaps wrong sorting method?")
	  

---
-- the numbers of vars and params of an given IneqSystem is restored
-- according to the given number of variables
restoreDims :: Integer -> IneqSystem -> IneqSystem
restoreDims vars is = 
    ineqSystem ((vars, (nVars is)-vars), getMatrix is)
	 
---
-- generates the code for the mapping (virtual procs -> mpi procs) for one ineq-system
oneMapping :: IneqSystem -> String
oneMapping is = concat $ (map (++ "\n") $ niceLoops loopCode body)
    where
    loopCode = lines (loopNest is)
    body = ["const int LOOPO_arrayTMP[] = {" ++
            concat (intersperse "," ["i" ++ show (i::Integer) | i <- [0..(nVars is)-1]]) ++
            "};"] ++
	   ["::std::vector<int> LOOPO_vectorTMP(LOOPO_arrayTMP," ++ 
            "LOOPO_arrayTMP+" ++ show (nVars is) ++ ");"] ++
	   [tileMapName ++ "[LOOPO_vectorTMP] = LOOPO_procNum++;"]

---
-- generates code for sending neccesary data according to a given dependenc
mcSenderCode ::[Access] -> [(DependenceID,AccessID,AccessID)] ->
               [(TStatementID,AffineFunction)] -> [Vector] -> 
               TrDependence  -> String
mcSenderCode accs acps inv_trafos lattice dependence
    =  --trace (generateProcMapping ispcs lattice) $
       concat $ map (++"\n") code
    where 

      srcAccess = srcAcc (trDepOrigID dependence) acps accs
      trSrcAccess = trafoAccessFkt invTrafo srcAccess
      indexCode = mcIndexFctCode (accIndexFct trSrcAccess)
      arraySym = "arr"++(show.accArrayID) srcAccess
      invTrafo = case lookup (trace "Warnung: StatementID vs. TStatementID in 'mcSenderCode'" $ trace "HACK!" $ integerToID $ integerFromID $ accStmtID srcAccess) inv_trafos 
              of Just t -> t
                 Nothing  -> error $ "mcSenderCode: Statement "++
                             (show.accStmtID) srcAccess++
                             " not found for access "++
                             (show.accID) srcAccess
		                 
      senderLoopCode = defaultTiledLoops
		       lattice (trDepRelation dependence)
			       [] [SrcTile]
      receiverLoopCode = defaultTiledLoops
			 lattice (trDepRelation dependence)
				 [SrcTile] [DestTile]

      loopCode = defaultTiledLoops 
                 lattice (trDepRelation dependence) 
                             [SrcTile,DestTile] [SrcCoord]

      outerLoopCode = defaultTiledLoops 
                      lattice (trDepRelation dependence) 
                                  [] [SrcTile,DestTile]

      outerDims = error "outerDims to be implemented..."
      senderDims = error "senderDims to be implemented..."
      receiverDims = error "receiverDims to be implemented..."

      initCode = ["init buffer" ++
                  concat ["[i" ++ show i ++ "]" | i <- [0..(outerDims-1)]] ++
                  ";"]

      code = [""] ++ 		 
             niceLoops senderLoopCode senderCode


      senderCode = genProcNumber name senderDims ++
	  ["if (" ++ name ++ ") == MPI_ID) {" ++ 
	   ""] ++
	  niceLoops receiverLoopCode receiverCode ++
	  ["}"]
	  where name = "LOOPO_sendProcNumber"
			     
      receiverCode =  {- MC 2003-09-07 -}
		      indexCode ++ --["// schreib in puffer..."]
		      genProcNumber recProcName receiverDims ++
					[sendBufferName ++ "[" ++ recProcName ++ 
					 "].push_back(" ++ tmpValueName ++ ");"]
			  where recProcName = "recProc"
		 

---
-- generates the code for a variable definition which holds the processor
-- number calculated from indices form a given index space using the 
-- tile/processor mapping.
--
-- @param name the name of the created variable holding the processor number
-- @param dim the dimension of the surrounding index space
genProcNumber :: String -> Integer -> [String]
genProcNumber name dim =
    ["",
     "const int LOOPO_arrayTMP[] = {" ++
     concat (intersperse "," ["i" ++ show (i::Integer) | 
			      i <- [0..(dim)-1]]) ++ 
     "};"
    ] ++ 
    ["::std::vector<int> LOOPO_vectorTMP(LOOPO_arrayTMP," ++ 
     "LOOPO_arrayTMP+" ++ show dim ++ ");"] ++
    ["const int " ++ name ++ " = " ++ tileMapName ++ "[LOOPO_vectorTMP];" ++ 
     ""]

niceLoops :: [String] -> [String] -> [String]
niceLoops l b = niceLoopsHelper l b ""

niceLoopsHelper [] body indent = map (indent++) body
niceLoopsHelper (l:loops) body indent = [indent++l++" {"]++
                                        niceLoopsHelper loops body (indent++"   ") ++
                                        [indent++"}"]



---
-- Build code for a statement which applies an affine function
-- 
-- @param iF affine function to be used
-- @returns Statement of the form "newcoeffs = matrix * coeffs^T"
mcIndexFctCode :: AffineFunction -> [String]
mcIndexFctCode iF 
{- MC 2003-09-07 -}
    = let varNames = ["i"++show i| i <- [0..nVars iE-1]] 
	  parNames = ["n"++show i| i <- [0..nParams iE-1]]
	  iE = toAffineExp iF
	  {-rightHandSide = concat ("(":intersperse "," (varNames++parNames++["1"]))++
			  ")^T"-}
	  tr = getMatrix iE
	  niceRows = rows' tr
	  homIndexVector = varNames++parNames++["1"]
	  products = map (\r -> zipWith genProduct r homIndexVector) niceRows
	  entries = map (intersperse " + ") products
      in ["const int " ++ tmpValueName ++ " = ai[" ++
	  concat (intersperse ", " (map concat entries)) ++
          "];"]

genProduct :: Rational -> String -> String
genProduct l r = "(" ++ show n ++ "*" ++ r ++ ")/" ++ show d
    where n = numerator l
	  d = denominator l


is1 = ineqSystem ((2,1), matrix[[1,0,0,0],
				[-1,0,1,5%1],
				[0,1,-1,1],
				[0,-1,0,4]])

is2 = ineqSystem ((2,1), matrix[[1,0,0,-3],
				[1,-1,0,-1%1],
				[1,1,0,-4],
				[-1,-1,0,10],
				[-1,1,0,6],
			        [0,-1,0,3],
			        [0,1,-1,0],
				[-1,0,0,8]])

is3 = ineqSystem ((2,1), matrix[[1,0,0,-6],
				[-1,0,0,8%1],
				[0,1,-1,2],
				[0,-1,0,5]])

isTiling = ineqSystem ((2,1), matrix [[1,0,0,0],
				      [-1,0,0,10],
				      [0,1,-1,0],
				      [1,-1,0,0%1]])

emptyIS vs ps = ineqSystem ((vs, ps), 
			   zeroMatrix 0 (vs+ps+1))


isMin = ineqSystem ((2,0), matrix[[1,0,0],
				  [-1,1,0],
				  [-1,-1,2%1]
				  ,[-1,0,1]
				 ])

{- Aufruf:
putStrLn $ concat $ map (\(Just x) -> oneMapping x) $ 
map (\x -> sortIneqs (restoreDims 2 x) PipMethod) (disjUnionAll [is1,is2,is3])
-}

{-
generateBackendCodeOnly "exs/example10_2_dv/" "example10_2.src.out" (Just lattice_2dim) True "" ""
-}
