module CompSpread.CloogInput
where

import List
import Ratio
import Maybe(isJust, fromJust)

import LMath.Types

import CompSpread.Data
import CompSpread.DomainGen
import CompSpread.DomainDescriptions
import CompSpread.ScatterFunctions
import CompSpread.CloogStatements
import CompSpread.Names
import CompSpread.Helpers


---
-- A more "generic" version of the interface for generating the input for CLooG.
genGenericCloogInput :: [[IneqSystem]] 
                     -> [String] 
                     -> [String]
                     -> [String]
                     -> Maybe Integer
                     -> Maybe [Matrix]
                     -> String
genGenericCloogInput domains iteratorNames scatterNames blobs minBlobSize scatterFcts = 
    concat ([ genCloogHeader blobs minBlobSize
            , genCloogDomainDescription domains iteratorNames
            ]
            ++ scatterFctDescription)
    where
    scatterFctDescription = 
        if (isJust scatterFcts)
        then [ "# --------------------- SCATTERING --------------------\n"
             , (show $ length domains) ++ " # Scattering functions\n"
             , concatMap showScatterMatrix2Cloog $ (fromJust scatterFcts)
             , "\n1 # we set the scattering dimension names\n"
             , (concat $ intersperse " " scatterNames)
             , "\n\n"
             ]
        else [ "# --------------------- NO SCATTERING --------------------\n"
             , "0 # no scatter functions\n"
             , "\n\n"
             ]


genMainCloogInput :: CompSpreadDS -> CloogStatements -> String
genMainCloogInput csDS cloogStatements =
        concat [genCloogHeader (csBlobNames csDS) (Just (csMinBlobSize csDS))
               , genMainDomainDescription csDS cloogStatements
               , genMainScatterFctDescription csDS cloogStatements]

genRPMapInput :: CompSpreadDS -> RPMapStatements -> String
genRPMapInput csDS rpMapData = cloogInput
    where
    rpMapStmts = rpmStmts rpMapData
    cloogInput =
        concat [genCloogHeader (csBlobNames csDS) (Just (csMinBlobSize csDS))
               , genRPMapDomainDescription csDS rpMapStmts
               , genRPMapScatterFctDescription csDS rpMapStmts]

genCloogHeader :: [String] -> Maybe Integer -> String
genCloogHeader blobs minBlobSize =
    "# created: " ++ getTime ++ "\n" ++
    "# ---------------------- CONTEXT ----------------------\n" ++
    "c # language is C\n\n" ++
    (if ((numBlobs > 0) && (isJust minBlobSize))
     then "1 # Context\n" ++
          (showIS2cloog $ ineqSystem ((0, numBlobs), 
				      unitMatrix numBlobs numBlobs `conc` 
				      blobSize_M)) ++ "\n\n"
     else "1 # Context (no parameter, so always true)\n" ++
          (showIS2cloog $ ineqSystem ((0, numBlobs), 
				      zeroMatrix 1 (numBlobs+1))) ++
          "\n\n"
    ) ++ "1 # set parameter names\n" ++
    (concat $ intersperse " " blobs) ++ "\n\n"
    where
    numBlobs = length blobs
    blobSize_M = (columnMatrix (genericReplicate numBlobs 
                                (-(fromJust minBlobSize)%1)))

--- 
-- Generic domain description function for use in other modules.
genCloogDomainDescription :: [[IneqSystem]] -> [String] -> String
genCloogDomainDescription cloogStmtDomains iteratorNames =
    concat $ ["# --------------------- STATEMENTS --------------------\n"
             , show numTotalStmts ++ "\n\n"
             , allDescriptions
             , (if null iteratorNames 
                then "\n0 # don't set the iterator names\n"
                else "\n1 # set the iterator names\n"
               )
             , (concat $ intersperse " " iteratorNames)
             , "\n\n"
             ]
    where
    numTotalStmts = genericLength cloogStmtDomains
    allDescriptions = concatMap cloogDomainDesc cloogStmtDomains

genMainDomainDescription :: CompSpreadDS -> CloogStatements -> String
genMainDomainDescription csDS CloogStatements { clCompStmts = comps
                                              , clWriteStmts = writes
                                              , clUnpackStmts = unpacks
                                              , clSendStmts = sends
                                              , clRecvStmts = recvs
                                              -- , clSpreadStmts = spreads
                                              } =
    concat ["# --------------------- STATEMENTS --------------------\n"
           , show numTotalStmts ++ "\n\n"
           , allDescriptions
--           , "\n0 # don't set the iterator names\n"
           , "\n1 # set the iterator names\n"
           , (concat $ intersperse " " $ getIteratorNames csDS) ++ "\n\n"
           ]
    where
    allDescriptions =
        concat [compStmtsDesc
               , writeStmtsDesc
               , unpackStmtsDesc
               , sendStmtsDesc
               , recvStmtsDesc
               -- , spreadStmtsDesc
               ]

    compStmtsDesc = 
        ("# " ++ (show $ length comps) ++ " computation stmts:\n" ++
	 concatMap domainDesc comps)

    writeStmtsDesc = 
        ("# " ++ (show $ length writes) ++ " write-buffer stmts:\n" ++
	 concatMap domainDesc writes)

    unpackStmtsDesc =
	("# " ++ (show $ length unpacks) ++ " unpack-buffer stmts:\n" ++
	 concatMap domainDesc unpacks)

    sendStmtsDesc =
	("# " ++ (show $ length sends) ++ " send stmts:\n" ++
	 concatMap domainDesc sends)

    recvStmtsDesc =
	("# " ++ (show $ length recvs) ++ " recv stmts:\n" ++
	 concatMap domainDesc recvs)

--     spreadStmtsDesc =
-- 	("# " ++ (show $ length spreads) ++ " data-spreading stmts:\n" ++
-- 	 concatMap domainDesc spreads)
        
    numTotalStmts = sum $ map genericLength [comps
                                            , writes
                                            , unpacks
                                            , sends
                                            , recvs
                                            -- , spreads
                                            ]
    

genRPMapDomainDescription :: CompSpreadDS -> [CloogStatement] -> String
genRPMapDomainDescription csDS rpMapStmts = 
    concat ["# --------------------- STATEMENTS --------------------\n"
           , show numTotalStmts ++ "\n\n"
           , rpMapDescriptions
           , "\n0 # don't set the iterator names\n"
--            , "\n1 # set the iterator names\n"
--            , (concat $ intersperse " " $ getIteratorNames csDS) ++ "\n\n"
           ]    
    where
    rpMapDescriptions =
	("# " ++ (show $ length rpMapStmts) ++ " rp-map stmts:\n" ++
	 concatMap domainDesc rpMapStmts)
    numTotalStmts = genericLength rpMapStmts
        



domainDesc :: CloogStatement -> String
domainDesc stmt =
    concat [show (length (clStmtDomain stmt)) ++ " # domains per statement\n"
           , concatMap (\is -> showIS2cloog is ++ "\n\n") (clStmtDomain stmt)
           , "0 0 0 # for future options...\n\n\n"
           ]

---
-- The generic version for domainDesc. (Intended for use in other modules)
cloogDomainDesc :: [IneqSystem] -> String
cloogDomainDesc dom = 
    concat [show (length dom) ++ " # domains per statement\n"
           , concatMap (\is -> showIS2cloog is ++ "\n\n") dom
           , "0 0 0 # for future options...\n\n\n"
           ]

genMainScatterFctDescription :: CompSpreadDS -> CloogStatements -> String
genMainScatterFctDescription csDS CloogStatements { clCompStmts = comps
                                                  , clWriteStmts = writes
                                                  , clUnpackStmts = unpacks
                                                  , clSendStmts = sends
                                                  , clRecvStmts = recvs
                                                  -- , clSpreadStmts = spreads
                                                  } = 
    "# --------------------- SCATTERING --------------------\n" ++
    (show numTotalStmts) ++ " # Scattering functions\n" ++
    scatterFctDesc ++
    "\n1 # we set the scattering dimension names\n" ++
    (concat $ intersperse " " scatterNames) ++ "\n\n"
    where
    compScatterFcts = map (genCompStmtScatterFct csDS) comps
    writeScatterFcts = map (genWriteStmtScatterFct csDS) writes
    unpackScatterFcts = map (genUnpackStmtScatterFct csDS) unpacks
    sendScatterFcts = map (genSendStmtScatterFct csDS) sends
    recvScatterFcts = map (genRecvStmtScatterFct csDS) recvs
    -- spreadScatterFcts = map (genSpreadStmtScatterFct csDS) spreads

    numTotalStmts = sum $ map genericLength [ comps, writes, unpacks, 
                                              sends, recvs
                                            -- , spreads 
                                            ]
    scatterNames = getScatterNames csDS

    scatterFctDesc = concatMap showScatterMatrix2Cloog $ 
                     concat[ compScatterFcts
                           , writeScatterFcts
                           , unpackScatterFcts
                           , sendScatterFcts
                           , recvScatterFcts
                           -- , spreadScatterFcts
                           ]
                     

genRPMapScatterFctDescription :: CompSpreadDS -> [CloogStatement] -> String
genRPMapScatterFctDescription csDS rpMapStmts =
    verbose "genRPMapScatterFctDescription" $
    "# --------------------- SCATTERING --------------------\n" ++
    (show numRPMappingStmts) ++ " # Scattering function\n" ++
    scatterFctDesc ++
    "\n1 # we set the scattering dimension names\n" ++
    scatterNames ++ "\n\n"    
    where    
    numRPMappingStmts = genericLength rpMapStmts
    scatterFcts = map (genRPMapStmtScatterFct csDS) rpMapStmts
    scatterFctDesc = concatMap showScatterMatrix2Cloog scatterFcts

    scatterNames = 
	concat $ intersperse " " $
        [procTile_t ++ show i | i <- [1 .. (getRPdim csDS)]]
