module CompSpread.Data
where

import Maybe
import Ratio
import Data.List
import CppInterface
import Foreign.C.String
import LoopoDS
import LMath.Types
import Debug.Trace

import CompSpread.Helpers
import ScannerParser.SymbolTable(getIndexLexems, getBlobLexems, getArrayLexems, arrayLexem)
import ScannerParser.ScannerParser(openParseFile)

data CompSpreadDS = CompSpreadDS
    { csTempDir          :: String
    , csProjFile         :: String
    , csTargoutExec      :: FilePath
    , csCloogExec        :: FilePath
    , csCloogOptions     :: String
    , csISPCs            :: [Statement]
    , csTISPCs           :: [TStatement]
    , csTrDeps           :: [TrDependence]
    , csTrafos           :: [(TStatementID, AffineFunction)]
    , csTRCs             :: [TRC]
    , csQualVects        :: [(TStatementID, [QualEntry])]
    , csIndexNames       :: [String]
    , csBlobNames        :: [String]
    , csArrayNames       :: [String]
    , csACPs             :: [AccessPair]
    , csAccesses         :: [Access]
    , csSLPs             :: SurroundingLoops
    , csLatVects         :: [Vector]
    , csTiledDims        :: [Integer]
    , csTilingUsedVect   :: [Bool]
    , csMinBlobSize      :: Integer
    , csNumSoakStmts     :: Integer
    -- ^ the number of soak statements
    , csNumDrainStmts    :: Integer
    -- ^ the number of drain statements
--    , csTimeTiling       :: Bool
--    , csSpaceTiling      :: Bool
--    , csCyclicMapping    :: Bool
    } deriving Show



-- |
-- Contains different types of identifiers that may occur in the original
-- program. (These identifiers are retrieved by the $targout$ tool.)
--
data Identifiers = Identifiers 
    { idIndices :: [String]
    , idBlobs :: [String]
    , idArrays :: [String]
    } deriving Show

-- |
-- Contains all command line options.
--
data CommandLine = CommandLine
    { clTempDir        :: String
    , clProjFile       :: String
    , clTargoutExec    :: FilePath
    , clCloogExec      :: FilePath
    , clCloogOptions   :: String
    , clLatVects       :: [Vector]
    , clTiledDims      :: [Integer]
    , clTilingUsedVect :: [Bool]
    , clMinBlobSize    :: Integer
    } deriving Show


unsafeCreateCompSpreadDS :: CommandLine -> CompSpreadDS
unsafeCreateCompSpreadDS commandLine  =
    trace ("properTISPCs: " ++ show properTISPCs) $
          trace ("(ispc, tispc:) " ++ show (ispcs, tispcs)) $
    verbose "unsafeCreateCompSpreadDS" $
    CompSpreadDS 
    { csTempDir = tempDir
    , csProjFile = projFile
    , csTargoutExec = clTargoutExec commandLine
    , csCloogExec = clCloogExec commandLine
    , csCloogOptions = clCloogOptions commandLine
    , csISPCs = ispcs
    , csTISPCs = properTISPCs
    , csTrDeps = trDeps
    , csTrafos = trafos
    , csTRCs = trcs
    , csQualVects = qualVects
    , csIndexNames = idIndices identifiers
    , csBlobNames = idBlobs identifiers
    , csArrayNames = idArrays identifiers
    , csACPs = acps
    , csAccesses = accs
    , csSLPs = slp
    , csLatVects = latVects
    , csTiledDims = tiledDims
    , csTilingUsedVect = tilingUsedVect
    , csMinBlobSize = clMinBlobSize commandLine
    , csNumSoakStmts = error "csNumSoakStmts not initialized!"
    , csNumDrainStmts = error "csNumDrainStmts not initialized!"
    }
    where    
    tempDir = clTempDir commandLine
    projFile = clProjFile commandLine
    loopoFile = tempDir ++ "/" ++ projFile
    loopoDS = 
	verbose "reading loopoDS" $
	unsafeReadLoopoDS loopoFile
    deps = map (trimDependence slp) (loopoDEP loopoDS)
    trDeps = transformDeps tispcs trafos deps
    trafos = makeProperTrafos (loopoTrafo loopoDS)
    ispcs = map (trimIndexSpace slp) (loopoISPC loopoDS)
    properTISPCs = map (\(x,y, is) -> (x,y, tightenIneqSystem is)) $ 
                   makeProperTISPCs ispcs tispcs trafos
    tispcs = loopoTISPC loopoDS
    slp = loopoSLP loopoDS
    acps = loopoACP loopoDS
    raccs = loopoACC loopoDS
    accs = map ((trimAccess slp). (rawAccessToAccess (loopoISPC loopoDS))) raccs
           -- acpsToAccs slp acps (loopoISPC loopoDS) allTrDeps raccs
    qualVects = loopoQUAL loopoDS
    trcs = loopoTRC loopoDS
    identifiers = getIdentifiers tempDir projFile
    latVects = clLatVects commandLine
    tiledDims = clTiledDims commandLine
    tilingUsedVect =  clTilingUsedVect commandLine


parseCommandLine :: [String] -> CommandLine
parseCommandLine args =
    trace "Warning: CompSpread.Data.parseCommandLine is deprecated" $
    case args of
      [] -> error usageInfo
      [tempDir, projFile, latVectsStr, tiledDimsStr, tileUsedVectStr,
       targoutExec, cloogExec, minBlobSizeStr, cloogOptions] ->
              CommandLine { clTempDir = tempDir
		          , clProjFile = projFile
		          , clTargoutExec = targoutExec
		          , clCloogExec = cloogExec
		          , clCloogOptions = cloogOptions
		          , clLatVects = latVects latVectsStr
                          , clTiledDims = (read tiledDimsStr)::[Integer]
		          , clTilingUsedVect = 
                              if ((length $ latVects latVectsStr) ==
                                  (length $ tileUsedVect tileUsedVectStr))
                              then tileUsedVect tileUsedVectStr
                              else error $ "wrong length of tileUsedVector!"
                            
                          , clMinBlobSize = (read minBlobSizeStr) :: Integer
                          }
      _ -> error $ "Could not parse command line parameters\n" ++
                   usageInfo
    where
    usageInfo :: String
    usageInfo =
        ("Usage: \n" ++
         "(insert your tool for \"executable\")" ++
         "executable <tmp-path> <project.src.out> <lattice-matrix> <tiledDims>\n" ++
	 "  <tile-used-vector> <targout-executable> <cloog-executable>\n" ++
         "  <minBlobSize> " ++
         "  <additional cloog parameters | \"\">\n\n" ++

         "example:\n" ++
         "executable \"/tmp/loopo23456/\" \"example.src.out\" " ++
         "\"[[10,0,0],[0,1,0],[0,0,200]]\" \"[0,1,2]\" \"[True,False,True]\" " ++
         "\"../loopo/targout/targout\" \"../cloog/gmp-bin/cloog\" " ++
         "1000 " ++
         "\"\""
        )
    tileUsedVect str = (read str) :: [Bool]

    latVectsIntRaw str =         
        (read str) :: [[Integer]]

    latVectsInt vects = if null vects
		  then [[]]
		  else vects
    latVects str = map (vector.map (\x -> x%1)) $ latVectsInt (latVectsIntRaw str)


getAllTStatementIDs :: CompSpreadDS -> [TStatementID]
getAllTStatementIDs csDS = fst3 $ unzip3 $ csTISPCs csDS

getAllStatementIDs :: CompSpreadDS -> [StatementID]
getAllStatementIDs csDS = nub $ stmtIDs
    where 
    (_, stmtIDs, _) = unzip3 $ csTISPCs csDS

getAllTrDepIDs :: CompSpreadDS -> [TrDependenceID]
getAllTrDepIDs csDS =
    map trDepID (csTrDeps csDS)

getTrafoFct :: CompSpreadDS -> TStatementID -> AffineFunction
getTrafoFct csDS trStmtID = trafoFct
    where
    trafoFct = 
	fromMaybe (error "getTrafoFct: no trafoFct for this TStatementID") $ 
		  lookup trStmtID (csTrafos csDS)

getTrDep :: CompSpreadDS -> TrDependenceID -> TrDependence
getTrDep csDS aTrDepID = trDep
    where
    trDep = fromMaybe (error "getTrDep: no TrDep for this TrDependenceID") $ 
	    find (\d -> trDepID d == aTrDepID) (csTrDeps csDS)
		  
    

getTRCis :: CompSpreadDS -> TStatementID -> IneqSystem
getTRCis csDS trStmtID = trcConstraints trc
    where
    trc = 
	fromJust $ 
	find (\oneTRC -> (trcTargetStmtID oneTRC) == trStmtID) (csTRCs csDS)


getISPC :: CompSpreadDS -> TStatementID -> IneqSystem
getISPC csDS trStmtID = 
    -- trace ("\ngetISPC: " ++ show is ++ "\n") $
    is
    where
    stmtID = getStmtID csDS trStmtID
    is = fromMaybe (error "getISPC: no ISPC for this StatementID found!") $ 
	 lookup stmtID (csISPCs csDS)

getTISPC :: CompSpreadDS -> TStatementID -> IneqSystem
getTISPC csDS trStmtID =
    fromMaybe (error "getTISPC: no TISPC for this TStatementID found!") $ 
	 lookup trStmtID tispcList
    where 
    tispcList = map (\(a, b, c) -> (a,c)) $ csTISPCs csDS

getStmtID :: CompSpreadDS -> TStatementID -> StatementID 
getStmtID CompSpreadDS { csTISPCs = tispcs } trID = stmtID
    where 
    ids' = foldr (\(t,s,ineq) ids -> if (t == trID) 
				     then s:ids 
				     else ids) [] tispcs
    stmtID = if length ids' == 1
	     then head ids' 
	     else error $ "couldn't find StatementID for TStatementID: " ++ 
		      (show trID)

getTStmtIDs :: CompSpreadDS -> StatementID -> [TStatementID]
getTStmtIDs CompSpreadDS { csTISPCs = tispcs } stmtID = trStmtIDs
    where 
    trStmtIDs' = foldr (\(t,s,ineq) trIDs -> if (s == stmtID) 
					     then t:trIDs 
					     else trIDs) [] tispcs
    trStmtIDs = if length trStmtIDs' > 0 
		then trStmtIDs'
		else error $ "couldn't find TStatementIDs for StatementID: " ++ 
			 (show stmtID)


getSrcAcc :: CompSpreadDS -> TrDependenceID -> Access
getSrcAcc csDS trDepNr = srcAcc
    where
    acps = csACPs csDS
    accs = csAccesses csDS
    depID = trDepOrigID trDep
    trDep = getTrDep csDS trDepNr
    acp = 
        fromMaybe (error "getSrcAcc: couldn't find accessPair!") $
        find (\acp -> acpDepID acp == depID) acps
    AccessPair {acpSrcAccID=srcAccID, acpDestAccID=destAccID} = acp
    srcAcc = 
        fromMaybe (error "getSrcAcc: couldn't find srcAccID!") $
        find (\a -> accID a == srcAccID) accs

---
-- Returns all TrDependenceIDs for a given srcAccessID
srcAccID2TrDepIDs :: CompSpreadDS -> AccessID -> [TrDependenceID]
srcAccID2TrDepIDs csDS srcAccID = trDepIDs
    where
    depIDs = map acpDepID $ filter (\acp -> acpSrcAccID acp == srcAccID) acps
    trDepIDs = [ trDepID trDep | trDep <- csTrDeps csDS, 
                                 trDepOrigID trDep `elem` depIDs]
    acps = csACPs csDS

---
-- Returns all TrDependenceIDs for a given destAccessID
destAccID2TrDepIDs :: CompSpreadDS -> AccessID -> [TrDependenceID]
destAccID2TrDepIDs csDS destAccID = trDepIDs
    where
    depIDs = map acpDepID $ filter (\acp -> acpDestAccID acp == destAccID) acps
    trDepIDs = [ trDepID trDep | trDep <- csTrDeps csDS, 
                                 trDepOrigID trDep `elem` depIDs]
    acps = csACPs csDS

getArrayName :: CompSpreadDS -> TrDependenceID -> String
getArrayName csDS trDepNr = arrayLexem symtab arrayID
    where
    (symtab,_,_,_) = unsafePerformIO $ openParseFile  (tempDir ++ "/" ++ projFile)
    tempDir = csTempDir csDS
    projFile = csProjFile csDS
    arrayID = accArrayID $ getSrcAcc csDS trDepNr

filterTrDeps :: [TrDependence] -> [DependenceType] -> [TrDependence]
filterTrDeps trueFlowDeps allowedTypes = 
  filter (\d -> (trDepType d) `elem` allowedTypes) trueFlowDeps
                           

undefinedCompSpreadDS :: CompSpreadDS
undefinedCompSpreadDS =   
    CompSpreadDS { csQualVects        = e
	         , csTempDir          = e
                 , csProjFile         = e
	         , csTargoutExec      = e
	         , csCloogExec        = e
                 , csCloogOptions     = e
	         , csISPCs            = e
	         , csTISPCs           = e
	         , csTrDeps           = e
	         , csTrafos           = e
	         , csTRCs             = e
	         , csIndexNames       = e
	         , csBlobNames        = e
                 , csArrayNames       = e
	         , csACPs             = e
	         , csAccesses         = e
	         , csSLPs             = e
	         , csLatVects         = e
                 , csTiledDims        = e
	         , csTilingUsedVect   = e
                 , csMinBlobSize      = e
                 , csNumSoakStmts     = e
                 , csNumDrainStmts    = e
--                 , csTimeTiling       = e
--                 , csSpaceTiling      = e
--	         , csCyclicMapping    = e
	         }
    where e = error "using undefinedCompSpreadDS"



-- warnung: derzeit wird implizit angenommen, dass
--          alle identifier von 1 weg durchgehend nummeriert sind!
getIdentifiers :: String -> String -> Identifiers
getIdentifiers tempDir projFile =
    Identifiers 
    { idIndices = indices,
      idBlobs = blobs,
      idArrays = arrays
    }		 
    where
    (symtab,_,_,_) = unsafePerformIO $ openParseFile  (tempDir ++ "/" ++ projFile)
    arrays = getArrayLexems symtab
    indices = getIndexLexems symtab
    blobs = getBlobLexems symtab
{-- MC 04.12.2006: sollte jetzt durch Benjamins symbol table funktionen ersetzt sein...
    indices = 
	map fromJust $ 
	takeWhile isJust $ map (getIndex tempDir projFile) [1..]
    blobs   = 
	map fromJust $ 
	takeWhile isJust $ map (getBlob tempDir projFile) [1..]
    arrays  =
	map fromJust $ 
	takeWhile isJust $ map (getArray tempDir projFile) [1..]
--}

foreign import ccall getIndex_c :: CString -> CString -> CInt ->
				   IO CString
getIndex :: String -> String -> Integer -> Maybe String
getIndex tempDir projFile indexNum = 
    unsafeCppiConvert3 getIndex_c tempDir projFile indexNum

foreign import ccall getBlob_c :: CString -> CString -> CInt ->
				   IO CString
getBlob :: String -> String -> Integer -> Maybe String
getBlob tempDir projFile blobNum = 
    unsafeCppiConvert3 getBlob_c tempDir projFile blobNum

foreign import ccall getArray_c :: CString -> CString -> CInt ->
				   IO CString
getArray :: String -> String -> Integer -> Maybe String
getArray tempDir projFile arrayID = 
    unsafeCppiConvert3 getArray_c tempDir projFile arrayID
