module CompSpread.DomainDescriptions
where

import Maybe
import List
import GHC.Base(assert)

import CompSpread.Data
import CompSpread.Helpers
import LoopoDS
import LMath.Types

{- data structures: -}

newtype StmtDomDescription = StmtDomDescription [StmtDomPart]
    deriving Show

data StmtDomPart = DpTile  DimInfo
		 | DpTISPC DimInfo
		 | DpISPC  DimInfo
		   deriving (Show, Eq)
			    
data DimInfo = DimInfo { dimQual :: QualEntry 
		       , dimRole :: Maybe CommSide
		       }
	       deriving (Show, Eq)

data CommSide = SendSide | RecvSide
		deriving (Show, Eq)

{- ================================================================ -}



-- |
-- Project @oldIneqs@ described by @oldDescr@ to the
-- new domain description @newDescr@. 
--
projectDomDescriptionTo :: CompSpreadDS 
			-> IneqSystem         -- ^ @oldIneqs@
			-> StmtDomDescription -- ^ @oldDescr@
			-> StmtDomDescription -- ^ @newDescr@
			-> Either TStatementID TrDependenceID
			-> IneqSystem
projectDomDescriptionTo csDS 
			oldIneqs 
			(StmtDomDescription oldDescr) 
			(StmtDomDescription newDescr) 
			someID = 
			    assert (length (concatMap snd oldDims) == nVars oldIneqs) newIneqs
    where    
    oldDims = 
	zip oldDescr 
		(getStmtDomDims csDS (StmtDomDescription oldDescr) someID)
    newDims = concatMap lookupPart newDescr

    newIneqs = projectToVarsMC oldIneqs newDims

    lookupPart p =
	fromMaybe (error ("projectDomDescriptionTo: " ++ 
			  "part of domain description not found: " ++ show p)) 
	              (lookup p oldDims)

-- |
-- Compute a list of dimensions for the statement description
-- given by @stmtDesc@ for either a given target statement
-- or a given target dependence.
--
getStmtDomDims :: CompSpreadDS 
	       -> StmtDomDescription 
	       -> Either TStatementID TrDependenceID 
	       -> [[Integer]]
getStmtDomDims csDS (StmtDomDescription stmtDesc) someID =
    snd $  mapAccumL (stmtDomDimHelper csDS someID) 0 stmtDesc

stmtDomDimHelper :: CompSpreadDS 
		 -> Either TStatementID TrDependenceID 
		 -> Integer 
		 -> StmtDomPart 
		 -> (Integer, [Integer])
stmtDomDimHelper csDS someID dimAccum (DpTile dimInfo) = 
    (dim + dimAccum, [dimAccum..dimAccum+dim-1])
      where
      qual = dimQual dimInfo
      dim = case qual of
	    Sched -> getTimeTileDim csDS
	    Place -> getProcTileDim csDS
            Additional -> error ("stmtDomHelper: no additional tiling" ++
				 " dimension possible!")      
stmtDomDimHelper csDS someID dimAccum (DpTISPC dimInfo) = 
    (dim + dimAccum, [dimAccum..dimAccum+dim-1])
      where
      dim = case (dimQual dimInfo) of
	    Sched -> getSchedDim csDS
	    Place -> getPlaceDim csDS
	    Additional -> (getCoordsDim csDS relevantStmtID -
                           (getSchedDim csDS + getPlaceDim csDS))
      qual = dimQual dimInfo
      relevantStmtID = getRelevantStmtID csDS someID dimInfo
stmtDomDimHelper csDS someID dimAccum (DpISPC dimInfo) = 
    (dim + dimAccum, [dimAccum..dimAccum+dim-1])
    where
    relevantStmtID = getRelevantStmtID csDS someID dimInfo
    dim = getISPCdim csDS relevantStmtID    

getRelevantStmtID :: CompSpreadDS 
		  -> Either TStatementID TrDependenceID 
		  -> DimInfo 
		  -> TStatementID
getRelevantStmtID csDS someID dimInfo = relevantStmtID
    where
    aTrStmtID = either id (const $ error "TStatementID should not be accessed!") someID
    aTrDepID  = either (const $ error "TrDependenceID should not be accessed!") id someID
    trDep = fromMaybe (error "getRelevantStmtID: no TrDep for this TrDependenceID") $ 
	    find (\d -> trDepID d == aTrDepID) (csTrDeps csDS)
    srcID = trDepSrcStmtID trDep
    destID = trDepDestStmtID trDep
    role = dimRole dimInfo
    relevantStmtID = 
	case role of
		 (Just SendSide) -> srcID
		 (Just RecvSide) -> destID
		 Nothing         -> aTrStmtID




getParamsDim :: CompSpreadDS -> Integer
getParamsDim csDS = paramsDim
    where
    (_, ispcs) = unzip $ csISPCs csDS
    paramsDims = map nParams ispcs
    paramsDim = if allEq paramsDims
		then head paramsDims
		else error ("all ISPCs should have" ++
			    " the same number of parameters!")

---
-- Returns the number of variables in the corresponding original index space
-- for a given target statement ID
getISPCdim :: CompSpreadDS -> TStatementID -> Integer
getISPCdim csDS = 
    nVars . getISPC csDS

getMaxISPCdim :: CompSpreadDS -> Integer
getMaxISPCdim csDS = maximum $ map (getISPCdim csDS) $ getAllTStatementIDs csDS

---
-- Counts the number of time dimensions in the target program
-- 
-- @return timeDim, the number of schedule dimensions in the target program
getSchedDim :: CompSpreadDS -> Integer
getSchedDim CompSpreadDS { csQualVects = qualVects } = 
    -- trace ("\ngetSchedDim: " ++ (show timeDim) ++ "\n") $
    timeDim
    where
    (_, qualEntries) = unzip qualVects
    schedEntries = map (filter (\e -> e == Sched)) qualEntries
    timeDimNumbers = map genericLength schedEntries
    timeDim = 
	if allEq timeDimNumbers
	then head timeDimNumbers
	else error ("number of schedule dimensions not equal for " ++
                    "all target statements!")

getPlaceDim :: CompSpreadDS -> Integer
getPlaceDim CompSpreadDS { csQualVects = qualVects } = 
    -- trace ("\ngetPlaceDim: " ++ (show $ head placeDims) ++ "\n") $
    if allEq placeDims
    then head placeDims
    else error ("number of placement dimensions not equal for " ++
                "all target statements!")
    where
    (_, qualEntries) = unzip qualVects
    placeEntries = map (filter (\e -> e == Place)) qualEntries
    placeDims = map genericLength placeEntries	

getCoordsDim :: CompSpreadDS -> TStatementID -> Integer
getCoordsDim CompSpreadDS { csQualVects = qualVects } trStmtID = 
    genericLength qualVect
    where 
    qualVect = 
	fromMaybe (error $ "getCoordsDim: no quality vector" ++
		   "found for this TStatementID") $ 
	lookup trStmtID $ qualVects

getCommonCoordsDim :: CompSpreadDS -> Integer
getCommonCoordsDim CompSpreadDS { csQualVects = qualVects } = 
    if allEq coordsDims 
    then head coordsDims 
    else error ("getCommonCoordsDim: varying number " ++ 
		"of schedule and placement dimensions")
    where
    (_, qualEntries) = unzip qualVects
    coordsDims = map (genericLength .
		      filter (\e -> e == Sched || e == Place)) qualEntries

getMaxCoordsDim :: CompSpreadDS -> Integer
getMaxCoordsDim CompSpreadDS { csQualVects = qualVects } = maxCoordsDim
    where
    maxCoordsDim = maximum $ map genericLength $ snd $ unzip qualVects

getCommonQualVect :: [(TStatementID, [QualEntry])] -> [QualEntry]
getCommonQualVect qualVects = head $ commonQualVects
    where
    (_, qualEntries) = unzip qualVects
    commonQualVects = map (filter (\e -> e == Sched || e == Place)) qualEntries


getGlobalTimeDim :: CompSpreadDS -> Integer
getGlobalTimeDim csDS = if True -- timeTiling 
			then 1 -- timeTileDim
			else getSchedDim csDS
    where
--    timeTiling = csTimeTiling csDS
    tilingDim = genericLength $ csLatVects csDS
    timeTileDim = min tilingDim (getSchedDim csDS)

-- returns the correct number of rp-dims in both cases of tiling or non-tiling
getRPdim :: CompSpreadDS -> Integer
getRPdim csDS = 
    -- trace ("\ngetRPdim: " ++ show rpDim ++ "\n") $
    rpDim
    where
    rpDim = if True -- procTiling 
	    then tilingDim - timeTileDim
	    else getPlaceDim csDS
--    procTiling = csSpaceTiling csDS
    tilingDim = getTilingDim csDS
    timeDim = getSchedDim csDS
    timeTileDim = min tilingDim timeDim

getTilingDim :: CompSpreadDS -> Integer
getTilingDim csDS = genericLength $ csLatVects csDS

getTimeTileDim :: CompSpreadDS -> Integer
getTimeTileDim csDS = min (getTilingDim csDS) (getSchedDim csDS)

getProcTileDim :: CompSpreadDS -> Integer
getProcTileDim csDS = 
    -- trace ("\ngetProcTileDim: " ++ show procTileDim ++ "\n") $
    procTileDim
    where
    procTileDim = max (getTilingDim csDS - getSchedDim csDS) 0

getArrAccessMaxDim :: CompSpreadDS -> Integer
getArrAccessMaxDim csDS = maximum accessTargetDims
    where
    accesses = csAccesses csDS
    accessTargetDims = map (targetDimIH.accIndexFct) accesses

getArrayDim :: CompSpreadDS -> ArrayID -> Integer
getArrayDim csDS arrayID = targetDimIH $ accIndexFct acc
    where
    accs = csAccesses csDS
    acc = 
      fromMaybe (error "in getArrayDim: wrong ArrayID!")
                    (find (\acc -> arrayID == accArrayID acc) accs)
			  
getDepDomainDims :: CompSpreadDS -> TrDependence -> (Integer, Integer)
getDepDomainDims CompSpreadDS { csISPCs = ispcs} trDep =
    -- verbose "getDepDomains" $
    -- trace (show dep) $
    (nVars srcDom, nVars destDom)
    where
    origSrcStmtID = trDepOrigSrcStmtID trDep
    origDestStmtID = trDepOrigDestStmtID trDep
    srcDom = 
        fromMaybe (error "getDepDomainDims: no ISPC for this StatementID found!") $
	          lookup origSrcStmtID ispcs
    destDom = 
        fromMaybe (error "getDepDomainDims: no ISPC for this StatementID found!") $
	          lookup origDestStmtID ispcs

