module CompSpread.Helpers
where

import List
import Maybe
import Ratio
import Debug.Trace(trace)
import GHC.Base(assert)
import System.Time
import System.IO.Unsafe

import LMath(LRV(..), dualOfSystem)
import LMath.Types
import LMath.Algo.Sort


--
-- Debugging function. Can be used to display the results of finished functions.
--
verbose :: String -> a -> a
verbose str x =
    trace ("$? " ++ str ++ "...") $
    let skipIt = False -- suppress debugging output
    in if skipIt then x
       -- else seq x $ trace ("Verbose: " ++ str ++ ": result=" ++ show x) x
       else seq x $ trace ("$! " ++ str ++ " done. ") x

getTime :: String
getTime = timeStr
    where
    clockTime = unsafePerformIO getClockTime
    calendarTime = unsafePerformIO $ toCalendarTime clockTime
    timeStr =  calendarTimeToString calendarTime

projectToVarsMC is vars = 
    verbose "projectToVarsMC" $
    fromMaybe (trace "warning: projectToVarsFM delivered empty IS" $
	       emptyIS) projected
    where
    emptyIS = ineqSystem ((varsDim, paramsDim), 
			  empty_M)
    empty_M = 
        -- contradicting inequality: -1 >= 0
	(zeroMatrix 1 (varsDim + paramsDim) `conc`
	 matrix [[-1%1]])
    varsDim = genericLength vars
    paramsDim = nParams is			  
    projected = projectToVars is Nothing vars

sortIneqsMC is =
    -- verbose "sortIneqsMC" $
     fromMaybe (trace "warning: sortIneqs delivered empty IS" $
		emptyIS) sortedIS
    where
    emptyIS = ineqSystem ((varsDim, paramsDim), 
			  empty_M)
    empty_M = 
        -- contradicting inequality: -1 >= 0
	(zeroMatrix 1 (varsDim + paramsDim) `conc`
	 matrix [[-1%1]])
    varsDim = nVars is
    paramsDim = nParams is			  
    sortedIS = sortIneqs is Nothing
    

--- 
-- shows a "LooPo style" IneqSystem in Cloog-conform syntax
showIS2cloog :: IneqSystem -> String
showIS2cloog is = output
    where
    rawM = getMatrix $ fmap fromIntegral $ integralIneqSystem is
    firstColumn = columnVectorsToMatrix1 [vector (replicate (nRows rawM) (1%1))]
    finalM = firstColumn `conc` rawM
    output = show (nRows finalM) ++ " " ++ show (nColumns finalM) ++ "\n" ++
	     intShowMat finalM False

--- 
-- shows a scatter function matrix in Cloog-conform syntax
showScatterMatrix2Cloog :: Matrix -> String
showScatterMatrix2Cloog m = output
    where
    zeroColumn = columnVectorsToMatrix1 [vector (replicate (nRows m) (0%1))]
    finalM = zeroColumn `conc` m
    output = show (nRows finalM) ++ " " ++ show (nColumns finalM) ++ "\n" ++
	     intShowMat finalM False ++ "\n"

myMultijoin1 ms = 
    assert (not $ null ms) $
    if ok 
    then multijoin1 ms
    else error ("multijoin1 failed with different columns in submatrices: " ++ 
		show cols)
    where
    ok = all (== head cols) cols
    cols = map nColumns ms

allEq :: (Eq a) => [a] -> Bool
allEq [] = False
allEq (x:xs) = all (==x) xs

fst3 :: (a, b, c) -> a
fst3 (x, _, _) = x

argMax f = maximumBy (\x y -> f x `compare` f y)

areIneqsUnbounded :: IneqSystem -> Bool
areIneqsUnbounded is = containsRays
    where
    containsRays = not $ null $ lrvRays dualLRV
    dualLRV = dualOfSystem eqs is
    eqs = eqSystem ((nVars is,nParams is), zeroMatrix 0 (nColumns is))


ineqsUnbounded :: IneqSystem -> [IneqSystem]
ineqsUnbounded is = projectedIS
    where
    varDim = nVars is
    projectedIS = map (projectToVarsMC is) [[d] | d <- [0.. (varDim - 1)]]

showLRV :: IneqSystem -> LRV
showLRV is = dualOfSystem eqs is
    where
    eqs = eqSystem ((nVars is,nParams is), zeroMatrix 0 (nColumns is))
