/* PROTOTYP: LASTAUSGLEICHSPROTOKOLL (LBP)
 * Zweck: Zum Ausschlachten fr die Codegenerierung
 * Hauptaugenmerk: 
 * - Serverschleife
 * - Umverteilungszyklus beim Client
 * - Kommunikatoren (MPI_COMM_SERVER / MPI_COMM_CLIENT) und Interkommunikatoren
 */

#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>
#include <assert.h>
#include <float.h>
// ^^^ wegen NaN, +Inf, -Inf
#include <math.h>
// ^^^ wegen floor
#include <list>
#include <vector>
#include "ProcessorListMap.cpp"
#include "NTupel.h"
#include "comp_spread_lib.h"

// Types
struct MapUpdate {
    ProcessorInterval VPinterval;
    int RPnum;
};

struct RedistributionInformation {
    ProcessorInterval VPinterval;
    int RPnum;
};

struct LoadInfo {
  double timeSpent;
  int globalTimeStep;
};

// a defined MPI_Datatype, consisting of 1 integer, followed by 1 double
// see initMPIDatatypes
MPI_Datatype MPI_LoadInfo; 


// Constants
#define LBP_MESSAGE_TAG MPI_TAG_UB
#define NaN (DBL_MAX/DBL_MIN) * 0
#define POSINF DBL_MAX
#define NEGINF DBL_MIN
const double UNBALANCE_RATIO_TRESHOLD = 0.1; //initiate load balance if unbalance exeeds this value

// Constants, temporary ones
//const int NUM_BALANCE_CYCLES = 10; //Anzahl Umverteilungszyklen
//const int SIZE_BALANCE_CYCLES = 10000; //Anzahl Schleifendurchlufe pro Umverteilungszyklus; Dummy-Wert
const int LBP_SERVER_ID = 0; //ID of responsible server in MPI_GROUP_WORLD; at this time, this is always =0, though there may once be support for more than one server
const int NUM_SERVERS = 1; //number of LBP servers; at this time =1 
int nVP; //number of virtual processors / processor tiles

// Ranks and sizes
int mpiRank; //rank in MPI_COMM_WORLD
int mpiSize; //size of MPI_COMM_WORLD
int selfLBP; //rank in MPI_COMM_SERVER or MPI_COMM_NODES
int nServers; //at this time, there is only one server (=1), its rank is LBP_SERVER_ID
int nNodes; //size of MPI_COMM_NODES

// Buffers
::std::vector<int> rdispls;
::std::vector<double> sendVect(1);
::std::vector<double> recvVect(1);


// Groups and communicators
MPI_Group MPI_GROUP_WORLD; //contains all processes in MPI_COMM_WORLD; used in LBP communication
MPI_Group MPI_GROUP_SERVER; //contains only the LBP servers; unused 
MPI_Comm MPI_COMM_SERVER;
MPI_Group MPI_GROUP_NODES; //contains all nodes, but not the server; used in communication between nodes
MPI_Comm MPI_COMM_NODES;

// Globals
MPI_Status status; //communication status


void initMPIDatatypes() {
  int blen[3]     = { 1, 1, 1 };
  int indices[3]  = { 0, sizeof(double), sizeof(LoadInfo) };
  int oldtypes[3] = { MPI_DOUBLE, MPI_INT, MPI_UB };
  MPI_Type_create_struct( 3, blen, indices, oldtypes, &MPI_LoadInfo );
}

/*
 * Initialises the two LBP specific communicators: 
 * - MPI_COMM_SERVER: contains LBP servers, but no nodes 
 * - MPI_COMM_NODES: contains all nodes, but no server
 * Notes:
 * - At this time, if new processors enter the computation after the call to
 *   this routine, the communicators may fail their specifications. 
 * - The routine is assumed to be called only once (memory leak). 
 * - If there are no nodes designated, an exception is thrown.
 * - Servers use the low indices of MPI_GROUP_WORLD, even if there will ever 
 *   be more than one of them.
 */
void initLbpCommGroups(int numServers) {
	int result;
	
	// Get group 'world'
	MPI_Comm_group(MPI_COMM_WORLD, &MPI_GROUP_WORLD); 
	
	MPI_Group_size(MPI_GROUP_WORLD, &result);
	assert(result == mpiSize); 
	
	// Create group and communicator 'server'
	int ranks[3] = {0,numServers-1,1}; 
	result = 
	  MPI_Group_range_incl(MPI_GROUP_WORLD, 1, &ranks, &MPI_GROUP_SERVER); 
	result = result ||
	  MPI_Group_size(MPI_GROUP_SERVER, &nServers);
	result = result ||
	  MPI_Comm_create(MPI_COMM_WORLD, MPI_GROUP_SERVER, &MPI_COMM_SERVER);
	
	if (result) 
	    throw "Failed to create communicator 'server'";
	
	// Create group and communicator 'nodes'
	result = 
	  MPI_Group_difference(MPI_GROUP_WORLD, MPI_GROUP_SERVER, &MPI_GROUP_NODES);
	result = result ||
	  MPI_Group_size(MPI_GROUP_NODES, &nNodes); 
	result = result ||
	  MPI_Comm_create(MPI_COMM_WORLD, MPI_GROUP_NODES, &MPI_COMM_NODES);
	
	if (result) 
	    throw "Failed to create communicator 'nodes'";
	
//	printf("%d: COMM_WORLD %d, COMM_SERVER %d, COMM_NODES %d\n", 
//		mpiRank, mpiSize, nServers, nNodes);
		
	// Assign IDs
	MPI_Group_rank(MPI_GROUP_NODES, &selfLBP);
	if (selfLBP == MPI_UNDEFINED) 
		MPI_Group_rank(MPI_GROUP_SERVER, &selfLBP);
	assert(selfLBP != MPI_UNDEFINED);
	
	// Check specs
	if (MPI_GROUP_NODES == MPI_GROUP_EMPTY)
		throw "Group 'nodes' is empty - no one left to do the calculations";
	
}


/*
 * loadBalance
 * Assigns changes to the ProcessorMap. 
 * @params: 
 * - IN: loadInformation: array of doubles; length = nNodes
 * - OUT: pMapChanges: reference to list of struct of ProcessorInterval
 *        and Processor; to be communicated and used in calls of 
 *        ProcessorMap::update; 
 * @uses: by execute_server()
 * @preconditions: 
 * - length(loadInformation) = nNodes
 * - nNodes >= 1
 * @TODO: loadBalance: substitute naive implementation with a better one
 */
void loadBalance(ProcessorMap &pMap,
				 LoadInfo* loadInformation, 
				 list<MapUpdate> &result) {
	// Dummy implementation: just print load Information
	printf("loadBalance: ");
	for (int i = 0; i < nNodes; i++) {
		printf("%4f ", loadInformation[i].spentTime);
	}
	printf("\n");

	printf("globalTimeStep: ");
	for (int i = 0; i < nNodes; i++) {
		printf("%d ", loadInformation[i].globalTimeStep);
	}
	printf("\n");
	
	// Naive implementation: Take load from slowest and add to fastest
	
	int slowestRP = -1;
	int fastestRP = -1;
	double slowestTime = NEGINF;
	double fastestTime = POSINF;
	
	for (int i = 0; i < nNodes; i++) {
		if (loadInformation[i].spentTime < fastestTime) { 
			fastestTime = loadInformation[i].spentTime;
			fastestRP = i;
		}
		if (loadInformation[i].spentTime > slowestTime) { 
			slowestTime = loadInformation[i].spentTime;
			slowestRP = i;
		}		
	}
	assert( (slowestRP != -1) && ( fastestRP != -1) );
	
	result.clear(); //make empty 
	
	// redistribute load only if unbalance treshold is exceeded
	if (slowestTime <= fastestTime * (1+UNBALANCE_RATIO_TRESHOLD)) return; 
	assert(slowestRP != fastestRP );
	
	if ((slowestTime == 0)) return; //avoid division by zero
	
	double overloadRatio = (slowestTime - fastestTime) / slowestTime;
	int *slowestVPs; //VPs of slowest RP; allocated by pMap.getVPs
	slowestVPs = 0;
	slowestVPs = pMap.getVPs(slowestRP); //slowestVPs[0] == length(slowestVPs)-1
	int numVPs = slowestVPs[0];
	// Assure that there is load at the slowest RP to redistribute
	if (numVPs == 0) {
		delete [] slowestVPs;
		return;
	}
	assert(numVPs > 0);
	// Take half of the overload and transfer it to fastest node
	int numVPsToMove = static_cast<int>(floor((overloadRatio * 0.5) * numVPs));
	
	ProcessorInterval tmpInterval;
	MapUpdate tmpUpdate;
	if (slowestRP < fastestRP) { //avoid fragmentation of VPs
		// Redistribute high VPs
		int i = (numVPs+1) - numVPsToMove;
		while(i <= numVPs) { //this loop counts intervals
			tmpInterval.minimumProcessor = slowestVPs[i];
			tmpInterval.maximumProcessor = tmpInterval.minimumProcessor;
			// Merge intervals to redistribute
			while(i < numVPs) { //this loop counts interval elements
				if (slowestVPs[i+1] != tmpInterval.maximumProcessor+1)
					break; //2nd loop condition (avoid need of lazy evaluation)
				tmpInterval.maximumProcessor++;
				i++; //next element
			}
			// Append VP interval to result
			assert(tmpInterval.minimumProcessor >= 0);
			assert(tmpInterval.maximumProcessor < nVP);
			tmpUpdate.VPinterval = tmpInterval;
			tmpUpdate.RPnum = fastestRP; 
			result.push_back(tmpUpdate);
			i++; //next interval
		}
	} else {
		// Redistribute low VPs
		int i = 1;
		while(i <= numVPsToMove) { //this loop counts intervals
			tmpInterval.minimumProcessor = slowestVPs[i];
			tmpInterval.maximumProcessor = tmpInterval.minimumProcessor;
			// Merge intervals to redistribute
			while(i < numVPsToMove) { //this loop counts interval elements
				if (slowestVPs[i+1] != tmpInterval.maximumProcessor+1)
					break; //2nd loop condition (avoid need of lazy evaluation)
				tmpInterval.maximumProcessor++;
				i++; //next element
			}
			// Append VP interval to result
			assert(tmpInterval.minimumProcessor >= 0);
			assert(tmpInterval.maximumProcessor < nVP);
			tmpUpdate.VPinterval = tmpInterval;
			tmpUpdate.RPnum = fastestRP; 
			result.push_back(tmpUpdate);
			i++; //next interval
		}
	}
	
	delete [] slowestVPs;
}


/*
 * Code executed by LBP server
 */
void execute_server(ProcessorMap &pMap) {
	
	printf("%i: Designated as server (#%i of %i); number of nodes is %i\n", 
		mpiRank, selfLBP, nServers, nNodes);
		
	if (selfLBP != LBP_SERVER_ID) {
		printf("%i, server%i: finished (no work for secondary servers at this time)\n",
			   mpiRank, selfLBP);
		return;
	}

	MPI_Request *request = new MPI_Request[nNodes]; //for pending communications
	double timeStarted, timeFinished; //statistics

	//load information of all nodes; important: ordering must be the same as in MPI_GROUP_NODES
	LoadInfo *loadInfo = new LoadInfo[nNodes]; 

	list<MapUpdate> pMapUpdate; //list of pMap updates, managed by 'loadBalance'
	list<MapUpdate>::iterator it;
	RedistributionInformation *redistInfo; //redistribution information send buffer

	// Set start time 
	timeStarted = MPI_Wtime();
	
	//for(int i = 0; i < NUM_BALANCE_CYCLES; i++) {
	while(true) {
		
		// STEP 1: gather load information from all nodes (blocking receive)
		
		// Reset load info buffer to visualise errors (dispensable)
		for (int j = 0; j < nNodes; j++) {
			loadInfo[j].spentTime = NaN;
			loadInfo[j].globalTimeStep = -100000;
		}
		
		LoadInfo tmpLoadInfo;
		for(int j = 0; j < nNodes; j++) {
			 // Receive from any node
			 // TODO: This code does not work correctly: not every value in loadTime is being set
			 MPI_Recv(&tmpLoadInfo, 1, MPI_LoadInfo, MPI_ANY_SOURCE, 
			          LBP_MESSAGE_TAG, MPI_COMM_WORLD, &status);
			 // Determine, from which node load information was received
			 assert(status.MPI_SOURCE >= NUM_SERVERS); //only receive from nodes
			 loadInfo[status.MPI_SOURCE - NUM_SERVERS] = tmpLoadInfo;
		}
	
		// STEP 2: calculate redistribution of load
		// TODO: As MPI needs redist buffer anyway, change interface to array
		loadBalance(pMap, loadInfo, pMapUpdate);
		
		// STEP 3: fill redistribution information send buffer
		// buffer will contain interval start & end, RPnum and time values
		redistInfo = new RedistributionInformation[pMapUpdate.size()];
		{ int k = 0;
			for(it = pMapUpdate.begin(); it != pMapUpdate.end(); it++) {
				redistInfo[k].VPinterval = it->VPinterval;
				redistInfo[k].RPnum = it->RPnum;
				//redistInfo[k].time = i * SIZE_BALANCE_CYCLES;
				k++;
			}
		}
		
		// STEP 4: send load redistribution information (nonblocking send or broadcast) 
		// Point-to-Point communication is used, as to not synchronise the internet
		for(int j = 0; j < nNodes; j++) {
			MPI_Isend(redistInfo, 
					  pMapUpdate.size() * sizeof(RedistributionInformation), 
					  MPI_INT, j + NUM_SERVERS, 
					  LBP_MESSAGE_TAG, MPI_COMM_WORLD, &request[j]); 
		}
		printf("%i, server%i: sent processor map update information (%i intervals)\n",
			mpiRank, selfLBP, pMapUpdate.size());
		
		// STEP 5: apply load redistribution information to ProcessorMap
		for(it = pMapUpdate.begin(); it != pMapUpdate.end(); it++)
			pMap.update(it->VPinterval, //VP interval to update
				    it->RPnum, //new RP to reside on
				    tmpLoadInfo.globalTimeStep+SIZE_BALANCE_CYCLES); //momentary time
		
		// STEP 6: Assure that sending load balancing information is completed
		// This step is crucial for the success of following MPI_Recvs with MPI_ANY_SOURCE
		for(int j = 0; j < nNodes; j++) {
			MPI_Wait(&request[j], &status); //wait for previously started Isend 
		}
		
		// STEP 7: Free memory
		// Dynamically allocated redistribution information send buffer
		delete [] redistInfo; 
		
	}
	
	delete [] request; //free Isend request status buffers
	delete [] loadInfo; //free load information buffer
	
	// Set end time 
	timeFinished = MPI_Wtime();
	
	printf("%i, server%i: finished; %6f seconds elapsed\n", 
		mpiRank, selfLBP, timeFinished - timeStarted);
}


/*
 * Code executed by LBP nodes
 * Note: Momentarily exactly one server is used
 */
void execute_node(ProcessorMap &pMap) {

  printf("%i: Designated as node (#%i of %i)\n", 
	 mpiRank, selfLBP, nNodes);

  MPI_Request request; //for pending communications
  double loadTime;
  int recvCount; //length of dynamic receive buffer
  ::std::vector<int> recvBuffer; //dynamic receive buffer for ProcessorMap update
  RedistributionInformation redistInfo; //redistribution information
  int recvBufferOffset;
		

  // insert project name for "project" here:
#include "project.src.out.defines"
#include "project.src.out.finalOutput"        
	
}

void initialize_node() {

  // insert project name for "project" here:
#include "project.src.out.rpDefines"
#include "project.src.out.rpMapOutput"


  nVP = loopo___bufferID;
  
  printf("nVPs: %d", nVP); 
  rdispls.resize(mpiSize);

  for (int i = 0; i < " ++ mpiSize_t ++ "; i++) {
    sendVect.push_back(::std::vector<double>());
  }
}


/*
 * main
 * Initialising, calling, exception handling
 */
int main(int argc, char *argv[]) {
	

	
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &mpiRank);
    MPI_Comm_size(MPI_COMM_WORLD, &mpiSize);
    
//    printf("%d: %d processes in world\n", mpiRank, mpiSize);
    	
	try {
    	
    	// Initialise communicators
    	initLbpCommGroups(NUM_SERVERS); //only one rank, namely 0, will be server

	initMPIDatatypes();
		
    	// Initialise ProcessorMap
    	ProcessorListMap pMap(nVP, mpiSize - NUM_SERVERS);
    	
    	// Determine if designated as server
    	int isServer;
    	MPI_Group_rank(MPI_GROUP_SERVER, &isServer);
    	
    	if (isServer != MPI_UNDEFINED) {
    		
    		execute_server(pMap);
    		
    	} else {
          initialize_node();
          execute_node(pMap);
    		
    	}
    	
    	// Print final pMap
    	printf("%i final pMap:\n",mpiRank); 
    	pMap.toString();

	} 
	catch(char const* e_string) {
		printf("%i: Exception encountered: %s\n", mpiRank, e_string);
	} //other exceptions: not caught in main
	
	MPI_Finalize();
	
	return 0;
}
