#include "ProcessorMap.cpp"
#include <vector>
#include <list>
#include <stdio.h> 
//^^ fr Test
#include <math.h>
//^^ fr ceil
#include <bitset>
//^^ fr update RPtoVP

using namespace std;

class InvalidArgumentException : public std::exception {};
class InvalidStateException : public std::exception {};

struct ProcessorTimeTuple {
    int time;
    int RPnum;
};

/*
 *
 */
class ProcessorListMap : public ProcessorMap {
    
    private: 
    
    /*
     * processorIntervals_insert
     * Designed to be able to insert an interval of processors into existing
     * intervals. It is expected that adjacent results are fusioned to keep 
     * the number resulting intervals small and duplicate free. 
     * Preconditions: 
     * - empty intersection between 'toChange' and 'toInsert' 
     *   (call 'processorIntervals_remove' if not sure)
     * - forall intervals of toChange a,b : succ(a) = b : a.max < b.min
     *   (intervals are in ascending order and "optimal")
     * - trivial: forall intervals a : a.minimumProcessor <= a.maximumProcessor
     */
    void processorIntervals_insert(list<ProcessorInterval> &toChange,
                                   ProcessorInterval toInsert) {
      
        // Find insertion spot
        list<ProcessorInterval>::iterator it;
        it = toChange.begin();
        while (it != toChange.end()
               && it->maximumProcessor < toInsert.minimumProcessor-1) { //lazy eval wichtig
            it++;
        }
        it = toChange.insert(it, toInsert); 
        
        // Now: toInsert.maximumProcessor >= succ(toInsert).minimumProcessor-1
        //   && (forall pred*(toInsert) a : a.maximumP < toInsert.minimumP)
        //   (all predecessors don't need to be changed)
        
        // Eliminate duplicates after 'toChange' in all successor intervals. 
        // Those relevant are: 
        //   pred*(toInsert) a : toInsert.maximumProc >= a.minimumProc-1
        list<ProcessorInterval>::iterator itToInsert;
        itToInsert = list<ProcessorInterval>::iterator(it);
        it++;
        
//        // Debug
//        deque<ProcessorInterval>::iterator itTest;
//        itTest = toChange.begin();
//        while (itTest != toChange.end()) {
//            itTest++;
//        } 
        
        while (it != toChange.end() //lazy eval wichtig
               && it->minimumProcessor-1 <= itToInsert->maximumProcessor) { 
            // Merge intervals into one
            itToInsert->minimumProcessor = min(itToInsert->minimumProcessor,
                                               it->minimumProcessor);
            itToInsert->maximumProcessor = max(itToInsert->maximumProcessor,
                                               it->maximumProcessor);
            it = toChange.erase(it);            
        }
    }
    
    /* 
     * processorIntervals_remove
     * Removes those processors that match from the given intervals. Those 
     * processors that don't match are returned as a list of intervals (as they
     * may have become split). 
     * Need of efficiency: 
     * - Overall runtime
     * - Minimal number of intervals at 'toChange'
     * - Minimal number of returned intervals
     * @param toChange: intervals of processors from which to remove processors
     * @param toRemove: processors to remove; precondition:
     *   forall intervals of toRemove a,b : succ(a) = b : a.max <= b.min
     *   (intervals are in ascending order)
     * @result: elements of 'toRemove', which did not exist in 'toChange'
     * Note: Beware of memory leaks: The return value (== toRemove) 
     * has to be deleted by the caller.
     */
    list<ProcessorInterval> &processorIntervals_remove(
                                   list<ProcessorInterval> &toChange, 
                                    list<ProcessorInterval> &toRemove) {
        
        list<ProcessorInterval>::iterator itToRemove;
        itToRemove = toRemove.begin();
        
        list<ProcessorInterval>::iterator itToChange;
        itToChange = toChange.begin();
        while (itToChange != toChange.end()
               && itToChange->minimumProcessor > itToRemove->maximumProcessor) { //lazy eval wichtig
            itToChange++;
        }
        // Now: itToChange->minimumProcessor <= itToRemove->maximumProcessor
        int eraseMinProcessor;
        int eraseMaxProcessor;
        int tmpProcessor;
        ProcessorInterval tmpInterval;

        while (itToChange != toChange.end() //lazy eval wichtig
               && itToRemove != toRemove.end() //lazy eval wichtig
               && itToChange->minimumProcessor <= itToRemove->maximumProcessor) { 
            eraseMinProcessor = max(itToChange->minimumProcessor,
                                    itToRemove->minimumProcessor);
            eraseMaxProcessor = min(itToChange->maximumProcessor,
                                    itToRemove->maximumProcessor);  
            // eraseMinProcessor <= eraseMaxProcessor                      
            // Step 1: Erasing in 'toChange'
            if (itToChange->minimumProcessor < eraseMinProcessor 
               && itToChange->maximumProcessor > eraseMaxProcessor) { 
                // Interval of 'toChange' is being split
                tmpProcessor = itToChange->maximumProcessor;
                itToChange->maximumProcessor = eraseMinProcessor-1;
                tmpInterval.minimumProcessor = eraseMaxProcessor+1;
                tmpInterval.maximumProcessor = tmpProcessor;
                itToChange++;
                itToChange = toChange.insert(itToChange, tmpInterval); //interval done
            } else if (itToChange->minimumProcessor < eraseMinProcessor) {
                // high end of interval is cut
                itToChange->maximumProcessor = eraseMinProcessor-1;  
                itToChange++; //interval done                
            } else if (itToChange->maximumProcessor > eraseMaxProcessor) {
                // low end of interval is cut
                itToChange->minimumProcessor = eraseMaxProcessor+1;  
                itToChange++; //interval done  
            } else {
                // interval is completely removed
                itToChange = toChange.erase(itToChange); //interval done
            }
            
            // Step 2: Erasing in 'toRemove'  
            if (itToRemove->minimumProcessor < eraseMinProcessor 
               && itToRemove->maximumProcessor > eraseMaxProcessor) { 
                // Interval of 'toChange' is being split
                tmpProcessor = itToRemove->maximumProcessor;
                itToRemove->maximumProcessor = eraseMinProcessor-1;
                tmpInterval.minimumProcessor = eraseMaxProcessor+1;
                tmpInterval.maximumProcessor = tmpProcessor;
                itToRemove++;
                itToRemove = toRemove.insert(itToRemove, tmpInterval);
            } else if (itToRemove->minimumProcessor < eraseMinProcessor) {
                // high end of interval is cut
                itToRemove->maximumProcessor = eraseMinProcessor-1;  
            } else if (itToRemove->maximumProcessor > eraseMaxProcessor) {
                // low end of interval is cut
                itToRemove->minimumProcessor = eraseMaxProcessor+1;  
            } else {
                // interval is completely removed
                itToRemove = toRemove.erase(itToRemove);
            }
            
        }

        
        return toRemove;
        
        /*
        delete(&toRemove);
        return *(new deque<ProcessorInterval>());
        */
    }    
    
    protected: 
    
    /**
     * VPtoRP
     * Holds the mapping history of VPs to RPs.
     * Every VP has a history of RPs it belonged to. 'time' is the starting 
     * time of affiliation. 
     * Type: Array of Lists of ProcessorTimeTuples
     */ 
    list< ProcessorTimeTuple > *VPtoRP; 
    
    /**
     * RPtoVP
     * Holds the momentary mapping of RPs to VPs.
     * Every RP may have several intervals of VPs. The number of intervals
     * as well as the number of RPs may change any time.
     * Constraint: Every VP is listed exactly once (over all Intervals over all
     *   RPs)
     * Type: Vector of Lists of ProcessorIntervals
     */
    vector< list< ProcessorInterval > > RPtoVP;
    
    public:
    
    //ProcessorListMap(); //do not use
    
    ProcessorListMap(const ProcessorListMap&);
    
    /**
     * Minimal constructor. 
     * It only specifies the number of virtual and real processors initially
     * in the system. 
     * The number of real processors may change during runtime.
     * VPs are initially uniformly distributed among RPs.
     * @param VPcount: number of virtual processors / tiles > 0
     * @param RPcount: number of real processors > 0 
     */ 
    ProcessorListMap(const int VPcount, const int RPcount) {
        ProcessorTimeTuple tmpTuple;        
        this->VPcount = VPcount;
        this->RPcount = RPcount; 
        
        int VPsPerRP = static_cast<int>(
                       ceil((static_cast<float>(VPcount))/RPcount));
        
        // VPtoRP initialization
        tmpTuple.time = 0;
        VPtoRP = new list<ProcessorTimeTuple> [VPcount];
        for(int i = 0; i < VPcount; i++) {
            tmpTuple.RPnum = i / VPsPerRP;
            VPtoRP[i].push_back(tmpTuple);
        }

        // RPtoVP initialization
        RPtoVP.resize(RPcount);
        ProcessorInterval tmpInterval;
        int j;
        for(int i = 0; i < RPcount; i++) {
            j = 0;
            while(j < VPcount) {
                if (VPtoRP[j].front().RPnum == i) {
                    tmpInterval.minimumProcessor = j; //minimum VP of interval
                    while(j < VPcount) {
                      if (VPtoRP[j].front().RPnum == i) { //avoid need of lazy eval 
                        tmpInterval.maximumProcessor = j; //maximum VP of interval
                        j++;
                      } else break;
                    }
                    RPtoVP.at(i).push_back(tmpInterval);                    
                }
                j++;
            }
        }   
    }   
       
    /**
     * Constructor for defining initial distribution of VPs 
     * (with some limitations).
     * @see: Minimal constructor
     * @param firstVPsOfInterval: length is equal to RPcount;
     *  [ firstVPsOfInterval[n]; firstVPsOfInterval[n+1]-1 ] belongs to 
     *  RP with index n. 
     */
    ProcessorListMap(const int VPcount, const int RPcount, 
                     const int* firstVPsOfInterval) {
        this->VPcount = VPcount;
        this->RPcount = RPcount;  
        
        // TODO: implement
        
        throw InvalidStateException(); //not implemented yet
    }
    
    /**
     * Destructor
     * Note: functionality not tested
     */
    ~ProcessorListMap() {
        delete [] VPtoRP;
    }
    
    virtual int getRP(const int VPnum, const int timeStep) {
        list<ProcessorTimeTuple>::reverse_iterator it;
        for( it = VPtoRP[VPnum].rbegin(); it != VPtoRP[VPnum].rend(); it++) {
            if (it->time <= timeStep) {
                return it->RPnum;
            }
        }
        throw InvalidStateException(); 
    }
         
    
    virtual int *getVPs(const int RPnum) const {
        list< ProcessorInterval > tmpList;
        tmpList = RPtoVP.at(RPnum);
        list< ProcessorInterval >::iterator it;
        
        // Count number of VPs
        int numberOfVPs = 0;
        for(it = tmpList.begin(); it != tmpList.end(); it++) {
            numberOfVPs += it->maximumProcessor - it->minimumProcessor + 1; 
        } 
        
        int *resultVPs = new int[numberOfVPs+1];
        resultVPs[0] = numberOfVPs;
        int doneVPsSucc = 1; //number of VPs done plus one
        for(it = tmpList.begin(); it != tmpList.end(); it++) {
            for(int i = it->minimumProcessor; i <= it->maximumProcessor; i++) {
                resultVPs[doneVPsSucc] = i;
                doneVPsSucc++;
            }
        } 
        return resultVPs;
    }
    
    virtual void update(const ProcessorInterval VPinterval, 
                        const int RPnum, const int time) {

        // Range check
        assert( (RPnum >= 0) && (RPnum < RPcount) );
        assert( (VPinterval.minimumProcessor >= 0) 
                && (VPinterval.maximumProcessor < VPcount) );

        // RPtoVP Update
        
        list<ProcessorInterval> tmpIntervals = list<ProcessorInterval>(1, VPinterval);
        
        // Search and delete old mappings of VPs
        int oldRP;
        while(tmpIntervals.size() != 0) {
            // Find RP of first VP to remove
            oldRP = getRP(tmpIntervals.front().minimumProcessor, time);
            // Remove as many VPs from this RP as possible
            tmpIntervals = processorIntervals_remove(
                                   RPtoVP.at(oldRP), 
                                   tmpIntervals);
        }

        // Set new mapping
        processorIntervals_insert(RPtoVP.at(RPnum), VPinterval);        

        // VPtoRP Update
        
        ProcessorTimeTuple tmpTuple;
        tmpTuple.RPnum = RPnum;
        tmpTuple.time = time;
        for(int i = VPinterval.minimumProcessor; i <= VPinterval.maximumProcessor; i++) {
            if (VPtoRP[i].back().time < time) {
                // Old location's timestamp is past: append new tuple
                if (VPtoRP[i].back().RPnum != RPnum) {
                    VPtoRP[i].push_back(tmpTuple);
                } //else: No location change: do nothing
            } else if (VPtoRP[i].back().time == time) {
                // Old location's timestamp is the same: overwrite last location
                VPtoRP[i].back().RPnum = RPnum;
            } else {
                // Updating in the past not allowed
                throw InvalidArgumentException();
            } 
        }
        
               
    }
    
    void toString() {
        // Step 1: print VPtoRP mapping
        list<ProcessorTimeTuple>::iterator it;
        for(int i = 0; i < VPcount; i++) {
            printf("VP%d:", i);
            for( it = VPtoRP[i].begin(); it != VPtoRP[i].end(); it++ ) {
                printf(" (%d, RP%d)", it->time, it->RPnum); 
            }
            printf("\n");
        }
        //Step 2: print RPtoVP mapping
        list< ProcessorInterval > tmpList;
        list<ProcessorInterval>::iterator it2;
        for(int i = 0; i < RPcount; i++) {
            printf("RP%d:", i);
            tmpList = RPtoVP.at(i);
            for(it2 = tmpList.begin(); it2 != tmpList.end(); it2++) {
                printf(" [VP%d, VP%d]", it2->minimumProcessor, it2->maximumProcessor); 
            }
            printf("\n");
        }
    }

          
};
