module CompSpread.Run
where

import System.Cmd(system)
import Debug.Trace(trace)
import GHC.IOBase(ExitCode(..))
import IO(openFile, hGetContents, IOMode(..))
import List

import LoopoDS
import CompSpread.Data
import CompSpread.CloogStatements
import CompSpread.DomainDescriptions
import CompSpread.Names
import CompSpread.CloogInput
import CompSpread.Defines
import CompSpread.Helpers

import InsertIfs.Parser(parseSource)
import InsertIfs.ParserDefs(ParserState(..))

import ScannerParser.SymbolTable
import ScannerParser.ScannerParser
import ScannerParser.ParseTree
import ScannerParser.PrintCode(printCode, TargetLanguage(..))

data RunMode = RunMainProg
	     | RunRealProcMapping { runRPdim :: Integer }
             | RunExternal { runInputExtension :: String 
                           , runOutputExtension :: String
                           }

writeMainInputFile :: CompSpreadDS -> CloogStatements -> RunMode -> IO ()
writeMainInputFile csDS cloogStmts RunMainProg = 
    do
    let mainCloogInput = genMainCloogInput csDS cloogStmts
        loopoFile = csTempDir csDS ++ "/" ++ csProjFile csDS
    putStrLn "writing main CLooG-input file..."
    writeFile (loopoFile ++
	       cloogInputSuffix RunMainProg) mainCloogInput
    putStrLn "ClooG-input file done."

writeRPMapInputFile :: CompSpreadDS -> RPMapStatements -> RunMode -> IO ()
writeRPMapInputFile csDS rpMapStmts RunRealProcMapping { runRPdim = rpD } =
    do
    let rpMapInput = genRPMapInput csDS rpMapStmts
        loopoFile = csTempDir csDS ++ "/" ++ csProjFile csDS
        rpDim = getRPdim csDS
    putStrLn "writing rpMap-input file..."
    writeFile (loopoFile ++
	       cloogInputSuffix (RunRealProcMapping rpDim)) rpMapInput
    putStrLn "rpMap-input file done."

runCloog :: CompSpreadDS -> RunMode -> IO ()
runCloog csDS runMode = 
    let 
    cloogOptions = csCloogOptions csDS     
    loopoFile = csTempDir csDS ++ "/" ++ csProjFile csDS
    cloogInputFile = loopoFile ++ cloogInputSuffix runMode
    cloogOutputFile = loopoFile ++ cloogOutputSuffix runMode
    cloogFlags = getCloogFlags runMode
    cloogCommand = 
	csCloogExec csDS ++ " " ++ cloogFlags ++ cloogOptions ++
        " -o " ++ cloogOutputFile ++ " " ++ cloogInputFile
    in do 
       exitCode <- do 
		   trace ("running CLooG: " ++ show cloogCommand) $ 
			 system cloogCommand
       let (ok, errCode) = checkResult exitCode
       if ok then do putStrLn "cloog completed."
             else do putStrLn ("cloog exited with error code: " ++ 
			       (show errCode))       
    
getCloogFlags :: RunMode -> String
getCloogFlags RunMainProg = "-block 1 -esp 0 -csp 0 -cpp 1"
getCloogFlags RunRealProcMapping { runRPdim = rpDim } = 
    "-block 1 -esp 0 -csp 0 -cpp 1 -stop " ++ (show (rpDim))
getCloogFlags RunExternal {} = "-block 1 -esp 0 -csp 0 -cpp 1"
    

cloogInputSuffix RunMainProg = ".cloogInput"
cloogInputSuffix RunRealProcMapping {} = ".rpMapInput"
cloogInputSuffix RunExternal { runInputExtension = inpExt } = inpExt

cloogOutputSuffix RunMainProg= ".cloogOutput"
cloogOutputSuffix RunRealProcMapping {} = ".rpMapOutput"
cloogOutputSuffix RunExternal { runOutputExtension = outExt } = outExt

finalOutputSuffix = ".finalOutput"

checkResult :: ExitCode -> (Bool, Int)
checkResult ExitSuccess = (True, 0)
checkResult (ExitFailure x) = (False, x)

runPostprocessing :: CompSpreadDS -> IO ()
runPostprocessing csDS =
  do
  let baseFile = csTempDir csDS ++ "/" ++ csProjFile csDS
      inpFile = baseFile ++ cloogOutputSuffix RunMainProg
      outputFile = baseFile ++ finalOutputSuffix
      rpDim = getRPdim csDS
      globalTimeDim = getGlobalTimeDim csDS
      srcProcNames = 
          (concat $ intersperse ", " 
           [srcProc_t ++ show i | i <- [1..rpDim]])
      destProcNames =
          (concat $ intersperse ", " 
           [destProc_t ++ show i | i <- [1..rpDim]])
 
      srcRP = srcProc_t ++ show rpDim
      destRP = destProc_t ++ show rpDim
      globalTime_t = glT_t ++ "1"

--       otherRPid = otherProc_t ++ show rpDim

      -- spreadStmtNr = getStmtTypeNr SpreadStatement {}
      writeStmtNr  = getStmtTypeNr WriteStatement {}
      unpackStmtNr = getStmtTypeNr UnpackStatement {}
      sendStmtNr   = getStmtTypeNr SendStatement {}
      recvStmtNr   = getStmtTypeNr RecvStatement {}
      compStmtNr   = getStmtTypeNr ComputeStatement {}

      ifConstraint_t = 
         -- srcRP = getRP(srcProc, srcTime); destRP = getRP(destProc, glT)  "oder glT+1 ?"
         -- if(stmtType == spread || stmtType == comm ||
         --    (stmtType == comp && destRP == rank) ||
         --    ((stmtType == write || stmtType == send) && srcRP == rank && destRP != rank) ||
         --    ((stmtType == unpack || stmtType == recv) && destRP == rank && srcRP != rank)
         --   )
         unlines $
         -- ACHTUNG: TODO: send/recv statements!
         [ ""
         , (srcRP_t ++ " = " ++ procMap_t ++ "." ++ lookupRP_t ++ 
            "(" ++ srcProcNames ++ ", " ++ srcTime_t ++ "1); " ++
           destRP_t ++ " = " ++ procMap_t ++ "." ++ lookupRP_t ++ 
            "(" ++ destProcNames ++ ", " ++ glT_t ++ "1);")

         , (" if(" )
            -- stmtType_t ++ " == " ++ (show spreadStmtNr) ++ " || ")

         , (" (" ++ stmtType_t ++ " == " ++ (show compStmtNr) ++ " && " ++ 
            destRP_t ++ " == " ++ nodeRank_t ++ ") ||")

         , (" ((" ++ stmtType_t ++ " == " ++ (show writeStmtNr) ++ 
            " || " ++ stmtType_t ++ " == " ++ (show sendStmtNr) ++ ")" ++
            " && " ++ srcRP_t ++ " == " ++ nodeRank_t ++ 
            " && " ++ destRP_t ++ " != " ++ nodeRank_t ++ ") ||")

         , (" ((" ++ stmtType_t ++ " == " ++ (show unpackStmtNr) ++ 
            " || " ++ stmtType_t ++ " == " ++ (show recvStmtNr) ++ ")" ++
            " && " ++ destRP_t ++ " == " ++ nodeRank_t ++ 
            " && " ++ srcRP_t ++ " != " ++ nodeRank_t ++ ")")           
         , ")"
         ]
         
      spreadCode_t = spreadCode csDS

--       recvIfConstraint_t =
--           -- if(stmtType != unpack || lookupRP(otherRP1, otherRP2) == nodeRank)
-- 	  " if(" ++ stmtType_t ++ " != " ++ (show unpackStmtNr) ++ " || " ++ 
--           lookupRP_t ++ "(" ++ otherRPnames ++ ") == " ++ nodeRank_t ++ ")"	  
  putStrLn "reading cloogOutput for parsing..."
  input <- readFile inpFile
  putStrLn ("parsing \"" ++ inpFile ++ "\"\n")
  let initState = 
          ParserState { psRPid = srcRP
		      , psIfConstraint = ifConstraint_t
                      , psUseCommunication = True
		      , psCloogStmtString = cloogStmt_t
		      }
      insertSpreadStmtState = 
           ParserState { psRPid = globalTime_t
 		       , psIfConstraint = spreadCode_t
                       , psUseCommunication = True
 		       , psCloogStmtString = cloogStmt_t
 		       }

      firstOutput = 
          trace "first parserPass" $
          parseSource input initState

      finalOutput = 
           trace "second parser pass" $
           parseSource firstOutput insertSpreadStmtState

      output = finalOutput
               --firstOutput
 
  writeFile outputFile output
  putStrLn ("parsing completed.\n")



writeDefinesFile :: CompSpreadDS -> CloogStatements -> IO ()
writeDefinesFile csDS clStmts =
    let loopoFile = csTempDir csDS ++ "/" ++ csProjFile csDS
        suffix = ".defines"
    in
    do origStmtsCode <- getOrigStatementsCode csDS
       let numComps   = genericLength $ clCompStmts clStmts
           numWrites  = genericLength $ clWriteStmts clStmts
           numUnpacks = genericLength $ clUnpackStmts clStmts
           numSends   = genericLength $ clSendStmts clStmts
           numRecvs   = genericLength $ clRecvStmts clStmts
           -- numSpreads = genericLength $ clSpreadStmts clStmts

	   writeStmtsNrs = [(numComps + 1) .. (numComps + numWrites)]
           unpackStmtsNrs = [(numComps + numWrites + 1) .. 
                             (numComps + numWrites + numUnpacks)]
           sendStmtsNrs = [(numComps + numWrites + numUnpacks + 1) ..
                           (numComps + numWrites + numUnpacks + numSends)]
           recvStmtsNrs = [(numComps + numWrites + numUnpacks + numSends + 1) ..
                           (numComps + numWrites + numUnpacks + numSends + numRecvs)]

           -- spreadStmtsNrs = [(numComps + numWrites + numUnpacks + numSends + numRecvs + 1) ..
--                              (numComps + numWrites + numUnpacks + numSends + numRecvs + numSpreads)]

           compStmtsDefines = 
	       genCompStmtsDefines csDS origStmtsCode $ clCompStmts clStmts
           writeStmtsDefines =
               unlines $ 
	       zipWith (genWriteStmtDefines csDS) writeStmtsNrs $ clWriteStmts clStmts
 	   unpackStmtsDefines = 
 	       unlines $
 	       zipWith (genUnpackStmtDefines csDS) unpackStmtsNrs $ clUnpackStmts clStmts
           sendStmtsDefines =
               unlines $ 
	       zipWith (genSendStmtDefines csDS) sendStmtsNrs $ clSendStmts clStmts
 	   recvStmtsDefines = 
 	       unlines $
 	       zipWith (genRecvStmtDefines csDS) recvStmtsNrs $ clRecvStmts clStmts
--  	   spreadStmtsDefines = 
--                trace "writing spreadStmtsDefines..." $
--  	       unlines $
--  	       zipWith (genSpreadStmtDefines csDS) spreadStmtsNrs $ clSpreadStmts clStmts

           defines = 
	       "/* created: " ++ getTime ++ " */\n" ++
	       unlines [ compStmtsDefines
                       , writeStmtsDefines
                       , unpackStmtsDefines
                       , sendStmtsDefines
                       , recvStmtsDefines
                       -- , spreadStmtsDefines
                       ]
       writeFile (loopoFile ++ suffix) defines
       return ()

writeRPMapDefinesFile :: CompSpreadDS -> RPMapStatements -> IO ()
writeRPMapDefinesFile csDS rpMapData =
    let loopoFile = csTempDir csDS ++ "/" ++ csProjFile csDS
        suffix = ".rpDefines"
        rpMapStmts = rpmStmts rpMapData
        rpMapStmtsDefines = 
            unlines $
            zipWith (genRPMapStmtDefines csDS) [1..(genericLength rpMapStmts)] rpMapStmts
        defines = 
	    "/* created: " ++ getTime ++ " */\n" ++ rpMapStmtsDefines
    in
    do writeFile (loopoFile ++ suffix) defines
       return ()
    
printOrigStatementsCode :: TargetLanguage 
                        -> CompSpreadDS 
                        -> IO [(StatementID, String)]
printOrigStatementsCode tl csDS@CompSpreadDS 
			    { csTempDir = tempDir
			    , csProjFile = projFile
                            } =
       do 
       let loopoFile = tempDir ++ "/" ++ projFile
       (symtab, pt, _, _) <- openParseFile loopoFile    
       let stmts = findMainNodes (\ln -> hasstatement ln && statement ln /= (integerToID 0)) pt
           code = map (\stmt -> (statement stmt, printCode tl symtab [stmt])) stmts
       putStrLn "stmts: \n"
       putStrLn $ show code
       putStrLn "end stmts."
       return code

getOrigStatementsCode = printOrigStatementsCode TL_PlainC

{-- MC: 2008-05-05: replaced by print function from ScannerParser...
getOrigStatementsCode :: CompSpreadDS -> IO [(StatementID, String)]
getOrigStatementsCode ccDS@CompSpreadDS 
			  { csTempDir = tempDir
			  , csProjFile = projFile
			  , csTargoutExec = targoutExec
			  } = 
    let
    tmpFile = projFile ++ ".stmts"
    langFile = "c_stmtsOnly.lang"
    targoutFlags = 
	" -temp " ++ tempDir ++
	" -proj " ++ tempDir ++ "/" ++ projFile ++
	" -pp " ++
	" -count " ++
	" -lang " ++ langFile ++
	" -f " ++ tmpFile
    in do
       exitCode <- do 
		   putStrLn ("calling targout for statement text...\n" ++
			     targoutExec ++ targoutFlags)
		   system (targoutExec ++ targoutFlags)
       let (ok, errCode) = checkResult exitCode
       if ok then do putStrLn "targout completed."
             else do error ("targout exited with error code: " ++
			    (show errCode))
       stmtsTextFile <- openFile (tempDir ++ "/" ++ tmpFile) ReadMode
       rawStmtsText <- hGetContents stmtsTextFile
       let stmtsLines = filter (isPrefixOf stmtPrefix) $
			lines rawStmtsText
       return $ map getStmtNumberAndText stmtsLines

--}

stmtPrefix :: String
stmtPrefix = "StatementNr: "

---
-- Reads the origStmtNr and the corresponding code for a statement from a 
-- given line of the temporary file generated by "targout".
getStmtNumberAndText :: String -> (StatementID, String)
getStmtNumberAndText line =
    (integerToID (read stmtNr::Integer), unwords stmtText)
    where
    (stmtNr:stmtText) = words $ drop (length stmtPrefix) line
