{-
    This file is part of the LooPo project, see
    <http://www.infosun.fim.uni-passau.de/cl/loopo/>.

    Copyright (C) 2007 Universitaet Passau, Germany
    Author: Armin Groesslinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-}

-- |
-- This module provides an interface to the C\/C++ parts of LooPo.
--
module CppInterface(module Foreign, module Foreign.C.String,
		    module Foreign.C.Types,
		    CPair, CTriple, CEnum(..), CBool(..),
		    cppiArgs1, cppiArgs2, cppiArgs3, cppiArgs4,
		    cppiConvert1, cppiConvert2, cppiConvert3, cppiConvert4,
		    unsafeCppiConvert1, unsafeCppiConvert2, unsafeCppiConvert3,
		    unsafeCppiConvert4, CConv(..),
                    PolylibValue,
		    setInitialRaySpace)
where

import Data.List ( genericLength )
import Data.Ratio

import Foreign
import Foreign.C.String
import Foreign.C.Types
import System.IO.Unsafe


foreign import ccall setInitialRaySpace_c :: CInt -> IO ()

setInitialRaySpace :: Maybe Int -> IO ()
setInitialRaySpace Nothing = setInitialRaySpace_c 0
setInitialRaySpace (Just rays)
    | rays > 0  =  setInitialRaySpace_c (fromIntegral rays)
    | otherwise =
	error "setInitialRaySpace: can't set ray space to non-positive value"


class CConv a b where
    toC   :: a -> (b -> IO c) -> IO c
    fromC :: b -> IO a

instance CConv () () where
    toC () action = action ()
    fromC () = return ()

--
-- Type Bool is represented as `int' where True is 1 on False is 0
-- (as usual).
--
newtype CBool = CBool CInt
instance CConv Bool CBool where
    toC   True  action = action (CBool 1)
    toC   False action = action (CBool 0)
    fromC (CBool b) = return (b /= 0)

--
-- Integral types are represented as `int' or 'long long'.
--
instance Integral a => CConv a CInt where
    toC   x f = f (fromIntegral x)
    fromC x   = return (fromIntegral x)
instance Integral a => CConv a CLLong where
    toC   x f = f (fromIntegral x)
    fromC x   = return (fromIntegral x)

--
-- Pointers can be represented "as is" on the Haskell side.
--
instance CConv (Ptr a) (Ptr a) where
    toC   x f = f x
    fromC x   = return x

newtype CEnum = CEnum CInt
type CPair a b = Ptr (a,b)
type CTriple a b c = Ptr (a,b,c)

--
-- Enum's are represented as `int' (using their enumeration value).
--
instance Enum a => CConv a CEnum where
    toC x action    = action (CEnum $ fromIntegral $ fromEnum x)
    fromC (CEnum x) = return (toEnum $ fromIntegral x)


-- type CString = Ptr CChar
instance CConv String CString where
    toC str action = withCString str action
    fromC ptr =
	do str <- peekCString ptr
	   free ptr
	   return str

--
-- When type `a' is represented as a pointer on the C side,
-- type (Maybe a) is also represented as a pointer (of the same type);
-- Nothing is represented by NULL.
--
instance CConv a (Ptr b) => CConv (Maybe a) (Ptr b) where
    toC Nothing  f = f nullPtr
    toC (Just x) f = toC x f

    fromC y 
	| y == nullPtr  =  return Nothing
	| otherwise     =  fromC y >>= \x -> return (Just x)


--
-- When type `a' is represented as a pointer on the C side, a list
-- of type [a] is represented as an array of pointers. The end of the
-- list is marked by a NULL pointer in the array.
--
instance (CConv a (Ptr b)) => CConv [a] (Ptr (Ptr b)) where
    toC xs action = f xs []
	where
	f []     ys = withArray (reverse (nullPtr:ys)) action
	f (x:xs) ys = toC x (\y -> f xs (y:ys))
    fromC ptr = do res <- getList 0
		   free ptr
		   return res
	where
	getList :: Int -> IO [a]
	getList n = do yp <- peekElemOff ptr n
		       if yp == nullPtr
			  then return []
			  else do ys <- getList (n+1)
				  y <- fromC yp
				  return (y:ys)

--
-- A list of integers is represented as an integer array on the
-- C side. The first element (with index 0) of the array is the
-- length of the list (not including the length element).
-- The first element of the integer list is therefore found at
-- offset 1.
--
instance (Integral a, Integral b, Storable b, CConv a b) => CConv [a] (Ptr b) where
    toC xs action = f xs []
	where
	f []     ys = withArray (reverse ys) action
	f (x:xs) ys = toC x (\y -> f xs (y:ys))
    fromC p =
	do len <- peek p
	   xs  <- peekArray (fromIntegral len + 1) p
	   free p
	   return $ map fromIntegral (tail xs)

--
-- When types `a' and `b' are represented as pointers on the C side,
-- the tuple type (a,b) is represented as a pointer to a two-element
-- pointer array where the {1st,2nd} element points the the {1st,2nd}
-- component of the pair.
--
instance (CConv a (Ptr a'), CConv b (Ptr b')) =>
    CConv (a,b) (CPair (Ptr a') (Ptr b')) where
    toC _ _ = undefined
    fromC p =
	do [p1', p2'] <- peekArray 2 (castPtr p :: Ptr (Ptr ()))
	   let p1 = castPtr p1' :: Ptr a'
	       p2 = castPtr p2' :: Ptr b'
	   obj1 <- fromC p1
	   obj2 <- fromC p2
	   free p
	   return (obj1, obj2)

--
-- Like CPair, but for triples.
--
instance (CConv a (Ptr a'), CConv b (Ptr b'), CConv c (Ptr c')) =>
    CConv (a,b,c) (CTriple (Ptr a') (Ptr b') (Ptr c')) where
    toC _ _ = undefined
    fromC p =
	do [p1', p2', p3'] <- peekArray 3 (castPtr p :: Ptr (Ptr ()))
	   let p1 = castPtr p1' :: Ptr a'
	       p2 = castPtr p2' :: Ptr b'
	       p3 = castPtr p3' :: Ptr c'
	   obj1 <- fromC p1
	   obj2 <- fromC p2
	   obj3 <- fromC p3
	   free p
	   return (obj1, obj2, obj3)


cppiArgs1 :: (CConv a a') => (a' -> IO b) -> (a -> IO b)
cppiArgs1 f x = toC x (\x' -> f x')

cppiArgs2 :: (CConv a a', CConv b b') =>
	 (a' -> b' -> IO c) -> (a -> b -> IO c)
cppiArgs2 f x y = toC x (\x' -> cppiArgs1 (f x') y)
 
cppiArgs3 :: (CConv a a', CConv b b', CConv c c') =>
	    (a' -> b' -> c' -> IO d) -> (a -> b -> c -> IO d)
cppiArgs3 f x y z = toC x (\x' -> cppiArgs2 (f x') y z)

cppiArgs4 :: (CConv a a', CConv b b', CConv c c', CConv d d') =>
	    (a' -> b' -> c' -> d' -> IO e) -> (a -> b -> c -> d -> IO e)
cppiArgs4 f w x y z = toC w (\w' -> cppiArgs3 (f w') x y z)
    
cppiConvert1 :: (CConv a a', CConv b b') => (a' -> IO b') -> (a -> IO b)
cppiConvert1 f x = cppiArgs1 f x >>= fromC

cppiConvert2 :: (CConv a a', CConv b b', CConv c c') =>
	    (a' -> b' -> IO c') -> (a -> b -> IO c)
cppiConvert2 f x y = cppiArgs2 f x y >>= fromC

cppiConvert3 :: (CConv a a', CConv b b', CConv c c', CConv d d') =>
	    (a' -> b' -> c' -> IO d') -> (a -> b -> c -> IO d)
cppiConvert3 f x y z = cppiArgs3 f x y z >>= fromC

cppiConvert4 :: (CConv a a', CConv b b', CConv c c', CConv d d', CConv e e') =>
	    (a' -> b' -> c' -> d' ->IO e') -> (a -> b -> c -> d -> IO e)
cppiConvert4 f w x y z = cppiArgs4 f w x y z >>= fromC

unsafeCppiConvert1 :: (CConv a a', CConv b b') => (a' -> IO b') -> (a -> b)
unsafeCppiConvert1 f x = unsafePerformIO ((cppiConvert1 f) x)

unsafeCppiConvert2 :: (CConv a a', CConv b b', CConv c c') =>
		  (a' -> b' -> IO c') -> (a -> b -> c)
unsafeCppiConvert2 f x y = unsafePerformIO ((cppiConvert2 f) x y)

unsafeCppiConvert3 :: (CConv a a', CConv b b', CConv c c', CConv d d') =>
		  (a' -> b' -> c' -> IO d') -> (a -> b -> c -> d)
unsafeCppiConvert3 f x y z = unsafePerformIO (cppiConvert3 f x y z)

unsafeCppiConvert4 :: (CConv a a', CConv b b', CConv c c', CConv d d', CConv e e')
		  => (a' -> b' -> c' -> d' -> IO e') -> (a -> b -> c -> d -> e)
unsafeCppiConvert4 f w x y z = unsafePerformIO (cppiConvert4 f w x y z)

foreign import ccall newFraction_c :: Ptr CLLong -> IO (Ptr Rational)
foreign import ccall decompFraction_c :: Ptr Rational -> IO (Ptr CLLong)
foreign import ccall deleteFraction_c :: Ptr Rational -> IO ()


--
-- `Rational' is represented as a pointer to a two-element `long long' array
-- with the numerator as first and the denominator as the second entry.
--
instance (Real a, Fractional a) => CConv a (Ptr Rational) where
    toC rat action =
	do frac <- withArray [num, denom] newFraction_c
	   res <- action frac
	   deleteFraction_c frac
	   return res
	where
	num   = fromIntegral $ numerator   $ toRational rat
	denom = fromIntegral $ denominator $ toRational rat
    fromC frac =
	do p <- decompFraction_c frac
	   [num, denom] <- peekArray 2 p
	   free p
	   return $ fromRational ((fromIntegral num) % (fromIntegral denom))

data PolylibValue

foreign import ccall polylibValueFromLimbs_c :: Ptr CUInt -> IO (Ptr PolylibValue)
foreign import ccall polylibValueToLimbsAndFree_c :: Ptr PolylibValue -> IO (Ptr CUInt)
foreign import ccall polylibValueFree_c :: Ptr PolylibValue -> IO ()

limbModulus :: Integer
limbModulus = 2 ^ fromIntegral (8 * sizeOf (0::CUInt) - 2)

instance Integral a => CConv a (Ptr PolylibValue) where
    toC i action =
        withArray limbs $ \limbPtr ->
           do pvPtr <- polylibValueFromLimbs_c limbPtr
              res <- action pvPtr
              polylibValueFree_c pvPtr
              return res
        where
        limbs = lenInd : li
            where
            li     = l (fromIntegral $ abs i)
            lenInd = 2*genericLength li + (if i < 0 then 1 else 0)
        l :: Integer -> [CUInt]
        l 0 = []
        l x = fromIntegral r : l q
            where
            (q,r) = x `quotRem` limbModulus

    fromC pvPtr =
        do limbPtr <- polylibValueToLimbsAndFree_c pvPtr
           lenInd <- peek limbPtr
           let len = fromIntegral (lenInd `quot` 2)
               neg = lenInd `rem` 2 /= 0
           limbs <- peekArray len (advancePtr limbPtr 1)
           free limbPtr
           let absVal = unlimb limbs
           return $ fromIntegral $ if neg then negate absVal else absVal
        where
        unlimb limbs = foldr (\l acc -> acc*limbModulus + fromIntegral l)
                             (0::Integer)
                             limbs
