#include <iostream>


#include "flmaths.h"
#include "CppSide.h"
#include "polylib.h"
extern "C" {
#include <assert.h>
}

/*
 * Interface to HsLooPo.
 */

/*
 * Arbitrary-sized integers are passed to/from the C side as an
 * array of limbs:  unsigned int *limbs;
 * Format:
 *    limbs[0]: Indicates the number of limbs and sign of the integer.
 *       The number of limbs ("len") is limbs[0]/2 and the integer is
 *       negative iff  (limbs[0]&1) != 0.
 *    limbs[1]..libs[len]: Store the absolute value of the integer.
 *       absValue = limbs[1] + limbs[2]*m + ... + limbs[len]*m^(len-1)
 *       where  m  is the modulus used to split the integer into limbs,
 *       limbModulus  in the implementation.
 */

/* must fit in a  signed int  (due to its use with value_set_si) */
static const unsigned int limbModulus = 1U << (8*sizeof(unsigned int) - 2);

Value *polylibValueFromLimbs_c(const unsigned int *limbs) {
    gmp_protect();

    unsigned len = *limbs / 2;
    unsigned neg = *limbs & 1;
    Value *v = (Value *)malloc(sizeof(Value));
    int i;

    value_init(*v);

    if (len >= 1) {
	Value modulus, l;
	value_init(modulus);
	value_init(l);
	value_set_si(modulus, limbModulus);
	value_set_si(*v, limbs[len]);
	for (i=len-1; i>=1; i--) {
	    value_set_si(l, limbs[i]);
	    value_multiply(*v, *v, modulus);
	    value_addto(*v, *v, l);
	}
	if (neg)
	    value_oppose(*v, *v);
	value_clear(modulus);
	value_clear(l);
    } else
	value_set_si(*v, 0);
    
    gmp_unprotect();
    return v;
}

Value *copy_value(Value &v) {
    Value *copy = (Value *)malloc(sizeof(Value));
    value_init(*copy);
    value_assign(*copy, v);
    return copy;
}

unsigned int *polylibValueToLimbsAndFree_c(Value *v) {
    Value modulus, vv, r;
    unsigned len;
    unsigned neg;
    int i;
    unsigned int *limbs;

    gmp_protect();

    value_init(modulus);
    value_init(vv);
    value_init(r);

    value_set_si(modulus, limbModulus);

    if (value_sign(*v) < 0) {
	neg = 1;
	value_oppose(*v, *v);
    } else
	neg = 0;

    value_assign(vv, *v);
    len = 0;
    while (value_sign(vv) > 0) {
	value_division(vv, vv, modulus);
	len++;
    }
    limbs = (unsigned int *)malloc(sizeof(unsigned int) * (len+1));
    limbs[0] = 2*len + neg;
    for (i=1; i<=len; i++) {
	value_modulus(r, *v, modulus);
	limbs[i] = VALUE_TO_INT(r);
	value_division(*v, *v, modulus);
    }
    value_clear(modulus);
    value_clear(vv);
    value_clear(r);
    value_clear(*v);
    free(v);

    gmp_unprotect();

    return limbs;
}

void polylibValueFree_c(Value *v) {
    gmp_protect();
    value_clear(*v);
    free(v);
    gmp_unprotect();
}

PolyMatrix *polylibMatrixNew_c(int rows, int columns, Value **entries) {
    PolyMatrix *pm;
    gmp_protect();
    pm = Matrix_Alloc(rows, columns);
    for (int r=0; r<rows; r++) {
	for (int c=0; c<columns; c++)
	    value_assign(pm->p[r][c], *entries[r*columns + c]);
    }
    gmp_unprotect();
    return pm;
}

int polylibMatrixRows_c(PolyMatrix *pm) {
    return pm->NbRows;
}

int polylibMatrixColumns_c(PolyMatrix *pm) {
    return pm->NbColumns;
}

Value *polylibMatrixEntry_c(int r, int c, PolyMatrix *pm) {
    gmp_protect();
    Value *v = copy_value(pm->p[r][c]);
    gmp_unprotect();
    return v;
}

void polylibMatrixFree_c(PolyMatrix *pm) {
    gmp_protect();
    Matrix_Free(pm);
    gmp_unprotect();
}

PolyMatrix *polylibMatrixCopy_c(PolyMatrix *pm) {
    PolyMatrix *copy;
    gmp_protect();
    copy = Matrix_Copy(pm);
    gmp_unprotect();
    return copy;
}

Polyhedron *polylibPolyhedronNew_c(PolyMatrix *pm) {
    gmp_protect();
    Polyhedron *p = Constraints2Polyhedron(pm, 0);
    gmp_unprotect();
    return p;
}

void polylibPolyhedronFree_c(Polyhedron *p) {
    gmp_protect();
    Polyhedron_Free(p);
    gmp_unprotect();
}

PolyMatrix *polylibPolyhedronMatrix_c(Polyhedron *p) {
    gmp_protect();
    PolyMatrix *pm = Polyhedron2Constraints(p);
    gmp_unprotect();
    return pm;
}

Polyhedron *polylibPolyhedronNext_c(Polyhedron *p) {
    return p->next;
}

void polylibPolyhedronSetNext_c(Polyhedron *p, Polyhedron *q) {
    p->next = q;
}


/*
 * Z-polyhedra functions
 */

static ZPolyhedron *makeZDomain(Polyhedron **ph, PolyMatrix **latt) {
    ZPolyhedron *domain, *last;
    
    domain = NULL;
    for ( ; *ph != NULL; ph++,latt++) {
	ZPolyhedron *zp = ZPolyhedron_Alloc(*latt, *ph);
	if (domain == NULL) {
	    domain = zp;
	} else {
	    last = domain;
	    domain = ZDomainUnion(last, zp);
	    ZDomain_Free(last);
	    ZDomain_Free(zp);
	}
    }
    return domain;
}

static Pair *zdomain2phlatt(ZPolyhedron *zp) {
    Pair *res;
    int i, len;
    ZPolyhedron *z;

    /* Count the number of polyhedra used to represent the
     * ZDomain. Note: each ZPolyhedron can contain a
     * union of polyhedra as preimage.
     */
    len = 0;
    for (z=zp; z!=NULL; z=z->next) {
	for (Polyhedron *p=z->P; p!=NULL; p=p->next)
	    len++;
    }
    res = (Pair *)malloc(sizeof(Pair) * (len+1));
    i = 0;
    for (z=zp; z!=NULL; z=z->next) {
	for (Polyhedron *p=z->P; p!=NULL; p=p->next) {
	    PolyMatrix *latt = Matrix_Copy(z->Lat);
	    Polyhedron *ph = Polyhedron_Copy(p);
	    res[i] = mkPair(ph, latt);
	    i++;
	}
    }

    res[len] = NULL;

    return res;
}

static Pair *zpOp2l_helper(Polyhedron **ph1, PolyMatrix **latt1, Polyhedron **ph2, PolyMatrix **latt2, ZPolyhedron *(*zdomainFunc)(ZPolyhedron *, ZPolyhedron *))
{
    ZPolyhedron *zp1, *zp2, *zp, *z;
    Pair *res;
    int i, len;

    if (*ph1 == NULL && *ph2 == NULL) {
	res = new Pair[1];
	res[0] = NULL;
	return res;
    }

    gmp_protect();

    zp1 = makeZDomain(ph1, latt1);
    zp2 = makeZDomain(ph2, latt2);
    zp = zdomainFunc(zp1, zp2);
    if (zp1 != NULL)
	ZDomain_Free(zp1);
    if (zp2 != NULL)
	ZDomain_Free(zp2);

    if (zp != NULL) {
	z = ZDomainSimplify(zp);
	ZDomain_Free(zp);
	zp = z;
    }

    res = zdomain2phlatt(zp);
    if (zp != NULL)
	ZDomain_Free(zp);

    gmp_unprotect();

    return res;
}

static ZPolyhedron *ZDomainDifference_safe(ZPolyhedron *zp1, ZPolyhedron *zp2) {
    if (zp1 == NULL)
	return NULL;
    if (zp2 == NULL)
	return ZDomain_Copy(zp1);

    return ZDomainDifference(zp1, zp2);
}

Pair *zpDifference_c(Polyhedron **ph1, PolyMatrix **latt1, Polyhedron **ph2, PolyMatrix **latt2) {
    return zpOp2l_helper(ph1, latt1, ph2, latt2, ZDomainDifference);
}

Pair *zpIntersection_c(Polyhedron **ph1, PolyMatrix **latt1, Polyhedron **ph2, PolyMatrix **latt2) {
    return zpOp2l_helper(ph1, latt1, ph2, latt2, ZDomainIntersection);
}

Pair *zpUnion_c(Polyhedron **ph1, PolyMatrix **latt1, Polyhedron **ph2, PolyMatrix **latt2) {
    return zpOp2l_helper(ph1, latt1, ph2, latt2, ZDomainUnion);
}

Pair *zpUnionDifference_c(Polyhedron **phIn, PolyMatrix **lattIn, Polyhedron **phOut, PolyMatrix **lattOut)
{
    return zpOp2l_helper(phIn, lattIn, phOut, lattOut, ZDomainDifference_safe);
}


int zpIncludes_c(Polyhedron *p1, PolyMatrix *latt1, Polyhedron *p2, PolyMatrix *latt2) {
    gmp_protect();

    ZPolyhedron *zp1 = ZPolyhedron_Alloc(latt1, p1);  
    ZPolyhedron *zp2 = ZPolyhedron_Alloc(latt2, p2);

    int incl = ZPolyhedronIncludes(zp2, zp1);
    ZDomain_Free(zp1);
    ZDomain_Free(zp2);

    gmp_unprotect();

    return incl;
}





/*
 * Old stuff
 */

static long long value_to_ll(Value &vv) {
    long long l = 0;
    Value v, v2, mask, div;
    int s = 0, neg = 0;

    value_init(v);
    if (value_neg_p(vv)) {
	value_oppose(v, vv);
	neg = 1;
    } else
	value_assign(v, vv);

    value_init(v2);
    value_init(mask);
    value_init(div);
    value_set_si(mask, 0xFFFF);
    value_set_si(div, 0x10000);
    while (value_notzero_p(v)) {
	value_andto(v2, v, mask);
	l |= ((long long)VALUE_TO_INT(v2)) << s;
	value_division(v, v, div);
	s = s + 16;
//	printf("s = %d\n", s);
    }

    return neg ? -l : l;
}

// out of memory in polylib
static void oomInPL() {
    std::cerr << "out of memory in Polylib" << std::endl;
    exit(1);
}

static int initialRays = 10000;

void setInitialRaySpace_c(int r) {
    if (r >= 0) 
	initialRays = r;
}

PolyMatrix *copyM(PolyMatrix *pm) {
    PolyMatrix *copy;
    copy = Matrix_Alloc(pm->NbRows, pm->NbColumns);
    if (copy == NULL)
	oomInPL();
    for (unsigned r=0; r<pm->NbRows; r++) {
	for (unsigned c=0; c<pm->NbColumns; c++)
	    value_assign(copy->p[r][c], pm->p[r][c]);
    }
    return copy;
}

Polyhedron *Constraints2Polyhedron_x(PolyMatrix *pm) {
    Polyhedron *ph;
    PolyMatrix *copy;
    int rays = (initialRays > 0) ? initialRays : pm->NbColumns;
    Pol_status = 0;
    copy = copyM(pm);
    ph = Constraints2Polyhedron(copy, rays);
    Matrix_Free(copy);
    while (Pol_status != 0) {
	rays *= 4;
	std::cerr << "increasing table space to " << rays
	          << " and trying again..." << std::endl;
	Pol_status = 0;
	copy = copyM(pm);
	ph = Constraints2Polyhedron(copy, rays);
	Matrix_Free(copy);
    }

    return ph;
}

Polyhedron *DomainUnion_x(Polyhedron *p1, Polyhedron *p2) {
    Polyhedron *p1copy, *p2copy, *result;

    int rays = (initialRays > 0) ? initialRays : 2*(p1->NbRays+p1->NbRays);

    do {
	Pol_status = 0;
	p1copy = Domain_Copy(p1);
	p2copy = Domain_Copy(p2);
	result = DomainUnion(p1copy, p2copy, rays);
	Domain_Free(p1copy);
	Domain_Free(p2copy);
	if (Pol_status != 0) {
	    rays *= 4;
	    std::cerr << "increasing table space to " << rays
		      << " and trying again..." << std::endl;
	}
    } while (Pol_status != 0);

    return result;
}

// put a FMatrix into a PolyMatrix starting in row `top', column `left'
void putFMatrixInPolyMatrix(PolyMatrix *pm, FMatrix *fm, int top, int left) {
    Value num, den, common;

    const int n_rows    = fm->get_rows();
    const int n_columns = fm->get_cols();
    
    value_init(num);
    value_init(den);
    value_init(common);
    for (int r=0; r<n_rows; r++) {
	// The current implementation (Oct 2001) guarantees that <lcm>s are
	// non-negative, and we rely on this so that the sign of
	// the numerator is corrected automatically for entries with
	// a negative denominator.
	denom_type commonDenom = (*fm)[r].lcm_of_entries();
	assert(commonDenom >= 0);

	value_set_si(common, commonDenom);
	for (int c=0; c<n_columns; c++) {
	    Fraction entry = fm->get_entry(r, c);
	    value_set_si(num, entry.get_num());
	    value_set_si(den, entry.get_denom());
	    value_division(den, common, den);
	    value_multiply(pm->p[top+r][left+c], num, den);
	}
    }
}

// put "1 >= 0" in the last row of the PolyMatrix `pm'
void positivityConstraint(PolyMatrix *pm) {
    const int n_columns = pm->NbColumns;
    const int n_rows    = pm->NbRows;

    for (int c = 1; c <= n_columns-2; c++)
	value_set_si(pm->p[n_rows-1][c], 0);
    value_set_si(pm->p[n_rows-1][0], 1);
    value_set_si(pm->p[n_rows-1][n_columns-1], 1);
}

const int Equation = 0, Inequation = 1;

void putFInEqSysInPolyMatrix(PolyMatrix *pm, int type,
			     FMatrix *fm, int top, int left) {
    putFMatrixInPolyMatrix(pm, fm, top, left+1);
    for (int r=0; r<fm->get_rows(); r++)
	value_set_si(pm->p[top+r][left], type);
}

Polyhedron *fesfis2ph(FEq_system *eqs, FIneq_system *ineqs) {
    assert(eqs->get_cols() == ineqs->get_cols());

    const int n_rows_eqs   = eqs->get_rows();
    const int n_rows_ineqs = ineqs->get_rows();
    const int n_columns    = eqs->get_cols();
    PolyMatrix *pm;
    Polyhedron *ph;

    pm = Matrix_Alloc(n_rows_eqs+n_rows_ineqs+1, n_columns+1);
    if (pm == NULL)
	oomInPL();

    putFInEqSysInPolyMatrix(pm, Equation,   eqs,   0,          0);
    putFInEqSysInPolyMatrix(pm, Inequation, ineqs, n_rows_eqs, 0);
    positivityConstraint(pm);
/*
    for (int r=0; r<pm->NbRows; r++) {
	for (int c=0; c<pm->NbColumns; c++)
	    std::cout << pm->p[r][c] << "  ";
	std::cout << std::endl;
    }
*/

    ph = Constraints2Polyhedron_x(pm);
    Matrix_Free(pm);
    if (ph == NULL)
	oomInPL();

    return ph;
}


Polyhedron *fis2ph(FIneq_system *fis) {
    PolyMatrix *pm;
    Polyhedron *ph;

    pm = Matrix_Alloc(fis->get_rows()+1, fis->get_cols()+1);
    if (pm == NULL)
	oomInPL();

    putFInEqSysInPolyMatrix(pm, Inequation, fis, 0, 0);
    positivityConstraint(pm);
/*
    int r,c;
    for (r=0; r<pm->NbRows; r++) {
	for (c=0; c<pm->NbColumns; c++)
	    std::cout << value_to_ll(pm->p[r][c]) << "  ";
	std::cout << std::endl;
    }
    std::cout << std::endl;
*/
    ph = Constraints2Polyhedron_x(pm);
    Matrix_Free(pm);
    if (ph == NULL)
	oomInPL();

    return ph;
}


FMatrix *pm2fm(PolyMatrix *pm) {
    const int n_rows    = pm->NbRows;
    const int n_columns = pm->NbColumns;

    FMatrix *fm = new FMatrix(n_rows, n_columns);
    if (fm == NULL)
	oomInLoopo();

    for (int r=0; r<n_rows; r++) {
	for (int c=0; c<n_columns; c++) {
	    Fraction entry(VALUE_TO_LONG(pm->p[r][c]), 1);
	    fm->set_entry(r, c, entry);
	}
    }

    return fm;
}

PolyMatrix *fm2pm(FMatrix *fm) {
    Matrix m(*fm);
    int denom = m.get_denom();
    int r, c;
    PolyMatrix *pm = Matrix_Alloc(m.get_rows(), m.get_cols());

    assert(denom == 1);

    if (pm == NULL) {
	std::cerr << "out of memory in polylib" << std::endl;
	exit(1);
    }
    for (r=0; r<m.get_rows(); r++) {
	for (c=0; c<m.get_cols(); c++) {
	    value_set_si(pm->p[r][c], m.entry(r, c)); // numerator
	}
    }

    return pm;
}

// convert vertices (given as *rows* in `m') to a Polyhedron
Polyhedron * vertices2polyhedron(Matrix m) {
    int denom = m.get_denom();
    int r, c;
    PolyMatrix *pm = Matrix_Alloc(m.get_rows(), m.get_cols()+2);
    if (pm == NULL) {
	std::cerr << "out of memory in polylib" << std::endl;
	exit(1);
    }
    for (r=0; r<m.get_rows(); r++) {
	value_set_si(pm->p[r][0], 1);  // vertex
	for (c=0; c<m.get_cols(); c++) {
	    value_set_si(pm->p[r][c+1], m.entry(r, c)); // numerator
	}
	value_set_si(pm->p[r][c+1], denom);  //denominator
    }
    Polyhedron *p = Rays2Polyhedron(pm, 10000);
    if (p == NULL) {
	std::cerr << "out of memory in polylib" << std::endl;
	exit(1);
    }
    Matrix_Free(pm);
    return p;
}

// determine whether the row in `coeffs' (which is in polylib's
// constraint format and has `n_vars' variables) represents
// a trivial equation (0 == 0) or inequation (c >= 0 with
// c constant and non-negative)
int isTrivial(Value *coeffs, int n_vars) {
    for (int v=1; v<=n_vars; v++) {
	if (value_notzero_p(coeffs[v]))
	    return 0;
    }
    return ((value_zero_p(coeffs[0]) &&
	     value_zero_p(coeffs[n_vars+1]))   // 0 == 0 
	    || (value_one_p(coeffs[0]) &&
		value_posz_p(coeffs[n_vars+1]))); // c >= 0
}


// convert a Polyhedron to a FIneq_system
FIneq_system polyhedron2ineqs(Polyhedron *p, int n_pars=0) {
    int c;
    unsigned r, rr;
    PolyMatrix *pm;

    pm = Polyhedron2Constraints(p);
    if (pm == NULL)
	oomInPL();

    const int n_vars = pm->NbColumns - 2;
    
    // count the number of non-trivial inequations
    int ineqs = 0;
    for (r=0; r<pm->NbRows; r++) {
	if (!isTrivial(pm->p[r], n_vars))
	    ineqs += value_zero_p(pm->p[r][0]) ? 2 : 1;
    }

/*    if (ineqs == 0) {
	cerr << "working around 0-rows-FMatrix bug..." << endl;
	ineqs = 1;
    }
*/    
    FMatrix m(ineqs, n_vars+1);
    rr = 0; // the current row in `m'
    for (r=0; r<pm->NbRows; r++) {
	if (!isTrivial(pm->p[r], n_vars)) {
	    // convert an inequation to one row, an equation to two rows
	    for (c=0; c<=n_vars; c++) {
//		std::cerr << pm->p[r][c+1] << std::endl;
		m.set_entry(rr, c, Fraction(value_to_ll(pm->p[r][c+1]), 1));
		if (value_zero_p(pm->p[r][0]))
		    m.set_entry(rr+1, c, Fraction(-value_to_ll(pm->p[r][c+1]), 1));
	    }
	    rr += value_zero_p(pm->p[r][0]) ? 2 : 1;
	}
    }
    Matrix_Free(pm);
    m.shrink_values();
    return FIneq_system(n_vars - n_pars, n_pars, &m);
}

//
//
// POLYLIB Interface
//
//

// calculate the dual representation of the vertices (represented as 
// *row* vectors)
FIneq_system *verticesAsRowVectorsToIneqs_c(FMatrix *fm) {
    Polyhedron *p;
    FIneq_system *fis;
    gmp_protect();
    p   = vertices2polyhedron(*fm);
    fis = new FIneq_system(polyhedron2ineqs(p));
    Polyhedron_Free(p);
    gmp_unprotect();
    return fis;
}


FIneq_system *convexUnionAll_c(FIneq_system **systems, int count) {
    Polyhedron *u;

    if (count == 0)
	return NULL;

    gmp_protect();

    u = fis2ph(systems[0]);

    for (int i=1; i<count; i++) {
	Polyhedron *p = fis2ph(systems[i]);
	Polyhedron *u2 = u;
	u = DomainUnion(p, u2, 10000);
	Polyhedron_Free(u2);
	Polyhedron_Free(p);
    }

    Polyhedron *c = DomainConvex(u, 10000);
    Polyhedron_Free(u);

    FIneq_system *res = new FIneq_system(polyhedron2ineqs(c,systems[0]->get_params()));
    Polyhedron_Free(c);

    gmp_unprotect();

    return res;
}

FMatrix* eqsIneqs2rays_c(FEq_system *fes, FIneq_system *fis) {
    Polyhedron *p;
    PolyMatrix *rays;
    
    gmp_protect();

    p = fesfis2ph(fes, fis);
    rays = Polyhedron2Rays(p);
    Polyhedron_Free(p);
    if (rays == NULL) 
	oomInPL();

/*
    for (int r=0; r<rays->NbRows; r++) {
	for (int c=0; c<rays->NbColumns; c++)
	    cout << rays->p[r][c] << "  ";
	cout << endl;
    }
*/
    FMatrix *frays = pm2fm(rays);
    Matrix_Free(rays);

    gmp_unprotect();

    return frays;
}


int polyhedronIncludes_c(FIneq_system *sys1, FIneq_system *sys2) {
    Polyhedron *p1, *p2;

    gmp_protect();

    p1 = fis2ph(sys1);
    p2 = fis2ph(sys2);

    int res = PolyhedronIncludes(p1, p2);
    Polyhedron_Free(p1);
    Polyhedron_Free(p2);

    gmp_unprotect();

    return res;
}


FIneq_system **union_c(FIneq_system **input) {
    Polyhedron *result, *ctx, *p, *r;
    FIneq_system **inp;
    FIneq_system **resultList;
    int n, i;

    if (input[0] == NULL) {
	resultList = (FIneq_system **) malloc(sizeof(FIneq_system *));
	if (resultList == NULL)
	    oomInLoopo();
	resultList[0] = NULL;
	return resultList;
    }

    gmp_protect();

    result = fis2ph(input[0]);
    ctx = Universe_Polyhedron(input[0]->get_vars() + input[0]->get_params());
    for (inp=&input[1]; *inp!=NULL; inp++) {
	p = fis2ph(*inp);
	r = result;
	result = DomainUnion_x(result, p);
	Domain_Free(r);
	r = result;
	result = DomainSimplify(result, ctx, 10000);
	Domain_Free(r);
	Polyhedron_Free(p);
    }

    p = result;
    n = 0;
    while (p != NULL) {
	p = p->next;
	n++;
    }

    resultList = (FIneq_system **)malloc(sizeof(FIneq_system *) * (n+1));
    if (resultList == NULL)
	oomInLoopo();
    for (i=0,p=result; i<n; i++,p=p->next) {
	resultList[i] = new FIneq_system(polyhedron2ineqs(p,input[0]->get_params()));
	if (resultList[i] == NULL)
	    oomInLoopo();
    }
    resultList[i] = NULL;

    Domain_Free(ctx);
    Domain_Free(result);

    gmp_unprotect();

    return resultList;
}


/* MC: disjUnionAll 
 * Computes the disjoint union of a given NULL terminated array of
 * inequality systems.
 */
FIneq_system** disjUnionAll_c(FIneq_system **input) {
  
  FIneq_system** res;
  Polyhedron *u;
  
  if (input[0] == NULL) {
    res = (FIneq_system**) malloc(sizeof(FIneq_system*));
    if (res == NULL)
      oomInLoopo();
    res[0] = NULL;
    return res;
  }

  gmp_protect();
    
  u= fis2ph(input[0]);
  
  for (int i=1; input[i] != NULL; i++) {
    Polyhedron *p = fis2ph(input[i]);
    Polyhedron *u2 = u;
    u = DomainUnion_x(p, u2);
    Polyhedron_Free(u2);
    Polyhedron_Free(p);
  }

  const int NO_INTERSECTS = 0;
  Polyhedron *c = Disjoint_Domain(u, NO_INTERSECTS, 10000);
  Polyhedron_Free(u);

  Polyhedron* p = c;
  int length = 0;
  while(p != NULL) {
    length++;
    p = p->next;
  }
      
  p = c;    

  // array of size (length+1), because it is terminated by NULL
  res = (FIneq_system**)malloc(sizeof(FIneq_system*) * (length+1)); 
  if (res == NULL)
    oomInLoopo();

  int i;
  for(i=0; i < length; i++) {
    assert(p != NULL);
    res[i] = new FIneq_system(polyhedron2ineqs(p,input[0]->get_params()));
    if (res[i] == NULL)
      oomInLoopo();

    //    Polyhedron_Print(stdout,"%3d",p);
    //    Matrix_Print(stdout,"%3d",Polyhedron2Constraints(p));
    p = p->next;
  }
  res[i] = NULL;

  Polyhedron_Free(c);

  gmp_unprotect();

  return res;
}



FIneq_system **disjointUnionDifference_c(FIneq_system *context, FIneq_system **fisIn, FIneq_system **fisOut)
{
    Polyhedron *set = NULL, *ctx, *zp, *newSet;
    FIneq_system **res;
    int len, i;

    gmp_protect();

    ctx = fis2ph(context);

    for (i=0; fisIn[i]!=NULL; i++) {
	Polyhedron *p  = fis2ph(fisIn[i]);

	if (set == NULL)
	    set = p;
	else {
	    newSet = DomainUnion(set, p, 10000);
	    Domain_Free(set);
	    set = newSet;
	}
    }

    if (set != NULL) {
	for (int i=0; fisOut[i]!=NULL; i++) {
	    Polyhedron *p  = fis2ph(fisOut[i]);
	    
	    newSet = DomainDifference(set, p, 10000);
	    Domain_Free(set);
	    set = newSet;
	}
    }

    newSet = DomainSimplify(set, ctx, 10000);
    Domain_Free(set);
    set = newSet;

    newSet = Disjoint_Domain(set, 0, 10000);
    Domain_Free(set);
    set = newSet;

    len = 0;
    for (zp=set; zp!=NULL; zp=zp->next) {
	len++;
    }
    res = (FIneq_system**)malloc(sizeof(FIneq_system*) * (len+1));
    for (i=0,zp=set; zp!=NULL; i++,zp=zp->next) {
	FIneq_system *f = new FIneq_system(polyhedron2ineqs(zp,context->get_params()));
	res[i] = f;
    }
    res[len] = NULL;
    Domain_Free(set);
    Domain_Free(ctx);

    gmp_unprotect();

    return res;
}


FIneq_system *simplifyIneqs_c(FIneq_system *context, FIneq_system *sys)
{
    Polyhedron *p, *s, *ctx;
    FIneq_system *res;

    gmp_protect();

    p = fis2ph(sys);
/*
    int r,c;
    for (r=0; r<p->NbConstraints; r++) {
	for (c=0; c<=p->Dimension+1; c++) {
	    std::cout << value_to_ll(p->Constraint[r][c]) << "  ";
	}
	std::cout << std::endl;
    }
*/
    ctx = (context != NULL) ? fis2ph(context) :
	      Universe_Polyhedron(sys->get_vars() + sys->get_params());
    s = DomainSimplify(p, ctx, 10000);
    Domain_Free(p);
    Domain_Free(ctx);

    res = new FIneq_system(polyhedron2ineqs(s,sys->get_params()));
    Domain_Free(s);

    gmp_unprotect();

    return res;
}

/*
FIneq_system **simplifyDomain_c(FIneq_system *context, FIneq_system **sys)
{
    Polyhedron *p, *dom, *s, *ctx;
    int i, count;
    FIneq_system **res;

    if (sys[0] == NULL) {
	res = (FIneq_system **)malloc(sizeof(FIneq_system *));
	res[0] = NULL;
	return res;
    }

    gmp_protect();

    ctx = (context != NULL) ? fis2ph(context) :
	      Universe_Polyhedron(sys[0]->get_vars() + sys[0]->get_params());
    p = NULL;
    for (i=0; sys[i] != NULL; i++) {
	dom = fis2ph(sys[i]);
	dom->next = p;
	p = dom;
    }
    s = DomainSimplify(dom, ctx, 10000);
    Domain_Free(dom);
    Domain_Free(ctx);

    count = 0;
    p = s;
    while (p != NULL) {
	p = p->next;
	count++;
    }
    res = (FIneq_system **)malloc(sizeof(FIneq_system *) * (count+1));

    p = s;
    for (i=0; i<count; i++) {
	res[i] = new FIneq_system(polyhedron2ineqs(p,sys[0]->get_params()));
	p = p->next;
    }
    Domain_Free(s);
    res[count] = NULL;

    gmp_unprotect();

    return res;
}
*/

Polyhedron *simplifyDomain_c(Polyhedron *context, Polyhedron *sys)
{
    Polyhedron *s, *ctx;
    int i, count;

    if (sys == NULL)
	return NULL;

    gmp_protect();

    if (context == NULL) {
	ctx =  Universe_Polyhedron(sys->Dimension);
    } else {
	ctx = context;
    }

    s = DomainSimplify(sys, ctx, 0);
    if (context == NULL)
	Domain_Free(ctx);

    gmp_unprotect();

    return s;
}

FIneq_system **intersectDomains_c(FIneq_system ***sys)
{
    Polyhedron *intersection, *p, *ctx;
    int i, domain, disj, count, nv, np;
    FIneq_system **res;

    gmp_protect();

    intersection = NULL;
    for (domain=0; sys[domain] != NULL; domain++) {
	FIneq_system **pDomain = sys[domain];
	Polyhedron *dom = NULL;
	for (disj=0; pDomain[disj] != NULL; disj++) {
	    Polyhedron *tmp = fis2ph(pDomain[disj]);
	    np = pDomain[disj]->get_params();
	    nv = pDomain[disj]->get_vars();
	    tmp->next = dom;
	    dom = tmp;
	}
	if (dom != NULL) {
	    if (intersection == NULL)
		intersection = dom;
	    else {
		Polyhedron *tmp = DomainIntersection(intersection, dom, 10000);
		Domain_Free(intersection);
		ctx = Universe_Polyhedron(nv+np);
		intersection = DomainSimplify(tmp, ctx, 10000);
		Domain_Free(ctx);
		Domain_Free(tmp);
		Domain_Free(dom);
	    }
	}
    }

    if (intersection == NULL) {
	res = (FIneq_system **)malloc(sizeof(FIneq_system *));
	res[0] = NULL;
    } else {
	p = intersection;
	count = 0;
	while (p != NULL) {
	    p = p->next;
	    count++;
	}
	res = (FIneq_system **)malloc(sizeof(FIneq_system *) * (count+1));

	p = intersection;
	for (i=0; i<count; i++) {
	    res[i] = new FIneq_system(polyhedron2ineqs(p, np));
	    p = p->next;
	}

	Domain_Free(intersection);
	res[count] = NULL;
    }

    gmp_unprotect();

    return res;
}




/**
 * MC 2003-02-25: this function is taken from polylib5.11.1:
 *
 * Computes the disjoint union of a union of polyhedra.
 * If flag = 0 the result is such that there are no intersections
 *                   between the resulting polyhedra,
 * if flag = 1 it computes a joint union, the resulting polyhedra are
 *                   adjacent (they have their facets in common).
 *
 * WARNING: if all polyhedra are not of same geometrical dimension
 *          duplicates may appear.
 */
Polyhedron *myDisjoint_Domain( Polyhedron *P, int flag, unsigned NbMaxRays )
{
  Polyhedron *lP, *tmp, *Result, *lR, *prec, *reste;
  Polyhedron *p1, *p2, *p3, *Pol1, *dx, *d1, *d2, *pi, *newpi;
  int i;

  if( flag!=0 && flag!=1 )
    {
      errormsg1("myDisjoint_Domain",
		"invalidarg", "flag should be equal to 0 or 1");
      return (Polyhedron*) 0;
    }
  if(!P) return (Polyhedron*) 0;

  gmp_protect();

  if(!P->next) return Polyhedron_Copy(P);

  Result = (Polyhedron *)0;

  for(lP=P;lP;lP=lP->next)
    {
      reste = Polyhedron_Copy(lP);
      prec = (Polyhedron *)0; /* preceeding lR */
      /* Intersection with each polyhedron of the current Result */
      lR=Result;
      while( lR && reste )
	{
	  /* dx = DomainIntersection(reste,lR->P,WS); */
	  dx = (Polyhedron *)0;
	  for( p1=reste; p1; p1=p1->next )
	    {
	      p3 = AddConstraints(lR->Constraint[0], lR->NbConstraints, p1,
				  NbMaxRays);
	      dx = AddPolyToDomain(p3,dx);
	    }

	  /* if empty intersection, continue */
	  if(!dx)
	    {	prec = lR;
	    lR=lR->next;
	    continue;
	    }
	  if (emptyQ(dx)) {	
	    Domain_Free(dx);
	    prec = lR;
	    lR=lR->next;
	    continue;
	  }

	  /* intersection is not empty, we need to compute the differences */
	  /* between the intersection and the two polyhedra, such that the */
	  /* results are disjoint unions (according to flag)               */
	  /* d1 = reste \ P = DomainDifference(reste,lR->P,WS);	*/
	  /* d2 = P \ reste = DomainDifference(lR->P,reste,WS); */

	  /* compute d1 */
	  d1 = (Polyhedron *)0;
	  for (p1=reste; p1; p1=p1->next)
	    {
	      pi = p1;
	      for (i=0; i<P->NbConstraints && pi ; i++)
		{

		  /* Add the constraint ( -P->constraint[i] [-1 if flag=0]) >= 0 in 'p1' */
		  /* and create the new polyhedron 'p3'.                      */
		  p3 = SubConstraint(P->Constraint[i], pi, NbMaxRays,2*flag);
		  /* Add 'p3' in the new domain 'd1' */
		  d1 = AddPolyToDomain (p3, d1);

		  /* If the constraint P->constraint[i][0] is an equality, then add   */
		  /* the constraint ( +P->constraint[i] [-1 if flag=0]) >= 0  in 'pi' */
		  /* and create the new polyhedron 'p3'.                              */
		  if( value_zero_p(P->Constraint[i][0]) ) /* Inequality */
		    {
		      p3 = SubConstraint(P->Constraint[i], pi, NbMaxRays,1+2*flag);
		      /* Add 'p3' in the new domain 'd1' */
		      d1 = AddPolyToDomain (p3, d1);

		      /* newpi : add constraint P->constraint[i]==0 to pi */
		      newpi = AddConstraints( P->Constraint[i], 1, pi, NbMaxRays);
		    }
		  else
		    {
		      /* newpi : add constraint +P->constraint[i] >= 0 in pi */
		      newpi = SubConstraint(P->Constraint[i], pi, NbMaxRays,3);
		    }
		  if( newpi && emptyQ( newpi ) )
		    {
		      Domain_Free( newpi );
		      newpi = (Polyhedron *)0;
		    }
		  if( pi != p1 )
		    Domain_Free( pi );
		  pi = newpi;
		}
	      if( pi != p1 )
		Domain_Free( pi );
	    }

	  /* and now d2 */
	  Pol1 = Polyhedron_Copy( lR );
	  for (p2=reste; p2; p2=p2->next)
	    {
	      d2 = (Polyhedron *)0;
	      for (p1=Pol1; p1; p1=p1->next)
		{
		  pi = p1;
		  for (i=0; i<p2->NbConstraints && pi ; i++)
		    {

		      /* Add the constraint ( -p2->constraint[i] [-1 if flag=0]) >= 0 in 'pi' */
		      /* and create the new polyhedron 'p3'.                      */
		      p3 = SubConstraint(p2->Constraint[i], pi, NbMaxRays,2*flag);
		      /* Add 'p3' in the new domain 'd2' */
		      d2 = AddPolyToDomain (p3, d2);

		      /* If the constraint p2->constraint[i][0] is an equality, then add   */
		      /* the constraint ( +p2->constraint[i] [-1 if flag=0]) >= 0  in 'pi' */
		      /* and create the new polyhedron 'p3'.                              */
		      if( value_zero_p(p2->Constraint[i][0]) ) /* Inequality */
			{
			  p3 = SubConstraint(p2->Constraint[i], pi, NbMaxRays,1+2*flag);
			  /* Add 'p3' in the new domain 'd2' */
			  d2 = AddPolyToDomain (p3, d2);

			  /* newpi : add constraint p2->constraint[i]==0 to pi */
			  newpi = AddConstraints( p2->Constraint[i], 1, pi, NbMaxRays);
			}
		      else
			{
			  /* newpi : add constraint +p2->constraint[i] >= 0 in pi */
			  newpi = SubConstraint(p2->Constraint[i], pi, NbMaxRays,3);
			}
		      if( newpi && emptyQ( newpi ) )
			{
			  Domain_Free( newpi );
			  newpi = (Polyhedron *)0;
			}
		      if( pi != p1 )
			Domain_Free( pi );
		      pi = newpi;
		    }
		  if( pi && pi!=p1 )
		    Domain_Free( pi );
		}
	      if( Pol1 )
		Domain_Free( Pol1 );
	      Pol1 = d2;
	    }
	  /* ok, d1 and d2 are computed */

	  /* now, replace lR by d2+dx (at least dx is nonempty) and set reste to d1 */
	  if( d1 && emptyQ(d1) )
	    {
	      Domain_Free( d1 );
	      d1 = NULL;
	    }
	  if( d2 && emptyQ(d2) )
	    {
	      Domain_Free( d2 );
	      d2 = NULL;
	    }

	  /* set reste */
	  Domain_Free( reste );
	  reste = d1;

	  /* add d2 at beginning of Result */
	  if( d2 )
	    {
	      for( tmp=d2 ; tmp->next ; tmp=tmp->next )
		;
	      tmp->next = Result;
	      Result = d2;
	      if( !prec )
		prec = tmp;
	    }

	  /* add dx at beginning of Result */
	  //			if( dx )
	  {
	    for( tmp=dx ; tmp->next ; tmp=tmp->next )
	      ;
	    tmp->next = Result;
	    Result = dx;
	    if( !prec )
	      prec = tmp;
	  }

	  /* suppress current lR */
	  if( !prec )
	    errormsg1( "myDisjoint_Domain","internalerror","internal error");
	  prec->next = lR->next;
	  Polyhedron_Free( lR );
	  lR = prec->next;
	} /* end for result */

      /* if there is something left, add it to Result : */
      if(reste)
	{
	  if(emptyQ(reste))
	    {
	      Domain_Free( reste );
	      reste = NULL;
	    }
	  else
	    {
	      Polyhedron *tnext;
	      for( tmp=reste ; tmp ; tmp=tnext )
		{
		  tnext = tmp->next;
		  tmp->next = NULL;
		  Result = AddPolyToDomain(tmp, Result);
		}
	    }
	}
    }

  gmp_unprotect();

  return( Result );
}
