-- -*- Mode:Haskell -*-
---
-- This module provides functions for accesses of dependencies
--
-- @author Tobias Langhammer, Martin Griebl
-- @version $Id: DepAcc.hs 3675 2007-07-10 12:09:54Z groessli $
--
module DepAcc 
where

import List
import Debug.Trace(trace)
import Loopo

import LoopoDS -- for testing only, to be deleted

sndOfThree :: (a,b,c) -> b
sndOfThree (_,y,_) = y

thrdOfThree :: (a,b,c) -> c
thrdOfThree (_,_,z) = z



errorFind ::  (a -> Bool) -> [a] -> a
errorFind f l = case find f l of
		Nothing -> error "Error: Element not found.\n"
		Just x  -> x




srcAcc :: DependenceID -> [AccessPair] -> [Access] -> Access 
srcAcc depNr acps accs = errorFindAcc  accs
		            (acpSrcAccID (errorFindAcp  acps depNr))
			   
dstAcc :: DependenceID -> [AccessPair] -> [Access] -> Access 
dstAcc depNr acps accs = errorFindAcc  accs
		            (acpDestAccID (errorFindAcp  acps depNr))
			   

errorFindAcp :: [AccessPair] -> DependenceID -> AccessPair
errorFindAcp acps depId = errorFind (\acp -> acpDepID acp==depId) acps

errorFindAcc :: [Access] -> AccessID -> Access 
errorFindAcc accs key = errorFind (\ a->(accID a)==key) accs



prefixSumAccuFct :: Integer -> Integer -> (Integer,Integer)
prefixSumAccuFct accu val = ( accu +  val,  accu )
superPrefixSum :: [Integer] -> [Integer]
superPrefixSum xs = snd res ++ [fst res]
       where res = mapAccumL prefixSumAccuFct 0 xs


joinZip' :: (Show a, Eq a) => [(a,b)] -> [(a,c)] -> [(a,(b,c))]
joinZip' [] _ = []
joinZip' ((key,dat):xs) l = 
      case ( find (\(s,d)-> key==s) l ) of
         Nothing    -> trace ("\nWarning: key " ++ show key ++
		              " in joinZip not found. Element dropped.")
		             (joinZip' xs l) 
         Just (s,d) -> (key,(dat,d)) : (joinZip' xs l) 


joinZip :: (Show a, Eq a) => [(a,b)] -> [(a,c)] -> [(b,c)]
joinZip xs ys = map snd (joinZip' xs ys)

combineDepAcp :: [Dependence] -> [(DependenceID,Integer,Integer)] -> [(Dependence,(Integer,Integer))]
combineDepAcp deps acps = joinZip  (zip (map depID deps) deps) 
			           (map (\ (a,b,c)->(a,(b,c))) acps)


findDep :: DependenceID -> [Dependence] -> Dependence
findDep id deps = 
    verbose "findDep" $
    case ( find (\d -> depID d==id ) deps ) of
         Nothing    -> error("dependence " ++ show id ++ " not found")
         Just d     -> d

findTrDep :: DependenceID -> [TrDependence] -> TrDependence
findTrDep id deps = 
    verbose "findTrDep" $
    case ( find (\d -> trDepOrigID d==id ) deps ) of
         Nothing    -> error("dependence " ++ show id ++ " not found")
         Just d     -> d

findIspc :: StatementID -> [Statement] -> IneqSystem
findIspc key  stmts = 
    case ( find (\(s,_) -> s==key ) stmts ) of
         Nothing    -> error("statement " ++ show key ++ " findIspc not found")
         Just (s,i) -> i


findAccess :: AccessID -> [RawAccess] -> RawAccess
findAccess key raccs =
    case ( find (\(_,_,n,_,_) -> n == integerFromID key ) raccs ) of
         Nothing    -> error ("access " ++ show key ++ " not found - do not use cffada") 
-- Reason for that problem:
-- with cffada the acc numbers (LaNode) do not match with the acp (Occ) numbers
         Just a -> a

-- findIndexFct :: Integer -> [RawAccess] -> AffineFunction
-- findIndexFct key raccs = f
--    where (_,_,_,_,f) = findAccess key raccs


data DepPart = Source | Dest deriving (Show,Eq)


depTypeToAccType :: DepPart -> DependenceType -> AccessType
depTypeToAccType dp dt = depTypeToAccType' dp 
			 (toEnum (if asInt > 6 then asInt-6 else asInt))
    where asInt = fromEnum dt


depTypeToAccType' Source UniformTrue = Write
depTypeToAccType' Source UniformFlow = Write
depTypeToAccType' Source UniformOutput = Write

depTypeToAccType' Source UniformAnti = Read
depTypeToAccType' Source UniformInput = Read

depTypeToAccType' Dest UniformTrue = Read
depTypeToAccType' Dest UniformFlow = Read
depTypeToAccType' Dest UniformInput = Read

depTypeToAccType' Dest UniformOutput = Write
depTypeToAccType' Dest UniformAnti = Write


-- die stmts mssen in der geschrumpften, nicht voll aufgeblasenen
-- Version vorliegen:
acpsToAccs :: SurroundingLoops -> [AccessPair] -> [Statement] -> [TrDependence] -> [RawAccess] -> [Access]
acpsToAccs slps acps stmts deps raccs = acpsToAccs' slps acps stmts deps raccs [] 


acpsToAccs' :: SurroundingLoops -> [AccessPair] -> [Statement] -> [TrDependence] -> [RawAccess] -> [AccessID] -> [Access]
acpsToAccs' _ [] _ _ _ _ = []
acpsToAccs' slps (AccessPair {acpDepID=dep,acpSrcAccID=src,acpDestAccID=dest} : xs) stmts deps raccs knowns =
    (if (elem src knowns) then [] else [accOf(src, dep, Source)]) ++
    (if (elem dest knowns) then [] else [accOf(dest, dep, Dest)]) ++
    acpsToAccs' slps xs stmts deps raccs (src : (dest : knowns )) 
    where
        accOf :: (AccessID,DependenceID,DepPart) -> Access
	accOf(accid,depid,dp) = 
	    let dep = findTrDep depid deps
		stmtid = if dp==Source then trDepOrigSrcStmtID dep
			 else trDepOrigDestStmtID dep
                (arrayId,_,_,_,f) = findAccess accid raccs
		in Access {
		  accID = accid,
		  accIndexFct = trimToSurroundingLoops slps stmtid f,
		  accStmtID = stmtid,
		  accDomain = findIspc stmtid stmts, -- reduce slps (stmtid, findIspc stmtid stmts),
		  accType = depTypeToAccType dp (trDepType dep),
		  accArrayID = integerToID arrayId
		  }




---
-- retrieve source and destination access of a given target depedency
--
-- @param accs List of Accesses
-- @param acps List of access pairs referencing accs
-- @param dep  Dependency to be used
-- @returns Pair of the source and destination accesses of dep
getSrcAndDstAccs :: [Access] -> [AccessPair] 
		 -> TrDependence -> (Access,Access)
getSrcAndDstAccs accs acps dep = (sAcc,dAcc)
    where  sAcc = srcAcc (trDepOrigID dep) acps accs 
	   dAcc = dstAcc (trDepOrigID dep) acps accs 


--
-- Debugging function. Can be used to display the results of finished functions.
--
verbose :: Show a => String -> a -> a
verbose str x =
    trace ("$? " ++ str ++ "...") $
    let skipIt = False -- suppress debugging output
    in if skipIt then x
       -- else seq x $ trace ("Verbose: " ++ str ++ ": result=" ++ show x) x
       else seq x $ trace ("$! " ++ str ++ " done. ") x
