module HsLoopoTest(main) where

import Control.Monad
import System.Random
import System.Exit
import Test.QuickCheck as QC

import Debug.Trace

import Loopo hiding ( kernel, inverseMatrix )

kernel = kernelOrdFrac
inverseMatrix = inverseMatrixOrdFrac

instance Integral a => Arbitrary (Ratio a) where
    arbitrary = sized $ \n -> let s = round $ sqrt $ fromIntegral n in
		resize s
		(let arb = arbitrary :: Gen Integer in
		 arb >>= \a ->
		 arb >>= \b ->
		 return (fromIntegral a + (fromIntegral b % fromIntegral s)))
    coarbitrary = error "coarbitrary: undefined for  Integral a => Arbitrary (Ratio a)"

instance Arbitrary (GMatrix Rational) where
    arbitrary = sized $ \n -> let s = round $ sqrt $ fromIntegral n in
		arbitrary >>= \r' -> let r = abs r' in
		arbitrary >>= \c' -> let c = abs c' in 
		sequence (replicate (r*c) arbitrary) >>= \coeffs ->
		return $ matrix (r,c,coeffs)
    coarbitrary = error "coarbitrary: undefined for  Arbitrary (GMatrix Rational)"

newtype SquareMatrix a = SquareMatrix (GMatrix a)
    deriving (Show)

instance Arbitrary (SquareMatrix Rational) where
    arbitrary = sized $ \n -> let s = round $ sqrt $ fromIntegral n in
		arbitrary >>= \r' -> let r = abs r' in
		sequence (replicate (r*r) arbitrary) >>= \coeffs ->
		return $ SquareMatrix $ matrix (r,r,coeffs)
    coarbitrary = error "coarbitrary: undefined for  Arbitrary (SquareMatrix Rational)"

instance Arbitrary (GIneqSystem Rational) where
    arbitrary =
	do let arb = arbitrary :: Gen Int
           nV <- liftM ((+1).abs) arb 
           nP <- liftM abs arb
	   nR <- liftM abs arb
	   let coeffs = do cs <- sequence $ replicate (nV+nP+1) arbitrary
		           if (any (/=0) $ take nV cs)
			      then return cs
			      else coeffs
	   co <- sequence $ replicate nR coeffs
	   let m = matrix co
           return (ineqSystem ((nV,nP), m))

    coarbitrary = error "coarbitrary: undefined for  Arbitrary (GIneqSystem Rational)"


{-
newtype FmIneqSystem = FmIneqSystem IneqSystem
    deriving Show

instance Arbitrary (FmIneqSystem) where
    arbitrary =
	do let arb = arbitrary :: Gen Int
           nV <- liftM abs arb 
           nP <- liftM abs arb
	   nR <- liftM abs arb
	   co <- sequence [makeRow i nR nV nP | i <- [1..nR]]
	   let m = matrix co
           trace (show co) $ return $ FmIneqSystem (ineqSystem ((nV,nP), m))
	where
	makeRow i nR nV nP =
	    do vs <- QC.vector (nV-offset)
	       let zeros = replicate offset 0
	       ps <- QC.vector (nP+1)
	       trace (show (nV,offset,nP,nR,zeros)) $ return (vs ++ zeros ++ ps)
	    where
	    offset = (i*nV) `div` nR
	    
    coarbitrary = error "coarbitrary: undefined for  Arbitrary FmIneqSystem"
-}

permute :: [a] -> Gen [a]
permute [] = return []
permute xs =
    do (i,y) <- oneof $ zipWith (\i x -> return (i,x)) [0..] xs
       let (xs1,_:xs2) = splitAt i xs
       ys <- permute (xs1 ++ xs2)
       return (y:ys)

checkKernel (m::Matrix) =
    and [all (==0) $ vectorEntries (m `mulMV` k) | k <- k1] &&
    and [any (/=0) $ vectorEntries (m `mulMV` k) | k <- k2] &&
    length k1 + length k2 == nColumns m
    where
    k1 = kernel m
    k2 = kernel $ rowVectorsToMatrix (nColumns m) k1

checkInverse (SquareMatrix m :: SquareMatrix Rational) =
--    trace (show m) $
    (null (kernel m)) ==>
        m * inverseMatrix m == unitMatrix (nRows m) (nRows m) &&
	inverseMatrix m * m == unitMatrix (nRows m) (nRows m)

checkSortIneqsFM (sys :: IneqSystem) =
    -- Qepcad fails to solve the 'impliesQf ...' for non-trivial examples!
    trace (show sys) $
    impliesQf [] ((s `qfImpl` r) `QfAnd` (r `qfImpl` s))
    where
    res = sortFn sys Nothing
    r = case res of
	  Nothing -> QfAt AtFalse
	  Just re -> ineqSystemToFormula $ fmap constPoly re
    s = ineqSystemToFormula $ fmap constPoly sys

    sortFn = sortIneqsFM


runTest :: Testable a => Int -> Int -> a -> String -> IO (Maybe Bool, String)
runTest count maxTriv theTest name =
    do stdgen <- newStdGen
       return $ rt stdgen 0 0
    where
    rt rnd pass triv
	| pass == count   = (Just True, name ++ ": passed " ++ show count ++ " tests")
	| triv >  maxTriv = (Nothing,   name ++ ": arguments exhausted after " ++
			                show triv ++ " iterations (passed " ++
			                show pass ++ " tests)")
	| otherwise =
	    case ok result of
	      Just True -> rt rnd2 (pass+1) triv
	      Nothing   -> rt rnd2 pass     (triv+1)
	      Just False -> (Just False, name ++": falsifiable after " ++ show pass ++
			            " tests:\n" ++
			            unlines (arguments result))
	where
	result = generate ((pass `div` 2)+3) rnd1 (evaluate theTest)
	(rnd1, rnd2) = split rnd


data Test = forall a. Testable a => Test a Int Int String


testsToPerform :: [Test]
testsToPerform =
    [ Test checkInverse      70 50 "checkInverse",
      Test checkKernel       70  0 "checkKernel",
      Test checkSortIneqsFM  20  0 "checkSortIneqsFM"
    ]

runTests :: [Test] -> IO (Bool, String)
runTests tests =
    do reslist <- mapM (\(Test prop passes maxTrivial name) ->
                        runTest passes maxTrivial prop name) tests
       let (results, strings) = unzip reslist
       return (Just False `notElem` results, unlines strings)

main :: IO ()
main = do (res, string) <- runTests testsToPerform
	  putStrLn string
	  unless res
	    (exitWith (ExitFailure 1))
	  return ()



-- kernel $ rowVectorsToMatrix $ kernel m
-- liefert falsches Ergebnis wegen int-Ueberlauf auf der C Seite...
m :: Matrix
m = matrix [[27 % 4,  7 % 6,  (-12) % 5,  (-20) % 3,  38 % 5],
	    [(-2) % 1,  (-13) % 2,  7 % 1,  (-47) % 8,  (-27) % 4],
	    [13 % 2,  6 % 1,  (-5) % 1,  (-11) % 3,  21 % 8]]
