module ISS.DataSplitting.SplitDependences(doSplitDependences) where

import ISS.DataSplitting.Info
import ISS.Language.Language
import ISS.Options
import ISS.Util

import LMath.Types.AffineExp
import LMath.Types.IneqSystem
import LMath.Types.LClasses
import LMath.Types.Matrix
import LoopoDS.Dependence
import LoopoDS.IntegralID
import LoopoDS.Statement


-- |Splits dependences according to given statement splitting information,
--  writes the result to a file given by the command-line options and returns
--  information about the performed splitting
doSplitDependences :: Options -> [Dependence] -> SplitInfo -> IO (DepSplitInfo)
doSplitDependences options dependences splitInfo = do
    let splittedDependencesInfo = splitDependences splitInfo dependences
    
    let newDependences = concat (map (map snd . snd . snd) splittedDependencesInfo)
    
    writeDependences (createFullFileName (outDirectory options) (outFile (depFileName options))) newDependences
    
    return splittedDependencesInfo
    

-- |Splits dependences according to given statement splitting information and
--  returns information about the performed splitting
splitDependences :: SplitInfo -> [Dependence] -> DepSplitInfo
splitDependences splitInfo deps = let depsMap = splitDependences' splitInfo [1..] deps
                                  in  [(depID dep, (dep, newDeps)) | (dep, depTarget) <- depsMap,
                                                                     let newDeps = [(depID newDep, newDep) | newDep <- depTarget]]


-- |Splits dependences according to given statement splitting information and
--  a list of free dependenceIDs and returns information about the performed
--  splitting
splitDependences' :: SplitInfo -> [Integer] -> [Dependence] -> [(Dependence, [Dependence])]
splitDependences' splitInfo free []         = []
splitDependences' splitInfo free (dep:deps) = let (newFree, newDeps) = splitDependence splitInfo dep free
                                              in  (dep, newDeps):splitDependences' splitInfo newFree deps


-- |Splits one single dependence according to given statement splitting
--  information and a list of free dependenceIDs and returns information about
--  the performed splitting
splitDependence :: SplitInfo -> Dependence -> [Integer] -> ([Integer], [Dependence])
splitDependence splitInfo dep = makeNewDeps [(src, dst) | src <- newStatements (depSrcStmtID dep),
                                                          dst <- newStatements (depDestStmtID dep)] where
    newStatements :: StatementID -> [(StatementID, IneqSystem)]
    newStatements oldStmtId = case find splitInfo oldStmtId of
                                  Nothing -> []
                                  Just list -> list
    
    makeNewDeps :: [((StatementID, GIneqSystem Rational), (StatementID, GIneqSystem Rational))] -> [Integer] -> ([Integer], [Dependence])
    makeNewDeps []                                             free         = (free, [])
    makeNewDeps (((srcStmt, srcSpace), (dstStmt, dstSpace)):t) (newId:free) = if areIneqsInfeasible newDstSpace || areIneqsInfeasible newSrcSpace then
                                                                                  makeNewDeps t (newId:free)
                                                                              else
                                                                                  let (newFree, newDeps) = makeNewDeps t free in
                                                                                  (newFree, newDep:newDeps) where
        newSrcSpace = simplifyIneqs Nothing (srcSpace `join` depSrcDomain dep)
        newDstSpace = simplifyIneqs Nothing ((ineqSystem ((vars,params),getMatrix (toAffineExp newSrcSpace) * getMatrix (depHTrafo dep))) `join` depDestDomain dep `join` dstSpace)
        newDep = dep {depSrcDomain = newSrcSpace, depDestDomain = newDstSpace, depSrcStmtID = srcStmt, depDestStmtID = dstStmt, depID = integerToID newId}
        vars = nVars newSrcSpace
        params = nParams newSrcSpace
        
