module ISS.DataSplitting.SplitOther (doSplitAccesses, doSplitAccessPairs, doSplitGroups,
                                     doWriteNewISPCs, doDetermineSurroundingLoops,
                                     doDetermineSurroundingIfs,
                                     SplitInfo, DepSplitInfo, LaNodeSplitInfo) where

import ISS.DataSplitting.Info
import ISS.Options
import ISS.Language.Language
import ISS.Util

import LMath.Types.IneqSystem
import LoopoDS.Access
import LoopoDS.AccessPair
import LoopoDS.Dependence
import LoopoDS.Groups
import LoopoDS.IntegralID
import LoopoDS.Statement
import LoopoDS.SurroundingIfs
import LoopoDS.SurroundingLoops
import ScannerParser.ParseTree



-- |Writes the new index spaces to a file given by the command-line options
doWriteNewISPCs :: Options -> SplitInfo -> IO ()
doWriteNewISPCs options splitInfo = writeISPC stmts filename where
    stmts = (concat . map snd) splitInfo
    filename = createFullFileName (outDirectory options) (outFile (ispcFileName options))


-- |Splits accesses according to given LaNode splitting information and
--  statement splitting information and writes the result to a file given by
--  the command-line options
doSplitAccesses :: Options -> LaNodeSplitInfo -> SplitInfo -> IO ()
doSplitAccesses options = doSplitAccesses' (opSplitAccesses options) where
    doSplitAccesses' Nothing                   _ splitInfo = return ()
    doSplitAccesses' (Just filenames) laNodeInfo splitInfo = do
        --read accesses from file and parse them
        maybeAccesses <- readACC (createFullFileName (inDirectory options) (inFile filenames))
        
        --check, if they could be parsed
        let oldAccesses = throwErrorIfWrong (languageParseError (languageText "accesses")) maybeAccesses
        
        --split the accesses
        let newAccesses = splitAccesses laNodeInfo splitInfo oldAccesses
        
        --write accesses to file
        writeACC newAccesses (createFullFileName (outDirectory options) (outFile filenames))
        
        return ()


-- |Splits access pairs according to given statement splitting information,
--  dependence splitting information and tree splitting information and writes
--  the result to a file given by the command-line options
doSplitAccessPairs :: Options -> SplitInfo -> DepSplitInfo -> LaNodeSplitInfo -> IO ()
doSplitAccessPairs options = doSplitAccessPairs' (opSplitAccessPairs options) where
    doSplitAccessPairs' Nothing          _         _            _               = return ()
    doSplitAccessPairs' (Just filenames) splitInfo depSplitInfo laNodeSplitInfo = do
        --read access-pairs from file and parse them
        maybeAccessPairs <- readACP (createFullFileName (inDirectory options) (inFile filenames))
        
        --check, if they could be parsed
        let oldAccessPairs = throwErrorIfWrong (languageParseError (languageText "accesspairs")) maybeAccessPairs
        
        --split the access-pairs
        let newAccessPairs = splitAccessPairs splitInfo depSplitInfo laNodeSplitInfo oldAccessPairs
        
        --write access-pairs to file
        writeACP newAccessPairs (createFullFileName (outDirectory options) (outFile filenames))
        
        return ()


-- |Splits statement groups according to given statement splitting information
--  and writes the result to a file given by the command-line options
doSplitGroups :: Options -> SplitInfo -> IO ()
doSplitGroups options = doSplitGroups' (opSplitGroups options) where
    doSplitGroups' Nothing          _         = return ()
    doSplitGroups' (Just filenames) splitInfo = do
        --read groups from file and parse them
        maybeGroups <- readGroups (createFullFileName (inDirectory options) (inFile filenames))
        
        --check, if they could be parsed
        let oldGroups = throwErrorIfWrong (languageParseError (languageText "groups")) maybeGroups
        
        --split the groups
        let newGroups = splitGroups splitInfo oldGroups
        
        --write groups to file
        writeGroups newGroups (createFullFileName (outDirectory options) (outFile filenames))
        
        return ()


-- |Determines the surrounding loops of every statement in the new parse-tree
--  and writes the result to a file given by the command-line options
doDetermineSurroundingLoops :: Options -> LambdaTree -> IO ()
doDetermineSurroundingLoops options = doDetermineSurroundingLoops' (opSplitSurrLoops options) where
    doDetermineSurroundingLoops' Nothing          _    = return ()
    doDetermineSurroundingLoops' (Just filename) tree = do
        --read surrLoops from file and parse them
        maybeLoops <- readSLP (createFullFileName (inDirectory options) filename)
        
        --check, if they could be parsed
        let oldLoops = throwErrorIfWrong (languageParseError (languageText "surrloops")) maybeLoops
        
        --determine new surrLoops
        let newSLPs = determineSurroundingLoops tree
        
        --write surrLoops to file
        writeSLP newSLPs (createFullFileName (outDirectory options) filename)
        
        return ()


-- |Determines the surrounding ifs of every statement in the new parse-tree
--  and writes the result to a file given by the command-line options
doDetermineSurroundingIfs :: Options -> LambdaTree -> IO ()
doDetermineSurroundingIfs options = doDetermineSurroundingIfs' (opSplitSurrIfs options) where
    doDetermineSurroundingIfs' Nothing          _    = return ()
    doDetermineSurroundingIfs' (Just filename) tree = do
        --read surrIfs from file and parse them
        maybeIfs <- readSIF (createFullFileName (inDirectory options) filename)
        
        --check, if they could be parsed
        let oldIfs = throwErrorIfWrong (languageParseError (languageText "surrifs")) maybeIfs
        
        --determine new surrIfs
        let newSIFs = determineSurroundingIfs tree
        
        --write surrIfs to file
        writeSIF newSIFs (createFullFileName (outDirectory options) filename)
        
        return ()


-- |Splits accesses according to given LaNode splitting information and
--  statement splitting information
splitAccesses :: LaNodeSplitInfo -> SplitInfo -> [RawAccess] -> [RawAccess]
splitAccesses laNodeSplit splitInfo = concat . (map (splitAccess splitInfo)) where
    splitAccess :: SplitInfo -> RawAccess -> [RawAccess]
    splitAccess splitInfo (arrayId,accType,oldLaNode,stmtNr,indexFct) = [(arrayId, accType, fromIntegral newLaNode, fromIntegral newStmtNr, indexFct)
        | (newLaNode, newStmtNr) <- getNodeAndNr (fromIntegral oldLaNode) (find splitInfo (integerToID stmtNr))] where
            getNodeAndNr :: Int -> Maybe [(StatementID, GIneqSystem Rational)] -> [(Int, Int)]
            getNodeAndNr oldLaNode Nothing     = []
            getNodeAndNr oldLaNode (Just list) = map (\(stmtId, _) -> (findNewNode stmtId oldLaNode, integerFromID stmtId)) list
            findNewNode stmtId oldLaNode = case find laNodeSplit (stmtId, oldLaNode) of
                                               Nothing -> error "node gone lost"
                                               Just newLaNode -> newLaNode


-- |Convert an AccessPair from the triple representation to the data structure
--  (necessary due to some inconsistency in the types used by loopo - some
--  functions use the triple, others use the data structure)
toPair :: (DependenceID, AccessID, AccessID) -> AccessPair
toPair (dep, acc1, acc2) = AccessPair {acpDepID = dep, acpSrcAccID = acc1, acpDestAccID = acc2}


-- |Splits access pairs according to given statement splitting information,
--  dependence splitting information and tree splitting information
splitAccessPairs :: SplitInfo -> DepSplitInfo -> LaNodeSplitInfo -> [AccessPair] -> [AccessPair]
splitAccessPairs splitInfo depSplitInfo laNodeSplitInfo = concat . (map (splitAccessPair splitInfo depSplitInfo laNodeSplitInfo)) where
    splitAccessPair :: SplitInfo -> DepSplitInfo -> LaNodeSplitInfo -> AccessPair -> [AccessPair]
    splitAccessPair splitInfo depSplitInfo laNodeSplitInfo pair = case find depSplitInfo (acpDepID pair) of
        Nothing                -> []
        Just (oldDep, newDeps) -> let constructPair :: DependenceID -> Maybe Int -> Maybe Int -> [AccessPair]
                                      constructPair _ Nothing _ = []
                                      constructPair _ _ Nothing = []
                                      constructPair newDepID (Just src) (Just dst) = [toPair (newDepID, integerToID src, integerToID dst)]
                                  in  concat [constructPair newDepID
                                                            (find laNodeSplitInfo (depSrcStmtID newDep, integerFromID (acpSrcAccID pair)))
                                                            (find laNodeSplitInfo (depDestStmtID newDep, integerFromID (acpDestAccID pair)))
                                              | (newDepID, newDep) <- newDeps]


-- |Splits statement groups according to given statement splitting information
splitGroups :: SplitInfo -> (Integer, [[Integer]]) -> (Integer, [[Integer]])
splitGroups splitInfo (oldNum, oldGroups) = (fromIntegral (sum (map length newGroups)), newGroups) where
    newGroups = map (splitGroup splitInfo) oldGroups
    splitGroup splitInfo = concat . (map (splitStmtNr splitInfo))
    splitStmtNr splitInfo stmtNr = case find splitInfo (integerToID stmtNr) of
        Nothing    -> []
        Just found -> [integerFromID newId | (newId,_) <- found]


-- |Determines the surrounding loops of every statement in the new parse tree
determineSurroundingLoops :: LambdaTree -> SurroundingLoops
determineSurroundingLoops tree = (SLP . map (fillUpSIF numLoops)) slps where
    
    (numLoops, slps) = detSLPNodes [] tree
    
    detSLPNode :: [Bool] -> LambdaNode -> (Integer, [(StatementID, [Bool])])
    detSLPNode loops node@(ForLoop {}) = (numLoops + 1, slps) where
        (numLoops, slps) = detSLPNodes (loops ++ [True]) (loopBody node)
    detSLPNode loops node@(WhileLoop {}) = (numLoops + 1, slps) where
        (numLoops, slps) = detSLPNodes (loops ++ [True]) (loopBody node)
    detSLPNode loops node@(FuncDecl {}) = detSLPNodes loops (funcbody node)
    detSLPNode loops node@(If {}) = (numLoops1 + numLoops2, slps1 ++ slps2) where
        (numLoops1, slps1) = detSLPNodes loops (thenbranch node)
        (numLoops2, slps2) = detSLPNodes (loops ++ false numLoops1) (elsebranch node)
    detSLPNode loops node =
        if hasstatement node then (0, [(statement node, loops)]) else (0, [])

    detSLPNodes :: [Bool] -> [LambdaNode] -> (Integer, [(StatementID, [Bool])])
    detSLPNodes loops [] = (0, [])
    detSLPNodes loops (node:nodes) =
        let (numLoops1, slps1) = detSLPNode loops node
            (numLoops2, slps2) = detSLPNodes (loops ++ false numLoops1) nodes
        in  (numLoops1 + numLoops2, slps1 ++ slps2)
    
    false :: Integer -> [Bool]
    false n = take' n (repeat False)

    fillUpSIF :: Integer -> (StatementID, [Bool]) -> (StatementID, [Bool])
    fillUpSIF n (sid, list) = (sid, fillUp n list)

    fillUp :: Integer -> [Bool] -> [Bool]
    fillUp n list = list ++ false (n - fromIntegral (length list))
