module ISS.DataSplitting.SplitTree(splitTreeWithElse, splitTreeWithoutElse) where

import ISS.DataSplitting.Info
import ISS.SplittingAlgorithm.ToolsGeneral
import ISS.Util

import LMath.Logic.Domain
import LMath.Types.IneqSystem
import LMath.Types.Matrix
import LoopoDS.Statement
import ScannerParser.ParseTree
import ScannerParser.SymbolTable

type SplitFunc = [Int] -> [StatementID] -> LambdaNode -> SplittedStatement -> ([Int], [StatementID], [LambdaNode], [(StatementID, GIneqSystem Rational)], LaNodeSplitInfo)


-- |Given a list of SplittedStatements this function splits the parse tree
--  according to this information using else-branches
splitTreeWithElse :: [(Int, LambdaNodeState)] -> Integer -> [SplittedStatement] -> LambdaTree -> (LambdaTree, SplitInfo, LaNodeSplitInfo)
splitTreeWithElse symtab params = splitTreeWithSplitFunc (splitFuncElse symtab params)


-- |Given a list of SplittedStatements this function splits the parse tree
--  according to this information not using else-branches
splitTreeWithoutElse :: [(Int, LambdaNodeState)] -> Integer -> [SplittedStatement] -> LambdaTree -> (LambdaTree, SplitInfo, LaNodeSplitInfo)
splitTreeWithoutElse symtab params = splitTreeWithSplitFunc (splitFuncWOElse symtab params)



splitTreeWithSplitFunc :: SplitFunc -> [SplittedStatement] -> LambdaTree -> (LambdaTree, SplitInfo, LaNodeSplitInfo)
splitTreeWithSplitFunc func splits tree = lastThree (splitTree splitsMap (freshNodeIDs tree) (freshStatementIDs tree) func tree) where
                                       splitsMap = map (splittedToMap) splits
                                       splittedToMap :: SplittedStatement -> (StatementID, SplittedStatement)
                                       splittedToMap splitted = (fst (originalStatement splitted), splitted)
                                       lastThree (_,_,x,y,z) = (x,y,z)


splitTree :: [(StatementID, SplittedStatement)] -> [Int] -> [StatementID] -> SplitFunc -> LambdaTree
             -> ([Int], [StatementID], LambdaTree, SplitInfo, LaNodeSplitInfo)
splitTree splits free freeSIDs splitFunc []     = (free, freeSIDs, [], [], [])
splitTree splits free freeSIDs splitFunc (x:xs) = let (newFree, newFreeSIDs, splittedNode, splitInfo1, lnsi1) = splitNode splits free freeSIDs splitFunc x
                                                      (newNewFree, newNewFreeSIDs, splittedTail, splitInfo2, lnsi2) = splitTree splits newFree newFreeSIDs splitFunc xs
                                                  in  (newNewFree, newNewFreeSIDs, splittedNode ++ splittedTail, splitInfo1 ++ splitInfo2, lnsi1 ++ lnsi2)


splitNode :: [(StatementID, SplittedStatement)] -> [Int] -> [StatementID] -> SplitFunc -> LambdaNode
             -> ([Int], [StatementID], [LambdaNode], SplitInfo, LaNodeSplitInfo)
splitNode splits free freeSIDs splitFunc node =
    let splitNode' node@(ForLoop {})   = let (newFree, newFreeSIDs, newSubNodes, splitInfo, lnsi) = splitTree splits free freeSIDs splitFunc (loopBody node)
                                         in  (newFree, newFreeSIDs, [node {loopBody = newSubNodes}], splitInfo, lnsi)
        splitNode' node@(If {})        = let (newFree, newFreeSIDs, newThenNodes, splitInfo1, lnsi1)       = splitTree splits free freeSIDs splitFunc (thenbranch node)
                                             (newNewFree, newNewFreeSIDs, newElseNodes, splitInfo2, lnsi2) = splitTree splits newFree newFreeSIDs splitFunc (elsebranch node)
                                         in  (newNewFree, newNewFreeSIDs, [node {thenbranch = newThenNodes, elsebranch = newElseNodes}], splitInfo1 ++ splitInfo2, lnsi1++lnsi2)
        splitNode' node@(WhileLoop {}) = let (newFree, newFreeSIDs, newSubNodes, splitInfo, lnsi) = splitTree splits free freeSIDs splitFunc (loopBody node)
                                         in  (newFree, newFreeSIDs, [node {loopBody = newSubNodes}], splitInfo, lnsi)
        splitNode' node@(FuncDecl {})  = let (newFree, newFreeSIDs, newSubNodes, splitInfo, lnsi) = splitTree splits free freeSIDs splitFunc (funcbody node)
                                         in  (newFree, newFreeSIDs, [node {funcbody = newSubNodes}], splitInfo, lnsi)
        splitNode' node                = (free, freeSIDs, [node], [], [])
    in  if isStatement node then
            case find splits (statement node) of
                Nothing    -> splitNode' node
                Just split -> let (freeNIDs, newFreeSIDs, nodes, info, lninfo) = splitFunc free freeSIDs node split
                              in  (freeNIDs, newFreeSIDs, nodes, [(statement node, info)], lninfo)
        else
            splitNode' node


-- |"SplitFunc"-generator for splitting WITH else-branches
splitFuncElse :: [(Int, LambdaNodeState)] -> Integer -> SplitFunc
splitFuncElse symtab params free freeSIDs node split = splitFuncElseSplit free freeSIDs node (splitting split) where
    splitFuncElseSplit :: [Int] -> [StatementID] -> LambdaNode -> Split -> ([Int], [StatementID], [LambdaNode], [(StatementID, GIneqSystem Rational)], LaNodeSplitInfo)
    splitFuncElseSplit free (sid:freeSIDs) node (NoSplit {}) = let (newFree, newNode, lninfo) = copyNode free sid node
                                                               in  (newFree, freeSIDs, [newNode], [], lninfo)
    splitFuncElseSplit free freeSIDs node s@(Split {}) = let space = dDisjointUnion (completeSpace (ifBranch s))
                                                             (newFree, newFreeSIDs, newThen, _, lninfo1)          = splitFuncElseSplit free freeSIDs node (ifBranch s)
                                                             (newNewFree, id:newNewFreeSIDs, newElse, _, lninfo2) = splitFuncElseSplit newFree newFreeSIDs node (elseBranch s)
                                                             (n:newNewNewFree, newCond) = ineqSystemsToCondition newNewFree id space symtab
                                                             dDisjointUnion = (map (polyIneqToIneq params)) . (disjointUnionD Nothing)
                                                         in  case space of
                                                                 (_:_) -> (newNewNewFree, newNewFreeSIDs, [If {nodeid = n,
                                                                                           occurence = 1, --TODO
                                                                                           statement = id,
                                                                                           targetStatement = TStatementID 0,
                                                                                           condition = Just newCond,
                                                                                           thenbranch = newThen,   
                                                                                           elsebranch = newElse
                                                                                          }], [], lninfo1 ++ lninfo2)
                                                                 [] -> error "empty space!"


-- |"SplitFunc"-generator for splitting WITHOUT else-branches
splitFuncWOElse :: [(Int, LambdaNodeState)] -> Integer -> SplitFunc
splitFuncWOElse symtab params free freeSIDs node split = splitFuncWOElseSplit free freeSIDs (splitToIneqs params (splitting split)) where
    splitFuncWOElseSplit :: [Int] -> [StatementID] -> [IneqSystem] -> ([Int], [StatementID], [LambdaNode], [(StatementID, GIneqSystem Rational)], LaNodeSplitInfo)
    splitFuncWOElseSplit free freeSIDs []     = (free, freeSIDs, [], [], [])
    splitFuncWOElseSplit free freeSIDs (x:xs) = (newNewFree, newNewFreeSIDs, newNode:newRest, (sid, x):splitInfo, lninfo1 ++ lninfo2) where
        (newFree, newFreeSIDs, newNode, sid, lninfo1)             = splitFuncWOElseIneq free freeSIDs x
        (newNewFree, newNewFreeSIDs, newRest, splitInfo, lninfo2) = splitFuncWOElseSplit newFree newFreeSIDs xs
        
        splitFuncWOElseIneq :: [Int] -> [StatementID] -> IneqSystem -> ([Int], [StatementID], LambdaNode, StatementID, LaNodeSplitInfo)
        splitFuncWOElseIneq free (sidif:sidstmt:freeSIDs) ineq = let (newFree, newCond)              = ineqSystemToCondition free sidif ineq symtab
                                                                     (n:newNewFree, newThen, lninfo) = copyNode newFree sidstmt node
                                                                 in  if nRows (getMatrix ineq) == 0 then (free, sidif:sidstmt:freeSIDs, node, statement node, [((statement node, nodeid node), nodeid node)])
                                                                     else (newNewFree, freeSIDs, If {
                                                                            nodeid = n,
                                                                            occurence = 1, --TODO
                                                                            statement = sidif,
                                                                            targetStatement = TStatementID 0,
                                                                            condition = Just newCond,
                                                                            thenbranch = [newThen],
                                                                            elsebranch = []
                                                                           }, sidstmt, lninfo)


-- |Creates a LambdaNode-Expression from several given inequation system
--  (or-combination of single boolean expressions, which represent single
--  inequation systems)
ineqSystemsToCondition :: [Int] -> StatementID -> [IneqSystem] -> [(Int, LambdaNodeState)] -> ([Int], LambdaNode)
ineqSystemsToCondition (n:free) sid (x:xs@(_:_)) symtab = let (newFree, nodeA) = ineqSystemToCondition free sid x symtab
                                                              (newNewFree, nodeB) = ineqSystemsToCondition newFree sid xs symtab
                                                          in  (newNewFree, Expression {nodeid = n,
                                                                                       occurence = 1, --TODO
                                                                                       statement = sid,
                                                                                       targetStatement = TStatementID 0,
                                                                                       operator = LNOp_Or,
                                                                                       condition = Nothing,
                                                                                       firstOperand = Just nodeA,
                                                                                       secondOperand = Just nodeB
                                                                                      })
ineqSystemsToCondition free sid [x] symtab = ineqSystemToCondition free sid x symtab
ineqSystemsToCondition free sid [] symtab = error "no ineqsystem given!!!"


-- |Creates a LambdaNode-Expression from a given inequation system
--  (and-combination of single boolean expressions, which represent single
--  inequations)
ineqSystemToCondition :: [Int] -> StatementID -> IneqSystem -> [(Int, LambdaNodeState)] -> ([Int], LambdaNode)
ineqSystemToCondition free sid ineqSys = ineqListToCondition free (rows (getMatrix (integralIneqSystem ineqSys))) where
    ineqListToCondition :: [Int] -> [GMatrix Int] -> [(Int, LambdaNodeState)] -> ([Int], LambdaNode)
    ineqListToCondition (n:free) (x:xs@(_:_)) symtab = let (newFree, nodeA) = ineqToCondition free sid x symtab
                                                           (newNewFree, nodeB) = ineqListToCondition newFree xs symtab
                                                       in  (newNewFree, Expression {nodeid = n,
                                                                             occurence = 1, --TODO
                                                                             statement = sid,
                                                                             targetStatement = TStatementID 0,
                                                                             operator = LNOp_And,
                                                                             condition = Nothing,
                                                                             firstOperand = Just nodeA,
                                                                             secondOperand = Just nodeB
                                                                            })
    ineqListToCondition free [x] symtab = ineqToCondition free sid x symtab


-- |Creates a LambdaNode-Expression from a given inequation
ineqToCondition :: [Int] -> StatementID -> GMatrix Int -> [(Int, LambdaNodeState)] -> ([Int], LambdaNode)
ineqToCondition (n1:n2:free) sid matr st = (newFree, ineq) where
    
    ineq = Expression {
                       nodeid = n1,
                       occurence = 1,
                       statement = sid,
                       targetStatement = TStatementID 0,
                       operator = LNOp_GEQ,
                       condition = Nothing,
                       firstOperand = Just formula,
                       secondOperand = Just zero
                      }
    (newFree, formula) = ineqToCondition' free (entries matr) st
    zero = ExprIndBlobConst {nodeid = n2, occurence = 1, name = 0, state = LNS_Constant, statement = sid, targetStatement = TStatementID 0}
    
    ineqToCondition' :: [Int] -> [Int] -> [(Int, LambdaNodeState)] -> ([Int], LambdaNode)
    ineqToCondition' (n:free)           [x]    _ = (free, constToNode n x)
    ineqToCondition' (n1:n2:n3:n4:free) (x:xs) ((symName,symType):symRest) = let (newFree, restNode) = ineqToCondition' free xs symRest
                                                                                 prod = Expression {nodeid = n1,
                                                                                                    occurence = 1, --TODO
                                                                                                    statement = sid,
                                                                                                    targetStatement = TStatementID 0,
                                                                                                    operator = LNOp_Times,
                                                                                                    condition = Nothing,
                                                                                                    firstOperand = Just constant,
                                                                                                    secondOperand = Just var
                                                                                                   }
                                                                                 constant = constToNode n2 x
                                                                                 var = ExprIndBlobConst {
                                                                                                         nodeid = n3,
                                                                                                         occurence = 1, --TODO
                                                                                                         name = symName,
                                                                                                         state = symType,
                                                                                                         statement = sid,
                                                                                                         targetStatement = TStatementID 0
                                                                                                        }
                                                                             in  (newFree,Expression {nodeid = n4,
                                                                                                      occurence = 1, --TODO
                                                                                                      statement = sid,
                                                                                                      targetStatement = TStatementID 0,
                                                                                                      operator = LNOp_Plus,
                                                                                                      condition = Nothing,
                                                                                                      firstOperand = Just prod,
                                                                                                      secondOperand = Just restNode
                                                                                                     })
    ineqToCondition' (n1:n2:n3:n4:free) (x:xs) _ = error "too few symbols!"
    
    constToNode :: Int -> Int -> LambdaNode
    constToNode id value = ExprIndBlobConst {
                                             nodeid = id,
                                             occurence = 1, --TODO
                                             name = value,
                                             state = LNS_Constant,
                                             statement = sid,
                                             targetStatement = TStatementID 0
                                            }


-- |Copies a LambdaNode and returns the list of free LaNodeIDs, the copies of
--  the LambdaNode and tree splitting information
copyNode :: [Int] -> StatementID -> LambdaNode -> ([Int], LambdaNode, LaNodeSplitInfo)
copyNode free newSID node@(Assign {})           = let (newFree, newIndices, lninfo1)   = copyNodes free    newSID (indices node)
                                                      (n:newNewFree, newExp, lninfo2)  = copyNode  newFree newSID (firstExpr node)
                                                  in  (newNewFree, node {nodeid = n, indices = newIndices, firstExpr = newExp, statement = newSID},
                                                       lninfo1 ++ lninfo2 ++ [((newSID, nodeid node), n)])
copyNode free newSID node@(ForLoop {})          = let (newFree, newStride, lninfo1)          = copyMaybeNode free          newSID (loopStride node)
                                                      (newNewFree, newUpper, lninfo2)        = copyMaybeNode newFree       newSID (loopUpperBound node)
                                                      (newNewNewFree, newLower, lninfo3)     = copyMaybeNode newNewFree    newSID (loopLowerBound node)
                                                      (n:newNewNewNewFree, newBody, lninfo4) = copyNodes     newNewNewFree newSID (loopBody node)
                                                  in  (newNewNewNewFree, node {nodeid = n, loopStride = newStride, loopUpperBound = newUpper,
                                                                               loopLowerBound = newLower, loopBody = newBody},
                                                       lninfo1 ++ lninfo2 ++ lninfo3 ++ lninfo4 ++ [((newSID, nodeid node), n)])
copyNode free newSID node@(If {})               = let (newFree, newCond, lninfo1)         = copyMaybeNode free       newSID (condition node)
                                                      (newNewFree, newThen, lninfo2)      = copyNodes     newFree    newSID (thenbranch node)
                                                      (n:newNewNewFree, newElse, lninfo3) = copyNodes     newNewFree newSID (elsebranch node)
                                                  in  (newNewNewFree, node {nodeid = n, condition = newCond, thenbranch = newThen, elsebranch = newElse,
                                                                            statement = newSID},
                                                       lninfo1 ++ lninfo2 ++ lninfo3 ++ [((newSID, nodeid node), n)])
copyNode free newSID node@(WhileLoop {})        = let (newFree, newStride, lninfo1)          = copyMaybeNode free          newSID (loopStride node)
                                                      (newNewFree, newCond, lninfo2)         = copyMaybeNode newFree       newSID (condition node)
                                                      (newNewNewFree, newLower, lninfo3)     = copyMaybeNode newNewFree    newSID (loopLowerBound node)
                                                      (n:newNewNewNewFree, newBody, lninfo4) = copyNodes     newNewNewFree newSID (loopBody node)
                                                  in  (newNewNewNewFree, node {nodeid = n, loopStride = newStride, condition = newCond,
                                                                               loopLowerBound = newLower, loopBody = newBody},
                                                       lninfo1 ++ lninfo2 ++ lninfo3 ++ lninfo4 ++ [((newSID, nodeid node), n)])
copyNode free newSID node@(FuncDecl {})         = let (newFree, newBody, lninfo1)      = copyNodes free    newSID (funcbody node)
                                                      (n:newNewFree, newArgs, lninfo2) = copyNodes newFree newSID (arguments node)
                                                  in  (newNewFree, node {nodeid = n, funcbody = newBody, arguments = newArgs},
                                                       lninfo1 ++ lninfo2 ++ [((newSID, nodeid node), n)])
copyNode free newSID node@(ProcCall {})         = let (n:newFree, newArgs, lninfo) = copyNodes free newSID (arguments node)
                                                  in  (newFree, node {nodeid = n, arguments = newArgs, statement = newSID},
                                                       lninfo ++ [((newSID, nodeid node), n)])
copyNode free newSID node@(ExprRead {})         = let (n:newFree, newInds, lninfo) = copyNodes free newSID (indices node)
                                                  in  (newFree, node {nodeid = n, indices = newInds, statement = newSID},
                                                       lninfo ++ [((newSID, nodeid node), n)])
copyNode free newSID node@(Expression {})       = let (newFree, newCond, lninfo1)           = copyMaybeNode free       newSID (condition node)
                                                      (newNewFree, newFirst, lninfo2)       = copyMaybeNode newFree    newSID (firstOperand node)
                                                      (n:newNewNewFree, newSecond, lninfo3) = copyMaybeNode newNewFree newSID (secondOperand node)
                                                  in  (newNewNewFree, node {nodeid = n, condition = newCond, firstOperand = newFirst, secondOperand = newSecond,
                                                                            statement = newSID},
                                                       lninfo1 ++ lninfo2 ++ lninfo3 ++ [((newSID, nodeid node), n)])
copyNode (n:free) newSID node                   = (free, if hasstatement node then node {nodeid = n, statement=newSID} else node {nodeid = n}, [((newSID, nodeid node), n)])


-- |Copies a LambdaNode and returns the list of free LaNodeIDs, the copy of the
--  LambdaNode and tree splitting information, if a LambdaNode is given.
copyMaybeNode :: [Int] -> StatementID -> Maybe LambdaNode -> ([Int], Maybe LambdaNode, LaNodeSplitInfo)
copyMaybeNode free newSID Nothing     = (free, Nothing, [])
copyMaybeNode free newSID (Just node) = let (newFree, newNode, lninfo) = copyNode free newSID node in (newFree, Just newNode, lninfo)


-- |Copies a list of LambdaNodes and returns the list of free LaNodeIDs, the
--  copies of the LambdaNodes and tree splitting information
copyNodes :: [Int] -> StatementID -> [LambdaNode] -> ([Int], [LambdaNode], LaNodeSplitInfo)
copyNodes free newSID []     = (free, [], [])
copyNodes free newSID (x:xs) = let (newFree, copiedNode, lninfo1)    = copyNode  free newSID x
                                   (newNewFree, copiedTail, lninfo2) = copyNodes free newSID xs
                               in  (newNewFree, copiedNode : copiedTail, lninfo1 ++ lninfo2)


-- |Checks if a LambdaNode is a statement
isStatement :: LambdaNode -> Bool
isStatement Assign {}           = True
isStatement If {}               = True
isStatement Skip {}             = True
isStatement ProcCall {}         = True
isStatement ExprIndBlobConst {} = True
isStatement ExprRead {}         = True
isStatement ExprFuncCall {}     = True
isStatement Expression {}       = True
isStatement _                   = False

