module ISS.Main where

import ISS.DataSplitting.Info
import ISS.DataSplitting.SplitDependences
import ISS.DataSplitting.SplitOther
import ISS.DataSplitting.SplitTree
import ISS.Debug.CodeGenerator
import ISS.Debug.Tests
import ISS.Language.Language
import ISS.Language.LanguageTools
import ISS.Options
import ISS.SplittingAlgorithm.SplitInitial
import ISS.SplittingAlgorithm.SplitPropagation
import ISS.SplittingAlgorithm.ToolsGeneral
import ISS.Util

import LMath.Logic.Domain
import LoopoDS.Access
import LoopoDS.AccessPair
import LoopoDS.Dependence
import LoopoDS.Groups
import LoopoDS.Statement
import ScannerParser.ParseTree
import ScannerParser.ScannerParser
import ScannerParser.SymbolTable

import System

-- |main ;-)
main :: IO()
main = do
    --welcome
    putStrLn (languageText "title")
    
    --read arguments
    options <- getArgs >>= (return . splitArgs)
    
    --either print help or print about or perform splitting
    doIf (about options) ((putStrLn . boxWrap 80 . languageText) "about")
    doIf (usage options) ((putStrLn . languageText) "options")
    doIf (not (usage options || about options)) (performSplitting options) where
    
    performSplitting options = do
          --parse files
          (deps, st, tree, lns, iv, stmts) <- doParseFiles options
          --convert dependencies
          let depsD = map depToDepD deps
          --initialize splitting
          let initStmts = mainInitializeSplit options stmts
          
          --initial splitting
          let firstSplittedStatements = mainSplitInitial options depsD initStmts
          
          --propagating splitting
          let splittedStatements = mainSplitPropagate options depsD firstSplittedStatements
          
          --splitting information distribution:
          
          --split tree
          (splittingInformation, laNodeInformation, newTree) <- doSplitTree options depsD splittedStatements tree st iv
          
          --split index-spaces
          doWriteNewISPCs options splittingInformation
          
          --split dependences (.dep)
          depInformation <- doSplitDependences options deps splittingInformation
          
          --split accesses (.acc)
          doSplitAccesses options laNodeInformation splittingInformation
          
          --split access pairs (.acp)
          doSplitAccessPairs options splittingInformation depInformation laNodeInformation
          
          --split statement groups (.grp)
          doSplitGroups options splittingInformation
          
          --determine new surrounding ifs (.sif)
          doDetermineSurroundingLoops options newTree
          
          --determine new surrounding loops (.slp)
          doDetermineSurroundingIfs options newTree
         
          return ()

-- |parse input files: dependences, parse tree and statement spaces
doParseFiles :: Options -> IO ([Dependence], SymTab, LambdaTree, [LambdaNode], [InfoVector], [StatementD])
doParseFiles options = do
    --read and parse dependences
    putStrLn (languageText "readingDeps")
    let fullDepFileName = createFullFileName (inDirectory options) (inFile (depFileName options))
    maybeParsedDependences <- readDependences fullDepFileName
    --check for errors
    let parsedDependences = throwErrorIfWrong (languageParseError (languageText "dependences")) maybeParsedDependences
    
    --read and parse parse tree
    putStrLn (languageText "readingTree")
    let fullTreeFileName = createFullFileName (inDirectory options) (inFile (treeFileName options))
    treeFileContents <- readFile fullTreeFileName
    let (symtab, parseTree, functions, infoVectors) = parseFile treeFileContents
    
    --read statement spaces
    putStrLn (languageText "readingSpaces")
    let fullIspcFileName = createFullFileName (inDirectory options) (inFile (ispcFileName options))
    maybeISPCs <- readISPCDomain fullIspcFileName
    --check for errors
    let ispcs = throwErrorIfWrong (languageParseError (languageText "statementSpaces")) maybeISPCs
    
    --print debug information
    doIf (showInputs options) (do
        putStrLn (languageText "inputSpaces")
        putStrLn (show ispcs)
        
        putStrLn (languageText "inputDeps")
        putStrLn (show parsedDependences)
        
        putStrLn (languageText "inputPTree")
        putStrLn (generateCode symtab 0 parseTree))
    
    return (parsedDependences, symtab, parseTree, functions, infoVectors, ispcs)

-- |perform initialization (conversion of loopo-statements to statements that are able to store splitting information)
mainInitializeSplit :: Options -> [StatementD] -> [SplittedStatement]
--this is the place for possibly converting "options" (command line parameters) to "function parameters" of the initialize function
mainInitializeSplit options = initializeSplit

-- |perform initial splitting (see documentation in ISS.Splitting.InitialSplit)
mainSplitInitial :: IsDomain d t => Options -> [GDependenceD d t] -> [GSplittedStatement d t] -> [GSplittedStatement d t]
--this is the place for possibly converting "options" (command line parameters) to "function parameters" of the split function
mainSplitInitial options = splitInitial

-- |perform splitting propagation (see documentation in ISS.Splitting.PropagationSplit)
mainSplitPropagate ::IsDomain d t => Options -> [GDependenceD d t] -> [GSplittedStatement d t] -> [GSplittedStatement d t]
--this is the place for possibly converting "options" (command line parameters) to "function parameters" of the split function
mainSplitPropagate options deps = if (propagate options) then splitPropagate (propagationDepth options) deps else id

-- |perform splitting of tree
doSplitTree :: Options -> [DependenceD] -> [SplittedStatement] -> LambdaTree -> SymTab -> [InfoVector] -> IO (SplitInfo, LaNodeSplitInfo, LambdaTree)
doSplitTree options deps splits tree symtab infoVectors = do
    --define a search function on the symtab
    let findSyms typ symtab = (map (\x -> (sym_name x, sym_type x)) (findAllSymbols ((==typ) . sym_type) symtab))
    
    --split the symtab into vars and constants
    let symbolsVars = findSyms LNS_Index symtab
    let symbolsConsts = findSyms LNS_Blob symtab
    let symbols = symbolsVars ++ symbolsConsts
    
    --retrieve the number of parameters of the program
    let params = length' (symbolsConsts)
    
    --split tree in one ore another way
    let (splitTree1, splitInfo1, lninfo1) = splitTreeWithElse symbols params splits tree
    let (splitTree2, splitInfo2, lninfo2) = splitTreeWithoutElse symbols params splits tree
    
    --save splitted tree
    let filename = createFullFileName (outDirectory options) (outFile (treeFileName options))
    writeFile filename (convertFile (symtab, splitTree2, [], infoVectors))
    
    --print debug information
    doIf (showResultParseTree options) (do
            putStrLn (generateCode symtab 0 splitTree1)
            putStrLn (generateCode symtab 0 splitTree2))
    doIf (showResultSplittingTree options) (do
            putStrLn (show splits))
            
    --return information about the tree splitting (necessary for splitting of other file types)
    return (splitInfo2, lninfo2, splitTree2)

-- |Monadic "if"
doIf :: Bool -> IO () -> IO ()
doIf condition operation = if condition then operation else return ()
