module ISS.Options (Options(..), InOutFile(..), splitArgs) where

import ISS.Language.Language

-- |Data structure for command line options
data Options = Options {inDirectory::String, -- ^Directory of input files
                        outDirectory::String, -- ^Directory to store the output files in
                        depFileName::InOutFile, -- ^Input and output filename of the dependences file (.dep)
                        treeFileName::InOutFile, -- ^Input and output filename of the parse tree file
                        ispcFileName::InOutFile, -- ^Input and output filename of the statement spaces file (.ispc)
                        usage::Bool, -- ^Show "usage" information
                        about::Bool, -- ^Show "about" information
                        propagate::Bool, -- ^Propagate
                        propagationDepth::Integer, -- ^Maximum iteration number to find the transitive closure
                        opSplitAccesses::Maybe InOutFile, -- ^Input and output filename of the accesses file (.acc)
                                                          --  or "Nothing" (if no splitting of this file is desired)
                        opSplitAccessPairs::Maybe InOutFile, -- ^Input and output filename of the access pairs file (.acp)
                                                             --  or "Nothing" (if no splitting of this file is desired)
                        opSplitGroups::Maybe InOutFile, -- ^Input and output filename of the groups file (.acc)
                                                        --  or "Nothing" (if no splitting of this file is desired)
                        opSplitSurrLoops::Maybe String, -- ^Output filename of the surrounding loops file (.slp)
                                                        --  or "Nothing" (if no splitting of this file is desired)
                        opSplitSurrIfs::Maybe String, -- ^Output filename of the surroinding ifs file (.sif)
                                                      --  or "Nothing" (if no splitting of this file is desired)
                        showInputs::Bool,             -- ^Debug: Show the read input files
                        showResultParseTree::Bool,    -- ^Debug: Show the resulting parse tree (after split)
                        showResultSplittingTree::Bool -- ^Debug: Show the splitting trees
                       }

-- |Option that tells us to read a given input file and store the splitted information into a given output file
data InOutFile = InOutFile { inFile::String, outFile::String }


-- |Parses the command line options
splitArgs :: [String] -> Options

--default options
splitArgs []                      = let err name = error ("\n" ++ languageText ("missingParam-" ++ name) ++ "\n")
                                    in Options {inDirectory = err "indir", depFileName = err "deps", treeFileName = err "tree",
                                                ispcFileName = err "ispc", outDirectory = err "outdir",
                                                about = False, usage = False,
                                                propagate = False, propagationDepth = 5,
                                                opSplitAccesses = Nothing, opSplitAccessPairs = Nothing, opSplitGroups = Nothing,
                                                opSplitSurrLoops = Nothing, opSplitSurrIfs = Nothing, showInputs = False,
                                                showResultParseTree = False, showResultSplittingTree = False}

--official options
splitArgs ("-debug"      :tail)                = (splitArgs tail) {showInputs = True, showResultParseTree = True, showResultSplittingTree = True}
splitArgs ("-showinputs" :tail)                = (splitArgs tail) {showInputs = True}
splitArgs ("-showresult" :tail)                = (splitArgs tail) {showResultParseTree = True}
splitArgs ("-showsplit"  :tail)                = (splitArgs tail) {showResultSplittingTree = True}
splitArgs ("-propagate"  :tail)                = (splitArgs tail) {propagate = True}
splitArgs ("-propdepth"  :num:tail)            = (splitArgs tail) {propagationDepth = read num}
splitArgs ("-options"    :tail)                = (splitArgs tail) {usage = True}
splitArgs ("-usage"      :tail)                = (splitArgs tail) {usage = True}
splitArgs ("-about"      :tail)                = (splitArgs tail) {about = True}
splitArgs ("-version"    :tail)                = (splitArgs tail) {about = True}
splitArgs ("--version"   :tail)                = (splitArgs tail) {about = True}
splitArgs ("-accesses"   :inFile:outFile:tail) = (splitArgs tail) {opSplitAccesses = Just InOutFile { inFile = inFile, outFile = outFile }}
splitArgs ("-accesspairs":inFile:outFile:tail) = (splitArgs tail) {opSplitAccessPairs = Just InOutFile { inFile = inFile, outFile = outFile }}
splitArgs ("-groups"     :inFile:outFile:tail) = (splitArgs tail) {opSplitGroups = Just InOutFile { inFile = inFile, outFile = outFile }}
splitArgs ("-surrloops"  :outFile:tail)        = (splitArgs tail) {opSplitSurrLoops = Just outFile}
splitArgs ("-surrifs"    :outFile:tail)        = (splitArgs tail) {opSplitSurrIfs = Just outFile}
splitArgs ("-deps"       :inFile:outFile:tail) = (splitArgs tail) {depFileName = InOutFile { inFile = inFile, outFile = outFile }}
splitArgs ("-parsetree"  :inFile:outFile:tail) = (splitArgs tail) {treeFileName = InOutFile { inFile = inFile, outFile = outFile }}
splitArgs ("-ispc"       :inFile:outFile:tail) = (splitArgs tail) {ispcFileName = InOutFile { inFile = inFile, outFile = outFile }}
splitArgs ("-inputdir"   :inDir:tail)          = (splitArgs tail) {inDirectory = inDir}
splitArgs ("-outputdir"  :outDir:tail)         = (splitArgs tail) {outDirectory = outDir}

--complete project
splitArgs ("-project"    :inDir:outDir:name:t) = (splitArgs t) {inDirectory = inDir, outDirectory = outDir,
                                                                opSplitAccesses = Just InOutFile { inFile = name ++ ".out.acc", outFile = name ++ ".out.acc" },
                                                                opSplitAccessPairs = Just InOutFile { inFile = name ++ ".out.acp", outFile = name ++ ".out.acp" },
                                                                opSplitGroups = Just InOutFile { inFile = name ++ ".out.grp", outFile = name ++ ".out.grp" },
                                                                opSplitSurrLoops = Just (name ++ ".out.slp"),
                                                                opSplitSurrIfs = Just (name ++ ".out.sif"),
                                                                depFileName = InOutFile { inFile = name ++ ".out.dep", outFile = name ++ ".out.dep" },
                                                                treeFileName = InOutFile { inFile = name ++ ".out", outFile = name ++ ".out" },
                                                                ispcFileName = InOutFile { inFile = name ++ ".out.ispc", outFile = name ++ ".out.ispc" }
                                                               }

--options, which tell us that the user doesn't know what to do
splitArgs ("-help"     :tail)     = (splitArgs tail) {usage = True, about = True}
splitArgs ("--usage"   :tail)     = (splitArgs tail) {usage = True, about = True}
splitArgs ("-?"        :tail)     = (splitArgs tail) {usage = True, about = True}
splitArgs ("--help"    :tail)     = (splitArgs tail) {usage = True, about = True}
splitArgs ("--?"       :tail)     = (splitArgs tail) {usage = True, about = True}
splitArgs ("/?"        :tail)     = (splitArgs tail) {usage = True, about = True}
splitArgs ("/help"     :tail)     = (splitArgs tail) {usage = True, about = True}
splitArgs ("/usage"    :tail)     = (splitArgs tail) {usage = True, about = True}
splitArgs (('-':option):tail)     = error ("\n" ++ languageBuildText "invalidParameter" [option] ++ "\n")


