module ISS.SplittingAlgorithm.SplitPropagation(splitPropagate) where

import List

import LMath.Logic.Domain
import LMath.Types.Polynomial
import LoopoDS.Dependence
import LoopoDS.Statement
import ISS.SplittingAlgorithm.ToolsGraph
import ISS.SplittingAlgorithm.ToolsGeneral


-- |Splits all given statements according to given dependences with propagation
splitPropagate :: (IsDomain d t) => Integer -> [GDependenceD d t] -> [GSplittedStatement d t] -> [GSplittedStatement d t]
splitPropagate iterations deps stmts = map (splitPropagateStmt iterations graph stmts) stmts where
    --build dependence graph
    graph = Graph {
        gAllEdges = deps,
        gAllNodes = map (fst . originalStatement) stmts,
        gEdges = (\source -> (\target -> filter (\dep -> depDSrcStmtID dep == source && depDDestStmtID dep == target) deps)),
        gSource = depDSrcStmtID,
        gTarget = depDDestStmtID
    }
    
    --find all paths in dependence graph by kleene algorithm
    paths = kleeneAll graph


-- |Splits one single statement according to given dependences with propagation
splitPropagateStmt :: (IsDomain d t) => Integer -> Graph StatementID (GDependenceD d t) -> [GSplittedStatement d t] -> GSplittedStatement d t -> GSplittedStatement d t
splitPropagateStmt iterations graph allStmts stmt = foldl splitPropagateStmt' stmt regexesToStmt where
    stmtID = fst (originalStatement stmt)
    regexesToStmt = [(domain, path) | sourceStmt <- allStmts,
                                      let (sourceStmtID, domain) = originalStatement sourceStmt,
                                      let path = kleene graph sourceStmtID stmtID]
    
    splitPropagateStmt' :: (IsDomain d t) => GSplittedStatement d t -> (d t, Regex (GDependenceD d t)) -> GSplittedStatement d t
    splitPropagateStmt' stmt (domain, regex) = splitStatementWithDomain stmt (propagate iterations domain regex)


-- |Calculates the target of a given domain under a composed relation (regex of
--  dependences)
propagate :: (IsDomain d t) => Integer -> d t -> Regex (GDependenceD d t) -> d t
propagate iterations = propagate' where
    propagate' domain (RegexConcat [x])    = propagate' domain x
    propagate' domain (RegexConcat (x:xs)) = propagate' (propagate' domain x) (RegexConcat xs)
    propagate' domain (RegexOr [x])        = propagate' domain x
    propagate' domain (RegexOr (x:xs))     = propagate' domain x `dUnion` propagate' domain (RegexOr xs)
    propagate' domain (RegexStar x)        = transitiveHull iterations domain x
    propagate' domain (RegexSingle x)      = propagateStep domain x
    propagate' domain RegexEmptyString     = domain
    propagate' domain RegexNoString        = dMakeEmpty domain


-- |Calculates the target of a given domain under the transitive hull of a
--  relation (= regex of dependences) by applying the relation several times to
--  the domain and checking, if the algorithm has reached a fixed point
transitiveHull :: (IsDomain d t) => Integer -> d t -> Regex (GDependenceD d t) -> d t
transitiveHull overallIterations = transitiveHull' overallIterations where
    transitiveHull' 0 domain paths = domain
    transitiveHull' i domain paths = if domain `dEquals` propDomain then domain else transitiveHull' (i-1) propDomain paths where
        propDomain = propagate overallIterations domain paths


-- |Calculates one step (one dependence in the path) of the propagation
propagateStep :: (IsDomain d t) => d t -> GDependenceD d t -> d t
propagateStep domain dependence = (dInverseImage hTrafo sourceSpace) `dIntersect` targetSpace where
    hTrafo = depDHTrafo dependence
    sourceSpace = depDSrcDomain dependence `dIntersect` domain
    targetSpace = depDDestDomain dependence




