module ISS.SplittingAlgorithm.ToolsGeneral(initializeSplit, SplittedStatement, Split, GSplittedStatement(..), GSplit(..),
                                           splitStatementWithDomain, depToDepD, StatementDomain, IsDomain(..),
                                           polyIneqToIneq, splitToIneqs, splitToDomains) where

import LMath.Logic.Domain
import LMath.Types
import LMath.Types.AffineFunction
import LMath.Types.IneqSystem
import LMath.Types.Polynomial
import LMath.Types.Matrix
import LoopoDS.Dependence
import LoopoDS.Statement

import Data.Ratio
import List

import ISS.Language.LanguageTools
import ISS.Util

-- |The usual type of Domain
type StatementDomain = Domain (Polynomial (Ratio Integer))

-- |Generic statement with splitting information
data GSplittedStatement d t = SplittedStatement {originalStatement::GStatementD d t,
                                                 directDependenceTargets::[d t],
                                                 splitting::GSplit d t}
                              deriving Show

-- |Statement with splitting information
type SplittedStatement = GSplittedStatement Domain (Polynomial (Ratio Integer))

instance Show SplittedStatement where
  show = show . splitting

instance Show Split where
  show split@(Split {})   = "if:\n" ++ indent 4 (show (ifBranch split)) ++ "\nelse:\n" ++ indent 4 (show (elseBranch split))
  show split@(NoSplit {}) = show (completeSpace split)


-- |A binary tree representing hierarchical splitting
data GSplit d t =
                  -- | Inner node
                  Split {completeSpace::d t,
                         ifBranch::GSplit d t,
                         elseBranch::GSplit d t}
                  -- | Leaf (not splitted any more)
                  | NoSplit {completeSpace::d t}
                    deriving Show

-- |The binary tree for the usual type of Domain
type Split = GSplit Domain (Polynomial (Ratio Integer))

-- |Class of types, which can be used as statement-domains in ISS
class IsDomain d t where
    dSimplify :: d t -> d t
    dSimplify = id
    dIntersect :: d t -> d t -> d t
    dUnion :: d t -> d t -> d t
    dComplement :: d t -> d t -> d t
    dInverseImage :: GAffineFunction t -> d t -> d t
    dIsEmpty :: d t -> Bool
    dEquals :: d t -> d t -> Bool
    dMakeEmpty :: d t -> d t
    dMakeEmpty domain = dComplement domain domain -- a very inefficient way of making a domain empty (but instances of this class may define a more efficient method)

-- |LoopoDS.Logic.Domain-implementation of "IsDomain"
instance (DomainCanBeSimplified t, BasicDomainElement t, PolyConvertableDomain t) => IsDomain Domain t where
    dSimplify = simplifyD'
    dIntersect = intersectD
    dUnion = unionD
    dComplement = complementD'
    dInverseImage = inverseImageD
    dIsEmpty = isAlwaysEmptyD'
    dEquals = equivalentD'


-- |Converts a list of DStatements to GSplittedStatement trees
initializeSplit :: [StatementD] -> [SplittedStatement]
initializeSplit = map initializeStatement where
                  initializeStatement stmt@(_, stmtSpace) = SplittedStatement {originalStatement = stmt,
                                                                               directDependenceTargets = [],
                                                                               splitting = NoSplit {completeSpace = stmtSpace}}

-- |Splits a statement further into two parts: the one that lies in the given domain and the one that doesn't
splitStatementWithDomain :: IsDomain d t => GSplittedStatement d t -> d t -> GSplittedStatement d t
splitStatementWithDomain statement polytope = statement {splitting = splitSplit (splitting statement) polytope} where
                         splitSplit :: IsDomain d t => GSplit d t -> d t -> GSplit d t
                         splitSplit split@(NoSplit {}) domain = if dIsEmpty ifDomain || dIsEmpty elseDomain then
                                                                  split
                                                                else
                                                                  Split {completeSpace = complete,
                                                                         ifBranch = NoSplit {completeSpace = ifDomain},
                                                                         elseBranch = NoSplit {completeSpace = elseDomain}
                                                                        } where
                                                                  complete = dSimplify (completeSpace split)
                                                                  ifDomain = dSimplify (complete `dIntersect` domain)
                                                                  elseDomain = dSimplify (complete `dIntersect` (dSimplify (dComplement complete domain)))
                         splitSplit split@(Split {ifBranch = ifB, elseBranch = elseB}) polytope = split {ifBranch = splitSplit ifB polytope,
                                                                                                         elseBranch = splitSplit elseB polytope}


-- |Converts a "normal" Dependence to a DependenceD (with Domains instead of IneqSystems for SrcDomain and DestDomain)
depToDepD :: Dependence -> DependenceD
depToDepD dep = DependenceD {depDID = depID dep,
                             depDType = depType dep,
                             depDSrcDomain  = ineqToDomain2 Nothing (nVars (toAffineExp (depSrcDomain dep)), nParams (toAffineExp (depSrcDomain dep))) (depSrcDomain dep),
                             depDDestDomain = ineqToDomain2 Nothing (nVars (toAffineExp (depSrcDomain dep)), nParams (toAffineExp (depDestDomain dep))) (depDestDomain dep),
                             depDSrcStmtID = depSrcStmtID dep,
                             depDDestStmtID = depDestStmtID dep,
                             depDHTrafo = affParametersToPolynomials (depHTrafo dep),
                             depDContext = depContext dep
                            }


-- |Converts an IneqSystem with linear polynomials over parameters in the homogenous column to an IneqSystem, in which
--  every parameter has got its own column
polyIneqToIneq :: Num a => Integer -> GIneqSystem (Polynomial a) -> GIneqSystem a
polyIneqToIneq nParams = fromAffineExp . polyAffineExpToAffineExp nParams . toAffineExp


-- |Converts an AffineExp with linear polynomials over parameters in the homogenous column to an AffineExp, in which
--  every parameter has got its own column
polyAffineExpToAffineExp :: Num a => Integer -> GAffineExp (Polynomial a) -> GAffineExp a
polyAffineExpToAffineExp nParams exp = let vars = nVars exp
                                       in  affineExp ((vars, nParams), polyMatrixToMatrix vars nParams (getMatrix exp))


-- |Converts a matrix with linear polynomials over parameters in the homogenous column to a matrix, in which
--  every parameter has got its own column
polyMatrixToMatrix :: Num a => Integer -> Integer -> GMatrix (Polynomial a) -> GMatrix a
polyMatrixToMatrix nVars nParams polyMatrix = let emptyMatrix = zeroMatrix lines (nVars + nParams + 1)
                                                  lines = nRows polyMatrix
                                                  updatesLeft = [(row,col,constantCoeff (entry polyMatrix row col))
                                                                 | row <- [0..fromIntegral (lines-1)], col <- [0..fromIntegral (nVars-1)]]
                                                  updatesRight = concat [polyToUpdates (fromIntegral row) poly
                                                                         | row <- [0..lines-1],
                                                                           let poly = entry polyMatrix (fromIntegral row) (fromIntegral nVars)]
                                                  polyToUpdates :: Num a => Integer -> Polynomial a -> [(Integer, Integer, a)]
                                                  polyToUpdates row poly = let Just coefficients = linearCoeffs poly
                                                                           in  coefficientsToUpdates row (fromIntegral nVars) coefficients
                                                  coefficientsToUpdates :: Integer -> Integer -> [a] -> [(Integer, Integer, a)]
                                                  coefficientsToUpdates _   _   []     = []
                                                  coefficientsToUpdates row col (x:xs) = (row, col, x):coefficientsToUpdates row (col+1) xs
                                              in  setEntriesST emptyMatrix (updatesLeft ++ updatesRight)

-- |Converts a linear multivariate polynomial to a list of the coeffizients in front of the polynomial variables
getCoefficients :: Num a => Polynomial a -> Integer -> [a]
getCoefficients poly nCoeffs = let Just coeffs = linearCoeffs poly
                                   zeroList = 0:zeroList
                                   coeffList = take' nCoeffs (init coeffs)
                                   fillUpList = take' (nCoeffs - (length' coeffList)) zeroList
                                   constList = [constantCoeff poly]
                               in  coeffList ++ fillUpList ++ constList

-- |Converts the hierarchical splitting tree to a list of IneqSystems
splitToIneqs :: Integer -> Split -> [IneqSystem]
splitToIneqs params = (map (polyIneqToIneq params)) . concat . (map (disjointUnionD Nothing)) . splitToDomains

-- |Converts the hierarchical splitting tree to a list of all domains at leafs
splitToDomains :: GSplit d t -> [d t]
splitToDomains s@(NoSplit {}) = [completeSpace s]
splitToDomains s@(Split {})   = splitToDomains (ifBranch s) ++ splitToDomains (elseBranch s)

