module ISS.SplittingAlgorithm.ToolsGraph(Regex(..), kleene, kleeneAll, Graph(..), cleanPaths) where

import List

import LoopoDS.Dependence
import LoopoDS.Statement
import ISS.DataSplitting.Info

-- |Regular expressions
data Regex a = RegexConcat [Regex a] | RegexOr [Regex a] | RegexStar (Regex a) | RegexSingle a | RegexNoString | RegexEmptyString

-- |Generic graph
data Graph node edge = Graph { gAllEdges :: [edge],
                               gAllNodes :: [node],
                               gEdges :: node -> node -> [edge],
                               gSource :: edge -> node,
                               gTarget :: edge -> node
                             }


-- |Output for regular expressions
instance (Show a) => Show (Regex a) where
    show = showPath where
        showPath (RegexConcat xs) = (concat . intersperse " " . map showPath) xs
        showPath (RegexOr [])     = "none"
        showPath (RegexOr [x])    = showPath x
        showPath (RegexOr (x:xs)) = "(" ++ showPath x ++ " | " ++ showPath (RegexOr xs) ++ ")"
        showPath (RegexStar x)    = "(" ++ showPath x ++ ")*"
        showPath (RegexSingle x)  = show x
        showPath RegexNoString    = "none"
        showPath RegexEmptyString = "empty"


-- |Simplify regular expressions (trivially)
cleanPaths :: Regex edge -> Regex edge
cleanPaths (RegexStar p) = case cleanPaths p of
    RegexNoString    -> RegexEmptyString
    RegexEmptyString -> RegexEmptyString
    p                -> RegexStar p
cleanPaths (RegexOr ps)  = case (filter notNoPath . map cleanPaths) ps of
    []  -> RegexNoString
    [p] -> p
    ps  -> RegexOr ps
cleanPaths (RegexConcat ps) = if not $ all notNoPath preCleaned then RegexNoString else cleaned where
    preCleaned = (filter notEmptyPath . map cleanPaths) ps
    cleaned = case preCleaned of
        []  -> RegexEmptyString
        [p] -> p
        ps  -> RegexConcat ps
cleanPaths p = p


-- |Checks, if a regular expression doesn't consist only of one "NoPath" expression
--  (used for simplification)
notNoPath :: Regex a -> Bool
notNoPath RegexNoString = False
notNoPath _             = True


-- |Checks, if a regular expression doesn't consist only of one "EmptyPath" expression
--  (used for simplification)
notEmptyPath :: Regex a -> Bool
notEmptyPath RegexEmptyString = False
notEmptyPath _                = True


-- |Apply the "Kleene Algorithm" to find all possible paths from any node to
--  any other node within a graph. The result is given as regular expression.
kleeneAll :: (Enum node) => Graph node edge -> [Regex edge]
kleeneAll graph = [cleanPaths $ kleene graph start end | start <- gAllNodes graph, end <- gAllNodes graph]


-- |Apply the "Kleene Algorithm" to find all possible paths from one node to
--  another within a graph. The result is given as regular expression.
kleene :: (Enum node) => Graph node edge -> node -> node -> Regex edge
kleene graph s d = cleanPaths $ kleene' maxNodeNum s d where
    maxNodeNum = maximum [fromEnum n | n <- gAllNodes graph]
    minNodeNum = minimum [fromEnum n | n <- gAllNodes graph]
    
    kleene' k from to = if k < minNodeNum then
                            RegexOr ([RegexSingle edge | edge <- gEdges graph from to] ++ if fromEnum from == fromEnum to then [RegexEmptyString] else [])
                        else
                            RegexOr [kleene' (k-1) from to, RegexConcat [kleene' (k-1) from (toEnum k),
                                                                       RegexStar (kleene' (k-1) (toEnum k) (toEnum k)),
                                                                       kleene' (k-1) (toEnum k) to]]

