-- -*- haskell -*-
{
module InsertIfs.Parser(parseSource) where

import Control.Monad.Error
import Control.Monad.State

import InsertIfs.DS
import InsertIfs.Scanner
import InsertIfs.ParserDefs
}

%name parse
%monad { P }
%tokentype { (Token,SourcePos) }

%token
    comment                { (TComment $$, _) }
    '{'                    { (TBraceBegin,$$) }
    '}'                    { (TBraceEnd,$$) }
    '('                    { (TParenBegin,$$) }
    ')'                    { (TParenEnd,$$) }
    '='                    { (TAssign,$$) }
    ';'                    { (TSemi,$$) }
    ','                    { (TComma,$$) }
    if                     { (TIf,$$) }
    for                    { (TFor,$$) }
    id                     { (TID $$, _) }
    single_code_char       { (TUnknownChar $$, _) }
    whitespaces            { (TWhitespaces $$, _) }

%%

Program :: { String }
        :  stmts { $1 }

stmts :: { String }
      : single_stmt stmts { $1 ++ $2 }
      | single_stmt { $1 }

single_stmt :: { String }
            : comment     { $1 }
	    | whitespaces { $1 }
            | for_block   { $1 }
            | if_block    { $1 }
	    | cloog_stmt  { $1 }
            | assignment  { $1 }

for_block :: { String }
          : for ws '(' ws id ws '=' code ';' code ';' code ')' ws '{' stmts '}'
	      {% do { s <- get
		    ; let rpID  = psRPid s
		    ; let ifCon = psIfConstraint s
                    ; let useComm = psUseCommunication s
		    ; return $
                      if useComm
                      then concat [ "for", $2, "(", $4, $5, $6, "=", $8, ";",
			            $10, ";", $12, ")", $14, "{",
			            if $5 == rpID then (ifCon ++ " { ") else "",
			            $16,
			            if $5 == rpID then "}" else "",
			            "}"
			          ]
                      else concat [ if $5 == rpID then (ifCon ++ "\n") else "",
                                    "for", $2, "(", $4, $5, $6, "=", $8, ";",
			            $10, ";", $12, ")", $14, "{",
			            $16,
                                    "}"
			          ]
		    }
	      }

if_block :: { String }
         : if ws '(' code ')' ws '{' stmts '}'
             { concat [ "if", $2, "(", $4, ")", $6, "{", $8, "}"] }


cloog_stmt :: { String }
           : id ws '(' code ')' ws ';'
               { concat [$1, $2, "(", $4, ")", $6, ";"] }

assignment :: { String }
	   : assign_block  { $1 }
           | simple_assign { $1 }

assign_block :: { String }
             : '{' ws simple_assign ws '}' { concat ["{", $2, $3, $4,"}"] }

simple_assign :: { String }
              : id ws '=' code ';' stmts
		  {% do { s <- get
			; let rpID  = psRPid s
			; let ifCon = psIfConstraint s
                        ; let useComm = psUseCommunication s
			; return $
                          if useComm
                          then concat [ $1, $2, "=", $4, ";",
				        if $1 == rpID then ifCon ++ " { " else "",
				        $6,
				        if $1 == rpID then "}" else ""
				      ]
                          else concat [ $1, $2, "=", $4, ";", $6 ]
			}
		  }


ws :: { String }
   : {- empty -} { "" }
   | whitespaces { $1 }

code :: { String }
     : single_code      { $1 }
     | single_code code { $1 ++ $2 }

-- helper production for production "code"
single_code :: { String }
            : single_code_char { [$1] }
            | whitespaces      { $1   }
            | id               { $1   }
            | '='              { "="  }
            | '(' code ')'     { "(" ++ $2 ++ ")" }
            | ','              { ","  }

{

--
-- 'happyError' is called, when the parser encounters an error.
--
happyError :: [(Token,SourcePos)] -> P a
happyError ((t,pos):_) =
    error ("happyError: parse error at " ++ showPos pos ++ ", token " ++ show t)
happyError [] = error "happyError: unexpected end of input"

--
-- Run the parser and do some processing on the result
-- to obtain a description of the input as a 'Program' value.
--
parseSource :: String      -- ^ the input file
            -> ParserState -- ^ the initial state
	    -> String      -- ^ the resulting file
parseSource str initState =
    evalState (parse (scan str)) initState
}
