module LGen ( main )
where

import Control.Exception
import System.Console.GetOpt
import System.Environment
import System.IO
import Text.ParserCombinators.Parsec

import Debug.Trace

import LMath
import LMath.Algo.LGen


data LGenOpt = Level Int
             | OutputFile FilePath
             | AlgoVariant LGenAlgoVariant
             | Progress
             | Debug
             | SimplifyIFs
             | SimplifyOuterLoops
             | SimplifyAll
             | ConvexUnion
             | Simplex
             | UseCAD
             | CADNoScatter
             | UseGEO
               deriving (Eq, Show)

options :: [OptDescr LGenOpt]
options = [Option ['l'] ["level"] (ReqArg (Level . read) "N") "pull IFs N levels out"
          ,Option []    ["simplify-ifs"] (NoArg SimplifyIFs) "simplify if statements"
          ,Option []    ["simplify-outer-loops"] (NoArg SimplifyOuterLoops) "simplify outer for loops"
          ,Option []    ["simplify"] (NoArg SimplifyAll) "turn on all --simplify-* options"
          ,Option []    ["simplex"] (NoArg Simplex) "use Simplex to solve LP questions"
          ,Option []    ["convex"] (NoArg ConvexUnion) "use convex description for outer loops"
          ,Option []    ["original"] (NoArg (AlgoVariant LGenOriginal)) "use original (slow but correct) algorithm"
          ,Option []    ["slightly-improved"] (NoArg (AlgoVariant LGenSlightlyImproved)) "use a slightly improved correct algorithm (default)"
          ,Option []    ["experimental"] (NoArg (AlgoVariant LGenExperimental)) "use experimental algorithm"
          ,Option []    ["progress"] (NoArg Progress) "show progress"
          ,Option []    ["debug"]    (NoArg Debug)    "show debugging output"
          ,Option []    ["cad"]      (NoArg UseCAD)   "use CAD algorithm (and ignore other options!)"
          ,Option []    ["cad-no-scatter"]      (NoArg CADNoScatter)   "ignore scatter functions for CAD"
          ,Option []    ["geo"]      (NoArg UseGEO)   "use Geoform (and ignore other options!)"
          ,Option ['o'] ["output"] (ReqArg OutputFile "OUTPUT") "write output to OUTPUT"
          ]

main :: IO ()
main = do args <- getArgs
          mainA args

mainA :: [String] -> IO()
mainA args =
          case getOpt Permute options args of
            (opts,[infile],[]  ) -> run infile opts
--            (opts,_,[]) ->
--                hPutStrLn stderr "Must give exactly one input file."
            (_,_,errs) ->
                do binary <- getProgName
                   let header = "Usage: " ++ binary ++
                                " [OPTION...] file.cloogInput"
                   hPutStrLn stderr (concat errs ++ usageInfo header options)

run :: FilePath -> [LGenOpt] -> IO ()
run input opts =
    do p <- parseFromFile parseLGenInput input
       case p of
           Right (lgp,lgoo)
{-
              | useCAD    ->
                  handleToolIOError $
                  do outputFn $
                       code2text nP (lgooParamNames lgoo ++ lgooVarNames lgoo) $
                       codeSimpl $
                       cadgen ctx tispcs
-}
              | useCAD    ->
                  handleToolIOError $
                  do -- print (lgp,lgoo)
                     -- print fullctx
                     -- print fulltispcs
                     outputFn $
                       code2text nP (lgooParamNames lgoo ++ lgooVarNames lgoo ++ lgooIterNames lgoo) $
                       codeSimpl $
                       cadgen fullctx fulltispcs
              | useGEO    ->
                  handleToolIOError $
                  do outputFn $
                       code2text nP (lgooParamNames lgoo ++ lgooVarNames lgoo) $
                       codeSimpl $
                       geogen ctx tispcs                  
              | otherwise ->
               let c  = runLGen lgo lgp lgoo !! dims
               in
               outputFn c
              where
              xs = lgpStatements lgp
              (qepcadF,_) = qfFormula2qepcad $ disjunctionQf tispcs
              ctx = fmap (revVarsAndParams nVtarget) $ ineqSystemToFormula $ insertZeroVars (fmap constPoly $ lgpContext lgp) 0 nVtarget
              tispcs = [disjunctionQf [fmap (revVarsAndParams nVtarget) $ ineqSystemToFormula $ fmap constPoly $ image (fmap fromIntegral af) is | (is,af) <- x] | x <- xs]

              (fullctx,fulltispcs) = (fmap (revVarsAndParams nV') $ ineqSystemToFormula $ insertZeroVars (fmap constPoly $ lgpContext lgp) 0 nV', [disjunctionQf [fmap (revVarsAndParams nV') $ ineqSystemToFormula $ fmap constPoly is | is <- iss] | iss <- fullIspcs noscatter lgp])
                  where
                    nV' = if noscatter then nV else nVfull
                    noscatter = CADNoScatter `elem` opts          

              nP = nParams $ head [is | x <- xs, (is,_) <- x]
              nVtarget = targetDimIH $ head [af | x <- xs, (_,af) <- x]
              nVfull = nVtarget + nV
              nV = maximum [nVars is | stmt <- lgpStatements lgp, (is,af) <- stmt]
           Left err ->
               hPutStrLn stderr ("Parse error: " ++ show err)
    where
    dims = head ([l | Level l <- opts] ++ [0])
    useCAD  = UseCAD `elem` opts
    useGEO  = UseGEO `elem` opts
    outputFn = head ([writeFile out | OutputFile out <- opts] ++ [putStrLn])
    algo = head ([a | AlgoVariant a <- opts] ++ [LGenSlightlyImproved])
    lgo = lgenOptionsStd { lgoAlgoVariant = algo, 
                           lgoLpVariant = if Simplex `elem` opts then LGenSimplex else LGenQE,
                           lgoProgressToStderr = Progress `elem` opts,
                           lgoDebug = Debug `elem` opts,
                           lgoSimplifyIFs = any (`elem` opts) [SimplifyIFs,SimplifyAll],
                           lgoSimplifyOuterLoops =
                               any (`elem` opts) [SimplifyOuterLoops,SimplifyAll],
                           lgoConvexUnion = ConvexUnion `elem` opts
                         }

fullIspcs :: Bool -> LGenParams -> [[IneqSystem]]
fullIspcs ignoreScatter lgp =
    if ignoreScatter then
        [[sys | (sys,af) <- stmt] | stmt <- stmts]
    else
        [[combine (nViter+nVtarget) sys (fmap fromIntegral af) | (sys,af) <- stmt] | stmt <- stmts]
    where
    nVtarget = targetDimIH $ head [af | stmt <- stmts, (_,af) <- stmt]
    nViter   = maximum [nVars sys | stmt <- stmts, (sys,_) <- stmt]
    stmts = lgpStatements lgp

combine :: Num a => Integer -> GIneqSystem a -> GAffineFunction a -> GIneqSystem a
combine n sys af =
    ineqSystem ((n,nParams sys), sys2)
    where
    sys2 = multijoin1 $
           map multiconc1 $
           [[-u, afV, z2a, afP, afC]
           ,[ u,-afV, z2a,-afP,-afC]
           ,[ z, isV, z2b, isP, isC]
           ]
    (afV,afP,afC) = partsVPCih af
    (isV,isP,isC) = partsVPC   sys
    tD = targetDimIH af
    u = unitMatrix tD tD
    z = zeroMatrix (nRows sys) tD
    z2a = zeroMatrix tD (n-tD-nVars sys)
    z2b = zeroMatrix (nRows sys) (n-tD-nVars sys)

loadFormulas :: FilePath -> IO [QfFormula (Polynomial Rational)]
loadFormulas input =
    do p <- parseFromFile parseLGenInput input
       case p of
           Right (lgp,lgoo) -> return tispcs
               where
               xs = lgpStatements lgp
               tispcs = [disjunctionQf [fmap (revVarsAndParams nVtarget) $ ineqSystemToFormula $ fmap constPoly $ image (fmap fromIntegral af) is | (is,af) <- x] | x <- xs]
               nVtarget = targetDimIH $ head [af | x <- xs, (_,af) <- x]
           Left err ->
               error ("Parse error: " ++ show err)

loadFormulasFull :: FilePath -> IO [QfFormula (Polynomial Rational)]
loadFormulasFull input =
    do p <- parseFromFile parseLGenInput input
       case p of
           Right (lgp,lgoo) -> return fulltispcs
               where
               xs = lgpStatements lgp
               fulltispcs = [disjunctionQf [fmap (revVarsAndParams nVfull) $ ineqSystemToFormula $ fmap constPoly is | is <- iss] | iss <- fullIspcs False lgp]
               nVtarget = targetDimIH $ head [af | x <- xs, (_,af) <- x]
               nVfull = nVtarget + maximum [nVars is | stmt <- lgpStatements lgp, (is,af) <- stmt]
           Left err ->
               error ("Parse error: " ++ show err)

revVarsAndParams :: Num a => Integer -> Polynomial a -> Polynomial a
revVarsAndParams nV p =
    createPolyFromList $ [(co,f es) | (co,es) <- polyToList p]
    where
    f es = reverse es1 ++ reverse es2
        where
        (es1,es2) = splitAt (fromIntegral nV) es

handleToolIOError :: IO a -> IO ()
handleToolIOError action =
    catchDyn (action >> return ())
             (\(tio :: ToolIOError) ->
                 do hPutStrLn stderr ("Tool IO error: " ++ show tio)
             )
