-- |
-- This module re-exports all math modules of /LooPo/.
--
-- Author: Armin Groesslinger
--
-- Version: $Id: LMath.hs 4156 2009-07-07 17:36:39Z groessli $
--
module LMath
    ( module Data.Ratio,
      module LMath.Algo,
      module LMath.Logic,
      module LMath.Numbers,
      module LMath.Tools,
      module LMath.Types,
      subsets,
      inversePermutation,

      LRV(..), dualOfSystem,

      -- * Experimental functions
      kernelOrdFrac, inverseMatrixOrdFrac
    )
where

import Data.List
import Data.Ratio
import Data.Maybe
import Control.Exception as CE

import CppInterface
import LMath.Algo
import LMath.Logic
import LMath.Numbers
import LMath.Tools
import LMath.Types


-------------------------
-- DTreeGen related stuff
-------------------------


leastEntry :: (Num a, Num cond)
           => GVector a
           -> DTreeGen a cond a
leastEntry v = le $ vectorEntries v
    where
    le [] = dLeaf 0
    le [x] = dLeaf x
    le (x:y:xs) =
	geCond (x-y) (le (y:xs))
	             (le (x:xs))

leastEntryOF :: (Ord a, Fractional a)
	     => GVector a -> a
leastEntryOF v = case makeDTree numOrdDTreeContext $ leastEntry v of
		   DLeaf res -> res
		   _         -> error "numOrdDTreeGen should never return a tree with a case distinction!"

leastEntryFP :: Integral a
             => [QfFormula (Polynomial (Ratio a))]
             -> GVector (FracPoly (Ratio a))
             -> DTree (Polynomial (Ratio a)) (FracPoly (Ratio a))
leastEntryFP ctx v = makeDTree (fracPolyDTreeContext ctx) $ leastEntry v 

-- Tests:
testLeastEntryOF = leastEntryOF $ vector [4,7,1,2,3]
testLeastEntryFP1 = leastEntryFP [QfAt (AtNeq p)] $ vector [p%%1,1%%p]
    where
    p = indeterminate 1 0
testLeastEntryFP2 = leastEntryFP [QfAt (AtGe (p-1))] $ vector [p%%1,1%%p]
    where
    p = indeterminate 1 0

-------------


data LRV = LRV { lrvLines    :: [Vector],
		 lrvRays     :: [Vector],
		 lrvVertices :: [Vector]
	       }
	   deriving (Eq, Show)

foreign import ccall eqsIneqs2rays_c :: Ptr EqSystem -> Ptr IneqSystem ->
				  IO (Ptr Matrix)

eqsIneqs2rays :: EqSystem -> IneqSystem -> Matrix
eqsIneqs2rays = unsafeCppiConvert2 eqsIneqs2rays_c


dualOfIS :: IneqSystem -> LRV
dualOfIS is = dualOfSystem eqs is
    where
    eqs = eqSystem ((nVars is,nParams is), zeroMatrix 0 (nColumns is))

dualOfSystem :: EqSystem -> IneqSystem -> LRV
dualOfSystem eqs ineqs =
    (\(ls,rs,vs) -> LRV { lrvLines = ls, lrvRays = rs, lrvVertices = vs }) $
    foldl extract ([],[],[]) (rows' $ eqsIneqs2rays eqs ineqs)
    where
    extract (ls,rs,vs) t =
	if s == 0
	   then (vector coeffs:ls, rs, vs)
	   else if k == 0
		   then (ls, vector coeffs:rs, vs)
		   else (ls, rs, vector (map (/k) coeffs) : vs)
	where
	s = head t
	k = last t
	coeffs = take (length t-2) $ drop 1 t


-- calculate the dual of a polyhedron with `n' rays (for n>=1)
testWithNrays :: Integer -> LRV
testWithNrays n = dualOfIS sys
    where
    sys = ineqSystem ((3,0), matrix ([mk i | i <- [1..(n-1)]] ++
				     [[0,0,1,0]]))
    mk i = [(-i)%1, 1, (i*i)%1, 0]

eq = eqSystem ((3,0), matrix [[0,0,1,-1%1]])

is = ineqSystem ((3,0), matrix [[1,0,0,0],
				[0,1,0,0],
				[-2%3,-1,0,2],
				[0,0,1,1]])

subsets :: [a] -> [[a]]
subsets []     = [[]]
subsets (x:xs) = subsets xs ++ [x : ys | ys <- subsets xs]

-- |
-- Computes the inverse of a permutation given as a list of numbers, 
-- where each number represents the index that is permutated to the 
-- corresponding position in the list.
inversePermutation :: Num a => [a] -> [a]
inversePermutation perm = [ myIndex i perm | i <- [0 .. (length perm)-1] ]
    where
    myIndex i xs = (fromIntegral $ fromJust $ elemIndex (fromIntegral i) xs)


-- |
-- Transform a matrix into reduced row echelon form.
-- The the leaves of the resulting DTree carry a pair @(m,xs)@ where
-- @m@ the matrix in reduced row echelon an @xs@ is a list of (row,column)
-- pairs denoting the positions of the pivots (in left-to-right order).
--
reducedRowEchelon :: Fractional a => GMatrix a -> DTreeGen a a (GMatrix a, [(Integer,Integer)])
reducedRowEchelon matr = elim 0 0 matr []
    where
    elim row col matr zs
	| col == nColumns matr  =  dLeaf (matr, reverse zs)
	| row == nRows    matr  =  dLeaf (matr, reverse zs)
	| otherwise           =
	    eqCond x
		   (findNZcol row col matr zs)
		   (elim (row+1) (col+1) (mkCol0 row col matr) ((row,col):zs))
	    where
	    x = entry matr row col

    mkCol0 row col matr =
	setEntries matr ((row,col,1) : [(row,c,entry matr row c / s) | c <- [col+1..maxColumn matr]] ++
			 [(r,c, entry matr r c - entry matr r col * entry matr row c / s) | r <- [0..maxRow matr], r/=row, c<-[col+1..maxColumn matr]] ++
			 [(r,col,0) | r <- [0..maxRow matr], r/=row])
	    where
	    s = entry matr row col

    findNZcol row col matr zs = findNZ row
	where
	findNZ r
	    | r == nRows matr  =  elim row (col+1) matr zs
	    | otherwise =
		eqCond (entry matr r col)
		       (findNZ (r+1))
		       (elim (row+1) (col+1) (mkCol0 row col matr') ((row,col):zs))
	    where
	    matr' = setEntries matr [(rr,c,entry matr (r+row-rr) c) | rr <- [row,r], c <- [0..maxColumn matr]]

inverseM :: Fractional a => GMatrix a -> DTreeGen a a (Maybe (GMatrix a))
inverseM matr =
    assert (nColumns matr == s) $
    compDTreeG (\(m,zs) ->
	          dLeaf $ if genericLength zs == s then Just (subMatrixVer m s (s+s-1)) else Nothing)
	       (reducedRowEchelon (matr `conc` unitMatrix s s))
    where
    s = nRows matr

inverseMatrixOrdFrac :: (Fractional a, Ord a) => GMatrix a -> GMatrix a
inverseMatrixOrdFrac matr =
    case makeDTree numOrdDTreeContext (inverseM matr) of
      DLeaf Nothing    -> error "inverseMatrixOrdFrac: singular matrix"
      DLeaf (Just inv) -> inv
      _                -> error "inverseMatrixOrdFrac: result is not (DLeaf _); should never happen!"


kernelM :: Fractional a => GMatrix a -> DTreeGen a a [GVector a]
kernelM matr =
    mapDTreeG
        (\(m,zs) ->
	    let stufen = genericLength zs
		ext = nColumns m - stufen
		ks = [c | (_,c) <- zs]
		reord = zip [0..nC-1] (ks ++ ([0..nC-1] \\ ks))
		m' = subMatrix (reorderColumns m [c | (_,c) <- reord]) 0 stufen (stufen-1) (nC-1) `join`
		     (negate $ unitMatrix ext ext)
		mm = columnVectors $
		     reorderRows m' [r | (r,c) <- sortBy (\(_,a) (_,b) -> compare a b) reord]
		in
	    mm
	)
	(reducedRowEchelon matr)
    where
    nC = nColumns matr
    nR = nRows matr
    ext = max (nC-nR) 0

    swapRows m k n
	| k == n     =  m
	| otherwise  =  upper ++ g:(middle ++ (f:lower))
	where
	(a,b) = if k < n then (k,n-k-1) else (n,k-n-1)
	(upper, f:rest) = genericSplitAt a m
	(middle, g:lower) = genericSplitAt b rest

kernelOrdFrac :: (Fractional a, Ord a) => GMatrix a -> [GVector a]
kernelOrdFrac matr =
    case makeDTree numOrdDTreeContext (kernelM matr) of
      DLeaf kern -> kern
      _          -> error "kernelMatrixOrdFrac: result is not (DLeaf _); should never happen!"
