---
-- For calculating the image under an affine mapping.
-- @author Armin Gr&ouml;&szlig;linger
-- @version $Id: AffineImage.hs 4184 2009-08-13 15:59:24Z groessli $
--
module LMath.Algo.AffineImage
    ( image, inverseImage,
      imageG,
      inverseAffine, pseudoInverse, fctComposition
    )
where

import Data.List (genericDrop, genericTake)
import Data.Maybe
import Data.Ratio

import LMath.Types
import LMath.Algo.Sort


-- import Trace

-- return the image of `is' under `af'
image :: (Fractional a, Real a)
      => GAffineFunction a -> GIneqSystem a -> GIneqSystem a
image af is
    | nV /= nVars is = error ("image: the AffineFunction's " ++
			      "source space dimensionality and " ++
			      "the IneqSystem's nubmer of " ++
			      "variables differ")
    | nP /= nParams is = error ("image: AffineFunction and " ++
				"IneqSystem differ in number " ++
				"of parameters")
    | targetDimIH af == nV && getMatrix (composeAffine af pi) == unitMatrix (nV+nP+1) (nV+nP+1) =
        inverseImage pi is
    | otherwise =
	case projectToVars extIS (Just extES) [0 .. (targetDimIH af - 1)] of
	   Nothing  -> infeasibleIneqSystem (targetDimIH af) (nParams af)
	   Just res -> res
    where
    (extIS, extES) = extIS_ES af is

    pi = pseudoInverse af
    nV = nVars af
    nP = nParams af

imageG :: (Fractional a, Num cond)
       => GAffineFunction a
       -> GIneqSystem a
       -> DTreeGen a cond (GIneqSystem a)
imageG af is
    | nV /= nVars is = error ("imageG: the GAffineFunction's " ++
			      "source space dimensionality and " ++
			      "the GIneqSystem's nubmer of " ++
			      "variables differ")
    | nP /= nParams is = error ("imageG: GAffineFunction and " ++
				"GIneqSystem differ in number " ++
				"of parameters")
    | otherwise =
        do p <- projectToVarsG extIS (Just extES) Nothing [0 .. (tD - 1)] 
	   case p of
	     Nothing  -> return $ infeasibleIneqSystem tD nP
	     Just res -> return res
    where
    (extIS, extES) = extIS_ES af is
    nV = nVars af
    nP = nParams af
    tD = targetDimIH af

--
-- Helper for image and imageG to compute the extended
-- system.
--
extIS_ES :: Num a 
         => GAffineFunction a -> GIneqSystem a -> (GIneqSystem a, GEqSystem a)
extIS_ES af is = (extIS, extES)
    where
    extIS  = ineqSystemVPC (vars, params, consts)
    extES  = eqSystemVPC   (varsE, paramsE, constsE)
    vars   = (  zeroMatrix (nRows       is) (targetDimIH af)  `conc` iV)
    varsE  = ((-unitMatrix (targetDimIH af) (targetDimIH af)) `conc` fV)

    params  = iP
    paramsE = fP
    consts  = iC
    constsE = fC
    (fV, fP, fC) = partsVPCih af
    (iV, iP, iC) = partsVPC   is



-- |
-- Calculate a description of the inverse image of 
-- an inequality system.
inverseImage :: Num a
	     => GAffineFunction a  -- ^ affine mapping
             -> GIneqSystem a      -- ^ image
             -> GIneqSystem a      -- ^ inverse image
inverseImage af is
    | targetDimIH af /= nVars is  =
        error ("inverseImage: number of dimensions does not match: " ++
               show (targetDimIH af) ++ " vs. " ++ show (nVars is))
    | nParams is /= nParams af  =
	error "inverseImage: number of parameters does not match"
    | otherwise = ineqSystem ((nVars af, nParams af),
			      getMatrix is * getMatrix af)
{-
--	trace (show extMatrix ++ "\n") $
	projectToVars (ineqSystem ((nV+nV', nP), extMatrix)) [0 .. (nV-1)]
    where
    extMatrix = multijoin (zeroMatrix 0 (nV+nV'+nP+1)) $ 
      [zeroMatrix nI nV `conc` vI                   `conc`   pI  `conc`   cI,
       vF               `conc` (-unitMatrix tD nV') `conc`   pF  `conc`   cF,
       (-vF)            `conc`   unitMatrix tD nV'  `conc` (-pF) `conc` (-cF)]
    nV  = nVars     af
    nP  = nParams   af
    tD  = targetDimIH af
    nV' = nVars is
    (vF, pF, cF) = partsVPCih af
    (vI, pI, cI) = partsVPC   is
    nI = numberOfInequations is
-}

-- the image and inverseImage
testInverseImage af destIS =
    (areIneqsEquivalent destIS  destIS'  Nothing,
     areIneqsEquivalent destIS' destIS'' Nothing,
     (destIS,srcIS,destIS',srcIS',destIS''))
    where
    srcIS    = inverseImage af destIS
    destIS'  = image af srcIS
    srcIS'   = inverseImage af destIS'
    destIS'' = image af srcIS'


-- check whether af2(af1(af2(af1(t)))) == af2(af1(t))
-- hold for some t's (and some parameter values) 
testPseudoInverse :: AffineFunction -> AffineFunction -> Bool
testPseudoInverse af1 af2 =
    and [let x = app af2 (app af1 (t `join` p) `join` p)
	 in app af2 (app af1 (x `join` p) `join` p) == x
	 | t <- points, p <- pcs]
    where
    points = columns $ unitMatrix sD sD
    sD = nVars af1
    pcs = columns $ unitMatrix (nParams af1 + 1) (nParams af1 + 1)

    app :: AffineFunction -> Matrix -> Matrix
    app f m = f `apply` m

-- |
-- Calculate an inequality system describing the inverse
-- relation of an affine function.
-- Take an affine function and the description of its definition set
-- and return an inequality system suitable for enumerating all points
-- of the inverse image of a given point from the destination set:
-- the returned inequality system has new variables for a point
-- from the destination set (which are /appended/ to the existing
-- variables); the inequality system is sorted for the original
-- variables, i.e. it contains the lower and upper bounds for
-- the variables representing the inverse image, but no bounds for
-- the newly inserted variables representing the point from the
-- image. 
--
inverseAffine :: Integral a
              => GAffineFunction (Ratio a)
	      -> GIneqSystem (Ratio a)
	      -> GIneqSystem (Ratio a)
inverseAffine af is
    | nVars is /= nV  = 
	error "inverseAffine: number of variables does not match"
    | nParams is /= nP =
	error "inverseAffine: number of parameters does not match"
    | otherwise = filterIneqs f sortedIneqs
    where
    -- at least one of the source variables has a non-zero coefficient
    f rs = any (/=0) (genericTake nV rs)

    -- better use projectToVars...
    sortedIneqs = case sortIneqs newIneqs (Just newEqs) of
		    Nothing     -> newIneqs
		    Just sorted -> sorted
    newIneqs  = ineqSystem ((nV+tD, nP), partI)
    newEqs    = eqSystem   ((nV+tD, nP), partE)
    partI = vI    `conc` zeroMatrix nI tD    `conc`   pI  `conc`   cI
    partE = vF    `conc` (-unitMatrix tD tD) `conc`   pF  `conc`   cF
    nV = nVars     af
    nP = nParams   af
    tD = targetDimIH af
    (vF, pF, cF) = partsVPCih af
    (vI, pI, cI) = partsVPC   is
    nI = numberOfInequations is


-- |
-- Return a (pseudo-)inverse of a given affine function.
-- When @f@ is injective, the a real inverse function @g@ is
-- returned, otherwise the resulting function @g@ will map
-- points from @f@'s target space to /one/ of its pre-images.
-- 
pseudoInverse :: (Real a, Fractional a)
              => GAffineFunction a  -- ^ f
	      -> GAffineFunction a  -- ^ g
pseudoInverse af = affineFunction ((tD,nParams af), inverseRows) Inhomogeneous
    where
    (v, p, c) = partsVPCih af
    (_, gjTransform, colsWithStairs) = gaussJordanTransform v
    tD = targetDimIH af
    targetInGJ = gjTransform * (unitMatrix tD tD `conc` (-p) `conc` (-c))

    -- keep a row of `m' when the corresponging column holds a stair,
    -- replace by a zero row otherwise
    inverseRows
	| nVars af == 0 = zeroMatrix 0 (nColumns targetInGJ)
	| otherwise     = multijoin1 [selectRow r | r <- [0 .. (nVars af - 1)]]
    selectRow r
	| r `elem` colsWithStairs  =  subMatrixHor targetInGJ r r
	| otherwise                =  zeroMatrix 1 (nColumns targetInGJ)



instance Applicable AffineFunction IneqSystem IneqSystem where
    af `apply` is = image af is


fm = matrix (2, 4,
	     [ 0,  1,  0,  0,
	       1,  1,  0,  0%1])

im = matrix (3, 4,
	     [ 1,  0,  0, -2,
	       0,  1,  0,  0,
	       0, -1,  1,  0%1])

f   = affineFunction ((2, 1), fm) Inhomogeneous
sys = ineqSystem ((2, 1), im)

f2   = affineFunction ((2,1), matrix [[1,0,1,0%1],[0,1,0,0],[0,0,0,0]])
       Inhomogeneous
sys2 = ineqSystem ((3,1),matrix [[1,0,0,0,-2%1], [0,1,0,0,-3],[0,-1,-1,0,5]])



-- berechnet ag(af(x))
fctComposition :: AffineFunction -> AffineFunction -> AffineFunction
fctComposition ag af = composeAffine ag af

instance Applicable AffineFunction AffineFunction AffineFunction where
    ag `apply` af = fctComposition ag af

fctCompositionTestOk = fctComposition f2 f 
fctCompositionTestFail = fctComposition f f2 
                            -- f2 has 5 rows, but f needs 4 entries


