-- |
-- Simple implementation of Cylindrical Algebraic Decomposition.
--
module LMath.Algo.CAD
    ( computeCAD,

      gcdfactFree, proj, bmcProj,
      CADMethod(..),

      CADStack(..), CADSection(..), SectionDef(..), CellBound(..),
      cadStack, cadLen, (!!!),
      cadgen, genStack, code2text, codeSimpl, Code(..),
      geogen,
      exEra, exEraCode, exEraCode2, exEraCode4,
      exAZ22, exAZ22Code, ex3',
      exT2Code,
      paperEx

    , reviterate, guardedProjectionOp, GuardedPolys(..)
    )
where

import Control.Exception ( assert, evaluate )
import Data.List
import Data.Maybe
import Data.Ratio
-- import Data.Time.Clock

import Debug.Trace

import LMath.Algo.DisjointUnion
import LMath.Algo.Sort
import LMath.Types
import LMath.Types.Polynomial.AlgNum
import LMath.Types.Polynomial.Roots
import LMath.Tools.Reduce
import LMath.Tools.Qepcad


mytrace a b = b


-- |
-- Factorize all polynomials and report every factor exactly
-- once. Make all polynomials (factors) monic.
--
gcdfactFree :: RealFrac a => [Polynomial a] -> [Polynomial a]
gcdfactFree ps = nub $ map (monic . mapPoly fromIntegral) fs
    where
    fs = concatMap (fact.snd.ratPoly2intPoly.mapPoly toRational) ps

    fact f =
        case totalDegree f of
           0  ->  []
           1  ->  [f]
           _  ->  map fst $ snd $ reduceFactoriseP f

------

-- the main CAD algorithm:

proj :: Fractional a => [Polynomial a] -> [Polynomial a]
proj aps = proj1 ++ proj2
    where
    proj1 = concat [leadingCoeff (univariate g) : subresultants g (derivative g)
		   | r <- reds, g <- r]
    proj2 = concat [subresultants gi gj
		   | (ri:rjs) <- tails reds, gi <- ri, rj <- rjs, gj <- rj]
    reds = map reducta aps


betterProj :: Fractional a => [Polynomial a] -> [Polynomial a]
betterProj aps = proj1 ++ proj2
    where
    proj1 = concat [leadingCoeff (univariate g) : subresultants g (derivative g)
		   | r <- map reducta aps, g <- r]
    proj2 = concat [subresultants f' g
		   | (f:gs) <- tails aps, f' <- reducta f, g <- gs]

--
-- Brown-McCallum Projection Operator
-- NOTE: This is _almost_ a projection operator, i.e., there
-- are cases where it does not yield a correct projection
--
bmcProj :: Fractional a => [Polynomial a] -> [Polynomial a]
bmcProj aps = [leadingCoeff (univariate f) | f <- aps] ++
              [resultant f (derivative f)  | f <- aps, degree f >= 2] ++
              [resultant f g | f:gs <- tails aps, degree f >= 1, g <- gs, degree g >= 1]

onlyPolysWithZeros :: RealFrac a
            => QfFormula (Polynomial a) -> [Polynomial a] -> [Polynomial a]
onlyPolysWithZeros ctx ps = [p | p <- ps, nondisj p]
    where
    nondisj p = case reduceEliminateQuantifiers [] cond of
                  QfAt AtTrue  -> True
                  QfAt AtFalse -> False
                  _            -> error "filterPolys"
        where
        cond = foEx $ QfAnd ctx $ QfAt $ AtEq p


gcdFilter :: Fractional a
	  => [Polynomial a] -> [Polynomial a] -> [Polynomial a]
gcdFilter filter ps = foldl filt ps filter
    where
    filt ps f = map shrink ps
	where
	shrink p
	    | totalDegree g == 0  =  p
	    | otherwise           =  p `quotPoly` g
	    where
	    g = gcdPoly p f


computeCAD :: RealFrac a => [Polynomial a] -> [[AlgNum a]]
computeCAD [] = [[]] --error "cad: undefined for []"
computeCAD aps = computeCAD' (unNIndeterminates $ nIndeterminatesN aps) aps


computeCAD' :: RealFrac a
	    => Integer
	    -> [Polynomial a]
	    -> [[AlgNum a]]
computeCAD' 0 _   = [[]]
computeCAD' 1 aps = map (\s -> [s]) $ samplePoints $
		    sort $ concatMap parToAlgnums (rootsND aps)
computeCAD' nIndet aps = pts2
    where
    pts = computeCAD' (nIndet-1) $ gcdfactFree $ betterProj aps
    pts2 = [ pn ++ p
           | p <- pts, pn <- firstDim p]
    firstDim p = computeCAD' 1 $
                 filter ((>=1).totalDegree) $
                 concat $
                 [map prePoly $ algSubstP p a | a <- aps]


cadStack :: (RealFrac a)
         => [Polynomial a] -> CADStack (AlgNum a) a ()
cadStack [] = error "cadStack: undefined for []"
cadStack ps =
    computeCADStack2 (unNIndeterminates $ nIndeterminatesN ps) $
    gcdfactFree ps

data SectionDef a   = RootOf Integer (Polynomial a)
                    | ConstDef (FracPoly a)
                    | Sqrt { sqrtMB :: Polynomial a, sqrtAddD :: Bool,
                             sqrtD  :: Polynomial a,
                             sqrt2A :: Polynomial a
                           }
                      deriving (Eq, Show)
data CADStack alg a b = CADSector { cadSectorSamplePoint :: a,
                                    cadSectorQf  :: QfFormula (TarskiPRE a),
                                    cadSectorSub :: CADStack alg a b,
                                    cadNeighbour :: Maybe (CADSection alg a b)
                                  }
                      | CADLeaf b
                        deriving (Eq, Show)
data CADSection alg a b = CADSection { cadSectionSamplePoint :: alg
                                     , cadSectionQf  :: QfFormula (TarskiPRE a)
                                     , cadSectionDef :: [SectionDef a]
                                     , cadSectionSub :: CADStack alg a b
                                     , cadSectionNeighbour :: CADStack alg a b
                                     }
                    deriving (Eq, Show)

-- |
-- A "test point" for CAD construction.
-- "Point" lies in the region described by the test point,
-- "Qf" describes a sub-set of the respective region,
-- i.e., it can be used to identify the region as can be
-- the test point.
-- 
data TestPoint alg a = TP { tpPoint :: [alg]
                          , tpQf    :: QfFormula (TarskiPRE a)
                          }
                       deriving Show

initialTestPoint :: (Alg alg, RealFrac a) => TestPoint alg a
initialTestPoint = TP { tpPoint = [], tpQf = QfAt AtTrue }


onlyOutermost :: Integer -> CADStack alg a () -> CADStack alg a ()
onlyOutermost n s@(CADSector { cadSectorSub=ss, cadNeighbour=mbN }) =
    s { cadSectorSub = if n==1 then CADLeaf () else onlyOutermost (n-1) ss
      , cadNeighbour = fmap (onlyOutermostSection n) mbN 
      }

onlyOutermostSection :: Integer -> CADSection alg a () -> CADSection alg a ()
onlyOutermostSection n s@(CADSection { cadSectionSub=ss, cadSectionNeighbour=st }) =
    s { cadSectionSub = newSS, cadSectionNeighbour = onlyOutermost n st }
    where
      newSS = if n == 1 then CADLeaf () else onlyOutermost (n-1) ss

-- |
-- Return whether we know for sure that a section is integral
-- (and never evaluates to a fraction).
-- May return FALSE even if the section evaluates to integers only.
--
sectionDefIsIntegral :: RealFrac a => SectionDef a -> Bool
sectionDefIsIntegral (ConstDef fp)
    | Just c <- isConstFP fp  =  denominator (toRational c) == 1
sectionDefIsIntegral _ = False

-- |
-- Return whether we know for sure that a section is not integral
-- (and never evaluates to an integer).
-- May return FALSE even if the section never evaluates to an integer.
--
sectionDefIsNotIntegral :: RealFrac a => SectionDef a -> Bool
sectionDefIsNotIntegral (ConstDef fp)
    | Just c <- isConstFP fp  =  denominator (toRational c) /= 1
sectionDefIsNotIntegral _ = False

sectionDefs :: CADStack alg a b -> [([SectionDef a],alg)]
sectionDefs (CADSector { cadNeighbour = Just (CADSection {cadSectionNeighbour=s,cadSectionDef=sd,cadSectionSamplePoint=sp})}) = (sd,sp) : sectionDefs s
sectionDefs (CADSector { cadNeighbour = Nothing}) = []

subStacks :: CADStack alg a b -> [CADStack alg a b]
subStacks (CADSector { cadSectorSub=ss, cadNeighbour=Nothing }) =
    [s | s@(CADSector {}) <- [ss]]
subStacks (CADSector { cadSectorSub=ss, cadNeighbour=Just (CADSection { cadSectionSub=ss2, cadSectionNeighbour=next}) }) =
    [s | s@CADSector {} <- [ss,ss2]] ++ subStacks next

cssp :: (Alg alg, RealFrac a) => CADStack alg a () -> [[alg]]
cssp = cssp' []
    where
    cssp' pts (CADSector { cadSectorSamplePoint=pt, cadSectorSub = s@(CADSector {}), cadNeighbour=Just s2 }) = cssp' (toAlg pt:pts) s ++ cstsp pts s2
    cssp' pts (CADSector { cadSectorSamplePoint=pt, cadSectorSub = CADLeaf _, cadNeighbour=Just s2 }) = (toAlg pt:pts) : cstsp pts s2
    cssp' pts (CADSector { cadSectorSamplePoint=pt, cadSectorSub = s@CADSector {}, cadNeighbour=Nothing }) = cssp' (toAlg pt:pts) s
    cssp' pts (CADSector { cadSectorSamplePoint=pt, cadSectorSub = CADLeaf _, cadNeighbour=Nothing }) = [toAlg pt:pts]

    cstsp pts (CADSection {cadSectionSamplePoint=pt,cadSectionSub = s@CADSector {},cadSectionNeighbour = s2 }) = cssp' (pt:pts) s ++ cssp' pts s2
    cstsp pts (CADSection {cadSectionSamplePoint=pt,cadSectionSub = CADLeaf {}, cadSectionNeighbour = s2 }) = (pt:pts) : cssp' pts s2


cadLen :: CADStack alg a () -> Integer
cadLen (CADSector { cadNeighbour = Nothing }) = 1
cadLen (CADSector { cadNeighbour = Just (CADSection {cadSectionNeighbour=s}) }) = 2 + cadLen s

(!!!) :: CADStack alg a () -> Integer -> CADStack alg a ()
(!!!) (CADSector {cadSectorSub=s@(CADSector {})}) 0 = s
(!!!) (CADSector {cadNeighbour=Just (CADSection {cadSectionSub=s@(CADSector {})})}) 1 = s
(!!!) (CADSector {cadNeighbour=Just (CADSection {cadSectionNeighbour=s})}) n = s !!! (n-2)

cnt (CADSector {cadNeighbour=Nothing,cadSectorSub=CADLeaf _}) = 1
cnt (CADSector {cadNeighbour=Nothing,cadSectorSub=s}) = cnt s
cnt (CADSector {cadNeighbour=Just (CADSection {cadSectionNeighbour=s,cadSectionSub=CADLeaf _}),cadSectorSub=CADLeaf _}) = cnt s + 2
cnt (CADSector {cadNeighbour=Just (CADSection {cadSectionNeighbour=s,cadSectionSub=s2}),cadSectorSub=s3}) = cnt s + cnt s2 + cnt s3


extend :: (Alg alg, RealFrac a)
       => (TestPoint alg a -> CADStack alg a b)
       -> CADStack alg a b
       -> CADStack alg a b
extend f = extendSector initialTestPoint
    where
    extendSector tp sector =
        case cadSectorSub sector of
           CADLeaf {} -> sector { cadSectorSub = f newTp,
                              cadNeighbour = newNeighbour
                            }
           ss@(CADSector {}) -> 
               sector { cadSectorSub = newSubstack ss,
                        cadNeighbour = newNeighbour
                      }
        where
        newTp     =
            TP { tpPoint = toAlg (cadSectorSamplePoint sector) : tpPoint tp
               , tpQf    = cadSectorQf sector `QfAnd`
                           fmap liftPRE (tpQf tp)
               }
        newNeighbour   = fmap (extendSection tp) $ cadNeighbour sector
        newSubstack ss = extendSector newTp ss

    extendSection tp section =
        case cadSectionSub section of
           CADLeaf {} -> section { cadSectionSub = f newTp,
                                   cadSectionNeighbour = newNeighbour
                                 }
           ss@(CADSector {}) -> 
               section { cadSectionSub = newSubstack ss,
                         cadSectionNeighbour = newNeighbour
                       }
        where
        newTp     =
            TP { tpPoint = cadSectionSamplePoint section : tpPoint tp
               , tpQf    = cadSectionQf section `QfAnd` fmap liftPRE (tpQf tp)
               }
        newNeighbour   = extendSector tp $ cadSectionNeighbour section
        newSubstack ss = extendSector newTp ss


computeCADStack2 :: (RealFrac a)
	    => Integer          -- ^ number of dimensions
	    -> [Polynomial a]   -- ^ input polynomials
	    -> CADStack (AlgNum a) a ()
computeCADStack2 nInd ps =
    trace ("projs = " ++ show projs) $
    trace ("mtd = " ++ show (maximum (map totalDegree $ concat projs))) $
    trace ("map length projs = " ++ show (map length projs)) $
    trace (show $ onlyOutermost 1 stack) $
    stack
    where
    stack = old_genStack 1 [] (QfAt AtTrue) projs
    projs = reverse $ genericTake nInd $ iterate projOp ps

    projOp = gcdfactFree . bmcProj

computeCADStack4 :: (RealFrac a)
	    => Integer          -- ^ number of dimensions
            -> [QfFormula (Polynomial a)]    -- ^ guard formulas
	    -> [Polynomial a]   -- ^ input polynomials
	    -> CADStack (AlgNum a) a ()
computeCADStack4 nInd guardFs ps =
--  trace ("projs = " ++ show projs) $
--  trace ("mtd = " ++ show (maximum (map totalDegree $ concat projs))) $
--  trace ("map length projs = " ++ show (map length projs)) $
--    trace (show $ length $ cssp $ onlyOutermost 4 stack) $
--    undefined  
--    trace (show $ head guardFs) $
    stack

    where
    stack = old_genStack 1 [] (QfAt AtTrue) projs
    projs = reverse $ genProj ps guardFs

    -- eliminate polynomials from the projection which have no zeros
    genProj _ [] = []
    genProj qs (gF:gFs) =
--        trace ("proj = " ++ show qs) $
        (qs : genProj proj gFs)
        where
        proj = [p | p <- projOp qs, not (implNeq gF p)]

    projOp = gcdfactFree . bmcProj



computeCADStackGeo :: RealFrac a
                   => Integer                    -- ^ number of dimensions
                   -> QfFormula (Polynomial a)   -- ^ context formula
                   -> [QfFormula (Polynomial a)] -- ^ input polynomials
                   -> CADStack (AlgNum a) a ()
computeCADStackGeo nIndets (ctx :: QfFormula (Polynomial a)) formulas =
    stack
    where
    stack = old_genStack 1 [] (QfAt AtTrue) projs

    projs = map gpPolys gps
    gps = qepcadGuardedProjections nIndets ctx $
          map domainToGP formulas


old_genStack :: RealFrac a
         => Integer 
         -> [AlgNum a]
         -> QfFormula (TarskiPRE a)
         -> [[Polynomial a]]
         -> CADStack (AlgNum a) a ()
old_genStack nInd sample ctx [] = CADLeaf ()
old_genStack nInd (sample :: [AlgNum a]) ctx (ps:projs) =
--    trace ("ps = " ++ show ps) $
--    trace ("sampleAndRoots = " ++ show (map snd sampleAndRoots)) $
    gen sampleAndRoots
    where
    sampleAndRoots :: [(AlgNum a, [(Polynomial a, Integer)])]
    sampleAndRoots = uniqSortBy compare (nubBy (==)) $
                     concatMap tp ps
        where
        tp p = [(a,(p,i))
               | p' <- pp'
               -- substP may return a polynomial with more zeros than p,
               -- but if p' has the same degree as p (in the outermost
               -- variable), there are no additional zeros.
               , let exact  =  degree p == degree p'
               , (a,i) <- zip [a | a <- concatMap parToAlgnums $ rootsND [p']
                                 , let s = substAlgs p (a:sample)
                                 , exact || s == 0
                              ]
                              [1..]
               ]
            where
            pp' = map prePoly $ algSubstP sample p            

    liftCtx = fmap liftPRE ctx

    gs s newCtx = old_genStack (nInd+1) (s:sample) newCtx projs

    gen []           = CADSector 0 liftCtx (gs 0 liftCtx) Nothing
    gen ((a,(p,i):pis):rs) = CADSector left nctx (gs (toAlgNum left) nctx) $
                             Just $ gen2 a ((p,i):pis) rs
        where
        left = algSmallerRat a
        nctx = liftCtx `QfAnd` (QfAt $ AtGt $ PRERoot False i p)

    gen2 left ((p,i):pis) [] =
        CADSection left nctx (map mkRoot ((p,i):pis)) (gs left nctx) $
        CADSector right nctx2 (gs (toAlgNum right) nctx2) $
        Nothing
        where
        right = algBiggerRat left
        nctx  = liftCtx `QfAnd` (QfAt $ AtEq $ PRERoot True i p)
        nctx2 = liftCtx `QfAnd` (QfAt $ AtGt $ PRERoot True i p)
    gen2 left ((p,i):pis) ((a,(p2,i2):pis2):rs) =
        CADSection left nctx (map mkRoot ((p,i):pis)) (gs left nctx) $
        CADSector middle nctx2 (gs (toAlgNum middle) nctx2) $
        Just $
        gen2 a ((p2,i2):pis2) rs
        where
        middle = {- trace (show (left,a)) $ -} algRatBetween left a
        nctx  = liftCtx `QfAnd` (QfAt $ AtEq $ PRERoot True i p)
        nctx2 = liftCtx `QfAnd` (QfAt $ AtGt $ PRERoot True i p)
                        `QfAnd` (QfAt $ AtGt $ PRERoot False i2 p2)

    mkRoot (p,n)
        | totalDegree u == 1  =  ConstDef (negate c %% b)
{-
        | totalDegree u == 2  =
            if d==0
            then ConstDef (negate b %% (2*a))
            else Sqrt { sqrtMB = -b, sqrtAddD =  ((n==1) == positiveA),
                        sqrtD  =  d, sqrt2A = 2*a
                      }
-}
        | otherwise = RootOf n p
        where
        u = univariate p
        a = coefficient u 2
        b = coefficient u 1
        c = coefficient u 0
        d = b*b-4*a*c

--        positiveA = substAlgs a sample > 0

--
-- degree? totalDegree? (rev)lex. order on exponents?
--
selectAlg f1@(a1,p1,_) f2@(a2,p2,_)
--  | degree (algNumPoly a1) < degree (algNumPoly a2)  =  f1
-- compare the degrees of the polynomials defining the section, not the
-- polynomialsr defining the (algebraic) roots
  | totalDegree p1 <= totalDegree p2  =  f1
  | otherwise                        =  f2


-- sort and group duplicates
uniqSortBy :: (a -> a -> Ordering) -> ([b] -> [b]) -> [(a,b)] -> [(a,[b])]
uniqSortBy cmp equiv list = s list []
    where
    s []            acc = acc
    s (x@(xa,_):xs) acc = s lt (newX : s gt acc)
        where
        newX = (xa, equiv $ map snd (x:eq))
        (lt,eq,gt) = part3 [] [] [] xs

        part3 lt eq gt []     = (lt,eq,gt)
        part3 lt eq gt (y@(ya,_):ys) =
            case ya `cmp` xa of
              LT -> part3 (y:lt)    eq     gt  ys
              EQ -> part3    lt  (y:eq)    gt  ys
              GT -> part3    lt     eq  (y:gt) ys

-- sort and eliminate duplicates
uniqSort :: (a -> a -> Ordering) -> [a] -> [a]
uniqSort cmp list = s list []
    where
    s []     acc = acc
    s (x:xs) acc = s ys (x : s zs acc)
        where
          (ys,zs) = part2 [] [] xs

          part2 lt gt [] = (lt,gt)
          part2 lt gt (a:as) =
              case a `cmp` x of
                LT -> part2 (a:lt)    gt  as
                GT -> part2    lt  (a:gt) as
                EQ -> part2    lt     gt  as


sgnAlgPoly :: RealFrac a => [AlgNum a] -> Polynomial a -> Ordering
sgnAlgPoly algs poly
    | Nothing `notElem` mbRats  =
        evalPoly poly rats `compare` 0
    | reduceEliminateQuantifiers [] (f AtEq) == QfAt AtTrue  =  EQ
    | reduceEliminateQuantifiers [] (f AtGt) == QfAt AtTrue  =  GT
    | otherwise                               =  LT
    where
    f op = FoEx [0..genericLength algs-1] $ FoQf $
           (conjunctionQf [a i | i <- [0..length algs-1]]
            `QfAnd`
            QfAt (op poly)
           )
    
    a i = algNumToFormula nV i (algs!!i)
    nV = length algs
    mbRats = map algIsRational algs
    rats = [r | Just r <- mbRats]
    

    
algNumToFormula :: RealFrac a => Int -> Int -> AlgNum a -> QfFormula (Polynomial a)
algNumToFormula nV varNum alg =
    case algNumRoots alg of
      (RootAt x):_       -> QfAt $ AtEq (indeterminate nV varNum - constPoly x)
      (RootIn a b _ _):_ ->
          conjunctionQf $
          map QfAt $
          [AtEq $ liftPolyN varNum $ appendIndeterminates (nV-varNum-1) $
                algNumPoly alg
          ,AtGt (indeterminate nV varNum - constPoly a)
          ,AtGt (constPoly b - indeterminate nV varNum)
          ]
    




samplePoints :: RealFrac a => [AlgNum a] -> [AlgNum a]
samplePoints [] = [0]
samplePoints (p:ps) = toAlgNum (algSmallerRat p) : sps (p:ps)
    where
    sps [x] = [x, toAlgNum (algBiggerRat x)]
    sps (x:y:rest) = x : toAlgNum (algRatBetween x y) : sps (y:rest)


{-
computeCAD [(indeterminate 2 0 ^ 2 - 5)*(indeterminate 2 1 ^ 3 + indeterminate 2 1 + 4)]

let { u0=indeterminate 2 0; u1=indeterminate 2 1 } in [u0^2*(u1*(-1)+2)+u0*(u1^2*(-1)+u1*2+2)+(u1^2*1+u1*1),u0*1+(u1*1+1),u0*(-1)+(-1),u0*u1*1+2,u0*u1*(-1)+(-2)]
-}


data CADMethod = CADInternal
               | CADQepcadGeo
                 deriving (Eq, Show)

-- variables/parametrs in "inverse" order:
-- first indeterminate is innermost dimension
old_cadgen :: RealFrac a
       => CADMethod
       -> QfFormula (Polynomial a) -> [QfFormula (Polynomial a)] -> [Code a]
old_cadgen cadmethod ctxFormula (formulas :: [QfFormula (Polynomial a)]) =
    -- trace ("guardFs = " ++ show guardFs) $
--    trace ("cadstack = " ++ show cadstack) $
    execute [] (QfAt AtTrue) cadstack
    where
    loops samples lb (CADSector {cadSectorSub=ss,cadSectorSamplePoint=s,cadNeighbour=Nothing,cadSectorQf=qf}) =
        ne (CFor lb ToPosInfty) (execute (toAlgNum s:samples) qf ss) $ []
    loops samples lb (CADSector {cadSectorSub=ss,cadSectorQf=qf,cadSectorSamplePoint=s,cadNeighbour=Just (CADSection {cadSectionSub=ss2,cadSectionQf=qf2,cadSectionSamplePoint=s2,cadSectionDef=def,cadSectionNeighbour=sector2})}) =
        ne (CFor lb (SectionBound False def)) (execute (toAlgNum s:samples) qf ss) $
        ne (CIf def) (execute (s2:samples) qf2 ss2) $
        loops samples (SectionBound False def) sector2
           
    ne :: ([b] -> c) -> [b] -> [c] -> [c]
    ne f [] cs = cs
    ne f bs cs = f bs : cs
 
    execute pt qf (CADLeaf {}) =
        [CStatement i 0 | (f,i) <- zip formulas [1..],
                          evalQf (\p -> sgnAlgPoly pt p) f
--                       qepcadEtfImpliesQf [qf] $ fmap PREPoly f
        ]

    execute pt qf (substack@CADSector {})
        | evalQf (\p -> sgnAlgPoly pt p) (guardFs!!length pt)  =
            loops pt FromNegInfty substack
        | otherwise  =  -- trace (show pt ++ " (empty)") $
                       []

    polynomials = termsQf formula
    nIndets = unNIndeterminates $ nIndeterminatesN polynomials

--    cadstack = cadStack polynomials
    cadstack =
        case cadmethod of
          CADInternal ->
              computeCADStack4 nIndets (reverse guardFs) $
              gcdfactFree polynomials
          CADQepcadGeo ->
              computeCADStackGeo nIndets ctxFormula formulas


    guardFs = reverse $ contexts nV formula

    formula = ctxFormula `QfAnd` disjunctionQf formulas
    nV = unNIndeterminates $ indeterminatesInQf formula


contexts :: Real a
         => Integer -> QfFormula (Polynomial a) -> [QfFormula (Polynomial a)]
contexts nV formula =
    [fmap (deleteIndeterminates (n+1)) $
     reduceEliminateQuantifiers [] (FoEx [0..n] $ FoQf formula)
    | n <- [0..nV-1]
    ]



sortRoots :: RealFrac a
          => Integer
          -> QfFormula (TarskiPRE a)
          -> [(Polynomial a, Integer)]
          -> [(Polynomial a, Integer)]
sortRoots nInd ctx roots = sortBy order roots
    where
    order = orderDisjointRoots nInd ctx

orderDisjointRoots :: RealFrac a
                   => Integer
                   -> QfFormula (TarskiPRE a)
                   -> (Polynomial a, Integer)
                   -> (Polynomial a, Integer)
                   -> Ordering
orderDisjointRoots nInd ctx (f,nf) (g,ng) =
--    trace (show (nInd,ctx,(f,nf),(g,ng))) $
    case (qepcadEtfEliminateQuantifiers [] (cond AtGt)
         ,qepcadEtfEliminateQuantifiers [] (cond (AtGt . negate))
         ) of
          (QfAt AtTrue ,QfAt AtFalse) -> GT
          (QfAt AtFalse,QfAt AtTrue ) -> LT
          (QfAt AtFalse,QfAt AtFalse) -> EQ
          _            -> error ("orderDisjointRoots: " ++ show (cond AtGt))
    where
    cond rel =
            FoEx [0..nInd] $ FoQf $
            conjunctionQf [ctx',
                           QfAt $ AtEq (PRERoot True nf f'),
                           QfAt $ AtEq (PRERoot True ng g'),
                           QfAt $ rel (PREPoly (indeterminate (nInd+1) (nInd-1) - indeterminate (nInd+1) nInd))
                          ]
    f' = appendIndeterminates 1 $ reverseIndeterminates nInd f
    g' = insertIndeterminates (nInd-1) 1 $ reverseIndeterminates nInd g
    ctx' = fmap (mapPREpoly revLift2) ctx

    revLift2 = reverseIndeterminates (nInd+1) . liftPolyN 2

howManyRoots :: RealFrac a
             => Integer      -- ^ roots in this indeterminate
             -> QfFormula (TarskiPRE a)
             -> Polynomial a
             -> Integer
howManyRoots dim ctx (f :: Polynomial a) = hmr 0
    where
    degI = degreeIn dim f

    hmr :: Integer -> Integer
    hmr n'
        | n' == degI = n'
        | atleastN   = hmr n
        | otherwise  = n'
        where
        n = n' + 1
        atleastN =
--            trace (show (ordCond, [root k| k <- [0..n-1]])) $
            case qepcadEtfEliminateQuantifiers [] cond of
              QfAt AtTrue  -> True
              QfAt AtFalse -> False
              qf -> error ("howManyRoots: " ++ show qf)

        cond :: FoFormula (TarskiPRE a)
        cond = FoEx ([0..nInd+n-2]) $ FoQf $
               conjunctionQf (ctx' : ordCond : [root k | k <- [0..n-1]])

        root :: Integer -> QfFormula (TarskiPRE a)
        root 0 = QfAt $ AtEq $ PREPoly f'
        root k = QfAt $ AtEq $ PREPoly $
                 swapIndeterminates dim (nInd+k-1) f'

        ordCond :: QfFormula (TarskiPRE a)
        ordCond = conjunctionQf [ordC k | k <- [0..n-2]]
        ordC k  = QfAt $ AtGt $ PREPoly
                  (indeterminate (nInd+n) (pos (k+1))
                   - indeterminate (nInd+n) (pos k)
                  )

        pos 0 = dim
        pos k = nInd + k - 1

        f' :: Polynomial a
        f'   = appendIndeterminates n f

        ctx' :: QfFormula (TarskiPRE a)
        ctx' = fmap (mapPREpoly (appendIndeterminates n')) ctx
    
    nInd :: Integer
    nInd = unNIndeterminates $ nIndeterminatesN (f : map prePoly (termsQf ctx))



howManyRoots' :: RealFrac a
             => Integer      -- ^ roots in this indeterminate
             -> QfFormula (TarskiPRE a)
             -> Polynomial a
             -> [(Integer, QfFormula (Polynomial a))]
howManyRoots' dim ctx (f :: Polynomial a) = hmr 0
    where
    degI = degreeIn dim f

    hmr :: Integer -> [(Integer, QfFormula (Polynomial a))]
    hmr n'
        | n' == degI  =  []
        | otherwise =
            case atleastN of
              QfAt AtFalse -> []
              qf           -> (n,qf) : hmr n
        where
        n = n' + 1
        atleastN =
--            trace (show (ordCond, [root k| k <- [0..n-1]])) $
            fmap (deleteEndIndeterminates (n-1)) $
            qepcadEtfEliminateQuantifiers [ctx'] cond

        cond :: FoFormula (TarskiPRE a)
        cond = FoEx (dim:[nInd..nInd+n-2]) $
               FoAll [nInd+n-1] $ FoQf $
               (
                conjunctionQf (ordCond : ((ordC (n-1) `QfOr` before) `qfImpl` nonroot) : [root k | k <- [0..n-1]])
               )

        root :: Integer -> QfFormula (TarskiPRE a)
        root 0 = QfAt $ AtEq $ PREPoly f'
        root k = QfAt $ AtEq $ PREPoly $
                 swapIndeterminates dim (nInd+k-1) f'

        nonroot = QfAt $ AtNeq $ PREPoly $
                  swapIndeterminates dim (nInd+n-1) f'

        ordCond :: QfFormula (TarskiPRE a)
        ordCond = conjunctionQf [ordC k | k <- [0..n-2]]
        ordC k  = QfAt $ AtGt $ PREPoly
                  (indeterminate (nInd+n) (pos (k+1))
                   - indeterminate (nInd+n) (pos k)
                  )
        before =  QfAt $ AtGt $ PREPoly
                  (indeterminate (nInd+n) (pos 0)
                   - indeterminate (nInd+n) (pos n)
                  )

        pos 0 = dim
        pos k = nInd + k - 1

        f' :: Polynomial a
        f'   = appendIndeterminates n f

        ctx' :: QfFormula (TarskiPRE a)
        ctx' = fmap (mapPREpoly (appendIndeterminates n)) ctx
--        revLift = reverseIndeterminates (nInd+n-1) . liftPolyN n
    
    nInd :: Integer
    nInd = unNIndeterminates $ nIndeterminatesN (f : map prePoly (termsQf ctx))



----------------------------------------------

data GuardedPolys a = GP { gpDim   :: Integer
                         , gpGuard :: QfFormula (Polynomial a)
                         , gpPolys :: [Polynomial a]
                         }
                      deriving Show

domainToGP :: QfFormula (Polynomial a) -> GuardedPolys a
domainToGP d = GP { gpPolys = termsQf d
                  , gpGuard = d
                  , gpDim   = unNIndeterminates $ indeterminatesInQf d
                  }

unionGPs :: RealFrac a => [GuardedPolys a] -> GuardedPolys a
unionGPs [] = error "unionGPs: empty union"
unionGPs gps =
    assert (length (nub $ map gpDim gps) == 1) $
    GP { gpDim = gpDim $ head gps
       , gpPolys = gcdfactFree $ concatMap gpPolys gps
       , gpGuard = disjunctionQf $ map gpGuard gps
       }

evalGuard :: (Alg alg, RealFrac a) => TestPoint alg a -> GuardedPolys a -> Bool
evalGuard tp gp =
    case guardQf of
      QfAt AtTrue  -> True
      QfAt AtFalse -> False
      _ ->
          assert (length (tpPoint tp) ==
                  unNIndeterminates (indeterminatesInQf guardQf)) $
          evalQf (\p -> algOrderingP (tpPoint tp) p) guardQf
--  maybe use Qepcad: ?
--    qepcadImpliesQf [tpQf tp] guardQf
    where
    guardQf = eliminateAtTF $ gpGuard gp

--
-- scatter dims first, then iterators, then parameters
cadgen :: (Alg alg, RealFrac a)
       => CADMethod
       -> Integer                      -- ^ number of parameter dimensions
       -> QfFormula (Polynomial a)     -- ^ context (in parameters only)
       -> Integer                      -- ^ number of scattering dimensions
       -> [QfFormula (Polynomial a)]   -- ^ domain descriptions
       -> alg
       -> [Code a]
--       -> CADStack a [(Integer,CADStack a ())]
--       -> [[GuardedPolys a]]
cadgen cadMethod nP (ctx :: QfFormula (Polynomial a)) nS fulldoms
           (algDummy :: alg) =
    assert (indsInQf ctx <= nP) $
    assert (all (\d -> indsInQf d >= nS+nP) fulldoms) $
    assert (all ((==nS+nP).gpDim) $ map head $ projectionsI)
--    trace (show scatterStack) $
--    projectionsI
--    scatterStack
    code
    where
    code = cad2loops (\tp val -> concatMap (innerCode tp) val)
                     initialTestPoint
                     projectionsPS
                     scatterStack

    innerCode :: TestPoint alg a -> (Integer,CADStack alg a ()) -> [Code a]
    innerCode tp (stmtNum,innerStack) =
        cad2loops (\_ _ -> [CStatement stmtNum (genericLength thisStmtProjs)])
                  tp
                  thisStmtProjs
                  innerStack
        where
        thisStmtProjs = tail (projectionsI !! (fromIntegral stmtNum - 1))

    -- construct the inner stack for a given statement number
    iStack :: TestPoint alg a -> Integer -> [(Integer, CADStack alg a ())]
    iStack tp stmtNum
        -- is the current cell part of this statements domain?
        | evalGuard tp commProj =
            [(stmtNum, genStack (\tp -> CADLeaf ()) tp innerProjs)]
        | otherwise = []
        where
        commProj:innerProjs = projectionsI !! (fromIntegral stmtNum - 1)

    iteratorStacks :: TestPoint alg a
                   -> CADStack alg a [(Integer,CADStack alg a ())]
    iteratorStacks tp = CADLeaf $ concatMap (iStack tp) [1..nStmts]

    -- construct the outer (common) stack for the parameters and
    -- scatter dimensions and call 'iteratorStacks' to construct
    -- the inner stacks for the iterators of each statement
    scatterStack :: CADStack alg a [(Integer, CADStack alg a ())]
    scatterStack = genStack iteratorStacks initialTestPoint projectionsPS 

    projectionsPS :: [GuardedPolys a]
    projectionsPS =
        case cadMethod of
          CADInternal  -> tail projectionsP ++ tail projectionsS
          CADQepcadGeo -> qepcadGuardedProjections (nP+nS) rctx' $
                          map head projectionsI
        where
          rctx' = fmap (reverseIndeterminates (nP+nS)) ctx

    projectionsP = reviterate nP guardedProjectionOp paramGP

    -- add parameter constraints to guard
    paramGP  = paramGP1 { gpGuard = rctx `QfAnd` gpGuard paramGP1
                        , gpPolys = gcdfactFree (termsQf rctx ++ gpPolys paramGP1)
                        }
    paramGP1 = head projectionsS

    projectionsS :: [GuardedPolys a]
    projectionsS = reviterate nS guardedProjectionOp commGP

    -- projection to the common dimensins (scatter dims)
    commGP :: GuardedPolys a
    commGP = unionGPs $ map head projectionsI
        
    -- project on the scatter dimensions

    -- projections in the iterator dimensions
    -- 
    projectionsI :: [ [GuardedPolys a] ]
    projectionsI = [reviterate (gpDim gp - nS - nP) guardedProjectionOp gp
                   | gp <- polys]

    -- polynomials in the domains and the number of indeterminates
    polys :: [GuardedPolys a]
    polys = [domainToGP d | d <- revfulldoms]

    -- doms in scatter and iterator dims with reversed ordering
    -- on indeterminates
    revfulldoms :: [QfFormula (Polynomial a)]
    revfulldoms = [fmap (revVarsAndParams nSI) d
                  | d <- fulldoms
                  -- number of scatter dim + iterator dims
                  -- (can be different for each statement)
                  , let nSI = indsInQf d - nP  
                  ]

    rctx :: QfFormula (Polynomial a)
    rctx = fmap (reverseIndeterminates nP) ctx

    indsInQf qf = unNIndeterminates $ indeterminatesInQf qf

    -- the number of statements
    nStmts :: Integer
    nStmts = genericLength fulldoms


-- |
-- Reverse first nV indeterminates and reverse remaining indeterminates.
--
revVarsAndParams :: Num a
                 => Integer       -- ^ nV
                 -> Polynomial a
                 -> Polynomial a
revVarsAndParams nV p =
    createPolyFromList $ [(co,f es) | (co,es) <- polyToList p]
    where
    f es = reverse es1 ++ reverse es2
        where
        (es1,es2) = splitAt (fromIntegral nV) es

guardedProjectionOp :: RealFrac a
                    => GuardedPolys a
                    -> GuardedPolys a
guardedProjectionOp gp =
    GP { gpPolys = [p | p <- projectionOp (gpPolys gp)
                      , not (implNeq newGuard p)]
       , gpGuard = newGuard
       , gpDim = gpDim gp - 1
       }
    where
    newGuard = fmap (deleteIndeterminates 1) $
               reduceEliminateQuantifiers [] $ FoEx [0] $ FoQf $ gpGuard gp

-- |
-- Check if a polynomial is non-zero under
-- a given context.
--
implNeq :: Real a
        => QfFormula (Polynomial a)  -- ^ context
        -> Polynomial a              -- ^ polynomial to test
        -> Bool
implNeq ctx p = signumNeq $ reduceImpliesSignum [ctx] p


projectionOp :: RealFrac a => [Polynomial a] -> [Polynomial a]
projectionOp = gcdfactFree . bmcProj

reviterate :: Integer -> (a -> a) -> a -> [a]
reviterate 0 f x = [x]
reviterate n f x
    | n < 0      =  error "reviterate: n<0"
    | otherwise  =  f y : rest
    where
      rest@(y:_) = reviterate (n-1) f x

-- Loop Generation

cad2loops :: (Alg alg, RealFrac a, Show b)
          => (TestPoint alg a -> b -> [Code a])
          -> TestPoint alg a
          -> [GuardedPolys a]
          -> CADStack alg a b
          -> [Code a]
cad2loops leafFn startTp gps stack =
--    trace ("cad2loops: " ++ show (tp,gps)) $
    loops [] FromNegInfty stack
    where
    loops samples lb (CADSector {cadSectorSub=ss
                                ,cadSectorSamplePoint=s
                                ,cadNeighbour=Nothing
                                ,cadSectorQf=qf}) =
        nonemptyAppCons (CFor lb ToPosInfty)
                        (execute (toAlg s:samples) qf ss)
                        []
    loops samples lb (CADSector {cadSectorSub=ss
                                ,cadSectorQf=qf
                                ,cadSectorSamplePoint=s
                                ,cadNeighbour=
                                 Just (CADSection {cadSectionSub=ss2
                                                  ,cadSectionQf=qf2
                                                  ,cadSectionSamplePoint=s2
                                                  ,cadSectionDef=def
                                                  ,cadSectionNeighbour=sector2
                                                  }
                                      )}) =
        nonemptyAppCons (CFor lb (SectionBound False def))
           (execute (toAlg s:samples) qf ss) $
        nonemptyAppCons (CIf def) (execute (s2:samples) qf2 ss2) $
        loops samples (SectionBound False def) sector2
    -- For a zero-dimensional CAD (samples == []), we can
    -- reach a leaf immediately. Inner leaves are reached in 'execute'.
    loops [] lb (CADLeaf val) = leafFn startTp val
           
    execute samples qf sector
        | evalGuard newTp gp =
            -- trace (show (length $ tpPoint newTp,True)) $
            case sector of
              CADLeaf val           -> leafFn newTp val
              substack@CADSector {} -> loops samples FromNegInfty substack
        | otherwise = -- trace (show (length $ tpPoint newTp,False)) $
                      []
        where
        newTp = TP { tpPoint = samples ++ tpPoint startTp, tpQf = qf }
        gp = gps !! (length samples - 1)

nonemptyAppCons :: ([b] -> c) -> [b] -> [c] -> [c]
nonemptyAppCons f [] cs = cs
nonemptyAppCons f bs cs = f bs : cs

-- Stack Construction

genStack :: (Alg alg, RealFrac a)
         => (TestPoint alg a -> CADStack alg a b)
         -> TestPoint alg a
         -> [GuardedPolys a]
         -> CADStack alg a b
genStack leafvalue tp                      []       = leafvalue tp
genStack leafvalue (tp :: TestPoint alg a) (gp:gps) =
--  trace ("(ps,sample) = " ++ show (ps,sample)) $
--    trace ("sampleAndRoots = " ++ show sampleAndRoots) $
--    trace ("genStack: " ++ show (tp,gp)) $
    gen sampleAndRoots
    where
    sample :: [alg]
    sample = tpPoint tp
    ctx    = tpQf    tp
    ps     = [p | p <- gpPolys gp, degree p > 0]

    sampleAndRoots :: [(alg, [(Polynomial a, Integer)])]
    sampleAndRoots =
        uniqSortBy compare (nubBy (==)) $
        concatMap pointsAndRoots $
        ps
        where
        pointsAndRoots :: Polynomial a
                       -> [(alg, (Polynomial a, Integer))]
        pointsAndRoots p =
            [(a,(monic p,i)) | (a,i) <- zip algs [1..]]
            where
            algs = uniqSort compare $ concatMap f $ algSubstP sample p
            f (PRERoot _ n poly) = [zeroToAlg poly n]
            f (PREPoly poly)
                | degree poly == -1  =  []
                | otherwise          =
                    [a | poly2 <- factors poly
                       , i <- [1..nRoots poly2]
                       , let a = zeroToAlg poly2 i
                       , exact || algEvalP (a:sample) p == 0
                    ]
                where
                -- substP may return a polynomial with more zeros than p,
                -- but if p' has the same degree as p (in the outermost
                -- variable), there are no additional zeros.
                exact =  degree p == degree poly

    factors :: Polynomial a -> [Polynomial a]
    factors p =
        map (fmap' fromIntegral) $ map fst $ snd $ reduceFactoriseP $ snd $
        ratPoly2intPoly p 

    liftCtx = fmap liftPRE ctx

    gs s newCtx = genStack leafvalue newTp gps
        where
          newTp = TP { tpPoint = s : sample
                     , tpQf    = newCtx
                     }

    gen []           = CADSector 0 liftCtx (gs 0 liftCtx) Nothing
    gen ((a,(p,i):pis):rs) = CADSector left nctx (gs (toAlg left) nctx) $
                             Just $ gen2 a ((p,i):pis) rs
        where
        left = algSmallerRat a
        nctx = liftCtx `QfAnd` (QfAt $ AtGt $ PRERoot False i p)

    gen2 left ((p,i):pis) [] =
        CADSection left nctx (map mkRoot ((p,i):pis)) (gs left nctx) $
        CADSector right nctx2 (gs (toAlg right) nctx2) $
        Nothing
        where
        right = algBiggerRat left
        nctx  = liftCtx `QfAnd` (QfAt $ AtEq $ PRERoot True i p)
        nctx2 = liftCtx `QfAnd` (QfAt $ AtGt $ PRERoot True i p)
    gen2 left ((p,i):pis) ((a,(p2,i2):pis2):rs) =
        CADSection left nctx (map mkRoot ((p,i):pis)) (gs left nctx) $
        CADSector middle nctx2 (gs (toAlg middle) nctx2) $
        Just $
        gen2 a ((p2,i2):pis2) rs
        where
        middle = {- trace (show (left,a)) $ -} algRatBetween left a
        nctx  = liftCtx `QfAnd` (QfAt $ AtEq $ PRERoot True i p)
        nctx2 = liftCtx `QfAnd` (QfAt $ AtGt $ PRERoot True i p)
                        `QfAnd` (QfAt $ AtGt $ PRERoot False i2 p2)

    mkRoot (p,n)
        | totalDegree u == 1  =  ConstDef (negate c %% b)
{-
        | totalDegree u == 2  =
            if d==0
            then ConstDef (negate b %% (2*a))
            else Sqrt { sqrtMB = -b, sqrtAddD =  ((n==1) == positiveA),
                        sqrtD  =  d, sqrt2A = 2*a
                      }
-}
        | otherwise = RootOf n p
        where
        u = univariate p
        a = coefficient u 2
        b = coefficient u 1
        c = coefficient u 0
        d = b*b-4*a*c

--        positiveA = substAlgs a sample > 0


-- falsch ?
qepcadGuardedProjections :: RealFrac a
                         => Integer                  -- ^ number of dimensions
                         -> QfFormula (Polynomial a)   -- ^ context formula
                         -> [GuardedPolys a]           -- ^ input polynomials
                         -> [GuardedPolys a]
qepcadGuardedProjections nIndets ctx (gpsIn :: [GuardedPolys a]) =
    assert (all (\p -> usedIndeterminates p == [nIndets]) formulaNums) $
    [GP { gpDim=d, gpGuard=guard, gpPolys=ps }
    | (d,guard,ps) <- zip3 [1..] guards projs
    ]
    where
    guards = map gpGuard $ reviterate (nIndets-1) guardedProjectionOp $
             (uGP { gpGuard = gpGuard uGP `qfAnd` ctx })

    uGP = unionGPs gpsIn

    formulas = map gpGuard gpsIn

    projs, projs1 :: [[Polynomial a]]
    projs = map gcdFree projs1

    projs1 = zipWith del (init geoPss) [0..]
        where
        del ps i =
            [reverseIndeterminates (i+1) $ deleteEndIndeterminates (nIndets-i) p
                 | p <- ps]
    formulaNums :: [Polynomial a]
    formulaNums = last geoPss

    geoPss :: [[Polynomial a]]
    geoPss = [[p | (p,j) <- pAndUsed, i==j] | i <- [0..nIndets]]

    pAndUsed = [(p, last $ usedIndeterminates p)
               | t <- termsQf geoF, let p = prePoly t
               ]
    geoF = qepcadGeoformEtf [ctxG] (FoQf $ disjunctionQf formulasG)

    ctxG = fmap fixIndets ctx
    formulasG = zipWith mkFormulaG formulas [0..]
    mkFormulaG f i =
        fmap fixIndets f
        `QfAnd`
        QfAt (AtEq (indeterminate (nIndets+1) nIndets - fromIntegral i))

    fixIndets = reverseIndeterminates (nIndets+1) . liftPoly



----------------------------------------------
-- Representation for Code Generated by CAD --
----------------------------------------------

data CellBound a = FromNegInfty
                 | ToPosInfty
                 | SectionBound Bool [SectionDef a]
                   deriving (Eq, Show)

data Code a = CFor (CellBound a) (CellBound a) [Code a]
            | CIf  [SectionDef a] [Code a]
            | CStatement Integer Integer
              deriving (Eq, Show)

body :: Code a -> [Code a]
body (CFor _ _ cs) = cs
body (CIf  _   cs) = cs
body _             = []

codeSimpl :: RealFrac a => [Code a] -> [Code a]
codeSimpl (c:rest) 
    | fractionalIf c  =  codeSimpl rest
   -- if no match, fall through
codeSimpl (c1:c2:rest) =
    case codeCompat c1 c2 of
        Right c   -> codeSimpl (c ++ rest)
        Left  c1s -> c1s ++ codeSimpl (c2:rest)
codeSimpl [CFor lb ub cs] =
    nonemptyAppCons (CFor lb ub) (codeSimpl cs) []
codeSimpl [CIf  sd    cs] =
    nonemptyAppCons (CIf  sd   ) (codeSimpl cs) []
codeSimpl rest = rest

--
-- check whether a CIf has a provably non-integral section,
-- i.e., in a program the corresponding conditional can never evaluate
-- to true, because the parameters and variables are always integral.
--
fractionalIf :: RealFrac a => Code a -> Bool
fractionalIf (CIf defs _) = any sectionDefIsNotIntegral defs
fractionalIf _            = False

codeCompat :: RealFrac a => Code a -> Code a -> Either [Code a] [Code a]
codeCompat (CFor lb sb@(SectionBound False sd1) cs1) (CIf sd2 cs2)
    | not (null $ intersect sd1 sd2)
    , Just cs <- matchCodes cs1s cs2s  =
         Right $
         nonemptyAppCons (CFor lb (SectionBound True $ union sd1 sd2)) cs []
    | otherwise  =  Left $ nonemptyAppCons (CFor lb sb) cs1s []
    where
    cs1s = codeSimpl cs1
    cs2s = codeSimpl cs2
codeCompat (CIf sd1 cs1) (CFor sb@(SectionBound False sd2) ub cs2) 
    | not (null $ intersect sd1 sd2)
    , Just cs <- matchCodes cs1s cs2s  =
        Right $
        nonemptyAppCons (CFor (SectionBound True $ union sd1 sd2) ub) cs []
    | otherwise  =  Left $ nonemptyAppCons (CIf sd1) cs1 []
    where
    cs1s = codeSimpl cs1
    cs2s = codeSimpl cs2
codeCompat (CFor lb sb1@(SectionBound incl1 sd1) cs1)
           (CFor sb2@(SectionBound incl2 sd2) ub cs2) 
    | incl1 /= incl2 , not (null $ intersect sd1 sd2)
    , Just cs <- matchCodes cs1s cs2s  =
        Right $ nonemptyAppCons (CFor lb ub) cs []
    | otherwise  =  Left $ nonemptyAppCons (CFor lb sb1) cs1s []
    where
    cs1s = codeSimpl cs1
    cs2s = codeSimpl cs2
codeCompat (CIf  sd1   cs1) c2 =
    Left $ nonemptyAppCons (CIf sd1)    (codeSimpl cs1) []
codeCompat (CFor lb ub cs1) c2 =
    Left $ nonemptyAppCons (CFor lb ub) (codeSimpl cs1) []
codeCompat c1 c2 = Left [c1]

matchCodes :: Num a => [Code a] -> [Code a] -> Maybe [Code a]
matchCodes = mc []
    where
    mc acc [] [] = Just $ reverse acc
    mc acc (c1:cs1) (c2:cs2)
        | Just c <- matchCode c1 c2  =  mc (c:acc) cs1 cs2
    mc acc _ _  =  Nothing

matchCode :: Num a => Code a -> Code a -> Maybe (Code a)
matchCode (CStatement x iters1) (CStatement y iters2)
    | x == y && iters1 == iters2  =  Just (CStatement x iters1)
matchCode (CIf sd1 cs1) (CIf sd2 cs2)
    | not (null sd)
    , Just cs <- matchCodes cs1 cs2  =
          Just $ CIf sd cs
    where
    sd = intersect sd1 sd2
matchCode (CIf sd1 cs1)
  (CFor (SectionBound True sd2l) (SectionBound True sd2u) cs2) 
     | not (null sdl), not (null sdu)
     , Just cs <- matchCodes cs1 cs2  =
         Just $ CFor (SectionBound True sdl) (SectionBound True sdu) cs
     where
     sdl = intersect sd1 sd2l
     sdu = intersect sd1 sd2u
matchCode c1@(CFor _ _ _) c2@(CIf _ _) = matchCode c2 c1
matchCode (CFor (SectionBound incl1l sd1l) (SectionBound incl1u sd1u) cs1)
  (CFor (SectionBound incl2l sd2l) (SectionBound incl2u sd2u) cs2)
    | incl1l == incl2l, incl1u == incl2u, not (null sdl), not (null sdu)
    , Just cs <- matchCodes cs1 cs2 =
           Just $ CFor (SectionBound incl1l sdl) (SectionBound incl1u sdu) cs
    where
    sdl = intersect sd1l sd2l
    sdu = intersect sd1u sd2u
matchCode (CFor FromNegInfty (SectionBound incl1u sd1u) cs1)
  (CFor FromNegInfty (SectionBound incl2u sd2u) cs2)
    | incl1u == incl2u, not (null sdu)
    , Just cs <- matchCodes cs1 cs2 =
           Just $ CFor FromNegInfty (SectionBound incl1u sdu) cs
    where
    sdu = intersect sd1u sd2u
matchCode (CFor (SectionBound incl1l sd1l) ToPosInfty cs1)
  (CFor (SectionBound incl2l sd2l) ToPosInfty cs2)
    | incl1l == incl2l, not (null sdl)
    , Just cs <- matchCodes cs1 cs2 =
           Just $ CFor (SectionBound incl1l sdl) ToPosInfty cs
    where
    sdl = intersect sd1l sd2l
matchCode _ _ = Nothing


statements :: Code a -> [Integer]
statements (CFor _ _ cs) = concatMap statements cs
statements (CIf  _   cs) = concatMap statements cs
statements (CStatement n _) = [n]



--
-- incorrect
--
geogen :: RealFrac a
       => QfFormula (Polynomial a)
       -> [QfFormula (Polynomial a)]
       -> [Code a]
geogen ctx' initialDomains = 
{-
    trace ("ctx' = " ++ show ctx') $
    trace ("initialDomains = " ++ show initialDomains) $
    trace ("exI = " ++ show (reduceEliminateQuantifiers [] (foEx $ disjunctionQf initialDomains))) $
    trace ("ctx = " ++ show ctx) $
    trace ("domains = " ++ show domains) $
    trace ("ex = " ++ show (reduceEliminateQuantifiers [] (foEx $ disjunctionQf domains))) $
-}
    trace ("domains = " ++ show (qfFormula2qepcad (disjunctionQf domains))) $
    trace ("geo = " ++ show (qfFormulaPRE2qepcad geo)) $
    case geo of
        QfAt AtFalse -> []
        _            -> geotree2code $ geo2tree nV geo
    where
    geo = qepcadGeoformEtf [ctx] (FoQf $ disjunctionQf domains)
    ctx = fmap liftAndRev ctx'
    domains = [fmap liftAndRev dom `QfAnd` isI i 
              | (dom,i) <- zip initialDomains [0..]]
    isI i = QfAt (AtEq (domVar - fromIntegral i))
    liftAndRev = reverseIndeterminates (nV+1) . liftPoly

    -- add variable for statement number as last indeterminate
    domVar = indeterminate (nV+1) nV
    nV = unNIndeterminates nI
    nI = nIndeterminatesN $ concatMap termsQf initialDomains

data GeoTree a = GNodes [( AtFormula (TarskiPRE a)
                         , AtFormula (TarskiPRE a)
                         , GeoTree a
                         )
                        ]
               | GStatements [Integer]
                 deriving Show

geotree2code :: RealFrac a
             => GeoTree a -> [Code a]
geotree2code (GStatements stmts) = [CStatement s 0 | s <- stmts]
geotree2code (GNodes nodes) = [gn2code l u inner | (l,u,inner) <- nodes]

gn2code :: RealFrac a
        => AtFormula (TarskiPRE a)
        -> AtFormula (TarskiPRE a)
        -> GeoTree a
        -> Code a
gn2code l u inner =   -- (inner :: GeoTree a) fails with GHC 6.8.2
    CFor lower upper $ geotree2code inner
    where
--    lower, upper :: CellBound a
    lower = mkL l
    upper = mkU u

--    mkU :: AtFormula (TarskiPRE a) -> CellBound a
    mkU AtTrue = ToPosInfty
    mkU at     = mkB at

--    mkL :: AtFormula (TarskiPRE a) -> CellBound a
    mkL AtTrue = FromNegInfty
    mkL at     = mkB at

--    mkB :: AtFormula (TarskiPRE a) -> CellBound a
    mkB (AtEq (PREPoly     p)) = SectionBound True  $ solve p
    mkB (AtEq (PRERoot _ n p)) = SectionBound True  $ rootOf n p
    mkB (AtGe (PREPoly     p)) = SectionBound True  $ solve p
    mkB (AtGe (PRERoot _ n p)) = SectionBound True  $ rootOf n p
    mkB (AtGt (PREPoly     p)) = SectionBound False $ solve p
    mkB (AtGt (PRERoot _ n p)) = SectionBound False $ rootOf n p

--    rootOf :: Integer -> Polynomial a -> [SectionDef a]
    -- rootOf 2 p = Sqrt { ... }
    rootOf n p = [RootOf n $ delRevInds p]

--    solve :: Polynomial a -> [SectionDef a]
    solve p =
--        trace (show (nInds,p,co)) $
        [ConstDef (negate constCo %% linCo)]
        where
        [constCo,linCo] = co
        co =
            coefficients $
            {- swapIndeterminates 0 (nInds-1) -} p'
        p' = delRevInds p
        nInds = unNIndeterminates $ nIndeterminates p'

    delRevInds p =
        reverseIndeterminates (usedInds+1) $
        deleteEndIndeterminates (nInds-usedInds-1) p
        where
        usedInds = last $ usedIndeterminates p
        nInds = unNIndeterminates $ nIndeterminates p


geo2tree :: RealFrac a => Integer -> QfFormula (TarskiPRE a) -> GeoTree a
geo2tree nV (geo :: QfFormula (TarskiPRE a)) =
    g2c 0 (QfAt AtTrue) (positiveDnfQf geo)
  -- nV is without the additional variable for the statement number
  where
  g2c :: Integer
      -> QfFormula (TarskiPRE a)
      -> [[AtFormula (TarskiPRE a)]]
      -> GeoTree a
  g2c dim ctx dnf
      | dim == nV  =
          assert (all ((==1).length) dnf) $
          assert (all ((==1).totalDegree) stmtPolys) $
          assert (all ((==1).denominator) stmts) $
          GStatements [numerator r | r <- stmts]
      | otherwise  =
          GNodes [ (lb,ub, g2c (dim+1) nctx inners)
                 | (cur,inners) <- sorted
                 , let nctx = conjunctionQf (ctx : map QfAt cur)
                 , let (lb,ub) = extractBounds cur
                 ]
      where
      sorted :: [([AtFormula (TarskiPRE a)], [[AtFormula (TarskiPRE a)]])]
      sorted =
          sortBy (\(ats1,_) (ats2,_) -> cmpAtSet dim ctx ats1 ats2) grouped

      grouped :: [([AtFormula (TarskiPRE a)], [[AtFormula (TarskiPRE a)]])]
      grouped =
          [(cur', inners) | (cur,inners) <- grouped1
                          , cur' <- map concat $ cartesian $ map splitZ cur
          ]

      grouped1 :: [([AtFormula (TarskiPRE a)], [[AtFormula (TarskiPRE a)]])]
      grouped1 = xgroupBy atfSetEq p

      p :: [([AtFormula (TarskiPRE a)], [AtFormula (TarskiPRE a)])]
      p = map (partDim dim) dnf

      splitZ at =
          case termsAt at of
            []           -> [[]]
            [PRERoot {}] -> [[at]]
            [PREPoly p ] -> splitzeros dim ctx p at

      stmts = sort $ map (negate . toRational . constantCoeff) stmtPolys
      stmtPolys = map prePoly $ concatMap termsAt $ concat dnf

      extractBounds :: [AtFormula (TarskiPRE a)]
                    -> (AtFormula (TarskiPRE a), AtFormula (TarskiPRE a))
      extractBounds [at@(AtEq pre)] = (at, at)
      extractBounds [at] =
          case classifyBound at of
            False -> (at, AtTrue)
            True  -> (AtTrue, at)
      extractBounds [at1, at2] =
          case (classifyBound at1, classifyBound at2) of
            (False,True) -> (at1, at2)
            (True,False) -> (at2, at1)
            c            -> error ("extractBounds:" ++ show c)
      extractBounds ats = error ("extractBounds: " ++ show ats)

      classifyBound :: AtFormula (TarskiPRE a) -> Bool -- False: lower bound
      classifyBound (AtGt pre) = classify pre
      classifyBound (AtGe pre) = classify pre
      classifyBound at = error ("classifyBound: " ++ show at)

      classify :: TarskiPRE a -> Bool
      -- lower bound, when x >= ...,i.e., +(x-...) >= 0
      classify (PRERoot mp _ _) = not mp
      classify (PREPoly p) =
          assert (degreeIn dim p == 1) $
          not isLower
          where
          isLower = qepcadEtfImpliesQf [ctx] $ QfAt $ AtGt $ PREPoly co
          co = swapIndeterminates 0 dim $
               liftPoly $
               leadingCoeff $
               univariate $ 
               swapIndeterminates 0 dim p

  atfSetEq :: [AtFormula (TarskiPRE a)] -> [AtFormula (TarskiPRE a)] -> Bool
  atfSetEq xs ys = all (`elem` ys) xs && all (`elem` xs) ys
  
  partDim dim ats =
          partition (all ((<= dim) . last . usedIndeterminatesPRE) . termsAt)
                    ats

cmpAtSet :: RealFrac a
         => Integer
         -> QfFormula (TarskiPRE a)
         -> [AtFormula (TarskiPRE a)]
         -> [AtFormula (TarskiPRE a)]
         -> Ordering
cmpAtSet dim ctx ats1 ats2
    | qepcadEtfImpliesQf fullctx ltcond  =  LT
    | otherwise                            =  GT
    where
    fullctx = ctx' : map QfAt (ats1'++ats2')

    ats1' = map (fmap (mapPREpoly (appendIndeterminates 1))) ats1
    ats2' = map (fmap (mapPREpoly (swapIndeterminates dim nInds . appendIndeterminates 1))) ats2
    ctx'  = fmap (mapPREpoly (appendIndeterminates 1)) ctx

    ltcond = QfAt $ AtGt $ PREPoly 
             (indeterminate (nInds+1) nInds - indeterminate (nInds+1) dim)

    nInds = unNIndeterminates $ nIndeterminatesN $
            map prePoly (concatMap termsAt (ats1++ats2) ++ termsQf ctx)

cartesian :: [[a]] -> [[a]]
cartesian []       = [[]]
cartesian (xs:xss) = concat [map (x :) $ cartesian xss | x <- xs]

xgroupBy :: (a -> a -> Bool) -> [(a,b)] -> [(a,[b])]
xgroupBy eq [] = []
xgroupBy eq ((xa,xb):xs) =
    (xa, xb:map snd eqXs) : xgroupBy eq neqXs
    where
    (eqXs,neqXs) = partition (\(a,b) -> a `eq` xa) xs

splitzeros :: RealFrac a
           => Integer
           -> QfFormula (TarskiPRE a)
           -> Polynomial a
           -> AtFormula (TarskiPRE a)
           -> [[AtFormula (TarskiPRE a)]]
splitzeros dim ctx (p :: Polynomial a) origAt
    | degreeIn dim p <= 1  =  [[origAt]]
    | otherwise            =  trace (show (dim,ctx,p,nzeros)) $ sat1
    where
    nzeros = -42 -- howManyRoots dim ctx p

    sat1 = {- simpl -} sat

    sat = [ ats
          | ats <- cands
          , let qf = conjunctionQf $ map QfAt ats
          , qepcadEtfImpliesQf [ctx,qf] $ QfAt origAt
          ]

    cands :: [[AtFormula (TarskiPRE a)]]
    cands = [[AtGt $ PRERoot False 1 p]]   -- x <= _root_1 p
            ++
            concat [[[AtEq $ PRERoot True i p]
                    , [AtGt $ PRERoot True i p, AtGt $ PRERoot False (i+1) p]
                    ]
                   | i <- [1 .. nzeros-1]
                   ]
            ++
            [[AtEq $ PRERoot True nzeros p]
            ,[AtGt $ PRERoot True nzeros p]
            ]
 


code2text :: Integral a => Integer -> [String] -> [Code (Ratio a)] -> String
code2text nP allnames = text allnames []
    where
    text :: Integral a => [String] -> [String] -> [Code (Ratio a)] -> String
    text _ _ [] = ""
    text (n:names) outerNames (CFor lb ub code2 : code) =
        (if genericLength outerNames < nP
         then case (lb, ub) of
               (FromNegInfty, ToPosInfty) -> "if (1) {\n"
               (FromNegInfty, _) ->
                   "if (" ++ n ++ " <= " ++ ubCode ++ ") {\n"
               (_, ToPosInfty) ->
                   "if (" ++ lbCode ++ " <= " ++ n ++ ") {\n"
               (_, _) ->
                   "if (" ++ lbCode ++ " <= " ++ n ++ " && " ++
                   n ++ " <= " ++ ubCode ++ ") {\n"
         else "for (" ++ n ++ "=" ++ lbCode ++ "; " ++
              n ++ "<=" ++ ubCode ++ "; " ++ n ++ "++) {\n"
        ) ++
        indent 2 (text names (n:outerNames) code2) ++
        "}\n" ++
        text (n:names) outerNames code
        where
        lbCode = showBound outerNames False lb
        ubCode = showBound outerNames True  ub
    text (n:names) outerNames (CIf def code2 : code) =
        (if genericLength outerNames < nP
         then if ceilCode == floorCode
              then "if (" ++ n ++ " == " ++ ceilCode ++ ") {\n"
              else "if (" ++ n ++ " == " ++ ceilCode ++ " && " ++
                   n ++ " == " ++ floorCode ++ ") {\n"
         else {- "if ((" ++ n ++ "=" ++ ceilCode ++ ") == " ++
                 floorCode ++ ") {\n"
               -}
              "for (" ++ n ++ "=" ++ ceilCode ++ "; " ++
              n ++ "<=" ++ floorCode ++ "; " ++ n ++ "++) {\n"
        ) ++
        indent 2 (text names (n:outerNames) code2) ++
        "}\n" ++
        text (n:names) outerNames code
        where
        ceilCode  = sectionDef outerNames True  def
        floorCode = sectionDef outerNames False def
    text names outerNames (CStatement i nIters : code) =
        "S" ++ show i ++ "(" ++ iters ++ ")" ++ ";\n" ++
        text names outerNames code
        where
        iters = concat $ intersperse "," $ reverse $
                genericTake nIters outerNames

    indent count s = unlines $ map (ind++) $ lines s
        where
        ind = replicate count ' '

    showBound names _ FromNegInfty = "-oo"
    showBound names _ ToPosInfty   = "oo"
    showBound names isUpperBound (SectionBound False sd) =
        sectionDef names isUpperBound sd ++
        (if isUpperBound then "-1" else "+1")
    showBound names isUpperBound (SectionBound True sd) =
        sectionDef names (not isUpperBound) sd

    sectionDef names ceild [def] = sectionDef1 names ceild def
    sectionDef names ceild defs =
        "OneOf(" ++ concat (intersperse "," $ map (sectionDef1 names ceild) defs) ++ ")"

    sectionDef1 names ceild (ConstDef fp) 
        | Just r <- isConstFP fp =
               show $
               (if ceild then ceiling else floor) $
               r
        | den == 1  =  showPolyC names num
        | otherwise =
            (if ceild then "ceild(" else "floord(") ++
            showPolyC names num ++ "," ++
            showPolyC names den ++ ")"
        where
        ifp = fracPolyToIFracPoly fp
        num = ifpNumerator ifp
        den = ifpDenominator ifp
    sectionDef1 names ceild s@(Sqrt {}) = 
        concat [fn, "(", show a2Den, ".0*((", sp mb, ")/", show mbDen, ".0",
                plusminus, "sqrt((", sp d, ")/", show dDen, ".0)))/(", sp a2, "))"]
        where
        (a2Den, a2) = ratPoly2intPoly (sqrt2A s)
        (mbDen, mb) = ratPoly2intPoly (sqrtMB s)
        (dDen,  d ) = ratPoly2intPoly (sqrtD  s)
        fn = if ceild then "ceil(" else "floor("
        plusminus = if sqrtAddD s then "+" else "-"
        sp = showPolyC names
    sectionDef1 names ceild (RootOf n p) =
        (if ceild then "ceilRootOf(" else "floorRootOf(") ++
        show n ++ "," ++ show (degree p) ++
        concat ["," ++ c i | i <- [0..degree p]] ++ ")"
        where
        c i = showPolyC names $ coefficient (univariate p') i
        (_, p') = ratPoly2intPoly p
{-
timeit :: Show a => a -> IO (NominalDiffTime, a)
timeit x =
    do startTime <- getCurrentTime
       evaluate (length str)
       endTime   <- getCurrentTime
       let diff = diffUTCTime endTime startTime
       print (diff,x)
       return (diff,x)
    where
    str = show x
-}

cad :: RealFrac a => QfFormula (Polynomial a) -> CADStack QAlg a ()
cad (d :: QfFormula (Polynomial a)) =
    genStack (\_ -> CADLeaf ()) (initialTestPoint :: TestPoint QAlg a) projs
    where
    projs = reviterate (gpDim gp - 1) guardedProjectionOp gp
    gp =  domainToGP d


ex1 :: [Polynomial Rational]
ex1 = [x-2, x-4, y, y-3, (y-1)^2+3-x]
    where
    x = indeterminate 2 1
    y = indeterminate 2 0

ex1' :: [QfFormula (Polynomial Rational)]
ex1' = map (conjunctionQf . map (QfAt . AtGe)) $
       [[x-2, -x+4, y, -y+3, (y-1)^2+3-x]
       ]
    where
    x = indeterminate 2 1
    y = indeterminate 2 0

paperEx :: [QfFormula (Polynomial Rational)]
paperEx = map (conjunctionQf . map (QfAt . AtGe)) $
          [[x-10, -x+70, y-10, -y+90, (y-40)^2+1200-30*x]
          ]
    where
    x = indeterminate 2 1
    y = indeterminate 2 0
paperExNames = ["x", "y"]
paperExCode  = code2text 1 paperExNames $ codeSimpl $
               old_cadgen CADInternal (QfAt $ AtTrue) paperEx


ex2' :: [QfFormula (Polynomial Rational)]
ex2' = map (conjunctionQf . map (QfAt . AtGe)) $
       [[(y-1)^2+3-x]
       ]
    where
    x = indeterminate 2 1
    y = indeterminate 2 0


ex3' :: [QfFormula (Polynomial Rational)]
ex3' = map (conjunctionQf . map (QfAt . AtGe)) $
       [[i, j, n-i, m-j]
       ]
    where
    j = indeterminate 4 0
    i = indeterminate 4 1
    m = indeterminate 4 2
    n = indeterminate 4 3
ex3names = ["n", "m", "i", "j"]


exEra  :: [QfFormula (Polynomial Rational)]
exEra = map (conjunctionQf . map (QfAt . AtGe)) $
        [[ii, n-(2*ii+3)^2, jj, n-(2*(2*ii+3)*jj+(2*ii+3)^2)]
        ]
    where
    jj = indeterminate 3 0
    ii = indeterminate 3 1
    n = indeterminate 3 2
exEraNames = ["n", "ii", "jj"]
exEraCode = code2text 1 exEraNames $ old_cadgen CADInternal (QfAt AtTrue) exEra

exEra2  :: [QfFormula (Polynomial Rational)]
exEra2 = map (conjunctionQf . map (QfAt . AtGe)) $
         [[ii, n-(2*ii+3)^2, jj, n-(2*(2*ii+3)*jj+(2*ii+3)^2)]
         ]
    where
    jj = indeterminate 3 1
    ii = indeterminate 3 0
    n = indeterminate 3 2
exEraNames2 = ["n", "jj", "ii"]

exEraCode2 = code2text 1 exEraNames2 $ old_cadgen CADInternal (QfAt AtTrue) exEra2


exEra3  :: [QfFormula (Polynomial Rational)]
exEra3 = map (conjunctionQf . map (QfAt . AtGe)) $
         [[2*ii-3, n-(2*ii)^2, 2*jj-ii, n-4*ii*jj]
         ]
    where
    jj = indeterminate 3 1
    ii = indeterminate 3 0
    n = indeterminate 3 2
exEraNames3 = ["n", "jj", "ii"]

exEraCode3 = code2text 1 exEraNames2 $ codeSimpl $
             old_cadgen CADInternal (QfAt AtTrue) exEra3

exEra4  :: [QfFormula (Polynomial Rational)]
exEra4 = map (conjunctionQf . map (QfAt . AtGe))
         [[i-2, n-i^2, j-i^2, n-j]
         ]
    where
    j = indeterminate 3 1
    i = indeterminate 3 0
    n = indeterminate 3 2
exEraNames4 = ["n", "j", "i"]
exEraCode4 = code2text 1 exEraNames4 $ codeSimpl $
             old_cadgen CADInternal (QfAt AtTrue) exEra4

exEra4k  :: [QfFormula (Polynomial Rational)]
exEra4k = [conjunctionQf (map (QfAt . AtGe) [i-2, n-i^2, k-i, n-i*k] ++
           [QfAt (AtEq (j-k*i))])
          ]
    where
    k = indeterminate 4 0
    i = indeterminate 4 1
    j = indeterminate 4 2
    n = indeterminate 4 3
exEraNames4k = ["n", "j", "i", "k"]
exEraCode4k = code2text 1 exEraNames4k $ codeSimpl $
              old_cadgen CADInternal (QfAt AtTrue) exEra4k


exAZ22 :: [QfFormula (Polynomial Rational)]
exAZ22 = [conjunctionQf (map (QfAt . AtGe) [x-2, n-x, y-4, y-(n-x), n-y]
                         ++
                         map (QfAt . AtEq) [t-((n-3)*x+y)]
                        )
         ]
    where
    y = indeterminate 4 2
    x = indeterminate 4 1
    t = indeterminate 4 0
    n = indeterminate 4 3
{-
    y = indeterminate 4 0
    x = indeterminate 4 1
    t = indeterminate 4 2
    n = indeterminate 4 3
-}
exAZ22Names = ["n", "t", "x", "y"]
exAZ22Code = code2text 1 exAZ22Names $ codeSimpl $
             old_cadgen CADInternal (QfAt $ AtGe (indeterminate 4 3 - 7)) exAZ22



exT2 :: [QfFormula (Polynomial Rational)]
exT2 = map (conjunctionQf . map (QfAt . AtGe)) $
       [[t, p, -t+n, -p+n, t-m*tt, -(t-m*tt)+m-1, p-m*pp, -(p-m*pp)+m-1]
       ]
    where
    p  = indeterminate 6 0
    t  = indeterminate 6 1
    pp = indeterminate 6 2
    tt = indeterminate 6 3
    m  = indeterminate 6 4
    n  = indeterminate 6 5
exT2Ctx = conjunctionQf
          [QfAt $ AtGe (indeterminate 6 4 - 1)
          ,QfAt $ AtGe (indeterminate 6 5 - 1)
          ]
exT2Names = ["n", "m", "tt", "pp", "t", "p"]
exT2Code = code2text 2 exT2Names $ codeSimpl $ old_cadgen CADInternal exT2Ctx exT2

exAZ22fm :: RealFrac a => DTreeGen (FracPoly a) (Polynomial a) (Maybe (SortedSystem (FracPoly a))) 
exAZ22fm = (\sys -> projectToVarsGss sys Nothing Nothing [2,0,1]) $
           fmap (%%1) $
           ineqSystem ((3,0), matrix [[ 1, 0,0,-2]
                                     ,[-1, 0,0, n]
                                     ,[ 0, 1,0,-4]
                                     ,[ 1, 1,0,-n]
                                     ,[ 0,-1,0, n]
                                     ,[n-3,1,-1,0]
                                     ,[-n+3,-1,1,0]
                                     ]
                      )
    where
    n = indeterminate 1 0


exQuil :: [QfFormula (Polynomial Rational)]
exQuil = [conjunctionQf $ map (QfAt . AtGe) [i-1, n-i, j-1, m-j]
         ,conjunctionQf $ map QfAt [AtEq (i-j), AtGe (j-3), AtGe (n-j)]
         ]
    where
    m = indeterminate 4 3
    n = indeterminate 4 2
    i = indeterminate 4 1
    j = indeterminate 4 0
exQuilNames = ["m", "n", "i", "j"]
exQuilCode = code2text 2 exQuilNames $ codeSimpl $
             old_cadgen CADInternal (QfAt AtTrue) exQuil


exTriangle :: [QfFormula (Polynomial Rational)]
exTriangle = [conjunctionQf $ map (QfAt . AtGe) [y-1, x-y, -x+p-y]
         ]
    where
    p = indeterminate 3 2
    x = indeterminate 3 1
    y = indeterminate 3 0
exTriangleNames = ["p", "x", "y"]
exTriangleCode = code2text 1 exTriangleNames $ codeSimpl $
                 old_cadgen CADInternal (QfAt AtTrue) exTriangle


exRect :: [QfFormula (Polynomial Rational)]
exRect = [conjunctionQf $ map (QfAt . AtGe) [y-1, x-1, -x+n, -y+m]
         ]
    where
    m = indeterminate 4 3
    n = indeterminate 4 2
    x = indeterminate 4 1
    y = indeterminate 4 0
exRectNames = ["m", "n", "x", "y"]
exRectCode = code2text 2 exRectNames $ codeSimpl $
             old_cadgen CADInternal (QfAt AtTrue) exRect


exCirc :: [QfFormula (Polynomial Rational)]
exCirc = [QfAt (AtEq (c-(x+y))) `qfAnd` QfAt (AtGe (p-(x^2+y^2)))]
    where
    p = indeterminate 4 3
    c = indeterminate 4 2
    x = indeterminate 4 1
    y = indeterminate 4 0
