{-
    This file is part of the LooPo project, see
    <http://www.infosun.fim.uni-passau.de/cl/loopo/>.

    Copyright (C) 2009 Armin Groesslinger
    Author: Armin Groesslinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-}

-- |
-- An implementation of Chernikova's algorithm for computing
-- a generator reperesentation of a polyhedron given by
-- constraints (a system of inequalities) and vice versa.
--
-- The implementation in this module is based on the paper
-- "A note on Chernikova's Algorithm" by Harvey Le Verge (1994),
-- available at http:\/\/www.irisa.fr\/polylib\/document\/cher.ps.gz
--
-- Author: Armin Groesslinger
--
module LMath.Algo.Chernikova
  ( 
  -- * Chernikova's Algorithm: Duality between Constraints and Generators
    constraintsToGeneratorsG
  , generatorsToConstraintsG
  , constraintsToGenerators
  , generatorsToConstraints

  -- * Simplifying Systems of Inequalities
  , chernikovaSimplifyIneqsG
  , chernikovaSimplifyIneqs

  -- * Sorting Systems of Inequalities (similar to Fourier-Motzkin
  --   Elimination)
  , chernikovaSortIneqsGss
  , chernikovaSortIneqsG
  , chernikovaSortIneqs

  -- * Projections of Systems of Inequalities
  , chernikovaProjectToVarsG
  , chernikovaProjectToVars
    
  -- * Maximising and Minimising a Target Function over a Polyhedron
  , chernikovaMaximiseG
  , chernikovaMinimiseG
  , chernikovaMaximise
  , chernikovaMinimise

  -- * Handling Domains (Unions of Polyhedra)
  , chernikovaSimplifyDomainG
  , chernikovaSimplifyDomain
  )
where

import LMath.Types

import Control.Exception ( assert )
import Control.Monad ( filterM, foldM )
import Data.List
import Data.Maybe ( fromMaybe )

import Debug.Trace ( trace )


-- |
-- Compute the generator representation (lines,rays) of an
-- inequality system from its constraint representation.
--
constraintsToGeneratorsG :: (Fractional a, Num cond)
                         => GIneqSystem a
                         -> DTreeGen a cond ([GVector a], [GVector a])
constraintsToGeneratorsG ineqs =
    do (lines,rays) <- chernikovaG (pos1 : rowVectors (getMatrix ineqs))
       --
       -- Look for a ray with a positive entry in the constant column.
       -- Only when such a ray exists, the system is feasible.
       -- (Otherwise, 1 is not positive...).
       -- We first look for a positive constant, then
       -- we make case distinctions.
       --
       dtc <- getDTreeContext
       let rOnes = map (last . vectorEntries) rays
           rSigs = map (dtcContextImpliesSignum dtc) rOnes
           pos = any signumGt rSigs
           nonConsts = [r | (sgn,r) <- zip rSigs rOnes, not (signumLe sgn)]
       if pos
          then return (lines,rays)
          else do pos' <- anyM (\n -> leCond (negate n) (return False) (return True))
                               nonConsts
                  if pos'
                     then return (lines,rays)
                     else return ([],[])
    where
    -- we add 1 >= 0 to the system
    pos1 = vector (replicate (nVars ineqs + nParams ineqs) 0 ++ [1])

-- |
-- Special case of 'constraintsToGeneratorsG' for non-parametric
-- coefficients.
--
constraintsToGenerators :: (Real a, Fractional a)
                        => GIneqSystem a
                        -> ([GVector a], [GVector a])
constraintsToGenerators sys =
    runNumOrdDTreeGenAlg $ constraintsToGeneratorsG sys

-- |
-- Comptue the constraint representation of an inequality system
-- from its generator representation (lines and rays).
--
generatorsToConstraintsG :: (Fractional a, Num cond, Integral b)
                         => b            -- ^ number of variables
                         -> [GVector a]  -- ^ lines
                         -> [GVector a]  -- ^ rays
                         -> DTreeGen a cond (Maybe (GIneqSystem a))
generatorsToConstraintsG _  [] [] = return Nothing
generatorsToConstraintsG nV lines rays =
    -- we must not add 1 >= 0 (i.e., a ray in the direction of the
    -- constant dimension) to the input
    do (l,r) <- chernikovaG (lines ++ map negate lines ++ rays)
       let r' = filter (any (/=0) . init . vectorEntries) r
           m  = rowVectorsToMatrix (nV+nP+1) (l ++ map negate l ++ r')
       return $ Just $ ineqSystem ((nV,nP), m)
    where
    nP = vectorLength (head (lines++rays)) - nV - 1

-- |
-- Special case of 'generatorsToConstraintsG' for
-- non-parametric coefficients.
--
generatorsToConstraints :: (Real a, Fractional a, Integral b)
                        => b
                        -> [GVector a]
                        -> [GVector a]
                        -> Maybe (GIneqSystem a)
generatorsToConstraints nV lines rays =
    runNumOrdDTreeGenAlg $ generatorsToConstraintsG nV lines rays
    

-- |
-- Apply Chernikova's Algorithm.
-- All dimensions are equal, i.e., to treat one dimension as the
-- constant dimension, a constraint 1 >= 0 has to be present in the input.
--
chernikovaG :: (Fractional a, Num cond)
            => [GVector a]
            -> DTreeGen a cond ([GVector a], [GVector a])
chernikovaG constraints =
    -- Iterate over all constraints.
    -- Note that the enumeration of the constraint indices must
    -- be backwards (from maxIdx to dim) because chernikovaStep
    -- may delete the column with index c if the corresponding
    -- constraint is found to be redundant.
    do (lines,rays) <- foldM step (initLines,[]) $
                       [maxIdx,maxIdx-1 .. dim]
       return (map deleteConstrs lines, map deleteConstrs rays)
    where
    step (lines,rays) c = chernikovaStep dim lines rays c

    --
    -- The initial lines: a line for each of the canonical
    -- base vectors of the space.
    -- We append the respective column of the constraint matrix
    -- to the line. This way, for any line or ray y in the computation,
    -- the value c*y (where c is a constraint) can be found at
    -- index (#c+dim) where #c is the row index of c in the
    -- constraint matrix
    --
    initLines = columnVectors $ (unitMatrix dim dim `join`
                                 rowVectorsToMatrix1 constraints
                                )
    
    dim    = vectorLength $ head constraints  -- dimensionality of the space
    maxIdx = length constraints + dim - 1     -- index of the last constraint

    --
    -- Remove the appended constraint columns from a line or ray.
    --
    deleteConstrs v = subVector v 0 dim


-- |
-- Adapt the sets of lines and rays to the the restrictions imposed by
-- one constraint given by its index c.
-- The indices c must come in descending order (in successive calls to
-- chernikovaStep), because line\/ray entries at index c may be deleted
-- when the constraint at index c is redundant.
--
chernikovaStep :: (Fractional a, Num cond)
               => Int          -- ^ dimensionality of the space
               -> [GVector a]  -- ^ lines (extended with constraint columns)
               -> [GVector a]  -- ^ rays  (extended with constraint columns)
               -> Int          -- ^ constraint to handle
               -> DTreeGen a cond ([GVector a]  -- new lines
                                  ,[GVector a]  -- new rays
                                  )
chernikovaStep dim lines (rays :: [GVector a]) c =
    do mbGtLine <- findGtLine lines c
       case mbGtLine of
         Just (gtLine,otherLines) ->
             -- if there is a line L which does not saturate the constraint,
             -- the positive part of the line L becomes a ray and
             -- the other lines are projected onto the hyperplane defined
             -- by the constraint along the line L.
             return (newLines, newRays)
             where
             newLines = map project otherLines
             newRays  = gtLine : map project rays
             c_gtLine = vectorEntry gtLine c
             project y = y - ((vectorEntry y c / c_gtLine) `mulSV` gtLine)
         Nothing ->
             do -- partition the set of rays into the rays which violate
                -- the constraint at c (neg),
                -- saturate it (zero), and verify it (pos)
                (neg,zero,pos) <- partitionBySign rays c
                -- if no ray violates the constraint at c, then the
                -- constraint is redundant w.r.t. the constraints
                -- at [c+1 .. maxIdx] and we delete it.
                if null neg then
                   return (map deleteC lines, map deleteC (pos++zero))
                 else
                   do -- compute which constraints are saturated by each ray
                      satRows <- mapM s rays
                      -- compute positive combinations of the rays
                      -- verifying the constraint and the rays violating
                      -- the constraints (such that the combination
                      -- saturates the constraint).
                      let combs = [(y1,y2,combine y1 y2) | y1<-pos, y2<-neg]
                      qbar1 <- -- trace (show (length pos,length neg)) $
                               filterM (adjacent satRows) combs
                      -- qbar constains exactly the extremal rays to add
                      -- to the set of rays
                      let qbar = [y | (y1,y2,y) <- qbar1]
                      return $ (lines, pos++zero++qbar)
    where
    t      = length lines    -- the number of lines
    -- index of last constraint (never used if  lines++rays==[])
    maxIdx = vectorLength (head (lines ++ rays)) - 1 

    -- remove constraint at index c from line/ray
    deleteC v = deleteEntries v c 1

    -- Compute a positive combination of y1 and y2 such that
    -- the combination saturates the constraint C at index c.
    -- (Note that y1*C >0 and y2*C < 0.)
    combine y1 y2 =
        y1 - ((vectorEntry y1 c / vectorEntry y2 c) `mulSV` y2)

    --
    -- Check whether y1 and y2 are adjacent, i.e., whether their positive
    -- combination y is an extremal ray.
    -- y is an extremal ray, if and only if, there is no y' in rays
    -- (y' /= y1, y' /= y2) such that y' saturates all the constraints
    -- that y saturates (and possibly more).
    -- Le Verge's criterion (to exclude some y immediately) is that
    -- if y satisfies at most dim-t-3
    -- constraints, then y is not an extremal ray.
    --
    adjacent :: Num cond
             => [[Int]]
             -> (GVector a,GVector a,GVector a)
             -> DTreeGen a cond Bool
    adjacent satRows (y1,y2,y) =
        do sy <- s y
           if length sy <= dim-t-3 then  -- Le Verge's criterion
             return False
            else
             return $ all (\sa -> not $ null (sy\\sa)) sat -- general criterion
        where
        -- constraints saturated by each ray (apart from y1, y2)
        sat = [sa | (y',sa) <- zip rays satRows, y' /= y1, y' /= y2]

    -- Return the list of constraints (represented by their indices)
    -- which y saturates.
    s y = filterM f [c+1..maxIdx]
        where
        f i = eqCond (vectorEntry y i) (return True) (return False)


-- |
-- Find a line which does not saturate the given constraint.
-- Return the line (represented by a vector which verifies the
-- constraint) and all the other lines, if such a line exists.
--
findGtLine :: (Fractional a, Num cond)
            => [GVector a]    -- lines (extended with constraint columns)
            -> Int            -- constraint index
            -> DTreeGen a cond (Maybe (GVector a,[GVector a]))
findGtLine lines c = f lines []
    where
    f [] _ = return Nothing
    f (l:ls) rest =
        sCond (vectorEntry l c)
              (return $ Just (negate l,ls++rest)) -- -l verifies the constr.
              (f ls (l:rest))                     --  l saturates the constr.
              (return $ Just (l, ls++rest))       --  l verifies the constr.

-- |
-- Partition the given rays by their sign w.r.t. the given
-- constraint (represented by its index).
--
partitionBySign :: (Fractional a, Num cond)
                => [GVector a]  -- rays (extended with constraint columns)
                -> Int          -- constraint index
                -> DTreeGen a cond ([GVector a]  -- rays violating the constr.
                                   ,[GVector a]  -- rays saturating the constr.
                                   ,[GVector a]  -- rays verifying the constr.
                                   )
partitionBySign [] c = return ([],[],[])
partitionBySign (r:rays) c =
    do (neg,zero,pos) <- partitionBySign rays c
       sCond (vectorEntry r c)
             (return (r:neg,   zero,   pos))
             (return (  neg, r:zero,   pos))
             (return (  neg,   zero, r:pos))


-----------------------------------------
-- Simplifying Systems of Inequalities --
-----------------------------------------


-- |
-- Simplify a given inequality system by computing
-- its generator representation and a minimal
-- constraint representation from the generators.
-- Returns @Nothing@ if the system is found to be
-- infeasible.
--
chernikovaSimplifyIneqsG :: (Fractional a, Num cond)
                         => Maybe (GIneqSystem a) -- ^ context incl. vars
                         -> GIneqSystem a
                         -> DTreeGen a cond (Maybe (GIneqSystem a))

chernikovaSimplifyIneqsG mbCtx sys =
    assert (maybe True (\ctx -> nVars ctx  ==nVars   sys) mbCtx) $
    assert (maybe True (\ctx -> nParams ctx==nParams sys) mbCtx) $
    -- Chernikova generates case distinctions which can carry
    -- the same result, so try to reduce the number of case
    -- distinctions here.
    combineSubtreesDTreeG' mix $
    do (lines,rays) <- constraintsToGeneratorsG sys'
       mbSys <- generatorsToConstraintsG nV lines rays
       case (mbSys,mbCtx) of
         (_, Nothing) -> return mbSys
         (Nothing, _) -> return Nothing
         (Just nSys, Just ctx) ->
             do rs <- filterM (filt ctx) $ rowVectors $ getMatrix nSys
                return $ Just $ ineqSystem ((nV,nP)
                                           , rowVectorsToMatrix (nV+nP+1) rs
                                           )
    where
    nV = nVars sys
    nP = nParams sys

    mix (Just is1) (Just is2)
        | getMatrix is1 == getMatrix is2  =  Just (Just is1)
    mix _           _                     =  Nothing

    sys' = case mbCtx of
             Just ctx -> sys `join` ctx
             Nothing  -> sys

    filt ctx v =
        do dtc <- getDTreeContext
           return $ not $ any (eqV dtc) $ rowVectors $ getMatrix ctx
        where
        eqV dtc r
            | r == v     =  True
            | otherwise  =
                  case vrcs of
                    (vc,rc):_ -> (rc `mulSV` v) == (vc `mulSV` r)
                    []        -> False
            where
            vrcs = [(vc,rc)
                   | (vc,rc) <- zip (vectorEntries v) (vectorEntries r)
                   , signumGt $ dtcContextImpliesSignum dtc (vc*rc)
                   ]
--        return $ find (==v) (rowVectors $ getMatrix ctx) == Nothing 


anyM :: Monad m => (a -> m Bool) -> [a] -> m Bool
anyM f [] = return False
anyM f (x:xs) =
    do b <- f x
       if b then return True else anyM f xs
           

-- |
-- Special case of 'chernikovaSimplifyIneqsG' for non-parametric
-- coefficients.
--
chernikovaSimplifyIneqs :: (Real a, Fractional a)
                        => Maybe (GIneqSystem a)
                        -> GIneqSystem a
                        -> Maybe (GIneqSystem a)
chernikovaSimplifyIneqs mbCtx sys =
    runNumOrdDTreeGenAlg $ chernikovaSimplifyIneqsG mbCtx sys


-- |
-- Remove bounds from a system which are redundant w.r.t. a given
-- context (in the variables and linear parameters).
-- Only simplifications w.r.t. the context are performed,
-- i.e., the system is not simplified w.r.t. itself.
--
chernikovaSimplifyWrtContextG :: (Fractional a, Num cond)
    => GIneqSystem a  -- ^ context (in variables and lin. parameters)
    -> GIneqSystem a  -- ^ system to simplify
    -> DTreeGen a cond (GIneqSystem a)
chernikovaSimplifyWrtContextG ctx sys =
    do (lines,rays) <- constraintsToGeneratorsG ctx
       chernikovaSimplifyWrtGeneratorsG lines rays sys

-- |
-- Remove bounds from a system which are redundant w.r.t. a given
-- context (in the variables and linear parameters). The context
-- is given in generator representation, i.e.,
-- defined by its lines and rays.
-- Only simplifications w.r.t. the context are performed,
-- i.e., the system is not simplified w.r.t. itself.
--
chernikovaSimplifyWrtGeneratorsG :: (Fractional a, Num cond)
    => [GVector a]    -- ^ lines of the context
    -> [GVector a]    -- ^ rays of the context
    -> GIneqSystem a  -- ^ system to simplify
    -> DTreeGen a cond (GIneqSystem a)
chernikovaSimplifyWrtGeneratorsG lines rays sys =
    do resRows <- filterM (elim lines rays) $ rowVectors $ getMatrix sys
       return $ ineqSystem ((nV,nP), rowVectorsToMatrix (nV+nP+1) resRows)
    where
    nV = nVars sys
    nP = nParams sys

    -- bound' is redundant w.r.t. the context, if
    -- its minimal value is >= 0.
    -- So, it is redundant, if bound's maximal value is <= 0
    elim lines rays bound' =
        do opt <- chernikovaMaximiseWrtGeneratorsG lines rays target
           case opt of
             Unbounded     -> return True  -- not redundant
             Nonexistent   -> return False -- redundant
             Optimum (v,_) ->
                 leCond (v+addconst)
                        (return False)  -- redundant
                        (return True)   -- not redundant
        where
          bound = negate bound'
          target = deleteEntries bound (nV+nP) 1
          addconst = vectorEntry bound (nV+nP)


-------------------------------------------------------------
-- Sorting inequality systems using Chernikova's algorithm --
-------------------------------------------------------------


-- |
-- Sort a system of inequalities (with optional equalities)
-- in a manner similar to Fourier-Motzkin elimination.
--
chernikovaSortIneqsGss :: (Fractional a, Integral b, Num cond)
                       => GIneqSystem a        -- ^ inequalities
	               -> Maybe (GEqSystem a)  -- ^ optional equalities
	               -> Maybe b              -- ^ number of "outer" variables
	               -> DTreeGen a cond (Maybe (SortedSystem a))
chernikovaSortIneqsGss (sys' :: GIneqSystem a) mbEqs mbOuterVars =
    do generators <- constraintsToGeneratorsG sys
       case generators of
         ([],[]) -> return Nothing
         _       -> do ss <- proj varsToElim generators
                       return $ Just ss
    where
    sys = (\s -> reorderVars s [nV-1,nV-2 .. 0]) $
          joinMbEqualities sys' mbEqs

    outerVars  = fromMaybe 0 mbOuterVars
    varsToElim = nV - outerVars

    nV = nVars sys'
    nP = nParams sys'
    
    proj 0   _            = return []
    proj vte (lines,rays) =
        do Just constrs <- generatorsToConstraintsG (outerVars+vte) lines rays
           currbnds <- extractBounds $ rowVectors $ getMatrix constrs
           let l' = projRays lines
               r' = projRays rays
           bnds <- proj (vte-1) (l',r')
           return (bnds++[currbnds])

    projRays rs = filter (any (/=0) . vectorEntries) $
                  map (\v -> deleteEntries v 0 1) rs

    extractBounds :: Num cond
                  => [GVector a] -> DTreeGen a cond (Bounds (GVector a))
    extractBounds constrs =
        do (neg,zero,pos) <- partitionBySign constrs 0
           return $ LowerUpper (map divByHead pos) (map divByHead neg)

    divByHead v = fmap (\e -> e / h) $ deleteEntries v 0 1
        where
          h = negate $ vectorEntry v 0

-- |
-- Sort a given system of (in)equalities. Returns a
-- sorted inequality system.
--
chernikovaSortIneqsG :: (Fractional a, Integral b, Num cond)
                     => GIneqSystem a        -- ^ inequalities
	             -> Maybe (GEqSystem a)  -- ^ optional equalities
	             -> Maybe b              -- ^ number of "outer" variables
	             -> DTreeGen a cond (Maybe (GIneqSystem a))
chernikovaSortIneqsG sys mbEqs mbOuterVars =
    do mbSS <- chernikovaSortIneqsGss sys mbEqs mbOuterVars
       return $ fmap (sortedSystemToIneqSystem nV nP) mbSS
    where
    nV = nVars sys
    nP = nParams sys

-- |
-- Sort a given system of (in)equalities. Returns a
-- sorted inequality system.
-- (Special case for non-parametric coefficients.)
--
chernikovaSortIneqs :: (Real a, Fractional a)
                    => GIneqSystem a
                    -> Maybe (GEqSystem a)
                    -> Maybe (GIneqSystem a)
chernikovaSortIneqs sys mbEqs =
    runNumOrdDTreeGenAlg $ chernikovaSortIneqsG sys mbEqs Nothing


-- |
-- Project a system of inequalities to given dimensions.
--
chernikovaProjectToVarsG :: (Fractional a, Num cond, Integral b)
                         => GIneqSystem a          -- ^ inequalities
                         -> Maybe (GEqSystem a)    -- ^ optional equalities
                         -> [b]                    -- ^ dimensions
                         -> DTreeGen a cond (Maybe (GIneqSystem a))
chernikovaProjectToVarsG sys' mbEqs vars =
    assert (all (\v -> v>=0 && v<nV) vars) $
    do (lines,rays) <- constraintsToGeneratorsG sys
       let ls = project lines
           rs = project rays
       generatorsToConstraintsG toKeep ls rs
    where
    sys = reorderVars (joinMbEqualities sys' mbEqs) vars
    nV  = nVars sys
    toKeep = genericLength vars
    toDel  = nV - toKeep

    project vs = filter (any (/=0) . vectorEntries) $ map del vs
    del v      = deleteEntries v toKeep toDel

-- |
-- Special case of 'chernikovaProjectToVarsG' for
-- non-parametric coefficients.
--
chernikovaProjectToVars :: (Real a, Fractional a, Integral b)
                        => GIneqSystem a
                        -> Maybe (GEqSystem a)
                        -> [b]
                        -> Maybe (GIneqSystem a)
chernikovaProjectToVars sys mbEqs vars =
    runNumOrdDTreeGenAlg $ chernikovaProjectToVarsG sys mbEqs vars

-- |
-- Maximise a target function w.r.t. a polyhedron.
-- Note that the system of inequalities (which defined the
-- polyhedron) must _not_ have linear parameters.
--
chernikovaMaximiseG :: (Fractional a, Num cond)
                    => GIneqSystem a
                    -> Maybe (GEqSystem a)
                    -> GVector a      -- ^ target function (in variables only)
                    -> DTreeGen a cond (Optimum (a,[a]))
chernikovaMaximiseG sys' mbEqs (target :: GVector a) =
    assert (nParams sys' == 0) $
    assert (vectorLength target == nV) $
    do (lines,rays) <- constraintsToGeneratorsG sys
       chernikovaMaximiseWrtGeneratorsG lines rays target
    where
    sys = joinMbEqualities sys' mbEqs
    nV  = nVars sys'

chernikovaMaximiseWrtGeneratorsG :: (Fractional a, Num cond)
     => [GVector a]  -- ^ lines
     -> [GVector a]  -- ^ rays
     -> GVector a    -- ^ target function (in variables only)
     -> DTreeGen a cond (Optimum (a,[a]))
chernikovaMaximiseWrtGeneratorsG [] []      (target :: GVector a) =
    return Nonexistent
chernikovaMaximiseWrtGeneratorsG lines rays (target :: GVector a) =
    assert (vectorLength (head (lines++rays)) == vectorLength target + 1) $
    checkLines lines
    where
    checkLines :: Num cond => [GVector a] -> DTreeGen a cond (Optimum (a,[a]))
    checkLines []     = checkRays Nothing rays
    checkLines (l:ls) =
        eqCond (initV l `mulVV` target)
               (checkLines ls)
               (return Unbounded)

    checkRays :: Num cond 
              => Maybe (a,GVector a) -> [GVector a]
              -> DTreeGen a cond (Optimum (a,[a]))
    checkRays Nothing           [] = return Nonexistent
    checkRays (Just (v,vertex)) [] = return $ Optimum (v, vectorEntries vertex)
    checkRays mbValVertex (r:rs)   =
        eqCond h
               (geCond (negate (ray `mulVV` target)) -- ray
                       (checkRays mbValVertex rs)    -- target*ray <= 0
                       (return Unbounded)            -- target*ray > 0
               )
               (case mbValVertex of                  -- vertex
                  Nothing           -> checkRays mbValVertex' rs
                  Just (val,vertex) ->
                      geleCond (val-val')                
                               (checkRays mbValVertex  rs) -- val >= val'
                               (checkRays mbValVertex' rs) -- val <= val'
                      
               )
        where
        mbValVertex' = Just (val',vertex')
        val'    = vertex' `mulVV` target
        vertex' = fmap (/h) ray
        ray     = initV r
        h       = lastV r

    initV v = deleteEntries v (vectorLength v-1) 1
    lastV v = last $ vectorEntries v

-- |
-- Minimise a target function w.r.t. a polyhedron.
-- Note that the system of inequalities (which defined the
-- polyhedron) must _not_ have linear parameters.
--
chernikovaMinimiseG :: (Fractional a, Num cond)
                    => GIneqSystem a
                    -> Maybe (GEqSystem a)
                    -> GVector a      -- ^ target function (in variables only)
                    -> DTreeGen a cond (Optimum (a,[a]))
chernikovaMinimiseG sys mbEqs (target :: GVector a) =
    do res <- chernikovaMaximiseG sys mbEqs (fmap negate target)
       return $ fmap (\(v,xs) -> (negate v, xs)) res

-- |
-- Maximise a target function w.r.t. to a non-parametric
-- polyhedron (no linear parameter, either).
--
chernikovaMaximise :: (Real a, Fractional a)
                   => GIneqSystem a
                   -> Maybe (GEqSystem a)
                   -> GVector a
                   -> Optimum (a, [a])
chernikovaMaximise sys mbEqs target =
    runNumOrdDTreeGenAlg $ chernikovaMaximiseG sys mbEqs target

-- |
-- Minimise a target function w.r.t. to a non-parametric
-- polyhedron (no linear parameter, either).
--
chernikovaMinimise :: (Real a, Fractional a)
                   => GIneqSystem a
                   -> Maybe (GEqSystem a)
                   -> GVector a
                   -> Optimum (a, [a])
chernikovaMinimise sys mbEqs target =
    runNumOrdDTreeGenAlg $ chernikovaMinimiseG sys mbEqs target



----------------------------------
-- Domain (Unions of Polyhedra) --
----------------------------------

-- |
-- Simplify a domain.
-- Note: At the moment, the simplification is
-- suboptimal, it applies chernikovaSimplifyIneqsG to each
-- disjunct. No simplification w.r.t. several disjunctions
-- is performed.
--
chernikovaSimplifyDomainG :: (Fractional a, Num cond)
                          => Maybe (GIneqSystem a)   -- ^ context with vars
                          -> [GIneqSystem a]
                          -> DTreeGen a cond [GIneqSystem a]
chernikovaSimplifyDomainG mbCtx syss =
    assert (length (nub $ map nVars   syss) == 1) $
    assert (length (nub $ map nParams syss) == 1) $
    assert (maybe nV nVars   mbCtx == nV) $
    assert (maybe nP nParams mbCtx == nP) $
    do syss' <- mapM (chernikovaSimplifyIneqsG mbCtx) syss
       let syss2 = [sys | Just sys <- syss']
       gens  <- mapM constraintsToGeneratorsG syss2
       res   <- simpl $ zip gens syss2
       syss3 <- mapM (\((l,r),_) -> generatorsToConstraintsG nV l r) res
       let simpl = [sys | Just sys <- syss3]
       return {- $ trace ("domsimpl: " ++ show (mbCtx,syss,simpl)) -} simpl
    where
    nV = nVars   $ head syss
    nP = nParams $ head syss

    simpl []     = return []
    simpl (x:xs) = simpl2 x xs

    simpl2 x [] = return [x]
    simpl2 x@((l1,r1),s1) ys@(y@((l2,r2),s2) : rest) =
        do s2' <- chernikovaSimplifyWrtGeneratorsG l1 r1 s2
           if nRows s2' == 0
              then simpl ys
              else do s1' <- chernikovaSimplifyWrtGeneratorsG l2 r2 s1
                      if nRows s1' == 0
                         then simpl2 x rest 
                         else do zs <- simpl2 x rest
                                 return (y:zs)

{-
    do syss' <- mapM (chernikovaSimplifyIneqsG mbCtx) syss
       let syss'' = [sys | Just sys <- syss']
       -- Check if there is a disjunct which is the universe
       -- polyhedron. If so, return the universe polyhedron only.
       case filter ((==0) . nRows) syss'' of
         []  -> return syss''
         u:_ -> return [u]
-}
    


chernikovaSimplifyDomain :: (Real a, Fractional a)
                         => Maybe (GIneqSystem a)
                         -> [GIneqSystem a]
                         -> [GIneqSystem a]
chernikovaSimplifyDomain mbCtx syss =
    runNumOrdDTreeGenAlg $ chernikovaSimplifyDomainG mbCtx syss
