-- |
-- Functions to count integral points in unions of polyhedra,
-- extending the functionality in 'LMath.Types.Ehrhart'.
--
module LMath.Algo.CountPolyhedra
  (
    lexicographicIndex, 
    -- * just for testing, do not use
    lexicographicIndex2
  )
where

import LMath.Types
import LMath.Algo.EhrhartArith
import LMath.Algo.OmegaTest
import LMath.Algo.Sort

import Data.List
import Debug.Trace


{-
-- |
-- Count how many integral points in a polyhedron precede a given
-- reference point in the specified dimensions.
--
lexicographicIndex1 :: RealFrac a
    => Bool                 -- ^ assume reference point is in the given system
    -> [Integer]                -- ^ selected dimensions
    -> Maybe (GIneqSystem a)    -- ^ optional context (including variables)
    -> GIneqSystem a            -- ^ the system to count
    -> [(GIneqSystem a, EValue a)]
lexicographicIndex1 assumeInside selectedDims mbCtx0 sys0 =
    cps
    where
    -- add parameters for the reference point (at front)
    sys = insertZeroParams (reorderVars sys0 selectedDims) 0 nSelDims

    -- copy the restrictions on the selected variables
    -- to the newly introduced parameters
    sysP = foldl (\s d -> insertZeroVars s d 1)
                 (moveVarsToParams (reorderVars sys0 selectedDims)
                                   0 nSelDims 0) 
                 selectedDims
    nV = nVars   sys0
    nP = nParams sys0
    nSelDims = genericLength selectedDims

    --
    -- Compute a tagged disjoint union of the conditions computed
    -- in cps1; since no duplicate tags are returned, each tagged
    -- partition contains at most one Ehrhart polynomial for
    -- each lexCond d.
    --
    cps = [(cond, sum ps)
          | (cond, xpolys) <- taggedDisjointUnion Nothing (q++cps1),
            let ps = [condneg d p | X d p <- xpolys]
          ]          
    condneg d p = if odd d then negate p else p
    -- substract 1 inside the system (the "1" will be negated
    -- by condneg due to odd (-1) )
    q = [(s, [X (-1) 1]) | odd nSelDims, Just s <- [projectToVars sysP Nothing []], trace ("s: " ++ show s) True]

    --
    -- Count the integral points in the system restricted to
    -- lexCond d  for all dimensions d and tag each result with
    -- the dimension d.
    --
    cps1 = [(cond, [X d ehp])
           | d <- [0..nSelDims-1],
             let fullsys = lexCond d `join` sys,
             (cond,ehp) <- countIntegralPoints [fullsys] mbCtxP
           ]
    --
    -- project the context to the parameter dimensions
    -- (countIntegralPoint takes a context in the parameters only)
    --
    mbCtxP = mbCtx >>= \s -> projectToVars s Nothing []

    --
    -- insert the parameters for the reference point into the context
    -- and add the 'assume inside' assumptions, if appropriate
    --
    mbCtx = case (mbCtx0, assumeInside) of
               (Nothing, False)  -> Nothing
               (Nothing, True)   -> Just sysP
               (Just ctx, False) -> Just $ insertZeroParams ctx 0 nSelDims
               (Just ctx, True)  -> Just (sysP `join` insertZeroParams ctx 0 nSelDims)

    --
    -- lexicographic condition in dimension d
    --
    lexCond 0 = ineqSystem ((nV,nSelDims+nP), leCond 0)
    lexCond d = -- foldl1 (+) leConds  is undefined for d==0
      ineqSystem ((nV,nSelDims+nP),
                  multijoin (geCond d) (negate (foldl1 (+) leConds) : leConds)
                 )
        where
        leConds = [leCond i | i <- [0..d-1]]
       
    -- v <= refpoint in dimension d
    leCond d = matrix [put nV d (-1) ++ put (nSelDims+nP) d 1 ++ [0]]
    geCond d = fmap negate $ leCond d

    -- v < refpoint in dimension d
    ltCond d = matrix [put nV d (-1) ++ put (nSelDims+nP) d 1 ++ [-1]]
    gtCond d = fmap negate $ ltCond d
-}

--
-- put 'value' at 'pos' (0-based) into a list of length 'len',
-- other entries 0.
--
put :: Num a => Integer -> Integer -> a -> [a]
put len pos value =
    genericReplicate pos 0 ++ [value] ++ genericReplicate (len-pos-1) 0


-- |
-- Count how many integral points in a union of polyhedra precede a given
-- reference point in the specified dimensions.
-- Returns a list of @(condition,polynomial)@
-- pairs where @polynomial@ evaluates to the number of integral
-- points in the given union that are lexicographically less 
-- than (strict less than) a reference point. To specify the
-- reference point new parameters (as leading parameters, i.e.,
-- before the other parameters) are introduced (the @condition@s
-- and @polynomial@s have these additional parameters).
-- If the reference point is always chosen from the given polyhedron,
-- a simpler description can be computed. For example, for the system
-- @
--    0 <= x <= p
-- @
-- under the assumption @p >= 0@, the result is
-- @
--    [(xx <= p, xx), (xx >= p+1, p+1)]
-- @
-- without assuming that the reference point (@xx@) is from the
-- given system (i.e., @xx >= p+1@ is possible). With the assumption
-- that @xx@ is in the given polyhedron, only the case @xx <= p@
-- is returned.
-- The @condition@s may overlap, if there are cases which can be
-- expressed with different polynomials.
--
lexicographicIndex :: RealFrac a
    => Bool                 -- ^ assume reference point is in the given system
    -> [Integer]                -- ^ selected dimensions
    -> Maybe (GIneqSystem a)    -- ^ optional context (including variables)
    -> [GIneqSystem a]          -- ^ the union to count
    -> EPartition a
lexicographicIndex _ _ _ [] = epartition []
lexicographicIndex assumeInside selectedDims mbCtx syss0 =
    epartition cps
    where
    -- add parameters for the reference point (at front)
    syss = map (\s -> insertZeroParams (reorderVars s selectedDims) 0 nSelDims) syss0

    -- copy the restrictions on the selected variables
    -- to the newly introduced parameters
    syssP = map (\s -> foldl (\s d -> insertZeroVars s d 1)
                             (moveVarsToParams (reorderVars s selectedDims)
                                               0 nSelDims 0)
                             selectedDims
                )
                syss0
    syssPonly = [s | Just s <- [projectToVars s Nothing [] | s <- syssP]]
    nV = nVars   $ head syss0
    nP = nParams $ head syss0
    nSelDims = genericLength selectedDims

    cps = [(cond, ehp)
          | (cond1,ehp) <- epPartitions $ countIntegralPoints lexSystems mbCtxP,
            cond <- if assumeInside then simplifyDomain syssPonly [cond1] else [cond1],
            omegaTest (Just cond) Nothing
          ]

    lexSystems = concat [map (`join` lexCond d) syss | d <- [0..nSelDims-1]]

    --
    -- project the context to the parameter dimensions
    -- (countIntegralPoint takes a context in the parameters only)
    --
    mbCtxP = mbCtx >>= \s ->
             projectToVars (insertZeroParams s 0 nSelDims) Nothing []

    --
    -- lexicographic condition in dimension d
    --
    lexCond 0 = ineqSystem ((nV,nSelDims+nP), ltCond 0)
    lexCond d = -- foldl1 (+) leConds  is undefined for d==0
      ineqSystem ((nV,nSelDims+nP),
                  multijoin (ltCond d) (negate (foldl1 (+) leConds) : leConds)
                 )
        where
        leConds = [leCond i | i <- [0..d-1]]
       
    -- v <= refpoint in dimension d
    leCond d = matrix [put nV d (-1) ++ put (nSelDims+nP) d 1 ++ [0]]
    geCond d = fmap negate $ leCond d

    -- v < refpoint in dimension d
    ltCond d = matrix [put nV d (-1) ++ put (nSelDims+nP) d 1 ++ [-1]]
    gtCond d = fmap negate $ ltCond d

lexicographicIndex2 :: RealFrac a
    => Bool                 -- ^ assume reference point is in the given system
    -> [Integer]                -- ^ selected dimensions
    -> Maybe (GIneqSystem a)    -- ^ optional context (including variables)
    -> [GIneqSystem a]          -- ^ the system to count
    -> EPartition a
lexicographicIndex2 _ _ _ [] = epartition []
lexicographicIndex2 assumeInside selectedDims mbCtx syss0 =
    cps posDims
    -- cps1
    where
    -- add parameters for the reference point (at front)
    syss = map (\s -> insertZeroParams (reorderVars s selectedDims) 0 nSelDims) syss0

    -- copy the restrictions on the selected variables
    -- to the newly introduced parameters
    syssP = map (\s -> foldl (\s d -> insertZeroVars s d 1)
                             (moveVarsToParams (reorderVars s selectedDims)
                                               0 nSelDims 0)
                             selectedDims
                )
                syss0
    syssPonly = [s | Just s <- [projectToVars s Nothing [] | s <- syssP]]
    nV = nVars   $ head syss0
    nP = nParams $ head syss0
    nSelDims = genericLength selectedDims

    cps1 = -- trace ("cpsP=" ++ show pos ++ "\ncpsN=" ++ show neg)$
           pos + negate neg
    pos = cps posDims
    neg = cps negDims

    (posDims, negDims) 
        | odd  nSelDims  =  ([0,2..nSelDims-1], [-1,1..nSelDims-1])
        | even nSelDims  =  ([1,3..nSelDims-1], [0,2..nSelDims-1])

    -- should we simplify _after_ computing the disjoint union
    -- (in addCEv), because simplifying first reduces the number
    -- of inequalities and, hence, creates more possibilities
    -- for overlap in the union (outside the context)?
    cps dims =
       epartition
          [(cond, ehp)
          | (cond,ehp) <- epPartitions $ countIntegralPoints ((\x -> trace (show x) x) $ concatMap lexSystems dims) mbCtxP,
--            cond <- if assumeInside then simplifyDomain syssPonly [cond1] else [cond1],
            omegaTest (Just cond) Nothing
          ]

    lexSystems d = map (lexCond d `join`) syss

    --
    -- project the context to the parameter dimensions
    -- (countIntegralPoint takes a context in the parameters only)
    --
    mbCtxP = mbCtx >>= \s ->
             projectToVars (insertZeroParams s 0 nSelDims) Nothing []

    --
    -- lexicographic condition in dimension d
    --
    lexCond (-1) = ineqSystem ((nV,nSelDims+nP), zeroMatrix 0 (nV+nSelDims+nP+1))
    lexCond 0 = ineqSystem ((nV,nSelDims+nP), geCond 0)
    lexCond d = -- foldl1 (+) leConds  is undefined for d==0
      ineqSystem ((nV,nSelDims+nP),
                  multijoin (geCond d) (negate (foldl1 (+) leConds) : leConds)
                 )
        where
        leConds = [leCond i | i <- [0..d-1]]
       
    -- v <= refpoint in dimension d
    leCond d = matrix [put nV d (-1) ++ put (nSelDims+nP) d 1 ++ [0]]
    geCond d = fmap negate $ leCond d

    -- v < refpoint in dimension d
    ltCond d = matrix [put nV d (-1) ++ put (nSelDims+nP) d 1 ++ [-1]]
    gtCond d = fmap negate $ ltCond d


