module LMath.Algo.DisjointUnion
    ( disjointUnionFunction,
      disjointUnionAlg,
      GDUTag(..),
      genericDisjointUnion, genericDisjointUnionBF,
      taggedDisjointUnion, staggedDisjointUnion
    )
where

import Control.Exception ( assert )
import Data.IntMap ( IntMap )
import qualified Data.IntMap as IntMap
import Data.List
import Data.Maybe
import Data.Ratio

import Debug.Trace

import LMath.Types


ineqSystemIsFeasible :: Integral a =>
			([QfFormula (Polynomial (Ratio a))] -> QfFormula (Polynomial (Ratio a)) -> Bool) ->
		        ([QfFormula (Polynomial (Ratio a))] -> Polynomial (Ratio a) -> Signum) ->
		        [QfFormula (Polynomial (Ratio a))] ->
		        GIneqSystem (FracPoly (Ratio a)) ->
		        Bool
ineqSystemIsFeasible implies impliesSgn context sys =
    not $
    implies [] (QfNot (conjunctionQf context' `QfAnd`
		       ineqSystemToFormula sys'))
    where
    sys' = unfracIneqSystem (Just $ impliesSgn context) sys
    context' = map (fmap (liftPolyN (nVars sys'))) context


-- |
-- Generic helper function for disjoint union computation.
--
disjointUnionFunction
              :: (Real a, Fractional a)
              => [QfFormula (Polynomial a)]
	      -> ([QfFormula (Polynomial a)] ->
		  QfFormula (Polynomial a) ->
		  Bool)
	      -> (Polynomial a -> Signum)
	      -> ([QfFormula (Polynomial a)] ->
		  ([QfFormula (Polynomial a)] ->
		   QfFormula (Polynomial a) ->
		   Bool) ->
		  [GIneqSystem (Polynomial a)] ->
	          [GIneqSystem (Polynomial a)]
		 )
	      -> [GIneqSystem (FracPoly a)]
	      -> [GIneqSystem (FracPoly a)]
disjointUnionFunction ctx implies decideSgn unionFn systems =
    map (fmap (%%1)) $
    unionFn ctx implies $
    map (unfracIneqSystem $ Just decideSgn) systems


-- |
-- Compute a disjoint union of polyhedra.
-- Use with care, mostly untested!
--
disjointUnionAlg :: (Real a, Fractional a)
		 => [QfFormula (Polynomial a)]
	         -> ([QfFormula (Polynomial a)] ->
		     QfFormula (Polynomial a) ->
		     Bool)
		 -> [GIneqSystem (Polynomial a)]
		 -> [GIneqSystem (Polynomial a)]
disjointUnionAlg ctx implies (systems :: [GIneqSystem (Polynomial a)]) =
    concatMap (fromJust . atFormulasToIneqSystems nV) $
    du $
    map ineqSystemToAtFormulas systems
    where
    du :: [[AtFormula (Polynomial a)]] ->
	  [[AtFormula (Polynomial a)]]
    du dnf' =
	case addNegs dnf of
	  []     ->  []
	  (_:ds) -> (head dnf) : du (concatMap positiveDnfQf ds) 
	where
	dnf = filter (\psi -> not (implies liftedCtx (QfNot $ conjunctionQf $ map QfAt psi))) dnf'

        addNegs :: [[AtFormula q]] -> [QfFormula q]
	addNegs [] = []
	addNegs (d:ds) = d' : [QfNot d' `QfAnd` e | e <- addNegs ds]
	    where
	    d' = conjunctionQf $ map QfAt d

    nV = nVars $ head systems
    liftedCtx = map (fmap (liftPolyN nV)) ctx


data GDUTag a = GDUTag Int a

instance Eq (GDUTag a) where
    GDUTag x _  == GDUTag y _  =  x == y

instance Ord (GDUTag a) where
    compare (GDUTag x _) (GDUTag y _)  =  compare x y

instance Show (GDUTag a) where
    showsPrec _ (GDUTag x _) =
        showString "(GDUTag " . shows x . showString "_ )"


-- |
-- Compute a disjoint union.
-- The given objects are tagged with a list of tags. In the result,
-- each object is tagged with the tags of the input objects that
-- are part of the respective result object.
-- The function argument is a function which returns the
-- intersection of two objects A and B, the difference A without B, and
-- the difference B without A, in this order (can return @Nothing@
-- to signal that the objects A and B are already disjoint).
--
genericDisjointUnion :: (Ord tag, Show tag, Show a)
   => (a -> a -> Maybe ([a],[a],[a]))   -- ^ intersection, A without B, B without A
   -> [(a,[tag])]   -- ^ objects and tags
   -> [(a,[tag])]
genericDisjointUnion interDiffFn (inputs :: [(a,[tag])]) =
    gdu startKey toCheck m
    where
    m :: IntMap (a,[tag])
    m = IntMap.fromList $ zip [0..] inputs
    tags = [tag | (_,tag) <- inputs]
    startKey = length tags
    toCheck = [(k1,k2) | k1 <- [0..startKey-1], k2 <- [k1+1..startKey-1]]

    gdu nextKey [] m = IntMap.elems m
    gdu nextKey ((a,b):rest) m =
     let Just (aObj,aTags) = IntMap.lookup a m
         Just (bObj,bTags) = IntMap.lookup b m
     in
     case interDiffFn aObj bObj of
      Nothing -> gdu nextKey rest m
      Just (inters,aWithoutB, bWithoutA) ->
                 gdu nextKey3 newCheck newM
        where
        nextKey1 = nextKey  + length inters
        nextKey2 = nextKey1 + length aWithoutB
        nextKey3 = nextKey2 + length bWithoutA
        newCheck = [(k1,k2) | k1 <- [nextKey..nextKey1-1], k2 <- [k1+1..nextKey1-1]] ++
                   [(k1,k2) | k1 <- [nextKey1..nextKey2-1], k2 <- [k1+1..nextKey2-1]] ++
                   [(k1,k2) | k1 <- [nextKey2..nextKey3-1], k2 <- [k1+1..nextKey3-1]] ++
                   sortUniq (replaceAB rest)

        addM = [(k,(i,tagUnion aTags bTags)) | k <- [nextKey..nextKey1-1] | i <- inters] ++
               [(k,(i,aTags)) | k <- [nextKey1..nextKey2-1] | i <- aWithoutB] ++
               [(k,(i,bTags)) | k <- [nextKey2..nextKey3-1] | i <- bWithoutA]
        newM = IntMap.union (IntMap.fromList addM) $ xM
               
        xM = IntMap.delete b $ IntMap.delete a m

        replaceAB [] = []
        replaceAB ((x,y):rs) = replAB ++ repl
           where
           repl = replaceAB rs
           replAB
               | x == y  =  error "genericDisjointUnion: reflexive"
               | x == a && y == b  =  mix
               | x == b && y == a  =  mix
               | x == a  =  [(y,aa) | aa <- aKeys]
               | y == a  =  [(x,aa) | aa <- aKeys]
               | x == b  =  [(y,bb) | bb <- bKeys]
               | y == b  =  [(x,bb) | bb <- bKeys]
               | otherwise  =  [(x,y)]
        aKeys = [nextKey..nextKey2-1]
        bKeys = [nextKey..nextKey1-1] ++ [nextKey2..nextKey3-1]
        mix   = [(aa,bb) | aa <- aKeys, bb <- bKeys, aa < bb] 

    tagUnion t1s t2s = sortUniq (t1s++t2s)

sortUniq :: Ord a => [a] -> [a]
sortUniq = onlyOne . sort
    where
    onlyOne []  = []
    onlyOne [x] = [x]
    onlyOne (x:xs@(y:ys))
        | x == y     =  onlyOne xs
        | otherwise  =  x : onlyOne xs

check :: Int -> [(Int,Int)] -> IntMap a -> Bool
check nextKey todo m =
    nextKey == maximum (IntMap.keys m) + 1  &&
    all (`elem` IntMap.keys m) (map fst todo ++ map snd todo)

genericDisjointUnionBF :: (Ord tag, Show tag, Show a)
   => (a -> a -> Maybe ([a],[a],[a])) -> [(a,[tag])] -> [(a,[tag])]
genericDisjointUnionBF _ [] = []
genericDisjointUnionBF intersDiffsFn ((zp,tags):rest) = bf [] rest
    where
    bf acc [] = (zp,tags) : bf [] acc
    bf acc ((zp1,tags1):rest1) =
      case intersDiffsFn zp zp1 of
        Nothing -> bf ((zp1,tags1):acc) rest1
        Just (inters, diffs0, diffs1) ->
             bf []
                ([(i, tagUnion tags tags1) | i <- inters] ++
                 [(d, tags) | d <- diffs0] ++
                 [(d, tags1) | d <- diffs1] ++
                 acc ++ rest1
                )

    tagUnion a b = sortUniq (a ++ b)


taggedDisjointUnion :: (RealFrac a, Ord t, Show t)
                    => Maybe (GIneqSystem a)   -- ^ context with variables
                    -> [(GIneqSystem a, [t])]
                    -> [(GIneqSystem a, [t])]
taggedDisjointUnion mbVctx xs =
    genericDisjointUnion intersDiff xs
    where
    intersDiff sys1 sys2
      | null inters  =  Nothing
      | otherwise    =
          Just ( inters
               , f $ disjointUnionDifference Nothing [sys1] [sys2]
               , f $ disjointUnionDifference Nothing [sys2] [sys1]
               )
        where
        inters = f [sys1 `join` sys2]

    f iss = [is2 | is <- iss, let is2 = simplifyIneqs Nothing is,
                   not (areIneqsInfeasible is2)
            ]
{-
    ds = disjUnionAll [i | (i,_) <- xs]

    res = [(sys2, tag) | (sys,tag) <- xs, d <- ds,
                          let sys2 = simplifyIneqs mbVctx (sys `join` d),
                          not (areIneqsInfeasible sys2)
          ]
-}

-- |
-- Disjoint union of polyhedra with single tags.
-- The result polyhedra carry exactly one of the tags of
-- the input polyhedra which they are part of.
--
staggedDisjointUnion :: (RealFrac a, Ord t, Show t)
                     => Maybe (GIneqSystem a)   -- ^ context with variables
                     -> [(GIneqSystem a, t)]
                     -> [(GIneqSystem a, t)]
staggedDisjointUnion mbVctx xts =
    [(x,tag) | (x,tag:_) <- genericDisjointUnion intersDiff $
                            [(x,[tag]) | (x,tag) <- xts]
    ]
    where
    intersDiff sys1 sys2
       | null $ f [sys1 `join` sys2] = Nothing
       | otherwise =
          Just ([]
               ,f [sys1]
               ,f $ disjointUnionDifference Nothing [sys2] [sys1]
               )

    f iss = [is2 | is <- iss, let is2 = simplifyIneqs Nothing is,
                   not (areIneqsInfeasible is2)
            ]
