-- |
-- Polyhedra and Polytopes in implicit and parametric representation
--
-- Author: Tilmann Rabl
--

module LMath.Algo.DoubleDescription
    (
      -- Chernikova's Algorithm
      -- chernikova,

      -- * Volume Computation (probably incorrect due to the use of
      -- an incorrect implementation of Chernikova's algorithm)
      volume
    )
where

import LMath.Types

import Control.Monad ( filterM )

--
-- Unfortunately, the implementation of Chernikova's algorithm in
-- this module is incorrect. It is left here for reference.
--
-- A test case that shows the implementation to be incorrect:
--
t :: IneqSystem
t = ineqSystem ((3,0), matrix [[ 0, 0, 0, 1 ]
                              ,[ 1, 1,-1, 0 ]
                              ,[ 0, 1, 0, 0 ]
                              ,[ 1, 1, 0, 0 ]   -- (+)
                              ,[ 1, 0, 0, 0 ]
                              ]
               )
DLeaf (l,r) = makeDTree numOrdDTreeContext $
              chernikova $ getMatrix t
--
-- The dual of t is
--   ( []                          -- no lines
--   , [vector [ 0, 0, 0, 1]
--     ,vector [ 0, 0,-1, 0]
--     ,vector [ 0, 1, 1, 0]
--     ,vector [ 1, 0, 1, 0]  (*)
--     ]
--   )
-- but the ray (*) is missing from the result computed by
-- the function chernikova in this module.
-- (The constraint (+) is redundant in the system t and if
-- (+) is removed, function chernikova computes the correct result.)
--
--     Armin Groesslinger, 2009-07-28
--


-- |
-- Calculates the dual of a polytope in homogeneous form.
-- Result is a DTreeGen with possible solutions at its leaves.
-- The values at the leaves are tuples of the resulting line- and rayspace.
-- The line- and ray space can be interpreted as follows:
-- An empty line and rayspace means there is no sollution to the
-- system of constraints. The length of the lists in the tuple is
-- the number of lines\/rays\/vertices, i.e. |linespace| = #lines
-- and |rayspace| = #vertices + #rays.
-- The lines\/rays\/vertices are in homogeneous form.
--
chernikova :: (Fractional a,Num cond)
           => GMatrix a
           -> DTreeGen a cond ([GVector a],[GVector a])
chernikova mat =
    do (lines,rays) <- __chernikova mat
       return (restrict lines, restrict rays)
    where
    dim = nColumns mat
    restrict = map (\v -> subVector v 0 dim)

-- |
-- Returns the rays and lines enhanced with the constraint vectors
__chernikova mat
    | dim == 0   =  dLeaf ([],[])
    | otherwise  =
        _chernikova indices [lambda] n dim nbConst   
    where
    dim = nColumns mat
    nbConst = nRows mat
    uni = unitMatrix dim dim
    bid = insertColumns uni dim (mtranspose mat)
    n   = nColumns bid - 1
    bidvecs = rowVectors bid
    indices = take (dim-1) bidvecs
    lambda  = last bidvecs
    




_chernikova :: (Fractional a,Num cond) 
            => [GVector a] -- Lines in homogeneous representation,
                           -- extended with the constraint dimensions
            -> [GVector a] -- Rays in homogeneous representation,
                           -- extended with the constraint dimensions
            -> Int -- position of current constraint, counted backwards
                   -- from length of vectors to dimension of rays and lines
            -> Int -- dimension of rays and lines
            -> Int -- number of constraints,
                   -- initially n = dim + nbConstraint - 1
            -> DTreeGen a cond ([GVector a],[GVector a])
_chernikova (bidray :: [GVector a]) ray n dim nbConstraint
    | n < dim    =  dLeaf (bidray,ray)
    | otherwise  =
        do (x,y) <- caseAOrB
           _chernikova x y (n-1) dim nbConstraint
    where
    -- Filtere die Line b mit b*aktuelles Constraint /= 0 und == 0 
    bidrayEqUneqTree :: Num cond => DTreeGen a cond ([GVector a],[GVector a])
    bidrayEqUneqTree = filterVecEq bidray n

    -- Gibt es einen Line b mit b*aktuelles Constraint /= 0 ?
    -- Ja, dann CaseA sonst CaseB
    caseAOrB :: Num cond => DTreeGen a cond ([GVector a],[GVector a])
    caseAOrB =
        do (x,y) <- bidrayEqUneqTree
           if null y then caseB x else caseA x y

    -- Es gibt eine Line curr mit curr*aktuelles Constraint /= 0
    caseA :: Num cond
          => [GVector a]
          -> [GVector a]
          -> DTreeGen a cond ([GVector a],[GVector a])
    caseA satBidRay (curr:notSatBidRay) = 
      -- die Line curr ist an der Stelle n positiv oder negativ -> geCond
      geCond (vectorEntry curr n)
             -- die neue Menge der Lines wird aus der alten berechnet, 
             -- so dass alle Lines im aktuellen Constraint liegen:
             -- satBidray wird unveraendert uebernommen und notSatBidray wird 
             -- mit curr kombiniert, curr wird aus bidray entfernt
             (prod2DTreeG 
               (prodDTreeG (map (\x -> combineVecs x curr n) notSatBidRay
                            ++ map dLeaf satBidRay
                           )
               )
               -- der positive Anteil der curr Line wird zu ray hinzugefuegt
               -- und alle rays werden mit curr kombiniert
               (prodDTreeG (map (\x -> combineVecs x curr n) ray
                            ++ [dLeaf curr]
                           )
               )
             )

             (prod2DTreeG
               (prodDTreeG (map (\x -> combineVecs x curr n) notSatBidRay
                            ++ map dLeaf satBidRay
                           )
               )
               (prodDTreeG (map (\x -> combineVecs x (negate curr) n) ray
                            ++ [dLeaf $ negate curr]
                           )
               )
             )


    -- Ein DTreeGen, der die Mengen verify, saturates, violates enthaelt,
    -- die die Punkte enthalten, die auf der richtigen Seite,
    -- auf dem aktuellen Constraint und auf der falschen Seite liegen
    rayVerSatVio :: Num cond
                 => DTreeGen a cond ([GVector a], [GVector a], [GVector a])
    rayVerSatVio = filterVecS ray n

    -- Es gibt keine Line b mit b*aktuelles Constraint /= 0 
    caseB :: Num cond 
          => [GVector a] -> DTreeGen a cond ([GVector a], [GVector a])
    caseB satBidRay = prod2DTreeG (dLeaf satBidRay) newRays

    -- Wie viele und welche Constraints werden von beiden Strahlen erfuellt?
    common :: Num cond
           => GVector a
           -> GVector a 
           -> Int
           -> Int
           -> Int
           -> [Int]
           -> DTreeGen a cond (GVector a, GVector a, Int, [Int])
    common vec1 vec2 lowerb ind nbCom sats
        | lowerb == ind  =  dLeaf (vec1,vec2,nbCom,sats)
        | otherwise      =
            eqCond (vectorEntry vec1 ind)
                   (eqCond (vectorEntry vec2 ind)
                           (common vec1 vec2 lowerb (ind-1) (nbCom+1) (sats++[ind]))
                           (common vec1 vec2 lowerb (ind-1) nbCom sats)
                   )
                   (common vec1 vec2 lowerb (ind-1) nbCom sats)

    -- Strahlenpaare, die genuegend gleiche Constraints erfuellen
    -- Bedeutung: (verify,saturate,Kartesisches Produkt aus verify und 
    -- violates mit Anzahl von beiden erfuellter Constraints und
    -- entsprechende n's)
    commonConstraints :: Num cond
                      => DTreeGen a cond 
                                  ( [GVector a]
                                  , [GVector a]
                                  , [(GVector a, GVector a, Int, [Int])]
                                  )

    commonConstraints =
        do (x,y,z) <- rayVerSatVio
           res <- mapDTreeG
                    (filter (\(_,_,nb,_) -> nb + length bidray >= dim-2))
                    (prodDTreeG
                       --     should this n here: v   be (n-1) ?
                       (map (\(a,b) -> common a b n (nbConstraint+dim-1) 0 [])
                            (cartesianProdukt x z)
                       )
                    )        
           return (x,y,res)
                        

    -- Erfuellt der Strahl die selben Constraints?
    notcommonzero :: Num cond => GVector a -> [Int] -> DTreeGen a cond Bool
    notcommonzero _ []       = dLeaf False
    notcommonzero r (st:sts) = eqCond (vectorEntry r st)
                                      (notcommonzero r sts)
                                      (dLeaf True)
    -- Ist der Strahl neu?
    isNewRay :: Num cond
             => GVector a -> GVector a -> [Int] -> [GVector a]
             -> DTreeGen a cond [GVector a]
    isNewRay x y _ []        =
        do vec <- combineVecs y x n
           return [vec]
    isNewRay x y sats (r:rs)
        | x /= r && y /= r  =
            do notCom <- notcommonzero r sats
               if notCom then isNewRay x y sats rs else return []
        | otherwise  =  isNewRay x y sats rs

    -- Neue Menge der Strahlen: verify ++ saturates ++ neueStrahlen
    newRays :: Num cond => DTreeGen a cond [GVector a]
    newRays =
        do (ver,sat,coms) <- commonConstraints
           newRay <- prodDTreeG (map (\(x,y,_,z) -> isNewRay x y z ray) coms)
           return (ver ++ sat ++ concat newRay)
                    




-- |
-- Splits a set of vectors with non-linear coefficients into three sets,
-- where the n-th coefficient is less then, equal and greater than 0.
-- Linear implementation:
-- (violate,saturate,verify) =
--    ((filter (\x -> (vectorEntry x n) < 0) ray),
--     (filter (\x -> (vectorEntry x n) == 0) ray),
--     (filter (\x -> (vectorEntry x n) > 0) ray))
filterVecS :: (Num a, Num cond)
           => [GVector a] 
           -> Int 
           -> DTreeGen a cond ([GVector a],[GVector a],[GVector a])
filterVecS vecs n = 
  _filter vecs n [] [] []
    where _filter [] _ ver sat vio = dLeaf (ver,sat,vio)
          _filter (v:vs) n ver sat vio = 
                   sCond (vectorEntry v n) 
                        (_filter vs n ver sat (vio++[v]))
                        (_filter vs n ver (sat++[setVectorEntry v n 0]) vio) 
                        (_filter vs n (ver++[v]) sat vio) 


-- |
-- Splits a set of vectors with non-linear coefficients into two sets,
-- where the n-th coefficient is equal and unequal 0.
-- Linear implementation:
-- (equ0,uneq0) =
--    ((filter (\x -> (vectorEntry x n) == 0) ray),
--     (filter (\x -> (vectorEntry x n) <> 0) ray))
--
filterVecEq :: (Num a, Num cond)
            => [GVector a] 
            -> Int 
            -> DTreeGen a cond ([GVector a],[GVector a])
filterVecEq vecs n = 
  _filter vecs n [] []
    where _filter [] _ vec0 vecmp = dLeaf (vec0,vecmp)
          _filter (v:vs) n vec0 vecmp = 
                   eqCond (vectorEntry v n) 
                        (_filter vs n (vec0++[setVectorEntry v n 0]) vecmp) 
                        (_filter vs n vec0 (vecmp++[v]))


-- |
-- Calculates a linear combination c of two vectors where the n-th entry of
-- c is 0. Does no normalisation.
--
combineVecs :: (Fractional a, Num cond)
            => GVector a -> GVector a -> Int -> DTreeGen a cond (GVector a)
combineVecs ray1 ray2 n =
  let
    a1 = negate $ vectorEntry ray1 n
    a2 = vectorEntry ray2 n
    newRay = mulVS ray1 a2 + mulVS ray2 a1
  in normalizeFP newRay

-- |
-- Normalizes a vector so that the first non empty entry is 1,
-- the result is a DTreeGen with all posible cases.
--
normalizeFP :: (Fractional a, Num cond) 
            => GVector a 
            -> DTreeGen a cond (GVector a)
normalizeFP vec =
  _normalize vec 0 (vectorLength vec)
  where
  _normalize v i lenVec =
      if i < lenVec 
         then sCond (vectorEntry v i) 
                    (dLeaf (mulVS v (-1/(vectorEntry v i))))
                    (_normalize v (i+1) lenVec) 
                    (dLeaf (mulVS v (1/(vectorEntry v i)))) 
         else dLeaf vec



-- |
-- Builds the cartesian produkt of to lists.
-- The resulting list contains all combinations of elements of the give lists.
cartesianProdukt :: [a] -> [b] -> [(a,b)]
cartesianProdukt [] _ = []
cartesianProdukt (x:xs) ys = (map (\z -> (x,z)) ys) ++ (cartesianProdukt xs ys)



------------------------
-- Volume Computation --
------------------------


-- |Calculates the volume of a polytope, given a matrix containing the constraints.
-- Returns a DTreeGen with the resulting volumes at its leaves. For unbounded polytopes
-- the tree has a leaf with value -1 and for polytopes of lower dimension it returns 0
volume :: (Fractional a, Num cond) => GMatrix a -> DTreeGen a cond a
volume constraints =
  let
    cherDTree = __chernikova constraints
    volumes = compDTreeG (\(x,y) -> volumeCase y x) cherDTree
    volumeCase :: (Fractional a, Num cond) => [GVector a] -> [GVector a] -> DTreeGen a cond a
    volumeCase rays bidray = 
      if ((length bidray) > 0)
       then dLeaf (-1)
        -- there is a line => the polytope is unbounded
        else if (length rays > 1)
          then
            let 
              dim = (nColumns constraints) - 1
            in compDTreeG (\x -> if x then dLeaf(-1) else _volume rays dim [] dim) (unbounded rays (dim))
          else dLeaf 0
            -- there are no rays => zero polytope => zero volume
          where
            -- checks if any ray of a list is unbounded
    unbounded :: (Fractional a, Num cond) => [GVector a] -> Int -> DTreeGen a cond Bool
    unbounded [] _ = dLeaf False
    unbounded (r:rs) n = eqCond (vectorEntry r n)
                                  (dLeaf True)
                                  (unbounded rs n)
   in volumes

-- |Calculates the volume of a polytope, given a matrix containing the constraints, the edges 
-- and the dimensionality. Expects a bounded polytope with full dimensionality.
-- __volume :: (Fractional a, Num cond) => GMatrix a -> [GVector a] -> DTreeGen a cond a
-- __volume hyperplanes rays = 
--   let dim = (nColumns hyperplanes) - 1
--       vertices = map (\x -> mulVS (subVector x 0 dim) (1/ (vectorEntry x dim))) rays
--    in if (isSimplex vertices dim) 
--         then simplexVolume vertices dim
--         else _volume (rowVectors hyperplanes) vertices dim []
--   in _volume (rowVectors hyperplanes) vertices dim []

_volume :: (Fractional a, Num cond)
       => [GVector a] 
       -> Int 
       -> [GVector a]
       -> Int 
       -> DTreeGen a cond a
_volume rays n vertices dim =
    if ((length rays) == 0)
      then (dLeaf 0)
      else if (n == 1)
        then 
          let 
            allVertices = vertices++rays
            result = simplexVolume allVertices ((length allVertices)-1)
            --result = dLeaf (int2FracPoly (length allVertices))
          in result
        else 
          let 
            centerVertex = centerOfMass rays
            newVertices = mapDTreeG (\x -> vertices++[x]) centerVertex
            facets = getFacets rays dim n
            volumes = mapDTreeG (map (\x -> compDTreeG (\y -> _volume x (n-1) y dim) newVertices)) facets
                  --vols = map absolute volumes
            simplexSum = mapDTreeG (foldl (+) 0) (compDTreeG prodDTreeG volumes)
          in simplexSum


-- |Converts an Int to a FracPoly Rational
int2FracPoly :: (Integral a, Fractional b) => a -> FracPoly b
int2FracPoly n = fromPoly (constPoly ((fromIntegral n)))

-- |Gives all (n-1)-dimenional facets of a n-dimensional polytope and their restricting hyperplanes
getFacets :: (Fractional a, Num cond) => [GVector a] -> Int -> Int -> DTreeGen a cond [[GVector a]]
getFacets rays dim n =
  let
    constraintNbs = [dim..((vectorLength (head rays))-1)]
    nbRays = length rays
    facetList = map (\x -> _getFacets rays x []) constraintNbs
    filteredFacetList = map (mapDTreeG (\x -> if ((length x) == nbRays) || ((length x)< n) then [] else x)) facetList
    facetTree = prodDTreeG filteredFacetList
  in facetTree

_getFacets [] n vertices = dLeaf vertices
_getFacets (r:rs) n vertices =
  eqCond (vectorEntry r n)
         (_getFacets rs n (vertices++[r]))
         (_getFacets rs n vertices)

-- getFacets :: Fractional a => [GVector a] -> [GVector a] -> [[GVector a]]
-- getFacets hyperplanes edges = 
--     let facets = getNotEmptyFacets hyperplanes edges []
--           where getNotEmptyFacets [] _ fcts = fcts
--                 getNotEmptyFacets (pl:pls) edgs fcts = 
--                   let fct = edgesOnHyperplane pl edgs
--                    in if (length fct) == 0 
--                       then getNotEmptyFacets pls edgs fcts
--                       else getNotEmptyFacets pls edgs (fcts++[fct])
--         {-result = getNewPlanes fcts 0
--           where getNewPlanes [] _ _ = []
--                 getNewPlanes (fc:fcs) n = (fc,(take n cutPlanes)++(drop (n+1) cutPlanes)):(getNewPlanes fcs cutPlanes (n+1))-}
--      in facets --result

-- -- |Returns all edges on a hyperplane
-- edgesOnHyperplane :: Fractional a => GVector a -> [GVector a] -> [GVector a]
-- edgesOnHyperplane _ [] = []
-- edgesOnHyperplane hyp (edge:edges) = 
--     if (isEdgeOnHyperplane hyp edge) 
--       then edge:(edgesOnHyperplane hyp edges)
--       else (edgesOnHyperplane hyp edges)
-- 
-- -- |Checks if an edge is on a hyperplane
-- isEdgeOnHyperplane :: Fractional a => GVector a -> GVector a -> Bool
-- isEdgeOnHyperplane hyp edge = 
--   let 
--     edgeN = insertEntries edge (vectorLength edge) [(int2FracPoly 1)]
--     multVec = mulVV hyp edgeN
--    in (multVec == 0)


-- |Calculates the center of mass of a polytope given by its edges.
centerOfMass :: (Fractional a, Num cond) => [GVector a] -> DTreeGen a cond (GVector a)
centerOfMass rays = 
  let 
    divBy     = 1 / (fromIntegral (length rays))
    vertexSum = foldl addVV (head rays) (tail rays)
    result    = normalizeFP (mulVS vertexSum divBy)
  in result

-- |Calculates the volume of a simplex by 
-- generating a matrix of vectors from one vertex to all the other vertices
-- calculating the matrixes volume with the determinant
-- and dividing the determinant by the faculty of the simplexes dimensionality
simplexVolume :: (Fractional a, Num cond) => [GVector a] -> Int -> DTreeGen a cond a
simplexVolume rays dim = 
  let
    vertices = map (\x -> mulVS (subVector x 0 dim) (1/ (vectorEntry x dim))) rays
  in _simplexVolume vertices dim
  
_simplexVolume :: (Fractional a, Num cond)
               => [GVector a] -> Int -> DTreeGen a cond a
_simplexVolume (rootEdge:edges) n = 
  let 
    negRoot = negV rootEdge
    vectors = map (addVV negRoot) edges
    dim = (vectorLength rootEdge) -1
    --vectors = map (\x -> mulVS (subVector x 0 (dim-1)) (1/ (vectorEntry x dim))) _vectors
    vMatrix = rowVectorsToMatrix1 vectors
    det     = determinantFP (rows' vMatrix)
    divBy   = fac n
    result  = det / fromIntegral divBy
    absResult = geCond result (dLeaf result) (dLeaf (-result))
  in absResult

-- |Calculates the faculty
fac :: Int -> Int
fac n = foldl (*) 1 [1..n] 

-- |Calculates the determinant of a matrix conaining fractions of polynomes
determinantFP :: Fractional a => [[a]] -> a
determinantFP [] = fromIntegral 0
determinantFP ((a:[]):[]) = a
determinantFP ((a:b:[]):(c:d:[]):[]) = (a * d) - (b * c)
determinantFP ((a:b:c:[]):(d:e:f:[]):(g:h:i:[]):[]) = (a * e * i) + (d * h * c) + 
                                                      (g * b * f) - (c * e * g) - 
                                                      (f * h * a) - (i * b * d)
determinantFP (x:xs) = _determinantFP x xs 0 1

_determinantFP :: Fractional a => [a] -> [[a]] -> Int -> Int -> a
_determinantFP [] a n sign = 0
_determinantFP (x:xs) a n sign = ((fromIntegral sign) * x * determinantFP (delCol a n)) + 
                                 (_determinantFP xs a (n+1) (sign * (-1)))

-- |Deletes a column in a row major order 2 dimensional array
delCol :: [[a]] -> Int -> [[a]]
delCol x n = map (delNthElem n) x

-- |Deletes the n-th element of a list
delNthElem :: Int -> [a] -> [a]
delNthElem n [] = []
delNthElem 0 (x:xs) = xs
delNthElem n (x:xs) = x:(delNthElem (n-1) xs)

-- Tests sind raus weil sonst zyklische Abhaengigkeiten entstehen (wg fracPolyDTreeContext)
-- test1 =
--   let
--     a = fromPoly (indeterminate 1 0)
--     hyp = matrix [[1,0,-1],[0,1,-1],[-1,0,6],[0,-1,5],[1,-1,-a]]
--     result = chernikova hyp
--   in makeDTree (fracPolyDTreeContext () []) result
-- 
-- test2 = 
--   let
--     p = fromPoly (indeterminate 1 0)
--     hyp = matrix [[1,0,0],
--                   [-1,0,4],
--                   [0,1,0],
--                   [0,-1,4],
--                   [p,-1,2-(2*p)]]
--     result = chernikova hyp
--   in makeDTree (fracPolyDTreeContext () []) result
-- 
-- test3 =
--   let
--     hyp = matrix [[1,0,-1],[0,1,-1],[-1,0,6],[0,-1,5],[1,-1,1]]
--     result = chernikova hyp
--   in makeDTree (fracPolyDTreeContext () []) result
-- 
-- test4 =
--   let
--     hyp = matrix [[1,0,2],[0,1,2],[-1,0,2],[0,-1,2]]
--     result = chernikova hyp
--   in makeDTree (fracPolyDTreeContext () []) result
-- 
-- test5 = 
--   let
--     p = fromPoly (indeterminate 1 0)
--     hyp = matrix [[1,0,0],
--                   [0,1,0],
--                   [0,-1,4],
--                   [p,-1,2-(2*p)]]
--     result = chernikova hyp
--   in makeDTree (fracPolyDTreeContext () []) result
-- 
-- testVol1 = 
--   let
--     a = fromPoly (indeterminate 1 0)
--     hyp = matrix [[1,0,-1],[0,1,-1],[-1,0,6],[0,-1,5],[1,-1,-a]]
--     vol = volume hyp
--   in makeDTree (fracPolyDTreeContext () []) vol
-- 
-- testVol2 = 
--   let
--     p = fromPoly (indeterminate 1 0)
--     hyp = matrix [[1,0,0],
--                   [-1,0,4],
--                   [0,1,0],
--                   [0,-1,4],
--                   [p,-1,2-(2*p)]]
--     vol = volume hyp
--   in makeDTree (fracPolyDTreeContext () []) vol
-- 
-- testVol5 =
--   let
--     p = fromPoly (indeterminate 1 0)
--     hyp = matrix [[1,0,0],
--                   [0,1,0],
--                   [0,-1,4],
--                   [p,-1,2-(2*p)]]
--     vol = volume hyp
--   in makeDTree (fracPolyDTreeContext () []) vol
