module LMath.Algo.EhrhartArith
  ( epPartsDisjoint, evSubstParamSum, epSubstParamSum, binom, evSubstParam, evCanon )
where

import Data.List
import Debug.Trace

import LMath.Types
import LMath.Algo.DisjointUnion

epPartsDisjoint :: RealFrac a => EPartition a -> [(GIneqSystem a, EValue a)]
epPartsDisjoint ep =
    [(sys, evs !! idx)
    | (sys,idx) <- staggedDisjointUnion Nothing $ zip syss [0..]
    ]
    where
    (syss, evs) = unzip $ epPartitions ep

instance Eq a => Eq (EValue a) where
    EConst a == EConst b  =  a == b
    _ == _  =  False
--    (==) = error "(==) not defined for Eq (EValue a)"

instance Num a => Num (EValue a) where
    x + y  =  evCanon $ eplus x y
    negate ev  =  fmap negate ev
    
    EConst x * y = fmap (*x) y
    x * EConst y = fmap (*y) x
    ev1 * ev2  =  error "(*) only defined for at least one constant factor in Num (EValue a)"

    fromInteger x  =  EConst $ fromInteger x
    signum = error "signum not implemented for Num (EValue a)"
    abs    = error "abs not implemented for Num (EValue a)"

instance RealFrac a => Eq (EPartition a) where
    (==) = error "(==) not defined for Eq [(GIneqSystem a, EValue a)]"

instance RealFrac a => Num (EPartition a) where
    negate xs = epartition [(cond, negate ehp) | (cond,ehp) <- epPartitions xs]
    (+) = arithCEv (+) 0
    (*) = arithCEv (*) 1
    fromInteger  =  error "fromInteger not implemented for Num (EPartition a)"
    signum = error "signum not implemented for Num (EPartition a)"
    abs    = error "abs not implemented for Num (EPartition a)"

eplus :: Num a => EValue a -> EValue a -> EValue a
eplus x y
  | isZ x  =  y
  | isZ y  =  x
eplus (EConst x) (EConst y) = EConst (x+y)
eplus (ESum xs)  (ESum ys)  = ESum (xs++ys)
eplus (ESum xs)  y          = ESum (y:xs)
eplus x          (ESum ys)  = ESum (x:ys)
{-
eplus (EPoly p xs) (EConst y) =
    EPoly p (init xs ++ [eplus (last xs) (EConst y)])
eplus (EConst x) (EPoly q ys) =
    EPoly q (init ys ++ [eplus (last ys) (EConst x)])
eplus (EPoly p xs) (EPoly q ys)
    | p == q  =  EPoly p $ reverse $ xzipWith eplus (reverse xs) (reverse ys)
    | p <  q  =  EPoly p (init xs ++ [eplus (last xs) (EPoly q ys)])
    | p >  q  =  EPoly q (init ys ++ [eplus (last ys) (EPoly p xs)])
eplus (EPeriodic p xs) (EPeriodic q ys)
    | p == q  =  EPeriodic p $ periodicValues (periodic xs + periodic ys)
-}
eplus x          y          = ESum [x,y]

xzipWith :: (a -> a -> a) -> [a] -> [a] -> [a]
xzipWith _ [] ys = ys
xzipWith _ xs [] = xs
xzipWith f (x:xs) (y:ys) = f x y : xzipWith f xs ys

isZ :: Num a => EValue a -> Bool
isZ (EConst 0) = True
isZ (EConst _) = False
isZ (EPoly _ xs) = all isZ xs
isZ (EPeriodic _ xs) = all isZ xs
isZ (ESum xs) = all isZ xs

--
-- The equality and ordering on X is defined by the
-- first argument of the constructor, only. This is
-- used below when computing tagged disjoint unions.
-- 
data X a b = X a b deriving Show
instance Eq a => Eq (X a b) where
    X a1 _ == X a2 _  =  a1 == a2
instance Ord a => Ord (X a b) where
    compare (X a1 _) (X a2 _) = compare a1 a2

arithCEv :: RealFrac a
  => (EValue a -> EValue a -> EValue a)
  -> EValue a
  -> EPartition a
  -> EPartition a
  -> EPartition a
arithCEv f e cevs1 cevs2 =
    epartition
      [(cond, foldl f e ps)
      | (cond, xpolys) <- taggedDisjointUnion Nothing cevs,
        let ps = [p | X _ p <- xpolys]
      ]          
    where
    cevs = [(cond, [X 1 ehp]) | (cond,ehp) <- epPartitions cevs1] ++
           [(cond, [X 2 ehp]) | (cond,ehp) <- epPartitions cevs2]

-- |
-- Substitute a parameter @p@ by @p + val@.
--
evSubstParamSum :: (Num a, Integral b, Integral v)
   => b           -- ^ parameter number
   -> v           -- ^ @val@
   -> EValue a
   -> EValue a
evSubstParamSum p' val = s
    where
    s (EConst c)       = EConst c
    s (EPeriodic q xs)
        | p == q    = EPeriodic q $ map s $ take l (drop off (xs++xs))
        | otherwise = EPeriodic q $ map s xs
        where
        off = fromIntegral val `rem` l
        l = length xs
    s (EPoly     q xs)
        | p == q    = sum [EPoly q [binom e i * fromIntegral (val^i) * ehp
                                   | i <- [0..e]]
                          | (ehp,e) <- zip xs [l-1,l-2 .. 0]]
        | otherwise = EPoly q $ map s xs
        where
        l = length xs
    s (ESum xs) = ESum $ map s xs

    p = fromIntegral p'

epSubstParamSum :: (Num a, Integral b, Integral v)
  => b
  -> v
  -> EPartition a
  -> EPartition a
epSubstParamSum p' val epart =
    epartition
           [(isSubstParamAE p pPlusVal cond, evSubstParamSum p val ehp)
           | (cond,ehp) <- xs]
    where
    xs = epPartitions epart
    pPlusVal = vector (replicate p 0 ++ [1] ++ replicate (nV+nP-p-1) 0 ++ [fromIntegral val])
    p = fromIntegral p'
    nV = nVars   asys
    nP = nParams asys
    asys = fst $ head xs

binom :: (Integral a, Num b) => a -> a -> b
binom n k
  | n >= k     =  fromIntegral $ product [kk+1 .. nn] `div` product [1..nn-kk]
  | otherwise  =  0
  where
  nn = fromIntegral n :: Integer
  kk = fromIntegral k :: Integer 


--
-- try to do _some_ canonicalisation; there's no guarantee
-- that the returned EValue is really canonical
--
evCanon :: Num a => EValue a -> EValue a
evCanon (EPeriodic _ [e]) = evCanon e
evCanon (EPoly _ [e])     = evCanon e
evCanon e
  | Just v <- evAffine (evParams e) e  =  evFromAffine v
evCanon (EPeriodic p xs)  = EPeriodic p $ map evCanon xs
evCanon (EPoly     p xs)  = EPoly     p $ map evCanon xs
evCanon (ESum xs)         = ESum $ map evCanon xs
evCanon e = e

-- |
-- Substitute an affine expression for a parameter.
-- MAY RETURN EVALUES WHICH ARE NOT IN CANONICAL FORM !
--
evSubstParam :: (Num a, Integral b, Integral c)
  => b
  -> GVector c
  -> EValue a
  -> EValue a
evSubstParam p' expr = evCanon . s
    where
    s :: Num a => EValue a -> EValue a
    s (EConst c) = EConst c
    s (ESum  xs) = ESum $ map s xs
    s (EPoly q xs :: EValue a)
        | p == q    = toEV 0 p2
        | otherwise = EPoly q $ map s xs
        where
        p1,p2 :: Polynomial (EValue a)
        p2 = sum [constPoly co * p1^e | (e,co) <- zip [0..] (reverse xs)]
        p1 = sum (constPoly (fromIntegral off) : [constPoly (fromIntegral val) * indeterminate numberOfParams n | (n,val) <- zip [0..] exprInit])

        toEV n poly = EPoly n $ map fixEV $ reverse $ coefficients poly
            where
            fixEV co = case isConstPoly co of
                           Just c  -> c
                           Nothing -> toEV (n+1) co
        numberOfParams = length exprInit

    s (EPeriodic q xs)
        | p == q    =  
             case nextP of
                 Just nP -> EPeriodic p [evSubstParam nP nextExpr $ EPeriodic nP $ take (ll `div` nextPcoeff) xs | xs <- nths]
                     where
                     nextPcoeff = fromIntegral $ vectorEntry expr nP
                     nextExpr = setVectorEntries expr [(p,0), (vectorLength expr-1,0)] 
                 Nothing -> EPeriodic q $ map head nths
        | otherwise = EPeriodic q $ map s xs
        where
        ll = lcm l $ length xs
            
        nths =
            take (if pCoeff == 0 then 1 else (ll `div` pCoeff)) $
            everyNth pCoeff $
            drop (off `mod` length xs) $
            cycle xs

    p = fromIntegral p'

    pCoeff :: Int
    pCoeff = fromIntegral $ vectorEntry expr p

    exprInit = init $ vectorEntries expr

    l :: Int
    l = fromIntegral $ foldl lcm 1 $ filter (/=0) exprInit

    off :: Int
    off = fromIntegral $ last $ vectorEntries expr

    nextP = case filter (/=p) $ findIndices (\i -> i/=0) $ exprInit of
              i:_ -> Just i
              []  -> Nothing

everyNth :: Int -> [a] -> [[a]]
everyNth _ [] = []
everyNth n xs = xs : everyNth n (drop n xs)
