{-
    This file is part of the LooPo project, see
    <http://www.infosun.fim.uni-passau.de/cl/loopo/>.

    Copyright (C) 2008 Universitaet Passau, Germany
    Author: Armin Groesslinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-}

module LMath.Algo.FM
    ( -- * Sorting functions
      fourierMotzkin
    , fourierMotzkinP
    ,  fmToLinSort
    , sortIneqsFM
    , FMErbAlgo(..)

      -- * Test functions for very primitive code generation
    , testCode, testCode2, dtreeSorted2code, dtreeSorted2hs, sorted2code
    )
where

import Control.Exception (assert)
import Control.Monad ( liftM )
import Data.IORef
import Data.List
import Data.Maybe
import Data.Ratio
import qualified Data.Set as Set
import Data.Set ( Set )
import Data.Typeable ( Typeable )
import System.IO.Unsafe

import Config ( debugTrace )
import LMath.Types
import LMath.Algo.Chernikova ( chernikovaSimplifyIneqsG )
import LMath.Algo.Simplex


import Debug.Trace


-- |
-- Available algorithms to eliminate redundant bounds (ERB)
-- after each projection.
--
data FMErbAlgo = FMErbNone          -- ^ no elimination of redundant bounds
               | FMErbChernikova    -- ^ check using Chernikova
               | FMErbSimplex       -- ^ check using Simplex
                 deriving (Eq, Show, Typeable)

erbSimplifierName :: FMErbAlgo -> String
erbSimplifierName FMErbNone       = "(none)"
erbSimplifierName FMErbChernikova = "Chernikova"
erbSimplifierName FMErbSimplex    = "Simplex"

-- |
-- Fourier-Motzkin Elimination.
-- Perform Fourier-Motzkin elimination on the given inequality
-- system (with additional equalities, if given). The
-- elimination starts with the variable with highest index
-- and finishes with variable number 0.
-- When the number of "outer" variables is @Just n@, the
-- elimination stops when only @n@ variables remain in the system
-- and the returned 'SortedSystem's do not contain bounds for the
-- outer @n@ variables. When it is @Nothing@, the elimination of the
-- last variables is followed by a satisfyability check for the
-- context, i.e., leaves of the form @Just ...@ are guarded by
-- conditions which guarantee the nonemptyness of the
-- system. Note that @Just 0@ performs an elimination for all the
-- variables, but does not check the context as @Nothing@ does.
--
fourierMotzkin :: (Fractional a, Integral b, Num cond)
               => GIneqSystem a        -- ^ inequalities
	       -> Maybe (GEqSystem a)  -- ^ optional equalities
	       -> Maybe b              -- ^ number of "outer" variables
	       -> DTreeGen a cond (Maybe (SortedSystem a))
fourierMotzkin ineqs' mbEqs ctxVars =
    assert (fmap (>=0) ctxVars /= Just False) $
    assert (fmap ((==nP).nParams) mbEqs /= Just False) $
    do mbSS' <- cFM (nColumns ineqs - 2)
                    (rowVectors $ getMatrix ineqs)
                    []
                    eqs
                    (fmap ((+nP).fromIntegral) ctxVars)
       let mbSS = case ctxVars of
                    Nothing -> fmap (genericDrop nP) mbSS'
                    -- if we stop prematurely (for some context variables)
                    -- we must not try to drop the parameter dimensions
                    -- because they are not represented in the result
                    -- SortedSystem
                    Just _  -> mbSS'
       return $ fmap (map (fmap paramsBack)) mbSS
    where
    eqs = rowVectors (getMatrix eqs1) ++
          case mbEqs of
	    Nothing -> []
	    Just es -> rowVectors $ getMatrix $ paramsToFront es
    (ineqs,eqs1) = findEqualities $ paramsToFront ineqs'

    paramsToFront ae = moveParamsToVars ae 0 nP 0
    paramsBack v = vector (vars ++ params ++ [co])
        where
          ents = vectorEntries v
          (params, rest) = genericSplitAt nP ents
          vars = init rest
          co   = last rest
    nP = nParams ineqs'

--
-- combinator based implementation of FM algorithm
--
cFM :: (Fractional a, Num cond)
    => Integer       -- position of variable to eliminate next
    -> [GVector a]   -- ineqs on current var
    -> [GVector a]   -- ineqs on vars lower than current var
    -> [GVector a]   -- equalities
    -> Maybe Integer -- stop when only ... variables are left
    -> DTreeGen a cond (Maybe (SortedSystem a))
cFM varPos (ineqs1 :: [GVector a]) ineqs2 eqs ctxVars =
    do erbAlgo <- lookupDTGOptionWithDefault FMErbNone
       ineqs <- cERB erbAlgo varPos ineqs1 ineqs2
       res erbAlgo ineqs
    where
    res erbAlgo ineqs
     | ctxVars == Just (varPos+1)  = 
	debugTrace ("fm: finished eliminating vars (" ++ show (varPos+1) ++
		    " context vars remain)") $
	dLeaf (Just [])
     | varPos <  -1  =  error "cFM: varPos < -1 ?"
     | varPos == -1  =
	debugTrace ("fm: finished eliminating vars, " ++ infoStr) $
	let eqConds   = map (\x -> QfAt $ AtEq $ vectorEntry x 0) eqs
	    ineqConds = map (\x -> QfAt $ AtGe $ vectorEntry x 0) ineqs in
	foldl (\st cond -> fCond cond st (dLeaf Nothing))
	      (dLeaf $ Just [])
	      (eqConds ++ ineqConds)
     | otherwise =
	debugTrace ("fm: " ++ show (varPos+1) ++ " vars remaining, " ++
                    infoStr) $
--	debugTrace (show ineqs) $
	foldr (\(e,es) acc ->
	       eqCond (vectorEntry e varPos)
		    acc
	            (mapDTreeG (fmap (++ [Equation $ mkEq e])) $
		     cFM (varPos-1)
		         []
		         (nub $ substitute varPos e ineqs)
		         (nub $ substitute varPos e es)
		         ctxVars
		    )
	      )
              (fmStep ineqs)
	      (selectOne eqs)
     where
     infoStr = show (length ineqs) ++ " inequalities" 
               ++ simplInfo ++ ", " ++
               show (length eqs) ++ " equalities"
     simplInfo =
         case erbAlgo of
           FMErbNone -> ""
           _         -> " (" ++ show (length ineqs1 + length ineqs2) ++
                        " before simpl. via " ++
                        erbSimplifierName erbAlgo ++ ")"

    mkEq e = deleteEntries e varPos 1 `mulVS`
	     recip (negate $ vectorEntry e varPos)

    fmStep :: Num b => [GVector a] -> DTreeGen a b (Maybe (SortedSystem a))
    fmStep ins =
        do (low,upp,rest) <- cBnds varPos ins
           r <- cFM (varPos-1)
			   [u-l | l <- low, u <- upp]
			   rest
			   newEqs
			   ctxVars
           return $ fmap (insertBounds (low,upp)) r
	where
	newEqs = nub $ map (\e -> deleteEntries e varPos 1) eqs
	insertBounds (low,upp) res = res ++ [LowerUpper low upp]


cBnds :: (Fractional a, Num b) =>
	Integer -> [GVector a] -> DTreeGen a b ([GVector a], [GVector a], [GVector a])
cBnds varPos []     = dLeaf ([], [], [])
cBnds varPos (r:rs) = withDTreeContext cbnds
    where
    cbnds dtc
	| signumLt  signum = tm
	| signumEq  signum = t0
	| signumGt  signum = tp
	| signumLe  signum = eqCond coeff t0 tm
	| signumGe  signum = eqCond coeff t0 tp
	| signumNeq signum = geCond coeff tp tm
	| otherwise        = sCond coeff tm t0 tp
	where
	sgn c = dtcContextImpliesSignum dtc c
	signum = sgn coeff
	coeff = vectorEntry r varPos 
	nobound = deleteEntries r varPos 1
	bound   = nobound `mulVS` recip (negate coeff)
	q = cBnds varPos rs

        tm = mapDTreeG (\(l1,u1,r1) -> (l1, ordIns False sgn bound u1, r1)) q
	t0 = mapDTreeG (\(l1,u1,r1) -> (l1, u1, nubIns nobound r1)) q
	tp = mapDTreeG (\(l1,u1,r1) -> (ordIns True sgn bound l1, u1, r1)) q


-- (c)combinator based version of (E)liminate (R)edundant (B)ounds
--
cERB :: (Fractional a, Num cond)
     => FMErbAlgo
     -> Integer
     -> [GVector a]
     -> [GVector a]
     -> DTreeGen a cond [GVector a]
cERB erbAlgo varPos bounds rest =
    do dtc <- getDTreeContext
       case erbAlgo of
         FMErbChernikova ->
          do mbSys <- chernikovaSimplifyIneqsG Nothing $
                      ineqSystem ((nV,0), rowVectorsToMatrix (nV+1) allBounds)
             case mbSys of
               Just sys -> return $ rowVectors $ getMatrix sys
               Nothing  -> return []
         FMErbSimplex -> return $ elim dtc bounds rest
         FMErbNone    -> return allBounds
    where
    allBounds = bounds ++ rest

    elim dtc [] new = new
    elim dtc (b:bs) new
	|  not $ and $ leavesDTree $ makeDTree dtc $ isRedundant (bs++new) b =
	    elim dtc bs (b:new)
	| otherwise  =  elim dtc bs new
	where
	system = negate b : bs ++ new
	ineqs  = ineqSystem ((nV,0), rowVectorsToMatrix (varPos+2) system)

    nV = varPos + 1

--
-- Turn a Fourier-Motzkin (or equivalent) function into
-- a sorting function for linearly parametrised inequality systems.
-- Note: the sorting function must return a DTree with a single
-- leaf (no case distinctions).
--
fmToLinSort :: (Fractional a, Ord a)
	    => (GIneqSystem a ->
		Maybe (GEqSystem a) ->
		Maybe (SortedSystem a)
	       )
	    -> GIneqSystem a
	    -> Maybe (GEqSystem a)
	    -> Maybe (GIneqSystem a)
fmToLinSort sortFn ineqs' mbEqs' =
    do sortedSys <- sortFn ineqs mbEqs
       let sys1 = sortedSystemToIneqSystem nVP 0 $
                  genericTake nVP $
                  sortedSys
       return $ moveVarsToParams sys1 0 nP 0
    where
    -- fourierMotzkin does not sort the parameter dimensions
    -- (like "outer variables"), so we move the parameters to
    -- the variables before sorting and move them back
    -- afterwards
    ineqs = moveParamsToVars ineqs' 0 nP 0
    mbEqs = fmap (\e -> moveParamsToVars e 0 nP 0) mbEqs'

    nV = nVars ineqs'
    nP = nParams ineqs'
    nVP = nV + nP


-- |
-- Sort a linear system using Fourier-Motzkin elimination.
-- This function respects the conditions on the parameters
-- present in the given system, i.e., sorting
--
-- >  1 <= x <= y
-- >  2 <= y <= p
--
-- yields
--
-- >  1 <= x <= p
-- >  2 <= y <= p
-- >  x <= y
-- >  2 <= p
--
sortIneqsFM :: (Fractional a, Ord a)
	    => GIneqSystem a          -- ^ input inequality system
	    -> Maybe (GEqSystem a)    -- ^ optional input equality system
            -> Maybe (GIneqSystem a)  -- ^ sorted system
sortIneqsFM = fmToLinSort (\is mbEs ->
			   runNumOrdDTreeGenAlg $
			   fourierMotzkin is mbEs Nothing
			  )

substitute :: Fractional a
           => Integer -> GVector a -> [GVector a] -> [GVector a]
substitute varPos e = 
    map (\r -> deleteEntries (r -  mulSV (vectorEntry r varPos * recip e1) e)
	                     varPos 1)
    where
    e1 = vectorEntry e varPos


-- selectOne [1,2,3] = [(1,[2,3]), (2,[1,3]), (3,[1,2])]
selectOne :: [a] -> [(a,[a])]
selectOne [] = []
selectOne (x:xs) = (x,xs) : [(y,x:ys) | (y,ys) <- selectOne xs]


----------------------------------------


nubIns :: Eq a => a -> [a] -> [a]
nubIns x [] = [x]
nubIns x (y:ys)
    | x == y     =  y : ys
    | otherwise  =  y : nubIns x ys

ordIns :: Num a =>
	  Bool -> (a -> Signum) -> GVector a -> [GVector a] -> [GVector a]
ordIns _     _     x [] = [x]
ordIns isLow mbSgn x (y:ys) =
    case isLow of
      True
	  | signumLe diff -> y:ys
	  | signumGe diff -> cont
	  | otherwise     -> y : cont
      False
	  | signumGe diff -> y : ys
	  | signumLe diff -> cont
	  | otherwise     -> y : cont
    where    
    cont = ordIns isLow mbSgn x ys
    -- 'diff' is the signum of (x-y) if constant
    diff
	| all (==0) vcs = mbSgn c
	| otherwise     = allSignumsFalse
	where
	(c:vcs) = reverse $ vectorEntries (x-y)


fourierMotzkinP :: (Integral a, Integral b)
		=> (   [QfFormula (Polynomial (Ratio a))]
		    -> Polynomial (Ratio a)
		    -> Signum
		   )
		-> [QfFormula (Polynomial (Ratio a))]
		-> GIneqSystem (FracPoly (Ratio a))
		-> Maybe (GEqSystem (FracPoly (Ratio a)))
		-> Maybe b
		-> DTree (Polynomial (Ratio a))
		         (Maybe (SortedSystem (FracPoly (Ratio a))))
fourierMotzkinP implSign ctx ineqs mbEqs mbCtxVars =
    makeDTree (implSignDTreeContext
	          ctx
	          implSign
	          (\fp -> signMonic (polyNumerator fp * polyDenominator fp))
	      ) $
    fourierMotzkin ineqs mbEqs mbCtxVars
    where
    mbSgn ctx2 p
	| Just p' <- isConstPoly p  =  orderingToSignum (p' `compare` 0)
	| otherwise                 =  implSign ctx2 p


-------------------------------------------

testCode nUnk dtree =
    testCode2 nUnk ['x' : show i | i <- [0..]] ['u' : show i | i <- [0..]] dtree

testCode2 :: (Integral a) =>
	     Int ->
	     [String] ->
	     [String] ->
	     DTree (Polynomial (Ratio a)) (Maybe (SortedSystem (FracPoly (Ratio a)))) ->
	     String
testCode2 nUnk varNames paramNames dtree =
    "#include <stdio.h>\n\
    \#include <stdlib.h>\n\
    \\n\
    \inline int min(int a, int b) { return a<b ? a : b; }\n\
    \inline int max(int a, int b) { return a<b ? b : a; }\n\
    \\n\
    \int ceild(int x, int y) {\n\
    \  if (y < 0) { x=-x; y=-y; }\n\
    \  return (x>=0) ? (x+y-1)/y : x/y;\n\
    \}\n\
    \int floord(int x, int y) {\n\
    \  if (y < 0) { x=-x; y=-y; }\n\
    \  return (x>=0) ? x/y : (x-y+1)/y;\n\
    \}\n\
    \\n\
    \int main(int argc, char *argv[]) {\n\
    \  int " ++
    concat (intersperse ", " ([varNames   !! i | i <- [0..nV-1]] ++
			      [paramNames !! i | i <- [0..nUnk-1]]) ) ++
    ";\n" ++
    concat ["  " ++ (paramNames!!i) ++ "=atoi(argv[" ++ show (i+1) ++ "]);\n" | i <- [0..nUnk-1]] ++
    "  printf(\"" ++ concat (intersperse "\\t" [varNames!!i | i <- [firstVar..nV+firstVar-1]])
    ++ "\\n\");\n" ++
    dtreeSorted2code varNames paramNames body dtree ++
    "  return 0;\n}\n"

    where
    nV = case [length l | Just l <- leavesDTree dtree] of
	   []  -> 0
	   n:_ -> n
    firstVar = case head [l | Just l <- leavesDTree dtree] of
		  Equation b : _         -> vectorLength b - 1
		  LowerUpper lbs ubs : _ -> vectorLength (head (lbs++ubs)) - 1
    body = "printf(\"" ++
	   concat (intersperse "," ((concat (replicate nV "%4d\\t") ++ "\\n\"")
				    : map (\x -> varNames!!x) [firstVar..nV+firstVar-1]))
	   ++ ");\n"


dtreeSorted2code :: (Integral a) =>
		    [String] -> [String] -> String ->
		    DTree (Polynomial (Ratio a)) (Maybe (SortedSystem (FracPoly (Ratio a)))) ->
		    String
dtreeSorted2code _ _ _ (DLeaf Nothing) = "printf(\"no solution\");"
dtreeSorted2code varNames paramNames body (DLeaf (Just ss)) =
    sorted2code varNames paramNames body ss
    where
    firstVar s = case s of
	          -- 'n' for [] is just a guess to make dtreeSorted2code
		  -- work for polyhedra which are not polytopes.
		  []                       -> n
		  Equation b : _           -> vectorLength b - 1
		  LowerUpper lbs ubs : s'
		      | null lbs && null ubs ->
			  firstVar s' - 1
		      | otherwise            -> vectorLength (head (lbs++ubs)) - 1
    fV = firstVar ss
    n = length ss
dtreeSorted2code vN pN body (SCond c tm t0 tp) =
    "if (" ++ qf2c pN (QfAt (AtGt (negate c))) ++ ") {\n" ++
    dtreeSorted2code vN pN body tm ++ "} else if (" ++
    qf2c pN (QfAt (AtEq c)) ++ ") {\n" ++
    dtreeSorted2code vN pN body t0 ++ "} else {\n" ++
    dtreeSorted2code vN pN body tp ++ "}\n"
dtreeSorted2code vN pN body (EqCond c t0 tpm) =
    "if (" ++ qf2c pN (QfAt (AtEq c)) ++ ") {\n" ++
    dtreeSorted2code vN pN body t0 ++ "} else {\n" ++
    dtreeSorted2code vN pN body tpm ++ "}\n"
dtreeSorted2code vN pN body (GeCond _ c t0p tm) =
    "if (" ++ qf2c pN (QfAt (AtGe c)) ++ ") {\n" ++
    dtreeSorted2code vN pN body t0p ++ "} else {\n" ++
    dtreeSorted2code vN pN body tm ++ "}\n"
dtreeSorted2code vN pN body (FCond qf tt tb) =
    "if (" ++ qf2c pN qf ++ ") {\n" ++
    dtreeSorted2code vN pN body tt ++ "} else {\n" ++
    dtreeSorted2code vN pN body tb ++ "}\n"				     


-- unvollstaending !!!
dtreeSorted2hs varNames paramNames (DLeaf (Just ss)) =
    sorted2hs varNames paramNames body ss
    where
    firstVar = case ss of
		  Equation b : _         -> vectorLength b - 1
		  LowerUpper lbs ubs : _ -> vectorLength (head (lbs++ubs)) - 1
    body = "(" ++
	   concat (intersperse "," (map (\x -> varNames!!x) [firstVar..n+firstVar-1])) ++ ")"
    n = length ss


qf2c :: Integral a => [String] -> QfFormula (Polynomial (Ratio a)) -> String
qf2c pN qf = qfFormulaToC (\p -> poly2c pN show (snd $ ratPoly2intPoly p)) qf


sorted2code :: (Integral a) =>
	       [String] -> [String] ->
	       String -> SortedSystem (FracPoly (Ratio a)) -> String
sorted2code vN pN body ss = 
    unlines (zipWith bnds2code ss [fV..]) ++
    replicate (2*length ss+4) ' ' ++ body ++ "\n" ++
    unlines [replicate (2*n) ' ' ++ "}" | n <- [length ss,length ss-1 .. 1]]
    where
    fV = firstVar ss
    firstVar [] = length ss  -- just a guess
    firstVar s =
	case head s of
	   Equation b                 -> vectorLength b - 1
	   LowerUpper lbs ubs
	       | null lbs && null ubs -> firstVar (tail s) - 1
	       | otherwise            -> vectorLength (head (lbs ++ ubs)) - 1
    bnds2code (Equation b) n =
	replicate (2*(n-fV)+2) ' ' ++ (vN!!n) ++ "=" ++
	mkBnd "" [b] ++ "; {"
    bnds2code (LowerUpper lbs ubs) n =
	replicate (2*(n-fV)+2) ' ' ++
	"for (" ++ (vN!!n)  ++ "=" ++ mkBnd "max" lbs ++ "; " ++
	(vN!!n) ++ "<=" ++ mkBnd "min" ubs ++ "; " ++ (vN!!n) ++ "++) {"

    mkBnd "max"  [] = "-infinity"
    mkBnd _      [] = "infinity"
    mkBnd minmax bs =
	foldl1 (\b1 b2 -> minmax ++ "(" ++ b1 ++ "," ++ b2 ++ ")")
	       (map (bnd2code minmax .vectorEntries) bs)

    bnd2code minmax coeffs
	| fact == 1 = expr
	| otherwise = divF ++ "(" ++ expr ++ "," ++ {-poly2c pN show fact-} showPolyStd pN fact ++ ")"
	where
	divF = if minmax=="max" then "ceild" else "floord"
--	expr = concat (zipWith (\c n -> if c/=0 then (poly2c pN show c ++ "*" ++ (vN!!n) ++ "+") else "") co [0..]) ++ poly2c pN show l
	expr = showPolyStd (take (length coeffs-1) vN ++ pN) $ affineToPoly vals
	l  = last vals
	co = init vals

	fact' = foldl lcmPoly 1 $ map polyDenominator coeffs
	vals' = map (\p -> (fact' `quotPoly` polyDenominator p) * polyNumerator p) coeffs
	fact2 = foldl lcm 1 $ map denominator $ concatMap collectCoeffs (fact':vals')

        (fact:vals) = map (mapPoly (\c -> (fact2 `quot` denominator c) * numerator c)) (fact':vals')

    fp2code fp
	| polyDenominator fp' == 1  =
	    "(" ++ poly2c pN show (polyNumerator fp') ++ ")"
	| otherwise  =
	    "(" ++ poly2c pN show (polyNumerator fp') ++ ")/(" ++
	    poly2c pN show (polyDenominator fp') ++ ")"
	where
	fp' :: FracPoly Double
	fp' = mapFP fromIntegral $ intFracPoly fp

sorted2hs :: (Integral a) =>
	       [String] -> [String] ->
	       String -> SortedSystem (FracPoly (Ratio a)) -> String
sorted2hs vN pN body ss = 
    "[" ++ body ++ " | " ++ concat (intersperse ", " $ zipWith bnds2code ss [firstVar..]) ++ "]"
    where
    firstVar = case head ss of
	         Equation b         -> vectorLength b - 1
		 LowerUpper lbs ubs -> vectorLength (head (lbs ++ ubs)) - 1
    bnds2code (Equation b) n =
	"let " ++ (vN!!n) ++ "=" ++
	mkBnd "" [b]
    bnds2code (LowerUpper lbs ubs) n =
	(vN!!n)  ++ " <- [" ++ mkBnd "maximum" lbs ++ " .. " ++
	mkBnd "minimum" ubs ++ "]"

    mkBnd minmax bs =
	minmax ++ "[" ++
	concat (intersperse "," $ map (bnd2code minmax . vectorEntries) bs) ++
	"]"

    bnd2code minmax coeffs
	| fact == 1 = expr
	| otherwise = divF ++ " (" ++ expr ++ ") (" ++ showPolyStd pN fact ++ ")"
	where
	divF = if minmax=="maximum" then "ceild" else "floord"
	expr = showPolyStd (take (length coeffs-1) vN ++ pN) $ affineToPoly vals
	l  = last vals
	co = init vals

	fact' = foldl lcmPoly 1 $ map polyDenominator coeffs
	vals' = map (\p -> (fact' `quotPoly` polyDenominator p) * polyNumerator p) coeffs
	fact2 = foldl lcm 1 $ map denominator $ concatMap collectCoeffs (fact':vals')

        (fact:vals) = map (mapPoly (\c -> (fact2 `quot` denominator c) * numerator c)) (fact':vals')

    fp2code fp
	| polyDenominator fp' == 1  =
	    "(" ++ poly2c pN show (polyNumerator fp') ++ ")"
	| otherwise  =
	    "(" ++ poly2c pN show (polyNumerator fp') ++ ")/(" ++
	    poly2c pN show (polyDenominator fp') ++ ")"
	where
	fp' :: FracPoly Double
	fp' = mapFP fromIntegral $ intFracPoly fp

poly2c :: Num a => [String] -> (a -> String) -> Polynomial a -> String
poly2c pN showC poly = "(" ++ p2c 0 poly ++ ")"
    where
    p2c n p
	| Just c <- isConstPoly p = showC c
	| otherwise =
	    concat $
	    intersperse "+" $
	    [showU (pN !! n) e ++ "(" ++ p2c (n+1) c ++ ")"
	    | (c,e) <- ces]
	where
	ces = coeffsAndExps $ univariate p

        showU u 0 = ""
	showU u 1 = u ++ "*"
	showU u e = u ++ "*" ++ showU u (e-1)
