{-
    This file is part of the LooPo project, see
    <http://www.infosun.fim.uni-passau.de/cl/loopo/>.

    Copyright (C) 2009 Armin Groesslinger
    Author: Armin Groesslinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-}


-- |
-- Computing integral zeros and signs of univariate polynomials.
--
module LMath.Algo.IntZeros
    ( IntSignList(..)
    , intSignList
    , IntInterval(..), IntIntervals
    , isEmptyIntIntervals
    , intersectIntIntervals, unionIntIntervals, complementIntIntervals
    , intPolySolutions
    , intPolyImpliesQf
--    , consistentXQf, impliesXQf, impliesXSignum, isIntegralCtxX
    )
where

import Debug.Trace

import Data.List ( sort )
import Data.Ratio

import LMath.Types

import LMath.Types.Polynomial.AlgNum
import LMath.Types.Polynomial.Roots

data IntSignList = SignTo Integer Ordering IntSignList
                 | SignToInfty    Ordering
                   deriving (Eq,Show)

simplifyIntSignList :: IntSignList -> IntSignList
simplifyIntSignList (SignTo a ord1 rest@(SignTo b ord2 _))
    | ord1 == ord2 = simplifyIntSignList rest
simplifyIntSignList (SignTo a ord1 rest@(SignToInfty ord2))
    | ord1 == ord2 = simplifyIntSignList rest
simplifyIntSignList (SignTo a ord rest) =
    SignTo a ord $ simplifyIntSignList rest
simplifyIntSignList (SignToInfty ord) = SignToInfty ord

intSignList :: Real a => Polynomial a -> IntSignList
intSignList f' = simplifyIntSignList $ zerosToIntSignList f zeros
    where
      f = fmap' toRational f'
      zeros = sort $ concatMap parToAlgnums $ rootsND [f]

signs :: Polynomial Rational -> [AlgNum Rational] -> [Ordering]
signs f [] = []
signs f (z:zs) = ordAt f (floorAlg z-1) : sgns (z:zs)
    where
      sgns [x] = [ordAt f (ceilingAlg z+1)]
      sgns (x:y:xs) =
          ordAt f (algRatBetween x y) : sgns (y:xs)

zerosToIntSignList :: Polynomial Rational -> [AlgNum Rational] -> IntSignList
zerosToIntSignList f [] = SignToInfty (ordAt f 0)
zerosToIntSignList f (z1:zs) =
    SignTo fz1 (ordAt f fz1) $ ztsl fz1 z1 (z1:zs)    
    where
      fz1 = ceilingAlg z1 - 1  -- the int left of z

      ztsl :: Integer -> AlgNum Rational -> [AlgNum Rational] -> IntSignList
      ztsl ft t [] = SignToInfty (ordAt f (floorAlg t+1))
      ztsl ft t (z:zs)
          | isInt z    =  (if ft+1 < fz
                           then SignTo (ceilingAlg z-1) (ordAt f (algRatBetween t z))
                           else id
                          ) $
                          SignTo fz EQ $
                          ztsl fz z zs
          | ft < fz    =  leftPart $ ztsl fz z zs
          | otherwise  =  ztsl ft t zs
          where
            fz = floorAlg z
            leftPart = if ft < fz
                       then SignTo fz (ordAt f (algRatBetween t z))
                       else id

ordAt :: Real a => Polynomial Rational -> a -> Ordering
ordAt f x = evalPoly f [toRational x] `compare` 0

isInt :: AlgNum Rational -> Bool
isInt alg = fromIntegral (floorAlg alg) == alg


data IntInterval = IntInterval (Maybe Integer) (Maybe Integer)
                   deriving (Eq, Show)

type IntIntervals = [IntInterval]

isEmptyIntIntervals :: IntIntervals -> Bool
isEmptyIntIntervals [] = True
isEmptyIntIntervals _  = False

simplifyIntIntervals :: [IntInterval] -> [IntInterval]
simplifyIntIntervals [] = []
simplifyIntIntervals (IntInterval a (Just b) : r@(IntInterval (Just c) d : rest))
    | b+1 == c   =  simplifyIntIntervals (IntInterval a d : rest)
simplifyIntIntervals (a@(IntInterval _ _) : rest) =
    a : simplifyIntIntervals rest

rLt :: Maybe Integer -> Maybe Integer -> Bool
rLt (Just _) Nothing  = True
rLt (Just x) (Just y) = x < y
rLt _        _        = False

rlLt :: Maybe Integer -> Maybe Integer -> Bool
rlLt Nothing  Nothing  = False
rlLt (Just _) Nothing  = False
rlLt (Just x) (Just y) = x < y
rlLt Nothing  (Just _) = False

lMax :: Maybe Integer -> Maybe Integer -> Maybe Integer
lMax Nothing y = y
lMax x Nothing = x
lMax (Just x) (Just y) = Just $ max x y

intersectIntIntervals :: [IntInterval] -> [IntInterval] -> [IntInterval]
intersectIntIntervals as bs = simplifyIntIntervals $ iii as bs
    where
      iii _  [] = []
      iii [] _  = []
      iii i1@(IntInterval l1 r1:rest1) i2@(IntInterval l2 r2:rest2)
          | r1 `rlLt` l2  =  iii rest1 i2
          | r2 `rlLt` l1  =  iii i1    rest2
          | r1 `rLt` r2   =  IntInterval (lMax l1 l2) r1 : iii rest1 i2
          | r2 `rLt` r1   =  IntInterval (lMax l1 l2) r2 : iii i1 rest2
          | r1 == r2      =  IntInterval (lMax l1 l2) r1 : iii rest1 rest2

complementIntIntervals :: [IntInterval] -> [IntInterval]
complementIntIntervals [] = [IntInterval Nothing Nothing]
complementIntIntervals [IntInterval Nothing Nothing]  = []
complementIntIntervals [IntInterval (Just a) Nothing] = [IntInterval Nothing (Just (a-1))]
complementIntIntervals (IntInterval _ (Just a) : rest) = compl a rest
    where
      compl left [] = [IntInterval (Just (left+1)) Nothing]
      compl left [IntInterval (Just b) Nothing]
          | b-1 >= left+1  =  [IntInterval (Just (left+1)) (Just (b-1))]
          | otherwise      =  []
      compl left (IntInterval (Just b) (Just c) : r)
          | b-1 >= left+1  =  IntInterval (Just (left+1)) (Just (b-1)) :
                              compl c r
          | otherwise      =  compl c r


unionIntIntervals :: [IntInterval] -> [IntInterval] -> [IntInterval]
unionIntIntervals as bs =
    simplifyIntIntervals $
    complementIntIntervals $
    intersectIntIntervals (complementIntIntervals as) 
                          (complementIntIntervals bs)

intPolySolutions :: Real a => Polynomial a -> [Ordering] -> [IntInterval]
intPolySolutions f ord = simplifyIntIntervals $ toInters ord $ intSignList f

toInters :: [Ordering] -> IntSignList -> [IntInterval]
toInters reqOrd (SignToInfty ord)
    | ord `elem` reqOrd = [IntInterval Nothing Nothing]
    | otherwise         = []
toInters reqOrd list = inters Nothing list
    where
      inters left (SignToInfty ord)
          | ord `elem` reqOrd  =  [IntInterval left Nothing]
          | otherwise          =  []
      inters left (SignTo x ord rest)
          | ord `elem` reqOrd  =  IntInterval left (Just x) : inters (Just (x+1)) rest
          | otherwise          =  inters (Just (x+1)) rest

intPolyQfSolutions :: Real a => QfFormula (Polynomial a) -> IntIntervals
intPolyQfSolutions (QfAnd f g) =
    intersectIntIntervals (intPolyQfSolutions f) (intPolyQfSolutions g)
intPolyQfSolutions (QfOr f g) =
    unionIntIntervals (intPolyQfSolutions f) (intPolyQfSolutions g)
intPolyQfSolutions (QfAt at) = intPolyAtSolutions at

intPolyAtSolutions :: Real a => AtFormula (Polynomial a) -> IntIntervals
intPolyAtSolutions at = intPolySolutions f os
    where
      (f,os) = ords at

      ords (AtEq f)  = (f,[EQ])
      ords (AtGt f)  = (f,[GT])
      ords (AtGe f)  = (f,[EQ,GT])
      ords (AtNeq f) = (f,[LT,GT])

intPolyImpliesQf :: Real a
                 => [QfFormula (Polynomial a)]
                 -> QfFormula (Polynomial a)
                 -> Bool
intPolyImpliesQf ctx p =
    isEmptyIntIntervals $
    intPolyQfSolutions (conjunctionQf (qfNot p : ctx))

{-
--
-- impliesQf for X
--

--
-- we evaluate the formulas in the order
-- given by the list for the "other" cases (between the left and
-- right FPP) such that we do not run into undefinedness
-- in the later formulas if the earlier formulas are
-- inconsistent!
--
consistentXQf :: Integral a => [QfFormula (X a)] -> Bool
consistentXQf (formulas :: [QfFormula (X a)]) =
--    trace (show (leftCases,rightCases,otherCases)) $
    not inconsistent
    where
      formula = conjunctionQf formulas

      formulaL, formulaR :: QfFormula (PP a)
      formulaL = fmap (denomFree . leftFppX)  $ formula
      formulaR = fmap (denomFree . rightFppX) $ formula
      termsL   = termsQf formulaL
      termsR   = termsQf formulaR
      periL = foldl lcm 1 $ map periodPP termsL
      periR = foldl lcm 1 $ map periodPP termsR

      inconsistent =
          and [all null leftCases 
              ,all null rightCases
              ,null otherCases
              ]
      leftCases  = [solsL i | i <- [0..periL-1]]
      rightCases = [solsR i | i <- [0..periR-1]]
      otherCases =
          [i | i <- [left+1 .. right-1]
             , let i' = fromIntegral i
             , let ev f = if isdefX f i'
                          then Just (evalX f i' `compare` 0)
                          else Nothing
             , Just True == evalQfMb ev formula
          ]

      solsL i = intPolyQfSolutions (fmap (\f -> constituentPP periL f i) formulaL)
                `intersectIntIntervals`
                [IntInterval Nothing (Just $ leftBound i)]
      leftBound i = (left-i) `div` periL

      solsR i = intPolyQfSolutions (fmap (\f -> constituentPP periR f i) formulaR)
                `intersectIntIntervals`
                [IntInterval (Just $ rightBound i) Nothing]
      rightBound i = (right-i) `div` periR

      left  = minimum $ map leftX  $ termsQf formula
      right = maximum $ map rightX $ termsQf formula

      denomFree fpp = numeratorFPP fpp * denominatorFPP fpp

impliesXQf :: Integral a
           => [QfFormula (X a)]
           -> QfFormula (X a)
           -> Bool
impliesXQf ctx f = not $ consistentXQf (qfNot f : ctx)

impliesXSignum :: Integral a
               => [QfFormula (X a)]
               -> X a
               -> Signum
impliesXSignum ctx f =
--    trace (show (ctx,f)) $
--    (\s -> trace (show s) s) $
    genericImpliesSignum impliesXQf ctx f


isIntegralCtxX :: Integral a
               => [QfFormula (X a)]
               -> X a
               -> Bool
isIntegralCtxX ctx f = signumEq $ impliesXSignum ctx (f - floorX f)
-}