{-
    This file is part of the LooPo project, see
    <http://www.infosun.fim.uni-passau.de/cl/loopo/>.

    Copyright (C) 2009 Armin Groesslinger
    Author: Armin Groesslinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-}

module LMath.Algo.LGen
    ( LCode(..),
      LGenParams(..), LGenOptions(..), LGenOutputOptions(..),
      LGenAlgoVariant(..), LGenLpVariant(..),
      lgenOptionsStd, emptyLGenOutputOptions,
      lgen, codeWithStatements, parseLGenInput,
      writeLGenInput,
      runLGenWithStatements, runLGen,
      loopGenIF, loopGenSelfNP
    )
where

import LMath.Types
import LMath.Algo.AffineImage
import LMath.Algo.CAD
import LMath.Algo.OmegaTest
import LMath.Algo.Simplex
import LMath.Algo.Sort
import LMath.Logic

import LMath.Tools.Qepcad
import LMath.Tools.Reduce

import Control.Exception ( assert, evaluate )
import Control.Monad     ( liftM, when )
import Data.Graph as Graph
import Data.List
import Data.Map ( Map )
import qualified Data.Map as Map
import Data.Maybe
import Data.Ratio
import Data.Set ( Set)
import qualified Data.Set as Set
import Data.Tree as Tree
import Text.ParserCombinators.Parsec

import Debug.Trace
import System.IO
import System.IO.Unsafe
import System.Time


----------------
-- data types --
----------------

-- [Conj] represents a disjunction
type Conj = [Vector]
type Disj = [Conj]

trueConj :: Conj
trueConj = []    -- empty conjunction

data LCode = LCode { lcodeBounds     :: [([[Vector]],[[Vector]])],
                     lcodeParamConds :: [IneqSystem],
                     lcodeKloetze    :: [Klotz]
                   }
             deriving Show

data Klotz = Klotz { outerIneqs :: IneqSystem,
                     outerConds :: [Conj],
                     innerPart  :: KlotzInner
                   }
	     deriving Show

data KlotzInner = Loop { kiLower :: [Vector]   -- lower bounds
                       , kiUpper :: [Vector]   -- upper bounds
		       , kiConds :: [Conj]     -- disjunction of conds
                       , kiInner :: [([Conj],KlotzInner)]
                       }
                | NoLoop { kiInner :: [([Conj],KlotzInner)]
                         }
                | Statement { kiStmtID :: Integer
                            , kiPreImage :: AffineFunction
                            }
		  deriving Show

statements :: KlotzInner -> [Integer]
statements (Statement {kiStmtID=sid}) = [sid]
statements (Loop {kiInner=inners}) = nub $ concatMap (statements.snd) inners


--------------------
-- main functions --
--------------------


lgen :: LGenOptions
     -> IneqSystem
     -> [[(IneqSystem,GAffineFunction Integer)]]
     -> [LCode]
lgen lgo ctx input =
    genericTake (tD+1) $
    map (mkOuterLoops lgo ctx) $
    map fst $
    iterate (step lgo ctx) $
    ([
      Klotz { outerIneqs = image (fmap fromIntegral af) is,
	      outerConds = [map (\v -> deleteEntries v tD nI) $
                            filter (isRelevant tD nI) $
			    rowVectors $ getMatrixIH inv],
	      innerPart  = -- trace (show (is,af,fmInv)) $
                           mkInner nP tD fmInv $
                           Statement { kiStmtID   = id,
				       kiPreImage = inv -- affineFunction ((nVtarget,nP),inv) Homogeneous
				     }
	    }
      | (id,xs) <- zip [1..] input, (is,af) <- xs,
        let nI = nVars is,        -- number of iterator dimensions
        let fmInv = fmPreImage is (fmap fromIntegral af),
        let inv   = mkInv nP tD nI fmInv
--        let inv = someInv $ fmap fromIntegral $ getMatrix af,
     ]
    ,[])
    where
    isRelevant :: Int -> Int -> Vector -> Bool
    isRelevant tD nI r =
        all (==0) (take nI (drop tD ents)) &&  -- not a condition on iterators
        foldl lcm 1 (map denominator $ ents) /= 1
        where
          ents = vectorEntries r

{-
    -- compute "some" inverse for the scatter function, i.e. for
    --   (i) |-> (i,i)
    -- possible inverses are
    --   (x,y) |-> x
    --   (x,y) |-> y
    --   (x,y) |-> x/2+y/2
    -- etc.
    someInv :: Matrix -> Matrix
    someInv m = subMatrixHor inv 0 (nV-1) `join`
		subMatrixHor inv (nRows inv - nP - 1) (nRows m-1)
	where
	inv = inverseMatrix $
	      insertColumns m nV $
	      columnVectorsToMatrix (nRows m) $ baseExt m
        nV  = nColumns m - nP - 1
-}
    (someIS,someAF) = head $ concat input
    nP = nParams someIS
    tD = targetDimIH someAF  -- number of target dimensions

--
-- [params, target dims, <iters>, 1] -> [target dims, <iters>, params, 1]
-- (number of iters varies)
--
mkInner :: Int -> Int -> SortedSystem Rational -> KlotzInner -> KlotzInner
mkInner nP tD ss stmt = mk 0 ss
    where
      mk :: Int -> SortedSystem Rational -> KlotzInner
      mk d (LowerUpper ls us : rest) =
          Loop { kiLower = map fixOrdering ls
               , kiUpper = map fixOrdering us
               , kiConds = [[]]
               , kiInner = [ ([[]], mk (d+1) rest) ]
               }
      mk d (Equation e : rest)
          | isRel  =
              NoLoop { kiInner = [ ([[fixOrdering e]], inner) ] }
          | otherwise                    =
              NoLoop { kiInner = [([[]], inner)] }
          where
            inner = mk (d+1) rest
            -- isRel: condition relevant, i.e., in the iterators
            isRel = any ((/=1) . denominator) ents &&
                    any (/=0) (init $ drop (nP+tD) ents)
            ents = vectorEntries e
      mk d [] = stmt
          
      fixOrdering v =
          vector (targets ++ init r2 ++ params ++ [last r2])
          where
            (params,  r1) = splitAt nP ents
            (targets, r2) = splitAt tD r1
            ents = vectorEntries v

--
-- Kommt bei   x  |->  (2*x, x)
-- immer die Bedingung  0.5*x  heraus, oder kommt
-- auch mal  x  heraus und die Aufzaehlung ist dann falsch?
--

--
-- [parameters, target dims, <iters>, 1] -> [target dims, iterator dims, parameters, 1]
mkInv :: Int -> Int -> Int -> SortedSystem Rational -> AffineFunction
mkInv nP tD nI ss = affineFunction ((tD+nI,nP), 
                                    matrix $ zipWith mkDim [0..] ss
                                   ) Inhomogeneous
    where
      mkDim d (LowerUpper ls us) =
          replicate (tD+d) 0 ++ 1 : replicate (nI-d-1+nP+1) 0
      mkDim d (Equation e) = 
          tPart ++ take d vPart2 ++ replicate (nI-d) 0 ++ pPart ++ [last vPart2]
          where
            cs = vectorEntries e
            (pPart, vPart1) = splitAt nP cs
            (tPart, vPart2) = splitAt tD vPart1

--
-- returns an enumeration of the iterator dimensions (i.e.,
-- of 'sys') in dependence of the parameters and
-- the target dimensions of 'af'. (parameters in front)
--
fmPreImage :: IneqSystem -> AffineFunction -> SortedSystem Rational
fmPreImage sys af =
    fixSelfRefs $
    case
     makeDTree numOrdDTreeContext $
     sortIneqsGss ins (Just eqs) (Just 0)
    of
      DLeaf (Just x) -> drop (nP+tD) x
      _              -> error "fmPreImage"
    where
      eqs = eqSystemVPC (afP `conc` negate (unitMatrix tD tD) `conc` afV, zeroMatrix tD 0, afC)
      ins = ineqSystemVPC (sysP `conc` zeroMatrix (nRows sys) tD `conc` sysV, zeroMatrix (nRows sys) 0, sysC)

      (afV, afP, afC)    = partsVPCih af
      (sysV, sysP, sysC) = partsVPC   sys
      nV = nVars sys
      nP = nParams sys
      tD = targetDimIH af

--
-- A dimension which is defined by an equation is not
-- enumerated (i.e., no loop is generated for it).
-- Therefore, we must change all references to this dimension
-- by substituting its definition (i.e., the expression for it)
-- in inner dimensions.
--
-- [Equation Vector 0 0 0 1.0  0
-- ,Equation Vector 0 0 0  0  1.0 0
-- ]
--  -->
-- [Equation Vector 0 0 0 1.0 0
-- ,Equation Vector 0 0 0 1.0 0 0
-- ]
--
fixSelfRefs :: Num a => SortedSystem a -> SortedSystem a
fixSelfRefs ss = map fixsr ss
    where
    fixsr bnd = foldl (\b i -> fmap (fix i) b) bnd [nOuter .. len-1]
        where
          len = case bnd of
                  Equation e       -> vectorLength e
                  LowerUpper [] [] -> nOuter
                  LowerUpper ls us -> vectorLength $ head (ls++us)

    -- lookup the definition of dimension i (i.e., the bound with index
    -- i-nOuter) and if it is an equation, substitute its definition
    -- into v.
    fix i v =
        case (vectorEntry v i, ss !! (i-nOuter)) of
          (0, _)    ->  v
          (c, Equation w) -> v' + c `mulSV` w'
              where
              v' = setVectorEntry v i 0
              w' = insertEntries w (wlen - 1) (replicate (vlen-wlen) 0)
              wlen = vectorLength w
              vlen = vectorLength v
          _  ->  v

    Just nOuter = ssOuterDims ss

{-
--
-- add unit vectors to the columns of 'm' to achieve
-- full row rank
--
baseExt :: Matrix -> [Vector]
baseExt m
    | [0..nColumns m-1] `isPrefixOf` cols  =
	[columnVectors unit !! (c-nColumns m) | c <- drop (nColumns m) cols]
    | otherwise = error ("baseExt: given matrix does not have full column rank"
                         ++ "\nThis is a deficiency in LGen :-(\n" ++ show m)
    where
    cols = [c | (_,c) <- findPivots ech]
    (_,ech) = echelonReduction ext
    ext = m `conc` unit
    unit = unitMatrix (nRows m) (nRows m)
-}

--
-- generate the outer loops
--
mkOuterLoops :: LGenOptions -> IneqSystem -> [Klotz] -> LCode
mkOuterLoops lgo ctx ks =
    LCode { lcodeBounds     = bounds,
            lcodeParamConds = paramConds,
            lcodeKloetze    = ks'
          }
    where
    ks'
      | lgoSimplifyIFs lgo  =  concatMap (postOptK ctx) ks
      | otherwise           =  ks

    simp is = is'
        where
        [is'] = simplifyDomain extCtxs [is]

    hull = {- filterSub $ -} simplifyDomain [extCtx] $
           nondisjointUnion $ map outerIneqs ks

    bounds = [(lower n,upper n) | n <- [0..nV-1]]

    sorted :: [IneqSystem]
    sorted
        | lgoConvexUnion lgo = [cvxSorted]            
        | otherwise          =
            [s | sys <- hull, let Just s = sortIneqs sys Nothing]

    paramConds = nondisjointUnion $
		 [deleteVars (filterIneqs (\i -> all (==0) $ take nV i) s)
                            0 nV
		  | s <- sorted]

    Just cvxSorted = sortIneqs cvxUnion Nothing
    Just cvxUnion  = convexUnionAll hull

    lower n = simpl $ map (\s -> deleteVars s (n+1) (nV-n-1)) (lowers!!n)
    upper n = simpl $ map (\s -> deleteVars s (n+1) (nV-n-1)) (uppers!!n)

    -- extCtxs describes the iteration space enumerated by the
    -- outer for loops.
    extCtxs :: [IneqSystem]
    extCtxs = lctx !! nV -- timeIt "extCtxs" $ intersectDomains (lowers++uppers)

    -- lowers and uppers have only inequalities where the n-th variable
    -- has a non-zero coefficient
    lowers, uppers :: [[IneqSystem]]
    lowers = [simplifyDom n $ map (filterIneqs (\i -> i!!n/=0)) (lowers0!!n) | n <- [0..nV-1]]
    uppers = [simplifyDom n $ map (filterIneqs (\i -> i!!n/=0)) (uppers0!!n) | n <- [0..nV-1]]

    simplifyDom n d = timeIt lgo ("simplifyDom " ++ show n) $
                      if lgoSimplifyOuterLoops lgo
                         then simplifyDomain (lctx !! n) d
                         else d

    lctx :: [[IneqSystem]]
    lctx = [extCtx] : [timeIt lgo ("lctx " ++ show (n+1)) $ intersectDomains [lowers!!n,uppers!!n,prev] | prev <- lctx | n <- [0..nV-1]]

    -- lowers0 and uppers0 have inequalities where the n-th variable can be 0.
    lowers0 = [[filterIneqs (\i -> part n i && i!!n>=0) s | s <- sorted]
              | n <- [0..nV-1]]
    uppers0 = [[filterIneqs (\i -> part n i && i!!n<=0) s | s <- sorted]
              | n <- [0..nV-1]]
--    lctx 0 = [extCtx]
--    lctx n = timeIt ("lctx " ++ show n) $ intersectDomains ([extCtx] : take n lowers ++ take n uppers)

    part n i = all (==0) $ take (nV-n-1) $ drop (n+1) i

    simpl :: [IneqSystem] -> [[Vector]]
    simpl [] = []
    simpl is = -- (\r -> trace (show ("simpl: " ++ show (is,r))) r) $
	       nub $ map solveLastVar is

    solveLastVar :: IneqSystem -> [Vector]
    solveLastVar is = rs
	where
	f v = fmap (/ negate (vectorEntry v lv)) $ deleteEntries v lv 1
	rs = map f $
--	     filter (\v -> vectorEntry v lv /= 0) $
	     rowVectors $ getMatrix is
	lv = nVars is - 1

    nV = nVars   $ head hull
    nP = nParams $ head hull

    extCtx :: IneqSystem
    extCtx = insertVars ctx 0 (zeroMatrix (nRows ctx) (nVars $ outerIneqs $ head ks))

--    Just sys = convexUnionAll [outerIneqs k | k <- ks]
--    context  = insertVars ctx 0 $ zeroMatrix (nRows ctx) (nVars sys)


--
-- return the lower and upper bounds for a given variable
-- in a given inequality system
--
extractBounds :: (Fractional a, Ord a)
	      => GIneqSystem a -> Integer -> ([GVector a], [GVector a])
extractBounds sys var = (lower, upper)
    where
    lower = [fmap (/ negate (vectorEntry l var)) $
             deleteEntries l var 1 | l <- ls]
    upper = [fmap (/ negate (vectorEntry u var)) $
             deleteEntries u var 1 | u <- us]
    (ls,us) = partition (\r -> vectorEntry r var > 0) $
	      filter (\r -> vectorEntry r var /= 0) $
              rowVectors $ getMatrix sys





runLGenWithStatements :: LGenOptions -> LGenParams -> LGenOutputOptions -> [String]
runLGenWithStatements lgo lgp lgoo =
    map (codeWithStatements lgoo) lg
    where
    lg = lgen lgo (lgpContext lgp) (lgpStatements lgp)


runLGen :: LGenOptions -> LGenParams -> LGenOutputOptions -> [String]
runLGen lgo lgp lgoo =
    runLGenWithStatements lgo lgp lgoo

lgenStdStmtGen :: LGenOutputOptions -> Integer -> [[Rational]] -> String
lgenStdStmtGen lgoo i varexps =
    'S' : show i ++ "(" ++ concat (intersperse "," $ map gen varexps) ++ ");"
    where
    gen :: [Rational] -> String
    gen rs = if l /= 1
               then "(" ++ s ++ ")/" ++ show l
               else s
	where
	(s,l) = ratl2code viNames paramNames nP rs

    viNames = genericTake nVI (lgooVarNames lgoo ++ lgooIterNames lgoo)
    paramNames = lgooParamNames lgoo
    nP = genericLength paramNames
    nVI = genericLength (head varexps) - nP - 1

{-
code :: [String] -> [String] -> LCode -> String
code varNames parNames lcode = codeWithStatements varNames parNames stmts lcode
    where
    stmts i varexps = 'S' : show i ++ "(" ++
                      concat (intersperse "," varexps) ++ ");"
-}

codeWithStatements :: LGenOutputOptions -> LCode -> String
codeWithStatements lgoo (LCode {lcodeBounds=bounds,lcodeKloetze=ks,lcodeParamConds=paramConds}) =
--    trace (show hull) $
    "if ((" ++ 
    concat (intersperse ") || (" $ [ineqs2c lgoo pc | pc <- paramConds]) ++ ")) {\n" ++
    unlines [replicate (2*n+2) ' ' ++ genFor lgoo nP (fromIntegral n) var (lower n) (upper n) stmtIDs ++ " {" | n <- [0..nV-1] | var <- varNames] ++
    indent (2*nV+2) (concatMap (klotz2code lgoo nP (fromIntegral nV)) ks) ++
    unlines [replicate (2*n) ' ' ++ "}" | n <- [nV,nV-1..1]] ++
    "}\n"
    where
    varNames   = lgooVarNames lgoo
    paramNames = lgooParamNames lgoo
    stmtGen = case lgooStmtGen lgoo of
                  Nothing   -> lgenStdStmtGen lgoo
                  Just sGen -> sGen
    lower n = fst (bounds!!n)
    upper n = snd (bounds!!n)

    stmtIDs = nub $ concatMap (statements . innerPart) ks

    nV = length bounds
    -- nP = (vectorLength $ head $ concat $ fst $ head bounds) - 1
    nP = nParams $ head (paramConds ++ [outerIneqs (head ks)])

indent :: Int -> String -> String
indent n str =
    unlines [replicate n ' ' ++ l | l <- lines str]

klotz2code :: LGenOutputOptions -> Integer -> Integer -> Klotz -> String
klotz2code lgoo nP n k =
    "if (" ++ condition ++ ") {\n" ++
    indent 2 (klotzInner2code lgoo nP n (innerPart k)) ++
    "}\n"
    where
    condition
        | null ineqConds = -- ineqConds is the empty union (shouldn't happen)
            "0"       
        | otherwise      =
	    concat [conds2c lgoo nP (outerConds k),
                    " && ",
		    concat $ intersperse " || " $
                        map (ineqs2c lgoo) ineqConds
	           ]
    oi = outerIneqs k
    ineqConds = [oi]

klotzInner2code :: LGenOutputOptions -> Integer -> Integer -> KlotzInner -> String
klotzInner2code lgoo nP n ki@(Loop {kiLower=ls,kiUpper=us,kiConds=bcs,kiInner=inners}) =
    genFor lgoo nP n var [ls] [us] (statements ki) ++ " {\n" ++
    indent 2 (bodyCond $ klotzInner2code lgoo nP n $ NoLoop {kiInner=inners})
    ++ "}\n"
    where
    bodyCond :: String -> String
    bodyCond body
	| bcs == [[]]  =  body
	| otherwise    =
	    "if (" ++ conds2c lgoo nP bcs ++ ") {\n" ++
	    indent 2 body ++
	    "}\n"

    varNames   = lgooVarNames lgoo ++ lgooIterNames lgoo
    paramNames = lgooParamNames lgoo
    var = varNames !! fromIntegral (nV-1)
    nV = vectorLength (head (ls++us)) - nP

klotzInner2code lgoo nP n (NoLoop    {kiInner=inners}) =
    concat $ map genInner inners
    where
    genInner :: ([Conj],KlotzInner) -> String
    genInner ([[]],ki) = klotzInner2code lgoo nP (n+1) ki
    genInner (cs,ki) =
	"if (" ++ conds2c lgoo nP cs ++ ") {\n" ++
	indent 2 (klotzInner2code lgoo nP (n+1) ki) ++
	"}\n"
    
klotzInner2code lgoo nP n (Statement {kiStmtID=sid,kiPreImage=pi}) =
    (stmtGen sid $ rows' $ getMatrixIH pi) ++ "\n"
    where
    stmtGen = fromMaybe (lgenStdStmtGen lgoo) (lgooStmtGen lgoo)
    varNames   = lgooVarNames lgoo ++ lgooIterNames lgoo
    paramNames = lgooParamNames lgoo


genFor :: LGenOutputOptions -> Integer -> Integer -> String -> [[Vector]] -> [[Vector]] -> [Integer] -> String
genFor lgoo nP n var ls us stmtIDs =
    case (lgooLoopGen lgoo) n var ls us lb ub stmtIDs of
       Just loophead -> loophead
       Nothing       -> concat ["for (", var, "=", lb, "; ", var, "<=", ub, "; ", var, "++)"]
    where
    varNames   = genericTake n (lgooVarNames lgoo ++ lgooIterNames lgoo)
    paramNames = lgooParamNames lgoo

    lb = bounds "min" "max" "ceild" ceiling ls
    ub = bounds "max" "min" "floord" floor us

    bounds :: String -> String -> String -> (Rational -> Integer) -> [[Vector]] -> String
    bounds outerMinmax minmax divname rndfn [] = "INFEASIBLE"
    bounds outerMinmax minmax divname rndfn [b] = bound minmax divname rndfn b
    bounds outerMinmax minmax divname rndfn (b:bs) =
	outerMinmax ++ "(" ++ bound minmax divname rndfn b ++ "," ++
	bounds outerMinmax minmax divname rndfn bs ++ ")"

    bound :: String -> String -> (Rational -> Integer) -> [Vector] -> String
    bound minmax divname rndfn []  = "UNBOUNDED"
    bound minmax divname rndfn [b] = gen divname rndfn b
    bound minmax divname rndfn (b:bs) =
        minmax ++ "(" ++ gen divname rndfn b ++ "," ++
        bound minmax divname rndfn bs ++ ")"

    gen divname rndfn v
        | all (==0) $ init ents  = show $ rndfn $ last ents
	| l == 1     =  s
	| otherwise  =  divname ++ '(':s ++ ',':show l ++ ")"
	where
	(s,l) = ratl2code varNames paramNames nP ents
        ents  = vectorEntries v


ineqs2c :: LGenOutputOptions -> IneqSystem -> String
ineqs2c lgoo sys
    | nRows sys == 0  =  "1"
    | otherwise       =
	concat $
	intersperse " && " (map (gen ">=0") (rows' is) ++
			    map (gen "==0") (rows' es)
			   )
    where
    (is,es) = findEqualities sys

    gen :: String -> [Rational] -> String
    gen comp rs = s ++ comp
	where
	(s,_) = ratl2code varNames (lgooParamNames lgoo) nP rs
    nP = nParams is

    varNames = take (nVars sys) (lgooVarNames lgoo ++ lgooIterNames lgoo)

ratl2code :: [String] -> [String] -> Integer -> [Rational] -> (String,Integer)
ratl2code varNames paramNames nP rs =
--    trace (show (varNames,paramNames,rs,nP)) $
    assert (length (varNames ++ paramNames) == length rs - 1) $
    ( if head str == '+' then tail str else str
    , l)
    where
    str = concat [showP v var   | v <- vs | var   <- varNames] ++
          concat [showP p param | p <- ps | param <- paramNames] ++
          (if co /= 0 || all (==0) (vs++ps) then showWithSign co else "")
    (vs,ps')  = genericSplitAt nV coeffs
    (ps,[co]) = genericSplitAt nP ps'
    coeffs = map (\r -> (l `quot` denominator r)*numerator r) rs
    l = foldl lcm 1 $ map denominator rs
    nV = genericLength rs - nP - 1

conds2c :: LGenOutputOptions -> Integer -> [Conj] -> String
conds2c _ nP []   = "0"
conds2c _ nP [[]] = "1"
conds2c lgoo nP cs =
    "(" ++ concat (intersperse ") || (" [concat $ intersperse " && " [gen (vectorEntries v) | v <- c] | c <- cs]) ++ ")"
    where
    gen :: [Rational] -> String
    gen rs = "(" ++ numExpr ++ ")%" ++ show denom ++ " == 0"
             -- l == 1 does not happen for module conditions
	where
          (numExpr,denom) = ratl2code varNames paramNames nP rs
          varNames = genericTake (genericLength rs - nP - 1)
                                 (lgooVarNames lgoo ++ lgooIterNames lgoo)
          paramNames = lgooParamNames lgoo


showP :: Integer -> String -> String
showP 0 _ = ""
showP 1    s = '+' : s
showP (-1) s = '-' : s
showP n s
    | n > 0      =  '+' : show n ++ '*':s
    | otherwise  =  show n ++ '*':s

showWithSign :: Integer -> String
showWithSign n
    | n >= 0     =  '+':show n
    | otherwise  =  show n


postOptK :: IneqSystem -> Klotz -> [Klotz]
postOptK ctx k =
--    trace ("postOptK: extCtx'=" ++ show extCtx') $
    case postOptKI extCtx' (innerPart k) of
        Just ki -> [k { innerPart  = ki,
		        outerIneqs = simplifyIneqs (Just extCtx) oi
		      }
                   ]
        Nothing -> []
    where
    extCtx' = extCtx `join` oi
    extCtx  = insertVars ctx 0 (zeroMatrix (nRows ctx) (nVars oi)) 
    oi = outerIneqs k

postOptKI :: IneqSystem -> KlotzInner -> Maybe KlotzInner
postOptKI extCtx ki@(Statement {}) = Just ki
postOptKI extCtx (NoLoop { kiInner=inners })
    | null newInner  =  Nothing
    | otherwise      =  Just $ NoLoop { kiInner = newInner }
    where
    newInner = [(cs,newI) | (cs,ki) <- inners
                          , Just newI <- [postOptKI extCtx' ki]
               ]
    extCtx' = insertVars extCtx (nVars extCtx) (zeroMatrix (nRows extCtx) 1)
postOptKI extCtx ki@(Loop {})
    | notEmpty  =
--       trace ("postOptKI: extCtx=" ++ show extCtx ++ "," ++
--              show (kiLower ki,kiUpper ki, newLower, newUpper,extCtx',luSys,bounds)) $
       Just $ ki { kiLower=newLower, kiUpper=newUpper, kiInner=newInner }
    | otherwise = Nothing
    where
    newInner = [(cs,newI)| (cs,i) <- kiInner ki, Just newI <- [postOptKI newExtCtx i]]
    newExtCtx = extCtx' `join` bounds
    (newLower,newUpper) = extractBounds bounds nV
    luSys = ineqSystem ((nV+1,nP),ls `join` us)
    bounds = simplifyIneqs (Just extCtx') luSys
    notEmpty = omegaTest (Just luSys) Nothing
    nV = nVars extCtx
    nP = nParams extCtx
    extCtx' = insertVars extCtx nV (zeroMatrix (nRows extCtx) 1)

    us = (\m -> insertColumns m nV (columnMatrix $ replicate (nRows m) (negate 1))) $
	 rowVectorsToMatrix (nV+nP+1) (kiUpper ki)
    ls = (\m -> insertColumns m nV (columnMatrix $ replicate (nRows m) 1)) $
	 negate $
	 rowVectorsToMatrix (nV+nP+1) (kiLower ki)


sysout :: [[(IneqSystem,GAffineFunction Integer)]]
       -> String
sysout input = s
    where
    (s, _) = qfFormula2qepcad f
    f  = disjunctionQf fs
    fs = map (ineqSystemToFormula . fmap constPoly) imgs
    imgs = [image (fmap fromIntegral af) is | xs <- input, (is,af) <- xs]
    






timeIt :: Show a => LGenOptions -> String -> a -> a
timeIt lgo msg x
    | lgoProgressToStderr lgo =
        unsafePerformIO $
        do hPutStrLn stderr ("begin: " ++ msg)
           t1   <- getClockTime
           evaluate (length $ show x)
           t2   <- getClockTime
           let totalSecs = fromIntegral (picoDiff t1 t2) / 10^12
           hPutStrLn stderr (msg ++ ": " ++ show totalSecs ++ " s")
           return x
    | otherwise = x
    where
    picoDiff :: ClockTime -> ClockTime -> Integer
    picoDiff (TOD s1 p1) (TOD s2 p2) = s*10^12 + p
        where
        s = s2 - s1
        p = p2 - p1

intersectDomainsReduce :: [[IneqSystem]] -> [IneqSystem]
intersectDomainsReduce doms =
    [moveVarsToParams (fmap unConst r) nV nP 0 | r <- res]
    where
    unConst p = let Just c = isConstPoly p in c
    -- Just res = formulaToIneqSystems (nV+nP) simplQf
    res = concat [s | d <- reduceDnfQf [] qf, let Just s = atFormulasToIneqSystems (nV+nP) d]
    simplQf = fastSimplifyQf [] qf
    qf = conjunctionQf [disjunctionQf [ineqSystemToFormula $ fmap (constPolyNI (Exactly 0)) i | i <- d] | d <- doms]

    nV = nVars   $ head $ concat doms
    nP = nParams $ head $ concat doms

simplifyDomainSlfq :: [IneqSystem] -> [IneqSystem] -> [IneqSystem]
simplifyDomainSlfq ectx sys =
    [moveVarsToParams (fmap unConst r) nV nP 0 | r <- res]
    where
    unConst p = let Just c = isConstPoly p in c
    Just res = formulaToIneqSystems (nV+nP) simplQf
    simplQf = slfqSimplifyQf [ctxQf] sysQf
    sysQf = disjunctionQf $ map i2f sys
    ctxQf = disjunctionQf $ map i2f ectx

    i2f ineq = ineqSystemToFormula $ fmap (constPolyNI (Exactly 0)) ineq

    nV = nVars   $ head (ectx ++ sys)
    nP = nParams $ head (ectx ++ sys)

step :: LGenOptions -> IneqSystem -> ([Klotz],[(Int,Int)]) -> ([Klotz],[(Int,Int)])
step lgo ctx (kloetze,initOrder) =
    (if lgoProgressToStderr lgo
        then trace ("project: " ++ show (length kloetze) ++ " -> " ++ show (length result))
        else id
    ) $
--    trace (show doms) $
    (result, order)
    where
    result = map project sets

    project :: (IneqSystem,[([Conj],KlotzInner)]) -> Klotz
    project (ineqs,innerParts) =
        Klotz { outerIneqs = simplifyIneqs (Just vctx) $
		             ineqSystem ((nV-1,nP),
                                         rowVectorsToMatrix (nV+nP) outer
                                        ),
                outerConds = [outerCs], -- [trueConj],
                innerPart  = ip
              }
        where
	vctx = insertVars ctx 0 (zeroMatrix (nRows ctx) (nV-1))
        ip = Loop { kiLower = map (staticIntDiv ceiling) lower,
                    kiUpper = map (staticIntDiv floor)   upper,
		    kiConds = [bodyCs],
                    kiInner = innerPartsNew
                  }
	innerPartsNew = zip condsNew2 kis
	(condsNew2,bodyCs) = commonConds (\v -> True) condsNew
	outerCs = map (\v -> deleteEntries v (nV-1) 1) outerCs'
	(condsNew,outerCs') = commonConds (\v -> vectorEntry v (nV-1) == 0) conds
	(conds,kis) = unzip innerParts

        outer = map shrink ([deleteEntries o curVar 1 | o <- outer'] ++
			    [u-l | l<-lower, u<-upper]
			   )
        lower = [fmap (/ negate (vectorEntry l curVar)) $
                 deleteEntries l curVar 1 | l <- ls]
        upper = [fmap (/ negate (vectorEntry u curVar)) $
                 deleteEntries u curVar 1 | u <- us]
        -- Just iSorted = sortIneqs ineqs Nothing
	iSorted = simplifyIneqs Nothing ineqs
        (bounds,outer') = partition (\r -> vectorEntry r curVar /= 0) $
                          rowVectors $ getMatrix iSorted
        (ls,us) = partition (\r -> vectorEntry r curVar > 0) bounds
        curVar = nV - 1
        nV = nVars ineqs
        nP = nParams ineqs

    sets :: [(IneqSystem,[([Conj],KlotzInner)])]
    order :: [(Int,Int)]
    (sets,order) = xsort lgo ctx kloetze initOrder

staticIntDiv :: (Rational -> Integer) -> Vector -> Vector
staticIntDiv roundFn v
    | all (==1) denoms  =
        vector (init ents ++ [fromIntegral $ roundFn $ last $ ents])
    | otherwise  =  v
    where
    ents = vectorEntries v
    denoms = init $ map denominator ents


shrink :: Vector -> Vector
shrink v
    | g /= 0     =  fmap (*(l % g)) v
    | otherwise  =  fmap (*(l % 1)) v
    where
    l = foldl lcm 1 $ fmap denominator $ vectorEntries v
    g = foldl gcd 0 $ filter (/=0) $ fmap numerator   $ vectorEntries v

commonConds :: (Vector -> Bool) -> [Disj] -> ([Disj],[Vector])
commonConds canMove vsss =
    (map (map remCommon) vsss, common)
    where
    remCommon ws = filter (\w -> w `notElem` common) ws
    common = nub $ filter (\v -> canMove v && all (v `elem`) vss) vs
    vss = concat vsss
    vs  = concat vss


xsort :: LGenOptions
      -> IneqSystem
      -> [Klotz]
      -> [(Int,Int)]
      -> ( [(IneqSystem,[([Conj],KlotzInner)])], [(Int,Int)] )
xsort lgo ctx ks initOrder =
    lgentrace ("xsort: initOrder=" ++ show initOrder ++ "; length ks = " ++ show (length ks)) $
    xsort' initOrder [] initialTodo initialKmap initialSysMap
    where
    lgentrace = if lgoDebug lgo then trace else flip const

    initialTodo = [(m,n) | m:ns <- tails [0..nKs-1], n <- ns]
    initialKmap   = Map.fromAscList [(n,Set.singleton n) | n <- [0..nKs-1]]
    initialSysMap = Map.fromAscList [(n,outerIneqs k) | k <- ks | n <- [0..]]
                   
    innerA = [(outerConds k, innerPart k) | k <- ks]
    nKs = length ks

    extCtx = insertVars ctx 0 (zeroMatrix (nRows ctx) (nVars $ head $ map outerIneqs ks))

    check :: [(Int,IneqSystem)] -> [(Int,Int)]
    check [] = []
    check ((n,is):rest) = concatMap cmpWith rest
        where
        cmpWith (n2,is2)
            | lexLess lgo ctx is is2 =
                if lexLess lgo ctx is2 is
                   then []
                   else [(n,n2)]
            | lexLess lgo ctx is2 is = [(n2,n)]
            | otherwise  =  error "check: not ordered?"

    -- kMap has _only_ vertex numbers that represent used polyhedra!
    xsort' :: [Edge]
           -> [Edge]
           -> [(Int,Int)]
           -> Map Vertex (Set Int)
           -> Map Vertex IneqSystem
           -> ( [(IneqSystem,[([Conj],KlotzInner)])], [(Int,Int)] )
    xsort' initOrder order [] kMap sysMap =
--        trace (show (order, scc graph)) $
        assert (not isCyclic) $
--        assert (null (c\\g)) $
--        trace ("xsort' 1: " ++ show (c,c\\g,g\\c)) $
--        trace ("xsort' 2: " ++ if isCyclic then "CYCLE" else "no cycle") $
        ([(is,inner) | idx <- ts, let Just is = Map.lookup idx sysMap,
                      let Just ks = Map.lookup idx kMap,
                      let inner = [innerA!!i | i <- Set.toAscList ks]
         ]
        ,[(a,b) | (x,y) <- order,
                  let Just a = findIndex (==x) ts,
                  let Just b = findIndex (==y) ts
         ]
        )
        where
        ts = filter (`elem` verts) $ Graph.topSort graph
        graph = buildG (0,maxVert) order
        verts = Map.keys kMap
        maxVert = maximum verts

        c = sort $ check (Map.toList sysMap)
        g = sort $ edges graph

        isCyclic = any (not . null . Tree.subForest) $ scc graph
             
    xsort' initOrder order ((a,b):todo) kMap sysMap
        | progress ("xsort': " ++ show (length todo+1) ++ " to go") $
          -- trace (show (a,b,Map.keys sysMap)) $
          ll isA isB a b =
            if ll isB isA b a
               then xsort' initOrder order   todo kMap sysMap
               else xsort' initOrder orderAB todo kMap sysMap
        | ll isB isA b a =
            xsort' initOrder orderBA todo kMap sysMap
        | otherwise =
--            trace (show (hasConjAB,length aWithoutB, length bWithoutA,conjAB) ++ "\n" ++ is2f isA ++ "\n" ++ is2f isB) $
            xsort' initOrder' order' todo' kMap' sysMap'
                              
        where
        progress :: String -> a -> a
        progress s val
            | lgoProgressToStderr lgo = trace s val
            | otherwise               = val

        is2f i = fst $ qfFormula2qepcad $ ineqSystemToFormula (fmap constPoly i)

        ll isAA isBB aa bb =
            case lgoAlgoVariant lgo of
               LGenSlightlyImproved     -> lex
               LGenOriginal        -> lex
               LGenExperimental
                   | lex && not o  ->  lgentrace ("ll: " ++ show (aa,bb))
                                       False
                   | otherwise     ->  lex && o
            where
            lex = lexLess lgo ctx isAA isBB
            o   = (bb,aa) `notElem` initOrder
        Just isA = Map.lookup a sysMap
        Just isB = Map.lookup b sysMap
        orderAB = (a,b) : order
        orderBA = (b,a) : order

        maxVert = maximum $ Map.keys kMap

        conjAB    = simplifyIneqs (Just extCtx) (isA `join` isB)
        hasConjAB = hasIntegralPoint (Just ctx) conjAB

        aWithoutB = filter (hasIntegralPoint (Just ctx)) $
                    disjointUnionDifference (Just extCtx) [isA] [isB]
        bWithoutA = filter (hasIntegralPoint (Just ctx)) $
                    disjointUnionDifference (Just extCtx) [isB] [isA]

        newVertices = [maxVert+1 ..]
        aWoBverts = take (length aWithoutB) newVertices
        bWoAverts = take (length bWithoutA) $ drop (length aWithoutB) newVertices
        conjABvert = head $ drop (length aWithoutB + length bWithoutA) newVertices

        orderWoA  = [(x,y) | (x,y) <- order, x /= a, y /= a]
        orderWoB  = [(x,y) | (x,y) <- order, x /= b, y /= b]
        orderWoAB = [(x,y) | (x,y) <- orderWoA, x /= b, y /= b]
        order' = map head $ group $ sort $
                 orderWoAB

        todoWoA     = [(x,y) | (x,y) <- todo,  x /= a, y /= a]
        todoWoB     = [(x,y) | (x,y) <- todo,  x /= b, y /= b]
        todoWoAB    = [(x,y) | (x,y) <- todoWoA, x /= b, y /= b]
        other = Map.keys kMap \\ [a,b]
        ab = (if hasConjAB then (conjABvert:) else id) (aWoBverts ++ bWoAverts)
        todo' = map head $ group $ sort todoX
        todoX = concat
                [todoWoAB,
                 [(x,y) | (x:rest) <- tails ab, y <- rest++other]
                ]

        initOrderWoA  = [(x,y) | (x,y) <- initOrder, x /= a, y /= a]
        initOrderWoB  = [(x,y) | (x,y) <- initOrder, x /= b, y /= b]
        initOrderWoAB = [(x,y) | (x,y) <- initOrderWoA, x /= b, y /= b]
        initOrder' = concat
                     [[(x,y) | (x,y) <- initOrder, x /= a, x /= b, y /= a, y /= b],
                      [(n,y) | (x,y) <- initOrderWoB, x == a, n <- aWoBverts],
                      [(x,n) | (x,y) <- initOrderWoB, y == a, n <- aWoBverts],
                      [(n,y) | (x,y) <- initOrderWoA, x == b, n <- bWoAverts],
                      [(x,n) | (x,y) <- initOrderWoA, y == b, n <- bWoAverts],
                      [(conjABvert,y) | hasConjAB, (x,y) <- initOrderWoB, x == a],
                      [(x,conjABvert) | hasConjAB, (x,y) <- initOrderWoB, y == a],
                      [(conjABvert,y) | hasConjAB, (x,y) <- initOrderWoA, x == b],
                      [(x,conjABvert) | hasConjAB, (x,y) <- initOrderWoA, y == b]
                ]

        sysMap' = Map.fromAscList (zip newVertices (aWithoutB++bWithoutA++(if hasConjAB then [conjAB] else [])))
                  `Map.union` (Map.delete b $ Map.delete a sysMap)
   
        Just aSet = Map.lookup a kMap
        Just bSet = Map.lookup b kMap
        kMap' = Map.fromAscList ([(n,aSet) | n <- aWoBverts] ++
                                 [(n,bSet) | n <- bWoAverts] ++
                                 if hasConjAB
                                    then [(conjABvert,aSet `Set.union` bSet)]
                                    else []
                                )
                `Map.union` (Map.delete b $ Map.delete a kMap)



--
-- Make sure the relative ordering among the "Klotz" is retained!
--
makeDomains :: IneqSystem -> [Klotz] -> [([IneqSystem],[Klotz])]
makeDomains ctx kloetze =
    (\r -> trace ("makeDomains: " ++ show (length kloetze) ++ " -> ") $
           trace ("makeDomains: -> " ++ show (length r, length [i | (is,_) <- r, i <- is])) r) $
    md kloetze univ [] []
    where
    nV = nVars $ head [outerIneqs k | k <- kloetze]
    univ = ineqSystem ((nV, nParams ctx), zeroMatrix 0 (nV+nParams ctx+1))
    extCtx = insertVars ctx 0 (zeroMatrix (nRows ctx) nV)

    md :: [Klotz] -> IneqSystem -> [IneqSystem] -> [Klotz] -> [([IneqSystem],[Klotz])]
    md []     _    _     []   =  []
    md []     isIn isOut ksIn =
	-- The call to 'reverse' retains the relative ordering of
	-- the "Klotz".
	if null dus then [] else [(dus, reverse ksIn)]
	where
	dus = filter (hasIntegralPoint (Just ctx)) $
	      disjointUnionDifference (Just extCtx) [isIn] isOut
    md (k:ks) isIn isOut ksIn =
--	trace (show ("md: feasible=" ++ show feasible)) $
	(if not feasible then [] else md ks is1 isOut ksIn1) ++
	md ks isIn (outerIneqs k : isOut) ksIn
	where
	feasible = hasIntegralPoint (Just ctx) is1
	is1   = simplifyIneqs (Just extCtx) (outerIneqs k `join` isIn)
	ksIn1 = k : ksIn

hasIntegralPoint :: Maybe IneqSystem -> IneqSystem -> Bool
hasIntegralPoint mbCtx sys =
--    (\b -> trace ("hasIntegralPoint " ++ show mbCtx ++ " " ++ show sys ++ " --> " ++ show b) b) $
    omegaTest (Just is) (Just es)
    where
    nV = nVars sys
    nP = nParams sys
    (is,es) = findEqualities sys'
    sys' = case mbCtx of
	      Nothing  -> sys
	      Just ctx -> sys `join`
			  insertVars ctx 0 (zeroMatrix (nRows ctx) nV)


-- i1 is less than i2 in the last variable
lexLess :: LGenOptions -> IneqSystem -> IneqSystem -> IneqSystem -> Bool
lexLess lgo ctx i1 i2 =
--    (\res -> trace (show res) res) $
{-    trace (show((geQE,geSimplex),(disjQE,disjSimplex))) $
    if (geQE/=geSimplex || disjQE/=disjSimplex)
       then error $ show [show simplexRes, show ctx, show i1, show i2]
       else -}
        case lgoAlgoVariant lgo of
          LGenOriginal          -> gt
          LGenExperimental      -> ge
          LGenSlightlyImproved  -> ge && (disj || (ns1 && ns2))
    where
    ns1 = decideFormula [] nonSing1
    ns2 = decideFormula [] nonSing2

    gt = case lgoLpVariant lgo of
           LGenQE      -> gtQE
           LGenSimplex -> gtSimplex
    gtSimplex =
               case simplexRes of
                  Unbounded     -> False
                  Nonexistent   -> True
                  Optimum (x,_) -> x < 0
    gtQE = impliesQf [c,f1,f2] (lessCond AtGt)
    ge = case lgoLpVariant lgo of
           LGenQE      -> geQE
           LGenSimplex -> geSimplex
    geSimplex = case simplexRes of
                  Unbounded     -> False
                  Nonexistent   -> True
                  Optimum (x,_) -> x <= 0
    geQE = impliesQf [c,f1,f2] (lessCond AtGe)
    disj =
        case lgoLpVariant lgo of
          LGenQE      -> disjQE
          LGenSimplex -> disjSimplex

    disjQE = not $ decideFormula [] $
             FoEx [0..nV+nP] $ FoQf $ conjunctionQf [c,f1a,f2a]
    disjSimplex = not $
        case makeDTree numOrdDTreeContext $
             isFeasibleSys is (Just es) of
           DLeaf leaf -> leaf
           _          -> error "lexLess: simplexResDisj: ???"
        where
        (is,es) = findEqualities $ moveParamsToVars x 0 nP nV
        x = multijoin1 [insertZeroVars ctx 0 nV, i1, i2]

    simplexRes = case makeDTree numOrdDTreeContext $
                      simplex stdSimplexParams is (Just es) tf of
           DLeaf leaf -> leaf
           _          -> error "lexLess: simplexRes: ???"
         where
         (is,es) = findEqualities $ moveParamsToVars x 0 nP (nV+1)
         tf = vector (replicate (fromIntegral (nV-1)) 0 ++ [1,-1] ++
                      replicate (fromIntegral nP) 0
                     )
         x = multijoin1 [insertZeroVars ctx 0 (nV+1),
                         ins nV i1, ins (nV-1) i2]

    -- question = conjunctionQf [c,f1,f2] `qfImpl` lessCond
    lessCond atrel = QfAt $ atrel (indeterminate (nV+nP+1) nV - indeterminate (nV+nP+1) (nV-1))
    c  = toFormula $ insertVars ctx 0 (zeroMatrix (nRows ctx) (nV+1))

    f1 = toFormula $ ins nV     i1
    f2 = toFormula $ ins (nV-1) i2
    ins pos sys = insertZeroVars sys pos 1

    f1a = toFormula $ ins nV i1
    f1b = toFormula $ ins (nV-1) i1
    f2a = toFormula $ ins nV i2
    f2b = toFormula $ ins (nV-1) i2

    nonSing1 = FoEx [0..nV+nP] $ FoQf $
               conjunctionQf [c, f1a, f1b, QfAt (AtNeq (indeterminate (nV+nP+1) nV - indeterminate (nV+nP+1) (nV-1)))]
    nonSing2 = FoEx [0..nV+nP] $ FoQf $
               conjunctionQf [c, f2a, f2b, QfAt (AtNeq (indeterminate (nV+nP+1) nV - indeterminate (nV+nP+1) (nV-1)))]

    toFormula = ineqSystemToFormula . fmap (constPolyNI $ Exactly 0)
    nV = nVars i1
    nP = nParams i1



data LGenAlgoVariant = LGenSlightlyImproved
                     | LGenOriginal
                     | LGenExperimental
                       deriving (Eq, Show)

data LGenLpVariant = LGenQE
                   | LGenSimplex
                     deriving (Eq, Show)

data LGenOptions = LGO { lgoAlgoVariant :: LGenAlgoVariant,
                         lgoLpVariant   :: LGenLpVariant,
                         lgoSimplifyIFs        :: Bool,
                         lgoSimplifyOuterLoops :: Bool,
                         lgoConvexUnion        :: Bool,
                         lgoProgressToStderr   :: Bool,
                         lgoDebug              :: Bool
                       }
                   deriving Show

lgenOptionsStd = LGO { lgoAlgoVariant = LGenSlightlyImproved,
                       lgoLpVariant   = LGenQE,
                       lgoProgressToStderr  = False,
                       lgoDebug             = True,
                       lgoSimplifyIFs       = False,
                       lgoSimplifyOuterLoops = False,
                       lgoConvexUnion       = False
                     }

data LGenParams = LGP { lgpContext :: IneqSystem,
			lgpStatements :: [[(IneqSystem,GAffineFunction Integer)]]
		      }
		  deriving Show

data LGenOutputOptions = LGOO { lgooParamNames :: [String]
			      , lgooVarNames   :: [String]
                              , lgooIterNames  :: [String]
                              , lgooStmtGen    :: Maybe (Integer -> [[Rational]] -> String)
                              , lgooLoopGen :: Integer -> String -> [[Vector]] -> [[Vector]] -> String -> String -> [Integer] -> Maybe String
                              }

instance Show LGenOutputOptions where
    showsPrec prec lgoo =
        showString "LGOO {lgooParamNames=" .
        shows (lgooParamNames lgoo) .
        showString ", lgooVarNames=" .
        shows (lgooVarNames lgoo) .
        showString ", lgooIterNames=" .
        shows (lgooIterNames lgoo) .
        showString ", lgooStmtGen=_, lgooLoopGen=_}"

emptyLGenOutputOptions = LGOO { lgooParamNames = []
                              , lgooVarNames   = []
                              , lgooIterNames  = []
                              , lgooStmtGen    = Nothing
                              , lgooLoopGen    = \_ _ _ _ _ _ _ -> Nothing
                              }

loopGenIF :: Integer -> String -> [[Vector]] -> [[Vector]] -> String -> String -> String
loopGenIF n var ls us lb ub =
        case (ls,us) of
          ([[]],[[]]) -> "if (1)"
          (_,[[]]) -> concat ["if (", lb, "<=", var, ")"]
          ([[]],_) -> concat ["if (", var, "<=", ub, ")"]
          (_,_)  -> concat ["if (", lb, "<=", var, " && ", var, "<=", ub, ")"]

loopGenSelfNP :: String -> String -> Integer -> String -> [[Vector]] -> [[Vector]] -> String -> String -> String
loopGenSelfNP self np n var ls us lb ub =
    case (ls,us) of
      ([[]],[[]]) -> "if (1)"
      (_,_) ->
            concat ["for (", var, "=", lb, "+", self, "; ", var, "<=", ub, "; ", var, "+=", np, ")"]



parseLGenInput :: Parser (LGenParams, LGenOutputOptions)
parseLGenInput =
    do spacesOrComment
       oneOf "cf"
       spacesOrComment
       ctx <- parseContext
       paramNames <- parseNames [[c] | c <- ['M'..]] (nParams ctx)
       doms <- parseDomains (nParams ctx)
       let stdIterNames = [[c] | c <- ['i'..]]
       iterNames <- parseNames stdIterNames $ maximum $ map nVars $ concat doms
       scatter' <- parseScatterFns (nParams ctx)
       when (not (null scatter') && length scatter' /= length doms) $
            fail "Number of scattering functions does not match number of statements"
       let scatter = if null scatter' then makeIdentityScatterFns doms
                                      else scatter'
       let tD = targetDimIH $ head scatter
       let scatterNames = ['c':show i | i <- [1..tD]]
       varNames <- if null scatter'
                   then return scatterNames
                   else parseNames scatterNames tD
       spacesOrComment
       eof
       -- when no scattering functions are given, we use the names
       -- of the iterators for the scatter dimensions; this makes the
       -- output more CLooG-like.
       let (vNames,iNames) = if null scatter'
                             then (iterNames, varNames )
                             else (varNames,  iterNames)
       return (LGP { lgpContext    = ctx,
		     lgpStatements = [[(is,s) | is <- dom] | dom <- doms | s <- scatter]
		   }
              ,emptyLGenOutputOptions { lgooParamNames = paramNames
		                      , lgooVarNames   = vNames
                                      , lgooIterNames  = iNames
                                      }
              )

makeIdentityScatterFns :: [[IneqSystem]] -> [GAffineFunction Integer]
makeIdentityScatterFns doms =
    map mkS doms
    where
    tDim = maximum $ map nVars $ concat doms

    mkS (sys:_) = affineFunctionVPC (unitMatrix  tDim nV
                                    , zeroMatrix tDim nP
                                    , zeroMatrix tDim 1
                                    )
                                    Inhomogeneous
        where
        nV = nVars sys
        nP = nParams sys


parseNames :: [String] -> Int -> Parser [String]
parseNames stdnames n =
    do spacesOrComment
       flag <- oneOf "01"
       if flag == '1'
	  then count n name
	  else return $ take n stdnames

ignoreNames :: Parser ()
ignoreNames =
    do spacesOrComment
       flag <- oneOf "01"
       if flag == '1'
	  then many name >>
	       return ()
	  else return ()

name :: Parser String
name = spacesOrComment >>
       many1 (letter <|> alphaNum <|> char '_')

parseContext :: Parser IneqSystem
parseContext =
    do spacesOrComment
--       char '1'
       (bs,m) <- parsePMatrix
       return $ pm2ineqSystem 0 (nColumns m-1) (bs,m)

parseDomains :: Integer -> Parser [[IneqSystem]]
parseDomains nP =
    do spacesOrComment
       n <- liftM readI $ many1 digit
       count n (parseDomain nP)

parseDomain :: Integer -> Parser [IneqSystem]
parseDomain nP =
    do spacesOrComment
       n <- liftM readI $ many1 digit
       res <- count n $ do (bs,m) <- parsePMatrix
			   return $ pm2ineqSystem (nColumns m - nP - 1) nP (bs,m)
       count 3 (spacesOrComment >> char '0')
       return $ {- minDisjUnion res -} {- disjUnionAll' -} res

disjUnionAll' xs
   | or [oT z z2 | z:zs <- tails ys, z2 <- zs] =
       trace ("disjUnionAll failed for " ++ show xs ++ "Result: " ++ show ys) ys
   | otherwise = ys
   where
   ys = disjUnionAll xs

   oT z z2 = omegaTest (Just ineqs) (Just eqs)
        where
        (ineqs,eqs) = findEqualities (z `join` z2)

{-
pairwiseDisjoint :: [IneqSystem] -> Bool
pairwiseDisjoint [] = True
pairwiseDisjoint [_] = True
pairwiseDisjoint (i:is) =
    all areIneqsInfeasible [i `join` j | j <- is] && pairwiseDisjoint is

disjUnionAll2 :: [IneqSystem] -> [IneqSystem]
disjUnionAll2 ineqs =
    [moveVarsToParams i nV nP 0 | i <- ineqs'']
    where
    ineqs'' = map (fmap (fromJust . isConstFP)) $
              disjointUnion [] $ map (fmap fromRational) ineqs'
    ineqs' = [moveParamsToVars i 0 nP nV | i <- ineqs]
    nV = nVars   $ head ineqs
    nP = nParams $ head ineqs
-}

{-
minDisjUnion :: [IneqSystem] -> [IneqSystem]
minDisjUnion [ineq] = [ineq]
minDisjUnion ineqs =
    trace ("min: " ++ show (length du1))
    mdu du1 ps
    where
    p:ps = permutsA ineqs
    du1 = disjUnionAll p

    mdu du [] = du
    mdu du (x:xs) = mdu du'' xs
        where
        du' = disjUnionAll x
        du'' = if length du' < length du
                  then trace ("min: " ++ show (length du')) du'
                  else du

permutsA :: [a] -> [[a]]
permutsA xs = unord (length ps) ps
    where
    ps = permuts xs

    unord _ [] = []
    unord l xs = x : unord (l-1) (xs1++xs2)
        where
        idx = 768745 `mod` l
        (xs1,x:xs2) = splitAt idx xs

permuts :: [a] -> [[a]]
permuts [] = [[]]
permuts xs = concat [map (y:) $ permuts (ys1++ys2) | i <- [0..length xs-1], let (ys1,y:ys2) = splitAt i xs]
-}

parseScatterFns :: Integer -> Parser [GAffineFunction Integer]
parseScatterFns nP =
    do spacesOrComment
       n <- liftM readI $ many1 digit
       count n (parseScatterFn nP)

parseScatterFn :: Integer -> Parser (GAffineFunction Integer)
parseScatterFn nP =
    do spacesOrComment
       m <- parsePMatrix
       return $ pm2scatterFn nP m

parsePMatrix :: Parser ([Bool],Matrix)
parsePMatrix =
    do spacesOrComment
       r <- liftM readI $ many1 digit
       spacesOrComment
       c <- liftM readI $ many1 digit
       ents <- count r (count c (spacesOrComment >> number))
       return (map ((/=0).head) ents, matrix (r,c-1,concatMap tail ents))

pm2scatterFn :: Integer -> ([Bool],Matrix) -> GAffineFunction Integer
pm2scatterFn nP (bs,m) =
    assert (all (==False) bs) $
    assert (u == unitMatrix nVtarget nVtarget) $
    assert (all (==1) $ map denominator $ entries $ getMatrix fn) $
    affineFunction ((nV,nP), fmap (negate . numerator) fn) Inhomogeneous
    where
    nVtarget = nRows m
    nV       = nColumns m - nVtarget - nP - 1
    u        = subMatrixVer m2 0 (nVtarget-1)
    fn       = subMatrixVer m2 nVtarget (nColumns m-1)
    (m2,_,_) = gaussJordanTransform m
    

pm2ineqSystem :: Integer -> Integer -> ([Bool],Matrix) -> IneqSystem
pm2ineqSystem nV nP (bs,m) =
    ineqSystem ((nV,nP), rowVectorsToMatrix (nV+nP+1) $
		         concat [f b r | b <- bs | r <- rowVectors m]
	       )
    where
    f False r = [r,-r]
    f True  r = [r]

readI :: String -> Int
readI = read

number :: Parser Rational
number = (char '-' >> liftM (fromIntegral . negate . readI) (many1 digit))
	 <|> liftM (fromIntegral . readI) (many1 digit)

spacesOrComment :: Parser ()
spacesOrComment =
    do many $ ((many1 space >> return ()) <|>
	       (char '#' >> manyTill anyToken ((newline >> return ())
					       <|> eof
					      ) >>
		return ()
	       )
	      )
       return ()

writeLGenInput :: LGenParams -> LGenOutputOptions -> String
writeLGenInput lgp lgoo =
    "c\n" ++ writeIneqs (lgpContext lgp) ++ "1\n" ++
    concat (intersperse " " (lgooParamNames lgoo)) ++ "\n" ++
    show (length $ lgpStatements lgp) ++ "\n" ++
    concatMap writeStatement (lgpStatements lgp) ++ "0\n" ++
    show (length $ lgpStatements lgp) ++ "\n" ++
    concatMap writeScatterFn (lgpStatements lgp) ++ "1\n" ++
    concat (intersperse " " (lgooVarNames lgoo)) ++ "\n"

writeIneqs :: IneqSystem -> String
writeIneqs sys = writeMatrix 1 $ getMatrix $ integralIneqSystem sys

writeMatrix :: Integer -> GMatrix Integer -> String
writeMatrix first m =
    show (nRows m) ++ " " ++ show (nColumns m + 1) ++ "\n" ++
    unlines [show first ++ concat [' ':show e | e <- r] | r <- rows' m]

writeStatement :: [(IneqSystem, GAffineFunction Integer)] -> String
writeStatement [] = error "writeLGenInput: empty union"
writeStatement stmts
    | length (nub $ map getMatrixIH fs) == 1  =  writeDomain ineqs
    | otherwise  =
        error "writeLGenInput: unions of domains with different scattering functions not supported"
    where
    (ineqs,fs) = unzip stmts

writeDomain :: [IneqSystem] -> String
writeDomain ineqs =
    show (length ineqs) ++ "\n" ++
    concatMap writeIneqs ineqs ++
    "0 0 0\n"

writeScatterFn :: [(IneqSystem, GAffineFunction Integer)] -> String
writeScatterFn stmts = writeFn f
    where
    f = head $ snd $ unzip stmts

writeFn :: GAffineFunction Integer -> String
writeFn af = writeMatrix 0 (negate (unitMatrix tD tD) `conc` getMatrixIH af)
    where
    tD = targetDimIH af


------------------------------------
-- not so useful things
------------------------------------

{-
checkedLGen :: IneqSystem -> [[(IneqSystem,GAffineFunction Integer)]] -> [Integer] -> [([([Integer], Integer)], [([Integer], Integer)], Bool)]
checkedLGen ctx input paramVals =
    [(i,o,i==o) | (i,o) <- res]
    where
    output = lgen lgo ctx input
    lgo = lgenOptionsStd { lgoAlgoVariant=LGenExperimental }
    res = map (checkLGen paramVals input) output

-- checkLGen :: [Integer] -> [[(IneqSystem,GAffineFunction Integer)]] -> [LCode] -> Bool
checkLGen paramVals input lcode = (orig, stmtSort $ mk hull ks)
    where
    stmtSort :: [([Integer],Integer)] -> [([Integer],Integer)]
    stmtSort l = concat $
	         map (sortBy (\(_,s1) (_,s2) -> compare s1 s2)) $
		 groupBy (\(x1,s1) (x2,s2) -> x1==x2) l

    mk :: [IneqSystem] -> [Klotz] -> [([Integer],Integer)]
    mk outerISs ks =
        [q | x <- points, k <- ks, and $ outerIneqs k `apply` (fmap fromIntegral (x++paramVals++[1]) :: [Rational]), checkConds x (outerConds k), q <- genKI x (innerPart k)]
        where
	points = map (\(g:_) -> g) $ group $ sort $ concatMap (enumerate paramVals (affineFunction ((nVars $ head outerISs,nParams $ head outerISs),unitMatrix (nVars $ head outerISs) (nColumns $ getMatrix $ head outerISs)) Inhomogeneous)) outerISs

    genKI :: [Integer] -> KlotzInner -> [([Integer],Integer)]
    genKI x ki@(Loop {}) =
        [q | y <- [ceiling $ maximum (map (subst x) $ kiLower ki) .. floor $ minimum (map (subst x) $ kiUpper ki)], let p = x++[y], checkConds p (kiConds ki), (cs,ki2) <- kiInner ki, checkConds p cs, q <- genKI p ki2]
    genKI surr (Statement {kiStmtID=id}) = [(surr,id)]

    checkConds :: [Integer] -> [Conj] -> Bool
    checkConds x cs =
        any (all (\cv -> denominator (v `mulVV` cv) == 1)) cs
        where
        v = vector $ map fromIntegral (x++paramVals++[1])

    orig = sort [(p, stmtID)
                 | (stmtID,inp) <- zip [1..] input, (sys,fn) <- inp, p <- enumerate paramVals fn sys]

    subst x ys = ys `mulVV` vector (map fromIntegral (x ++ paramVals ++ [1]))
-}

{-
enumerate :: [Integer] -> GAffineFunction Integer -> IneqSystem -> [[Integer]]
enumerate paramVals fn sys =
--    (\e -> trace (show (fn,sys,nV,paramVals)) e) $
    en 0 [[]]
       where
       en n xs
           | n == nV    =  -- trace (show (n,last xs)) $
			   map (\x -> vectorEntries $ applyIH fn $ vector (x++paramVals)) xs
           | otherwise  =
               en (n+1) [x++[y] | x <- xs, y <- [ceiling $ maximum (map (subst x) $ lower n) .. floor $ minimum (map (subst x) $ upper n)]]

       subst x ys = ys `mulVV` vector (map fromIntegral (x ++ paramVals ++ [1]))

       Just sorted = sortIneqs sys Nothing
       lower n = rowVectors $
                 fmap negate $ getMatrix $
                 deleteVars (filterIneqs (\i -> part n i && i!!n>0) sorted) n (nV-n)
       upper n = rowVectors $ getMatrix $
                  deleteVars (filterIneqs (\i -> part n i && i!!n<0) sorted) n (nV-n)
       part n i = all (==0) $ take (nV-n-1) $ drop (n+1) i
       nV = nVars sys
-}

{-
postOpt :: IneqSystem -> LCode -> LCode
postOpt ctx lcode =
    lcode { lcodeKloetze = map (postOptK ctx) $ lcodeKloetze lcode }
-}





{-
lgen (ineqSystem ((0,0),zeroMatrix 0 1)) [[(ineqSystem ((2,0),matrix [[1,0,0],[0,1,0],[-2,-1,10]]),affineFunction ((2,0),matrix [[2,0,0],[0,1,0]]) Inhomogeneous)]]

lgen (ineqSystem ((0,0),zeroMatrix 0 1)) [[(ineqSystem ((2,0),matrix [[1,0,0],[0,1,0],[-2,-1,10]]),affineFunction ((2,0),matrix [[2,0,0],[0,1,0]]) Inhomogeneous)],[(ineqSystem ((2,0),matrix [[1,0,0],[0,1,0],[-2,-1,10]]),affineFunction ((2,0),matrix [[2,0,0],[0,1,0]]) Inhomogeneous)]]

lgen (ineqSystem ((0,0),zeroMatrix 0 1)) [[(ineqSystem ((2,0),matrix [[1,0,0],[0,1,0],[-2,-1,10]]),affineFunction ((2,0),matrix [[2,0,0],[0,1,0]]) Inhomogeneous)],[(ineqSystem ((2,0),matrix [[1,0,0],[0,1,0],[-2,-1,8]]),affineFunction ((2,0),matrix [[2,0,1],[0,1,0]]) Inhomogeneous)]]

lgen (ineqSystem ((0,1),zeroMatrix 0 2)) [[(ineqSystem ((2,1),matrix [[1,0,0,0],[0,1,0,0],[-2,-1,1,0]]),affineFunction ((2,1),matrix [[2,0,0,0],[0,1,0,0]]) Inhomogeneous)],[(ineqSystem ((2,1),matrix [[1,0,0,0],[0,1,0,0],[-2,-1,1,-2]]),affineFunction ((2,1),matrix [[2,0,0,1],[0,1,0,0]]) Inhomogeneous)]]

lgen (ineqSystem ((0,0),zeroMatrix 0 1)) [[(ineqSystem ((1,0),matrix [[1,0],[-1,10]]),affineFunction ((1,0),matrix [[1,0],[1,0]]) Inhomogeneous)]]


Right l <- parseFromFile parseInput "/home/clstaff/groessli/michael-cholesky1/cholesky1.src.out.cloogInput"
let c = runLGen l
c!!6
map (\(_,_,b) -> b) $ checkedLGen (lgpContext l) (lgpStatements l) [17]



xsort True (ineqSystem ((0,0),zeroMatrix 0 1)) [Klotz{outerIneqs=ineqSystem((1,0),matrix [[1,0],[-1,10]]),outerConds=[],innerPart=Statement{kiStmtID=1,kiPreImage=affineFunction((0,0),zeroMatrix 0 1) Inhomogeneous}}, Klotz{outerIneqs=ineqSystem((1,0),matrix [[1,-5],[-1,15]]),outerConds=[],innerPart=Statement{kiStmtID=2,kiPreImage=affineFunction((0,0),zeroMatrix 0 1) Inhomogeneous}}]

xsort True (ineqSystem ((0,2),matrix [[1,0,0],[0,1,0]])) [Klotz{outerIneqs=ineqSystem((1,2),matrix [[1,0,0,0],[-1,1,0,0]]),outerConds=[],innerPart=Statement{kiStmtID=1,kiPreImage=affineFunction((0,0),zeroMatrix 0 1) Inhomogeneous}}, Klotz{outerIneqs=ineqSystem((1,2),matrix [[1,0,-1,0],[-1,0,0,15]]),outerConds=[],innerPart=Statement{kiStmtID=2,kiPreImage=affineFunction((0,0),zeroMatrix 0 1) Inhomogeneous}}]

xsort True (ineqSystem ((0,0),zeroMatrix 0 1)) [Klotz{outerIneqs=ineqSystem((2,0),matrix [[1,0,0],[-1,0,10],[0,1,0],[0,-1,10]]),outerConds=[],innerPart=Statement{kiStmtID=1,kiPreImage=affineFunction((0,0),zeroMatrix 0 1) Inhomogeneous}}, Klotz{outerIneqs=ineqSystem((2,0),matrix [[1,0,-5],[-1,0,15],[0,1,-5],[0,-1,15]]),outerConds=[],innerPart=Statement{kiStmtID=2,kiPreImage=affineFunction((0,0),zeroMatrix 0 1) Inhomogeneous}}]

lgen (ineqSystem ((0,0),zeroMatrix 0 1)) [[(ineqSystem((3,0),matrix [[1,0,0,0],[-1,0,0,10],[0,0,1,0],[0,0,-1,10],[0,1,0,0],[0,-1,0,0]]),affineFunction((3,0),unitMatrix 3 3 `conc` zeroMatrix 3 1) Inhomogeneous)],[(ineqSystem((3,0),matrix [[1,0,0,-5],[-1,0,0,15],[0,0,1,-5],[0,0,-1,15],[0,1,0,0],[0,-1,0,0]]),affineFunction((3,0),unitMatrix 3 3 `conc` zeroMatrix 3 1) Inhomogeneous)]]
-}

test2 = lgen lgenOptionsStd
             (ineqSystem ((0,1), zeroMatrix 0 2))
             [[(ineqSystem ((1,1), matrix [[1,0,0], [-1,1,0]])
               ,affineFunction ((1,1), matrix [[2, 0, 0]]) Inhomogeneous)
              ]
             ]

noninjTest = lgen lgenOptionsStd
                  (ineqSystem ((0,1), zeroMatrix 0 2))
                  [[(ineqSystem ((2,1), matrix [[1,0,0,0], [-1,0,1,0], [0,1,0,0],[0,-1,0,10]])
                    ,affineFunction ((2,1), matrix [[1, 1, 0, 0]]) Inhomogeneous)
                   ]
                  ]
