{-
    This file is part of the LooPo project, see
    <http://www.infosun.fim.uni-passau.de/cl/loopo/>.

    Copyright (C) 2009 Armin Groesslinger
    Author: Armin Groesslinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-}

module LMath.Algo.NLGen
    ( LCode(..)
    , LGenParams(..), LGenOptions(..), LGenOutputOptions(..)
    , LGenAlgoVariant(..), LGenLpVariant(..)
    , LGenFunctions(..), ScatterFn
    , lgenOptionsStd, emptyLGenOutputOptions
    , lgen
    , parseLGenInput, writeLGenInput
    , runLGenWithStatements
    , loopGenIF, loopGenSelfNP
    , lgfRealFrac
    , realFracLGOO, fracPolyLGOO
    )
where

import LMath.Types
import LMath.Algo.AffineImage
import LMath.Algo.Chernikova
import LMath.Algo.DisjointUnion
import LMath.Algo.OmegaTest
import LMath.Algo.Simplex
import LMath.Algo.Sort

import Control.Exception ( assert, evaluate )
import Control.Monad     ( liftM, when )
import Data.Graph as Graph
import Data.List
import Data.Map ( Map )
import qualified Data.Map as Map
import Data.Maybe
import Data.Ratio
import Data.Set ( Set)
import qualified Data.Set as Set
import Data.Tree as Tree
import Text.ParserCombinators.Parsec

import Debug.Trace
import System.IO
import System.IO.Unsafe
import System.Time


----------------
-- data types --
----------------

-- [Conj] represents a disjunction
type Conj a = [GVector a]
type Disj a = [Conj a]

trueConj :: Conj a
trueConj = []    -- empty conjunction

data LCode a = LCode { lcodeBounds     :: [([[GVector a]],[[GVector a]])],
                       lcodeParamConds :: [GIneqSystem a],
                       lcodeLNests    :: [LNest a]
                     }
             deriving Show

data LNest a = LNest { outerIneqs :: GIneqSystem a,
                       outerConds :: [Conj a],
                       innerPart  :: LNestInner a
                     }
	     deriving Show

data LNestInner a = Loop { kiLower :: [GVector a]   -- lower bounds
                         , kiUpper :: [GVector a]   -- upper bounds
		         , kiConds :: [Conj a]     -- disjunction of conds
                         , kiInner :: [([Conj a], LNestInner a)]
                         }
                  | NoLoop { kiInner :: [([Conj a], LNestInner a)]
                           }
                  | Statement { kiStmtID :: Integer
                              , kiPreImage :: GAffineFunction a
                            }
		  deriving Show

statements :: LNestInner a -> [Integer]
statements (Statement {kiStmtID=sid}) = [sid]
statements (Loop {kiInner=inners})    = nub $ concatMap (statements.snd) inners
statements (NoLoop {kiInner=inners})  = nub $ concatMap (statements.snd) inners


--------------------
-- main functions --
--------------------

iterateM :: (Integral b, Monad m) => b -> (a -> m a) -> a -> m a
iterateM 0 f x = return x
iterateM n f x =
    do y <- f x
       iterateM (n-1) f y


lgen :: (Fractional a, Num cond)
     => LGenOptions
     -> LGenFunctions a cond
     -> Integer
     -> GIneqSystem a
     -> [([GIneqSystem a], ScatterFn a)]
     -> DTreeGen a cond (LCode a)
lgen lgo (lgf :: LGenFunctions a cond) level ctx input =
    assert (nVars ctx == 0) $
    assert (all ((==nP).nParams) allISs) $
    assert (all ((==nP).nParams) allSCs) $
    assert (length (nub targetDims) <= 1) $
    assert (level >= 0) $
    assert (level <= head targetDims) $
    do lnests1 <- mapM mkLNests [(id, iss, sc)
                                  | (id,(iss,sc)) <- zip [1..] input
                                  ]
       let lnests = concat lnests1
       res <- iterateM level (step lgo lgf ctx) lnests
       lcodes <- mkOuterLoops lgo lgf ctx res
       return lcodes
    where
    targetDims = [nVars sc - nVars is | (iss,sc) <- input, is <- iss]
    --
    -- Compute the initial parts.
    -- Make the images of the iteratior domains under the scattering
    -- "functions" disjoint. We need to keep track of (at least one)
    -- preimage domain such that we can correctly enumerate the
    -- preimage if the scattering is non-injective (computing the
    -- inverse image of the image under the scattering enumerates
    -- a superset, e.g., 0<=x,y<=10, (x,y) -> x+y yields an image of
    -- 0<=x+y<=20 and the preimage of this is unbounded:
    -- x\in\Z, -x<=y<=20-x).
    --
    mkLNests :: (Integer, [GIneqSystem a], ScatterFn a)
              -> DTreeGen a cond [LNest a]
    mkLNests (id, dom, sc) =
        do outerISs <- mapM (\is -> scatterImage lgf is sc) dom
           dtc <- getDTreeContext
           let disjISs' =
                   genericDisjointUnion iddFn
                            [(img,[GDUTag i preimg])
                            | (Just img,preimg,i) <- zip3 outerISs dom [0..]
                            ]
               nonempty = lgfNonEmpty lgf dtc Nothing
               iddFn a b
                   | nonempty inters = Just ([inters], aWithoutB, bWithoutA)
                   | otherwise       = Nothing
                   where
                     inters = a `join` b
                     aWithoutB = lgfDisjointUnionDifference lgf Nothing [a] [b]
                     bWithoutA = lgfDisjointUnionDifference lgf Nothing [b] [a]

               disjISs = [(img,preimg) | (img,GDUTag _ preimg:_) <- disjISs']
           mapM (\(img,preimg) -> mkLNest id img preimg sc) disjISs

    mkLNest :: Integer
            -> GIneqSystem a  -- projection onto target side (disj part)
            -> GIneqSystem a  -- preimage
            -> ScatterFn a
            -> DTreeGen a cond (LNest a)
    mkLNest id outerIs sys sc =
        do fmInv <- fmPreImage lgf sys sc
           let inv = mkInv nP tD nI fmInv
           inner <- -- trace ("fmInv = " ++ show fmInv) $
                    mkInner lgf nP tD fmInv $
                    Statement { kiStmtID   = id
			      , kiPreImage = inv
                              }
           return $ LNest { outerIneqs = outerIs
	                  , outerConds = [map (\v -> deleteEntries v tD nI) $
                                          filter (isRelevant tD nI) $
			                  rowVectors $ getMatrixIH inv]
	                  , innerPart  = inner
	                  }
        where
        nI = nVars sys

    -- with lgfUnitDenom lgf we check whether the denominators
    -- are units.

    isRelevant :: Int -> Int -> GVector a -> Bool
    isRelevant tD nI r =
        all (==0) (take nI (drop tD ents)) &&  -- not a condition on iterators
        any (not . lgfUnitDenom lgf) ents
        where
          ents = vectorEntries r

    (allISss,allSCs) = unzip input
    allISs           = concat allISss
    someIS           = head allISs
    someSC           = head allSCs

    tD = nVars someSC - nVars someIS  -- number of target dimensions

    nP = nParams ctx



--
-- [params, target dims, <iters>, 1] -> [target dims, <iters>, params, 1]
-- (number of iters varies)
--
mkInner :: (Fractional a, Num cond)
        => LGenFunctions a cond
        -> Int 
        -> Int
        -> SortedSystem a
        -> LNestInner a
        -> DTreeGen a cond (LNestInner a)
mkInner lgf nP tD (ss :: SortedSystem a) stmt = mk 0 ss
    where
      mk :: Int -> SortedSystem a -> DTreeGen a cond (LNestInner a)
      mk d (LowerUpper ls us : rest) =
          do inner <- mk (d+1) rest
             return $ Loop { kiLower = map fixLower ls
                           , kiUpper = map fixUpper us
                           , kiConds = [trueConj]
                           , kiInner = [ ([trueConj], inner) ]
                           }
      mk d (Equation e : rest) =
          do inner <- mk (d+1) rest
             return $ NoLoop { kiInner = [ ([cond], inner) ] }
          where
            cond = if isRel then [fixOrdering e] else trueConj
            -- isRel: condition relevant, i.e., in the iterators
            isRel = any (not . isUnitDenom) ents &&
                    any (/=0) (init $ drop (nP+tD) ents)
            ents = vectorEntries e
      mk d [] = return stmt

      fixLower = tighten' ceiling . fixOrdering
      fixUpper = tighten' floor   . fixOrdering
          
      fixOrdering v =
          -- vector (init r2 ++ targets ++ params ++ [last r2])
          vector (targets ++ init r2 ++ params ++ [last r2])
          where
            (params,  r1) = splitAt nP ents
            (targets, r2) = splitAt tD r1
            ents = vectorEntries v

      isConst = lgfIsConst lgf  -- is a value a rational?
      tighten' = tighten isConst

      isUnitDenom = lgfUnitDenom lgf


--
-- [parameters, target dims, <iters>, 1] -> [target dims, iterator dims, parameters, 1]
mkInv :: Fractional a
      => Int -> Int -> Int -> SortedSystem a -> GAffineFunction a
mkInv nP tD nI ss = affineFunction ((tD+nI,nP), 
                                    matrix $ zipWith mkDim [0..] ss
                                   ) Inhomogeneous
    where
      mkDim d (LowerUpper ls us) =
          replicate (tD+d) 0 ++ 1 : replicate (nI-d-1+nP+1) 0
      mkDim d (Equation e) = 
          tPart ++ take d vPart2 ++ replicate (nI-d) 0 ++ pPart ++ [last vPart2]
          where
            cs = vectorEntries e
            (pPart, vPart1) = splitAt nP cs
            (tPart, vPart2) = splitAt tD vPart1

--
-- Compute the image under the scattering.
-- Return Nothing if the iteration domain is empty.
--
scatterImage :: (Fractional a, Num cond)
             => LGenFunctions a cond
             -> GIneqSystem a
             -> ScatterFn a
             -> DTreeGen a cond (Maybe (GIneqSystem a))
scatterImage lgf sys sc =
    (lgfProjectToVarsG lgf) fullsys Nothing Nothing [0..tD-1]
    where
      fullsys = insertZeroVars sys 0 tD `join` sc
      tD = nVars sc - nVars sys
      nP = nParams sc

--
-- returns an enumeration of the iterator dimensions (i.e.,
-- the preimage of 'sys') in dependence of the parameters and
-- the target dimensions of 'sc'. (parameters in front)
--
-- Take care to change self-referencing equations, such that
-- they reference the outer dimensions directly
-- instead of refering to another equation dimension which
-- then refers to the outer dimensions. See  fixSelfRefs.
--
fmPreImage :: (Fractional a, Num cond)
           => LGenFunctions a cond
           -> GIneqSystem a
           -> ScatterFn a
           -> DTreeGen a cond (SortedSystem a)
fmPreImage lgf sys sc =
    do mbSS <- (lgfSortIneqsGss lgf) ins Nothing (Just 0)
       return $ fixSelfRefs $ fixMbSS mbSS
    where
      fixMbSS (Just ss) = drop (nP+tD) ss
      fixMbSS Nothing   = error "fmPreImage: Nothing not handled :-("

      ins = moveParamsToVars sc 0 nP 0
            `join`
            moveParamsToVars (insertZeroVars sys 0 tD) 0 nP 0

      tD = nVars sc - nV  -- vars on the target side
      nV = nVars sys      -- vars on the preimage side (iterators)
      nP = nParams sys


--
-- [Equation Vector Poly 0 Poly 0 Poly 0 Poly 1.0 Poly 0
-- ,Equation Vector Poly 0 Poly 0 Poly 0 Poly 0 Poly 1.0 Poly 0
-- ]
--  -->
-- [Equation Vector Poly 0 Poly 0 Poly 0 Poly 1.0 Poly 0
-- ,Equation Vector Poly 0 Poly 0 Poly 0 Poly 1.0 Poly 0 Poly 0
-- ]
--
fixSelfRefs :: Num a => SortedSystem a -> SortedSystem a
fixSelfRefs ss =
--  (\r -> trace (show (ss,r)) r) $
--    map fixsr ss
    fixsr ss
    where
    fixsr [] = []
    fixsr (Equation e : rest) =
        Equation e : fixsr [fmap (fix e) bnds | bnds <- rest]
    fixsr (lu : rest) = lu : fixsr rest

    fix e v =
--        trace (show (e,v,v0,e')) $
        (vectorEntry v varIdx `mulSV` e')  +  v0
        where
        varIdx = vectorLength e - 1  -- index of the dimension described by e
        v0     = setVectorEntry v varIdx 0
        e'     = insertEntries e varIdx $
                 replicate (vectorLength v - vectorLength e) 0

--
-- generate the outer loops
--
mkOuterLoops :: (Fractional a, Num cond)
             => LGenOptions
             -> LGenFunctions a cond
             -> GIneqSystem a     -- ^ context in linear parameters only
             -> [LNest a]
             -> DTreeGen a cond (LCode a)
mkOuterLoops lgo lgf ctx (ks :: [LNest a]) =
--    trace ("mkOuterLoops " ++ show ks) $
    do 
       ksOpt <- if lgoSimplifyBounds lgo
                then do kss <- mapM (postOptK lgf ctx) ks
                        return $ concat kss
                else return ks
       let hull1 = map outerIneqs ksOpt
       hull <- return hull1
               -- does not seem to improve anything:
               -- (lgfSimplifyDomainG lgf) (Just extctx) hull1
       
       -- Inefficiency: we compute the projection onto the parameter
       -- space twice, in lgfSortIneqsGss und lgfProjectToVarsG
       boundsMBs <- mapM (\sys -> lgfSortIneqsGss lgf sys Nothing Nothing) hull
       let bnds1 :: [[([GVector a],[GVector a])]]
           bnds1 = [toBounds ss | Just ss <- boundsMBs]
       let bnds2 = f bnds1

       bnds3 <- return bnds2
                -- simplify bnds2 ?

       paramCondss <- mapM (\sys -> lgfProjectToVarsG lgf sys Nothing (Just 0) [0..nP-1]) $
                      [moveParamsToVars h 0 nP 0 | h <- hull]
       let paramConds' = [pc | Just pc <- paramCondss]
       paramConds <- (lgfSimplifyDomainG lgf) (Just ctx') paramConds'
              
       return $ LCode { lcodeBounds = bnds3
                      , lcodeParamConds = [moveVarsToParams c 0 nP 0
                                           | c <- paramConds]
                      , lcodeLNests = ksOpt
                      }
    where
    nP   = nParams $ outerIneqs $ head ks
    nV   = nVars $ outerIneqs $ head ks

    extctx = insertZeroVars ctx 0 nV
    ctx' = moveParamsToVars ctx 0 nP 0

    f :: [ [([GVector a],[GVector a])] ] -> [ ([[GVector a]],[[GVector a]]) ]
    f []     = []
    f ([]:_) = []
    f xs = unzip (map head xs) : f (map tail xs)
        

    toBounds :: SortedSystem a -> [([GVector a],[GVector a])]
    toBounds []                        = []
    toBounds (Equation e       : rest) = ([e],[e]) : toBounds rest
    toBounds (LowerUpper ls us : rest) = (ls, us)  : toBounds rest




step :: (Fractional a, Num cond)
     => LGenOptions
     -> LGenFunctions a cond
     -> GIneqSystem a
     -> [LNest a]
     -> DTreeGen a cond [LNest a]
step lgo (lgf :: LGenFunctions a cond) ctx (lnests :: [LNest a]) =
    (if lgoProgressToStderr lgo
        then trace ("project: " ++ show (length lnests))
        else id
    ) $
--    trace (show doms) $
    do s <- sets
       projRes <- mapM project s
       return [p | Just p <- projRes]
    where
    project :: (GIneqSystem a,[([Conj a],LNestInner a)])
            -> DTreeGen a cond (Maybe (LNest a))
    project (ineqs,innerParts) =
        do mbSorted <- (lgfSimplifyIneqsG lgf)  Nothing ineqs
           case mbSorted of
             Just iSorted -> proj iSorted >>= \ln -> return $ Just ln
             Nothing      -> return Nothing
        where
        proj iSorted = do
           bnds <- (lgfSortIneqsGss lgf) iSorted Nothing (Just (nV-1))
           let (lower,upper) = -- trace ("bnds = " ++ show bnds) $
                               case bnds of
                                 Just [Equation e]       -> ([e],[e])
                                 Just [LowerUpper ls us] ->
                                     (map (tighten isConst ceiling) ls
                                     ,map (tighten isConst floor)   us
                                     )
                                 Nothing      -> -- error "project: Nothing"
                                                 ([1], [0]) -- ?
                                 Just (xs@(_:_:_)) ->
                                     error ("project: " ++ show xs)
           Just outer <- (lgfProjectToVarsG lgf) iSorted Nothing (Just 0) [0..nV-2]
           let ip = Loop { kiLower = lower
                         , kiUpper = upper
		         , kiConds = [bodyCs]
                         , kiInner = innerPartsNew
                         }
           
           return $ LNest { outerIneqs = outer
                          , outerConds = [outerCs] -- [trueConj]
                          , innerPart  = ip
                          }

	innerPartsNew = zip condsNew2 kis
	(condsNew2,bodyCs) = commonConds (\v -> True) condsNew
	outerCs = map (\v -> deleteEntries v (nV-1) 1) outerCs'
	(condsNew,outerCs') = commonConds (\v -> vectorEntry v (nV-1) == 0) conds
	(conds,kis) = unzip innerParts

        curVar = nV - 1
        nV = nVars ineqs
        nP = nParams ineqs

    sets :: DTreeGen a cond [(GIneqSystem a,[([Conj a],LNestInner a)])]
    sets = xsort lgo lgf ctx lnests

    isConst = lgfIsConst lgf  -- is a coefficient a rational number?

tighten :: Num a
        => (a -> Maybe Rational)
        -> (Rational -> Integer)
        -> GVector a
        -> GVector a
tighten isConst tightenFn v =
    case mbEnts of
      Just (coeffs,constCo) ->
          vector $ fmap fromIntegral (coeffs ++ [tightenFn constCo])
      _ -> v
    where
    mbEnts :: Maybe ([Integer], Rational)
    mbEnts =
        do ents <- mapM isConst $ vectorEntries v
           let co:cs = reverse ents
           if (all ((==1).denominator) cs)
              then Just (reverse $ map numerator cs, co)
              else Nothing


commonConds :: Fractional a
            => (GVector a -> Bool) -> [Disj a] -> ([Disj a],[GVector a])
commonConds canMove vsss =
    (map (map remCommon) vsss, common)
    where
    remCommon ws = filter (\w -> w `notElem` common) ws
    common = nub $ filter (\v -> canMove v && all (v `elem`) vss) vs
    vss = concat vsss
    vs  = concat vss


xsort :: (Fractional a, Num cond)
      => LGenOptions
      -> LGenFunctions a cond
      -> GIneqSystem a
      -> [LNest a]
      -> DTreeGen a cond ( [(GIneqSystem a, [([Conj a],LNestInner a)])]
                            )
xsort lgo (lgf :: LGenFunctions a cond) ctx (ks :: [LNest a]) =
    lgentrace ("xsort: length ks = " ++ show (length ks)) $
    xsort' [] initialTodo initialKmap initialSysMap
    where
    lgentrace = if lgoDebug lgo then trace else flip const

    initialTodo = [(m,n) | m:ns <- tails [0..nKs-1], n <- ns]
    initialKmap   = Map.fromAscList [(n,Set.singleton n) | n <- [0..nKs-1]]
    initialSysMap = Map.fromAscList [(n,outerIneqs k) | k <- ks | n <- [0..]]
                   
    innerA = [(outerConds k, innerPart k) | k <- ks]
    nKs = length ks

    extCtx = insertVars ctx 0 (zeroMatrix (nRows ctx) (nVars $ head $ map outerIneqs ks))

    orderIsAcyclic :: [Edge] -> Bool
    orderIsAcyclic order =
        all (null . Tree.subForest) $ scc graph
        where
        graph = buildG (0,maxVertex) order
        maxVertex = maximum (-1 : verts)
        verts = let (as,bs) = unzip order in as++bs

    -- kMap has _only_ vertex numbers that represent used polyhedra!
    xsort' :: [Edge]
           -> [Edge]
           -> Map Vertex (Set Int)
           -> Map Vertex (GIneqSystem a)
           -> DTreeGen a cond ( [(GIneqSystem a,[([Conj a],LNestInner a)])]
                                 )
    xsort' order [] kMap sysMap =
--        trace (show (order,dff graph)) $
        assert (not isCyclic) $
        return $
         [(is,inner) | idx <- ts, let Just is = Map.lookup idx sysMap,
                      let Just ks = Map.lookup idx kMap,
                      let inner = [innerA!!i | i <- Set.toAscList ks]
         ]

        where
        ts = filter (`elem` verts) $ Graph.topSort graph
        graph = buildG (0,maxVert) order
        verts = Map.keys kMap
        maxVert = maximum (-1 : verts)

        g = sort $ edges graph
        isCyclic = not $ orderIsAcyclic order
             
    xsort' order ((a,b):todo) kMap sysMap =
--        trace ("order = " ++ show order) $
        assert (orderIsAcyclic order) $
        progress ("xsort': " ++ show (length todo+1) ++ " to go") $
        do mbOrd <- ordering lgo lgf ctx isA isB
           let (ab,ba) = case mbOrd of
                           Nothing -> (False,False)
                           Just LT -> (True, False)
                           Just EQ -> (True, True)
                           Just GT -> (False, True)
           if ab
             then
              do -- ba <- ll isB isA b a
                 if ba then
                    xsort' order   todo kMap sysMap
                  else
                    xsort' orderAB todo kMap sysMap
             else
              do -- ba <- ll isB isA b a
                 if ba then
                     xsort' orderBA todo kMap sysMap
                  else
                     withDTreeContext separate
                              
        where
        progress :: String -> z -> z
        progress s val
            | lgoProgressToStderr lgo = trace s val
            | otherwise               = val

        Just isA = Map.lookup a sysMap
        Just isB = Map.lookup b sysMap
        orderAB = (a,b) : order
        orderBA = (b,a) : order

        maxVert = maximum $ Map.keys kMap

        simplIneqsG = lgfSimplifyIneqsG lgf
        hasIntegralPointG = lgfNonEmpty lgf
        disjointUnionDifferenceG = lgfDisjointUnionDifference lgf
 
        separate dtc =
         do mbConjAB <- simplIneqsG (Just extCtx) (isA `join` isB)
            let (hasConjAB,conjAB) =
                    case mbConjAB of
                      Just conj -> (hasIntegralPointG dtc (Just ctx) conj
                                   ,conj
                                   )
                      Nothing   -> (False, error "should not evaluate conjAB")
                aWithoutB = filter (hasIntegralPointG dtc (Just ctx)) $
                            disjointUnionDifferenceG (Just extCtx) [isA] [isB]
                bWithoutA = filter (hasIntegralPointG dtc (Just ctx)) $
                            disjointUnionDifferenceG (Just extCtx) [isB] [isA]

                newVertices = [maxVert+1 ..]
                aWoBverts = take (length aWithoutB) newVertices
                bWoAverts = take (length bWithoutA) $ drop (length aWithoutB) newVertices
                conjABvert = head $ drop (length aWithoutB + length bWithoutA) newVertices

                orderWoA  = [(x,y) | (x,y) <- order, x /= a, y /= a]
                orderWoB  = [(x,y) | (x,y) <- order, x /= b, y /= b]
                orderWoAB = [(x,y) | (x,y) <- orderWoA, x /= b, y /= b]
                order' = map head $ group $ sort $
                         orderWoAB

                todoWoA     = [(x,y) | (x,y) <- todo,  x /= a, y /= a]
                todoWoB     = [(x,y) | (x,y) <- todo,  x /= b, y /= b]
                todoWoAB    = [(x,y) | (x,y) <- todoWoA, x /= b, y /= b]
                other = Map.keys kMap \\ [a,b]
                ab = (if hasConjAB then (conjABvert:) else id) (aWoBverts ++ bWoAverts)
                todo' = map head $ group $ sort todoX
                todoX = concat
                        [todoWoAB,
                         [(x,y) | (x:rest) <- tails ab, y <- rest++other]
                        ]

                sysMap' = Map.fromAscList (zip newVertices (aWithoutB++bWithoutA++(if hasConjAB then [conjAB] else [])))
                          `Map.union` (Map.delete b $ Map.delete a sysMap)
   
                Just aSet = Map.lookup a kMap
                Just bSet = Map.lookup b kMap
                kMap' = Map.fromAscList ([(n,aSet) | n <- aWoBverts] ++
                                         [(n,bSet) | n <- bWoAverts] ++
                                         if hasConjAB
                                         then [(conjABvert,aSet `Set.union` bSet)]
                                         else []
                                        )
                        `Map.union` (Map.delete b $ Map.delete a kMap)
            -- in
            xsort' order' todo' kMap' sysMap'

ordering :: (Fractional a, Num cond)
         => LGenOptions
         -> LGenFunctions a cond
         -> GIneqSystem a
         -> GIneqSystem a
         -> GIneqSystem a
         -> DTreeGen a cond (Maybe Ordering)
ordering lgo lgf ctx isA isB =
    do dtc <- getDTreeContext
       let aLess = lessCond lgo lgf dtc ctx isA isB
           bLess = lessCond lgo lgf dtc ctx isB isA
           implAB = dtcImpliesQf dtc [aLess] bLess
           implBA = dtcImpliesQf dtc [bLess] aLess
{- incorrect
   we must not impose an ordering if the polyhedra are not
   comparable (and both orderings are fine) -> return $ Just EQ
       case (implAB, implBA) of
         (_, True) -> fCond aLess (dLeaf $ Just LT) (dLeaf Nothing)
         (True, _) -> fCond bLess (dLeaf $ Just GT) (dLeaf Nothing)
         (False, False) ->
-}
       fCond' aLess 
              (fCond' bLess (dLeaf $ Just EQ) (dLeaf $ Just LT))
              (fCond' bLess (dLeaf $ Just GT) (dLeaf Nothing))

lessCond :: (Fractional a, Num cond)
        => LGenOptions
        -> LGenFunctions a cond
        -> DTreeContext a cond
        -> GIneqSystem a
        -> GIneqSystem a
        -> GIneqSystem a
        -> QfFormula cond
lessCond lgo (lgf :: LGenFunctions a cond) dtc ctx i1 i2 =
--    trace ("lessCond: " ++ show (i1,i2,lt)) $
    case lgoAlgoVariant lgo of
          LGenOriginal          -> lt
--          LGenSlightlyImproved  -> ge && (disj || (ns1 && ns2))
    where
--    ns1 = decideFormula [] nonSing1
--    ns2 = decideFormula [] nonSing2

    -- is i2 greater than i1 in the last variable?
    lt :: QfFormula cond
    lt = case lgoLpVariant lgo of
           LGenQE      -> ltQE
           LGenSimplex -> ltSimplex
    ltQE :: QfFormula cond
    ltQE = fmap (dtcCondGen dtc) $
           (lgfImplies lgf) dtc (multijoin1 [c,f1,f2]) False leCondition

    ltSimplex :: QfFormula cond
    ltSimplex =
        dtreeToFormula $
        makeDTree dtc $
        do r <- simplexRes
           case r of
                  Unbounded     -> f
                  Nonexistent   -> t
                  Optimum (x,_) -> geCond x f t
        where
        t = dLeaf (QfAt AtTrue)
        f = dLeaf (QfAt AtFalse)

    simplexRes :: DTreeGen a cond (Optimum (a,[a]))
    simplexRes = -- simplex stdSimplexParams
                 chernikovaMaximiseG
                        is (Just es) tf
         where
         (is,es) = findEqualities $ moveParamsToVars x 0 nP (nV+1)
         x = multijoin1 [c, f1, f2]

    tf = vector (replicate (fromIntegral (nV-1)) 0 ++ [1,-1] ++
                 replicate (fromIntegral nP) 0
                )

    leCondition = fmap negate geCondition
    geCondition = 
        ineqSystem ((nV+1,nP), matrix [vectorEntries tf ++ [0]])

    c  = insertZeroVars ctx 0 (nV+1)
    f1 = ins nV     i1
    f2 = ins (nV-1) i2
    ins pos sys = insertZeroVars sys pos 1

    nV = nVars i1
    nP = nParams i1



{-
-- i1 is less than i2 in the last variable
lexLess :: (Fractional a, Num cond)
        => LGenOptions
        -> LGenFunctions a cond
        -> GIneqSystem a
        -> GIneqSystem a
        -> GIneqSystem a
        -> DTreeGen a cond Bool
lexLess lgo (lgf :: LGenFunctions a cond) ctx i1 i2 =
    trace ("lexLess: i1="++show i1 ++", i2="++show i2) $
    case lgoAlgoVariant lgo of
          LGenOriginal          -> lt
--          LGenSlightlyImproved  -> ge && (disj || (ns1 && ns2))
    where
--    ns1 = decideFormula [] nonSing1
--    ns2 = decideFormula [] nonSing2

    -- is i2 greater than i1 in the last variable?
    lt :: DTreeGen a cond Bool
    lt = case lgoLpVariant lgo of
           LGenQE      -> ltQE
           LGenSimplex -> ltSimplex
    ltQE :: DTreeGen a cond Bool
    ltQE = withDTreeContext $ \dtc ->
           fCond ((lgfImplies lgf) dtc (multijoin1 [c,f1,f2]) False leCondition) (dLeaf True) (dLeaf False)

    ltSimplex :: DTreeGen a cond Bool
    ltSimplex =
        do r <- simplexRes
           case r of
                  Unbounded     -> return False
                  Nonexistent   -> return True
                  Optimum (x,_) -> geCond x (dLeaf False) (dLeaf True)

    le = case lgoLpVariant lgo of
           LGenQE      -> leQE
           LGenSimplex -> leSimplex
    leQE = withDTreeContext $ \dtc ->
           fCond ((lgfImplies lgf) dtc (multijoin1 [c,f1,f2]) True leCondition) (dLeaf True) (dLeaf False)
    leSimplex =
        do r <- simplexRes
           case r of
                  Unbounded     -> return False
                  Nonexistent   -> return True
                  Optimum (x,_) -> geCond (negate x) (dLeaf True) (dLeaf False)

{-
    disj =
        case lgoLpVariant lgo of
--          LGenQE      -> disjQE
          LGenSimplex -> disjSimplex

    disjSimplex =
           do f <- isFeasibleSys is (Just es)
              return $ not f
        where
        (is,es) = findEqualities $ moveParamsToVars x 0 nP nV
        x = multijoin1 [insertZeroVars ctx 0 nV, i1, i2]
-}

    simplexRes :: DTreeGen a cond (Optimum (a,[a]))
    simplexRes = simplex stdSimplexParams is (Just es) tf
         where
         (is,es) = findEqualities $ moveParamsToVars x 0 nP (nV+1)
         tf = vector (replicate (fromIntegral (nV-1)) 0 ++ [1,-1] ++
                      replicate (fromIntegral nP) 0
                     )
         x = multijoin1 [c, f1, f2]

    leCondition = fmap negate geCondition
    geCondition = 
        ineqSystem ((nV+1,nP), matrix [replicate (nV-1) 0 ++ [1,-1] ++
                                       replicate (nP+1) 0
                                      ]
                   )

    c  = insertZeroVars ctx 0 (nV+1)
    f1 = ins nV     i1
    f2 = ins (nV-1) i2
    ins pos sys = insertZeroVars sys pos 1

    nV = nVars i1
    nP = nParams i1
-}


runLGenWithStatements :: (Fractional a, Num cond)
                      => LGenOptions
                      -> LGenFunctions a cond
                      -> LGenParams a
                      -> LGenOutputOptions a
                      -> DTreeContext a cond
                      -> (cond -> String)
                      -> Integer
                      -> String
runLGenWithStatements lgo lgf lgp lgoo (dtc :: DTreeContext a cond) showCond level =
    assert (nParams (lgpContext lgp) == genericLength (lgooParamNames lgoo)) $

    showTree showCond lg
    where
    lg :: DTree cond String
    lg = makeDTree dtc $
         do lc <- lgen lgo lgf level (lgpContext lgp) (lgpStatements lgp)
            return $ codeWithStatements lgoo lc

showTree :: (cond -> String) -> DTree cond String -> String
showTree s (DLeaf code) = code
showTree s (SCond cond tm t0 tp) =
    "if (" ++ s cond ++ " > 0) {\n" ++
    indent 2 (showTree s tp) ++
    "} else if (" ++ s cond ++ " < 0) {\n" ++
    indent 2 (showTree s tm) ++
    "} else {\n" ++
    indent 2 (showTree s t0) ++
    "}\n"
showTree s (EqCond cond t0 tpm) =
    "if (" ++ s cond ++ " == 0) {\n" ++
    indent 2 (showTree s t0) ++
    "} else {\n" ++
    indent 2 (showTree s tpm) ++
    "}\n"
showTree s (GeCond _ cond t0p tm) =
    "if (" ++ s cond ++ " >= 0) {\n" ++
    indent 2 (showTree s t0p) ++
    "} else {\n" ++
    indent 2 (showTree s tm) ++
    "}\n"
showTree s (FCond qf tt tf) =
    "if (" ++ qfFormulaToC s qf ++ ") {\n" ++
    indent 2 (showTree s tt) ++
    "} else {" ++
    indent 2 (showTree s tf) ++
    "}\n"



lgenStdStmtGen :: LGenOutputOptions a -> Integer -> [GVector a] -> String
lgenStdStmtGen lgoo i (iterExps :: [GVector a]) =
    'S' : show i ++ "(" ++ concat (intersperse "," $ map gen iterExps) ++ ");"
    where
    gen :: GVector a -> String
    gen rs = case l of
               Right 1 -> s
               _       -> "(" ++ s ++ ")/(" ++ lstr ++ ")"
	where
	(s,l) = (lgooAffineExpToString lgoo) (viNames++paramNames) nP rs
        lstr = case l of
                 Left  str -> str
                 Right i   -> show i

    viNames = genericTake nVI (lgooVarNames lgoo ++ lgooIterNames lgoo)
    paramNames = lgooParamNames lgoo
    nP = genericLength paramNames
    nVI = vectorLength (head iterExps) - nP - 1



codeWithStatements :: Fractional a => LGenOutputOptions a -> LCode a -> String
codeWithStatements lgoo (LCode {lcodeBounds=bounds,lcodeLNests=ks,lcodeParamConds=paramConds}) =
--    trace (show hull) $
    if null paramConds2 then
        loopNest
    else
        "if ((" ++ 
        concat (intersperse ") || (" $ [ineqs2c lgoo pc | pc <- paramConds2])
        ++ ")) {\n"
        ++ indent 2 loopNest
        ++ "}\n"
    where
    loopNest =
        unlines
        [replicate (2*n) ' ' ++
         genFor lgoo nP (fromIntegral n) var (lower n) (upper n) stmtIDs ++
         " {" | n <- [0..nV-1] | var <- varNames
        ] ++
        indent (2*nV) (concatMap (lnest2code lgoo nP (fromIntegral nV)) ks) ++
        unlines [replicate (2*n) ' ' ++ "}" | n <- [nV-1,nV-2..0]]
    varNames   = lgooVarNames lgoo ++ lgooIterNames lgoo
    paramNames = lgooParamNames lgoo
    lower n = fst (bounds!!n)
    upper n = snd (bounds!!n)

    stmtIDs = nub $ concatMap (statements . innerPart) ks

    nV = length bounds
    -- nP = (vectorLength $ head $ concat $ fst $ head bounds) - 1
    nP = nParams $ head (paramConds ++ [outerIneqs (head ks)])

    -- drop universal contexts (1 >= 0)
    paramConds2 = filter ((/=0). nRows) paramConds

indent :: Int -> String -> String
indent n str =
    unlines [replicate n ' ' ++ l | l <- lines str]

lnest2code :: Fractional a 
           => LGenOutputOptions a -> Integer -> Integer -> LNest a -> String
lnest2code lgoo nP n k
    | condition == ""  =  innerCode
    | otherwise        =
        concat ["if (", condition, ") {\n"
               ,indent 2 innerCode, "}\n"
               ]
    where
    innerCode = lnestInner2code lgoo nP n (innerPart k)
    condition
        | null outerCondsStr  =  outerIneqsStr
        | null outerIneqsStr  =  outerCondsStr
        | otherwise =
            "(" ++ outerCondsStr ++ ") && (" ++ outerIneqsStr ++ ")"

    outerCondsStr
        | [] `elem` outerConds k     =  ""
        | otherwise                  =  conds2c lgoo nP (outerConds k)
    outerIneqsStr
        | nRows (outerIneqs k) == 0  =  ""
        | otherwise                  =  ineqs2c lgoo $ outerIneqs k
    

lnestInner2code :: Fractional a
                => LGenOutputOptions a 
                -> Integer  -- ^ number of parameters
                -> Integer  -- ^ index of this loop
                -> LNestInner a
                -> String
lnestInner2code lgoo nP n ki@(Loop {kiLower=ls,kiUpper=us,kiConds=bcs,kiInner=inners}) =
    genFor lgoo nP n var [ls] [us] (statements ki) ++ " {\n" ++
    indent 2 (bodyCond $ lnestInner2code lgoo nP n $ NoLoop {kiInner=inners})
    ++ "}\n"
    where
    bodyCond :: String -> String
    bodyCond body
        | [] `elem` bcs = body
        | otherwise     =
              concat ["if (", conds2c lgoo nP bcs, ") {\n"
	             ,indent 2 body, "}\n"
                     ]

    varNames   = lgooVarNames lgoo ++ lgooIterNames lgoo
    paramNames = lgooParamNames lgoo
    var = varNames !! fromIntegral (nV-1)
    nV = n + 1

lnestInner2code (lgoo :: LGenOutputOptions a) nP n (NoLoop {kiInner=inners}) =
    concat $ map genInner inners
    where
    genInner :: ([Conj a],LNestInner a) -> String
    genInner ([[]],ki) = lnestInner2code lgoo nP (n+1) ki
    genInner (cs,ki) =
	"if (" ++ conds2c lgoo nP cs ++ ") {\n" ++
	indent 2 (lnestInner2code lgoo nP (n+1) ki) ++
	"}\n"
    
lnestInner2code lgoo nP n (Statement {kiStmtID=sid,kiPreImage=pi}) =
    stmt ++ "\n"
    where
    stmt = fromMaybe (lgenStdStmtGen lgoo sid inv)
                     (lgooStmtGen lgoo sid inv)
    inv = rowVectors $ getMatrixIH pi


genFor :: Fractional a
       => LGenOutputOptions a 
       -> Integer 
       -> Integer 
       -> String 
       -> [[GVector a]] 
       -> [[GVector a]] 
       -> [Integer] 
       -> String
genFor (lgoo :: LGenOutputOptions a) nP n var ls us stmtIDs =
--    trace ("genFor " ++ show (ls,us)) $
    case (lgooLoopGen lgoo) n var ls us lb ub stmtIDs of
       Just loophead -> loophead
       Nothing       -> concat ["for (", var, "=", lb, "; ", var, "<=", ub, "; ", var, "++)"]
    where
    varNames   = genericTake n (lgooVarNames lgoo ++ lgooIterNames lgoo)
    paramNames = lgooParamNames lgoo

    lb = bounds "min" "max" "ceild" ceiling ls
    ub = bounds "max" "min" "floord" floor us

    bounds :: String -> String -> String -> (Rational -> Integer) -> [[GVector a]] -> String
    bounds outerMinmax minmax divname rndfn [] = "INFEASIBLE"
    bounds outerMinmax minmax divname rndfn [b] = bound minmax divname rndfn b
    bounds outerMinmax minmax divname rndfn (b:bs) =
	outerMinmax ++ "(" ++ bound minmax divname rndfn b ++ "," ++
	bounds outerMinmax minmax divname rndfn bs ++ ")"

    bound :: String -> String -> (Rational -> Integer) -> [GVector a] -> String
    bound minmax divname rndfn []  = "UNBOUNDED"
    bound minmax divname rndfn [b] = gen divname rndfn b
    bound minmax divname rndfn (b:bs) =
        minmax ++ "(" ++ gen divname rndfn b ++ "," ++
        bound minmax divname rndfn bs ++ ")"

    gen divname rndfn v
--        | all (==0) $ init ents  = show $ rndfn $ last ents
	| l == Right 1 =  s
	| otherwise    =  divname ++ '(':s ++ ',':lstr ++ ")"
	where
	(s,l) = (lgooAffineExpToString lgoo) (varNames ++ paramNames) nP v
        lstr = case l of
                 Left  str -> str
                 Right i   -> show i
        ents  = vectorEntries v


ineqs2c :: Fractional a => LGenOutputOptions a -> GIneqSystem a -> String
ineqs2c lgoo (sys :: GIneqSystem a) =
--    trace ("sys = " ++ show sys) $
    (lgooIneqSystemToString lgoo) names sys
    where
      names = take (nVars sys) (lgooVarNames lgoo ++ lgooIterNames lgoo)
              ++
              lgooParamNames lgoo


conds2c :: LGenOutputOptions a -> Integer -> [Conj a] -> String
conds2c _ nP []   = "0==1"
conds2c _ nP [[]] = "1==1"
conds2c lgoo nP (cs :: [Conj a]) =
    (if (length conds > 1) then "(" else "")
    ++ concat (intersperse ") || (" conds) ++
    (if (length conds > 1) then ")" else "")
    where
    conds = [concat $ intersperse " && " [gen v | v <- c] | c <- cs]

    gen :: GVector a -> String
    gen rs = "(" ++ numExpr ++ ")%" ++ "(" ++ denomStr ++ ") == 0"
             -- l == 1 does not happen for module conditions
	where
          (numExpr,denom) = (lgooAffineExpToString lgoo) names nP rs
          denomStr = case denom of
                       Left  str -> str
                       Right i   -> show i
          names = genericTake (vectorLength rs - nP - 1) 
                              (lgooVarNames lgoo ++ lgooIterNames lgoo) ++
                  lgooParamNames lgoo


showP :: Integer -> String -> String
showP 0 _ = ""
showP 1    s = '+' : s
showP (-1) s = '-' : s
showP n s
    | n > 0      =  '+' : show n ++ '*':s
    | otherwise  =  show n ++ '*':s

showWithSign :: Integer -> String
showWithSign n
    | n >= 0     =  '+':show n
    | otherwise  =  show n



postOptK :: (Fractional a, Num cond)
         => LGenFunctions a cond
         -> GIneqSystem a -> LNest a -> DTreeGen a cond [LNest a]
postOptK lgf ctx k =
--    trace ("postOptK: extCtx'=" ++ show extCtx') $
    do mbKI <- postOptKI lgf extCtx' (innerPart k)
       case mbKI of
        Just ki -> do mbOi' <- simplIneqsG (Just extCtx) oi
                      case mbOi' of
                        Just oi' -> return [k { innerPart  = ki
		                              , outerIneqs = oi'
		                              }
                                           ]
                        Nothing  -> return []
        Nothing -> return []
    where
    extCtx' = extCtx `join` oi
    extCtx  = insertVars ctx 0 (zeroMatrix (nRows ctx) (nVars oi)) 
    oi = outerIneqs k

    simplIneqsG = lgfSimplifyIneqsG lgf

postOptKI :: (Fractional a, Num cond)
          => LGenFunctions a cond
          -> GIneqSystem a
          -> LNestInner a
          -> DTreeGen a cond (Maybe (LNestInner a))
postOptKI lgf extCtx ki@(Statement {}) = return $ Just ki
postOptKI lgf extCtx (NoLoop { kiInner=inners }) =
    do mbNewInner <- sequence [do newI <- postOptKI lgf extCtx' ki
                                  return (cs,newI) 
                              | (cs,ki) <- inners
                              ]
       let newInner = [(cs,newI) | (cs,newI') <- mbNewInner
                                 , Just newI <- [newI']
                      ]
       if null newInner then return Nothing
                        else return $ Just $ NoLoop { kiInner = newInner }
    where
    extCtx' = insertVars extCtx (nVars extCtx) (zeroMatrix (nRows extCtx) 1)
postOptKI lgf extCtx ki@(Loop {}) =
    do mbLuSysSimpl <- (lgfSimplifyIneqsG lgf) (Just extCtx') luSys
       case mbLuSysSimpl of
         Nothing         -> return Nothing
         Just luSysSimpl ->
           do let newExtCtx = extCtx' `join` luSysSimpl
              mbNewInner <- sequence [do newI <- postOptKI lgf newExtCtx i
                                         return (cs,newI) 
                                      | (cs,i) <- kiInner ki
                                     ]
              let newInner = [(cs,newI) | (cs,newI') <- mbNewInner
                                        , Just newI <- [newI']
                             ]
              -- simplify under  (Just extCtx')  ?
              -- check integral non-emptiness ?
              mbBounds <- -- trace (show (extCtx,luSysSimpl,nV)) $
                          (lgfSortIneqsGss lgf) luSysSimpl Nothing (Just nV)
              case mbBounds of
                Nothing  -> return Nothing
                Just bounds ->
                    trace (show bounds) $
                    return $ Just $ ki { kiLower=newLower
                                       , kiUpper=newUpper
                                       , kiInner=newInner
                                       }
                    where
                      (newLower,newUpper) = case bounds of
                                              [Equation e]       -> ([e],[e])
                                              [LowerUpper ls us] -> (ls,us)
    where
    luSys = ineqSystem ((nV+1,nP),ls `join` us)

    nV = nVars extCtx
    nP = nParams extCtx
    extCtx' = insertVars extCtx nV (zeroMatrix (nRows extCtx) 1)

    us = (\m -> insertColumns m nV (columnMatrix $ replicate (nRows m) (-1))) $
	 rowVectorsToMatrix (nV+nP+1) (kiUpper ki)
    ls = (\m -> insertColumns m nV (columnMatrix $ replicate (nRows m) 1)) $
	 negate $
	 rowVectorsToMatrix (nV+nP+1) (kiLower ki)

{-
timeIt :: Show a => LGenOptions -> String -> a -> a
timeIt lgo msg x
    | lgoProgressToStderr lgo =
        unsafePerformIO $
        do hPutStrLn stderr ("begin: " ++ msg)
           t1   <- getClockTime
           evaluate (length $ show x)
           t2   <- getClockTime
           let totalSecs = fromIntegral (picoDiff t1 t2) / 10^12
           hPutStrLn stderr (msg ++ ": " ++ show totalSecs ++ " s")
           return x
    | otherwise = x
    where
    picoDiff :: ClockTime -> ClockTime -> Integer
    picoDiff (TOD s1 p1) (TOD s2 p2) = s*10^12 + p
        where
        s = s2 - s1
        p = p2 - p1
-}



--------------------------------------------------------


------------------------------------------
-- Functions to Test the Generated Code --
------------------------------------------




---------------------
-- Data Structures --
---------------------


data LGenAlgoVariant = LGenSlightlyImproved
                     | LGenOriginal
                     | LGenExperimental
                       deriving (Eq, Show)

data LGenLpVariant = LGenQE
                   | LGenSimplex
                     deriving (Eq, Show)

data LGenOptions = LGO { lgoAlgoVariant :: LGenAlgoVariant,
                         lgoLpVariant   :: LGenLpVariant,
                         lgoSimplifyBounds     :: Bool,
                         lgoConvexUnion        :: Bool,
                         lgoProgressToStderr   :: Bool,
                         lgoDebug              :: Bool
                       }
                   deriving Show

lgenOptionsStd = LGO { lgoAlgoVariant = LGenOriginal,
                       lgoLpVariant   = LGenQE,
                       lgoProgressToStderr  = False,
                       lgoDebug             = True,
                       lgoSimplifyBounds    = False,
                       lgoConvexUnion       = False
                     }

data LGenFunctions a cond =
  LGF { lgfDisjointUnionDifference :: Maybe (GIneqSystem a)
                                   -> [GIneqSystem a]
                                   -> [GIneqSystem a]
                                   -> [GIneqSystem a]
      , lgfNonEmpty :: DTreeContext a cond
                    -> Maybe (GIneqSystem a)
                    -> GIneqSystem a
                    -> Bool
      , lgfUnitDenom :: a -> Bool  -- denominator is a unit (+/- 1)
      , lgfIsConst :: a -> Maybe Rational
      , lgfImplies :: DTreeContext a cond
                   -> GIneqSystem a
                   -> Bool
                   -> GIneqSystem a
                   -> QfFormula a

      , lgfProjectToVarsG :: GIneqSystem a
                          -> Maybe (GEqSystem a)
                          -> Maybe Integer
                          -> [Integer]
                          -> DTreeGen a cond (Maybe (GIneqSystem a))
      , lgfSortIneqsGss :: GIneqSystem a
                        -> Maybe (GEqSystem a)
                        -> Maybe Integer
                        -> DTreeGen a cond (Maybe (SortedSystem a))
      , lgfSimplifyIneqsG :: Maybe (GIneqSystem a)   -- ^ context with vars
                         -> GIneqSystem a
                         -> DTreeGen a cond (Maybe (GIneqSystem a))
      , lgfSimplifyDomainG :: Maybe (GIneqSystem a)   -- ^ context with vars
                           -> [GIneqSystem a]
                           -> DTreeGen a cond [GIneqSystem a]
      }


------------------------------------

--
-- LGenFunctions for RealFrac a => a
--

lgfRealFrac :: RealFrac a => LGenFunctions a a
lgfRealFrac = LGF { lgfDisjointUnionDifference = disjointUnionDifference
                  , lgfNonEmpty = hasIntegralPointRF
                  -- check the denominator of the corresponding IFracPoly,
                  -- because in (1%2) %% 1 we want 2 as the denominator
                  , lgfUnitDenom = (\d -> d == 1 || d == -1) .
                                   denominator .
                                   toRational
                  , lgfIsConst = Just . toRational
                  , lgfImplies = impliesIneqs

                  , lgfProjectToVarsG = projectToVarsG
                  , lgfSortIneqsGss   = sortIneqsGss
                  , lgfSimplifyIneqsG = simplifyIneqsRF
                  , lgfSimplifyDomainG = chernikovaSimplifyDomainG
                  }


hasIntegralPointRF :: RealFrac a
                   => DTreeContext a a
                   -> Maybe (GIneqSystem a) 
                   -> GIneqSystem a
                   -> Bool
hasIntegralPointRF dtc mbCtx sys =
--    (\b -> trace ("hasIntegralPoint " ++ show mbCtx ++ " " ++ show sys ++ " --> " ++ show b) b) $
    omegaTest (Just is) (Just es)
    where
    nV = nVars sys
    nP = nParams sys
    (is,es) = findEqualities sys'
    sys' = case mbCtx of
	      Nothing  -> sys
	      Just ctx -> sys `join`
			  insertVars ctx 0 (zeroMatrix (nRows ctx) nV)

simplifyIneqsRF :: RealFrac a
                => Maybe (GIneqSystem a) -- context with variables
                -> GIneqSystem a
                -> DTreeGen a a (Maybe (GIneqSystem a))
simplifyIneqsRF mbCtx sys =
    return $ Just $ simplifyIneqs mbCtx sys

impliesIneqs :: RealFrac a
             => DTreeContext a a
             -> GIneqSystem a
             -> Bool
             -> GIneqSystem a
             -> QfFormula a
impliesIneqs dtc assump weak conds =
    assert (nVars   assump == nVars   conds) $
    assert (nParams assump == nParams conds) $
    if all (impliesIneq assump weak) $ rowVectors $ getMatrix conds
    then QfAt AtTrue
    else QfAt AtFalse

impliesIneq :: RealFrac a => GIneqSystem a -> Bool -> GVector a -> Bool
impliesIneq assump weak cond =
    trace ("impliesIneq: " ++ show (assump,cond)) $
    case opt of
      Unbounded     -> True
      Nonexistent   -> True
      Optimum (o,_) -> if weak then (val>=0) else (val>0)
          where
          val = o + addconst
    where
    opt = runNumOrdDTreeGenAlg $
          simplex stdSimplexParams assump' Nothing target

    assump' = moveParamsToVars assump 0 nP nV
    nV = nVars assump
    nP = nParams assump

    target   = vector $ init $ vectorEntries cond
    addconst = last $ vectorEntries cond



------------------------------------

type ScatterFn a = GIneqSystem a

data LGenParams a =
    LGP { lgpContext :: GIneqSystem a
	, lgpStatements :: [([GIneqSystem a], ScatterFn a)]
	}
		  deriving Show

data LGenOutputOptions a =
    LGOO { lgooNonlinParamNames :: [String]
         , lgooParamNames :: [String]
	 , lgooVarNames   :: [String]
         , lgooIterNames  :: [String]
         , lgooStmtGen    :: Integer -> [GVector a] -> Maybe String
         , lgooLoopGen :: Integer
                       -> String
                       -> [[GVector a]]
                       -> [[GVector a]]
                       -> String
                       -> String
                       -> [Integer]
                       -> Maybe String
         , lgooIneqSystemToString :: [String]
                                  -> GIneqSystem a
                                  -> String
         , lgooAffineExpToString :: [String]
                                 -> Integer  -- number of linear parameters
                                 -> GVector a
                                 -> (String, Either String Integer)
         }

instance Show a => Show (LGenOutputOptions a) where
    showsPrec prec lgoo =
        showString "LGOO {lgooParamNames=" .
        shows (lgooParamNames lgoo) .
        showString ", lgooVarNames=" .
        shows (lgooVarNames lgoo) .
        showString ", lgooIterNames=" .
        shows (lgooIterNames lgoo) .
        showString ", lgooStmtGen=_, lgooLoopGen=_}"

emptyLGenOutputOptions :: Num a => LGenOutputOptions a
emptyLGenOutputOptions =
    LGOO { lgooNonlinParamNames = []
         , lgooParamNames = []
         , lgooVarNames   = []
         , lgooIterNames  = []
         , lgooStmtGen    = \_ _ -> Nothing
         , lgooLoopGen    = \_ _ _ _ _ _ _ -> Nothing
         , lgooIneqSystemToString = emptyLgooIneqSystemToString
         , lgooAffineExpToString = emptyLgooAffineExpToString
         }


emptyLgooIneqSystemToString :: Num a
                            => [String]
                            -> GIneqSystem a
                            -> String
emptyLgooIneqSystemToString linNames sys =
    concat (intersperse " && " $
            map (\v -> let (s1,Right 1) = emptyLgooAffineExpToString linNames 0 v
                       in  s1 ++ " >= 0") aes
           )
    where
    aes = rowVectors $ getMatrix sys

emptyLgooAffineExpToString :: Num a
                           => [String]
                           -> Integer
                           -> GVector a
                           -> (String, Either String Integer)
emptyLgooAffineExpToString linNames nP v =
    assert (length linNames == vectorLength v - 1) $
    (case concat $ intersperse "+" $ filter (not . null) $
              (zipWith showFn vp linNames ++ [showFn2 co]) of
       "" -> "0"
       s  -> s
    , Right 1
    )
    where


    showFn 0 n = ""
    showFn c n = "("++show c++")*"++n

    showFn2 0 = ""
    showFn2 c = "(" ++ show c ++ ")"

    ents = vectorEntries v
    vp = init ents
    co = last ents

realFracLGOO :: RealFrac a
             => [String]
             -> [String]
             -> [String]
             -> LGenOutputOptions a
realFracLGOO iterNames varNames linParamNames =
    fpLGOO { lgooStmtGen    = \_ _ -> Nothing
           , lgooLoopGen    = \_ _ _ _ _ _ _ -> Nothing
           ,  lgooIneqSystemToString =
                 \linNames is -> lgooIneqSystemToString fpLGOO linNames $
                                 fmap (fromRational . toRational) is
           , lgooAffineExpToString =
               \linNames nP v -> lgooAffineExpToString fpLGOO linNames nP $
                                 fmap (fromRational . toRational) v
           }
    where
    fpLGOO :: LGenOutputOptions (FracPoly Rational)
    fpLGOO = fracPolyLGOO iterNames varNames linParamNames []

fracPolyLGOO :: RealFrac a
             => [String]
             -> [String]
             -> [String]
             -> [String]
             -> LGenOutputOptions (FracPoly a)
fracPolyLGOO iterNames varNames linParamNames nlParamNames =
    LGOO { lgooNonlinParamNames = nlParamNames
         , lgooParamNames = linParamNames
         , lgooVarNames   = varNames
         , lgooIterNames  = iterNames
         , lgooStmtGen    = \_ _ -> Nothing
         , lgooLoopGen    = \_ _ _ _ _ _ _ -> Nothing
         , lgooIneqSystemToString = ineqs2str
         , lgooAffineExpToString = ae2str
         }
    where
    ae2str :: RealFrac a
           => [String]
           -> Integer
           -> GVector (FracPoly a)
           -> (String, Either String Integer)
    ae2str linNames nP v =
--        trace ("ae2str " ++ show (names, nP, v, numPoly, denomPoly)) $
        assert (length linNames == vectorLength v - 1) $
        case isConstPoly denomPoly of
          Just c  -> (numStr, Right c)
          Nothing -> (numStr, Left denomStr)
        where
        -- numerator is an affine expression with polynomial coefficients
        -- (-> names), denominator is a LCM of the coefficient denominators
        -- only (-> nlParamNames).
        numStr    = showPolyC names        numPoly
        denomStr  = showPolyC nlParamNames denomPoly

        numPoly   = constPoly coDenom * numPoly'
        denomPoly = constPoly coNum   * denomPoly'

        (coNum, numPoly') = 
            ratPoly2intPoly $
            flatPoly $
            affinePolyFromList $
            fmap (\e -> polyNumerator e * (d `quotPoly` polyDenominator e)) ents

        (coDenom, denomPoly') = ratPoly2intPoly d
        d    = foldl lcmPoly 1 $ map polyDenominator ents
        ents = vectorEntries v

        names = linNames ++ nlParamNames

    ineqs2str :: RealFrac a
              => [String]
              -> GIneqSystem (FracPoly a)
              -> String
    ineqs2str linNames sys =
        assert (length linNames == nVars sys + nParams sys) $
        qfFormulaToC (showPolyC names . snd . ratPoly2intPoly) $
        ineqSystemToFormula sys'
        where
        sys' = unfracIneqSystem Nothing sys
        names = linNames ++ nlParamNames


loopGenIF :: Integer
          -> String
          -> [[GVector a]]
          -> [[GVector a]]
          -> String 
          -> String
          -> String
loopGenIF n var ls us lb ub =
        case (ls,us) of
          ([[]],[[]]) -> "if (1)"
          (_,[[]]) -> concat ["if (", lb, "<=", var, ")"]
          ([[]],_) -> concat ["if (", var, "<=", ub, ")"]
          (_,_)  -> concat ["if (", lb, "<=", var, " && ", var, "<=", ub, ")"]

loopGenSelfNP :: String
              -> String
              -> Integer
              -> String
              -> [[GVector a]]
              -> [[GVector a]]
              -> String
              -> String
              -> String
loopGenSelfNP self np n var ls us lb ub =
    case (ls,us) of
      ([[]],[[]]) -> "if (1)"
      (_,_) ->
            concat ["for (", var, "=", lb, "+", self, "; ", var, "<=", ub
                   , "; ", var, "+=", np, ")"]



parseLGenInput :: RealFrac a
               => Parser ( LGenParams (FracPoly a)
                         , LGenOutputOptions (FracPoly a)
                         , [QfFormula (Polynomial a)]
                         )
parseLGenInput =
    do spacesOrComment
       oneOf "cf"
       spacesOrComment
       variant <- try (string "nlgen") <|> return ""
       (nlParamNames,nlContext) <-
           if variant == "nlgen" then
               do nlN   <- parseNlParamNames
                  nlCtx <- parseNlContext nlN
                  return (nlN,nlCtx)
           else
               return ([],[])
       ctx <- parseContext nlParamNames
       paramNames <- parseNames (Just [[c] | c <- ['M'..]]) (nParams ctx)
       doms <- parseDomains nlParamNames (nParams ctx)
       let stdIterNames = [[c] | c <- ['i'..]]
       iterNames <- parseNames (Just stdIterNames) $
                               maximum (0 : map nVars (concat doms))
       scatter' <- parseScatterFns nlParamNames (nParams ctx)
       when (not (null scatter') && length scatter' /= length doms) $
            fail "Number of scattering functions does not match number of statements"
       let scatter = if null scatter' then makeIdentityScatterFns doms
                                      else scatter'
       let tD = nVars (head scatter) - nVars (head $ head doms)
       let scatterNames = ['c':show i | i <- [1..tD]]
       varNames <- if null scatter'
                   then return scatterNames
                   else parseNames (Just scatterNames) tD
       spacesOrComment
       eof
       -- when no scattering functions are given, we use the names
       -- of the iterators for the scatter dimensions; this makes the
       -- output more CLooG-like.
       let (vNames,iNames) = if null scatter'
                             then (iterNames, varNames )
                             else (varNames,  iterNames)
       return (LGP { lgpContext    = ctx,
		     lgpStatements = [(dom,s) | dom <- doms | s <- scatter]
		   }
              ,fracPolyLGOO iNames vNames paramNames nlParamNames
              ,nlContext
              )

makeIdentityScatterFns :: Fractional a
                       => [[GIneqSystem a]]
                       -> [ScatterFn a]
makeIdentityScatterFns doms =
    map mkS doms
    where
    tDim = maximum $ map nVars $ concat doms

    mkS (sys:_) = ineqSystemVPC
                    (eqsV `join` negate eqsV
                    ,zeroMatrix (2*tDim) nP
                    ,zeroMatrix (2*tDim) 1
                    )
        where
        eqsV = negate (unitMatrix tDim tDim) `conc` unitMatrix  tDim nV
        nV = nVars sys
        nP = nParams sys

parseNlParamNames :: Parser [String]
parseNlParamNames =
    do spacesOrComment
       n <- liftM readI $ many1 digit
       count n name

parseNlContext :: Fractional a => [String] -> Parser [QfFormula (Polynomial a)]
parseNlContext nlNames =
    do spacesOrComment
       n <- liftM readI $ many1 digit
       count n (nlContext nlNames)

nlContext :: Fractional a => [String] -> Parser (QfFormula (Polynomial a))
nlContext nlNames =
    do spacesOrComment
       qf1 <- parseQfFormula (parsePoly nlNames)
              :: Parser (QfFormula (Polynomial Integer))
       let qf = fmap (fmap' fromIntegral) qf1
       spaces
       char ';'
       return qf

parseNames :: Maybe [String] -> Int -> Parser [String]
parseNames stdnames n =
    do spacesOrComment
       flag <- oneOf "01"
       case (flag,stdnames) of
         ('0', Nothing) -> fail "Must specify names for NLGen"
         ('0', Just ns) -> return $ take n ns
         (  _, _      ) -> count n name

name :: Parser String
name = spacesOrComment >>
       many1 (letter <|> alphaNum <|> char '_')

parseContext :: Fractional a => [String] -> Parser (GIneqSystem (FracPoly a))
parseContext nlParamNames =
    do spacesOrComment
--       char '1'
       (bs,m) <- parsePMatrix nlParamNames
       return $ pm2ineqSystem 0 (nColumns m-1) (bs,m)

parseDomains :: Fractional a
             => [String] -> Integer -> Parser [[GIneqSystem (FracPoly a)]]
parseDomains nlNames nP =
    do spacesOrComment
       n <- liftM readI $ many1 digit
       count n (parseDomain nlNames nP)

parseDomain :: Fractional a
            => [String] -> Integer -> Parser [GIneqSystem (FracPoly a)]
parseDomain nlNames nP =
    do spacesOrComment
       n <- liftM readI $ many1 digit
       res <- count n $ do (bs,m) <- parsePMatrix nlNames
			   return $ pm2ineqSystem (nColumns m - nP - 1) nP (bs,m)
       count 3 (spacesOrComment >> char '0')
       return res

parseScatterFns :: Fractional a
                => [String] -> Integer -> Parser [ScatterFn (FracPoly a)]
parseScatterFns nlNames nP =
    do spacesOrComment
       n <- liftM readI $ many1 digit
       count n (parseScatterFn nlNames nP)

parseScatterFn :: Fractional a
               => [String] -> Integer -> Parser (ScatterFn (FracPoly a))
parseScatterFn nlNames nP =
    do spacesOrComment
       m <- parsePMatrix nlNames
       return $ pm2scatterFn nP m

parsePMatrix :: Fractional a
             => [String] -> Parser ([Bool], GMatrix (FracPoly a))
parsePMatrix nlNames =
    do spacesOrComment
       r <- liftM readI $ many1 digit
       when (r < 0) $ fail "negative number of rows for matrix"
       spacesOrComment
       c <- liftM readI $ many1 digit
       when (c <= 0) $ fail "non-positive number of columns for matrix"
       ents <- count r (count c (spacesOrComment >> parseCoeff nlNames))
       return (map ((/=0).head) ents, matrix (r,c-1,concatMap tail ents))

pm2scatterFn :: Fractional a
             => Integer -> ([Bool],GMatrix a) -> ScatterFn a
pm2scatterFn nP (bs,m) =
    pm2ineqSystem (nColumns m-nP-1) nP (bs,m)

pm2ineqSystem :: Fractional a
              => Integer
              -> Integer
              -> ([Bool],GMatrix a)
              -> GIneqSystem a
pm2ineqSystem nV nP (bs,m) =
    ineqSystem ((nV,nP), rowVectorsToMatrix (nV+nP+1) $
		         concat [f b r | b <- bs | r <- rowVectors m]
	       )
    where
    f False r = [r,-r]
    f True  r = [r]

readI :: String -> Int
readI = read

parseCoeff :: Fractional a => [String] -> Parser (FracPoly a)
parseCoeff nlNames =
    (char '-' >> liftM (fromIntegral . negate . readI) (many1 digit))
    <|> liftM (fromIntegral . readI) (many1 digit)
    <|>
    do char '('
       p <- parsePoly nlNames
       char ')'
       return (p%%1)

spacesOrComment :: Parser ()
spacesOrComment =
    do many $ ((many1 space >> return ()) <|>
	       (char '#' >> manyTill anyToken ((newline >> return ())
					       <|> eof
					      ) >>
		return ()
	       )
	      )
       return ()

writeLGenInput :: RealFrac a => LGenParams a -> LGenOutputOptions a -> String
writeLGenInput lgp lgoo =
    "c\n" ++ writeIneqs (lgpContext lgp) ++ "1\n" ++
    concat (intersperse " " (lgooParamNames lgoo)) ++ "\n" ++
    show (length $ lgpStatements lgp) ++ "\n" ++
    concatMap writeStatement (lgpStatements lgp) ++ "1\n" ++
    concat (intersperse " " (lgooIterNames lgoo)) ++ "\n" ++
    show (length $ lgpStatements lgp) ++ "\n" ++
    concatMap writeScatterFn (lgpStatements lgp) ++ "1\n" ++
    concat (intersperse " " (lgooVarNames lgoo)) ++ "\n"

writeIneqs :: RealFrac a => GIneqSystem a -> String
writeIneqs sys =
    show (nRows eqs + nRows ineqs) ++ " " ++ show (nColumns intSys + 1) ++
    "\n" ++
    writeMatrix 0 (getMatrix eqs) ++
    writeMatrix 1 (getMatrix ineqs)
    where
      intSys :: GIneqSystem Integer
      intSys = integralIneqSystem sys

      (ineqs,eqs) = findEqualities intSys

writeMatrix :: Integral a => Integer -> GMatrix a -> String
writeMatrix first m =
    unlines [show first ++ concat [' ':show e | e <- r] | r <- rows' m]

writeStatement :: RealFrac a
               => ([GIneqSystem a], ScatterFn a) -> String
writeStatement ([],_)    = error "writeLGenInput: empty union"
writeStatement (ineqs,_) = writeDomain ineqs

writeDomain :: RealFrac a => [GIneqSystem a] -> String
writeDomain ineqs =
    show (length ineqs) ++ "\n" ++
    concatMap writeIneqs ineqs ++
    "0 0 0\n"

writeScatterFn :: RealFrac a => ([GIneqSystem a], ScatterFn a) -> String
writeScatterFn (_, f) = writeIneqs f




