-- |
-- This module contains a straight forward implementation of the
-- Omega test as described in /The Omega test: a fast and practical
-- integer programming algorithm for dependence analysis/ by William Pugh
-- published in /Supercomputing/, pages 4-13, 1991 (availale at
-- @http:\/\/citeseer.ist.psu.edu\/pugh92omega.html@).
-- It is _not_ as smart as the paper suggests in some
-- aspects of the implementation. For example, it does not
-- try to find an exact projection but simply projects by
-- the first variale always.
--
module LMath.Algo.OmegaTest
    ( omegaTest, omegaTest' )
where

import Control.Exception
import Data.List
import Data.Ratio
import Debug.Trace

import LMath.Types

-- |
-- Perform the Omega test.
-- Returns @True@ if the given system of inequalities and
-- equalities is feasile in the integers (i.e., there exists
-- an assignment of integral values 
-- for the parameters and the variables that satisfies the
-- given system).
--
omegaTest :: Real a => Maybe (GIneqSystem a) -> Maybe (GEqSystem a) -> Bool
omegaTest Nothing Nothing = True
omegaTest (Just sys) Nothing   = omegaTest' ineqs []
    where
    ineqs = rowVectors $ getMatrix $ integralIneqSystem sys
omegaTest Nothing (Just eqsys) = omegaTest' [] eqs
    where
    eqs   = rowVectors $ getMatrix $ integralEqSystem   eqsys
omegaTest (Just sys) (Just eqsys) =
    assert (nVars sys   == nVars eqsys) $
    assert (nParams sys == nParams eqsys) $
    omegaTest' ineqs eqs
    where
    ineqs = rowVectors $ getMatrix $ integralIneqSystem sys
    eqs   = rowVectors $ getMatrix $ integralEqSystem   eqsys

--
-- Jump into the omega test. The inequalities (first parameter)
-- and equalities (second parameter) are given as
-- vectors, the variables preceeding the constant coefficient.
--
omegaTest' :: [GVector Integer] -> [GVector Integer] -> Bool
omegaTest' ineqs eqs 
    | Just startEqs <- normalizeEqs eqs  =
        case handleEqs (normalizeIneqs ineqs) startEqs of
	  Nothing       -> False
	  Just newIneqs ->
	      let simpleIneqs = simplify newIneqs in
	      if hasRealSolution simpleIneqs
		 then handleIneqs simpleIneqs
		 else False
    | otherwise  =  False
    

--
-- Normalise the given inequalities.
-- Divide the coefficients of the variales and parameters
-- by their greatest common divisor @g@ and floor divide
-- the constant coefficient by @g@ (i.e., make the
-- inequality as tight as possible in the integers).
--
normalizeIneqs :: [GVector Integer] -> [GVector Integer]
normalizeIneqs ineqs =
    map head $          -- select the tightest of the ineqs
    groupBy coeffEq $   -- all the ineqs with the same var-part in one group
    sort $
    map normalizeIneq ineqs
    where
    coeffEq v1 v2 = init (vectorEntries v1) == init (vectorEntries v2)

-- 
-- Normalise one inequality.
--
normalizeIneq :: GVector Integer -> GVector Integer
normalizeIneq ineq
    | g == 0     =  ineq
    | otherwise  =
	fmap (\c -> floor (c % g)) ineq
    where
    g = foldl gcd0 0 $ init $ vectorEntries ineq

--
-- Normalise the given set of equalities.
-- Divide the coefficients of an equality by their
-- greatest common divisor. If an equality is trivially
-- infeasible in the integers (i.e., the greatest common divisor
-- of the coefficient of the variables and parameters does not
-- divide the constant coefficient), @Nothing@ is returned.
-- Otherwise, the nomalised equalities are returned (with
-- the trivial equalities @0=0@ removed).
--
normalizeEqs :: [GVector Integer] -> Maybe [GVector Integer]
normalizeEqs eqs
    | Left False `elem` rs  =  Nothing
    | otherwise  =
	Just [eq | Right eq <- rs]
    where
    rs = filter (/=Left True) $ map normalizeEq eqs

--
-- Normalise a single equality.
-- Return @Left True@ or @Left False@ if the equality is
-- determined to be true or false for trivial reasons,
-- respectively.
--
normalizeEq :: GVector Integer -> Either Bool (GVector Integer)
normalizeEq eq
    | g == 0            =  Left (c==0)
    | c `mod` g  ==  0  =  Right $ fmap (`div` g) eq
    | otherwise         =  Left False
    where
    g = foldl gcd0 0 $ init $ vectorEntries eq
    c = last $ vectorEntries eq

--
-- Handle the equalities one after the other.
--
handleEqs :: [GVector Integer] -> [GVector Integer] -> Maybe [GVector Integer]
handleEqs ineqs [] = Just ineqs
handleEqs ineqs (eq:eqs) =
    case handleEq ineqs eqs eq of
      Nothing                -> Nothing
      Just (newIneqs,newEqs) -> handleEqs newIneqs newEqs

--
-- Handle one equality.
-- Substitute into the other equalities and inequalities
-- when eq has a coefficient +1 or -1, replace one variable
-- otherwise. Return the set of new inequalities and equalities
-- (or @Nothing@ if the system is determined to be infeasible).
--
handleEq :: [GVector Integer]
	 -> [GVector Integer]
	 -> GVector Integer
	 -> Maybe ([GVector Integer],[GVector Integer])
handleEq ineqs eqs eq
    | Just j <- findIndex ((==1).abs) $ init $ vectorEntries eq  =
      case normalizeEqs $ map (substitute j eq) eqs of
        Nothing     -> Nothing
	Just subEqs ->
          let subIneqs = normalizeIneqs $ map (substitute j eq) ineqs in
          Just (subIneqs, subEqs)
    | Just newEqs <- mbNewEqs = Just (newIneqs, newEqs)
    | otherwise  =  Nothing
    where
    (repl,k) = replaceExpr eq
    mbNewEqs = normalizeEqs $ map (replace k repl) (eq:eqs)
    newIneqs = normalizeIneqs $ map (replace k repl) ineqs


--
-- Determine which variale to replace by a fresh variable
-- in the equality handling phase of the Omega test.
-- Return the expression to substitute for the
-- selected variable (called @x_k@ in the paper)
-- and its index. Note that @k=0@ means the first
-- variable here (in the paper, @k=0@ refers to the
-- constant coefficient).
--
replaceExpr :: GVector Integer -> (GVector Integer, Int)
replaceExpr eq = (xk_expr, k)
    where
    xk_expr = setVectorEntry (fmap (\c -> sgn_ak * mod_m c) eq)
	                     k
			     (negate sgn_ak * m)

    Just k = elemIndex abs_ak abs_coeffs
    abs_ak = minimum $ filter (>0) $ abs_coeffs
    sgn_ak = signum $ vectorEntry eq k
    m = abs_ak + 1

    mod_m a = a - m * floor(a % m + 1%2)

    abs_coeffs = map abs $ init $ vectorEntries eq

--
-- Handle the inequalities and return whether
-- they are feasible in the integers.
--
handleIneqs :: [GVector Integer] -> Bool
handleIneqs [] = True
handleIneqs ineqs
    -- The cases for 0 and 1 variales are handled directly.
    | nV == 0  =  all (>=0) [vectorEntry v 0 | v <- ineqs]
    | nV == 1  =
        -- Determine if the ceiling of the maximum of the lower bounds is
        -- less than the floor of the minimum of the upper bounds.
	let ls = [ceiling (negate b % a) | v <- ineqs,
		  let [a,b] = vectorEntries v, a > 0
		 ]
            us = [floor (b % negate a) | v <- ineqs,
		  let [a,b] = vectorEntries v, a < 0
		 ]
	    cs = [b | v <- ineqs, let [a,b] = vectorEntries v, a == 0]
        in all (>=0) cs && (null ls || null us || (maximum ls <= minimum us))
	     
    | otherwise  =
        -- We could check if the exact shadow has a real
        -- solution before proceeding with the test, but
        -- it seems that this consumes more time than it
        -- saves on average.
        --
        -- Check if there is an exact projection; if there is not,
        -- use the non-exact projection which needs to test the
        -- fewest sub-problems.
        case (exactShadows,nonexactShadows) of
          (exactShadow:_, _) ->
--	      hasRealSolution exactShadow &&
              handleIneqs exactShadow
          ([], (realShadow,darkShadow,_,testEqs):_) ->
--	      hasRealSolution realShadow &&
              (handleIneqs darkShadow ||
	       any (\e -> omegaTest' ineqs [e]) testEqs
	      )
    where
    nV = vectorLength (head ineqs) - 1

    -- Compute the projections w.r.t. every variable and
    -- partition them in exact and non-exact projections.
    -- The non-exact projections are sorted (ascending) by the number of
    -- sub-problems they require.
    candidates   = [shadows elimVar ineqs | elimVar <- [0..nV-1]]
    exactShadows = [exS | Left exS <- candidates]
    nonexactShadows =
        sortBy (\(_,_,nTEs1,_) (_,_,nTEs2,_) -> nTEs1 `compare` nTEs2) $
        [r | Right r <- candidates]

--
-- Determine the real and dark shadows of the given inequalities
-- when projecting away the given variable and the
-- additional equalities needed to check for integral
-- solutions close to the bounds of the polyhedron.
-- Returns either @Left shadow@ if the real and dark shadow are
-- identical or @Right real dark testEqs@ if they differ.
-- @testEqs@ is the set of equalities that (each for itself) has
-- to be added to the system when the dark shadow does not have
-- integral solutions but the real shadow has.
-- 
shadows :: Integer
	-> [GVector Integer]
	-> Either [GVector Integer]
	        ([GVector Integer],[GVector Integer],Integer,[GVector Integer])
shadows elimVar ineqs
    | realShadow == darkShadow  =  Left $ simplify realShadow
    | otherwise                 =
	Right (simplify realShadow, simplify darkShadow, nTestEqs, testEqs)
    where
    nTestEqs = sum [maxI b + 1 | (b,_) <- lower]
    testEqs = [insertEntries (addLast i beta) elimVar [negate b]
	      | (b,beta) <- lower, i <- [0 .. maxI b]
	      ]
    maxI b = floor ((max_a*b-max_a-b) % max_a)

    max_a = maximum [a | (a,alpha) <- upper]
    realShadow = [mulSV b alpha `subVV` mulSV a beta
		  | (b,beta) <- lower, (a,alpha) <- upper
		 ]
		 ++ other
    darkShadow = [addLast ((1-a)*(b-1)) (mulSV b alpha `subVV` mulSV a beta)
		  | (b,beta) <- lower, (a,alpha) <- upper
		 ]
		 ++ other

    lower = [spl1 v | v <- ineqs, vectorEntry v elimVar > 0]
    upper = [spl2 v | v <- ineqs, vectorEntry v elimVar < 0]
    other = [deleteEntries v elimVar 1
	     | v <- ineqs, vectorEntry v elimVar == 0
	    ]

    spl1 v = (vectorEntry v elimVar, negate $ deleteEntries v elimVar 1)
    spl2 v = (negate $ vectorEntry v elimVar, deleteEntries v elimVar 1)

--
-- Add a value to the last entry of a vector.
--
addLast :: Num a => a -> GVector a -> GVector a
addLast x v =
    setVectorEntry v off (vectorEntry v off + x)
    where
    off = vectorLength v - 1

--
-- Simplify the given inequalities.
-- Calls normalizeIneqs.
-- Should this also call a simplifier which simplifies
-- the inequalities in the reals?
--
simplify :: [GVector Integer] -> [GVector Integer]
simplify = normalizeIneqs

--
-- Determine if the given inequalities have a solution
-- in the reals.
--
hasRealSolution :: [GVector Integer] -> Bool
hasRealSolution [] = True
hasRealSolution ineqs =
    not $
    areIneqsInfeasible $
    ineqSystem ((nV,0), fmap fromIntegral $ rowVectorsToMatrix1 ineqs)
    where
    nV = vectorLength (head ineqs) - 1

--
-- Solve @eq@ for entry @j@ (which has to be @+1@ or @-1@) and
-- substitute into @vect@ (making @vect@ one entry shorter).
--
substitute :: Int -> GVector Integer -> GVector Integer -> GVector Integer
substitute j eq vect =
    deleteEntries
      (addVV vect (eq `mulVS` negate (vectorEntry eq j * vectorEntry vect j)))
      j 1

--
-- Replace variable @j@ in @vect@ with @eq@ whereby
-- introducing a new variable (@j@ in @eq@).
-- (I.e., @eq@ is the right-hand side of the equation
-- a_@j@ = @eq@).
--
replace :: Int -> GVector Integer -> GVector Integer -> GVector Integer
replace j eq vect =
    setVectorEntry vect j 0 `addVV` (eq `mulVS` vectorEntry vect j)


--
-- For the Haskell Prelude, gcd 0 0 is undefined, but
-- it is more comfortable (and consistent) to define it 
-- to be 0.
--
gcd0 :: Integer -> Integer -> Integer
gcd0 0 0 = 0
gcd0 x y = gcd x y


--
-- This is the "Omega test nightmare" example from the paper.
-- @omegaTest' nightmareIneqs []@ should be @False@.
--
nightmareIneqs :: [GVector Integer]
nightmareIneqs = [vector [11,13,-27], vector [-11,-13,45],
		  vector [7,-9,10],  vector [-7,9,4]
		 ]
