module LMath.Algo.Simplex
    ( SimplexParams(..), stdSimplexParams,
      simplex, lexmin, lexmax, branchAndCut,
      integrateEqs, isFeasibleSys,
      getPivotingCounter, resetPivotingCounter,

      isRedundant
    )
where

import Control.Exception ( assert )
import Data.IORef
import Data.List
import qualified Data.Set as Set
import Data.Set ( Set )
import System.IO.Unsafe

import Debug.Trace ( trace )

import LMath.Types

isRedundant :: (Fractional a, Num cond)
	    => [GVector a]
            -> GVector a
	    -> DTreeGen a cond Bool
isRedundant sys ineq =
--    trace (show (sys,ineq,matr,nV)) $
    compDTreeG checkRed $
    simplex stdSimplexParams (ineqSystem ((nV,0), matr)) Nothing target
    where
    matr    = rowVectorsToMatrix (nV+1) sys
    nV      = vectorLength ineq - 1
    target  = negate $ deleteEntries ineq nV 1
    addTerm = vectorEntry ineq nV

    checkRed Nonexistent = dLeaf True
    checkRed Unbounded   = dLeaf False
    checkRed (Optimum (v,xs)) =
	withDTreeContext $ \dtc ->
	dLeaf (signumGe $ dtcContextImpliesSignum dtc (addTerm-v))

-- 
-- Eliminate redundant bounds from a given system by
-- using the Simplex algorithm (to test each bound).
--
simplexEliminateRedundantBounds = undefined
		   

-- |
-- Test a system for feasibility.
-- Returns a tree (generator) which carries @True@ at
-- a leaf if the system is feasible under the respective context
-- and @False@ otherwise.
--
isFeasibleSys :: (Fractional a, Num cond)
	      => GIneqSystem a          -- ^ inequality system
	      -> Maybe (GEqSystem a)    -- ^ optional equality system
	      -> DTreeGen a cond Bool
isFeasibleSys ineqs mbEqs = withDTreeContext ifs
    where
    ifs dtc =
	mapDTreeG toBool $
	cFeasibleDictionary stdSimplexParams ineqs''
		        (vector $ replicate (nVars ineqs'') 0)
	where
	-- ineqs'' has a variable for the "negative part"
	ineqs'' = ineqSystemVPC (insertColumns v (nColumns v) $ negate $
	          columnMatrix $ map sum (rows' v), p, c)
        (v,p,c) = partsVPC ineqs'
	(ineqs', _, _, _) = integrateEqs isNeq mbEqs ineqs
			          (vector $ replicate (nVars ineqs) 0)
	isNeq = signumNeq . dtcContextImpliesSignum dtc

    toBool Unbounded   = True
    toBool (Optimum _) = True
    toBool Nonexistent = False

--------------------------------
-- Lexmin/max based on Simpex --
--------------------------------


lexmin :: (Fractional a, Num cond)
       => SimplexParams
       -> GIneqSystem a
       -> Maybe (GEqSystem a)
       -> DTreeGen a cond (Optimum [a])
lexmin params ineqs mbEqs =
    mapDTreeG (fmap (map negate)) $
    lexmax params (mirrorIneqs ineqs) (fmap mirrorEquations mbEqs)


lexmax :: (Fractional a, Num cond)
       => SimplexParams
       -> GIneqSystem a
       -> Maybe (GEqSystem a)
       -> DTreeGen a cond (Optimum [a])
lexmax params = simplex2lexmax (simplex params)

simplex2lexmax :: (Fractional a, Num cond)
	       => (GIneqSystem a ->
		   Maybe (GEqSystem a) ->
		   GVector a ->
		   DTreeGen a cond (Optimum (a,[a]))
		  )
	       -> GIneqSystem a
	       -> Maybe (GEqSystem a)
	       -> DTreeGen a cond (Optimum [a])
simplex2lexmax simpl ineqs mbEqs = withDTreeContext s2l
    where
    s2l dtc =
	mapDTreeG (fmap (\vs -> vectorEntries (resubAF `applyIH` vector vs))) $
	simplex2lexmax' (\ins tg -> simpl ins Nothing tg) ineqs'
	where
	(ineqs', _, _, resubAF) = integrateEqs isNeq mbEqs ineqs
			                (vector $ replicate (nVars ineqs) 0)
	isNeq = signumNeq . dtcContextImpliesSignum dtc


simplex2lexmax' :: (Fractional a, Num cond)
	         => (GIneqSystem a -> GVector a -> DTreeGen b cond (Optimum (a,[a])))
		 -> GIneqSystem a
		 -> DTreeGen b cond (Optimum [a])
simplex2lexmax' simpl ineqs
    | nV == 0   = mapDTreeG (fmap (\_ -> [])) $ simpl ineqs (vector [])
    | otherwise = compDTreeG recurse $ simpl ineqs target
    where
    recurse Nonexistent = dLeaf Nonexistent
    recurse Unbounded   = dLeaf Unbounded
    recurse (Optimum (res,_)) =
	mapDTreeG (fmap (res:)) $ lm $ substVar ineqs 0 res
    lm = simplex2lexmax' simpl
    nV     = fromIntegral $ nVars ineqs
    target = vector (1 : replicate (nV-1) 0)


-- 

branchAndCut :: (Fractional a, Num cond)
             => (a -> a)
             -> SimplexParams
             -> GIneqSystem a
             -> Maybe (GEqSystem a)
             -> GVector a
             -> DTreeGen a cond (Optimum (a,[a]))
branchAndCut floorFn simplexParams ineqs mbEqs target =
    bac ineqs
    where
    nV = nVars ineqs

    bac   sys = compDTreeG branch $ splex sys
        where
        branch Nonexistent = dLeaf Nonexistent
        branch Unbounded   = dLeaf Unbounded
        branch (Optimum (opt,vars)) =
            withDTreeContext $ \dtc -> check dtc vars
            where
            check dtc [] = dLeaf $ Optimum (opt,vars)
            check dtc (v:vs)
                | signumEq (dtcContextImpliesSignum dtc (v - vFloored))  =
                    check dtc vs
                | otherwise =
                   trace ("newSys1 = " ++ show newSys1) $
                   bac newSys2
                                
--                    compDTreeG select $
--                    prod2DTreeG (bac newSys1) (bac newSys2)
                    
                where
                n = nV - length vs - 1
                vFloored   = floorFn v
                vNegCeiled = {- negate $ negate $ -} floorFn $ negate v
                
                ineq1 = replicate n 0 ++ [-1] ++ replicate (nV-n-1) 0 ++
                        [vFloored]
                ineq2 = replicate n 0 ++ [1] ++ replicate (nV-n-1) 0 ++
                        [vNegCeiled]

                newSys1 = sys `join` ineqSystem ((nV,0), matrix [ineq1])
                newSys2 = sys `join` ineqSystem ((nV,0), matrix [ineq2])

            -- Unbounded should never happen
            select (o@(Optimum _), Nonexistent) = dLeaf o
            select (Nonexistent, o@(Optimum _)) = dLeaf o
            select (o1@(Optimum (opt1,_)), o2@(Optimum (opt2,_))) =
                geCond (opt2-opt1) (dLeaf o2) (dLeaf o1)
            select _ = error "branchAndCut: Unbounded in select?"

        splex sys = simplex simplexParams sys mbEqs target


------------------------------
-- Combinator based Simplex --
------------------------------

data SimplexParams = SimplexParams { simplexParamOptimisticGE :: Bool }
		     deriving (Eq, Show)

stdSimplexParams = SimplexParams { simplexParamOptimisticGE = False }

--
-- columns:
--  coefficients of variables (#vars many)
--  1 constant
--
data Tableau a = Tableau (GVector a) [(GVector a, Int)]
		 deriving Show

simplex :: (Fractional a, Num cond)
        => SimplexParams
	-> GIneqSystem a
        -> Maybe (GEqSystem a)
	-> GVector a
	-> DTreeGen a cond (Optimum (a,[a]))
simplex = posSimplex2simplex (performSimplex2positiveSimplex performSimplex)

posSimplex2simplex :: (Fractional a, Num cond)
		   => (SimplexParams ->
		       GIneqSystem a ->
		       GVector a ->
		       DTreeGen a cond (Optimum (a,[a]))
		      )
		   -> SimplexParams
		   -> GIneqSystem a
		   -> Maybe (GEqSystem a)
		   -> GVector a
		   -> DTreeGen a cond (Optimum (a,[a]))
posSimplex2simplex posSimplex params ineqs mbEqs target =
--    trace ("posSimplex2simplex: " ++ show (ineqs,target,newIneqs,newTarget)) $
    withDTreeContext ps2s
    where
    ps2s dtc =
        mapDTreeG (unEq . resubLastVar) $
        posSimplex params newIneqs newTarget
        where
        newTarget = insertEntries target' (vectorLength target')
		     [negate $ sum $ vectorEntries target']
        newIneqs = ineqSystemVPC (insertColumns v (nColumns v) $ negate $
			      columnMatrix $ map sum (rows' v), p, c)

        (ineqs', target', targetOff, resubAF) = integrateEqs isNeq mbEqs ineqs target
        (v,p,c) = partsVPC ineqs'
        isNeq = signumNeq . dtcContextImpliesSignum dtc

        unEq Unbounded   = Unbounded
        unEq Nonexistent = Nonexistent
        unEq (Optimum (opt,vars)) =
            Optimum (opt + targetOff, vectorEntries vars')
            where
            vars' = resubAF `applyIH` vector vars

    resubLastVar Unbounded            = Unbounded
    resubLastVar Nonexistent          = Nonexistent
    resubLastVar (Optimum (opt,vars)) =
	Optimum (opt, map (subtract xneg) $ init vars)
	where
	xneg = last vars



    


performSimplex2positiveSimplex :: (Fractional a, Num cond)
			       => (SimplexParams
				   -> Set [Int]
				   -> Bool
				   -> Tableau a
				   -> DTreeGen a cond (Tableau a, Maybe a)
				  )
			       -> SimplexParams
			       -> GIneqSystem a
			       -> GVector a
			       -> DTreeGen a cond (Optimum (a,[a]))
performSimplex2positiveSimplex perfSimpl params ineqs target =
    -- trace ("pS2pS: " ++ show (ineqs',target',resubAF)) $
    compDTreeG performS $ cFeasibleDictionary params ineqs target
    where
    performS (Optimum tab) =
	    mapDTreeG (\r -> case r of
	                       (_,       Nothing) -> Unbounded
	                       (lastTab, Just x ) ->
		                   Optimum (x, optPoint lastTab)
	              ) $
		      perfSimpl params Set.empty False tab
    performS Unbounded   = dLeaf Unbounded
    performS Nonexistent = dLeaf Nonexistent

    optPoint (Tableau _ eqs) = map value [0 .. nVars ineqs - 1]
	    where
	    value varNr = case lookup varNr (map (\(a,b)->(b,a)) eqs) of
			    Just row -> vectorEntry row (vectorLength row - 1)
			    Nothing  -> 0



cFeasibleDictionary :: (Fractional a, Num cond)
		    => SimplexParams
		    -> GIneqSystem a
		    -> GVector a
		    -> DTreeGen a cond (Optimum (Tableau a))
cFeasibleDictionary params ineqs target =
--    debugTrace (show (nColumns ineqs, vectorLength target)) $
    assert (nColumns ineqs == vectorLength target + 1) $
--    debugTrace ("fD: " ++ show wTableau) $
    case wEquations of
      []         -> dLeaf $ Optimum $ initSimplex ineqs target
      (e1,_):eqs -> makeInitialLeave 1 (vectorEntry e1 constCol,0) eqs
    where
    makeInitialLeave _   (b,brow) [] =
	geleCond b
                 (dLeaf $ Optimum $ initSimplex ineqs target) -- origin is feasible
	         (compDTreeG mkDict $ performSimplex params Set.empty True $
		  enterLeave wTableau 0 brow)
    makeInitialLeave row (b,brow) ((e,_):eqs) =
	geCond (b'-b)
	         (makeInitialLeave (row+1) (b,brow) eqs)
	         (makeInitialLeave (row+1) (b',row) eqs)
	where
	b' = vectorEntry e constCol

    mkDict (Tableau _ eqs, Just _)
	| 0 /= head [bv | (_,bv) <- eqs] =     -- x0 is not basic
	    (dLeaf $ Optimum $ Tableau newTarg newEqs)
	where
	newEqs = map (\(v,bv) -> if bv == 0 then error "feasibleDictionary: x0 is basic? (internal error)"
		                            else (deleteEntries v 0 1,bv-1)
		     ) eqs
	newTarg = foldl1 (+) $ map look [0..nV-1]
	look v = (case lookup v (map (\(c,d)->(d,c)) newEqs) of
		    Nothing -> vector (replicate v 0 ++ 1 : replicate (nV-v+nR) 0)
		    Just e  -> negate $ setVectorEntry e v 0
		 ) `mulVS` vectorEntry target v
    mkDict _       = dLeaf Nonexistent

    wTableau@(Tableau wTarget wEquations) =
	initSimplex (insertVars ineqs 0 ones) w
    w = vector (-1 : replicate nV 0)
    ones = columnMatrix $ replicate nR 1
    nV = fromIntegral $ nVars ineqs
    nR = fromIntegral $ nRows ineqs
    constCol = nV + 1 + nR

-- create an initial tableau
initSimplex :: Fractional a
	    => GIneqSystem a
	    -> GVector a
	    -> Tableau a
initSimplex ineqs target =
--    debugTrace (show ineqs ++ "\n" ++ show target) $
    assert (nColumns ineqs == vectorLength target + 1) $
    assert (nParams  ineqs == 0) $
    Tableau (target `conc` vector (replicate (fromIntegral nR+1) 0))
	    (zip (rowVectors tableau) [fromIntegral (nVars ineqs)..])
    where
    (vs, _, consts) = partsVPC ineqs
    tableau = (negate vs `conc` unitMatrix nR nR `conc` consts)
	      
    nR = nRows ineqs
    nV = nVars ineqs

--
-- equations _must_ be sorted (ascending) by the index
-- of its basic vars !
-- Leaving and entering variables are chosen such that their
-- index is minimal among all possible choiches!
--
-- optimisticGE == True  can lead to non-termination of the Simplex.
-- You should know what you are doing if you set this to True.
--
performSimplex :: (Fractional a, Num cond)
	       => SimplexParams
	       -> Set [Int]
               -> Bool
	       -> Tableau a
	       -> DTreeGen a cond (Tableau a, Maybe a)
performSimplex params@(SimplexParams {simplexParamOptimisticGE=optimisticGE})
	       seenBasic auxiliaryStop tableau@(Tableau target equations)  =
--    trace (show $ Set.toList seenBasic) $
--    debugTrace (show (cardinality seenBasic) ++ "  " ++ show auxiliaryStop) $
--    debugTrace (show tableau) $
    findPivotCol 0
    where
    findPivotCol pivotCol
	| pivotCol == nV   =
	    dLeaf (tableau, Just $ negate $ vectorEntry target nV)
	| otherwise =
	    geCond (negate $ vectorEntry target pivotCol)
		   (findPivotCol (pivotCol+1))
	           (findPivotRow equations 0 Nothing)
	where
	findPivotRow [] _ Nothing = dLeaf (tableau, Nothing)
	findPivotRow [] _ (Just (_,_,pivotRow))
	    | pivotCol == 0 && auxiliaryStop  =
		dLeaf (newTableau, Just $ negate $ vectorEntry nT nV)
		-- x0 left basis
	    | otherwise  =
		performSimplex params newSeenBasic auxiliaryStop newTableau
	    where
	    newTableau@(Tableau nT nEqs) = enterLeave tableau pivotCol pivotRow
	    newSeenBasic = basicVars `Set.insert` seenBasic
	findPivotRow ((eq',_):eqs) row blub =
	    geCond (negate r')
		   (findPivotRow eqs (row+1) blub)
		   (case blub of
		      Nothing   -> findPivotRow eqs (row+1) (Just (r',s',row))
		      Just (r,s,_) ->
--		          (if optimisticGE then trace ("optimistic: " ++ show (s'/r') ++ " - " ++ show (s/r) ++ " = " ++ show (s'/r'-s/r)) else id) $
		          geCond' (optimisticGE && not inCycle)
		                  (s'/r' - s/r)
			          (findPivotRow eqs (row+1) blub)
		                  (findPivotRow eqs (row+1) (Just (r',s',row)))
		   )
	    where
	    r' = vectorEntry eq' pivotCol
	    s' = vectorEntry eq' constCol
    nV = fromIntegral $ vectorLength target - 1
    constCol  = nV
    inCycle = basicVars `Set.member` seenBasic
    basicVars = [basic | (_,basic) <- equations]

{-# NOTINLINE pivotingCounter #-}
pivotingCounter :: IORef Int
pivotingCounter = unsafePerformIO $ newIORef 0

resetPivotingCounter :: IO ()
resetPivotingCounter = writeIORef pivotingCounter 0

getPivotingCounter :: IO Int
getPivotingCounter = readIORef pivotingCounter

enterLeave :: Fractional a
           => Tableau a
	   -> Int
	   -> Int
	   -> Tableau a
enterLeave (Tableau target equations) pivotCol pivotRow =
--    trace ("enterLeave") $
    unsafePerformIO (modifyIORef pivotingCounter (+1) >>
		     return newTableau)
    where
    newTableau = Tableau newTarget newEquations
    (ups, eq@(eq',_):eqs) = splitAt pivotRow equations

    pivot = vectorEntry eq' pivotCol
    newTarget = subst' target
    newEquations =
	insertBy (\(_,bv1) (_,bv2) -> compare bv1 bv2) newEq $
	map subst (ups ++ eqs)
    newEq' = eq' `mulVS` recip pivot
    newEq  = (newEq', pivotCol)
    
    subst  (row,bv) = (subst' row, bv)
    subst' row = row - (vectorEntry row pivotCol `mulSV` newEq')



integrateEqs :: Fractional a
	     => (a -> Bool)
             -> Maybe (GEqSystem a)
	     -> GIneqSystem a
	     -> GVector a
	     -> (GIneqSystem a,     -- new inequalities
                 GVector a,         -- new target function
                 a,                 -- delta between new and old target function
                 GAffineFunction a  -- resub function (map new vars to old vars)
                )
integrateEqs isNeq Nothing    ineqs target =
    (ineqs, target, 0,
     affineFunction ((nV,0),unitMatrix nV (nV+1)) Inhomogeneous)
    where
    nV = nVars ineqs
integrateEqs isNeq (Just eqs) ineqs target =
    (ineqSystem ((newVars,0),
		 rowVectorsToMatrix (newVars+1) (is++es++(map negate es))),
     deleteEntries t (vectorLength t-1) 1,
     vectorEntry t (vectorLength t-1),
     resubAF
    )
    where
    newVars = nV-nRows eqs+genericLength es
    (t:is, es, resubAF) =
	staticSubstitute isNeq (target0 : rowVectors (getMatrix ineqs)) eqs
    nV = nVars ineqs
    constCol = nV
    target0 = insertEntries target constCol [0]

staticSubstitute :: Fractional a
		 => (a -> Bool)           -- ^ @\/=0@ oracle
		 -> [GVector a]           -- ^ original affine expressions
                 -> GEqSystem a           -- ^ equations to substitute
	         -> ([GVector a],         -- resulting expresions
		     [GVector a],         -- unused equations
		     GAffineFunction a)   -- resubstitution function
staticSubstitute isNeq affines eqs =
    sub affines [] (rowVectors $ unitMatrix nV (nV+1))
        (rowVectors $ getMatrix eqs)
    where
    sub afs unused resub [] =
	(afs, unused, affineFunction ((nV', 0), matr) Inhomogeneous)
	where
	nV' = nV - numberOfEquations eqs + genericLength unused
	matr = rowVectorsToMatrix (nV'+1) resub
    sub afs unused resub (e:es) =
	case findNeCoeff e of
	  Nothing  -> sub afs (e:unused) resub es
	  Just idx -> sub (map (substitute e idx) afs)
                          (map (substitute e idx) unused)
			  (map (substitute e idx) resub)
                          (map (substitute e idx) es)

    findNeCoeff e = findIndex isNeq $ init $ vectorEntries e

    substitute e idx aff = deleteEntries (aff + e') idx 1
	where
	e' = (negate $ vectorEntry aff idx / vectorEntry e idx) `mulSV` e

    nV = nVars eqs
