-- |
-- Interface to the various tools for sorting inequality systems.
--
-- The following suffixes on the function names are used:
--
--   * a prime ' means that the function is parametrised 
--     by a tool or sorting function.
--
--   * @G@ denotes a general function which works using 'DTreeGen'.
--
--   * @P@ denotes a specialised function for 'Polynomial' or 'FracPoly'.
--
--   * @ss@ means that the function returns a 'SortedSystem' instead of
--     an 'IneqSystem'.
--
module LMath.Algo.Sort
    ( -- * Sorting and Projecting for Rational Coefficients
      SortTool(..),
      sortIneqs, sortIneqs', projectToVars, projectToVars',

      -- * Sorting and Projecting for 'DTreeGen'
      sortIneqsG, sortIneqsGss, projectToVarsG, projectToVarsG',
      projectToVarsGss, projectToVarsGss',

      -- * Projecting for FracPoly Coefficients
      projectToVarsPss',

      -- * Other sorting related functions
      enumeratePolytope
    )
where

import Data.Ratio
import Data.Maybe

import LMath.Algo.Chernikova
import LMath.Algo.FM
import LMath.Types

-- |
-- Select a sorting tool.
--
data SortTool = CSideFM   -- ^ C side Fourier-Motzkin
              | CSidePip  -- ^ C side PIP
              | HsFM      -- ^ Haskell Fourier-Motzkin which respects
                          --   conditions on the parameters
              | HsChernikova  -- ^ Haskell implementation of Chernikova's
                              --   algorithm
                deriving (Eq, Show)

sortIneqs :: (Fractional a, Real a)
	  => GIneqSystem a
	  -> Maybe (GEqSystem a)
	  -> Maybe (GIneqSystem a)
sortIneqs = sortIneqs' HsFM
            -- sortIneqs' CSidePip
sortIneqs' :: (Fractional a, Real a)
	   => SortTool
           -> GIneqSystem a
           -> Maybe (GEqSystem a)
	   -> Maybe (GIneqSystem a)
sortIneqs' CSideFM  sys mbEqs =
    csideSort FmMethod sys mbEqs
sortIneqs' CSidePip sys mbEqs =
    csideSort PipMethod sys mbEqs
sortIneqs' HsFM sys mbEqs = sortIneqsFM sys mbEqs
sortIneqs' HsChernikova sys mbEqs = chernikovaSortIneqs sys mbEqs

csideSort :: (Fractional a, Real a)
	  => ToolMethod
          -> GIneqSystem a
	  -> Maybe (GEqSystem a)
	  -> Maybe (GIneqSystem a)
csideSort tm ineqs mbEqs =
    fmap (fmap fromRational) $
    csideSortIneqs (fmap toRational (addEqs ineqs mbEqs))
                   tm

-- |
-- Project a given system to some variables using
-- the default projection tool (currently @HsFM@).
--
projectToVars :: (Fractional a, Real a, Integral b)
	      => GIneqSystem a
              -> Maybe (GEqSystem a)
	      -> [b]
	      -> Maybe (GIneqSystem a)
projectToVars = projectToVars' HsFM

-- |
-- Project a given system to some variables using the
-- specified tool for sorting the system.
-- If the specified tool respects conditions on the parameters
-- (i.e., it outputs some inequalities with only parameters),
-- the result of the projection will also contain the conditions
-- on the parameters.
--
projectToVars' :: (Fractional a, Real a, Integral b)
	       => SortTool
               -> GIneqSystem a
               -> Maybe (GEqSystem a)
	       -> [b]
	       -> Maybe (GIneqSystem a)
projectToVars' tool ineqs' mbEqs' vars =
    case sortIneqs' tool ineqs mbEqs of
       Nothing  -> Nothing
       Just sys -> Just $ deleteVars (filterIneqs filt sys) projV dropV
    where
    ineqs = reorderVars ineqs' vars
    mbEqs = fmap (\eqs -> reorderVars eqs vars) mbEqs'

    projV = length vars
    dropV = nVars ineqs - projV
    filt coeffs = all (==0) $ take dropV $ drop projV coeffs


addEqs :: Num a => GIneqSystem a -> Maybe (GEqSystem a) -> GIneqSystem a
addEqs ineqs Nothing = ineqs
addEqs ineqs (Just eqs) =
    multijoin ineqs [fromAffineExp $ toAffineExp eqs,
		     fromAffineExp $ fmap negate $ toAffineExp eqs
		    ]

-- |
-- Project an inequality system to given dimensions.
-- @projectToVarsPss'@ takes a sorting function
-- (like @sortIneqsPss@ in LMath.Logic.QE),
-- a context, the list of variables to project onto, and
-- an inequality system (with optional equalities) as input. It then
-- computes the projection of the input system (as 'SortedSystem') to the
-- variables listed in
-- the given variable list where the first element of the list is the
-- outermoust dimension in the projection.
-- 
projectToVarsPss' :: (Integral a, Integral b)
	        => ([QfFormula (Polynomial (Ratio a))] ->
		    GIneqSystem (FracPoly (Ratio a)) ->
		    Maybe (GEqSystem (FracPoly (Ratio a))) ->
		    Maybe b ->
		    DTree (Polynomial (Ratio a))
		          (Maybe (SortedSystem (FracPoly (Ratio a))))
		   )  -- ^ sorting function
		-> [QfFormula (Polynomial (Ratio a))]  -- ^ context
		-> GIneqSystem (FracPoly (Ratio a))    -- ^ input inequality system
		-> Maybe (GEqSystem (FracPoly (Ratio a)))  -- ^ optional input equality system
		-> Maybe b                       -- ^ number of "outer" variables
		-> [b]                           -- ^ variables to project onto
		-> DTree (Polynomial (Ratio a))
		         (Maybe (SortedSystem (FracPoly (Ratio a))))
projectToVarsPss' sortFn ctx ineqs mbEqs mbOuterVars vars =
    fmap (fmap $ take (length vars - outers)) $
    sortFn ctx ineqs' mbEqs' mbOuterVars
    where
    ineqs' = reorderVars ineqs vars
    mbEqs' = fmap (\es -> reorderVars es vars) mbEqs
    outers = fromIntegral $ fromMaybe 0 mbOuterVars


--
-- number of parameters can be >0.
--
sortIneqsG :: (Fractional a, Integral b, Num cond)
             => GIneqSystem a
             -> Maybe (GEqSystem a)
	     -> Maybe b     -- ^ number of "outer" variables
	     -> DTreeGen a cond (Maybe (GIneqSystem a))
sortIneqsG ineqs mbEqs mbCtxVars =
    fmap (fmap toIneqs) $
    sortIneqsGss ineqs' mbEqs' mbCtxVars
    where
    toIneqs ss = moveVarsToParams (sortedSystemToIneqSystem (nV+nP) 0 ss)
                                  0 nP 0
    ineqs' = moveParamsToVars ineqs 0 nP 0
    mbEqs' = fmap (\es -> moveParamsToVars es 0 nP 0) mbEqs

    nV = nVars ineqs
    nP = nParams ineqs

--
-- number of parameters can be >0.
--
sortIneqsGss :: (Fractional a, Integral b, Num cond)
             => GIneqSystem a
             -> Maybe (GEqSystem a)
	     -> Maybe b
	     -> DTreeGen a cond (Maybe (SortedSystem a))
sortIneqsGss = fourierMotzkin


projectToVarsG :: (Fractional a, Integral b, Num cond)
	       => GIneqSystem a
	       -> Maybe (GEqSystem a)
	       -> Maybe b
	       -> [b]
	       -> DTreeGen a cond (Maybe (GIneqSystem a))
projectToVarsG = projectToVarsG' sortIneqsG

projectToVarsG' :: (Fractional a, Integral b, Num cond)
                => ( GIneqSystem a ->
		     Maybe (GEqSystem a) ->
		     Maybe b ->
		     DTreeGen a cond (Maybe (GIneqSystem a))
		   ) 			-- ^ sorting function
	        -> GIneqSystem a    	-- ^ input inequality system
	        -> Maybe (GEqSystem a)  -- ^ optional input equality system
		-> Maybe b              -- ^ number of "outer" variables
		-> [b] 		        -- ^ variables to project onto
		-> DTreeGen a cond (Maybe (GIneqSystem a))
projectToVarsG' sortFn ineqs' mbEqs' mbOuterVars vars =
    do mbIs <- sortFn ineqs mbEqs mbOuterVars
       return $ fmap (\res -> deleteVars (filterIneqs filt res) projV dropV) mbIs
    where
    ineqs = reorderVars ineqs' vars
    mbEqs = fmap (\eqs -> reorderVars eqs vars) mbEqs'

    projV = length vars
    dropV = nVars ineqs - projV
    filt coeffs = all (==0) $ take dropV $ drop projV coeffs

projectToVarsGss :: (Fractional a, Integral b, Num cond)
	         => GIneqSystem a    	 -- ^ input inequality system
	         -> Maybe (GEqSystem a)  -- ^ optional input equality system
		 -> Maybe b              -- ^ number of "outer" variables
		 -> [b] 		 -- ^ variables to project onto
		 -> DTreeGen a cond (Maybe (SortedSystem a))
projectToVarsGss = projectToVarsGss' sortIneqsGss

-- |
-- Project an inequality system to given dimensions.
-- @projectToVarsGss'@ takes a sorting function (like 'sortIneqsGss'),
-- the list of variables to project onto, the number of \"outer\" variables,
-- and an inequality system (with optional equalities) as input. It then
-- computes the projection of the input system (as 'SortedSystem') to the
-- variables listed in
-- the given variable list where the first element of the list is the
-- outermoust dimension in the projection.
-- 
projectToVarsGss' :: (Fractional a, Integral b, Num cond)
                => ( GIneqSystem a ->
		     Maybe (GEqSystem a) ->
		     Maybe b ->
		     DTreeGen a cond (Maybe (SortedSystem a))
		   ) 			-- ^ sorting function
	        -> GIneqSystem a    	-- ^ input inequality system
	        -> Maybe (GEqSystem a)  -- ^ optional input equality system
		-> Maybe b              -- ^ number of "outer" variables
		-> [b] 		        -- ^ variables to project onto
		-> DTreeGen a cond (Maybe (SortedSystem a))
projectToVarsGss' sortFn ineqs mbEqs mbOuterVars vars =
    mapDTreeG (fmap $ take (length vars - outers)) $
    sortFn ineqs' mbEqs' mbOuterVars
    where
    ineqs' = reorderVars ineqs vars
    mbEqs' = fmap (\es -> reorderVars es vars) mbEqs
    outers = fromIntegral $ fromMaybe 0 mbOuterVars


-- |
-- Enumerate the integral points contained in the
-- given polytope for the given values of the parameters.
--
enumeratePolytope :: (RealFrac a, Integral b)
                  => [b]           -- ^ parameter values
                  -> GIneqSystem a -- ^ polytope
                  -> [[b]]
enumeratePolytope paramVals sys
    | paramsOk   =  en 0 [[]]
    | otherwise  =  []
    where
    en n xs
           | n == nV    =  xs
           | otherwise  =
               en (n+1) [x++[y] | x <- xs, y <- [ceiling $ maxi (map (subst x) $ lower n) ..
                                                 floor $ mini (map (subst x) $ upper n)]]

    subst x ys = ys `mulVV` vector (map fromIntegral (x ++ paramVals ++ [1]))

    maxi [] = error "enumeratePolytope: not a polytope (unbounded!)"
    maxi xs = maximum xs
    mini [] = error "enumeratePolytope: not a polytope (unbounded!)"
    mini xs = minimum xs

    Just sorted = sortIneqs sys Nothing
    lower n = rowVectors $
                 fmap negate $ getMatrix $
                 deleteVars (filterIneqs (\i -> part n i && i!!n>0) sorted) n (nV-n)
    upper n = rowVectors $ getMatrix $
                  deleteVars (filterIneqs (\i -> part n i && i!!n<0) sorted) n (nV-n)
    part n i = all (==0) $ take (nV-n-1) $ drop (n+1) i
    nV = nVars sys

    -- check whether the parameters satisfy the conditions 
    -- on the parameters (implicitly) contained in the given system
    paramsOk =
        all (>= 0) [mulVV r paramValsV | r <- rows]
        where
        rows = rowVectors $ getMatrix $
               deleteVars (filterIneqs (all (==0) . take nV) sorted) 0 nV
        paramValsV = vector (map fromIntegral paramVals ++ [1])
