-- |
-- Topological sorting algorithms.
--
module LMath.Algo.TopSort
    ( -- * topological sorting algorithms (for varying input formats)
      topSortGraph, topSort, topSort', topSortByPartialOrder,

      -- * graph constructors
      Graph (), mkGraph, mkGraphFromPartialOrder,

      -- * additional functions which are useful for debugging
      createPartialOrder, isTopSorted
    )
where

import Data.Array
import Data.List
import qualified Data.Map as Map
    (empty, fromAscList, insertWith, findWithDefault,
     mapWithKey, toList, findMax, delete, elems
    )
import Data.Map (Map)
import Debug.Trace

-- |
-- A graph which contains a list of outgoing edges for each node.
--
-- See also: 'mkGraph', 'mkGraphFromPartialOrder'
--
newtype Graph = Graph (Map Int [Int])
                deriving Show

-- |
-- Counts all incoming edges for a specific node.
-- This data structure is used by the algorithm to avoid
-- scanning through the whole graph, which is too expensive.
--
-- See also: 'initIncomingEdgeCounter'
--
newtype IncomingEdgesCounter = IncomingEdgesCounter (Map Int Int)
                               deriving Show

-- |
-- Creates a graph from a list of edges.
-- The list of edges must not contains loops, i.e. edges
-- with the same begin and end node.
--
-- See also: 'mkGraphFromPartialOrder'
--
mkGraph :: [(Int, Int)] -- ^ a list of all edges @[(fromNode, toNode)]@
        -> Graph
mkGraph [] = Graph Map.empty
mkGraph xs
    | any (uncurry (==)) xs =
	error "mkGraph: input contains loops, i.e. an edge from n to n"
    | otherwise = let addEdge mp (x,y) = Map.insertWith append x [y] mp
		      append [y] ys = y:ys
		      fastNub = unique . sort
		      unique (x1:x2:xs) | x1 == x2  = unique (x2:xs)
					| otherwise = x1 : unique (x2:xs)
		      unique xs = xs
                  in Graph $ fmap fastNub $ foldl addEdge Map.empty xs

-- |
-- Creates a graph from a given partial order and a list of elements.
--
-- Warning: This function is quite expensive:
-- @O(n^2)@ calls of the partial order functions, where @n@ is the
-- number of elements.
--
-- See also: 'mkGraph'
--
mkGraphFromPartialOrder :: (a -> a -> Maybe Bool) -- ^ the partial order
			-> [a]                    -- ^ a list of elements
			-> Graph
mkGraphFromPartialOrder order xs = mkGraph edges
    where
    xs' = zip xs [0..]
    edges = [(n1, n2) | (x1, n1) <- xs', (x2, n2) <- xs', n1 /= n2,
	                order x1 x2 == Just True]

-- |
-- Returns a list of all outgoing edges of node @n@, i.e. all
-- nodes @x@ for which exists an edge @(n, x)@.
--
edges :: Int   -- ^ the node @n@
      -> Graph -- ^ the graph
      -> [Int] -- ^ the ID of all nodes which can be directly reached from @n@
edges n (Graph mp) = Map.findWithDefault [] n mp

-- |
-- Returns a list of all incoming edges of node @n@, i.e. all
-- nodes @x@ for which exists an edge @(x, n)@.
--
incomingEdges :: Int   -- ^ the node @n@
	      -> Graph -- ^ the graph
	      -> [Int] -- ^ the ID of all nodes which can reach @n@ directly
incomingEdges n (Graph mp) = [x | (x,y) <- Map.toList mp, n `elem` y]

-- |
-- Looks up in the incoming edge counter whether
-- the node @n@ has incoming edges.
--
hasIncomingEdges :: IncomingEdgesCounter -- ^ the incoming edge counter
		 -> Int                  -- ^ the node @n@
		 -> Bool
hasIncomingEdges (IncomingEdgesCounter mp) n = (Map.findWithDefault 0 n mp) == 0

-- |
-- Takes a list of nodes which must not have incoming edges
-- and remove these nodes with their outgoing edges from the graph.
-- The function will return the new graph and the updated
-- incoming edge counter.
--
removeNodesWithoutIncomingEdges :: [Int] -- ^ the node which should be tested
                                -> Graph -- ^ the graph
	                        -> IncomingEdgesCounter
				-- ^ the incoming edge counter
                                -> (Graph, IncomingEdgesCounter)
removeNodesWithoutIncomingEdges nodes graph inCounter = (newGraph, newInCounter)
    where
    (Graph gMP) = graph
    (IncomingEdgesCounter incMP) = inCounter

    newGraph = Graph $ foldl (flip Map.delete) gMP nodes
    missingInEdges = concat [edges n graph | n <- nodes]
    countMissing = accumArray (+) 0 (0, fst $ Map.findMax incMP) $
		   [(n,1) | n <- missingInEdges]
    newInCounter = IncomingEdgesCounter $
		   Map.mapWithKey (\n c -> c - (countMissing ! n)) incMP

-- |
-- Initializes the incoming edge counter for the given graph.
--
initIncomingEdgeCounter :: Graph -- ^ the graph
		        -> Int   -- ^ the number of nodes
		        -> IncomingEdgesCounter
initIncomingEdgeCounter g numNodes =
    IncomingEdgesCounter $
    Map.fromAscList [(n, length $ incomingEdges n g) | n <- [0..numNodes-1]]

-- |
-- Constructs a function which represents the partial order,
-- which is given as a list of edge of a directed acyclic graph (DAG).
--
-- Warning: The resulting function is not very efficient.
--
createPartialOrder :: (Eq a)
                   => [(a, a)]
		   -- ^ the partial order represented as edges of a DAG
		   -> (a -> a -> Maybe Bool)
		   -- ^ the partial order represented as a function
createPartialOrder facts x y
    | (x,y) `elem` facts = Just True
    | (y,x) `elem` facts = Just False
    | otherwise          = Nothing

-- |
-- Computes a topological sorting for the given graph.
-- If the graph contains cycles, 'Nothing' will be returned instead.
--
topSortGraph :: Graph       -- ^ the graph
	     -> Int         -- ^ the number of nodes
	     -> Maybe [Int] -- ^ the resulting linear order
topSortGraph _ 0 = --trace "topSortGraph: empty graph" $
		   Just []
topSortGraph initGraph l = --trace (let (Graph mp, edges) = (initGraph, length $ concat $ Map.elems mp) in "topSortGraph: graph with " ++ show l ++ " nodes and " ++ show edges ++ " edges found.") $
			   recSort initGraph initInCounter [0..l-1]
    where
    initInCounter = initIncomingEdgeCounter initGraph l

    recSort _ _ [] = Just []
    recSort graph inCounter remNodes =
	let (noInEdges, withInEdges) =
		partition (hasIncomingEdges inCounter) remNodes
	    (newGraph, newInCounter) =
		removeNodesWithoutIncomingEdges noInEdges graph inCounter
        in if null noInEdges
	   then Nothing -- cycle found!
	   else case recSort newGraph newInCounter withInEdges of
		  Just xs -> Just (noInEdges ++ xs)
		  Nothing -> Nothing -- input contains cycles!

-- |
-- Computes a topological sorting for the given graph,
-- which is represented as a list of nodes.
-- If the graph contains cycles, 'Nothing' will be returned instead.
--
-- The input edges must not contain loops, i.e. edges
-- with the same begin and end node.
--
topSort :: [(Int, Int)] -- ^ a list of all edges @[(fromNode, toNode)]@
	-> Int          -- ^ the number of nodes
	-> Maybe [Int]  -- ^ the resulting linear order
topSort = topSortGraph . mkGraph

-- |
-- Computes a topological sorting for the given graph,
-- which is represented as a list of nodes.
-- If the graph contains cycles, 'Nothing' will be returned instead.
--
-- The input edges must not contain loops, i.e. edges
-- with the same begin and end node.
--
-- Warning:
-- This version should be less efficient than 'topSort'.
-- Do not use it to topologically sort larger graphs.
--
-- See also: 'topSort'
--
topSort' :: (a -> a -> Bool) -- ^ the test for equality
         -> [(a, a)]         -- ^ the list of all edges
         -> Maybe [a]        -- ^ the resulting linear order
topSort' eq order = fmap (map (\x -> intToA ! x)) $
                    topSort [(aToInt x, aToInt y) | (x,y) <- order] l
    where
    l   = length xs
    xs  = nubBy eq $ concat [[x,y] | (x,y) <- order]
    aToInt y = head [n | (x,n) <- zip xs [0..], x `eq` y]
    intToA = listArray (0, l-1) xs

-- |
-- Computes a topological sorting for the given graph,
-- which is represented as a list of nodes.
-- If the graph contains cycles, 'Nothing' will be returned instead.
--
topSortByPartialOrder :: (a -> a -> Maybe Bool) -- ^ the partial order
		      -> [a]                    -- ^ the elements to be sorted
		      -> Maybe [a]              -- ^ the resulting linear order
topSortByPartialOrder order xs =
    fmap (map (\n -> xs' ! n)) $
    topSortGraph (mkGraphFromPartialOrder order xs) (length xs)
    where
    xs' = listArray (0, length xs - 1) xs

-- |
-- Checks whether the linear order @xs@ is consistent
-- with the given partial order.
--
isTopSorted :: (a -> a -> Maybe Bool) -- ^ the partial order
            -> [a]                    -- ^ the linear order @xs@
            -> Bool                   -- ^ @True@ iff the order is legal
isTopSorted _ [] = True
isTopSorted order (x:xs) = all (\y -> order x y /= Just False) xs &&
			   isTopSorted order xs
