-- |
-- Representation of and operations on Z-polyhedra.
--
module LMath.Algo.ZPolyhedron
  ( -- * Data type for Z-polyhedra
    ZPolyhedron(..), ZDomain,
    zpolyhedron, zpFromIneqs, zpVars, zpParams, zpTargetDim,
    zpPreimage, zpLattice,
    zpWithLattice,
    zpIncludes,
    zpInsertZeroParams,
    zpRealImage, zpToIneqs,
    zpMirror, zdMirror, zpMirrorDim, zpMirrorDims,

    zpAddTargetConstraints, zpSubstParam,
    zpMoveTargetDimsToParams, zpMoveParamsToTargetDims,

    -- * Image under affine functions
    zpImage,

    -- * Difference, union, disjoint union
    zdDifference, zdIntersection, zdUnion,
    zdDisjointUnion, zdDisjointUnionDifference,
    zdTaggedDisjointUnion,

    -- * Counting and enumerating integral points
    zpCountIntegralPoints, zpCountIntegralPointsDisj,
    zpLexicographicIndex, zdLexicographicIndex,
    zpEnumerate
  )
where

import Control.Exception ( assert )
import Data.Ratio

import CppInterface
import LMath.Types
import LMath.Algo.AffineImage
import LMath.Algo.CountPolyhedra
import LMath.Algo.DisjointUnion
import LMath.Algo.OmegaTest
import LMath.Algo.Sort

-- Z-polyhedra implementation
import LMath.Algo.ZPolyhedronPL

import Debug.Trace

-- |
-- Return a Z-polyhedron which describes exactly the integral
-- points of a given polyhedron.
--
zpFromIneqs :: (Fractional a, Real a) => GIneqSystem a -> ZPolyhedron
zpFromIneqs ineqs = zpolyhedron ineqs af
    where
    af = affineFunction ((nV,nP), unitMatrix nV (nV+nP+1)) Inhomogeneous
    nV = nVars ineqs
    nP = nParams ineqs

zpFromIneqsDims :: (Fractional a, Real a, Integral b)
  => [b] -> GIneqSystem a -> ZPolyhedron
zpFromIneqsDims dims ineqs = zpolyhedron ineqs af
    where
    af = affineFunction ((nV,nP), m) Inhomogeneous
    m = multijoin (zeroMatrix 0 w) [unitRowVect w d | d <- dims]
    w = nV+nP+1
    nV = nVars ineqs
    nP = nParams ineqs


zpWithLattice :: (Real a)
  => (GAffineFunction a -> GAffineFunction a)
  -> ZPolyhedron
  -> ZPolyhedron
zpWithLattice f zp = zpolyhedron (zpPreimage zp) (f $ zpLattice zp)

-- |
-- Compute the image of a Z-polyhedron under an affine function.
--
zpImage :: Real a => GAffineFunction a -> ZPolyhedron -> ZPolyhedron
zpImage f zp = zpolyhedron (zpPreimage zp) (composeAffine f $ zpLattice zp)

zpInverseImage :: Real a => GAffineFunction a -> ZPolyhedron -> ZPolyhedron
zpInverseImage f zp = error "zpInverseImage not implemented"

zpMirror :: ZPolyhedron -> ZPolyhedron
zpMirror zp = zpolyhedron (zpPreimage zp) (afMirror $ zpLattice zp)

zdMirror :: ZDomain -> ZDomain
zdMirror = map zpMirror

zpMirrorDim :: (Integral b) => b -> ZPolyhedron -> ZPolyhedron
zpMirrorDim dim zp = zpolyhedron (zpPreimage zp) $ afMirrorDim dim $ zpLattice zp

zpMirrorDims :: (Integral b)
  => [b] -> ZPolyhedron -> ZPolyhedron
zpMirrorDims dims zp = zpolyhedron (zpPreimage zp) $ afMirrorDims dims $ zpLattice zp


zpMoveTargetDimsToParams :: (Integral b)
 => b
 -> b
 -> b
 -> ZPolyhedron
 -> ZPolyhedron
zpMoveTargetDimsToParams tdIdx count paramIdx zp =
    assert (tdIdx >= 0 && tdIdx+count <= tD) $
    assert (paramIdx >= 0 && paramIdx <= nP) $
    zpFromIneqsDims dims $ moveVarsToParams sys (nV+tdIdx) count paramIdx
    where
    sys = zpToIneqs zp
    dims = [nV .. nV+tD-1-count]
    tD = zpTargetDim zp
    nP = zpParams zp
    nV = zpVars zp

zpMoveParamsToTargetDims :: (Integral b)
 => b
 -> b
 -> b
 -> ZPolyhedron
 -> ZPolyhedron
zpMoveParamsToTargetDims paramIdx count tdIdx zp =
    assert (tdIdx >= 0 && tdIdx <= tD) $
    assert (paramIdx >= 0 && paramIdx+count <= nP) $
    zpFromIneqsDims dims $ moveParamsToVars sys paramIdx count (nV+tdIdx)
    where
    sys = zpToIneqs zp
    dims = [nV .. nV+tD-1+count]
    tD = zpTargetDim zp
    nP = zpParams zp
    nV = zpVars zp

zpSubstParam :: (Real a, Fractional a)
 => Integer
 -> GVector a
 -> ZPolyhedron
 -> ZPolyhedron
zpSubstParam idx expr zp =
    zpolyhedron (ineqSystemVPC (sSys,pSys',cSys'))
                (affineFunctionVPC (vAF,pAF',cAF') Inhomogeneous)
    where
    sys = zpPreimage zp
    af  = zpLattice zp
    (sSys, pSys, cSys) = partsVPC sys
    (vAF,pAF,cAF) = partsVPCih af

    (pSys',cSys') = subst pSys cSys
    (pAF', cAF' ) = subst pAF  cAF

    subst p c = (new_p, new_c)
        where
        pc_rows = rowVectors (p `conc` c)
        new_rows = rowVectorsToMatrix1 $ map substRow pc_rows
        new_p = subMatrixVer new_rows 0 (nP-1)
        new_c = subMatrixVer new_rows nP nP
    substRow r = setVectorEntry r idx 0 + (vectorEntry r idx `mulSV` expr)
    nP = nParams af

-- |
-- Write the Z-polyhedron as an inequality system with
-- dimensions for the preimage domain (leading variables)
-- and the image (trailing variables).
--
zpToIneqs :: Real a => ZPolyhedron -> GIneqSystem a
zpToIneqs zp = sys
    where
    is = zpPreimage zp
    af = zpLattice zp
    (isV,isP,isC) = partsVPC   is
    (afV,afP,afC) = partsVPCih af
    nV = nVars   is
    nP = nParams is
    nVt = nRows afV

    sys = ineqSystem ((nV+nVt,nP), m)
    m = multijoin1 $ map multiconc1
          [[ isV, z, isP, isC ]
          ,[ afV,-u, afP, afC ]
          ,[-afV, u,-afP,-afC ]
          ]
    u = unitMatrix nVt nVt
    z = zeroMatrix (nRows isV) nVt

-- |
-- Number of dimensions in the target of the Z-polyhedron.
--
zpTargetDim :: Integral b => ZPolyhedron -> b
zpTargetDim zp = targetDimIH $ zpLattice zp

-- |
-- Number of variables in the preimage of the Z-polyhedron.
--
zpVars :: Integral b => ZPolyhedron -> b
zpVars zp = nVars $ zpPreimage zp

-- |
-- Number of parameters of the Z-polyhedron.
--
zpParams :: Integral b => ZPolyhedron -> b
zpParams zp = nParams $ zpPreimage zp

-- |
-- Add constraints to the image domain of a Z-polyhedron.
--
zpAddTargetConstraints :: RealFrac a
                       => GIneqSystem a -> ZPolyhedron -> ZPolyhedron
zpAddTargetConstraints targetConstr zp =
    assert (nParams targetConstr == nParams af) $
    assert (nVars targetConstr == targetDimIH af) $
    zpolyhedron (simplifyIneqs Nothing (sys `join` constr)) af
    where
    sys = zpPreimage zp
    af  = zpLattice  zp
    constr = inverseImage af targetConstr

zpInsertZeroParams :: (Integral b)
                   => ZPolyhedron -> b -> b -> ZPolyhedron
zpInsertZeroParams zp pos count =
    zpolyhedron (insertZeroParams is pos count) (affineFunction ((nV,nP+count), m) Inhomogeneous)
    where
    is = zpPreimage zp
    af = zpLattice zp
    m = insertZeroColumns (getMatrixIH af) (nV+pos) count
    nV = nVars   af
    nP = nParams af



zdDisjointUnionDifference :: ZDomain -> ZDomain -> ZDomain
zdDisjointUnionDifference ins outs = zdDisjointUnion $ zdDifference ins outs

zdDisjointUnion :: ZDomain -> ZDomain
zdDisjointUnion zps =
    map fst $ zdTaggedDisjointUnion $ zip zps (repeat [] :: [[Int]])

zpTaggedRealImageDU :: [(ZPolyhedron,[tag])] -> [[(ZPolyhedron,[tag])]]
zpTaggedRealImageDU zpts = res
    where
    images = [(zpRealImage zp, zpLattice zp, tag) | (zp, tag) <- zpts]
    ds = disjUnionAll [i | (i,_,_) <- images]

    res = [ [ (zpolyhedron (simplifyIneqs Nothing $ inverseImage af sys2) af, tag) | (tsys,af,tag) <- images, let sys2 = tsys `join` d, not (areIneqsInfeasible sys2)] | d <- ds]

    nP = head [zpParams zp | (zp, _) <- zpts]

zdTaggedDisjointUnion :: (Ord tag, Show tag)
                      => [(ZPolyhedron, [tag])]
                      -> [(ZPolyhedron, [tag])]
zdTaggedDisjointUnion inputs =
    concatMap (genericDisjointUnion interDiffFn) $ zpTaggedRealImageDU inputs
    where
    interDiffFn a b
        | null inters  =  Nothing
        | not $ null $ zdIntersection inters aWithoutB  =
             error "1"
        | not $ null $ zdIntersection inters bWithoutA  =
             error "2"
        | not $ null $ zdIntersection aWithoutB bWithoutA =
             error "3"
        | not $ null $ zdIntersection aWithoutB inters =
             error "4"
        | not $ null $ zdIntersection bWithoutA inters  =
             error "5"
        | not $ null $ zdIntersection bWithoutA aWithoutB =
             error "6"
        | otherwise    =  Just (inters
                               ,aWithoutB
                               ,bWithoutA
                               )
        where
        inters = zdIntersection [a] [b]
        aWithoutB = zdDifference [a] [b]
        bWithoutA = zdDifference [b] [a]



zpCountIntegralPoints :: RealFrac a
                      => Maybe (GIneqSystem a)
                      -> ZPolyhedron
                      -> EPartition a
zpCountIntegralPoints mbCtx zp =
    countIntegralPoints [zpToIneqs zp] mbCtx

zpCountIntegralPointsDisj :: RealFrac a
                          => Maybe (GIneqSystem a)
                          -> ZPolyhedron
                          -> EPartition a
zpCountIntegralPointsDisj mbCtx zp =
    epartition $
    (\zs -> [(cond,ehps!!idx) | (cond,idx) <- zs]) $
    staggedDisjointUnion mbCtx $ zip conds [0..]
    where
    (conds,ehps) = unzip $ epPartitions $ zpCountIntegralPoints mbCtx zp


zpLexicographicIndex :: RealFrac a
    => Bool                   -- ^ assume reference point is inside
    -> Maybe (GIneqSystem a)  -- ^ optional context in the target dimensions and parameters
    -> ZPolyhedron          -- ^ the Z-polyhedron to count
    -> EPartition a
zpLexicographicIndex assumeInside mbCtx0 zp =
    zdLexicographicIndex assumeInside mbCtx0 [zp]

    
zdLexicographicIndex :: RealFrac a
    => Bool                   -- ^ assume reference point is inside
    -> Maybe (GIneqSystem a)  -- ^ optional context in the target dimensions and parameters
    -> ZDomain              -- ^ the Z-domain to count
    -> EPartition a
zdLexicographicIndex assumeInside mbCtx0 [] = epartition []
zdLexicographicIndex assumeInside mbCtx0 zd =
    lexicographicIndex assumeInside selectedDims mbCtx syss
    where
    syss = map zpToIneqs zd
    selectedDims = [sdims .. sdims+tdims-1]
    sdims = zpVars $ head zd
    tdims = zpTargetDim $ head zd
    mbCtx = fmap (\ctx -> insertZeroVars ctx 0 sdims) mbCtx0


-- |
-- Enumerate the integral points in a 'ZPolyhedron' in
-- lexicographic order for given values of the parameters.
--
zpEnumerate :: (Integral b)
            => [b]                  -- ^ parameter values
            -> ZPolyhedron        -- ^ the 'ZPolyhedron' to enumerate
            -> [[b]]
zpEnumerate paramValues zp =
    [c | c <- cands, all ((==1).denominator.toRational) $
                     vectorEntries $
                     applyIH invFn $
                     vector $
                     fmap fromIntegral (c++paramValues)
    ]
    where
    sys = zpPreimage zp
    af  = zpLattice zp
    invFn = pseudoInverse af 
    tsys  = inverseImage invFn sys
    cands = enumeratePolytope paramValues tsys



--
-- Test
--
xy2 = zpolyhedron (ineqSystem ((2,2), matrix [[ 1, 0, 0,  0, 0]
                                             ,[-1, 0, 1,  0, 0]
                                             ,[ 0, 1, 0,  0, 0]
                                             ,[ 0,-1, 0,1%2, 0]
                                             ]
                              )
                  )
                  (affineFunction ((2,2), matrix [[1, 0, 0, 0, 0]
                                                 ,[0, 2, 0, 0, 0]
                                                 ]
                                  )
                                  Inhomogeneous
                  )
