-- |
-- Z-polyhedra implemented via Polylib
--
module LMath.Algo.ZPolyhedronPL
  ( ZPolyhedron(..), ZDomain, zpolyhedron,
    zpPreimage, zpLattice, zpRealImage,
    zpIncludes, zdIntersection, zdDifference, zdUnion
  )
where

import CppInterface
import LMath.Types
import LMath.Algo.AffineImage
import LMath.Algo.OmegaTest

import Control.Exception ( assert )
import Data.Ratio
import Debug.Trace ( trace )

data ZPolyhedron = ZP IneqSystem (GAffineFunction Integer) deriving Show
type ZDomain = [ZPolyhedron]

zpolyhedron :: (Real a, Real b)
            => GIneqSystem a -> GAffineFunction b -> ZPolyhedron
zpolyhedron is_ af_ =
--    trace (show (is,is')) $
    assert (all ((==1).denominator) (entries af')) $ 
    assert (nVars is == nVars af) $
    assert (nParams is == nParams af) $
--    trace (show (is2, (keepColumns m indepIdxs, afP, afC))) $
    ZP is' $ fmap numerator af'
    where
    is = fmap toRational is_
    af = fmap toRational af_
    (isV,isP,isC) = partsVPC is
    (afV,afP,afC) = partsVPCih af
    nV  = nVars is
    nP  = nParams is
    nVt = nRows afV
    m = afV `conc` unitMatrix nVt nVt
    (_,_,indepIdxs) = gaussJordanTransform m
    af' = affineFunctionVPC (keepColumns m indepIdxs, afP, afC) Inhomogeneous
    is2 = tightenIneqSystem $ image af is
    is' = tightenIneqSystem $
          if indepIdxs == [0..nV-1]
             then is
             else inverseImage af' is2


zpolyhedron' = ZP

-- |
-- Return the preimage domain used to represent
-- a Z-polyhedron.
--
zpPreimage :: Real a => ZPolyhedron -> GIneqSystem a
zpPreimage (ZP sys af) = fmap fromIntegral $ integralIneqSystem sys

-- |
-- Return the lattice of a Z-polyhedron.
--
zpLattice :: Real a => ZPolyhedron -> GAffineFunction a
zpLattice (ZP sys af) = fmap fromIntegral af

-- |
-- Compute the real image of the preimage domain
-- under the lattice.
--
zpRealImage :: Real a => ZPolyhedron -> GIneqSystem a
zpRealImage (ZP sys af) = fmap fromIntegral $ integralIneqSystem $
                          image (fmap fromIntegral af) sys


foreign import ccall zpIncludes_c ::
     Ptr PolylibPolyhedron
  -> Ptr PolylibMatrix
  -> Ptr PolylibPolyhedron
  -> Ptr PolylibMatrix
  -> IO CBool
zpIncludes :: ZPolyhedron -> ZPolyhedron -> Bool
zpIncludes zp1@(ZP is1 _) zp2@(ZP is2 _) =
    unsafeCppiConvert4 zpIncludes_c
         is1
         (getMatrix $ zpLattice zp1)
         is2
         (getMatrix $ zpLattice zp2)
    

foreign import ccall zpDifference_c
 :: Ptr (Ptr PolylibPolyhedron)
 -> Ptr (Ptr PolylibMatrix)
 -> Ptr (Ptr PolylibPolyhedron)
 -> Ptr (Ptr PolylibMatrix)
 -> IO (Ptr (CPair (Ptr PolylibPolyhedron) (Ptr PolylibMatrix)))
foreign import ccall zpIntersection_c
 :: Ptr (Ptr PolylibPolyhedron)
 -> Ptr (Ptr PolylibMatrix)
 -> Ptr (Ptr PolylibPolyhedron)
 -> Ptr (Ptr PolylibMatrix)
 -> IO (Ptr (CPair (Ptr PolylibPolyhedron) (Ptr PolylibMatrix)))
foreign import ccall zpUnion_c
 :: Ptr (Ptr PolylibPolyhedron)
 -> Ptr (Ptr PolylibMatrix)
 -> Ptr (Ptr PolylibPolyhedron)
 -> Ptr (Ptr PolylibMatrix)
 -> IO (Ptr (CPair (Ptr PolylibPolyhedron) (Ptr PolylibMatrix)))
foreign import ccall zpUnionDifference_c
 :: Ptr (Ptr PolylibPolyhedron)
 -> Ptr (Ptr PolylibMatrix)
 -> Ptr (Ptr PolylibPolyhedron)
 -> Ptr (Ptr PolylibMatrix)
 -> IO (Ptr (CPair (Ptr PolylibPolyhedron) (Ptr PolylibMatrix)))

zdDifference :: ZDomain -> ZDomain -> ZDomain
zdDifference zp1 []  = zp1
zdDifference []  _   = []
zdDifference zp1 zp2 = zpOp2l_helper zpDifference_c zp1 zp2

zdIntersection :: ZDomain -> ZDomain -> ZDomain
zdIntersection [] zp2  = []
zdIntersection zp1 []  = []
zdIntersection zp1 zp2 = zpOp2l_helper zpIntersection_c zp1 zp2

zdUnion :: ZDomain -> ZDomain -> ZDomain
zdUnion zp1 zp2 = zpOp2l_helper zpUnion_c zp1 zp2

zpOp2l_helper :: (Ptr (Ptr PolylibPolyhedron) ->
                  Ptr (Ptr PolylibMatrix) ->
                  Ptr (Ptr PolylibPolyhedron) ->
                  Ptr (Ptr PolylibMatrix) ->
                  IO (Ptr (CPair (Ptr PolylibPolyhedron) (Ptr PolylibMatrix)))
                 )
              -> ZDomain
              -> ZDomain
              -> ZDomain
zpOp2l_helper zpFunc_c zp1s (zp2s :: ZDomain) =
    -- trace (show (is1,af1,is2,af2)) $
--    trace (show (length zp1s, length zp2s)) $
    [zpolyhedron' is af | (is,af) <- res, notSillyInfeasible is]
    where
    res = [(ineqSystem ((nV,nP),ineqs), makeAf latt) | (ineqs,latt) <- res1]
    res1 :: [(GMatrix Rational, GMatrix Integer)]
    res1 = unsafeCppiConvert4 zpFunc_c
             (map zpPreimage zp1s)
             (map (getMatrix . zpLattice)  zp1s)
             (map zpPreimage zp2s)
             (map (getMatrix . zpLattice)  zp2s)

    makeAf :: GMatrix Integer -> GAffineFunction Integer
    makeAf m = affineFunction ((nVars af1,nParams af1), mIH) Inhomogeneous
        where
        mIH = fmap fromIntegral $ subMatrixHor m 0 (nVt-1)
    af1 = zpLattice $ head (zp1s ++ zp2s)
    nVt = targetDimIH af1
    nV  = nVars af1
    nP  = nParams af1

--
-- Polylib returns ZPolyhedra with  -1 >= 0  as constraint.
-- We filter them out...
--
notSillyInfeasible :: Real a => GIneqSystem a -> Bool
notSillyInfeasible is =
    all rowCondition [0..maxRow m] && oT
    where
    m = getMatrix is
    mC = maxColumn m
    nV = nVars is
    rowCondition r = entry m  r mC >= 0 ||
                     or [entry m r c /= 0 | c <- [0..mC-1]]

    oT = omegaTest (Just ineqs) (Just eqs)
    (ineqs,eqs) = findEqualities is


