-- |
-- Provides elementary functions to work with domains, which are
-- finite unifications of polyhedrons.
--
-- As some operations will produce decision trees, you have
-- to extend the functionality of the function in this
-- module to work on decision trees of domains.
-- This has been done in 'LMath.Logic.DomainTree'.
--
-- See also: 'LMath.Logic.DomainTree'
--
module LMath.Logic.Domain
    ( -- * Constructors
      Domain, DomainID, domain, domain', domain2, domain2',
      ineqToDomain, ineqToDomain', ineqToDomain2,
      emptyDomain, universalDomain, singlePointD,

      -- * Deprecated functions
      fmapD,

      -- * Accessors
      nVarsD, nParamsD, domainToIneqs,

      -- * Extending the number of parameters
      setParamsD,

      -- * Basic operations
      intersectD, unionD, unionsD, complementD, complementD',
      includesD, includesD', equivalentD, equivalentD',
      isEmptyD, isEmptyD', isAlwaysEmptyD, isAlwaysEmptyD',
      imageD, inverseImageD, projectToVarsD_QE, projectToVarsD,

      -- * Extended operations
      lPartitionD, convexHullD, incConvexHullD, disjointUnionD,

      -- * Conversions
      toRatioDomain, toRatioDomain_checked,
      toPolyDomain, toPolyDomain', toFracPolyDomain,
      BasicDomainElement, PolyConvertableDomain,

      -- * PolyLib compatibility
      canBeHandledByPolyLib, toPolyLibIneqsD, toPolyLibIneqsD_checked,

      -- * Simplifications
      simplifyTriviallyD, simplifyD, simplifyD', simplifyDTreeD,
      DomainCanBeSimplified, substituteParamsD, substituteAllConstantsD,
      domainSimplifier,

      -- * Logic part
      domainToFormula, formulaToDomain,

      -- * Topological sorting
      lessEqD, strictLessD, lTopSortD,

      -- * Experimental
      domainToIneqTree,


{-
      -- * Testing function (only exported for testing, but should not be used)
      -- none
-}
    )
where

import Data.Array
import Data.Maybe
import Data.List (genericLength, genericTake, maximumBy,
		  intersperse, sort, nub, nubBy, delete, (\\))
import Data.Ratio
import Debug.Trace

import LMath.Algo
import LMath.Logic.DomainContext
import LMath.Logic.QE
import LMath.Logic.Sort
import LMath.Tools
import LMath.Types
import LMath.Logic.DTreeContexts

-- |
-- Domains are finite unifications of polyhedrons.
-- Do not use this constructor directly, prefer 'domain', 'domain2'
-- and 'ineqToDomain', 'ineqToDomain2' instead.
-- (Reason: There are some constraits, e.g. the
--  internal list must be non-null.)
--
data Domain a = Domain { domPolys   :: [GIneqSystem a]
		       , domFormula :: QfFormula (Polynomial Rational)
		       , domVars    :: Integer
		       , domParams  :: Integer
		       , domIsSimplified :: Bool
		       , domSimplContext :: Maybe DomainContext
		       }

-- |
-- An integer which identifies a certain domain.
-- This type alias is used by some algorithms working on lists of domains.
--
type DomainID = Integer

--
-- Initialize helper functions:
--

-- |
-- Converts a domain to a formula which is in disjunctive normal form.
-- (Each single polyhedron is a conjunction and the list of all polyhedrons
-- is a disjunction of this conjunctions).
--
-- See also: 'domainToFormula'
--
domainToAtFormulas :: (Num a)
                   => Domain (Polynomial a) -- ^ the domain to be converted
		   -> [[AtFormula (Polynomial a)]]
domainToAtFormulas d | checkPrecondD1_weak d "domainToAtFormulas" =
			 map ineqSystemToAtFormulas (domPolys d)

-- |
-- Converts a formula @f@ to a domain.
--
formulaToIneqsD_ :: Maybe ShiftedDomainContext -- ^ the (optional) context
                 -> (Integer, Integer) -- ^ number of variables and parameters
		 -> QfFormula (Polynomial Rational) -- ^ @f@
		 -> [GIneqSystem (Polynomial Rational)]
formulaToIneqsD_ sContext (nV, nP) f =
    atFormulasToIneqsD_ sContext nV (dnfQfD sContext (nV, nP) f)

-- |
-- Converts a formula @dnfF@ to a domain.
-- The formula has to be in disjunctive normal form (DNF).
--
atFormulasToIneqsD_ :: (Integral a)
                   => Maybe ShiftedDomainContext -- ^ the (optional) context
                   -> Integer -- ^ number of variables
		   -> [[AtFormula (Polynomial (Ratio a))]] -- ^ @dnfF@
		   -> [GIneqSystem (Polynomial (Ratio a))]
atFormulasToIneqsD_ sContext nV dnfF
    | nV < 0 = error "atFormulasToIneqsD_: negative number of variables"
    | null dnfF = [infeasibleIneqSystem nV 0]
    | otherwise = ineqs
    where
    ineqs = concatMap toIneq dnfF
    toIneq f = maybe nonLinearError id (atFormulasToIneqSystems nV f)
    nonLinearError =
	error ("atFormulasToIneqsD_: One formula is not linear " ++
	       "in the variables.\nnV=" ++ show nV ++ ", dnfF=" ++ show dnfF)

-- |
-- This function does exactly the same thing as 'formulaToIneqsD_'.
--
-- See also: 'formulaToIneqsD_'
--
initPolysD_ :: Maybe ShiftedDomainContext -- ^ the (optional) context
            -> (Integer, Integer) -- ^ the number of variables and parameters
            -> QfFormula (Polynomial Rational) -- ^ @f@
            -> [GIneqSystem (Polynomial Rational)]
initPolysD_ = formulaToIneqsD_

-- |
-- Decides if it is sensitive to make effort to simplify the
-- domain @d@ under the given context.
--
-- See also: 'simplifyDomainsOnlyOnce'
--
skipSimplificationD_ :: Domain a -- ^ @d@
		     -> Maybe DomainContext -- ^ the context
		     -> Bool
skipSimplificationD_ d context
  | not (domIsSimplified d) = False
  | simplifyDomainsOnlyOnce = True
  | otherwise =
      heuristic_DC_implies_DC' (domSimplContext d) context == (Just True)

-- |
-- Returns a domain @d@, where the internal flag has been set,
-- which indicates that the domain has been already simplified
-- under the given context.
--
-- See also: 'clearSimplifiedFlag_'
--
setSimplifiedFlag_ :: Domain a -- ^ @d@
		   -> Maybe DomainContext -- ^ @context@
                   -> Domain a
setSimplifiedFlag_ d context =
    d { domIsSimplified = True
      , domSimplContext = context
      }

-- |
-- Warning: This function is deprecated!
--
-- Returns a domain @d@, where the internal flag has been cleared.
--
-- See also: 'setSimplifiedFlag_'
--
clearSimplifiedFlag_ :: Domain a -- ^ @d@
                     -> Domain a
clearSimplifiedFlag_ d =
    warnIfUsingDeprecatedFunctionD ("clearSimplifiedFlag_ has been decrecated") $
    d { domIsSimplified = False
      , domSimplContext = error "domain has not been simplified"
      }

--
-- (end of initialize helper functions)
--

-- |
-- Tests if the internal representation of the domain @d@
-- consists of a single polyhedron.
--
containsSinglePolyD :: Domain a -- ^ @d@
		    -> Bool
containsSinglePolyD d = case domPolys d of
			  [_] -> True
			  _   -> False

class (Num a) => BasicDomainElement a where
    -- |
    -- Counts the total number of indeterminates in @x@.
    --
    countIndetsD :: (Integral b)
 		 => a -- ^ @x@
 	         -> b -- ^ the number of indeterminates in @x@

    -- |
    -- Counts the number of used indeterminates in @x@.
    --
    countUsedIndetsD :: (Integral b)
                     => a -- ^ @x@
                     -> b -- ^ the number of used indeterminates in @x@
    -- |
    -- Set the number of indeterminates of @p@ to @x@.
    --
    setNIndetsD :: (Integral b)
  	        => b -- ^ @x@
                -> a -- ^ @p@
                -> a

instance (Integral a) => BasicDomainElement (Ratio a) where
    countIndetsD = const 0
    countUsedIndetsD = const 0
    setNIndetsD 0 x = x
    setNIndetsD n x = -- trace "Warning: setNINdetsD (Ratio a): Cannot be extended." $
		      x

instance (Integral a) => BasicDomainElement (Polynomial (Ratio a)) where
    countIndetsD p = unNIndeterminates $ nIndeterminates p

    countUsedIndetsD p = case usedIndeterminates p of
		           [] -> 0
		           xs -> 1 + last xs

    setNIndetsD n p | n == c = p
		    | n > c = appendIndeterminates    (n-c) p
		    | n < c = deleteEndIndeterminates (c-n) p
	where
	c = unNIndeterminates $ nIndeterminates p

instance (Integral a) => BasicDomainElement (FracPoly (Ratio a)) where
    countIndetsD p = let pNum = polyNumerator p
 			 pDen = polyDenominator p
                     in max (countIndetsD pNum) (countIndetsD pDen)

    countUsedIndetsD p = let pNum = polyNumerator p
                             pDen = polyDenominator p
                         in max (countUsedIndetsD pNum) (countUsedIndetsD pDen)

    setNIndetsD n p | n < 0 = error "setNIndetsD (FracPoly a): n < 0"
		    | otherwise = let pNum = polyNumerator p
				      pDen = polyDenominator p
 				  in (setNIndetsD n pNum) %% (setNIndetsD n pDen)

instance (BasicDomainElement a) => Show (Domain a) where
    show d | checkPrecondD1 d "Domain.show" =
        let xs = domPolys d
	    showPoly c x = show c ++ ". polyhedron:\n" ++ show x
	    polyStr = (concat $ zipWith showPoly [1..] xs)
        in "Domain of " ++ show (length xs) ++ " polyhedrons\n" ++
           show (nVarsD d) ++ " variables, " ++ show (nParamsD d) ++
           " parameters\n" ++ polyStr
           -- ++ debugInfo
        where
        debugInfo = if domIsSimplified d
	            then "simplified=true [context=" ++ show (domSimplContext d) ++ "]\n"
	            else "simplified=false\n"

-- |
-- Applies the operation @f@ to each element of the matrix,
-- which belongs to the domain @d@.
--
-- Note that the dual formula representation will be invalidated.
-- If you do not want that, e.g. if you are converting all elements
-- from 'Ratio' to 'Polynomial', you can try to use the function
-- 'unsafe_fmapD'.
--
-- See also: 'unsafe_fmapD'
--
fmapD :: PolyConvertableDomain b
       => (a -> b) -- ^ @f@
       -> Domain a -- ^ @d@
       -> Domain b
fmapD f d | checkPrecondD1_weak d "fmapD" =
	      logWarningD "fmapD is deprecated" $
	      let (nV, nP) = (nVarsD d, nParamsD d)
              in initDomainByPolys' (nV, nP) (map (fmap f) (domPolys d))

-- |
-- Same as 'fmapD', but it does not invalidate the formula representation.
-- So the resulting domain may become inconsistent, if the conversion 
-- changes the semantic of the original domain @d@.
-- If you are not sure what this means, you should use 'fmapD' instead.
--
-- Note:
-- Currently this function is implemented by calling 'fmapD',
-- because the former implementation was too "unsafe" to use.
--
-- See also: 'fmapD'
--
unsafe_fmapD :: PolyConvertableDomain b
             => (a -> b) -- ^ @f@
             -> Domain a -- ^ @d@
             -> Domain b
unsafe_fmapD f d | checkPrecondD1_weak d "unsafe_fmapD" =
		     logWarningD "unsafe_fmapD is deprecated" $
		     -- d { domPolys = map (fmap f) (domPolys d) }
		     --
		     -- The upper version may lead to errors, because
		     -- the number of indeterminates of the domain
		     -- and its corresponding formula may differ.
		     -- That's why we use the "safe" version instead.
		     fmapD f d

-- |
-- Domains of 'PolyConvertableDomain' elements can be converted
-- to @Domain (Polynomial (Ratio a))@.
--
class PolyConvertableDomain a where
    -- |
    -- Constructs a domain out of a list of polyhedrons
    -- represented by IneqSystems.
    -- 
    -- See also: @initDomainByPolys'@
    --
    initDomainByPolys :: (Integral b)
                      => Maybe DomainContext -- ^ the (optional) context
		      -> (b, b) -- ^ the number of variables and parameters
		      -> [GIneqSystem a]     -- ^ list of polyhedrons
                      -> Domain a
    -- |
    -- Constructs a domain out of a list of polyhedrons
    -- represented by IneqSystems.
    -- (If you have a non-empty context, you should use 'initDomainByPolys')
    --
    -- See also: 'initDomainByPolys'
    --
    initDomainByPolys' :: (Integral b)
                       => (b, b) -- ^ the number of variables and parameters
		       -> [GIneqSystem a] -- ^ list of polyhedrons
                       -> Domain a        -- ^ @d@
    initDomainByPolys' = initDomainByPolys Nothing

    -- |
    -- Initialize a domain by its formula @f@, which can be any type
    -- of quantifier-free formula.
    --
    -- See also: @initDomainByFormula'@
    --
    initDomainByFormula :: (Integral b, Integral c)
                        => Maybe DomainContext -- ^ the (optional) context
	                -> (b, b) -- ^ the number of variables and parameters
			-> QfFormula (Polynomial (Ratio c)) -- ^ @f@
                        -> Domain a

    -- |
    -- Initialize a domain by its formula @f@, which can be any type
    -- of quantifier-free formula.
    --
    -- If you have a special context for this domain, use the
    -- function 'initDomainByFormula'.
    --
    -- See also: 'initDomainByFormula'
    --
    initDomainByFormula' :: (Integral b, Integral c)
                         => (b, b) -- ^ the number of variables and parameters
			 -> QfFormula (Polynomial (Ratio c)) -- ^ @f@
                         -> Domain a
    initDomainByFormula' = initDomainByFormula Nothing

    -- |
    -- Initialize a domain by a formula @f@ and a list of polyhedrons.
    --
    -- You should prefer this function to 'initDomainByPolys' and
    -- 'initDomainByFormula' if you can effienctly compute both
    -- representations (i.e. formula and list of polyhedrons) effienctly.
    -- But if you have to compute one representations explicitly by
    -- converting from the other one, you should not use this function.
    --
    -- Preconditions:
    -- 
    --   (1) The formula @f@ and the list of polyhedrons must describe
    --       the same domain. Otherwise the results are undefined.
    --
    -- See also: 'initDomainByFormula', 'initDomainByPolys'
    --
    initDomainByFormulaAndPolys :: (Integral b, Integral c)
                                => (b, b) -- ^ the number of variables and parameters
                                -> QfFormula (Polynomial (Ratio c)) -- ^ @f@
				-> [GIneqSystem a] -- ^ list of polyhedrons
                                -> Domain a
    -- |
    -- Converts a domain of Polynomials @p@ into a domain of
    -- a more general type ('PolyConvertableDomain').
    --
    fromPolyDomain :: (Integral b)
                   => Domain (Polynomial (Ratio b)) -- ^ @p@
		   -> Domain a

    -- |
    -- Converts a domain to a equivalent domain of polynomials.
    -- An optional context can be passed to improve the
    -- quality of the conversion in some case.
    --
    -- See also: @toPolyDomain'@
    --
    toPolyDomain :: (Integral b)
	         => Maybe DomainContext -- ^ the (optional) context
		 -> Domain a            -- ^ the domain to be converted
                 -> Domain (Polynomial (Ratio b)) -- ^ the converted domain

    -- |
    -- Converts a domain to a equivalent domain of polynomials.
    -- (If you have a non-empty context, you should use 'toPolyDomain')
    --
    -- See also: 'toPolyDomain'
    --
    toPolyDomain' :: (Integral b)
	          => Domain a -- ^ the domain to be converted
                  -> Domain (Polynomial (Ratio b)) -- ^ the converted domain
    toPolyDomain' = toPolyDomain Nothing

instance (Integral c) => PolyConvertableDomain (Ratio c) where
    
    -- context will be ignored
    initDomainByPolys _ (nV, nP) xs =
	checkPostcondD1 d "initDomainByPolys (Ratio c)"
	where
	polyToFormula = ineqSystemToFormula . (fmap convertToPoly) 
	formula = fmap (setNIndetsD (nV+nP)) $
		  disjunctionQf (map polyToFormula xs)
	d = Domain { domPolys   = xs
		   , domFormula = formula
		   , domVars    = fromIntegral nV
		   , domParams  = fromIntegral nP
		   , domIsSimplified = False
		   , domSimplContext = error "domain has not been simplified"
		   }

    initDomainByFormula context (nV, nP) formula =
	checkPostcondD1 d "initDomainByFormula (Ratio c)"
	where
	nV' = fromIntegral nV
	nP' = fromIntegral nP
	sContext = mkShiftedDomainContext' context nV
	formula' = fmap (setNIndetsD (nV+nP) . convertToPoly) formula
	polys    = map (fmap ((setNIndetsD nP) . convertToRatio))
		   (initPolysD_ sContext (nV', nP') formula')
	d = Domain { domPolys   = polys
		   , domFormula = formula' 
		   , domVars    = nV'
		   , domParams  = nP'
		   , domIsSimplified = False
		   , domSimplContext = error "domain has not been simplified"
		   }

    initDomainByFormulaAndPolys (nV, nP) formula polys =
	checkPostcondD1 d "initDomainByFormulaAndPolys (Ratio c)"
	where
	formula' = fmap (setNIndetsD (nV+nP) . convertToPoly) formula
	polys'   = [fmap (setNIndetsD nP) p | p <- polys]
	d = Domain { domPolys   = polys'
		   , domFormula = formula'
		   , domVars    = fromIntegral nV
		   , domParams  = fromIntegral nP
		   , domIsSimplified = False
		   , domSimplContext = error "domain has not been simplified"
		   }

    -- Note: 'convertToRatio' may throw an error
    fromPolyDomain d = let polys = [fmap convertToRatio p | p <- domPolys d]
                       in d { domPolys = polys }

    -- context will be ignored
    toPolyDomain _ d = let polys = [fmap convertToPoly p | p <- domPolys d]
                       in d { domPolys = polys }

instance (Integral c) => PolyConvertableDomain (Polynomial (Ratio c)) where
    -- context will be ignored
    initDomainByPolys _ (nV, nP) xs =
	checkPostcondD1 d "initDomainByPolys (Polynomial (Ratio c))"
	where
	formula = fmap ({-setNIndetsD (nV+nP) . -} convertToPoly) $
		  disjunctionQf (map ineqSystemToFormula polys)
	polys   = [fmap (setNIndetsD nP) p | p <- xs]
	d = Domain { domPolys   = polys
		   , domFormula = formula
		   , domVars    = fromIntegral nV
		   , domParams  = fromIntegral nP
		   , domIsSimplified = False
		   , domSimplContext = error "domain has not been simplified"
		   }

    initDomainByFormula context (nV, nP) formula =
	checkPostcondD1 d "initDomainByFormula (Polynomial (Ratio c))"
	where
	nV' = fromIntegral nV
	nP' = fromIntegral nP
	sContext = mkShiftedDomainContext' context nV
	formula' = fmap (setNIndetsD (nV+nP) . convertToPoly) formula
	polys    = map (fmap ((setNIndetsD nP) . convertToPoly))
		   (initPolysD_ sContext (nV', nP') formula')
	d = Domain { domPolys   = polys
		   , domFormula = formula' 
		   , domVars    = nV'
		   , domParams  = nP'
		   , domIsSimplified = False
		   , domSimplContext = error "domain has not been simplified"
		   }

    initDomainByFormulaAndPolys (nV, nP) formula polys =
	checkPostcondD1 d "initDomainByFormulaAndPolys (Polynomial (Ratio c))"
	where
	formula' = fmap (setNIndetsD (nV+nP) . convertToPoly) formula
	polys'   = [fmap (setNIndetsD nP) p | p <- polys]
	d = Domain { domPolys   = polys'
		   , domFormula = formula'
		   , domVars    = fromIntegral nV
		   , domParams  = fromIntegral nP
		   , domIsSimplified = False
		   , domSimplContext = error "domain has not been simplified"
		   }

    fromPolyDomain d = let polys = [fmap convertToPoly p | p <- domPolys d]
                       in d { domPolys = polys }

    -- context will be ignored
    toPolyDomain _ d = let polys = [fmap convertToPoly p | p <- domPolys d]
                       in d { domPolys = polys }

instance (Integral c) => PolyConvertableDomain (FracPoly (Ratio c)) where
    initDomainByPolys context (nV, nP) xs =
	checkPostcondD1 d "initDomainByPolys (FracPoly (Ratio c))"
	where
	polyToFormula = ineqSystemToFormula .
			(convertIneqFromFracToPoly context) .
			(fmap convertToFracPoly)
	formula = fmap (setNIndetsD (nV+nP) . convertToPoly) $
		  disjunctionQf (map polyToFormula polys)
	polys   = [fmap (setNIndetsD nP) p | p <- xs]
	d = Domain { domPolys   = polys
		   , domFormula = formula
		   , domVars    = fromIntegral nV
		   , domParams  = fromIntegral nP
		   , domIsSimplified = False
		   , domSimplContext = error "domain has not been simplified"
		   }

    initDomainByFormula context (nV, nP) formula =
	checkPostcondD1 d "initDomainByFormula (FracPoly (Ratio c))"
	where
	nV' = fromIntegral nV
	nP' = fromIntegral nP
	sContext = mkShiftedDomainContext' context nV
	formula' = fmap (setNIndetsD (nV+nP) . convertToPoly) formula
	polys    = map (fmap ((setNIndetsD nP) . convertToFracPoly))
		   (initPolysD_ sContext (nV', nP') formula')
	d = Domain { domPolys   = polys
		   , domFormula = formula' 
		   , domVars    = nV'
		   , domParams  = nP'
		   , domIsSimplified = False
		   , domSimplContext = error "domain has not been simplified"
		   }

    initDomainByFormulaAndPolys (nV, nP) formula polys =
	checkPostcondD1 d "initDomainByFormulaAndPolys (FracPoly (Ratio c))"
	where
	formula' = fmap (setNIndetsD (nV+nP) . convertToPoly) formula
	polys'   = [fmap (setNIndetsD nP) p | p <- polys]
	d = Domain { domPolys   = polys'
		   , domFormula = formula'
		   , domVars    = fromIntegral nV
		   , domParams  = fromIntegral nP
		   , domIsSimplified = False
		   , domSimplContext = error "domain has not been simplified"
		   }

    fromPolyDomain d = let polys = [fmap convertToFracPoly p | p <- domPolys d]
                       in d { domPolys = polys }

    toPolyDomain context d
	| checkPrecondD1 d "PolyConvertableDomain (FracPoly (Ratio c)): toPolyDomain" =
	    initDomainByPolys context (nV, nP) polys
	where
	(nV, nP) = (nVarsD d, nParamsD d)
	polys = map (fmap convertToPoly .
		     convertIneqFromFracToPoly context) (domPolys d)

-- |
-- Domains over instances of this class can be simplified.
--
class DomainCanBeSimplified a where
    -- |
    -- Makes best efford to simplify the domain @d@.
    --
    -- Note: If the @d@ has been already simplified, then no
    -- effort is made to simplify it further.
    --
    -- See also: @simplifyD'@, 'newly_simplifyD'
    --
    simplifyD :: Maybe DomainContext -- ^ the (optional) context
	      -> Domain a            -- ^ the domain @d@ to be simplified
              -> Domain a            -- ^ a simplified version of @d@

    -- |
    -- Makes best efford to simplify the domain @d@.
    --
    -- If you have a non-empty context, you should use 'simplifyD' instead.
    --
    -- Note: If the @d@ has been already simplified, then no
    -- effort is made to simplify it further.
    --
    -- See also: 'simplifyD', 'newly_simplifyD'
    --
    simplifyD' :: Domain a -- ^ the domain @d@ to be simplified
	       -> Domain a -- ^ a simplified version of @d@
    simplifyD' = simplifyD Nothing

    --
    -- Makes best efford to simplify the domain @d@.
    --
    -- Idea:
    -- If the context is getting more specialized, then the
    -- simplification can be done again to get better results.
    --
    -- See also: 'simplifyD', @simplifyD'@
    --
    newly_simplify :: Maybe DomainContext
                   -> Domain a
                   -> Domain a
    newly_simplify Nothing d = simplifyD' d -- no new context information
    newly_simplify context d =
	if null $ fromDomainContext' context
	then simplifyD' d -- no new context information
	else simplifyD context (clearSimplifiedFlag_ d)

-- |
-- Simplifies a domain of rational numbers.
--
-- See also: 'simplifyTriviallyD'
--
instance (Integral a) => DomainCanBeSimplified (Ratio a) where
    simplifyD context d
	| checkPrecondD1 d "simplifyD (Ratio a)" =
 	    if skipSimplificationD_ d context
	    then d -- domain has been already simplified
	    else let polyD = simplifyD context (toPolyDomain context d)
                 in setSimplifiedFlag_ (toRatioDomain polyD) context

-- |
-- Simplifies a domain of 'Polynomial's.
--
instance (Integral a) => DomainCanBeSimplified (Polynomial (Ratio a)) where
    simplifyD context d
 	| checkPrecondD1 d "simplifyD (Polynomial (Ratio a))" =
 	    if skipSimplificationD_ d context
	    then d -- domain has been already simplified
	    else let (nV, nP) = (nVarsD d, nParamsD d)
		     sContext = mkShiftedDomainContext' context nV
		     fPre   = domainToFormula d
		     fSimpl = simplifyQfD sContext (nV, nP) fPre
		     simplDomain = if fPre == fSimpl
				   then d -- domain could not be simplified
				   else initDomainByFormula context (nV, nP) fSimpl
                 in setSimplifiedFlag_ simplDomain context

-- |
-- Simplifies a domain of 'FracPoly's.
--
instance (Integral a) => DomainCanBeSimplified (FracPoly (Ratio a)) where
    simplifyD context d
 	| checkPrecondD1 d "simplifyD (FracPoly (Ratio a))" =
 	    if skipSimplificationD_ d context
	    then d -- domain has been already simplified
 	    else let polyD = simplifyD context (toPolyDomain context d)
                 in setSimplifiedFlag_ (toFracPolyDomain polyD) context

-- |
-- Create a 'Domain' which contains all the given polyhedrons.
--
-- Preconditions:
--
--   (1) The list @xs@ must not be empty.
--
--   (2) The number of variables of all inequality systems @xs@ must be equal.
--
--   (3) All inequality systems must not contain parameters,
--       but they may contain polynomials.
--
-- If the third precondition is too restrictive, you should use 'domain2'.
--
-- See also: @domain'@, 'domain2'
--
domain :: (BasicDomainElement a, PolyConvertableDomain a, Integral b)
       => Maybe DomainContext -- ^ the (optional) context
       -> (b, b)          -- ^ the number of variables and parameters
       -> [GIneqSystem a] -- ^ a non-null list of polyhedrons @xs@
       -> Domain a        -- ^ a domain which consists of all polyhedrons @xs@
domain context (nV, nP) xs
    | null xs = error "domain: empty list"
    | assertVarsAreEqual xs =
	verifyResult $ unionsD (map (ineqToDomain context (nV, nP)) xs)
    where
    assertVarsAreEqual [] = error "domain: empty domain list"
    assertVarsAreEqual (x:xs) =
	if domain_NDEBUG || all (\y -> nVars y == nVars x) xs
	then True else error ("domain: number of variables of all "++
			      "GIneqSystems must be equal")
    verifyResult x | checkPrecondD1 x "domain (postcondition)" = x

-- |
-- Create a 'Domain' which contains all the given polyhedrons.
--
-- Preconditions:
--
--   (1) The list @xs@ must not be empty.
--
--   (2) The number of variables of all inequality systems @xs@ must be equal.
--
--   (3) All inequality systems must not contain parameters,
--       but they may contain polynomials.
--
-- If the third precondition is too restrictive, you should use 'domain2'.
--
-- If you have a special context for the domain, you should
-- use the function 'ineqToDomain' instead.
--
-- See also: @domain'@, 'domain2'
--
domain' :: (BasicDomainElement a, PolyConvertableDomain a, Integral b)
        => (b, b)          -- ^ the number of variables and parameters
        -> [GIneqSystem a] -- ^ a non-null list of polyhedrons @xs@
        -> Domain a        -- ^ a domain which consists of all polyhedrons @xs@
domain' = domain Nothing

-- |
-- Create a 'Domain' which contains all the given polyhedrons.
--
-- Preconditions:
--
--   (1) The list @xs@ must not be empty.
--
--   (2) The number of variables of all inequality systems @xs@ must be equal.
--
-- See also: @domain2'@, 'domain'
--
domain2 :: (PolyConvertable a, Integral b, Integral c)
        => Maybe DomainContext -- ^ the (optional) context
	-> (b, b) -- ^ the number of variables and parameters
        -> [GIneqSystem a] -- ^ a non-null list of polyhedrons @xs@
        -> Domain (Polynomial (Ratio c))
	-- ^ a domain which consists of all polyhedrons @xs@
domain2 context (nV, nP) xs
    | assertVarsAreEqual xs = unionsD $ map (ineqToDomain2 context (nV, nP)) xs
    where
    assertVarsAreEqual [] = error "domain2: empty domain list"
    assertVarsAreEqual (x:xs) =
	if domain_NDEBUG || all (\y -> nVars y == nVars x) xs
	then True else error ("domain2: number of variables of all "++
			      "GIneqSystems must be equal")

-- |
-- Create a 'Domain' which contains all the given polyhedrons.
--
-- Preconditions:
--
--   (1) The list @xs@ must not be empty.
--
--   (2) The number of variables of all inequality systems @xs@ must be equal.
--
-- If you have a special context for the domain, you should
-- use the function 'domain2' instead.
--
-- See also: @domain2'@, 'domain'
--
domain2' :: (PolyConvertable a, Integral b, Integral c)
         => (b, b) -- ^ the number of variables and parameters
         -> [GIneqSystem a] -- ^ a non-null list of polyhedrons @xs@
         -> Domain (Polynomial (Ratio c))
	 -- ^ a domain which consists of all polyhedrons @xs@
domain2' = domain2 Nothing

-- |
-- Constructs a 'Domain' that consists of a single polyhedron @x@.
--
-- See also: @ineqToDomain'@
--
ineqToDomain :: (BasicDomainElement a, PolyConvertableDomain a,
		 Integral b)
             => Maybe DomainContext -- ^ the (optional) context
             -> (b, b) -- ^ the number of variables and parameters
	     -> GIneqSystem a -- ^ @x@
	     -> Domain a
ineqToDomain context (nV, nP) x
    | nV /= nVars   x = error "ineqToDomain: nV /= nVars x"
    | otherwise = 
	verifyResult $
	if nParams x == 0
	then initDomainByPolys context (nV, nP) [x]
	else error ("ineqToDomain: 'nParams x == 0' was not met " ++
		    "by the input IneqSystem x:\n" ++ show x)
    where
    nV = nVars x
    -- debugging functions:
    verifyResult y | checkPrecondD1 y "ineqToDomain (postcondition)" = y

-- |
-- Constructs a 'Domain' that consists of a single polyhedron @x@.
--
-- If you have a special context for the domain, you should
-- use the function 'ineqToDomain' instead.
--
-- See also: 'ineqToDomain'
--
ineqToDomain' :: (BasicDomainElement a, PolyConvertableDomain a,
		  Integral b)
              => (b, b) -- ^ the number of variables and parameters
              -> GIneqSystem a -- ^ @x@
	      -> Domain a
ineqToDomain' = ineqToDomain Nothing

-- |
-- Constructs a 'Domain' that consists of a single polyhedron @x@.
--
-- Should be the default function, because it offers the user more
-- comfort than 'ineqToDomain'.
--
ineqToDomain2 :: (PolyConvertable a, Integral b, Integral c)
              => Maybe DomainContext -- ^ the domain's special context
	      -> (b, b) -- ^ the number of variables and parameters
              -> GIneqSystem a -- ^ @x@
 	      -> Domain (Polynomial (Ratio c))
ineqToDomain2 context (nV, nP) x_
    | nV /= nVars x = error "ineqToDomain2: nV /= nVars x"
    | nParams x /= 0 && nParams x /= nP =
	error "ineqToDomain2: nParams x must be zero or equal to 'nP'"
    | otherwise =
	verifyResult $ toPolyDomain' $
	if nParams x == 0
	then initDomainByPolys context (nV, nP) [x]
	else if allEntriesRational
	     then initDomainByPolys context (nV, nP) [x']
	     else err
    where
    x  = fmap convertToPoly x_

    -- debugging functions:
    verifyResult y | checkPrecondD1 y "ineqToDomain (postcondition)" = y
    err = error ("ineqToDomains: if the input IneqSystem 'p' contains " ++
		 "non-constant values, then 'nParams p' must be 0.\n" ++
		 "p:" ++ show x)

    -- extract information about the IneqSystem 'x':
    -- test if the polyhedron contains only rational numbers
    allEntriesRational = all (isJust . isConstPoly) (entries x)

    -- eliminate the parameter columns of the IneqSystem 'x';
    -- this can be done by "moving" the parameters into the "constant 1-column"
    x' = ineqSystem ((nV, 0), multijoin emptyMatrix entries')
    emptyMatrix = zeroMatrix 0 (nV + 1)
    entries' = map convertRow (rows x)
    convertRow row = let (varsE, remE) = splitAt (fromIntegral nV) (entries row)
			 (paramsE, oneE) = (init remE, last remE)
			 newOneE = oneE + sum (zipWith (*) paramsE indets)
                     in rowMatrix (varsE ++ [newOneE])
    indets = map (indeterminate nP) [0..nP - 1]

-- |
-- Creates an empty domain.
--
-- See also: 'universalDomain'
--
emptyDomain :: (BasicDomainElement a, PolyConvertableDomain a, Integral b)
	    => (b, b) -- ^ the number of variables and parameters
	    -> Domain a
emptyDomain (nV, nP) =
    if enableMatrixFormulaOptimization
    then initDomainByFormulaAndPolys (nV, nP) formula [emptyPoly]
    else ineqToDomain Nothing (nV, nP) emptyPoly
    where
    emptyPoly = infeasibleIneqSystem nV 0
    formula   = QfAt AtFalse

-- |
-- Creates a domain without constraints.
--
-- See also: 'emptyDomain'
--
universalDomain :: (BasicDomainElement a, PolyConvertableDomain a, Integral b)
		 => (b, b) -- ^ the number of variables and parameters
		 -> Domain a
universalDomain (nV, nP) = ineqToDomain' (nV, nP) emptyIneqSystem
    where
    emptyIneqSystem = ineqSystem ((nV, 0), emptyMatrix)
    emptyMatrix     = zeroMatrix 0 (nV + 1)

-- |
-- Creates an 'IneqSystem' which contains only one point, which has been given
-- as a vector. This vector must have an entry for every variable.
--
-- See also: 'singlePointD'
--
singlePointIneq :: (Integral b, Num a)
                => (b, b) -- ^ the number of variables and parameters
                -> [a]    -- ^ the coordinates of the point @v@
                -> GIneqSystem a
		-- ^ an IneqSystem which contains only the point @v@
singlePointIneq (nV, nP) v
    | nP < 0 = error "singlePointIneq: nP < 0"
    | genericLength v /= nV =
	error ("singlePointD: length of input vector must " ++
               "be equal to the number of variables")
    | otherwise = ineqSystem ((nV, 0), matrix rows)
    where
    rows = concat $
	   zipWith (\x p -> let r = [if y == p then 1 else 0 | y <- [0..nV-1]]
		                    ++ [-x]
                            in [r, map negate r]) v [0..]

-- |
-- Creates a domain which contains only one point, which has been given
-- as a vector. This vector must have an entry for every variable.
--
-- See also: 'singlePointIneq'
--
singlePointD :: (PolyConvertableDomain a, BasicDomainElement a, Integral b)
             => (b, b) -- ^ the number of variables and parameters
             -> [a]    -- ^ the coordinates of the point @v@
             -> Domain a -- ^ a domain which contains only the point @v@
singlePointD (nV, nP) v = ineqToDomain' (nV, nP) (singlePointIneq (nV, nP) v)

-- |
-- Returns the number of variables of the domain @d@.
--
nVarsD :: Integral b
       => Domain a -- ^ @d@
       -> b
nVarsD = fromIntegral . domVars

-- |
-- Returns the number of parameters of the domain @d@.
--
nParamsD :: Integral b
	 => Domain a -- ^ @d@
         -> b
nParamsD = fromIntegral . domParams

-- |
-- Divides the domain @d@ into a set of polyhedrons and
-- returns it as a list.
--
-- Notes:
-- 
--   (1) No assumption can be made that this decomposition is minimal.
--
--   (2) This function will most likely return the
--       internal representation of @d@, which is always non-empty.
--
domainToIneqs :: Domain a        -- ^ @d@
	      -> [GIneqSystem a] -- ^ a non-empty list of 'GIneqSystem'
domainToIneqs d | checkPrecondD1_weak d "domainToIneqs" = domPolys d

-- |
-- Internal helper function. Should not be used outside this module!
--
-- Returns all elements of the matrices of each polyhedron of the domain @d@.
--
domainEntries :: Domain a -- ^ @d@
	      -> [a]      -- ^ a list of all elements of the matrices
domainEntries d = concatMap entries (domPolys d)

-- |
-- Test if the domains internal list contains at least one polyhedron.
--
isDomainListEmpty :: Domain a -- ^ the domain
		  -> Bool     -- ^ @True@ iff the internal list is empty
isDomainListEmpty = null . domPolys

-- |
-- Displays a warning message that has something to do
-- with the Domain-Library.
--
-- This function can be used exactly like 'trace'.
--
-- See also: 'warnIfUsingDeprecatedFunctionD'
--
logWarningD :: String -- ^ the warning
            -> a      -- ^ the value @x@ that is being passed through
            -> a      -- ^ the (unmodified) value @x@
logWarningD = trace

-- |
-- Displays a warning whenever a deprecated function is being used.
--
-- This function can be used exactly like 'trace'.
--
-- See also: 'logWarningD'
--
warnIfUsingDeprecatedFunctionD
    :: String -- ^ the name of the deprecated function
    -> a      -- ^ the value @x@ that is being passed through
    -> a      -- ^ the (unmodified) value @x@
warnIfUsingDeprecatedFunctionD = trace

-- |
-- Controls whether the assertions which are used in this module
-- should be disabled (like NDEBUG in C\/C++).
--
--   [@True@]   skip assertions
--
--   [@False@]  test assertions (recommended)
--
domain_NDEBUG :: Bool
domain_NDEBUG = True

-- |
-- Controls whether domains, which have been already simplified,
-- should be resimplified, if additional context information
-- is available.
--
--   [@True@]   Do not resimplify the domain
--
--   [@False@]  If new context is available, resimplify the domain
--
simplifyDomainsOnlyOnce :: Bool
simplifyDomainsOnlyOnce = False

-- |
-- This switch enables the use of the dual representation of domains
-- for optimization. (Strongly recommended to be @True@.)
--
enableMatrixFormulaOptimization :: Bool
enableMatrixFormulaOptimization = True

-- |
-- Performs some checks on the domain @d@, unless 'domain_NDEBUG' is 'True'.
-- Always returns the input domain @d@, but it may throw an error exception.
--
-- Includes all checks of 'checkPostcondD1_weak'.
--
checkPostcondD1 :: (BasicDomainElement a)
                => Domain a -- ^ @d@
		-> String
		-- ^ the caller's name (will be used in error messages)
		-> Domain a -- ^ always the input domain @d@
checkPostcondD1 d funcName
    | domain_NDEBUG = d
    | checkPrecondD1 d (funcName ++ " (postcondition)") = d
    | otherwise = error (funcName ++ " (postcondition): internal error")

-- |
-- Performs some checks on the domain @d@, unless 'domain_NDEBUG' is 'True'.
-- Always returns the input domain @d@, but it may throw an error exception.
--
checkPostcondD1_weak :: Domain a -- ^ @d@
		     -> String
		     -- ^ the caller's name (will be used in error messages)
		     -> Domain a -- ^ always the input domain @d@
checkPostcondD1_weak d funcName
    | domain_NDEBUG = d
    | checkPrecondD1_weak d (funcName ++ " (postcondition)") = d
    | otherwise = error (funcName ++ " (postcondition): internal error")

-- |
-- Performs some checks on the domain @d@, unless 'domain_NDEBUG' is 'True'.
-- Always returns 'True', but it may throw an error exception.
--
checkPrecondD1_weak :: Domain a -- ^ @d@
	            -> String -- ^ the caller's name (will be used in error messages)
	            -> Bool   -- ^ always @True@
checkPrecondD1_weak d funcName
    | domain_NDEBUG       = True -- skip all tests if NDEBUG is defined
    | isDomainListEmpty d = err "empty domain list"
    | nVarsD   d < 0 = err "nVarsD   d < 0"
    | nParamsD d < 0 = err "nParamsD d < 0"
    | any (\x -> nParams x /= 0) (domPolys d) =
	err ("the underlying IneqSystems of the domain must not contain parameters.")
    | any (\x -> nVars x /= nVarsD d) (domPolys d) =
	err ("all underlying IneqSystems must have the same number of variables.")
    | otherwise = True
    where
    err message = error (funcName ++ ": " ++ message)

-- |
-- Performs some checks on the domain @d@, unless 'domain_NDEBUG' is 'True'.
-- Always returns 'True', but it may throw an error exception.
--
-- Includes all checks of 'checkPrecondD1_weak'.
--
checkPrecondD1 :: BasicDomainElement a
               => Domain a -- ^ @d@
	       -> String -- ^ the caller's name (will be used in error messages)
	       -> Bool   -- ^ always @True@
checkPrecondD1 d funcName
    | domain_NDEBUG = True -- skip all tests if NDEBUG is defined
    | not (checkPrecondD1_weak d funcName) =
	error "checkPrecondD1: internal error, should not be reached"
    | otherwise = True

-- |
-- Performs some checks on the domain @d1@ and @d2@,
-- unless 'domain_NDEBUG' is 'True'.
-- Always returns 'True', but may throw an error exception.
--
checkPrecondD2_weak :: Domain a -- ^ @d1@
	            -> Domain a -- ^ @d2@
	            -> String -- ^ the caller's name (will be used in error messages)
	            -> Bool   -- ^ always @True@
checkPrecondD2_weak d1 d2 funcName
    | domain_NDEBUG          = True -- skip all tests if NDEBUG is defined
    | nVarsD d1 /= nVarsD d2 = err ("number of variables differ (nVarsD d1 == " ++
				    show (nVarsD d1) ++ ", nVarsD d2 == " ++
				    show (nVarsD d2) ++ ")")
    | nParamsD d1 /= nParamsD d2 =
	err ("number of parameters diff (nParamsD d1 == " ++
	     show (nParamsD d1) ++ ", nParamsD d2 ==" ++
	     show (nParamsD d2) ++ ")")
    | not (checkPrecondD1_weak d1 funcName && checkPrecondD1_weak d2 funcName) = 
	error "checkPrecondD2_weak: internal error, should not be reached"
    | otherwise = True
    where
    err message = error (funcName ++ ": " ++ message)

-- |
-- Performs some checks on the domain @d1@ and @d2@,
-- unless 'domain_NDEBUG' is 'True'.
-- Always returns 'True', but may throw an error exception.
--
-- Includes all checks of 'checkPrecondD2_weak'.
--
checkPrecondD2 :: BasicDomainElement a
               => Domain a -- ^ @d1@
	       -> Domain a -- ^ @d2@
	       -> String -- ^ the caller's name (will be used in error messages)
	       -> Bool   -- ^ always @True@
checkPrecondD2 d1 d2 funcName
    | domain_NDEBUG = True -- skip all tests if NDEBUG is defined
    | not (checkPrecondD2_weak d1 d2 funcName) =
	error "checkPrecondD2: internal error, should not be reached"
    | otherwise = checkPrecondD1 d1 funcName && checkPrecondD1 d2 funcName

-- |
-- Checks if all given domains are pairwise disjoint
-- (under the current context). The test will be skipped if
-- 'domain_NDEBUG' is 'True'.
--
-- This function will always return 'True', but it may throw an error exception.
--
checkPrecondAllDisjoint :: (PolyConvertableDomain a, BasicDomainElement a)
                        => Maybe DomainContext -- ^ the (optional) context
                        -> [Domain a] -- ^ the list of domains to be tested
                        -> String
			-- ^ the caller's name (will be used in error messages)
                        -> Bool -- ^ always @True@
checkPrecondAllDisjoint context ds funcName
    | domain_NDEBUG = True -- skip all tests if NDEBUG is defined
    | otherwise = case candidates of
		    ((d1,d2):_) -> errMessage d1 d2
		    [] -> True
    where
    l = length ds 
    candidates = [(d1,d2) | i <- [0..l-1], j <- [i+1..l-1],
	                    let d1 = ds !! i, let d2 = ds !! j,
		            notDisjoint d1 d2]
    notDisjoint d1 d2 = not $ isAlwaysEmptyD context (d1 `intersectD` d2)
    errMessage d1 d2 = error (funcName ++ ": precondition was not met: " ++
			      "d1 and d2 were not disjoint:\nd1:\n" ++
			      show d1 ++ "\nd2:\n" ++ show d2 ++
			      "\nContext:\n" ++ show context)

-- |
-- Modifies the number of parameters.
-- Returns a domain which is equivalent to the input domain @d@,
-- but has the requested number of parameters.
--
setParamsD :: (BasicDomainElement a, Integral b)
           => b -- ^ the number of parameters
           -> Domain a -- ^ @d@
           -> Domain a
setParamsD nP d
    | nP == nP_old = d -- nothing to do
    | otherwise = d { domPolys   = polys
		    , domFormula = formula
		    , domParams  = fromIntegral nP
		    }
    where
    (nV, nP_old) = (nVarsD d, nParamsD d)
    polys   = [fmap (setNIndetsD nP) p | p <- domPolys d]
    formula = fmap (setNIndetsD (nV+nP)) (domFormula d)

-- |
-- Warning: This function is deprecated.
-- Use 'setParamsD' instead.
--
-- Takes two domains @d1@ and @d2@ and returns a pair of two
-- domains which are equivalent of @d1@ and @d2@ respectively,
-- but they have at least as much parameters as @d1@ and @d2@.
--
-- See also: @extendParamsD'@
--
extendParamsD :: (BasicDomainElement a, PolyConvertableDomain a)
	      => Domain a -- ^ @d1@
	      -> Domain a -- ^ @d2@
	      -> (Domain a, Domain a) -- ^ the "extended" domain @d1@ and @d2@
extendParamsD d1 d2
    | checkPrecondD2 d1 d2 "extendParamsD" =
	logWarningD "extendParamsD is deprecated" $
	(setParamsD nP_max d1, setParamsD nP_max d2)
    where
    nP_max = (nParamsD d1) `max` (nParamsD d2)

-- |
-- Warning: This function is deprecated.
-- Use 'setParamsD' instead.
--
-- Takes a list of domains @ds@ and returns a list of equivalent domains
-- which have at least as much parameters as each domain in @ds@.
--
-- This is a more generalized version of 'extendParamsD'.
--
-- See also: 'extendParamsD'
--
extendParamsD' :: (BasicDomainElement a, PolyConvertableDomain a)
               => [Domain a] -- ^ @ds@ list of domains
	       -> [Domain a] -- ^ list of "extended" domains
extendParamsD' ds
    | all (\x -> checkPrecondD1 x "extendParamsD'") ds =
	logWarningD "extendParamsD' is deprecated" $
	map (setParamsD nP_max) ds
    | otherwise = error ("extendParamsD': one of the input domains did " ++
			 "not satisfy the preconditions")
    where
    nP_max = maximum $ map nParamsD ds

-- |
-- Warning: This function is deprecated.
-- Use 'setParamsD' instead.
--
-- Reduce the number of indeterminates of @p@ to @x@.
--
shrinkParamsD :: (BasicDomainElement a, PolyConvertableDomain a, Integral b)
              => b        -- ^ @x@ 
              -> Domain a -- ^ a domain @p@
	      -> Domain a -- ^ @p@ with only @x@ intedeterminates
shrinkParamsD c d
    | c < 0 = error ("shrinkParamsD: argument must be >= 0")
    | checkPrecondD1 d "shrinkParamsD" =
	logWarningD "shrinkParamsD is deprecated" $
	setParamsD c d

-- |
-- Intersects a domain with a single polyhedron.
--
intersectD_polyhedron :: (PolyConvertableDomain a)
                      => Domain a      -- ^ the domain
		      -> GIneqSystem a -- ^ the polyhedron
		      -> Domain a      -- ^ the intersection
intersectD_polyhedron d poly
    | checkPrecondD1_weak d "intersectD_polyhedron" =
	let (nV, nP) = (nVarsD d, nParamsD d)
	in initDomainByPolys' (nV, nP) (map (join poly) (domPolys d))

-- |
-- Intersects two domains.
--
intersectD :: (BasicDomainElement a, PolyConvertableDomain a)
	   => Domain a -- ^ the first domain  (@d1@)
	   -> Domain a -- ^ the second domain (@d2@)
	   -> Domain a -- ^ the intersection of @d1@ and @d2@
intersectD d1 d2
    | checkPrecondD2 d1 d2 "intersectD" =
	if enableMatrixFormulaOptimization
	then initDomainByFormulaAndPolys (nV, nP) formula polys
	else unionsD [intersectD_polyhedron d2 p | p <- (domPolys d1)]
    where
    (nV, nP) = (nVarsD d1, nParamsD d1)
    formula  = (domFormula d1) `QfAnd` (domFormula d2)
    polys    = concat [map (join p) (domPolys d2)  | p <- domPolys d1]

-- |
-- Returns the union of two domains.
--
unionD :: (BasicDomainElement a, PolyConvertableDomain a)
       => Domain a -- ^ the first domain  (@d1@)
       -> Domain a -- ^ the second domain (@d2@)
       -> Domain a -- ^ the union of @d1@ and @d2@
unionD d1 d2
    | checkPrecondD2 d1 d2 "unionD" =
	let xs' = domPolys d1
	    ys' = domPolys d2
	    polys = xs' `append'` ys'
	    formula = (domFormula d1) `QfOr` (domFormula d2)
        in if enableMatrixFormulaOptimization
	   then initDomainByFormulaAndPolys (nV, nP) formula polys
	   else initDomainByPolys' (nV, nP) polys
    where
    (nV, nP) = (nVarsD d1, nParamsD d1)
    -- same as (++), but avoids the worst case of (foldl1 unionD xs)
    -- (The end user does not know that it is better to
    -- use foldr1 here than foldl1)
    append' (x:xs) (y:ys) = x:y:append' xs ys
    append' [] ys = ys
    append' xs [] = xs

-- |
-- Returns the union of a non-empty set of domains.
--
-- (Semantically equal to @foldr1 unionD@ and @foldl1 unionD@,
-- but it may be faster because of the Matrix\/Formula optimization.)
--
unionsD :: (BasicDomainElement a, PolyConvertableDomain a)
        => [Domain a] -- ^ list of domains to be unified
	-> Domain a   -- ^ the unified domain
--unionsD = foldr1 unionD
unionsD [] = error "unionsD: empty list"
unionsD xs = if enableMatrixFormulaOptimization
	     then initDomainByFormulaAndPolys (nV, nP) formula polys
	     else initDomainByPolys' (nV, nP) polys
     where
     (nV, nP) = (nVarsD (head xs), nParamsD (head xs))
     polys    = concatMap domPolys xs
     formula  = disjunctionQf [domFormula x | x <- xs]

-- |
-- Computes the complement of a domain @d1@ under another domain @d2@
--
-- Warning:
-- This function does not work for rational (or real) solutions,
-- but only for integral points.
--
-- See also: @complementD'@
--
complementD ::  (BasicDomainElement a, PolyConvertableDomain a)
            => Maybe DomainContext -- ^ the (optional) context
	    -> Domain a -- ^ @d1@
	    -> Domain a -- ^ @d2@
	    -> Domain a -- ^ all points in @d1@ which are not in @d2@
complementD context d1 d2
    | checkPrecondD2 d1 d2 "complementD" =
	let f1 = domFormula d1
	    f2 = domFormula d2
	    compl = f1 `QfAnd` (qfNot f2) -- all points which are in 'd1' but not in 'd2'
	    (nV, nP) = (nVarsD d1, nParamsD d1)
	in initDomainByFormula context (nV, nP) compl

-- |
-- Computes the complement of a domain @d1@ under another domain @d@.
--
-- If you have a non-empty context, you should prefer 'complementD'.
--
-- Warning:
-- This function does not work for rational (or real) solutions,
-- but only for integral points.
--
-- See also: 'complementD'
--
complementD' :: (BasicDomainElement a, PolyConvertableDomain a)
	     => Domain a
             -> Domain a
             -> Domain a
complementD' = complementD Nothing
        
-- |
-- Tests if the first domain @d1@ is part of the second domain @d2@.
--
-- Both domains are converted to quantifier-free formulas.
-- Then @d1@ is part of @d2@ if the logical expression of @d2@
-- implies the expression of @d1@.
--
-- See also: @includesD'@
--
includesD :: (BasicDomainElement a)
          => Maybe DomainContext -- ^ the (optional) context
          -> Domain a -- ^ @d1@
          -> Domain a -- ^ @d2@
          -> Bool     -- ^ @True@ iff @d2@ includes @d1@
includesD context d1 d2
    | checkPrecondD2 d1 d2 "includesD" =
	let l1 = domainToFormula d1
	    l2 = domainToFormula d2
	    (nV, nP) = (nVarsD d1, nParamsD d1)
	    sContext = mkShiftedDomainContext' context nV
        in impliesQfD sContext (nV, nP) (qfImpl l1 l2)

-- |
-- Tests if the first domain @d1@ is part of the second domain @d2@.
--
-- If you have a special context, you should prefer the function 'includesD'.
--
-- See also: 'includesD'
--
includesD' :: (BasicDomainElement a)
           => Domain a -- ^ @d1@
	   -> Domain a -- ^ @d2@
           -> Bool -- ^ @True@ iff @d2@ includes @d1@
includesD' = includesD Nothing

-- |
-- Tests if the two domains @d1@ and @d2@ are equivalent
-- under the given context.
--
-- First both domains are converted to quantifier-free formulas.
-- Then @d1@ and @d2@ are equvalent if the logical expression of @d2@
-- implies the expression of @d1@ and vice versa.
--
-- See also: @equivalentD'@
--
equivalentD :: (BasicDomainElement a)
	    => Maybe DomainContext -- ^ the (optional) context
	    -> Domain a -- ^ @d1@
	    -> Domain a -- ^ @d2@
	    -> Bool -- ^ @True@ iff @d1@ is equivalent to @d2@
equivalentD context d1 d2
    | checkPrecondD2 d1 d2 "equivalentD" =
	let l1 = domainToFormula d1
  	    l2 = domainToFormula d2
 	    f  = (qfImpl l1 l2) `QfAnd` (qfImpl l2 l1)
	    (nV, nP) = (nVarsD d1, nParamsD d1)
	    sContext = mkShiftedDomainContext' context nV
        in impliesQfD sContext (nV, nP) f
        -- alternative: includesD context d1 d2 && includesD context d2 d1

-- |
-- Tests if the two domains @d1@ and @d2@ are equivalent.
--
-- If you have a special context, you should prefer the function 'equivalentD'.
--
-- See also: 'equivalentD'
--
equivalentD' :: (BasicDomainElement a)
	     => Domain a -- ^ @d1@
	     -> Domain a -- ^ @d2@
	     -> Bool -- ^ @True@ iff @d1@ is equivalent to @d2@
equivalentD' = equivalentD Nothing


-- |
-- Tests if the domain @d@ is empty under the given context.
-- There are 3 possible results:
--
--   [@QfAt AtTrue@]  @d@ is always empty (for every choice of the parameters)
--
--   [@QfAt AtFalse@] @d@ is never empty (for every choice of the parameters)
--
--   [@cond@]         @d@ is empty, iff the condition @cond@ is satisfied.
--
-- See also: @isEmptyD'@
--
isEmptyD :: (BasicDomainElement a, Integral b)
         => Maybe DomainContext -- ^ the (optional) context
         -> Domain a            -- ^ the domain @d@
	 -> QfFormula (Polynomial (Ratio b))
	 -- ^ the condition for which @d@ is empty
isEmptyD context d
    | checkPrecondD1 d "isEmptyD" =
	let (nV, nP)   = (nVarsD d, nParamsD d)
	    sContext   = mkShiftedDomainContext' context nV
	    nonEmpty   = FoEx [0..nV - 1] (FoQf (domainToFormula d))
	    nonEmptyQf = eliminateQuantifiersD sContext (nV, nP) nonEmpty
        in qfNot nonEmptyQf

-- |
-- Tests if the domain @d@ is empty.
-- There are 3 possible results:
--
--   [@QfAt AtTrue@]  @d@ is always empty (for every choice of the parameters)
--
--   [@QfAt AtFalse@] @d@ is never empty (for every choice of the parameters)
--
--   [@cond@]         @d@ is empty, iff the condition @cond@ is satisfied.
--
-- If you have a special context, you should prefer the function 'isEmptyD'.
--
-- See also: 'isEmptyD'
--
isEmptyD' :: (BasicDomainElement a, Integral c)
          => Domain a -- ^ the domain @d@
	  -> QfFormula (Polynomial (Ratio c))
isEmptyD' = isEmptyD Nothing

-- |
-- Decides whether the domain @d@ is (always) empty under the current
-- context, i.e. means for every choice of the parameters.
--
-- See also: 'isEmptyD', @isAlwaysEmptyD'@
--
isAlwaysEmptyD :: (BasicDomainElement a)
               => Maybe DomainContext -- ^ the (optional) context
               -> Domain a -- ^ the domain @d@
	       -> Bool
isAlwaysEmptyD context d
    | checkPrecondD1 d "isAlwaysEmptyD" =
	let (nV, nP) = (nVarsD d, nParamsD d)
	    sContext = mkShiftedDomainContext' context nV
	in impliesQfD sContext (nV, nP) (isEmptyD context d)

-- |
-- Decides whether the domain @d@ is (always) empty,
-- i.e. means for every choice of the parameters.
--
-- If you have a context which should be considered,
-- you should use the function 'isAlwaysEmptyD'.
--
-- See also: 'isAlwaysEmptyD', 'isEmptyD'
--
isAlwaysEmptyD' :: (BasicDomainElement a)
                => Domain a -- ^ the domain @d@
	        -> Bool
isAlwaysEmptyD' = isAlwaysEmptyD Nothing

-- |
-- Return the image of @is@ under @af@.
-- This is a more generalized version of the function 'image'.
--
imageP :: (Integral a)
       => Maybe DomainContext -- ^ the (optional) context
       -> GAffineFunction (FracPoly (Ratio a)) -- ^ @af@
       -> GIneqSystem (FracPoly (Ratio a))     -- ^ @is@
       -> DTree (Polynomial (Ratio a)) (GIneqSystem (FracPoly (Ratio a)))
imageP context af is
    | nVars af /= nVars is = error ("image: the AffineFunction's " ++
				    "source space dimensionality and " ++
				    "the IneqSystem's nubmer of " ++
				    "variables differ")
    | nParams af /= nParams is = error ("image: AffineFunction and " ++
					"IneqSystem differ in number " ++
					"of parameters")
    | otherwise = fmap ((sortedSystemToIneqSystem nV 0) . fromJust) $
		  projectToVars' is Nothing Nothing [0 .. (targetDimIH af - 1)]
    where
    nV = nVars is -- number of variables

    -- projectToVarsP  should do the "context-shifting"
    projectToVars' =
	projectToVarsPss' fmSortIneqs (fromDomainContext' context) 

    -- The following code is taken from the function 'image'
    -- (which is located in module 'LMath.LinAlg.AffineImage')
    extIS  = ineqSystemVPC (vars, params, consts)
    vars   = (  zeroMatrix (nRows       is) (targetDimIH af)  `conc` iV) `join`
	     ((-unitMatrix (targetDimIH af) (targetDimIH af)) `conc` fV) `join`
	     (  unitMatrix (targetDimIH af) (targetDimIH af)  `conc` (-fV))
    params = iP `join`
	     fP `join` (-fP)
    consts = iC `join`
	     fC `join` (-fC)
    (fV, fP, fC) = partsVPCih af
    (iV, iP, iC) = partsVPC   is

-- |
-- Computes the image of a domain @d@ under an affine mapping @af@.
--
imageD :: (Integral a, Num cond)
       => Maybe DomainContext -- ^ the (optional) context
       -> DTreeContext (Polynomial (Ratio a)) cond -- ^ the 'DTree' context
       -> GAffineFunction (FracPoly (Ratio a)) -- ^ @f@
       -> Domain (FracPoly (Ratio a))          -- ^ @d@
       -> DTree cond (Domain (FracPoly (Ratio a)))
imageD context tree_context af d
    | checkPrecondD1 d "imageD" =
	let dtrees = map (toDomain.(imageP context af)) (domPolys d)
	    toDomain = fmap (ineqToDomain context (nV, nP))
	    (nV, nP) = (nVarsD d, nParamsD d)
	in mergeListOfDTrees tree_context unionD dtrees

-- |
-- Computes the inverse image of a domain @d@ under an affine mapping @af@.
--
inverseImageD :: (Num a, PolyConvertableDomain a)
	      => GAffineFunction a  -- ^ @af@
	      -> Domain a           -- ^ @d@
	      -> Domain a
inverseImageD af d
    | checkPrecondD1_weak d "inverseImageD" =
	let (nV, nP) = (nVarsD d, nParamsD d)
	in initDomainByPolys' (nV, nP) (map (inverseImage af) (domPolys d))

-- |
-- Projects a domain @d@ to the given dimensions using quantifier elimination.
-- This function does not generate a decision tree.
--
-- See also: 'projectToVarsD'
--
projectToVarsD_QE :: (BasicDomainElement a, Integral b)
		  => Maybe DomainContext -- ^ the (optional) context
		  -> [Integer] -- ^ variables to project onto
		  -> Domain a  -- ^ the input domain @d@
		  -> Domain (Polynomial (Ratio b))
projectToVarsD_QE context pVars d
    | checkPrecondD1 d "projectToVarsD_QE" =
	autoSimplD context $
	formulaToDomain context (nV, nP) $
        slfqSimplifyQfD sContext (nV, nP) $
	eliminateQuantifiersD sContext (nV, nP) $
	FoEx ([0..fromIntegral (nV-1)] \\ pVars) $ FoQf $ domainToFormula d
    where
    (nV, nP) = (nVarsD d, nParamsD d)
    sContext = mkShiftedDomainContext' context nV

-- |
-- Projects a domain @d@ to the given dimensions.
-- This is a more generalized version of the function @projectToVarsPss'@.
--
-- The current implementation uses 'projectToVarsD_QE'.
--
-- See also: @LMath.Algo.Sort.projectToVarsPss'@, 'projectToVarsD_QE'
--
projectToVarsD :: (BasicDomainElement a, PolyConvertable a,
		   FracPolyConvertable a, Integral b, Integral c)
               => DTreeContext (Polynomial (Ratio b)) (Polynomial (Ratio b))
	       -- ^ the 'DTreeContext'
               -> Maybe DomainContext -- ^ the (optional) context
	       -> [Integer] -- ^ variables to project onto
               -> Domain a  -- ^ the input domain @d@
               -> DTree (Polynomial (Ratio b)) (Domain (Polynomial (Ratio c)))
projectToVarsD tree_context context pVars d
    | checkPrecondD1 d "projectToVarsD" = DLeaf $ projectToVarsD_QE context pVars d
{-  
    --
    -- Alternative: Use FM to project:
    --
    | checkPrecondD1 d "projectToVarsD" =
	mytrace ("projectToVarsD context=" ++ show context ++
		 ", pVars=" ++ show pVars ++ ", d=" ++ show d) $
	simplifyDTreeD context $
        fmap (toPolyDomain context) $
	mergeListOfDTrees tree_context unionD $
	map (fmap ((ineqToDomain2 Nothing (nV, nP)).(convertIneqFromFracToPoly context).sortedIneqsToIneqs)) $
	mytrace ("xs=" ++ show xs) xs
    where
    xs = mytrace "calculating: map project (domPolys $ toFracPolyDomain d)" $
	 map project (domPolys $ toFracPolyDomain d)

    -- projectToVarsP should do the "context-lifting"
    -- (i.e. shifting the context by the number of variables)
    project is = projectToVarsPss' fmSortIneqs (fromDomainContext' context) is Nothing Nothing pVars

    (nV, nP) = (nVarsD d, nParamsD d)
    sortedIneqsToIneqs (Just x) = sortedSystemToIneqSystem nV x
    sortedIneqsToIneqs Nothing  = head $ domPolys (emptyDomain (nV, nP))

    mytrace message x = --trace message $
			--seq x $
                        x
-}


-- |
-- Test a system for feasibility.
-- Returns a tree (generator) which carries @True@ at
-- a leaf if the system is feasible under the respective context
-- and @False@ otherwise.
--
-- This function is a more generalized version of 'isFeasibleSys'.
-- 'isFeasibleSys' works on only on IneqSystems, where
-- 'isFeasibleSysD' will work on domains.
--
-- See also: 'LMath.Logic.FM.isFeasibleSys'
--
isFeasibleSysD :: (Integral b, Num cond)
               => DTreeContext (FracPoly (Ratio b)) cond
	       -- ^ the DTreeContext for 'FracPoly'
               -> Domain (FracPoly (Ratio b))
               -> DTreeGen (FracPoly (Ratio b)) cond Bool
isFeasibleSysD treeContext d
    | checkPrecondD1 d "isFeasibleSysD" =
	-- TODO: Laesst sich wahrscheinlich noch stark optimieren.
	-- Bsp: d = [is1, is2], wobei is1 immer loesbar ist,
	-- also besteht der Baum nur aus einem Blatt "True"
	-- is2 soll dagegen einen Riesenbaum erzeugen.
	-- Die aktuelle Implementierung wuerde bei beiden Baeume alles mit
	-- allem kombinieren, aber in unserem Fall koennte man evtl. viel
	-- Arbeit sparen, indem man ausnutzt, dass (||) nicht-strikt ist.
	let (nV, nP) = (nVarsD d, nParamsD d)
	    dt_gen = map (\is' -> let is = head $ domPolys (simplifyD' (ineqToDomain' (nV, nP) is'))
			              x = isFeasibleSys is Nothing
			          in mytrace ("calling 'isFeasibleSys is Nothing'\n" ++
					      "where is=" ++ show is) $
			             mytrace ("isFeasibleSys returned x=" ++
					      show (makeDTree treeContext x)) $
			             x
			 ) (domPolys d)
	    mytrace message x = --trace message $
				--seq x $
				x
        in mytrace ("isFeasibleSysD called") $
	   mergeListOfDTreeG (||) dt_gen

-- |
-- Computes a partitioning for the given list of domains.
-- It will return a list of pairs @(d, ids)@
-- where each pair contains the domain @d@ and a list of IDs (@ids@).
--
-- The domain @d@ is one part of the partitioning and
-- the list of IDs contains the IDs of all the input domains
-- which lie with the domain @d@.
--
-- Each of the returned domains is simply connected, i.e. it contains no holes.
-- (In the current implementation all domains are even convex.)
--
-- Notes:
--
--   (1) The ID of a domain is its position in the list of input domains.
--
--   (2) It is recommended that you prefer the "labeled" version of
--       this function, which is named 'lPartitionD'.
--
-- See also: 'lPartitionD', 'DomainID'
--
partitionD :: (BasicDomainElement a, PolyConvertableDomain a,
	       PolyConvertableDomain b)
           => Maybe DomainContext -- ^ the (optional) context
           -> [Domain a] -- ^ the input domains
           -> [(Domain b, [DomainID])] -- ^ the partitioning
partitionD _ [] = error "partitionD: empty list of domains"
partitionD context ds' = mytrace ("partitionD called.\ncontext=" ++ show context) $
                         map (\(x, ids) -> (fromPolyDomain x, ids)) $
                         splitAllNonConnected $
			 rec (unionsD ds) (zip ds [0..])
    where
    -- initial conversion:
    ds = map (toPolyDomain context) ds'
    (nV, nP) = (nVarsD (head ds'), nParamsD (head ds'))

    -- optimization: eliminate empty domains
    rec universe _ | isAlwaysEmptyD context universe = []
    rec universe [] = [(universe, [])] -- base case
    rec universe ((d, dID):rem) =
	let withD    = autoSimplD context $ intersectD          universe d
	    withoutD = autoSimplD context $ complementD context universe d
	    addID (d', ids) = (d', dID:ids)
        in [addID x | x <- rec withD rem] ++ rec withoutD rem

    -- Assure that all returned domains are convex
    -- (simply connected would be enough, but that is hard to test)
    -- Also assure that all domains are split into disjoint polyhedrons.
    splitAllNonConnected = concatMap split
    split (d, ids)
	| containsSinglePolyD d = [(d, ids)] -- nothing to do
	| otherwise = [(initDomainByPolys context (nV, nP) [ineq], ids)
		       | ineq <- disjU]
	where
	disjU = disjointUnionD context (toPolyDomain context d)

    mytrace message x = --trace message $
			--seq x $
			x

-- |
-- Calculates a partitioning for the given list of domains.
-- This is the "labeled" version of the function 'partitionD' and
-- it should be prefered (i.e. use 'lPartitionD' and not 'partitionD').
--
-- Advantage:
-- Each domain is associated with a label of you choice (e.g. a string),
-- whereas 'partitionD' assigns each domain a new 'DomainID'.
--
-- See also: 'partitionD'
--
lPartitionD :: (BasicDomainElement a, PolyConvertableDomain a,
		PolyConvertableDomain b)
            => Maybe DomainContext   -- ^ the (optional) context
            -> [(Domain a, label)]   -- ^ the input domains and their labels
            -> [(Domain b, [label])] -- ^ the partitioning
lPartitionD = liftUnlabeledFuncD2 partitionD

-- |
-- Converts a functions @op@ which operates on 'DomainID's to
-- a function which operates on labels.
--
-- See also: 'liftUnlabeledFuncD2', 'liftUnlabeledFuncDT1'
--
liftUnlabeledFuncD1 :: (Maybe DomainContext -> [a] -> [(b, DomainID)])
                    -- ^ the unlifted operation @op@
		    -> (Maybe DomainContext -> [(a, label)] -> [(b, label)])
		    -- ^ the lifted operation
liftUnlabeledFuncD1 unlabeledOp context lDs =
    let ds  = removeLabel lDs
	ds' = unlabeledOp context ds
    in addLabel (lookupLabel lDs) ds'

-- |
-- Converts a functions @op@ which operates on 'DomainID's to
-- a function which operates on labels.
-- 
-- See also: 'liftUnlabeledFuncD1', 'liftUnlabeledFuncDT2'
--
liftUnlabeledFuncD2 :: (Maybe DomainContext -> [a] -> [(b, [DomainID])])
                    -- ^ the unlifted operation @op@
		    -> (Maybe DomainContext -> [(a, label)] -> [(b, [label])])
		    -- ^ the lifted operation
liftUnlabeledFuncD2 unlabeledOp context lDs =
    let ds  = removeLabel lDs
	ds' = unlabeledOp context ds
    in addLabels (lookupLabel lDs) ds'

-- |
-- Converts a functions @op@ which operates on 'DomainID's to
-- a function which operates on labels.
--
-- See also: 'liftUnlabeledFuncDT2', 'liftUnlabeledFuncD1'
--
liftUnlabeledFuncDT1 :: (Maybe DomainContext -> [a] -> DTree cond [(b, DomainID)])
                     -- ^ the unlifted operation @op@
		     -> (Maybe DomainContext -> [(a, label)] -> DTree cond [(b, label)])
		     -- ^ the lifted operation
liftUnlabeledFuncDT1 unlabeledOp context lDs =
    let ds  = removeLabel lDs
	ds' = unlabeledOp context ds
    in fmap (addLabel (lookupLabel lDs)) ds'

-- |
-- Converts a functions @op@ which operates on 'DomainID's to
-- a function which operates on labels.
--
-- See also: 'liftUnlabeledFuncDT1', 'liftUnlabeledFuncD2'
--
liftUnlabeledFuncDT2 :: (Maybe DomainContext -> [a] -> DTree cond [(b, [DomainID])])
                     -- ^ the unlifted operation @op@
		     -> (Maybe DomainContext -> [(a, label)] -> DTree cond [(b, [label])])
		     -- ^ the lifted operation
liftUnlabeledFuncDT2 unlabeledOp context lDs =
    let ds  = removeLabel lDs
	ds' = unlabeledOp context ds
    in fmap (addLabels (lookupLabel lDs)) ds'

-- |
-- Creates a lookup function which can be used as an argument
-- to the functions 'addLabel' and 'addLabels'.
-- It will just use the 'DomainID' as an index to the original labels,
-- i.e. 0 maps to the first label, 1 maps to the second label, etc.
--
-- See also: 'addLabel', 'addLabels'
--
lookupLabel :: [(a, label)] -- ^ a list of labeled domains
	    -> DomainID     -- ^ the domain ID
	    -> label        -- ^ the original label
lookupLabel lDs id
    | id < 0 || id >= genericLength lDs =
	error ("lookupLabel: index out of bounds: " ++ show id)
    | otherwise = snd $ lDs !! (fromIntegral id)

-- |
-- Removes the labels from a list of labeled domains.
-- (It is nothing more than a simple @map fst@).
--
-- See also: 'addLabel', 'addLabels'
--
removeLabel :: [(a, label)] -- ^ a list of labeled domains
	    -> [a]          -- ^ a list of unlabeled domains
removeLabel = map fst

-- |
-- Transforms a list of domains with IDs to a list of labeled domains.
-- The mapping will be done by the given lookup function.
--
-- See also: 'lookupLabel', 'addLabels'
--
addLabel :: (DomainID -> label) -- ^ the lookup function
	 -> [(a, DomainID)]     -- ^ a list of domains with IDs
	 -> [(a, label)]        -- ^ a list of labeled domains
addLabel lookup xs = [(d, lookup id) | (d,id) <- xs]

-- |
-- Transforms a list of domains with IDs to a list of labeled domains.
-- The mapping will be done by the given lookup function.
--
-- See also: 'lookupLabel', 'addLabel'
--
addLabels :: (DomainID -> label) -- ^ the lookup function
	  -> [(a, [DomainID])]
	  -- ^ a list of domains, where each domain has a list of associated IDs
	  -> [(a, [label])]
	  -- ^ a list of domains, where each domain has a list of associated labels
addLabels lookup xs = [(d, map lookup ids) | (d,ids) <- xs]

-- |
-- Computes the convex hull of a domain @d@.
--
-- Warning:
--   In the most general cases you should not use 'convexHullD' directly,
--   because for more complex domains 'incConvexHullD' seems to be better.
--   'incConvexHullD' should not have any disadvantage
--   compared to 'convexHullD', too.
--
-- See also: 'incConvexHullD'
--
convexHullD :: (Integral a)
            => Maybe DomainContext           -- ^ the (optional) context
            -> Domain (Polynomial (Ratio a)) -- ^ @d@
	    -> Domain (Polynomial (Ratio a)) -- ^ the convex hull of @d@
convexHullD context d'
    | checkPrecondD1 d' "convexHullD" =

        mytrace ("============================================\nconvexHullD called\n============================================\ncontext=" ++ show context ++ "\nd'=" ++ show d') $
	--mytrace ("trying to simplify d' without the context...") $
	--mytrace ("d=" ++ show (simplifyD' d')) $
	--mytrace ("trying to simplify d' using the context...") $
	--mytrace ("d=" ++ show (simplifyD context d')) $

	--
        -- This algorithm is described by Armin in his diploma theses
	-- on page 69 (Theorem 4.5).
	--
	-- It has been slightly extended to support empty polyhedrons,
	-- because the original algorithmus defined the convex hull
	-- in such a way, that it will always return an empty domain
	-- if at least one of the polyhedrons (of the input domain) is empty.
        --
	let -- (We no longer have to remove empty polyhedrons, because the
	    --  the extended algorithm works with empty polyhedrons, too)
	    d  = autoSimplD context d'
	    xs = domPolys d
	    (nV, nP) = (nVarsD d, nParamsD d)

            -- collect information about the variables of the domain
	    -- and convert each (single) polyhedron 'P_i' into a
            -- quantorfree formula 'w_i'
            w = listArray (1, k) (map (extendPolys.ineqSystemToFormula) xs)
	    k = genericLength xs
            n = nV
	    m = nP
	    extendPolys = fmap (setNIndetsD nIndet)

	    -- original variables (of the domain 'd')
	    x, p, r :: Array Integer Integer
	    x  = listArray (1, n) [0..n - 1]
	    p  = listArray (1, m) [n..n+m - 1]
	    -- new variables r={r_1..r_k} and x'={x_11..x_kn}
	    r  = listArray (1, k) [n+m..n+m+k - 1]
	    x' :: Array (Integer, Integer) Integer
	    x' = listArray ((1,1), (k,n)) [n+m+k..n+m+k + k*n - 1]
	    -- total number of variables:
	    -- |{x_1..x_n}| + |{p_1..p_m}| + |{r_1..r_k}| + |{x_11..x_kn}|
	    nOriginalIndet = n + m
	    nExtraIndet    = k + k*n
	    nIndet = nOriginalIndet + nExtraIndet

	    -- the free variables used in the existence quantors are:
	    -- {r_1..r_k} and {x_11..x_kn}
	    exVars = elems r ++ elems x'
	    claim1 = conjunctionQf [(renameQf (w ! i) [(x!j, x'!(i,j)) | j <- [1..n]])
				     `QfOr`
				     (QfAt $ AtEq (poly (r!i))) -- extension: r_i == 0
				    | i <- [1..k]]
	    claim2 = conjunctionQf [QfAt $ AtGe $ poly (r!i)
				    | i <- [1..k]] -- for all i: r_i >= 0
		     `QfAnd`
		     (QfAt ((sum [poly (r!i) | i <- [1..k]]) `isEqual` 1)) -- sum r_i = 1
	    claim3 = conjunctionQf (map (QfAt . sumClaim) [1..n])
            -- helper functions to construct claims1, claim2 and claim3
	    poly :: (Integral a) => Integer -> Polynomial (Ratio a)
	    poly = indeterminate nIndet
	    isEqual :: (Integral a) => Polynomial (Ratio a)
		    -> Polynomial (Ratio a) -> AtFormula (Polynomial (Ratio a))
	    isEqual p1 p2 = AtEq (p1 - p2)
	    sumClaim j = (sum [poly (r!i) * poly (x'!(i,j)) | i <- [1..k]])
                         `isEqual` (poly (x!j))

            -- put all claims together to get a single formula
            -- that describes the convex hull of the domain
	    shiftedContext = mkShiftedDomainContext' context n
	    convFormula = simplifyQfD shiftedContext (nV, nP) $
			  claim1 `QfAnd` claim2 `QfAnd` claim3
	    convHull = FoEx exVars (FoQf convFormula)

	    -- eliminate the existence quantors
	    qfFree = eliminateQuantifiersD shiftedContext (nV, nP) convHull
	    linearQfFree = assertLinearity n $
			   slfqSimplifyQfD shiftedContext (nV, nP) qfFree

        -- simplify the resulting quantifier-free formula
	-- and convert to the domain representation
        in if containsSinglePolyD d
	   then d' -- special case: the domain is already a single polyhedron,
                   --               which is always convex
	   else initDomainByFormula context (nV, nP) linearQfFree
    where
    assertLinearity nV f
	| domain_NDEBUG = f
	| otherwise = 
            if isWeakLinearlyParameterized f nV
	    then f
	    else error ("convexHullD: internal error\n" ++
			"simplifier did not produce a weak linearly " ++
			"parameterized formula:\n" ++ show f)
    mytrace message x = --trace message $
			--seq x $
	                x

    renameQf :: (Integral a)
	     => QfFormula (Polynomial (Ratio a))
	     -> [(Integer, Integer)]
	     -> QfFormula (Polynomial (Ratio a))
    renameQf qfF xs = fmap (renameSeq xs) qfF

    -- Note: The variables are not renamed in parallel, but fortunately
    --       this is not really necessary here.
    renameSeq :: (Integral a)
	      => [(Integer, Integer)]
	      -> Polynomial (Ratio a)
	      -> Polynomial (Ratio a)
    renameSeq [] p = p
    renameSeq (x:xs) p = renameSeq xs (renamePolyVariable x p)

-- |
-- Incremential convex hull algorithm.
--
-- Idea:
-- Do not calculate the convex union of all input polyhedrons at once,
-- but only take two polyhedrons at each pass, until all polyhedrons
-- have been considered.
--
-- It should produce better results than the "standard" algorithm
-- ('convexHullD'), when it is applied to complex domains which
-- contain of more than two polyhedrons.
--
-- See also: 'convexHullD'
--
incConvexHullD :: (Integral a)
	       => Maybe DomainContext           -- ^ the (optional) context
               -> Domain (Polynomial (Ratio a)) -- ^ @d@
	       -> Domain (Polynomial (Ratio a)) -- ^ the convex hull of @d@
incConvexHullD context d' | checkPrecondD1 d' "incConvexHullD" =
    let d = autoSimplD context d'
	(nV, nP) = (nVarsD d, nParamsD d)
        (d1:ds) = map (ineqToDomain context (nV, nP)) (domPolys d)
    in foldl (\x y -> convexHullD context (unionD x y)) d1 ds
    --
    -- Weitere mögliche Optimierungen:
    --
    -- 1) Zuerst alle Polyeder vereinigen, die von der
    --    PolyLib verarbeitet werden koennen.
    --
    -- 2) Divide-and-Conquer wäre wohl noch besser
    --    als der inkrementielle Ansatz
    --


-- |
-- Computes a disjoint union for the input domain @d@.
--
-- Warning:
-- This function does not work for rational (or real) solutions,
-- but only for integral points.
-- (cmp. documentation of 'LMath.Logic.QE.disjointUnion').
--
-- See also: 'disjointUnionD_FracPoly', 'LMath.Logic.QE.disjointUnion'
--
disjointUnionD :: (BasicDomainElement a, FracPolyConvertable a,
		   DomainCanBeSimplified a, Integral b)
               => Maybe DomainContext -- ^ the (optional) context
	       -> Domain a            -- ^ @d@
	       -> [GIneqSystem (Polynomial (Ratio b))]
	       -- ^ a disjoint union of polyhedrons which describe @d@
disjointUnionD context d
  | checkPrecondD1 d "disjointUnionD" =
      let convert = convertIneqFromFracToPoly context
	  d' = simplifyD context d
      in map convert (disjointUnionD_FracPoly context d')

{-
sortedDisjointUnionD_gen ::
  (BasicDomainElement a, FracPolyConvertable a,
   DomainCanBeSimplified a, Integral b, Num cond)
  => Maybe DomainContext
  -> Domain a
  -> DTreeGen (Polynomial (Ratio b)) cond
              [GIneqSystem (Polynomial (Ratio b))]
sortedDisjointUnionD context d
  | checkPrecondD1 d "sortedDisjointUnionD" =
      let polys = disjointUnion context d
	  polysWithID = zip polys [0..]
      in topSortD_qe context 
-}

-- |
-- Computes a disjoint union for the input domain @d@.
--
-- Warning:
-- This function does not work for rational (or real) solutions,
-- but only for integral points.
-- (cmp. documentation of 'LMath.Logic.QE.disjointUnion').
--
-- See also: 'disjointUnionD', 'LMath.Logic.QE.disjointUnion'
--
disjointUnionD_FracPoly :: (BasicDomainElement a, FracPolyConvertable a,
			    Integral b)
                        => Maybe DomainContext -- ^ the (optional) context
	                -> Domain a            -- ^ @d@
	                -> [GIneqSystem (FracPoly (Ratio b))]
disjointUnionD_FracPoly context d
    | checkPrecondD1 d "disjointUnionD_FracPoly" =
	let d' = simplifyD context $ toFracPolyDomain d
            result = disjointUnion (fromDomainContext' context) (domPolys d')
	in {-trace ("disjointUnionD_FracPoly context=" ++ show context ++
		  "\nd:\n" ++ show d) $
	   trace ("disjointUnionD_FracPoly finished:\n" ++ show result) $ -}
	   checkPostcondition result
    where
    (nV, nP) = (nVarsD d, nParamsD d)
    -- assertions: 
    checkPostcondition :: (Integral c)
                       => [GIneqSystem (FracPoly (Ratio c))]
                       -> [GIneqSystem (FracPoly (Ratio c))]
    checkPostcondition ineqs
	| domain_NDEBUG = ineqs -- skip all tests
	| otherwise =
	    let convert = convertIneqFromFracToPoly context
                ds  = [ineqToDomain2 context (nV,nP) x | x <- map convert ineqs]
		dsI = zip ds [0..]
	        isUnion = if null ds
			  then isAlwaysEmptyD context d
			  else equivalentD context
			       (toFracPolyDomain $ unionsD ds)
			       (toFracPolyDomain d)
		allDisjoint =
		    and [isAlwaysEmptyD context (intersectD d1 d2)
			 | (d1,i1) <- dsI, (d2,i2) <- dsI, i1 < i2]
            in if allDisjoint -- do not test "isUnion"
	       then ineqs -- all tests passed
	       else error ("disjointUnionD_FracPoly: postcondition violation" ++
			   "\ncontext=" ++ show context ++
			   ",\n input domain d:\n" ++ show d ++
			   "result:\n" ++ show ineqs ++
			   "\nisUnion=" ++ show isUnion ++
			   ", allDisjoint=" ++ show allDisjoint)

-- |
-- This function tries to convert a domain to a decision tree
-- that has only single polyhedrons at its leaves.
-- The domain @d@ must be a single polyhedron for every choice of
-- its parameters, i.e. it has to be convex.
--
-- If the algorithm was not able to solve the problem,
-- it will return 'Nothing'.
--
-- Warning:
--   This algorithm is highly experimental and we think, that it
--   should not be capable of solving the most general cases.
--
domainToIneqTree :: (Integral a, Num cond)
                 => Maybe DomainContext -- ^ the (optional) context
                 -> DTreeContext (Polynomial (Ratio a)) cond
		 -- ^ the 'DTree' context
                 -> Domain (Polynomial (Ratio a)) -- ^ @d@
		 -> Maybe (DTree cond (GIneqSystem (Polynomial (Ratio a))))
domainToIneqTree optC treeC d =
    let polyD = toPolyDomain optC d
    in fmap (makeDTree treeC) (domainToIneqTree_Gen optC polyD)

--
-- This function tries to convert a domain to a decision tree
-- that has only single polyhedrons at its leaves.
-- The domain @d@ must be a single polyhedron for every choice of
-- its parameters, i.e. it has to be convex.
--
-- If the algorithm was not able to solve the problem,
-- it will return 'Nothing'.
--
-- Warning:
--   This algorithm is highly experimental and we think, that it
--   should not be capable of solving the most general cases.
--
-- See also: 'domainToIneqTree'
--
domainToIneqTree_Gen :: (Num cond, Integral a)
                     => Maybe DomainContext -- ^ the (optional) context
                     -> Domain (Polynomial (Ratio a)) -- ^ @d@
		     -> Maybe (DTreeGen (Polynomial (Ratio a)) cond
			                (GIneqSystem (Polynomial (Ratio a))))
domainToIneqTree_Gen context d =
    fmap (fmap toIneqs) (domainToIneqTree_GenImpl context d)
    where
    toIneqs x = case domainToIneqs x of
		  [y] -> y
		  _   -> error ("domainToIneqTree_Gen: One leaf contains more " ++
				"than one IneqSystem:\n" ++ show x)

--
-- Helper function used by 'domainToIneqTree_Gen'.
--
-- Warning:
--   This algorithm is highly experimental and we think, that it
--   should not be capable of solving the most general cases.
--
-- See also: 'domainToIneqTree', 'domainToIneqTree_Gen'
--
domainToIneqTree_GenImpl :: (Num cond, Integral a)
                         => Maybe DomainContext -- ^ the (optional) context
			 -> Domain (Polynomial (Ratio a)) -- ^ @d@
			 -> Maybe (DTreeGen (Polynomial (Ratio a)) cond
			                    (Domain (Polynomial (Ratio a))))
domainToIneqTree_GenImpl _ d | containsSinglePolyD d =
				 Just (dLeaf d) -- here is nothing to do
domainToIneqTree_GenImpl initContext d =
    mytrace "domainToIneqTree_GenImpl called" $
    mytrace ("initContext=" ++ show initContext) $
    mytrace ("initContext=" ++ show initContext) $
    buildTree initContext (domainToIneqs (simplifyD initContext d))
    where
    mytrace s x = trace s $
		  seq x $
                  x
    -- converts an 'IneqSystem' to a 'DTree Domain' leaf
    leaf context x = dLeaf (initDomainByPolys context (nV, nP) [x])
    (nV, nP) = (nVarsD d, nParamsD d)

    buildTree _ []  = error "domainToIneqTree_GenImpl: internal error (buildTree [])"
    buildTree context [x] = mytrace ("buildTree (2): x=" ++ show x) $
			    Just (leaf context x)
    buildTree context xs =
	mytrace "buildTree (3)" $
	let comb = genericTake (length xs) (iterate (\(y:ys) -> ys ++ [y]) xs)
        in listToMaybe $ catMaybes (map (buildTreeHelper context) comb)
	
    buildTreeHelper context (x:xs) =
	trace "buildTree (4)" $
	let -- convert the single IneqSystem 'x' and the union
            -- of IneqSystems 'xs' to a quantifier-free formula
            sContext = mkShiftedDomainContext' context nV
	    f_x  = simplifyQfD sContext (nV, nP) $ ineqSystemToFormula x
	    f_xs = simplifyQfD sContext (nV, nP) $
		   disjunctionQf (map ineqSystemToFormula xs)
	    -- let the quantifier elimination find out under
	    -- which constraints 'x' implies 'xs'
	    f_impl = FoAll [0..nV-1] (FoQf (qfImpl f_x f_xs))
	    f_cond = mytrace ("buildTreeHelper: calling elimininateQuantifiers") $
		     eliminateQuantifiersD sContext (nV, nP) f_impl
        in mytrace ("f_cond=" ++ show f_cond) $
	   case f_cond of
	     -- if 'x' implies 'xs', we can just return 'x'
	     QfAt AtTrue  -> Just (leaf context x)
	     QfAt AtFalse -> Nothing
	     _ -> let extContext = trace "extending context (1)" $
			           Just $ extendDomainContext' (qfNot f_cond) context
		      recTree    = buildTree extContext xs
                  in case recTree of
		       Nothing  -> Nothing
		       (Just t) -> Just (fCond f_cond (leaf extContext x) t)

{-
domainToIneqTree2 :: (Integral a)
                  => Maybe DomainContext
		  -> DTreeContext (Polynomial (Ratio a)) cond
                  -> Domain (Polynomial (Ratio a))
                  -> DTree cond (GIneqSystem (Polynomial (Ratio a)))
domainToIneqTree2 context treeC d
    | not domain_NDEBUG && not (equivalentD context d (incConvexHullD context d)) =
	error ("domainToIneqTree2: input domain is not convex\nd=" ++ show d)
    | otherwise = makeDTree treeC $
	          case disjointUnionD context d of
                    []  -> dLeaf (head $ domainToIneqs $ emptyDomain (nVarsD d, nParamsD d))
                    [p] -> dLeaf p
                    xs  -> eliminateEmptyPolys xs
    where
    eliminateEmptyPolys [p] = dLeaf []
-}

n_ = indeterminate 2 0
m_ = indeterminate 2 1

testD1 :: Domain (Polynomial Rational)
testD1 = singlePointD (1, 2) [n_]

testD2 :: Domain (Polynomial Rational)
testD2 = singlePointD (1, 2) [m_]

testD3 = incConvexHullD Nothing (testD1 `unionD` testD2)

-- |
-- Converts the domain @d@ to 'Domain (Ratio b)'.
--
-- This function is considered unsafe, because it will fail
-- if the domain @d@ contains any non-constant polynomials!
--
-- See also: 'toRatioDomain_checked', 'toPolyLibIneqsD'
--
toRatioDomain :: (RatioConvertable a, CheckedRatioConvertable a, Integral b)
              => Domain a -- ^ @d@
	      -> Domain (Ratio b) -- ^ @d@ but with rational coefficients
toRatioDomain d
    | domain_NDEBUG = d { domPolys = polys }
    | otherwise = fromJust' $ toRatioDomain_checked d
    where
    polys = [fmap convertToRatio p | p <- domPolys d]
    fromJust' (Just x) = x
    fromJust' Nothing  =
	error "toRatioDomain: domain contained non-constant polynomials"

-- |
-- Converts the domain @d@ to 'Domain (Ratio b)', but returns
-- 'Nothing' if the domain @d@ contains any non-constant polynomials.
--
-- See also: 'toRatioDomain', 'toPolyLibIneqsD'
--
toRatioDomain_checked :: (RatioConvertable a, CheckedRatioConvertable a,
			  Integral b)
		      => Domain a -- ^ @d@
	              -> Maybe (Domain (Ratio b))
		      -- ^ the converted domain or @Nothing@ if the conversion failed
toRatioDomain_checked d =
    if all (\x -> isJust (convertToRatio_checked x)) (domainEntries d)
    then Just $ d { domPolys = polys }
    else Nothing
    where
    polys = [fmap convertToRatio p | p <- domPolys d]

-- |
-- Converts the domain @d@ to 'Domain (FracPoly (Ratio b))'.
--
toFracPolyDomain :: (FracPolyConvertable a, Integral b)
		 => Domain a -- ^ @d@
                 -> Domain (FracPoly (Ratio b)) -- ^ the converted domain
toFracPolyDomain d =
    let polys = [fmap convertToFracPoly p | p <- domPolys d]
    in d { domPolys = polys }

-- |
-- Substitutes constant values for the parameters of the domain @d@.
-- If the value of some parameter is unknown, you can pass 'Nothing'.
--
-- Example:
-- You have a domain @d@ and want to substitute the parameters
-- @u1 = 7@ and @u3 = 4@, but you know nothing about @u0@ and @u2@.
--
-- @substituteParams d [Nothing, Just 7, Nothing, Just 4]@
--
-- See also: 'substituteAllConstantsD'
--
substituteParamsD :: (Integral a)
                  => Domain (Polynomial (Ratio a))  -- ^ @d@
		  -> [Maybe (Ratio a)] -- ^ list of parameters
		  -> Domain (Polynomial (Ratio a))
substituteParamsD d params
    | checkPrecondD1 d "substituteParamsD" =
	if all isNothing params
	then d -- nothing to do (all parameters are unknown)
	else initDomainByPolys' (nV, nP) polys
    where
    (nV, nP) = (nVarsD d, nParamsD d)
    polys = [fmap (replacePoly1 substMonomial) p | p <- domPolys d]

    -- assume all parameters which are not in the "param"-list
    -- to be non-constant (by setting them to Nothing).
    extParams = params ++ repeat Nothing
    -- calculates the polynomial that you get when you
    -- substitute all known paramters in the given monomial.
    substMonomial exps =
	let onePoly = (1,[])
	    (coeff, revExps) = foldl subst onePoly (zip exps extParams)
	    subst (c, xs) (exp, Nothing)  = (c, exp:xs)
	    subst (c, xs) (exp, (Just p)) = (c * (p ^ exp), 0:xs)
	in createPolyFromList [(coeff, reverse revExps)]

-- |		      
-- If the context implies that some parameters are constant,
-- this function tries to determine this constant values and
-- substitute them in the input domain @d@.
--
-- See also: 'substituteParamsD'
--
substituteAllConstantsD :: (Integral a)
                        => Maybe DomainContext -- ^ the context
			-> Domain (Polynomial (Ratio a)) -- ^ @d@
			-> Domain (Polynomial (Ratio a)) -- ^ simplified @d@
substituteAllConstantsD Nothing d = d -- nothing to do
substituteAllConstantsD (Just context) d =
    mytrace ("This is the context: " ++ show context ++
	     "\nThese are the constant parameters:\n" ++ show params) $
    substituteParamsD d params
    where
    (nV, nP) = (nVarsD d, nParamsD d)
    entries' = domainEntries d
    params   = map determineValue [0..nP - 1]
    determineValue n = if paramIsUsed n
		       then paramIsConstant n
		       else Nothing -- if the parameter is not needed,
		                    -- we stop here and save a lot of work
    paramIsUsed n =
	let n' = fromIntegral n
	    containsIt = any (\(_,exps) -> (mytrace ("    containsIt called: exps=" ++ show exps) $
					    if n >= genericLength exps || n < 0
					    then outOfBoundsError exps
					    else exps !! n') /= 0)
	    outOfBoundsError exps =
		error ("substituteAllConstantsD: paramIsUsed: " ++
		       "containsIt: 'n' out of bounds (internal error)\n" ++
		       "n=" ++ show n ++ ", length(exps)=" ++
		       show (length exps))
	    nonConstantEntries = filter (isNothing.isConstPoly) entries'
        in mytrace ("  paramIsUsed called (n=" ++ show n ++ "):\nentries'=" ++ show entries' ++ "\nnonConstantEntries=" ++ show nonConstantEntries) $
	   any (containsIt.polyToList) nonConstantEntries

    -- Idea: Project on each parameter and test whether the defining
    --       set contains only a single point.
    simplContext = simplifyQfD Nothing (0, nP) -- number of variables: 0
		   (conjunctionQf $ fromDomainContext context)
    projContext n = if nP == 1
		    then simplContext -- we have only one dimension,
                                      -- so the projection is not necessary
		    else eliminateQuantifiers [] $
			 FoEx (delete n [0..nP - 1]) (FoQf simplContext)
    paramIsConstant n = case simplifyQf [] (projContext n) of
			  QfAt (AtEq p) -> parseEq n p
			  _ -> Nothing

    -- Test whether the polynomial is of the following form:
    -- "c1*x - c2" where x is the n-th indeterminate and
    -- c1 and c2 are some constant.
    -- If that is true, try to solve the equation "c1*x - c2 = 0"
    -- and return the result. Otherwise return 'Nothing'.
    parseEq n p =
	let totDeg (_, xs) = sum xs
	    isVar (c, exps) =
		c /= 0 && and (zipWith (\x n' -> if n == n'
					         then x == 1
						 else x == 0) exps [0..])
	    isZero (_, exps) = all (==0) exps
	    test v1@(c1, _) v2@(c2, _)
		| isVar v1 && isZero v2 =
		    if c1 == 1
		    then Just (-c2)
		    else if c1 == -1
			 then Just c2
			 else Nothing
	in case polyToList p of
	     [x, y] -> if isZero x
		       then test y x -- 'x' is the candidate for the constant
		       else test x y -- 'y' is the candidate for the constant
             [x]    -> -- there is no candidate for the constant,
	               -- but we can take 0 instead.
		       -- Then we get the equation "c*p = 0" which
		       -- has the solution: "p = 0" (if c /= 0)
		       if isVar x then Just 0 else Nothing
	     _      -> Nothing

    mytrace message x = --trace message $
			--seq x $
                        x

-- |
-- Describes the type of simplifier that is used in the Domain module.
--
-- See also: 'domainSimplifier'
--
data SimplifierType = DefaultSimplifierD
                    | RedlogSimplifierD
		    | SlfqSimplifierD
		    | NoSimplifierD
		    deriving (Eq, Show)

-- |
-- Returns the type of standard simplifier that is used in the Domain module.
--
domainSimplifier :: SimplifierType
domainSimplifier = DefaultSimplifierD
                   --RedlogSimplifierD
                   --SlfqSimplifierD
		   --NoSimplifierD

-- |
-- Controls whether the intermediate formulas, that are generated
-- within this module, should be immediately simplified.
--
-- (Simplification is done by the standard simplifier 'simplifyQfD'.)
--
-- See also: 'simplifyQfD', 'autoSimplD', 'domainSimplifier'
--
simplifyAfterEachStep :: Bool
simplifyAfterEachStep = True

-- |
-- Simplifies the domain @d@, if the flag 'simplifyAfterEachStep' is set.
-- Otherwise it just returns @d@ (without any simplification).
--
-- See also: 'simplifyAfterEachStep', @autoSimpl'@
--
autoSimplD :: (DomainCanBeSimplified a)
           => Maybe DomainContext -- ^ the (optional) context
	   -> Domain a -- ^ @d@
	   -> Domain a -- ^ @d@ or a simplified version of @d@
autoSimplD context d | checkPrecondD1_weak d "autoSimplD" =
			 if simplifyAfterEachStep
			 then simplifyD context d
			 else d -- do nothing

-- |
-- Simplifies the domain @d@ if the flag 'simplifyAfterEachStep' is set.
-- Otherwise it just returns @d@ (without any simplification).
--
-- If you have a non-empty context, you should use 'autoSimplD' instead.
--
-- See also: 'simplifyAfterEachStep', 'autoSimpl'
--
autoSimplD' :: (DomainCanBeSimplified a)
	   => Domain a -- ^ @d@
	   -> Domain a -- ^ @d@ or a simplified version of @d@
autoSimplD' = autoSimplD Nothing

----------------------------------------------------------------------
-- Begin of "lifted" versions of already existing functions, which
-- demand the context to be of the form @[QfFormula (Polynomial (Ratio a))]@.
--
-- Our "lifted" functions demand the context to be of the
-- form 'DomainContext' and 'ShiftedDomainContext'.
----------------------------------------------------------------------

-- |
-- Standard simplifier used within this module.
--
-- It does not make any assumptions about the quality of the simplification,
-- i.e., if you need special guarantees (like 'sqlfSimplifyQfD')
-- you should not use this general function.
--
-- See also: 'domainSimplifier', 'slfqSimplifyQfD'
--
simplifyQfD :: (Integral a, Integral b)
	    => Maybe ShiftedDomainContext       -- ^ the (optional) context
            -> (b, b) -- ^ the number of variables and parameters
            -> QfFormula (Polynomial (Ratio a)) -- ^ the formula to be simplified
	    -> QfFormula (Polynomial (Ratio a)) -- ^ the simplified formula
simplifyQfD sContext' (nV, nP) f
    | nV < 0    = error "simplifyQfD: nV < 0"
    | nP < 0    = error "simplifyQfD: nP < 0"
    | otherwise = simplifier (fromShiftedContext' sContext nV) f
    where
    sContext = -- Nothing
               sContext'
    simplifier = case domainSimplifier of
                   DefaultSimplifierD -> simplifyQf
		   RedlogSimplifierD  -> fastSimplifyQf
		   SlfqSimplifierD    -> slfqSimplifyQf
		   NoSimplifierD      -> (\_ f -> f)
    mytrace message x = --trace message $
	                --seq x $
                        x


-- |
-- slfq-simplifier used within this module.
--
-- If you do not need the guarantees of the slfq-simplifier, you should
-- use the standard simplififier 'simplifyQfD'.
--
-- Warning: Should only be used in the context of the domain module!
--
-- See also: 'simplifyQfD'
--
slfqSimplifyQfD :: (Integral a, Integral b)
                => Maybe ShiftedDomainContext       -- ^ the (optional) context
                -> (b, b) -- ^ the number of variables and parameters
                -> QfFormula (Polynomial (Ratio a)) -- ^ the formula to be simplified
		-> QfFormula (Polynomial (Ratio a)) -- ^ the simplified formula
slfqSimplifyQfD sContext (nV, nP)
    | nV < 0    = error "slfqSimplifyQfD: nV < 0"
    | nP < 0    = error "slfqSimplifyQfD: nP < 0"
    | domainSimplifier == RedlogSimplifierD && contextSeemsTooBigForSlfq =
	logWarningD "Huge context detected. Calling SLFQ with empty context." $
	slfqSimplifyQf [] -- workaround: Slfq fails if the context
			  --             becomes too complicated
    | otherwise = slfqSimplifyQf context
    where
    context = fromShiftedContext' sContext nV
    contextSeemsTooBigForSlfq = length (show context) > 500 -- ugly :-(

-- |
-- Computes a (positive) disjunctive normal form of the given formula @f@.
-- The given context may be used by the implementation to
-- simplify the result (e.g., drop irrelevant disjuncts).
--
-- Warning: Should only be used in the context of the domain module!
--
-- See also: 'LMath.Logic.QE.dnfQf'
--
dnfQfD :: (Integral a, Integral b)
       => Maybe ShiftedDomainContext -- ^ the shifted context
       -> (b, b) -- ^ number of variables and parameters
       -> QfFormula (Polynomial (Ratio a)) -- ^ @f@
       -> [[AtFormula (Polynomial (Ratio a))]]
dnfQfD sContext (nV, nP)
    | nV < 0    = error "dnfQfD: nV < 0"
    | nP < 0    = error "dnfQfD: nP < 0"
    | otherwise = dnfQf (fromShiftedContext' sContext nV)

-- |
-- Decides whether the formulas in a given context @sContext@
-- imply that the formula @f@ holds in the real numbers.
--
-- Warning: Should only be used in the context of the domain module!
--
-- See also: 'LMath.Logic.QE.impliesQf'
--
impliesQfD :: (Integral a, Integral b)
           => Maybe ShiftedDomainContext -- ^ the shifted context @sContext@
           -> (b, b) -- ^ number of variables and parameters
           -> QfFormula (Polynomial (Ratio a)) -- ^ @f@
           -> Bool
impliesQfD sContext (nV, nP)
    | nV < 0    = error "impliesQfD: nV < 0"
    | nP < 0    = error "impliesQfD: nP < 0"
    | otherwise = impliesQf (fromShiftedContext' sContext nV)

-- |
-- Computes a quantifier-free equivalent of a first-order logical formula.
--
-- Warning: Should only be used in the context of the domain module!
--
-- See also: 'LMath.Logic.QE.eliminateQuantifiers'
--
eliminateQuantifiersD :: (Integral a, Integral b)
                      => Maybe ShiftedDomainContext
                      -> (b, b) -- ^ number of variables and parameters
                      -> FoFormula (Polynomial (Ratio a))
                      -> QfFormula (Polynomial (Ratio a))
eliminateQuantifiersD sContext (nV, nP) f
    | nV < 0    = error "eliminateQuantifiers: nV < 0"
    | nP < 0    = error "eliminateQuantifiers: nP < 0"
    | otherwise = eliminateQuantifiers (fromShiftedContext' sContext nV) f

-- |
-- Simplifies a decision tree by cutting irrelevant branches,
-- which cannot be reached under the given context.
--
-- This function will not unify identical branches of the tree.
--
-- See also: 'LMath.Logic.QE.simplifyDTree'
--
simplifyDTreeD :: (Integral a)
               => Maybe DomainContext -- ^ the (optional)context
               -> DTree (Polynomial (Ratio a)) val -- ^ the original DTree
               -> DTree (Polynomial (Ratio a)) val -- ^ the simplified DTree
simplifyDTreeD context = simplifyDTree (fromDomainContext' context)

-- |
-- Converts an 'IneqSystem' of 'FracPoly's to an 'IneqSystem' of 'Polynomial's.
--
-- The important part of the work is done by "unfracIneqSystem"
-- which converts a "GIneqSystem (FracPoly (Ratio a))" into
-- GIneqSystem (Polynomial (Ratio a)).
--
convertIneqFromFracToPoly :: (Integral a, Integral b)
                          => Maybe DomainContext -- ^ the (optional) context
		          -> GIneqSystem (FracPoly (Ratio a))
			  -- ^ the original FracPoly IneqSystem
                          -> GIneqSystem (Polynomial (Ratio b))
			  -- ^ the converted Polynomial IneqSystem
convertIneqFromFracToPoly context sys =
    fmap convertToPoly $ unfracIneqSystem (Just $ impliesSignum context') sys
    where
    -- unfracIneqSystem should do the "context-shifting"
    context' = fromDomainContext' context

----------------------------------------------------------------------
-- End of the "lifted" functions.
----------------------------------------------------------------------


-- |
-- Simplifies each polyhedron of the domain, but does not simplify globally.
--
-- Works only for 'Rational'.
--
simplifyTriviallyD :: Domain Rational -- ^ the domain to be simplified
		   -> Domain Rational
simplifyTriviallyD d
    | checkPrecondD1 d "simplifyTrivially" =
        let (nV, nP) = (nVarsD d, nParamsD d)
	in initDomainByPolys' (nV, nP) (map simplifyTrivially (domPolys d))

-- |
-- Converts a domain to a quantifier-free formula
-- which is in disjunctive normal form.
-- (Each single polyhedron is a conjunction and the list of all polyhedrons
-- is a disjunction of these conjunctions).
--
-- See also: 'formulaToDomain'
--
domainToFormula :: (Integral b)
                => Domain a -- ^ the domain to be converted
                -> QfFormula (Polynomial (Ratio b))
domainToFormula d
    | checkPrecondD1_weak d "domainToFormula" =
        fmap convertToPoly (domFormula d)

-- |
-- Converts a formula @f@ to a domain.
--
-- See also: 'domainToFormula'
--
formulaToDomain :: (PolyConvertableDomain a, Integral b, Integral c)
                => Maybe DomainContext -- ^ the domain's special context
                -> (b, b) -- ^ number of variables and parameters
                -> QfFormula (Polynomial (Ratio c)) -- ^ @f@
                -> Domain a
formulaToDomain context (nV, nP) f = initDomainByFormula context (nV, nP) f

-- |
-- Creates a quantifier-free formula which describes a lexicographic ordering
-- of points of two domains. Both domains must have the same number of
-- variables which have to be passed as a list of indices.
-- @xs@ are the indices for the first domain and @ys@ of the second domain.
-- Additionally a function must be provided which maps the indices
-- on the indeterminates in the resulting formula.
--
-- If you would like to restrict the lexical order to specific dimensions,
-- e.g. consider only the last dimension, you can pass the list of the
-- relevant dimensions. The resulting formula will contain an additional
-- precondition, that all the variables of all ignored dimension are equal.
--
-- If you want to consider all dimensions, just pass 'Nothing'.
--
-- Preconditions:
--
--   (1) @length xs == length ys@
--
-- See also: 'lessEqD'
--
weakLexOrderFormula :: (Integral a, Integral b)
                    => Maybe [a] -- ^ a list of relevant dimensions
                    -> [Integer] -- ^ a list of indices @xs@
                    -> [Integer] -- ^ a list of indices @ys@
                    -> (Integer -> Polynomial (Ratio b))
	            -- ^ a function which maps an index to its corresponding indeterminate
		    -> QfFormula (Polynomial (Ratio b))
weakLexOrderFormula relDims xs ys toVar
    | not domain_NDEBUG && length xs /= length ys =
	error "weakLexOrderFormula: length xs /= length ys"
    | otherwise =
        --trace ("relDims=" ++ show relDims) $
	--trace ("xs=" ++ show xs) $
	--trace ("ys" ++ show ys) $
	case relDims of
	  Nothing -> -- all dimensions are relevant
		     makeLexOrder xs ys initDim
	  Just ds -> -- some dimensions should be ignored by assuming that the
		     -- variables in these dimensions are equal
		     let precond = conjunctionQf
				   [dimEqualFormula x | x <- ignoredDims]
			 ignoredDims = filter (not.isRelevant) [0..genericLength xs - 1]
			 lookupVar n xs
			     | n < 0 || n >= genericLength xs =
				 error "weakLexOrderFormula: lookupVar: 'n' out of bounds"
			     | otherwise = toVar (xs !! (fromIntegral n))
			 dimEqualFormula n =
			     QfAt $ AtEq (lookupVar n xs - lookupVar n ys)
		     in --qfImpl precond (makeLexOrder xs ys 0)
		        makeLexOrder xs ys initDim

    where
    initDim = fromIntegral (checkedHead xs)
    checkedHead [] = error "weakLexOrderFormula: internal error (head failed)"
    checkedHead (x:_) = x
    makeLexOrder [] [] _ = QfAt AtTrue
    makeLexOrder [x] [y] dim =
        if isRelevant dim
	then QfAt $ AtGe (toVar y - toVar x)
        else QfAt AtTrue
    makeLexOrder (x:xs) (y:ys) dim =
	let recFormula = makeLexOrder xs ys (dim+1)
	in if isRelevant dim
	   then (QfAt $ AtGt (toVar y - toVar x))
		`QfOr`
		((QfAt $ AtEq (toVar y - toVar x)) `QfAnd` recFormula)
	   else recFormula
    makeLexOrder _ _ _ = error "weakLexOrderFormula: length xs /= length ys"

    isRelevant n = case relDims of
		     Nothing   -> True
		     (Just ds) -> n `elem` ds

-- |
-- Tests if the domain @d1@ is lexicographically less or equal to
-- the second domain @d2@.
-- It will return a quantifier-free formula for this condition, because
-- in the most general case, the answer will depend on the parameters.
--
-- @dim@ is the first dimension that has not been sorted.
--
-- Preconditions:
--
--   (1) @d1@ and @d@ must be disjoint for every choice of the parameters
--       (but the context will be considered)
--
-- See also: 'strictLessD'
--
lessEqD :: (PolyConvertableDomain a, BasicDomainElement a,
	    Integral b, Integral c)
        => Maybe DomainContext -- ^ the (optional) context
        -> c                   -- ^ the current dimension
 	-> Domain a            -- ^ @d1@
        -> Domain a            -- ^ @d2@
        -> QfFormula (Polynomial (Ratio b))
	-- ^ the condition under which @d1@ is lexicographically before @d@
lessEqD context dim' d1 d2
    | checkPrecondD2 d1 d2 "lessEqD" &&
      checkPrecondAllDisjoint context [d1,d2] "lessEqD" =
	let -- collect information about the variables of the two domains 
            (nV, nP) = (nVarsD d1, nParamsD d1)
	    nV_d1 = [0..nV-1]
	    nV_d2 = map (+nV) nV_d1
	    nIndet = 2*nV + nP
            dim   = fromIntegral dim'

            -- renaming and shifting of the parameters
	    toPoly x | x >= 0 && x < nIndet = indeterminate nIndet x
		     | otherwise = error "lessEqD (toPoly): internal error"
	    extendPolys = mytrace ("extend indeterminates to " ++ show nIndet) $
		          fmap (setNIndetsD nIndet)
	    shiftParams p   = renameQf p
			      (reverse -- "hack": we don't have parallel substitution
                               [(x + nV, x + 2*nV) | x <- [0..nP-1]]
			      )
	    unshiftParams p = renameQf p [(x + 2*nV, x) | x <- [0..nP-1]]
	    shiftedContext  = mkShiftedDomainContext' context (2*nV)

            -- construct quantifier-free formulars which describe
	    -- 1) inD1: the point is in the domain d1
	    -- 2) inD2: the point is in the domain d2
	    -- 3) d1_leq_d2: all points in d1 and d2 that are equal on the
            --               first d dimension are lexicographically
	    --               before all points in d2
            inD1 = shiftParams $ extendPolys $ domainToFormula d1
	    inD2 = let qfF = extendPolys $ domainToFormula d2
                   in renameQf qfF
		      (reverse -- "hack": we don't have parallel substitution
		       [(x, x + nV) | x <- [0..nV + nP - 1]]
		       )
		   
	    d1_leq_d2 = (conjunctionQf [QfAt $ AtEq (toPoly i - toPoly (i + nV))
                                       | i <- [0..dim-1]])
                        `qfImpl`
                        (weakLexOrderFormula (Just [dim..nV-1]) [dim..nV-1] [dim+nV..2*nV-1] toPoly)

	    renameQf :: (Integral a)
		     => QfFormula (Polynomial (Ratio a))
		     -> [(Integer, Integer)]
		     -> QfFormula (Polynomial (Ratio a))
	    renameQf qfF xs = fmap (renameSeq xs) qfF

            -- Warning: The variables are not renamed in parallel!
	    renameSeq :: (Integral a)
		      => [(Integer, Integer)]
	              -> Polynomial (Ratio a)
	              -> Polynomial (Ratio a)
	    renameSeq [] p = p
	    renameSeq (x:xs) p = renameSeq xs (renamePolyVariable x p)

            -- combine the formulas above
	    ltF_without_qf = simplifyQfD shiftedContext (2*nV, nP)  $
			     qfImpl (inD1 `QfAnd` inD2) d1_leq_d2
	    ltFormula = FoAll (nV_d1 ++ nV_d2) $ FoQf $ ltF_without_qf

            -- eliminate all quantifiers, so that the resulting
	    -- formula contains only parameters (or is 'True' or 'False')
            qfFormula = eliminateQuantifiersD shiftedContext (2*nV, nP) ltFormula
	    qfFormulaSimpl = simplifyQfD shiftedContext (2*nV, nP) qfFormula
	    qfFormulaUnshifted = fmap (setNIndetsD nP) $
				 unshiftParams qfFormulaSimpl

            mytrace message x = --trace message $
				--seq x $
				x
	in fmap convertToPoly qfFormulaUnshifted
    

-- |
-- Tests if the domain @d1@ is lexicographically before the second domain @d2@.
-- It will return a quantifier-free formula for this condition, because
-- in the most general case, the answer will depend on the parameters.
--
-- @dim@ is the first dimension that has not been sorted.
--
-- Preconditions:
--
--   (1) @d1@ and @d@ must be disjoint for every choice of the parameters
--       (but the context will be considered)
--
-- See also: 'lessEqD'
--
strictLessD :: (PolyConvertableDomain a, BasicDomainElement a,
		Integral b, Integral c)
            => Maybe DomainContext -- ^ the (optional) context
            -> c                   -- ^ the current dimension
 	    -> Domain a            -- ^ @d1@
            -> Domain a            -- ^ @d2@
            -> QfFormula (Polynomial (Ratio b))
strictLessD context dim d1 d2 = qfNot $ lessEqD context dim d2 d1

-- |
-- Topologically sorts a list of domains where each domain has an unique ID.
-- The result is a decision tree where all leaves contains a list of
-- the sorted domains. This list is a pair of the domain and its original ID.
--
-- @dim@ is the first dimension that has not been sorted.
--
-- Preconditions:
--
--   (1) All input domains must be disjoint (under the given context)
--
-- See also: 'topSortD', 'DomainID'
--
topSortD_qe :: (BasicDomainElement a, PolyConvertableDomain a,
                DomainCanBeSimplified a, Integral b, Integral c, Num cond)
	    => Maybe DomainContext -- ^ the (optional) context
            -> c -- ^ the current dimension
	    -> [(Domain a, DomainID)]
	    -- ^ a (labeled) list of domains to be sorted
	    -> DTreeGen (Polynomial (Ratio b)) cond [(Domain a, DomainID)]
topSortD_qe initContext dim ds
    | not domain_NDEBUG && sort (map snd ds) /= [0..genericLength ds - 1] =
	error "topSortD_qe: IDs must be unique number between [0..length ds-1]"
    | otherwise =
    let mkGraphTree context [] edgeGraph =
	    dLeaf $ extractEdgesFromGraph edgeGraph
	mkGraphTree context (edge@((d1,id1), (d2,id2)) : xs) edgeGraph' =
	    --
	    -- four possibilities:
	    -- 1) x <= y &&     y <= x     -> no edge (uncomparable)
	    -- 2) x <= y && not y <= x     -> add edge from x to y
	    -- 3) not x <= y && y <= x     -> add edge from y to x
	    -- 4) not x <= y && not y <= x -> error
	    let ex_XY = lessEqD (if useContext then Just context else Nothing) dim d1 d2
		useContext = True
		case_1_or_2 =
		    let context' = extendDomainContext ex_XY context
		        ex_YX  = lessEqD (if useContext then Just context' else Nothing) dim d2 d1
			case_1 = mkGraphTree (extendDomainContext ex_YX context') xs
				 (removeEdge (id1,id2) edgeGraph)
			case_2 = mkGraphTree (extendDomainContext (qfNot ex_YX) context') xs
				 (addEdge (id1,id2) edgeGraph)
		    in fCond ex_YX case_1 case_2
		case_3_or_4 =
		    let context' = extendDomainContext (qfNot ex_XY) context
			ex_YX  = lessEqD (if useContext then Just context' else Nothing) dim d2 d1
			case_3 = mkGraphTree (extendDomainContext ex_YX context') xs
				 (addEdge (id2,id1) edgeGraph)
			case_4 = mkGraphTree (extendDomainContext (qfNot ex_YX) context') xs
				 edgeGraph
		    in if domain_NDEBUG || impliesQf (fromDomainContext context') ex_YX
		       then case_3
		       else fCond ex_YX case_3 case_4
                edgeGraph = removeEdgesWhichLeadToCycles edgeGraph'
	    in case edgeGraph' ! (id1,id2) of
                      Just True  -> mytrace ("topSortD_qe: transitive edge found!") $
				    mkGraphTree context xs edgeGraph
		      Just False -> mytrace ("topSortD_qe: transitive anti-edge found!") $
				    mkGraphTree context xs edgeGraph
		      Nothing    -> fCond ex_XY case_1_or_2 case_3_or_4
	eqIDs (_,id1) (_,id2) = id1 == id2
	topSortLeaves edges =
	    case topSort' eqIDs edges of
	      Just order -> nubBy eqIDs (order ++ ds)
	      Nothing    -> error "topSortD_qe: internal error (cycles found)"
        edgesCandidates = [(x,y) | x <- ds, y <- ds, snd x < snd y]
	initContext' = maybe emptyDomainContext id initContext

        -- edge graph to find the transitive closure:
	l = genericLength ds
	emptyEdgeGraph = array ((0,0), (l-1,l-1))
			 [((x,y), if x == y then Just False else Nothing)
			  | x <- [0..l-1], y <- [0..l-1]]
        addEdge (n1, n2) edgeGraph =
	    computeClosure $
	    edgeGraph // [((n1,n2), Just True), ((n2,n1), Just False)]
	removeEdge (n1,n2) edgeGraph =
	    edgeGraph // [((n1,n2), Just False)]
	extractEdgesFromGraph edgeGraph =
	    [((d1, n1), (d2, n2)) |
	     (d1, n1) <- ds, (d2, n2) <- ds, edgeGraph ! (n1,n2) == Just True]
	computeClosure edgeGraph =
	    let transitiveEdges =
		    [(n1,n2) | n1 <- [0..l-1], n2 <- [0..l-1],
                               edgeGraph ! (n1,n2) == Nothing,
			       x <- [0..l-1], edgeGraph ! (n1,x) == Just True,
                               edgeGraph ! (x,n2) == Just True]
	    in case transitiveEdges of
		 (e:_) -> addEdge e edgeGraph
		 []    -> edgeGraph
	removeEdgesWhichLeadToCycles edgeGraph =
	    let edges = [(fromInteger x, fromInteger y)
			 | ((_,x),(_,y)) <- extractEdgesFromGraph edgeGraph]
		removeable = [((n1, n2), Just False)
			      | n1 <- [0..l-1], n2 <- [0..l-1],
			      edgeGraph ! (n1,n2) == Nothing,
			      topSort ((fromInteger n1, fromInteger n2):edges) (fromIntegral l) == Nothing
			     ]
            in edgeGraph // removeable

        mytrace message x = --trace message $
	                    --seq x $
                            x

    in fmap topSortLeaves $
       mkGraphTree initContext' edgesCandidates emptyEdgeGraph

-- |
-- Topologically sorts a list of domains. The result is a decision tree
-- where all leaves contains a list of the sorted domains.
-- This list is a pair of the domain and its ID, which is its original
-- position in the input list.
--
-- @dim@ is the first dimension that has not been sorted.
--
-- Preconditions:
--
--   (1) All input domains must be disjoint (under the given context)
--
-- See also: 'lTopSortD', 'topSortD_mergesort', 'DomainID'
--
topSortD :: (BasicDomainElement a, PolyConvertableDomain a,
	     DomainCanBeSimplified a, Integral b, Integral c)
         => Maybe DomainContext -- ^ the (optional) context
	 -> c                   -- ^ the current dimension
         -> [Domain a]          -- ^ a list of domains to be sorted
         -> DTree (Polynomial (Ratio b)) [(Domain a, DomainID)]
topSortD context dim ds
    | all (\x -> checkPrecondD1 x "topSortD") ds &&
      checkPrecondAllDisjoint context ds "topSortD" =
	let tree_context     = polyDTreeContext (fromDomainContext' context)
	    mergesort_result = topSortD_qe context dim (zip ds [0..])
        in assertSorted $
	   simplifyDTree (fromDomainContext' context) $
	   makeDTree tree_context mergesort_result
    where
    assertSorted dt
	| domain_NDEBUG = dt
	| any (not . isPermutation) [map snd xs | xs <- leavesDTree dt] =
	    error "topSortD: internal error (sorting condition failed)"
	| otherwise = dt
    isPermutation xs = sort (nub xs) == [0..genericLength ds - 1]

-- |
-- Topologically sorts a list of of domains.  The result is a decision tree
-- where all leaves contains a list of the sorted domains.
-- This list is a pair of the domain and its ID, which is its original
-- position in the input list.
--
-- @dim@ is the first dimension that has not been sorted.
--
-- This is the "labeled" version of the function 'topSortD' and 
-- it should be prefered (i.e. use 'lTopSortD' and not 'topSortD').
--
-- Advantage:
-- Each domain is associated with a label of you choice (e.g. a string),
-- whereas 'topSortD' assigns each domain a new 'DomainID'.
--
-- Preconditions:
--
--   (1) All input domains must be disjoint (under the given context)
--
-- See also: 'topSortD'
--
lTopSortD :: (BasicDomainElement a, PolyConvertableDomain a,
	      DomainCanBeSimplified a, Integral b, Integral c)
          => Maybe DomainContext -- ^ the (optional) context
	  -> c                   -- ^ the current dimension
          -> [(Domain a, label)] -- ^ a list of domains to be sorted
          -> DTree (Polynomial (Ratio b)) [(Domain a, label)]
lTopSortD context dim =
    liftUnlabeledFuncDT1 (\c -> topSortD c dim) context

-- |
-- Helper function which tests, if the formula @f@ is weak linearly
-- parameterized, i.e. all variables must be linear, but the
-- parameters may appear as non-linear polynomials.
--
-- Convention: The first @nV@ indeterminates are variables
-- and the remaining indeterminates are parameters.
--
isWeakLinearlyParameterized :: Integral b
                            => QfFormula (Polynomial a) -- ^ @f@
                            -> b -- ^ the number of variables @nV@
			    -> Bool
isWeakLinearlyParameterized f nV
    | nV < 0 = error "isWeakLinearlyParameterized: nV < 0"
    | otherwise = let terms       = map polyToList (termsQf f)
		      extractVars = genericTake nV
		      isLinear xs = sum (extractVars xs) <= 1
		  in all (all (\(_, xs) -> isLinear xs)) terms

-- |
-- Tests if all parameters of the domain @d@ are only linearly
-- parameterized there are only univariate polynomials.
-- 
-- If that is true, they can be handled by the PolyLib library.
--
-- See also: 'toPolyLibIneqsD'
--
canBeHandledByPolyLib :: (BasicDomainElement a, PolyConvertableDomain a)
                      => Domain a -- ^ @d@
		      -> Bool
canBeHandledByPolyLib d =
    let isOK = isJust $ toPolyLibIneqsD_checked d
        regressionTestFailed = isOK && not regression
    in if not domain_NDEBUG && regressionTestFailed
       then error ("canBeHandledByPolyLib: internal error\nd=" ++ show d)
       else isOK
    where
    regression = let d' = toPolyDomain' d
		     entries = map polyToList (domainEntries d')
		     isLinear xs = sum xs <= 1
		 in domParams d == 0 ||
		    all (all (\(_, xs) -> isLinear xs)) entries

-- |
-- Converts the domain @d@ to a list of IneqSystems which contains
-- only rational coefficients. This is only possible if all
-- parameters of @d@ are only linearly parameterized and
-- there are only univariate polynomials (cmp. precondition).
--
-- Preconditions:
--
--   (1) @canBeHandledByPolyLib d == True@
--
-- See also: 'canBeHandledByPolyLib'
--
toPolyLibIneqsD :: (BasicDomainElement a, PolyConvertableDomain a, Integral b)
                => Domain a -- ^ @d@
                -> [GIneqSystem (Ratio b)]
toPolyLibIneqsD d =
    case toPolyLibIneqsD_checked d of
      Just ineqs -> ineqs
      Nothing -> error ("toPolyLibIneqsD: the input domain d cannot be " ++
			"converted to a PolyLib compatible representation:\n" ++
			show d)

-- |
-- Converts the domain @d@ to a list of IneqSystems which contains
-- only rational coefficients. This is only possible if all
-- parameters of @d@ are only linearly parameterized and
-- there are only univariate polynomials.
-- Otherwise the function will return 'Nothing'.
--
-- See also: 'canBeHandledByPolyLib', 'toPolyLibIneqsD'
--
toPolyLibIneqsD_checked :: (BasicDomainElement a, PolyConvertableDomain a,
			    Integral b)
			=> Domain a -- ^ @d@
                        -> Maybe [GIneqSystem (Ratio b)]
toPolyLibIneqsD_checked d
    | checkPrecondD1 d "toPolyLibIneqsD_checked" =
	case () of
	  _ | nP == 0 -> Just $ domPolys $ toRatioDomain $ toPolyDomain' d
          _ | isOK -> -- generalize the result to [GIneqSystem (Ratio c)]
		      Just [fmap convertToRatio ineq | (_, ineq) <- convertedIneqs]
	  _ | otherwise -> Nothing
    where
    nP = nParamsD d
    nV = nVarsD d

    -- convert each polyhedron and return a pair (ok, ineq) where
    --   'ok':   if polyhedron can be handled by PolyLib
    --   'ineq': the PolyLib compatible IneqSystem (only defined if 'ok')
    convertedIneqs = map toPolyLibIneqD_impl (domPolys (toPolyDomain' d))
    isOK = all fst convertedIneqs -- all conversions must be OK

    toPolyLibIneqD_impl is = let convRows = map convertRow (rows is)
				 conversionOK = all isJust convRows
				 rowMatrices = map (rowMatrix.fromJust) convRows
			         mEntries = multijoin emptyMatrix rowMatrices
			     in (conversionOK, ineqSystem ((nV, nP), mEntries))
    convertRow row = let (varsE', [constantE]) = splitAt nV (entries row)
			 paramsE = uncompressParam nP constantE
			 varsE = map convertToRatio_checked varsE'
	             in if all isJust varsE
			then Just (map fromJust varsE ++ paramsE)
			else Nothing
    emptyMatrix = zeroMatrix 0 (nV + nP + 1)

-- |
-- Helper function used by 'toPolyLibIneqsD_checked'.
--
-- @p@ must be an univariate polynomial.
--
uncompressParam :: (Integral a, Integral b, Integral c)
                => a -- ^ the number of parameters
                -> (Polynomial (Ratio b)) -- ^ the polynomial @p@
                -> [Ratio c] -- ^ a list of parameters with rational coefficients
uncompressParam nP p
    | nP < 0 = error "uncompressParam: nP < 0"
    | otherwise = sumRows $ map conv (polyToList p)
    where
    nP' :: Int
    nP' = fromIntegral nP

    -- In this function we have assumed that the input parameter p
    -- is an univariate polynomial.
    -- So every parameter in p can appear only linearly, or p is a constant.
    conv :: (RatioConvertable a, Integral b) => (a,[b]) -> [Ratio b]
    conv (c, xs) = let pos = case filter (\(x,_) -> x == 1) (zip xs [0..]) of
			       [(_,x)] -> if x >= nP'
					  then errOutOfBounds
					  else x
			       [] -> nP' -- the "1" columns is at the last position
			       _ -> errMulti
                       c' = convertToRatio c
                   in if any (>=2) xs
		      then errMulti
		      else [if x == pos then c' else 0 | x <- [0..nP']]
    sumRows = foldl (zipWith (+)) zeroRow
    zeroRow = replicate (nP' + 1) 0
    errMulti = error ("uncompressParam: p is an multivariate polynomial, but " ++
	              "this function can only deal with univariate polynomials." ++
                      "\np=" ++ show p)
    errOutOfBounds = error ("uncompressParam: the input polynomial p " ++
		            "contains more parameters than expected." ++
			    "\np=" ++ show p ++ "\nnP=" ++ show nP)

-- |
-- If a fast PolyLib implementation is available, then use it.
-- This will of course only work in the linear case.
-- Otherwise fall back to our own (non-linear) implementation.
--
-- This version of the 'tryPolyLibx' functions deals with
-- operations which have one argument.
--
-- See also: 'tryPolyLib2'
--
tryPolyLib1 :: (BasicDomainElement a, PolyConvertableDomain a)
            => (Domain a -> Maybe b) -- ^ the polylib implementation
            -> (Domain a -> b)       -- ^ our own implementation
            -> Domain a -- ^ the input domain @d@
            -> b
tryPolyLib1 polyImpl ownImpl d
    | canBeHandledByPolyLib d =
	mytrace "PolyLib implementation was chosen" $
	case polylib_result of
          (Just x) -> x
	  Nothing  -> logWarningD ("tryPolyLib1: PolyLib failed. " ++
				   "Falling back to default implementation.") $
		      own_result
    | otherwise = own_result
    where
    polylib_result = polyImpl d
    own_result     = ownImpl d
    mytrace message x = -- trace message
			x

-- |
-- If a fast PolyLib implementation is available, then use it.
-- This will of course only work in the linear case.
-- Otherwise fall back to our own (non-linear) implementation.
--
-- This version of the 'tryPolyLibx' functions deals with
-- operations which have two arguments.
--
-- See also: 'tryPolyLib1'
--
tryPolyLib2 :: (BasicDomainElement a, PolyConvertableDomain a)
            => (Domain a -> Domain a -> Maybe b) -- ^ the polylib implementation
            -> (Domain a -> Domain a -> b)       -- ^ our own implementation
            -> Domain a -- ^ the first input domain @d1@
            -> Domain a -- ^ the second input domain @d2@
            -> b
tryPolyLib2 polyImpl ownImpl d1 d2
    | canBeHandledByPolyLib d1 && canBeHandledByPolyLib d2 =
	mytrace "polylib implementation was chosen" $
	case polylib_result of
	  (Just x) -> x
	  Nothing  -> logWarningD ("tryPolyLib2: PolyLib failed. " ++
				   "Falling back to default implementation.") $
		      own_result
    | otherwise = own_result
    where
    polylib_result = polyImpl d1 d2
    own_result     = ownImpl  d1 d2
    mytrace message x = --trace message $
			--seq x $
			x

-- |
-- So koennte eine Funktion aussehen, die falls moeglich
-- die PolyLib Funktion benutzt, und anderenfalls
-- auf die Domain Implementierung zurueckgreift.
--
equivalent_test :: (BasicDomainElement a, PolyConvertableDomain a)
                => Domain a
		-> Domain a
		-> Bool
equivalent_test = tryPolyLib2 equivalent_PolyLib equivalentD'

-- |
-- Versucht, die beiden Eingabe-Domains in PolyLib kompatible
-- Darstellung zu bringen, und falls dies erfolgreich war,
-- die PolyLib Funktion aufzurufen.
--
-- Andernfalls wird Nothing zurueckgegeben, um zu signalisieren,
-- dass man die Domain Implementierung benutzen soll.
--
equivalent_PolyLib :: (BasicDomainElement a, PolyConvertableDomain a)
                   => Domain a
                   -> Domain a
                   -> Maybe Bool
equivalent_PolyLib d1 d2 =
    case (toPolyLibIneqsD d1, toPolyLibIneqsD d2) of
      ([x], [y]) -> Just (areIneqsEquivalent x y Nothing)
      _ -> Nothing -- this implementation cannot handle domains,
                   -- which contain more than one polyhedron.

-- |
-- Debugging function. Can be used to display the results of finished functions.
--
verbose :: Show a => String -> a -> a
verbose str x
    | skipVerbose = x
    | otherwise = seq x $ trace ("Verbose: " ++ str ++ ": result=" ++ show x) x

skipVerbose = False


--
-- let d2 = convexHullD Nothing $ (unionsD [singlePointD nV_ x | x <- [[0,0], [0,indeterminate 2 0], [indeterminate 2 1, 0]]])::(Domain (Polynomial Rational))
--
-- let p1 = head (domainToIneqs d2)
--
-- ineqSystemToAtFormulas p1
--
----------------------------------------------------------------------
--
-- ineqSystemToFormula $ ineqSystem ((0,0), matrix [[(indeterminate 2 1 * indeterminate 2 0)::Polynomial Rational]])
--
----------------------------------------------------------------------
--
-- simplifyD' $ convexHullD Nothing $ (unionsD [singlePointD nV_ x | x <- [[0,0], [0,indeterminate 2 0], [indeterminate 2 1, 0]]])::(Domain (Polynomial Rational))
--
