module LMath.Logic.DomainBenchmarks
    (main, runBenchmarks, showElapsedTime, printElapsedTime,
     stopElapsedPicoseconds, stopElapsedPicoseconds'
    )
where

import Monad
import Debug.Trace
import Time

import LMath.Logic.QE
import LMath.Logic.Sort
import LMath.Tools
import LMath.Types

import LMath.Logic.Domain
import LMath.Logic.DomainExamples

type Benchmark = IO ()

verbose :: Bool
verbose = False

mytrace :: String -> b -> b
mytrace message x = if verbose
		    then trace message x
		    else x

main :: IO ()
main = do
       putStrLn "0 parameters"
       runBenchmarks (benchmarkCollection 0)
       putStrLn "1 parameter"
       runBenchmarks (benchmarkCollection 1)
       return ()

benchmarkCollection :: Integer -> [Benchmark]
benchmarkCollection p =
    --forceEvaluation $
    [
     bmTemplate "convexHullD"                 (convexHullD Nothing)    conv
     ,bmTemplate "convexHullD (simplified)"   (convexHullD Nothing)    simpl
     ,bmTemplate "incConvexHullD"             (incConvexHullD Nothing) conv
    ,bmTemplate "incConvexHullD (simplified)" (incConvexHullD Nothing) simpl
    ,bmTemplate "convexUnion (Polylib)"        convexUnionAll     (map fst raw)
    ,bmTemplate2 "includesD"                   includesD'         conv2
    ,bmTemplate2 "includesD (simplified)"      includesD'         simpl2
    ,bmTemplate2 "isSuperset (Polylib)"        isSuperset         (map fst raw2)
    ]
    where
    raw   = case p of
	    0 -> testDomains_raw_p0
	    1 -> testDomains_raw_p1
            _ -> error "benchmarkCollection: p invalid"
    conv  = case p of
	    0 -> testDomains_conv_p0
	    1 -> testDomains_conv_p1
            _ -> error "benchmarkCollection: p invalid"
    simpl = case p of
	    0 -> testDomains_simpl_p0
	    1 -> testDomains_simpl_p1
            _ -> error "benchmarkCollection: p invalid"

    raw2   = let z' = take 10 [((x,y), nVP) | (z, nVP) <- raw,  x <- z, y <- z]
             in mytrace ("length z': " ++ show (length z')) z'
    conv2  = map (\((x,y), nVP) -> (domain2' nVP [x], domain2' nVP [y])) raw2
    simpl2 = map (\(x,y) -> (simplifyD' x, simplifyD' y)) conv2

    forceEvaluation x = if show simpl2 /= show simpl2 ||
                           show simpl  /= show simpl
		        then error "benchmarkCollection: internal error!"
		        else x

runBenchmarks :: [Benchmark] -> IO Integer
runBenchmarks xs =
    do
    putStrLn "Initializing benchmarks..."
    (diff, results) <- stopElapsedPicoseconds (mapM runSingleBenchmark xs)
    putStrLn "\nAll benchmarks have finished."
    printElapsedTime diff
    return diff

runSingleBenchmark :: Benchmark -> IO Integer
runSingleBenchmark bm =
    do
    -- "bm": runs the specific benchmark
    diff_pico <- stopElapsedPicoseconds' bm
    putStrLn ("done (" ++ showElapsedTime diff_pico ++ ")")
    return diff_pico

stopElapsedPicoseconds :: IO a -> IO (Integer, a)
stopElapsedPicoseconds action =
    do
    t1 <- getClockTime
    r  <- action
    t2 <- getClockTime
    let diffT     = diffClockTimes t2 t1
	diff_day  = 24*60*60*1000*1000000000 * (fromIntegral $ tdDay diffT)
	diff_hour = 60*60*1000*1000000000 * (fromIntegral $ tdHour diffT)
	diff_min  = 60*1000*1000000000 * (fromIntegral $ tdMin diffT)
	diff_sec  = 1000*1000000000 * (fromIntegral $ tdSec diffT)
        diff_pico = tdPicosec diffT
    return (diff_day + diff_hour + diff_min + diff_sec + diff_pico, r)

stopElapsedPicoseconds' :: IO a -> IO Integer
stopElapsedPicoseconds' action = do (diff, _) <- stopElapsedPicoseconds action
				    return diff

showElapsedTime :: Integer -> String
showElapsedTime t_pico
    | t_pico < 0 = error "showElapsedTime: t_pico < 0"
    | otherwise = let t_millis = t_pico `div` 1000000000
	          in if t_millis > 1000
		     then show (t_millis `div` 1000) ++ " sec, " ++ show (t_millis `mod` 1000) ++ " ms"
		     else show t_millis ++ " ms"

printElapsedTime :: Integer -> IO ()
printElapsedTime t_pico =
    putStrLn ("Elapsed time: " ++ showElapsedTime t_pico)

-- 0 parameters
testDomains_raw_p0 :: [([GIneqSystem Rational], (Integer, Integer))]
testDomains_raw_p0 =
    [(c [testPoly1, testPoly2, testPoly3], (2,0))
    ,(c [testPoly2, testPoly3, testPoly4], (2,0))
    ,(c [testPoly3, testPoly5], (2,0))
    ,([testPolyPoint [2,3,4,5], testPolyPoint [1,2,5,-2]], (4,0))
    ,([testPolyPoint [2,3,5,6], testPolyPoint [1,2,-2,7], testPolyPoint [1,3,4,8]], (4,0))
    ]
    where
    c = map (fmap convertToRatio)

testDomains_conv_p0 :: [Domain (Polynomial Rational)]
testDomains_conv_p0 = map (\(xs, nVP) -> domain2' nVP xs) testDomains_raw_p0

testDomains_simpl_p0 :: [Domain (Polynomial Rational)]
testDomains_simpl_p0 = map simplifyD' testDomains_conv_p0

-- 1 parameter
testDomains_raw_p1 :: [([GIneqSystem Rational], (Integer, Integer))]
testDomains_raw_p1 =
    [(c [testPoly6, testPoly7], (2,1))
    ,(c [testPoly9, testPoly8], (2,1))
    ]
    where
    c = map (fmap convertToRatio)

testDomains_conv_p1 :: [Domain (Polynomial Rational)]
testDomains_conv_p1 = map (\(xs, nVP) -> domain2' nVP xs) testDomains_raw_p1

testDomains_simpl_p1 :: [Domain (Polynomial Rational)]
testDomains_simpl_p1 = map simplifyD' testDomains_conv_p1

bmTemplate :: (Show b) => String -> (a -> b) -> [a] -> Benchmark
bmTemplate opName op input =
    do
    putStr (opName ++ "...")
    let xs = map op input
    when (show xs /= show xs) (error "bmTemplate: Internal error!")
    when verbose (print xs)

bmTemplate2 :: (Show c) => String -> (a -> b -> c) -> [(a,b)] -> Benchmark
bmTemplate2 opName op input =
    do
    putStr (opName ++ "...")
    let xs = map (uncurry op) input
    when (show xs /= show xs) (error "bmTemplate: Internal error!")
    when verbose (print xs)
