-- |
-- This module contains the representation of the context as it is used
-- within the domain modules.
--
-- Usage: In most of the cases you just have to convert the standard
-- representation of the context (@[QfFormula (Polynomial (Ratio a))]@)
-- to @Maybe DomainContext@ by using the function 'mkDC'.
--
-- See also: 'LMath.Logic.Domain', 'DomainContext', 'mkDC'
--
module LMath.Logic.DomainContext
    ( -- * Constructors
      DomainContext, mkDomainContext, mkDC, emptyDomainContext,
      ShiftedDomainContext, mkShiftedDomainContext, mkShiftedDomainContext',

      -- * Accessors
      fromDomainContext, fromDomainContext',
      fromShiftedContext, fromShiftedContext',

      -- * Operations
      extendDomainContext, extendDomainContext',

      -- * Extended operations
      heuristic_DC_implies_DC, heuristic_DC_implies_DC',

      -- * Helper functions
      generalizeContext, shiftContext
    )
where

import Data.Array
import Data.List
import Data.Maybe
import Data.Ratio

import LMath.Types
import LMath.Logic.QE

-- |
-- The standard domain context. You should create a domain context
-- by using the function 'mkDC' which returns @Maybe DomainContext@.
--
-- See also: 'ShiftedDomainContext', 'mkDC', 'fromDomainContext'
--
data DomainContext =
    DomainContext { dcContext    :: [QfFormula (Polynomial Rational)]
		    -- ^ the context in its usual representation
		  , dcShiftedCxt :: Int -> [QfFormula (Polynomial Rational)]
		    -- ^ a function which returns a shifted context
		  }

-- |
-- The shifted domain context, i.e. all parameters have been shifted
-- to the right (e.g. u0 -> u2, u1 -> u3, u2 -> u4, ...).
-- The shifted context will remember the number of variables and
-- cannot be used with different values. This should help to avoid some bugs,
-- because the number of variables should be constant in most situations
-- inside the domain module.
--
-- This representation is only used deep within the domain module.
--
-- See also: 'DomainContext', 'mkShiftedDomainContext', 'fromShiftedContext'
--
data ShiftedDomainContext = ShiftedDomainContext Int DomainContext


instance Show DomainContext where
    show = show . dcContext

instance Show ShiftedDomainContext where
    show (ShiftedDomainContext nV context) =
	"ShiftedDomainContext: nV = " ++ show nV ++
        ", context = " ++ show context

-- |
-- Converts the standard representation of the context
-- (as it is used in all other parts of hsloopo) into the
-- 'DomainContext' representation.
--
-- See also: 'mkDomainContext'
--
mkDomainContext :: (Integral a)
                => [QfFormula (Polynomial (Ratio a))]
		-- ^ the context in the standard representation
                -> DomainContext
		-- ^ the context in the domain representation
mkDomainContext context =
    let context' = generalizeContext context
	shiftFun nV | nV < 0 = error ("DomainContext: internal shifting " ++
				      "function received: nV < 0")
                    | otherwise = [shiftContext x context' | x <- [0..]] !! nV

    in checkPostcondition $
       DomainContext { dcContext    = context'
		     , dcShiftedCxt = shiftFun
		     }

-- |
-- Converts the standard representation of the context
-- (as it is used in all other parts of hsloopo) into the
-- 'DomainContext' representation.
--
-- See also: 'mkDomainContext'
--
mkDC :: (Integral a)
     => [QfFormula (Polynomial (Ratio a))]
     -- ^ the context in the standard representation
     -> Maybe DomainContext
     -- ^ the context in the domain representation
mkDC [] = Nothing
mkDC xs = Just $ mkDomainContext xs

-- |
-- An empty domain context.
-- In the standard representation this would correspond to @[]@.
--
emptyDomainContext :: DomainContext
emptyDomainContext = mkDomainContext []

-- |
-- Converts a domain context into a 'ShiftedDomainContext'
-- where all parameters have been right-shifted by @b@ positions.
--
-- See also: @mkShiftedDomainContext'@
--
mkShiftedDomainContext :: (Integral b)
		       => DomainContext        -- ^ the context
		       -> b                    -- ^ the number of variables
		       -> ShiftedDomainContext -- ^ the shifted context
mkShiftedDomainContext context nV
    | nV < 0 = error "mkShiftedDomainContext: nV < 0"
    | otherwise = ShiftedDomainContext (fromIntegral nV) context

-- |
-- Converts a domain context into a 'ShiftedDomainContext'
-- where all parameters have been right-shifted by @b@ positions.
--
-- See also: 'mkShiftedDomainContext'
--
mkShiftedDomainContext' :: (Integral b)
                        => Maybe DomainContext  -- ^ the context
                        -> b                    -- ^ the number of variables
                        -> Maybe ShiftedDomainContext -- ^ the shifted context
mkShiftedDomainContext' Nothing _ = Nothing
mkShiftedDomainContext' (Just context) nV =
    Just $ mkShiftedDomainContext context nV

-- |
-- Converts a context, which is in domain representation ('DomainContext'),
-- into a context which is in the usual representation
-- (@[QfFormula (Polynomial (Ratio a))]@).
--
-- See also: @fromDomainContext'@
--
fromDomainContext :: (Integral a)
		  => DomainContext
		  -- ^ the context in the domain representation
                  -> [QfFormula (Polynomial (Ratio a))]
		  -- ^ the context in the usual representation
fromDomainContext = generalizeContext . dcContext

-- |
-- Converts a context, which is in domain representation ('DomainContext'),
-- into a context which is in the usual representation
-- (@[QfFormula (Polynomial (Ratio a))]@).
--
-- See also: 'fromDomainContext'
--
fromDomainContext' :: (Integral a)
                   => Maybe DomainContext
		   -- ^ the context in the domain representation
                   -> [QfFormula (Polynomial (Ratio a))]
		   -- ^ the context in the usual representation
fromDomainContext' = maybe [] fromDomainContext

-- |
-- Converts the shifted context to the usual representation of the
-- context (@[QfFormula (Polynomial (Ratio a))]@).
-- All variables will be shifted @nV@ positions to the right.
--
-- The number of variables @nV@ must match the number of variables
-- which is stored inside the shifted context.
--
-- See also: @fromShiftedContext'@
--
fromShiftedContext :: (Integral a, Integral b)
                   => ShiftedDomainContext -- ^ the shifted context
                   -> b                    -- ^ the number of variables
		   -> [QfFormula (Polynomial (Ratio a))]
		   -- ^ the shifted context in its usual representation
fromShiftedContext (ShiftedDomainContext nV' context) nV
    | nV < 0 = error "fromShiftedContext: nV < 0"
    | nV /= fromIntegral nV' =
	error ("fromShiftedContext: number of variables is not the " ++
	       "same as in the the given context:\n" ++
	       "context_nV=" ++ show nV' ++ ", arg_nV=" ++ show nV)
    | otherwise = generalizeContext $ dcShiftedCxt context nV'

-- |
-- Converts the shifted context to the usual representation of the
-- context (@[QfFormula (Polynomial (Ratio a))]@).
-- All variables will be shifted @nV@ positions to the right.
--
-- The number of variables @nV@ must match the number of variables
-- which is stored inside the shifted context.
--
-- See also: 'fromShiftedContext'
--
fromShiftedContext' :: (Integral a, Integral b)
		    => Maybe ShiftedDomainContext -- ^ the shifted context
		    -> b                          -- ^ the number of variables
		    -> [QfFormula (Polynomial (Ratio a))]
		    -- ^ the shifted context in its usual representation
fromShiftedContext' sContext nV =
    maybe [] (\x -> fromShiftedContext x nV) sContext


-- |
-- Takes an existing context and returns a new context with
-- an additional condition @f@.
--
-- See also: @extendDomainContext'@
--
extendDomainContext :: (Integral a)
                    => QfFormula (Polynomial (Ratio a)) -- ^ the new condition @f@
                    -> DomainContext -- ^ the old context
                    -> DomainContext -- ^ the old context plus the new condition @f@
extendDomainContext f oldContext@(DomainContext { dcContext = context } ) =
    case autoSimplify (Just oldContext) (fmap convertToPoly f) of
      Nothing -> oldContext -- f contains no new information
      Just f' -> mkDomainContext (f' : context)

-- |
-- Takes an existing context and returns a new context with
-- an additional condition @f@.
--
-- See also: 'extendDomainContext'
--
extendDomainContext' :: (Integral a)
                     => QfFormula (Polynomial (Ratio a)) -- ^ the new condition @f@
                     -> Maybe DomainContext -- ^ the old context
                     -> DomainContext -- ^ the old context plus the new condition @f@
extendDomainContext' f Nothing =
    mkDomainContext $ maybeToList $ autoSimplify Nothing f
extendDomainContext' f (Just context) = extendDomainContext f context
              
-- |
-- Reconstructs the original context from the shifted context.
--
-- Warning: This function should be avoided (if possible),
-- because it somewhat violates the whole context of the seperatation
-- between 'DomainContext' and 'ShiftedDomainContext'.
--
unshiftDomainContext :: ShiftedDomainContext -- ^ the shifted context
                     -> DomainContext        -- ^ the original context
unshiftDomainContext (ShiftedDomainContext _ context) = context

-- |
-- Test if the domain context @dc1@ implies the domain context @dc2@.
--
-- All possible results:
--
--     [@Just True@]   @dc1@ implies @dc2@
--
--     [@Just False@]  @dc2@ does not imply @dc2@
--
--     [@Nothing@]     the heuristic cannot decide the problem
--
-- See also: @heuristic_DC_implies_DC'@
--
heuristic_DC_implies_DC :: DomainContext -- ^ @dc1@
			-> DomainContext -- ^ @dc2@
			-> Maybe Bool
heuristic_DC_implies_DC dc1 dc2 =
    case (dcContext dc1, dcContext dc2) of
      (_,  []) -> Just True
      ([], ys) -> Nothing -- Just False would be too optimistic,
                          -- e.g. ys could be [QfAt AtTrue]
      (xs, ys) -> if xs == ys || all (\y -> y `elem` xs) ys
		  then Just True
		  else Nothing

-- |
-- Test if the domain context @dc1@ implies the domain context @dc2@.
--
-- All possible results:
--
--     [@Just True@]   @dc1@ implies @dc2@
--
--     [@Just False@]  @dc2@ does not imply @dc2@
--
--     [@Nothing@]     the heuristic cannot decide the problem
--
-- See also: 'heuristic_DC_implies_DC'
--
heuristic_DC_implies_DC' :: Maybe DomainContext -- ^ @dc1@
                         -> Maybe DomainContext -- ^ @dc2@
                         -> Maybe Bool
heuristic_DC_implies_DC' dc1 Nothing = Just True
heuristic_DC_implies_DC' Nothing (Just dc2) =
    heuristic_DC_implies_DC emptyDomainContext dc2
heuristic_DC_implies_DC' (Just dc1) (Just dc2) =
    heuristic_DC_implies_DC dc1 dc2

-- |
-- Helper function which transforms one representation of the context
-- into another form and converts between the types,
-- i.e. @(Ratio a) -> (Ratio b)@.
--
-- See also: 'checked_generalizeContext'
--
generalizeContext :: (Integral a, Integral b)
		  => [QfFormula (Polynomial (Ratio a))]
		  -- ^ the original context
		  -> [QfFormula (Polynomial (Ratio b))]
		  -- ^ the transformed context
generalizeContext = map (fmap convertToPoly)

-- |
-- Shift all indeterminates in the given formula to the right
-- to avoid conflicts between the parameters and the variables.
--
-- @t' = t[x(nV) \ x(0), x(nV+1) \ x(1), ..., x(nV+n) \ x(n)]@
--
-- Example:
-- @t = [QfAt (AtGe (u1*2%1))]@
-- @shiftContext 2 t = [QfAt (AtGe (u2*2%1))]@
--
-- (This function seems to be the same as 'liftPolyN'.)
--
shiftContext :: (Integral a, Integral b)
             => b -- the number of variables @nV@
             -> [QfFormula (Polynomial (Ratio a))] -- ^ original context
             -> [QfFormula (Polynomial (Ratio a))] -- ^ shifted context
shiftContext nV context
    | nV < 0 = error "shiftContext: nV < 0"
    | otherwise = map (fmap shiftIndets) context
    where
    zeros = replicate (fromIntegral nV) 0
    shiftIndets p =
	createPolyFromList [(c, zeros ++ xs) | (c, xs) <- polyToList p]

-- |
-- Test whether the given context is consistent.
--
-- See also: 'domainContext_NDEBUG'
--
checkPostcondition :: DomainContext -- ^ context @d@
	           -> DomainContext -- ^ unmodified context @d@
checkPostcondition context
    | domainContext_NDEBUG = context -- check has been disabled
    | impliesQf (dcContext context) (QfAt AtFalse) =
	error ("DomainContext: inconsistent theory " ++
	       "(i.e. context implies False)\ncontext=" ++
	       show (dcContext context))
    | otherwise = context -- context is consistent

----------------------------------------------------------------------
--                             Flags:
----------------------------------------------------------------------

-- |
-- Controls whether the assertions which are used in this module
-- should be disabled (like NDEBUG in C\/C++).
--
--   [@True@]   skip assertions
--
--   [@False@]  test assertions (recommended)
--
domainContext_NDEBUG :: Bool
domainContext_NDEBUG = True

-- |
-- Controls whether every new piece of context should be
-- simplified before it will be inserted.
--
--   [@True@]   simplify new context
--
--   [@False@]  insert context without simplification
--
autoSimplifyFlagSet :: Bool
autoSimplifyFlagSet = True

-- |
-- Simplifies the given formula under the current context
-- and returns the new context, which is the old context
-- and the new (simplified) formula.
--
-- However, if the formula is implied by the context,
-- Nothing will be returned instead.
--
autoSimplify :: (Integral a)
             => Maybe DomainContext              -- ^ the current context
             -> QfFormula (Polynomial (Ratio a)) -- ^ the original formula
             -> Maybe (QfFormula (Polynomial (Ratio a))) -- ^ simplified formula
autoSimplify context f
  | not domainContext_NDEBUG &&
    impliesQf (fromDomainContext' context) (QfAt AtFalse) =
	error ("autoSimplify: context is inconsistent:\n" ++ show context)
  | autoSimplifyFlagSet =
      case fastSimplifyQf (fromDomainContext' context) f of
        QfAt AtTrue  -> Nothing -- f contains no new information
        QfAt AtFalse ->
	    error ("autoSimplify: Inconsistent context extension detected\n" ++
		   "context=" ++ show context ++ "\nf=" ++ show f)
        simpl_f -> Just simpl_f
  | otherwise = Just f
