module LMath.Logic.DomainExamples where

import Data.Ratio
import Data.Maybe
import Data.List (genericLength, genericTake, maximumBy, intersperse)
import Debug.Trace

import LMath.Algo.Simplex
import LMath.Logic.QE
import LMath.Logic.Sort
import LMath.Tools
import LMath.Types
import LMath.Logic.DTreeContexts

import LMath.Logic.Domain
import LMath.Logic.DomainContext

import VisualPolylib

----------------------------------------------------------------------
-- Testing stuff
----------------------------------------------------------------------

test_p1 :: GIneqSystem Rational
test_p1 = ineqSystem ((mVars, mParams), mEntries)
    where
    mVars    = 2
    mParams  = 0
    mEntries = rowMatrix [1, 0, 0] `join`
	       rowMatrix [0, 1, 0] `join`
	       rowMatrix [2,-3, 3]

test_p2 :: GIneqSystem Rational
test_p2 = ineqSystem ((mVars, mParams), mEntries)
    where
    mVars    = 2
    mParams  = 0
    mEntries = rowMatrix [1, 0,  0] `join`
	       rowMatrix [-1,0,5%2] `join`
	       rowMatrix [0, 1,  1] `join`
	       rowMatrix [0,-1,  2]

test_domain1 :: Domain Rational
test_domain1 = ineqToDomain' (2,0) test_p1

test_domain2 :: Domain Rational
test_domain2 = ineqToDomain' (2,0) test_p2

test_mapping1 :: Num b => GAffineFunction b
test_mapping1 = constantAffineFunction 2 0 (matrix [[1],[0]])

n' :: Polynomial Rational
n' = indeterminate 1 0   -- Unbekannte (=Parameter) Nr. 0 von 1

n :: FracPoly Rational
n  = n' %% 1

-- Kontext
ctx :: [QfFormula (Polynomial Rational)]
ctx = map QfAt [AtGe (n'-1)]   -- n >= 1

-- Dreieck mit den Ecken (0,0), (0,10), (n,0)
sys :: GIneqSystem (FracPoly Rational)
sys = ineqSystem ((2,0), matrix [[ 1,     0,  0],   -- x >= 0
                                 [ 0,     1,  0],   -- y >= 0
                                 [-10/n, -1, 10]    -- y <= (-10/n) * x + 10
                                ]
		 )

-- Dreieck mit den Ecken (0,0), (0,10), (n,0)
sysP :: GIneqSystem (Polynomial Rational)
sysP = ineqSystem ((2,0), matrix [[ 1,      0,  0],   -- x >= 0
                                  [ 0,      1,  0],   -- y >= 0
                                  [-10*n', -1, 10]    -- y <= (-10*n) * x + 10
                                 ]
		  )

test_conversion n = error "test_conversion has been disabled"
                    -- WAS: atFormulasToDomain 2 $ domainToAtFormulas test_domain
--    where test_domain = foldl1 unionD $ replicate n (ineqToDomain' sysP)


----------------------------------------------------------------------
--
----------------------------------------------------------------------

testPolySimple :: (Integral a)
	       => GIneqSystem (Polynomial (Ratio a))
testPolySimple = ineqSystem ((2,0), matrix [[1,  0,  0]   -- x >= 0
                                           ]
			    )

-- contains an infinite number of integer points
testPoly1 :: (Integral a)
	  => GIneqSystem (Polynomial (Ratio a))
testPoly1 = ineqSystem ((2,0), matrix [[1,  0,  0],   -- x >= 0
                                       [0,  1,  0],   -- y >= 0
                                       [1, -1, 10]    -- y <= x + 10
                                      ]
                       )

-- singular point: (1,2)
testPoly2 :: (Integral a)
	  => GIneqSystem (Polynomial (Ratio a))
testPoly2 = ineqSystem ((2,0), matrix [[-1, 0,  1],   -- x <= 1
                                       [0,  1, -2],   -- y >= 2
                                       [2, -1,  0]    -- y <= 2 * x
                                      ]
                       )

-- triangle: (0,0), (3,0), (0,3)
testPoly3 :: (Integral a)
	  => GIneqSystem (Polynomial (Ratio a))
testPoly3 = ineqSystem ((2,0), matrix [[1,  0, 0],   -- x >= 0
				       [-1, 0, 3],   -- x <= 3
				       [0,  1, 0],   -- y >= 0
				       [0, -1, 3],   -- y <= 3
				       [-1,-1, 3]    -- x + y <= 3
				      ]
		       )

-- triangle: (4,0), (4,4), (0,4)
testPoly4 :: (Integral a)
	  => GIneqSystem (Polynomial (Ratio a))
testPoly4 = ineqSystem ((2,0), matrix [[1,  0, 0],   -- x >= 0
				       [-1, 0, 4],   -- x <= 4
				       [0,  1, 0],   -- y >= 0
				       [0, -1, 4],   -- y <= 4
				       [1,  1,-4]    -- x + y >= 4
				      ]
		       )

-- square: (0,0), (4,4)
testPoly5 :: (Integral a)
          => GIneqSystem (Polynomial (Ratio a))
testPoly5 = ineqSystem ((2,0), matrix [[1, 0, 0],    -- x >= 0
				       [-1,0, 4],    -- x <= 4
				       [0, 1, 0],    -- y >= 0
				       [0,-1, 4]     -- y <= 4
				      ]
		       )

-- random example
testPoly6 :: (Integral a)
          => GIneqSystem (Polynomial (Ratio a))
testPoly6 = ineqSystem ((2,1), matrix [[1, 0, 0, 2],
				       [0, 1, 0, -1],
				       [0,-1, 4, 5]
				      ]
		       )

-- random example
testPoly7 :: (Integral a)
          => GIneqSystem (Polynomial (Ratio a))
testPoly7 = ineqSystem ((2,1), matrix [[1, 0, 1, 2],
				       [-1,0, 4, 3]
				      ]
		       )

-- random example
testPoly8 :: (Integral a)
          => GIneqSystem (Polynomial (Ratio a))
testPoly8 = ineqSystem ((2,1), matrix [[3, 2, 0, 4],
--				       [-1,2, 0, 2],
--				       [1, 2, 4, 1],
--				       [2, 4, 7, -1],
				       [1,-1, 5, 3]
				      ]
		       )

-- random example
testPoly9 :: (Integral a)
          => GIneqSystem (Polynomial (Ratio a))
testPoly9 = ineqSystem ((2,1), matrix [[1, 3, 1, 2],
--				       [-1,1,-4, 5],
--				       [2, 1, 7, -1],
				       [5,-1, 3, 5]
				      ]
		       )

-- random example
testPoly10 :: (Integral a)
           => GIneqSystem (Polynomial (Ratio a))
testPoly10 = ineqSystem ((1,1), matrix [[1, 0, 0],
					[-1,1, 0]
				       ]
			)

testPoly10_d = ineqToDomain2 Nothing (1,1) testPoly10

testPoly10_context ::Integral b => Maybe [QfFormula (Polynomial (Ratio b))]
testPoly10_context = Just [QfAt $ AtEq m]
    where
    -- m = indeterminate 1 0
    m = indeterminate 2 1

-- singular point: (1,2)
testPolyPoint :: (Integral b)
	      => [b] -- ^ coordinates of the single point
              -> GIneqSystem Rational
testPolyPoint coords = ineqSystem ((l, 0), matrix rows)
    where
    l = genericLength coords
    rows = concat $
	   zipWith (\x p -> let r = [if y == p then 1 else 0 | y <- [0..l-1]] ++
		                    [fromIntegral (-x)]
		            in [r, map negate r]) coords [0..]


testDomain1 :: (Integral a)
            => Domain (Polynomial (Ratio a))
testDomain1 = ineqToDomain' (2,0) testPoly1

testDomain2 :: (Integral a)
            => Domain (Polynomial (Ratio a))
testDomain2 = ineqToDomain' (2,0) testPoly2

testDomain3 :: (Integral a)
            => Domain (Polynomial (Ratio a))
testDomain3 = testDomain1 `unionD` testDomain2

testDomain4 :: (Integral a)
            => Domain (Polynomial (Ratio a))
testDomain4 = testDomain3 `complementD'` testDomain2

testDomain5 :: (Integral a)
	    => Domain (Polynomial (Ratio a))
testDomain5 = testDomain2 `complementD'` testDomain2 -- emptyDomain

testDomain6 :: (Integral a)
            => Domain (Polynomial (Ratio a))
testDomain6 = ineqToDomain' (2,0) testPoly3

testDomain7 :: (Integral a)
            => Domain (Polynomial (Ratio a))
testDomain7 = ineqToDomain' (2,0) testPoly4

testDomain8 :: (Integral a)
            => Domain (Polynomial (Ratio a))
testDomain8 = testDomain6 `unionD` testDomain7

testDomain9 :: (Integral a)
	    => Domain (Polynomial (Ratio a))
testDomain9 = ineqToDomain' (2,0) testPoly5

testDomain10 :: (Integral a)
             => Domain (Polynomial (Ratio a))
testDomain10 = testDomain1 `intersectD` testDomain2

-- parameterized domains
testDomain11 :: (Integral a)
             => Domain (Polynomial (Ratio a))
testDomain11 = let d = toPolyDomain' ineqTreeExample_3
               in fmapD (\x -> x + (x * indeterminate 1 0)) d

testDomain12 :: (Integral a)
             => Domain (Polynomial (Ratio a))
testDomain12 = toPolyDomain' ineqTreeExample_2

----------------------------------------------------------------------
-- Test Mergetree
----------------------------------------------------------------------

zipDTree :: (Num cond, BasicDomainElement a)
         => (Domain a -> Domain a -> Domain a)
         -> DTree cond (Domain a)
         -> DTree cond (Domain a)
         -> DTree cond (Domain a)
zipDTree = mergeDTrees (identityDTreeContext [])
 
bug5_domain_fp = toFracPolyDomain bug5_domain
-- All those DTrees are (semantically) equal
testDTree_1 = DLeaf bug5_domain

testDTree_2 = DLeaf (simplifyD' bug5_domain)
testDTree_3 = FCond p_strict_positive (DLeaf t1) (DLeaf t2)
    where
    p_strict_positive = QfAt $ AtGt (indeterminate 3 2)
    t1 = simplifyD (mkDC [p_strict_positive])       bug5_domain
    t2 = simplifyD (mkDC [QfNot p_strict_positive]) bug5_domain
-- fmap simplifyD' $ simplifyDTree [] (let x = zipDTree complementD testDTree_3 testDTree_3 in zipDTree complementD x x)
-- fmap (simplifyD (Just [QfAt $ AtGt (indeterminate 3 2)])) $ simplifyDTree [QfAt $ AtGt (indeterminate 3 2)] (let x = zipDTree complementD testDTree_3 testDTree_3 in let y = zipDTree complementD x x in zipDTree complementD y y)



testDomainToIneqTree = domainToIneqTree Nothing (identityDTreeContext [])

-- is a polyhedron for arbitrary parameters? I don't think so.
testIneqTree_1 =
    (ineqToDomain' (2,1) testIneq4) `unionD` (ineqToDomain' (2,1) testIneq3)
testIneqTree_2 = testIneqTree_1 `unionD` (ineqToDomain' (2,1) testIneq5)
testIneqTree_3 = testIneqTree_1 `intersectD` (ineqToDomain' (2,1) testIneq5)
testIneqTree_4 = complementD Nothing testIneqTree_2
		 (ineqToDomain' (2,1)testIneq5)
-- (laesst sich ohne Kontext bisher noch nicht berechnen)
-- mit incConvexHullD sollte es jetzt klappen
testIneqTree_5 = convexHullD Nothing testIneqTree_2
testIneqTree_6 = convexHullD Nothing testIneqTree_3
testIneqTree_7 = ineqToDomain' (2,2) testIneq6
testIneqTree_8 = simplifyD' testIneqTree_7
testIneqTree_9 = testIneqTree_8 `unionD` testIneqTree_1
testIneqTree_10 = simplifyD' testIneqTree_9
testIneqTree_11 = convexHullD Nothing testIneqTree_10
testIneqTree_12 = convexHullD (mkDC [QfAt $ AtGt (indeterminate 3 2)]) testIneqTree_2
testIneqTree_13 = incConvexHullD (mkDC [QfAt $ AtGt (indeterminate 3 2)]) testIneqTree_2
(testIneqTree_14_a, testIneqTree_14_b) =
    let subst x = fmapD (\y -> evalPoly y [x])
	d1 = incConvexHullD (mkDC [QfAt $ AtGt (indeterminate 3 2)])
	     (toPolyDomain' (subst 2 testIneqTree_2))
	d2 = subst 2 testIneqTree_13
    in (d1, d2)

[ineq1, ineq2] = domainToIneqs testIneqTree_13

-- Sollte jetzt erledigt sein.
slfqBug1 = let x = indeterminate 1 0 in slfqSimplifyQf [] ((((QfAt $ AtGe 0) `QfAnd` (QfAt $ AtGe (-1)))) `QfOr` ((QfAt $ AtEq x) `QfAnd` (QfAt $ AtGe x) `QfAnd` (QfAt $ AtEq (x-1))))



--
-- Martins Beispiele
--
-- TODO: Anpassen oder entfernen
--
-- context: p == -1 \/ p == 1
ineqTreeExample_context :: Maybe DomainContext
ineqTreeExample_context =
    trace ("ineqTreeExample_context: Vorsicht! Falscher Kontext wird verwendet!") $
    mkDC [QfAt $ AtEq (p*p - 1)]
    where
    p = indeterminate 2 1 -- Dieser Kontext ist sicherlich falsch!

-- Originalversion (p = 1 \/ p = -1)
ineqTreeExample_1_formula =
    (QfAt $ AtEq (x - p)) `QfAnd`
    ((QfAt $ AtEq (p - 1)) `QfOr` (QfAt $ AtEq (p + 1)))
    where
    x = indeterminate 2 0
    p = indeterminate 2 1

ineqTreeExample_1 :: Integral a => Domain (Polynomial (Ratio a))
ineqTreeExample_1 = formulaToDomain Nothing (1,1) ineqTreeExample_1_formula

-- kompaktere Version (p^2 = 1)
ineqTreeExample_2_poly :: Integral a => GIneqSystem (Polynomial (Ratio a))
ineqTreeExample_2_poly =
    ineqSystem ((1,0), matrix [[1, -p], -- x >= p
			       [-1, p], -- x <= p
			       [0, p*p - 1] -- p*p = 1
			      ])
    where
    p = indeterminate 1 0

ineqTreeExample_2 = ineqToDomain' (1,1) ineqTreeExample_2_poly

-- Orginalversion, wobei der Parameter p als Variable betrachtet wird
ineqTreeExample_3 :: Integral a => Domain (Polynomial (Ratio a))
ineqTreeExample_3 = formulaToDomain Nothing (1,1) ineqTreeExample_1_formula


----------------------------------------------------------------------
-- Known bugs and limitations:
----------------------------------------------------------------------

-- bug1: non-deterministic behaviour (one of the two errors occurs)
-- 1) calculations does not finish in time (SOLVED!)
-- 2) ghc internal error:
-- <interactive>: internal error: scavenge_stack: weird activation record found on stack: 10
--     Please report this as a bug to glasgow-haskell-bugs@haskell.org,
--     or http://www.sourceforge.net/projects/ghc/
-- make: *** [ghci] Error 254
--
-- (both problems are solved)
--
bug1_context :: (Integral a) => QfFormula (Polynomial (Ratio a))
bug1_context = (QfAt $ AtGe (indeterminate 2 0)) `QfAnd`
             (QfAt $ AtGe (indeterminate 2 1))

bug1_domain = testDomain3

bug1 = convexHullD (mkDC [bug1_context]) bug1_domain

-- Seems to work now
bug2 = nParamsD $ convexHullD Nothing $ convexHullD Nothing testDomain1

bug3 = convexHullD Nothing (emptyDomain (0,0))

-- bug4 = test5 testDomain4

-- Ergebnisse scheinen zu passen.
-- Dreieck: (0,0)-(p,0)-(0,10)
bug5_domain = toPolyDomain' (ineqToDomain' (2,1) sys)
bug5_simpl = simplifyD' bug5_domain
-- TODO: Parameter duerfen parametrisch auftreten
-- -> Test abschwaechen (WURDE MITTLERWEILE GELOEST)
bug5 = convexHullD Nothing bug5_domain

testParam1 :: (Integral a)
           => GIneqSystem (Polynomial (Ratio a))
testParam1 = ineqSystem ((2,1), matrix [[1, 0, 1, 0],
					[-1,0, 1, 4],
					[0, 1, 1, 0],
					[0,-1, 1, 4] 
                                       ])

domainParam1 = ineqToDomain' (2,1) testParam1

-- Verwirrender Bug (WURDE MITTLERWEILE GELOEST):
-- Der folgende Test klappt, ...
bug6_passt  = includesD' testDomain4 (convexHullD Nothing (simplifyD' testDomain4))
bug6_passt2 = includesD' (simplifyD' testDomain4) (convexHullD Nothing (simplifyD' testDomain4))
-- ... aber die folgende Zeile funktioniert wiederum nicht!
bug6 = includesD' testDomain4 (convexHullD Nothing testDomain4)
bug6_alternative = includesD' (simplifyD' testDomain4) (convexHullD Nothing testDomain4)
-- am Simplifizieren sollte es aber nicht liegen, denn:
bug6_simplifyPasst = equivalentD' testDomain4 (simplifyD' testDomain4)

bug7_domain :: (Integral a) => Domain (Polynomial (Ratio a))
bug7_domain = unionD (emptyDomain (2,0)) (universalDomain (2,0))
bug7_hull = convexHullD Nothing $
	    (unionD (emptyDomain (2,0)) (universalDomain (2,0)))
bug7  = not $ equivalentD' bug7_hull (emptyDomain (2,0)) -- sollte wahr sein
bug7' = equivalentD' bug7_hull (universalDomain (2,0))   -- sollte wahr sein

testIneq1 :: (Num a) => GIneqSystem a
testIneq1 = ineqSystem ((2,0), matrix [[1, 0, 0], -- x >= 0
                                       [0, 1, 0]  -- y >= 0
                                      ])

testDomainX :: (BasicDomainElement a, PolyConvertableDomain a) => Domain a
testDomainX = ineqToDomain' (2,1) testIneq1

--testDomainY :: (Integral b, Num a) => b -> Domain a
testDomainY x = unionsD $
		replicate (fromIntegral x) (ineqToDomain' (2,1) testIneq1)


testIneq2 :: (Num a) => GIneqSystem a
testIneq2 = ineqSystem ((2,1), matrix [[1, 0, -1, 0], -- x >= p
                                       [0, 1,  0, 0]  -- y >= 0
                                      ])

--testDomainZ :: (Integral b, Num a) => b -> Domain a
testDomainZ x = unionsD $
		replicate (fromIntegral x) (ineqToDomain' (2,1) testIneq2)

testIneq3 :: (Integral a) => GIneqSystem (Polynomial (Ratio a))
testIneq3 = ineqSystem ((2,0), matrix [[1, 0, -p], -- x >= p
                                       [0, 1,  0]  -- y >= 0
                                      ])
    where
    p = indeterminate 1 0

testDomainA x = unionsD $
		replicate (fromIntegral x) (ineqToDomain' (2,1) testIneq3)

testIneq4 :: (Integral a) => GIneqSystem (Polynomial (Ratio a))
testIneq4 = ineqSystem ((2,0), matrix [[1, 0, -2*p + 4],  -- x >= 2*p - 4
                                     [0, 1,  -3*p - 1]  -- y >= 3*p + 1
                                    ])
    where
    p = indeterminate 1 0

testIneq5 :: (Integral a) => GIneqSystem (Polynomial (Ratio a))
testIneq5 = ineqSystem ((2,0), matrix [[2, -3, 7*p + 1],  -- random values
                                       [3, 11, -3*p + 85] -- random values
                                      ])
    where
    p = indeterminate 1 0

testIneq6 :: (Integral a) => GIneqSystem (Polynomial (Ratio a))
testIneq6 = ineqSystem ((2,0), matrix [[3, -2, 2*p^3 - 3*q^4 + 21], -- random values
                                       [5, -3, -3*p^4 + 7*q^5 - 5]  -- random values
                                      ])
    where
    p = indeterminate 2 0
    q = indeterminate 2 1


----------------------------------------------------------------------

--bug8_context  = mkDC [QfAt $ AtGe (indeterminate 3 2)] -- das hier ist dubios...
--bug8_context' = mkDC [QfAt $ AtGt (indeterminate 3 2)] -- das hier ist dubios...
--bug8  = simplifyD bug8_context  bug5_domain
--bug8' = simplifyD bug8_context' bug5_domain

-- SOLVED:
-- ToolFailed "/public/packages/scientific/qepcad-B.1.41/bin/slfq failed: \"Qepcad Output\" (line 1, column 1):\nunexpected end of input\nexpecting \"TRUE\", \"FALSE\", \"-\", \"+\", digit, \"x\", \">\", \"<\", \"/=\", \"=\" or \"[\""
-- *** Exception: (unknown)

--
-- Test topSortD:
--
topEx_1 = ineqToDomain' (2,0) $ testPolyPoint [0,0]
topEx_2 = ineqToDomain' (2,0) $ testPolyPoint [3,3]
topEx_3 = ineqToDomain' (2,0) $ testPolyPoint [4,1]

topsort_bug1_input = map (ineqToDomain2 Nothing (1,1))
		     [ineqTreeExample_2_poly,
		      fmap convertToPoly $ testPolyPoint [0]
		     ]

bug_topsort_1_d1 :: (Integral a) => Domain (Polynomial (Ratio a))
bug_topsort_1_d1 = ineqToDomain' (2,0) $
                   ineqSystem ((2,0), matrix [[1,0,-1]
					     ,[-1,0,5]
					     ,[0,1,-1]
					     ,[0,-1,6]
					     ])

bug_topsort_1_d2 :: (Integral a) => Domain (Polynomial (Ratio a))
bug_topsort_1_d2 = ineqToDomain' (2,0) $
                   ineqSystem ((2,0), matrix [[1,0,-1]
					     ,[-1,0,5]
					     ,[0,1,0]
					     ,[0,-1,0]
					     ])

-- funktioniert: (Richtig: ["S1", "S3", "S4"] -> ["S4"])
bug_topsort_1a = lTopSortD Nothing 1
		 [(bug_topsort_1_d1, ["S4"])
		 ,(bug_topsort_1_d2, ["S1", "S3", "S4"])
		 ]
bug_topsort_1b = lTopSortD Nothing 1
		 [(bug_topsort_1_d2, ["S1", "S3", "S4"])
		 ,(bug_topsort_1_d1, ["S4"])
		 ]

-- Um eine VisualPolylib Darstellung zu bekommen:
-- putStrLn $ concat $ map (\(x,_) -> showDomain2VisualPolylib (toRatioDomain x) (nVarsD x + nParamsD x) []) $ head $ leavesDTree bug_topsort_1b


isFeasibleSys_bugIneqSys :: GIneqSystem (FracPoly Rational)
isFeasibleSys_bugIneqSys =
    ineqSystem ((2,0), matrix [[1, 0, -p]
			      ,[1, 0, 0]
			      ,[-1,0, 0]
                              ])
    where
    p = (indeterminate 1 0) %% 1

isFeasibleSys_bug = makeDTree (fracPolyDTreeContext () []) $
		    isFeasibleSys isFeasibleSys_bugIneqSys Nothing
