-- |
-- Contains a testsuite for the Domain-module.
--
-- See also: 'LMath.Logic.Domain'
--
module DomainTests
    (main, runTests
    )
where

import Data.Ratio
import Data.Maybe
import Data.List (genericLength, genericTake, maximumBy,
		  intersperse, nub, sort)
import Debug.Trace

import LMath.Logic.QE
import LMath.Logic.Sort
import LMath.Tools
import LMath.Types

import LMath.Logic.Domain
import LMath.Logic.DomainExamples
import LMath.Logic.DomainContext

----------------------------------------------------------------------
-- Testcases:
----------------------------------------------------------------------

-- |
-- Runs all testcases.
--
runTests :: Bool
runTests = singleTests [("test11", test11)
		       ,("test12 (ineqToDomain + parameters)", test12)
		       ,("test13 (topsort [single example])", test13)
		       ,("test14_1 (topsort [sort rational numbers])", test14_1)
		       ,("test14_2 (topsort [sort pairs of rational numbers])", test14_2)
		       ] &&
           all (fullTestN ds) [("test7", test7), ("test8", test8),
			       ("test9", test9)] &&
	   all (fullTest1 ds) [("test4", test4)
			      ,("test5 (convexHullD)", test5)
			      ,("test5_ext (convexHullD extended)", test5_ext)
			      ,("test5_context (convexHullD extended)", test5_context)
			      ,("test5_inc (incConvexHullD)", test5_inc)
			      ,("test5_subst_1 (convexHullD)", test5_subst 3 (convexHullD Nothing))
			      ,("test5_subst_2 (incConvexHullD)", test5_subst 3 (incConvexHullD Nothing))
			      ,("test10", test10)] &&
	   all (fullTest2 ds) [("test1", test1), ("test2", test2),
			       ("test3", test3)]
    where
    ds = [testDomain1, testDomain2, testDomain3, testDomain4, testDomain5]
	 ++ extended
         -- withParams
    extended = [testDomain6, testDomain7, testDomain8, testDomain9, testDomain10]
    withParams = [testDomain11, testDomain12]

singleTests :: [(String, Bool)]
	    -> Bool
singleTests ((name, test) : xs) = trace ("testing " ++ name ++ "...") $
				  seq test (test && singleTests xs)
singleTests [] = True

fullTest1 :: [Domain a]
	  -> (String, (Domain a -> Bool))
	  ->  Bool
fullTest1 xs (name, op) = trace ("testing " ++ name ++ "...") $
			  all runOp (zip xs [1..])
    where
    l = genericLength xs
    runOp (x, num) = trace ("  " ++ show num ++ "/" ++ show l ++
			    " (" ++ name ++ ")") $
		     let res = op x in seq res res

fullTest2 :: [Domain a]
	  -> (String, (Domain a -> Domain a -> Bool))
	  -> Bool
fullTest2 xs (name, op) = trace ("testing " ++ name ++ "...") $
			  and [runOp x y | x <- xs', y <- xs']
    where
    xs' = zip xs [1..]
    runOp (x, num1) (y, num2) = trace' ("  " ++ show num1 ++ ":" ++ show num2 ++
					" (" ++ name ++ ")") $
				op x y
    verbose = True
    trace' = if verbose then trace else (flip const)

fullTestN :: [Domain a]
	  -> (String, ([Domain a] -> Bool))
	  -> Bool
fullTestN xs (name, op) = trace ("testing " ++ name ++ "...") $
			  op xs

-- Test if 'equivalentD d1 d2 == (includesD d1 d2 && includesD d2 d1)'
test1 d1 d2 = equivalentD' d1 d2 == (includesD' d1 d2 && includesD' d2 d1)

-- Test if 'unionD' is commutative (weak test case!).
test2 d1 d2 = equivalentD' (d1 `unionD` d2) (d2 `unionD` d1)

-- Test if 'intersectD' is commutative (weak test case!).
test3 d1 d2 = equivalentD' (d1 `intersectD` d2) (d2 `intersectD` d1)

-- Test conversion: domain <-> formula
test4 d1 = let f = domainToFormula d1
	       d1' = formulaToDomain Nothing (nV, nP) f
	       (nV, nP) = (nVarsD d1, nParamsD d1)
           in equivalentD' d1 d1'

-- Test convex hull
test5 :: (Integral a)
      => Domain (Polynomial (Ratio a))
      -> Bool
test5 d1 = let hull = convexHullD Nothing d1
	   in includesD' d1 hull

-- extended convex hull tests
test5_ext :: (Integral a)
          => Domain (Polynomial (Ratio a))
          -> Bool
test5_ext d1 = let hull  = convexHullD Nothing d1
		   hull2 = convexHullD Nothing hull
		   params1 = nParamsD d1
		   params2 = nParamsD hull
		   params3 = nParamsD hull2
	       in params1 <= params2 &&
		  params2 <= params3 &&
		  equivalentD' hull hull2

-- Test if an additional context, which could be ignored,
-- does not lead to errors.
test5_context :: (Integral a)
              => Domain (Polynomial (Ratio a))
              -> Bool
test5_context d1 = trace ("test5_context skipped.") $
		   True
{-
                   --trace ("test5_context: results=\n" ++ show results) $
                   allEqual (map nParamsD results) &&
		   and (zipWith equivalentD' results (tail results))
-}
    where
    p = nParamsD d1
    context n = mkDC [QfAt $ AtGe (indeterminate n i) | i <- [p..n-1]]
    results = [convexHullD (context n) d1 | n <- [p..p+1]]

    allEqual [] = True
    allEqual (x:xs) = -- trace ("testing allEqual: (x:xs)=" ++ show (x:xs)) $
		      all (==x) xs

-- Test if @convexHullD [] d == incConvexHullD [] d@.
test5_inc :: (Integral a)
          => Domain (Polynomial (Ratio a))
          -> Bool
test5_inc d1 = let hull1 = convexHullD    Nothing d1
                   hull2 = incConvexHullD Nothing d1
               in equivalentD' hull1 hull2

-- Test if it doesn't matter, if parameters are replaced by values
-- before the call to 'convexHullD' or after the call.
test5_subst :: (Integral a)
            => Int
	    -> (Domain (Polynomial Rational) -> Domain (Polynomial Rational))
            -> Domain (Polynomial (Ratio a))
            -> Bool
test5_subst numTests convexHullAlgo d1 =
    if nP == 0
    then True
    else all check (take numTests params)
    where
    -- yes, this is no good random number generator...
    pseudo_rands = map fromIntegral $
	           zipWith (*) [abs (x `mod` 61) | x <- [1,71..]] (cycle [1,-1])
    nP = nParamsD d1
    params = tail $ map snd (iterate (\(x1,_) -> let (y1,y2) = splitAt nP x1
			                         in (y2, y1)) (pseudo_rands, []))
    substP xs d = toPolyDomain' $
		  substituteParamsD d [Just x | x <- xs]
		  --fmapD (\x -> evalPoly x xs) d
    check xs = let pre   = convexHullAlgo (substP xs (toPolyDomain' d1))
		   after = substP xs (convexHullAlgo (toPolyDomain' d1))
               in equivalentD' pre after

-- Test if @unionsD xs == unionsD (reverse xs)@
test7 xs = equivalentD' (unionsD xs) (unionsD (reverse xs))

-- Test if @unionsD xs == foldl1 unionD xs@
test8 xs = equivalentD' (unionsD xs) (foldl1 unionD xs)

-- Test if @unionsD xs == foldr1 unionD xs@
test9 xs = equivalentD' (unionsD xs) (foldr1 unionD xs)

-- Test @simplifyD' d1 == d1@
test10 d1 = d1 `equivalentD'` (simplifyD' d1)

-- Test 'convexHullD' on an empty domain
test11 = let hull = convexHullD Nothing (emptyDomain (0,0))
         in nParamsD hull == 0

test12_ineq_fail :: Integral a => GIneqSystem (Polynomial (Ratio a))
test12_ineq_fail = ineqSystem ((2,1), matrix [[1, 0, n, 2],
					      [0, 1, n^2, 2 + n]
					     ]
			      )
    where
    n = indeterminate 1 0

test12_ineq_ok :: Integral a => GIneqSystem (Polynomial (Ratio a))
test12_ineq_ok = ineqSystem ((3,0), matrix [[1, 0, n, 2],
					    [0, 1, n^2, 2 + n]
					   ]
			    )
    where
    n = indeterminate 1 0

test12_domain_fail :: (Integral a)
                   => Domain (Polynomial (Ratio a))
test12_domain_fail = ineqToDomain2 Nothing (2,1) test12_ineq_fail

test12_domain_ok :: (Integral a)
                 => Domain (Polynomial (Ratio a))
test12_domain_ok = toPolyDomain' $
		   ineqToDomain2 Nothing (3,1) test12_ineq_ok

test12_ineq1 :: (Integral a)
             => GIneqSystem (Polynomial (Ratio a))
test12_ineq1 = ineqSystem ((2,2), matrix [[2, 3, 4, 5, 2]
					 ,[-2, 0, 0, -1, 3]
					 ]
			  )
    where
    p = indeterminate 2 0
    q = indeterminate 2 1

test12_ineq2 :: (Integral a)
             => GIneqSystem (Polynomial (Ratio a))
test12_ineq2 = ineqSystem ((2,0), matrix [[ 2, 3, 4*p + 5*q + 2]
					 ,[-2, 0, -q + 3]
					 ]
			  )
    where
    p = indeterminate 2 0
    q = indeterminate 2 1

-- ineqToDomain: elimination of the parameters
test12 :: Bool
test12 = nParamsD test12_domain_ok == 1 && d1 `equivalentD'` d2
    where
    d1 = ineqToDomain2 Nothing (2,2) test12_ineq1
    d2 = ineqToDomain2 Nothing (2,2) test12_ineq2

-- (1,0), (2,0), (3,0), (4,0), (5,0)
test13_ineq1 = ineqSystem ((2,0), matrix [[7, 1, -7]
					 ,[-7, -1, 100]
					 ,[7, 0, -1]
					 ,[-1, 0, 5]
					 ,[0, 1, 0]
					 ,[0, -1, 0]
					 ]
			  )

-- ineq system which does not contain an integer point
test13_ineq2 = ineqSystem ((2,0), matrix [[-7, -1, 6]
					 ,[7,0,-1]
					 ,[-1,0,5]
					 ,[0,1,0]
					 ,[0,-1,0]
					 ]
			  )

-- rectangle: (1,1) - (5,6), (but contains more non-integer points)
test13_ineq3 = ineqSystem ((2,0), matrix [[7,1,-7]
					 ,[-7,-1,100]
					 ,[7,0,-1]
					 ,[-1,0,5]
					 ,[0,1,-1]
					 ,[0,-1,6]
					 ]
			  )

test13_d1 = ineqToDomain2 Nothing (2,0) test13_ineq1
test13_d2 = ineqToDomain2 Nothing (2,0) test13_ineq2
test13_d3 = ineqToDomain2 Nothing (2,0) test13_ineq3

-- topsort (single example)
test13 :: Bool
test13 = isOk test13a && isOk test13b
    where
    test13a = lTopSortD Nothing 1 $
	      [(test13_d1, "d1"), (test13_d3, "d3")]
    test13b = lTopSortD Nothing 1 $
	      reverse [(test13_d1, "d1"), (test13_d3, "d3")]
    isOk t = case leavesDTree t of
	       [[(_, "d1"), (_, "d3")]] -> True
	       _ -> False

-- use lTopSortD for sorting rational numbers
sortByTopSortD :: (Integral a)
               => [Ratio a]
               -> [Ratio a]
sortByTopSortD xs
    | length xs == length (nub xs) = parseResult resultTree
    | otherwise = error "sortByTopSortD: no duplicates are allowed"
    where
    parseResult = (map snd) . head . leavesDTree
    input = [(singlePointD (1,0) [x], x) | x <- xs]
    resultTree = lTopSortD Nothing 0 input

-- use lTopSortD for sorting pairs of rational numbers
sortByTopSortD_2 :: (Integral a)
                 => [(Ratio a, Ratio a)]
                 -> [(Ratio a, Ratio a)]
sortByTopSortD_2 xs
    | length xs == length (nub xs) = parseResult resultTree
    | otherwise = error "sortByTopSortD: no duplicates are allowed"
    where
    parseResult = (map snd) . head . leavesDTree
    input = [(singlePointD (2,0) [x1,x2], (x1,x2)) | (x1,x2) <- xs]
    resultTree = lTopSortD Nothing 0 input

-- use lTopSortD for sorting rational numbers
test14_1 :: Bool
test14_1 = sortByTopSortD xs == sort xs
    where
    xs = [3,1,5,7,2,6]

-- use lTopSortD for sorting pairs of rational numbers
test14_2 :: Bool
test14_2 = sortByTopSortD_2 input == output
    where
    input  = [(1,1), (2,2), (1,2), (2,1), (30,-30), (-30,30)]
    output = [(-30,30), (1,1), (1,2), (2,1), (2,2), (30,-30)]

-- |
-- Runs all testcases and prints its result.
--
main :: IO ()
main = catchToolIOError run
       (\x -> do
	      putStrLn "\n-------------begin error message ----------------\n" 
	      putStrLn (show x)
	      putStrLn "\n------------- end error message -----------------\n")
    where
    run = putStrLn (if runTests
		    then "All tests were successful!"
		    else "At least one test failed!")

----------------------------------------------------------------------
-- Test Mergetree
----------------------------------------------------------------------

zipDTree :: (Num cond, BasicDomainElement a)
         => (Domain a -> Domain a -> Domain a)
         -> DTree cond (Domain a)
 	 -> DTree cond (Domain a)
 	 -> DTree cond (Domain a)
zipDTree = mergeDTrees (identityDTreeContext [])

bug5_domain_fp = toFracPolyDomain bug5_domain
-- All those DTrees are (semantically) equal
testDTree_1 = DLeaf bug5_domain
testDTree_2 = DLeaf (simplifyD' bug5_domain)
testDTree_3 = FCond p_strict_positive (DLeaf t1) (DLeaf t2)
    where
    p_strict_positive = QfAt $ AtGt (indeterminate 3 2)
    t1 = simplifyD (mkDC [p_strict_positive])       bug5_domain
    t2 = simplifyD (mkDC [QfNot p_strict_positive]) bug5_domain

-- fmap simplifyD' $ simplifyDTree [] (let x = zipDTree complementD testDTree_3 testDTree_3 in zipDTree complementD x x)

-- fmap (simplifyD (Just [QfAt $ AtGt (indeterminate 3 2)])) $ simplifyDTree [QfAt $ AtGt (indeterminate 3 2)] (let x = zipDTree complementD testDTree_3 testDTree_3 in let y = zipDTree complementD x x in zipDTree complementD y y)



testDomainToIneqTree = domainToIneqTree Nothing (identityDTreeContext [])
