--
-- This module contains function which do not work on domains,
-- but on decision trees of domains. So in this respect it is
-- an extension of the 'LMath.Logic.Domain' module.
--
-- See also: 'LMath.Logic.Domain'
--
module LMath.Logic.DomainTree
    ( -- * extended domain operations (for 'DTree's)
      lTopSortDT, lPartitionDT, projectToVarsDT,
      simplifyDT, simplifyDT',

      -- * helper functions
      lifted_fmapDTree_with_context
    )
where

import Data.Ratio
import Data.Maybe
import Data.List (genericLength, genericTake, maximumBy, intersperse)
import Array
import Debug.Trace

import LMath.Logic.QE
import LMath.Logic.Sort
import LMath.Tools
import LMath.Types

import LMath.Logic.Domain
import LMath.Logic.DomainContext

-- |
-- Same as 'fmapDTree_with_context', but it will work with
-- operations @f@ which uses 'DomainContext's instead of
-- @[QfFormula (Polynomial (Ratio a))]@.
--
-- See also: 'fmapDTree_with_context'
--
lifted_fmapDTree_with_context :: (Integral c)
                              => Maybe DomainContext
			      -- ^ initial context
			      -> (Maybe DomainContext -> a -> b)
			      -- ^ the transformation operations @f@
			      -> DTree (Polynomial (Ratio c)) a
			      -- ^ the original tree @dt@
			      -> DTree (Polynomial (Ratio c)) b
lifted_fmapDTree_with_context initContext f =
    fmapDTree_with_context (fromDomainContext' initContext) liftedF
    where
    liftedF context = f (mkDC context)

-- |
-- Topologically sorts a decision tree of domains.
--
-- See also: 'LMath.Logic.Domain.lTopSortD
--
lTopSortDT :: (BasicDomainElement a, PolyConvertableDomain a,
	       DomainCanBeSimplified a, Integral b, Integral c)
           => Maybe DomainContext -- ^ the (optional) context
           -> c                   -- ^ the current dimension
           -> DTree (Polynomial (Ratio b)) [(Domain a, label)]
           -> DTree (Polynomial (Ratio b)) [(Domain a, label)]
lTopSortDT initContext relDims dt =
    flattenDTree $
    lifted_fmapDTree_with_context initContext (\c -> lTopSortD c relDims) dt

-- |
-- Calculates a partitioning for the list of domains,
-- which are the leaves of the given decision tree.
--
-- See also: 'LMath.Logic.Domain.lPartitionD'
--
lPartitionDT :: (BasicDomainElement a, PolyConvertableDomain a,
		 PolyConvertableDomain b, Integral c)
	     => Maybe DomainContext
             -> DTree (Polynomial (Ratio c)) [(Domain a, label)]
	     -> DTree (Polynomial (Ratio c)) [(Domain b, [label])]
lPartitionDT initContext =
    lifted_fmapDTree_with_context initContext lPartitionD

-- |
-- Projects a domain @d@ to the given dimensions.
-- This is a more generalized version of the function @projectToVarsPss'@.
--
-- See also: 'LMath.Logic.Domain.projectToVarsD',
--           @LMath.Algo.Sort.projectToVarsPss'@
--
projectToVarsDT :: (BasicDomainElement a, PolyConvertable a,
		   FracPolyConvertable a, Integral b, Integral c)
		=> DTreeContext (Polynomial (Ratio b)) sc (Polynomial (Ratio b))
		-> Maybe DomainContext
		-> [Integer] -- Dimensionen, auf die projeziert werden soll
		-> DTree (Polynomial (Ratio b)) (Domain a)
		-> DTree (Polynomial (Ratio b)) (Domain (Polynomial (Ratio c)))
projectToVarsDT tree_context initContext pVars dt =
    flattenDTree $
    lifted_fmapDTree_with_context initContext
    (\context d -> projectToVarsD tree_context context pVars d) dt

-- |
-- Simplifies the domains on the leaves of the given decision tree @dt@.
--
-- Note:
-- This function does not simplify the decision tree itself,
-- i.e. it does not change the structure of the tree.
--
-- See also: @simplifyDT'@, 'LMath.Logic.Domain.simplifyD'
--
simplifyDT :: (DomainCanBeSimplified a, Integral b)
           => Maybe DomainContext -- ^ the initial context
           -> DTree (Polynomial (Ratio b)) (Domain a) -- ^ @dt@
           -> DTree (Polynomial (Ratio b)) (Domain a)
simplifyDT initContext =
    lifted_fmapDTree_with_context initContext simplifyD

-- |
-- Simplifies the domains on the leaves of the given decision tree @dt@.
-- If you have a special context, you should prefer the function 'simplifyDT'.
--
-- Note:
-- This function does not simplify the decision tree itself,
-- i.e. it does not change the structure of the tree.
--
-- See also: 'simplifyDT', 'LMath.Logic.Domain.simplifyD'
--
simplifyDT' :: (DomainCanBeSimplified a, Integral b)
            => DTree (Polynomial (Ratio b)) (Domain a) -- ^ @dt@
            -> DTree (Polynomial (Ratio b)) (Domain a)
simplifyDT' = simplifyDT Nothing
