module LMath.Logic.Examples.CloogExample where

import Data.Ratio
import Data.Maybe
import Data.List (genericLength, genericTake, maximumBy, intersperse)
import Debug.Trace

import LMath.Logic.QE
import LMath.Logic.Sort
import LMath.Tools
import LMath.Types
import LMath.Logic.DTreeContexts

import LMath.Logic.Domain
import LMath.Logic.DomainContext
import LMath.Logic.DomainTree
import LMath.Logic.DomainBenchmarks (stopElapsedPicoseconds', printElapsedTime)
import LMath.Logic.Examples.Quillere
import VisualPolylib

----------------------------------------------------------------------
-- Martins "ClooG-per-Hand" Beispiel:
----------------------------------------------------------------------

-- parameters:
m_ :: Polynomial Rational
m_ = indeterminate nP_ 0

params_ :: [Polynomial Rational]
params_ = [m_]
nP_ = genericLength params_
nV_ = 2 -- [t, r]

--
-- The context that is used to simplify the intermediate results.
--
context_ :: Maybe DomainContext
context_ = --mkDC [m_strict_positive_]
           --mkDC [m_strict_positive_, m_leq_100_]
	   --mkDC [m_greater_eq_ 12, m_leq_100_]
           --mkDC [m_greater_than_100_]
           --mkDC [m_equal_99_]
	   --mkDC [m_equal_6_]
           --mkDC [m_equal_7_]
	   --mkDC [m_equal_12_]
           --mkDC [m_strict_negative_]
           --mkDC [m_int_ [1..100]]
           --mkDC [m_int_ [12..100]]
           --mkDC [m_greater_eq_ 1, m_leq_100_]
           Nothing

--
-- The context that is used to simplify the final decision tree.
--
final_context_ :: Maybe DomainContext
final_context_ = Nothing
                 --mkDC [m_strict_positive_]
		 --mkDC [m_int_ [12..100]]
		 --mkDC [m_equal_7_]
                 --mkDC [m_greater_eq_ 1, m_leq_100_]
		 --mkDC [m_int_ [1..100]]

m_strict_positive_ = QfAt $ AtGe (m_ - 1)   -- M >= 1
m_greater_than_100_ = QfAt $ AtGe (m_ - 101) -- M >= 101
m_greater_than_1000_ = QfAt $ AtGe (m_ - 1000) -- M >= 1000
m_greater_eq_ n    = QfAt $ AtGe (m_ - n)   -- M >= "n"
m_leq_100_         = QfAt $ AtGe (-m_ + 100) -- M <= 100
m_equal_7_         = QfAt $ AtEq (m_ - 7)   -- M == 7
m_equal_6_         = QfAt $ AtEq (m_ - 6)   -- M == 6
m_equal_12_        = QfAt $ AtEq (m_ - 12)  -- M == 12
m_equal_99_        = QfAt $ AtEq (m_ - 99)  -- M == 99
m_strict_negative_ = QfAt $ AtGe (-1  - m_) -- M <= -1
m_int_ interval = disjunctionQf [QfAt $ AtEq (m_ - constPoly x) | x <- interval]

polynomial_tree_context_ =
    polyDTreeContext () (fromDomainContext' context_)
fracPoly_tree_context_ =
    fracPolyDTreeContext () (fromDomainContext' context_)

--
-- Construct the inequality systems for each statement (S1, S2, S3, S4)
--
s1_IneqSys :: (Integral a) => GIneqSystem (Ratio a)
s1_IneqSys = ineqSystem ((nV_,nP_), matrix m)
    where
    --    t   r  M   1
    m = [[1,  0, 0, 0] -- t >= 0
	,[-1, 0, 0, 5] -- t <= 5
	,[0,  1, 0, 0] -- r >= 0
	,[0, -1, 0, 0] -- r <= 0
	]

s2_IneqSys :: (Integral a) => GIneqSystem (Ratio a)
s2_IneqSys = ineqSystem ((nV_,nP_), matrix m)
    where
    --    t   r  M   1
    m = [[1,  0, 0, -6] -- t >= 6
	,[-1, 0, 0, 10] -- t <= 10
	,[0,  1, 0,  0] -- r >= 0
	,[0, -1, 0,  0] -- r <= 0
	]

s3_IneqSys :: (Integral a) => GIneqSystem (Ratio a)
s3_IneqSys = ineqSystem ((nV_,nP_), matrix m)
    where
    --    t   r  M   1
    m = [[1,  0, 0,  0] -- t >= 0
	,[-1, 0, 1, -1] -- t <= M-1
	,[0,  1, 0,  0] -- r >= 0
	,[0, -1, 0,  0] -- r <= 0
	]

s4_IneqSys :: (Integral a) => GIneqSystem (Polynomial (Ratio a))
s4_IneqSys = fmap convertToPoly $ ineqSystem ((nV_, 0), matrix m)
    where
    --    t    r    1
    m = [[m_,  1,  -m_^2]          -- M*t -M^2 + M + r >= M
	,[-m_, -1, m_^2 -m_ + 100] -- M*t -M^2 + M + r <= 100
	,[0,   1,      0]          -- r >= 0
	,[0, -1,    m_-1]          -- r <= M-1
	]

--
-- Convert the polyhedra to domains
--
s1_domain = (ineqToDomain2 context_ (nV_, nP_) s1_IneqSys, "S1")
s2_domain = (ineqToDomain2 context_ (nV_, nP_) s2_IneqSys, "S2")
s3_domain = (ineqToDomain2 context_ (nV_, nP_) s3_IneqSys, "S3")
s4_domain = (ineqToDomain2 context_ (nV_, nP_) s4_IneqSys, "S4")

allDomains = [s1_domain, s2_domain, s3_domain, s4_domain]

proj1 (d,l) = fmap (\x -> (x,l))
	      (projectToVarsD polynomial_tree_context_ context_ [0] d)
proj12 (d,l) = fmap (\x -> (x,l))
	       (projectToVarsD polynomial_tree_context_ context_ [0,1] d)


--
-- Project on the 1. dimension (which is "t")
--
s1_project_1, s2_project_1, s3_project_1, s4_project_1 ::
    DTree (Polynomial Rational) (Domain (Polynomial Rational), String)
s1_project_1 = proj1 s1_domain
s2_project_1 = proj1 s2_domain
s3_project_1 = proj1 s3_domain
s4_project_1 = proj1 s4_domain

--
-- Partitionate the results of the projections and
-- find the overlapping parts.
--
partitions_1 :: DTree (Polynomial Rational)
		[(Domain (Polynomial Rational), [String])]
partitions_1 = lPartitionDT context_ projections
    where
    projections = mergeListOfDTrees polynomial_tree_context_ (++) $
		  map (fmap (\x -> [x])) projList
    projList = [s1_project_1, s2_project_1, s3_project_1, s4_project_1]

--
-- Topologically sort the partitions.
--
topsort_1 :: DTree (Polynomial Rational)
	     [(Domain (Polynomial Rational), [String])]
topsort_1 = lTopSortDT context_ 0 partitions_1


test_topsort_1 = catchToolIOError (print topsort_1) print

--
-- Project on the 1st and 2nd dimension
--
s1_project_12 = proj12 s1_domain
s2_project_12 = proj12 s2_domain
s3_project_12 = proj12 s3_domain
s4_project_12 = proj12 s4_domain

projections_2 :: DTree (Polynomial Rational)
                 [(OuterLoop, [(Domain (Polynomial Rational), String)])]
projections_2 = fmapDTree_with_context initContext proj2 topsort_1
    where
    initContext = fromDomainContext' context_
    proj2 c ps = [(OuterLoop d, [(simplifyD (mkDC c) ((getDomain l) `intersectD` d), l) | l <- ls])
		  | (d,ls) <- ps]
    -- Wichtig:
    -- An dieser Stelle muesste eigentlich die Projektion auf die 1+2
    -- Komponente von S1,..,S4 gemacht werden. Weil wir in diesem Beispiel
    -- aber nur 2 Dimensionen haben, habe ich das hier weggelassen.
    -- Im Allgemeinen muesste man das aber schon machen und bekommt dann
    -- nicht einfache Domains, sondern Baeume von Domains.
    -- Es ist zwar moeglich, diese zusaetzliche Ebene wieder zu
    -- entfernen, aber das ist eine recht laestige Arbeit...
    --
    -- Fuer unsere Zwecke sollte es OK sein, wenn man das hier weglaesst.
    --
    getDomain l = fst $ head $ filter (\(_,l') -> l == l') allDomains

partitions_2 :: DTree (Polynomial Rational)
                [(OuterLoop, [(Domain (Polynomial Rational), [String])])]
partitions_2 = fmapDTree_with_context initContext (\c -> map (partition c)) projections_2
    where
    initContext = fromDomainContext' context_
    partition c (outLoop, ds) = (outLoop, lPartitionD (mkDC c) ds)

topsort_2 :: DTree (Polynomial Rational)
             [(OuterLoop, [(Domain (Polynomial Rational), [String])])]
topsort_2 = simplifyDTree (fromDomainContext' final_context_) $
	    flatten $ fmapDTree_with_context initContext
	    (\c -> (map (\(outLoop, xs) -> fmap (\x -> (outLoop, x)) (lTopSortD (mkDC c) 1 xs)
                        )
                   )
            )
	    partitions_2
    where
    initContext = fromDomainContext' context_
    flatten1 dt = fmap (\t -> if null t
			      then DLeaf []
                              else mergeListOfDTrees polynomial_tree_context_ mergeOp $
                                   map (fmap (\x -> [x])) t
		       ) dt
    flatten2 = flattenDTree
    flatten = flatten2 . flatten1
    mergeOp = (++)

polyLibRep topsort_result = -- z.B. (subst_topsort_2 [Just 99] Nothing)
    putStrLn $ unlines $
    map (\(d,l) -> "\nLabel " ++ show l ++ ":\n" ++
		   showDomain2VisualPolylib d (nVarsD d) []) $
    concat $ head $ leavesDTree topsort_result

--
-- Replace the parameter @m@ by a numerical constant.
--
subst_context_ = mkDC [m_equal_7_]
simpl_domain_  = (simplifyD subst_context_) .
		 (substituteAllConstantsD subst_context_)
simpl_dtree_   = simplifyDTree (fromDomainContext' subst_context_)

--
-- Helper functions (for the substitution and simplification):
--
simpl_proj1_ = simpl_dtree_ . fmap (\(x,y) -> (simpl_domain_ x, y))
simpl_partitions1_ =
    simpl_dtree_ .
    fmap (filter (not.isAlwaysEmptyD'.fst) . (map (\(x,y) -> (simpl_domain_ x, y))))
simpl_topsort1_ = simpl_partitions1_

--
-- Substituted projections on the 1. dimension.
--
subst_s1_project_1, subst_s2_project_1, subst_s3_project_1, subst_s4_project_1 ::
    DTree (Polynomial Rational) (Domain (Polynomial Rational), String)
subst_s1_project_1 = simpl_proj1_ s1_project_1
subst_s2_project_1 = simpl_proj1_ s2_project_1
subst_s3_project_1 = simpl_proj1_ s3_project_1
subst_s4_project_1 = simpl_proj1_ s4_project_1

--
-- Substituted partitions.
--
subst_partitions_1 :: DTree (Polynomial Rational)
		      [(Domain (Polynomial Rational), [String])]
subst_partitions_1 = simpl_partitions1_ partitions_1

--
-- Substituted topologically sorted partitions.
--
subst_topsort_1 :: DTree (Polynomial Rational)
		   [(Domain (Polynomial Rational), [String])]
subst_topsort_1 = simpl_topsort1_ topsort_1

subst_topsort_2 params context =
    let subst x = substituteParamsD x params
    in fmapDTree_with_context (fromDomainContext' context)
	   (\c -> map (\(OuterLoop outLoop, xs) -> (OuterLoop (simplifyD (mkDC c) (subst outLoop)),
						    map (\(x,y) -> (simplifyD (mkDC c) x,y)) xs))) $
       fmap (map (\(outLoop, xs) -> (outLoop, map (\(x,y) -> (subst x,y)) xs))) topsort_2

-- M == 7
subst_m7_topsort_2 = simplifyResult (mkDC [m_equal_7_]) $
                     subst_topsort_2 [Just 7] (mkDC [m_equal_7_])

simplifyResult context dt =
    fmapDTree_with_context (fromDomainContext' context)
    (\context -> (map (\(OuterLoop outLoop, xs) -> (OuterLoop (simplifyD (mkDC context) outLoop), map (\(x,y) -> (simplifyD (mkDC context) x, y)) xs)) )) $
    simplifyDTree (fromDomainContext' context) dt


-- putStrLn $ head $ leavesDTree $ fmap (concatMap (concatMap (\(d,s) -> show s ++ "\n" ++ showDomain2VisualPolylib d 2 [] ++ "\n" ++ replicate 70 '-' ++ "\n"))) subst_m7_topsort_2

targetCode :: String
targetCode = codegen topsort_2 final_context_

main = do
       printLine
       putStrLn "calculating projections:"
       putStrLn "s1_project_1:"
       putStrLn (show s1_project_1)
       putStrLn "s2_project_1:"
       putStrLn (show s2_project_1)
       putStrLn "s3_project_1:"
       putStrLn (show s3_project_1)
       putStrLn "s4_project_1:"
       putStrLn (show s4_project_1)
       printLine
       putStrLn "calculating partitions_1:"
       putStrLn (show partitions_1)
       printLine
       putStrLn "calculating topsort_1:"
       putStrLn (show topsort_1)
       printLine
       putStrLn "calculating projections_2:"
       putStrLn (show projections_2)
       printLine
       putStrLn "calculating partitions_2"
       putStrLn (show partitions_2)
       printLine
       putStrLn "calculating topsort_2"
       putStrLn (show topsort_2)
       printLine
       putStrLn "generating target code"
       putStrLn (show targetCode)
    where
    printLine = putStrLn (replicate 70 '-')

-- |
-- Returns time in picoseconds.
--
benchmark :: IO Integer
benchmark = do
	    putStrLn "Generating target code..."
            diff_pico <- stopElapsedPicoseconds'
			 (putStrLn targetCode >> putStrLn "done.")
            putStrLn ("context: " ++ show context_)
	    putStrLn ("final context: " ++ show final_context_)
	    putStrLn ("simplifier: " ++ show domainSimplifier)
            putStrLn ("#leaves: " ++ show (length $ leavesDTree topsort_2))
            putStrLn ("lines of code: " ++ show (countLines targetCode))
            putStrLn ("#if-statements: " ++ show (countIfs targetCode))
            putStrLn ("#expressions: " ++ show (countExpr targetCode))
	    printElapsedTime diff_pico
	    return diff_pico

{- Orginalbeispiel:

Beispielprogramm:


DO i=0,5
  A = ...
END DO

DO i=6,10
  A = ...
END DO

DO i=0, M-1
  B[i] = B[i-1]
END DO

DO i=M, 100
  B[i] = B[i-M]
END DO

und

t1 = i
t2 = i
t3 = i
t4 = floor(i/M), also: i = M * t4 + r mit 0 <= r <= M-1

Zielsysteme:

S1:
0 <= t <= 5
0 = r

S2:
6 <= t <= 10
0 = r

S3:
0 <= t <= M-1
0 = r

S4:
M <= M*t + r <= 100
0 <= r <= M-1

Projektionen auf 1. Dim (alias t):

S1:
0 <= t <= 5

S2:
6 <= t <= 10

S3:
0 <= t <= M-1

S4:
M <= M*t+r <= 100
0 <= r <= M-1


Überlappungsbereiche / Partitionierung:

t=0; t=1..min(M,5); Größenordnung: 20 Fälle (Blätter)
Beispielfall: M=7 => Partitionen:
  0: S1, S3
  1..5: S1, S3, S4
  6: S2, S3, S4
  7-10: S2, S4
  11-14? : S4

Top. Sort der 1. Dim = fade.

-------------------------------

Proj. auf 1.+2.Dim (= alles), aber für jede Partition (in jedem Blatt)
separat: Beispielfall M=7 und  Partition  1..5: S1, S3, S4

Context:
1 <= t <= 5

S1:
0 = r

S3:
0 = r

S4:
0 <= r <= 6

Überlappungsbereiche / Partitionen:

r=0: S1, S3, S4 (a)
1<=r<=6 : S4    (b)

TopSortierung:

(a) vor (b)

Resultat in diesem Fall:

 1 <= t <= 5
     r=0: {S1, S3, S4}

     1<=r<=6 : S4

-}
