module LMath.Logic.Examples.CloogExample2 where

import Data.Ratio
import Data.Maybe
import Data.List (genericLength, genericTake, maximumBy, intersperse)
import Debug.Trace

import LMath.Logic.QE
import LMath.Logic.Sort
import LMath.Tools
import LMath.Types
import LMath.Logic.DTreeContexts

import LMath.Logic.Domain
import LMath.Logic.DomainContext
import LMath.Logic.DomainTree
import LMath.Logic.DomainBenchmarks (stopElapsedPicoseconds', printElapsedTime)
import LMath.Logic.Examples.Quillere
import VisualPolylib

----------------------------------------------------------------------
-- X-"CLooG-per-Hand" Beispiel:
----------------------------------------------------------------------
{-
Das "X" besteht aus 2 Parallelogrammen (S1 und S2), die sich überlappen.
S2 ist fest gewählt: (10,0) - (15,0) - (0,10) - (5,10)
S1 ist parameterisiert duch die Parameter { p, q, r }:
(0,0) - (p,0) - (q,r) - (q+p,r)

Hier ist ein Diagramm für p=5, q=10, r=10:

(0,10)          (p+q,r)
      ____  ____
      \ S2\/ S1/
       \  /\  /
	\/  \/ 
	/\  /\
       /  \/  \
      /___/\___\
(0,0)            (15,0)


Wenn man aber die Parameter verändert, ändert sich damit auch der
Überlappungsbereich, bis er schließlich ganz verschwindet.
-}

-- parameters:
p_, q_, r_ :: Polynomial Rational
p_ = indeterminate nP_ 0
q_ = indeterminate nP_ 1
r_ = indeterminate nP_ 2

params_ :: [Polynomial Rational]
params_ = [p_, q_, r_]

nP_, nV_ :: Integer
nP_ = genericLength params_
nV_ = 2

--
-- The context that is used to simplify the intermediate results.
--
context_ :: Maybe DomainContext
context_ = Nothing
           --mkDC all_params_strict_positive_
           --mkDC p5_q10_r10_context_
           --mkDC p5_qGT_rGT_context_
           --mkDC pGT_q10_r10_context_
           --mkDC pGT_qGT_r10_context_

--
-- The context that is used to simplify the final decision tree.
--
final_context_ :: Maybe DomainContext
final_context_ = mkDC p5_qGT_rGT_context_

-- context: p >= 1, q >= 1, r >= 1
all_params_strict_positive_ = [gt p_, gt q_, gt r_]
    where
    gt p = QfAt $ AtGe (p - 1)

-- context: p=5, q=10, r=10
p5_q10_r10_context_ = [eq p_ 5, eq q_ 10, eq r_ 10]
    where
    eq p x = QfAt $ AtEq (p - x)

p5_qGT_rGT_context_ = [eq p_ 5, gt q_, gt r_]
    where
    eq p x = QfAt $ AtEq (p - x)
    gt p = QfAt $ AtGe (p - 1)

-- context: p >= 1, q=10, r=10
pGT_q10_r10_context_ = [gt p_, eq q_ 10, eq r_ 10]
    where
    gt p = QfAt $ AtGe (p - 1)
    eq p x = QfAt $ AtEq (p - x)

-- context: p >= 1, q >= 1, r=10
pGT_qGT_r10_context_ = [gt p_, gt q_, eq r_ 10]
    where
    gt p = QfAt $ AtGe (p - 1)
    eq p x = QfAt $ AtEq (p - x)

polynomial_tree_context_ =
    polyDTreeContext () (fromDomainContext' context_)
fracPoly_tree_context_ =
    fracPolyDTreeContext () (fromDomainContext' context_)

--
-- Construct the systems for each statement (S1, S2) by
-- calculating the convex hull of the four edge points.
--
s1_points :: [[Polynomial Rational]]
s1_points = [[0,  0]
	    ,[p_, 0]
	    ,[q_, r_]
	    ,[q_+p_, r_]
	    ]

s2_points :: [[Polynomial Rational]]
s2_points = [[10, 0]
	    ,[15, 0]
	    ,[0, 10]
	    ,[5, 10]
	    ]

pointsToDomain context xs = incConvexHullD context_ $
			    substituteAllConstantsD context $
			    unionsD [singlePointD (nV_, nP_) x | x <- xs]

s1_domain = if isNothing context_ && isNothing final_context_
	    then (pointsToDomain final_context_ s1_points, "S1")
	    else trace "Using handcrafted s1_domain" $
		 (simplifyD' (fst s1_domain_handcrafted), "S1")
s2_domain = (pointsToDomain final_context_ s2_points, "S2")

allDomains = [s1_domain, s2_domain]

--
-- context: p >= 1, q >= 1, r >= 1
--
s1_domain_handcrafted :: (Domain (Polynomial Rational), String)
s1_domain_handcrafted = (d, "S1 (handcrafted)")
    where
    context = mkDC all_params_strict_positive_
    d = ineqToDomain2 context (nV_, nP_) ineqSys
    ineqSys = ineqSystem ((nV_, 0), matrix m)
    --    i    j      1
    m = [[0,   1,     0 ] -- j >= 0
	,[0,  -1,     r_] -- j <= r
	,[r_, -q_,    0 ] -- q*j <= r*i
	,[-r_, q_, r_*p_] -- q*j >= r * (i-p)
	]
-- Test: (if context_ == all_params_strict_positive_)
-- equivalentD context_ (fst s1_domain) (fst s1_domain_handcrafted)

main = do
       putStrLn "Increasing SLFQ Heap (ca. 800 Mb)..."
       setSlfqCells 200
       putStrLn "s1_domain:"
       putStrLn (show s1_domain)
       putStrLn "s2_domain:"
       putStrLn (show s2_domain)
       printLine
       putStrLn "calculating projections:"
       putStrLn "s1_project_1:"
       putStrLn (show s1_project_1)
       putStrLn "s2_project_1:"
       putStrLn (show s2_project_1)
       printLine
       putStrLn "calculating partitions_1:"
       putStrLn (show partitions_1)
       printLine
       putStrLn "calculating topsort_1:"
       putStrLn (show topsort_1)
       printLine
       putStrLn "calculating projections_2:"
       putStrLn (show projections_2)
       printLine
       putStrLn "calculating partitions_2"
       putStrLn (show partitions_2)
       printLine
       putStrLn "calculating topsort_2"
       putStrLn (show topsort_2)
    where
    printLine = putStrLn (replicate 70 '-')

--
-- Helper functions for the projection phase:
--
proj1 (d,l) = fmap (\x -> (x,l))
	      (projectToVarsD polynomial_tree_context_ context_ [0] d)
proj12 (d,l) = fmap (\x -> (x,l))
	       (projectToVarsD polynomial_tree_context_ context_ [0,1] d)


--
-- Project on the 1. dimension (which is "t")
--
s1_project_1, s2_project_1 ::
    DTree (Polynomial Rational) (Domain (Polynomial Rational), String)
s1_project_1 = proj1 s1_domain
s2_project_1 = proj1 s2_domain

--
-- Partitionate the results of the projections and
-- find the overlapping parts.
--
partitions_1 :: DTree (Polynomial Rational)
		[(Domain (Polynomial Rational), [String])]
partitions_1 = lPartitionDT context_ projections
               --fmap (map (\(x,y) -> (ineqToDomain2 context_ (nV_,nP_) x, y))) $
	       --lPartitionDT2 context_ projections
    where
    projections = mergeListOfDTrees polynomial_tree_context_ (++) $
		  map (fmap (\x -> [x])) projList
    projList = [s1_project_1, s2_project_1]

--
-- Topologically sort the partitions.
--
topsort_1 :: DTree (Polynomial Rational)
	     [(Domain (Polynomial Rational), [String])]
topsort_1 = lTopSortDT context_ 0 partitions_1


test_topsort_1 = catchToolIOError (print topsort_1) print

--
-- Project on the 1st and 2nd dimension
--
s1_project_12 = proj12 s1_domain
s2_project_12 = proj12 s2_domain

projections_2 :: DTree (Polynomial Rational)
                 [(OuterLoop, [(Domain (Polynomial Rational), String)])]
projections_2 = fmapDTree_with_context initContext proj2 topsort_1
    where
    initContext = fromDomainContext' context_
    proj2 c ps = [(OuterLoop d, [(simplifyD (mkDC c) ((getDomain l) `intersectD` d), l) | l <- ls])
		  | (d,ls) <- ps]
    -- Wichtig:
    -- An dieser Stelle muesste eigentlich die Projektion auf die 1+2
    -- Komponente von S1,..,S4 gemacht werden. Weil wir in diesem Beispiel
    -- aber nur 2 Dimensionen haben, habe ich das hier weggelassen.
    -- Im Allgemeinen muesste man das aber schon machen und bekommt dann
    -- nicht einfache Domains, sondern Baeume von Domains.
    -- Es ist zwar moeglich, diese zusaetzliche Ebene wieder zu
    -- entfernen, aber das ist eine recht laestige Arbeit...
    --
    -- Fuer unsere Zwecke sollte es OK sein, wenn man das hier weglaesst.
    --
    getDomain l = fst $ head $ filter (\(_,l') -> l == l') allDomains

partitions_2 :: DTree (Polynomial Rational)
                [(OuterLoop, [(Domain (Polynomial Rational), [String])])]
partitions_2 = fmapDTree_with_context initContext (\c -> map (partition c)) projections_2
    where
    initContext = fromDomainContext' context_
    partition c (outLoop, ds) = (outLoop, lPartitionD (mkDC c) ds)

topsort_2 :: DTree (Polynomial Rational)
             [(OuterLoop, [(Domain (Polynomial Rational), [String])])]
topsort_2 = simplifyDTree (fromDomainContext' final_context_) $
	    flatten $ fmapDTree_with_context initContext
	    (\c -> (map (\(outLoop, xs) -> fmap (\x -> (outLoop, x)) (lTopSortD (mkDC c) 1 xs)
                        )
                   )
            )
	    partitions_2
    where
    initContext = fromDomainContext' context_
    flatten1 dt = fmap (\t -> if null t
			      then DLeaf []
                              else mergeListOfDTrees polynomial_tree_context_ mergeOp $
                                   map (fmap (\x -> [x])) t
		       ) dt
    flatten2 = flattenDTree
    flatten = flatten2 . flatten1
    mergeOp = (++)

targetCode :: String
targetCode = codegen topsort_2 final_context_

polyLibRep topsort_result = -- z.B. (subst_topsort_2 [Just 99] Nothing)
    putStrLn $ unlines $
    map (\(d,l) -> "\nLabel " ++ show l ++ ":\n" ++
		   showDomain2VisualPolylib d (nVarsD d) []) $
    concat $ head $ leavesDTree topsort_result

--
-- Replace the parameter @m@ by a numerical constant.
--
subst_context_ = mkDC p5_q10_r10_context_
simpl_domain_  = (simplifyD subst_context_) .
		 (substituteAllConstantsD subst_context_)
simpl_dtree_   = simplifyDTree (fromDomainContext' subst_context_)

--
-- Helper functions (for the substitution and simplification):
--
simpl_proj1_ = simpl_dtree_ . fmap (\(x,y) -> (simpl_domain_ x, y))
simpl_partitions1_ =
    simpl_dtree_ .
    fmap (filter (not.isAlwaysEmptyD'.fst) . (map (\(x,y) -> (simpl_domain_ x, y))))
simpl_topsort1_ = simpl_partitions1_

--
-- Substituted projections on the 1. dimension.
--
subst_s1_project_1, subst_s2_project_1 ::
    DTree (Polynomial Rational) (Domain (Polynomial Rational), String)
subst_s1_project_1 = simpl_proj1_ s1_project_1
subst_s2_project_1 = simpl_proj1_ s2_project_1

--
-- Substituted partitions.
--
subst_partitions_1 :: DTree (Polynomial Rational)
		      [(Domain (Polynomial Rational), [String])]
subst_partitions_1 = simpl_partitions1_ partitions_1

--
-- Substituted topologically sorted partitions.
--
subst_topsort_1 :: DTree (Polynomial Rational)
		   [(Domain (Polynomial Rational), [String])]
subst_topsort_1 = simpl_topsort1_ topsort_1

subst_topsort_2 params context =
    let subst x = substituteParamsD x params
    in fmapDTree_with_context (fromDomainContext' context)
	   (\c -> map (\(OuterLoop outLoop, xs) -> (OuterLoop (simplifyD (mkDC c) (subst outLoop)),
						    map (\(x,y) -> (simplifyD (mkDC c) x,y)) xs))) $
       fmap (map (\(outLoop, xs) -> (outLoop, map (\(x,y) -> (subst x,y)) xs))) topsort_2

subst_p5_q10_r10_topsort_2 = simplifyAllDomains $
			     simplifyDTree p5_q10_r10_context_ $
			     subst_topsort_2 [Just 5, Just 10, Just 10] context_
    where
    simplifyAllDomains = fmapDTree_with_context p5_q10_r10_context_
			 (\context -> (map (\(outLoop, xs) -> (outLoop, map (\(x,y) -> (simplifyD (mkDC context) x, y)) xs)) ))

-- Zum Anzeigen:
-- putStrLn $ concat $ concatMap (map (\(x,y) -> (showDomain2VisualPolylib x 2 [] ++ show y ++ "\n" ++ replicate 70 '-' ++ "\n"))) $ head $ leavesDTree subst_p5_q10_r10_topsort_2

-- |
-- Returns time in picoseconds.
--
benchmark :: IO Integer
benchmark = do
	    putStrLn "Generating target code..."
            diff_pico <- stopElapsedPicoseconds'
			 (putStrLn targetCode >> putStrLn "done.")
            putStrLn ("context: " ++ show context_)            
            putStrLn ("#leaves: " ++ show (length $ leavesDTree topsort_2))
            putStrLn ("lines of code: " ++ show (countLines targetCode))
	    printElapsedTime diff_pico
	    return diff_pico
