module LMath.Logic.Examples.CloogExample4 where

import Data.Ratio
import Data.Maybe
import Data.List (genericLength, genericTake, maximumBy, intersperse)
import Debug.Trace

import LMath.Logic.QE
import LMath.Logic.Sort
import LMath.Tools
import LMath.Types
import LMath.Logic.DTreeContexts

import LMath.Logic.Domain
import LMath.Logic.DomainContext
import LMath.Logic.DomainTree
import LMath.Logic.Examples.Quillere
import VisualPolylib

----------------------------------------------------------------------
-- Quillere-Beispiel
----------------------------------------------------------------------

-- parameters:
n_ :: Polynomial Rational
n_ = indeterminate nP_ 0

params_ :: [Polynomial Rational]
params_ = [n_]
nP_ = genericLength params_
nV_ = 2 -- [i, j]

--
-- the context: assume that the parameters are (strict) positive
--
context_ = --mkDC [n_strict_positive_]
           --mkDC [n_equal_12_]
           mkDC [n_greater_than_12_]
           --Nothing

n_strict_positive_ = QfAt $ AtGe (n_ - 1)  -- N >= 1
n_greater_than_12_ = QfAt $ AtGe (n_ - 12) -- N >= 12
n_equal_12_        = QfAt $ AtEq (n_ - 12) -- N == 12

polynomial_tree_context_ =
    polyDTreeContext () (fromDomainContext' context_)
fracPoly_tree_context_ =
    fracPolyDTreeContext () (fromDomainContext' context_)

--
-- Construct the inequality systems for each statement (S1, S2, S3)
--
s1_IneqSys :: (Integral a) => GIneqSystem (Ratio a)
s1_IneqSys = ineqSystem ((nV_,nP_), matrix m)
    where
    --    i   j  N   1
    m = [[0,  1, 0,-1] -- j >= 1
	,[1, -1, 0, 0] -- j <= i
	,[-1, 0, 1, 0] -- i <= N
	]

s2_IneqSys :: (Integral a) => GIneqSystem (Ratio a)
s2_IneqSys = ineqSystem ((nV_,nP_), matrix m)
    where
    --    i   j  N   1
    m = [[1,  1, 0, -6] -- i+j >= 6
	,[-1,-1, 0, 18] -- i+j <= 18
	,[1, -1, 0,  2] -- i-j >= -2
	,[-1, 1, 0,  2] -- i-j <= 2
	]

s3_IneqSys :: (Integral a) => GIneqSystem (Ratio a)
s3_IneqSys = ineqSystem ((nV_,nP_), matrix m)
    where
    --    i   j  N   1
    m = [[1,  0, 0, -5] -- i >= 5
	,[-1, 0, 0, 10] -- i <= 10
	,[0,  1, 0, -6] -- j >= 6
	,[0, -1, 0, 11] -- j <= 11
	]

--
-- Convert the polyhedrons to domains
--
s1_domain = (ineqToDomain2 context_ (nV_, nP_) s1_IneqSys, "S1")
s2_domain = (ineqToDomain2 context_ (nV_, nP_) s2_IneqSys, "S2")
s3_domain = (ineqToDomain2 context_ (nV_, nP_) s3_IneqSys, "S3")

allDomains = [s1_domain, s2_domain, s3_domain]

--
-- Assertions:
--
test_s1 = equivalentD context_ (fst s1_domain) $
	  pointsToDomain context_ [[1,1], [n_,1], [n_,n_]]
test_s2 = equivalentD context_ (fst s2_domain) $
	  pointsToDomain context_ [[2,4], [4,2], [8,10], [10,8]]
test_s3 = equivalentD context_ (fst s3_domain) $
	  pointsToDomain context_ [[10,6], [10,6+b], [10-b,6], [10-b,6+b]]
    where
    b = 5

-- TODO: Nach domain.hs verschieben
pointsToDomain context xs = incConvexHullD context_ $
			    substituteAllConstantsD context $
			    unionsD [singlePointD (nV_, nP_) x | x <- xs]

proj1 (d,l) = fmap (\x -> (x,l))
	      (projectToVarsD polynomial_tree_context_ context_ [0] d)
proj12 (d,l) = fmap (\x -> (x,l))
	       (projectToVarsD polynomial_tree_context_ context_ [0,1] d)


--
-- Project on the 1. dimension (which is "i")
--
s1_project_1, s2_project_1, s3_project_1 ::
    DTree (Polynomial Rational) (Domain (Polynomial Rational), String)
s1_project_1 = proj1 s1_domain
s2_project_1 = proj1 s2_domain
s3_project_1 = proj1 s3_domain

--
-- Partitionate the results of the projections and
-- find the overlapping parts.
--
partitions_1 :: DTree (Polynomial Rational)
		[(Domain (Polynomial Rational), [String])]
partitions_1 = lPartitionDT context_ projections
    where
    projections = mergeListOfDTrees polynomial_tree_context_ (++) $
		  map (fmap (\x -> [x])) projList
    projList = [s1_project_1, s2_project_1, s3_project_1]

--
-- Topologically sort the partitions.
--
topsort_1 :: DTree (Polynomial Rational)
	     [(Domain (Polynomial Rational), [String])]
topsort_1 = lTopSortDT context_ 0 partitions_1


test_topsort_1 = catchToolIOError (print topsort_1) print

--
-- Project on the 1st and 2nd dimension
--
s1_project_12 = proj12 s1_domain
s2_project_12 = proj12 s2_domain
s3_project_12 = proj12 s3_domain

projections_2 :: DTree (Polynomial Rational)
                 [(OuterLoop, [(Domain (Polynomial Rational), String)])]
projections_2 = fmapDTree_with_context initContext proj2 topsort_1
    where
    initContext = fromDomainContext' context_
    proj2 c ps = [(OuterLoop d, [(simplifyD (mkDC c) ((getDomain l) `intersectD` d), l) | l <- ls])
		  | (d,ls) <- ps]
    -- Wichtig:
    -- An dieser Stelle muesste eigentlich die Projektion auf die 1+2
    -- Komponente von S1,..,S4 gemacht werden. Weil wir in diesem Beispiel
    -- aber nur 2 Dimensionen haben, habe ich das hier weggelassen.
    -- Im Allgemeinen muesste man das aber schon machen und bekommt dann
    -- nicht einfache Domains, sondern Baeume von Domains.
    -- Es ist zwar moeglich, diese zusaetzliche Ebene wieder zu
    -- entfernen, aber das ist eine recht laestige Arbeit...
    --
    -- Fuer unsere Zwecke sollte es OK sein, wenn man das hier weglaesst.
    --
    getDomain l = fst $ head $ filter (\(_,l') -> l == l') allDomains

partitions_2 :: DTree (Polynomial Rational)
                [(OuterLoop, [(Domain (Polynomial Rational), [String])])]
partitions_2 = fmapDTree_with_context initContext (\c -> map (partition c)) projections_2
    where
    initContext = fromDomainContext' context_
    partition c (outLoop, ds) = (outLoop, lPartitionD (mkDC c) ds)

topsort_2 :: DTree (Polynomial Rational)
             [(OuterLoop, [(Domain (Polynomial Rational), [String])])]
topsort_2 = simplifyDTree (fromDomainContext' context_) $
	    flatten $ fmapDTree_with_context initContext
	    (\c -> (map (\(outLoop, xs) -> fmap (\x -> (outLoop, x)) (lTopSortD (mkDC c) 1 xs)
                        )
                   )
            )
	    partitions_2
    where
    initContext = fromDomainContext' context_
    flatten1 dt = fmap (\t -> if null t
			      then DLeaf []
                              else mergeListOfDTrees polynomial_tree_context_ mergeOp $
                                   map (fmap (\x -> [x])) t
		       ) dt
    flatten2 = flattenDTree
    flatten = flatten2 . flatten1
    mergeOp = (++)

polyLibRep topsort_result = -- z.B. (subst_topsort_2 [Just 12] Nothing)
    putStrLn $ unlines $
    map (\(d,l) -> "\nLabel " ++ show l ++ ":\n" ++
		   showDomain2VisualPolylib d (nVarsD d) []) $
    concat $ head $ leavesDTree topsort_result

--
-- Replace the parameter @m@ by a numerical constant.
--
subst_context_ = mkDC [n_equal_12_]
simpl_domain_  = (simplifyD subst_context_) .
		 (substituteAllConstantsD subst_context_)
simpl_dtree_   = simplifyDTree (fromDomainContext' subst_context_)

--
-- Helper functions (for the substitution and simplification):
--
simpl_proj1_ = simpl_dtree_ . fmap (\(x,y) -> (simpl_domain_ x, y))
simpl_partitions1_ =
    simpl_dtree_ .
    fmap (filter (not.isAlwaysEmptyD'.fst) . (map (\(x,y) -> (simpl_domain_ x, y))))
simpl_topsort1_ = simpl_partitions1_

--
-- Substituted projections on the 1. dimension.
--
subst_s1_project_1, subst_s2_project_1, subst_s3_project_1 ::
    DTree (Polynomial Rational) (Domain (Polynomial Rational), String)
subst_s1_project_1 = simpl_proj1_ s1_project_1
subst_s2_project_1 = simpl_proj1_ s2_project_1
subst_s3_project_1 = simpl_proj1_ s3_project_1

--
-- Substituted partitions.
--
subst_partitions_1 :: DTree (Polynomial Rational)
		      [(Domain (Polynomial Rational), [String])]
subst_partitions_1 = simpl_partitions1_ partitions_1

--
-- Substituted topologically sorted partitions.
--
subst_topsort_1 :: DTree (Polynomial Rational)
		   [(Domain (Polynomial Rational), [String])]
subst_topsort_1 = simpl_topsort1_ topsort_1

subst_topsort_2 params context =
    let subst x = substituteParamsD x params
    in fmapDTree_with_context (fromDomainContext' context)
	   (\c -> map (\(OuterLoop outLoop, xs) -> (OuterLoop (simplifyD (mkDC c) (subst outLoop)),
						    map (\(x,y) -> (simplifyD (mkDC c) x,y)) xs))) $
       fmap (map (\(outLoop, xs) -> (outLoop, map (\(x,y) -> (subst x,y)) xs))) topsort_2

-- N == 12
subst_n12_topsort_2 = simplifyAllDomains $
                     simplifyDTree [n_equal_12_] $
                     subst_topsort_2 [Just 12] (mkDC [n_equal_12_])
    where
    simplifyAllDomains = fmapDTree_with_context [n_equal_12_]
			 (\context -> (map (\(outLoop, xs) -> (outLoop, map (\(x,y) -> (simplifyD (mkDC context) x, y)) xs)) ))


-- putStrLn $ head $ leavesDTree $ fmap (concatMap (concatMap (\(d,s) -> show s ++ "\n" ++ showDomain2VisualPolylib d 2 [] ++ "\n" ++ replicate 70 '-' ++ "\n"))) subst_n12_topsort_2


main = do
       printLine
       putStrLn "calculating projections:"
       putStrLn "s1_project_1:"
       putStrLn (show s1_project_1)
       putStrLn "s2_project_1:"
       putStrLn (show s2_project_1)
       putStrLn "s3_project_1:"
       putStrLn (show s3_project_1)
       printLine
       putStrLn "calculating partitions_1:"
       putStrLn (show partitions_1)
       printLine
       putStrLn "calculating topsort_1:"
       putStrLn (show topsort_1)
       printLine
       putStrLn "calculating projections_2:"
       putStrLn (show projections_2)
       printLine
       putStrLn "calculating partitions_2"
       putStrLn (show partitions_2)
       printLine
       putStrLn "calculating topsort_2"
       putStrLn (show topsort_2)
    where
    printLine = putStrLn (replicate 70 '-')
