module LMath.Logic.Examples.Quillere
    (OuterLoop(..), codegen,

     countLines, countIfs, countExpr
    )
where

import Data.Ratio
import Data.Maybe
import Data.List
import Debug.Trace

import LMath.Logic.QE
import LMath.Logic.Sort
import LMath.Tools
import LMath.Types
import LMath.Logic.DTreeContexts

import LMath.Logic.Domain
import LMath.Logic.DomainContext
import LMath.Logic.DomainTree
import VisualPolylib

newtype OuterLoop = OuterLoop (Domain (Polynomial Rational))

instance Show OuterLoop where
    show (OuterLoop d) = "\n" ++
	                 "-------------\n" ++
	                 "  OuterLoop: \n" ++
			 "-------------\n" ++
			 show d ++ "\n"

data LBounds = LBounds { bLowerbounds :: [([Polynomial Rational], Polynomial Rational)]
                       , bUpperbounds :: [([Polynomial Rational], Polynomial Rational)]
                       , bGuards :: [[Polynomial Rational]]
                       , bVar :: Int
                       }

-- |
-- All input domains must be single polyhedra!
--
codegen :: DTree (Polynomial Rational)
                 [(OuterLoop, [(Domain (Polynomial Rational), [String])])]
        -> Maybe DomainContext
        -> String
codegen tree context = showTree nV $
		       simplifyDTree (fromDomainContext' context) $
		       fmap codegen_impl $
		       makeDTree treeContext $
		       sortTree $
		       filterEmptyLoops tree
  where
  nV = 2
  treeContext = polyDTreeContext () (fromDomainContext' context)
  filterEmptyLoops = fmap (filter (\(_, xs) -> not (null xs)))

codegen_impl :: [(LBounds, [(LBounds, [String])])]
             -> String
codegen_impl = concatMap genOut
    where
    nV = 2
    genOut (outBounds, innerBounds) =
      let innerLoopBody = concatMap genIn innerBounds
          genIn (innerBounds, stmts) =
	      genLoop innerBounds 1 (concat [x ++ ";\n" | x <- stmts])
      in genLoop outBounds 0 innerLoopBody
    genLoop bounds var body =
	let lowerbound = case lb of
			   [x] -> x
			   _   -> "max(" ++ concat (intersperse ", " lb) ++ ")"
	    upperbound = case ub of
                           [x] -> x
                           _   -> "min(" ++ concat (intersperse ", " ub) ++ ")"
            lb = [showLowerbound x nV | x <- bLowerbounds bounds]
            ub = [showUpperbound x nV | x <- bUpperbounds bounds]
	    (guardsBegin, guardsEnd) =
		case bGuards bounds of
	          [] -> ("", "")
	          xs -> let expr = concat (intersperse " && "
                                           [ineqevalcode x nV ++ " >= 0" | x <- xs])
                        in ("if (" ++ expr ++ ") {\n", "}\n")
        in guardsBegin ++
	   "for (int " ++ showVar var ++ " = " ++ lowerbound ++ "; " ++
                 showVar var ++ " <= " ++ upperbound ++ "; " ++
                 showVar var ++ "++) {\n" ++
           body ++
           "}\n" ++
           guardsEnd

sortTree :: DTree (Polynomial Rational)
                  [(OuterLoop, [(Domain (Polynomial Rational), [String])])]
         -> DTreeGen (Polynomial Rational) sc (Polynomial Rational)
                  [(LBounds, [(LBounds, [String])])]
sortTree tree = flattenDTreeG (fmap sortLeaves (convertToDTreeGen tree))

sortLeaves :: [(OuterLoop, [(Domain (Polynomial Rational), [String])])]
           -> DTreeGen (Polynomial Rational) sc (Polynomial Rational)
                    [(LBounds, [(LBounds, [String])])]
sortLeaves [] = dLeaf []
sortLeaves bounds = mergeListOfDTreeG (++) $
                    map (fmap (\x -> [x])) $
                    map sort' bounds
  where
  sort' (OuterLoop outerLoop, innerLoop) =
      case domainToIneqs outerLoop of
        [out] -> flattenDTreeG $
		 fmap (\x -> sortInnerLoops (x, innerLoop)) (sortPolyIneqs out 0)
        _ -> error "sortLeaves: All input domains must be single polyhedra."

sortInnerLoops :: (LBounds, [(Domain (Polynomial Rational), [String])])
               -> DTreeGen (Polynomial Rational) sc (Polynomial Rational)
                       (LBounds, [(LBounds, [String])])
sortInnerLoops (outD, xs) =
      let sort' (inD, ids) =
           case domainToIneqs inD of
	     [poly] -> fmap (\x -> [(x, ids)]) (sortPolyIneqs poly 1)
             _ -> error "sortInnerLoops: All input domains must be single polyhedra"
      in if null xs
         then dLeaf (outD, [])
         else fmap (\x -> (outD, x)) $
              mergeListOfDTreeG (++) $
	      map sort' xs

sortPolyIneqs :: GIneqSystem (Polynomial Rational)
              -> Int -- ^ variables
              -> DTreeGen (Polynomial Rational) sc (Polynomial Rational) LBounds
sortPolyIneqs p var = foldr sort' initBounds (map entries (rows p))
  where
  initBounds = dLeaf (LBounds { bLowerbounds = []
			      , bUpperbounds = []
			      , bGuards = []
			      , bVar = var
			      })
  sort' ineq boundTree = flattenDTreeG $ fmap (sortImpl ineq) boundTree
  sortImpl ineq bounds =
    let coeff = ineq !! var
        rem   = [if pos == var then 0 else c | (c,pos) <- zip ineq [0..]]
        tm = bounds { bUpperbounds = (rem, -coeff) : bUpperbounds bounds }
        t0 = bounds { bGuards = rem : (bGuards bounds) }
        tp = bounds { bLowerbounds = (map negate rem, coeff) :
                                     bLowerbounds bounds }
    in sCond coeff (dLeaf tm) (dLeaf t0) (dLeaf tp)

showVar var = "i" ++ show var
showIneq op (vect, div) nV
    | div == 1 = code
    | otherwise = code ++ " " ++ op ++ " " ++ show div
    where
    code = ineqevalcode vect nV

showLowerbound = showIneq "ceildiv"
showUpperbound = showIneq "floordiv"
showTree _ (DLeaf x) = x
showTree nV (SCond cond tm t0 tp) =
    "if (" ++ polyevalcode cond nV ++ " > 0) {\n" ++ showTree nV tp ++
    "} else if (" ++ polyevalcode cond nV ++ " == 0) {\n" ++ showTree nV t0 ++
    "} else { // " ++ polyevalcode cond nV ++ " < 0\n" ++ showTree nV tm ++
    "}\n" 
showTree nV (EqCond cond t0 tpm) =
    "if (" ++ polyevalcode cond nV ++ " == 0) {\n" ++ showTree nV t0 ++
    "} else {\n" ++ showTree nV tpm ++ "}\n"
showTree nV (GeCond _ cond t0p tm) =
    "if (" ++ polyevalcode cond nV ++ " >= 0) {\n" ++ showTree nV t0p ++
    "} else {\n" ++ showTree nV tm ++ "}\n"
showTree nV (FCond qfCond tt tf) =
    "if (" ++ condevalcode qfCond ++ ") {\n" ++ showTree nV tt ++
    "} else {\n" ++ showTree nV tf ++ "}\n"

findBounds :: Integral a
           => GIneqSystem (Polynomial (Ratio a))
           -> Int -- ^ variable
           -> ([String], [String]) -- (lowerbound, upperbound)
findBounds p var = verify $
		   foldl nextBound ([], []) (rows p)
    where
    nextBound (lb, ub) row = undefined -- entries row

    verify ([], _) = error "findBounds: empty lowerbounds"
    verify (_, []) = error "findBounds: empty upperbounds"
    verify xs = xs -- all tests passed

polyevalcode :: (Integral a, Integral b)
             => Polynomial (Ratio a)
             -> b -- ^ number of variables
             -> String
polyevalcode p nV = show p

ineqevalcode :: (Integral a, Integral b)
             => [Polynomial (Ratio a)]
             -> b -- ^ number of variables
             -> String
ineqevalcode xs nV
    | genericLength xs /= nV + 1 = error "ineqevalcode: length xs /= nV + 1"
    | otherwise = code
  where
  code = if all (==0) xs
	 then "0"
	 else concat $ intersperse " + " $
	      catMaybes $ zipWith f xs [0..]
  f x n | x == 0    = Nothing
	| n == nV   = Just (show x)
	| otherwise = Just (showVar n ++ " * (" ++ show x ++ ")")

condevalcode :: Integral a
             => QfFormula (Polynomial (Ratio a))
             -> String
condevalcode cond = "eval_qf_condition(<" ++ show cond ++ ">)"

-- |
-- Counts the number of newlines in the input string.
--
countLines :: Integral b
           => String -- ^ the input string
           -> b      -- ^ the number of newlines
countLines = genericLength . lines

countIfs :: Integral b
         => String
         -> b
countIfs = genericLength . (filter ("if" `isPrefixOf`)) . lines

countExpr :: Integral b
          => String
          -> b
countExpr = sum . map findExpr . lines
  where
  findExpr s | "if"  `isPrefixOf` s =
		 1 + genericLength (filter (== "&&") (words s))
	     | "for" `isPrefixOf` s =
		 3 + genericLength (filter (== ',') s)
	     | "S"   `isPrefixOf` s = 1
	     | otherwise = 0

--
-- Tests
--
example1 = DLeaf []

example2 :: DTree (Polynomial Rational)
                  [(OuterLoop, [(Domain (Polynomial Rational), [String])])]
example2 = DLeaf [(outerLoop, [(innerLoop, ["S1"])])]
  where
  outerIneq :: GIneqSystem (Polynomial Rational)
  outerIneq = ineqSystem ((2, 0), matrix [[1,   0,  0] -- i >= 0
					 ,[-1,  0, 10] -- i <= 10
					 ,[-p0, 0, p1] -- p0*i <= p1
					 ]
			 )
  innerIneq :: GIneqSystem (Polynomial Rational)
  innerIneq = ineqSystem ((2, 0), matrix [[0,  1,  -2] -- j >= 2
					 ,[0, -1,  20] -- j <= 20
					 ,[1, -1,   0] -- j <= i
					 ]
                         )
  p0 = indeterminate  2 0
  p1 = indeterminate  2 1

  outerLoop = OuterLoop $ ineqToDomain Nothing (2, 2) outerIneq
  innerLoop = ineqToDomain Nothing (2, 2) innerIneq

testIt ex = codegen ex Nothing
