-- |
-- Interface to algorithms using first-order logic.
-- This module provides access to algorithms which use
-- some helper functions implemented by different logic modules.
--
module LMath.Logic.QE
    ( -- * Logic
      simplifyQf, fastSimplifyQf,
      impliesQf,
      impliesSignum,
      eliminateQuantifiers,
      decideFormula,
      dnfQf,

      -- * Decision Tree Simplification
      simplifyDTree, fmapDTree_with_context,

      -- * Disjoint Unions
      disjointUnion, disjointUnionQf,
      disjointPolyhedra,

      -- * Other functions
      qepcadDisjointUnion,

      -- * Experimental functions
      impliesQfFC, impliesSignumFC
    )
where

import Control.Exception
import Control.Monad (liftM)
import Data.IORef
import Data.List
import Data.Maybe
import Data.Ratio
import System.IO.Unsafe (unsafePerformIO)
import System.IO ( hPutStrLn, stderr )
import Debug.Trace

import Config
import LMath.Algo
import LMath.Types
import LMath.Tools

mytrace _ a = a


sel :: [(Bool,a)] -> a
sel cvs = head [v | (c,v) <- cvs, c]

haveQepcad, haveReduce, haveSlfq :: Bool
haveQepcad = externalToolAvailable QEPCAD
haveReduce = externalToolAvailable REDUCE
haveSlfq   = externalToolAvailable SLFQ

selectOnFirst :: [(Bool,a)] -> [a]
selectOnFirst xs = [x | (True,x) <- xs]


reduceDecideQf :: Integral a => QfFormula (Polynomial (Ratio a)) -> Bool
reduceDecideQf = reduceImpliesQf []

-- |
-- Simplify a quantifier-free formula wrt. a given context.
-- When several simplification tools are available, the
-- simplifier selected by this function is (on average)
-- at least as fast as the simplifier called by 'simplifyQf', but
-- may not simplify as much as the simplifier used by 'simplifyQf'.
-- 'simplifyQf' and 'fastSimplifyQf' may use the same simplifier.
--
fastSimplifyQf :: RealFrac a
	   => [QfFormula (Polynomial a)]   -- ^ the context
           -> QfFormula (Polynomial a)     -- ^ the formula to simplify
	   -> QfFormula (Polynomial a)
fastSimplifyQf ctx qf =
    tryTools "fastSimplifyQf" $
    selectOnFirst [(haveReduce, (reduceFastSimplifyQf ctx $! qf, "Reduce")),
		   (haveQepcad, (qepcadSimplifyQf ctx $! qf, "Qepcad")),
		   (haveSlfq, (slfqSimplifyQf ctx $! qf, "SLFQ")),
		   (True, (warnNoQE qf, "(identity)"))]

-- |
-- Simplify a quantifier-free formula wrt. a given context.
-- Uses the currently selected simplifier to simplify a quantifier-free
-- formula with respect to the formulas given in the context.
--
simplifyQf :: RealFrac a
	   => [QfFormula (Polynomial a)]   -- ^ the context
           -> QfFormula (Polynomial a)     -- ^ the formula to simplify
	   -> QfFormula (Polynomial a)
simplifyQf ctx qf =
    tryTools "simplifyQf" $
    selectOnFirst [(haveSlfq, (slfqSimplifyQf ctx $! qf, "SLFQ")),
		   (haveReduce, (reduceSimplifyQf ctx $! qf, "Reduce")),
		   (haveQepcad, (qepcadSimplifyQf ctx $! qf, "Qepcad")),
		   (True, (warnNoQE qf, "(identity)"))]

--
-- Has the user been warned about missing QE tools?
--
{-# NOTINLINE warnedAboutMissingQE #-}
warnedAboutMissingQE :: IORef Bool
warnedAboutMissingQE = unsafePerformIO $ newIORef False

warnNoQE :: a -> a
warnNoQE x = printWarning `seq` x
    where
    printWarning = unsafePerformIO $
                   do w <- atomicModifyIORef warnedAboutMissingQE
                                             (\v -> (True,v))
                      if w
                         then return ()
                         else hPutStrLn stderr "WARNING: formula simplification (simplifyQf, fastSimplifyQf, etc.)\nis ineffective when configured without QE tools.\nConsider configuring HsLooPo with REDUCE, QEPCAD and/or SLFQ support."
    

-- |
-- Check a logical implication.
-- @impliesQf@ checks whether the formulas in a given context @C@
-- imply that the formula @f@ holds in the real numbers, i.e.,
-- @impliesQf@ decides whether the universal closure of
-- @\/\\C -> f@ is a theorem, or not. Uses the currently selected
-- implementation of @impliesQf@ from the global 'LogicOptions'.
--
impliesQf :: RealFrac a
          => [QfFormula (Polynomial a)]  -- ^ context @C@
          -> QfFormula (Polynomial a)    -- ^ formula @f@ to check
          -> Bool
impliesQf = --getLogicOption Nothing logicImpliesQf
            defaultImpliesQf

-- |
-- Default implementation for 'impliesQf'.
-- Tries Redlog first, then Qepcad, and CAD last.
--
defaultImpliesQf :: RealFrac a
                 => [QfFormula (Polynomial a)]
                 -> QfFormula (Polynomial a)
		 -> Bool
defaultImpliesQf ctx qf =
    tryTools "impliesQf" $
    selectOnFirst [(haveReduce, (reduceImpliesQf ctx qf, "Reduce"))
		  ,(haveQepcad, (qepcadImpliesQf ctx qf, "Qepcad"))
		  ]

-- |
-- Check whether a given context determines the sign of a polynomial.
-- @impliesSignum@ checks whether the context @C@ implies that the sign
-- of @p@ is determined, i.e., that @\/\\ C@ implies that @p<0@, @p<=0@,
-- @p=0@, @p\/=0@, @p>=0@, or @p>0@. The result is returned as a
-- 'Signum' record. Uses the currently selected
-- implementation of @impliesQf@ from the global 'LogicOptions'.
--
impliesSignum :: RealFrac a =>
		 [QfFormula (Polynomial a)]  -- ^ context @C@
	      -> Polynomial a                -- ^ polynomial @p@
              -> Signum
impliesSignum ctx p = --getLogicOption Nothing logicImpliesSignum
    tryTools "impliesSignum" $
    selectOnFirst [(haveReduce, (reduceImpliesSignum ctx p, "Reduce"))
		  ,(haveQepcad, (qepcadImpliesSignum ctx p, "Qepcad"))
		  ]
{-
                sel [-- (haveReduce&&haveQepcad, rqcImpliesSignum),
		     (haveQepcad, qepcadImpliesSignum),
		     (haveReduce, redCadImpliesSignum),
		     (True, cadImpliesSignum)]
-}

-- |
-- Compute a quantifier-free equivalent of a first-order logical
-- formula. A quantifier elimination tool is called to compute a
-- quantifier-free formula which is equivalent to the given
-- first-order logical formula in the real numbers.
-- A context can be given which may help the quantifier elimination
-- tool to produce a simpler result formula. There is no guarantee
-- that the quantifier elimination tool will use any information
-- present in the context when computing the result formula.
-- Uses the currently selected
-- implementation of @impliesQf@ from the global 'LogicOptions'.
--
eliminateQuantifiers :: RealFrac a
		     => [QfFormula (Polynomial a)]  -- ^ context
		     -> FoFormula (Polynomial a)    -- ^ input formula
		     -> QfFormula (Polynomial a)
		                 -- ^ equivalent quantifier-free formula
eliminateQuantifiers = --getLogicOption Nothing logicEliminateQuantifiers
                       defaultEliminateQuantifiers

-- |
-- Default implementation for 'eliminateQuantifiers':
-- Try Redlog first, and if Redlog fails, call Qepcad.
--
defaultEliminateQuantifiers :: RealFrac a
		        => [QfFormula (Polynomial a)]
		        -> FoFormula (Polynomial a)
		        -> QfFormula (Polynomial a)
defaultEliminateQuantifiers ctx fo =
    fo `seq`
    tryTools "eliminateQuantifiers" $
    selectOnFirst [(haveReduce, (reduceEliminateQuantifiers ctx fo, "Reduce")),
		   (haveQepcad, (qepcadEliminateQuantifiers ctx fo, "QEPCAD"))]

-- |
-- Decide a formula under a given context. Returns
-- whether the given formula is equivalent to @True@ or @False@.
-- (When the formula is not equivalent to @True@ or @False@,
-- an error is thrown.)
--
decideFormula :: Integral a
              => [QfFormula (Polynomial (Ratio a))]
              -> FoFormula (Polynomial (Ratio a))
              -> Bool
decideFormula ctx foformula =
    case eliminateQuantifiers ctx foformula of
        QfAt AtTrue  -> True
        QfAt AtFalse -> False
        qf -> error ("decideFormula: formula " ++ show foformula ++
                     " is not equivalent to True or False under the context "
                     ++ show ctx ++ " (is equivalent to " ++ show qf ++ ")")
-- |
-- Simplify a decision tree.
-- A given decision tree is simplified under the given context,
-- i.e., irrelevant branches of the tree are cut by checking whether
-- some in the tree (and the context) imply that some other conditions
-- further below in the tree are either redundant or contradictive.
-- Uses the currently selected
-- implementation of @impliesQf@ from the global 'LogicOptions'.
--
simplifyDTree :: Integral a =>
		 [QfFormula (Polynomial (Ratio a))] ->
		 DTree (Polynomial (Ratio a)) val ->
		 DTree (Polynomial (Ratio a)) val
simplifyDTree = --getLogicOption Nothing logicSimplifyDTree
                topdownSimplifyDTree isConstPoly
		   impliesSignum impliesQf simplifyQf

-- |
-- Simplifies the decision tree @dt@ and applies the operation @f@
-- to all leaves. Apart from the initial simplification, the only
-- difference between this function and a simple 'fmap' is that fmap
-- does not consider the context.
--
-- In this function the context will be collected and passed to the
-- transformation operation @f@. You can additionally define an
-- initial context which will be added to the collected context.
--
-- See also: 'fmap', 'simplifyDTree',
--           'LMath.Types.DTree.unsafe_fmapDTree_with_context'
--
fmapDTree_with_context :: (Integral c)
		       => [QfFormula (Polynomial (Ratio c))]
		       -- ^ initial context
                       -> ([QfFormula (Polynomial (Ratio c))] -> a -> b)
		       -- ^ the transformation operation @f@
		       -> DTree (Polynomial (Ratio c)) a
		       -- ^ the original tree @dt@
		       -> DTree (Polynomial (Ratio c)) b
fmapDTree_with_context context f dt =
    let simpl_dt = simplifyDTree context dt
    in unsafe_fmapDTree_with_context context f simpl_dt

-- |
-- Compute a disjoint union of polyhedra.
-- A union of the given polyhedra (represented by inequality systems)
-- is calculated such that every /integral/ point which lies in one of
-- the input polyhedra lies in exactly one output polyhedron and the output
-- polyhedra do not contain integral points which are not part of any
-- input polyhedron. Note that the just specified property does not
-- hold for rational (or real) points. In general, the disjoint union
-- of polyhedra requires strict inequalities in the output conjunctive sets,
-- so the output can not be represented by polyhedra. A representation as
-- polyhedra is only possible due to the restriction to integral points
-- (by converting a strict inequality @e \> 0@ with a denominator-free
-- expression @e@ to @e-1 \>= 0@).
-- Uses the currently selected
-- implementation of @impliesQf@ from the global 'LogicOptions'.
--
disjointUnion :: Integral a
	      => [QfFormula (Polynomial (Ratio a))]  -- ^ context
              -> [GIneqSystem (FracPoly (Ratio a))]  -- ^ input inequality systems
	      -> [GIneqSystem (FracPoly (Ratio a))]
disjointUnion = --getLogicOption Nothing logicDisjointUnion
                defaultDisjointUnion

-- |
-- Compute a \"disjoint union\" of a quantifier-free formula.
-- The function returns a list @[a1,a2,...,aN]@ of lists of
-- atomic formulas. The following properties hold:
--
-- >  R  |=  (/\a1) \/ (/\a2) \/ ... \/ (/\aN)   <->   f
-- >  R  |=  (/\aK) /\ (/\aL)  <->  0       for K /= L
--
-- Uses the currently selected
-- implementation of @impliesQf@ from the global 'LogicOptions'.
--
disjointUnionQf :: Integral a
		=> [QfFormula (Polynomial (Ratio a))]    -- ^ context
	        -> QfFormula (Polynomial (Ratio a))      -- ^ input formula
	        -> [[AtFormula (Polynomial (Ratio a))]]
disjointUnionQf = --getLogicOption Nothing logicDisjointUnionQf
                  sel [(haveReduce, reduceDisjointUnionQf),
		       (True, error "`disjointUnionQf' not available (no \
				    \suitable tool configured)")]

-- |
-- Compute a (positive) disjunctive normal form of a given formula.
-- The given context may be used by the implementation to
-- simplify the result (e.g., drop irrelevant disjuncts).
--
dnfQf :: Integral a
      => [QfFormula (Polynomial (Ratio a))]    -- ^ context
      -> QfFormula (Polynomial (Ratio a))      -- ^ input formula
      -> [[AtFormula (Polynomial (Ratio a))]]
dnfQf = --getLogicOption Nothing logicDisjunctiveNormalFormQf
        sel [(haveReduce, reduceDnfQf),
	     (True, \ctx f -> positiveDnfQf f)]


tryTools :: String -> [(a, String)] -> a
tryTools functionName [] =
    error ("tryTool: no tools available for function " ++ functionName)
tryTools functionName tools = unsafePerformIO $ tt tools
    where
    tt [] = error ("tryTools: all tools failed for " ++ functionName)
    tt (v:rest) =
--	trace (functionName ++ ": " ++ snd v) $
	toolIOError Nothing {- Just functionName -} (evaluate $ fst v) (tt rest)


{-# NOTINLINE deg2counter #-}
deg2counter :: IORef (Int,Int)
deg2counter = unsafePerformIO $ newIORef (0,0)


qepcadDisjointUnion :: Integral a =>
	    	       [QfFormula (Polynomial (Ratio a))] ->
		       [GIneqSystem (FracPoly (Ratio a))] ->
		       [GIneqSystem (FracPoly (Ratio a))]
qepcadDisjointUnion ctx =
    disjointUnionFunction ctx qepcadImpliesQf
                          (qepcadImpliesSignum ctx)
			  disjointUnionAlg

-- |
-- Default implementation for 'disjointUnion'. Uses
-- the currently selected functions from the
-- global 'LogicOptions'.
--
defaultDisjointUnion :: Integral a =>
		    [QfFormula (Polynomial (Ratio a))] ->
		    [GIneqSystem (FracPoly (Ratio a))] ->
		    [GIneqSystem (FracPoly (Ratio a))]
defaultDisjointUnion ctx =
    disjointUnionFunction ctx impliesQf
			  (impliesSignum ctx)
			  disjointUnionAlg

applyMbList :: ([a] -> b) -> [Maybe a] -> Maybe b
applyMbList f xs
    | any (isNothing) xs = Nothing
    | otherwise          = Just $ f $ catMaybes xs


-- |
-- Compute intersection and differences of two polyhedra.
-- 
disjointPolyhedra :: Integral a
		  => [QfFormula (Polynomial (Ratio a))]  -- ^ context
		  -> GIneqSystem (Polynomial (Ratio a))  -- ^ polyhedron @A@
		  -> GIneqSystem (Polynomial (Ratio a))  -- ^ polyhedron @B@
		  -> ([GIneqSystem (Polynomial (Ratio a))],
		      GIneqSystem (Polynomial (Ratio a)),
		      [GIneqSystem (Polynomial (Ratio a))]
		     )
disjointPolyhedra ctx ineqsA ineqsB =
    (concatMap (fromJust . atFormulasToIneqSystems nV) aAndNotB,
     ineqsA `join` ineqsB,
     concatMap (fromJust . atFormulasToIneqSystems nV) bAndNotA
    )
    where
    nV = nVars ineqsA
    a = ineqSystemToFormula ineqsA
    b = ineqSystemToFormula ineqsB
    aAndNotB = disjointUnionQf ctx (a `QfAnd` QfNot b)
    bAndNotA = disjointUnionQf ctx (b `QfAnd` QfNot a)



impliesSignumFC :: Integral a
		=> [QfFormula (IFracPolyFC a)]
		-> IFracPolyFC a
		-> Signum
impliesSignumFC = genericImpliesSignum impliesQfFC

impliesQfFC :: Integral a
	    => [QfFormula (IFracPolyFC a)]
	    -> QfFormula (IFracPolyFC a)
	    -> Bool
impliesQfFC ctx qf =
--    case eliminateQuantifiers [] (FoAll [0..nIndet-1] formula) of
{-    case eliminateQuantifiers [] (FoAll [nPseudos..nPseudos+nIndet-1] formula) of
      QfAt AtTrue  -> True
      QfAt AtFalse -> False
      _ -> error "impliesQfFC: result is not AtTrue or AtFalse ?!"
-}
--    trace ("Context: " ++ show ctx ++ "\nQuestion: " ++ show qf ++ " ? " ++ show res) $
    res
    where
    res = impliesQf [] formula
    (nIndet, nPseudos, formula) = fcQf2fo (conjunctionQf ctx `qfImpl` qf)

fcQf2fo :: Integral a
	=> QfFormula (IFracPolyFC a)
        -> (Integer, Integer, QfFormula (Polynomial (Ratio a)))
fcQf2fo inputQf =
--    trace (show inputQf) $
    (nIndets-nPseudos, nPseudos, fo)
    where
    fo = --FoAll [0..nPseudos-1] $ FoQf $
	 (conjunctionQf [QfAnd (QfAt $ AtGe p)
                               (QfOr (QfAt $ AtGe (unFP(denom-1-(p%%1)))) (QfAt $ AtGe (unFP (-denom-1-(p%%1)))))
--			         (QfAt $ AtGt (unFP (denom-(p%%1))))
			| i <- [0..nPseudos-1], let p = indeterminate nIndets i
                        | denomI <- floorDenoms, let denom = ifracPolyToFracPoly denomI]
	  `qfImpl` qf')
    preparedQf = prepare inputQf

    qf' = fmap (unIFP.pseudoExpression.unifyPs) preparedQf
    (floorDenoms, unifyPs) = unifyPseudos $ termsQf preparedQf
    nPseudos = genericLength floorDenoms

    nIndets = unNIndeterminates $ nIndeterminatesN $ termsQf qf'

    unIFP fp = mapPoly (% 1) (ifpNumerator fp * ifpDenominator fp)

    unFP fp = polyNumerator fp * polyDenominator fp

    prepare qf = qfMapAt prepFloors qf

    prepFloors (AtGe p)
	| Just q <- refact p  =  AtGe q
    prepFloors (AtEq p)
	| Just q <- refact p  =  AtEq q
    prepFloors qf = qf

    -- das hier soll wohl irgendwas bei Ausdruecken der Form
    -- a*floor(b)^c  optimieren (wenn ich nur wuesste, was
    -- ich mir dabei mal gedacht habe...)       AG
    -- 
    refact p = case isFloorMultipleFC p of
	         Nothing -> Nothing
		 Just (ps,co) ->
		     Just (mkFCi co * product [floorFC (f p) ^ e | (p,e) <- ps])
        where
	f p = case refact p of
	        Nothing -> p
		Just q  -> q

{-
fcQf2fo intparams qf = (nIndet, nPseudos, univQuantPs $ gen (nIndet+nPseudos))
    where
    univQuantPs f =
	FoAll [nIndet..nIndetAndPseudos-1] $ FoQf $
	(conjunctionQf [QfAnd (QfAt $ AtGe p) (QfAt $ AtGt (1-p))
		       | i<-[nIndet..nIndetAndPseudos-1],
		         let p = indeterminate nIndetAndPseudos i]
	   `qfImpl` f)
    (gen, (_,nIndetAndPseudos)) = conv ([],nIndet) qf
    nPseudos = nIndetAndPseudos - nIndet
    nIndet = unNIndeterminates $ nIndeterminatesN $
	     concatMap polysFC $ termsQf qf

    conv i (QfNot a) = (QfNot . a', i')
	where
	(a',i') = conv i a
    conv i (QfAnd a b) = (\n -> QfAnd (a' n) (b' n), i'')
	where
	(a',i')  = conv i  a
	(b',i'') = conv i' b
    conv i (QfOr a b) = (\n -> QfOr (a' n) (b' n), i'')
	where
	(a',i')  = conv i  a
	(b',i'') = conv i' b
    conv i (QfAt a) = (QfAt . a', i')
	where
	(a',i') = convAt i a

    convAt i AtTrue  = (const AtTrue,i)
    convAt i AtFalse = (const AtFalse,i)
    convAt i (AtEq fc) = let (fc',i') = convFC i fc in (AtEq . fc', i')
    convAt i (AtNeq fc) = let (fc',i') = convFC i fc in (AtNeq . fc', i')
    convAt i (AtGe fc) = let (fc',i') = convFC i fc in (AtGe . fc', i')
    convAt i (AtGt fc) = let (fc',i') = convFC i fc in (AtGt . fc', i')

    convFC i (Ceil (FP fp))
	| Just poly <- isIntFP fp  =
		       (const $ appendIndeterminates nPseudos poly, i)
    convFC i (Floor (FP fp))
    	| Just poly <- isIntFP fp  =
		       (const $ appendIndeterminates nPseudos poly, i)
    convFC i (FP fp) = (const $ appendIndeterminates nPseudos (num*denom), i)
	where
	num   = polyNumerator fp
	denom = polyDenominator fp
    convFC i (Sum [x]) = convFC i x
    convFC i (Sum (x:xs)) = convFC2 (+) i x (Sum xs)
    convFC i (Prod [x]) = convFC i x
    convFC i (Prod (x:xs)) = convFC2 (*) i x (Prod xs)
    convFC i (Quot a b) = convFC2 (*) i a b
    convFC i (Floor a) =
	case findPseudo (Floor a) pl of
            Just a'' -> (a'', (pl,i'))
	    Nothing ->
		let a'' = \n -> a' n - indeterminate n i'
		in (a'', ((Floor a,a',\n -> indeterminate n i'):pl, i'+1))
	where
	(a',(pl,i')) = convFC i a
    convFC i (Ceil a)  =
	case findPseudo (Ceil a) pl of
            Just a'' -> (a'', (pl,i'))
	    Nothing ->
		let a'' = \n -> a' n + indeterminate n i'
		in (a'', ((Ceil a,a',\n -> indeterminate n i'):pl, i'+1))
	where
	(a',(pl,i')) = convFC i a

    findPseudo (Floor a) pl =
	case catMaybes $ map f pl of
	    []   -> Nothing
	    a':_ -> Just a'
	where
	f (Floor b, b', ind)
		| a == b = Just (\n -> b' n - ind n)
	f (Ceil b, b', ind)
		| a == negate b = Just (\n -> b' n - ind n)
	f _ = Nothing
    findPseudo (Ceil a) pl =
	case catMaybes $ map f pl of
	    []   -> Nothing
	    a':_ -> Just a'
	where
	f (Ceil b, b', ind)
		| a == b = Just (\n -> b' n + ind n)
	f (Floor b, b', ind)
		| a == negate b = Just (\n -> ind n - b' n)
	f _ = Nothing

    convFC2 op i a b = (\n -> a' n `op` b' n, i'')
	where
	(a',i')  = convFC i  a
	(b',i'') = convFC i' b

    isIntFP :: Integral b => FracPoly (Ratio b) -> Maybe (Polynomial (Ratio b))
    isIntFP fp =
	case isConstPoly $ polyDenominator fp of
	    Just cdenom
		| all ((==1).denominator) $ collectCoeffs num',
		  all (`elem` intparams) $ usedIndeterminates num' -> Just num'
		| otherwise   ->  Nothing
		where
		num' = mapPoly (/cdenom) $ polyNumerator fp
	    Nothing -> Nothing


    unFP fp = polyNumerator fp * polyDenominator fp
-}
