module LMath.Logic.Sort
    ( -- * Sorting Inequality Systems
      sortIneqsPss,
      fmSortIneqs,
      qeSortIneqs,

      -- * Specialised Functions
      fmSortIneqsReduce, fmSortIneqsQepcad,

      -- * Experimental Functions
      qeSortIneqsGss
    )
where

import Data.Maybe
import Data.Ratio

import Debug.Trace

import LMath.Algo
import LMath.Types
import LMath.Tools
import LMath.Logic.QE

-- |
-- Sort an inequality system.
--
-- The result is a sorted inequality system (represented by a 'SortedSystem'
-- structure). The first variable of the system will be the
-- outermost variable in the projection, i.e., its bounds will
-- depend on the parameters only.
-- The number of \"outer\" variables specified as second argument
-- denotes for how many variables (counted from outside in) no projection
-- is calculated. With @outer=Just 2@, for example, the result will start
-- with the bounds of the third variable in dependence of the first and
-- second variable (and the parameters, of course); no bounds for the
-- first and second variable will be calculated.
--
-- NOTE: If the number of \"outer\" variables is set to @Nothing@, the
-- projection will be calculated as for @Just 0@ (i.e., the result starts
-- with bounds for the first variable in dependece of the parameters), but
-- it is guaranteed that all non-@Nothing@ leaves of the result
-- represent feasible inequality systems. To put it in other words:
-- when the number of \"outer\" variables is set to @Nothing@, no
-- infeasible solutions are present in the result (they are replaced
-- by @Nothing@).
--
sortIneqsPss
    :: (Integral a, Integral b)
    => [QfFormula (Polynomial (Ratio a))]     -- ^ context
    -> GIneqSystem (FracPoly (Ratio a))       -- ^ input inequality system
    -> Maybe (GEqSystem (FracPoly (Ratio a))) -- ^ optional input equality system
    -> Maybe b                          -- ^ number of \"outer\" variables
    -> DTree (Polynomial (Ratio a))
             (Maybe (SortedSystem (FracPoly (Ratio a))))
sortIneqsPss = fmSortIneqs

-- |
-- Perfom Fourier-Motzkin elimination.
-- This function calls the Fourier-Motzkin elimination method.
--
-- See 'sortIneqsP' for a specification.
--
fmSortIneqs
    :: (Integral a, Integral b)
    => [QfFormula (Polynomial (Ratio a))]     -- ^ context
    -> GIneqSystem (FracPoly (Ratio a))       -- ^ input inequality system
    -> Maybe (GEqSystem (FracPoly (Ratio a))) -- ^ optional input equality system
    -> Maybe b                                -- ^ number of \"outer\" variables
    -> DTree (Polynomial (Ratio a))
             (Maybe (SortedSystem (FracPoly (Ratio a))))
fmSortIneqs ctx ineqs mbEqs ctxVars =
    fourierMotzkinP impliesSignum ctx ineqs mbEqs ctxVars


fmSortIneqsQepcad :: (Integral a, Integral b) =>
		     [QfFormula (Polynomial (Ratio a))] ->
		     GIneqSystem (FracPoly (Ratio a))   ->
		     Maybe (GEqSystem (FracPoly (Ratio a))) ->
		     Maybe b ->
		     DTree (Polynomial (Ratio a))
			   (Maybe (SortedSystem (FracPoly (Ratio a))))
fmSortIneqsQepcad ctx ineqs mbEqs mbCtxVars =
    fourierMotzkinP qepcadImpliesSignum ctx ineqs mbEqs mbCtxVars

fmSortIneqsReduce :: (Integral a, Integral b) =>
		     [QfFormula (Polynomial (Ratio a))] ->
		     GIneqSystem (FracPoly (Ratio a))   ->
		     Maybe (GEqSystem (FracPoly (Ratio a))) ->
		     Maybe b ->
		     DTree (Polynomial (Ratio a))
			   (Maybe (SortedSystem (FracPoly (Ratio a))))
fmSortIneqsReduce ctx ineqs mbEqs mbCtxVars =
    fourierMotzkinP reduceImpliesSignum ctx ineqs mbEqs mbCtxVars


qeSortIneqs :: (Integral a, Integral b)
    => GIneqSystem (Ratio a)       -- ^ input inequality system
    -> Maybe (GEqSystem (Ratio a)) -- ^ optional input equality system
    -> Maybe b                     -- ^ number of \"outer\" variables
    -> Maybe (SortedSystem (FracPoly (Ratio a)))
qeSortIneqs ineqs mbEqs = undefined
    where
    

qeSortIneqsGss :: (Integral a, Integral b)
    => GIneqSystem (FracPoly (Ratio a))       -- ^ input inequality system
    -> Maybe (GEqSystem (FracPoly (Ratio a))) -- ^ optional input equality system
    -> Maybe b                                -- ^ number of \"outer\" variables
    -> DTreeGen (Polynomial (Ratio a)) (Polynomial (Ratio a))
             (Maybe (SortedSystem (FracPoly (Ratio a))))
qeSortIneqsGss ineqs mbEqs mbCtxVars =
    withDTreeContext (\dtc ->
		      let projs = startFormula (dtcContext dtc) :
		                  [elim dtc v p | p <- projs | v <- [nV-1,nV-2..0]]
		      in
		      trace (show projs) $
		      bounds projs (nV - 1)
		     )
    where
    startFormula ctx =
	conjunctionQf $ map QfAt $
	case mbEqs of
	  Just eqs  ->  startIneqs ctx ++ startEqs
	      where
	      startEqs = eqSystemToAtFormulas $ unfracEqSystem eqs
	  Nothing   ->  startIneqs ctx
    startIneqs ctx = ineqSystemToAtFormulas $
		     unfracIneqSystem (Just $ impliesSignum ctx) ineqs

    nV = nVars ineqs

    elim dtc var qf = 
	eliminateQuantifiers (dtcContext dtc) $
	FoEx [var] $ FoQf qf

    cv = fromIntegral $ fromMaybe 0 mbCtxVars

    bounds :: Integral a
	   => [QfFormula (Polynomial (Ratio a))]
	   -> Integer
	   -> DTreeGen (Polynomial (Ratio a)) (Polynomial (Ratio a)) (Maybe (SortedSystem (FracPoly (Ratio a))))
    bounds (formula:rest) var = 
	trace ("var = " ++ show var ++ ", formula = " ++ show formula ++ ":") $
	if var >= cv then
	     withDTreeContext (\dtc ->
			       let ctx = map (fmap (liftPolyN nV)) $ dtcContext dtc
			           formula' = slfqSimplifyQf ctx $
			                      formula
			       in
			       compDTreeG (\conj ->
					   compDTreeG (q conj) $
					   bounds rest (var-1)
					  ) $
			       makeConj formula' var nV
			      )
	else dLeaf $ Just []
	where
	q _    Nothing   = dLeaf Nothing
	q conj (Just ss) =
	    case qfFormulaToBounds conj var nV of
	        Nothing  -> dLeaf Nothing
		Just dtg -> fmap (\bnds -> Just (ss ++ [bnds])) dtg
			   

makeConj :: Integral a
	 => QfFormula (Polynomial (Ratio a))
	 -> Integer
         -> Integer
	 -> DTreeGen (Polynomial (Ratio a)) (Polynomial (Ratio a)) (QfFormula (Polynomial (Ratio a)))
makeConj qf var nV
    | isConjunctionQf qf = dLeaf qf
    | otherwise =
	withDTreeContext (\dtc ->
			  let conds = map (implCond (dtcContext dtc)) ats
			  in  trace (show $ zip ats conds) $ make [] $ zip ats conds
			 )
    where
    ats = extractAt $ eliminateAtTF qf
    implCond ctx atf =
	fmap (deleteIndeterminates nV) $
	eliminateQuantifiers ctx (FoAll [0..nV-1] $ FoQf (qf `qfImpl` QfAt atf))
    make bnds [] =
        withDTreeContext (\dtc ->
			  let f  = conjunctionQf $ map QfAt bnds
			      ctx = map (fmap (liftPolyN nV)) $ dtcContext dtc
			      f' = slfqSimplifyQf ctx f
			  in
			  trace (show (bnds,f')) $  dLeaf f'
			 )
    make bnds ((a,cond):rest) =
	fCond cond (make (a:bnds) rest) (make bnds rest)
