-- |
-- Interface to the Qepcad quantifier elimination tool.
--
module LMath.Tools.Qepcad
    ( qepcadImpliesQf, qepcadImpliesSignum,
      qepcadSimplifyQf, qepcadEliminateQuantifiers,
      qepcadGeoformEtf,
      slfqSimplifyQf,
      runSlfqIO,
      setQepcadCells, setSlfqCells,
      parseQepcadQfFormula, isNonETF,
      qfFormula2qepcad, qfFormulaPRE2qepcad,
      parseEtfFormula, parseFormula,
      qepcadEtfEliminateQuantifiers
    , qepcadEtfImpliesQf

    -- * Algebraic Number Arithmetic Using Qepcad
    , QAlg
    )
where

import qualified Config
import LMath.Types
import LMath.Types.Polynomial.AlgNum ( parToAlgnums, algNumRoots )
import LMath.Types.Polynomial.Roots  ( allRoots, rootInterval )
import LMath.Tools.ToolIO

import Control.Exception ( assert, catchDyn, throwDyn )
import Control.Monad
import Data.Char
import Data.Graph
import Data.Ratio
import Data.IORef
import Data.List
import Data.Maybe ( fromMaybe, fromJust )
import System.IO
import System.IO.Unsafe ( unsafePerformIO )
import Text.ParserCombinators.Parsec as P

import Debug.Trace


data QepcadOpts = QepcadOpts { qepcadCells :: Integer
			     , qepcadSlfqCells :: Integer
			     }

getQepcadCommand :: IO FilePath
getQepcadCommand = Config.externalToolPathOrError Config.QEPCAD

getSlfqCommand :: IO FilePath
getSlfqCommand = Config.externalToolPathOrError Config.SLFQ

{-# NOTINLINE qepcadConn #-}
qepcadConn :: ToolConnection
qepcadConn = unsafePerformIO $
	     newToolConnection True getQepcadCommand args "Qepcad" "" False
    where
    args = do cells <- getQepcadCells
	      let cellArg = "+N" ++ show (cells*10^6)
              return [cellArg]

{-# NOTINLINE qepcadOpts #-}
qepcadOpts :: IORef QepcadOpts
qepcadOpts = unsafePerformIO $
	     newIORef ( QepcadOpts { qepcadCells = 100,
				     qepcadSlfqCells = 10
				   }
		      )

getQepcadCells :: IO Integer
getQepcadCells = liftM qepcadCells (readIORef qepcadOpts)

setQepcadCells :: Integer -> IO ()
setQepcadCells n = modifyIORef qepcadOpts (\ro -> ro { qepcadCells = n })

getSlfqCells :: IO Integer
getSlfqCells = liftM qepcadSlfqCells (readIORef qepcadOpts)

setSlfqCells :: Integer -> IO ()
setSlfqCells n = modifyIORef qepcadOpts (\ro -> ro { qepcadSlfqCells = n })

varnames :: [String]
varnames = ['x' : show i | i <- [0..]]

--
-- Output a formula in Qepcad/SLFQ format.
-- Get rid of constant expression (e.g., 2 >= 0),
-- eliminate AtTrue and AtFalse.
-- Return also the number of indeterminates in the _simplified_
-- formula (which may be different from the number of indeterminates
-- in the input formula since some terms may have been deleted).
-- When the given formula is determined to be equivalent to
-- AtTrue for trivial reasons,  ("", AtLeast 0)  is returned.
--
qfFormula2qepcad :: Real a
		 => QfFormula (Polynomial a)
		 -> (String, NIndeterminates)
qfFormula2qepcad f = qfFormulaPRE2qepcad (fmap PREPoly f)

qfFormulaPRE2qepcad :: Real a
		    => QfFormula (TarskiPRE a)
		    -> (String, NIndeterminates)
qfFormulaPRE2qepcad f =
    (f2q simplF, nIndeterminatesN $ map prePoly $ termsQf simplF)
    where
    --
    -- Simplify the given formula and bring it into the
    -- right format before determining the the number of indeterminates
    -- ( QfAt (AtGe (constPolyN 2 42))  is simplified to  QfAt AtTrue;
    -- the number of indeterminates _must_ be returned as 0 then, not 2!)
    --
    simplF = eliminateAtTF $ positiveQf $
	     eliminateAtSignum constPRESignum $ intPolys f
    intPolys = fmap preRatPoly2intPoly

    preRatPoly2intPoly (PREPoly p) = PREPoly $ snd $ ratPoly2intPoly p
    preRatPoly2intPoly (PRERoot sgn n p) =
        PRERoot sgn n $ snd $ ratPoly2intPoly p

    f2q :: Integral a => QfFormula (TarskiPRE a) -> String
    f2q (QfAt at)     = a2q at
    f2q (QfAnd f1 f2) = "[" ++ f2q f1 ++ "] /\\ [" ++ f2q f2 ++ "]" 
    f2q (QfOr  f1 f2) = "[" ++ f2q f1 ++ "] \\/ [" ++ f2q f2 ++ "]" 
    f2q (QfNot _)     = error "internal error: LMath.Logic.Qepcad.f2q \
			      \should never be called with (QfNot _)"

    -- be careful with "/=" operator:
    -- a "/" can also start a fraction, i.e.
    -- in "2/=0" Qepcad expects a digit after "2/"
    -- => insert a space before "/=" !
    --
    a2q AtTrue    = ""  -- only reached when f is trivial (then AtTrue only atomic formula)
    a2q AtFalse   = "dummy*dummy < 0"  --error "qfFormula2qepcad: AtFalse should not occur"
    a2q (AtEq  (PREPoly p)) = p2q 0 p ++ "=0"
    a2q (AtNeq (PREPoly p)) = p2q 0 p ++ " /=0"
    a2q (AtGe  (PREPoly p))
	| leadingCoeff p >= 0 = p2q 0 p ++ ">=0"
	| otherwise           = p2q 0 (negate p) ++ "<=0"
    a2q (AtGt  (PREPoly p))
	| leadingCoeff p >= 0 = p2q 0 p ++ ">0"
	| otherwise           = p2q 0 (negate p) ++ "<0"

    a2q (AtEq  (PRERoot sgn   n p)) = pre "="  n p
    a2q (AtNeq (PRERoot sgn   n p)) = pre "/=" n p
    a2q (AtGe  (PRERoot True  n p)) = pre ">=" n p
    a2q (AtGt  (PRERoot True  n p)) = pre ">"  n p
    a2q (AtGe  (PRERoot False n p)) = pre "<=" n p
    a2q (AtGt  (PRERoot False n p)) = pre "<"  n p

    pre op n p =
        varnames!!(fromIntegral k) ++ op ++ "_root_" ++
        show n ++ " " ++ p2q 0 p
        where 
        k = last $ usedIndeterminates p

    p2q n p = mkT1 (head r) ++ concatMap mkT (tail r)
	where
	r = p2q' n p
	mkT (True,v)  = '+':v
	mkT (False,v) = '-':v
	mkT1 (True,v)  = v
	mkT1 (False,v) = '-':v
    p2q' n p
	| Just c <- isConstPoly p = [(c>=0, show (abs c))]
	| otherwise  =
	    [(sgn,v ++ expo (varnames!!n) e)
	     | (c,e) <- ces, (sgn,v) <- p2q' (n+1) c]
        where
	ces = coeffsAndExps (univariate p)
	expo name e
	    | e == 0 = ""
	    | e == 1 = ' ':name
	    | e >  1 = ' ':name ++ '^':show e


qepcadImpliesQf :: Real a
                => [QfFormula (Polynomial a)]
                -> QfFormula (Polynomial a)
		-> Bool
qepcadImpliesQf ctx formula =
    case callQepcad Nothing $ FoAll [0..nUnk-1] (FoQf question) of
        QfAt AtTrue  -> True
	QfAt AtFalse -> False
	_              ->
	    error "QEPCAD's answer to a decision problem is neither \
		  \TRUE nor FALSE ?!"
    where
    question = conjunctionQf ctx `qfImpl` formula
    nUnk = unNIndeterminates $ nIndeterminatesN $ termsQf question


qepcadImpliesSignum :: Real a
		    => [QfFormula (Polynomial a)]
		    -> Polynomial a
		    -> Signum
qepcadImpliesSignum = genericImpliesSignum qepcadImpliesQf


qepcadSimplifyQf :: Real a
		 => [QfFormula (Polynomial a)]
	         -> QfFormula (Polynomial a)
		 -> QfFormula (Polynomial a)
qepcadSimplifyQf ctx formula =
    callQepcad (Just $ conjunctionQf ctx) (FoQf formula)

qepcadEtfImpliesQf :: Real a
                   => [QfFormula (TarskiPRE a)]
                   -> QfFormula (TarskiPRE a)
                   -> Bool
qepcadEtfImpliesQf ctx formula =
    case qepcadEtfEliminateQuantifiers ctx foQuestion of
      QfAt AtTrue  -> True
      QfAt AtFalse -> False
      qf           -> error ("qepcadEtfImpliesQf: answer to a decision \
                             \problem is " ++ show qf ++ " !?")
    where
    foQuestion = FoAll [0..nUnk-1] $ FoQf question
    nUnk = unNIndeterminates $ nIndeterminatesN $ map prePoly $
           termsQf question
    question = conjunctionQf ctx `qfImpl` formula

qepcadEliminateQuantifiers :: Real a
			   => [QfFormula (Polynomial a)]
			   -> FoFormula (Polynomial a)
			   -> QfFormula (Polynomial a)
qepcadEliminateQuantifiers ctx (FoAll [] formula) =
    qepcadEliminateQuantifiers ctx formula
qepcadEliminateQuantifiers ctx (FoQf formula) = formula
qepcadEliminateQuantifiers ctx formula =
    callQepcad (Just $ conjunctionQf ctx) formula

qepcadEtfEliminateQuantifiers :: Real a
			      => [QfFormula (TarskiPRE a)]
			      -> FoFormula (TarskiPRE a)
			      -> QfFormula (Polynomial a)
qepcadEtfEliminateQuantifiers ctx (FoAll [] formula) =
    qepcadEtfEliminateQuantifiers ctx formula
qepcadEtfEliminateQuantifiers ctx formula =
    callEtfQepcad qepModeTarski Nothing (Just $ conjunctionQf ctx) formula

qepcadEtfEliminateQuantifiersEtf :: Real a
			      => [QfFormula (TarskiPRE a)]
			      -> FoFormula (TarskiPRE a)
			      -> QfFormula (TarskiPRE a)
qepcadEtfEliminateQuantifiersEtf ctx (FoAll [] formula) =
    qepcadEtfEliminateQuantifiersEtf ctx formula
qepcadEtfEliminateQuantifiersEtf ctx formula =
    callEtfQepcad qepModeEtf Nothing (Just $ conjunctionQf ctx) formula

qepcadGeoformEtf :: Real a
                 => [QfFormula (Polynomial a)]
                 -> FoFormula (Polynomial a)
                 -> QfFormula (TarskiPRE a)
qepcadGeoformEtf []  fo =
    callEtfQepcad qepModeGeoformEtf Nothing Nothing $ fmap PREPoly fo
qepcadGeoformEtf ctx fo =
    callEtfQepcad qepModeGeoformEtf Nothing
       (Just $ fmap PREPoly $ conjunctionQf ctx) $
    fmap PREPoly fo

--
-- mbContext may not be (Just (QfAt AtTrue)) or
-- (Just (QfAt AtFalse)) !
--
callQepcad :: Real a
           => Maybe (QfFormula (Polynomial a))
           -> FoFormula (Polynomial a)
	   -> QfFormula (Polynomial a)
callQepcad mbContext foQuestion =
    callEtfQepcad qepModeTarski Nothing
        (fmap (fmap PREPoly) mbContext) (fmap PREPoly foQuestion)

qepModeTarski     :: Real a => QepMode (Polynomial a)
qepModeTarski     = QepUsual      parseFormula
qepModeEtf        :: Real a => QepMode (TarskiPRE a)
qepModeEtf        = QepEtf        parseEtfFormula
qepModeGeoformEtf :: Real a => QepMode (TarskiPRE a)
qepModeGeoformEtf = QepGeoformETF parseEtfFormula

data QepMode a = QepUsual      { parseFn :: NIndeterminates -> String -> Either [Char] (QfFormula a) }
               | QepEtf        { parseFn :: NIndeterminates -> String -> Either [Char] (QfFormula a) }
               | QepGeoformETF { parseFn :: NIndeterminates -> String -> Either [Char] (QfFormula a) }

qepMode2commands :: QepMode a -> String
qepMode2commands (QepUsual _)      =
    "solution-extension T\n"
qepMode2commands (QepEtf _)      =
    "solution-extension I\nE\nY\nS\nP\n.\n"
qepMode2commands (QepGeoformETF _) =
    "solution-extension I\nE\nY\nG\nP\n.\n"
     
callEtfQepcad :: Real a
	      => QepMode term
              -> Maybe [Int]  -- ^ variable ordering
              -> Maybe (QfFormula (TarskiPRE a))
              -> FoFormula (TarskiPRE a)
	      -> QfFormula term
callEtfQepcad mode mbVarOrder mbContext foQuestion
    -- QEPCAD is unreliable with corner cases:
    --   - it does not like formulas with 0 variables,
    --   - quantifier-free tautologies/contradicitions sometimes
    --     generate an empty output (no "TRUE" or "FALSE").
    | unNIndeterminates nUnk == 0  =  truthValueAtZero
    | otherwise  =
        -- trace ("Qepcad input:\n" ++ input) $
	case communicateWithQepcad (parseFn mode) nUnk input of
	  Nothing -> truthValueAtZero
          Just f  -> f
    where
    input = "[ HsLooPo calling QEPCAD ]\n" ++
	    "(" ++ concat (intersperse "," $ map (varnames!!) varOrder) ++
            ")\n" ++
	    show (length free) ++ "\n" ++
	    quantText ++ "[" ++ qepcadFormula ++ "].\n" ++
            maybe "" assumptionF mbContext ++
	    proj_op ++ "go\ngo\ngo\n" ++
	    qepMode2commands mode ++
            (if Config.reuseQepcadSession ()
                then "continue\n"
                else "quit\n"
            )
    nonfree = nub [fromIntegral i | (i,_) <- quants]
    free    = [0..unNIndeterminates nUnk-1] \\ nonfree

    varOrder = case mbVarOrder of
                 Just order -> order
                 Nothing    -> free ++ nonfree

    assumptionF ctx = case eliminateAtTF ctx of
                        QfAt AtTrue  -> ""
                        QfAt AtFalse -> ""
                        c            -> "assume [" ++ aF ++ "]\n"
                            where
                            (aF,_) =  qfFormulaPRE2qepcad c

    --
    -- The McCallum projection operator is only guaranteed to be correct
    -- in 2 and 3 dimensions, so make sure that Hong's projection operator
    -- is used for additional dimensions.
    -- Note: The proj-op Qepcad command has n-1 arguments for n dimensions,
    -- since there are only n-1 projections; therefore, there are 2 "m"s at
    -- at the beginning for McCallum projections for 3->2 and 2->1.
    --
    proj_op = case unNIndeterminates nUnk of
	        1 -> ""
		2 -> ""
		n -> "proj-op (m,m" ++ concat (replicate (n-3) ",h") ++ ")\n"

    quantText = concat $
		['(' : qu q : ' ' : (varnames!!fromIntegral i) ++ ")"
		| (i,q) <- quants]
	where
	qu True  = 'A'
	qu False = 'E'
    (quants, question)    = matrixFo foQuestion
    (qepcadFormula, nUnk) = qfFormulaPRE2qepcad question
    truthValueAtZero =
       case eliminateAtSignum ev question of
           QfAt AtTrue  -> QfAt AtTrue
           QfAt AtFalse -> QfAt AtFalse
           qf           -> error ("truthValueAtZero = " ++ show qf)
       where
       zeros = replicate (unNIndeterminates $ nIndeterminatesN $ map prePoly $
                          termsQf question) 0
       ev (PREPoly p) = orderingToSignum (evalPoly p zeros `compare` 0)
       ev (PRERoot mp p n) = allSignumsFalse


communicateWithQepcad :: (NIndeterminates
                          -> String
                          -> Either String (QfFormula term)
                         )
                      -> NIndeterminates
	              -> String
	              -> Maybe (QfFormula term)
communicateWithQepcad parseFn nV input =
    case Config.reuseQepcadSession () of
      False -> unsafePerformIO $ runQepcadIO parseFn nV input
      True  -> talkToQepcad parseFn nV input

talkToQepcad :: (NIndeterminates
                 -> String
                 -> Either String (QfFormula term)
                )
             -> NIndeterminates
             -> String
	     -> Maybe (QfFormula term)
talkToQepcad parseFn nV input =
--    trace ("Qepcad input: " ++ input) $
    talkToTool qepcadConn input (readFn mkTioe parseFn nV)
    where
    mkTioe e = ToolIOError { tioeTool = "Qepcad", tioeInput = input, tioeError = e }


readFn :: (TIOError -> err)
       -> (NIndeterminates -> String -> Either String (QfFormula term))
       -> NIndeterminates
       -> String
       -> Either err (Maybe (QfFormula term),String)
readFn mkTioe parseFn nV str =
    case find (preText `isPrefixOf`) (tails str) of
	  Nothing   ->
	      case find (endText `isPrefixOf`) (tails str) of
	        Nothing -> Left $ mkTioe $
			   UnparsableResult ("did not find \"" ++ preText
					     ++"\" nor \"" ++
					     endText)
		Just rest -> Right (Nothing, drop (length endText) rest)
	  Just str2 ->
	      let str3 = drop (length preText) str2
                  unparsable = Left $ mkTioe$ 
			       UnparsableResult ("cannot interpret:\n" ++ str3)
              in
	      case findIndex (endText `isPrefixOf`) (tails str3) of
	        Just endIdx ->
		  let (formStr,postStr) = splitAt endIdx str3 in
                  case parseFn nV formStr of
		      Right x -> Right (Just x, postStr)
		      Left  x -> Left  (mkTioe $ UnparsableResult x)
		Nothing   -> unparsable
    where
    -- We always use "solution-extension X" followed
    -- by "continue" or "quit", so there is no "=  The End  ="
    --
    preText = "equivalent quantifier-free formula:\n\n"
    endText = "\n\n"
--    postText = "=  The End  ="



runQepcadIO :: (NIndeterminates
                -> String
                -> Either String (QfFormula term)
               )
            -> NIndeterminates
            -> String
	    -> IO (Maybe (QfFormula term))
runQepcadIO parseFn nV input =
    do -- putStrLn input
       qepcadBin   <- getQepcadCommand
       qepcadCells <- getQepcadCells
       hFlush stdout
       res <- runToolIO True qepcadBin ["+N" ++ show (qepcadCells*10^6)] input myReadFn
--       putStrLn ("qepcad: " ++ show res)
       return res
    where
    myReadFn = either (Left . id) (Right . fst) . readFn id parseFn nV


slfqSimplifyQf :: Real a
	       => [QfFormula (Polynomial a)]
	       -> QfFormula (Polynomial a)
	       -> QfFormula (Polynomial a)
slfqSimplifyQf ctx formula =
    case formula' of
       QfAt AtTrue  -> QfAt AtTrue
       QfAt AtFalse -> QfAt AtFalse
       _ -> -- (\res -> trace ("slfqSimplifyQf " ++ show (ctx,formula',res)) res) $
            unsafePerformIO $
	    (runSlfqIO (conjunctionQf ctx) formula'
{-	     `catchDyn` \(ex :: ToolIOError) ->
	                  hPutStrLn stderr (show ex) >> throwDyn ex
-}
	    )
    where
    formula' = eliminateAtTF $ eliminateAtSignum constPolySignum formula



--
-- Older SLFQ verions requires that every variables occurs in the
-- formula and in the assumptions!
--
runSlfqIO :: Real a
	  => QfFormula (Polynomial a)
	  -> QfFormula (Polynomial a)
	  -> IO (QfFormula (Polynomial a))
runSlfqIO assumptions formula =
    do slfqBin <- getSlfqCommand
       slfqCells <- getSlfqCells
--       hPutStrLn stderr ("input: " ++ input ++ "\n" ++ "assumptions: " ++ assumptionA)
--       hPutStrLn stderr (show $ length assumptionA)
       -- Fix the variable ordering; if we have to call Qepcad
       -- afterwards to get rid of _root_N expressions, we must use
       -- the same variable ordering (the formula may not be proper
       -- otherwise!)
       res <- runToolIO False slfqBin
		 ("-q" :
		  (if assumptionA=="" then [] else ["-a", assumptionA]) ++
--                  ["-L", concat $ intersperse "," [varnames!!i |
--                                        i <- [0..unNIndeterminates nV-1]] ] ++
		  ["-N", show slfqCells])
		 input
		 readFn
--     hPutStrLn stderr ("res: " ++ show res)
       return res
    where
    varsInF = foldl union [] $ map usedIndeterminates $ termsQf formula
    varsInA = foldl union [] $ map usedIndeterminates $ termsQf assumptions
    nV = nIndeterminatesN $ termsQf (QfAnd assumptions formula)
    (input,_)  = qfFormula2qepcad
		    (formula `QfAnd`
		     conjunctionQf [QfAt $ AtGe (indeterminateNI nV i ^ 2)
				   | i <- varsInA\\varsInF]
		    )
    (assumptionA,_) = qfFormula2qepcad assumptions

    readFn :: Real a
	   => String -> Either TIOError (QfFormula (Polynomial a))
    readFn res
	| any ("Too few cells reclaimed" `isPrefixOf`) (tails res) =
	    Left HeapSpaceLow
	| otherwise =
	    case parseFormula nV dotFreeRes of
		   Right f  -> Right f
		   Left err -> Left $ ToolFailed err
	where
        dotFreeRes = nodot res
	nodot str
	    | (_,'.':new) <- break (=='.') str = nodot new
	    | otherwise                        = str


parseEtfFormula :: Real a
                   => NIndeterminates
                   -> String
                   -> Either String (QfFormula (TarskiPRE a))
parseEtfFormula nInd str =
    case rp of
      Left  l  -> Left (show l)
      Right qf -> Right $ fmap (mapPRE fromIntegral) qf
    where
    rp =
	runParser (spaces >> parseQepcadQfFormula nInd varnames >>= \res ->
                   eof >> return res
                  )
             ()
	     ("Qepcad/SLFQ Output (which is \"" ++ str ++ "\")")
	     str

    varnames = ['x':show i | i <- [0..unNIndeterminates nInd-1]]


--
-- Parse a formula in Qepcad notation.
-- If the forumla is an extended tarski formula,
-- call Qepcad to get an equivalent non-extended formula.
--
parseFormula :: Real a
		=> NIndeterminates
	        -> String
		-> Either String (QfFormula (Polynomial a))
parseFormula nV str =
    case parseEtfFormula nV str of
      Left  l  -> Left l
      Right qf
        | Just f <- isNonETF qf  -> Right f
	| Config.externalToolAvailable Config.QEPCAD ->
              Right qf2
	| otherwise  ->
	      Left "SLFQ returned _root_<N>, but Qepcad is not available."
	  where
          qf2 = callEtfQepcad qepModeTarski (Just allvars) Nothing (FoQf qf)

          allvars = topSort g `union` [0..unNIndeterminates nV-1]

          g = buildG (0,unNIndeterminates nV-1) $ properVarOrd qf


isNonETF :: QfFormula (TarskiPRE a) -> Maybe (QfFormula (Polynomial a))
isNonETF qf
    | all termNotETF $ termsQf qf  =  Just $ fmap unPRE qf
    | otherwise                    =  Nothing
    where
    termNotETF (PREPoly _) = True
    termNotETF _           = False

    unPRE (PREPoly p) = p
    unPRE (PRERoot _ _ _) = error "isNonETF: internal error"


properVarOrd :: QfFormula (TarskiPRE a) -> [(Int,Int)]
properVarOrd qf = concatMap varOrd $ termsQf qf
    where
    varOrd (PREPoly _      ) = []
    varOrd (PRERoot _ _ rhs) = [(r,k) | r <- usedIndeterminates rhs, r/=k]
        where
        k = maximum $ usedIndeterminates rhs

--
-- Parse a formula and Qepcad notation and
-- return the parsed formula if it is _not_ a restricted
-- extended tarski formula.
-- If the formula is an extended tarski formula, return a variable
-- ordering under which the formula is proper.
-- 
parseQepcadQfFormula :: (Integral a, Read a)
                     => NIndeterminates
                     -> [String]
                     -> Parser (QfFormula (TarskiPRE a))
parseQepcadQfFormula nInd varnames =
    pQf
    where
    pQf =((atFormula nInd varnames' >>= \a -> return (QfAt a)) <|>
          (char '[' >> spaces >> pQf >>= \i ->
           char ']' >> spaces >> return i
          )
         ) >>= \f ->
         (P.try (junctor   >>= \j ->
	         pQf >>= \g ->
	         return (j f g)
	        ) <|>
          return f
         )

    junctor = ((string "\\/" >> spaces >> return QfOr) <|>
	       (string "/\\" >> spaces >> return QfAnd)
              )
    -- an earlier variable name must not be a prefix of a later
    -- variable name; therefore, we sort (descending) by the
    -- length of the variable name
    varnames' = sortBy (\(v1,_) (v2,_) -> compare (length v2) (length v1)) $
                zip varnames [0..]

atFormula :: (Integral a, Read a)
	  => NIndeterminates
          -> [(String, Int)]
          -> Parser (AtFormula (TarskiPRE a))
atFormula nInd varnames =
    (
     (string "TRUE"  >> spaces >> return AtTrue) <|>
     (string "FALSE" >> spaces >> return AtFalse) <|>
     (polynomial nInd varnames >>= \left ->
      relop      >>= \r ->
      rhs        >>= \right ->
      case right of
	        Right right -> return $ r $ PREPoly (left-right)
	        Left  (n,right)
                    | left == indeterminateNI (AtLeast (k+1)) k ->
                        return $ r $ PRERoot True n right
                    | otherwise -> fail "illegal ETF atomic formula"
                    where
                    k = maximum $ usedIndeterminates right
     )
    )
    where
    relop =
        (
            (char '>' >>
	     ((char '=' >> return AtGe) <|> return AtGt)
	    ) <|>
            (char '<' >>
	     ((char '=' >> return (\f -> AtGe $ negate f)) <|>
	      return (\f -> AtGt $ negate f))
	    ) <|>
	    (string "/=" >> return AtNeq) <|>
	    (char   '='  >> return AtEq )
        ) >>= \op -> spaces >> return op

    rhs =
        ( 
           (do string "_root_"
	       n <- integer
	       p <- polynomial nInd varnames
	       return $ Left (n, p)
	   ) <|>
           (polynomial nInd varnames >>= \p -> return $ Right p)
        )
					       

polynomial :: (Integral a, Read a)
           => NIndeterminates
           -> [(String, Int)]
           -> Parser (Polynomial a)
polynomial nInd varnames =
    do c <- coeff
       t <- term  nInd varnames
       next <- ((lookAhead (oneOf "+-") >> polynomial nInd varnames) <|>
                return 0)
       return ((c `pmul` t) + next)


integer :: (Integral a, Read a) => CharParser z a
integer = ((char '-' >> return (-1)) <|> (optional (char '+') >> return 1)) >>= \s ->
          many1 digit >>= \ds ->
          spaces >>
          return (s * read ds)

coeff :: (Integral a, Read a) => CharParser z a
coeff =
    do s <- (char '-' >> return (-1)) <|> (optional (char '+') >> return 1)
       spaces
       ds <- many digit
       spaces
       return (s * fromIntegral (number ds))
    where
    number :: String -> Integer
    number [] = 1
    number ds = read ds

term :: Integral a
     => NIndeterminates
     -> [(String, Int)]
     -> Parser (Polynomial a)
term nInd varnames =
    (
     (var  >>= \v ->
      expo >>= \e ->
      term nInd varnames >>= \t ->
      return (indeterminateNI nInd v ^ e * t)
     ) <|> return 1
    )
    where
    var = do n <- choice (map (try . string . fst) varnames)
             let Just i = lookup n varnames
             spaces
             return i

    expo = 
        (do char '^'
            spaces
	    ds <- many1 digit
            spaces
	    return (read ds)
	) <|> return 1

-----------


data QAlg = QAlg (Polynomial Integer) Integer [(Rational,Rational)]
          | QRat Rational
            deriving Show

qalg :: (Real a, Integral b) => Polynomial a -> b -> QAlg
qalg p' n' =
    assert (nIndeterminates p' == Exactly 1) $
    assert (abs n >= 1 && abs n <= nR) $
    if degree p == 1 then
        QRat (negate (constantCoeff p) % leadingCoeff p)
    else
        QAlg p n rs
    where
      rs = map rootInterval (allroots `genericIndex` m)
      allroots = allRoots ratP
      nR = genericLength allroots
      m  = if n>0 then n - 1 else nR + n 

      (_,p) = ratPoly2intPoly ratP
      ratP  = fmap' toRational p'
      n = fromIntegral n'

-- |
-- Indeterminate @dim@ will equal the algebraic number.
--
qalgToFormula :: Real a
              => Integer
              -> Integer
              -> QAlg
              -> QfFormula (TarskiPRE a)
qalgToFormula nI dim (QRat r)   =
    qalgToFormula nI dim $
    QAlg (fromIntegral (denominator r) * indeterminate 1 0 -
          fromIntegral (numerator r)
         )
         1
         undefined
qalgToFormula nI dim (QAlg p n _) =
    assert (nIndeterminates p == Exactly 1) $
    QfAt $ AtEq $ PRERoot True n (liftPolyN dim $ appendIndeterminates e $
                                  fmap' fromIntegral p)
    where
    e = nI - dim - 1

qalgOrdering :: QAlg -> Ordering
qalgOrdering (QRat r) = compare r 0
qalgOrdering qalg
    | qepcadEtfImpliesQf' [algF] $ question AtGt  =  GT
    | qepcadEtfImpliesQf' [algF] $ question AtEq  =  EQ
    | otherwise                                   =  LT
    where
    algF = qalgToFormula 1 0 qalg
    question atrel = QfAt $ atrel $ PREPoly $ indeterminate 1 0

qalgSubstP :: Real a => [QAlg] -> Polynomial a -> [TarskiPRE a]
qalgSubstP [] p  =  [PREPoly p]
qalgSubstP algs p
    | Just c <- isConstPoly p  =
        [PREPoly $ constPolyNI (Exactly 1) c]
    | Just rats <- mbRats  =
        let (_,newP) = ratPoly2intPoly $
                       fmap' (\q -> evalPoly q rats) $
                       univariate $
                       fmap' toRational p
        in [PREPoly $ fmap' fromIntegral newP]
    | otherwise  =
        case res of
          QfAt AtFalse    -> []
          QfAt AtTrue     -> [PREPoly $ constPolyNI (Exactly 1) 0]
          qf  ->
              assert (nIndeterminatesN (map prePoly $ termsQf qf) == Exactly 1) $
              let at2pre (QfAt (AtEq pre)) = pre
                  at2pre _ =
                      error ("qalgSubstP: unexpected result " ++ show qf ++
                             "\nfor input: " ++ show (algs,p))
              in map at2pre $ disjunctsQf qf
    where
    nInds = genericLength algs + 1

    res = fmap (mapPREpoly (deleteEndIndeterminates (nInds-1))) $ res1
    res1= qepcadEtfEliminateQuantifiersEtf' [] fo

    fo = FoEx [1..nInds-1] $ FoQf q
    q = (algsQf `qfAnd` pQf)

    pQf = QfAt $ AtEq $ PREPoly p
    algsQf = conjunctionQf
             [qalgToFormula nInds dim alg | (dim,alg) <- zip [1..] algs]

    mbRats :: Maybe [Rational]
    mbRats = sequence $ map algIsRat algs


instance Alg QAlg where
    algIsRat (QRat r) = Just r
    algIsRat _        = Nothing

    toAlg = QRat . toRational
    zeroToAlg p n = qalg p n

    algOrdering = qalgOrdering
    algSubstP = qalgSubstP

    fromAlg (QRat r) = repeat (r',r')
        where
          r' = fromRational r
    fromAlg (QAlg p n rs) = rs

    algDefiningPoly (QRat r)     = indeterminate 1 0 - constPoly r
    algDefiningPoly (QAlg p n _) = fmap' fromIntegral p

qalgOps :: AlgOps QAlg
qalgOps = genericAlgOps

instance Eq QAlg where
    (==) = algOpEq qalgOps

instance Ord QAlg where
    compare = algOpCompare qalgOps

instance Num QAlg where
    (+) = algOpAdd qalgOps
    (-) = algOpSub qalgOps
    (*) = algOpMul qalgOps
    negate = algOpNegate qalgOps
    abs = algOpAbs qalgOps
    signum = algOpSignum qalgOps
    fromInteger = algOpFromInteger qalgOps

instance Fractional QAlg where
    recip = algOpRecip qalgOps
    (/)   = algOpDiv   qalgOps
    fromRational = algOpFromRational qalgOps

qqtrace a b = {- trace a -} b
qepcadEtfImpliesQf' x = qqtrace ("qepcadEtfImpliesQf") qepcadEtfImpliesQf x
qepcadEtfEliminateQuantifiersEtf' x = qqtrace ("qepcadEtfEliminateQuantifiersEtf") qepcadEtfEliminateQuantifiersEtf x
