-- |
-- Interface to the Reduce computer algebra system.
--
module LMath.Tools.Reduce
    ( reduceSimplifyQf, reduceFastSimplifyQf, reduceSortIneqs,
      reduceImpliesQf, reduceImpliesSignum,
      reduceEliminateQuantifiers,
      reduceFactorise,
      reduceFactoriseP, reduceGcdPoly, reduceResultant,
      reduceConvexUnion,
      reduceDisjointUnion, reduceDisjointUnionDifference,
      reduceDisjointUnionQf,
      reduceDnfQf,
      reduceLexmin, reduceLexmax
    )
where

import Control.Exception ( assert )
import Control.Monad
import Data.Char
import Data.IORef
import Data.List as List
import Data.Maybe
import Debug.Trace

import Data.Ratio
import System.IO.Unsafe

import LMath.Tools.ToolIO
import LMath.Types.AffineExp
import LMath.Types.DTree
import LMath.Types.EqSystem
import LMath.Types.Formula
import LMath.Types.IneqSystem
import LMath.Types.Matrix
import LMath.Types.LClasses
import LMath.Types.Polynomial
import LMath.Types.Signum
import LMath.Types.SortedSystem
import LMath.Types.Vector
import LMath.Tools.ReduceFunctions (reduceFunctions)
import qualified Config
import LP

import System.Posix
import Text.ParserCombinators.Parsec



mytrace a b = b


--
-- Expressions with +,-,*, and ^ (power) with non-negative exponent,
-- and integral coefficients.
-- Variables are  Var (n,i)  which weans "variable i of n" where
--  i is from {0,1,...,n-1}.
-- NOTE: Var (Exactly 0,i) denotes the "linear" variable i,
-- in contrast to the (possibly) non-linear variables (i.e., parameters)
-- denoted by  Var (ni,i)  where ni /= Exactly 0
--
data Exp = Add Exp Exp
	 | Neg Exp
	 | Mul Exp Exp
	 | Pwr Exp Integer
	 | Var   (NIndeterminates, Integer)
	 | Const (NIndeterminates, Integer)
	   deriving (Eq,Show,Read)

parseExp :: Parser Exp
parseExp =
      (do char '('
          spaces
          e <- parseExp
          char ')'
          spaces
          return e
      ) <|>
      (do string "Add"
          spaces
          e1 <- parseExp
          e2 <- parseExp
          return $ Add e1 e2
      ) <|>
      (do string "Neg"
          spaces
          e <- parseExp
          return $ Neg e
      ) <|>
      (do string "Mul"
          spaces
          e1 <- parseExp
          e2 <- parseExp
          return $ Mul e1 e2
      ) <|>
      (do string "Pwr"
          spaces
          e <- parseExp
          i <- parseInteger
          return $ Pwr e i
      ) <|>
      (do string "Var"
          spaces
          char '('
          spaces
          ni <- nind
          char ','
          spaces
          i <- parseInteger
          char ')'
          spaces
          return $ Var (ni, i)
      ) <|>
      (do string "Const"
          spaces
          char '('
          spaces
          ni <- nind
          char ','
          spaces
          i <- parseInteger
          char ')'
          spaces
          return $ Const (ni, i)
      )

nind :: Parser NIndeterminates
nind = (do string "Exactly"
           spaces
           i <- parseInteger
           return $ Exactly i
       ) <|>
       (do string "AtLeast"
           spaces
           i <- parseInteger
           return $ AtLeast i
       )

class ExpConv e where
    toExp   :: e -> Exp
    fromExp :: Exp -> e

isConstExp :: Exp -> Maybe Integer
isConstExp (Add x y) =
    do xx <- isConstExp x
       yy <- isConstExp y
       return (xx+yy)
isConstExp (Mul x y) =
    do xx <- isConstExp x
       yy <- isConstExp y
       return (xx*yy)
isConstExp (Neg x) = fmap negate $ isConstExp x
isConstExp (Pwr x e) =
    do xx <- isConstExp x
       return (xx ^ e)
isConstExp (Var _) = Nothing
isConstExp (Const (_,val)) = Just val


getUsedIndets :: Exp -> [Integer]
getUsedIndets (Add x y) = List.union (getUsedIndets x) (getUsedIndets y)
getUsedIndets (Mul x y) = List.union (getUsedIndets x) (getUsedIndets y)
getUsedIndets (Neg x  ) = getUsedIndets x
getUsedIndets (Pwr x _) = getUsedIndets x
getUsedIndets (Var   (_,i)) = [i]
getUsedIndets (Const _    ) = []

getNIndets :: Exp -> NIndeterminates
getNIndets (Add x y) = max (getNIndets x) (getNIndets y)
getNIndets (Neg x)   = getNIndets x
getNIndets (Mul x y) = max (getNIndets x) (getNIndets y)
getNIndets (Pwr x _) = getNIndets x
getNIndets (Var   (ni,_)) = ni
getNIndets (Const (ni,_)) = ni

setNIndets :: NIndeterminates -> Exp -> Exp
setNIndets ni (Add x y) = Add (setNIndets ni x) (setNIndets ni y)
setNIndets ni (Neg x)   = Neg $ setNIndets ni x
setNIndets ni (Mul x y) = Mul (setNIndets ni x) (setNIndets ni y)
setNIndets ni (Pwr x p) = Pwr (setNIndets ni x) p
setNIndets ni v@(Var (Exactly 0,_)) = v
setNIndets ni (Var (ni',n))
    | canExtendToNI ni' ni = Var (ni,n)
    | otherwise = error ("setNIndets: cannot extend indeterminates from " ++
			 show ni' ++ " to " ++ show ni)
setNIndets ni (Const (ni',n))
    | ni >= ni' = Const (ni,n)
    | otherwise = error ("setNIndets: cannot extend indeterminates from " ++
			 show ni' ++ " to " ++ show ni)


instance ExpConv Exp where
    toExp   = id
    fromExp = id


showNI :: NIndeterminates -> String
showNI (Exactly n) = show n
showNI (AtLeast n) = show $ negate (n+1)

polynomial2exp :: Integral a => Polynomial a -> Exp
polynomial2exp poly = intpoly2exp poly 0
    where
    nInd = nIndeterminates poly
    intpoly2exp :: Integral a => Polynomial a -> Integer -> Exp
    intpoly2exp p varNr =
	case isConstPoly p of
	  Just c  -> Const (nInd, fromIntegral c)
	  Nothing -> foldl1 Add sumands
	      where
	      ces = coeffsAndExps (univariate p)
	      sumands = map sumand ces
              sumand (c,e) =
		  case e of
		    0 -> co
		    1 -> Mul co var
		    _ -> Mul co (Pwr var (fromIntegral e))
		  where 
		  co  = intpoly2exp c (varNr+1)
		  var = Var (nInd, varNr)

instance Real a => ExpConv (Polynomial a) where
    toExp p = let (_,p') = ratPoly2intPoly (mapPoly toRational p) in polynomial2exp p'
    fromExp (Add x y) = fromExp x + fromExp y
    fromExp (Neg x)   = negate $ fromExp x
    fromExp (Mul x y) = fromExp x * fromExp y
    fromExp (Pwr x e) = potPoly (fromExp x) e
    fromExp (Var (n,i)) = indeterminateNI n i
    fromExp (Const (n,x)) = constPolyNI n $ fromIntegral x   

readExp :: ExpConv e => String -> Maybe e
readExp str =
    case parse p "" str of
      Right e -> Just $ fromExp e
      Left  _ -> Nothing
    where
    p = do spaces
           res <- parseExp
           spaces
           eof
           return res


readQfFormula :: ExpConv e => String -> Maybe (QfFormula e)
readQfFormula str =
    case (reads trimstr :: [(QfFormula Exp,String)]) of
      (res, ""):_ -> Just $ fmap fromExp res
      _           -> Nothing
    where
      trimstr = trim str

trim :: String -> String
trim = reverse . dropWhile isSpace . reverse . dropWhile isSpace

showsVar :: (NIndeterminates, Integer) -> ShowS
showsVar (a,b) = showsV (unNIndeterminates a, b)
    where
    showsV (0,i) = showString "x" . shows i
    showsV (n,i) = showString "u" . shows n . showString "_" . shows i

showVar :: (NIndeterminates, Integer) -> String
showVar arg = showsVar arg ""

exp2red :: Exp -> String -> String
exp2red (Add e1 e2) = showString "(" . exp2red e1 . showString "+" .
		      exp2red e2 . showString ")"
exp2red (Neg e)     = showString "( -" . exp2red e . showString ")"
exp2red (Mul e1 e2) = showString "(" . exp2red e1 . showString "*" .
		      exp2red e2 . showString ")"
exp2red (Pwr e1 p)  = showString "(" . exp2red e1 . showString "^" .
		      shows p . showString ")"
exp2red (Var (n,i)) = showsVar (n,i)
exp2red (Const (_,i)) = shows i

expc2red :: ExpConv e => e -> String -> String
expc2red = exp2red . toExp

rel2red :: ExpConv e => AtFormula e -> String -> String
rel2red (AtEq e) = expc2red e . showString "=0"
rel2red (AtNeq e) = expc2red e . showString "<>0"
rel2red (AtGt e) = expc2red e . showString ">0"
rel2red (AtGe e) = expc2red e . showString ">=0"
rel2red AtTrue   = showString "true"
rel2red AtFalse  = showString "false"

-- convert a formula to Redlog's format.
-- Make sure the number of indeterminates is the same for
-- every term (i.e. polynomial) of the formula!!
-- Return the number of indeterminates in the formula and a
-- string representation of the formula in Redlog's format.
--
qfFormula2red' :: ExpConv e => QfFormula e -> (NIndeterminates, String)
qfFormula2red' f = (nInd, qf2red (fmap (setNIndets nInd) f') "")
    where
    f' = fmap toExp f
    nInd = maximum (AtLeast 0 : fmap getNIndets (termsQf f'))

    qf2red :: ExpConv e => QfFormula e -> String -> String
    qf2red (QfAt a) = rel2red a
    qf2red (QfNot f)     = showString "not (" . qf2red f . showString ")"
    qf2red (QfAnd f1 f2) = showString "(" . qf2red f1 . showString " and " .
			   qf2red f2 . showString ")"
    qf2red (QfOr f1 f2) =  showString "(" . qf2red f1 . showString " or " .
			   qf2red f2 . showString ")"

qfFormula2red :: ExpConv e => QfFormula e -> String
qfFormula2red = snd . qfFormula2red'

foFormula2red :: ExpConv a => FoFormula a -> (NIndeterminates, String)
foFormula2red fo = (nInd, tored fo)
    where
    (nInd, str) = qfFormula2red' $ snd $ matrixFo fo

    tored (FoQf _) = str
    tored (FoAll vs f) =
	"all({" ++ concat (intersperse "," (map showV vs))
        ++ "}," ++ tored f ++ ")"
    tored (FoEx vs f) =
	"ex({" ++ concat (intersperse "," (map showV vs)) ++
        "}," ++ tored f ++ ")"

    showV i = showVar (nInd, i)


reduceGcdPoly :: RealFrac a
	      => Polynomial a
	      -> Polynomial a
	      -> Polynomial a
reduceGcdPoly p q =
    monic $
    runReduce ("m=>printFormula(m," ++ showNI nInd ++ ")")
	      ((showString "gcd(" . exp2red p1 . showString "," .
		exp2red q1 . showString ")") "")
	      (fmap fromExp . readExp)
    where
    p1 = setNIndets nInd $ toExp p
    q1 = setNIndets nInd $ toExp q
    nInd = nIndeterminatesN [p,q]

{-
reducePP :: Integral a
	 => Polynomial (Ratio a)
	 -> Polynomial (Ratio a)
	 -> FracPoly (Ratio a)
reducePP p q =
    where
    
    runReduce ("m=>printFormula(m," ++ showNI nInd ++ ")")
	      ((showString "gcd(" . exp2red p1 . showString "," .
		exp2red q1 . showString ")") "")
	      (fmap fromExp . readExp)
    where
    p1 = setNIndets nInd $ toExp p
    q1 = setNIndets nInd $ toExp q
    nInd = nIndeterminatesN [p,q]    
-}

rlqe :: (Show a, ExpConv a, ExpConv b)
     => FoFormula a
     -> [AtFormula b]
     -> QfFormula a
rlqe fo theo =
    case eliminateAtSignumQ (getUsedIndets . toExp) sgn fo of
      FoQf qf ->
          -- trace ("eliminateAtSignumQ: " ++ show fo ++ " -> " ++ show qf) $
          qf
      fo'     ->
          -- qf' may have a different quantification structure,
          -- but that does not matter, because we eliminate every
          -- quantifier.
          runReduce ("m=>printFormula(m," ++ showNI nInd ++ ")")
		    ("rlqe(" ++ formulaStr ++ "," ++
		     theo2red theo ++ ")")
		    readQfFormula
          where
          (nInd, formulaStr) = foFormula2red fo'
    where
    sgn a = case isConstExp $ toExp a of
              Just i  -> orderingToSignum (i `compare` 0)
              Nothing -> allSignumsFalse

rlqe_bool :: Real a =>
	     FoFormula (Polynomial a) ->
	     [AtFormula (Polynomial a)] -> Bool
rlqe_bool formula theo =
    case rlqe formula theo of
      QfAt AtTrue  -> mytrace "rlqe_bool:true" $ True
      QfAt AtFalse -> mytrace "rlqe_bool:false" $ False
      res          -> error ("rlqe_bool: result is not `true' or `false':\n"
			     ++ show res)



reduceEliminateQuantifiers :: Real a
			   => [QfFormula (Polynomial a)]
			   -> FoFormula (Polynomial a)
			   -> QfFormula (Polynomial a)
reduceEliminateQuantifiers ctx formula = rlqe formula (contextToTheo ctx)

-------------

data Sorted = SolOk
	    | Sorted [(QfFormula Exp, [[(Exp,Exp)]], [[(Exp,Exp)]], Sorted)]
	      deriving (Read, Show)
--
-- linear parameters of the affine expression are handled
-- like additional variables
--
affineExpToExp :: RealFrac a =>
		  (Maybe (QfFormula (Polynomial a) -> Bool)) ->
		  GAffineExp (FracPoly a) -> [Exp]
affineExpToExp decideSqFact exp = map (aeToE . getMatrix) (unjoin exp)
    where
    nVP = nVars exp + nParams exp
    aeToE m = foldl Add (polynomial2exp add)
	        [Mul (Var (Exactly 0,i)) (polynomial2exp c)
	        | (c,i) <- zip coeffs [nVP-1,nVP-2 .. 0]]
	where
	nums   = map polyNumerator   (entries m)
	denoms = map polyDenominator (entries m)
	fact'  = foldl lcmPoly 1 denoms
	fact   = case decideSqFact of
		   Nothing     -> fact'
		   Just decide -> if decide (QfAt (AtGt fact'))
				     then fact'
				     else (fact' * fact')
	(factsAndCoeffs) =
	    reverse $
	    map ratPoly2intPoly $
	    map (\f -> polyNumerator f * (fact `quotPoly` polyDenominator f)) (entries m)
	fsLcm = foldl1 lcm $ fst $ unzip factsAndCoeffs
	add:coeffs = map (\(f,co) -> (fsLcm `quot` f) `pmul` co) factsAndCoeffs
	

ineqsToRFormula ::RealFrac a => 
		  [QfFormula (Polynomial a)] ->
		  GIneqSystem (FracPoly a) -> 
		  Maybe (GEqSystem (FracPoly a)) -> QfFormula Exp
ineqsToRFormula ctx ineqs mbEqs =
    foldl QfAnd eqsF [QfAt (AtGe e)
		      | e <- affineExpToExp (Just decide) (toAffineExp ineqs)]
    where
    decide f = reduceImpliesQf ctx f
    eqsF = case mbEqs of
             Nothing  -> QfAt AtTrue
             Just eqs -> foldl QfAnd (QfAt AtTrue)
			       [QfAt (AtEq e)
			       | e <- affineExpToExp Nothing (toAffineExp eqs)]

readSorted :: String -> Maybe Sorted
readSorted str =
    case reads str of
	     []        -> Nothing
	     ((s,_):_) -> Just s

ineqsToSorted :: Integral a => 
		 [QfFormula (Polynomial (Ratio a))] ->
		 GIneqSystem (FracPoly (Ratio a)) -> 
		 Maybe (GEqSystem (FracPoly (Ratio a))) -> Sorted 
ineqsToSorted ctx ineqs mbEqs = runReduce printer input readSorted
    where
    printer = "m=>printSorted(m,0," ++ showNI parN ++ ")"
    input = "sortIneqs(" ++ qfFormula2red (ineqsToRFormula ctx ineqs mbEqs) ++
	    ",{" ++ concat (intersperse "," (map showVar [(Exactly 0,i) | i <- [0..nVars ineqs-1]])) ++ "}," ++ context2red ctx ++ ")"
    parN = nIndeterminatesN polys

    polys = map polyNumerator (entries ineqs) ++ map polyDenominator (entries ineqs)


atFormula2red :: ExpConv e => AtFormula e -> String
atFormula2red f = qfFormula2red (QfAt f)

sortedToDTree :: Integral a =>
		 Sorted -> DTree (Polynomial (Ratio a)) (Maybe (SortedSystem (FracPoly (Ratio a))))
sortedToDTree = std []
    where
    std acc SolOk = DLeaf $ Just $ map mkLU $ reverse acc
    std acc (Sorted []) = DLeaf Nothing
    std acc (Sorted ((cond,lbs,ubs,sub):next)) =
	case cond of
	  QfAt AtTrue -> subTree
	  _           -> FCond (fmap fromExp cond)
		               subTree
			       (std acc (Sorted next))
	where
	subTree = (std ((map (map expexpToFP) lbs,
			 map (map expexpToFP) ubs):acc)
		       sub)
    mkLU ([l],[u])
	| l == u    = Equation $ vector l
	| otherwise = LowerUpper [vector l] [vector u]
    mkLU (lbs,ubs) = LowerUpper (map vector lbs) (map vector ubs)

    expexpToFP :: Integral a => (Exp,Exp) -> FracPoly (Ratio a)
    expexpToFP (num,den) = fromExp num %% fromExp den

reduceSortIneqs :: Integral a =>
		   [QfFormula (Polynomial (Ratio a))] ->
		   GIneqSystem (FracPoly (Ratio a)) ->
		   Maybe (GEqSystem (FracPoly (Ratio a))) ->
		   Maybe Integer ->
		   DTree (Polynomial (Ratio a)) (Maybe (SortedSystem (FracPoly (Ratio a))))
reduceSortIneqs ctx ineqs mbEqs ctxVars =
    assert (nParams ineqs == 0) $
    assert (fmap ((==0).nParams) mbEqs /= Just False) $
    fmap (fmap (drop $ fromIntegral ctxVars')) $
    sortedToDTree $
    ineqsToSorted ctx ineqs mbEqs
    where
    ctxVars' = fromMaybe 0 ctxVars



-------------


theo2red :: ExpConv a => [AtFormula a] -> String
theo2red theo = "{" ++ concat (intersperse "," $ map (atFormula2red . fmap toExp) theo)
		++ "}"

context2red :: Real a => [QfFormula (Polynomial a)] -> String
context2red = theo2red . contextToTheo

contextToTheo :: Real a =>
		 [QfFormula (Polynomial a)] -> [AtFormula Exp]
contextToTheo ctx = map (fmap toExp) $ concatMap prepare ctx
    where
    prepare qf = case eliminateAtTF $ eliminateAtSignum constPolySignum qf of
		   QfAt AtTrue  -> []
		   QfAt AtFalse -> []
		   qf'          -> extract qf'
    extract (QfAt  f)   = [f]
    extract (QfAnd f g) = extract f ++ extract g
    extract _           = []


reduceFastSimplifyQf :: Real a => [QfFormula (Polynomial a)] -> QfFormula (Polynomial a) -> QfFormula (Polynomial a)
reduceFastSimplifyQf ctx formula = runReduce printer input readQfFormula
    where
    printer = "m=>printFormula(m," ++ showNI (indeterminatesInQf formula) ++ ")"
    input = "rlsimpl(" ++ qfFormula2red  formula ++ "," ++
	    context2red ctx ++ ")"

reduceSimplifyQf :: Real a => [QfFormula (Polynomial a)] -> QfFormula (Polynomial a) -> QfFormula (Polynomial a)
reduceSimplifyQf ctx formula = runReduce printer input readQfFormula
    where
    printer = "m=>printFormula(m," ++ showNI (indeterminatesInQf formula) ++ ")"
    input = "rlsimpl(" ++ qfFormula2red  formula ++ "," ++
	    context2red ctx ++ ")"

------------

reduceImpliesQf :: Real a =>
                   [QfFormula (Polynomial a)] ->
                   QfFormula (Polynomial a)   ->
                   Bool
reduceImpliesQf ctx qf = rlqe_bool formula []
    where
    formula   = FoAll [0..unNIndeterminates (indeterminatesInQf qfFormula)-1]
		      (FoQf qfFormula)
    qfFormula = conjunctionQf ctx `qfImpl` qf

reduceImpliesSignum :: Real a
		    => [QfFormula (Polynomial a)]
		    -> Polynomial a
		    -> Signum
reduceImpliesSignum = genericImpliesSignum reduceImpliesQf

------------

--
-- Only handles *finite* lexicographic minima !!!
--
reduceLexmin :: Integral a
	     => [QfFormula (Polynomial (Ratio a))]
	     -> GIneqSystem (FracPoly (Ratio a))
	     -> DTree (Polynomial (Ratio a)) (Optimum [FracPoly (Ratio a)])
reduceLexmin ctx ineqs = runReduce printer input (liftM convLM . mbRead)
    where
    printer = "m=>printList(m,n=>print1Lexminmax(n," ++ showNI parN ++ "))"
    input = "lexmin(" ++ qfFormula2red formula ++
	    ",{" ++ concat (intersperse "," (map showVar [(Exactly 0,i) | i <- [0..nVars ineqs-1]])) ++ "}," ++ context2red ctx ++ ")"
    parN = nIndeterminatesN polys
    polys = map polyNumerator (entries ineqs) ++ map polyDenominator (entries ineqs)

    formula = ineqsToRFormula ctx ineqs Nothing
 
    nonexistent = rlqe (FoAll [0..nVars ineqs-1] $ FoQf formula) (contextToTheo ctx)

    convLM :: Integral a
           => [(QfFormula Exp, [(Exp,Exp)])]
           -> DTree (Polynomial (Ratio a))
                    (Optimum [FracPoly (Ratio a)])
    convLM [] = 
	FCond (fmap fromExp nonexistent) (DLeaf Nonexistent) (DLeaf Unbounded)
    convLM ((cond,vs):rest) =
	FCond (fmap fromExp cond) (DLeaf $ Optimum $ map mkFP vs) (convLM rest)
    mkFP (num,den) = fromExp num %% fromExp den


reduceLexmax :: Integral a
	     => [QfFormula (Polynomial (Ratio a))]
	     -> GIneqSystem (FracPoly (Ratio a))
	     -> DTree (Polynomial (Ratio a))
		      (Optimum [FracPoly (Ratio a)])
reduceLexmax ctx ineqs =
    fmap (fmap (map negate)) $ reduceLexmin ctx (mirrorIneqs ineqs)

------------

-- |
-- Factorise a polynomial.
-- Return the content of the given polynomial and
-- the primitive polynomial factors together with their
-- multiplicity.
--
reduceFactoriseP :: Integral a
		 => Polynomial a
		 -> (a, [(Polynomial a, Integer)])
reduceFactoriseP (poly :: Polynomial a)
  | Just c <- isConstPoly poly  =  (c, [])
  | totalDegree poly == 1       =  (cont, [(fmap' (`div` cont) poly, 1)])
  | otherwise  =
    (product [constantCoeff e ^ f | (e,f) <- coeffs], polys)
    where
    cont = foldl1 gcd $ collectCoeffs poly
    factors = runReduce (printer "") (input "") readFs
    (coeffs,polys) = partition (\(p,_) -> totalDegree p <= 0) factors

    printer = showString "res=>printList(res,m=>printPair(m,n=>printFormula(n,"
	      . shows nV . showString "),o=>write(o" . showString ")))"
    input = showString "factorize(" . expc2red poly . showString ")"

    nV = unNIndeterminates $ nIndeterminates poly
    
    readFs :: String -> Maybe [(Polynomial a,Integer)]
    readFs str = case reads str of
		   (res :: [(Exp,Integer)], ""):_ -> Just $
				 map (\(e,expo) -> (fromExp e,expo)) res
		   _         -> Nothing

reduceFactorise :: Integral a
		=> a
		-> [(a, Integer)]
reduceFactorise (value :: a) =
    factors
    where
    factors = runReduce (printer "") (input "") readFs

    printer = showString "res=>printList(res,m=>printPair(m,(o=>write(o)),(p=>write(p))))"
    input = showString "factorize(" . shows value . showString ")"
    
    readFs :: String -> Maybe [(a,Integer)]
    readFs str = case reads str of
		   (res :: [(Integer,Integer)], ""):_ -> Just $
				 map (\(e,expo) -> (fromIntegral e,expo)) res
		   _          -> Nothing

reduceResultant :: RealFrac a => Polynomial a -> Polynomial a -> Polynomial a
reduceResultant p q = -- trace (show (totalDegree p, totalDegree q)) $
    ((1/(fromIntegral a^degree q * fromIntegral b^degree p)) `pmul`) $
    mapPoly fromIntegral $
    deleteIndeterminates 1 $
    runReduce ("m=>printFormula(m," ++ showNI nInd ++ ")")
	      ((showString "resultant(" . exp2red p1 . showString "," .
		exp2red q1 . showString "," . showsVar (nInd,0) .
                showString ")") "")
	      (fmap fromExp . readExp)
    where
    p1 = setNIndets nInd $ toExp p'
    q1 = setNIndets nInd $ toExp q'
    nInd = nIndeterminatesN [p,q]

    (a,p') = ratPoly2intPoly $ mapPoly toRational p
    (b,q') = ratPoly2intPoly $ mapPoly toRational q


reduceConvexUnion :: Integral a =>
		     [QfFormula (Polynomial (Ratio a))] ->
		     [GIneqSystem (FracPoly (Ratio a))] ->
		     DTree (Polynomial (Ratio a))
		           (Maybe (SortedSystem (FracPoly (Ratio a))))
reduceConvexUnion ctx [] = error "reduceConvexUnion not defined for []"
reduceConvexUnion ctx ineqs =
    assert (all ((==0).nParams) ineqs) $
    sortedToDTree $ runReduce printer input readSorted
    where
    printer = "m=>printSorted(m,0," ++ showNI parN ++ ")"
    input    = "convexUnion(" ++ formulas ++ ",{" ++
	       concat (intersperse "," $
		       map showVar [(Exactly 0,i) | i <- [0..varN-1]]) ++
	       "}," ++ context2red ctx ++ ")"
    formulas = "{" ++ concat (intersperse ","
			      (map (\is -> qfFormula2red $ ineqsToRFormula ctx is Nothing) ineqs))
	       ++ "}"
    varN = nVars $ head ineqs
    parN = nIndeterminatesN polys

    polys = map polyNumerator allEntries ++ map polyDenominator allEntries
    allEntries = concatMap entries ineqs


mbRead :: Read a => String -> Maybe a
mbRead str = case reads str of
	       []      -> Nothing
	       (r,_):_ -> Just r

-- |
-- Compute a disjoint union of the given polyhedra
-- (linear parameters allowed).
--
reduceDisjointUnion :: RealFrac a
		    => [QfFormula (Polynomial a)]  -- ^ context
		    -> [GIneqSystem (FracPoly a)]  -- ^ union of polyhedra
		    -> [GIneqSystem (FracPoly a)]
reduceDisjointUnion ctx ins = reduceDisjointUnionDifference ctx ins []

-- |
-- Compute a disjoint union of @\/A - \/B@
-- (linear parameters allowed).
--
reduceDisjointUnionDifference :: RealFrac a
                              => [QfFormula (Polynomial a)] -- ^ context
                              -> [GIneqSystem (FracPoly a)] -- ^ A
                              -> [GIneqSystem (FracPoly a)] -- ^ B
                              -> [GIneqSystem (FracPoly a)]
reduceDisjointUnionDifference _ [] _   = []
reduceDisjointUnionDifference ctx ins outs =
    assert (length (nub $ map nVars   (ins++outs)) == 1) $
    assert (length (nub $ map nParams (ins++outs)) == 1) $
    matricesToIneqs $ runReduce printer input (mbRead :: String -> Maybe [[[(Exp,Exp)]]])
    where
    matricesToIneqs :: RealFrac a => 
		       [[[(Exp,Exp)]]] -> [GIneqSystem (FracPoly a)]
    matricesToIneqs esss =
	[ineqSystem ((nV,nP), matrix $ map (map mkEl) ess) | ess <- esss]
	where
	mkEl (n,d) = fromExp n %% fromExp d
  
    printer = "m=>printList(m, el=>printBounds(el," ++ show nVP ++ "," ++
	      showNI parN ++"))";
    input = "map(conjToBounds, disjointUnion(" ++ formula ++ "," ++ 
	    qfFormula2red (conjunctionQf ctx) ++ "))"
    formula = "(" ++
              (concat $ intersperse " or " $
	       map (\is -> qfFormula2red $ ineqsToRFormula ctx is Nothing) ins
              ) ++ ")" ++
              if null outs
               then ""
               else
                (" and not (" ++
                 (concat $ intersperse " or " $
	          map (\is -> qfFormula2red $ ineqsToRFormula ctx is Nothing) outs
                 ) ++ ")"
                )

    nV = nVars (head ins)
    nP = nParams (head ins)
    nVP = nV + nP
    parN = nIndeterminatesN polys

    polys = map polyNumerator (concatMap entries ins) ++
	    map polyDenominator (concatMap entries ins)

reduceDisjointUnionQf :: Real a
		      => [QfFormula (Polynomial a)]
		      -> QfFormula (Polynomial a)
		      -> [[AtFormula (Polynomial a)]]
reduceDisjointUnionQf ctx formula =
    case unNIndeterminates varN of
      0   -> if evalQf ((`compare`0).fromJust.isConstPoly) formula
		      then [[]]   -- disjunction of one empty conjunction (i.e., true)
		      else []     -- empty disjunction (i.e., false)
      _ ->
	  map (map unQf) $ runReduce printer input (mbRead :: String -> Maybe [[QfFormula Exp]])
	  where
	  unQf (QfAt atf) = fmap fromExp atf
	  unQf _          = error "reduceDisjointUnionQf: interal error: non-atomic formula found"
  
          printer = "n=>printList(n,m=>printList(m, el=>printFormula(el," ++ showNI varN ++ ")))"
	  input = "map(d=>rlatl(d),disjointUnion(" ++ qfFormula2red formula ++
		  "," ++ qfFormula2red (conjunctionQf ctx) ++ "))"
    where
    varN = nIndeterminatesN $ termsQf formula

------------

-- |
-- Compute the disjunctive normal form of a given formula under
-- a given context.
-- Return a list of disjuncts. Each disjunct is represented by
-- the list of its conjuncts (which are atomic formulas).
--
reduceDnfQf :: Integral a
	  => [QfFormula (Polynomial (Ratio a))]
          -> QfFormula (Polynomial (Ratio a))
	  -> [[AtFormula (Polynomial (Ratio a))]]
reduceDnfQf ctx formula =
    map (map (unQf . fmap fromExp)) $
    runReduce printer input (mbRead :: String -> Maybe [[QfFormula Exp]])
    where
    (varN, formulaR) = qfFormula2red' formula
    contextR = context2red ctx
    ctxR = qfFormula2red $ conjunctionQf ctx
    input = "map(f=>extract(f,and),findDisjuncts(rlsimpl(" ++ formulaR ++ ","
	    ++ contextR ++ ")," ++ ctxR ++ "))"
    printer = "l=>printList(l,m=>printList(m,n=>printFormula(n," ++ showNI varN ++ ")))"
    unQf (QfAt f)         = f
    unQf (QfNot (QfAt f)) = notAt f
    unQf _        =
	error "internal error: non-atomic formula met in 'reduceDnf'"

------------


runReduce :: String -> String -> (String -> Maybe a) -> a
runReduce printer inputExpr readFn =
    if Config.reuseReduceSession ()
       then runReduceIO2 printer inputExpr readFn
       else runReduceIO  printer inputExpr readFn
    

reduceCmd :: IO FilePath
reduceCmd = Config.externalToolPathOrError Config.REDUCE

{-# NOTINLINE reduceConn #-}
reduceConn :: ToolConnection
reduceConn = unsafePerformIO $
	     newToolConnection True  -- merge stderr with stdout
                               reduceCmd
			       (return [])
			       "Reduce"
			       reduceFunctions
			       True

runReduceIO2 :: String -> String -> (String -> Maybe a) -> a
runReduceIO2 printer inputExpr (readFn :: String -> Maybe a) =
--    incrementInvocationCounter >>
--    talkToReduce input readFn2
    talkToTool reduceConn input readFn2
    where
    readFn2 :: String -> Either ToolIOError (a,String)
    readFn2 str =
	case resOrError of
	  Right result ->
	      case readFn result of
		  Just value -> Right (value, rest)
		  Nothing    -> Left $ ToolIOError { tioeTool  = "Reduce",
						     tioeInput = input,
						     tioeError = UnparsableResult str
						   }
	  Left tioerr -> Left tioerr
	where
	(rest, resOrError) = findResOrError input str
    input = freeInputAndResultList ++
	    "write \"&\", printResult(" ++ printer ++ "," ++
	    inputExpr ++ "), \"#\";\n"
    freeInputAndResultList = "symbolic (inputbuflis!* := nil); \
			     \symbolic (resultbuflis!* := nil);"

-- look for the leading '&' and return the text
-- between the '&' and '#' (with whitespace before and after removed),
-- or return the error message after '*'.
--
findResOrError :: String -> String -> (String, Either ToolIOError String)
findResOrError input str =
    case break (\c -> c == '&' || c == '*') str of
       (_, '&':res) ->
	  case break (\c -> c == '#' || c == '*') res of
	     (result,'#':rest) -> (rest, Right $ trim result)
	     (_,'*':rest) -> (rest2,
			      Left $ mkTioe $ ToolFailed errline
			     )
	      where
	      (errline,rest2) = break (=='\n') $ dropWhile (=='*') rest
             _            -> ("", Left $ mkTioe $ ToolFailed "unexpected end of input" )  -- should never happen
       (_, '*':res) ->
           case [tio | (errmsg,tio) <- knownErrors
                     , any (errmsg `isPrefixOf`) msgtails] of
             tio:_ -> (rest2, Left $ mkTioe tio)
             []    -> (rest2, Left $ mkTioe $ ToolFailed msg)
	   where
           msgtails = tails $ map toLower msg
	   (rest2,msg) = extractMsg res
       _            -> ("", Left $ mkTioe $ ToolFailed "unexpected end of input")
    where
    extractMsg rest@('*':_) =
	(therest, line1 ++ "\n" ++ msg)
	where
	(line1,'\n':rest2) = break (=='\n') $ dropWhile (=='*') rest
	(therest, msg) = extractMsg (dropWhile isSpace rest2)
    extractMsg rest = (rest, "")

    mkTioe e = ToolIOError { tioeTool = "Reduce", tioeInput = input, tioeError = e }

knownErrors :: [(String, TIOError)]
knownErrors =
    [("head space low", HeapSpaceLow)
    ,(" , invalid after ", RestartableStrangeError ", invalid")
    ]

--
-- remove whitespace before passing the result string
-- to 'readFn'.
--
runReduceIO :: String -> String -> (String -> Maybe a) -> a
runReduceIO printer inputExpr readFn =
    unsafePerformIO $
    do reduceBin <- reduceCmd
       runToolIO True reduceBin [] input readFn2
    where
    readFn2 str =
	case break (=='&') str of
            (_,'&':rest) ->
		let (result, _)  = break (=='#') rest in
		case readFn (trim result) of
		   Just value -> Right value
		   Nothing    -> Left $ UnparsableResult str
	    _ -> Left (ToolFailed ("unexpected output (missing delimiters):\n" ++
				   str))
    input = reduceFunctions ++
	    "\nwrite \"&\", printResult(" ++ printer ++ "," ++ inputExpr ++ "), \"#\"; bye$"


-- ex(y, x*y > 0)
xyTest :: FoFormula (Polynomial Rational)
xyTest = FoEx [1] (FoQf $ QfAt $ AtGt $ indeterminate 2 0 * indeterminate 2 1)
