module LMath.Tools.ToolIO
    ( ToolConnection, ToolIOError(..), TIOError(..),
      newToolConnection, talkToTool,
      runToolIO, runToolIO',
      toolIOError, catchToolIOError,

      getToolPicos, resetToolPicos,
      getToolInvocationCounter, resetToolInvocationCounter
    )
where

import Control.Concurrent
import Control.Exception as CE
import Data.Dynamic
import Data.IORef
import Data.List
import Prelude
import System.IO
import System.IO.Unsafe
import System.Cmd
import System.Directory
import System.Exit
import System.Mem.Weak
import System.Posix    ( getProcessID )
import System.Posix.Signals ( installHandler, processStatusChanged,
			      openEndedPipe,
			      Handler(Ignore,Catch) )
import System.Process
import System.Time

import qualified Config

type ToolConnection = MVar (TA, Maybe TC)

data TC = TC { tcWriteHdl :: Handle,
	       tcOutput   :: String,
	       tcPH       :: ProcessHandle
	     }

data TA = TA { taStartCounter :: Integer,
	       taTalkCounter  :: Integer,
	       taName         :: String,
	       taInitialInput :: String,
	       taRestartOnHeapSpaceLow :: Bool,
               taMergeStderr  :: Bool,
	       taBinary    :: IO FilePath,
	       taArguments :: IO [String]
	     }

data ToolIOError = ToolIOError { tioeTool  :: String,
				 tioeInput :: String,
				 tioeError :: TIOError
			       }
		   deriving (Typeable, Show)

data TIOError = UnparsableResult String
	      | ToolFailed String
	      | InvocationError String
	      | HeapSpaceLow
              | RestartableStrangeError String
		deriving (Typeable, Eq, Show)

newToolConnection :: Bool
                  -> IO FilePath
		  -> IO [String]
		  -> String
		  -> String
		  -> Bool
		  -> IO ToolConnection
newToolConnection mergeStderr binary arguments name initialInput restartOnHeapSpaceLow =
    newMVar (TA { taStartCounter = 0, taTalkCounter = 0,
		  taName = name, taInitialInput = initialInput,
		  taRestartOnHeapSpaceLow = restartOnHeapSpaceLow,
                  taMergeStderr = mergeStderr,
		  taBinary = binary, taArguments = arguments },
	     Nothing)

--
-- Be careful to evaluate 'input' before aquiring the
-- toolconn MVar's lock; evaluating 'input' may trigger
-- another call to the same tool, and a second attempt to
-- aquire the MVar's lock would be made!
--
talkToTool :: ToolConnection
	   -> String
	   -> (String -> Either ToolIOError (a, String))
	   -> a
talkToTool toolconn unevalInput readFn =
    unsafePerformIO $
    do input <- mapM evaluate unevalInput
       modifyMVar toolconn (talk input)
    where
    talk input tconn =
	do tconn2 <- st tconn
	   (tconn3@(ta,_),res) <- ttt input tconn2
	   (tconnF,resF) <- case res of
			        Left (ToolIOError { tioeError = HeapSpaceLow })
				    | taRestartOnHeapSpaceLow ta  ->
					heapLowRetry (taName ta) input tconn3
                                Left (ToolIOError { tioeError = RestartableStrangeError msg }) ->
                                    strangeErrorRetry msg (taName ta) input tconn3
				_ -> return (tconn3,res)
	   case resF of
	       Right r     -> return (tconnF,r)
	       Left tioerr -> return (tconnF,throwDyn tioerr)

    strangeErrorRetry msg name input tconn =
	do hPutStrLn stderr ("A known error occurred with " ++ name ++ " (\""
                             ++ msg ++ "\"), retrying")
	   tconn' <- st tconn
	   ttt input tconn'

    heapLowRetry name input tconn =
	do hPutStrLn stderr ("Heap space low in " ++ name ++ ", retrying")
	   tconn' <- st tconn
	   ttt input tconn'

    st tconn@(ta,Just _) = return tconn
    st (ta,Nothing) =
	do binary <- taBinary ta
	   arguments <- taArguments ta
	   (ph, stdinHdl, output) <- startProcess mergeStderr binary arguments
	   hPutStr stdinHdl (taInitialInput ta)
--	   forkIO (mapM_ (\c -> appendFile "/tmp/toolio-output" [c]) output)
	   return (ta { taStartCounter = taStartCounter ta + 1 },
		   Just $ TC { tcWriteHdl = stdinHdl, tcOutput = output,
			       tcPH = ph })
        where
          mergeStderr = taMergeStderr ta
    ttt input (ta,Nothing) = error "internal error"
    ttt input (ta,Just tc) =
	do forkIO (hPutStr (tcWriteHdl tc) input >> hFlush (tcWriteHdl tc))
	   case readFn (tcOutput tc) of
	     Right (res,newOutput) -> return ((ta',Just (tc { tcOutput = newOutput })), Right res)
             Left tioerr ->
		  do CE.try (hClose $ tcWriteHdl tc)
		     CE.try (terminateProcess $ tcPH tc)
		     return ((ta',Nothing), Left tioerr)
	where
	ta' = ta { taTalkCounter = taTalkCounter ta + 1 }

--
-- Merge streams and return the result _lazily_.
--
lazyMerge :: [Handle] -> IO String
lazyMerge hdls =
    do ch <- newChan
       mapM_ (\hdl -> forkIO (copy2chan ch hdl)) hdls
       res <- getChanContents ch
       return (filterNothing (length hdls) res)
    where
    filterNothing 0 _ = []
    filterNothing n (Just x:rest)  = x : filterNothing n rest
    filterNothing n (Nothing:rest) = filterNothing (n-1) rest

    copy2chan ch hdl =
	do hSetBuffering hdl NoBuffering
	   s <- hGetContents hdl
	   mapM_ (writeChan ch . Just) s
	   writeChan ch Nothing

startProcess :: Bool -> FilePath -> [String] -> IO (ProcessHandle, Handle, String)
startProcess mergeStderr command arguments =
    do installHandler processStatusChanged Ignore Nothing
       installHandler openEndedPipe (Catch sigpipe) Nothing
       (stdinHdl, stdoutHdl, stderrHdl, ph) <-
	   runInteractiveProcess command arguments Nothing Nothing
       result <- if mergeStderr then
                   lazyMerge [stdoutHdl,stderrHdl]
                 else
                   -- when we do not merge standard error,
                   -- we copy it to the parent's standard error
                   -- not to block the process by a full standard error
                   -- channel
                   do forkIO (hGetContents stderrHdl >>= hPutStr stderr)
                      hGetContents stdoutHdl
       addFinalizer ph $
		    do -- Config.debugPutStrLn ("stdout/err finalizer for " ++
		       --                       command)
		       terminateProcess ph
       return (ph, stdinHdl, result)
    where
    sigpipe :: IO ()
    sigpipe = do hPutStrLn stderr
	          ("Ouch, SIGPIPE has been raised for '" ++ command ++
		   "'.\n   Subprocess exited prematurely?\n")

--
-- Run an external tool, communication via pipes.
-- 'readFn' needs to create enough strictness (by looking at
-- the characters in its argument) such that the output has
-- been consumed before runToolIO returns (and 'terminateProcess'
-- is called!).
--
runToolIO :: Bool -> FilePath -> [String] -> String -> (String -> Either TIOError a) -> IO a
runToolIO mergeStderr toolBin arguments input readFn =
    do (cnt,keep) <- incrementInvocationCounter
       startTime <- getPicoTime
       (ph, stdinHdl, output) <- startProcess mergeStderr toolBin arguments
       (do forkIO (hPutStr stdinHdl input `finally` hClose stdinHdl)
           case readFn output of
              Left  err -> throwIO $ DynException $ toDyn $
	                   ToolIOError { tioeTool  = toolBin,
					 tioeInput = input,
					 tioeError = err
				       }
	      Right res ->
	        do endTime <- getPicoTime
	           addToolPicos (endTime-startTime)
		   return res
        ) `finally` CE.try (terminateProcess ph)


catchToolIOError :: IO a -> (ToolIOError -> IO a) -> IO a
catchToolIOError = catchDyn

-- |
-- If the invoction of @action@ raises a 'ToolIOError', an
-- error message describing the problem is sent to standard error and
-- @alternative@ is executed.
--
toolIOError :: Maybe String  -- ^ description for error message
	    -> IO a    -- ^ action
	    -> IO a    -- ^ alternative
	    -> IO a
toolIOError descr action alternative =
    catchToolIOError action $ \(tioe :: ToolIOError) ->
    do case descr of
         Nothing -> return ()
	 Just d  -> hPutStr stderr (d ++ ": ")
       hPutStr stderr ("Tool " ++  tioeTool tioe  ++ " failed for input: \"" ++
		  tioeInput tioe ++ "\": ")
       case tioeError tioe of
	  UnparsableResult msg -> hPutStrLn stderr ("Unparsable result: \"" ++ msg ++ "\"")
	  ToolFailed msg -> hPutStrLn stderr msg
	  InvocationError msg -> hPutStrLn stderr ("Erroneous invocation: \"" ++ msg ++ "\"")
	  HeapSpaceLow   -> hPutStrLn stderr "Insufficient heap space."
          RestartableStrangeError msg ->
              hPutStrLn stderr ("Strange errer: \"" ++ msg ++
                                "\" (restartable)")
       alternative

{-# NOTINLINE toolPicos #-}
toolPicos :: IORef Integer
toolPicos = unsafePerformIO $ newIORef 0

getToolPicos :: IO Integer
getToolPicos = readIORef toolPicos

resetToolPicos :: IO ()
resetToolPicos = writeIORef toolPicos 0

addToolPicos :: Integer -> IO ()
addToolPicos inc = atomicModifyIORef toolPicos (\p -> (p+inc,()))

{-# NOTINLINE invocationCounter #-}
invocationCounter :: IORef (Integer, Bool)
invocationCounter = unsafePerformIO $
		    newIORef (0,False)

incrementInvocationCounter :: IO (Integer,Bool)
incrementInvocationCounter =
    atomicModifyIORef invocationCounter
		      (\(count,keep) -> ((count+1,keep),(count,keep)))

getToolInvocationCounter :: IO Integer
getToolInvocationCounter =
    readIORef invocationCounter >>= \(count, keep) ->
    return count

resetToolInvocationCounter :: IO ()
resetToolInvocationCounter =
    modifyIORef invocationCounter (\(_, keep) -> (0, keep))

setKeepToolIOfiles :: Bool -> IO ()
setKeepToolIOfiles keep =
    modifyIORef invocationCounter (\(count,_) -> (count,keep))


-- Run an external tool: write 'input' to a temporary file,
-- run the external tool 'toolBin' which reads the temporary file via stdin
-- and writes to another temporary file via stdout,
-- read the tool's output from the second temporary file, and
-- parse it using 'readFn'.
-- If the tool runs successfully and 'readFn' returns (Right _),
-- the temporary files are deleted immediately; otherwise the
-- temporary files are removed when 'invocationCounter' is finalised.
--
--
-- DO NOT USE
-- 'system' seems to be incompatible with 'runInteractiveProcess'
--
runToolIO' :: FilePath -> String -> (String -> Either TIOError a) -> IO a
runToolIO' toolBin input readFn =
    do (cnt,keep) <- incrementInvocationCounter
       pid <- getProcessID
       let ext = show pid ++ "." ++ show cnt
	   toolInFile  = "/tmp/hsloopo.tool-input."  ++ ext
	   toolOutFile = "/tmp/hsloopo.tool-output." ++ ext
           cleanUp = CE.try (removeFile toolInFile)  >>
                     CE.try (removeFile toolOutFile) >>
		     return ()
	   delayedCleanUp = mkWeakIORef invocationCounter cleanUp >> return ()
       ((do startTime <- getPicoTime
	    writeFile toolInFile input
            res <- system (toolBin ++ " <" ++ toolInFile ++ " >" ++ toolOutFile)
            case res of
             ExitSuccess     ->
	       do str <- strictReadFile toolOutFile
		  case readFn str of
	              Left tio -> throwIO $ DynException $ toDyn $ mkTioe tio
	              Right res ->
    	                do if keep then delayedCleanUp else cleanUp
	                   endTime <- getPicoTime
	                   addToolPicos (endTime-startTime)
	                   return res
	     ExitFailure err ->
	        throwIO $ DynException $ toDyn $ mkTioe $
	         ToolFailed (toolBin ++ " failed with exit code " ++ show err)
        )
        `CE.catch`
        (\ex -> delayedCleanUp >> throwIO ex)
        )
    where
    mkTioe e = ToolIOError { tioeTool = toolBin, tioeInput = input, tioeError = e }

strictReadFile :: FilePath -> IO String
strictReadFile fp = 
    do h   <- openFile fp ReadMode
       str <- getFully h
       hClose h
       return str

strictGetContents :: Handle -> IO String
strictGetContents hdl = getFully hdl `finally` hClose hdl

getFully :: Handle -> IO String
getFully hdl = hSetBuffering hdl LineBuffering >> input
    where
    input =
	do eof <- hIsEOF hdl
	   if eof then return ""
	          else do l    <- hGetLine hdl
			  rest <- input
			  return (l ++ '\n' : rest)

getPicoTime :: IO Integer
getPicoTime =
    getClockTime >>= \(TOD sec picos) ->
    return (sec*10^12+picos)


-- Test
catConn :: ToolConnection
catConn = unsafePerformIO $ newToolConnection False (return "/bin/cat") (return []) "cat" "" False

catReadFn str = let (a,b) = break (=='\n') str in Right (a,tail b)
