-- |
-- Provides a versatile data type for affine expressions.
-- Author: Armin Groesslinger
--
module LMath.Types.AffineExp
    ( GAffineExp, AffineExp, AffineExpBased(..),
      affineExp, affineExpVPC, vpcVersion,
      nVars, nParams, partsVPC,
      deleteAffineRows,
      insertZeroVars, insertZeroParams,
      insertVars, moveVars, deleteVars,
      insertParams, moveParams, deleteParams,
      reorderVars, reorderParams,
      moveVarsToParams, moveParamsToVars,
      substVar, substParam,
      integralAffineExp,
      afeParametersToPolynomials,
      affineExpToString, affineExpToString'
    )
where

import Data.Ratio

import LMath.Types.LClasses
import LMath.Types.Matrix
import LMath.Types.Polynomial
import LMath.Types.Vector


-- |
-- General data type for affine expressions.
-- A (multi-dimensional) affine expression is represented as
-- a matrix. Each row is a list of coefficients for the
-- variables, the parameters, and a constant term.
-- For example,
-- 
-- >   2   3   4   5
-- >   6   7   8   9
-- 
-- represents (assuming there are 2 variables "x", "y"
-- and 1 parameter "p"),
-- the expression
-- @ (2x+3y+4p+5, 6x+7y+8p+9) @
--
data GAffineExp a = AE { matr   :: GMatrix a,
			 vars   :: Integer,
			 params :: Integer }

type AffineExp = GAffineExp Rational

instance Functor GAffineExp where
    fmap f ae = ae { matr = fmap f (matr ae) }

class MatrixBased t entries => AffineExpBased t entries where
    fromAffineExp :: GAffineExp entries -> t
    toAffineExp   :: t -> GAffineExp entries

nVars, nParams :: (AffineExpBased aeb el, Integral a)
	       => aeb
	       -> a
nVars   aeb = fromIntegral $ vars   $ toAffineExp aeb
nParams aeb = fromIntegral $ params $ toAffineExp aeb

partsVPC :: AffineExpBased t a => t -> (GMatrix a, GMatrix a, GMatrix a)
partsVPC aeb =
    (subMatrixVer (matr ae)    0                  (vars ae - 1),
     subMatrixVer (matr ae) (vars ae)             (vars ae + params ae - 1),
     subMatrixVer (matr ae) (vars ae + params ae) (vars ae + params ae))
    where
    ae = toAffineExp aeb

instance AffineExpBased (GAffineExp a) a where
    fromAffineExp = id
    toAffineExp   = id

-- |
-- Constructs an affine expression from the number of variables,
-- parameters, and the coefficient matrix.
--
affineExp :: (Integral a, Integral b)
	  => ((a,b), GMatrix entries)  -- ^ @((vars,params), coeffs)@
          -> GAffineExp entries
affineExp ((v, p), mat)
    | nColumns mat == toInteger v + toInteger p + 1   =
	AE { matr = mat, vars=toInteger v, params=toInteger p }
    | nColumns mat == 0 && nRows mat == 0  =
	AE { matr = matrix (0, v'+p'+1, []),
	     vars = v', params = p' }
    | otherwise  = error ("affineExp:  number of columns in " ++
			  "matrix does not match number of " ++
			  "variables, parameters and constants")
    where
    v' = toInteger v
    p' = toInteger p

--
-- Turn a function like `affineExp' into a variable-parameter-constant
-- variant like `affineExpVPC', i.e. the ((vars,params),matrix) parameter
-- of the function is constructed from a (varMatr,paramMatr,constMatr)
-- parameter.
--
vpcVersion :: String ->
	      (((Integer,Integer), GMatrix entries) -> res) ->
	      ((GMatrix entries, GMatrix entries, GMatrix entries) -> res)
vpcVersion funName f (mV, mP, mC)
    | nRows mV == nRows mP && nRows mP == nRows mC && nColumns mC == 1 =
	f ((nColumns mV,nColumns mP), multiconc1 [mV, mP, mC])
    | otherwise =
	case (nColumns mC) of
           0 -> error (funName ++ ":  no constants column")
	   1 -> error (funName ++ ":  number of rows does not match ")
           _ -> error (funName ++ ":  more than one constants column")

-- |
-- Constructs an affine expression from the coefficients for
-- the variables, parameters, and the constant column.
--
affineExpVPC :: (GMatrix a,GMatrix a, GMatrix a) -> GAffineExp a
affineExpVPC = vpcVersion "affineExpVPC" affineExp
      

instance Num ent =>
    Applicable (GAffineExp ent) (GMatrix ent) (GMatrix ent) where
    apply ae m
     | nColumns ae == nRows m && nColumns m == 1 =
	 columnMatrix [entry (r*m) 0 0 | r <- rows ae]
     | otherwise = error "apply: AffineExp and Matrix don't match!"

instance Num ent => Applicable (GAffineExp ent) [ent] [ent] where
    apply ae rs
	| toInteger (length rs) == nColumns ae =
	    [vectp rs row | row <- rows' ae]
	| otherwise = error "apply:  GAffineExp and [ent] don't match"
	where
	vectp xs ys = sum $ zipWith (*) xs ys



instance Joinable (GAffineExp ent) where
    join is1 is2
	| nVars is1 == nVars is2 && nParams is1 == nParams is2  =
	    affineExp ((nVars is1, nParams is1), matr is1 `join` matr is2)
	| otherwise = error ("join:  affine expressions must have same " ++
			     "number of variables and parameters!")
    unjoin is = [affineExp ((nVars is, nParams is), matr) | matr <- rows is]


instance MatrixBased (GAffineExp ent) ent where
    setEntry is row col value = is { matr = setEntry (matr is) row col value }
    getMatrix is = matr is

deleteAffineRows :: (Integral b, AffineExpBased e ent) => e -> b -> b -> e
deleteAffineRows aeb idx count =
    fromAffineExp $ affineExp ((nVars aeb, nParams aeb), matr)
    where
    matr = deleteRows (getMatrix aeb) (fromIntegral idx) (fromIntegral count)

insertZeroVars :: (Num ent, AffineExpBased e ent, Integral b)
               => e -> b -> b -> e
insertZeroVars aeb idx width =
    insertVars aeb idx (zeroMatrix (nRows aeb) width)

insertZeroParams :: (Num ent, AffineExpBased e ent, Integral b)
                 => e -> b -> b -> e
insertZeroParams aeb idx width =
    insertParams aeb idx (zeroMatrix (nRows aeb) width)

insertVars :: (AffineExpBased e ent, Integral b) => e -> b -> GMatrix ent -> e
insertVars aeb idx coeffs
    | nRows coeffs /= nRows vs  =
	error "insertVars:  number of rows doesn't match"
    | idx < 0 || idx > nColumns vs =
	error ("insertVars:  index out of bounds: " ++ show idx ++
               ", but only " ++ show (nColumns vs) ++ " variables")
    | otherwise = fromAffineExp $ affineExpVPC (vs', ps, cs)
    where
    vs' = insertColumns vs idx coeffs
    (vs, ps, cs) = partsVPC aeb

insertParams :: (AffineExpBased e ent, Integral b)
             => e -> b -> GMatrix ent -> e
insertParams aeb idx coeffs
    | nRows coeffs /= nRows ps =
	error "insertParams:  number of rows doesn't match"
    | idx < 0 || idx > nColumns ps =
	error "insertParams:  index out of bounds"
    | otherwise = fromAffineExp $ affineExpVPC (vs, ps', cs)
    where
    ps' = insertColumns ps idx coeffs
    (vs, ps, cs) = partsVPC aeb

moveVars :: AffineExpBased e ent => e -> Integer -> Integer -> Integer -> e
moveVars aeb fromIdx toIdx count
    | fromIdx >= 0 && toIdx >= 0 && fromIdx + count <= nColumns vs &&
      toIdx + count <= nColumns vs && count >= 0 = fromAffineExp newAE
    | otherwise = error "moveVars:  index out of bounds"
    where
    newAE = affineExpVPC (vs', ps, cs)
    vs' = moveColumns vs fromIdx toIdx count
    (vs, ps, cs) = partsVPC aeb

moveParams :: AffineExpBased e ent => e -> Integer -> Integer -> Integer -> e
moveParams aeb fromIdx toIdx count
    | fromIdx >= 0 && toIdx >= 0 && fromIdx + count <= nColumns ps &&
      toIdx + count <= nColumns ps && count >= 0 = fromAffineExp newAE
    | otherwise = error "moveParams:  index out of bounds"
    where
    newAE = affineExpVPC (vs, ps', cs)
    ps' = moveColumns ps fromIdx toIdx count
    (vs, ps, cs) = partsVPC aeb

deleteVars :: (Integral b, AffineExpBased e ent) => e -> b -> b -> e
deleteVars aeb idx count
    | idx < 0 || count < 0 || idx + count > nVars aeb  =
	error "deleteVars:  index out of bounds"
    | otherwise = fromAffineExp $ affineExpVPC (vs', ps, cs)
    where
    vs' = deleteColumns vs (fromIntegral idx) (fromIntegral count)
    (vs, ps, cs) = partsVPC aeb

deleteParams :: AffineExpBased e ent => e -> Integer -> Integer -> e
deleteParams aeb idx count
    | idx < 0 || count < 0 || idx + count > nParams aeb  =
	error "deleteParams:  index out of bounds"
    | otherwise = fromAffineExp newAE
    where
    newAE = affineExp ((nVars aeb, nParams aeb - count), newCoeffs)
    newCoeffs = deleteColumns (getMatrix aeb) (nVars aeb + idx) count


reorderVars :: (AffineExpBased e ent, Integral b) => e -> [b] -> e
reorderVars aeb vs
    | not (null vs) && (minimum vs < 0 || maximum vs >= nVars aeb)  =
	error ("reorderVars:  index (" ++ show (maximum vs) ++ 
	       ") out of bounds [0.." ++ show (nVars aeb - 1) ++ "]")
    | otherwise = fromAffineExp newAE
    where
    newAE  = affineExpVPC (reorderColumns vars vs, params, consts)
    (vars, params, consts) = partsVPC aeb

reorderParams :: (AffineExpBased e ent, Integral b) => e -> [b] -> e
reorderParams aeb ps
    | not (null ps) && (minimum ps < 0 || maximum ps >= nParams aeb)  =
	error "reorderVars:  index out of bounds"
    | otherwise = fromAffineExp newAE
    where
    newAE  = affineExpVPC (vars, reorderColumns params ps, consts)
    (vars, params, consts) = partsVPC aeb

moveVarsToParams :: (AffineExpBased e ent, Integral b)
		 => e -> b -> b -> b -> e
moveVarsToParams aeb varIdx varCount paramIdx =
    fromAffineExp newAE
    where
    (vars, params, consts) = partsVPC aeb
    newAE = affineExpVPC (deleteColumns vars varIdx varCount,
			  insertColumns params paramIdx
			        (subMatrixVer vars varIdx (varIdx+varCount-1)),
			  consts)

moveParamsToVars :: (AffineExpBased e ent, Integral b)
		 => e -> b -> b -> b -> e
moveParamsToVars aeb paramIdx paramCount varIdx =
    fromAffineExp newAE
    where
    (vars, params, consts) = partsVPC aeb
    newAE = affineExpVPC (insertColumns vars varIdx
			                (subMatrixVer params paramIdx
					            (paramIdx+paramCount-1)),
			  deleteColumns params paramIdx paramCount,
			  consts)

---
-- Substitutes a value for a variable.
-- The variable <code>varNr</code> is replaced by <code>value</code>,
-- for example
-- <pre>
--   2x + 3y -  p + 1 &gt;= 0
--    x - 6y + 2p - 3 &gt;= 0
-- </pre>
-- with <code>substVar 1 (1%3)</code> becomes
-- <pre>
--    2x -  p + 2 &gt;= 0
--     x + 2p - 5 &gt;= 0
-- </pre>
--
-- @param aeb   - the original object
-- @param varNr - the number of the variable to substitute
--                <code>value</code> for
-- @param value - the value to use for <code>varNr</code>
-- @return - the original object with the given variable replaced
--           by the given value.
--
substVar :: (AffineExpBased e ent, Num ent) => e -> Integer -> ent -> e
substVar aeb varNr value
    | varNr < 0 || varNr >= nVars aeb  =
	error "substVar:  index out of bounds"
    | otherwise = fromAffineExp $
		  affineExpVPC (deleteColumns v varNr 1, p, c + c')
    where
    (v, p, c) = partsVPC aeb
    c' = subMatrixVer v varNr varNr `smul` value

--- Substitutes a value for a parameter.
-- The parameter <code>paramNr</code> is replaced by <code>value</code>.
--
-- @param aeb - the original object
-- @param paramNr - the number of the parameter to substitute
--                  <code>value</code> for.
-- @param value - the value to use for parameter <code>paramNr</code>.
-- @return - the original object with the given parameter replaced
--           by the given value.
--
-- @see substVar
--
substParam :: (AffineExpBased e ent, Num ent, Integral b)
  => e -> b -> ent -> e
substParam aeb paramNr value
    | paramNr < 0 || paramNr >= nParams aeb  =
	error "substParam:  index out of bounds"
    | otherwise = fromAffineExp $
		  affineExpVPC (v, deleteColumns p paramNr 1, c + c')
    where
    (v, p, c) = partsVPC aeb
    c' = subMatrixVer p paramNr paramNr `smul` value

-- |
-- Make the given affine expression integral by
-- multiplying each row with the least common multiple
-- of its denominators.
--
integralAffineExp :: (Real a, Integral b) =>
		     GAffineExp a -> GAffineExp b
integralAffineExp e = 
    affineExp ((nVars e, nParams e), 
	       rowVectorsToMatrix (nColumns m) $ map row2Integral $
	           rowVectors m
	      )
    where
    m = getMatrix e
	    
row2Integral :: (Real a, Integral b) => GVector a -> GVector b
row2Integral v = 
    fmap (\x -> fromIntegral (numerator x * (denomLCM `div` denominator x))) v'
    where
    -- lcm always returns a positive number
    denomLCM = foldl lcm 1 $ map denominator $ vectorEntries v'
    v' = fmap toRational v

afeParametersToPolynomials :: (Num a, AffineExpBased aeb1 a, AffineExpBased aeb2 (Polynomial a)) => aeb1 -> aeb2
afeParametersToPolynomials exp =
    fromAffineExp $ affineExp ((nV,0), v' `conc` c')
    where
    v' = fmap constPoly v
    c' = columnMatrix [affinePolyFromList (pr++cr)
                      | (pr,cr) <- zip (rows' p) (rows' c)]
    (v,p,c) = partsVPC exp

    nV = nVars exp
    nP = nParams exp

affineExpToString :: Bool -> Bool -> AffineExp -> String
affineExpToString nums inhRows exp = (if nums then show (vars exp) ++ "\n" ++ show (params exp) ++ "\n" else "")
    ++ matrixToString True (matr exp) (if inhRows then nRows exp else nRows exp - nParams exp - 1)

affineExpToString' :: Bool -> Bool -> AffineExp -> String
affineExpToString' nums inhRows exp = show rows ++ "\n"
    ++ (if nums then show (vars exp) ++ "\n" ++ show (params exp) ++ "\n" else "")
    ++ matrixToString False (matr exp) rows where
        rows = if inhRows then nRows exp else nRows exp - nParams exp - 1
