-- |
-- Affine mappings and related functions.
--
-- Author: Armin Groesslinger
--
module LMath.Types.AffineFunction
    ( GAffineFunction, AffineFunction, CoordinateSystem(..),
      affineFunction, affineFunctionVPC,
      targetDim, targetDimIH, getMatrixIH, partsVPCih,
      applyIH, composeAffine,
      affParametersToPolynomials,
      constantAffineFunction,
      affineFunctionToString,
      afMirror, afMirrorDim, afMirrorDims,
      afIdentity,
      afRelation
    )
where

import Data.List
import Data.Ratio

import LMath.Types.LClasses
import LMath.Types.Matrix
import LMath.Types.AffineExp
import LMath.Types.EqSystem
import LMath.Types.IneqSystem
import LMath.Types.Polynomial
import LMath.Types.Vector

-- |
-- Data type to represent affine mappings.
-- The repesentation is based on 'AffineExp'.
-- Note that the matrix representing the function is stored in
-- homogeneous coordinates,
-- i.e., the matrix for the function
-- 
-- >   (x,y)  |->  (2x+3y+4p+5, 6x+7y+8p+9)
-- 
-- is really
-- 
-- >   2   3   4   5
-- >   6   7   8   9
-- >   1   0   0   0
-- >   0   1   0   0
-- >   0   0   1   0
-- >   0   0   0   1
-- 
-- Use 'getMatrixIH', 'targetDimsIH', 'partsVPCih' to
-- extract information about the "interesting" part of
-- the matrix (i.e., without the homogenisation).
--
data GAffineFunction ent = AF { affineE :: GAffineExp ent }
type AffineFunction = GAffineFunction Rational

instance Functor GAffineFunction where
    fmap f af = fromAffineExp $ fmap f $ toAffineExp af

-- |
-- Data type to determine whether the matrix of an affine
-- function is given in homogeneous coordinates or not
-- when creating an affine function.
-- (After construction, the function is always described by
-- a matrix in homogeneous coordinates.)
--
data CoordinateSystem = Homogeneous | Inhomogeneous

-- |
-- Constructs a new affine function. Note that the affine functions is
-- always stored as a matrix in the homogeneous representation of
-- the affine mapping. The source coordinates must be given in
-- homogeneous coordinates. When @destCoordSys@ is
-- @Inhomogeneous@ then the target dimension are extended
-- with the identitiy in parameters and 1.
--
affineFunction :: (Num entries, Integral a, Integral b)
               => ((a,b), GMatrix entries)   -- ^ @((vars,params), matrix)@
               -> CoordinateSystem           -- ^ @destCoordSys@
               -> GAffineFunction entries
affineFunction arg Homogeneous
    | tD <= nP  =  error ("affineFunction: there's error in " ++
			  "the\n   dimensionality of the *homogeneous*" ++
			  " target space of the affine function:\n" ++
			  "number of variables:  " ++ show nV ++
			  "\nnumber of parameters: " ++ show nP ++
			  "\ndimensionality of target space: " ++ show tD ++
			  "\n" ++ show (getMatrix af))
    | homPart /= unitMatrix (nP+1) (nP+1) =
	error ("affineFunction: a homogeneous representation of an\n" ++
	       "   affine function must map the parameters and \"1\" onto\n" ++
	       "   themselves.\n\n" ++ show af)
    | otherwise  =  af
    where
    af = AF { affineE = affineExp arg }
    nV = nVars     af
    nP = nParams   af
    tD = targetDim af
    homPart = subMatrix (getMatrix af) (tD-nP-1) nV (tD-1) (nV+nP)
affineFunction arg Inhomogeneous =
    affineFunction ((nV, nP), matr `join` homPart) Homogeneous
    where
    af = AF { affineE = affineExp arg }
    nV = nVars   af
    nP = nParams af
    matr    = getMatrix af
    homPart = zeroMatrix (nP+1) nV  `conc`  unitMatrix (nP+1) (nP+1)
    
affineFunctionVPC :: Num entries =>
		     (GMatrix entries, GMatrix entries, GMatrix entries) ->
		     CoordinateSystem -> GAffineFunction entries
affineFunctionVPC = vpcVersion "affineFunctionVPC" affineFunction


-- |
-- Returns the dimensionality of the (homogeneous) target space.
--
targetDim :: Integral b => GAffineFunction ent -> b
targetDim af = nRows af

targetDimIH :: Integral b => GAffineFunction ent -> b
targetDimIH af = nRows af - nParams af - 1

getMatrixIH :: GAffineFunction ent -> GMatrix ent
getMatrixIH af = subMatrixHor (getMatrix af) 0 (targetDimIH af - 1)

partsVPCih :: GAffineFunction ent -> (GMatrix ent, GMatrix ent, GMatrix ent)
partsVPCih af = (subMatrixVer m 0 (nV-1), subMatrixVer m nV (nV+nP-1),
		 subMatrixVer m (nV+nP) (nV+nP))
    where
    m  = getMatrixIH af
    nV = nVars   af
    nP = nParams af

-- |
-- Make a constant affine function.
-- The value of the function is (regardless of actual variable or parameter
-- values) constantly the value represented by @constMatr@ (which
-- must be a column vector, of course). @vars@ and
-- @params@ are the number of variables and parameters, present
-- in the returned affine function, respectively.
--
constantAffineFunction :: Num ent
		       => Integer        -- ^ @vars@
                       -> Integer        -- ^ @params@
                       -> GMatrix ent    -- ^ @constMatrix@, constant part
                       -> GAffineFunction ent
constantAffineFunction vars params constMatr =
    affineFunctionVPC (zeroMatrix r vars,
		       zeroMatrix r params, 
		       constMatr)
                      Inhomogeneous
    where
    r = nRows constMatr


-- |
-- Apply an affine function and return the part of the result
-- tuple corresponding to the target dimension (and cut off
-- the parameters and the constant).
--
applyIH :: Num ent
        => GAffineFunction ent    -- ^ an affine function
        -> GVector ent            -- ^ values for the variables and parameters
        -> GVector ent
applyIH af vs = getMatrixIH af `mulMV` insertEntries vs (vectorLength vs) [1]



-- |
-- Compute the composition \"@f@ after @g@\" of two affine functions
-- @f@ and @g@
composeAffine :: Num a
              => GAffineFunction a  -- ^ @f@
              -> GAffineFunction a  -- ^ @g@
              -> GAffineFunction a
composeAffine f g
    | nParams f /= nParams g  =
        error ("composeAffine: different number of parameters when " ++
               "trying to compose " ++ show f ++ " and " ++ show g)
    | targetDimIH g /= nVars f  =
        error ("composeAffine: dimensionality mismatch when " ++
               "trying to compose " ++ show f ++ " and " ++ show g)
    | otherwise  =
        affineFunction ((nVars g, nParams g), getMatrix f * getMatrix g) 
                       Homogeneous

instance AffineExpBased (GAffineFunction ent) ent where
    toAffineExp      = affineE
    fromAffineExp ae = AF { affineE=ae }

instance MatrixBased (GAffineFunction ent) ent where
    getMatrix af = getMatrix (affineE af)
    setEntry af row col val = AF {affineE = setEntry (affineE af) row col val }

instance Num ent =>
    Applicable (GAffineFunction ent) (GMatrix ent) (GMatrix ent) where
    af `apply` m = (affineE af) `apply` m


instance Show ent => Show (GAffineFunction ent) where
    show af = "Affine function with " ++ show (nVars af) ++ " variables " ++
	      "and " ++ show (nParams af) ++ " parameters. Coefficient " ++
	      "matrix:\n" ++ show (getMatrix $ affineE af)

affParametersToPolynomials :: Num a => GAffineFunction a -> GAffineFunction (Polynomial a)
affParametersToPolynomials = fromAffineExp . afeParametersToPolynomials . toAffineExp

affineFunctionToString :: Bool -> Bool -> AffineFunction -> String
affineFunctionToString nums inhRows = affineExpToString nums inhRows . affineE


afMirror :: Num a => GAffineFunction a -> GAffineFunction a
afMirror af = affineFunction ((nV,nP), negate $ getMatrixIH af) Inhomogeneous
    where
    nV = nVars af
    nP = nParams af

afMirrorDim :: (Num a, Integral b)
  => b -> GAffineFunction a -> GAffineFunction a
afMirrorDim dim af = affineFunction ((nV,nP), m) Inhomogeneous
    where
    m = matrix (upper ++ [map negate middle] ++ lower)
    (upper, middle:lower) = genericSplitAt dim $ rows' $ getMatrixIH af
    nV = nVars af
    nP = nParams af

afMirrorDims :: (Num a, Integral b)
  => [b] -> GAffineFunction a -> GAffineFunction a
afMirrorDims dims af = foldl (\f d -> afMirrorDim d f) af dims

afIdentity :: (Integral b, Num a) => b -> b -> GAffineFunction a
afIdentity nV nP =
    affineFunctionVPC (unitMatrix nV nV, zeroMatrix nV nP, zeroMatrix nV 1) Inhomogeneous

-- |
-- Express an affine function as an equality system
-- between the source and target variables.
-- The target variables are in leading position
-- if the 'Bool' parameter is True, and in trailing position
-- otherwise.
--
afRelation :: Num a
           => Bool   -- ^ whether target variables are in leading position
           -> GAffineFunction a
           -> GEqSystem a
afRelation targetInFront af =
    eqSystemVPC (newV, afP, afC)
    where
    newV = if targetInFront
           then negU `conc` afV
           else afV  `conc` negU

    negU = negate $ unitMatrix tD tD

    (afV, afP, afC) = partsVPCih af
    nV = nVars af
    nP = nParams af
    tD = targetDimIH af