module LMath.Types.AlgNum 
  ( 
  -- * (terms for) Proper Restricted Extended Tarski Formulas
    TarskiPRE(..)
  , prePoly, mapPRE, mapPREpoly, liftPRE
  , usedIndeterminatesPRE, constPRESignum
  , levelQfEtf, invlevelQfEtf
  -- * Algebraic Numbers
  , Alg(..), AlgOps(..)
  , genericAlgOps
  , algOrderingP
  , algBiggerRat
  , algSmallerRat
  , algRatBetween
  )
where

import LMath.Types.Formula
import LMath.Types.LClasses
import LMath.Types.Polynomial
import LMath.Types.Signum

import Control.Exception ( assert )
import Data.List
import Data.Ratio

default (Integer)

data TarskiPRE a =
    PREPoly (Polynomial a)
  -- True: +, False: -, N=1,...
  | PRERoot Bool Integer (Polynomial a) -- +/-(x_k - _root_N f)
                   deriving (Eq, Show)

usedIndeterminatesPRE :: Integral b => TarskiPRE a -> [b]
usedIndeterminatesPRE = usedIndeterminates . prePoly

liftPRE :: TarskiPRE a -> TarskiPRE a
liftPRE (PREPoly p) = PREPoly $ liftPoly p
liftPRE (PRERoot sgn n p) =
    PRERoot sgn n $ liftPoly p

prePoly :: TarskiPRE a -> Polynomial a
prePoly (PREPoly p) = p
prePoly (PRERoot _ _ p) = p

mapPRE :: Num b => (a -> b) -> TarskiPRE a -> TarskiPRE b
mapPRE f (PREPoly p)       = PREPoly (mapPoly f p)
mapPRE f (PRERoot sgn n p) = PRERoot sgn n (mapPoly f p)

mapPREpoly :: (Polynomial a -> Polynomial b) -> TarskiPRE a -> TarskiPRE b
mapPREpoly f (PREPoly p) = PREPoly (f p)
mapPREpoly f (PRERoot sgn n p) = PRERoot sgn n (f p)

instance Num a => Num (TarskiPRE a) where
    negate (PREPoly p)      =  PREPoly $ negate p
    negate (PRERoot sgn n p) = PRERoot (not sgn) n p

    fromInteger = PREPoly . fromInteger

    (+) = error "(+) not defined vor TarskiPRE a"
    (*) = error "(*) not defined vor TarskiPRE a"
    abs = error "abs not defined vor TarskiPRE a"
    signum = error "signum not defined for TarskiPRE a"

constPRESignum :: (Num a, Ord a) => TarskiPRE a -> Signum
constPRESignum (PREPoly p)
    | Just c <- isConstPoly p  =  orderingToSignum (c `compare` 0)
constPRESignum _               =  allSignumsFalse

levelQfEtf :: (Num a, Integral b) => QfFormula (TarskiPRE a) -> b
levelQfEtf f = levelQf $ fmap prePoly f

invlevelQfEtf :: (Num a, Integral b) => QfFormula (TarskiPRE a) -> b
invlevelQfEtf f = invlevelQf $ fmap prePoly f



class (Num alg, Ord alg) => Alg alg where
    algIsRat :: alg -> Maybe Rational
    toAlg :: Real a => a -> alg
    zeroToAlg :: Real a => Polynomial a -> Integer -> alg
    algSubstP :: Real a => [alg] -> Polynomial a -> [TarskiPRE a]
    algOrdering :: alg -> Ordering
    fromAlg :: alg -> [(Rational,Rational)]
    algDefiningPoly :: alg -> Polynomial Rational

    -- implement algEvalP if algSubstP may not return a PRERoot
    -- for "arithmetic polynomials"
    algEvalP :: Real a => [alg] -> Polynomial a -> alg
    algEvalP algs p =
        preToAlg "algEvalP" $
        case nub $ algSubstP algs (liftPoly p' - indeterminate (nInd+1) 0) of
          [res] -> res
          xs    -> error ("algEvalP: multiple results: " ++ show xs ++
                          " for " ++ show (algs,p))
        where
        nInd = length algs
        (_,p'::Polynomial Integer) = ratPoly2intPoly $ fmap' toRational p

algOrderingP :: (Alg alg, Real a) => [alg] -> Polynomial a -> Ordering
algOrderingP algs p = algOrdering' $ algEvalP algs p

algSmallerRat :: (Alg alg, RealFrac a) => alg -> a
algSmallerRat alg
    | Just r <- algIsRat alg  =
        fromRational $
        if denominator r == 1  then  r-1  else  toRational (floor r)
    | (x,_):_ <- fromAlg alg  =  fromRational x - 1

algBiggerRat :: (Alg alg, RealFrac a) => alg -> a
algBiggerRat alg
    | Just r <- algIsRat alg  =
        fromRational $
        if denominator r == 1  then  r+1  else  toRational (ceiling r)
    | (_,y):_ <- fromAlg alg  =  fromRational y + 1

algRatBetween :: (Alg alg, RealFrac a) => alg -> alg -> a
algRatBetween alg1 alg2 =
    fromRational $
    f (map snd $ fromAlg alg1) (map fst $ fromAlg alg2)
    where
      f (x:xs) (y:ys)
        | x < y  =  (x+y)/2
        | otherwise  =  f xs ys

data AlgOps alg =
    AlgOps { algOpAdd :: alg -> alg -> alg
           , algOpMul :: alg -> alg -> alg
           , algOpSub :: alg -> alg -> alg
           , algOpNegate :: alg -> alg
           , algOpRecip  :: alg -> alg
           , algOpDiv :: alg -> alg -> alg
           , algOpSignum :: alg -> alg
           , algOpAbs :: alg -> alg
           , algOpFromInteger :: Integer -> alg
           , algOpFromRational :: Rational -> alg
           , algOpEq :: alg -> alg -> Bool
           , algOpCompare :: alg -> alg -> Ordering
           }

genericAlgOps :: Alg alg => AlgOps alg
genericAlgOps = ops
    where
    ops =
        AlgOps { algOpAdd = arithHelp (+) (\a b c -> a+b-c)
               , algOpMul = arithHelp (*) (\a b c -> a*b-c)
               , algOpSub = arithHelp (-) (\a b c -> a-b-c)
               , algOpNegate = \x -> arithHelp (\a b -> negate a)
                                               (\a b c -> a+c) x (toAlg 0)
               , algOpFromInteger = toAlg . fromInteger
               , algOpAbs = \x -> if algOrdering' x == LT
                                  then algOpNegate ops x
                                  else x
               , algOpSignum = \x -> case algOrdering' x of
                                       LT -> toAlg (-1)
                                       EQ -> toAlg 0
                                       GT -> toAlg 1
               , algOpCompare = algCompare'
               , algOpEq      = \a b -> algOrdering' (algOpSub ops a b) == EQ
               , algOpRecip = \x -> arithHelp (\a b -> recip a)
                                              (\a b res -> a*res-1) x (toAlg 0)
               , algOpDiv = arithHelp (/) (\a b res -> b*res-a)
               , algOpFromRational = toAlg
               }

algOrdering' :: Alg alg => alg -> Ordering
algOrdering' alg
    | Just r <- algIsRat alg  =  compare r 0
    | p <- algDefiningPoly alg, constantCoeff p /= 0  =
        findSign $ fromAlg alg
    | otherwise               =  algOrdering alg
    where
      findSign ((a,b):rest)
          | b < 0  =  LT
          | a > 0  =  GT
          | otherwise =  findSign rest

algCompare' :: Alg alg => alg -> alg -> Ordering
algCompare' alg1 alg2
    | Just r1 <- algIsRat alg1, Just r2 <- algIsRat alg2  =
        compare r1 r2
    | common    = algOrdering' $ algOpSub genericAlgOps alg1 alg2
    | otherwise = cmp (fromAlg alg1) (fromAlg alg2)
    where
      p1 = algDefiningPoly alg1
      p2 = algDefiningPoly alg2

      common = degree (gcdPoly p1 p1) > 0

      cmp ((a,b):rest1) ((c,d):rest2)
          | b < c  =  LT
          | d < a  =  GT
          | otherwise  =  cmp rest1 rest2

arithHelp :: Alg alg
          => (Rational -> Rational -> Rational)
          -> (   Polynomial Integer
              -> Polynomial Integer
              -> Polynomial Integer
              -> Polynomial Integer
             )
          -> alg
          -> alg
          -> alg
arithHelp opRat _ a b
    | Just aa <- algIsRat a, Just bb <- algIsRat b  =  toAlg (aa `opRat` bb)
arithHelp _ opP a b =
    case algSubstP [a,b] poly of
      [PRERoot _ n p] -> zeroToAlg p n
      [PREPoly p]
          | degree p == 1   ->
              toAlg (negate (constantCoeff p) % leadingCoeff p)
      pre  ->  error ("arithHelp: unexpected " ++ show pre)
    where
    -- opP a b res
    --  -->
    -- res = a `op` b
    --  -->
    -- (res - (a `op` b)) == 0
    poly = opP (indeterminate 3 1) (indeterminate 3 2) (indeterminate 3 0)

preToAlg :: Alg alg => String -> TarskiPRE Integer -> alg
preToAlg caller pre =
    case pre of
      PRERoot _ n p -> zeroToAlg p n
      PREPoly p
          | degree p == 1   ->
              toAlg (negate (constantCoeff p) % leadingCoeff p)
      pre  ->  error (caller ++ ": unexpected " ++ show pre)
