module LMath.Types.DTree.Contexts
    ( identityDTreeContext,
      numOrdDTreeContext, runNumOrdDTreeGenAlg,
      implSignDTreeContext,
      testPointDTreeContext
    )
where

import Data.Ratio

import LMath.Types.DTree.DTree
import LMath.Types.DTree.DTreeGen
import LMath.Types.Polynomial
import LMath.Types.Signum
import LMath.Types.Formula

-- |
-- Simple 'DTreeContext'.
--
-- Some algorithms like 'LMath.Algo.FM.isFeasibleSys' may not
-- work correctly if you use this DTreeContext. The reason is
-- that identityDTreeContext does not eliminate infinite parts
-- of the tree which cannot be reached. This can lead to infinite
-- recusions. So use this DTreeContext with care.
--
identityDTreeContext :: DTreeContext cond cond
identityDTreeContext =
    newDTreeContext []
     (\ctx f -> False)
     (\ctx p -> allSignumsFalse)
     (\ctx f -> f)
     id
     (\p -> Nothing)

-- |
-- Build a 'DTreeContext' which works for condition types
-- which are members of the 'Num' and 'Ord' type classes.
--
numOrdDTreeContext :: (Num cond, Ord cond)
                   => DTreeContext cond cond
numOrdDTreeContext =
    newDTreeContext []
     (\ctx f -> evalQf (`compare` 0) f)
     (\ctx p -> orderingToSignum (p `compare` 0))
     (\ctx f -> QfAt (if evalQf (`compare` 0) f then AtTrue else AtFalse))
     id
     (\p -> Just $ compare p 0)

-- |
-- With 'numOrdDTreeContext', there can never be case distinctions
-- in the result. Run the 'DTreeGen' and return the single leaf.
--
runNumOrdDTreeGenAlg :: (Num cond, Ord cond) => DTreeGen cond cond a -> a
runNumOrdDTreeGenAlg dtg = result
    where
    DLeaf result = makeDTree numOrdDTreeContext $ dtg

-- |
-- Build a 'DTreeContext' from a function which
-- evaluates the signum of a formula under a given context.
--
implSignDTreeContext :: [QfFormula cond]
		     -> ([QfFormula cond] -> cond -> Signum)
		     -> (c -> cond)
		     -> DTreeContext c cond
implSignDTreeContext ctx implSign condGen =
    newDTreeContext ctx
     (\ctx f -> implQf ctx f)
     (\ctx p -> implSign ctx p)
     (\ctx f -> f)
     condGen
     isConst
    where
    implQf ctx2 f =
	case evalQfSignum (implSign ctx2) f of
	  Just True -> True
	  _         -> False
    isConst p
        | signumEq sgn  =  Just EQ
        | signumGt sgn  =  Just GT
        | signumLt sgn  =  Just LT
        | otherwise     =  Nothing
        where
          sgn = implSign ctx p
          

-- |
-- A 'DTreeContext' which checks the conditions by
-- evaluating them at a given test point (assuming that the
-- condition has the same truth value for all allowed parameter
-- values including the test point).
--
testPointDTreeContext :: Integral a
                      => [Ratio a]
                      -> DTreeContext (FracPoly (Ratio a)) (Polynomial (Ratio a))
testPointDTreeContext testpoint =
    implSignDTreeContext []
       (\ctx cond -> orderingToSignum (evalPoly cond testpoint `compare` 0))
       (\fp -> polyNumerator fp * polyDenominator fp)

