-- |
-- Decision Trees and combinators for building them.
--
-- Author: Armin Groesslinger
--
-- Version: $Id: DTree.hs 4165 2009-07-23 17:16:50Z groessli $
--
module LMath.Types.DTree.DTree
  ( -- * Decision Tree Datatype
    DTree(..), unsafe_fmapDTree_with_context,
    compDTree, cmapDTree, showDTree,
    leavesDTree, flattenDTree,
    dtreeToFormula,
    combineSubtrees, combineSubtrees',
    simplifyDTreeTestPoint,
    simplifyDTreeOrd, simplifyDTreeMbOrd, simplifyDTreeSignum,
    unfracDTree,

    -- * Optimum
    Optimum(..),

    -- * Experimental Functions
    restructDTree
  )
where

import Control.Monad ( liftM )
import Data.Ratio
import Data.List (intersperse)
import Debug.Trace

import LMath.Types.Signum
import LMath.Types.Polynomial
import LMath.Types.Formula

-- |
-- Representation of optimal values.
--
data Optimum a = Optimum a
	       | Unbounded
	       | Nonexistent
		deriving (Eq, Show)

instance Functor Optimum where
    fmap f (Optimum x) = Optimum (f x)
    fmap _ Unbounded   = Unbounded
    fmap _ Nonexistent = Nonexistent


-- |
-- Decision Trees.
--
data DTree cond val = DLeaf val
		    | SCond  cond (DTree cond val) (DTree cond val) (DTree cond val)
		    | EqCond cond (DTree cond val) (DTree cond val)
		    | GeCond Bool cond (DTree cond val) (DTree cond val)
		    | FCond (QfFormula cond) (DTree cond val) (DTree cond val)
		      deriving (Eq)

instance Functor (DTree cond) where
    fmap f (DLeaf x) = DLeaf (f x)
    fmap f (SCond c tm t0 tp) = SCond  c (fmap f tm)  (fmap f t0) (fmap f tp)
    fmap f (EqCond c t0 tpm)  = EqCond c (fmap f t0)  (fmap f tpm)
    fmap f (GeCond leOk c t0p tm)  = GeCond leOk c (fmap f t0p) (fmap f tm)
    fmap f (FCond  c tt tb)   = FCond  c (fmap f tt)  (fmap f tb)

eqDTree :: (cond -> cond -> Bool)
        -> (val -> val -> Bool)
        -> DTree cond val
        -> DTree cond val
        -> Bool
eqDTree eqCond eqVal = eq
    where
    eq (DLeaf v1) (DLeaf v2) = v1 `eqVal` v2
    eq (SCond c1 tm1 t01 tp1) (SCond c2 tm2 t02 tp2) =
        (c1 `eqCond` c2) && eq tm1 tm2 && eq t01 t02 && eq tp1 tp2
    eq (EqCond c1 t01 tpm1) (EqCond c2 t02 tpm2) =
        (c1 `eqCond` c2) && eq t01 t02 && eq tpm1 tpm2
    eq (GeCond leOk1 c1 t0p1 tm1) (GeCond leOk2 c2 t0p2 tm2) =
        leOk1 == leOk2 && (c1 `eqCond` c2) &&
        eq t0p1 t0p2 && eq tm1 tm2
    eq (FCond c1 tt1 tb1) (FCond c2 tt2 tb2) =
        (eqQfFormula eqCond c1 c2) && eq tt1 tt2 && eq tb1 tb2
    eq _ _ = False

-- |
-- Applies the operation @f@ to all leaves of the decision tree @dt@.
-- The only difference between this function and a simple 'fmap' is
-- that fmap does not consider the context.
--
-- In this function the context will be collected and passed to the
-- transformation operation @f@. You can additionally define an
-- initial context which will be added to the collected context.
--
-- Warning:
-- It is strongly recommendend to simplify the input tree with the
-- current context. Otherwise @f@ may be called with an inconsistent
-- context, e.g. a context which implies @False@.
--
-- If you cannot guarantee that the tree is already simplified,
-- you should the function 'fmapDTree_with_context', which
-- simplifies the tree first.
--
-- See also: 'fmap', 'LMath.Logic.fmapDTree_with_context',
--           'LMath.Logic.QE.simplifyDTree'
--
unsafe_fmapDTree_with_context :: (Integral c)
		              => [QfFormula (Polynomial (Ratio c))]
			      -- ^ initial context
                              -> ([QfFormula (Polynomial (Ratio c))] -> a -> b)
			      -- ^ the transformation operation @f@
			      -> DTree (Polynomial (Ratio c)) a
			      -- ^ the original tree @dt@
			      -> DTree (Polynomial (Ratio c)) b
unsafe_fmapDTree_with_context initContext f =
    rec_impl initContext
    where
    rec_impl context (DLeaf x) = DLeaf (f context x)
    rec_impl context (SCond c tm t0 tp) =
	let tm' = rec_impl (QfAt        (AtGt c)  : context) tm
            t0' = rec_impl (QfAt        (AtEq c)  : context) t0
            tp' = rec_impl (QfNot (QfAt (AtGt c)) : context) tp
        in SCond c tm' t0' tp'
    rec_impl context (EqCond c t0 tpm) =
	let t0'  = rec_impl (QfAt (AtEq  c) : context) t0
            tpm' = rec_impl (QfAt (AtNeq c) : context) tpm
        in EqCond c t0' tpm'
    rec_impl context (GeCond leOk c t0p tm) =
	let t0p' = rec_impl (QfAt        (AtGe  c)  : context) t0p
            tm'  = rec_impl (QfNot (QfAt (AtGe c)) : context) tm
        in GeCond leOk c t0p' tm'
    rec_impl context (FCond c tt tf) =
	let tt' = rec_impl (      c : context) tt
            tf' = rec_impl (QfNot c : context) tf
        in FCond c tt' tf'

-- |
-- Collect all the leaves of a decision tree.
--
leavesDTree :: DTree cond val -> [val]
leavesDTree (DLeaf x) = [x]
leavesDTree (SCond _ tm t0 tp) = leavesDTree tm ++ leavesDTree t0 ++ leavesDTree tp
leavesDTree (EqCond _ t0 tpm) = leavesDTree t0 ++ leavesDTree tpm
leavesDTree (GeCond _ _ t0p tm) = leavesDTree t0p ++ leavesDTree tm
leavesDTree (FCond _ tt tb) = leavesDTree tt ++ leavesDTree tb

dtreeToFormula :: Num cond => DTree cond (QfFormula cond) -> QfFormula cond
dtreeToFormula (DLeaf qf) = qf
dtreeToFormula (SCond c tm t0 tp) =
    disjunctionQf [QfAt (AtGt $ negate c) `QfAnd` dtreeToFormula tm
                  ,QfAt (AtEq c) `QfAnd` dtreeToFormula t0
                  ,QfAt (AtGt c) `QfAnd` dtreeToFormula tp
                  ]
dtreeToFormula (EqCond c t0 tpm) =
    disjunctionQf [QfAt (AtEq  c) `QfAnd` dtreeToFormula t0
                  ,QfAt (AtNeq c) `QfAnd` dtreeToFormula tpm
                  ]
dtreeToFormula (GeCond elseOk c t0p tm) =
    disjunctionQf [QfAt (AtGe  c) `QfAnd` dtreeToFormula t0p
                  ,QfAt (atL   c) `QfAnd` dtreeToFormula tm
                  ]
    where
    atL co = (if elseOk then AtGe else AtGt) $ negate co
dtreeToFormula (FCond c tt tf) =
    disjunctionQf [      c `QfAnd` dtreeToFormula tt
                  ,QfNot c `QfAnd` dtreeToFormula tf
                  ]    

-- |
-- Flattens a decision tree, i.e. convert a tree of trees
-- into a simple tree.
--
-- See also: 'flattenDTreeG'
--
flattenDTree :: DTree cond (DTree cond val)
             -> DTree cond val
flattenDTree (DLeaf x) = x
flattenDTree (SCond c tm t0 tp) =
    SCond c (flattenDTree tm) (flattenDTree t0) (flattenDTree tp)
flattenDTree (EqCond c t0 tpm) =
    EqCond c (flattenDTree t0) (flattenDTree tpm)
flattenDTree (GeCond leOk c t0p tm) =
    GeCond leOk c (flattenDTree t0p) (flattenDTree tm)
flattenDTree (FCond c tt tf) =
    FCond c (flattenDTree tt) (flattenDTree tf)

-- |
-- Replace the fractions at the inner nodes of
-- a decision tree by polynomials. The resulting
-- decision tree is equivalent to the original tree.
--
unfracDTree :: Fractional a => DTree (FracPoly a) b -> DTree (Polynomial a) b
unfracDTree (DLeaf l) = DLeaf l
unfracDTree (SCond fp tm t0 tp) =
    SCond (signMonic (polyNumerator fp * polyDenominator fp))
          (unfracDTree tm)
	  (unfracDTree t0)
	  (unfracDTree tp)
unfracDTree (EqCond fp t0 tpm) =
    EqCond (signMonic (polyNumerator fp * polyDenominator fp))
	   (unfracDTree t0)
	   (unfracDTree tpm)
unfracDTree (GeCond leOk fp t0p tm) =
    GeCond leOk
           (signMonic (polyNumerator fp * polyDenominator fp))
	   (unfracDTree t0p)
	   (unfracDTree tm)
unfracDTree (FCond qf tt tb) =
    FCond (qfMapAt unfracAt qf) (unfracDTree tt) (unfracDTree tb)

unfracAt :: Num a => AtFormula (FracPoly a) -> AtFormula (Polynomial a)
unfracAt AtTrue = AtTrue
unfracAt AtFalse = AtFalse
unfracAt (AtEq  fp) = AtEq  (polyNumerator fp)
unfracAt (AtNeq fp) = AtNeq (polyNumerator fp)
unfracAt (AtGe  fp) = AtGe  (polyNumerator fp * polyDenominator fp)
unfracAt (AtGt  fp) = AtGt  (polyNumerator fp * polyDenominator fp)

instance (Show a, Show cond) => Show (DTree cond a) where
    show t = unlines $ fst $ showDTree t 1


showDTree :: (Show a, Show cond) =>
         DTree cond a -> Int -> ([String], Int)
showDTree (DLeaf m)  n =
    ([show n ++ ". " ++ show m], n+1)
{-    ([concat $ intersperse "\n" $ ((nr ++ r) : map (ind++) rs)], n+1)
    where
    (r:rs) = map (concat . intersperse "  " . map show . vectorEntries) m
    nr     = show n ++ ". "
    ind    = replicate (length nr) ' '
-}
showDTree (SCond cs tm t0 tp) n = (c:tms ++ t0s ++ tps, tl)
    where
    c = show n ++ ". SIGN " ++ show cs ++ " SWITCH " ++ show tmn ++
	", " ++ show t0n ++ ", " ++ show tpn
    tmn = n + 1
    (tms, t0n) = showDTree tm tmn
    (t0s, tpn) = showDTree t0 t0n
    (tps, tl ) = showDTree tp tpn
showDTree (EqCond cs t0 tpm) n = (c:t0s ++ tpms, tl)
    where
    c = show n ++ ". IF " ++ show cs ++ " == 0 THEN " ++ show t0n ++
	" ELSE " ++ show tpmn
    t0n = n + 1
    (t0s,  tpmn) = showDTree t0  t0n
    (tpms, tl  ) = showDTree tpm tpmn
showDTree (GeCond leOk cs t0p tm) n = (c:t0ps ++ tms, tl)
    where
    c = show n ++ ". IF " ++ show cs ++ " >= 0 THEN " ++ show t0pn ++
	" ELSE" ++ (if leOk then "0 " else " ") ++ show tmn
    t0pn = n + 1
    (t0ps,  tmn) = showDTree t0p  t0pn
    (tms,   tl ) = showDTree tm tmn
showDTree (FCond cs tt tb) n = (c:tts ++ tbs, tl)
    where
    c = show n ++ ". IF " ++ show cs ++ " THEN " ++ show ttn ++
	" ELSE " ++ show tbn
    ttn = n + 1
    (tts, tbn) = showDTree tt  ttn
    (tbs, tl ) = showDTree tb tbn


restructDTree :: DTree c (Maybe v) -> DTree c (Maybe v)
restructDTree l@(DLeaf _) = l
restructDTree (GeCond _ p (GeCond _ q t (DLeaf Nothing)) (DLeaf Nothing)) =
    restructDTree (FCond (QfAt (AtGe p) `QfAnd` QfAt (AtGe q)) t (DLeaf Nothing))
restructDTree (SCond c tm t0 tp) =
    SCond c (restructDTree tm) (restructDTree t0) (restructDTree tp)
restructDTree (GeCond leOk c t0p tm) =
    GeCond leOk c (restructDTree t0p) (restructDTree tm)
restructDTree (EqCond c t0 tpm) =
    EqCond c (restructDTree t0) (restructDTree tpm)
restructDTree (FCond c tt tb) =
    FCond c (restructDTree tt) (restructDTree tb)



compDTree :: (v -> DTree c w) -> DTree c v -> DTree c w
compDTree f (DLeaf x)  = f x
compDTree f (SCond c tm t0 tp) = SCond c (compDTree f tm) (compDTree f t0) (compDTree f tp)
compDTree f (EqCond c t0 tpm) = EqCond c (compDTree f t0)  (compDTree f tpm)
compDTree f (GeCond leOk c t0p tm) = GeCond leOk c (compDTree f t0p) (compDTree f tm)
compDTree f (FCond  c tt tb)  = FCond  c (compDTree f tt)  (compDTree f tb)


cmapDTree :: (c -> d) -> DTree c val -> DTree d val
cmapDTree f (DLeaf val) = DLeaf val
cmapDTree f (SCond c tm t0 tp) =
    SCond (f c) (cmapDTree f tm) (cmapDTree f t0) (cmapDTree f tp)
cmapDTree f (EqCond c t0 tpm) = EqCond (f c) (cmapDTree f t0) (cmapDTree f tpm)
cmapDTree f (GeCond leOk c t0p tm) =
    GeCond leOk (f c) (cmapDTree f t0p) (cmapDTree f tm)
cmapDTree f (FCond qf tt tb)  =
    FCond (fmap f qf) (cmapDTree f tt) (cmapDTree f tb)


mixDTrees :: Eq cond
          => (val -> val -> Maybe val)
          -> DTree cond val
          -> DTree cond val
          -> Maybe (DTree cond val)
mixDTrees mixVals t1 t2 = mix t1 t2
    where
    mix (DLeaf a) (DLeaf b) = liftM DLeaf $ mixVals a b
    mix (EqCond c1 t01 tpm1) (EqCond c2 t02 tpm2)
        | c1 == c2   =  mix2 (EqCond c1) t01 t02 tpm1 tpm2
    mix (GeCond leOk1 c1 t0p1 tm1) (GeCond leOk2 c2 t0p2 tm2)
        | leOk1 == leOk2 && c1 == c2  =
            mix2 (GeCond leOk1 c1) t0p1 t0p2 tm1 tm2
    mix (FCond qf1 tt1 tb1) (FCond qf2 tt2 tb2)
        | eqQfFormula (==) qf1 qf2  =  mix2 (FCond qf1) tt1 tt2 tb1 tb2
    mix (SCond c1 tm1 t01 tp1) (SCond c2 tm2 t02 tp2)
        | c1 == c2  = mix3 (SCond c1) tm1 tm2 t01 t02 tp1 tp2
    mix _ _ = Nothing

    mix2 fun a b c d =
        case (mix a b, mix c d) of
          (Just e, Just f) -> Just $ fun e f
          _                -> Nothing

    mix3 fun a b c d e f =
        case (mix a b, mix c d, mix e f) of
          (Just g, Just h, Just i) -> Just $ fun e f i
          _                        -> Nothing

    eqVal a b = case mixVals a b of
                  Just _  -> True
                  Nothing -> False


combineSubtrees :: (Num cond, Eq val)
		=> DTree cond val
                -> DTree cond val
combineSubtrees =
    combineSubtrees' (\a b -> if a == b then Just a else Nothing)

combineSubtrees' :: Num cond
                 => (val -> val -> Maybe val)
                 -> DTree cond val
                 -> DTree cond val
combineSubtrees' mixVals = comb
    where
    mix = mixDTrees mixVals

    comb l@(DLeaf _) = l
    comb (SCond c tm t0 tp) =
        case (mix tm t0, mix t0 tp, mix tm tp) of
          (Just tm0, _, _)
              | Just t <- mix tm0 tp  ->  t
          (Just tm0, _, _) -> GeCond False (negate c) tm0 tp'
          (Nothing, Just t0p, _) -> GeCond False c t0p tm'
          (Nothing, Nothing, Just tmp) -> EqCond c t0' tmp
          _                            -> SCond c tm' t0' tp'
        where
          tm' = comb tm
          t0' = comb t0
          tp' = comb tp
    comb (EqCond c t0 tpm) =
        case mix t0 tpm of
          Just t  -> t
          Nothing -> EqCond c t0' tpm'
        where
          t0'  = comb t0
          tpm' = comb tpm
    comb (GeCond leOk c t0p tm) =
        case mix t0p tm of
          Just t  -> t
          Nothing -> GeCond leOk c t0p' tm'
        where
          t0p' = comb t0p
          tm'  = comb tm
    comb (FCond f tt tb) =
        case mix tt tb of
          Just t  -> t
          Nothing -> FCond f tt' tb'
        where
          tt' = comb tt
          tb' = comb tb



simplifyDTreeTestPoint :: (Num a, Ord a)
		       => DTree (Polynomial a) val
		       -> [a] 
		       -> DTree (Polynomial a) val
simplifyDTreeTestPoint tree points =
    simplifyDTreeMbOrd (\p -> Just (evalPoly p points `compare` 0)) tree



simplifyDTreeOrd :: (Num cond, Ord cond) => DTree cond a -> DTree cond a
simplifyDTreeOrd dtree = simplifyDTreeMbOrd (Just.(`compare`0)) dtree

simplifyDTreeMbOrd :: (cond -> Maybe Ordering) -> DTree cond a -> DTree cond a
simplifyDTreeMbOrd _ l@(DLeaf _) = l
simplifyDTreeMbOrd ordFn (SCond c tm t0 tp) =
    case ordFn c of
      Just LT -> simplifyDTreeMbOrd ordFn tm
      Just EQ -> simplifyDTreeMbOrd ordFn t0
      Just GT -> simplifyDTreeMbOrd ordFn tp
      Nothing -> SCond c (simplifyDTreeMbOrd ordFn tm)
		         (simplifyDTreeMbOrd ordFn t0)
			 (simplifyDTreeMbOrd ordFn tp)
simplifyDTreeMbOrd ordFn (EqCond c t0 tpm) =
    case ordFn c of
      Nothing -> EqCond c (simplifyDTreeMbOrd ordFn t0)
		          (simplifyDTreeMbOrd ordFn tpm)
      Just EQ -> simplifyDTreeMbOrd ordFn t0
      _       -> simplifyDTreeMbOrd ordFn tpm
simplifyDTreeMbOrd ordFn (GeCond leOk c t0p tm) =
    case ordFn c of
      Nothing -> GeCond leOk c (simplifyDTreeMbOrd ordFn t0p)
		               (simplifyDTreeMbOrd ordFn tm)
      Just LT -> simplifyDTreeMbOrd ordFn tm
      _       -> simplifyDTreeMbOrd ordFn t0p
simplifyDTreeMbOrd ordFn (FCond f tt tb) =
    case evalQfMb ordFn f of
	Just True  -> simplifyDTreeMbOrd ordFn tt
	Just False -> simplifyDTreeMbOrd ordFn tb
	Nothing    -> FCond f (simplifyDTreeMbOrd ordFn tt)
		              (simplifyDTreeMbOrd ordFn tb)

simplifyDTreeSignum :: Num cond
		    => (cond -> Signum)
		    -> DTree cond a
		    -> DTree cond a
simplifyDTreeSignum sgn l@(DLeaf _) = l
simplifyDTreeSignum sgn (SCond c tm t0 tp)
    | signumLe s =
	if signumLt s
	   then simplifyDTreeSignum sgn tm
	   else if signumEq s
		   then simplifyDTreeSignum sgn t0
		   else GeCond True (negate c) (simplifyDTreeSignum sgn tm)
			                       (simplifyDTreeSignum sgn t0)
    | signumGe s =
	if signumGt s
	   then simplifyDTreeSignum sgn tp
	   else GeCond True c (simplifyDTreeSignum sgn tp)
		              (simplifyDTreeSignum sgn t0)
    | signumNeq s = GeCond True c (simplifyDTreeSignum sgn tp)
		                  (simplifyDTreeSignum sgn tm)
    | otherwise = SCond c (simplifyDTreeSignum sgn tm)
		          (simplifyDTreeSignum sgn t0)
			  (simplifyDTreeSignum sgn tp)
    where
    s = sgn c
simplifyDTreeSignum sgn (EqCond c t0 tpm)
    | signumEq s  = simplifyDTreeSignum sgn t0
    | signumNeq s = simplifyDTreeSignum sgn tpm
    | otherwise = EqCond c (simplifyDTreeSignum sgn t0)
		           (simplifyDTreeSignum sgn tpm)
    where
    s = sgn c
simplifyDTreeSignum sgn (GeCond leOk c t0p tm)
    | signumGe s = simplifyDTreeSignum sgn t0p
    | leOk =
	if signumLe s then simplifyDTreeSignum sgn tm else cont
    | otherwise = 
	if signumLt s then simplifyDTreeSignum sgn tm else cont
    where
    cont = GeCond leOk c (simplifyDTreeSignum sgn t0p)
		         (simplifyDTreeSignum sgn tm)
    s = sgn c
simplifyDTreeSignum sgn (FCond f tt tb) =
    case evalQfSignum sgn f of
	Just True  -> simplifyDTreeSignum sgn tt
	Just False -> simplifyDTreeSignum sgn tb
	Nothing    -> FCond f (simplifyDTreeSignum sgn tt)
		              (simplifyDTreeSignum sgn tb)


