module LMath.Types.DTree.DTreeGen
    ( -- * Combinators for DTree generation
      DTreeGen,
      dLeaf,
      sCond, eqCond, geCond, geCond', leCond, geleCond, fCond, fCond',
      compDTreeG, mapDTreeG, cmapDTreeG,
      prod2DTreeG, prodDTreeG, flattenDTreeG,
      DTreeContext,
      dtcCondGen, dtcSimplifyIsWeak, dtcSimplifyConditions, dtcContext,
      getDTreeContext,
      newDTreeContext,
      withDTreeContext, makeDTree, makeDTreeSimplifyConds,
      dtcImpliesQf, dtcImpliesSignum, dtcContextImpliesSignum,
      combineSubtreesDTreeG, combineSubtreesDTreeG'

    -- * Setting Options of Algorithms
    , setDTGOption, lookupDTGOption, lookupDTGOptionWithDefault
    )
where

import Data.Dynamic
import Data.List  ( nub )
import Data.Maybe ( listToMaybe )
import Debug.Trace

import LMath.Types.DTree.DTree
import LMath.Types.Signum
import LMath.Types.Formula

-- |
-- A DTree \"context\".
-- A 'DTreeContext' represents the context of a node in a 'DTree', i.e.,
-- the conditions which hold on the path to the respective node.
-- The conditions stored in the context are of type @QfFormula cond@.
-- The values of type @cond@ are generated (by a function stored
-- in the 'DTreeContext') from values of type @c@, which
-- is the type on which case distinctions are initially made. For example,
-- when dealing with @GIneqSystem (FracPoly Rational)@, type @c@ is
-- @FracPoly Rational@ and type @cond@ is (usually) @Polynomial Rational@.
-- In addition to that, a 'DTreeContext' also stores functions
-- to check for logical implication (whether some conditions and 
-- the context together imply another condition), and simplify formulas
-- under the context.
--
-- To communicate options to algorithms, the context can also store
-- option values (identified by their types). See 'setDTGOption' and
-- friends for this feature.
--
-- See 'numOrdDTreeContext' and
-- 'LMath.Logic.DTreeContexts.fracPolyDTreeContext' for how to
-- get an initial 'DTreeContext'.
--
data DTreeContext c cond =
    DTreeContext
    { dtcContext        :: [QfFormula cond]
    , dtcSimplifyQfS    :: [QfFormula cond] -> QfFormula cond -> QfFormula cond
    , dtcImpliesQfS     :: [QfFormula cond] -> QfFormula cond -> Bool
    , dtcImpliesSignumS :: [QfFormula cond] -> cond -> Signum
    , dtcCondGen        :: c -> cond

      -- experimental support for condition simplification
    , dtcSimplifyConditions :: Bool
    , dtcSimplifyIsWeak :: Bool
    , dtcFactoriseS :: cond -> [cond]
    , dtcIsConstant :: cond -> Maybe Ordering

      -- conditions which are known to be true (experimental)
    , dtcImpliedConds :: [QfFormula cond]

    , dtcAlgorithmOptions :: [Dynamic]
    }

-- |
-- Check whether two contexts are to be considered equivalent.
-- Note that _only_ the context formulas (dtcContext) are
-- compared.
--
eqDTreeContext :: Eq cond
               => DTreeContext c cond -> DTreeContext c cond -> Bool
eqDTreeContext dtc1 dtc2 =
    dtcContext dtc1 == dtcContext dtc2

-- |
-- A function mapping a 'DTreeContext' to an non user visible
-- (i.e., opaque) representation of a 'DTree' carrying values of
-- type @val@ at the leaves.
-- All the combinators return a 'DTreeGen' function.
--
newtype DTreeGen c cond val = DTreeGen (DTreeContext c cond -> DTree cond (val, DTreeContext c cond))


instance Monad (DTreeGen c cond) where
    return = dLeaf
    f >>= g  = compDTreeG g f

-- |
-- Get the DTree context.
-- Intended to be used with monadic DTreeGen construction.
-- For example:
--
-- @
--   do dtc <- getDTreeContext
--      ... dtcImpliesQf dtc ...
-- @
--
getDTreeContext :: DTreeGen c cond (DTreeContext c cond)
getDTreeContext = withDTreeContext $ \dtc -> return dtc


updateDTreeContext :: (DTreeContext c cond -> DTreeContext c cond)
                   -> DTreeGen c cond ()
updateDTreeContext f =
    DTreeGen (\dtc -> DLeaf ((), f dtc))

-- |
-- Ternary case distinction between @t \< 0@, @t == 0@, and @t > 0@.
-- 
sCond :: (Show c, Num cond)
      => c -- ^ term @t@
      -> DTreeGen c cond val -- ^ sub-tree for @t <  0@
      -> DTreeGen c cond val -- ^ sub-tree for @t == 0@
      -> DTreeGen c cond val -- ^ sub-tree for @t >  0@
      -> DTreeGen c cond val 
sCond p' (DTreeGen tm) (DTreeGen t0) (DTreeGen tp) = DTreeGen s
    where
    s dtc
	| signumGe sgn =
	    if      signumGt sgn then tp (newDtc $ QfAt $ AtGt p)
	    else if signumEq sgn then t0 (newDtc $ QfAt $ AtEq p)
	    else EqCond p
		        (t0 $ newDtc2 $ QfAt $ AtEq p)
		        (tp $ newDtc2 $ QfAt $ AtGt p)
	| signumLe sgn =
	    if signumLt sgn
	    then tm (newDtc $ QfAt $ AtGt $ negate p)
	    else EqCond p
		        (t0 $ newDtc2 $ QfAt $ AtEq p)
		        (tm $ newDtc2 $ QfAt $ AtGt $ negate p)
	| signumNeq sgn =
	    GeCond False p
		         (tp $ newDtc2 $ QfAt $ AtGt p)
		         (tm $ newDtc2 $ QfAt $ AtGt $ negate p)
	| otherwise = 
	    SCond p
		  (tm $ newDtc2 $ QfAt $ AtGt $ negate p)
	          (t0 $ newDtc2 $ QfAt $ AtEq p)
		  (tp $ newDtc2 $ QfAt $ AtGt p)
	where
	p = removeSgnFactors $ dtcCondGen dtc p'
        implSgnS = dtcImpliesSignumS dtc (dtcContext dtc)
	sgn = implSgnS p
	newDtc  f = addImpliedConds dtc [elim f]
        elim f = eliminateAtSignum (mbOrderingToSignum . dtcIsConstant dtc) f

	newDtc2 f = dtc { dtcContext = f : dtcContext dtc
                        }

        removeSgnFactors p = facts $ dtcFactoriseS dtc p
        facts [] = 1
        facts (f:fs)
            | signumGt fsgn = facts fs
            | signumLt fsgn = negate $ facts fs
            | otherwise     = f * facts fs
            where
            fsgn = implSgnS f

-- |
-- Binary case distinction between @t == 0@ and @t \/= 0@.
--
eqCond :: (Show c, Num cond)
       => c  -- ^ term @t@
       -> DTreeGen c cond val -- ^ sub-tree for @t == 0@
       -> DTreeGen c cond val -- ^ sub-tree for @t \/= 0@
       -> DTreeGen c cond val 
eqCond p = condOpt (QfAt $ AtEq p) False

-- |
-- Binary case distinction between @t >= 0@ and @t < 0@.
--
geCond :: (Show c, Num cond)
       => c  -- ^ term @t@
       -> DTreeGen c cond val -- ^ sub-tree for @t >= 0@
       -> DTreeGen c cond val -- ^ sub-tree for @t <  0@
       -> DTreeGen c cond val
geCond = geCond' False

-- |
-- Binary case distinction between @t <= 0@ and @t > 0@.
--
leCond :: (Show c, Num c, Num cond)
       => c  -- ^ term @t@
       -> DTreeGen c cond val -- ^ sub-tree for @t >= 0@
       -> DTreeGen c cond val -- ^ sub-tree for @t <  0@
       -> DTreeGen c cond val
leCond cond t1 t2 = geCond (negate cond) t1 t2

-- |
-- Binary case distinction between @t >= 0@ and @t <= 0@.
-- Use this combinator, if
-- in the case @t == 0@ both sub-trees hold the desired result.
-- A simplifier will then be able to replace the case distinction
-- by the @t <= 0@ branch if the context implies @t <= 0@.
--
geleCond :: (Show c, Num cond)
         => c  -- ^ term @t@
         -> DTreeGen c cond val  -- ^ sub-tree for @t >= 0@
         -> DTreeGen c cond val  -- ^ sub-tree for @t <= 0@
         -> DTreeGen c cond val
geleCond = geCond' True

-- |
-- Binary case distinction between @t >= 0@ and @t < 0@ or @t <= 0@.
-- The 'Bool' parameter decides
-- whether this combinator behaves as 'geCond' (@False@)
-- or 'geleCond' (@True@).
--
geCond' :: (Show c, Num cond)
        => Bool  -- ^ @t < 0@ branch is valid also for @t == 0@.
	-> c     -- ^ term @t@
        -> DTreeGen c cond val  -- ^ sub-tree for @t >= 0@
        -> DTreeGen c cond val  -- ^ sub-tree for @t <= 0@ or @t < 0@.
        -> DTreeGen c cond val
geCond' leOk p = condOpt (QfAt $ AtGe p) leOk

-- |
-- Binary case distinction on a given formula.
--
fCond :: (Show c, Num cond)
      => QfFormula c  -- ^ condition @p@
      -> DTreeGen c cond val  -- ^ applicable sub-tree if @p@ holds
      -> DTreeGen c cond val  -- ^ applicable sub-tree if @p@ does not hold
      -> DTreeGen c cond val
fCond f = condOpt f False

fCond' :: (Show c, Num cond)
       => QfFormula cond  -- ^ condition @p@
       -> DTreeGen c cond val  -- ^ applicable sub-tree if @p@ holds
       -> DTreeGen c cond val  -- ^ applicable sub-tree if @p@ does not hold
       -> DTreeGen c cond val
fCond' f = condOpt' f False


condOpt :: (Show c, Num cond)
	=> QfFormula c
	-> Bool
        -> DTreeGen c cond val
	-> DTreeGen c cond val
	-> DTreeGen c cond val
condOpt inputQf elseOkForGe dtgT dtgF =
    withDTreeContext $ \dtc ->
    condOpt' (fmap (dtcCondGen dtc) inputQf) elseOkForGe dtgT dtgF

condOpt' :: (Show c, Num cond)
	=> QfFormula cond
	-> Bool
        -> DTreeGen c cond val
	-> DTreeGen c cond val
	-> DTreeGen c cond val
condOpt' (QfAt AtTrue)  _ dtgT _    = dtgT
condOpt' (QfAt AtFalse) _ _    dtgF = dtgF
condOpt' qf elseOkForGe (DTreeGen dtgTrue) (DTreeGen dtgFalse) =
    DTreeGen (\dtc -> cond dtc qf)
    where
    cond dtc unsimpQf =
      case signsimpl dtc unsimpQf of
        QfAt AtTrue  -> dtgTrue  dtc
        QfAt AtFalse -> dtgFalse dtc
        qf           -> cond2 dtc qf

    signsimpl dtc qf =
        eliminateAtSignum (mbOrderingToSignum . dtcIsConstant dtc) qf

    cond2 dtc unsimpQf =
      case syntSimpl (dtcContext dtc ++ dtcImpliedConds dtc) unsimpQf of
        QfAt AtTrue  -> dtgTrue dtc
        QfAt AtFalse -> dtgFalse dtc
        qf -> cond3 dtc qf

    cond3 dtc unsimpQf =
--        trace (show (simplifyConds, unsimpQf, simpQf)) $
	case theQf of
	    QfAt AtTrue  -> dtgTrue  newDtcT
	    QfAt AtFalse -> dtgFalse newDtcF
	    QfAt (AtGt p)
		| signumGt s -> dtgTrue  $ newDtcT
		| signumLe s -> dtgFalse $ newDtcF
		| otherwise  ->
		    GeCond elseOkForGe (negate p) tFalse tTrue
		where
		s = sign p
	    QfAt (AtGe p)
		| signumGe s -> dtgTrue  $ newDtcT
		| signumLt s -> dtgFalse $ newDtcF
		| otherwise  ->
		    GeCond elseOk p tTrue tFalse
		where
		s = sign p
		-- we need to make sure that we don't get an Eq case
		-- when elseOk==True and there was no Eq case in the
		-- original AtGe.
		elseOk
		    | elseOkForGe =
			 (case unsimpQf of
			    QfAt (AtGe p1) ->
			       dtcImpliesQfS dtc (dtcContext dtc)
			               (QfAt (AtEq p) `qfImpl` QfAt (AtEq p1))
			    _ -> False
			 )
		    | otherwise  = False
            QfAt (AtEq p)
		| signumEq  s -> dtgTrue  $ newDtcT
		| signumNeq s -> dtgFalse $ newDtcF
		| otherwise  ->
		    EqCond p tTrue tFalse
		where
		s = sign p
            QfAt (AtNeq p)
		| signumNeq s -> dtgTrue  $ newDtcT
		| signumEq  s -> dtgFalse $ newDtcF
		| otherwise  ->
		    EqCond p tFalse tTrue
		where
		s = sign p
            qf
		| simplifyIsWeak && impl      -> dtgTrue  $ newDtcT
		| simplifyIsWeak && impl_not  -> dtgFalse $ newDtcF
		| otherwise -> FCond qf tTrue tFalse
		where
	        impl     = dtcImpliesQf dtc (dtcContext dtc) theQf
	        impl_not = dtcImpliesQf dtc (dtcContext dtc) notTheQf
	where
	simpQf   = dtcContextSimplifyQf dtc unsimpQf

        theQf    = if simplifyConds then simpQf else unsimpQf
	notTheQf = qfNot theQf
    
        tTrue  = dtgTrue  $ dtc { dtcContext = theQf    : dtcContext dtc
                                }
        tFalse = dtgFalse $ dtc { dtcContext = notTheQf : dtcContext dtc
                                }

        sign p
	    | simplifyConds && not simplifyIsWeak =
		-- when using a strong simplifier, there is no
		-- point in checking whether the simplified formula
		-- has constant sign under the context.
		allSignumsFalse
	    | otherwise = dtcImpliesSignumS dtc (dtcContext dtc) p

        --
        -- New DTreeContext for the cases that the the condition
        -- is implied (newDtcT) or inconsistent with the context
        -- (newDtcF). There is no need to add the condition (or its
        -- negation, respectively) to the context as the it does
        -- not carry any information not already implied by
        -- the original context.
        --
        condsToAdd = if simplifyConds && simpQf /= QfAt AtTrue
                     then [unsimpQf,simpQf] else [unsimpQf]
        newDtcT = addImpliedConds dtc $ map elim condsToAdd
        newDtcF = addImpliedConds dtc $ map (elim . qfNot) condsToAdd
        elim = eliminateAtSignum (mbOrderingToSignum . dtcIsConstant dtc)

        simplifyConds  = dtcSimplifyConditions dtc
        simplifyIsWeak = dtcSimplifyIsWeak dtc

addImpliedConds :: Num cond
                => DTreeContext c cond
                -> [QfFormula cond]
                -> DTreeContext c cond
addImpliedConds dtc fs =
    dtc { dtcImpliedConds = nub (fs' ++ dtcImpliedConds dtc) }
    where
      fs' = filter (/= QfAt AtTrue) fs

-- |
-- Construct a leaf.
-- 
dLeaf :: val  -- ^ value to represent as a leaf
      -> DTreeGen c cond val
dLeaf val = DTreeGen $ \dtc -> DLeaf (val,dtc)


-- 
-- Syntactically simplyfy qf.
-- Assumes that ctx is consistent.
--
syntSimpl :: Num cond
          => [QfFormula cond]  -- implied conditions
          -> QfFormula cond
          -> QfFormula cond
syntSimpl ctx qf
    | qf       `elem` ctx  =  QfAt AtTrue
    | qfNot qf `elem` ctx  =  QfAt AtFalse
syntSimpl ctx (QfAt (AtEq f))
    | QfAt (AtGt f)          `elem` ctx  =  QfAt AtFalse
    | QfAt (AtGt $ negate f) `elem` ctx  =  QfAt AtFalse
syntSimpl ctx (QfAt (AtNeq f))
    | QfAt (AtGt f)          `elem` ctx  =  QfAt AtTrue
    | QfAt (AtGt $ negate f) `elem` ctx  =  QfAt AtTrue
syntSimpl ctx (QfAt (AtGe f))
    | QfAt (AtGt f)  `elem` ctx           =  QfAt AtTrue
    | QfAt (AtNeq f) `elem` ctx           =  QfAt $ AtGt f
    | QfAt (AtNeq $ negate f) `elem` ctx  =  QfAt $ AtGt f
syntSimpl ctx (QfAt (AtGt f))
    | QfAt (AtGt $ negate f) `elem` ctx   =  QfAt AtFalse
    | QfAt (AtEq $ negate f) `elem` ctx   =  QfAt AtFalse
    | QfAt (AtEq          f) `elem` ctx   =  QfAt AtFalse
syntSimpl ctx qf = qf


-- |
-- Check whether the conjunction of the given formulas implies
-- the given formula.
--
dtcImpliesQf :: DTreeContext c cond 
             -> [QfFormula cond]
             -> QfFormula cond 
             -> Bool
dtcImpliesQf dtc ctx qf = dtcImpliesQfS dtc (ctx ++ dtcContext dtc) qf

-- |
-- Check whether the context together with some additional conditions
-- implies a signum for a certain expression.
--
dtcImpliesSignum :: DTreeContext c cond
		 -> [QfFormula cond]
		 -> c
		 -> Signum
dtcImpliesSignum dtc ctx p =
    dtcImpliesSignumS dtc (ctx ++ dtcContext dtc) (dtcCondGen dtc p)

-- |
-- Check whether the context implies a signum for a given expression.
--
dtcContextImpliesSignum :: DTreeContext c cond
			-> c
			-> Signum
dtcContextImpliesSignum dtc p = dtcImpliesSignum dtc [] p


-- |
-- Simplify a quantifier-free formula under the current context.
--
dtcContextSimplifyQf :: DTreeContext c cond
		     -> QfFormula cond
		     -> QfFormula cond
dtcContextSimplifyQf dtc qf =
    dtcSimplifyQfS dtc (dtcContext dtc) qf


-- |
-- Build a new 'DTreeContext'.
-- Please do not use this function to build a 'DTreeContext' for now.
-- The interface my change!
--
-- Look in LMath for some predefined 'DTreeContext's.
--
newDTreeContext :: [QfFormula cond]
	    -> ([QfFormula cond] -> QfFormula cond -> Bool)
	    -> ([QfFormula cond] -> cond -> Signum)
	    -> ([QfFormula cond] -> QfFormula cond -> QfFormula cond)
	    -> (c -> cond)
            -> (cond -> Maybe Ordering)
	    -> DTreeContext c cond
newDTreeContext ctx implQfS implSignumS simplQfS condGen isConst =
    DTreeContext { dtcContext        = ctx,
	           dtcImpliesQfS     = implQfS,
	           dtcImpliesSignumS = implSignumS,
	           dtcSimplifyQfS    = simplQfS,
	           dtcCondGen        = condGen,
	           dtcSimplifyConditions = True,
	           dtcSimplifyIsWeak     = True,
                   dtcFactoriseS         = \p -> [p]
                 , dtcIsConstant   = isConst
                 , dtcImpliedConds    = []
                 , dtcAlgorithmOptions = []
	         }


{-
nullDTreeContext :: DTreeContext cond cond
nullDTreeContext =
    DTreeContext { dtcSpecialContext = undefined,
	       dtcContext        = [],
	       dtcSimplifyQfS = \sc ctx f -> f,
	       dtcImpliesQfS  = \sc ctx f -> (False, sc),
	       dtcImpliesSignumS = \sc ctx p -> (allSignumsFalse, sc),
	       dtcCondGen = id
	     }
-}


-- |
-- Access to the 'DTreeContext'.
-- 
withDTreeContext :: (DTreeContext c cond -> DTreeGen c cond v)
		 -> DTreeGen c cond v
withDTreeContext f = DTreeGen (\dtc -> let DTreeGen res = f dtc in res dtc)

-- |
-- Apply a 'DTreeGen' to all the leaves of a 'DTree' (to be) generated by
-- another 'DTreeGen', i.e., the leaves of the input tree are
-- replaced by sub-trees generated by @f@.
--
compDTreeG :: (v -> DTreeGen c cond w)  -- ^ function @f@
	   -> DTreeGen c cond v         -- ^ input tree
	   -> DTreeGen c cond w
compDTreeG f (DTreeGen g) =
    DTreeGen (\dtc ->
	      compDTree (\(val,dtc2) -> let DTreeGen res = f val in res dtc2) $
                        g dtc
             )

-- |
-- Compute the \"product\" of a given list of trees.
-- All the (ordered) combinations of leaves from the given trees
-- are computed and each list of leaves is guarded by the
-- respective conditions in the given trees, i.e., if
-- (for an input with two trees) a leaf
-- @l1@ is guarded by @f@ in the first tree, and a leaf @l2@ is
-- guarded by @g@ in the second tree, the leaf
-- @[l1,l2]@ in the result is guarded by @f \/\\ g@.
--
prodDTreeG :: [DTreeGen c cond v] -> DTreeGen c cond [v]
prodDTreeG [] = dLeaf []
prodDTreeG (t:ts) =
    compDTreeG (\xs -> mapDTreeG (:xs) t) $
    prodDTreeG ts

-- |
-- Compute the \"product\" of two trees.
-- The set of leaves of the result is the
-- product of the two sets of leaves of the given trees.
-- Each leaf in the result is guarded by the conditions
-- guarding the two respective leaves in the given trees.
-- For example, if @l1@ is guarded by @f@ in the first tree,
-- and @l2@ is guarded by @g@ in the second tree,
-- the leaf @(l1,l2)@ in the result is guarded by @f \/\\ g@.
--
prod2DTreeG :: DTreeGen c cond v
	    -> DTreeGen c cond w
	    -> DTreeGen c cond (v,w)
prod2DTreeG t1 t2 = compDTreeG (\v -> mapDTreeG (\w -> (v,w)) t2) t1

-- |
-- Apply a function to the leaves of a (to be generated) 'DTree'.
--
mapDTreeG :: (v -> w) -> DTreeGen c cond v -> DTreeGen c cond w
mapDTreeG f (DTreeGen gen) =
    DTreeGen (\dtc -> fmap (\(v,dtc2) -> (f v, dtc2)) $ gen dtc)

-- |
-- Apply a function to the conditions on a (to be generated) 'DTree'.
--
cmapDTreeG :: (cond -> cond) -> DTreeGen c cond v -> DTreeGen c cond v
cmapDTreeG f (DTreeGen gen) = DTreeGen (\dtc -> cmapDTree f $ gen dtc)


-- |
-- Flattens a decision tree, i.e. convert a tree of trees
-- into a simple tree.
--
-- See also: 'flattenDTree'
--
flattenDTreeG :: DTreeGen c cond (DTreeGen c cond val)
              -> DTreeGen c cond val
flattenDTreeG = compDTreeG id

instance Functor (DTreeGen c cond) where
    fmap = mapDTreeG


combineSubtreesDTreeG :: (Num cond, Eq val)
                      => DTreeGen c cond val
                      -> DTreeGen c cond val
combineSubtreesDTreeG =
    combineSubtreesDTreeG' (\a b -> if a == b then Just a else Nothing)

combineSubtreesDTreeG' :: Num cond
                       => (val -> val -> Maybe val)
                       -> DTreeGen c cond val
                       -> DTreeGen c cond val
combineSubtreesDTreeG' mixVals (DTreeGen gen) =
    DTreeGen (\dtc -> combineSubtrees' mixVs $ gen dtc)
    where
    mixVs (v1,dtc1) (v2,dtc2) 
        | Just v <- mixVals v1 v2  =  Just (v, newDtc)
        | otherwise                =  Nothing
        where
        newCtx = conjunctionQf (dtcContext dtc1) `QfOr`
                 conjunctionQf (dtcContext dtc2)
        newCtx' = dtcSimplifyQfS dtc1 [] newCtx
        newDtc = dtc1 { dtcContext = [newCtx'] }


-- |
-- Set an optional value in the context of a 'DTreeGen'.
-- The options can be thought of as (key,value) pairs where
-- the key of a value is its type. That is why the type of 
-- the option has to be a member of the 'Typeable' class.
-- Using the type as key implies that there
-- can be at most one value per type. For example, a DTreeGen
-- can only transport one value of type Int. So, to transport 
-- option values via DTreeGen, create a new datatype for each
-- option.
-- See, for example, the data type FMErbAlgo in LMath.Algo.FM.
--
setDTGOption :: Typeable a => a -> DTreeGen c cond ()
setDTGOption opt =
    updateDTreeContext (\dtc -> dtc { dtcAlgorithmOptions =
                                          toDyn opt : dtcAlgorithmOptions dtc
                                    }
                       )

-- |
-- Look up a DTreeGen option with a default value.
--
lookupDTGOptionWithDefault :: Typeable a => a -> DTreeGen c cond a
lookupDTGOptionWithDefault def =
    do mb <- lookupDTGOption
       case mb of
         Nothing  -> return def
         Just res -> return res

-- |
-- Look up a DTreeGen option value.
--
lookupDTGOption :: Typeable a => DTreeGen c cond (Maybe a)
lookupDTGOption =
    do dtc <- getDTreeContext
       let aos = dtcAlgorithmOptions dtc
           res = [x | ao <- aos, Just x <- [fromDynamic ao]]
       return $ listToMaybe res

-- |
-- Produce a 'DTree' from a 'DTreeContext' and a 'DTreeGen'.
-- 
makeDTree :: DTreeContext c cond -> DTreeGen c cond v -> DTree cond v
makeDTree dtc (DTreeGen gen) = fmap (\(val,_) -> val) $ gen dtc

-- |
-- Produce a 'DTree' from a 'DTreeContext' and a 'DTreeGen'.
-- Turn on experimental simplification of the conditions in
-- the resulting 'DTree'. The boolean parameter tells
-- whether the simplifier used by the 'DTreeContext' is
-- weak, i.e., there is no guarantee that a formula
-- which holds for all parameter values is simplified
-- to true, or not.
--
makeDTreeSimplifyConds :: Bool   -- ^ weak simplifier in use
                   -> DTreeContext c cond
		   -> DTreeGen c cond v
		   -> DTree cond v
makeDTreeSimplifyConds simplifyIsWeak dtc dtg =
    makeDTree (dtc { dtcSimplifyIsWeak=simplifyIsWeak,
		     dtcSimplifyConditions=True }
	      )
              dtg
