-- |
-- Contains functions to merge 'DTree's.
--
module LMath.Types.DTree.Merge
    ( -- * merge functions for two trees
      mergeDTrees, mergeDTrees', mergeDTreesG,

      -- * merge functions for a list of trees
      mergeListOfDTrees, mergeListOfDTrees', mergeListOfDTreeG,

      -- * helper functions
      convertToDTreeGen
    )
where

import LMath.Types.DTree.DTree
import LMath.Types.DTree.DTreeGen

-- |
-- Merges two decision trees @d1@ and @d2@ into one decision tree
-- according to a given merge operation.
--
-- See also: 'mergeListOfDTrees', 'mergeDTreesG'
--
mergeDTrees :: (Show c, Num cond)
	    => DTreeContext c cond  -- ^ the context
	    -> (val1 -> val2 -> val3)  -- ^ the merge operation
	    -> DTree c    val1         -- ^ @d1@
 	    -> DTree c    val2         -- ^ @d2@
 	    -> DTree cond val3         -- ^ the resulting tree
mergeDTrees context op d1 d2 = makeDTree context $ mergeDTrees_Impl1 op d1 d2

mergeDTrees' :: (val1 -> val2 -> val3)
             -> DTree c val1
             -> DTree c val2
             -> DTree c val3
mergeDTrees' op d1 d2 = merge d1
    where
    merge (DLeaf val)               = fmap (op val) d2
    merge (SCond cond tm t0 tp)     = SCond cond (merge tm) (merge t0) (merge tp)
    merge (EqCond cond t0 tpm)      = EqCond cond (merge t0) (merge tpm)
    merge (GeCond leOk cond t0p tm) = GeCond leOk cond (merge t0p) (merge tm)
    merge (FCond qfCond tt tf)      = FCond qfCond (merge tt) (merge tf)


-- |
-- Merge two decision tree generators @d1G@ and @d2G@ into one decision tree
-- generator according to a given merge operation.
--
-- See also: 'mergeListOfDTreeG', 'mergeDTree'
--
mergeDTreesG :: (val1 -> val2 -> val3)  -- ^ the merge operation
             -> DTreeGen c cond val1 -- ^ @d1G@
             -> DTreeGen c cond val2 -- ^ @d2G@
             -> DTreeGen c cond val3 -- ^ @d3G@
mergeDTreesG op d1G d2G = mapDTreeG (uncurry op) (prod2DTreeG d1G d2G)

-- |
-- Merges two decision trees @d1@ and @d2@ into one decision tree generator
-- according to a given merge operation.
--
-- This is a helper function of 'mergeDTrees'.
--
-- See also: 'mergeDTrees', 'mergeDTrees_Impl2'
--
mergeDTrees_Impl1 :: (Show c, Num cond)
	          => (val1 -> val2 -> val3)  -- ^ the merge operation
		  -> DTree c val1            -- ^ @d1@
		  -> DTree c val2            -- ^ @d2@
 		  -> DTreeGen c cond val3 -- ^ the resulting tree
mergeDTrees_Impl1 op d1 d2 = mergeDTrees_Impl2 op d1 d2G
    where
    -- Convert the second DTree @d2@ into a DTreeGen
    d2G = convertToDTreeGen d2

-- |
-- Merges two decision trees @d1@ and @d2@ into one decision tree generator
-- according to a given merge operation.
--
-- This is a helper function of 'mergeDTrees'.
--
-- See also: 'mergeDTrees', 'mergeDTrees_Impl1'
--
mergeDTrees_Impl2 :: (Show c, Num cond)
	          => (val1 -> val2 -> val3)  -- ^ the merge operation
		  -> DTree c val1            -- ^ @d1@
		  -> DTreeGen c cond val2 -- ^ @d2G@
 		  -> DTreeGen c cond val3 -- ^ the resulting tree
mergeDTrees_Impl2 op d1 d2G = merge d1
    where
    -- Recursively descent the first DTree @d1@ and replace each of its
    -- leafs by the tree that you get, if you apply the merge operation
    -- (with the current leaf as the first parameter) on the second tree @d2G@.
    merge (DLeaf val)               = mapDTreeG (op val) d2G
    merge (SCond cond tm t0 tp)     = sCond cond (merge tm) (merge t0) (merge tp)
    merge (EqCond cond t0 tpm)      = eqCond cond (merge t0) (merge tpm)
    merge (GeCond leOk cond t0p tm) = geCond' leOk cond (merge t0p) (merge tm)
    merge (FCond qfCond tt tf)      = fCond qfCond (merge tt) (merge tf)

-- |
-- Merges a non-empty list of decision trees into one decision tree
-- according to a given merge operation.
--
-- See also: 'mergeDTrees'
--
mergeListOfDTrees :: (Show c, Num cond)
                  => DTreeContext c cond -- ^ the tree context
                  -> (val -> val -> val)    -- ^ the merge operation
                  -> [DTree c val]
		  -- ^ a non-empty list of trees that should be merged
                  -> DTree cond val
mergeListOfDTrees _ _ [] = error "mergeListOfDTrees: empty list of DTrees"
mergeListOfDTrees treeContext op xs =
    makeDTree treeContext $ mergeListOfDTreeG op $ map convertToDTreeGen xs

mergeListOfDTrees' :: (val -> val -> val)
                   -> [DTree cond val]
                   -> DTree cond val
mergeListOfDTrees' _  [] = error "mergeListOfDTrees': empty list of DTrees"
mergeListOfDTrees' op xs = foldl1 (mergeDTrees' op) xs

-- |
-- Merges a non-empty list of decision trees into one decision tree
-- according to a given merge operation.
--
-- See also: 'mergeDTreesG'
--
mergeListOfDTreeG :: (val -> val -> val) -- ^ the merge operation
                  -> [DTreeGen a cond val]
		  -- ^ a non-empty list of tree generators that should be merged
                  -> DTreeGen a cond val
mergeListOfDTreeG _ []  = error "mergeListOfDTreeG: empty list of DTrees"
mergeListOfDTreeG op xs = foldl1 (mergeDTreesG op) xs

-- |
-- Reconvert the 'DTree' @d@ into a 'DTreeGen'.
--
convertToDTreeGen :: (Show c, Num cond)
 		  => DTree c val -- ^ @d@
 		  -> DTreeGen c cond val
convertToDTreeGen (DLeaf val) = dLeaf val
convertToDTreeGen (SCond cond tm t0 tp) =
    sCond cond (convertToDTreeGen tm) (convertToDTreeGen t0) (convertToDTreeGen tp)
convertToDTreeGen (EqCond cond t0 tpm) =
    eqCond cond (convertToDTreeGen t0) (convertToDTreeGen tpm)
convertToDTreeGen (GeCond leOk cond t0p tm) =
    geCond' leOk cond (convertToDTreeGen t0p) (convertToDTreeGen tm)
convertToDTreeGen (FCond qfCond tt tf) =
    fCond qfCond (convertToDTreeGen tt) (convertToDTreeGen tf)
