module LMath.Types.Ehrhart
    ( EValue(..), EPartition,
      evParam, epartition, epPartitions,
      evalEValue,
      evAffine, evFromAffine, evParams, evInsertZeroParams,
      countIntegralPoints,
      evalueToCExpression, evaluesWithCondsToCExpression
    )
where

import Data.Ratio
import Data.List
import Debug.Trace

import CppInterface
import LMath.Types.AffineExp
import LMath.Types.IneqSystem
import LMath.Types.LClasses
import LMath.Types.Matrix
import LMath.Types.PeriodicNumber
import LMath.Types.Vector

--
-- The Int argument denotes the (0-based) number of the parameter.
-- The coefficients of a polynomial are given in decreasing exponent order
-- (i.e., leading coefficient first).
--
data EValue a = EConst a
              | EPeriodic Int [EValue a]
              | EPoly     Int [EValue a]
              | ESum [EValue a]     -- temporary until arithmetic ops are implemented
                deriving Show

data EPartition a = EPart [(GIneqSystem a, EValue a)] deriving Show

epartition :: [(GIneqSystem a, EValue a)] -> EPartition a
epartition = EPart

epPartitions :: EPartition a -> [(GIneqSystem a, EValue a)]
epPartitions (EPart xs) = xs

-- |
-- Monomial  p .
--
evParam :: (Num a, Integral b) => b -> EValue a
evParam p = EPoly (fromIntegral p) [EConst 1, EConst 0]

instance Functor EValue where
    fmap f (EConst x)        = EConst $ f x
    fmap f (EPeriodic p evs) = EPeriodic p $ map (fmap f) evs
    fmap f (EPoly p evs)     = EPoly p $ map (fmap f) evs
    fmap f (ESum evs)        = ESum $ map (fmap f) evs

evalEValue :: (Integral b, Real a) => [b] -> EValue a -> b
evalEValue params evalue
    | denominator value == 1  =  fromIntegral $ numerator value
    | otherwise  =
       error "evalEValue: EValue does not evaluate to an integer"
    where
    value = toRational $ ev evalue

    ev (EConst x) = x
    ev (EPeriodic p coeffs) =
        ev $ genericIndex coeffs ((params!!p) `mod` genericLength coeffs)
    ev (EPoly p coeffs) =
        foldl (\acc co -> acc*pv+ev co) 0 coeffs
        where
        pv = fromIntegral (params !! p)
    ev (ESum es) = sum $ map ev es


ratfreeEValue :: (Real a, Integral b) => EValue a -> (EValue b, b)
ratfreeEValue ev' = (fmap mkInt ev, fromIntegral commonDen)
    where
    cs = evalueConsts ev
    commonDen = foldl lcm 1 $ map denominator cs
    mkInt r = fromIntegral (numerator r * (commonDen `div` denominator r))
    ev = fmap toRational ev'

evalueConsts :: EValue a -> [a]
evalueConsts (EConst x) = [x]
evalueConsts (EPeriodic _ evs) = concatMap evalueConsts evs
evalueConsts (EPoly _ evs)     = concatMap evalueConsts evs
evalueConsts (ESum    evs)     = concatMap evalueConsts evs


evAffine :: (Integral b, Num a) => b -> EValue a -> Maybe (GVector a)
evAffine nP ev = aff ev
    where
    aff (EConst c) = Just $ vector (genericReplicate nP 0 ++ [c])
    aff (ESum [])  = aff (EConst 0)
    aff (ESum xs)  = do vs <- mapM aff xs
                        return $ foldl1 (+) vs
    aff (EPeriodic _ [x]) = aff x
    aff (EPeriodic _ _  ) = Nothing
    aff (EPoly _ [x])   = aff x
    aff (EPoly p [x,y]) =
        case (aff x, aff y) of
           (Just vx, Just vy)
              | all (==0) (init $ vectorEntries vx) ->
                  Just (last (vectorEntries vx) `mulSV` vp + vy)
           _ -> Nothing
        where
        vp = vector (genericReplicate p 0 ++ 1 : genericReplicate (fromIntegral nP-p) 0)
    aff (EPoly _ _) = Nothing

evFromAffine :: Num a => GVector a -> EValue a
evFromAffine v =
    foldr mkPoly (EConst $ last coeffs) $ zip (init coeffs) [0..]
    where
    mkPoly (0, p) acc = acc
    mkPoly (co,p) acc = EPoly p [EConst co, acc]
    coeffs = vectorEntries v

evParams :: Integral b => EValue a -> b
evParams (EConst _) = 0
evParams (ESum []) = 0
evParams (ESum xs) = maximum $ map evParams xs
evParams (EPoly     p xs) = maximum (fromIntegral (p+1) : map evParams xs)
evParams (EPeriodic p xs) = maximum (fromIntegral (p+1) : map evParams xs)

evInsertZeroParams :: Integral b
  => b   -- ^ where to insert (first parameter to add)
  -> b   -- ^ number of parameters to insert
  -> EValue a -> EValue a
evInsertZeroParams n' count' ev = ins ev
    where
    ins c@(EConst _) = c
    ins (ESum xs) = ESum $ map ins xs
    ins (EPoly p xs) = EPoly p' $ map ins xs
        where
        p' = if p>=n then p+count else p
    ins (EPeriodic p xs) = EPeriodic p' $ map ins xs
        where
        p' = if p>=n then p+count else p

    n     = fromIntegral n'
    count = fromIntegral count'

-- |
-- Write Ehrhart polynomials together with their guarding conditions
-- (obtained, e.g., form 'countIntegralPoints') as a valid
-- C/C++/Java expression which evaluates the conditions and
-- computes the value of the respective Ehrhart polynomial.
--
evaluesWithCondsToCExpression
    :: Integral a
    => [String]                         -- ^ names of the parameters
    -> EPartition (Ratio a)             -- ^ conditions and Ehrhart polynomials
    -> String
evaluesWithCondsToCExpression _ (EPart []) = "0"
evaluesWithCondsToCExpression paramNames (EPart cevs) =
    concat ['(' : ineqs2c c ++ ") ? (" ++ ev2c ev ++ "):" | (c,ev) <- cevs] ++
    "0"
    where
    ev2c :: Integral a => EValue (Ratio a) -> String
    ev2c = evalueToCExpression paramNames

    ineqs2c :: Integral a => GIneqSystem (Ratio a) -> String
    ineqs2c sys
        | nRows sys == 0  =  "(1 == 1)"
        | otherwise  =
            concat $
            intersperse "&&" 
             ((map (row2c show paramNames ">=") $ rows' $ getMatrix $ is)
              ++
              (map (row2c show paramNames "==") $ rows' $ getMatrix $ es)
             )
        where
        (is,es) = findEqualities $ integralIneqSystem sys
    row2c :: Integral a => (a -> String) -> [String] -> String -> [a] -> String
    row2c showFn _ op [c] = op ++ show (negate c)
    row2c showFn (pn:pns) op (0:cs) = row2c showFn pns op cs
    row2c showFn (pn:pns) op (c:cs) = showFn c ++ "*" ++ pn ++ row2c show' pns op cs
    row2c _ [] _ _ =
        error "evaluesWithCondsToCExpression: not enough parameter names"

    show' :: Integral a => a -> String
    show' x
        | x < 0      =  show x
        | otherwise  =  '+' : show x

-- |
-- Write an Ehrhart polynomial as a valid C/C++/Java expression.
--
evalueToCExpression
    :: Real a
    => [String]      -- ^ names of the parameters
    -> EValue a      -- ^ Ehrhart polynomial to write as expression
    -> String
evalueToCExpression paramNames ev
    | denom == 1  =  expr
    | otherwise   =  concat ["(", expr, ")/", show denom]
    where
    expr = evalueToExpr' paramNames ev'
    (ev',denom) = ratfreeEValue ev

evalueToExpr' :: Integral a => [String] -> EValue a -> String
evalueToExpr' paramNames (EConst x)
    | x < 0      =  concat ["(", show x, ")"]
    | otherwise  =  show x
evalueToExpr' paramNames(EPeriodic p evs)
    | p >= length paramNames =
        error "evalueToExpr': not enough parameter names"
    | otherwise =
        concat ["(" ++ modExpr ++ show n ++ "?" ++ e ++ ":"
                    | e <- exprs | n <- [1..]] ++
        expr0 ++ replicate (length evs - 1) ')'
    where
    modExpr = (paramNames !! p) ++ "%" ++ show (length evs) ++ "=="
    expr0 : exprs = [evalueToExpr' paramNames ev | ev <- evs]
evalueToExpr' paramNames (EPoly p evs)
    | p >= length paramNames =
        error "evalueToExpr': not enough parameter names"
    | otherwise =
        replicate (length evs - 1) '(' ++
        concat (intersperse (")*" ++ (paramNames !! p) ++ "+")
                            [evalueToExpr' paramNames ev | ev <- evs]
               )
evalueToExpr' paramNames (ESum [])  = "0"
evalueToExpr' paramNames (ESum evs) =
    concat $
    intersperse "+" ["(" ++ evalueToExpr' paramNames ev ++ ")" | ev <- evs]
    
    
-- |
-- Count the number of integral points in the polytope
-- described by the given inequality system. It is an error
-- if the given inequality system describes an unbounded
-- polyhedron instead of a polytope.
-- Returns a list of pairs @(c,ep)@ where @c@ is a condition
-- condition on the parameters and @ep@ an Ehrhart polynomial
-- which represents the number of integral points in the
-- polytope if @c@ holds.
--
{-
countIntegralPoints
    :: RealFrac a
    => GIneqSystem a         -- ^ polytope to count
    -> Maybe (GIneqSystem a) -- ^ optional context (conditions on the parameters only)
    -> [(GIneqSystem a, EValue a)]
countIntegralPoints sys Nothing = countIntegralPoints' [sys] ctx
    where
    ctx = ineqSystem ((0,nParams sys), zeroMatrix 0 (nParams sys+1))
countIntegralPoints sys (Just ctx) = countIntegralPoints' [sys] ctx
-}

-- |
-- Count the number of integral points in the union of polytopes
-- described by the given inequality systems. It is an error
-- if the given inequality systems describe an unbounded
-- domain.
-- Returns a list of pairs @(c,ep)@ where @c@ is a condition
-- condition on the parameters and @ep@ an Ehrhart polynomial
-- which represents the number of integral points in the
-- union if @c@ holds.
--
countIntegralPoints
    :: RealFrac a
    => [GIneqSystem a]       -- ^ union of polytopes to count
    -> Maybe (GIneqSystem a) -- ^ optional context (conditions on the parameters only)
    -> EPartition a
countIntegralPoints []   _       = EPart []
countIntegralPoints syss Nothing = EPart $ countIntegralPoints' syss ctx
    where
    ctx = ineqSystem ((0,nP), zeroMatrix 0 (nP+1))
    nP = nParams $ head syss
countIntegralPoints syss (Just ctx) = EPart $ countIntegralPoints' syss ctx


foreign import ccall countIntegralPoints_c
    :: Ptr PolylibDomain        -- inequalities
    -> Ptr PolylibPolyhedron    -- context, may not contain variables
    -> IO (Ptr (CPair (Ptr PolylibDomain) (Ptr PolylibEValue)))

countIntegralPoints'
    :: RealFrac a
    => [GIneqSystem a]  -- ^ polytopes to count
    -> GIneqSystem a  -- ^ optional context (conditions on the parameters only)
    -> [(GIneqSystem a, EValue a)]
countIntegralPoints' []   _   =  []
countIntegralPoints' syss (ctx :: GIneqSystem a)
    | length (nub $ map nVars syss) /= 1  =
        error "countIntegralPoints: systems must have the same number of variables"
    | length (nub $ map nParams syss) /= 1  =
        error "countIntegralPoints: systems must have the same number of parameters"
    | nP /= nParams ctx =
        error "countIntegralPoints: number of parameters does not match"
    | nVars ctx /= 0  =
        error "countIntegralPoints: context must not contain variables"
    | otherwise  = 
        trace ("countIntegralPoints: " ++ show syss ++ "ctx=" ++ show ctx) $
        [(ineqSystem ((0,nP),m), ev) | (ms,ev) <- res, m <- ms]
    where
    res :: [([GMatrix a], EValue a)]
    res = unsafeCppiConvert2 countIntegralPoints_c syss ctx
    nP = nParams $ head syss


data PolylibEValue

instance RealFrac a => CConv (EValue a) (Ptr PolylibEValue) where
    toC _ _ = error "toC: undefined for EValue"

    -- Note: evalue structs are not allocated on the C side
    -- (they are part of the Enumeration structure), so we
    -- do not free them explicitly.
    fromC evptr =
        do ev <- convertEValue evptr
           evalueFree_c evptr
           return $ fmap fromRational ev
        where
        convertEValue ptr =
            do d <- evalueGetD ptr
               if d == 0
                 then
                   do p <- evalueGetP ptr
                      valPtrs <- enodeGetValues p
                      vals <- mapM convertEValue valPtrs
                      periodic <- enodeIsPeriodic p
                      case vals of
                        -- a single coefficient (= constant) or a
                        -- a periodic number with a single element list
                        -- (i.e. "mod 1") is better represented as the
                        -- single value than a polynomial or perodic.
                        [v] -> return v
                        _   ->
                          if periodic
                             then do paramPos <- enodeGetParameterPos p
                                     return $ EPeriodic (paramPos-1) vals
                             else do paramPos <- enodeGetParameterPos p
                                     return $ EPoly (paramPos-1) $ reverse vals
                 else
                   do n <- evalueGetN ptr
                      return $ EConst (n % d)

data ENode  -- only for  Ptr ENode

foreign import ccall evalueFree_c :: Ptr PolylibEValue -> IO ()

foreign import ccall evalueGetD_c :: Ptr PolylibEValue -> IO (Ptr PolylibValue)
evalueGetD :: Ptr PolylibEValue -> IO Integer
evalueGetD ptr = evalueGetD_c ptr >>= fromC

foreign import ccall evalueGetN_c :: Ptr PolylibEValue -> IO (Ptr PolylibValue)
evalueGetN :: Ptr PolylibEValue -> IO Integer
evalueGetN ptr = evalueGetN_c ptr >>= fromC

foreign import ccall evalueGetP_c :: Ptr PolylibEValue -> IO (Ptr ENode)
evalueGetP :: Ptr PolylibEValue -> IO (Ptr ENode)
evalueGetP = evalueGetP_c

-- get enode->arr[..] _without_ calling 'free' on the evalues
-- (i.e., return [Ptr PolylibEValue], not [EValue])
foreign import ccall enodeGetValues_c :: Ptr ENode -> IO (Ptr (Ptr PolylibEValue))
enodeGetValues :: Ptr ENode -> IO [Ptr PolylibEValue]
enodeGetValues ptr = enodeGetValues_c ptr >>= fromC

foreign import ccall enodeGetParameterPos_c :: Ptr ENode -> IO CInt
enodeGetParameterPos :: Ptr ENode -> IO Int
enodeGetParameterPos ptr = enodeGetParameterPos_c ptr >>= fromC

foreign import ccall enodeIsPeriodic_c :: Ptr ENode -> IO CBool
enodeIsPeriodic :: Ptr ENode -> IO Bool
enodeIsPeriodic ptr = enodeIsPeriodic_c ptr >>= fromC
