---
-- High level functions for equation systems.
-- @author Armin Gr&ouml;&szlig;linger
-- @version $Id: EqSystem.hs 4188 2009-08-17 16:42:26Z groessli $
--
module LMath.Types.EqSystem
    ( GEqSystem, EqSystem, eqSystem, eqSystemVPC,
      numberOfEquations, mirrorEquations,
      integralEqSystem
    )
where

import Data.List
import Data.Ratio

import CppInterface
import LMath.Types.LClasses
import LP
import LMath.Types.AffineExp
import LMath.Types.Matrix

-- |
-- Data type for representing inequality systems.
--
newtype GEqSystem ent = ES { affineE :: GAffineExp ent }
type EqSystem = GEqSystem Rational

instance Functor GEqSystem where
    fmap f es = es { affineE = fmap f (affineE es) }


-- |
-- Constructs an equality system.
--
eqSystem :: (Integral a, Integral b)
	 => ((a,b), GMatrix entries)   -- ^  the parts to construct the equation system from.
         -> GEqSystem entries
eqSystem arg =  ES { affineE = affineExp arg }


eqSystemVPC = vpcVersion "eqSystemVPC" eqSystem

-- return the number of equations in the given equation system
numberOfEquations :: GEqSystem ent -> Integer
numberOfEquations = nRows

-- |
-- Mirror an affine sub-space at the origin.
--
mirrorEquations :: Num a
		=> GEqSystem a
		-> GEqSystem a
mirrorEquations eqs = eqSystemVPC (negate v, p, c)
    where
    (v, p, c) = partsVPC eqs


instance AffineExpBased (GEqSystem ent) ent where
    toAffineExp      = affineE
    fromAffineExp ae = ES { affineE=ae }

instance MatrixBased (GEqSystem ent) ent where
    setEntry is row col value = is { affineE = setEntry (affineE is)
				     row col value }
    getMatrix is = getMatrix (affineE is)

instance Joinable (GEqSystem ent) where
    join is1 is2 = ES { affineE = (affineE is1) `join` (affineE is2) }
    unjoin is = [eqSystem ((nVars is, nParams is), matr) | matr <- rows is]

instance (Num ent, Ord ent) => Applicable (GEqSystem ent) [ent] [Bool] where
    is `apply` rs = map (>=0) $ (affineE is `apply` rs :: [ent])

instance (Num ent, Ord ent) =>
    Applicable (GEqSystem ent) (GMatrix ent) [Bool] where
    is `apply` m = map (>=0) $ entries ((affineE is `apply`  m) :: GMatrix ent)

instance Show ent => Show (GEqSystem ent) where
    show is = "Equality system in " ++ show (nVars is) ++ " variables and "++
	      show (nParams is) ++ " parameters\n" ++
	      show (getMatrix (affineE is))

-- entfernt alle Brüche aus den Gleichungen in is durch 
-- Erweitern der Gleichung
integralEqSystem :: (Real a, Integral b) => GEqSystem a -> GEqSystem b
integralEqSystem es = fromAffineExp $ integralAffineExp $ toAffineExp es

-------

--
-- Achtung: Die C Seite kein kein Eq_system mit Parametern !!!

foreign import ccall newFEqSys_c    :: Ptr Matrix   -> IO (Ptr EqSystem)
foreign import ccall deleteFEqSys_c :: Ptr EqSystem -> IO ()
foreign import ccall matrixFEqSys_c :: Ptr EqSystem -> IO (Ptr Matrix)

instance CConv EqSystem (Ptr EqSystem) where
    toC sys action =
	do fes <- cppiArgs1 newFEqSys_c (getMatrix sys)
	   res <- action fes
	   deleteFEqSys_c fes
	   return res
    fromC fes =
	do matr   <- matrixFEqSys_c fes >>= fromC
	   deleteFEqSys_c fes
	   return $ eqSystem ((nColumns matr - 1, 0), matr :: Matrix)


