module LMath.Types.Formula
    ( -- * Atomic Formulas
      AtFormula(..), 
      notAt, evalAt, evalAtMb, evalAtSignum,
      indeterminatesInAt, termsAt,
      eqAtFormula,

      -- * Quantifier-free Formulas
      QfFormula(..), 
      qfNot, qfImpl, qfOr, qfAnd,
      positiveQf, positiveDnfQf, conjunctionQf, disjunctionQf,
      qfMapAt, qfMapAtQf, evalQf, evalQfMb, evalQfSignum,
      indeterminatesInQf, termsQf, extractAt,
      isConjunctionQf, conjunctsQf, disjunctsQf,
      levelQf, invlevelQf,
      eliminateAtTF, eliminateAtSignum,
      genericImpliesSignum,
      eqQfFormula,

      -- * First-order Formulas
      FoFormula(..),
      indeterminatesInFo,
      foEx,
      matrixFo, eliminateAtSignumQ,

      -- * Printing Functions
      atFormulaToC, qfFormulaToC,

      -- * Parsing Formulas
      parseAtFormula, parseQfFormula,

      -- * Conversion between Inequality Systems and Formulas
      ineqSystemToFormula, ineqSystemToStrictFormula,
      ineqSystemToAtFormulas,
      eqSystemToFormula, eqSystemToAtFormulas,
      formulaToIneqSystems, atFormulasToIneqSystems,
    )
where

import Control.Monad
import Data.List as List ( nubBy, span, union )
import Data.Maybe
import Data.Ratio
import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Expr

import LMath.Types.LClasses
import LMath.Types.Matrix
import LMath.Types.AffineExp
import LMath.Types.EqSystem
import LMath.Types.IneqSystem
import LMath.Types.Polynomial
import LMath.Types.Signum

-- |
-- Atomic formulas: relational expression with =, \/=, >, >=, and
-- constants True, False.
--
data AtFormula a = AtEq  a
		 | AtNeq a
		 | AtGt  a
		 | AtGe  a
		 | AtTrue
		 | AtFalse
		   deriving (Eq, Read, Show)

notAt :: Num a => AtFormula a -> AtFormula a
notAt (AtEq  x) = AtNeq x
notAt (AtNeq x) = AtEq x
notAt (AtGt  x) = AtGe (negate x)
notAt (AtGe  x) = AtGt (negate x)
notAt AtTrue    = AtFalse
notAt AtFalse   = AtTrue

evalAt :: (a -> Ordering) -> AtFormula a -> Bool
evalAt _  AtTrue     =  True
evalAt _  AtFalse    =  False
evalAt ev (AtGe  x)  =  ev x >= EQ
evalAt ev (AtGt  x)  =  ev x == GT
evalAt ev (AtEq  x)  =  ev x == EQ
evalAt ev (AtNeq x)  =  ev x /= EQ

evalAtMb :: (a -> Maybe Ordering) -> AtFormula a -> Maybe Bool
evalAtMb _  AtTrue     =  Just True
evalAtMb _  AtFalse    =  Just False
evalAtMb ev (AtGe  x)  =  liftM (>= EQ) $ ev x
evalAtMb ev (AtGt  x)  =  liftM (== GT) $ ev x 
evalAtMb ev (AtEq  x)  =  liftM (== EQ) $ ev x
evalAtMb ev (AtNeq x)  =  liftM (/= EQ) $ ev x

evalAtSignum :: (a -> Signum) -> AtFormula a -> Maybe Bool
evalAtSignum _  AtTrue      =  Just True
evalAtSignum _  AtFalse     =  Just False
evalAtSignum sgn (AtGe  x)  =  sgnCheck sgn (signumGe,signumLt) x
evalAtSignum sgn (AtGt  x)  =  sgnCheck sgn (signumGt,signumLe) x
evalAtSignum sgn (AtEq  x)  =  sgnCheck sgn (signumEq,signumNeq) x
evalAtSignum sgn (AtNeq x)  =  sgnCheck sgn (signumNeq,signumEq) x

sgnCheck :: (a -> Signum) -> (Signum -> Bool, Signum -> Bool) -> a -> Maybe Bool
sgnCheck sgn (f,g) x = case (f s, g s) of
		        (True, False) -> Just True
		        (False, True) -> Just False
		        _             -> Nothing
    where
    s = sgn x

termsAt :: AtFormula a -> [a]
termsAt AtTrue    = []
termsAt AtFalse   = []
termsAt (AtEq  t) = [t]
termsAt (AtNeq t) = [t]
termsAt (AtGe  t) = [t]
termsAt (AtGt  t) = [t]

instance Functor AtFormula where
    fmap f (AtEq  x) = AtEq  (f x)
    fmap f (AtNeq x) = AtNeq (f x)
    fmap f (AtGt  x) = AtGt  (f x)
    fmap f (AtGe  x) = AtGe  (f x)
    fmap _ AtTrue    = AtTrue
    fmap _ AtFalse   = AtFalse

data QfFormula a = QfAt  (AtFormula a)
		 | QfNot (QfFormula a)
		 | QfAnd (QfFormula a) (QfFormula a)
		 | QfOr  (QfFormula a) (QfFormula a)
		   deriving (Eq, Read, Show)

instance Functor QfFormula where
    fmap f (QfAt a)  = QfAt (fmap f a)
    fmap f (QfNot q) = QfNot (fmap f q)
    fmap f (QfAnd g h) = QfAnd (fmap f g) (fmap f h)
    fmap f (QfOr  g h) = QfOr  (fmap f g) (fmap f h)


-- |
-- Make a conjunction of two formulas.
-- The result may be (trivially) simplified.
--
qfAnd :: QfFormula a -> QfFormula a -> QfFormula a
qfAnd qf1 qf2 = eliminateAtTF $ QfAnd qf1 qf2

-- |
-- Make a disjunction of two formulas.
-- The result may be (trivially) simplified.
--
qfOr :: QfFormula a -> QfFormula a -> QfFormula a
qfOr qf1 qf2 = eliminateAtTF $ QfOr qf1 qf2

-- |
-- Make an implication between two formulas.
-- The result may be (trivially) simplified.
--
qfImpl :: QfFormula a -> QfFormula a -> QfFormula a
qfImpl a b = eliminateAtTF (b `QfOr` QfNot a)


conjunctionQf :: [QfFormula a] -> QfFormula a
conjunctionQf []     = QfAt AtTrue
conjunctionQf [f]    = f
conjunctionQf (f:fs) = foldl QfAnd f fs

disjunctionQf :: [QfFormula a] -> QfFormula a
disjunctionQf []     = QfAt AtFalse
disjunctionQf [f]    = f
disjunctionQf (f:fs) = foldl QfOr f fs

qfMapAt :: (AtFormula a -> AtFormula b) -> QfFormula a -> QfFormula b
qfMapAt f (QfAt a)  = QfAt (f a)
qfMapAt f (QfNot g) = QfNot (qfMapAt f g)
qfMapAt f (QfAnd g h) = QfAnd (qfMapAt f g) (qfMapAt f h)
qfMapAt f (QfOr  g h) = QfOr  (qfMapAt f g) (qfMapAt f h)

qfMapAtQf :: (AtFormula a -> QfFormula b) -> QfFormula a -> QfFormula b
qfMapAtQf f (QfAt a)  = f a
qfMapAtQf f (QfNot g) = QfNot $ qfMapAtQf f g
qfMapAtQf f (QfAnd g h) = QfAnd (qfMapAtQf f g) (qfMapAtQf f h)
qfMapAtQf f (QfOr  g h) = QfOr  (qfMapAtQf f g) (qfMapAtQf f h)

termsQf :: QfFormula a -> [a]
termsQf (QfAt  f) = termsAt f
termsQf (QfNot f) = termsQf f
termsQf (QfAnd f g) = termsQf f ++ termsQf g
termsQf (QfOr  f g) = termsQf f ++ termsQf g

isConjunctionQf :: QfFormula a -> Bool
isConjunctionQf (QfAnd f g) = isConjunctionQf f && isConjunctionQf g
isConjunctionQf (QfAt  _  ) = True
isConjunctionQf _           = False

conjunctsQf :: QfFormula a -> [QfFormula a]
conjunctsQf (QfAt AtTrue) = []
conjunctsQf (QfAnd f g) = conjunctsQf f ++ conjunctsQf g
conjunctsQf f           = [f]

disjunctsQf :: QfFormula a -> [QfFormula a]
disjunctsQf (QfAt AtFalse) = []
disjunctsQf (QfOr f g) = disjunctsQf f ++ disjunctsQf g
disjunctsQf f          = [f]

-- |
-- Eliminate @AtTrue@ and @AtFalse@ from a
-- quantifier-free formula.
-- The resulting formula is equivalent to the
-- given formula. The result is either a
-- formula without
-- @AtTrue@ and @AtFalse@, or @QfAt AtTrue@,
-- or @QfAt AtFalse@.
--
eliminateAtTF :: QfFormula a -> QfFormula a
eliminateAtTF f@(QfAt _) = f
eliminateAtTF (QfAnd f g) =
    case (eliminateAtTF f, eliminateAtTF g) of
      (QfAt AtFalse, _) -> QfAt AtFalse
      (_, QfAt AtFalse) -> QfAt AtFalse
      (QfAt AtTrue, h)  -> h
      (h, QfAt AtTrue)  -> h
      (k, l)            -> QfAnd k l
eliminateAtTF (QfOr f g) =
    case (eliminateAtTF f, eliminateAtTF g) of
      (QfAt AtTrue, _)  -> QfAt AtTrue
      (_, QfAt AtTrue)  -> QfAt AtTrue
      (QfAt AtFalse, h) -> h
      (h, QfAt AtFalse) -> h
      (k, l)            -> QfOr k l
eliminateAtTF (QfNot f) =
    case eliminateAtTF f of
      QfAt AtTrue  -> QfAt AtFalse
      QfAt AtFalse -> QfAt AtTrue
      h            -> QfNot h

-- |
-- Eliminate @AtTrue@ and @AtFalse@ from a
-- quantifier-free formula and all atomic formulas
-- which can be decided using @sgn@.
-- The resulting formula is equivalent to the
-- given formula. The result is either a
-- formula without
-- @AtTrue@, @AtFalse@ and atomic formulas
-- with known signum (wrt. @sgn@), or @QfAt AtTrue@,
-- or @QfAt AtFalse@.
--
eliminateAtSignum :: (a -> Signum)  -- ^ signum function @sgn@
	          -> QfFormula a    -- ^ input formula
	          -> QfFormula a
eliminateAtSignum sgn = eac
    where
    eac (QfAt atF) =
	case evalAtSignum sgn atF of
	   Nothing    -> QfAt atF
	   Just True  -> QfAt AtTrue
	   Just False -> QfAt AtFalse
    eac (QfAnd f g) =
	case (eac f, eac g) of
	  (QfAt AtFalse, _) -> QfAt AtFalse
	  (_, QfAt AtFalse) -> QfAt AtFalse
	  (QfAt AtTrue, h)  -> h
	  (h, QfAt AtTrue)  -> h
	  (k, l)            -> QfAnd k l
    eac (QfOr f g) =
	case (eac f, eac g) of
	  (QfAt AtTrue, _)  -> QfAt AtTrue
	  (_, QfAt AtTrue)  -> QfAt AtTrue
	  (QfAt AtFalse, h) -> h
	  (h, QfAt AtFalse) -> h
	  (k, l)            -> QfOr k l
    eac (QfNot f) =
	case eac f of
	  QfAt AtTrue  -> QfAt AtFalse
	  QfAt AtFalse -> QfAt AtTrue
	  h            -> QfNot h


eqAtFormula :: (a -> a -> Bool) -> AtFormula a -> AtFormula a -> Bool
eqAtFormula eqExpr = eq
    where
    eq (AtEq a)  (AtEq b)  = a `eqExpr` b
    eq (AtNeq a) (AtNeq b) = a `eqExpr` b
    eq (AtGe a)  (AtGe b)  = a `eqExpr` b
    eq (AtGt a)  (AtGt b)  = a `eqExpr` b
    eq AtTrue    AtTrue    = True
    eq AtFalse   AtFalse   = True
    eq _         _         = False

eqQfFormula :: (a -> a -> Bool) -> QfFormula a -> QfFormula a -> Bool
eqQfFormula eqExpr = eq
    where
    eq (QfAnd f1 g1) (QfAnd f2 g2) = eq f1 f2 && eq g1 g2
    eq (QfOr  f1 g1) (QfOr  f2 g2) = eq f1 f2 && eq g1 g2
    eq (QfNot f1)    (QfNot f2)    = eq f1 f2
    eq (QfAt  f1)    (QfAt  f2)    = eqAtFormula eqExpr f1 f2
    eq _             _             = False

-- |
-- Prenex first-order formulas.
--
data FoFormula a = FoQf  (QfFormula a)
		 | FoAll [Integer] (FoFormula a)
		 | FoEx  [Integer] (FoFormula a)
		   deriving (Eq, Show)

instance Functor FoFormula where
    fmap f (FoQf qf)     = FoQf (fmap f qf)
    fmap f (FoAll vs fo) = FoAll vs (fmap f fo)
    fmap f (FoEx  vs fo) = FoEx  vs (fmap f fo)


-- |
-- Return the quantification information and the contained
-- quantifier-free formula of a first-order formula.
-- If a variable is quantified more than once, only the
-- innermost quantification is reported.
--
matrixFo :: FoFormula a -> ([(Integer,Bool)], QfFormula a)
matrixFo fo = (uniqQuants , qf)
    where
    -- If a variable is quantified more than once, only
    -- the innermost quantifier has an effect; therefore, we
    -- remove all but the last occurrence of each variable.
    uniqQuants = reverse $ nubBy (\(v1,_) (v2,_) -> v1 == v2) $ reverse quants

    (quants, qf) = mfo [] fo

    mfo qs (FoQf f)     = (qs, f)
    mfo qs (FoAll vs f) = mfo (qs ++ [(v,True)  | v <- vs]) f
    mfo qs (FoEx  vs f) = mfo (qs ++ [(v,False) | v <- vs]) f

-- |
-- Eliminate True, False and constant terms from a quantified formula
-- (see 'eliminateAtSignum') and simplify the quantification
-- structure by removing quantifications for variables which do
-- not occur in the formula.
-- 
eliminateAtSignumQ :: Integral b
                   => (a -> [b])     -- ^ used indeterminates
                   -> (a -> Signum)  -- ^ sign of term
                   -> FoFormula a
                   -> FoFormula a
eliminateAtSignumQ getUsedIndets sgnFn fo =
    mkQuants newQuants $ FoQf newQf
    where
    (quants, qf) = matrixFo fo
    newQf   = eliminateAtSignum sgnFn qf
    used    = fmap fromIntegral $
              foldl List.union [] $
              map getUsedIndets $
              termsQf newQf
    newQuants = [q | q@(v,_) <- quants, v `elem` used]

    mkQuants [] = id
    mkQuants qs
        | null allq  =  FoEx  (map fst exq ) . mkQuants rest2
        | otherwise  =  FoAll (map fst allq) . mkQuants rest1
        where
        (allq, rest1) = span (\(_,typ) -> typ) qs
        (exq,  rest2) = span (\(_,typ) -> not typ) rest1

    exQuant   = case [v | (v,False) <- newQuants] of
                  [] -> id
                  vs -> FoEx vs
    unQuant   = case [v | (v,True ) <- newQuants] of
                  [] -> id
                  vs -> FoAll vs


positiveQf :: Num a => QfFormula a -> QfFormula a
positiveQf a@(QfAt _)           = a
positiveQf (QfNot (QfAt a))     = QfAt $ notAt a
positiveQf (QfNot (QfNot f))    = positiveQf f
positiveQf (QfNot (QfAnd f g))  =
    QfOr  (positiveQf $ QfNot f) (positiveQf $ QfNot g)
positiveQf (QfNot (QfOr  f g))  =
    QfAnd (positiveQf $ QfNot f) (positiveQf $ QfNot g)
positiveQf (QfAnd f g) = QfAnd (positiveQf f) (positiveQf g)
positiveQf (QfOr  f g) = QfOr  (positiveQf f) (positiveQf g)


positiveDnfQf :: Num a => QfFormula a -> [[AtFormula a]]
positiveDnfQf (QfAt a)    = [[a]]
positiveDnfQf f@(QfNot _) = positiveDnfQf (positiveQf f)
positiveDnfQf (QfOr f g)  = positiveDnfQf f ++ positiveDnfQf g
positiveDnfQf (QfAnd f g) =
    [xs++ys | xs <- positiveDnfQf f, ys <- positiveDnfQf g]

-- |
-- Negates the formula @f@. The result is equivalent to @QfNot f@,
-- but it guarantees that atomic formulas (i.e., formulas which start
-- with @QfAt@) remain atomic, e.g., @qfNot (QfAt AtTrue) == QfAt AtFalse@,
-- @qfNot (QfAt AtFalse) == QfAt AtTrue@,
-- @qfNot (QfAt (AtGe x)) == QfAt (AtGt (negate x))@
--
qfNot :: Num a
      => QfFormula a -- ^ @f@
      -> QfFormula a -- ^ the negation of @f@
qfNot (QfAt AtTrue)  = QfAt AtFalse
qfNot (QfAt AtFalse) = QfAt AtTrue
qfNot (QfAt (AtEq  t)) = QfAt $ AtNeq t
qfNot (QfAt (AtNeq t)) = QfAt $ AtEq t
qfNot (QfAt (AtGe  t)) = QfAt $ AtGt $ negate t
qfNot (QfAt (AtGt  t)) = QfAt $ AtGe $ negate t
qfNot f              = QfNot f

evalQf :: (a -> Ordering) -> QfFormula a -> Bool
evalQf ev (QfAt a)     =  evalAt ev a
evalQf ev (QfNot f)    =  not $ evalQf ev f
evalQf ev (QfAnd f g)  =  evalQf ev f && evalQf ev g
evalQf ev (QfOr  f g)  =  evalQf ev f || evalQf ev g

evalQfMb :: (a -> Maybe Ordering) -> QfFormula a -> Maybe Bool
evalQfMb ev (QfAt a)     =  evalAtMb ev a
evalQfMb ev (QfNot f)    =  liftM not $ evalQfMb ev f
evalQfMb ev (QfAnd f g)  =  case (evalQfMb ev f, evalQfMb ev g) of
                              (Just False, _)       -> Just False
                              (Just True,  mbY)     -> mbY
                              (Nothing, Just False) -> Just False
			      _                     -> Nothing
evalQfMb ev (QfOr  f g)  =  case (evalQfMb ev f, evalQfMb ev g) of
                              (Just True,  _)      -> Just True
                              (Just False, mbY)    -> mbY
                              (Nothing, Just True) -> Just True
			      _                    -> Nothing

evalQfSignum :: (a -> Signum) -> QfFormula a -> Maybe Bool
evalQfSignum sgn (QfAt a)  = evalAtSignum sgn a
evalQfSignum sgn (QfNot f)    =  liftM not $ evalQfSignum sgn f
evalQfSignum sgn (QfAnd f g)  =
    case (evalQfSignum sgn f, evalQfSignum sgn g) of
	     (Just x, Just y) -> Just (x && y)
	     _                -> Nothing
evalQfSignum sgn (QfOr  f g)  =
    case (evalQfSignum sgn f, evalQfSignum sgn g) of
	     (Just x, Just y) -> Just (x || y)
	     _                -> Nothing

genericImpliesSignum :: Num term
                     => ([QfFormula term] ->
                         QfFormula term   ->
                         Bool)
                     -> [QfFormula term]
                     -> term
                     -> Signum
genericImpliesSignum impliesQf ctx p =
    Signum { signumEq  = impl (AtEq  p),
             signumNeq = impl (AtNeq p),
             signumLt  = impl (AtGt $ negate p),
             signumLe  = impl (AtGe $ negate p),
             signumGe  = impl (AtGe p),
             signumGt  = impl (AtGt p)
           }
    where
    impl atf = impliesQf ctx (QfAt atf)


--nIndet0 :: Polynomial a -> Integer
--nIndet0 p = unNIndeterminates $ nIndeterminates p

indeterminatesInAt :: AtFormula (Polynomial a) -> NIndeterminates
indeterminatesInAt AtTrue  = AtLeast 0
indeterminatesInAt AtFalse = AtLeast 0
indeterminatesInAt (AtEq  p) = nIndeterminates p 
indeterminatesInAt (AtNeq p) = nIndeterminates p 
indeterminatesInAt (AtGt  p) = nIndeterminates p
indeterminatesInAt (AtGe  p) = nIndeterminates p 

indeterminatesInQf :: QfFormula (Polynomial a) -> NIndeterminates
indeterminatesInQf (QfAt a)  = indeterminatesInAt a
indeterminatesInQf (QfNot f) = indeterminatesInQf f
indeterminatesInQf (QfAnd f g) =
    max (indeterminatesInQf f) (indeterminatesInQf g)
indeterminatesInQf (QfOr  f g) =
    max (indeterminatesInQf f) (indeterminatesInQf g)

indeterminatesInFo :: FoFormula (Polynomial a) -> NIndeterminates
indeterminatesInFo (FoQf f)    = indeterminatesInQf f
indeterminatesInFo (FoEx  _ f) = indeterminatesInFo f
indeterminatesInFo (FoAll _ f) = indeterminatesInFo f

-- |
-- Return the number of the lowest used indeterminate.
-- Only defined if the formula contains at least one
-- non-constant polynomial.
--
levelQf :: (Num a, Integral b) => QfFormula (Polynomial a) -> b
levelQf f =
    minimum $
    concatMap (take 1) $
    map usedIndeterminates $
    termsQf f

-- |
-- Return the number of the highest used indeterminate.
-- Only defined if the formula contains at least one
-- non-constant polynomial.
--
invlevelQf :: (Num a, Integral b) => QfFormula (Polynomial a) -> b
invlevelQf f = 
    maximum $
    concatMap (take 1) $
    map (reverse . usedIndeterminates) $
    termsQf f



-- |
-- Existentially quantify all variables in the formula.
--
foEx :: QfFormula (Polynomial a) -> FoFormula (Polynomial a)
foEx f = FoEx [0..nInd-1] $ FoQf f
    where
    nInd = unNIndeterminates $ indeterminatesInQf f

-- |
-- Extract all the atomic formulas from the given quantifier-free
-- formula.
--
extractAt :: Num a => QfFormula a -> [AtFormula a]
extractAt ctx = ext ctx
    where
    ext (QfAt a)    = [a]
    ext (QfAnd f g) = ext f ++ ext g
    ext (QfOr  f g) = ext f ++ ext g
    ext (QfNot f)   = ext f


atFormulaToC :: (a -> String) -> AtFormula a -> String
atFormulaToC _ AtTrue  = "1==1"
atFormulaToC _ AtFalse = "0==1"
atFormulaToC s (AtEq p)  = s p ++ " == 0"
atFormulaToC s (AtNeq p) = s p ++ " != 0"
atFormulaToC s (AtGe p)  = s p ++ " >= 0"
atFormulaToC s (AtGt p)  = s p ++ " > 0"

qfFormulaToC :: (a -> String) -> QfFormula a -> String
qfFormulaToC s = snd . qf2c
    where
      -- "Bool" tells whether parenthesis are needed if joined
      -- with something different than "||"
      -- QfFormula a -> (Bool, String)
      qf2c (QfAt a)    = (False, atFormulaToC s a)
      qf2c (QfNot f)   = (False, "!(" ++ subf ++ ")")
          where
          (_,subf) = qf2c f
      qf2c (QfAnd f g) =
          (False, condParens pf subf ++ " && " ++ condParens pg subg)
          where
          (pf,subf) = qf2c f
          (pg,subg) = qf2c g
      qf2c (QfOr  f g) = (True, subf ++ " || " ++ subg)
          where
          (_,subf) = qf2c f
          (_,subg) = qf2c g

      condParens True  x = "(" ++ x ++ ")"
      condParens False x = x


-- |
-- Convert a inequality system to a quantifier-free formula.
--
ineqSystemToFormula :: Num a
		    => GIneqSystem (Polynomial a) -- ^ input inequality system
		    -> QfFormula (Polynomial a)
ineqSystemToFormula ineqs =
    conjunctionQf $ map QfAt $ ineqSystemToAtFormulas ineqs

ineqSystemToStrictFormula :: Num a
                          => GIneqSystem (Polynomial a)
                          -> QfFormula (Polynomial a)
ineqSystemToStrictFormula ineqs =
    conjunctionQf $ map QfAt $ affineExpToAtFormulas AtGt ineqs

eqSystemToFormula :: Num a
		  => GEqSystem (Polynomial a) -- ^ input inequality system
		  -> QfFormula (Polynomial a)
eqSystemToFormula ineqs =
    conjunctionQf $ map QfAt $ eqSystemToAtFormulas ineqs

ineqSystemToAtFormulas :: Num a =>
		     GIneqSystem (Polynomial a) -> [AtFormula (Polynomial a)]
ineqSystemToAtFormulas = affineExpToAtFormulas AtGe


eqSystemToAtFormulas :: Num a =>
		     GEqSystem (Polynomial a) -> [AtFormula (Polynomial a)]
eqSystemToAtFormulas = affineExpToAtFormulas AtEq

affineExpToAtFormulas :: (Num a, AffineExpBased aeb (Polynomial a))
		      => (Polynomial a -> AtFormula (Polynomial a))
		      -> aeb
                      -> [AtFormula (Polynomial a)]
affineExpToAtFormulas atRel aeb = map ineq2f $ unjoin $ getMatrix aeb
    where
    nVP :: Integer
    nVP = nVars aeb + nParams aeb
    ineq2f line = atRel $ 
		  flatPoly
		  (constPolyN nVP con +
		   sum [c `pmul` indeterminate nVP i
		       | (c,i) <- zip coeffs [nVP-1,nVP-2 .. 0]]
		  )
	where
	con:coeffs = reverse $ entries line


-- |
-- Convert a formulas to a set of inequality systems.
-- Non-disjoint! Returns 'Nothing' when the input formula
-- is not linear in the variables.
--
formulaToIneqSystems :: Integral a
	             => Integer        -- ^ number of variables
		     -> QfFormula (Polynomial (Ratio a))  -- ^ formula describing a union of polyhedra
		     -> Maybe [GIneqSystem (Polynomial (Ratio a))]
formulaToIneqSystems n formula =
    applyMbList concat $ map (atFormulasToIneqSystems n) dnf
    where
    dnf = positiveDnfQf formula

applyMbList :: ([a] -> b) -> [Maybe a] -> Maybe b
applyMbList f xs
    | any (isNothing) xs = Nothing
    | otherwise          = Just $ f $ catMaybes xs

-- |
-- Convert a list of atomic formulas (interpreted as a conjunction)
-- to an equivalent 'GIneqSystem'. If one of the formulas is not
-- linear in the variables, 'Nothing' is returned. Each disequation
-- @f \/=  0@ is converted to @f > 0  \\\/  f < 0@ and
-- doubles the number of 'GIneqSystem's returned (that's why a list
-- of 'GIneqSystem's is returned).
-- Strict inequalities @f > 0@ and @f < 0@ are converted to weak
-- inequalities @f >= 1@ and @f <= -1@.
--
atFormulasToIneqSystems :: (Real a, Fractional a)
			=> Integer                              -- ^ number of variables
			-> [AtFormula (Polynomial a)]   -- ^ conjuncts of the formula to be converted
			-> Maybe [GIneqSystem (Polynomial a)]
atFormulasToIneqSystems n [] = Just [ineqSystem ((n,0), zeroMatrix 0 (n+1))]
atFormulasToIneqSystems n (formulas :: [AtFormula (Polynomial a)]) =
    fmap (map (fmap (fmap' fromRational))) $
    applyMbList id $
    map (applyMbList multijoin1 . map f2ineq) $
    unNeq $
    map (fmap (fmap' toRational)) $
    formulas
    where
    unNeq [] = [[]]
    unNeq (AtNeq p : fs) = 
	map (AtGt p :) un ++ map (AtGt (negate p) :) un
	where 
	un = unNeq fs
    unNeq (f : fs) = map (f:) $ unNeq fs

    f2ineq f =
	case f of
	    AtTrue  -> Just $ ineqSystem ((n,0), zeroMatrix 0 (n+1))
	    AtFalse -> Just $ ineqSystem ((n,0), zeroMatrix 1 n `conc` matrix [[-1]])
	    AtEq p  ->
		fmap (\cs -> ineqSystem ((n,0), matrix [cs, map negate cs])) $
		     coeffs n p
	    AtGe p  -> fmap (\cs -> ineqSystem ((n,0), matrix [cs])) $
		       coeffs n p
	    AtGt p  -> let g = case collectCoeffs p of
			         [] -> -1
				 cs -> constPoly $ (1%) $ negate $ abs $ foldl1 gcd $ map denominator cs
		       in fmap (\cs -> ineqSystem ((n,0), matrix [lst (+g) cs ])) $ coeffs n p
	    AtNeq _ -> Nothing


    lst f [x]    = [f x]
    lst f (x:xs) = x : lst f xs

    coeffs :: Integer -> Polynomial Rational -> Maybe [Polynomial Rational]
    coeffs 0 p = Just [p]
    coeffs k p =
	case degree uni of
          -1 -> Just $ replicate (fromIntegral k + 1) 0
	  0  -> fmap (0:) $ coeffs (k-1) (constantCoeff uni)
	  1  -> liftM2 (:) (dropP (k-1) $ leadingCoeff uni)
	                   (coeffs (k-1) $ constantCoeff uni)
	  _  -> Nothing
	where
	uni = univariate p

    dropP 0 p = Just p
    dropP k p = case degree uni of
		  0 -> dropP (k-1) (constantCoeff uni)
		  _ -> Nothing
	where
	uni = univariate p


parseQfFormula :: Num a
               => CharParser st a
               -> CharParser st (QfFormula a)
parseQfFormula (parseT :: CharParser st a) =
    expr
    where
    expr :: CharParser st (QfFormula a)
    expr = buildExpressionParser table formula
           <?> "logical formula"

    table :: OperatorTable Char st (QfFormula a) 
    table = [[op2 "/\\" (QfAnd) AssocLeft]
            ,[op2 "\\/" (QfOr ) AssocLeft]
            ]          
        where
        op2 s f assoc
             = Infix (do{ spaces; string s; return f} <?> "logical operator")
                     assoc

    formula :: CharParser st (QfFormula a)
    formula = spaces >> formula'
        where
        formula' =
            do char '['
               qf <- expr
               spaces
               char ']'
               return qf
            <|>
            do atF <- parseAtFormula parseT
               return $ QfAt atF

parseAtFormula :: Num a 
               => CharParser st a
               -> CharParser st (AtFormula a)
parseAtFormula parseT =
    do l    <- parseT
       atOp <- parseRel
       r    <- parseT
       return $ atOp (l-r)

parseRel :: Num a => CharParser st (a -> AtFormula a)
parseRel =
    spaces >>
    (
     (char '=' >> return AtEq) <|>
     (char '>' >> ((char '=' >> return AtGe) <|> return AtGt)) <|>
     (char '<' >> ((char '>' >> return AtNeq) <|>
                   (char '=' >> return (AtGe . negate)) <|>
                   return (AtGt . negate)
                  )
     ) <|>
     (string "/=" >> return AtNeq)
    )
