-- |
-- High level functions for inequation systems and access to the
-- C++ side of /LooPo/.
--
-- Authors: Matrin Griebl, Armin Groesslinger
--
-- Version: $Id: IneqSystem.hs 4201 2009-08-30 19:18:33Z groessli $
--
module LMath.Types.IneqSystem
    ( GIneqSystem, IneqSystem,
      ineqSystem, ineqSystemVPC,
      infeasibleIneqSystem, numberOfInequations,
      convexUnionAll, isSuperset, nondisjointUnion,
      disjUnionAll, disjointUnionDifference,
      convertIneqEntries,
      simplifyIneqs, simplifyDomain, intersectDomains,
      equalitiesToInequalities,
      findEqualities, joinEqualities, joinMbEqualities,
      findContext,
      Restriction(..), ToolMethod(..),
      csideSortIneqs, eliminateRedundant, isIneqRedundant,
      areIneqsEquivalent, areIneqsInfeasible,
      joinVarRest, joinParamRest, joinRestriction,
      oldProjectToVars, eliminateRedundantSorted,
      mirrorIneqs, verticesToIneqs,
      BoundaryType(..), boundsOfDim,
      integralIneqSystem, makeDivisorfree, makeMinIntegral,
      mainDenom, normalizeCol, dropContextOnlyInfo,
      simplifyTrivially, perhapsRelevant, isIn, filterIneqs,
      ineqSystemToString, ineqSystemToString',
      tightenIneqSystem,
      isSubstParamAE
    )    
where

import Data.List
import Data.Ratio

import CppInterface
import LP
import LMath.Types.LClasses
import LMath.Types.AffineExp
import LMath.Types.EqSystem
import LMath.Types.Matrix
import LMath.Types.Vector

import Debug.Trace (trace)

-- |
-- Data type for representing inequation systems.
--
newtype GIneqSystem ent = IS { affineE :: GAffineExp ent }
type IneqSystem = GIneqSystem Rational

instance Functor GIneqSystem where
    fmap f is = is { affineE = fmap f (affineE is) }

-- |
-- Constructs an inequality system.
--
ineqSystem :: (Integral a, Integral b)
	   => ((a,b), GMatrix entries)  -- ^ ((variables, linear parameters), coefficient matrix)
           -> GIneqSystem entries
ineqSystem arg = IS { affineE = affineExp arg }

ineqSystemVPC :: (GMatrix entries, GMatrix entries, GMatrix entries) ->
		 GIneqSystem entries
ineqSystemVPC = vpcVersion "ineqSystemVPC" ineqSystem


-- |
-- Construct an infeasible inequality system.
--
infeasibleIneqSystem :: (Num ent, Integral b)
		     => b               --' number of variables
		     -> b               --' number of parameters
		     -> GIneqSystem ent
infeasibleIneqSystem nV nP =
    ineqSystemVPC (negate (unitMatrix 1 nV) `join` unitMatrix 1 nV,
		   zeroMatrix 2 nP,
		   columnMatrix [0, -1]
		  )

-- return the number of inequations in the given inequation system
numberOfInequations :: GIneqSystem ent -> Integer
numberOfInequations = nRows

instance AffineExpBased (GIneqSystem ent) ent where
    toAffineExp      = affineE
    fromAffineExp ae = IS { affineE=ae }

instance MatrixBased (GIneqSystem ent) ent where
    setEntry is row col value = is { affineE = setEntry (affineE is)
				     row col value }
    getMatrix is = getMatrix (affineE is)

instance Joinable (GIneqSystem ent) where
    join is1 is2 = IS { affineE = (affineE is1) `join` (affineE is2) }
    unjoin is = [ineqSystem ((nVars is, nParams is), matr) | matr <- rows is]

instance (Num ent, Ord ent) => Applicable (GIneqSystem ent) [ent] [Bool] where
    is `apply` rs = map (>=0) $ (affineE is `apply` rs :: [ent])

instance (Num ent, Ord ent) =>
    Applicable (GIneqSystem ent) (GMatrix ent) [Bool] where
    is `apply` m = map (>=0) $ entries ((affineE is `apply` m :: GMatrix ent))

instance Show ent => Show (GIneqSystem ent) where
    show is = "Inequality system in " ++ show (nVars is) ++ " variables and "++
	      show (nParams is) ++ " parameters\n" ++
	      show (getMatrix (affineE is))


-- |
-- Find equations in an inequality system.
-- For every inequality @I@ in the input inequality system, it is
-- checked, whether @-I@ is also part of the system. Every such @I@ is
-- put into the resulting equality system.
-- The returned inequality system contains all the inequalities which are
-- not present in both negated and unnegated form.
--
findEqualities:: (Num a, Eq a) =>
		 GIneqSystem a -> (GIneqSystem a, GEqSystem a)
findEqualities sys = 
    (ineqSystem ((nVars sys, nParams sys), rowVectorsToMatrix width newIS),
     eqSystem   ((nVars sys, nParams sys), rowVectorsToMatrix width newES)
    )
    where
    width = nColumns sys
    (newIS, newES) = foldl ins ([],[]) ineqs
    ineqs = nub $ rowVectors $ getMatrix sys

    ins (is,es) r
	| r `elem` es || negate r `elem` es  =  (is, es)
	| negate r `elem` ineqs = (is,   r:es)
	| otherwise             = (r:is, es)

equalitiesToInequalities :: Num a => GEqSystem a -> GIneqSystem a
equalitiesToInequalities eqs =
    joinEqualities (ineqSystem ((nV,nP),zeroMatrix 0 (nV+nP+1))) eqs
    where
    nV = nVars   eqs
    nP = nParams eqs


-- |
-- Add equality constraints to a system of inequalities.
-- Note that the equalities e are _not_ added e and -e
-- (i.e., the number of added inequalities is twice the number
-- of equalities), but the added inequalities are e
-- and -sum(e), i.e., the number of added inequalities is
-- one plus the number of equality constraints.
-- Note that, in general, 'findEqualities' will not find the
-- equalities in the resulting system.
--
joinEqualities :: Num a => GIneqSystem a -> GEqSystem a -> GIneqSystem a
joinEqualities sys eqs
    | nRows eqs == 0  =  sys
    | otherwise       =  sys `join` eqs'
    where
    eqs'   = ineqSystem ((nV,nP), m `join` negSum)
    m      = getMatrix eqs
    negSum = rowVectorsToMatrix1 [negate $ foldl1 (+) $ rowVectors m]
    nV = nVars sys
    nP = nParams sys

joinMbEqualities :: Num a
                 => GIneqSystem a -> Maybe (GEqSystem a) -> GIneqSystem a
joinMbEqualities sys Nothing    = sys
joinMbEqualities sys (Just eqs) = joinEqualities sys eqs


findContext :: (Num a, Eq a)
            => GIneqSystem a -> (GIneqSystem a, GIneqSystem a)
findContext sys =
    (ineqSystem ((nV,nP), rowVectorsToMatrix (nV+nP+1) nonctx)
    ,ineqSystem (( 0,nP), rowVectorsToMatrix (nP+1)    ctx   )
    )
    where
    rs = rows' sys
    soc = map sysOrCtx rs
    ctx    = [vector r | Right r <- soc]
    nonctx = [vector r | Left r <- soc]

    sysOrCtx r
        | all (==0) $ take nV r  =  Right $ drop nV r
        | otherwise              =  Left  r

    nV = nVars sys
    nP = nParams sys



-------


---
-- Possible restrictions for inequation system.
--
data Restriction = NoRest      --' no assumptions are made.
		 | VarsPos     --' assume that all variables are non-negative.
		 | ParamsPos   --' assume that all parameters are non-negative.
		 | AllPos      --' consider every variable and parameter
			       --  to be non-negative.
		   deriving (Enum, Eq)

---
-- The available tools.
--
data ToolMethod = DefaultMethod   --' use the default tool.
		| PipMethod       --' use P. Feautrier's PIP tool.
		| FmMethod        --' use the Fourier/Motzkin algorithm.
		  deriving (Enum, Eq)


foreign import ccall newFIneqSys_c :: CInt -> CInt -> Ptr Matrix ->
				IO (Ptr IneqSystem)
foreign import ccall deleteFIneqSys_c :: Ptr IneqSystem -> IO ()
foreign import ccall varsFIneqSys_c :: Ptr IneqSystem -> IO CInt
foreign import ccall paramsFIneqSys_c :: Ptr IneqSystem -> IO CInt
foreign import ccall matrixFIneqSys_c :: Ptr IneqSystem -> IO (Ptr Matrix)

instance (Real a, Fractional a) => CConv (GIneqSystem a) (Ptr IneqSystem) where
    toC sys action =
	do fsys <- cppiArgs3 newFIneqSys_c (nVars sys) (nParams sys)
		                       (getMatrix sys)
	   res <- action fsys
	   deleteFIneqSys_c fsys
	   return res
    fromC fsys =
	do vars   <- varsFIneqSys_c fsys
	   params <- paramsFIneqSys_c fsys
	   matr   <- matrixFIneqSys_c fsys >>= fromC
	   deleteFIneqSys_c fsys
	   return $ ineqSystem ((vars,params), matr :: GMatrix a)

	   
foreign import ccall sortIneqs_c :: Ptr IneqSystem -> CEnum ->
                                    IO (Ptr IneqSystem)
csideSortIneqs :: IneqSystem -> ToolMethod -> Maybe IneqSystem
csideSortIneqs = 
    trace "sortIneqs called!" $
    unsafeCppiConvert2 sortIneqs_c

foreign import ccall eliminateRedundant_c :: Ptr IneqSystem ->
					     Ptr IneqSystem ->
					     CEnum -> IO (Ptr IneqSystem)
eliminateRedundant :: IneqSystem -> Maybe IneqSystem ->
                      ToolMethod -> IneqSystem
eliminateRedundant = unsafeCppiConvert3 eliminateRedundant_c

foreign import ccall isIneqRedundant_c :: Ptr IneqSystem -> CInt ->
				    Ptr IneqSystem -> CEnum -> IO CBool
isIneqRedundant :: IneqSystem -> Integer -> Maybe IneqSystem ->
		   ToolMethod -> Bool
isIneqRedundant = unsafeCppiConvert4 isIneqRedundant_c


foreign import ccall verticesAsRowVectorsToIneqs_c
    :: Ptr Matrix -> IO (Ptr IneqSystem)
verticesToIneqs :: [Vector] -> IneqSystem
verticesToIneqs vs =
    unsafeCppiConvert1 verticesAsRowVectorsToIneqs_c $ rowVectorsToMatrix1 vs


foreign import ccall convexUnionAll_c :: Ptr (Ptr IneqSystem)
				      -> CInt
				      -> IO (Ptr IneqSystem)
convexUnionAll_hs :: Real a => [GIneqSystem a] -> Int -> Maybe (GIneqSystem a)
convexUnionAll_hs syss n =
    fmap (fmap fromIntegral . integralIneqSystem) $
    unsafeCppiConvert2 convexUnionAll_c (map (fmap toRational) syss) n

-- |
-- Compute a convex union of a set of polyhedra.
--
-- Note: Be careful with convex unions of parametrised polyhedra!
-- E.g., the convex union of {0 <= x <= p, p <= x} (for a variable
-- x and a parameter p) is the universe polyhedron, i.e.,
-- x is not restricted, since the parametrised convex union
--   if  p >=0  then  0 <= x  else  p <= x
-- is not convex in the 2-dimensional space of the
-- variable _and_ the parameter.
--
convexUnionAll :: Real a => [GIneqSystem a] -> Maybe (GIneqSystem a)
convexUnionAll iss = convexUnionAll_hs iss (length iss)


foreign import ccall polyhedronIncludes_c :: Ptr IneqSystem ->
					     Ptr IneqSystem -> IO CBool
isSuperset :: RealFrac a => GIneqSystem a -> GIneqSystem a -> Bool
isSuperset = unsafeCppiConvert2 polyhedronIncludes_c


foreign import ccall union_c :: Ptr (Ptr IneqSystem) -> IO (Ptr (Ptr IneqSystem))
nondisjointUnion :: [IneqSystem] -> [IneqSystem]
nondisjointUnion = unsafeCppiConvert1 union_c


-- MC: disjUnionAll:
foreign import ccall disjUnionAll_c :: Ptr (Ptr IneqSystem) -> 
				       IO (Ptr (Ptr IneqSystem))

disjUnionAll :: RealFrac a => [GIneqSystem a] -> [GIneqSystem a]
disjUnionAll = unsafeCppiConvert1 disjUnionAll_c


----

foreign import ccall simplifyIneqs_c :: Ptr IneqSystem
				     -> Ptr IneqSystem
				     -> IO (Ptr IneqSystem)
--
-- The optional context is in the variables and parameters
-- (not only the parameters)!
--
simplifyIneqs :: Real a
              => Maybe (GIneqSystem a) -> GIneqSystem a -> GIneqSystem a
simplifyIneqs mbCtx sys =
    fmap fromIntegral $
    integralIneqSystem $
    unsafeCppiConvert2 simplifyIneqs_c (fmap (fmap toRational) mbCtx) (fmap toRational sys)

foreign import ccall simplifyDomain_c :: Ptr PolylibDomain
				      -> Ptr PolylibDomain
				      -> IO (Ptr PolylibDomain)
simplifyDomain :: Real a
  => [GIneqSystem a] -> [GIneqSystem a] -> [GIneqSystem a]
simplifyDomain _   []  =  []
simplifyDomain dom1 (dom2 :: [GIneqSystem a]) =
    [ineqSystem ((nV,nP),m) | m <- ms, notSillyInfeasible m]
    where
    ms :: [GMatrix a]
    ms = unsafeCppiConvert2 simplifyDomain_c dom1 dom2
    nV = nVars   $ head dom2
    nP = nParams $ head dom2

    notSillyInfeasible m
       | any sillyInf $ rows' m = False
       | otherwise              = True
    sillyInf rs = all (==0) (init rs) && last rs < 0

foreign import ccall intersectDomains_c :: Ptr (Ptr (Ptr IneqSystem))
                                        -> IO (Ptr (Ptr IneqSystem))
intersectDomains :: [[IneqSystem]] -> [IneqSystem]
intersectDomains = unsafeCppiConvert1 intersectDomains_c

foreign import ccall disjointUnionDifference_c :: Ptr IneqSystem
					       -> Ptr (Ptr IneqSystem)
					       -> Ptr (Ptr IneqSystem)
					       -> IO (Ptr (Ptr IneqSystem))

-- |
-- Compute a disjoint union.
-- Takes two list of polyhedra, @A@ and @B@, and
-- an optional context (i.e., assupmtions about the
-- variables and parameters) and returns a
-- list @C@ of polyhedra such that in the given context
-- @
--    union(C) = union(A) - union(B)
-- @
-- holds, i.e., the union of the returned polyhedra is
-- the same set as the difference between the union of
-- @A@ and the union of @B@.
-- In addition, the polyhedra in @C@ are pairwise disjoint.
--
disjointUnionDifference :: (Real a)
                        => Maybe (GIneqSystem a)  -- ^ context (with variables!)
			-> [GIneqSystem a]        -- ^ A
			-> [GIneqSystem a]        -- ^ B
			-> [GIneqSystem a]
disjointUnionDifference _ [] [] = []
disjointUnionDifference mbCtx ins' outs' =
    map (fmap fromIntegral . integralIneqSystem) $
    unsafeCppiConvert3 disjointUnionDifference_c ctx ins outs
    where
    ins  = map (fmap toRational) ins'
    outs = map (fmap toRational) outs'
    ctx   = case mbCtx of
               Nothing -> ineqSystem ((nV,nP), zeroMatrix 0 (nV+nP+1))
               Just c  -> fmap toRational c
    nV    = nVars   $ head (ins ++ outs)
    nP    = nParams $ head (ins ++ outs)


-- |
-- Applies the conversion operation @f@ on all entries of
-- the 'GIneqSystem' and returns the resulting matrix.
--
convertIneqEntries :: (a -> b)      -- ^ @f@
		   -> GIneqSystem a -- ^ @is@
		   -> GIneqSystem b
convertIneqEntries f is = IS affExp
    where
    affExp = affineExp ((nVars is, nParams is), matrix)
    matrix = convertMatrixEntries f (getMatrix is)

-------

--- MC: 2003-02-25 tests
mcTestSys1 :: IneqSystem
mcTestSys1 = ineqSystem ((2, 0), matrix [[ 1, 0, 0],
					 [-1, 0, 2],
					 [ 0, 1, 0],
					 [ 0,-1, 2%1]])

mcTestSys2 = ineqSystem ((2, 0), matrix [[ 1, 0,-1],
					 [-1, 0, 3],
					 [ 0, 1,-1],
					 [ 0,-1, 3%1]])

simpleTest = disjUnionAll [
			   ineqSystem ((1, 0), matrix [[ 1, 0],
						       [ -1,  4%1]]),
			   ineqSystem ((1, 0), matrix [[ 1, -5],
						       [ -1,  7%1]]),
			   ineqSystem ((1, 0), matrix [[ 1, -2],
						       [ -1,  6%1]])
	     ]

simpleTest2 = disjUnionAll [
			    ineqSystem ((2,1), matrix [[1, 0, 0%1, 0]]),
			    ineqSystem ((2,1), matrix [[0, 1, 0%1, 0]])
			   ]


mcTestUnion = [mcTestSys1, mcTestSys2]

mcTestDisjUnion = disjUnionAll mcTestUnion

---- MC: end tests


testsys = ineqSystem ((3, 2), matrix [[1,0,3,4,5,6],[1,2,0,0,0,0],
				      [1,4,5,3,3,1],[0,1,2,0,3,1%1]])


-- are is1 and is2 equivalent under the given context?
areIneqsEquivalent :: IneqSystem -> IneqSystem -> Maybe IneqSystem -> Bool
areIneqsEquivalent is1 is2 Nothing =
    isSuperset is1 is2 && isSuperset is2 is1
areIneqsEquivalent is1 is2 (Just ctx) =
    isSuperset is1' is2' && isSuperset is2' is1'
    where
    is1' = is1 `join` ectx
    is2' = is2 `join` ectx
    ectx = insertZeroVars ctx 0 (nVars is1)

areIneqsInfeasible :: RealFrac a => GIneqSystem a -> Bool
areIneqsInfeasible is = infeasibleIneqSystem nv np `isSuperset` is
    where
    nv = nVars is
    np = nParams is

joinVarRest :: IneqSystem -> Restriction -> IneqSystem
joinVarRest is rest
    | rest == VarsPos || rest == AllPos =
	is `join`
	ineqSystemVPC (unitMatrix (nVars is) (nVars is),
		       zeroMatrix (nVars is) (nParams is),
		       zeroMatrix (nVars is) 1)
    | otherwise = is

joinParamRest :: IneqSystem -> Restriction -> IneqSystem
joinParamRest is rest
    | rest == ParamsPos || rest == AllPos =
	is `join`
	ineqSystemVPC (zeroMatrix (nParams is) (nVars is),
		       zeroMatrix (nParams is) (nParams is),
		       zeroMatrix (nParams is) 1)
    | otherwise = is

joinRestriction :: IneqSystem -> Restriction -> IneqSystem
joinRestriction is rest = joinParamRest (joinVarRest is rest) rest

---
-- Project the solution of an inequation system to given dimensions.
-- @param is - the original inequation system
-- @param vars - the variables (=dimensions) to project onto
-- @return projIS - an inequation system describing the projection of the
--   system <code>is</code> onto the dimensions <code>vars</code>.
--
oldProjectToVars :: IneqSystem -> [Integer] -> IneqSystem
oldProjectToVars is vars = 
    trace "warning: old projectToVars called!" $
    is''
    where
    is'  = reorderVars is vars
    is'' = case csideSortIneqs is' PipMethod of
                Nothing  ->
		    ineqSystemVPC (eNv `join` (-eNv),
				zeroMatrix (2*newV) (nParams is'),
				matrix (genericTake newV (repeat [1%1])) `join`
				zeroMatrix newV 1)
		Just sys -> deleteVars (filterIneqs f sys) newV (origV - newV)
    origV = nVars is
    newV  = genericLength vars
    eNv   = unitMatrix newV newV
    f coeffs = all (==0) (genericTake (origV - newV) $ genericDrop newV coeffs)


-- only marginally better than `simplifyTrivially'
eliminateRedundantSorted :: IneqSystem -> Maybe IneqSystem ->
			    ToolMethod -> IneqSystem
eliminateRedundantSorted is context method =
    case bnds of
	      [] -> is
	      _  -> multijoin1 [eliminateRedundant b context method
				| b <- bnds]
    where
    vs1  = nVars is - 1
    -- extract upper / lower bounds (in right order!)
    bnds = [boundsOfDim b is dim | dim <- [vs1,(vs1-1)..0], b <- [Lower, Upper]]



-- |
-- Mirror a polyhedron at the origin.
--
mirrorIneqs :: Num a
            => GIneqSystem a
	    -> GIneqSystem a
mirrorIneqs ineqs = ineqSystemVPC (negate v, p, c)
    where
    (v, p, c) = partsVPC ineqs




----------------------------------

-- functions by Martin Griebl

data BoundaryType = Upper | Lower | Any
		    deriving Eq

boundsOfDim :: BoundaryType -> IneqSystem -> Integer -> IneqSystem
boundsOfDim b is dim = ineqSystem ((nVars is, nParams is), suitable_submat)
    where suitable_submat = foldr join emptyMat suitable_rows
	  emptyMat = zeroMatrix 0 (nVars is + nParams is +1)
	  suitable_rows = filter filter_func all_rows
	  all_rows = rows (getMatrix (toAffineExp (normalizeCol dim is))) 
	  filter_func m =  zero_part m && act_dim_filter m
	  zero_part m =  ((subMatrixVer m (dim+1) (nVars is-1)) == (zeroMatrix 1 (nVars is -dim-1)))
			 
	  act_dim_filter m = if b==Any then (entry m 0 (dim)) /= 0 
			     else if b==Upper then (entry m 0 (dim)) < 0
				  else (entry m 0 (dim)) > 0




-- general function for row manipulation
normalize :: ([Rational] -> [Rational]) -> IneqSystem -> IneqSystem
normalize normfunc is = ineqSystem((nVars is, nParams is), 
			  matrix (map normfunc  
				  (rows' (getMatrix (toAffineExp is)))))


-- entfernt alle Br"uche aus den Gleichungen in is durch 
-- Erweitern der Gleichung
integralIneqSystem :: (Real a, Integral b) => GIneqSystem a -> GIneqSystem b
integralIneqSystem es = fromAffineExp $ integralAffineExp $ toAffineExp es


-- isToInteger entfernt alle Br"uche aus den Ungleichungen in is durch 
-- Erweitern der Ungleichung
makeIntegral :: IneqSystem -> IneqSystem
makeIntegral = --normalize ineqRow2Integer
               error ("makeIntegral ist veraltet! Verwende stattdessen " ++ 
		      "makeIneqIntegral")

ineqRow2Integer :: [Rational] -> [Rational]
ineqRow2Integer v = map (* (toRational (abs(mainDenom v)))) v
	 
mainDenom :: [Rational] -> Integer
mainDenom v = foldr lcm 1 (map denominator v)



makeDivisorfree :: IneqSystem -> IneqSystem
makeDivisorfree = normalize divisorFreeRow

divisorFreeRow :: [Rational] -> [Rational]
divisorFreeRow r = map (/toRational(row_gcd r)) r
   where row_gcd :: [Rational] -> Integer
         row_gcd r = foldr gcd maxentry (map numerator r)
	 maxentry = foldr max 0 (map abs (map numerator r))



makeMinIntegral :: IneqSystem -> IneqSystem
makeMinIntegral ies = makeDivisorfree $ makeIntegral ies 



-- normalizeCol assures the coefficients of column c to be -1 or 0 or 1
-- by dividing the whole row by the absolute value of column c's coefficient
normalizeCol :: Integer -> IneqSystem -> IneqSystem
normalizeCol c = normalize (normForCoeff c)

normForCoeff :: Integer -> [Rational] -> [Rational]
normForCoeff col v = if coeff == 0 then v
		   else map (/ coeff) v
    where coeff = abs (v!!(fromIntegral col))


----------------------------------------------------------------------


-- general function for row selection
filterIneqs :: ([a] -> Bool) -> GIneqSystem a -> GIneqSystem a
filterIneqs filtfunc  is =
    ineqSystem((nVars is, nParams is), 
	       matrix (filter filtfunc  
		       (rows' (getMatrix (toAffineExp is)))))


dropContextOnlyInfo :: IneqSystem -> IneqSystem
dropContextOnlyInfo ies = filterIneqs filtfunc ies
     where filtfunc ::  [Rational] -> Bool
	   filtfunc r = or [(r!!(fromIntegral pos))/=0 | pos <- [0..(nVars ies)-1]]



simplifyTrivially :: IneqSystem -> IneqSystem
simplifyTrivially ies = ineqSystem((nVars ies, nParams ies), res_mat)
	where res_mat :: Matrix
	      res_mat = matrix res_rows
              orig_rows = rows' (getMatrix (toAffineExp ies))
	      res_rows :: [[Rational]]
	      res_rows = [ orig_rows!!row_number | 
		row_number <- [0..((length orig_rows)-1)], 
		is_minimal row_number orig_rows, 
		not (is_redundant (orig_rows!!row_number) orig_rows)]

	      is_minimal :: Int -> [[Rational]] -> Bool
	      is_minimal rownr allrows = and [(allrows !!rownr) /= (allrows !! (fromIntegral curr)) | curr <- [0..rownr-1]]

	      is_redundant :: [Rational] -> [[Rational]] -> Bool
	      is_redundant row allrows = or [is_const_pos (fzip (-) row (allrows !! curr))  | curr <- [0..(length allrows)-1]]

	      is_const_pos :: [Rational] -> Bool
	      is_const_pos rs = 
		and [(rs!!pos)==0 | pos <- [0..(length rs)-2]] &&
		(rs!!((length rs)-1)) >0

fzip :: (a->b->c) -> [a] -> [b] -> [c]
fzip f v1 v2 = map  (f_pair f) (zip v1 v2)
  where    f_pair f (a,b) = f a b

test_simplifyTrivially = simplifyTrivially (ineqSystem ((2,0), matrix [[0%1,1,0],[0,1,-2],[0,-1,42],[0,-1,50],[1,0,-3],[1,0,-4]])) 
-- should be == ineqSystem ((2,0), matrix [[0%1,1,-2],[0,-1,42],[1,0,-4]])




inBoth :: IneqSystem -> IneqSystem -> IneqSystem
inBoth iea ieb = if (ok) then 
		  ineqSystem((nVars iea,nParams ieb), 
			     matrix intersected_rows)
		  else error ("Ineqs in inBoth do not match!")
  where ok = (nVars iea == nVars ieb) && (nParams iea == nParams ieb)
        rowsa :: [[Rational]] 
        rowsa = rows' (getMatrix (toAffineExp iea))
        rowsb :: [[Rational]]  
        rowsb = rows' (getMatrix (toAffineExp ieb))
	intersected_rows :: [[Rational]]  
        intersected_rows = filter filterfunc rowsa 
	filterfunc :: [Rational] -> Bool
	filterfunc r0 = or [ r==r0 | r <- rowsb ]



firstWithoutSecond :: IneqSystem -> IneqSystem -> IneqSystem
firstWithoutSecond iea ieb = if (ok) then 
		  ineqSystem((nVars iea,nParams ieb), 
			     matrix intersected_rows)
		  else error ("Ineqs in inBoth do not match!")
  where ok = (nVars iea == nVars ieb) && (nParams iea == nParams ieb)
        rowsa :: [[Rational]] 
        rowsa = rows' (getMatrix (toAffineExp iea))
        rowsb :: [[Rational]]  
        rowsb = rows' (getMatrix (toAffineExp ieb))
	intersected_rows :: [[Rational]]  
        intersected_rows = filter filterfunc rowsa 
	filterfunc :: [Rational] -> Bool
	filterfunc r0 = and [ r/=r0 | r <- rowsb ]



perhapsRelevant :: IneqSystem -> IneqSystem -> IneqSystem
perhapsRelevant context ineqs = firstWithoutSecond (simplifyTrivially ineqs) context


isIn :: IneqSystem -> Matrix -> Bool
isIn i m = and (apply i m)

test_in :: Rational -> Bool
test_in n = isIn (ineqSystem ((1,0), matrix [[1%1,-2],[-1,5]])) (matrix [[n],[1]])



ineqSystemToString :: Bool -> Bool -> IneqSystem -> String
ineqSystemToString showNums showInhRows = affineExpToString showNums showInhRows . affineE

ineqSystemToString' :: Bool -> Bool -> IneqSystem -> String
ineqSystemToString' showNums showInhRows = affineExpToString' showNums showInhRows . affineE


-- |
-- Integer-tighten the given inequalities.
-- Divide the coefficients of the variales and parameters
-- by their greatest common divisor @g@ and floor divide
-- the constant coefficient by @g@ (i.e., make each
-- inequality as tight as possible in the integers).
--
tightenIneqSystem :: Real a => GIneqSystem a -> GIneqSystem a
tightenIneqSystem sys =
    ineqSystem ((nV,nP), m)
    where
    m = fmap fromIntegral $
        rowVectorsToMatrix (nV+nP+1) $
        tightenIneqs $
        rowVectors $ getMatrix $ integralIneqSystem sys
    nV = nVars sys
    nP = nParams sys

tightenIneqs :: [GVector Integer] -> [GVector Integer]
tightenIneqs ineqs =
    map head $          -- select the tightest of the ineqs
    groupBy coeffEq $   -- all the ineqs with the same var-part in one group
    sort $
    map tightenIneq ineqs
    where
    coeffEq v1 v2 = init (vectorEntries v1) == init (vectorEntries v2)

-- 
-- Integer-tighten one inequality.
--
tightenIneq :: GVector Integer -> GVector Integer
tightenIneq ineq
    | g == 0     =  ineq
    | otherwise  =
	fmap (\c -> floor (c % g)) ineq
    where
    g = foldl gcd0 0 $ init $ vectorEntries ineq

--
-- For the Haskell Prelude, gcd 0 0 is undefined, but
-- it is more comfortable (and consistent) to define it 
-- to be 0.
--
gcd0 :: Integral a => a -> a -> a
gcd0 0 0 = 0
gcd0 x y = gcd x y



-- |
-- Substitute a parameter by an affine expression in the variables and
-- parameters (and 1). The parameter with the given index is replaced by
-- the affine expression (in the same number of parameters as the
-- given affine system; this allows substitutions like [t:=t+1]).
--
isSubstParamAE :: (Num a, Integral b)
    => b             -- ^ index of parameter to substitute
    -> GVector a     -- ^ replacement expression
    -> GIneqSystem a
    -> GIneqSystem a
isSubstParamAE idx expr aeb =
    (\m -> ineqSystem ((nV,nP), m)) $
    rowVectorsToMatrix (nV+nP+1) $
    map newRow $
    rowVectors $
    getMatrix aeb
    where
    newRow r = vectorEntry r (nV+idx) `mulSV` expr +
               setVectorEntry r (nV+idx) 0

    nV = nVars aeb 
    nP = nParams aeb



--
-- A GIneqSystem or a union of GIneqSystems can be represented as
-- a Polyhedron structure in polylib. There is no way to convert
-- back directly to a GIneqSystem, because polylib does not
-- distinguish between variables and parameters.
--
instance Real a => CConv (GIneqSystem a) (Ptr PolylibPolyhedron) where
    toC sys = toC (getMatrix sys)
    fromC = error "fromC cannot be defined for  Real a => CConv (GIneqSystem a) (Ptr PolylibPolyhedron)"

instance Real a => CConv [GIneqSystem a] (Ptr PolylibDomain) where
    toC syss = toC (map getMatrix syss)
    fromC = error "fromC cannot be defined for  Real a => CConv [GIneqSystem a] (Ptr PolylibPolyhedron)"
