-- |
-- Some type classes and functions.
--
-- Author: Armin Groesslinger
--
-- Version: $Id: LClasses.hs 4115 2009-04-24 14:41:29Z groessli $
--
module LMath.Types.LClasses
    ( Joinable(..), Concatenatable(..), Applicable(..),
      multijoin, multiconc,
      FunctorArg(..)
    )
where

---
-- For things that can be joined.
-- @see Concatenatable
--
class Joinable a where
    ---
    -- Joins two objects. Joining means
    -- that <code>o1</code> is put above <code>o2</code> (whatever that
    -- means).
    -- @param o1       -  the upper object.
    -- @param o2       -  the lower object.
    -- @return joined  -  the joined object.
    --
    join :: a -> a -> a

    ---
    -- Splits on object into the smallest
    -- possible parts. The given object <code>whole</code> must be non-empty
    -- (whatever that means for objects of type <code>a</code>).
    -- For every non-empty object <code>obj :: a</code> the following
    -- condition must hold:
    --  <pre>
    --  multijoin (unjoin obj) = obj
    --  </pre>
    -- @param whole    -  the non-empty object to split.
    -- @return pieces  -  the smallest possible pieces of
    --                    <code>whole</code>.
    --
    unjoin :: a -> [a]

    --
    -- Joins multiple objects.
    -- The list of objects to join must not be empty. The first element
    -- of the list is put into the top position, the second element below
    -- the first one and so on.
    --
    multijoin1 :: [a] -> a
    multijoin1 [] = error "multijoin1: empty list"
    multijoin1 xs = foldl1 join xs

multijoin :: Joinable a => a -> [a] -> a
multijoin x xs = multijoin1 (x:xs)

---
-- For things that can be concatenated.
-- @see Joinable
--
class Concatenatable a where
    ---
    -- Concatenates two objects.
    -- Concatenating <code>o1</code> and <code>o2</code> means that
    -- <code>o2</code> is "appended" to <code>o1</code> on the right side
    -- (whatever left and right mean then).
    -- @param o1     -  the left object.
    -- @param o2     -  the right object.
    -- @return cced  -  the concatenated object.
    --
    conc      :: a -> a -> a

    ---
    -- Splits on object into the smallest
    -- possible parts. The given object <code>whole</code> must be non-empty
    -- (whatever that means for objects of type <code>a</code>).
    -- For every non-empty object <code>obj :: a</code> the following
    -- condition must hold:
    --  <pre>
    --  multiconc (unconc obj) = obj
    --  </pre>
    -- @param whole    -  the non-empty object to split.
    -- @return pieces  -  the smallest possible pieces of
    --                    <code>whole</code>.
    --
    unconc :: a -> [a]

    --
    -- Concatenates multiple objects.
    -- The list of objects to concatenate must not be empty. The first
    -- element of the list is put into the leftmost position,
    -- the second element is placed to the right of the first one and so on.
    --
    multiconc1 :: [a] -> a
    multiconc1 [] = error "multiconc1: empty list"
    multiconc1 xs = foldl1 conc xs


multiconc :: Concatenatable a => a -> [a] -> a
multiconc x xs = multiconc1 (x:xs)

{-
--
-- reduction over a list in a tree-like fashion;
-- `f' must (should) be associative.
--
fold1tree :: (a -> a -> a) -> [a] -> a
fold1tree f list = f1T list (length list)
    where
    f1T [x]   _ = x
    f1T [x,y] _ = f x y
    f1T xs len  = f (f1T left lenL) (f1T right lenR)
	where
	lenL = len `div` 2
	lenR = len - lenL
	(left, right) = splitAt lenL xs
-}

---
-- For anything that can be applied to something else.
--
class Applicable a b c where
    ---
    -- Apply on object to another. 
    -- Sometimes it is useful to "apply" a data value (which is not
    -- a function) to another value, for example <em>LooPo</em>'s
    -- <code>IneqSystem</code> can be "applied" to a
    -- list of rationals, giving a list of booleans. In that case a
    -- declaration like
    --  <pre>
    --  instance Applyable IneqSystem [Rational] [Bool] where
    --      is `apply` ratlist = ...
    --  </pre>
    -- is used to make <code>IneqSystem</code>s applyable to lists of
    -- rationals for one can write
    --  <pre>
    --  boollist = (is :: IneqSystem) `apply` [1%2, 2%3]  :: [Bool]
    --  </pre>
    -- @param f     -  the object which acts as a function.
    -- @param arg   -  the argument for the "function".
    -- @return res  -  the result of applying <code>f</code> to
    --                 <code>arg</code>.
    --
    apply :: a -> b -> c

{-
runTests :: [(a -> Bool, [a], String)] -> String
runTests tlist = unlines $ zipWith singleTest tlist [1..]
    where
    singleTest :: (a -> Bool, [a], String) -> Integer -> String
    singleTest (testfun, inputs, name) nr =
	"Running test #" ++ show nr ++ " (" ++ name ++")... " ++
	if null failedlist
	   then " passed"
	   else " failed for input " ++
		unwords ["#" ++ show nr | nr <- failedlist]
	where
	results    = map testfun inputs
	failedlist = [nr | (nr, False) <- zip [1..] results]
-}


-- |
-- A class similar to 'Functor', but with the additional
-- possibility to constrain the type arguments of the
-- type constructor (not possible in Haskell 98).
--
class FunctorArg a b c where
    fmap' :: (b -> c) -> a b -> a c

-- |
-- Every type that belongs to the 'Functor' class can
-- belong to the 'FunctorArg' class, too.
--
instance Functor a => FunctorArg a b c where
    fmap' = fmap
