-- |
-- This module provides the high level functions for the
-- 'Matrix' data type and exports the function interface
-- to /LooPo/'s C++ matrix implementation.
--
-- Author: Martin Griebl, Armin Groesslinger
--
-- Version: $Id: Matrix.hs 4136 2009-06-26 17:02:14Z groessli $ 
--
module LMath.Types.Matrix
    ( MatrixBased(..), MGen(..), GMatrix, Matrix,
      zeroMatrix, unitMatrix, negUnitMatrix, rowMatrix, columnMatrix,
      mtranspose, nRows, nColumns, entry, mmap,
      subMatrix, subMatrixHor, subMatrixVer,
      maxRow, maxColumn, indexes,
      rows, columns, rows', columns', entries, smul,
      multiplyRow, multiplyColumn, subtractRow, subtractColumn,
      insertZeroRows, insertZeroColumns,
      insertRows, insertColumns, deleteRows, deleteColumns,
      keepRows, keepColumns,
      moveRows, moveColumns, reorderRows, reorderColumns,
      interchangeColumns, interchangeRows,
      integralMatrix,
      convertMatrixEntries,

      determinant, gaussJordan,
      echelonReduction, inverseMatrix,
      gaussJordanTransform, findPivots,
      diagonalMatrix, unitRowVect, colOneMatrix,
      translationMatrix,
      filterRows, filterColumns,

      mulMV, mulVM,
      rowVectorsToMatrix, columnVectorsToMatrix,
      rowVectorsToMatrix1, columnVectorsToMatrix1,
      rowVectors, columnVectors,
      kernel,

      setEntriesST,
      showMatrix, intShowMat,
      showPlainMatrix,canonicShowMat,
      
      matrixToString,

      PolylibMatrix, PolylibPolyhedron, PolylibDomain, polylibMatrixCopy_c
    )
where

import Control.Exception
import Control.Monad ( when )
import Control.Monad.ST
import Data.Array.IArray
import Data.Array.MArray
import Data.Array.ST
import Data.List
import Data.Ratio

import CppInterface
import LMath.Types.LClasses
import LMath.Types.Vector
import LP

 
-- |
-- Type class for all data structures based on a matrix.
--
class MatrixBased m ent | m -> ent where
    -- |
    -- Returns the contained matrix.
    --
    getMatrix ::    m            -- ^ the object based on a matrix.
		 -> GMatrix ent  -- ^ the matrix contained in the object.

    -- |
    -- Sets a single entry in a matrix based data object. The entry in the
    -- specified row and column is set to the given value and an updated
    -- version of the original data object is returned.
    --
    setEntry :: Integral b
             => m    -- ^ the object to manipulate.
	     -> b    -- ^ the row index.
	     -> b    -- ^ the column index.
	     -> ent  -- ^ the new value.
	     -> m    -- ^ the updated object.
    setEntry mb row col value = setEntries mb [(row,col,value)]

    ---
    -- Modifies multiple entries at once. For each entry in the given list,
    -- the value at <code>(r,c)</code> is set to <code>val</code>. The list
    -- is processed from left to right, that is if there is more than one
    -- list element with the same value for <code>r</code> and <code>c</code>,
    -- then the last one determines the new value of the corresponding
    -- entry.
    -- @param mb  -  the matrix based object.
    -- @param [(r :: Integer, c :: Integer, val :: Rational)]  - 
    -- the list of modifications.
    -- @return newMb  -  the modified matix based object.
    --
    setEntries :: Integral b => m -> [(b, b, ent)] -> m
    setEntries mb = foldl (\ma (r,c,v) -> setEntry ma r c v) mb

---
-- Returns the nubmer of rows in a matrix based object.
-- @param mb -  the matrix based oject.
-- @return rows  -  the number of rows.
--
nRows :: (MatrixBased m a, Integral b) => m -> b
nRows mb = let (_,(max_r,_)) = bounds (coeffs (getMatrix mb))
	   in fromIntegral max_r + 1

---
-- Returns the nubmer of columnss in a matrix based object.
-- @param mb -  the matrix based oject.
-- @return columns  -  the number of columns.
--
nColumns :: (MatrixBased m a, Integral b) => m -> b
nColumns mb = let (_,(_, max_c)) = bounds (coeffs (getMatrix mb))
	      in fromIntegral max_c + 1

---
-- Gets a single entry from a matrix based object.
-- @param mb      -  the matrix based object.
-- @param row     -  the row index.
-- @param column  -  the column index.
-- @return value  -  the value of the entry at <code>(row,column)</code>.
--
entry :: (MatrixBased m a, Integral b) => m -> b -> b -> a
entry mb row col = coeffs (getMatrix mb) ! (fromIntegral row, fromIntegral col)

---
-- Returns a sub matrix. A matrix of width <code>left-right+1</code> and
-- height <code>bottom-top+1</code> is made from the entries between
-- <code>(top,left)</code> and <code>(bottom,right)</code> (including the
-- borders).
-- @param mb      -  the matrix based object.
-- @param top     -  the top row.
-- @param left    -  the left column.
-- @param bottom  -  the bottom row.
-- @param right   -  the right column.
-- @return matr   -  the sub matrix.
--
subMatrix :: (MatrixBased m a, Integral b) =>
	     m -> b -> b -> b -> b -> GMatrix a
subMatrix matr top left bottom right =
    assert (left <= right+1) $
    assert (top  <= bottom+1) $
    GMatrix { coeffs = ixmap bnds f (coeffs m) }
	where
	m    = getMatrix matr
	bnds = ((0,0),(fromIntegral (bottom-top), fromIntegral (right-left)))
	f (r,c) = (r+fromIntegral top,c+fromIntegral left)



---
-- Maps over a mtrix. The given function is applied to each element
-- of the matrix based object's matrix.
-- @param f       -  the function to map over the matrix.
-- @param mb      -  the matrix based object.
-- @return newMb  -  the matrix based object with the modified matrix.
-- 
mmap :: MatrixBased m a => (a -> a) -> m -> m
mmap f mb = setEntries mb [(r, c, f (entry mb r c)) | (r,c) <- indexes mb] 

---
-- Returns the maximum row number of the matrix based object.
-- @param mb       -  the matrix based object.
-- @return maxrow  -  the maximum row number.
--
maxRow :: (MatrixBased m a, Integral b) => m -> b
maxRow matr = nRows matr - 1

---
-- Returns the maximum column number of the matrix based object.
-- @param  mb      -  the matrix based object.
-- @return maxcol  -  the maximum column number.
--
maxColumn :: (MatrixBased m a, Integral b) => m -> b
maxColumn matr = nColumns matr - 1

-- |
-- Returns the list of indexes of the matrix based object's matrix.
--
indexes :: (Integral b, MatrixBased m a)
	=> m        -- ^ the matrix based object
	-> [(b, b)] -- ^ the list of indexes
indexes mb = [(fromIntegral row, fromIntegral col)
	     | row <- [top..bottom], col <- [left..right]]
    where
    ((top,left),(bottom,right)) = bounds (coeffs $ getMatrix mb)

-- |
-- Extract columns from a matrix based object's matrix. The columns
-- between @left@ and @right@ (including the borders)
-- are taken from the original matrix and returned as a new matrix. The
-- returned matrix has @right-left+1@ columns.
--
subMatrixVer :: (MatrixBased m a, Integral b)
	     => m  -- ^ the matrix based object 
             -> b  -- ^ the left column index 
             -> b  -- ^ the right column index
             -> GMatrix a
subMatrixVer matr left right = subMatrix matr 0 left (maxRow matr) right

---
-- Extracts rows from a matrix based object's matrix. The rows
-- between <code>top</code> and <code>bottom</code> (including the borders)
-- are taken from the original matrix and returned as a new matrix. The
-- returned matrix has <code>bottom-top+1</code> rows.
-- @param mb        -  the matrix based object.
-- @param top       -  the top column index.
-- @param bottom    -  the bottom column index.
-- @return submatr  -  the horizontal sub matrix.
--
subMatrixHor :: (MatrixBased m a, Integral b)
	     => m -> b -> b -> GMatrix a
subMatrixHor matr top bottom = subMatrix matr top 0 bottom
			       (maxColumn matr)

---
-- Returns a list of rows of a matrix. Each entry of the returned list is a
-- matrix with a single row with the same values as the corresponding row in
-- the matrix based object's matrix.
-- @param mb            -  the matrix based object.
-- @return rowmatrices  -  the list of rows.
--
rows :: MatrixBased m a => m -> [GMatrix a]
rows matr = [subMatrixHor matr r r | r <- [0 .. maxRow matr]]

---
-- Returns a list of columns of a matrix. Each entry of the returned list
-- is a matrix with a single column with the same values as the corresponding
-- column in the matrix based object's matrix.
-- @param mb               -  the matrix based object.
-- @return columnmatrices  -  the list of columns.
--
columns :: MatrixBased m a => m -> [GMatrix a]
columns matr = [subMatrixVer matr c c | c <- [0 .. maxColumn matr]] 

---
-- Returns a list of rows of a matrix. Each entry of the returned list is
-- a list of rationals with the same values as the corresponding row in the
-- matrix based object's matrix.
-- @param mb        -  the matrix based object.
-- @return rowlist  -  the list of rows.
--
rows' :: MatrixBased m a => m -> [[a]]
rows' mb = [[coeffs m ! (r,c) | c <- [0 .. maxColumn m]]
           | r <- [0 .. maxRow m]
           ]
    where
    m = getMatrix mb

---
-- Returns a list of columns of a matrix. Each entry of the returned list
-- is a list of rationals with the same values as the corresponding column
-- in the matrix based object's matrix.
-- @param mb            -  the matrix based object.
-- @return columnlist   -  the list of columns.
--
columns' :: MatrixBased m a => m -> [[a]]
columns' mb = [[coeffs m ! (r,c) | r <- [0 .. maxRow m]]
              | c <- [0 .. maxColumn m]
              ]
    where
    m = getMatrix mb

---
-- Returns a list of all the entries of a matrix. The entries of the
-- matrix based object's matrix are put into a list, from left to right
-- and from top to bottom. To put it in other words, <code>entries</code>
-- is equivalent to <code>(concat . rows')</code>.
-- @param mb       -  the matrix based object.
-- @return values  -  the list of entries.
--
entries :: MatrixBased m a => m -> [a]
entries = -- concat . rows'
          -- more space efficient:
          elems . coeffs . getMatrix

---
-- Scalar multiplication.
-- All entries in the matrix based object's matrix are multiplied
-- by <code>factor</code>.
-- @param mb      -  the matrix based object.
-- @param factor  -  the factor to multiply the matrix with.
-- @return newMb  -  the modified matrix based object.
--
smul :: (Num a, MatrixBased m a) => m -> a -> m
smul m s = mmap (*s) m

-- multiply row `r' in `m' with factor `fact'
multiplyRow :: (Num a, MatrixBased m a, Integral b) => m -> b -> a -> m
multiplyRow m r fact =
    setEntries m [(r,c,entry m r c * fact) | c <- [0..maxColumn m]]

-- multiply column `c' in `m' with factor `fact'
multiplyColumn :: (Num a, MatrixBased m a, Integral b) => m -> b -> a -> m
multiplyColumn m c fact =
    setEntries m [(r,c,entry m r c * fact) | r <- [0..maxRow m]]

-- | Zieht in Matrix m Zeile k*z von Zeile i ab
subtractRow :: (Num a, Integral b) => GMatrix a -> b -> b -> a -> GMatrix a
subtractRow m k i z = foldl (\m j -> setEntry m i j ((entry m i j) - z*(entry m k j))) m [0..((nColumns m)-1)]

-- | Zieht in Matrix m Spalte k*z von Spalte j ab
subtractColumn :: (Num a, Integral b) => GMatrix a -> b -> b -> a -> GMatrix a
subtractColumn m k j z = foldl (\m i -> setEntry m i j ((entry m i j) - z*(entry m i k))) m [0..((nRows m)-1)]


---
-- <em>LooPo</em>'s data structure for representing matrices.
-- Note that row and columns indexes start with <b>0</b> (and not 1).
--
data GMatrix a = GMatrix { coeffs :: Array (Integer,Integer) a }

instance Eq a => Eq (GMatrix a) where
    (GMatrix x) == (GMatrix y)  =  x == y

instance Functor GMatrix where
    fmap f (GMatrix {coeffs=c}) = GMatrix { coeffs = fmap f c } 

type Matrix = GMatrix Rational


---
-- Show functions for matrices.
--
instance Show a => Show (GMatrix a) where
    ---
    -- Prints a matrix. The matrix is
    -- converted to a string, each row on a line by itself. A short
    -- introduction text tells the number of rows and columns.
    -- @param matr  -  the matrix to show.
    -- @return s    -  a human readable representation of the matrix.
    --
    show matr = "Matrix with " ++ show (nRows matr) ++ " rows and " ++
		show (nColumns matr) ++ " columns\n" ++
		unlines (map showRow (rows' matr))
	where
	showRow :: Show b => [b] -> String
	showRow rs = " " ++
		     concat (intersperse "  " (map show rs))

showMatrix :: Integral a => GMatrix (Ratio a) -> String
showMatrix matr =
    "Matrix with " ++ show (nRows matr) ++ " rows and " ++
    show (nColumns matr) ++ " columns\n" ++
    showPlainMatrix matr

showPlainMatrix :: Integral a => GMatrix (Ratio a) -> String
showPlainMatrix matr = unlines (map showRow (rows' matr))
    where showRow :: Integral a => [Ratio a] -> String
	  showRow rs = " " ++
		       concat (intersperse "  "
			       [show (numerator r) ++ "/" ++
				show (denominator r) | r <- rs])

intShowMat :: Integral a => GMatrix (Ratio a) -> Bool -> String
intShowMat matr header = 
    (case header of 
		True  -> "Matrix with " ++ show (nRows matr) ++ " rows and " ++
			 show (nColumns matr) ++ " columns\n" 
		False -> ""
    ) ++ unlines (map showIntRow (rows' matr))
	     where
	     showIntRow :: Integral a => [Ratio a] -> String
	     showIntRow rs = " " ++
		     concat (intersperse " "
			     [show (numerator r)  ++  (if denominator r == 1 then [] 
                                                       else error "non-integer matrix")
			      | r <- rs])

canonicShowMat :: Integral a => GMatrix (Ratio a) -> Bool -> String
canonicShowMat matr header = 
    (case header of 
		True  -> "Matrix with " ++ show (nRows matr) ++ " rows and " ++
			 show (nColumns matr) ++ " columns\n" 
		False -> ""
    ) ++ unlines (map showIntRow (rows' matr))
	     where
	     showIntRow :: Integral a => [Ratio a] -> String
	     showIntRow rs = " " ++
		     concat (intersperse " "
			     [show (numerator r)  ++  (if denominator r == 1 then [] 
                                                       else "/" ++ show (denominator r))
			      | r <- rs])

---
-- Constructs a zero matrix.
-- This function makes a matrix with <code>r</code> rows and
-- <code>c</code> columns and all entries set to 0.
-- @param r      -  the number of rows.
-- @param c      -  the number of columns.
-- @return zero  -  the zero matrix.
--
zeroMatrix :: (Num a, Integral b) => b -> b -> GMatrix a
zeroMatrix r' c' =
    GMatrix { coeffs = array ((0,0),(r-1,c-1))
			    [((y,x),0) | y <- [0 .. r-1], x <- [0 .. c-1]] }
    where
    r = fromIntegral r'
    c = fromIntegral c'

---
-- Constructs an identity matrix.
-- A matrix with 1's on the diagonal (and 0 elsewhere) with
-- a width of <code>rows</code> and a height of <code>columns</code>
-- is returned. To put it in other words, for any <code>r,c&gt;=0</code>
-- and <code>0&lt;=i&lt;r</code> and <code>0&lt;=j&lt;c</code>,
-- <code>unitMatrix</code> satisfies
-- the property
-- <pre>
-- (entry (unitMatrix r c) i j == 1) = (i == j)
-- </pre>
-- @param rows     -  the number of rows.
-- @param columns  -  the number of columns.
-- @return uni     -  the unit matrix.
--
unitMatrix :: (Num a, Integral b) => b -> b -> GMatrix a
unitMatrix r' c' =
    GMatrix { coeffs = array ((0,0),(r-1,c-1))
			    [((y,x),diag (y,x)) | y <- [0 .. r-1],
						  x <- [0 .. c-1]] }
    where
    r = fromIntegral r'
    c = fromIntegral c'
    diag (x, y)
     | x == y    = 1
     | otherwise = 0


negUnitMatrix :: (Num a, Integral b) => b -> b -> GMatrix a
negUnitMatrix r c = convertMatrixEntries (*(-1)) (unitMatrix r c)


---
-- Generator function for matrices.
--
class MGen a b | a -> b where
    ---
    -- Constructs a matrix.
    -- @param input  -  the input for constructing a matrix.
    -- @return m     -  the constructed matrix.
    --
    matrix :: a -> GMatrix b


instance (Integral a, Integral b) => MGen (a, b, [c]) c where
    matrix (r,c,entries)
	| r < 0 || c < 0  =
	    error "matrix: negative number of rows or columns"
	| genericLength entries == toInteger r * toInteger c  =
	    GMatrix { coeffs = listArray bnds entries }
	| otherwise = error ("matrix: number of rows and columns does not " ++
			     "match the list of entries.")
	  where
	  bnds = ((0, 0), (toInteger r - 1, toInteger c - 1))

instance MGen [[a]] a where
    matrix  [] = matrix (0, 0, [] :: [a])
    matrix rss = matrix (length rss, length (head rss), concat rss)


---
-- Produces a row matrix.
-- The entries in the row matrix are taken from the given list. That means
-- that the resulting matrix has as many columns as the given list has
-- entries.
-- @param rs     -  the entries for the row matrix.
-- @return matr  -  the row matrix.
--
rowMatrix :: [a] -> GMatrix a
rowMatrix entries = matrix (1, length entries, entries)

---
-- Produces a column matrix.
-- The entries in the column matrix are taken from the given list. That means
-- that the resulting matrix has as many rowss as the given list has
-- entries.
-- @param rs     -  the entries for the column matrix.
-- @return matr  -  the column matrix.
--
columnMatrix :: [a] -> GMatrix a
columnMatrix entries = matrix (length entries, 1, entries)

---
-- Transposes a matrix.
-- For any matrix <code>m</code> and <code>0&lt;=i&lt;=maxRow m</code>
-- and <code>0&lt;=j&lt;=maxColumn m</code>, <code>transpose</code>
-- satiesfies the following properties:
-- <pre>
-- entry m i j = entry (mtranspose m) j i
--
-- mtranspose . mtranspose  =  id
-- </pre>
-- @param m      -  the matrix to transpose.
-- @return newM  -  the transposed matrix.
--
mtranspose :: GMatrix a -> GMatrix a
mtranspose m = matrix (nColumns m, nRows m, concat (columns' m))

---
-- Moves some rows within a matrix.
-- The rows <code>from</code> to <code>from+count-1</code> are moved to
-- the new position <code>to</code>. After the move, the original row
-- with the index <code>from</code> is then at index <code>to</code>.
-- @param m            -  the input matrix.
-- @param from         -  the first folumn to move.
-- @param to           -  the destination of the rows.
-- @param count        -  the number of rows to move.
-- @return newMatrix   -  the modified matrix.
--
moveRows :: GMatrix a -> Integer -> Integer -> Integer -> GMatrix a
moveRows matr fromIdx toIdx count
    | fromIdx >= 0 && toIdx >= 0 && fromIdx + count <= nrows &&
      toIdx + count <= nrows && count >= 0 =
	  matrix $ move (rows' matr) fromIdx toIdx count
    | otherwise = error "moveRows:  index out of bounds"
    where
    nrows   = nRows matr

---
-- Move columns within a matrix.
-- The columns <code>from</code> to <code>from+count-1</code> are moved
-- to the new position <code>to</code>. After the move the column with
-- (old) index <code>from</code> has the new index <code>to</code>.
-- @param m     - the input matrix.
-- @param from  - the first column to move.
-- @param to    - the destination index.
-- @param count - the number of columns to move.
-- @return newMatrix - the modified matrix.
--
moveColumns :: GMatrix a -> Integer -> Integer -> Integer -> GMatrix a
moveColumns matr fromIdx toIdx count
    | fromIdx >= 0 && toIdx >= 0 && fromIdx + count <= ncols &&
      toIdx + count <= ncols && count >= 0 =
	  matrix (transpose $ move (columns' matr) fromIdx toIdx count)
    | otherwise = error "moveColumns:  index out of bounds"
    where
    ncols = nColumns matr

-- helper function for `moveRows' and `moveColumns'
move :: [[a]] -> Integer -> Integer -> Integer -> [[a]]
move list fromIdx toIdx count = part1 ++ part2 ++ part3
    where
    (part1a, tmp   ) = splitAt (fromInteger fromIdx) list
    (part2,  part3a) = splitAt (fromInteger count)   tmp

    (part1, part3) = splitAt (fromInteger toIdx) (part1a ++ part3a)


insertZeroColumns :: (Num a, Integral b) => GMatrix a -> b -> b -> GMatrix a
insertZeroColumns m idx count =
    insertColumns m idx (zeroMatrix (nRows m) count)

insertZeroRows :: (Num a, Integral b) => GMatrix a -> b -> b -> GMatrix a
insertZeroRows m idx count = insertRows m idx (zeroMatrix count (nColumns m))

-- |
-- Insert rows into a matrix.
--
insertRows :: Integral b
           => GMatrix a   -- ^ the original matrix
           -> b           -- ^ the position for the uppermost new row
           -> GMatrix a   -- ^ the rows to insert
           -> GMatrix a
insertRows matr idx middle
    | nColumns middle /= nColumns matr =
	error "insertRows:  number of columns doesn't match"
    | idx >= 0 && idx <= nRows matr  =
	multijoin (subMatrixHor matr 0 (idx-1))
		      [middle, subMatrixHor matr idx (maxRow matr)]
    | otherwise = error "insertRows:  index out of bounds"
    where
    (upper, lower) = splitAt (fromIntegral idx) (rows' matr)

insertColumns :: Integral b => GMatrix a -> b -> GMatrix a -> GMatrix a
insertColumns matr idx middle
    | nRows middle /= nRows matr =
	error "insertColumns:  number of rows doesn't match"
    | idx >= 0 && idx <= nColumns matr  =
	multiconc (subMatrixVer matr 0 (idx-1))
		      [middle, subMatrixVer matr idx (maxColumn matr)]
    | otherwise = error "insertColumns:  index out of bounds"
    where
    (left, right) = splitAt (fromIntegral idx) (columns' matr)

deleteRows :: Integral b => GMatrix a -> b -> b -> GMatrix a
deleteRows matr idx count
    | idx >= 0 && idx + count <= nRows matr && count >= 0  =
	matrix newRows
    | otherwise  = error  "deleteRows:  index out of bounds"
    where
    rs = rows' matr
    newRows = take (fromIntegral idx) rs ++
	      drop (fromIntegral (idx+count)) rs

deleteColumns :: Integral b => GMatrix a -> b -> b -> GMatrix a
deleteColumns matr idx count =
    assert (idx >= 0) $
    assert (idx + count <= nColumns matr) $
    assert  (count >= 0) $
    left `conc` right
    where
    left  = subMatrixVer matr 0           (idx-1)
    right = subMatrixVer matr (idx+count) (nColumns matr-1)

reorderRows :: Integral b => GMatrix a -> [b] -> GMatrix a
reorderRows m idxs
    | not (null idxs) && (minimum idxs < 0 || maximum idxs >= nRows m) =
	error "reorderRows:  index out of bounds"
    | otherwise =
	multijoin (matrix (0, nColumns m, []))
		      [rows m !! fromIntegral i | i <- all_idxs]
    where
    all_idxs = idxs ++ ([0..maxRow m] \\ idxs)

reorderColumns :: Integral b => GMatrix a -> [b] -> GMatrix a
reorderColumns m idxs = keepColumns m (idxs ++ ([0..maxColumn m] \\ idxs))

keepRows :: Integral b => GMatrix a -> [b] -> GMatrix a
keepRows  m idxs
    | not (null idxs) && (minimum idxs < 0 || maximum idxs >= nRows m) =
	error "keepRows:  index out of bounds"
    | otherwise =
	multijoin (matrix (0, nColumns m, []))
		      [(rows m)!!(fromIntegral i) | i <- idxs]


-- | Vertauscht zwei Zeilen einer Matrix
interchangeRows :: Integral a => GMatrix b -> a -> a -> GMatrix b
interchangeRows m f t = reorderRows m (map (\i -> if i == f then t else if i == t then f else i) [0..((nRows m)-1)])

-- | Vertauscht zwei Spalten einer Matrix
interchangeColumns :: Integral a => GMatrix b -> a -> a -> GMatrix b
interchangeColumns m f t = reorderColumns m (map (\i -> if i == f then t else if i == t then f else i) [0..((nColumns m)-1)])


keepColumns :: Integral b => GMatrix a -> [b] -> GMatrix a
keepColumns  m idxs
    | not (null idxs) && (minimum idxs < 0 || maximum idxs >= nColumns m) =
	error "keepColumns:  index out of bounds"
    | otherwise =
	multiconc (matrix (nRows m, 0, []))
		      [(columns m)!!(fromIntegral i) | i <- idxs]

instance Concatenatable (GMatrix a) where
    conc m1 m2
	| nRows m1 == nRows m2  =
	    GMatrix { coeffs = listArray ((0,0),(maxRow m1, newCols)) 
		      (concat (zipWith (++) (rows' m1) (rows' m2))) }
	| otherwise = error "conc: matrices must have same number of rows"
        where
	newCols = nColumns m1 + nColumns m2 - 1
    unconc = columns

instance Joinable (GMatrix a) where
    join m1 m2
	| nColumns m1 == nColumns m2  =
	    GMatrix { coeffs = listArray ((0,0), (newRows, maxColumn m1))
		      (elems (coeffs m1) ++ elems (coeffs m2)) }
	| otherwise = error "join: matrices must have same number of columns"
        where
	newRows = nRows m1 + nRows m2 - 1
    unjoin = rows 

-- wrong for matrices with 0 rows!
--    multijoin1 ms = matrix (concatMap rows' ms)

instance MatrixBased (GMatrix a) a where
    getMatrix = id

    setEntry matr row col value =
	matr {coeffs = (coeffs matr)//[((fromIntegral row, fromIntegral col),value)]} 


instance Num a => Num (GMatrix a) where
    m1 + m2
     | nRows m1 == nRows m2 && nColumns m1 == nColumns m2 =
	 GMatrix { coeffs = array ((0,0),(row,col))
		              [((r,c),(coeffs m1)!(r,c) + (coeffs m2)!(r,c)) |
			       r <- [0..row], c <- [0..col]]  }
     | otherwise = error "(+):  cannot add matrices with different sizes"
     where
     row = maxRow    m1
     col = maxColumn m1

    m1 * m2
     | nColumns m1 == nRows m2 =
	 GMatrix { coeffs = array ((0,0),(row,col))
                           [((r,c), inner (m1_rows!! fi r) (m2_cols!! fi c)) |
			    r <- [0..row], c <- [0..col]] }
     | otherwise = error "(*):  given matrices cannot be multiplied"
     where
     row = maxRow    m1
     col = maxColumn m2
     inner xs ys = sum (zipWith (*) xs ys)  -- vector product
     m1_rows = rows'    m1
     m2_cols = columns' m2
     fi = fromInteger

    negate = mmap (*(-1))
    abs    = mmap abs

    signum        = error "'signum' undefined in Num (GMatrix a)"
    fromInteger i = error ("'fromInteger' undefined in Num (GMatrix a), tried to call (fromInteger " ++ show i ++ ")")


-- |
-- Make the entries of a given matrix integral by
-- multiplying the entries with the least common multiple
-- of the denominators of the entries.
-- The entries are always multiplied by a positive factor,
-- i.e., the sign of the entries is retained.
--
integralMatrix :: (Real a, Integral b) => GMatrix a -> GMatrix b
integralMatrix matr' =
    fmap (\e -> fromIntegral (numerator e * (commonDen `quot` denominator e)))
	 matr
    where
    commonDen = foldl lcm 1 $ map denominator $ entries matr
    matr = fmap toRational matr'

-- |
-- Make the entries of a given matrix integral by
-- multiplying the entries with the least common multiple
-- of the denominators of the entries.
-- The entries are always multiplied by a positive factor,
-- i.e., the sign of the entries is retained.
--
integralMatrix' :: (Integral a) => GMatrix (Ratio a) -> (GMatrix a, a)
integralMatrix' matr' =
    (fmap (\e -> fromIntegral (numerator e * (commonDen `quot` denominator e)))
	 matr', commonDen)
    where
    commonDen = foldl lcm 1 $ map denominator $ entries matr'
    matr = fmap toRational matr'

-- |
-- Applies the conversion operation @f@ on all entries of the
-- matrix @matr@ and returns the resulting matrix.
--
convertMatrixEntries :: (a -> b)  -- ^ @f@
		     -> GMatrix a -- ^ @matr@
		     -> GMatrix b
convertMatrixEntries f (GMatrix { coeffs = matrCoeffs }) =
    let entries = map (\(x, y) -> (x, f y)) (assocs matrCoeffs)
    in GMatrix { coeffs = array (bounds matrCoeffs) entries }


foreign import ccall newFMatrix_c :: CInt -> CInt -> Ptr (Ptr Rational) ->
			       IO (Ptr Matrix)
foreign import ccall deleteFMatrix_c :: Ptr Matrix -> IO ()
foreign import ccall rowsFMatrix_c :: Ptr Matrix -> IO CInt
foreign import ccall colsFMatrix_c :: Ptr Matrix -> IO CInt
foreign import ccall entriesFMatrix_c :: Ptr Matrix -> IO (Ptr (Ptr Rational))

instance (Real a, Fractional a) => CConv (GMatrix a) (Ptr Matrix) where
    toC matr action =
	do fm <- (cppiArgs3 newFMatrix_c) (nRows matr) (nColumns matr)
		                         (entries matr)
	   res <- action fm
	   deleteFMatrix_c fm
	   return res
    fromC fm =
	do rs <- rowsFMatrix_c fm
	   cs <- colsFMatrix_c fm
	   coeffs <- entriesFMatrix_c fm
	   entries <- peekArray (fromIntegral (rs*cs)) coeffs >>= \xs ->
		      sequence (map fromC xs)
	   free coeffs
	   deleteFMatrix_c fm
	   return $ matrix (rs, cs, entries :: [a])

foreign import ccall determinant_c :: Ptr Matrix -> IO (Ptr Rational)
determinant :: Matrix -> Rational
determinant = unsafeCppiConvert1 determinant_c

foreign import ccall gaussJordan_c :: Ptr Matrix -> Ptr Matrix ->
				IO (CPair (Ptr Matrix) (Ptr CInt))
gaussJordan :: Matrix -> Matrix -> Maybe (Matrix, [Integer])
gaussJordan = unsafeCppiConvert2 gaussJordan_c


foreign import ccall echelonReduction_c :: Ptr Matrix ->
				     IO (CPair (Ptr Matrix) (Ptr Matrix))
echelonReduction :: Matrix -> (Matrix, Matrix)
echelonReduction = unsafeCppiConvert1 echelonReduction_c


foreign import ccall inverse_c :: Ptr Matrix -> IO (Ptr Matrix)
inverseMatrix :: Matrix -> Matrix
inverseMatrix = unsafeCppiConvert1 inverse_c


-----------------------------------------
-- Conversion to/from Polylib matrices --
-----------------------------------------

data PolylibMatrix

foreign import ccall polylibMatrixNew_c :: CInt -> CInt -> Ptr (Ptr PolylibValue) -> IO (Ptr PolylibMatrix)
foreign import ccall polylibMatrixColumns_c :: Ptr PolylibMatrix -> IO CInt
foreign import ccall polylibMatrixRows_c :: Ptr PolylibMatrix -> IO CInt
foreign import ccall polylibMatrixEntry_c :: CInt -> CInt -> Ptr PolylibMatrix -> IO (Ptr PolylibValue)
foreign import ccall polylibMatrixFree_c :: Ptr PolylibMatrix -> IO ()
foreign import ccall polylibMatrixCopy_c :: Ptr PolylibMatrix -> IO (Ptr PolylibMatrix)

instance Integral a => CConv (GMatrix a) (Ptr PolylibMatrix) where
    toC m action =
        toC (entries m) $ \pvsPtr ->
        do pmPtr <- polylibMatrixNew_c (nRows m) (nColumns m) pvsPtr
           res <- action pmPtr
           polylibMatrixFree_c pmPtr
           return res
    fromC pmPtr =
        do r <- polylibMatrixRows_c pmPtr
           c <- polylibMatrixColumns_c pmPtr
           ents <- sequence [polylibMatrixEntry_c i j pmPtr >>= fromC
                            | i <- [0..r-1], j <- [0..c-1]] :: IO [a]
           polylibMatrixFree_c pmPtr
           return $ matrix (r, c, ents)

data PolylibPolyhedron
data PolylibDomain

foreign import ccall polylibPolyhedronNew_c :: Ptr PolylibMatrix -> IO (Ptr PolylibPolyhedron)
foreign import ccall polylibPolyhedronFree_c :: Ptr PolylibPolyhedron -> IO ()
foreign import ccall polylibPolyhedronMatrix_c :: Ptr PolylibPolyhedron -> IO (Ptr PolylibMatrix)
foreign import ccall polylibPolyhedronNext_c :: Ptr PolylibDomain -> IO (Ptr PolylibDomain)
foreign import ccall polylibPolyhedronSetNext_c :: Ptr PolylibDomain -> Ptr PolylibDomain -> IO ()

--
-- A polyhedron in polylib does not store the number of variables/parameters,
-- only the total number of dimensions. So the conversion to/from
-- polylib polyhedra is defined for GMatrices instead of IneqSystems.
--
instance Real a => CConv (GMatrix a) (Ptr PolylibPolyhedron) where
    toC m action =
        toC m' $ \pmPtr ->
        do phPtr <- polylibPolyhedronNew_c pmPtr
           res <- action phPtr
           polylibPolyhedronFree_c phPtr
           return res
        where
        m' = (ones `conc` integralMatrix m) `join` positivity
        ones = columnMatrix $ replicate (nRows m) 1
        positivity = rowMatrix (1 : replicate (nColumns m - 1) 0 ++ [1])
    fromC phPtr =
        do ms <- fromC (castPtr phPtr :: Ptr PolylibDomain) :: IO [GMatrix a]
           when (length ms /= 1) (fail "fromC: not a single polyhedron")
           return $ head ms

instance Real a => CConv [GMatrix a] (Ptr PolylibDomain) where
    toC [] action = action nullPtr
    toC (m:ms) action =
       toC m $ \phPtr ->
       toC ms $ \pdPtr2 ->
       let pdPtr = castPtr (phPtr :: Ptr PolylibPolyhedron) :: Ptr PolylibDomain
       in do polylibPolyhedronSetNext_c pdPtr pdPtr2
             action pdPtr

    fromC pdPtr
        | pdPtr == nullPtr  =  return []
        | otherwise =
            do pdPtr2 <- polylibPolyhedronNext_c pdPtr
               ms <- fromC pdPtr2
               m  <- firstMatrix
               return (m:ms)
        where
        firstMatrix =
           do let phPtr = castPtr pdPtr
              pmPtr <- polylibPolyhedronMatrix_c phPtr
              phNext <- polylibPolyhedronNext_c phPtr
              polylibPolyhedronFree_c phPtr
              m <- fromC pmPtr :: IO (GMatrix Integer)
              return $ fmap fromIntegral $ matrixFromPolylib m
        -- convert (in)equalities from polylib format to a matrix
        -- suitable for an IneqSystem.
        matrixFromPolylib m =
            rowVectorsToMatrix (nColumns m - 1) $
            concat $
            zipWith fixEqs isEq $ rowVectors m1
            where
            isEq = map (==0) $ entries $ subMatrixVer m 0 0
            m1 = subMatrixVer m 1 (maxColumn m)
            fixEqs False r
                | constCoeff >= 0 && all (==0) varCoeffs  =  []    -- 1 >= 0
                | otherwise                               =  [r]
                where
                constCoeff:varCoeffs = reverse $ vectorEntries r
            fixEqs True  r = [r, negate r]

        
    

-----------------------------
-- Matrix-Vector functions --
-----------------------------

-- matrix * (column) vector
mulMV :: Num a => GMatrix a -> GVector a -> GVector a
mulMV m v 
    | nColumns m == vectorLength v =
	vector [sum $ zipWith (*) r (vectorEntries v) | r <- rows' m]
    | otherwise = error "mulMV: matrix and vector don't match"

-- (row) vector * matrix
mulVM :: Num a => GVector a -> GMatrix a -> GVector a
mulVM v m 
    | nRows m == vectorLength v =
	vector [sum $ zipWith (*) (vectorEntries v) c | c <- columns' m]
    | otherwise = error "mulVM: vector and matrix don't match"

rowVectorsToMatrix :: Integral b => b -> [GVector a] -> GMatrix a
rowVectorsToMatrix cols [] = matrix (0,cols,[])
rowVectorsToMatrix _    vs = rowVectorsToMatrix1 vs

columnVectorsToMatrix :: Integral b => b -> [GVector a] -> GMatrix a
columnVectorsToMatrix rows [] = mtranspose $ matrix (0,rows,[])
columnVectorsToMatrix _    vs = columnVectorsToMatrix1 vs

-- |
-- Construct a matrix from row vectors.
-- The list of row vectors must not be empty.
--
rowVectorsToMatrix1 :: [GVector a] -> GMatrix a
rowVectorsToMatrix1 [] = error "rowVectorsToMatrix1: called with []"
rowVectorsToMatrix1 vs = matrix [vectorEntries v | v <- vs]

-- |
-- Construct a matrix from column vectors.
-- The list of colmun vectors must not be empty.
--
columnVectorsToMatrix1 :: [GVector a] -> GMatrix a
columnVectorsToMatrix1 [] = error "columnVectorsToMatrix1: called with []"
columnVectorsToMatrix1 vs = mtranspose $ rowVectorsToMatrix1 vs
    
rowVectors :: GMatrix a -> [GVector a]
rowVectors m = [vector r | r <- rows' m]

columnVectors :: GMatrix a -> [GVector a]
columnVectors m = [vector c | c <- columns' m]



foreign import ccall kernel_c :: Ptr Matrix -> IO (Ptr Matrix)
kernel :: Matrix -> [Vector]
kernel = columnVectors . unsafeCppiConvert1 kernel_c



---------


-- |
-- Apply Gauss-Jordan transformation to the input matrix @m@.
-- Return a tuple @(m',t,v)@ where @m'@ is the transformed matrix,
-- @t * m = m'@ holds and @v@ lists the columns of @m'@ which
-- contain pivots in ascending order.
-- 
gaussJordanTransform :: (Fractional a, Real a)
                     => GMatrix a
                     -> (GMatrix a, GMatrix a, [Integer])
gaussJordanTransform input =
    (subMatrixVer jordanComposite 0          (inputWidth-1),
     subMatrixVer jordanComposite inputWidth (maxColumn jordanComposite),
     takeWhile (< inputWidth) (map snd pivotsInComp))
    where
    jordanComposite = foldl fixPivot composite pivotsInComp
    pivotsInComp    = findPivots composite
    composite       = echelon `conc` transform
    (transform',echelon') = echelonReduction $ fmap toRational input
    transform = fmap fromRational transform'
    echelon   = fmap fromRational echelon'
    inputWidth = nColumns input


-- handle the pivot at (`r',`c') in `m', i.e.
-- make that coefficient equal 1 and subtract
-- respective multiples from the rows above `r' to zero
-- out the coefficients in column `c' (above `r')
fixPivot :: (Fractional a, Real a)
         => GMatrix a -> (Integer, Integer) -> GMatrix a
fixPivot m (r,c) = foldl (addRowMultiple r) one rowsAndFacts
    where
    rowsAndFacts = [(r', - entry m r' c) | r' <- [0 .. (r-1)]]
    one = multiplyRow m r (1 / entry m r c)

-- |
-- Return the (row,column) locations of the pivots in @m@
-- (assuming that @m@ is an echelon matrix)
--
findPivots :: (Num a, Integral b)
	   => GMatrix a    -- ^ Matrix @m@
	   -> [(b, b)]
findPivots m = findPivots 0 0
    where
    findPivots r c
	| r > maxRow m      =  []
	| c > maxColumn m   =  []
	| entry m r c /= 0  =  (r,c) : findPivots (r+1) (c+1)
	| otherwise         =  findPivots r (c+1)


-- add the `fact' multiple of the `r1'th row to the `r2'th row
addRowMultiple :: Num a => Integer -> GMatrix a -> (Integer, a) -> GMatrix a
addRowMultiple r1 m (r2, fact) =
    subMatrixHor m 0 (r2-1)          `join`
    ((rw1 `smul` fact) + rw2)        `join`
    subMatrixHor m (r2+1) (maxRow m)
    where
    rw1 = subMatrixHor m r1 r1
    rw2 = subMatrixHor m r2 r2
    

{-
rowEchelonReduction :: Integral a
  => GMatrix a -> (GMatrix a, GMatrix a, [Integer])
rowEchelonReduction (m :: GMatrix a) =
    runST red
    where
    red :: ST s (GMatrix a, GMatrix a, [Integer])
    red = do
      a <- newArray ((0,0),(maxR,maxC+maxR)) (entries (m `conc` unitMatrix maxR maxR))
      reduceRow a 0

    reduceRow a r
        | r > maxRow  =  return ()
        | otherwise   =
            do 
      

    maxC = maxColumn m
    maxR = maxRow    m
-}



filterRows :: ([a] -> Bool) -> GMatrix a -> GMatrix a
filterRows f m = matrix filtered
    where
    filtered = filter f (rows' m)

filterColumns :: ([a] -> Bool) -> GMatrix a -> GMatrix a
filterColumns f m = mtranspose (matrix filtered)
    where
    filtered = filter f (columns' m)


-- some functions by Martin Griebl

diagonalMatrix :: [Rational] -> Matrix
diagonalMatrix rs = matrix [ (unitVect x (length rs)) | x <- [0 .. (length rs)-1]] 
    where
    unitVect :: Int -> Int -> [Rational]
    unitVect x len = [ diag (x,i) | i <- [0.. len-1]]
    diag :: (Int, Int) -> Rational
    diag (x, y)
     | x == y    = (rs!!x)
     | otherwise = 0



unitRowVect :: (Num a, Integral b) => b -> b -> GMatrix a
unitRowVect len pos = rowMatrix [ if i==pos then 1 else 0 | i <- [0..len-1] ]



colOneMatrix :: Integer -> Integer -> Integer -> Matrix
colOneMatrix r c onecol =
    foldr join (zeroMatrix 0 c) [ (unitRowVect c onecol) | j <- [0 .. r-1]]



translationMatrix :: Integer -> [Rational] -> Matrix
translationMatrix  nConsts rs = 
    conc 
       (join 
	  (unitMatrix (size-1) (size-1)) 
          (zeroMatrix 1 (size-1)) 
	)
       (multijoin1 [
		   mtranspose(matrix [rs]) ,
		   zeroMatrix (nConsts) 1 ,
		   unitMatrix 1 1 ]
	)
	 
    where size = toInteger(length rs) + nConsts + 1



--
-- use update-in-place for matrix updates (through the ST monad);
-- should be faster for update lists with more than 2 elements.
-- Maybe `setEntries' should be defined through `setEntriesST'.
--
setEntriesST :: GMatrix a -> [(Integer,Integer,a)] -> GMatrix a
setEntriesST (m :: GMatrix a) upd =
    GMatrix { coeffs = runST update }
    where
    update :: ST s (Array (Integer,Integer) a)
    update =
	do a <- thaw (coeffs m) :: ST s (STArray s (Integer,Integer) a)
	   sequence_ [writeArray a (r,c) val | (r,c,val) <- upd]
	   unsafeFreeze a

matrixToString :: (Show ent, Integral ent) => Bool -> GMatrix (Ratio ent) -> Int -> String
matrixToString showRows matrix ihRows = (if showRows then show ihRows ++ "\n" else "")
    ++ show denom ++ "\n" ++ intMatrixToString intMatrix ihRows where
        (intMatrix, denom) = integralMatrix' matrix

intMatrixToString :: (Integral ent) => GMatrix ent -> Int -> String
intMatrixToString matrix ihRows = unlines (map showRow (take ihRows (rows' matrix)))
    where
    showRow :: Show b => [b] -> String
    showRow rs = concat (intersperse " " (map show rs))
