module LMath.Types.PeriodicNumber
    ( Periodic,
      periodic, period, periodicValues, periodicWith
    )
where

import Data.List

import LMath.Types.LClasses

newtype Periodic a = Periodic [a] deriving Show

instance Eq b => FunctorArg Periodic a b where
    fmap' f (Periodic xs) = periodic $ map f xs

periodic :: Eq a => [a] -> Periodic a
periodic = Periodic . leastPeriod

period :: Integral b => Periodic a -> b
period (Periodic xs) = genericLength xs

periodicValues :: Periodic a -> [a]
periodicValues (Periodic xs) = xs

leastPeriod :: Eq a => [a] -> [a]
leastPeriod xs =
    case shorterPeriods of
       []     ->  xs
       (s:_)  ->  take s xs
    where
    shorterPeriods = [p | p <- [1..len `div` 2],
                          len `mod` p == 0, 
                          all allEqual [everyNth p (drop off xs)
                                        | off <- [0..p-1]]
                     ]   
    len = length xs

everyNth :: Int -> [a] -> [a]
everyNth n []     = []
everyNth n (x:xs) = x : everyNth n (drop (n-1) xs)

allEqual :: Eq a => [a] -> Bool
allEqual []              = True
allEqual [x]             = True
allEqual (x:rest@(y:xs)) = x == y && allEqual rest

periodicWith :: Eq c => (a -> b -> c) -> Periodic a -> Periodic b -> Periodic c
periodicWith f (Periodic xs) (Periodic ys) = Periodic $ commonWith f xs ys

commonWith :: Eq c => (a -> b -> c) -> [a] -> [b] -> [c]
commonWith f xs ys =
    leastPeriod $ take commonPeriod $ zipWith f (cycle xs) (cycle ys)
    where
    xsLen = length xs
    ysLen = length ys
    commonPeriod = lcm xsLen ysLen

instance Eq a => Eq (Periodic a) where
    Periodic xs == Periodic ys  =  and $ commonWith (==) xs ys

instance Num a => Num (Periodic a) where
    Periodic xs + Periodic ys  =  Periodic $ commonWith (+) xs ys
    Periodic xs * Periodic ys  =  Periodic $ commonWith (*) xs ys
    negate (Periodic xs)  =  Periodic $ map negate xs
    fromInteger x = Periodic [fromInteger x]
    abs    = fmap' abs
    signum = fmap' signum

instance Fractional a => Fractional (Periodic a) where
    recip (Periodic xs) = Periodic $ map recip xs
    fromRational r = Periodic [fromRational r]

