module LMath.Types.PipQuast
  ( GQuast(..), QVector(..), Quast,
    showQQuast, showGQuast,
    mapQuast,
    pipLexmin, pipLexmax,
    mirrorQuast
  )
where

import Data.List ( intersperse )
import Data.Ratio

import LMath.Types.IneqSystem
import LMath.Types.Matrix
import LMath.Types.Vector
import CppInterface

-- |
-- A quasi-affine solution tree (quast) delivered by PIP.
-- It contains two different kinds of numbers, fractions
-- (rat) and integers (int).
--
data GQuast qvec rat int = QNoResult
                         | QList [qvec rat]
	                 | QIf   (GVector int) (GQuast qvec rat int) (GQuast qvec rat int)
                         | QNewParam Int (GVector int) int (GQuast qvec rat int)
	            deriving Show

data QVector rat = QNegInfinity Int    -- number of coefficients (including 1)
                 | QFinite      (GVector rat)
                 | QPosInfinity Int    -- number of coefficients (including 1)
                 deriving Show


showQQuast :: (Real rat, Integral int)
           => [String] -> String -> GQuast QVector rat int -> String
showQQuast paramNames newParamPrefix quast =
    showGQuast showQFn showGFn paramNames newParamPrefix quast
    where
    showQFn names (QNegInfinity _) = "-oo"
    showQFn names (QPosInfinity _) = "+oo"
    showQFn names (QFinite v) = showGFn names v

    showGFn names v
        | l == 1  =  numStr
        | l /= 1  =  '(' : numStr ++ ")/" ++ show l
        where
        numStr = concat (zipWith showTerm varCoeffs names) ++ showCoeff constCoeff
        varCoeffs1  = map toRational $ init $ vectorEntries v
        constCoeff1 = toRational $ last $ vectorEntries v
        l = foldl lcm (denominator constCoeff1) $ map denominator varCoeffs1
        newNum r = numerator r * (l `div` denominator r)
        varCoeffs   = map newNum varCoeffs1
        constCoeff  = newNum constCoeff1

    showTerm   0  name = ""
    showTerm   1  name = '+' : name
    showTerm (-1) name = '-' : name
    showTerm   c  name = showCoeff c ++ "*" ++ name

    showCoeff c
        | c >= 0  =  '+' : show c
        | c <  0  =  show c

showGQuast :: (Show int)
           => ([String] -> qvec rat -> String)
           -> ([String] -> GVector int -> String)
           -> [String] -> String -> GQuast qvec rat int -> String
showGQuast showQVecFn showGVecFn paramNames newParamPrefix quast = s quast
    where
    s QNoResult = "_|_\n"
    s (QIf v t e) =
        "if " ++ showGVecFn paramNames' v ++ " >= 0\n" ++
        indent 2 (s t) ++
        "else\n" ++
        indent 2 (s e)
    s (QList vs) = '(' : concat (intersperse ", " (map (showQVecFn paramNames') vs)) ++ ")\n"
    s (QNewParam p def denom q) =
       newParamPrefix ++ show (p-nP) ++ " := floor((" ++
       showGVecFn paramNames' def ++ ")/" ++ show denom ++ ")\n" ++
       s q

    paramNames' = paramNames ++ [newParamPrefix ++ show i | i <- [0..]]

    nP = case quast of
           (QIf v _ _)           -> vectorLength v - 1
           (QNewParam _ def _ _) -> vectorLength def - 1
           _                     -> error "showQuast: nP should not be needed"

    indent n str = unlines [replicate n ' ' ++ l | l <- lines str]

instance Functor QVector where
    fmap f (QNegInfinity n) = QNegInfinity n
    fmap f (QPosInfinity n) = QPosInfinity n
    fmap f (QFinite      v) = QFinite $ fmap f v

-- |
-- A standard quast as delivered by PIP with
-- rational and integer coefficients.
--
type Quast = GQuast GVector Rational Integer


instance CConv (GQuast QVector Rational Integer) (Ptr Quast) where
  toC = error "toC not defined for Quast"
  -- fromC :: Ptr Quast -> IO Quast
  fromC ptr = do q <- getQuast ptr
                 freeQuast_c ptr
                 return q

-- |
-- Apply functions to all entries in conditions/results/new param definitions.
--
mapQuast :: Functor qvec
         => (rat -> rat2)
         -> (int -> int2)
         -> GQuast qvec rat int
         -> GQuast qvec rat2 int2
mapQuast f g = mq
    where
    mq QNoResult             = QNoResult
    mq (QList xs)            = QList $ map (fmap f) xs
    mq (QIf c q1 q2)         = QIf (fmap g c) (mq q1) (mq q2)
    mq (QNewParam p v den q) = QNewParam p (fmap g v) (g den) (mq q)

getQuast :: Ptr Quast -> IO (GQuast QVector Rational Integer)
getQuast ptr =
     do list <- isList ptr
        addParamsFn <- addNewParams ptr
        if list
          then do l <- getLeaf ptr
                  let res =
                       case l of
                         Nothing -> QNoResult
                         Just xs -> QList xs
                  return $ addParamsFn res
          else do c <- getCond ptr
                  left  <- getThen_c ptr >>= getQuast
                  right <- getElse_c ptr >>= getQuast
                  return $ addParamsFn $ QIf c left right

addNewParams :: Ptr Quast
             -> IO (GQuast qvec Rational Integer -> GQuast qvec Rational Integer)
addNewParams ptr =
    do (ranks,vs,denoms) <- getNewParams ptr
       return $ foldl (.) id $ zipWith3 QNewParam ranks vs denoms

foreign import ccall getLeaf_c :: Ptr Quast -> IO (Ptr (Ptr Matrix))
getLeaf :: Ptr Quast -> IO (Maybe [QVector Rational])
getLeaf ptr = do pm <- getLeaf_c ptr
                 mbMs <- (fromC pm :: IO (Maybe [Matrix]))
                 return $ fmap (map m2mbV) mbMs
    where
    -- An infinity in a coordinate is represented by a 0xN matrix
    -- in the C side's result.
    m2mbV :: Matrix -> QVector Rational
    m2mbV m 
        | nRows m == 0  =  QNegInfinity $ nColumns m
        | nRows m == 1  =  QFinite      $ head $ rowVectors m
        | nRows m == 2  =  QPosInfinity $ nColumns m
        | otherwise     =  error "getLeaf: unexpected number of rows"

foreign import ccall getThen_c :: Ptr Quast -> IO (Ptr Quast)
foreign import ccall getElse_c :: Ptr Quast -> IO (Ptr Quast)

foreign import ccall getCond_c :: Ptr Quast -> IO (Ptr Matrix)
getCond :: Ptr Quast -> IO (GVector Integer)
getCond ptr = do m <- getCond_c ptr >>= fromC
                 return (integralVector1 (rowVectors m !! 0))

foreign import ccall getNewParams_c :: Ptr Quast -> IO (CTriple (Ptr CInt) (Ptr (Ptr Matrix)) (Ptr CLLong))
getNewParams :: Ptr Quast -> IO ([Int],[GVector Integer],[Integer])
getNewParams ptr = do (ranks,vs,denoms) <- getNewParams_c ptr >>= fromC
                      return (ranks, [integralVector1 (rowVectors m !! 0) | m <- vs], denoms)
                             
foreign import ccall freeQuast_c :: Ptr Quast -> IO ()

foreign import ccall isList_c :: Ptr Quast -> IO CBool
isList :: Ptr Quast -> IO Bool
isList ptr = do b <- isList_c ptr
                fromC b

foreign import ccall pip_c :: Ptr IneqSystem -> Ptr IneqSystem -> CBool -> IO (Ptr Quast)
-- |
-- Call PIP to compute the lexicographic minimum of a given
-- inequality system.
--
pip :: IneqSystem        -- ^ inequality system
    -> Maybe IneqSystem  -- ^ optional assumptions on the parameters
    -> Bool              -- ^ integral solution?
    -> GQuast QVector Rational Integer
pip = unsafeCppiConvert3 pip_c

-- |
-- Compute the lexicographic minimum of a given inequality system
-- under (optional) given constraints of the parameters.
--
pipLexmin :: RealFrac a
       => GIneqSystem a          -- ^ inequality system
       -> Maybe (GIneqSystem a)  -- ^ optional assumptions on the parameters
       -> Bool                   -- ^ integral solution?
       -> GQuast QVector a Integer
pipLexmin sys mbCtx intSol =
    mapQuast fromRational id $
    pip (fmap toRational sys) (fmap (fmap toRational) mbCtx) intSol

-- |
-- Compute the lexicographic maximum of a given inequality system
-- under (optional) given constraints of the parameters.
--
pipLexmax :: RealFrac a
       => GIneqSystem a          -- ^ inequality system
       -> Maybe (GIneqSystem a)  -- ^ optional assumptions on the parameters
       -> Bool                   -- ^ integral solution?
       -> GQuast QVector a Integer
pipLexmax sys mbCtx intSol =
    mirrorQuast $
    pipLexmin (mirrorIneqs sys) mbCtx intSol

-- |
-- Mirror the solutions at the leaves of a given quast,
-- i.e. negate the entries in the result vectors.
--
mirrorQuast :: (Num a) => GQuast QVector a b -> GQuast QVector a b
mirrorQuast QNoResult     = QNoResult
mirrorQuast (QList xs)    = QList $ map mirror xs
    where
    mirror (QFinite v)      = QFinite $ fmap negate v
    mirror (QNegInfinity n) = QPosInfinity n
    mirror (QPosInfinity n) = QNegInfinity n
mirrorQuast (QIf c q1 q2) = QIf c (mirrorQuast q1) (mirrorQuast q2)
mirrorQuast (QNewParam n def denom q) =
    QNewParam n def denom (mirrorQuast q)

{-
x = pip (ineqSystem ((1,1),matrix [[1,-1,0]])) Nothing False
y = pip (ineqSystem ((1,1),matrix [[1,-1,0]])) (Just $ ineqSystem ((0,1),matrix[[1,0]])) False
z = pip (ineqSystem ((1,1),matrix [[1,-1%2,0],[1,-1%3,-6],[-1,0,100]])) (Just $ ineqSystem ((0,1),matrix[[1,0]])) True
-}
