module LMath.Types.Polynomial.AlgNum
    ( AlgNum,
      algNum, algNumPoly, algNumRoots, algIsRational,
      addAlg, multAlg, negAlg, cmpAlg,
      floorAlg, ceilingAlg,
      toAlgNum, parToAlgnums,
      simpleAlg, normalAlg, normalAlgP, substAlg, substAlgs,
      disjointInterval
    )
where

import Control.Exception ( assert )
import Data.Ratio
import Data.List  (find, sort, tails)

import Data.IORef
import System.IO.Unsafe

import LMath.Types.AlgNum
import LMath.Types.LClasses
import LMath.Types.Polynomial.Base hiding ( gcdPoly )
import qualified LMath.Types.Polynomial.Base as PB
import LMath.Types.Polynomial.Univariate
import LMath.Types.Polynomial.Roots
import LMath.Types.Polynomial.ModArith
import qualified LMath.Types.Polynomial.Resultant as R
import LMath.Tools.Reduce
import Debug.Trace


instance RealFrac a => Alg (AlgNum a) where
    algIsRat (RatNum x) = Just $ toRational x
    algIsRat _          = Nothing

    toAlg x = toAlgNum $ fromRational $ toRational x
    zeroToAlg p n
        | n > 0  =  algs !! fromIntegral (n-1)
        | n < 0  =  reverse algs !! fromIntegral (-n-1)
        where
          algs = sort $ concatMap parToAlgnums $ rootsND [p']
          p' = fmap' (fromRational . toRational) p
    algOrdering x = compare x 0
    algSubstP algs p = [PREPoly $ fmap' fromIntegral p2]
        where
          (_,p2) = ratPoly2intPoly $ substP algs $
                   fmap' (fromRational . toRational) p

    algEvalP algs p = substAlgs (fmap' (fromRational . toRational) p) algs

    fromAlg (RatNum x) = repeat (x',x')
        where
          x' = fromRational $ toRational x
    fromAlg (AlgNum rs _) = map f rs
        where
          f (RootAt x)       = let x' = fromRational $ toRational x in (x',x')
          f (RootIn x y _ _) = (fromRational $ toRational x
                               ,fromRational $ toRational y
                               )

    algDefiningPoly = fmap' toRational . algNumPoly

algOps :: RealFrac a => AlgOps (AlgNum a)
algOps = genericAlgOps


gcdPoly :: RealFrac a => Polynomial a -> Polynomial a -> Polynomial a
gcdPoly = reduceGcdPoly

{-# NOTINLINE resCache #-}
resCache :: IORef [((Polynomial Rational, Polynomial Rational), Polynomial Rational)]
resCache = unsafePerformIO $
           newIORef []

resultant' :: Polynomial Rational -> Polynomial Rational -> IO (Polynomial Rational)
resultant' f g = atomicModifyIORef resCache res
    where
    res :: [((Polynomial Rational, Polynomial Rational), Polynomial Rational)]
        -> ([((Polynomial Rational, Polynomial Rational), Polynomial Rational)], Polynomial Rational)
    res cache = case find pred cache of
                   Just ((_,_),r)  -> trace "found" $ (cache, r)
                   Nothing         -> trace "not found" $ (((f,g),resu):cache , resu)
    pred ((f',g'),_) = nIndeterminates f == nIndeterminates f' && nIndeterminates g == nIndeterminates g' && f' == f && g' == g
    resu = R.resultant f g

resultant :: RealFrac a => Polynomial a -> Polynomial a -> Polynomial a
resultant p q = -- R.resultant p q
    -- trace (show (reduceResultant p q, R.resultant p q)) $
--    trace (show (degree p, degree q, show p, show q)) $
    reduceResultant p q
    -- R.resultant2 p q
{-
    mapPoly fromRational $
    unsafePerformIO $
    resultant' (mapPoly toRational p) (mapPoly toRational q)
-}

mytrace = flip const --trace


data RealFrac a => AlgNum a = AlgNum [Root a] (Polynomial a)
			    | RatNum a

instance (RealFrac a, Show a) => Show (AlgNum a) where
    show (AlgNum (r:_) p) = "AlgNum [" ++ show r ++ ",...] " ++ show p
    show (RatNum x)       = "RatNum " ++ show x

algNum :: RealFrac a => Root a -> Polynomial a -> AlgNum a
algNum (RootAt r) _ = RatNum r
algNum r          p
  | degree p <= 0  =  error "algNum: constant polynomial"
  | degree p == 1  =  RatNum (negate (coefficient p 0) / coefficient p 1)
  | otherwise      =  AlgNum (iterate (isolate2 p) r) p

algNumPoly :: RealFrac a => AlgNum a -> Polynomial a
algNumPoly (AlgNum rs p) = p
algNumPoly (RatNum r)    = indeterminate 1 0 - constPoly r

algNumRoots :: RealFrac a => AlgNum a -> [Root a]
algNumRoots (AlgNum rs p) = rs
algNumRoots (RatNum r)    = repeat (RootAt r)

-- |
-- Check whether an algebraic number is _represented_ as a rational
-- number. An algebraic number may be rational although
-- @Nothing@ is returned!
--
algIsRational :: RealFrac a => AlgNum a -> Maybe a
algIsRational (RatNum a) = Just a
algIsRational _          = Nothing

toAlgNum :: RealFrac a => a -> AlgNum a
toAlgNum x = RatNum x

parToAlgnums :: RealFrac a => PolyAndRoots a -> [AlgNum a]
parToAlgnums par = [algNum r (parSqfree par) | r <- parRoots par]

near :: RealFrac a => AlgNum a -> a
near (AlgNum r p) = case dropWhile dcrit ({-iterate (isolate2 p)-} r) of
		      (RootAt x:_) -> x
		      (RootIn x y _ _:_) -> (x+y)/2
    where
    dcrit (RootAt _) = False
    dcrit (RootIn x y _ _) = y-x > 1/1000
near (RatNum x) = x

smallerRational :: RealFrac a => AlgNum a -> a
smallerRational (AlgNum (RootIn x _ _ _ : _) _) = x
smallerRational (AlgNum (RootAt x : _)       _) = x - 1
smallerRational (RatNum x)                      = x - 1

biggerRational :: RealFrac a => AlgNum a -> a
biggerRational (AlgNum (RootIn _ y _ _ : _) _) = y
biggerRational (AlgNum (RootAt y : _)       _) = y + 1
biggerRational (RatNum x)                      = x + 1

rationalBetween :: RealFrac a => AlgNum a -> AlgNum a -> a
rationalBetween (RatNum x) (RatNum y) = simpleRationalBetween x y
rationalBetween (RatNum x) alg =
    rationalBetween (AlgNum (repeat $ RootAt x) undefined) alg
rationalBetween alg (RatNum y) =
    rationalBetween alg (AlgNum (repeat $ RootAt y) undefined)
rationalBetween (AlgNum roots1 _) (AlgNum roots2 _) =
    between roots1 roots2
    where
    between (RootAt x:_) (RootAt y:_) = simpleRationalBetween x y
    between (RootIn _ x _ _:r1s) (RootIn y _ _ _:r2s)
	| x <= y    = simpleRationalBetweenIncl x y
	| otherwise = between r1s r2s
    between (RootAt x:r1s) (RootIn y _ _ _:r2s)
	| x < y     = simpleRationalBetween x y
	| otherwise = between r1s r2s
    between (RootIn _ x _ _:r1s) (RootAt y:r2s)
	| x < y     = simpleRationalBetween x y
	| otherwise = between r1s r2s

simpleRationalBetween :: RealFrac a => a -> a -> a
simpleRationalBetween x y = fromRational $ approxRational ((x+y)/2) (abs (x-y) / 3)

simpleRationalBetweenIncl :: RealFrac a => a -> a -> a
simpleRationalBetweenIncl x y = fromRational $ approxRational ((x+y)/2) (abs (x-y) / 2)

floorAlg :: (RealFrac a, Integral b) => AlgNum a -> b
floorAlg (RatNum x) = floor x
floorAlg (AlgNum roots p) = findFloor roots
    where
      findFloor (RootAt x : _) = floor x
      findFloor (RootIn x y _ _ : rs)
          | y-x >= 1          =  findFloor rs
          | x <= i && i <= y  && evalPoly p [i] == 0  =  floor y
          | otherwise  = findFloor2 rs 
          where
            i = fromIntegral $ floor y

      findFloor2 (RootAt x : _) = floor x
      findFloor2 (RootIn x y _ _ : rs)
          | i < x || y < i    =  floor x
          | otherwise         =  findFloor2 rs
          where
            i = fromIntegral $ floor y

ceilingAlg :: (RealFrac a, Integral b) => AlgNum a -> b
ceilingAlg = negate . floorAlg . negate
          

addAlg :: RealFrac a => AlgNum a -> AlgNum a -> AlgNum a
addAlg (RatNum x) (RatNum y) = RatNum (x+y)
addAlg (RatNum x) (AlgNum roots p) =
    AlgNum [addToRootInterval r x | r <- roots]
           (replacePoly1 (\[e] -> (indeterminate 1 0 - constPolyN 1 x) ^ e) p)
addAlg alg x@(RatNum _) = addAlg x alg
addAlg a b = mytrace ("addAlg (" ++ show a ++ ") (" ++ show b ++")\n") $
	     arithHelper a b addFn addRootIntervals
    where
    addFn Nothing    = 1
    addFn (Just [e]) = potPoly (indeterminate 2 1 - indeterminate 2 0) e

multAlg :: RealFrac a => AlgNum a -> AlgNum a -> AlgNum a
multAlg (RatNum x) (RatNum y) = RatNum (x*y)
multAlg (RatNum 0) _          = RatNum 0
multAlg (RatNum x) (AlgNum (RootAt y:_) _) = RatNum (x*y)
multAlg (RatNum x) (AlgNum (r : _) p) =
    algNum rt d
{-
    AlgNum [multRootInterval r x | r <- roots]
           (replacePoly1 (\[e] -> (mapPoly (/x) $ indeterminate 1 0) ^ e) p)
-}
    where
    d = replacePoly1 (\[e] -> (mapPoly (/x) $ indeterminate 1 0) ^ e) p
    RootIn a b _ _ = multRootInterval r x
    rt = makeRoot d a b
multAlg alg x@(RatNum _) = multAlg x alg
multAlg a@(AlgNum _ p) b = mytrace ("multAlg (" ++ show a ++ ") (" ++ show b ++")\n") $ mytrace (": multAlg (" ++ show a ++ ") (" ++ show b ++")\n") $
			   arithHelper a b multFn multRootIntervals
    where
    degA = degree p

    multFn Nothing    = potPoly (indeterminate 2 0) degA
    multFn (Just [e]) = potPoly (indeterminate 2 1) e *
			potPoly (indeterminate 2 0) (degA-e)

potAlg :: RealFrac a => AlgNum a -> Integer -> AlgNum a
potAlg (RatNum x)    expo = RatNum (x^expo)
potAlg (AlgNum rs p) expo = algNum rt d
    where
    (d,rt) = disjointInterval inters ds
    ds = concatMap (\par -> [(parSqfree par,r) | r <- parRoots par]) $ 
	 rootsND $ (\p -> [p]) $ -- sqff' $
	 resultant (replacePoly1 (\[e] -> potPoly (indeterminate 2 0) e) p)
	           (indeterminate 2 1 - potPoly (indeterminate 2 0) expo)
    inters = map (`potRootInterval` expo) rs

substAlg :: RealFrac a => Polynomial a -> AlgNum a -> AlgNum a
substAlg poly  (RatNum x)    = RatNum $ evalPoly poly [x]
substAlg poly' (AlgNum rs p) = algNum rt d
    where
    poly = poly' `remPoly` p
    (d,rt) = disjointInterval inters ds
    ds = concatMap (\par -> [(parSqfree par,r) | r <- parRoots par]) $ 
	 roots [resu]  -- must be disjoint (we call disjointInerval later)!
    resu=resultant (indeterminate 2 1 - replacePoly1 (\[e] -> potPoly (indeterminate 2 0) e) poly)
	           (replacePoly1 (\[e] -> potPoly (indeterminate 2 0) e) p)
    inters = map evPr rs

    evPr inter = reducePoly addRootIntervals (\co (Just [e]) -> multRootInterval (potRootInterval inter e) co) (RootAt 0) poly

substAlgs :: RealFrac a =>
	     Polynomial a -> [AlgNum a] -> AlgNum a
substAlgs poly algs =
    case sequence mbRats of
      Just rats -> toAlgNum $ evalPoly poly rats
      Nothing   -> substAlg newPoly prim
    where
    newPoly = replacePoly1 (\es -> product (zipWith (^) ps es)) poly
    (prim, ps) = simpleAlg algs
    mbRats = map algIsRational algs

substP :: RealFrac a =>
	   [AlgNum a] -> Polynomial a -> Polynomial a
substP []   (poly :: Polynomial a) =
--    trace (show poly) $
    assert (totalDegree poly <= 0 || unNIndeterminates (nIndeterminates poly) == 1) $
    poly
substP algs (poly :: Polynomial a)
    -- constant polynomials need not have exactly (length algs)
    -- indeterminates;  the number of indeterminates could be,
    -- e.g., AtLeast 0.
    | Just _ <- isConstPoly poly  =  poly
    | Just rats <- mbRats  =
        fmap' (\p -> evalPoly p rats) $
        univariate poly
    | otherwise  =
        assert (length algs+1 == unNIndeterminates (nIndeterminates poly)) $
        normalAlgP newPoly prim
    where
    uni = univariate poly
    newPoly = mapPoly newcoeff uni
    newcoeff co = replacePoly1 (\exps -> product (zipWith potPoly ps exps)) co
    (prim, ps) = simpleAlg algs

    mbRats :: Maybe [a]
    mbRats = sequence $ map algIsRational algs


arithHelper :: RealFrac a =>
	       AlgNum a -> AlgNum a -> (Maybe [Int] -> Polynomial a) ->
	       (Root a -> Root a -> Root a) ->
	       AlgNum a
arithHelper (AlgNum r1s p1) (AlgNum r2s p2) f1 intervalF =
    let (d,r) = disjointInterval (intervals r1s r2s) ds
    in  algNum r d
    where
    ds = concatMap (\par -> [(parSqfree par,r) | r <- parRoots par]) $ 
	 rootsND $ (\p -> [p]) $ -- sqff' $
	 mytrace (show (replacePoly f1 p1,replacePoly helpFn p2)) $
	 resultant (replacePoly f1 p1)
		   (replacePoly helpFn p2)

    helpFn Nothing    = 1
    helpFn (Just [e]) = potPoly (indeterminate 2 0) e

    intervals (i1:i1s) (i2:i2s) = intervalF i1 i2 : intervals i1s i2s
--		      intervals (isolate2 p1 i1) (isolate2 p2 i2)

disjointInterval :: RealFrac a =>
		    [Root a] -> [(Polynomial a, Root a)] ->
		    (Polynomial a, Root a)
disjointInterval _ [] = error "disjointInterval: [] -\
			      \- error in interval arithmetic?"
disjointInterval _ [(d,RootIn a b _ _)] = (d, makeRoot d a b)
disjointInterval _ [d] = d
disjointInterval (v:vs) ds =
    mytrace ("disjointInterval: " ++ show ds) $ disjointInterval vs (filter nonDisj ds)
    where
    nonDisj (_,r) = not $ rootsAreDisjoint v r


negAlg :: RealFrac a => AlgNum a -> AlgNum a
negAlg (RatNum x)    = RatNum (negate x)
negAlg (AlgNum rs p) = AlgNum rs' p'
    where
    rs' = map (\r -> case r of
                       RootAt x           -> RootAt (-x)
	               RootIn x1 x2 s1 s2 -> RootIn (-x2) (-x1) s2 s1
	      )
	     rs
    p' = replacePoly negFn p
    
    negFn Nothing    = 1
    negFn (Just [e]) = potPoly (negate $ indeterminate 1 0) e

instance RealFrac a => Num (AlgNum a) where
    (+)    = addAlg
    (*)    = multAlg
    negate = negAlg
    signum a = case compare a 0 of
	         LT -> -1
		 EQ ->  0
		 GT ->  1
    abs a = case compare a 0 of
	      LT -> -a
	      _  ->  a
    fromInteger = toAlgNum . fromInteger


instance RealFrac a => Eq (AlgNum a) where
    a1 == a2  =  cmpAlg a1 a2 == EQ


cmpAlg :: RealFrac a => AlgNum a -> AlgNum a -> Ordering
cmpAlg (RatNum x) (RatNum y) = x `compare` y
cmpAlg (RatNum x) alg        =
    cmpAlg (AlgNum (repeat $ RootAt x) (indeterminate 1 0 - constPolyN 1 x)) alg
cmpAlg alg y@(RatNum _)      =
    case cmpAlg y alg of
      LT -> GT
      EQ -> EQ
      GT -> LT
cmpAlg (AlgNum (RootAt x:_) _) (AlgNum (RootAt y:_) _)  =  compare x y
cmpAlg (AlgNum r1s p1) (AlgNum r2s p2) =
    case r of
        Nothing           -> compare (head r1s) (head r2s)
	Just (RootAt x)   ->
	    if signumPoly g x == EQ
	       then EQ
	       else cmpI
	Just (RootIn x y _ _) ->
	    if signumPoly g x /= signumPoly g y
	       then EQ
	       else cmpI
	where
	g = gcdPoly p1 p2
	r = intersectRoots (head r1s) (head r2s)
	roots = zip r1s r2s -- (r1,r2) : [(isolate2 p1 s1, isolate2 p2 s2) | (s1,s2) <- roots]

        cmpI = case dropWhile (\(a,b) -> not (rootsAreDisjoint a b)) roots of
		          (s1,s2):_ -> compare s1 s2

instance RealFrac a => Ord (AlgNum a) where
    compare = cmpAlg


instance RealFrac a => Fractional (AlgNum a) where
    fromRational = toAlgNum . fromRational
    recip (RatNum x)    = RatNum $ recip x
    recip (AlgNum rs p) = AlgNum rs' p'
	where
	p' = reflectPoly p
	rs' = map (\r -> case r of
		           RootAt 0 -> error "recip: zero"
		           RootAt x -> RootAt (1/x)
		           RootIn a1 a2 _ _
		             | a1 <= 0 && a2 >= 0 && signumPoly p 0 == EQ ->
			                error "recip: zero"
		             | otherwise ->
			         case isolateUntil zeroOk p r of
			           RootAt x         -> RootAt (1/x)
			           RootIn x1 x2 _ _ -> makeRoot p' (1/x2) (1/x1)
		  )
	          rs
	    where
	    zeroOk (RootAt _) = True
	    zeroOk (RootIn x1 x2 _ _) = x1 > 0 || x2 < 0

---------


sq :: Rational -> AlgNum Rational
sq x = algNum (makeRoot p 0 (max x (3%2))) p
    where
    p = sqfPart (indeterminate 1 0 * indeterminate 1 0 - constPolyN 1 x)

---------

simple :: RealFrac a =>
	  AlgNum a -> AlgNum a -> (AlgNum a, Polynomial a, Polynomial a)
simple (RatNum x) (RatNum y) = (RatNum 0, constPolyN 1 x, constPolyN 1 y)
simple (RatNum x) alg        = (alg, constPolyN 1 x, indeterminate 1 0)
simple alg (RatNum y)        = (alg, indeterminate 1 0, constPolyN 1 y)
simple a1@(AlgNum r1s a) a2@(AlgNum r2s b)
    | a1 == a2   =  (a1, indeterminate 1 0, indeterminate 1 0)  
    | otherwise  =  (gamma, c1, c2)
    where
    a' = replacePoly1 (\[e] -> potPoly (indeterminate 3 1 - indeterminate 3 2 * indeterminate 3 0) e) a
    b' = replacePoly1 (\[e] -> potPoly (indeterminate 3 0) e) b
    rt = resultant a' b'

    rs = [(replacePoly1 (\[e1,e2] -> (fromIntegral t ^ e2) `pmul`
	                              potPoly (indeterminate 1 0) e1)     rt
          ,t)
         | t <- [(1::Integer)..]] 

    Just (c0,t1) = find (\(r,_) -> degree (gcdPoly r (derivative r)) == 0) rs
    stSeq = sturm c0

--    r1s = iterate (isolate2 a) r1
--    r2s = iterate (isolate2 b) r2
    ks = zipWith (\i j -> i `addRootIntervals` (multRootInterval j (fromIntegral t1))) r1s r2s

    Just k' = find f ks
	where
	f (RootAt _)       = True
	f (RootIn x y _ _) = nstIn stSeq x y == 1
    k = case k' of
	  r@(RootAt _)   -> r
	  RootIn x y _ _ -> RootIn x y (signumPoly c0 x) (signumPoly c0 y)

    gamma = algNum k c0

    g_t1x = constPolyN 1 (modArith (indeterminate 1 0) (factorWithRoot k c0) {-k-}{-r2-}) - pmul (fromIntegral t1) (indeterminate 1 0)
    a'' = replacePoly1 (\[e] -> potPoly g_t1x e) (mapPoly (\x -> modArith (liftPoly x) (factorWithRoot k c0) {-k-}{-r2-}) $ univariate a)
    b'' = mapPoly (\x -> modArith (liftPoly x) (factorWithRoot k c0) {-k-}{-r2-}) $ univariate b
    b_ast = PB.gcdPoly a'' b''
    -- b_ast has the form a*X+b and beta=b/a
    c2 = negate $ getModArith (constantCoeff b_ast / leadingCoeff b_ast)
    c1 = (indeterminate 1 0 - pmul (fromIntegral t1) c2) `remPoly` c0


normalAlg :: RealFrac a => Polynomial (AlgNum a) -> Polynomial a
normalAlg algpoly = normalAlgP p alpha
    where
    p = sum $ zipWith (\b e -> potPoly (indeterminate 1 0) e *
	                       constPoly b)
                      bbars exps	
    (betas,exps)  = unzip $ coeffsAndExps algpoly 
    (alpha,bbars) = simpleAlg betas
{-
normal algpoly
    | isZeroPoly algpoly = trace "normal:" $ constPolyN (nIndeterminates algpoly) 0
    | otherwise          = monic c
    where
    (betas,exps)  = unzip $ coeffsAndExps algpoly
    (alpha,bbars) = simpleAlg betas
    bn = head bbars
    AlgNum _ a = alpha

    d = gcdPoly a bn

    abar = a `quotPoly` d
    bbars' = map (`remPoly` abar) bbars

    (a_touse,bs) = if degree d > 0 then (abar,bbars') else (a,bbars)

    c = resultant
	  (replacePoly1 (\[e] -> potPoly (indeterminate 2 0) e) a_touse)
          (foldl1 (+) $ zipWith (\bj e -> bj * potPoly (indeterminate 2 1) e) bjs exps)
        where
	bjs = map (replacePoly1 (\[e] -> potPoly (indeterminate 2 0) e)) bs
-}

normalAlgP :: RealFrac a =>
	      Polynomial (Polynomial a) -> AlgNum a -> Polynomial a
normalAlgP algpoly (RatNum x) = mapPoly (\p -> evalPoly p [x]) algpoly
normalAlgP algpoly (AlgNum (rt:_) a)
    | unNIndeterminates (nIndeterminates algpoly) /= 1 =
        error "normalAlgP: algpoly must have exactly 1 indeterminate"
    | null bbars = constPolyNp algpoly 0
    | otherwise  = monic c
    where
    (bbars,exps) = unzip $
		   filter (\(c,_) -> modArith c (factorWithRoot rt a) {-rt-} /= 0) $
		   coeffsAndExps algpoly
    bn = head bbars

    d = gcdPoly a bn

    abar = a `quotPoly` d
    bbars' = map (`remPoly` abar) bbars

    (a_touse,bs) = if degree d > 0 then (abar,bbars') else (a,bbars)

    c = resultant
	  (replacePoly1 (\[e] -> potPoly (indeterminate 2 0) e) a_touse)
          (foldl1 (+) $ zipWith (\bj e -> bj * potPoly (indeterminate 2 1) e) bjs exps)
        where
	bjs = map (replacePoly1 (\[e] -> potPoly (indeterminate 2 0) e)) bs


simpleAlg :: RealFrac a =>
	       [AlgNum a] -> (AlgNum a, [Polynomial a])
simpleAlg []  = (1, [])
simpleAlg [x] = (x, [indeterminate 1 0])
simpleAlg (x:rest) = (alpha, xPoly:newPs)
    where
    (al,ps) = simpleAlg rest
    newPs   = map (replacePoly1 (\[e] -> potPoly aPoly e)) ps
    (alpha,aPoly,xPoly) = simple al x

---------


testA1, testA2 :: Polynomial Rational
(testA1,testA2) =
    (pmul 144 (y*y) + pmul 96 (x*x*y) + pmul 9 (x*x*x*x) + pmul 105 (x*x) +
     pmul 70 x - constPolyN 2 98,
     x*y*y + pmul 6 (x*y) + x*x*x + pmul 9 x
    )
    where
    y = indeterminate 2 0
    x = indeterminate 2 1

testB :: Polynomial Rational
testB = y*y*y*y - pmul 2 (y*y*y) + y*y - pmul 3 (x*x*y) + pmul 2 (x*x*x*x)
    where
    y = indeterminate 2 0
    x = indeterminate 2 1

testC :: Polynomial Rational
testC = y*y*y - pmul 2 (y*y) + y - pmul 3 (x*x)
    where
    y = indeterminate 2 0
    x = indeterminate 2 1

x_y :: Polynomial Rational
x_y = x - y
    where
    y = indeterminate 2 0
    x = indeterminate 2 1

testU :: Polynomial Rational
testU = x*x + y*y + z*z - 1
    where
    y = indeterminate 3 0
    x = indeterminate 3 1
    z = indeterminate 3 2

a1,a2,a3,b2,b3,c3 :: Polynomial Rational
a1 = indeterminate 1 0
a2 = indeterminate 2 1
a3 = indeterminate 3 2
b2 = indeterminate 2 0
b3 = indeterminate 3 1
c3 = indeterminate 3 0

{-
let a = NonZeroPolynomial (Recursive [(1,Recursive [(0,Constant (Alg ((-1) % 1)))]),(0,Recursive [(1,Constant (Alg (1 % 1)))])])

let p = [MAlg (AlgNum (RootIn (Alg ((-2) % 1)) (Alg (2 % 1)) LT GT) (NonZeroPolynomial (Recursive [(1,Constant (Alg (1 % 1)))])))]

replacePoly1 (\coeffs -> product (zipWith (^) p coeffs) `pmul` unknown 0 1) a

substP p a

----

let (p1,p2) = (NonZeroPolynomial (Recursive [(0,Recursive [(1,Constant (Alg (1 % 1)))])]),NonZeroPolynomial (Recursive [(1,Recursive [(0,Constant (Alg (1 % 1)))]),(0,Recursive [(0,Constant (Alg ((-1) % 1)))])]))

resultant p1 p2

----

let a = MAlg (AlgNum (RootIn (Alg ((-391) % 256)) (Alg ((-187) % 128)) GT LT) (NonZeroPolynomial (Recursive [(6,Constant (Alg (2048 % 1))),(4,Constant (Alg ((-4608) % 1))),(2,Constant (Alg (37 % 1))),(0,Constant (Alg (12 % 1)))])))

let b = MAlg (AlgNum (RootIn (Alg ((-15874991) % 67108864)) (Alg ((-7937487) % 33554432)) LT GT) (NonZeroPolynomial (Recursive [(6,Constant (Alg (2048 % 1))),(4,Constant (Alg ((-4608) % 1))),(2,Constant (Alg (37 % 1))),(0,Constant (Alg (12 % 1)))])))

a+b


----------

gcdPoly (NonZeroPolynomial (Recursive [(2,Constant (ModArith (NonZeroPolynomial (Recursive [(0,Constant (-1 % 2))])) (NonZeroPolynomial (Recursive [(4,Constant (1 % 64)),(2,Constant (-3 % 16)),(0,Constant (1 % 16))])))),(1,Constant (ModArith (NonZeroPolynomial (Recursive [(1,Constant (1 % 1))])) (NonZeroPolynomial (Recursive [(4,Constant (1 % 64)),(2,Constant (-3 % 16)),(0,Constant (1 % 16))])))),(0,Constant (ModArith (NonZeroPolynomial (Recursive [(2,Constant (-1 % 2)),(0,Constant (1 % 1))])) (NonZeroPolynomial (Recursive [(4,Constant (1 % 64)),(2,Constant (-3 % 16)),(0,Constant (1 % 16))]))))])) (NonZeroPolynomial (Recursive [(2,Constant (ModArith (NonZeroPolynomial (Recursive [(0,Constant (-1 % 4))])) (NonZeroPolynomial (Recursive [(4,Constant (1 % 64)),(2,Constant (-3 % 16)),(0,Constant (1 % 16))])))),(0,Constant (ModArith (NonZeroPolynomial (Recursive [(0,Constant (1 % 1))])) (NonZeroPolynomial (Recursive [(4,Constant (1 % 64)),(2,Constant (-3 % 16)),(0,Constant (1 % 16))]))))]))

----

let a = AlgNum (RootIn (-2 % 1) (-1 % 1) LT GT) (NonZeroPolynomial (Recursive [(2,Constant (1 % 1)),(1,Constant (2 % 1)),(0,Constant (77 % 81))])) 
let b = AlgNum (RootIn (-2 % 1) (-1 % 1) LT GT) (NonZeroPolynomial (Recursive [(2,Constant (1 % 1)),(1,Constant (2 % 1)),(0,Constant (77 % 81))]))
multAlg a b

----

let a = AlgNum (RootIn (-1074028 % 531441) (-268507 % 177147) GT LT) (NonZeroPolynomial (Recursive [(4,Constant (1 % 1)),(3,Constant (340 % 81)),(2,Constant (3233230 % 531441)),(1,Constant (155221220 % 43046721)),(0,Constant (208422380089 % 282429536481))]))
let b = AlgNum (RootIn (83 % 81) (166 % 81) LT GT) (NonZeroPolynomial (Recursive [(2,Constant (1 % 1)),(1,Constant (-170 % 81)),(0,Constant (5929 % 6561))]))
a+b

----


let p = NonZeroPolynomial (Recursive [(8,Constant (1 % 1)),(6,Constant (-976144 % 531441)),(5,Constant (87040 % 531441)),(4,Constant (70238386208 % 94143178827)),(3,Constant (-5933342720 % 282429536481)),(2,Constant (-10617127727616256 % 150094635296999121)),(1,Constant (48955531018240 % 150094635296999121)),(0,Constant (83387401476414112000 % 79766443076872509863361))])
gcdPoly p (derivative p)


----

let algpoly = NonZeroPolynomial (Recursive [(2,Constant (AlgNum (RootIn ((-3) % 1) (3 % 1) LT GT) (NonZeroPolynomial (Recursive [(1,Constant (1 % 1)),(0,Constant ((-1) % 1))])))),(0,Constant (AlgNum (RootIn (3 % 7) (5 % 1) LT GT) (NonZeroPolynomial (Recursive [(2,Constant (1 % 1)),(1,Constant ((-3) % 1))]))))])

normal algpoly

---

let p = AlgNum (RootIn ((-3) % 1) (3 % 1) LT GT) (NonZeroPolynomial (Recursive [(1,Constant (1 % 1)),(0,Constant ((-1) % 1))]))

recip p  --   GT GT !!!



Nichttermination bei:
cad [NonZeroPolynomial (Recursive [(1,Recursive [(0,Constant ((-1) % 1))]),(0,Recursive [(1,Constant (1 % 1))])]),NonZeroPolynomial (Recursive [(1,Recursive [(0,Constant (1 % 1))]),(0,Recursive [(0,Constant ((-1) % 1))])]),NonZeroPolynomial (Recursive [(0,Recursive [(1,Constant (1 % 1)),(0,Constant ((-1) % 1))])]),NonZeroPolynomial (Recursive [(0,Recursive [(1,Constant (1 % 1)),(0,Constant (1 % 1))])])]
--> simple (AlgNum (RootIn ((-3) % 1) ((-1) % 3) GT LT) (NonZeroPolynomial (Recursive [(2,Constant (1 % 1)),(1,Constant (1 % 1))]))) (AlgNum (RootIn ((-3) % 1) (3 % 1) LT GT) (NonZeroPolynomial (Recursive [(1,Constant (1 % 1)),(0,Constant (1 % 1))])))
--> gcdPoly (NonZeroPolynomial (Recursive [(2,Constant (ModArith (NonZeroPolynomial (Recursive [(0,Constant (1 % 1))])) (NonZeroPolynomial (Recursive [(2,Constant (1 % 1)),(1,Constant (3 % 1)),(0,Constant (2 % 1))])) (RootIn ((-3) % 1) (3 % 1) LT GT))),(1,Constant (ModArith (NonZeroPolynomial (Recursive [(1,Constant ((-2) % 1)),(0,Constant ((-1) % 1))])) (NonZeroPolynomial (Recursive [(2,Constant (1 % 1)),(1,Constant (3 % 1)),(0,Constant (2 % 1))])) (RootIn ((-3) % 1) (3 % 1) LT GT)))])) (NonZeroPolynomial (Recursive [(1,Constant (ModArith (NonZeroPolynomial (Recursive [(0,Constant (1 % 1))])) (NonZeroPolynomial (Recursive [(2,Constant (1 % 1)),(1,Constant (3 % 1)),(0,Constant (2 % 1))])) (RootIn ((-3) % 1) (3 % 1) LT GT))),(0,Constant (ModArith (NonZeroPolynomial (Recursive [(0,Constant (1 % 1))])) (NonZeroPolynomial (Recursive [(2,Constant (1 % 1)),(1,Constant (3 % 1)),(0,Constant (2 % 1))])) (RootIn ((-3) % 1) (3 % 1) LT GT)))]))
--> geloest (intervall in 'simple' muss 'k' sein statt 'r2')


*** Exception: matrix: number of rows and columns does not match the list of entries.
cad [NonZeroPolynomial (Recursive [(1,Recursive [(0,Constant ((-1) % 1))]),(0,Recursive [(1,Constant (1 % 1))])]),NonZeroPolynomial (Recursive [(1,Recursive [(0,Constant (1 % 1))]),(0,Recursive [(0,Constant ((-1) % 1))])]),NonZeroPolynomial (Recursive [(0,Recursive [(1,Constant (1 % 1)),(0,Constant ((-1) % 1))])]),NonZeroPolynomial (Recursive [(2,Recursive [(0,Constant ((-1) % 1))]),(1,Recursive [(1,Constant (1 % 1)),(0,Constant ((-1) % 1))]),(0,Recursive [(1,Constant (1 % 1))])])]
--> c1 und c2 in 'simple' waren vertauscht


extreme Ineffizienz
cad [NonZeroPolynomial (Recursive [(1,Recursive [(0,Constant ((-1) % 1))]),(0,Recursive [(1,Constant (1 % 1))])]),NonZeroPolynomial (Recursive [(1,Recursive [(0,Constant (1 % 1))]),(0,Recursive [(0,Constant ((-1) % 1))])]),NonZeroPolynomial (Recursive [(0,Recursive [(1,Constant (1 % 1)),(0,Constant ((-1) % 1))])]),NonZeroPolynomial (Recursive [(3,Recursive [(1,Constant ((-8) % 1)),(0,Constant ((-8) % 1))]),(2,Recursive [(1,Constant ((-24) % 1)),(0,Constant ((-24) % 1))]),(1,Recursive [(3,Constant (8 % 1)),(2,Constant (24 % 1)),(0,Constant ((-16) % 1))]),(0,Recursive [(3,Constant (8 % 1)),(2,Constant (24 % 1)),(1,Constant (16 % 1))])])]
--> nomalP und substAlg geschrieben


extreme Ineffizienz
let { poly = NonZeroPolynomial (Recursive [(14,Constant ((-1075) % 6402060288)),(13,Constant ((-7525) % 1067010048)),(12,Constant ((-125831) % 800257536)),(11,Constant ((-209849) % 88917504)),(10,Constant ((-25359) % 971776)),(9,Constant ((-6576545) % 29639168)),(8,Constant ((-21959661) % 14819584)),(7,Constant ((-29034657) % 3704896)),(6,Constant ((-485513791) % 14819584)),(5,Constant ((-60987123) % 569984)),(4,Constant ((-56992643) % 213744)),(3,Constant ((-901773283) % 1852448)),(2,Constant ((-2247997365) % 3704896)),(1,Constant ((-835573707) % 1852448)),(0,Constant ((-133874829) % 926224))]); p = NonZeroPolynomial (Recursive [(16,Constant (1 % 1)),(15,Constant (48 % 1)),(14,Constant (1200 % 1)),(13,Constant (20160 % 1)),(12,Constant (250608 % 1)),(11,Constant (2417472 % 1)),(10,Constant (18569664 % 1)),(9,Constant (115153920 % 1)),(8,Constant (579708000 % 1)),(7,Constant (2366938368 % 1)),(6,Constant (7782324480 % 1)),(5,Constant (20316883968 % 1)),(4,Constant (41137106688 % 1)),(3,Constant (62163127296 % 1)),(2,Constant (65593857024 % 1)),(1,Constant (42466664448 % 1)),(0,Constant (12274073856 % 1))]) } in resultant (indeterminate 2 1 - replacePoly1 (\[e] -> potPoly (indeterminate 2 0) e) poly) (replacePoly1 (\[e] -> potPoly (indeterminate 2 0) e) p)
-}


