-- #hide
module LMath.Types.Polynomial.Base
    ( Polynomial(..), NZP(..),
      NIndeterminates(..), unNIndeterminates, addNIndeterminates,
      addToNIndeterminates, canExtendToNI,
      mapPoly,
      nIndeterminates, nIndeterminatesN,
      appendIndeterminates, insertIndeterminates, reverseIndeterminates,
      swapIndeterminates,
      liftPoly, liftPolyN, isZeroPoly, isConstPoly,
      deleteIndeterminates, deleteEndIndeterminates,
      ensureNIndets, ensureEndNIndets,
      indeterminate, indeterminateNI,
      constPoly, constPolyN, constPolyNI, constPolyNp,
      usedIndeterminates, levelPoly, invlevelPoly,
      coefficients, linearCoeffs,

      negPoly, addPoly, multPoly, quotPoly, remPoly, quotRemPoly, potPoly, pmul,
      gcdPoly, lcmPoly, euclid, extEuclid, gcdFree,

      evalPoly, replacePoly, replacePoly1, reducePoly,
      monomials,

      renamePolyVariable, createPolyFromList, polyToList,
      affinePolyFromList,
      indeterminateToFront,

      ratPoly2intPoly, univariate, flatPoly, splitPoly, affineToPoly,

      reductum, reducta,

      leadingCoeff, constantCoeff, collectCoeffs, 
      monic, signMonic, content, degree, degreeIn, totalDegree, derivative,
      constPolySignum,

      FracPoly, mapFP, applyFP, polyNumerator, polyDenominator,
      (%%), isConstFP, intFracPoly, fromPoly, flatFracPoly,
      nIndeterminatesFP, evalFP,

      IFracPoly, (%%%), ifpNumerator, ifpDenominator,
      fracPolyToIFracPoly, ifracPolyToFracPoly, flatIFracPoly,
      applyIFP, mapIFP, nIndeterminatesIFP,
      isConstIFP,

      showU
    )
      
where

import Control.Exception ( assert )
import Control.Monad
import Control.Monad.ST
import Data.Array
import Data.Array.ST
import Data.List
import Data.Maybe ( fromMaybe )
import Data.Ratio

import LMath.Types.LClasses
import LMath.Types.Signum

import Debug.Trace

-- for debugging
mytrace :: String -> a -> a
mytrace str io = {- trace str -} io

-- |
-- Representation of multivariate polynomials having coefficients
-- of type @coeff@.
-- The indeterminates of a polynomial are numbered @0, 1, 2, ..., n-1@, where
-- @n@ is the total number of indeterminates in the polynomial
-- (see 'indeterminate').
--
-- Note that the number of indeterminates in a polynomial
-- (see 'nIndeterminates') need not be fixed,
-- but /may/ be specified as \"at least ...\". For example, if a literal
-- (like @5@) appears in an expression with type @Polynomial a@,
-- the number of indeterminates is unknown: @nIndeterminates 5 == AtLeast 0@.
-- The arithmetic operations take care of appending additional
-- indeterminates to polynomials with an \"AtLeast\" specification of
-- the number of indeterminates such that the number of indeterminates
-- matches.
--
data Polynomial coeff = ZeroPolynomial    !MayExtend !Int
		      | NonZeroPolynomial !MayExtend !(NZP coeff)

type MayExtend = Bool

instance Show a => Show (Polynomial a) where
    showsPrec = showU showsPrec

instance Integral a => Show (Polynomial (Ratio a)) where
    showsPrec = showU showR
	where
	showR prec r
	    | d == 1    = showsPrec prec n
	    | otherwise = showsPrec prec n . showString "/" . showsPrec prec d
	    where
	    n = numerator r
	    d = denominator r

data Underscore = Underscore

instance Show Underscore where
    showsPrec _ Underscore = showChar '_'

showPolyStructure :: Polynomial a -> ShowS
showPolyStructure (ZeroPolynomial me n) =
    showString "ZeroPolynomial " . shows me . showString " " . shows n
showPolyStructure (NonZeroPolynomial me nzp) =
    showString "NonZeroPolynomial " . shows me . showString " " .
    shows (fmap (const Underscore) nzp)


--
-- Representation of a non-zero (multivariate) polynomial in
-- LEXICOGRAPHIC-RECURSIVE REPRESENTATION !
--
data NZP a = Recursive ![(Int, NZP a)]
	   | Constant !a
	     deriving (Eq,Show)

instance (Num a, Eq a) => Eq (Polynomial a) where
    (==) = unFrom2 "(==)" (===)
	where
	ZeroPolynomial True _ === ZeroPolynomial True _  =  True
	ZeroPolynomial _    x === ZeroPolynomial _    y
	    | x == y  =  True
	    | otherwise = error "(==): number of indeterminates does not match"

	p@(NonZeroPolynomial _ nzp1) === q@(NonZeroPolynomial _ nzp2) =
	    nzp1 == nzp2

	ZeroPolynomial    _ _ === NonZeroPolynomial _ _  =  False
	NonZeroPolynomial _ _ === ZeroPolynomial    _ _  =  False


instance (Num b) => FunctorArg Polynomial a b where
    fmap' = mapPoly

-- |
-- Return the indexes of the indeterminates which occur in
-- the given polynomial (with a non-zero coefficient).
-- The returned list is in ascending order.
--
usedIndeterminates :: Integral b => Polynomial a -> [b]
usedIndeterminates (ZeroPolynomial _ _) = []
usedIndeterminates (NonZeroPolynomial _ nzp) = sort $ ui 0 nzp
    where
    ui n (Constant _) = []
    ui n (Recursive [(0,nzp')]) = ui (n+1) nzp'
    ui n (Recursive rs) =
	n : foldl union [] [ui (n+1) nzp' | (_,nzp') <- rs]

-- |
-- Return the index of the used variable with the least index.
-- Only defined for non-constant polynomials.
--
levelPoly :: (Num a, Integral b) => Polynomial a -> b
levelPoly p = case usedIndeterminates p of
                []   -> error "levelPoly: constant polynomial"
                u:us -> u

-- |
-- Return the index of the used variable with the highest index.
-- Only defined for non-constant polynomials.
--
invlevelPoly :: (Num a, Integral b) => Polynomial a -> b
invlevelPoly p = case usedIndeterminates p of
                [] -> error "invlevelPoly: constant polynomial"
                us -> last us

instance Functor NZP where
    fmap f (Constant  x ) = Constant  (f x)
    fmap f (Recursive rs) = Recursive [(e, fmap f r) | (e,r) <- rs]

mapPoly :: Num b => (a -> b) -> Polynomial a -> Polynomial b
mapPoly f (ZeroPolynomial me n) = ZeroPolynomial me n
mapPoly f (NonZeroPolynomial me nzp) = normalizeNZP me $ fmap f nzp

showU :: (Int -> a -> ShowS) -> Int -> Polynomial a -> ShowS
showU showCo prec (ZeroPolynomial _ _) = showString "0"
showU showCo prec (NonZeroPolynomial _ nzp) =
    showNZP prec nzp 0
    where
    showNZP prec (Constant x) _ = showCo prec x
    showNZP prec (Recursive [(0,n)]) unk = showNZP prec n (unk+1)
    showNZP prec (Recursive rs) unk =
	showParen (if length rs == 1 then prec>7 else prec>6) $
	foldl (.) id $
	intersperse (showString "+")
	[sE e (showNZP (if e == 0 then 6 else 7) n (unk+1)) | (e,n) <- rs]

	where
	sE 0 sub = sub
	sE 1 sub = showString "u" . showsPrec 0 unk . showString "*" . sub
	sE e sub = showString "u" . showsPrec 0 unk . showString "^" .
		   showsPrec 0 e . showString "*" . sub



normalizeNZP :: Num a => MayExtend -> NZP a -> Polynomial a
normalizeNZP me p =
    case normalize p of
        Constant  0  -> ZeroPolynomial me 0
	Recursive []
	    | me         -> ZeroPolynomial me 0
	    | otherwise  -> ZeroPolynomial False (nIndeterminatesNZP p)
	nonZero      -> NonZeroPolynomial me nonZero
    where
    normalize (Recursive ms) =
	Recursive $ filter nn [(exp, normalize p) | (exp,p) <- ms]
    normalize c = c

    nn (_, Constant 0)   = False
    nn (_, Recursive []) = False
    nn _                 = True



-- |
-- Add one indeterminate to a polynomial.
-- @liftPoly@ is an abbreviation for @liftPolyN 1@.
-- See also 'liftPolyN'.
--
liftPoly :: Polynomial a -> Polynomial a
liftPoly (ZeroPolynomial    me u) = ZeroPolynomial    me (u+1)
liftPoly (NonZeroPolynomial me p) = NonZeroPolynomial me (Recursive [(0,p)])


-- |
-- Adds indeterminates to a polynomial.
-- @liftPoly@ increases the number of indeterminates of a given
-- polynomial @p@ by @n@. The new indeterminates are added
-- in the front, i.e.
-- @
--     liftPolyN 2 (indeterminate 2 0 + indeterminate 2 1)
--  ==
--     indeterminate 4 2 + indeterminate 4 3
-- @
-- (and not @indeterminate 4 0 + indeterminate 4 1@).
--
liftPolyN :: Integral b
          => b              -- ^ n
          -> Polynomial a   -- ^ p
          -> Polynomial a
liftPolyN 0 p = p
liftPolyN 1 p = liftPoly p
liftPolyN n p = liftPolyN (n-1) (liftPoly p)


ensureNIndets :: Integral b => b -> Polynomial a -> Polynomial a
ensureNIndets n p
    | toLift > 0  =  liftPolyN toLift p
    | otherwise   =  p
    where
    toLift = n - unNIndeterminates (nIndeterminates p)

ensureEndNIndets :: (Num a, Integral b) => b -> Polynomial a -> Polynomial a
ensureEndNIndets n p
    | toAppend > 0  =  appendIndeterminates toAppend p
    | otherwise     =  p
    where
    toAppend = n - unNIndeterminates (nIndeterminates p)

-- |
-- Check if a polynomial is the 0-polynomial.
-- @isZeroPoly p@ has the same result as @p == 0@, but
-- @isZeroPoly p@ is slightly faster.
--
isZeroPoly :: Num a => Polynomial a -> Bool
isZeroPoly (ZeroPolynomial    _ _) = True
isZeroPoly (NonZeroPolynomial _ _) = False

-- |
-- Check if a given polynomial is constant.
-- If the given polynomial is a constant @c@, the value @Just c@
-- is returned. Otherwise (i.e., the polynomial is not constant)
-- the value @Nothing@ is returned.
-- 
isConstPoly :: Num a => Polynomial a -> Maybe a
isConstPoly (ZeroPolynomial _ _) = Just 0
isConstPoly (NonZeroPolynomial _ nzp) = isConstNZP nzp

isConstNZP :: Num a => NZP a -> Maybe a
isConstNZP (Constant c)        = Just c
isConstNZP (Recursive [(0,r)]) = isConstNZP r
isConstNZP _                   = Nothing

-- |
-- Construct a polynomial representing the @n@-th indeterminate in
-- a polynomial ring of @nIndet@ indeterminates.
--
indeterminate :: (Num a, Integral b) =>
		 b        -- ^ total number @nIndet@ of indeterminates
	      -> b        -- ^ the number @n@ of the indeterminate
	      -> Polynomial a
indeterminate nIndet n
    | n >= 0 && nIndet > n = NonZeroPolynomial False (indeterminateNZP nIndet n)
    | otherwise =
	error ("indeterminate: `indeterminate " ++ show nIndet ++ " " ++ show n ++
	       "' undefined")

indeterminateNI :: (Num a, Integral b)
		=> NIndeterminates
		-> b
		-> Polynomial a
indeterminateNI (Exactly nIndet) n = indeterminate (fromIntegral nIndet) n
indeterminateNI (AtLeast nIndet) n' =
    NonZeroPolynomial True $ indeterminateNZP nn n
    where
    nn = max nIndet (n + 1)
    n  = fromIntegral n'

indeterminateNZP :: (Integral b, Num a) => b -> b -> NZP a
indeterminateNZP 0 (-1) = Constant 1
indeterminateNZP u (-1) = Recursive [(0,indeterminateNZP (u-1) (-1))]
indeterminateNZP u   0  = Recursive [(1,indeterminateNZP (u-1) (-1))]
indeterminateNZP u   n  = Recursive [(0,indeterminateNZP (u-1) (n-1))]


-- |
-- Returns the number of indeterminates of a list of polynomials.
-- If the number of indeterminates is unknown for /every/ polynomial
-- in the given list, then @Nothing@ is returned. Otherwise, @Just n@,
-- where @n@ is the number of indeterminates of the first polynomial
-- with a known number of indeterminates, is returned.
-- See also 'indeterminate'.
--
nIndeterminatesN :: [Polynomial a] -> NIndeterminates
nIndeterminatesN ps = maximum (AtLeast 0 : map nIndeterminates ps)

data NIndeterminates = Exactly Integer
		     | AtLeast Integer
		       deriving (Read, Show)

addNIndeterminates :: NIndeterminates -> NIndeterminates -> NIndeterminates
addNIndeterminates (Exactly m) (Exactly n) = Exactly (m+n)
addNIndeterminates (AtLeast m) (Exactly n) = AtLeast (m+n)
addNIndeterminates (Exactly m) (AtLeast n) = AtLeast (m+n)
addNIndeterminates (AtLeast m) (AtLeast n) = AtLeast (m+n)

addToNIndeterminates :: Integral b => b -> NIndeterminates -> NIndeterminates
addToNIndeterminates m (Exactly n) = Exactly (fromIntegral m + n)
addToNIndeterminates m (AtLeast n) = AtLeast (fromIntegral m + n)

unNIndeterminates :: Integral a => NIndeterminates -> a
unNIndeterminates (Exactly a) = fromIntegral a
unNIndeterminates (AtLeast a) = fromIntegral a

instance Eq NIndeterminates where
    Exactly a == Exactly b  =  a == b
    AtLeast a == AtLeast b  =  a == b
    _ == _  =  False

instance Ord NIndeterminates where
    compare (Exactly a) (Exactly b) = compare a b
    compare (AtLeast a) (AtLeast b) = compare a b
    compare (AtLeast a) (Exactly b)
	| a == b     =  LT
	| otherwise  =  compare a b
    compare (Exactly a) (AtLeast b)
	| a == b     = GT
	| otherwise  = compare a b

-- |
-- Check whether polynomials with @a@ as their number of
-- indeterminates are allowed to extend their number of
-- indeterminates to @b@, i.e., @a <= b@ and @a@ is of type
-- @AtLeast@, or @a == b@ and @a@ is of type @Exactly@.
canExtendToNI :: NIndeterminates  -- ^ a
              -> NIndeterminates  -- ^ b
              -> Bool
canExtendToNI (AtLeast a) b  =  a <= unNIndeterminates b
canExtendToNI (Exactly a) b  =  a == unNIndeterminates b

-- |
-- Determine the number of indeterminates of a polynomial.
-- @nIndeterminates@ returns @Just n@ if the number of indeterminates
-- of a given polynomial is known. If the number is unknown
-- (see 'Polynomial') @Nothing@ is returned.
--
nIndeterminates :: Polynomial a -> NIndeterminates
nIndeterminates (ZeroPolynomial True  u)    = AtLeast $ fromIntegral u
nIndeterminates (ZeroPolynomial False u)    = Exactly $ fromIntegral u
nIndeterminates (NonZeroPolynomial False p) = Exactly $ nIndeterminatesNZP p
nIndeterminates (NonZeroPolynomial True  p) = AtLeast $ nIndeterminatesNZP p

nIndeterminatesNZP :: Integral b => NZP a -> b
nIndeterminatesNZP (Constant _) = 0
nIndeterminatesNZP (Recursive ((_,p):_)) = 1 + nIndeterminatesNZP p
nIndeterminatesNZP (Recursive []) =
    error ("nIndeterminatesNZP: illegal argument: Recursive [] ")

-- x^2(y+2) + (3y^2+4y-5)
testPoly :: Polynomial Integer
testPoly = NonZeroPolynomial False $
	   Recursive [(2, Recursive [(1,Constant 1), (0,Constant 2)]),
		      (0, Recursive [(2,Constant 3), (1,Constant 4), (0,Constant (-5))])]

unk1 :: Polynomial Integer
unk1 = NonZeroPolynomial False $ Recursive [(1,Constant 1)]

p1, p2 :: Polynomial Integer
p1 = NonZeroPolynomial False $
     Recursive [(2, Recursive [(0,Constant 1)]),
		(1, Recursive [(1,Constant (-2)), (0,Constant 1)]),
		(0, Recursive [(1,Constant (-2))])]
p2 = NonZeroPolynomial False $
     Recursive [(1, Recursive [(2,Constant 1), (0,Constant (-3))]),
		(0, Recursive [(3,Constant (-2)), (1, Constant 6)])]

q1, q2 :: Polynomial Integer
q1 = NonZeroPolynomial False $
     Recursive [(8,Constant 1),(6,Constant 1),(4,Constant (-3)),
		(3,Constant (-3)),(2,Constant 8),(1,Constant 2),
		(0,Constant (-5))]
q2 = NonZeroPolynomial False $
     Recursive [(6,Constant 3),(4,Constant 5),(2,Constant (-4)),
		(1,Constant (-9)),(0,Constant 21)]

a :: Polynomial Integer
a = NonZeroPolynomial False $
    Recursive [(1, Recursive [(1, Constant 1)]),
	       (0, Recursive [(0, Constant 1)])]


instance Num a => Num (Polynomial a) where
    (+) = addPoly
    (*) = multPoly
    negate = negPoly

    fromInteger 0 = ZeroPolynomial True 0
    fromInteger x = NonZeroPolynomial True $ Constant $ fromInteger x

    abs = error "abs: undefined in instance Num a => Num (Polynomial a)"
    signum = error "signum: undefined in instance Num a => Num (Polynomial a)"

unFrom2 :: Num a
	=> String
	-> (Polynomial a -> Polynomial a -> b)
        -> Polynomial a
        -> Polynomial a
        -> b
unFrom2 name f p@(NonZeroPolynomial me1 nzp1) q@(NonZeroPolynomial me2 nzp2)
    | me1 && ind1 < ind2 =
	f (NonZeroPolynomial me $ appendIndetsNZP nzp1 (ind2-ind1)) q
    | me2 && ind2 < ind1 =
	f p (NonZeroPolynomial me $ appendIndetsNZP nzp2 (ind1-ind2))
    | ind1 /= ind2  =  unFrom2err name ind1 ind2 p q
    | otherwise = f p q
    where
    me = me1 && me2
    ind1 = nIndeterminatesNZP nzp1
    ind2 = nIndeterminatesNZP nzp2
unFrom2 name f p@(ZeroPolynomial me1 n1) q@(ZeroPolynomial me2 n2)
    | me1 && n1 < n2 = f (ZeroPolynomial me n2) q
    | me2 && n2 < n1 = f p (ZeroPolynomial me n1)
    | n1 /= n2  =  unFrom2err name n1 n2 p q
    | otherwise      = f p q
    where
    me = me1 && me2
unFrom2 name f p@(NonZeroPolynomial me1 nzp1) q@(ZeroPolynomial me2 n2)
    | me1 && ind1 < n2 =
	f (NonZeroPolynomial me $ appendIndetsNZP nzp1 (n2-ind1)) q
    | me2 && n2 < ind1 =
	f p (ZeroPolynomial me ind1)
    | ind1 /= n2  =  unFrom2err name ind1 n2 p q
    | otherwise = f p q
    where
    me = me1 && me2
    ind1 = fromIntegral $ nIndeterminatesNZP nzp1
unFrom2 name f p@(ZeroPolynomial me1 n1) q@(NonZeroPolynomial me2 nzp2)
    | me1 && n1 < ind2 =
	f (ZeroPolynomial me ind2) q
    | me2 && ind2 < n1 =
	f p (NonZeroPolynomial me $ appendIndetsNZP nzp2 (n1-ind2))
    | n1 /= ind2  =  unFrom2err name n1 ind2 p q
    | otherwise = f p q
    where
    me = me1 && me2
    ind2 = nIndeterminatesNZP nzp2

unFrom2err :: String
           -> Int
           -> Int
	   -> Polynomial b
	   -> Polynomial b
           -> a
unFrom2err name ind1 ind2 p q=
    error ((showString name .
	    showString ": number of indeterminates does not match;\n\
		       \first operand has " .
	    shows ind1 .
	    showString " indeterminate(s),\nsecond operand has " .
	    shows ind2 .
	    showString " indeterminate(s).\nOperand structure:\n" .
	    showPolyStructure p .
	    showString "\n" .
	    showPolyStructure q
	   ) "")

appendIndeterminates :: (Num a, Integral b)
		     => b
		     -> Polynomial a
		     -> Polynomial a
appendIndeterminates n (ZeroPolynomial me m) =
    ZeroPolynomial me (fromIntegral n + m)
appendIndeterminates n (NonZeroPolynomial me nzp) =
    NonZeroPolynomial me $ appendIndetsNZP nzp (fromIntegral n)

appendIndetsNZP :: NZP a -> Int -> NZP a
appendIndetsNZP c@(Constant _) n
    | n >= 0  =  app n
    | n <  0  =  error "appendIndetsNZP: n<0"
    where
    app 0 = c
    app i = Recursive [(0, app (i-1))]
appendIndetsNZP (Recursive rs) n =
    Recursive [(e,appendIndetsNZP nzp n) | (e,nzp) <- rs]

-- |
-- Insert indeterminates in a polynomial.
--
insertIndeterminates :: (Num a, Integral b)
                     => b           -- ^ where to insert (inserts in front of this position)
                     -> b           -- ^ how many
                     -> Polynomial a
                     -> Polynomial a
insertIndeterminates pos count p =
    replacePoly1 ins p
    where
    ins es = product [indeterminateNI ni i ^ e | (e,i) <- zip (concat [es1,zeros,es2]) [0..]]
        where
        (es1,es2) = genericSplitAt pos es
    zeros = genericReplicate count 0
    ni = addToNIndeterminates count $ nIndeterminates p

-- |
-- Deletes the first @n@ indeterminates of the polynomial @p@.
--
-- Example:
-- @deleteIndeterminates 1 (indeterminate 2 1) == indeterminate 1 0@ is true
-- @deleteIndeterminates 1 (indeterminate 2 0) == indeterminate 1 0@ is false
--
-- See also: 'deleteEndIndeterminates'
--
deleteIndeterminates :: (Num a, Integral b)
		     => b -- ^ @n@
		     -> Polynomial a -- ^ @p@
		     -> Polynomial a
deleteIndeterminates n (ZeroPolynomial True  n') =
    ZeroPolynomial True (max 0 (n' - fromIntegral n))
deleteIndeterminates n (ZeroPolynomial False n') =
    assert (fromIntegral n <= n') $
    ZeroPolynomial False (n' - fromIntegral n)
deleteIndeterminates n' (NonZeroPolynomial me nzp') = di n' nzp'
    where
    di 0 nzp          = normalizeNZP me nzp
    di _ (Constant _) = assert me $ ZeroPolynomial me 0
    di n (Recursive r)
	| (0,nzp) <- last r = di (n-1) nzp
	| me  =  ZeroPolynomial True (max 0 (nI - fromIntegral n'))
	| otherwise = assert (fromIntegral n' <= nI) $
		      ZeroPolynomial False (nI - fromIntegral n')
    nI = nIndeterminatesNZP nzp'

-- |
-- Deletes the last @n@ indeterminates of the polynomial @p@.
--
-- Preconditions:
--
--  1) The last @n@ indeterminates may not occur in @p@.
--
-- Example:
-- @deleteEndIndeterminates 1 (indeterminate 2 0) == indeterminate 1 0@ is true
-- @deleteEndIndeterminates 1 (indeterminate 2 1)@ fails
--
-- See also: 'deleteIndeterminates'
--
deleteEndIndeterminates :: (Num a, Integral b)
                        => b -- ^ @n@
                        -> Polynomial a -- ^ @p@
                        -> Polynomial a
deleteEndIndeterminates n p
    | n < 0  = error "deleteEndIndeterminates: n < 0"
    | n == 0 = p -- nothing to do
    | Just c <- isConstPoly p =
		  constPolyN ((unNIndeterminates (nIndeterminates p)) - n) c
    | otherwise = -- Warning: this implementation is maybe inefficient
                  deleteIndeterminates n $ replacePoly1 shiftNRight p
    where
    shiftNRight xs =
	mytrace ("shiftNRight xs=" ++ show xs) $
	let l = length xs
	    dropLastN ys =
		case splitAt (fromIntegral n) (reverse ys) of
		  (del, ok) -> if genericLength del == n && all (==0) del
			       then reverse ok
			       else errMessage
        in product [(indeterminate l i) ^ (fromIntegral a)
		    | (a,i) <- zip (dropLastN xs) [fromIntegral n..]]
    errMessage =
      error ("deleteEndIndeterminates: cannot delete last " ++
	     show n ++ " indeterminates of p=" ++ show p ++
	     " without loss! (nIndeterminates p == " ++
	     show (nIndeterminates p) ++ ")")

reverseIndeterminates :: (Num a, Integral b)
                      => b
                      -> Polynomial a
                      -> Polynomial a
reverseIndeterminates ni p =
    replacePoly1 rev p
    where
    rev es = product [indeterminate ni (ni-i-1)^e | (e,i) <- zip (es ++ repeat 0) [0..ni-1]]

-- |
-- Add two polynomials.
--
addPoly :: Num a => Polynomial a -> Polynomial a -> Polynomial a
addPoly = unFrom2 "addPoly" addPoly'

addPoly' :: Num a => Polynomial a -> Polynomial a -> Polynomial a
addPoly' (ZeroPolynomial _ _) y = y
addPoly' x (ZeroPolynomial _ _) = x
addPoly' (NonZeroPolynomial me1 x) (NonZeroPolynomial me2 y) =
    normalizeNZP (me1&&me2) $ addNZP x y

addNZP :: Num a => NZP a -> NZP a -> NZP a
addNZP (Constant x)   (Constant y)   = Constant (x+y)
addNZP (Recursive ms) (Recursive ns) = Recursive (add ms ns)
    where
    add :: Num a => [(Int,NZP a)] -> [(Int,NZP a)] -> [(Int,NZP a)]
    add [] ns = ns
    add ms [] = ms
    add ms@((e1,p1):m) ns@((e2,p2):n)
	| e1 <  e2  =  (e2,p2) : add ms n
	| e1 == e2  =  (e1, addNZP p1 p2) : add m n
	| otherwise =  (e1,p1) : add m ns
addNZP _ _ =
    error "addNZP: different number of indeterminates."


-- |
-- Subtract two polynomials.
--
subPoly :: Num a => Polynomial a -> Polynomial a -> Polynomial a
subPoly p q = addPoly p (negPoly q)

-- |
-- Negate a polynomial.
--
negPoly :: Num a => Polynomial a -> Polynomial a
negPoly z@(ZeroPolynomial _ _) = z
negPoly (NonZeroPolynomial me p) = NonZeroPolynomial me (negP p)
    where
    negP (Constant  c)  = Constant (-c)
    negP (Recursive ms) = Recursive [(exp, negP p) | (exp,p) <- ms]

-- |
-- Multiply two polynomials.
--
multPoly :: Num a => Polynomial a -> Polynomial a -> Polynomial a
multPoly = unFrom2 "multPoly" multPoly'

multPoly' :: Num a => Polynomial a -> Polynomial a -> Polynomial a
multPoly' z@(ZeroPolynomial _ _) q = z
multPoly' p z@(ZeroPolynomial _ _) = z
multPoly' p@(NonZeroPolynomial me1 x) q@(NonZeroPolynomial me2 y) =
    normalizeNZP (me1&&me2) $ mult x y
    where
    mult (Constant c1)  (Constant c2)  = Constant (c1*c2)
    mult (Recursive ms) (Recursive ns) =
	Recursive $
	sortAddLevel [(e1+e2, mult p1 p2) | (e1,p1) <- ms, (e2,p2) <- ns]
    mult _ _ = error ("multPoly: polynomials have different number of indeterminates: " ++ show (nIndeterminates p) ++ " against " ++ show (nIndeterminates q) ++ "\n" ++ showPolyStructure p "" ++ "\n" ++ showPolyStructure q "")

sortAddLevel :: Num a => [(Int, NZP a)] -> [(Int, NZP a)]
sortAddLevel ls@((emax,_::NZP a):_) = runST sortL
    where
    sortL :: ST s [(Int, NZP a)]
    sortL =
	do a <- (newArray (0,emax) Nothing :: ST s (STArray s Int (Maybe (NZP a))))
	   mapM_ (upd a) ls
	   foldM (sumup a) [] [0..emax]
    upd a (e,p) =
	do old <- readArray a e
	   case old of
	     Nothing -> writeArray a e (Just p)
	     Just p' -> writeArray a e (Just (addNZP p p'))
    sumup a s e =
	do val <- readArray a e
	   case val of
	     Nothing -> return s
	     Just p  -> return ((e,p) : s)
       

-- |
-- Multiply a constant and a polynomial.
--
pmul :: Num a => a -> Polynomial a -> Polynomial a
pmul f p = mapPoly (*f) p


-- |
-- Raise a polynomial to given non-negative power.
--
potPoly :: (Num a, Integral b) => Polynomial a -> b -> Polynomial a
potPoly p 0 = constPolyNp p 1
potPoly p n
    | n < 0      =  error "potPoly: not defined for exponent < 0"
    | otherwise  =  p ^ n

-- |
-- Build a constant polynomial.
-- @constPoly@ returns a constant with the specified value.
-- The number of indeterminates of the polynomial is unspecified.
--
constPoly :: Num a
	  => a    -- ^ value of the polynomial
          -> Polynomial a
constPoly 0 = ZeroPolynomial True 0
constPoly x = NonZeroPolynomial True $ Constant x

-- |
-- Build a constant polynomial with a specific number of indeterminates.
-- @constPolyN@ returns a constant polynomial with the specified value and
-- @Exactly n@ indeterminates.
--
constPolyN :: (Num a, Integral b)
	   => b             -- ^ number @n@ of indeterminates
           -> a             -- ^ value of the polynomial
	   -> Polynomial a
constPolyN n c = constPolyNI (Exactly $ fromIntegral n) c

-- |
-- Return a constant polynomial with a given number of indeterminates.
--
constPolyNI :: Num a
	    => NIndeterminates
            -> a
	    -> Polynomial a
constPolyNI (AtLeast n) 0 = ZeroPolynomial True  $ fromIntegral n
constPolyNI (Exactly n) 0 = ZeroPolynomial False $ fromIntegral n
constPolyNI nn c =
    case nn of
      Exactly n -> NonZeroPolynomial False $ constP n
      AtLeast n -> NonZeroPolynomial True  $ constP n
    where
    constP n
	| n >  0  =  Recursive [(0, constP (n-1))]
	| n == 0  =  Constant c
	| otherwise =
	    error "constPolyNI: negative number of indeterminates"


constPolyME :: (Num a, Integral b)
	    => Bool
            -> b
	    -> a
	    -> Polynomial a
constPolyME True  n x = constPolyNI (AtLeast $ fromIntegral n) x
constPolyME False n x = constPolyNI (Exactly $ fromIntegral n) x

-- |
-- Build a constant polynomial with the same number of indeterminates
-- as a given polynomial.
-- @constPolyNp@ returns a constant polynomial with the same number of
-- indeterminates as the given polynomial @p@.
--
constPolyNp :: Num a
	    => Polynomial b     -- ^ polynomial @p@
            -> a                -- ^ value of the new polynomial
	    -> Polynomial a
constPolyNp p c = constPolyNI (nIndeterminates p) c


-- |
-- Polynomial division with remainder.
-- @quotRemPoly a b@ returns a pair @(q, r)@ of polynomials such that
-- @a=q*b+r@ holds and @r=0@ or @degree r \< degree b@. @b@ must not be
-- 0.
quotRemPoly :: Fractional a =>
	       Polynomial a -> Polynomial a -> (Polynomial a, Polynomial a)
quotRemPoly = unFrom2 "quotRemPoly" quotRemPoly'

quotRemPoly' :: Fractional a =>
	        Polynomial a -> Polynomial a -> (Polynomial a, Polynomial a)
quotRemPoly' _ (ZeroPolynomial _ _) =
    error "quotRem: division by 0-polynomial."
quotRemPoly' z@(ZeroPolynomial _ _) _ = (z, z)
quotRemPoly' a@(NonZeroPolynomial me1 (Constant p))
	     (NonZeroPolynomial me2 (Constant q)) =
    (normalizeNZP (me1&&me2) $ Constant (p/q),
     ZeroPolynomial (me1&&me2) 0
    )
quotRemPoly' p@(NonZeroPolynomial me1 (Recursive ((e1,p1):ms)))
	    q@(NonZeroPolynomial me2 (Recursive ((e2,p2):ns)))
    | e1 >= e2  =
	case quotRemPoly' (NonZeroPolynomial me p1) (NonZeroPolynomial me p2) of
	   (NonZeroPolynomial _ qu, _) ->
	       (quotient1 `addPoly` quotient2, remainder)
	       where
	       quotient1 = NonZeroPolynomial me (Recursive [(e1-e2,qu)])
	       (quotient2, remainder) = 
		   (p `subPoly` (quotient1 `multPoly'` q)) `quotRemPoly'` q
	   (ZeroPolynomial me' z, _) -> (ZeroPolynomial me' (z+1), p)
    | otherwise = (ZeroPolynomial me (nIndeterminatesNZP p1 + 1), p)
    where
    me = me1 && me2
quotRemPoly' p q =
    error "quotRemPoly: polynomials have different number of indeterminates"


-- |
-- Calculate the quotient of a polynomial division.
--
quotPoly :: Fractional a =>
	    Polynomial a -> Polynomial a -> Polynomial a
quotPoly p q = let (qu,rest) = quotRemPoly p q in qu

-- |
-- Calculate the remainder of a polynomial division.
--
remPoly :: Fractional a =>
	   Polynomial a -> Polynomial a -> Polynomial a
remPoly p q = let (qu,rest) = quotRemPoly p q in rest


-- |
-- Evaluate a polynomial at a given point.
-- @evalPoly@ substitutes the values from the list @values@ for
-- the indeterminates in the polynomial @p@ and thereby evaluates
-- the polynomial. The length of @values@ must match the
-- number of indeterminates in @p@. The i-th element in
-- @values@ is substitued for indeterminate number @i@.
--
evalPoly :: Num a
         => Polynomial a    -- ^ p
         -> [a]             -- ^ values
         -> a
evalPoly (ZeroPolynomial _ _)    _      = 0
evalPoly (NonZeroPolynomial me nzp) values = eval values nzp
    where
    eval []     (Constant  x ) = x
    eval vs     (Constant  x )
	 | me  =  x
    eval (v:vs) (Recursive rs) = sum [v^e * eval vs q | (e,q) <- rs]
    eval _ _ = error ("evalPoly: number of indeterminates doesn't match:\n\
		      \           (NonZeroPolynomial " ++ show me ++ " (" ++
		                  show nzp ++ ")) " ++ show values)


-- |
-- Make the coefficients of a polynomial integral.
-- @ratPoly2intPoly@ calculates a polynomial pair @(f, q)@ such that
-- @f `pmul` p == q@ and @f > 0@ hold.
--
ratPoly2intPoly :: (Real a, Integral b)
                => Polynomial a   -- ^ p
                -> (b, Polynomial b)
ratPoly2intPoly (ZeroPolynomial me n) = (1, ZeroPolynomial me n)
ratPoly2intPoly p' =
    (fromIntegral lcmDenom,
     mapPoly (\r -> fromIntegral ((numerator r * lcmDenom) `div` denominator r)) p)
    where
    coeffs = collectCoeffs p
    lcmDenom = abs $ foldl1 lcm (map denominator coeffs)
    p = mapPoly toRational p'

-- |
-- Returns the reductum of a polynomial (wrt. the first indeterminate).
--
reductum :: Num a => Polynomial a -> Polynomial a
reductum z@(ZeroPolynomial _ _) = z
reductum (NonZeroPolynomial me (Constant _)) =
    ZeroPolynomial me 0
reductum p@(NonZeroPolynomial me (Recursive (r:rs)))
    | null rs   = if me then ZeroPolynomial True 0 else constPolyNp p 0
    | otherwise = NonZeroPolynomial me (Recursive rs)

-- |
-- Calculates all non-zero reducta of a polynomial (wrt.
-- the first indeterminate).
--
reducta :: Num a => Polynomial a -> [Polynomial a]
reducta p = takeWhile (not.isZeroPoly) $ iterate reductum p


-- |
-- Return the leading coefficient of a polynomial.
--
leadingCoeff :: Num a => Polynomial a -> a
leadingCoeff (ZeroPolynomial _ _) = 0
leadingCoeff (NonZeroPolynomial _ nzp) = lc nzp
    where
    lc (Constant x)          = x
    lc (Recursive ((e,q):_)) = lc q

-- |
-- Return the additive constant of of a polynomial.
--
constantCoeff :: Num a => Polynomial a -> a
constantCoeff (ZeroPolynomial _ _) = 0
constantCoeff (NonZeroPolynomial _ nzp) = cc nzp
    where
    cc (Constant x)   = x
    cc (Recursive rs) = case last rs of
			    (0,q) -> cc q
			    _     -> 0
-- |
-- Consider a polynomial as a univariate polynomial in its first variable.
--
univariate :: Polynomial a -> Polynomial (Polynomial a)
univariate (ZeroPolynomial True _)                 = ZeroPolynomial False 1
univariate (ZeroPolynomial False n)
    | n >= 1     =  ZeroPolynomial False 1
    | otherwise  = 
	error "`univariate' undefined for polynomial in exactly 0 indeterminates"
univariate c@(NonZeroPolynomial True (Constant _)) =
    NonZeroPolynomial False (Recursive [(0, Constant c)])
univariate (NonZeroPolynomial False (Constant _)) =
    error "`univariate' undefined for polynomial in exactly 0 indeterminates"
univariate (NonZeroPolynomial me (Recursive rs)) =
    NonZeroPolynomial False (Recursive [(e, Constant (NonZeroPolynomial me p))
					| (e,p) <- rs])

splitPoly :: (Num a, Integral b)
          => b
          -> Polynomial a
          -> Polynomial (Polynomial a)
splitPoly n (NonZeroPolynomial me nzp1) =
    NonZeroPolynomial False $ splitNZP n nzp1
    where
    splitNZP 0 nzp = Constant $ NonZeroPolynomial me nzp
    splitNZP n (Recursive rs) =
	Recursive [(e, splitNZP (n-1) nzp) | (e,nzp) <- rs]
    splitNZP _ (Constant _) =
	error "splitPoly: insufficient number of indeterminates"
splitPoly n (ZeroPolynomial True _) =
    ZeroPolynomial False $ fromIntegral n
splitPoly n (ZeroPolynomial False unk)
    | unk >= fromIntegral n   =  ZeroPolynomial False $ fromIntegral n
    | otherwise  =  error "splitPoly: insufficient number of \
			  \indeterminates"


-- |
-- Flatten a polynomial in polynomials.
-- The function is undefined for polynomials of the form
-- @FromNum p@ where p is not @FromNum _@ 
-- nor @ZeroPolynomial _ _@. Note that 'constPoly' returns
-- such polynomials, for example. Use 'constPolyN' or 'constPolyNI'
-- if you intend to use 'flatPoly' on constant polynomials.
--
flatPoly :: Num a => Polynomial (Polynomial a) -> Polynomial a
flatPoly (ZeroPolynomial _ n) = ZeroPolynomial True n
flatPoly poly@(NonZeroPolynomial _ nzp) =
    NonZeroPolynomial me (mult nzp)
    where
    mult :: Num a => NZP (Polynomial a) -> NZP a
    mult (Constant (NonZeroPolynomial False nzp)) = nzp
    mult (Constant (NonZeroPolynomial True  nzp)) =
	appendIndetsNZP nzp (unNIndeterminates nInd - nIndeterminatesNZP nzp)
    mult (Recursive rs) = Recursive [(e,mult p) | (e,p) <- rs]

    nInd = nIndeterminatesN $ collectCoeffs poly
    me = case nInd of
	   Exactly _ -> False
	   AtLeast _ -> True


-- |
-- Return all the non-zero coefficients of a polynomials.
-- The leading coefficient is the first entry in the result
-- list, the constant coefficient is the last.
--
collectCoeffs :: Num a => Polynomial a -> [a]
collectCoeffs (ZeroPolynomial _ _) = []
collectCoeffs (NonZeroPolynomial _ nzp) = cc nzp
    where
    cc (Constant c) = [c]
    cc (Recursive rs) = concat [cc p | (_,p) <- rs]


-- |
-- Make a polynomial monic, i.e., divide each coefficient
-- by the leading coefficient. The leading coefficient
-- of the result polynomial is then (of course) @1@.
--
monic :: Fractional a => Polynomial a -> Polynomial a
monic z@(ZeroPolynomial _ _)    = z
monic p@(NonZeroPolynomial _ _) = mapPoly (/ leadingCoeff p) p

-- |
-- Make a polynomial \"monic\", but keep the signs of the coefficients.
-- @
--  monic ( "5*x*y+10")     =    "x*y+2"
--  monic ("-5*x*y+10")     =    "x*y-2"
--  signMonic ( "5*x*y+10") =    "x*y+2"
--  signMonic ("-5*x*y+10") = "-1*x*y+2"
-- @
--
signMonic :: Fractional a => Polynomial a -> Polynomial a
signMonic z@(ZeroPolynomial _ _) = z
signMonic p@(NonZeroPolynomial _ _) = mapPoly (/ abs (leadingCoeff p)) p

-- |
-- Return the 'Signum' for constant polynomials. For non-constant
-- polynomials all fields in the result are set to false.
--
constPolySignum :: (Num a, Ord a) => Polynomial a -> Signum
constPolySignum p
    | Just c <- isConstPoly p  =  orderingToSignum (c `compare` 0)
    | otherwise                =  allSignumsFalse

-- |
-- Perform Euclid's algorithm.
-- This function performs Euclid's algorithm in the /first/
-- indeterminate of the given polynomials. For univariate polynomials,
-- this calculates the greatest common divisor (gcd). The
-- result polynomial is always a monic polynomial.
--
euclid :: Fractional a => Polynomial a -> Polynomial a -> Polynomial a
euclid = unFrom2 "euclid" euclid'

-- a gcd is always a monic polynomial !!!
euclid' :: Fractional a => Polynomial a -> Polynomial a -> Polynomial a
euclid' (ZeroPolynomial _ _) (ZeroPolynomial _ _) =
    error "euclid: undefined for two 0-polynomials"
euclid' (ZeroPolynomial _ _) q = monic q
euclid' p (ZeroPolynomial _ _) = monic p
euclid'  (NonZeroPolynomial me1 (Constant p)) (NonZeroPolynomial me2 (Constant q)) =
    if me1&&me2 then constPolyNI (AtLeast 0) 1 else constPolyN 0 1
euclid' p@(NonZeroPolynomial _ (Recursive ((e1,p1):ms)))
       q@(NonZeroPolynomial _ (Recursive ((e2,p2):ns)))
    | e1 < e2   = euclid' q p
    | otherwise = euclid' q remainder
    where
    (_, remainder) = p `quotRemPoly'` q

-- |
-- Calculate the gcd of (multivariate) polynomials.
-- The gcd is always a monic polynomial.
--
gcdPoly :: Fractional a => Polynomial a -> Polynomial a -> Polynomial a
gcdPoly a b = mytrace (show ("gcdPoly (" ++ show a ++ ") (" ++ show b ++ ")")) $
	      unFrom2 "gcdPoly" gcdPoly' a b

-- a gcd is always a monic polynomial !!!
gcdPoly' :: Fractional a => Polynomial a -> Polynomial a -> Polynomial a
gcdPoly' (ZeroPolynomial me n) (ZeroPolynomial _ _) = ZeroPolynomial me n
gcdPoly' (ZeroPolynomial _ _) q = monic q
gcdPoly' p (ZeroPolynomial _ _) = monic p
gcdPoly' (NonZeroPolynomial me1 (Constant p)) (NonZeroPolynomial me2 (Constant q)) =
    if me1&&me2 then constPolyNI (AtLeast 0) 1 else constPolyN 0 1
gcdPoly' (NonZeroPolynomial me1 (Recursive [(e1,p1)]))
	 (NonZeroPolynomial me2 (Recursive [(e2,p2)])) =
    NonZeroPolynomial me (Recursive [(min e1 e2,g)])
    where
    (me,g) = case gcdPoly' (NonZeroPolynomial me1 p1)
                           (NonZeroPolynomial me2 p2) of
	         NonZeroPolynomial me_ g_ -> (me_,g_)
		 ZeroPolynomial _ _       ->
		     error "gcdPoly': ZeroPolynomial in recursive call??"
gcdPoly' p q =
    case unNIndeterminates $ nIndeterminates p of
        1 -> euclid' p q  -- for (very much) better efficiency
	_ -> monic euc `multPoly'` liftPoly (gcdPoly contP contQ)
    where
    contP = content p
    contQ = content q
    primP = mapPoly (`quotPoly` contP) $ univariate p   -- quotPoly p (liftPoly contP)
    primQ = mapPoly (`quotPoly` contQ) $ univariate q   -- quotPoly q (liftPoly contQ)
    euc   = toPoly $
	    euclid' (mapPoly fromPoly primP) (mapPoly fromPoly primQ)

-- |
-- Perform the extended version of Euclid's algorithm.
-- The following equalities hold:
-- 
-- > g = gcd(p,q)
-- > g = s*p + t*q
--
extEuclid :: Fractional a
	  => Polynomial a         -- ^ @p@
          -> Polynomial a         -- ^ @q@
	  -> (Polynomial a, Polynomial a, Polynomial a)  -- ^ @(g,s,t)@
extEuclid = unFrom2 "extEuclid" exteuclid'

-- a gcd is always a monic polynomial !!!
exteuclid' :: Fractional a => Polynomial a -> Polynomial a ->
              (Polynomial a, Polynomial a, Polynomial a)
exteuclid' (ZeroPolynomial _ _) (ZeroPolynomial _ _) =
    error "euclid: undefined for two 0-polynomials"
exteuclid' (ZeroPolynomial me n) q =
    (monic q, cp n 0, cp n (recip lc))
    where
    cp = constPolyNI . (if me then AtLeast else Exactly) . fromIntegral
    lc = leadingCoeff q
exteuclid' p (ZeroPolynomial me n) =
    (monic p, cp n (recip lc), cp n 0)
    where
    cp = constPolyNI . (if me then AtLeast else Exactly) . fromIntegral
    lc = leadingCoeff p
exteuclid'  (NonZeroPolynomial me1 (Constant p)) (NonZeroPolynomial me2 (Constant q)) =
    if me1&&me2
       then (constPolyNI (AtLeast 0) 1, constPolyNI (AtLeast 0) $ recip p,
             ZeroPolynomial True 0)
       else (constPolyN 0 1, constPolyN 0 (recip p), constPolyN 0 0)
exteuclid' p@(NonZeroPolynomial _ (Recursive ((e1,p1):ms)))
       q@(NonZeroPolynomial _ (Recursive ((e2,p2):ns)))
    | e1 < e2   = let (e,s,t) = exteuclid' q p in (e,t,s)
    | otherwise = let (e,s,t) = exteuclid' q remainder
		  in  (e,t,s-t*qu)	
    where
    (qu, remainder) = p `quotRemPoly'` q

-- |
-- Return the content of a polynomial, i.e.,
-- the gcd of its coefficients. N.B. the polynomials is
-- regarded as a univariate polynomial in the first indeterminate
-- and the gcd is calculated from the \"coefficients\" of the
-- polynomial in this form.
--
content :: Fractional a => Polynomial a -> Polynomial a
content (NonZeroPolynomial me (Recursive ms)) = foldl gcdPoly' c cs
    where
    (c:cs) = map (NonZeroPolynomial me . snd) ms
content (NonZeroPolynomial _ (Constant _)) =
    error "content: the contents of a constant is undefined"
content (ZeroPolynomial _ _) =
    error "content: the content of a 0-polynomial is undefined"

lcmPoly :: Fractional a => Polynomial a -> Polynomial a -> Polynomial a
lcmPoly p q = (p `quotPoly` gcdPoly p q) `multPoly` q


gcdFree :: Fractional a => [Polynomial a] -> [Polynomial a]
gcdFree ps = foldl insertGcdFree [] $ filter ((>0).totalDegree) ps

insertGcdFree :: Fractional a => [Polynomial a] -> Polynomial a -> [Polynomial a]
insertGcdFree [] n
    | totalDegree n > 0  =  [n]
    | otherwise     =  []
insertGcdFree (p:ps) n
    | totalDegree p == degree g  = insertGcdFree (insertGcdFree ps g) (n `quotPoly` g)
    | totalDegree g > 0  =
	insertGcdFree (insertGcdFree ((p `quotPoly` g) :  ps) g) (n `quotPoly` g)
    | otherwise     = p : insertGcdFree ps n 
    where
    g = gcdPoly p n



-- |
-- Returns the degree of a given polynomial in the /first/ indeterminate.
--
degree :: (Num a, Integral b) => Polynomial a -> b
degree (ZeroPolynomial _ _) = -1
degree (NonZeroPolynomial _ (Constant _))          = 0
degree (NonZeroPolynomial _ (Recursive ((e,_):_))) = fromIntegral e

-- |
-- Returns the total degree of a polynomial, i.e., the
-- maximum exponent sum of a term.
--
totalDegree :: (Num a, Integral b) => Polynomial a -> b
totalDegree (ZeroPolynomial _ _) = -1
totalDegree (NonZeroPolynomial _ nzp) = td nzp
    where
    td (Constant _)   = 0
    td (Recursive rs) = maximum [fromIntegral e + td nzp | (e,nzp) <- rs]

-- |
-- Return the degree of the polynomial in a given indeterminate.
--
degreeIn :: (Num a, Integral b)
         => b             -- ^ the number of the indeterminate
         -> Polynomial a  -- ^ the polynomial
         -> b
degreeIn _ (ZeroPolynomial _ _) = -1
degreeIn 0 p = degree p
degreeIn n p =
    maximum $ map (degreeIn (n-1)) $ coefficients p

-- |
-- Returns the derivative of a given polynomial wrt. the /first/
-- indeterminate.
--
derivative :: Num a => Polynomial a -> Polynomial a
derivative p@(ZeroPolynomial _ n) = p
derivative (NonZeroPolynomial me (Constant _))  = ZeroPolynomial me 0
derivative (NonZeroPolynomial me (Recursive nzp)) =
    normalizeNZP me $
    Recursive [(m1 e, fmap (* fromIntegral e) r) | (e,r) <- nzp]
    -- never produce "Recursive []" !!!
    where
    m1 0 = 0
    m1 e = e - 1


-- |
-- Replace every term of a polynomial by a polynomial.
-- (BETTER EXPLANATION NEEDED).
--
replacePoly :: (Num a, Integral b) => 
	       (Maybe [b] -> Polynomial a) -> Polynomial a -> Polynomial a
replacePoly _ (ZeroPolynomial _ n) = ZeroPolynomial True n
replacePoly f (NonZeroPolynomial _ nzp) = replace nzp []
    where
    replace (Constant c)   exps =
	let p = f (Just $ reverse exps) in p * constPolyNp p c
    replace (Recursive rs) exps =
	foldl1 (+) [replace r (fromIntegral e : exps) | (e,r) <- rs]

-- |
-- Special case of 'replacePoly'.
--
replacePoly1 :: (Num a, Integral b) => 
		([b] -> Polynomial a) -> Polynomial a -> Polynomial a
replacePoly1 f p = replacePoly mbF p
    where
    mbF Nothing   = 1
    mbF (Just es) = f es
    
-- |
-- Reduce \"over\" a polynomial.
-- (EXPLANATION MISSING).
--
reducePoly :: Integral e =>
	      (b -> b -> b) -> (a -> Maybe [e] -> b) -> b -> Polynomial a -> b
reducePoly plus mono start (ZeroPolynomial _ n) = start
reducePoly plus mono start (NonZeroPolynomial _ nzp) = reduce nzp []
    where
    reduce (Constant c)   exps = mono c (Just $ reverse exps)
    reduce (Recursive rs) exps =
	foldl plus start [reduce r (fromIntegral e : exps) | (e,r) <- rs]

-- |
-- Return the list of monomials of the given polynomial.
-- A monomial is represented by its coefficient and a
-- list of exponents of the indeterminates.
monomials :: Integral e => Polynomial a -> [(a,[e])]
monomials = reducePoly (++) f []
    where
    f co Nothing   = [(co,[])]
    f co (Just es) = [(co,es)]


-- |
-- Swap two indeterminates.
--
swapIndeterminates :: (Num a, Integral b)
                   => b
                   -> b
                   -> Polynomial a
                   -> Polynomial a
swapIndeterminates a b p
    | a >  b  =  swapIndeterminates b a p
    | a == b  =  p
    | otherwise  =
        replacePoly1 (\es -> product [indeterminate nInd i ^ e
                                     | e <- newexps es | i <- [0..]]
                     ) p
    where
    newexps es =
        concat [genericTake a es
               ,[es !! fromIntegral b]
               ,genericTake (b-a-1) $ genericDrop (a+1) es
               ,[es!!fromIntegral a]
               ,genericDrop (b+1) es
               ]

    nInd = unNIndeterminates $ nIndeterminates p


-- |
-- Replaces each indeterminate @i@ of @p@ by the indeterminate @j@.
--
-- Example:
-- 
-- 1) @rename (0,1) (u0^2*1 + u1*2 + u2*3) = u1^2*1 + u1*2 + u2*3@
--
-- 2) @rename (2,1) (u0^2*1 + u1*2 + u2*3) = u0^2*1 + u1*5@
--
renamePolyVariable :: (Num a, Integral b)
                   => (b, b) -- ^ the indeterminates (@i@, @j@) to be renamed
		   -> Polynomial a -- ^ @p@
		   -> Polynomial a
renamePolyVariable (i', j') p
    | i' == j'  = error "renamePolyVariable: i and j must differ"
    | otherwise = replacePoly1 renameFunc p
    where
    (i,j) = (fromIntegral i', fromIntegral j')
    -- This implementation is not very efficient:
    renameFunc xs = let arr  = listArray (0, genericLength xs - 1) xs
                        arr' = arr // [(i, 0), (j, (arr ! i) + (arr ! j))]
                    in createPolyFromList [(1, elems arr')]

indeterminateToFront :: (Num a, Integral b)
                     => Polynomial a -- ^ @p@
		     -> b -- ^ the indeterminates to move to position 0
		     -> Polynomial a
indeterminateToFront p i'
    | i >= 0 && i < unNIndeterminates (nIndeterminates p)  =
        replacePoly1 renameFunc p
    | otherwise  =  error ("indeterminateToFront: index " ++ show i ++ 
                           "illegal for " ++ show p
                          )
    where
    i = fromIntegral i'
    -- This implementation is not very efficient:
    renameFunc xs = let (ys,z:zs) = splitAt i xs
                        newExps   = z : (ys ++ zs)
                    in createPolyFromList [(1, newExps)]

-- |
-- Creates a multivariate polynomial from a list of terms which
-- are described by their factor and a list of the exponents
-- of the indeterminates. The resulting polynomial is the sum
-- of all terms which are converted into polynomials.
--
-- Examples:
--
-- 1) @createPolyFromList [ (42,[2,3]) ] = u0^2*u1^3*42@
--
-- 2) @createPolyFromList [ (42,[2,3]), (17,[0,3])] = u0^2*u1^3*42 + u1^3*17@
--
-- 3) @createPolyFromList [ (5,[0,1,0]), (7,[0,1,0]) ] = u1*12@
--
createPolyFromList :: (Num a, Integral b)
                   => [(a, [b])]
		   -- ^ a list of pairs (factor, exponents)
		   -> Polynomial a -- ^ the resulting polynomial
createPolyFromList xs = sum $ map createPoly xs
    where
    createPoly :: (Num a, Integral b) => (a, [b]) -> Polynomial a
    createPoly (coeff, exps) = pmul coeff (polyFromExpList exps)

    polyFromExpList :: (Num a, Integral b) => [b] -> Polynomial a
    polyFromExpList exps =
	let nIndet = genericLength exps
	    conv x i = (indeterminate nIndet i) ^ x
	in product $ zipWith conv exps [0..]

-- |
-- Converts a polynomial @p@ to a list representation.
--
-- See also: 'createPolyFromList'
--
polyToList :: (Integral b)
           => Polynomial a -- ^ @p@
           -> [(a, [b])]
polyToList (ZeroPolynomial _ _) = []
polyToList (NonZeroPolynomial _ nzp) = parseNZP nzp
    where
    parseNZP :: (Integral b) => NZP a -> [(a, [b])]
    parseNZP (Constant x) = [(x, [])]
    parseNZP (Recursive xs) =
	let conv c = map (\(x,y) -> (x, (fromIntegral c):y))
        in concatMap (\(c, nzp) -> conv c (parseNZP nzp)) xs


-- |
-- Construct an affine Polynomial (i.e., a polynomial with
-- total degree <= 1) from a list of coefficients.
-- Example:
--
-- @
--   affinePolyFromList [1,2,3,4]  = u0 + 2*u1 + 3*u2 + 4
-- @
--
affinePolyFromList :: Num a => [a] -> Polynomial a
affinePolyFromList coeffs =
    (constPolyN nInd (last coeffs) +
		sum [c `pmul` indeterminate nInd i
		         | (c,i) <- zip (init coeffs) [0..]]
    )
    where
      nInd = length coeffs - 1

-- not really Gauss: leading coefficients are divided by factors of
-- "factor" !!!
-- NOTE: Don't use Gauss to calculate determinants, Bareiss' algorithm
-- (see `bareiss' below) is much more efficient.
{-
utGauss :: Fractional a =>
	   GMatrix (Polynomial a) -> (Polynomial a, GMatrix (Polynomial a))
utGauss matr =
    let (fac,ut) = runST (mkut matr)
    in  (fac, matrix (nRows matr, nColumns matr, ut))
    where
    mkut m =
	do a <- newSTArray ((0,0),(maxRow m,maxColumn m)) undefined
	   sequence_ [writeArray a (r,c) (entry m r c) | (r,c) <- indexes m]
	   factor <- newSTRef one
	   ut factor a 0 0
	   els <- getElems a
	   f   <- readSTRef factor
	   return (f,els)
    ut factor a r0 c0
	| c0 > maxC  =  return ()
	| otherwise  =
	    sequence [readArray a (r,c0) | r <- [r0..maxR]] >>= \cs ->
	    case findIndex (not.isZeroPoly) cs of
	        Nothing -> ut factor a r0 (c0+1)
		Just i  -> do exch r0 (r0 + fromIntegral i)
			      sequence_ [mkZero r | r <- [r0+1 .. maxR]]
			      co <- readArray a (r0,c0)
			      f  <- readSTRef factor
			      let g = gcdPoly co f
			      when (isConstPoly g == Nothing)
			        (writeSTRef factor (f `quotPoly` g) >>
				 writeArray a (r0,c0) (co `quotPoly` g))
			      ut factor a (r0+1) (c0+1)
	where
	mkZero r =
	    do co0 <- readArray a (r0,c0)
	       co  <- readArray a (r, c0)
	       let l  = lcmPoly co0 co
		   f0 = l `quotPoly` co0
		   f  = if isZeroPoly co then one else l `quotPoly` co
	       oldFact <- readSTRef factor
	       writeSTRef factor (oldFact * f)
	       sequence_ [zero f0 f c | c <- [c0..maxC]]
	    where
	    zero f0 f c = readArray a (r, c) >>= \v  ->
			  readArray a (r0,c) >>= \v0 ->
			  writeArray a (r,c) (f*v - f0*v0)
	exch r1 r2 
	    | r1 /= r2  =
		sequence_ [readArray  a (r1,c) >>= \v1 ->
			   readArray  a (r2,c) >>= \v2 ->
			   writeArray a (r2,c) v1 >>
			   writeArray a (r1,c) v2 
			  | c <- [0..maxC]] >>
		readSTRef factor >>= \f ->
		writeSTRef factor (-f)
	    | otherwise = return ()

    maxR = maxRow matr
    maxC = maxColumn matr
    one  = constPoly (nIndeterminates (entry matr 0 0)) 1
-}

{-
determinantP matr = foldl1 (*) [entry ut i i | i <- [0..maxRow matr]]
		    `quotPoly` multiplier
    where
    (multiplier,ut) = utGauss matr
-}

-- |
-- Return the \"coefficients\" of the given polynomial when regarding the
-- polynomial as being univariate over R[X2,...,Xn], i.e.,
--
-- @coefficients \"x^2*(y*3) - 1*(y^5-3)\"  =  [\"-y^5+3\", \"0\", \"y*3\"]@
--
coefficients :: Num a => Polynomial a -> [Polynomial a]
coefficients (ZeroPolynomial _ _) = []
coefficients p@(NonZeroPolynomial True (Constant c)) = [p]
coefficients (NonZeroPolynomial False (Constant c)) =
    error "coefficients: not defined for polynomials in 0 indeterminates"
coefficients (NonZeroPolynomial me (Recursive cs)) = coeffs cs
    where
    coeffs [(e,c)] = replicate e zero ++ [NonZeroPolynomial me c]
    coeffs ((e1,c) : rest@((e2,_):_)) =
	coeffs rest ++ replicate (e1-e2-1) zero ++ [NonZeroPolynomial me c]
    zero
	| me        = ZeroPolynomial True 0
	| otherwise = ZeroPolynomial False $ nIndeterminatesNZP $ snd $ head cs


-- |
-- Checks whether the given polynomial is a linear term
-- in the indeterminates and returns the coefficients of the 
-- indeterminates (in the order of the indeterminates) and
-- the constant term, e.g.
-- @
--    linearCoeffs "2*u0 + 3*u1 + 0*u2 + 4"  =  Just [2, 3, 0 ,4]
--    linearCoeffs "2*u0 + 3"                =  Just [2, 3]
--    linearCoeefs "2*u0^2 + 3"              =  Nothing
-- @
--
linearCoeffs :: Num a => Polynomial a -> Maybe [a]
linearCoeffs (ZeroPolynomial _ n) = Just $ replicate n 0
linearCoeffs (NonZeroPolynomial _ nzp) = lc nzp
    where
    lc (Constant  c) = Just [c]
    lc (Recursive [(0,nzp')]) = liftM (0:) $ lc nzp'
    lc (Recursive [(1,nzp')]) =
	case lc nzp' of
	  Nothing -> Nothing
	  Just cs -> if all (==0) (init cs)
		        then Just (last cs : init cs ++ [0])
			else Nothing
    lc (Recursive [(1,nzp'),(0,nzp'')]) =
	case lc nzp' of
	  Nothing -> Nothing
	  Just cs -> if all (==0) (init cs)
		        then liftM (last cs :) $ lc nzp''
			else Nothing

-- |
-- Convert an affine expression to a polynomial.
-- The (1-dimensional) affine expression represented by the list of
-- coefficients @a@ is converted to a polynomial by introducing
-- new indeterminates which take the places of the variables of
-- the affine expression. The indeterminates in the coefficients
-- are \"shifted\" by the number of variables in the affine expression.

-- Example: convert the affine expression @(u0+1)*x0 + (u1+2)*x1 + (u2+3)@
-- to a polynomial:
--
-- > affineToPoly [u0+1,u1+2,u2+3] = (u2+1)*u0 + (u3+2)*u1 + (u4+3)
--
affineToPoly :: Num a
             => [Polynomial a]   -- ^ a
	     -> Polynomial a
affineToPoly polys =
    sum (liftPolyN nIndet (last polys) :
	 zipWith (*) [indeterminate (nIndet+nOldIndet) i | i <- [0..nIndet-1]]
	             (map (liftPolyN nIndet) polys)
	)
    where
    nOldIndet = unNIndeterminates $ nIndeterminatesN polys
    nIndet = genericLength polys - 1

---------------


-- |
-- A fraction of two polynomials.
--
data FracPoly a = FracPoly (Polynomial a) (Polynomial a)
		| Poly     (Polynomial a)

instance Show a => Show (FracPoly a) where
    showsPrec prec (FracPoly n d) =
	showParen (prec>=10)
                  (showString "FracPoly " . showsPrec 10 n .
		   showString " " . showsPrec 10 d)
    showsPrec prec (Poly p)       =
	showParen (prec>=10) (showString "Poly " . showsPrec 10 p)

instance Integral a => Show (FracPoly (Ratio a)) where
    show (FracPoly n d) = "(" ++ show n ++ ")/(" ++ show d ++ ")"
    show (Poly p)       = show p



nIndeterminatesFP :: Num a => FracPoly a -> NIndeterminates
nIndeterminatesFP (Poly p) = nIndeterminates p
nIndeterminatesFP (FracPoly a b) = max (nIndeterminates a) (nIndeterminates b)

-- |
-- Return the numerator of a 'FracPoly'.
--
polyNumerator :: FracPoly a -> Polynomial a
polyNumerator (FracPoly n _) = n
polyNumerator (Poly     p  ) = p

-- |
-- Return the denominator of a 'FracPoly'.
--
polyDenominator :: Num a => FracPoly a -> Polynomial a
polyDenominator (FracPoly _ d) = d
polyDenominator (Poly     p  ) = constPolyNp p 1


-- |
-- Map a function over the numerator and denominator of
-- a @FracPoly@.
--
mapFP :: (Num a, Fractional b) => (a -> b) -> FracPoly a -> FracPoly b
mapFP f = applyFP (mapPoly f)


-- |
-- Apply a function to the numerator and the denominator of
-- a @FracPoly@.
applyFP :: (Num a, Fractional b)
        => (Polynomial a -> Polynomial b) -> FracPoly a -> FracPoly b
applyFP f (FracPoly n d) = f n %% f d
applyFP f (Poly p)       = f p %% f 1

isConstFP :: Fractional a => FracPoly a -> Maybe a
isConstFP (FracPoly n d)
    | Just n' <- isConstPoly n, Just d' <- isConstPoly d  =  Just (n'/d')
isConstFP (Poly p)
    | Just p' <- isConstPoly p  =  Just p'
isConstFP _ = Nothing


{-
instance (Integral a, Show a) => Show (FracPoly a) where
    show fp
	| isConstPoly (polyDenominator fp) == Just 1 = show (polyNumerator fp)
	| otherwise = 
	    "(" ++ show (polyNumerator fp) ++ " %% " ++
	    show (polyDenominator fp) ++ ")"
-}
{-
instance (Fractional a, Show a) => Show (FracPoly a) where
    show fp
	| Just d <- isConstPoly (polyDenominator fp)  =
            show (fmap (/d) $ polyNumerator fp)
	| otherwise = 
	    "(" ++ show (polyNumerator fp) ++ " %% " ++
	    show (polyDenominator fp) ++ ")"
-}
-- wenn wir uns darauf verlassen, dass der Nenner _immer_
-- normiert ist, dann kann (==) effizienter definiert werden.
instance (Num a, Eq a) => Eq (FracPoly a) where
    (Poly p1)        == (Poly p2)         =  p1 == p2
    (Poly p)         == (FracPoly n d)    =  p*d == n
    (FracPoly n d)   == (Poly p)          =  n   == p*d
    (FracPoly n1 d1) == (FracPoly n2 d2)  =  n1*d2 == n2*d1

{-
eqFP :: (Eq a, Num a) => FracPoly a -> FracPoly a -> Bool
eqFP (Poly p1) (Poly p2) == p1 == p2
eqFP (Poly p) (FracPoly n d)    =  p*d == n
eqFP (FracPoly n d) (Poly p)          =  n   == p*d
    (FracPoly n1 d1) == (FracPoly n2 d2)  =  n1*d2 == n2*d1
-}

-- |
-- Construct a 'FracPoly'.
-- @p %% q@ represents the fraction with numerator @p@ and
-- denominator @q@.
--
(%%) :: Fractional a =>
	Polynomial a -> Polynomial a -> FracPoly a
(%%) _ (ZeroPolynomial _ _) = error "(%%): denominator is a 0-polynomial"
(%%) n' d' = unFrom2 "(%%)" mkFP n' d'
    where
    mkFP n d
	| g == d              =  Poly (n `quotPoly` d)
	| totalDegree g <= 0  =  FracPoly n d
	| otherwise           =  FracPoly (n `quotPoly` g) (d `quotPoly` g)
	where
	g = leadingCoeff d `pmul` gcdPoly n d

-- |
-- Make a 'FracPoly' with denominator 1.
-- @fromPoly p@ is the same as @p %% 1@, but more efficient.
--
fromPoly :: Fractional a => Polynomial a -> FracPoly a
fromPoly p = Poly p


-- |
-- Evaluate a 'FracPol' at a given point.
--
evalFP :: Fractional a => [a] -> FracPoly a -> a
evalFP pt fp =
    evalPoly (polyNumerator fp) pt / evalPoly (polyDenominator fp) pt

-- |
-- Make the coefficients in a @FracPoly@ integral.
-- Numerator and denominator of the given fraction are
-- multiplied by the least common multiple of the denominators
-- of the coefficients in the numerator and denominator
-- polynomials.
--
intFracPoly :: Integral a => FracPoly (Ratio a) -> FracPoly a
intFracPoly fp = FracPoly (mkInt pn) (mkInt pd)
    where
    mkInt p = mapPoly (\c -> (fact `div` denominator c) * numerator c) p

    fact  = abs $ foldl lcm 1 $ map denominator $
	    (collectCoeffs pn ++ collectCoeffs pd)
    pn = polyNumerator   fp
    pd = polyDenominator fp


instance Fractional a => Num (FracPoly a) where
    (+) (Poly p1) (Poly p2) = Poly (p1+p2)
    (+) (Poly p) (FracPoly n d) = (p*d+n) %% d
    (+) (FracPoly n d) (Poly p) = (p*d+n) %% d
    (+) (FracPoly n1 d1) (FracPoly n2 d2) = n %% d
        where
	n = n1 * d2 + n2 * d1
	d = d1 * d2
    negate (FracPoly n d) = FracPoly (negPoly n) d
    negate (Poly     p  ) = Poly     (negPoly p)
    (*) (Poly p1) (Poly p2) = Poly (p1*p2)
    (*) (Poly p) (FracPoly n d)
	| totalDegree g <= 0         =  FracPoly (p*n) d
	| g == d  =  Poly (n * (p `quotPoly` d))
	| otherwise             =  FracPoly ((p `quotPoly` g) * n) (d `quotPoly` g)
	where
	g = leadingCoeff d `pmul` gcdPoly p d
    (*) f1@(FracPoly _ _) f2@(Poly _) = f2 * f1
    (*) (FracPoly n1 d1) (FracPoly n2 d2) = n %% d
        where
	n = n1 * n2
	d = d1 * d2
    fromInteger n = fromPoly $ fromInteger n
    abs = undefined
    signum = undefined

instance Fractional a => Fractional (FracPoly a) where
    (/) (Poly p1) (Poly p2) = p1 %% p2
    (/) f@(Poly _) g@(FracPoly _ _) = f * recip g
    (/) (FracPoly n d) (Poly p)
	| totalDegree g <= 0         = 
	    FracPoly (recip (leadingCoeff p) `pmul` n) (d * monic p)
	| g == p  =  FracPoly (n `quotPoly` p) d
	| otherwise             =  FracPoly (n `quotPoly` g) (d * (p `quotPoly` g))
	where
	g = leadingCoeff p `pmul` gcdPoly n p
    (/) (FracPoly n1 d1) (FracPoly n2 d2) = n %% d
        where
	n = n1 * d2
	d = d1 * n2
    recip (FracPoly n d)
	| isZeroPoly n  =  error "recip: polynomial numerator is zero"
        | monN == 1     =  Poly num
        | otherwise     =  FracPoly num monN
	where
	num  = recip (leadingCoeff n) `pmul` d
	monN = monic n
    recip (Poly p)
	| totalDegree p <  0  =  error "recip: polynomial numerator is zero"
	| totalDegree p == 0  =  Poly (constPoly $ recip lc)
	| otherwise  =  FracPoly (constPoly $ recip lc) (monic p)
	where
	lc = leadingCoeff p
    fromRational r = fromInteger (numerator r) %% fromInteger (denominator r)


toPoly :: Fractional a => Polynomial (FracPoly a) -> Polynomial a
toPoly poly = flatPoly $ mapPoly mkPoly poly
    where
    (nums, denoms) = unzip $
		     map (\c -> (polyNumerator c, polyDenominator c)) $
		     collectCoeffs poly
    numGcd   = foldl1 gcdPoly nums
    denomGcd = foldl1 lcmPoly denoms

    mkPoly pf = (denomGcd `quotPoly` polyDenominator pf) * polyNumerator pf

flatFracPoly :: Fractional a => FracPoly (FracPoly a) -> FracPoly a
flatFracPoly (Poly p)       = flatFracPoly (FracPoly p $ constPolyNp p 1)
flatFracPoly (FracPoly n d) = flatPoly (mkP n) %% flatPoly (mkP d)
    where
    mkP = mapPoly (\fp -> polyNumerator fp * (lcm `quotPoly` polyDenominator fp))
    lcm = foldl1 lcmPoly $ map polyDenominator (collectCoeffs n ++ collectCoeffs d)




fracPolyToIFracPoly :: Integral a => FracPoly (Ratio a) -> IFracPoly a
fracPolyToIFracPoly fp =
    case intFracPoly fp of
      Poly a       -> IPoly a
      FracPoly a b -> IFracPoly a b

ifracPolyToFracPoly :: Integral a => IFracPoly a -> FracPoly (Ratio a)
ifracPolyToFracPoly (IPoly p) = Poly (mapPoly (% 1) p)
ifracPolyToFracPoly (IFracPoly p q) =
    mapPoly (% 1) p  %%  mapPoly (% 1) q
    

data IFracPoly a = IPoly     (Polynomial a)
		 | IFracPoly (Polynomial a) (Polynomial a)
		   deriving (Show)

{-
showIFracPolyWithNames :: [String] -> ShowS (IFracPoly a)
showIFracPolyWithNames names (IPoly 
-}

ifpNumerator :: Num a => IFracPoly a -> Polynomial a
ifpNumerator (IPoly a) = a
ifpNumerator (IFracPoly a b) = a

ifpDenominator :: Num a => IFracPoly a -> Polynomial a
ifpDenominator (IPoly a) = 1
ifpDenominator (IFracPoly a b) = b

(%%%) :: Num a => Polynomial a -> Polynomial a -> IFracPoly a
a %%% 1    = IPoly a
a %%% (-1) = IPoly $ negate a
a %%% b    = IFracPoly a b

applyIFP :: (Num a, Num b)
         => (Polynomial a -> Polynomial b) -> IFracPoly a -> IFracPoly b
applyIFP f (IPoly a)       = f a %%% f 1
applyIFP f (IFracPoly a b) = f a %%% f b

mapIFP :: (Num a, Num b) => (a -> b) -> IFracPoly a -> IFracPoly b
mapIFP f (IPoly a)       = mapPoly f a %%% mapPoly f 1
mapIFP f (IFracPoly a b) = mapPoly f a %%% mapPoly f b

nIndeterminatesIFP :: Num a => IFracPoly a -> NIndeterminates
nIndeterminatesIFP (IPoly a) = nIndeterminates a
nIndeterminatesIFP (IFracPoly a b) =
    max (nIndeterminates a) (nIndeterminates b)

isConstIFP :: Integral a => IFracPoly a -> Maybe (a,a)
isConstIFP (IPoly a) = fmap (\c -> (c,1)) $ isConstPoly a
isConstIFP (IFracPoly a b) =
    isConstPoly a >>= \a' -> isConstPoly b >>= \b' -> return (a', b')

flatIFracPoly :: Integral a => IFracPoly (IFracPoly a) -> IFracPoly a
flatIFracPoly ifp = fracPolyToIFracPoly $ flatFracPoly (num %% den)
    where
    num = mapPoly ifracPolyToFracPoly $ ifpNumerator   ifp
    den = mapPoly ifracPolyToFracPoly $ ifpDenominator ifp

instance Num a => Eq (IFracPoly a) where
    IPoly a == IPoly b              =  a == b
    IPoly a == IFracPoly b c        =  a*c == b
    IFracPoly b c == IPoly a        =  a*c == b
    IFracPoly a b == IFracPoly c d  =  a*d == b*c


instance Integral a => Num (IFracPoly a) where
    IPoly a + IPoly b = IPoly (a+b)
    IPoly a + b@(IFracPoly _ _) = IFracPoly a 1 + b
    b@(IFracPoly _ _) + IPoly a = IFracPoly a 1 + b
    IFracPoly a b + IFracPoly c d =
	fracPolyToIFracPoly ((mkR a %% mkR b) + (mkR c %% mkR d))
	where
	mkR = mapPoly (% 1)
    negate (IPoly a) = IPoly $ negate a
    negate (IFracPoly a b) = IFracPoly (negate a) b

    IPoly a * IPoly b       = IPoly (a*b)
    IPoly a * IFracPoly b c = IFracPoly (a*b) c
    IFracPoly a b * IPoly c = IFracPoly (a*c) b
    IFracPoly a b * IFracPoly c d = (a*c) %%% (b*d)

    fromInteger = IPoly . fromInteger

    signum = error "signum: undefined for Integral a => Num (IFracPoly a)"
    abs    = error "abs: undefined for Integral a => Num (IFracPoly a)"


instance Num a => Num (IFracPoly (Polynomial a)) where
    IPoly a + IPoly b = IPoly (a+b)
    IPoly a + b@(IFracPoly _ _) = IFracPoly a 1 + b
    b@(IFracPoly _ _) + IPoly a = IFracPoly a 1 + b
    IFracPoly a b + IFracPoly c d = (a*d+b*c) %%% (b*d)
    negate (IPoly a) = IPoly $ negate a
    negate (IFracPoly a b) = IFracPoly (negate a) b

    IPoly a * IPoly b       = IPoly (a*b)
    IPoly a * IFracPoly b c = IFracPoly (a*b) c
    IFracPoly a b * IPoly c = IFracPoly (a*c) b
    IFracPoly a b * IFracPoly c d = (a*c) %%% (b*d)

    fromInteger = IPoly . fromInteger

    signum = error "signum: undefined for Num a => Num (IFracPoly (Polynomial a))"
    abs    = error "abs: undefined for Num a => Num (IFracPoly (Polynomial a))"


instance Num a => Fractional (IFracPoly (Polynomial a)) where
    recip (IPoly a)       = 1 %%% a
    recip (IFracPoly a b) = b %%% a

    fromRational r = (fromIntegral $ numerator r) %%%
		     (fromIntegral $ denominator r)

instance Integral a => Fractional (IFracPoly a) where
    recip (IPoly a)       = 1 %%% a
    recip (IFracPoly a b) = b %%% a

    fromRational r = (fromIntegral $ numerator r) %%%
		     (fromIntegral $ denominator r)

--euclidQ :: Fractional a => FracPoly a -> FracPoly a -> FracPoly a
--euclidQ = undefined

{-
let (p,q) = (NonZeroPolynomial (Recursive [(2,Recursive [(0,Constant (1 % 1))]),(1,Recursive [(1,Constant (-2 % 1))]),(0,Recursive [(2,Constant (1 % 1)),(0,Constant (-2 % 1))])]),NonZeroPolynomial (Recursive [(2,Recursive [(0,Constant (1 % 1))]),(0,Recursive [(0,Constant (-4 % 1))])]))

    contP = content p
    contQ = content q
    primP = quotPoly p (liftPoly contP)
    primQ = quotPoly q (liftPoly contQ)
    euc   = toPoly $
	    euclid (fmap fromPoly $ univariate primP) (fmap fromPoly $ univariate primQ)




let p = NonZeroPolynomial (Recursive [(2,Constant (NonZeroPolynomial (Recursive [(0,Constant (1 % 1))]) %% NonZeroPolynomial (Recursive [(0,Constant (1 % 1))]))),(0,Constant (NonZeroPolynomial (Recursive [(0,Constant (-4 % 1))]) %% NonZeroPolynomial (Recursive [(0,Constant (1 % 1))])))])
let q = NonZeroPolynomial (Recursive [(1,Constant (NonZeroPolynomial (Recursive [(1,Constant (-2 % 1))]) %% NonZeroPolynomial (Recursive [(0,Constant (1 % 1))]))),(0,Constant (NonZeroPolynomial (Recursive [(2,Constant (1 % 1)),(0,Constant (2 % 1))]) %% NonZeroPolynomial (Recursive [(0,Constant (1 % 1))])))])

quotRemPoly p q


let quotient1 = NonZeroPolynomial (Recursive [(1,Constant (NonZeroPolynomial (Recursive [(0,Constant (1 % 1))]) %% NonZeroPolynomial (Recursive [(1,Constant (-2 % 1))])))])
let q = NonZeroPolynomial (Recursive [(1,Constant (NonZeroPolynomial (Recursive [(1,Constant (-2 % 1))]) %% NonZeroPolynomial (Recursive [(0,Constant (1 % 1))]))),(0,Constant (NonZeroPolynomial (Recursive [(2,Constant (1 % 1)),(0,Constant (2 % 1))]) %% NonZeroPolynomial (Recursive [(0,Constant (1 % 1))])))])

(quotient1 `multPoly` q)


let n = NonZeroPolynomial (Recursive [(2,Constant (1 % 1)),(0,Constant (2 % 1))])
let d = NonZeroPolynomial (Recursive [(1,Constant (-2 % 1))])
gcdPoly n d


let p = NonZeroPolynomial (Recursive [(1,Constant (NonZeroPolynomial (Constant (1 % 1)) %% NonZeroPolynomial (Constant (1 % 1))))])
let q = NonZeroPolynomial (Recursive [(0,Constant (NonZeroPolynomial (Constant (2 % 1)) %% NonZeroPolynomial (Constant (1 % 1))))])
quotRemPoly p q
-}
