-- |
-- This module provides functions to convert between these types:
--
--   (1) 'Ratio a' ('RatioConvertable')
--
--   (2) 'Polynomial (Ratio a)' ('PolyConvertable')
--
--   (3) 'FracPoly (Ratio a)' ('FracPolyConvertable')
--
-- It is always possible to /promote/ a type ((1) -> (2) -> (3)).
-- But the other direction may be undefined (e.g. FracPoly Polynomial)
-- or unsafe (e.g. Polynomial to Ratio fails if the polynomial
-- is non-constant).
--
module LMath.Types.Polynomial.Conversion
    (-- * Conversion type classes
     RatioConvertable(convertToRatio),
     CheckedRatioConvertable(convertToRatio_checked),
     PolyConvertable(convertToPoly),
     FracPolyConvertable(convertToFracPoly)
    )
where

import Data.Ratio
import LMath.Types.Polynomial.Base

-- |
-- An interface for all objects that can be converted to 'Ratio a'.
--
class RatioConvertable a where
    convertToRatio :: (Integral b)
	           => a -- ^ the object to be converted
		   -> Ratio b

-- |
-- An interface for all objects that can be converted to 'Ratio a'.
--
-- This class is similiar to 'RatioConvertable', but it will
-- return 'Nothing', if the element cannot be converted to 'Ratio a'.
--
-- Example: Non-constant polynomials cannot be converted to 'Ratio a'.
--
class CheckedRatioConvertable a where
    convertToRatio_checked :: (Integral b)
                           => a -- ^ the object to be converted
			   -> Maybe (Ratio b)

-- |
-- An interface for all objects that can be converted to 'Polynomial (Ratio a)'.
--
class PolyConvertable a where
    convertToPoly :: (Integral b)
		  => a -- ^ the object to be converted
		  -> Polynomial (Ratio b)

-- |
-- An interface for all objects that can be converted to 'FracPoly (Ratio a)'.
--
class FracPolyConvertable a where
    convertToFracPoly :: (Integral b)
		      => a -- ^ the object to be converted
		      -> FracPoly (Ratio b)


----------------------------------------------------------------------
-- The following code implements these type classes for
-- the most common data types in hsloopo.
--

----------------------------------------------------------------------
-- Integrals: Integer, Int
-- (all conversions are safe)

instance (Integral a) => RatioConvertable a where
    convertToRatio x = fromIntegral x % 1

instance (Integral a) => CheckedRatioConvertable a where
    convertToRatio_checked x = Just (fromIntegral x % 1)

instance (Integral a) => PolyConvertable a where
    convertToPoly = constPoly . fromIntegral

instance (Integral a) => FracPolyConvertable a where
    convertToFracPoly = fromPoly . convertToPoly


----------------------------------------------------------------------
-- Ratio:
-- (all conversions are safe)

instance (Integral a) => RatioConvertable (Ratio a) where
    convertToRatio x = let (n,d) = (numerator x, denominator x)
                       in (fromIntegral n) % (fromIntegral d)

instance (Integral a) => CheckedRatioConvertable (Ratio a) where
    convertToRatio_checked x = let (n,d) = (numerator x, denominator x)
			       in Just $ (fromIntegral n) % (fromIntegral d)

instance (Integral a) => PolyConvertable (Ratio a) where
    convertToPoly = constPoly . convertToRatio

instance (Integral a) => FracPolyConvertable (Ratio a) where
    convertToFracPoly = fromPoly . convertToPoly

----------------------------------------------------------------------
-- Polynomial (Ratio a)
-- (conversion to Ratio is unsafe!)

-- |
-- Converts a polynomial @p@ to 'Ratio a'.
-- This fails if @p@ is a non-constant polynomial!
--
instance (Num a, RatioConvertable a) => RatioConvertable (Polynomial a) where
    convertToRatio p =
	case isConstPoly p of
	  Just x -> convertToRatio x
	  Nothing  -> error ("RatioConvertable.convertToRatio: " ++
			     "cannot convert non-constant polynomials: " ++
			     show p)

-- |
-- Converts a polynomial @p@ to 'Ratio a'.
-- It will return 'Nothing', if @p@ is non-constant.
--
instance (Num a, CheckedRatioConvertable a) => CheckedRatioConvertable (Polynomial a) where
    convertToRatio_checked p =
	case isConstPoly p of
          (Just x) -> convertToRatio_checked x
	  Nothing  -> Nothing

instance (Num a, RatioConvertable a) => PolyConvertable (Polynomial a) where
    convertToPoly = mapPoly convertToRatio

instance (Num a, RatioConvertable a) => FracPolyConvertable (Polynomial a) where
    convertToFracPoly = fromPoly . convertToPoly

----------------------------------------------------------------------
-- FracPoly (Ratio a)
--
-- Conversion FracPoly to Polynomial is problematic, because it cannot be
-- done in general without multiplying with the most common denominator.
-- 
-- If you need this functionality take a look at 'unfracIneqSystem'
-- (in LMath.Logic.Qe), which needs the helper function 'impliesSignum'
-- (in LMath.Logic.LogicOptions).

instance (Num a, RatioConvertable a) => FracPolyConvertable (FracPoly a) where
    convertToFracPoly = mapFP convertToRatio
