-- |
-- Residue class field arithmetic (for arithmetic modulo the
-- minimal polynomial of an algebraic number).
--
-- Let @F@ be a sub-field of the real numbers. For any polynomial
-- @a@ from @F[X]@, the ring
-- @F[X]\/(a)@ is a field, if and only if @a@ is irreducible.
-- This module offers arithmetic in @F[X]\/(a)@, @a@ must be irreducible.
--
module LMath.Types.Polynomial.ModArith
    ( ModArith,
      modArith, getModArith
    )
where

import LMath.Types.Polynomial.Base  hiding ( euclid )
import LMath.Types.Polynomial.Roots
import LMath.Types.Polynomial.Univariate

import LMath.Tools.Reduce

import Debug.Trace

euclid :: RealFrac a => Polynomial a -> Polynomial a -> Polynomial a
euclid = reduceGcdPoly

data RealFrac a =>
    ModArith a = FromNumM a
	       | ModArith (Polynomial a) (Polynomial a)
		 deriving (Show)

modArith :: RealFrac a =>
	    Polynomial a -> Polynomial a -> ModArith a
modArith p modulus = -- ModArith (p `remPoly` modulus) modulus r
--    (\x -> trace (show x) x) $
    ModArith (p `remPoly` modulus) modulus



getModulus :: RealFrac a => ModArith a -> Polynomial a
getModulus (ModArith _ modulus) = modulus
getModulus (FromNumM _) = error "getModulus: undefined for FromNumM"

getModArith :: RealFrac a => ModArith a -> Polynomial a
getModArith (ModArith p _) = p
getModArith (FromNumM _) = error "getModArith: undefined for FromNumM"

addModArith :: RealFrac a => ModArith a -> ModArith a -> ModArith a
addModArith (FromNumM a) (FromNumM b) = FromNumM (a+b)
addModArith (ModArith p modulus) (ModArith q _) = ModArith s modulus
    where
    s = (p+q) `remPoly` modulus

negModArith :: RealFrac a => ModArith a -> ModArith a
negModArith (FromNumM a) = FromNumM (-a)
negModArith (ModArith p modulus) =
    ModArith ((modulus-p) `remPoly` modulus) modulus

multModArith :: RealFrac a => ModArith a -> ModArith a -> ModArith a
multModArith (FromNumM a) (FromNumM b) = FromNumM (a*b)
multModArith (ModArith p modulus) (ModArith q _) = ModArith pr modulus
    where
    pr = (p*q) `remPoly` modulus

recipModArith :: (RealFrac a, Ord a) => ModArith a -> ModArith a
recipModArith (FromNumM 0) = error "recipModArith: not defined for 0"
recipModArith (FromNumM a) = FromNumM (recip a)
recipModArith (ModArith p modulus)
    | isZeroPoly p  =  error "recipModArith: not defined for 0"
    | otherwise     =  ModArith inverse modulus
    where
    inverse = s `remPoly` modulus
    (e,s,t) = extEuclid p modulus
{-
    | isAroot g r  =  error "recipModArith: not defined for 0"
    | otherwise    =  ModArith inverse modulus r
    where
    g = euclid p modulus
    mod' = modulus `quotPoly` g
    p'   = p `remPoly` mod'
    inverse = s `remPoly` mod'
    (e,s,t) = extEuclid p' mod'
-}

{-
isZeroEquiv :: (RealFrac a, Ord a) =>
	       Polynomial a -> Polynomial a -> Root a -> Bool
isZeroEquiv p m r = isAroot (euclid p m) r
-}

eqModArith :: (RealFrac a, Ord a) => ModArith a -> ModArith a -> Bool
eqModArith (FromNumM a) (FromNumM b)          =  a == b
eqModArith (ModArith p1 m1) (ModArith p2 m2)  = 
    p1 == p2 && m1 == m2
--    m1 == m2 && r1 == r2 && isZeroEquiv (p1-p2) m1 r1

unFromM2 :: RealFrac a =>
	    (ModArith a -> ModArith a -> b) -> ModArith a -> ModArith a -> b
unFromM2 f (FromNumM a) m@(ModArith p modulus) =
    f (ModArith (constPolyNp p a) modulus) m
unFromM2 f m@(ModArith p modulus) (FromNumM a) =
    f m (ModArith (constPolyNp p a) modulus)
unFromM2 f a b = f a b

instance (RealFrac a, Ord a) => Eq (ModArith a) where
    (==) = unFromM2 eqModArith

instance (RealFrac a, Ord a) => Num (ModArith a) where
    (+) = unFromM2 addModArith
    (*) = unFromM2 multModArith
    negate = negModArith

    fromInteger = FromNumM . fromInteger
    abs = error "`abs' undefined for (ModArith a)"
    signum = error "`signum' undefined for (ModArith a)"

instance (RealFrac a, Ord a) => Fractional (ModArith a) where
    recip        = recipModArith
    fromRational = FromNumM . fromRational
