module LMath.Types.Polynomial.Roots
    ( Root(..)
    , allRoots, nRoots, rootInterval

    , PolyAndRoots(..),
      roots, rootsND,
      makeRoot, makeRootsDisjoint, rootsAreDisjoint,
      signumPoly,
      addRootIntervals, multRootIntervals, potRootInterval,
      addToRootInterval, multRootInterval,
      nstIn,
      intersectRoots,
      integralRoots,
      sturm,
      factorWithRoot,
      isolate, isolate2, isolateUntil
    )
where

import Data.Array
import Data.Ratio
import Data.List

import LMath.Types.Polynomial.Base
import LMath.Types.Polynomial.Univariate
import LMath.Tools.Reduce

import Debug.Trace ( trace )



-- |
-- Return all the real roots of a polynomial
-- (isolation intervals). The roots are returned
-- in ascending order
--
allRoots :: RealFrac a => Polynomial a -> [[Root a]]
allRoots p = [iterate (isolate2 s) r | r <- isolate s]
    where
      s = sqfPart p

nRoots :: RealFrac a => Polynomial a -> Integer
nRoots p = genericLength $ allRoots p

rootInterval :: Root a -> (a,a)
rootInterval (RootAt x)       = (x,x)
rootInterval (RootIn x y _ _) = (x,y)









data Root a = RootAt a
	    | RootIn a a Ordering Ordering
	      deriving (Show)

instance (Fractional a, Ord a) => Eq (Root a) where
    RootAt x == RootAt y  =  x == y
    RootIn x1 x2 _ _ == RootIn y1 y2 _ _  =  x1 == y1 && x2 == y2
    _ == _  =  False

addRootIntervals :: (Fractional a, Ord a) => Root a -> Root a -> Root a
addRootIntervals (RootAt a) (RootAt b) = RootAt (a+b)
addRootIntervals (RootAt a) (RootIn b c _ _) =
    RootIn (a+b) (a+c) undefined undefined
addRootIntervals (RootIn b c _ _) (RootAt a) =
    RootIn (a+b) (a+c) undefined undefined
addRootIntervals (RootIn a b _ _) (RootIn c d _ _) =
    RootIn (a+c) (b+d) undefined undefined

multRootIntervals :: (Fractional a, Ord a) => Root a -> Root a -> Root a
multRootIntervals r@(RootAt 0) _ = r
multRootIntervals _ r@(RootAt 0) = r
multRootIntervals (RootAt a) (RootAt b) = RootAt (a*b)
multRootIntervals (RootAt a) (RootIn b c _ _) =
    RootIn r1 r2 undefined undefined
    where
    (r1,r2)
	| a > 0      =  (a*b,a*c)
	| otherwise  =  (a*c,a*b)
multRootIntervals (RootIn b c _ _) (RootAt a) =
    RootIn r1 r2 undefined undefined
    where
    (r1,r2)
	| a > 0      =  (a*b,a*c)
	| otherwise  =  (a*c,a*b)
multRootIntervals (RootIn a b _ _) (RootIn c d _ _) =
    RootIn (minimum rs) (maximum rs) undefined undefined
    where
    rs = [a*c,b*d,a*d,b*c]

potRootInterval :: (Fractional a, Ord a) => Root a -> Integer -> Root a
potRootInterval (RootAt x) e     = RootAt (x^e)
potRootInterval (RootIn x y _ _) e
    | x >= 0           =  RootIn x' y' undefined undefined
    | x < 0 && y >= 0  =
        if even e then RootIn 0  (max x' y')  undefined undefined
	          else RootIn x' y'           undefined undefined
    | otherwise        =  RootIn (min x' y') (max x' y') undefined undefined
    where
    x' = x ^ e
    y' = y ^ e

addToRootInterval :: (Fractional a, Ord a) => Root a -> a -> Root a
addToRootInterval (RootAt x) a         = RootAt (x+a)
addToRootInterval (RootIn x y s1 s2) a = RootIn (x+a) (y+a) s1 s2

multRootInterval :: (Fractional a, Ord a) => Root a -> a -> Root a
multRootInterval r a = multRootIntervals r (RootAt a)


rootsAreDisjoint :: (Fractional a, Ord a) => Root a -> Root a -> Bool
rootsAreDisjoint r1 r2   =  intersectRoots r1 r2 == Nothing


intersectRoots :: (Fractional a, Ord a) => Root a -> Root a -> Maybe (Root a)
intersectRoots r@(RootAt x) (RootAt y)
    | x == y     =  Just r
    | otherwise  =  Nothing
intersectRoots r@(RootAt x) (RootIn y1 y2 _ _)
    | y1 <= x && x <= y2  =  Just r
    | otherwise           =  Nothing
intersectRoots (RootIn y1 y2 _ _) r@(RootAt x)
    | y1 <= x && x <= y2  =  Just r
    | otherwise           =  Nothing
intersectRoots (RootIn x1 x2 _ _) (RootIn y1 y2 _ _)
    | a == b     =  Just (RootAt a)
    | a <  b     =  Just (RootIn a b undefined undefined)
    | otherwise  =  Nothing
    where
    (a,b) = (max x1 y1, min x2 y2)

-- the signums in `parRoots' are signums of the square-free part
-- (of course) !!!
data (Fractional a, Ord a) =>
    PolyAndRoots a = PAR { -- parPoly   :: Polynomial a,
			   parSqfree :: Polynomial a,
			   parRoots  :: [Root a]
			 }
		     deriving Show
					     


sturm :: (Fractional a, Ord a) => Polynomial a -> [Polynomial a]
sturm p = takeWhile (not . isZeroPoly) sturmSeq
    where
    sturmSeq = p : derivative p : zipWith negRem sturmSeq (tail sturmSeq)
    negRem a b = negate rem
	where
	(_,rem) = a `quotRemPoly` b

nstIn :: (Fractional a, Ord a) => [Polynomial a] -> a -> a -> Integer
nstIn sturmSeq x1 x2 =
    variation sturmSeq x1 - variation sturmSeq x2

variation :: (Num a, Ord a) => [Polynomial a] -> a -> Integer
variation sturmSeq x =
    signChanges (filter (/=0) $ map (flip evalPoly [x]) sturmSeq) 

signChanges :: (Num a, Ord a) => [a] -> Integer
signChanges []  = 0
signChanges [_] = 0
signChanges (x:y:ys)
    | x < 0 && y > 0  =  1 + signChanges (y:ys)
    | x > 0 && y < 0  =  1 + signChanges (y:ys)
    | otherwise       =  signChanges (y:ys)


-- for every root `a' of `p'
--    |a| < rootBound p
-- holds (N.B. strictly less than!)
rootBound :: (Fractional a, Ord a) => Polynomial a -> a
rootBound p = 2 + (maximum (0:map abs coeffs) / abs lc)
    where
    lc:coeffs = collectCoeffs p


makeRoot :: (Fractional a, Ord a) => Polynomial a -> a -> a -> Root a
makeRoot p x y
    | x == y     =  RootAt x
    | otherwise  =  RootIn x y (signumPoly p x) (signumPoly p y)

signumPoly :: (Num a, Ord a) => Polynomial a -> a -> Ordering
signumPoly p x
    | value <  0  =  LT
    | value == 0  =  EQ
    | value >  0  =  GT
    where
    value = evalPoly p [x]


-- qfP *must* be square-free !!
isolate :: (Fractional a, Ord a) => Polynomial a -> [Root a]
isolate qfPr = recIso (-b) b (variation sturmSeq (-b)) (variation sturmSeq b)
    where
    b = rootBound qfPr
    sturmSeq = sturm qfPr

    recIso x1 x2 v1 v2
	| v1 - v2 == 0  =  []
	| v1 - v2 == 1  =  
	    [RootIn x1 x2 (signumPoly qfPr x1) (signumPoly qfPr x2)]
	| evalPoly qfPr [xm] == 0  =  recIso x1 xm1 v1 vm1 ++
				      [RootAt xm] ++ recIso xm2 x2 vm2 v2
	| otherwise = recIso x1 xm v1 vm ++ recIso xm x2 vm v2
	where
	xm = (x1+x2) / 2
	vm = variation sturmSeq xm

        -- nicht ganz richtig (?) ...
        minRoot = 1 / (rootBound (reflectPoly qfPr) + xm)

	xm1 = xm - minRoot
	xm2 = xm + minRoot
	vm1 = variation sturmSeq xm1
	vm2 = variation sturmSeq xm2


-- only well defined if NOT (one of the intervals is contained in
-- the other interval).
instance (Fractional a, Ord a) => Ord (Root a) where
    compare (RootAt x) (RootAt y)              =  compare x y
    compare (RootAt x) (RootIn y1 y2 _ _)
	| x <= y1  =  LT
	| x >= y2  =  GT
	| otherwise = EQ
    compare (RootIn x1 x2 _ _) (RootAt y)
	| x2 <= y  =  LT
	| y <= x1  =  GT
	| otherwise = EQ
    compare (RootIn x1 x2 _ _) (RootIn y1 y2 _ _)
	| x1 < y1 && x2 <= y2  =  LT
	| x2 > y2 && x1 >= x1  =  GT
	| otherwise            =  EQ


refineRoot :: (Fractional a, Ord a) =>
	      (Polynomial a -> Root a -> Either Bool (Root a)) ->
	      Polynomial a -> Root a -> Maybe (Root a)
refineRoot f qfP root@(RootAt x) =
    case f qfP root of
        Right r    -> Just r
	Left False -> Nothing
	Left True  ->
	    error "refineRoot: don't know how to continue with `RootAt'"    
refineRoot f qfP (RootIn a b sigA sigB) =
    refine a b sigA sigB
    where
    refine x1 x2 sig1 sig2 =
	case f qfP (RootIn x1 x2 sig1 sig2) of
	    Right r    -> Just r
	    Left False -> Nothing
	    Left True  ->
		if sigm == EQ
		   then refineRoot f qfP (RootAt xm)
		   else if sig1 /= sigm
			  then refine x1 xm sig1 sigm
			  else refine xm x2 sigm sig2
	where
	xm   = (x1 + x2) / 2
	sigm = signumPoly qfP xm


acceptIntRoot :: RealFrac a =>
		 Polynomial a
              -> Root a
              -> Either Bool (Root a)
acceptIntRoot _ root@(RootAt x)
    | frac == 0  =  Right root
    | otherwise  =  Left False
    where
    (_, frac) = properFraction x
acceptIntRoot qfP (RootIn x1 x2 _ _)
    | x2  - x1  >= 1  =  Left True   -- more than one integer
    | x1c > x2f       =  Left False  -- no integer
    | evalPoly qfP [fromIntegral x1c] == 0  =  Right $ RootAt $ fromIntegral x1c
    | otherwise                             =  Left False
    where
    x1c = ceiling x1
    x2f = floor   x2

acceptRootI1 :: RealFrac a =>
		Polynomial a -> Root a -> Either Bool (Root a)
acceptRootI1 _ r@(RootAt _) = Right r
acceptRootI1 qfP r@(RootIn x1 x2 _ _)
    | x2 - x1 < 1  =  
	if x1c > x2f
	   then Right r
	   else if evalPoly qfP [fromIntegral x1c] == 0
		   then Right $ RootAt x1
		   else Left True
    | otherwise    =  Left True
    where
    x1c = ceiling x1
    x2f = floor   x2

--intervalsLt1 :: RealFrac a => Polynomial a -> [Root a] -> [Root a]
--intervalsLt1 sqfPr rs = [r | Just r <- map (refineRoot acceptRootI1 sqfPr) rs]

--
-- try to find some "special" roots.
-- If 'p' is squarefree, the result polynomial is also squarefree!
--
guessRoots :: RealFrac a
           => Polynomial a
           -> (Polynomial a, [a])
guessRoots p
    | constantCoeff p == 0  =
	let (p', rs) = guessRoots (p `quotPoly` (indeterminate 1 0) ^ miniExp)
        in  (p', 0:rs)
    | degree p == 1  =  (constPolyN 1 1, [-constantCoeff p/leadingCoeff p])
    | degree p == 2  =
        let a  = coeffNr 2
            b  = coeffNr 1
            c  = coeffNr 0
            d  = toRational (b*b - 4*a*c)
            d' = case (intSqrt $ numerator d, intSqrt $ denominator d) of
                    (Just nu, Just de) -> Just (nu % de)
                    _                  -> Nothing
        in case d' of
            Nothing -> (p, [])
            Just 0  -> (constPolyN 1 1, [-b / (2*a)])
            Just s  -> (constPolyN 1 1, [(-b-fromRational s) / (2*a), (-b+fromRational s) / (2*a)])
    | otherwise = (p, []) 
    where
    miniExp = snd $ last cas
    cas = coeffsAndExps p

    coeffNr n =
        case find ((==n).snd) cas of
          Just (co,_) -> co
          Nothing     -> 0

intSqrt :: Integer -> Maybe Integer
intSqrt n = foldl halfmul (Just 1) $ reduceFactorise n
    where
    halfmul (Just v) (prime,expo)
        | even expo  =  Just (v * prime^(expo `div` 2))
    halfmul _ _  =  Nothing
{-
    trace (show n) $
    case span (\s -> s*s < n) [0..] of
              (_,x:_)
                 | x*x == n   ->  Just x
                 | otherwise  ->  Nothing
-}            

isolate2 :: (Fractional a, Ord a) =>
	    Polynomial a -> Root a -> Root a
isolate2 _ r@(RootAt x) = r
isolate2 sqfPoly (RootIn x1 x2 sig1 sig2)
    | sigm == EQ   = RootAt xm
    | sigm /= sig1 = RootIn x1 xm sig1 sigm
    | sigm /= sig2 = RootIn xm x2 sigm sig2
    | otherwise =
	error "isolate2: same signum"
    where
    xm   = (x1 + x2) / 2
    sigm = signumPoly sqfPoly xm

isolateUntil :: (Fractional a, Ord a) =>
		(Root a -> Bool) -> Polynomial a -> Root a -> Root a
isolateUntil f p root = iso root
    where
    iso r
	| f r       = r
	| otherwise = iso (isolate2 p r)


makeRootsDisjoint :: (Fractional a, Ord a) =>
		     (PolyAndRoots a, PolyAndRoots a) ->
		     (PolyAndRoots a, PolyAndRoots a)
makeRootsDisjoint (par1, par2) =
    let (roots1rev, roots2rev) = disj (parRoots par1) (parRoots par2) [] []
    in  (par1 { parRoots = reverse roots1rev },
	 par2 { parRoots = reverse roots2rev })
    where
    disj [] [] accx accy = (accx, accy)
    disj [] (y:yrest) accx accy = disj [] yrest accx (y:accy)
    disj (x:xrest) [] accx accy = disj xrest [] (x:accx) accy
    disj xs@(x@(RootAt x1):xrest) ys@(y@(RootAt y1):yrest) accx accy
	| x1 < y1 = disj xrest ys (x:accx) accy
	| y1 < x1 = disj xs yrest accx (y:accy) 
    disj xs@(x@(RootAt x1):xrest) ys@(y@(RootIn y1 y2 _ _):yrest) accx accy
	| y1 <= x1 && x1 <= y2 = disj xs (isolate2 sqfP2 y : yrest) accx accy
	| x1 < y1              = disj xrest ys (x:accx) accy
	| y2 < x1              = disj xs yrest accx (y:accy)
    disj xs@(x@(RootIn x1 x2 _ _):xrest) ys@(y@(RootAt y1):yrest) accx accy
	| x1 <= y1 && y1 <= x2 = disj (isolate2 sqfP1 x : xrest) ys accx accy
	| x2 < y1              = disj xrest ys (x:accx) accy
	| y1 < x1              = disj xs yrest accx (y:accy)
    disj xs@(x@(RootIn x1 x2 _ _):xrest) ys@(y@(RootIn y1 y2 _ _):yrest) 
	 accx accy
	| x2 < y1 = disj xrest ys (x:accx) accy
	| y2 < x1 = disj xs yrest accx (y:accy)
	| otherwise =
	    disj (isolate2 sqfP1 x : xrest) (isolate2 sqfP2 y : yrest)
		 accx accy
    disj xs ys accx accy =
	error ("disj: unhandled case " ++ show (xs,ys,accx,accy) ++ "\n"
	       ++ show (par1,par2))
    sqfP1 = parSqfree par1
    sqfP2 = parSqfree par2


-- retain order of list elements !!
mutatingPairs :: ((a,a) -> (a,a)) -> [a] -> [a]
mutatingPairs _ [] = []
mutatingPairs f (x:xs) = x' : mutatingPairs f xs'
    where
    (x',xs') = mapAccumL (curry f) x xs

intRoot :: RealFrac a =>
	   Polynomial a -> Root a -> Maybe (Root a)
intRoot p r = refineRoot acceptIntRoot p r
	          

-- 
-- may throw away a common factor of the coefficients,
-- i.e., simpleFactoriseP (2*x-2)  can be  [(x-1,1)]
simpleFactoriseP :: Integral a => Polynomial a -> [(Polynomial a, Integer)]
simpleFactoriseP f
    | deg <= 0   =  []
    | deg == 1   =  [(mapPoly (`div` cont) f, 1)]
    | minExp > 0 =  (polyX, fromIntegral minExp) : simpleFactoriseP rest
    | otherwise  =  snd $ reduceFactoriseP f
    where
      deg = totalDegree f
      cas = coeffsAndExps f
      minExp = snd $ last cas
      rest = sum [co `pmul` (polyX^(e-minExp)) | (co,e) <- cas]
      cont = foldl1 gcd $ map fst cas

-- |
-- Return the roots of the given polynomials.
-- It is guaranteed that the isolation intervals are
-- pairwise disjoint.
--
roots :: RealFrac a => [Polynomial a] -> [PolyAndRoots a]
roots = mutatingPairs makeRootsDisjoint . map polyAndRoots

-- |
-- Return the roots of a _factored_ polynomial.
--
polyAndRoots :: RealFrac a => Polynomial a -> PolyAndRoots a
polyAndRoots p =
    PAR { -- parPoly   = p,
	  parSqfree = sqfPr,
	  parRoots  = map RootAt someRoots ++
                      (if degree p' >= 1 then isolate p' else [])
	}
    where
    -- since we require 'p' to be factored, 'p' is also squarefree
    sqfPr  = p -- monic (p `quotPoly` {-gcdPoly-} euclid p (derivative p))
    (p', someRoots) = guessRoots sqfPr

    (_,ip) = ratPoly2intPoly $ mapPoly toRational p
    fs = simpleFactoriseP ip    


-- |
-- Factorise every given polynomial and return representations
-- for all the roots of the polynomials (without making the
-- isolation intervals necessarily disjoint). Ignores 0 polynomials.
--
rootsND :: RealFrac a => [Polynomial a] -> [PolyAndRoots a]
rootsND ps =
    --roots (gcdFree ps)
    -- gcdFree get rid of 0 polynomials
    -- Oh, we make the isolation intervals disjoint...
    map polyAndRoots $ gcdFree [p' | p <- ps, p' <- fact p]
    --roots $ map snd $ sqff $ foldl1 (*) $
    --filter (\p -> isConstPoly p == Nothing) ps
    where
    fact p = [mapPoly fromIntegral f | (f,_) <- fs]
        where
        (_,ip) = ratPoly2intPoly $ mapPoly toRational p
        fs = simpleFactoriseP ip


integralRoots :: RealFrac a => Polynomial a -> [Integer]
integralRoots p = [i | r <- roots, Just (RootAt i') <- [intRoot sqf r], let (i,_) = properFraction i']
    where
    pars = rootsND [p]
    roots = concatMap parRoots pars
    sqf   = parSqfree $ head pars

-- |
-- Factor the given polynomial and return the (squarefree part of the)
-- factor which has the given root.
--
factorWithRoot :: RealFrac a => Root a -> Polynomial a -> Polynomial a
factorWithRoot r p
    | eisensteinsCriterion p  =  monic p
    | otherwise               =
        monic $
        head [f | f' <- fs, let f = mapPoly fromIntegral f', isAroot f r] 
    where
    (_,fs') = reduceFactoriseP $ snd $ ratPoly2intPoly $ mapPoly toRational p
    fs = map fst fs'


eisensteinsCriterion :: Real a => Polynomial a -> Bool
eisensteinsCriterion p'
    | degree p' <= 1  =  True
    | cc == 0         =  False
    | otherwise       =  any eisenstein primes
    where
    (_,p) = ratPoly2intPoly $ mapPoly toRational p'
    (hc:lowerCoeffs,exps) = unzip $ coeffsAndExps p
    cc = constantCoeff p

    primes = map fst $ reduceFactorise $ foldl gcd 0 lowerCoeffs

    eisenstein prime = hc `mod` prime /= 0  &&  cc `mod` (prime*prime) /= 0

-- p is squarefree and has at most one root in the interval !!!
isAroot :: (RealFrac a, Ord a) =>
	   Polynomial a -> Root a -> Bool
isAroot p (RootAt x)        =  signumPoly p x == EQ
isAroot p (RootIn x y _ _)  =  signumPoly p x /= signumPoly p y



{-
rootsND [NonZeroPolynomial (Recursive [(0,Constant ((-1) % 1))]),FromNum (1 % 1),NonZeroPolynomial (Recursive [(1,Constant (1 % 1))]),FromNum (1 % 1),NonZeroPolynomial (Recursive [(0,Constant (1 % 1))]),FromNum (1 % 1),NonZeroPolynomial (Recursive [(0,Constant ((-1) % 1))]),FromNum (1 % 1),NonZeroPolynomial (Recursive [(1,Constant (1 % 1)),(0,Constant ((-1) % 1))]),FromNum (1 % 1),NonZeroPolynomial (Recursive [(0,Constant ((-1) % 1))]),NonZeroPolynomial (Recursive [(2,Constant (1 % 1)),(1,Constant (2 % 1)),(0,Constant (1 % 1))]),FromNum (1 % 1),NonZeroPolynomial (Recursive [(1,Constant (1 % 1)),(0,Constant ((-1) % 1))]),FromNum (1 % 1),NonZeroPolynomial (Recursive [(1,Constant (1 % 1))]),FromNum (1 % 1),NonZeroPolynomial (Recursive [(1,Constant ((-1) % 1)),(0,Constant (1 % 1))]),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),FromNum (0 % 1),FromNum (1 % 1),NonZeroPolynomial (Recursive [(2,Constant ((-1) % 1))]),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),NonZeroPolynomial (Recursive [(1,Constant (2 % 1)),(0,Constant ((-2) % 1))]),FromNum (1 % 1),NonZeroPolynomial (Recursive [(1,Constant (2 % 1)),(0,Constant ((-1) % 1))]),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1),FromNum (1 % 1)]


[u0, u0*1-1, u0*u0+u0*2+1, u0, u0*u0*(-1), u0*2-1]


-}
