module LMath.Types.Polynomial.Show
    ( showPoly, showPolyStd, showPolyC, showPolyExt
    , parsePoly
    )
where

import LMath.Types.Polynomial.Base
import LMath.Types.Polynomial.Univariate

import Control.Exception ( assert )
import Data.List
import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Expr
import qualified Text.ParserCombinators.Parsec.Token as P
import Text.ParserCombinators.Parsec.Language

data ParenNeed = Sum
	       | Product
	       | Power
	       | NoParen
		 deriving (Eq, Ord, Show)

data ShowPolyOptions = SPO { spoHatForExp :: Bool  -- use p^2 for p*p etc.
                           }
                       deriving Show

ensureLength :: Int -> [a] -> Bool
ensureLength n _ | n <= 0 = True
ensureLength n (x:xs)     = ensureLength (n-1) xs
ensureLength _ []         = False

showPoly :: Num a
	 => ShowPolyOptions
         -> (a -> (Bool, (ParenNeed, String)))
         -> [String]
	 -> Polynomial a
	 -> String
showPoly spo showCoeff varNames origPoly =
    assert (ensureLength (unNIndeterminates (nIndeterminates origPoly)) varNames) $
    let (_,(_,text)) = showP 0 origPoly in text
    where
    showP varNr p
	| Just const <- isConstPoly p  =  showCoeff const
	| otherwise  =  case monoms of
			  [m]       -> m
			  ((neg,m1):rest) -> (neg, (Product,
						    withParen Sum m1 ++ sumup rest))
	where
	monoms = map (\(c,e) -> withExpo c e varNr) $
		 coeffsAndExps $ univariate p

    sumup [] = ""
    sumup ((True, s1):rest) = withParen Sum s1 ++ sumup rest
    sumup ((False,s1):rest) = "+" ++ withParen Sum s1 ++ sumup rest

    withExpo c 0 varNr = showP (varNr+1) c
    withExpo 1 1 varNr = (False, (NoParen, varNames !! varNr))
    withExpo (-1) 1 varNr = (True, (Power, '-' : (varNames !! varNr)))
    withExpo 1 e varNr = (False, (Power, pot (varNames!!varNr) e))
    withExpo c 1 varNr = (isNeg && pn>Product,
			   (Power, withParen Product pt ++
			           ('*' : (varNames!!varNr))
			   )
			 )
	where
	(isNeg, pt@(pn,_)) = showP (varNr+1) c
    withExpo c e varNr = (isNeg && pn>Product,
			    (Power, withParen Product pt ++
			            ('*' : pot (varNames!!varNr) e)
			    )
			 )
	where
	(isNeg, pt@(pn,_)) = showP (varNr+1) c

    withParen parenUse (pn, text)
	| pn > parenUse = text
	| otherwise     = '(' : (text ++ ")")

    pot v e
        | spoHatForExp spo = v ++ ('^' : show e)
        | otherwise        = concat $ intersperse "*" $ replicate e v

showPolyStd' :: (Num a, Ord a)
	    => ShowPolyOptions
            -> [String]
            -> Polynomial a
            -> String
showPolyStd' spo names poly =
    showPoly spo (\c -> (c<0,(NoParen,show c))) names poly

-- |
-- Show a polynomial in "standard" syntax.
--
showPolyStd :: (Num a, Ord a)
	    => [String]
            -> Polynomial a
            -> String

-- |
-- Show a polynomial in a syntax suitable for C and C-like languages.
--
showPolyC   :: (Num a, Ord a)
	    => [String]
            -> Polynomial a
            -> String
showPolyStd = showPolyStd' (SPO { spoHatForExp = True })
showPolyC   = showPolyStd' (SPO { spoHatForExp = False })


showPolyExt :: (Num a)
	    => [String]
	    -> (a -> String)
	    -> (a -> Bool)
            -> Polynomial a
            -> String
showPolyExt names showCo isNeg poly =
    showPoly (SPO { spoHatForExp = True })
             (\c -> (isNeg c, (NoParen, showCo c))) names poly



parsePoly :: Num a => [String] -> CharParser st (Polynomial a)
parsePoly names = spaces >> embed names expr
    
embed :: st -> GenParser tok st a -> GenParser tok st' a
embed st inner =
    do input <- getInput
       pos   <- getPosition
       case runParser (inner' pos) st "" input of
         Right (res,input2,pos2) ->
             do setInput input2
                setPosition pos2
                return res
         Left  err          -> fail $ show err
    where
    inner' pos =
        do setPosition pos
           res <- inner
           inp2 <- getInput
           pos2 <- getPosition
           return (res,inp2,pos2)


lexer :: P.TokenParser [String]
lexer  = P.makeTokenParser 
         (haskellDef
          { reservedOpNames = ["*","+","-"]
          }
         )

identifier = P.identifier lexer
natural    = P.natural lexer
reservedOp = P.reservedOp lexer

expr    :: Num a => CharParser [String] (Polynomial a)
expr    = buildExpressionParser table factor
          <?> "expression"

table :: Num a => OperatorTable Char [String] (Polynomial a) 
table   = [[op1 "-" negate]
          ,[op2 "*" (*) AssocLeft]
          ,[op2 "+" (+) AssocLeft, op2 "-" (-) AssocLeft]
          ]          
        where
          op1 s f = Prefix (do{ reservedOp s; return f } <?> "operator")
          op2 s f assoc
             = Infix (do{ reservedOp s; return f} <?> "operator") assoc

factor :: Num a => CharParser [String] (Polynomial a)
factor = do f <- factor1
            e <- expo
            return (f ^ e)

factor1 :: Num a => CharParser [String] (Polynomial a)
factor1 = do char '('
             e <- expr
             char ')'
             return e
          <|>
          do names <- getState
             i <- identifier
             case findIndex (==i) names of
               Just idx -> return $ indeterminate (length names) idx
               Nothing  -> fail ("unknown indeterminate: " ++ show i ++
                                 ", expecting one of " ++
                                 concat (intersperse ", " $ map show names)
                                )
          <|>
          do i <- natural
             return $ fromIntegral i
            

expo :: CharParser [String] Integer
expo = (char '^' >> natural) <|> return 1
          