-- |
-- Functions for univariate polynomials.
--
module LMath.Types.Polynomial.Univariate
    ( polyX,
      univariateFromCoeffs,
      coefficient, coeffsAndExps, reflectPoly,
      sqfPart, sqff, sqff'
    )			    
where

import LMath.Types.Polynomial.Base

import Data.List ( find )

-- |
-- The indeterminate in a univariate polynomial ring.
--
polyX :: Num a => Polynomial a
polyX = indeterminate 1 0

-- |
-- Create a univariate polynomial from the list of coefficients;
-- the additive constant is the first element of the list, e.g.
-- @univariateFromCoeffs [1,2,3]@ represents the polynomial
-- @3*x^2 + 2*x + 1@.
--
univariateFromCoeffs :: Num a => [a] -> Polynomial a
univariateFromCoeffs coeffs = 
    case reverse (uvc 0 coeffs) of
        []  -> ZeroPolynomial False 1
	nzp -> NonZeroPolynomial False (Recursive nzp)
    where
    uvc _ []     = []
    uvc e (0:cs) = uvc (e+1) cs
    uvc e (c:cs) = (e,Constant c) : uvc (e+1) cs

-- |
-- Return the coefficient of @x^n@ (0 if there is no such coefficient).
--
coefficient :: Num a
            => Polynomial a -- ^ polynomial
            -> Integer      -- ^ @n@
            -> a
coefficient p n =
    case find (\(c,e) -> fromIntegral e == n) (coeffsAndExps p) of
       Just (coeff,_) -> coeff
       Nothing        -> 0

-- |
-- Return a (coefficient,exponent) list representing the given polynomial.
-- The (coefficient,exponent)-pairs are returned in decreasing exponent
-- order.
coeffsAndExps :: Num a => Polynomial a -> [(a,Int)]
coeffsAndExps (ZeroPolynomial True 0) = []
coeffsAndExps (ZeroPolynomial _    1) = []
coeffsAndExps (ZeroPolynomial _ _) =
    error "coeffsAndExps: not defined for /=1 indeterminates"
coeffsAndExps (NonZeroPolynomial True (Constant n)) = [(n,0)]
coeffsAndExps p@(NonZeroPolynomial _ nzp)
    | unNIndeterminates (nIndeterminates p) /= 1  = 
	error "coeffsAndExps: not defined for /=1 indeterminates"
    | otherwise = [(c,e) | let (Recursive rs) = nzp, (e,Constant c) <- rs]


-- |
-- Calculate the reflected polynomial.
-- The reflected polynomial @r@ of a given polynomial @p@ is derived
-- from @p@ by replacing every term @x^i@ by @x^(n-i)@ where @n@ is
-- the degree of @p@. The reflected polynomial of the 0-polynomial
-- is the 0-polynomial by definition.
--
reflectPoly :: Polynomial a -> Polynomial a
reflectPoly p@(ZeroPolynomial _ _) = p
reflectPoly p@(NonZeroPolynomial _ (Constant _)) = p
reflectPoly (NonZeroPolynomial me (Recursive (rs@((e,r):_)))) =
    NonZeroPolynomial me $ Recursive [(e-e',q) | (e',q) <- reverse rs]


-- |
-- Calculate the squarefree part of a polynomial.
--
sqfPart :: Fractional a => Polynomial a -> Polynomial a
sqfPart p = p `quotPoly` (gcdPoly p (derivative p)) 

-- |
-- Calculate a squarefree factorization of a polynomial.
-- Returns the squarefree factors paired with their respective
-- multiplicity.
--
sqff :: Fractional a => Polynomial a -> [(Integer, Polynomial a)]
sqff f = relevant (degree f) $ zip [1..] gs
    where
    gs = zipWith3 (\fi fi1 fi2 -> (fi*fi2) `quotPoly` (fi1*fi1))
	          fs (tail fs) (tail (tail fs))
    fs = f : zipWith euclid fs (map derivative fs)

    relevant 0 _          = []
    relevant d ((e,p):ps)
	| degree p == 0  =  relevant d ps
	| otherwise      =  (e,p) : relevant (d - e * degree p) ps

-- |
-- Calculate a squarefree factorization of the given polynomial.
-- Returns the squarefree factors only (see also 'sqff').
--
sqff' :: Fractional a => Polynomial a -> [Polynomial a]
sqff' = map snd . sqff


