{-
    This file is part of the LooPo project, see
    <http://www.infosun.fim.uni-passau.de/cl/loopo/>.

    Copyright (C) 2008 Universitaet Passau, Germany
    Author: Armin Groesslinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-}

-- |
-- Univariate quasi-polynomials.
--
module LMath.Types.QuasiPolynomial
    ( -- * Quasi-polynomials (univariate)
      PP, ppX,
      periodPP, constituentPP,
      evalPP, modPP, floorPP, quotPP, remPP, gcdPP,

      -- * Fractions of quasi-polynomials (univariate)
      FPP, fppX, toFPP, fpp, numeratorFPP, denominatorFPP, periodFPP,
      constituentFPP,

      -- * Entire quasi-polynomials
      EQP, toEQP,
       eqpX, constEQP, evalEQP,
      periodEQP, degreeEQP,
      componentEQP, componentsEQP,
      rootsEQP, dividesEQP, pseudoQuotEQP, pseudoRemEQP,
      quotRemEP, gcdEQP, extgcdEQP, ugcdEQP, ugcdEQPn,
      weakEchelonEQP, strongEchelonEQP,
      determinantEQP,
      backSubEQP,
      ldeSolveEQP, ldeSolve, rowEchelonReduction, columnEchelonReduction,
      EQPCases(..), EQPRoots(..),

{-
      -- * X
      X, leftX, rightX, leftFppX, rightFppX, floorX, mkX,
      periodX, isdefX,
      evalX, isIntegralX
-}
    )
where

import LMath.Numbers
import LMath.Types.AffineFunction
import LMath.Types.AlgNum
import LMath.Types.PeriodicNumber
import LMath.Types.Polynomial
import LMath.Types.Polynomial.AlgNum
import LMath.Types.Polynomial.Roots
import LMath.Types.LClasses
import LMath.Types.Matrix
import LMath.Types.Vector

import Control.Exception ( assert )
import Data.Ratio
import Data.List
import Data.Map ( Map )
import qualified Data.Map as Map
import Data.Maybe

import Debug.Trace

-- |
-- Type for the ring of quasi-polynomials with
-- coefficients of type @Ratio a@
--
type PP a = Polynomial (Periodic (Ratio a))

ppX :: Integral a => PP a
ppX = indeterminate 1 0

evalPP :: Integral a => PP a -> a -> Ratio a
evalPP f x = -- evalPeriodic (\f x -> evalPoly f [x%1])
             evalPoly (componentPP f (fromIntegral x)) [x%1]

integralPP :: Integral a
           => PP a -> (Polynomial (Periodic a), a)
integralPP f = (f' , factor)
    where
      f' = mapPoly (fmap' (\c -> numerator c * (factor `div` denominator c))) f
      factor = foldl lcm 1 $
               map denominator $
               concatMap periodicValues $
               map fst $ coeffsAndExps f

-- |
-- Compute f mod modulus.
-- The result contains only non-negative entries (less than modulus,
-- of coures).
--
modPP :: Integral a
      => PP a -> a -> Periodic (Ratio a)
modPP f modulus =
    assert (modulus > 0) $
    periodic [evalPP f i `rimod` modulus | i <- [0..period-1]]
    where
      (f', commonDenom) = integralPP f
      period = lcm commonDenom modulus

rimod :: Integral a => Ratio a -> a -> Ratio a
rimod r m =
    r - fromIntegral (floor (r / m') * m)
    where
      m' = fromIntegral m

-- |
-- Compute floor(f).
--
floorPP :: Integral a => PP a -> EQP
floorPP f = EQP (f' - constPoly (modPP f' 1))
    where
      f' = fmap' (fmap' toRational) f

quotPP :: Integral a
       => PP a
       -> PP a
       -> PP a
quotPP _ 0 = error "quotPP: division by zero"
quotPP a b = componentwisePP quotPoly0 a b
    where
      quotPoly0 0 0 = 0
      quotPoly0 _ 0 = error "quotPP: division by zero-divisor"
      quotPoly0 x y = quotPoly x y

remPP :: Integral a => PP a -> PP a -> PP a
remPP _ 0 = error "remPP: division by zero"
remPP a b = componentwisePP remPoly0 a b
    where
    remPoly0 0 0 = 0
    remPoly0 _ 0 = error "remPP: division by zero-divisor"
    remPoly0 x y = remPoly x y

gcdPP :: Integral a => PP a -> PP a -> PP a
gcdPP = componentwisePP gcdPoly

lcmPP :: Integral a => PP a -> PP a -> PP a
lcmPP a b = (a `quotPP` gcdPP a b) * b

periodPP :: (Integral a, Integral b) => PP a -> b
periodPP f = foldl lcm 1 $ map period $ collectCoeffs f

componentwisePP :: Integral a
                => (Polynomial (Ratio a) -> Polynomial (Ratio a) -> Polynomial (Ratio a))
                -> PP a
                -> PP a
                -> PP a
componentwisePP f a b = res
    where
    res = mixPP es
    es = [f (componentPP a i) (componentPP b i) | i <- [0..l-1]]
    l = lcm (periodPP a) (periodPP b)

mixPP :: Integral a => [Polynomial (Ratio a)] -> PP a
mixPP [] = 0
mixPP comps =
    univariateFromCoeffs [periodic [coefficient f i | f <- comps]
                         | i <- [0..deg]]
    where
      deg = maximum $ map degree comps

-- |
-- Return the effective leading coefficient of a quasi-polynomial.
-- Note that in [0,1]x + [2,0] the leading coefficient is [0,1], but
-- the effective leading coefficient is [2,1].
--
leadingCoeffPP :: Integral a => PP a -> Periodic (Ratio a)
leadingCoeffPP f = foldl (periodicWith comp) 0 cs
    where
      comp 0 y = y
      comp x _ = x

      cs = map fst $ coeffsAndExps f  -- in decreasing exponent order


-- |
-- Fractions of quasi-polynomials.
--
data Integral a => FPP a = FPP (PP a) (PP a)
                         | PP     (PP a)
                           deriving Show

periodFPP :: Integral a => FPP a -> Integer
periodFPP f = max (periodPP $ numeratorFPP f) (periodPP $ denominatorFPP f)

constituentFPP :: Integral a => Integer -> FPP a -> Integer -> FracPoly (Ratio a)
constituentFPP peri f i =
    constituentPP peri (numeratorFPP f) i %%
    constituentPP peri (denominatorFPP f) i

toFPP :: Integral a => PP a -> FPP a
toFPP num = PP num

fppX :: Integral a => FPP a
fppX = toFPP $ indeterminateNI (Exactly 1) 0

fpp :: Integral a => PP a -> PP a -> FPP a
fpp num den
    | den' == 1  =  PP  num'
    | otherwise  =  FPP num' den'
    where
      g = gcdPP num den
      num' = fmap' (*s) (num `quotPP` g)
      den' = fmap' (*s) den''
      den'' = den `quotPP` g
      s = fmap' recip0 $ leadingCoeffPP den''

      -- normalise leading coefficients of the denominator to 1;
      -- zero out the components in the numerator which are
      -- zero in the denominator to obtain a unique representation
      recip0 0 = 0
      recip0 x = recip x

numeratorFPP :: Integral a => FPP a -> PP a
numeratorFPP (PP  num  ) = num
numeratorFPP (FPP num _) = num

denominatorFPP :: Integral a => FPP a -> PP a
denominatorFPP (PP  _    ) = 1
denominatorFPP (FPP _ den) = den

evalFPP :: Integral a => FPP a -> a -> Ratio a
evalFPP f x
    | den == 0   =  error "0"
    | otherwise  =  num / den
    where
      num = evalPP (numeratorFPP f) x 
      den = evalPP (denominatorFPP f) x

instance Integral a => Eq (FPP a) where
    -- because we normalise (leading coefficient of the denominator
    -- is 1 except for zero-divisor components)
    -- equality is syntactic equality
    PP  num1      == PP  num2       =  num1 == num2
    FPP num1 den1 == FPP num2 den2  =  num1 == num2 && den1 == den2
    PP  num1      == FPP num2 den2  =  False
    FPP num1 den1 == PP  num2       =  False


instance Integral a => Num (FPP a) where
    PP     num1   + PP     num2    =  PP     (num1+num2)
    PP     num1   + FPP num2 den2  =  fpp (num1*den2+num2) den2
    FPP num1 den1 + PP     num2    =  fpp (num1+num2*den1) den1
    FPP num1 den1 + FPP num2 den2  =
        fpp (num1*f1+num2*f2) den
        where
          den = lcmPP den1 den2
          f1  = den `quotPP` den1
          f2  = den `quotPP` den2

    negate (FPP num den) = FPP (negate num) den
    negate (PP     num ) = PP     (negate num)

    PP     num1   * PP     num2   = PP     (num1*num2)
    PP     num1   * FPP num2 den2 = fpp (num1*num2) den2
    FPP num1 den1 * PP     num2   = fpp (num1*num2) den1
    FPP num1 den1 * FPP num2 den2 = fpp (num1*num2) (den1*den2)

    fromInteger = toFPP . fromInteger

    abs = error "'abs' not defined for instance Integral a => Num (FPP a)"
    signum = error "'signum' not defined for instance Integral a => Num (FPP a)"

instance Integral a => Fractional (FPP a) where
    recip (FPP num den) = fpp den num
    recip (PP     num ) = fpp (constPolyNI (Exactly 1) 1) num

    fromRational r = fpp (fromIntegral $ numerator r)
                         (fromIntegral $ denominator r)


--
-- xRestRight converges to 0 for x ->  oo and its absolute value is
--  <0 for x >= xRight,
-- xRestLeft  converges to 0 for x -> -oo and its absolute value is
--  <0 for x <= xLeft,
--
data Integral a => X a = X { xLeft  :: Integer
                           , xRight :: Integer
                           , xEqpLeft   :: EQP
                           , xRestLeft  :: FPP a
                           , xEqpRight  :: EQP
                           , xRestRight :: FPP a
                           , xOther :: Map Integer (Ratio a)
                           }

instance Integral a => Show (X a) where
    showsPrec _ f =
        showString "X{" . shows (Map.toList $ xOther f) . semi .
        shows (xLeft  f) . colon . shows (leftFppX f) . semi .
        shows (xRight f) . colon . shows (rightFppX f) .
        showString "}"
        where
          colon = showString ":"
          semi  = showString ";"

leftX :: Integral a => X a -> Integer
leftX = xLeft

rightX :: Integral a => X a -> Integer
rightX = xRight

leftFppX :: Integral a => X a -> FPP a
leftFppX f = toFPP (fromEQP $ xEqpLeft f) + xRestLeft f

rightFppX :: Integral a => X a -> FPP a
rightFppX f = toFPP (fromEQP $ xEqpRight f) + xRestRight f

isdefX :: Integral a => X a -> a -> Bool
isdefX f x = isJust $ evalMbX f x

evalMbX :: Integral a => X a -> a -> Maybe (Ratio a)
evalMbX f x
    | x' <= xLeft  f && evalPP (denominatorFPP $ xRestLeft  f) x /= 0  =
        Just $ evalFPP (leftFppX  f) x
    | x' >= xRight f && evalPP (denominatorFPP $ xRestRight f) x /= 0  =
        Just $ evalFPP (rightFppX f) x
    | otherwise      =  Map.lookup x' (xOther f)
    where
      x' = fromIntegral x

evalX :: Integral a => X a -> a -> Ratio a
evalX f x = 
    case evalMbX f x of
              Just v  -> v
              Nothing -> error "evalX: value undefined"

periodX :: Integral a => X a -> Integer
periodX f = lcm (periodFPP $ leftFppX f) (periodFPP $ rightFppX f)

isZeroX :: Integral a => X a -> Bool
isZeroX f =
    and [xEqpLeft   f == 0
        ,xEqpRight  f == 0
        ,xRestLeft  f == 0
        ,xRestRight f == 0
        ,all (==0) $ Map.elems $ xOther f
        ]

instance Integral a => Eq (X a) where
     a == b  =  isZeroX (a-b)

instance Integral a => Num (X a) where
    (+) = xArithOp False (+) (+)
    (*) = xArithOp False (*) (*)
    negate x =
        x { xEqpLeft = negate $ xEqpLeft x
          , xRestLeft = negate $ xRestLeft x
          , xEqpRight = negate $ xEqpRight x
          , xRestRight = negate $ xRestRight x
          , xOther = Map.map negate $ xOther x
          }
    fromInteger = mkX . fromInteger

    abs = error "abs: undefined in instance Integral a => Num (X a)"
    signum = error "signum: undefined in instance Integral a => Num (X a)"

instance Integral a => Fractional (X a) where
    recip = xRecip
    (/)   = xArithOp True (/) (/)
    fromRational = mkX . fromRational

xArithOp :: Integral a
         => Bool
         -> (Ratio a -> Ratio a -> Ratio a)
         -> (FPP a -> FPP a -> FPP a)
         -> X a
         -> X a
         -> X a
xArithOp isDiv op opP a (b :: X a) =
    mkX' (Just (l,r)) other fl fr
    where
      l = min (xLeft  a) (xLeft  b)
      r = max (xRight a) (xRight b)
      other :: Map Integer (Ratio a)
      other = Map.fromList [(x,evalX a x' `op` bval)
                           | x <- Map.keys (xOther a) `union`
                                  Map.keys (xOther b)
                           , let x' = fromIntegral x
                           , isdefX a x', isdefX b x'
                           , let bval = evalX b x'
                           , not isDiv || bval /= 0
                           ]
      fl = leftFppX a  `opP` leftFppX b
      fr = rightFppX a `opP` rightFppX b


xRecip :: Integral a
         => X a
         -> X a
xRecip (a :: X a) =
    mkX' (Just (l,r)) other fl fr
    where
      l = xLeft  a
      r = xRight a
      other :: Map Integer (Ratio a)
      other = Map.fromList [(x,recip aval)
                           | x <- Map.keys (xOther a)
                           , Just aval <- [evalMbX a (fromIntegral x)]
                           , aval /= 0
                           ]
      fl = recip $ leftFppX a
      fr = recip $ rightFppX a

xRightNeg, xLeftNeg :: Integral a => X a -> EQP
xRightNeg = negToInfty    . xRestRight
xLeftNeg  = negToNegInfty . xRestLeft

negToInfty, negToNegInfty :: Integral a => FPP a -> EQP
negToInfty f = constEQP $ fmap' (\x -> if x>=0 then 0 else 1) n
    where
      n = leadingCoeffPP $ numeratorFPP   f
negToNegInfty f = constEQP $ fmap' (\x -> if (x>0)/=inv || x==0 then 0 else 1) n
    where
      n = leadingCoeffPP $ numeratorFPP   f
      n_deg = degree $ numeratorFPP   f
      d_deg = degree $ denominatorFPP f
      inv  =  even n_deg /= even d_deg
      

floorX :: Integral a => X a -> X a
floorX x = x { xRestRight = 0, xRestLeft = 0
             , xEqpRight = xEqpRight x - xRightNeg x
             , xEqpLeft  = xEqpLeft  x - xLeftNeg  x
             , xOther = fmap (fromIntegral . floor) $ xOther x
             }

ceilingX :: Integral a => X a -> X a
ceilingX = negate . floorX . negate

mkX :: Integral a => FPP a -> X a
mkX f = mkX' Nothing Map.empty f f

mkX' :: Integral a
     => Maybe (Integer,Integer)   -- ^ maximal left, minimal right
     -> Map Integer (Ratio a)  -- ^ values between max. left und min. right (not including max. left und min. right)
     -> FPP a
     -> FPP a
     -> X a
mkX' mbMaxLeftMinRight other1 (fL :: FPP a) fR =
    X { xLeft = left, xRight = right
      , xEqpLeft = eL, xEqpRight = eR
      , xRestLeft = restL, xRestRight = restR
      , xOther = Map.fromAscList (case mbMaxLeftMinRight of
                                    Just (maxLeft,minRight) -> 
                                        [(x, evalFPP fL (fromIntegral x))
                                        | x <- [left+1 .. maxLeft]
                                        , evalPP denL (fromIntegral x) /= 0
                                        ] ++
                                        [(x, evalFPP fR (fromIntegral x))
                                        | x <- [minRight .. right-1]
                                        , evalPP denR (fromIntegral x) /= 0
                                        ]
                                    Nothing ->
                                        [(x, evalFPP fL (fromIntegral x))
                                        | x <- [left+1 .. right-1]
                                        , evalPP denL (fromIntegral x) /= 0
                                        ]
                                 )
                 `Map.union` other1
      }
    where
      numR = numeratorFPP fR
      denR = denominatorFPP fR
      qR = numR `quotPP` denR
      rR = numR `remPP`  denR
      eR = floorPP qR
      restR = toFPP (qR - fromEQP eR) + fpp rR denR

      numL = numeratorFPP fL
      denL = denominatorFPP fL
      qL = numL `quotPP` denL
      rL = numL `remPP`  denL
      eL = floorPP qL
      restL = toFPP (qL - fromEQP eL) + fpp rL denL


      left  = maybe left'  (min left' ) $ fmap fst mbMaxLeftMinRight
      right = maybe right' (max right') $ fmap snd mbMaxLeftMinRight
      left'  = minimum ls
      right' = maximum rs
      rs = [periR * safe01 True  (constituentFPP periR restR i :: FracPoly (Ratio a)) + i
                | i <- [0..periR-1]
           ]
      ls = [periL * safe01 False (constituentFPP periL restL i :: FracPoly (Ratio a)) + i
                | i <- [0..periL-1]
           ]
      periR = periodFPP restR
      periL = periodFPP restL

safe01 :: RealFrac a => Bool -> FracPoly a -> Integer
safe01 isRight fp
    | isRight    =  r
    | otherwise  =  l
    where
      roots1 = concatMap parToAlgnums $ rootsND [num-den,num+den]
      roots0 = concatMap parToAlgnums $ rootsND [num]
      roots1up   = map ((+ 1)       .floorAlg  ) roots1
      roots1down = map ((subtract 1).ceilingAlg) roots1
      roots0up   = map ceilingAlg roots0
      roots0down = map floorAlg   roots0

      l' = minimum (roots0down ++ roots1down)
      r' = maximum (roots0up   ++ roots1up  )
      (l,r)
          | null roots0 && null roots1  =  (-1  , 0 )
          | l' == r'                    =  (r'-1, r')
          | otherwise                   =  (l'  , r')

      num = polyNumerator fp
      den = polyDenominator fp

--
-- may be more to the right than necessary, since if
-- the rightmost zero of {num,num-den,num+den} is a zero of num and integral,
-- we need not add 1 to it. Analogously for the left side.
--
_safe01 :: RealFrac a => Bool -> FracPoly a -> Integer
_safe01 right fp =
    case allRoots of
      []  -> 0
      [x] -> if right 
             then ceiling (algBiggerRat  x) + 1
             else floor   (algSmallerRat x)
      xs  -> if right
             then maximum xs2
             else minimum xs1
          where
            xs1 = map (floor  . algSmallerRat) xs
            xs2 = map (ceiling . algBiggerRat) xs
    where
      allRoots = concatMap parToAlgnums $ rootsND [num,num-den,num+den]
      num = polyNumerator fp
      den = polyDenominator fp
                                     

isIntegralX :: Integral a => X a -> Bool
isIntegralX f = floorX f == f


--
-- Entire quasi-polynomials
--


-- |
-- The ring of entire quasi-polynomials (i.e., the ring of
-- univariate polynomials with coefficients from the
-- periodic rationals and integral function values at
-- integral locations).
--
data EQP = EQP (Polynomial (Periodic Rational))
           deriving Show

-- |
-- Regard an entire quasi-polynomial as a polynomial
-- with periodic number coefficients.
--
fromEQP :: Integral a => EQP -> PP a
fromEQP (EQP f) = mapPoly (fmap' fromRational) f



-- |
-- Test whether the given polynomial is integral for
-- every value of the form n*period+offset (n from the integers).
--
isEntire :: Integer  -- ^ offset
         -> Integer  -- ^ period
         -> Polynomial Rational -> Bool
isEntire offset period f =
    all ((==1).denominator) $
    [evalPoly f [x%1] | n <- [0..cases-1], let x = offset + period*n]
    where
    cases = foldl lcm 1 $ map denominator $ collectCoeffs f

-- |
-- The indeterminate in the ring of entire quasi-polynomials.
--
eqpX :: EQP
eqpX = EQP $ indeterminate 1 0

-- |
-- A constant entire quasi polynomial.
--
constEQP :: Real a => Periodic a -> EQP
constEQP coeffs
    | all ((==1).denominator) $ periodicValues rats  =
        EQP $ constPolyNI (Exactly 1) rats
    | otherwise  =  error "constEQP: coefficients not integral"
    where
    rats = fmap' toRational coeffs

-- |
-- (Maximal) degree of an entire quasi polynomial.
--
degreeEQP :: Integral b => EQP -> b
degreeEQP (EQP p) = degree p

-- |
-- Period of an entire quasi polynomial.
--
periodEQP :: Integral b => EQP -> b
periodEQP (EQP f) = foldl lcm 1 $ map period $ collectCoeffs f

-- |
-- Construct an entire quasi polynomial from its constituent polynomials
-- The constituents are given for offsets 0,...,period-1 (in that order).
toEQP :: [Polynomial Rational] -> EQP
toEQP fs
    | and [isEntire off p f | f <- fs | off <- [0..]]  =
        unsafeToEQP fs
    | otherwise  =  error "toEQP: constituents are not entire"
    where
    p = genericLength fs

unsafeToEQP :: [Polynomial Rational] -> EQP
unsafeToEQP fs =
    EQP $ univariateFromCoeffs [periodic [coefficient f expo | f <- fs]
                                | expo <- [0..d]]
    where
    d = maximum $ map degree fs

unsafeToEQP1 :: Polynomial Rational -> EQP
unsafeToEQP1 f = EQP $ mapPoly fromRational f

-- |
-- Evaluate an entire quasi polynomial at an integral location.
--
evalEQP :: EQP -> Integer -> Rational
evalEQP f x = evalPoly (componentEQP f x) [x%1]

instance Eq EQP where
    EQP a == EQP b  =  a == b

instance Num EQP where
    EQP a + EQP b  =  EQP (a+b)
    EQP a * EQP b  =  EQP (a*b)
    negate (EQP a) =  EQP $ negate a
    fromInteger x  =  EQP $ fromInteger x
    abs    = error "abs undefined for EQP"
    signum = error "signum undefined for EQP"


-- |
-- Pseudo division of (entire) polynomials.
-- Returns the (quotient,remainder) pair.
-- Properties for @(quot,rem) = quotRemEP a b@ and any integral @n@:
--
-- @
--   degreeEQP rem <= degree b
--   a(n) = quot(n)*b(n) + rem(n)
--   quot(n) is integral
--   rem(n) is integral, if a(n) and b(n) are integral
--   degreeEQP rem == degree b   implies  |HC(rem)| < |HC(b)|
-- @
--
-- There is no guarantee on the integrality of the
-- remainder values when @a@ or @b@ is not integral.
-- Note also that the degree of @rem@ may equal the degree of @b@;
-- in the general case, @degreeEQP rem < degree b@ is not possible.
--
quotRemEP :: Polynomial Rational -> Polynomial Rational -> (EQP, EQP)
quotRemEP a b
    | b == 0    =  error "quotRemEP: division by 0"
    | degDiff <  0  =  (0, EQP $ mapPoly fromRational a)
    | degDiff == 0  =
        (unsafeToEQP1 q0
        ,unsafeToEQP1 (a - q0 * b)
        )
    | lcR == 0  =
        case quotRemEP (a-q0*b) b of
          (quo,re) -> (quo + unsafeToEQP1 q0, re)
    | otherwise  =
        case componentwiseEQP' contDiv (EQP r) (unsafeToEQP1 b) of
          [quo,re] -> (quoFinal, re)
             where
             quoFinal = quo + EQP q
    where
    degDiff = degree a - degree b
    lcA = leadingCoeff a
    lcB = leadingCoeff b
    lcQ = fromIntegral $ floor (lcA/lcB)
    lcR = lcA - lcB*lcQ

    q0 = constPoly (fromRational lcQ) * indeterminate 1 0 ^ degDiff

    nCases' = abs (numerator lcB `div` gcd (numerator lcA) (numerator lcB))
    nCases  = product [f^e' | (f,e) <- snd $ factorise nCases',
                              let e' = ceiling (e % degDiff)
                      ]

    q1 = mapPoly (*(lcA/lcB)) (indeterminate 1 0 ^ degDiff)
    rests = periodic [evalPoly q1 [n%1] | n <- [0..nCases-1]]
    q = mapPoly fromRational q1 - constPoly rests
    r = mapPoly fromRational a - q * mapPoly fromRational b

    contDiv aa bb = let (qq,rr) = quotRemEP aa bb in [qq,rr]

-- |
-- Select the @n@-th component of an entire quasi polynomial
-- (@n >= 0@).
--
componentEQP :: Integral a => EQP -> Integer -> Polynomial (Ratio a)
componentEQP (EQP f) i = fmap' fromRational $ componentPP f i

componentPP :: Integral a => PP a -> Integer -> Polynomial (Ratio a)
componentPP f i = mapPoly select f
    where
    select pn = pvs !! fromIntegral (i `mod` genericLength pvs)
        where
        pvs = periodicValues pn

-- |
-- Return the constituent @i@ of a quasi-polynomial by
-- selecting the respective coefficient and
-- substituting @peri*X+i@ for @X@.
--
constituentPP :: Integral a => Integer -> PP a -> Integer -> Polynomial (Ratio a)
constituentPP peri f i
    | peri `mod` periodPP f == 0  =  replacePoly1 comp $ mapPoly select f
    | otherwise = error "constituentPP: given period not valid"
    where
      comp [expo] = newX ^ expo
      comp []     = 1
      newX = constPoly (fromIntegral peri)*polyX + constPoly (fromIntegral i)

      select pn = pvs !! fromIntegral (i `mod` genericLength pvs)
          where
            pvs = periodicValues pn

-- |
-- Return the components of an entire quasi polynomial.
--
componentsEQP :: EQP -> [Polynomial Rational]
componentsEQP f = [componentEQP f i | i <- [0..periodEQP f-1]]

data EQPRoots = EQPRoots { eqprPeriod :: Integer,
                           eqprPeriodicRoots :: [Integer],
                           eqprSingularRoots :: [Integer]
                         }
                deriving Show

unionEQPR :: EQPRoots -> EQPRoots -> EQPRoots
unionEQPR a b =
    EQPRoots { eqprPeriod = l,
               eqprPeriodicRoots = [i | i <- [0..l-1], unionCond i],
               eqprSingularRoots = sortUniq (eqprSingularRoots a ++
                                             eqprSingularRoots b)
             }
    where
    l = lcm (eqprPeriod a) (eqprPeriod b)
    unionCond i = ((i `mod` eqprPeriod a) `elem` eqprPeriodicRoots a) ||
                  ((i `mod` eqprPeriod b) `elem` eqprPeriodicRoots b)

isPeriodicRoot :: EQPRoots -> Integer -> Bool
isPeriodicRoot (EQPRoots {eqprPeriod=l, eqprPeriodicRoots=ps}) x =
    (x `mod` l) `elem` ps

-- |
-- Return all the (integral) roots of an entire quasi polynomial.
--
rootsEQP :: EQP -> EQPRoots
rootsEQP f =
    EQPRoots { eqprPeriod = l,
               eqprPeriodicRoots = periodicRoots,
               eqprSingularRoots =  sort singularRoots
             }
    where
    periodicRoots = [i | (0,i) <- zip cs [0..]]
    singularRoots = 
        [x | (p,i) <- zip cs [0..], p /= 0,
             x <- integralRoots p,
             x `mod` l == i
        ]

    cs = componentsEQP f
    l  = genericLength cs


componentwiseEQP :: (Polynomial Rational -> Polynomial Rational -> EQP)
                 -> EQP -> EQP -> EQP
componentwiseEQP f a b = res
    where
    res = mixComponents l es
    es = [f (componentEQP a i) (componentEQP b i) | i <- [0..l-1]]
    l = lcm (periodEQP a) (periodEQP b)

componentwiseEQP' :: (Polynomial Rational -> Polynomial Rational -> [EQP])
                  -> EQP -> EQP -> [EQP]
componentwiseEQP' f a b = res
    where
    res = [mixComponents l es | es <- ess]
    ess = transpose [f (componentEQP a i) (componentEQP b i)
                     | i <- [0..l-1]
                    ]
    l = lcm (periodEQP a) (periodEQP b)

mixComponents :: Integer -> [EQP] -> EQP
mixComponents l es =
    unsafeToEQP [componentEQP e (i `div` l)
                 | (e,i) <- zip (cycle es) [0..allLcm-1]
                ]
    where
    allLcm = foldl lcm l (map periodEQP es)

-- |
-- A pseudo division that yields 0 in the constituents of the
-- quotient that correspond to zero components in the divisor
-- (i.e., it is possible to divide by zero divisors, but the
-- quotient is, of course, not meaningful in the constituents
-- that correspond to the constituents where the divisor
-- vanishes).
--
pseudoQuotEQP :: EQP -> EQP -> EQP
pseudoQuotEQP a b = componentwiseEQP pquot a b
    where
    pquot :: Polynomial Rational -> Polynomial Rational -> EQP
    pquot _ 0 = 0
    pquot x y = fst $ quotRemEP x y


pseudoRemEQP :: EQP -> EQP -> EQP
pseudoRemEQP a b = componentwiseEQP prem a b
    where
    prem :: Polynomial Rational -> Polynomial Rational -> EQP
    prem x 0 = toEQP [x]
    prem x y = snd $ quotRemEP x y


-- |
-- Compute the greatest common divisor of two
-- entire quasi-polynomials.
--
gcdEQP :: EQP -> EQP -> EQP
gcdEQP aa bb = componentwiseEQP f aa bb
    where
    f :: Polynomial Rational -> Polynomial Rational -> EQP
    f a 0 = EQP $ mapPoly fromRational a
    f a b = componentwiseEQP f (EQP $ mapPoly fromRational b) r
        where
        (q,r) = quotRemEP a b

-- |
-- Perform Euclid's extended algorithm, i.e.,
-- compute a triple @(e,s,t)@ from the input polynomials @f@ and @g@,
-- such that @e@ is a GCD of @f@ and @g@ and the equation
-- @e=s*f+t*g@ holds.
--
extgcdEQP :: EQP  -- ^ @f@
          -> EQP  -- ^ @g@
          -> (EQP, EQP, EQP) -- ^ @(e,s,t)@
extgcdEQP aa bb =
    case componentwiseEQP' f aa bb of
        [e,s,t] -> (e,s,t)
    where
    f :: Polynomial Rational -> Polynomial Rational -> [EQP]
    f a 0 = [EQP $ mapPoly fromRational a, 1, 0]
    f a b = [e, t, s-t*q]
        where
        (q,r) = quotRemEP a b
        [e,s,t] = componentwiseEQP' f (EQP $ mapPoly fromRational b) r

-- |
-- Compute the gcd @d@ of the given entire quasi-polynomials @a@ and @b@
-- and return (the rows of) a 2x2 matrix @U@ such that 
-- @U*(a b)^T = (d 0)^T@ and @det U=1@.
--
ugcdEQP :: EQP -> EQP -> (EQP, ((EQP,EQP),(EQP,EQP)))
ugcdEQP aa bb =
    case componentwiseEQP' f aa bb of
        [e,w,x,y,z]  ->  (u*e, ((u*w,u*x),(y,z)))
-- Test mit  == -1  ist nicht richtig, da sich das Vorzeichen
-- der Determinante von [[w,x],[y,z]] von Konstituent zu Konstituent
-- unterscheiden kann (d.h. die Determinante ist periodisch in {1,-1}).
--            | w*z - y*x == -1  ->  (negate e, ((negate w,negate x),(y,z)))
--            | otherwise        ->  (e,((w,x),(y,z)))
            where
            u = w*z - y*x
    where
    f :: Polynomial Rational -> Polynomial Rational -> [EQP]
    f a 0 = [EQP $ mapPoly fromRational a, 1, 0, 0, 1]
    f 0 b = [EQP $ mapPoly fromRational b, 0, 1, -1, 0]
    f a b = [e, x, w-x*q, z, y-z*q]
        where
        (q,r) = quotRemEP a b
        [e,w,x,y,z] = componentwiseEQP' f (EQP $ mapPoly fromRational b) r

-- |
-- Compute the gcd @d@ of a list (vector) @v@ of
-- entire quasi-polynomials and a matrix @U@ such that
-- @U*v = (d 0 .. 0)@ and @det U=1@.
--
ugcdEQPn :: [EQP] -> (EQP, GMatrix EQP)
ugcdEQPn [] = (0, zeroMatrix 0 0)
ugcdEQPn (f:fs) = (d, u'' * u')
    where
    (d1, u1) = ugcdEQPn fs
    (d,  ((w,x),(y,z))) = ugcdEQP  f d1

    u'  = unitMatrix 1 n `join` (zeroMatrix (n-1) 1 `conc` u1)
    u'' = (matrix [[w,x],[y,z]] `conc` zeroMatrix 2 (n-2))
          `join`
          (zeroMatrix (n-2) 2 `conc` unitMatrix (n-2) (n-2))

    n = length fs + 1

-- |
-- Determine at which locations the function value of the
-- entire quasi polynomial @g@ divides the function value
-- of the entire quasi polynomial @f@.
--
dividesEQP :: EQP  -- ^ @g@
           -> EQP  -- ^ @f@
           -> EQPRoots
dividesEQP g f = rootsEQP (e+g) `unionEQPR` rootsEQP (e-g)
    where
    e = gcdEQP f g

-- |
-- Compute from a given matrix @A@ two matrices @U@ and @S@
-- with @UA=S@ and @det U=1@.
--
weakEchelonEQP :: GMatrix EQP                 -- ^ @A@
               -> (GMatrix EQP, GMatrix EQP)  -- ^ @(U,S)@
weakEchelonEQP a
    | n <= 1 || m == 0              =  (unitMatrix n n, a)
    | all (==0) firstColumnEntries  =  (ur, sr_res)
    | n >  1                        =  (u, s)
    where
    firstColumnEntries = vectorEntries $ head $ columnVectors a
    (d, u1) = ugcdEQPn firstColumnEntries

    a' = u1 * a
    (u', s') = weakEchelonEQP $ subMatrix a' 1 1 (n-1) (m-1)
    s = subMatrixHor a' 0 0
        `join`
        (zeroMatrix (n-1) 1 `conc` s')
    u = (unitMatrix 1 n `join` (zeroMatrix (n-1) 1 `conc` u')) * u1
 
    ar = subMatrixVer a 1 (m-1)
    (ur,sr) = weakEchelonEQP ar
    sr_res = zeroMatrix n 1 `conc` sr

    n = nRows a
    m = nColumns a


-- |
-- Return the determinant of a matrix with
-- entire quasi-polynomials as entries.
--
determinantEQP :: GMatrix EQP -> EQP
determinantEQP m
    | nRows m == nColumns m  =
        -- weakEchelonEQP guarantees that det(U)=1 (and not -1)
        product [entry s i i | i <- [0..maxRow s]]
    | otherwise =
        error "determinantEQP: matrix is not square"
    where
    (u,s) = weakEchelonEQP m

strongEchelonEQP :: GMatrix EQP
                 -> (GMatrix EQP, GMatrix EQP, Periodic [Integer], [Integer])
strongEchelonEQP a
    | nRows a == 0  =  (u, s, periodic [[]], [])
    | otherwise =
      case mblIdx of
         Nothing -> (u, s, periodic [[]], [])
         Just lIdx -> (allU, allS, periodic ci,
                       sortUniq (eqprSingularRoots eqpr ++ sings1 ++
                                 filter (isPeriodicRoot eqpr) rSings
                                )
                      )
    where
    (u,s) = weakEchelonEQP a
    mblIdx@(~(Just lIdx)) = findIndex (/=0) $ head $ rows' s 

    leader = entry s 0 lIdx
    eqpr   = rootsEQP leader
    uuu1 = fmap (*antiMask) ru0 + uu

    allS = lPart `conc` (rPart' + ss)
    allU = uuu1 * u
             
    lPart  = subMatrixVer s 0 lIdx
    rPart  = subMatrixVer s (lIdx+1) (maxColumn s)
    rPart' = fmap (*antiMask) rPart''

    row0  = subMatrixHor rPart 0 0
    row1p = subMatrixHor rPart 1 (maxRow rPart)
    (ru,rs,leaders1,sings1) = strongEchelonEQP row1p
    rPart'' = row0 `join` rs
    ru0   = unitMatrix 1 (nRows ru + 1) `join`
                     (zeroMatrix (nRows ru) 1 `conc` ru)
                           
    (uu',ss',rLeaders,rSings) = strongEchelonEQP rPart
    uu = fmap (*mask) uu'
    ss = fmap (*mask) ss'

    mask = constEQP $ periodic [if i `elem` eqprPeriodicRoots eqpr
                                then 1 else 0
                               | i <- [0..eqprPeriod eqpr-1]]
    antiMask = 1 - mask

    l = lcm (lcm (eqprPeriod eqpr) (period rLeaders)) (period leaders1)
    ci = [if (i `mod` eqprPeriod eqpr) `elem` eqprPeriodicRoots eqpr
           then map (+1) (periodicValues rLeaders !! ir)
           else (0 : map (+1) (periodicValues leaders1 !! i1))
          | i <- [0..l-1],
            let i1 = fromIntegral (i `mod` period leaders1),
            let ir = fromIntegral (i `mod` period rLeaders)
         ]


data EQPCases a b = EQPCases { eqpcPeriod :: Integer,
                               eqpcPeriodic :: [(Integer,a)],
                               eqpcSingular :: [(Integer,b)]
                             }
                    deriving Show

flatCasesEQP :: (Functor m, MatrixBased (m EQP) EQP,
                 MatrixBased (m (Polynomial Rational)) (Polynomial Rational)
                )
             => EQPCases (m EQP) b -> EQPCases (m (Polynomial Rational)) b
flatCasesEQP eqpc =
    EQPCases { eqpcPeriod = l2,
               eqpcPeriodic = concatMap flat [0..l2-1],
               eqpcSingular = eqpcSingular eqpc
             }
    where
    l  = eqpcPeriod eqpc
    l2 = foldl lcm l [periodEQP f | s <- sols, f <- entries s]
    sols = map snd $ eqpcPeriodic eqpc

    flat i = [(i,fmap (flip componentEQP i) sol)
              | (j,sol) <- eqpcPeriodic eqpc, (i `mod` l) == j]
                                                

-- |
-- Solve a linear Diophantine equation system with
-- entiere quasi-polynomials as coefficients.
-- Return the (parameterized) solutions for @x@ of @x*A=b@;
-- the variables of the affine functions returned are the
-- free variables (degrees of freedom) in the solution.
--
ldeSolveEQP :: GMatrix EQP  -- ^ @A@
            -> GVector EQP  -- ^ @b@
            -> EQPCases (GAffineFunction (Polynomial Rational))
                        (Maybe AffineFunction)
ldeSolveEQP a b =
    flatCasesEQP $
    eqpc { eqpcPeriodic = [(i, composeU u t_fn)
                           | (i,t_fn) <- eqpcPeriodic eqpc
                          ],
           eqpcSingular = [(p, solveSing p) | p <- sings]
         }
    where
    (u, s, colIdxs, sings1) = strongEchelonEQP a
    eqpc = backSubEQP s b colIdxs
    sings = sortUniq (sings1 ++ map fst (eqpcSingular eqpc))

    solveSing p =
        case sols of
          []        -> Nothing
          [(0,sol)] -> Just $ fmap (flip evalEQP p) (composeU u' sol)
        where
        a' = fmap (\e -> constEQP $ periodic [evalEQP e p]) a
        b' = fmap (\e -> constEQP $ periodic [evalEQP e p]) b
        
        (u',s') = weakEchelonEQP a'

        EQPCases { eqpcPeriod=1, eqpcPeriodic=sols, eqpcSingular=[] } =
            backSubEQP s' b' (periodic [colIdxs])
        colIdxs = [fromIntegral c | r <- rows' s',
                                    Just c <- [findIndex (/=0) r]]

    composeU uu t_fn = composeAffine uu_fn t_fn
        where
        uu_fn = affineFunctionVPC (mtranspose uu,
                                   zeroMatrix (nRows uu) 0,
                                   zeroMatrix (nRows uu) 1
                                  )
                                  Inhomogeneous

backSubEQP :: GMatrix EQP
           -> GVector EQP
           -> Periodic [Integer]
           -> EQPCases (GAffineFunction EQP) ()
backSubEQP s b colIdxs =
    EQPCases { eqpcPeriod = resL,
               eqpcPeriodic = [(i,sol) | i <- [0..resL-1], (eqpr,sol) <- res,
                               (i `mod` eqprPeriod eqpr) `elem`
                                                   eqprPeriodicRoots eqpr
                              ],
               eqpcSingular = map (flip (,) ()) $
                              sortUniq $
                              concat [eqprSingularRoots epqr | (epqr,_) <- res]
             }
    where
    l = period colIdxs

    res :: [(EQPRoots, GAffineFunction EQP)]
    res = zipWith bs [0..l-1] (periodicValues colIdxs)
    resL = foldl lcm 1 [eqprPeriod eqpr | (eqpr,_) <- res]

    bs i cis = bs2 cis [] leftIs0
        where
        -- The entries in b left of the first pivot column
        -- of S must be zero (because the columns left of the
        -- first pivot in S are zero) and
        -- the parameter must be congruent to i mod l.
        leftIs0 = foldl commonSol rootsI $
                  map rootsEQP $
                  genericTake firstPivot $ vectorEntries b
        firstPivot = head (cis ++ [vectorLength b])
        rootsI = EQPRoots { eqprPeriod=l,
                            eqprPeriodicRoots=[i],
                            eqprSingularRoots=[]
                          }

    bs2 [] ts solCases        =
        (solCases, affineFunctionVPC (v,p,c) Inhomogeneous)
        where
        v = zeroMatrix (genericLength ts) addTs `join`
            unitMatrix addTs addTs
        p = zeroMatrix (genericLength ts + addTs) 0
        c = columnMatrix ts `join` zeroMatrix addTs 1
        addTs = nRows s - genericLength ts
    bs2 (c1:rest) ts solCases =
        bs2 rest (ts ++ [head qs]) (commonSol sols solCases)
        where
        c2 = case rest of
               []   -> nColumns s
               c2:_ -> c2

        sols = foldl1 commonSol $
               (zipWith dividesEQP lhs rhs ++
                [rootsEQP (q1-q2) | (q1,q2) <- zip qs (tail qs)]
               )
        qs = zipWith pseudoQuotEQP rhs lhs

        lhs = entries $ subMatrix s tN c1 tN (c2-1)
        rhs = vectorEntries $
              addVV (subVector b c1 (c2-c1)) $
              vector $
              map (negate . mulVV (vector ts)) $
              columnVectors $
              subMatrix s 0 c1 (tN-1) (c2-1)
        tN = genericLength ts


--
-- (u,s,_) = rowEchelonReduction a   =>  u*a == s
--
rowEchelonReduction :: Integral a
  => GMatrix a -> (GMatrix a, GMatrix a, [Integer])
rowEchelonReduction m =
    assert (null exceps) $
    assert (period pcs == 1) $
    (fmap un u, fmap un s, head $ periodicValues pcs)
    where
    (u,s,pcs,exceps) = strongEchelonEQP $ fmap (\e -> constEQP $ periodic [e]) m
    un eqp = case componentsEQP eqp of
               [f] -> case isConstPoly f of
                        Just c 
                           | denominator c == 1  ->  fromIntegral $ numerator c

--
-- (u,s,_) = columnEchelonReduction a   =>   a*u == s  
--
columnEchelonReduction :: Integral a
  => GMatrix a -> (GMatrix a, GMatrix a, [Integer])
columnEchelonReduction m = (mtranspose u, mtranspose s, cols)
    where
    (u, s, cols) = rowEchelonReduction $ mtranspose m
         

ldeSolve :: Real a => GMatrix a -> GVector a -> Maybe (GAffineFunction a)
ldeSolve m' v' = ldeSolve2 m v
    where
    m'' = fmap toRational m'
    v'' = fmap toRational v'
    fact = foldl lcm 1 $ map denominator (entries m'' ++ vectorEntries v'')
    m = fmap (\e -> numerator e * (fact `div` denominator e)) m''
    v = fmap (\e -> numerator e * (fact `div` denominator e)) v''


ldeSolve2 m v
    | eqpcPeriod eqpc == 1 && null (eqpcSingular eqpc)  =
        trace "ldeSolve" $
        fmap (fmap fromPoly) $
        lookup 0 (eqpcPeriodic eqpc)
    | otherwise  =  error "ldeSolve: unexpected result"
    where
    eqpc :: EQPCases (GAffineFunction (Polynomial Rational)) (Maybe AffineFunction)
    eqpc = ldeSolveEQP (fmap mkEQP m) (fmap mkEQP v)
    mkEQP x = constEQP $ periodic [x]
    fromPoly p = case isConstPoly p of
                     Just c
                       | denominator c == 1  -> fromIntegral $ numerator c
                     _ -> error "ldeSolve: unexpected polynomial"

gcdn :: Integral a => [a] -> a
gcdn xs = case filter (/=0) xs of
             []    ->  0
             y:ys  ->  foldl gcd y ys

--
-- Compute the intersection of the periodic roots, and the
-- _union_ of the sigular roots.
--
commonSol :: EQPRoots -> EQPRoots -> EQPRoots
commonSol a b =
    EQPRoots { eqprPeriod = l,
               eqprPeriodicRoots = [i | i <- [0..l-1], intersectCond i],
               eqprSingularRoots = sortUniq (eqprSingularRoots a ++
                                             eqprSingularRoots b)
             }
    where
    l = lcm (eqprPeriod a) (eqprPeriod b)
    intersectCond i = ((i `mod` eqprPeriod a) `elem` eqprPeriodicRoots a) &&
                      ((i `mod` eqprPeriod b) `elem` eqprPeriodicRoots b)

--
-- sort the given list and eliminate duplicates
--
sortUniq :: Ord a => [a] -> [a]
sortUniq = map head . group . sort

h :: GMatrix EQP
h = fmap constEQP $
    matrix [[periodic [1,0], periodic [0,-1]],
            [periodic [0,1], periodic [1,-1]]
           ]


macisEx1_A :: GMatrix EQP
macisEx1_A = matrix [[eqpX], [2], [-eqpX], [-2]]

macisEx1_b :: GVector EQP
macisEx1_b = vector [0]

ex2_A :: GMatrix EQP
ex2_A = matrix [[eqpX, 2, eqpX^2+eqpX],
               [1, eqpX, eqpX],
               [constEQP (periodic [0,1]), eqpX, eqpX]
              ]

ex2_b :: GVector EQP
ex2_b = vector [eqpX,eqpX^2,eqpX^2]

