module LMath.Types.Signum
    ( Signum(..),
      allSignumsFalse,
      orderingToSignum, mbOrderingToSignum
    )

where


-- |
-- Representation of possible signums. Every field represents
-- one of the relations @== 0@, @\/= 0@, @\>= 0@, @\> 0@, @\<= 0@,@\< 0@.
-- For example, when @signumGe@ is @True@, the value represented
-- by the corresponding @Signum@ structure is always @>= 0@, i.e.,
-- it is never negative. It is possible that @signumEq@ (and @signumLe@) or
-- @signumGt@ (and @signumNeq@) are @True@, too, if the value can,
-- in fact, only be zero or positive, respectively. To put it in
-- other words: When @signumGt@ is @True@, then @signumGe@ and
-- @signumNeq@ are also @True@, since @p\>0@ implies @p\>=0@ and @p\/=0@.
--
data Signum = Signum { signumEq  :: Bool,   -- ^ @== 0@
		       signumNeq :: Bool,   -- ^ @\/= 0@
		       signumGe  :: Bool,   -- ^ @\>= 0@
		       signumGt  :: Bool,   -- ^ @\>  0@
		       signumLe  :: Bool,   -- ^ @<= 0@
		       signumLt  :: Bool    -- ^ @<  0@
		     }
	      deriving (Show)


-- |
-- A 'Signum' value whose fields are initialized to @False@.
--
allSignumsFalse :: Signum
allSignumsFalse = Signum { signumEq  = False,
			   signumNeq = False,
			   signumGe  = False,
			   signumLe  = False,
			   signumGt  = False,
			   signumLt  = False
			 }

-- |
-- Convert an 'Ordering' value to an equivalent 'Signum'
-- value.
--
orderingToSignum :: Ordering -> Signum
orderingToSignum LT = Signum { signumEq = False, signumNeq = True,
			       signumLe = True,  signumLt  = True,
			       signumGe = False, signumGt  = False }
orderingToSignum EQ = Signum { signumEq = True,  signumNeq = False,
			       signumLe = True,  signumLt  = False,
			       signumGe = True,  signumGt  = False }
orderingToSignum GT = Signum { signumEq = False, signumNeq = True,
			       signumLe = False, signumLt  = False,
			       signumGe = True,  signumGt  = True }

mbOrderingToSignum :: Maybe Ordering -> Signum
mbOrderingToSignum Nothing  = allSignumsFalse
mbOrderingToSignum (Just o) = orderingToSignum o
