module LMath.Types.SortedSystem
    ( Bounds(..), SortedSystem,
      ssOuterDims,
      qfFormulaToBounds,
      sortedSystemToIneqSystem,
      countBounds
    )
where

import Control.Monad ( liftM2 )
import Data.List

import Debug.Trace

import LMath.Types.DTree
import LMath.Types.Formula
import LMath.Types.IneqSystem
import LMath.Types.LClasses
import LMath.Types.Matrix
import LMath.Types.Vector
import LMath.Types.Polynomial
import LMath.Types.Signum

-- |
-- Representation of a sorted inequality system.
-- Each entry in the list represents the bounds for a variable.
-- The first entry corresponds to the first variable, etc.
-- The bounds for the first variable are represented in term
-- of constants, i.e., vectors with one entry. The bounds for
-- the second variable are represented by an affine function
-- depending on the first variable, i.e., by vectors with two
-- entries (the additive constant being the last entry, as usual),
-- etc.
--
type SortedSystem a = [Bounds (GVector a)]

-- |
-- A bound for a variable.
-- A bound is either a list of lower and upper bounds, or
-- an equation (when the variable's value is an affine function
-- in the surrounding variables). It is possible that an
-- equation is represented by lower and upper bounds, i.e.,
-- there is no guarantee that equations are detected when they
-- hold.
--
data Bounds a = LowerUpper [a] [a]
	      | Equation   a
		deriving (Eq, Show)

instance Functor Bounds where
    fmap f (Equation x) = Equation (f x)
    fmap f (LowerUpper ls us) = LowerUpper (map f ls) (map f us)


-- |
-- Determine the number of context (outer) dimensions
-- of a sorted system. For example, if the first bound in
-- a sorted system has 6 entries in (one of) its vector(s),
-- there are 5 outer dimensions (the sixth entry represents the
-- constant term).
--
ssOuterDims :: Integral b => SortedSystem a -> Maybe b
ssOuterDims [] = Nothing
ssOuterDims (Equation x : _) = Just (vectorLength x - 1)
ssOuterDims (LowerUpper [] [] : rest) = fmap (\d -> d-1) $ ssOuterDims rest
ssOuterDims (LowerUpper ls us : _   ) =
    Just (vectorLength (head (ls++us)) - 1)
            

qfFormulaToBounds :: (Fractional a, Num cond)
		  => QfFormula (Polynomial a)    -- ^ formula to convert
		  -> Integer                     -- ^ number of variable to solve for
		  -> Integer                     -- ^ total number of variables
		  -> Maybe (DTreeGen (Polynomial a) cond (Bounds (GVector (FracPoly a))))
qfFormulaToBounds (QfAt (AtEq p)) varNr nV =
    Just $
    solve p varNr nV $ \mbP' ->
    case mbP' of
      Nothing     -> dLeaf $ LowerUpper [] []
      Just (_,p') -> dLeaf $ Equation p'
qfFormulaToBounds f varNr nV
    | Just ats <- extract f  =
		  Just $
		  fmap (\(l,u) -> LowerUpper l u) $ solveN ats
    | otherwise  = Nothing
    where
    extract (QfAnd f g) = liftM2 (++) (extract f) (extract g)
    extract (QfAt  f@(AtEq _)) = Just [f]
    extract (QfAt  f@(AtGe _)) = Just [f]
    extract (QfAt  AtTrue    ) = Just []
    extract (QfAt  AtFalse   ) = Nothing
    extract (QfAt    (AtGt p)) = error ("qfFormulaToBounds: AtGt not handled (since not supported)")

    solveN :: (Fractional a, Num cond)
           => [AtFormula (Polynomial a)]
	   -> DTreeGen (Polynomial a) cond ([GVector (FracPoly a)],[GVector (FracPoly a)])
    solveN []       = dLeaf ([],[])
    solveN (at:ats) =
	compDTreeG (\(l,u) ->
	    case at of
	      AtGe p -> solve p varNr nV $ \mbP' ->
			case mbP' of
			  Nothing -> dLeaf (l,u)
			  Just (flip,p')
			    | flip       ->  dLeaf (l  , p':u)
			    | otherwise  ->  dLeaf (p':l, u)
	      AtEq p -> solve p varNr nV $ \mbP' ->
			case mbP' of
			  Nothing -> dLeaf (l,u)
			  Just (flip,p') -> dLeaf (p':l, p':u)
              atf -> error ("qfFormulaToBounds: internal error (" ++ show atf
                            ++ " should not occur")
	    )
	    (solveN ats)

solve :: (Fractional a, Num cond)
      => Polynomial a
      -> Integer
      -> Integer
      -> (Maybe (Bool,GVector (FracPoly a)) -> DTreeGen (Polynomial a) cond val)
      -> DTreeGen (Polynomial a) cond val
solve p varNr nV f =
    case linearCoeffs (splitPoly nV p) of
      Nothing -> f Nothing
      Just lc'
	  | all (==0) inner'' ->
	      sCond myvar (f $ Just (True,  solved))
			  (f Nothing)
			  (f $ Just (False, solved))
	  | otherwise -> f Nothing
	  where
	  lc = map id {- removeNIndets ... -} lc'
	  (outer,myvar:inner') = splitAt (fromIntegral varNr) lc
	  const:inner'' = {- trace (show (outer,myvar,inner')) $ -} reverse inner'
	  solved = vector $ map ((%%myvar).negate) (outer ++ [const])
    

-- |
-- Convert a 'SortedSystem' to an equivalent 'GIneqSystem'.
--
sortedSystemToIneqSystem :: (Num a, Integral b)
                         => b   -- ^ number of variables
                         -> b   -- ^ number of parameters
                         -> SortedSystem a
                         -> GIneqSystem a
sortedSystemToIneqSystem nV nP [] =
    ineqSystem ((nV,nP), zeroMatrix 0 (nV+nP+1))
sortedSystemToIneqSystem nV nP ss =
    ineqSystem ((nV,nP), multijoin1 $ reverse $ map mkI ss)
    where
    mkI (Equation x) = m `join` negate m
	where
	m = rowMatrix (vs ++ negate 1 : genericReplicate (nV+nP-l) 0 ++ paramsAndConst)
        (vs,paramsAndConst) = genericSplitAt (l-nP-1) $ vectorEntries x
        l = vectorLength x

    mkI (LowerUpper ls us) =
	case (ls, us) of
	  ([],[]) -> zeroMatrix 0 (nV+nP+1)
	  (ls,[]) -> lsM
	  ([],us) -> usM
	  (ls,us) -> lsM `join` usM
	where
	lsM = m (map negate ls) 1
	usM = m us              (negate 1)
	m xs v = multiconc1 [vars
			    ,matrix $ replicate (length xs) [v]
			    ,zeroMatrix (genericLength xs) (nV+nP-l)
                            ,paramsAndConst
			    ]
	    where
	    vars = rowVectorsToMatrix1 $ map (\v -> deleteEntries v (l-nP-1) (nP+1)) xs
            paramsAndConst =
                rowVectorsToMatrix1 $ map (\v -> deleteEntries v 0 (l-nP-1)) xs
	    l = vectorLength (head xs)

countBounds :: Integral b => SortedSystem a -> b
countBounds = fromIntegral . sum . map cb
    where
    cb (Equation _) = 1
    cb (LowerUpper ls us) = length ls + length us
