---
-- Vectors and manipulation functions for vectors.
-- @author Armin Gr&ouml;&szlig;linger
-- @version $Id: Vector.hs 4157 2009-07-08 16:57:30Z groessli $
module LMath.Types.Vector
    ( GVector, Vector, vector, zeroVector, vectorEntries, vectorEntry,
      vectorLength, setVectorEntries, setVectorEntry, subVector,
      modifyVectorEntry,
      insertEntries, deleteEntries,
      integralVector, integralVector1,
      mulVV, addVV, subVV, negV, mulVS, mulSV, isNegV 
)
where

import Data.Array
import Data.List
import Data.Ratio

import CppInterface
import LMath.Types.LClasses

newtype GVector a = GVector (Array Integer a)

type Vector = GVector Rational

instance Eq a => Eq (GVector a) where
    (GVector x) == (GVector y)  =  x == y

instance Functor GVector where
    fmap f (GVector v) = GVector (fmap f v)

instance Ord a => Ord (GVector a) where
    compare v1 v2  =  vectorEntries v1 `compare` vectorEntries v2

vector :: [a] -> GVector a
vector rs = GVector (listArray (0,genericLength rs-1) rs)

zeroVector :: Num a => Integer -> GVector a
zeroVector n = GVector (listArray (0,n-1) (genericReplicate n 0))

vectorEntries :: GVector a -> [a]
vectorEntries (GVector v) = elems v

vectorLength :: Integral b => GVector a -> b
vectorLength (GVector v) = fromIntegral n + 1
    where
    (_,n) = bounds v

vectorEntry :: Integral b => GVector a -> b -> a
vectorEntry (GVector v) off = v ! fromIntegral off

setVectorEntries :: Integral b => GVector a -> [(b,a)] -> GVector a
setVectorEntries (GVector v) upd =
    GVector (v//[(fromIntegral idx,v) | (idx,v) <- upd])

setVectorEntry :: Integral b => GVector a -> b -> a -> GVector a
setVectorEntry v idx value = setVectorEntries v [(idx,value)]

modifyVectorEntry :: GVector a -> Integer -> (a -> a) -> GVector a
modifyVectorEntry v idx f = setVectorEntry v idx (f $ vectorEntry v idx)

subVector :: Integral b => GVector a -> b -> b -> GVector a
subVector (GVector v)  idx' width' = GVector $ listArray (0,width-1) [v!(i+idx) | i<-[0..width-1]]
    where
    idx   = fromIntegral idx' 
    width = fromIntegral width'

addVV :: Num a => GVector a -> GVector a -> GVector a
addVV v1 v2
    | vectorLength v1 == vectorLength v2 =
	vector $ zipWith (+) (vectorEntries v1) (vectorEntries v2)
    | otherwise = error ("addVV: vectors do not match (" ++
			 show (vectorLength v1) ++ " vs. " ++
			 show (vectorLength v2) ++ " entries)")

subVV :: Num a => GVector a -> GVector a -> GVector a
subVV v1 v2
    | vectorLength v1 == vectorLength v2 =
	vector $ zipWith (-) (vectorEntries v1) (vectorEntries v2)
    | otherwise = error ("subVV: vectors do not match (" ++
			 show (vectorLength v1) ++ " vs. " ++
			 show (vectorLength v2) ++ " entries)")

-- | returns true if vector is negative
isNegV :: forall a. (Num a, Ord [a]) => GVector a -> Bool
isNegV v = ves < replicate (length ves) 0 where
           ves = vectorEntries v


negV :: Num a => GVector a -> GVector a
negV (GVector v) = GVector (fmap negate v)

mulVV :: Num a => GVector a -> GVector a -> a
mulVV v1 v2
    | vectorLength v1 == vectorLength v2 =
	sum $ zipWith (*) (vectorEntries v1) (vectorEntries v2)
    | otherwise = error "mulVV: vectors don't match"

-- vector * scalar
mulVS :: Num a => GVector a -> a -> GVector a
mulVS (GVector v) r = GVector (fmap (*r) v)

mulSV :: Num a => a -> GVector a -> GVector a
mulSV = flip mulVS


insertEntries :: Integral b => GVector a -> b -> [a] -> GVector a
insertEntries v idx es = vector (p1++es++p2)
    where
    (p1,p2) = splitAt (fromIntegral idx) (vectorEntries v)

deleteEntries :: (Integral b, Integral c) => GVector a -> b -> c -> GVector a
deleteEntries v idx count = vector (p1++p3)
    where
    (p1,p23) = splitAt (fromIntegral idx)   (vectorEntries v)
    p3       = drop    (fromIntegral count) p23

integralVector :: Integral a => GVector (Ratio a) -> GVector a
integralVector v = fmap (\e -> numerator e * (denom `div` denominator e)) v
    where
    denom = foldl lcm 1 $ map denominator $ vectorEntries v

integralVector1 :: Integral a => GVector (Ratio a) -> GVector a
integralVector1 v
    | all ((==1).denominator) $ vectorEntries v  =  fmap numerator v
    | otherwise  =  error ("integralVector1: denominators are not 1:\n" ++
                           show (fmap toRational v)
                          )


instance Concatenatable (GVector a) where
    conc v1 v2 = vector (vectorEntries v1 ++ vectorEntries v2)
    unconc v   = [vector [x] | x <- vectorEntries v]

instance Show a => Show (GVector a) where
    show v = "Vector" ++
	     concat [" " ++ show x | x <- vectorEntries v] ++ "\n"

showVector :: Integral a => GVector (Ratio a) -> String
showVector v = "Vector" ++
	       concat [" " ++ s x | x <- vectorEntries v] ++ "\n"
    where
    s x = if (denominator x == 1) then
	    show (numerator x) 
	  else 
	    show (numerator x) ++ "/" ++ show (denominator x)


instance Num a => Num (GVector a) where
    (+)    = addVV
    negate = negV
    (*)    = error "(*) undefined for GVector"
    signum = error "`signum' undefined for GVector"
    abs    = error "`abs' undefined for GVector"
    fromInteger = error "`fromInteger' undefined for GVector"

