{-
    This file is part of the LooPo project, see
    <http://www.infosun.fim.uni-passau.de/cl/loopo/>.

    Copyright (C) 2006 Universitaet Passau, Germany
    Author: Armin Groesslinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-}

-- |
-- This module provides a monad and the necessary combintors
-- to build parsers for /LooPo/'s data structures.
-- A parser is combined from the primitive parse action, like
-- 'parseInteger'. These functions (and the resulting parser)
-- have types like @Parser a@. The functions here are
-- based on 'Text.ParserCombinators.Parsec' and can be
-- freely mixed with functions from Parsec.
-- Here's an example to illustrate how to compose the parse actions:
-- 
-- > myParse :: Parser (Integer, [Integer], Integer)
-- > myParse =
-- >     do i    <- parseInteger
-- >        list <- parseIntegers 3
-- >        j    <- parseInteger
-- >        return (i, list, j)
-- 
-- To run the parser, simply invoke 'runLP'. After running
-- 
-- > res = runLP myParse "1 2 3 4 5"
-- 
-- @res@ is bound to the value
-- @ParseOk (1, [2, 3, 4], 5)@.
-- 
-- When the parser fails at any point (not enough input or wrong input),
-- the current and all subsequent calls to parse actions return values
-- bound to a call to 'error'; that means that accessing any of
-- these values will cause program termination. Whether the parser has failed
-- can be checked at any time with 'hasFailed'. In case of error,
-- 'runLP' returns @ParseFailed (msg, line)@ where
-- @msg@ is an error message and @line@ is the line
-- where the error occurred.
--
-- Author: Armin Groesslinger
-- Version: $Id: LP.hs 4188 2009-08-17 16:42:26Z groessli $
--
module LP (Parser, LPResult(..),
	   runLP, parseInteger, parseN, parseIntegers, parseRational,
	   parseFraction, parseFraction0, parseChar, attempt,
	   match, parseOrError, loadAndParse)


where

import Control.Monad
import Data.Char
import Data.List
import Data.Ratio

import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Error
import Text.ParserCombinators.Parsec.Pos


-- |
-- Parse an integer from the parser's input.
--
parseInteger :: Parser Integer
parseInteger = do f <- (char '-' >> return (-1)) <|> return 1
		  ds <- many1 digit
		  spaces
		  return (f * read ds)

-- |
-- Parse a single character.
-- 'parseChar' accepts any character from the given string
-- as next input symbol. If the next input symbol is not
-- contained in the given string, the parse action fails.
--
parseChar :: String  -- ^ characters accepted by the parse action 
          -> Parser Char
parseChar = oneOf

-- |
-- Parses a rational number.
-- Two integers are taken from the input (without any special separator
-- between them) and make into a rational number. The first integer is
-- used as numeration and the second serves as denominator of the
-- fraction.
--
parseRational :: Parser Rational
parseRational =
    do num   <- parseInteger
       denom <- parseInteger
       return (num % denom)

-- |
-- Parse a fraction.
-- Legal fractions have the form @n@ or @n \/ d@, where
-- @n@ is interpreted as the numerator and @d@ as the
-- denominator of the fraction. In the first from (i.e. the @\/@ and the
-- denominator are missing) the denominator is assumed to be 1.
--
parseFraction :: Integral a => Parser (Ratio a)
parseFraction =
    do mbF <- parseFraction0
       case mbF of
         Just f  -> return f
         Nothing -> fail "fraction with denominator 0 in input"

-- |
-- Parse a fraction or @0\/0@.
-- Like 'parseFraction', but return @Nothing@ if @0\/0@ is read
-- from the input.
--
parseFraction0 :: Integral a => Parser (Maybe (Ratio a))
parseFraction0 =
    do num   <- parseInteger
       slash <- attempt (match "/")
       case slash of
         Nothing -> return $ Just $ fromIntegral num
         Just _  -> spaces >> parseInteger >>= \denom ->
		    return $ if denom == 0
                                then Nothing
                                else Just (fromIntegral num % fromIntegral denom)

-- |
-- Repeats a parse action.
-- The action @parse@ is repeated @count@ times
-- and the results are returned as a list. The first element of the list
-- corresponds to the result of the first execution of @parse@
-- and so on.
--
parseN :: Integer       -- ^ @count@
       -> Parser a      -- ^ @parse@
       -> Parser [a]
parseN count op = sequence $ take (fromInteger count) (repeat op)

-- |
-- Parses a sequence of integers.
-- The parsed seqeunce of integers is returned as a list.
--
parseIntegers :: Integer -> Parser [Integer]
parseIntegers count = parseN count parseInteger


match :: String -> Parser ()
match str = string str >> return ()

-- |
-- Tries to run a sub-parser.
-- The given parse action is executed on the current state of the parser.
-- When it fails, the state of the main parser remains unchanged and
-- @Nothing@ is returned.
-- When it succeeds, its result is returned and the parser's new state
-- is the state after running the sub-parser.
--
attempt :: Parser a -> Parser (Maybe a)
attempt p = (liftM Just $ try p) <|> return Nothing

-- |
-- Represents a parsing result. A parse is either
-- successful and returns a value, or it fails with an error message
-- and the number of the line where the error occurrred.
--
data LPResult a = ParseOk a | ParseFailed (String, Integer) deriving (Show)

-- |
-- Run the a given parser on a given input.
--
runLP :: Parser a -> String -> LPResult a
runLP p str = case parse (spaces >> p >>= \r -> eof >> return r) "" str of
	        Right res -> ParseOk res
		Left  pe  ->
		    ParseFailed (unlines $ map messageString $
				           errorMessages pe,
				 fromIntegral $ sourceLine $ errorPos pe
				)

parseOrError :: String -> Parser a -> String -> a
parseOrError funname parseAction input =
    case runLP parseAction input of
    ParseOk     result      -> result
    ParseFailed (msg, line) ->
	error (funname ++ ": parser failed in line " ++
	       show line ++ " with message:\n\"" ++ msg ++
	       "\" for input\n" ++ input)

loadAndParse :: FilePath -> Parser a -> IO (LPResult a)
loadAndParse filename parser =
    do contents <- readFile filename
       return $ runLP parser contents
