-- |
-- This module contains the definitions of /LooPo/'s data structures.
-- Each data structure can be written to and read from disk
-- using the provided functions from this module.
--
-- Author: Armin Groesslinger
--
-- Version: $Id: LoopoDS.hs 3816 2007-11-16 14:45:30Z groessli $
--
module LoopoDS
    ( -- * Loading of LooPo data structures
      LoadError(..),
      loadParseTreeAndSymTab,
      loadStatementsAndAccesses, loadSurroundingLoops,
      loadSurroundingIfs,
      loadDependences, loadAccessPairs,
      loadCfGraph,

      -- * Unsafe interface for testing purposes
      LoopoDS(..), unsafeReadLoopoDS,
      
      -- * Modules defining LooPo data structures
      module LoopoDS.IntegralID,
      module LoopoDS.Access,
      module LoopoDS.AccessPair,
      module LoopoDS.Arrays,
      module LoopoDS.CfGraph,
      module LoopoDS.Dependence,
      module LoopoDS.Groups,
      module LoopoDS.Qual,
      module LoopoDS.Quast,
      module LoopoDS.SchedPlace,
      module LoopoDS.Statement,
      module LoopoDS.SurroundingIfs,
      module LoopoDS.SurroundingLoops,
      module LoopoDS.Trafo,
      module LoopoDS.TRC,
      module LoopoDS.TrDep,

      module ScannerParser.LambdaExpression,
      module ScannerParser.ParseTree,
      module ScannerParser.SymbolTable,

      -- * Modules with auxiliary definitions
      module LoopoDS.AGraph
    )
where

import Data.List
import Data.Ix
import Data.Dynamic
import Control.Exception as CE
import Control.Monad hiding (join)
import Data.Ratio
import System.IO.Unsafe (unsafePerformIO)
import Text.ParserCombinators.Parsec

import LP
import ParseBase
import LMath

import LoopoDS.IntegralID
import LoopoDS.Access
import LoopoDS.AccessPair
import LoopoDS.Arrays
import LoopoDS.CfGraph
import LoopoDS.Dependence
import LoopoDS.Groups
import LoopoDS.Qual
import LoopoDS.Quast
import LoopoDS.SchedPlace
import LoopoDS.Statement
import LoopoDS.SurroundingIfs
import LoopoDS.SurroundingLoops
import LoopoDS.Trafo
import LoopoDS.TRC
import LoopoDS.TrDep

import LoopoDS.AGraph

import ScannerParser.ScannerParser
import ScannerParser.LambdaExpression
import ScannerParser.ParseTree
import ScannerParser.SymbolTable

-- |
-- If a file cannot be loaded, a 'LoadError' value is
-- thrown as a dynamic exception.
--
data LoadError = LoadError String FilePath Integer String
		 deriving Typeable

instance Show LoadError where
    show (LoadError funname path line msg) =
	"Error loading " ++ show path ++ " in " ++
	funname ++ ", line " ++ show line ++ ":\n" ++ msg

loadOrException :: String -> (FilePath -> IO (LPResult a)) -> FilePath -> IO a
loadOrException funname readFn path =
    readFn path >>= \res ->
    case res of
       ParseOk     r -> return r
       ParseFailed (msg, line) -> throwDyn (LoadError funname path line msg)
	   

data ParseTree = ParseTree { ptMain        :: LambdaTree,
                             ptFunctions   :: [LambdaNode],
                             ptSymbolTable :: SymTab,
                             ptInfoVects   :: [InfoVector]
                           }
                 deriving (Show)

loadParseTreeAndSymTab :: FilePath   -- ^ base name of the project
                       -> IO ParseTree
loadParseTreeAndSymTab basepath =
    do res <- CE.try $ openParseFile (basepath ++ ".out")
       case res of
         Right (st, m, fs, infos) ->
             return $ ParseTree { ptMain = m, ptFunctions = fs,
                                  ptSymbolTable = st, ptInfoVects = infos
                                }
         Left ex -> throwDyn $ LoadError "loadParseTreeAndSymtab"
                                         file
                                         (-1)
                                         (show ex)
    where
    file = basepath ++ ".out"

-- |
-- Load the statements an the accesses.
--
loadStatementsAndAccesses :: FilePath -> IO ([Statement], [Access])
loadStatementsAndAccesses path =
    do stmts <- loadOrException "loadStatementsAndAccesses" readISPC (path++".ispc")
       raccs <- loadOrException "loadStatementsAndAccesses" readACC (path++".acc")
       return (stmts, map (rawAccessToAccess stmts) raccs)

-- |
-- Load the surrounding loops.
--
loadSurroundingLoops :: FilePath -> IO SurroundingLoops
loadSurroundingLoops path =
    loadOrException "loadSurroundingLoops" readSLP (path++".slp")

-- |
-- Load surrounding Ifs.
--
loadSurroundingIfs :: FilePath -> IO SurroundingIfs
loadSurroundingIfs path =
    loadOrException "loadSurroundingIfs" readSIF (path++".sif")

-- |
-- Load the dependences.
--
loadDependences :: FilePath -> IO [Dependence]
loadDependences path =
    loadOrException "loadDependences" readDependences (path ++ ".dep")

-- |
-- Load the access pairs.
--
loadAccessPairs :: FilePath -> IO [AccessPair]
loadAccessPairs path =
    loadOrException "loadAccessPairs" readACP (path ++ ".acp")


-- |
-- Load the control-flow graph.
--
loadCfGraph :: FilePath -> IO CfGraph
loadCfGraph path =
    loadOrException "loadCfGraph" readCfGraph (path ++ ".cfg")


-- |
-- A record holding all the data structures calculated by LooPo
-- for a given project.
--
data LoopoDS = LoopoDS
    {
     loopoACC   :: [RawAccess],  -- ^ raw Accesses
     loopoAccesses :: [Access],  -- ^ Accesses
     loopoACP   :: [AccessPair],                  -- ^ Access pairs
     loopoArrays:: [(String, [String], ArrayType)],          -- ^ arrays, their size identifiers for each dim and their IN/OUT/INOUT use information
     loopoSLP   :: SurroundingLoops,              -- ^ Surrounding loops
     loopoSIF   :: SurroundingIfs,                -- ^ Surrounding ifs
     loopoGRP   :: (Integer, [[Integer]]),        -- ^ Groups
     loopoDEP   :: [Dependence],                  -- ^ Dependences
     loopoCFG   :: CfGraph,                       -- ^ Control-flow graph
     loopoISPC  :: [Statement],                   -- ^ Index spaces
     loopoTISPC :: [(TStatementID, StatementID, IneqSystem)],  -- ^ Transformed index spaces
     loopoTrafo :: Trafo,              -- ^ Transformation
     loopoTRC   :: [TRC],              -- ^ Transformation constraints
     loopoQUAL  :: [(TStatementID, [QualEntry])], -- ^ Loop quality vectors for each target stmt
     loopoSCHED :: (Integer, Integer, Schedules), -- ^ vars, params, schedule
     loopoPLACE :: (Integer, Integer, Placements StatementID) -- ^ vars, params, placement
    }
	       deriving (Show)

-- |
-- Read all LooPo data structures for a given project.
-- The data structures are read using @System.IO.Unsafe.unsafePerformIO@.
-- Use this function for testing purposes only, i.e., as long as
-- you do not have a @main@ function in your prototype implementation
-- (and use GHCi to run your unfinished implementation), 'unsafeReadLoopoDS'
-- is a convenient way to quickly (and dirtily) load the required
-- files. In a real, finished program you should always use
-- the @load...@ functions which operate in the 'IO' monad.
-- If something breaks due to the use of @unsafePerformIO@, you've
-- got to keep the pieces.
-- The argument of the function is the
-- base name of the project files to be loaded, i.e., a file path
-- without the extension \".acc\", \".acp\", etc.
--
unsafeReadLoopoDS ::    FilePath   -- ^ base name of the project files. 
		     -> LoopoDS
unsafeReadLoopoDS basename = lds
    where
    uPI = unsafePerformIO
    lds = LoopoDS { loopoACC      = r ".acc" readACC,
		    loopoAccesses = accs,
		    loopoACP      = uPI $ loadAccessPairs basename,
                    loopoArrays   = r ".arrays" readArrays,
		    loopoSLP      = uPI $ loadSurroundingLoops basename,
                    loopoSIF      = uPI $ loadSurroundingIfs basename,
		    loopoGRP      = r ".grp" readGroups,
		    loopoDEP      = uPI $ loadDependences basename,
                    loopoCFG      = uPI $ loadCfGraph basename,
		    loopoISPC     = stmts,
		    loopoTISPC = r ".tispc" readTISPC,
		    loopoTrafo = r ".trafo" readTrafo,
		    loopoTRC   = r ".trc"   readTRC,
		    loopoQUAL  = r ".qual" readQUAL,
		    loopoSCHED = r ".sched" readSchedule,
                    loopoPLACE = r ".place" readPlacement
		  }
    (stmts,accs) = uPI $ loadStatementsAndAccesses basename
    r ext fn =
	unsafePerformIO $
        loadOrException "unsafeReadLoopoDS" fn (basename ++ ext)



