-- |
-- A simple general graph representation.
-- Nodes and edges have integral IDs and are annoted
-- with user-supplied data.
--
module LoopoDS.AGraph
    ( AGraph(..),
      NodeID (..), EdgeID (..), NodeInfo (..), EdgeInfo (..),               
      predEdges, succEdges, predIDs, succIDs, succs, preds,
      isTransSucc, getEdgeBetween, getEdgeInfoBetween, nodeInfo, edgeInfo,
      testgraph
    )
where
	      

newtype NodeID = NodeID Integer deriving (Eq)
instance Show (NodeID) where 
    show (NodeID a) = "N" ++ show(a)
instance Ord (NodeID) where
    compare (NodeID a) (NodeID b) = compare a b

newtype EdgeID = EdgeID Integer deriving (Eq)
instance Show (EdgeID) where 
    show (EdgeID a) = "E" ++ show(a)

data NodeInfo a = NodeInfo {nodeID :: NodeID, nodeData :: a}
instance Show a => Show (NodeInfo a) where 
    show (NodeInfo {nodeID = nodeID, nodeData = nodeData}) = 
	"{" ++ show(nodeID) ++ "," ++ show(nodeData) ++ "}" 
	
data EdgeInfo b = EdgeInfo {edgeID :: EdgeID, edgeData :: b, 
                            edgeSrc :: NodeID, edgeDst :: NodeID} deriving (Eq)
{-instance Show a => Show (EdgeInfo a) where 
    show (EdgeInfo {edgeID = edgeID, edgeData = edgeData, edgeSrc = edgeSrc,
	edgeDst= edgeDst}) = "{" ++ show(edgeID) ++ "," ++ show(edgeData) ++ ","
	    ++ show(edgeSrc) ++ "->" ++ show(edgeDst) ++ "}" 
-}
instance Show a => Show (EdgeInfo a) where
    show (EdgeInfo {edgeID = edgeID, edgeData = edgeData, edgeSrc = edgeSrc,
            edgeDst= edgeDst}) = show(edgeID)
--             edgeDst= edgeDst}) = show(edgeID) ++ "(" ++ show(edgeSrc) ++ "," ++ show(edgeDst) ++ ")"
--             edgeDst= edgeDst}) = show(edgeID) ++ "(" ++ show(edgeData) ++ ")"



data AGraph a b = AGraph { graphNodes :: [NodeInfo a],
			   graphEdges :: [EdgeInfo b]
                         } deriving (Show)
		 

nodeInfo :: AGraph a b -> NodeID -> NodeInfo a
nodeInfo graph nodId = 
    head $ filter (\nid->nodeID nid==nodId) (graphNodes graph)

edgeInfo :: AGraph a b -> EdgeID -> EdgeInfo b
edgeInfo graph edgId =
    head $ filter (\eid->edgeID eid==edgId) (graphEdges graph)


predEdges  :: AGraph a b -> NodeID -> [EdgeInfo b]
predEdges graph nodId =
    filter (\ein->edgeDst ein==nodId) (graphEdges graph)

succEdges  :: AGraph a b -> NodeID -> [EdgeInfo b]
succEdges graph nodId = filter (\ein->edgeSrc ein==nodId) (graphEdges graph)

predIDs  :: AGraph a b -> NodeID -> [NodeID]
predIDs graph nodId = map edgeSrc $ predEdges graph nodId

succIDs  :: AGraph a b -> NodeID -> [NodeID]
succIDs graph nodId = map edgeDst $ succEdges graph nodId

preds  :: AGraph a b -> NodeID -> [NodeInfo a]
preds graph nodId = map (nodeInfo graph) (predIDs graph nodId)

succs  :: AGraph a b -> NodeID -> [NodeInfo a]
succs graph nodId = map (nodeInfo graph) (succIDs graph nodId)

-- Checks if a node (check) is transitive successor of a node (current)
isTransSucc :: AGraph a b -> NodeID -> NodeID -> Bool
isTransSucc graph check current = 
    let 
	isTS :: AGraph a b -> NodeID -> [NodeID] -> NodeID -> Bool
	isTS graph check visited current 
	    | check `elem` (succIDs graph current) = True 
	    | check `notElem` (succIDs graph current) = 
		foldr ((||).(isTS graph check (current:visited))) False 
		    [ y | y <- (succIDs graph current), y `notElem` visited]
		    
    in isTS graph check [] current

-- Returns the edge which connects a node (pred) with another node (succ),
-- where pred is the predecessor of succ.
getEdgeInfoBetween :: (Eq b) => AGraph a b -> NodeID -> NodeID -> (EdgeInfo b)
getEdgeInfoBetween graph pred succ =
    let edge = [(EdgeInfo a b c d) | EdgeInfo a b c d <- succEdges graph pred,
		a `elem` [a | EdgeInfo a _ _ _ <- predEdges graph succ]]
    in
	if edge /= [] then
	    head edge
	else
	    error ("No edge between this nodes: " ++ show pred ++ show succ)

-- Returns the edge which connects a node (pred) with another node (succ),
-- where pred is the predecessor of succ.
getEdgeBetween :: AGraph a b -> NodeID -> NodeID -> EdgeID
getEdgeBetween graph pred succ =
    let edge = [a | EdgeInfo a _ _ _ <- succEdges graph pred,
                a `elem` [a | EdgeInfo a _ _ _ <- predEdges graph succ]]
    in
    	if edge /= [] then
		head edge
	else
		error ("No edge between this nodes: " ++ show pred ++ show succ)


testnode0 =  NodeInfo (NodeID 0) "allein"
testnode1 =  NodeInfo (NodeID 1) "erster"
testnode2 =  NodeInfo (NodeID 2) "zweiter"
testnode3 =  NodeInfo (NodeID 3) "dritter"
testnode4 =  NodeInfo (NodeID 4) "vierter"
testnode5 =  NodeInfo (NodeID 5) "selber"

testedge1 = EdgeInfo (EdgeID 101) (True,"eins")  (NodeID 1) (NodeID 3)
testedge2 = EdgeInfo (EdgeID 102) (False,"zwei") (NodeID 3) (NodeID 4)
testedge3 = EdgeInfo (EdgeID 103) (False,"drei") (NodeID 4) (NodeID 1)
testedge4 = EdgeInfo (EdgeID 104) (False,"vier") (NodeID 3) (NodeID 2)
testedge5 = EdgeInfo (EdgeID 105) (True,"fuenf") (NodeID 1) (NodeID 4)
testedge6 = EdgeInfo (EdgeID 106) (False,"loop") (NodeID 2) (NodeID 2)
testedge7 = EdgeInfo (EdgeID 107) (False,"self") (NodeID 5) (NodeID 5)

testgraph = AGraph [testnode0,testnode1,testnode2,testnode3,testnode4,
		    testnode5]
		   [testedge1,testedge2,testedge3,testedge4,testedge5,
		    testedge6,testedge7]
