module LoopoDS.Access
    ( Access(..), AccessType(..), AccessID, ArrayID, RawAccess,
      readACC, rawAccessToAccess, writeACC
    )
where

import LMath
import ParseBase

import LoopoDS.IntegralID
import LoopoDS.Statement

import Data.List

type RawAccess = (Integer, AccessType, Integer, Integer, AffineFunction)
-- arrayId,accType,laNode,stmtNr,indexFct

---
-- Reads the file .acc containing the accesses.
-- When the given file is parsed correctly, a list of tuples 
-- <code>(arrayId,accType,laNode,stmtNr,indexFct)</code> is returned. 
-- Every tuple consists of the the number of the variable/array
-- in the symbol table (<code>arrayId</code>), the access type
-- Read or Write (<code>accType</code>), the id of the parsre tree 
-- node (<code>laNode</code>), the number of statement 
-- (<code>stmtNr</code>) and the array's index function
-- (<code>indexFct</code>).
-- @param filename - the file to read.
-- @monadic r - <code>ParseOk </code> when parsing
--            is successsful, <code>ParseFailed </code>
--            otherwise.
--
readACC :: FilePath -> IO (LPResult [RawAccess])
readACC filename = loadAndParse filename parseACCs
    where
    parseACCs :: Parser [RawAccess]
    parseACCs =
	do numberOfVars <- parseInteger
           accs <-   parseN numberOfVars (parseInteger >>= 
                                           \n -> parseN n parseACC )
           return (concat accs)

    parseACC :: Parser RawAccess
    parseACC =
	do arrayId     <- parseInteger
           intAccType  <- parseInteger
           laNode      <- parseInteger
           stmtNr      <- parseInteger
           indexFct    <- parseAffineFunctionI
           return (arrayId,intToAccessType intAccType, laNode,
		   stmtNr,indexFct)

writeACC :: [RawAccess] -> FilePath -> IO ()
writeACC accesses filename = writeFile filename (accessesToString accesses)

accessesToString :: [RawAccess] -> String
accessesToString accs = show (length groupedAccs) ++ "\n" ++ concat (map accessesToString' groupedAccs)
    where
    groupedAccs = groupBy compareArrayID accs
    compareArrayID (x,_,_,_,_) (y,_,_,_,_) = x==y
    accessesToString' accs = show (length accs) ++ "\n" ++ (concat . map accessToString) accs

accessToString :: RawAccess -> String
accessToString (arrayId, accType, laNode, stmtNr, indexFct) =
    show arrayId ++ " " ++ show (accessTypeToInt accType) ++ " " ++ show laNode
    ++ " " ++ show stmtNr ++ "\n" ++ affineFunctionToString True False indexFct

rawAccessToAccess :: [Statement] -> RawAccess -> Access
rawAccessToAccess ispc (arrayId,accType,laNode,stmtNr,indexFn) =
    case lookup (integerToID stmtNr) ispc of
      Just dom -> Access { accID       = AccessID laNode,
			   accIndexFct = indexFn,
			   accStmtID   = integerToID stmtNr,
			   accDomain   = dom,
			   accType     = accType,
			   accArrayID  = ArrayID arrayId
			 }
      Nothing -> error ("rawAccessToAccess: Statement ID " ++ show stmtNr ++
			" not found")

data AccessType = Read | Write deriving(Show,Eq)

intToAccessType 0 = Read
intToAccessType 1 = Write
intToAccessType _ = error ("unknown access type")

accessTypeToInt :: AccessType -> Int
accessTypeToInt Read  = 0
accessTypeToInt Write = 1

newtype AccessID = AccessID Integer
    deriving (Show, Eq, Ord)

instance HasIntegralID AccessID where
    integerFromID (AccessID i) = fromIntegral i
    integerToID = AccessID . fromIntegral

newtype ArrayID = ArrayID Integer
    deriving (Show, Read, Eq, Ord)

instance HasIntegralID ArrayID where
    integerFromID (ArrayID i) = fromIntegral i
    integerToID = ArrayID . fromIntegral

data Access = Access { accID       :: AccessID,
		       accIndexFct :: AffineFunction,
		       accStmtID   :: StatementID,
		       accDomain   :: IneqSystem,
		       accType     :: AccessType,
		       accArrayID  :: ArrayID
		     } deriving Show


-- instance Show Access where
--   show acc = "Access with function " ++ show (getFunctionOfAccess acc) ++
--     " with " ++ show (getNrBoundsOfAccess acc) ++ " bounds.\n"

