module LoopoDS.AccessPair
    ( AccessPair(..),
      readACP, writeACP
    )
where

import ParseBase
import LoopoDS.Access
import LoopoDS.Dependence
import LoopoDS.IntegralID

data AccessPair = AccessPair { acpDepID :: DependenceID,
			       acpSrcAccID :: AccessID,
			       acpDestAccID :: AccessID
			     }
		  deriving (Show, Eq)

---
-- Reads the file .acp containing the access pairs.
-- When the given file is parsed correctly, a list of tuples 
-- <code>(depId,sourceAcc,destAcc)</code> is returned. 
-- Every tuple consists of the the number of the dependence
-- (<code>depId</code>), and the access identifier for source
-- (<code>sourceAcc</code>)  and destination (<code>destAcc</code>) of
-- the dependent accesses.
-- @param filename - the file to read.
-- @monadic r - <code>ParseOk </code> when parsing
--            is successsful, <code>ParseFailed </code>
--            otherwise.
--
readACP :: FilePath -> IO (LPResult [AccessPair])
readACP filename = loadAndParse filename parseACPs
    where
    parseACPs :: Parser [AccessPair]
    parseACPs =
	do numberOfPairs <- parseInteger
           parseN numberOfPairs parseACP

    parseACP :: Parser AccessPair
    parseACP =
	do depId       <- parseInteger
           sourceAcc   <- parseInteger
           destAcc     <- parseInteger
           return $ AccessPair { acpDepID = integerToID depId,
		                 acpSrcAccID = integerToID sourceAcc,
		                 acpDestAccID = integerToID destAcc
                               }


acpToTriple :: AccessPair -> (DependenceID, AccessID, AccessID)
acpToTriple acp = (acpDepID acp, acpSrcAccID acp, acpDestAccID acp)

tripleToACP :: (DependenceID, AccessID, AccessID) -> AccessPair
tripleToACP (depID, srcID, destID) = AccessPair { acpDepID = depID, acpSrcAccID = srcID, acpDestAccID = destID }

writeACP :: [AccessPair] -> FilePath -> IO ()
writeACP accessPairs filename = writeFile filename (accessPairsToString accessPairs)

accessPairsToString :: [AccessPair] -> String
accessPairsToString accessPairs = show (length accessPairs) ++ "\n" ++ unlines (map accessPairToString accessPairs)

accessPairToString :: AccessPair -> String
accessPairToString acp = (show $ integerFromID $ acpDepID acp) ++ " "
                         ++ (show $ integerFromID $ acpSrcAccID acp) ++ " " ++ (show $ integerFromID $ acpDestAccID acp)


