module LoopoDS.Arrays
    ( readArrays
    , ArrayType(..)
    )
where

import ParseBase
import qualified Text.ParserCombinators.Parsec.Token as P
import Text.ParserCombinators.Parsec.Language(javaStyle)
import Text.ParserCombinators.Parsec.Error

import Text.ParserCombinators.Parsec
import Data.Char

-- The lexer
lexer       = P.makeTokenParser javaStyle
    
identifier    = P.identifier lexer
squares       = P.squares lexer
symbol        = P.symbol lexer
commaSep      = P.commaSep lexer
stringLiteral = P.stringLiteral lexer

data ArrayType = ArrayIN | ArrayOUT | ArrayINOUT
                 deriving (Show, Eq)

-- |
-- This parser reads a filename and parses the content like 
-- in the following example:
-- @3@
-- @A, [n,m], OUT@
-- @B, [n], IN@
-- @C, [m], INOUT@
-- Here, the first line indicates the number of following array descriptions.
-- Each line consists of an array identifier, followed by a list of identifiers
-- indicating the size of each array dimension, then a description wether the
-- array is used for input, output or both (INOUT).
--
readArrays :: FilePath -> IO (LPResult [(String, [String], ArrayType)])
readArrays filename = loadAndParse filename parseArrays

parseArrayType :: Parser ArrayType
parseArrayType =
    do arrayTypeStr <- identifier <?> "array type: IN, OUT or INOUT"
       spaces
       let upperArrayTypeStr = map toUpper arrayTypeStr
           arrayType = case upperArrayTypeStr of
                       "IN" -> ArrayIN
                       "OUT" -> ArrayOUT
                       "INOUT" -> ArrayINOUT
                       otherwise -> error "wrong array type"
       return arrayType
       

parseArrays :: Parser [(String, [String], ArrayType)]
parseArrays = 
    do nLines <- parseInteger <?> "number of array information lines"
       spaces
       arrays <- parseN nLines parseArrayLine <?> "array information"
       spaces
       eof
       return arrays

parseArrayLine :: Parser (String, [String], ArrayType)
parseArrayLine =
    do arrayName <- identifier <?> "array identifier"
       spaces
       comma <- parseChar "," <?> "comma"
       spaces
       arrayDims <- squares (commaSep (many1 (noneOf ",]") <?> "size expression")) <?> "array dims"
       spaces
       comma <- parseChar "," <?> "comma"
       spaces
       arrayType <- parseArrayType <?> "array type"
       spaces
       return (arrayName, arrayDims, arrayType)


{-- test:			
*LoopoDS.Arrays> input <- readFile "LoopoDS/test.arrays"
*LoopoDS.Arrays> parseTest parseArrays input
--}
