-- |
-- Control-flow graph.
--
module LoopoDS.CfGraph
    ( CfgNodeData(..), CfgEdgeData(..), CfgNodeType(..),
      CfgEdgeType(..),
      CfGraph, Relation,
      getNodeType, getEdgeType, getAdvancedCfGraph,
      getSrcOfEdgeFromStatementID, getLoopIDFromLoopHead, getIntOfLoopID, 
      getMaybeOfEdgeData, getEdgeTypeOfEdge,
      readCfGraph
    )
where

import Debug.Trace
import LoopoDS.AGraph
import LoopoDS.IntegralID
import LoopoDS.SurroundingLoops (LoopID(..))
import LoopoDS.Statement (StatementID(..))

import LP

import Text.ParserCombinators.Parsec

-- Useful to debug CfGraph without all Loopo DS
-- newtype StatementID = StatementID Integer deriving (Show)


type CfGraph = AGraph CfgNodeData CfgEdgeData
type Relation = String

data CfgNodeData = CfgNodeData CfgNodeType deriving (Eq)
instance Show (CfgNodeData) where
    show (CfgNodeData a) = show (a)

data CfgEdgeData = CfgEdgeData CfgEdgeType (Maybe StatementID) Relation deriving (Eq)
instance Show (CfgEdgeData) where
    show (CfgEdgeData a Nothing c) = show (a) ++ "," ++ show (c)
    show (CfgEdgeData a (Just b) c) = show (a) ++ "," ++ show (b) ++ "," ++ show (c)

data CfgNodeType = Fork | Join | LoopHead | Standard | ForLoopHead | DoWhileLoopHead  deriving (Show, Eq)
  -- the information is directly taken from the .cfg file.
  -- Fork represents, e.g., an IF, Join an Endif.

data CfgEdgeType = StructuredBackward LoopID | UnstructuredBackward | BodyStatement |
                   Forward deriving (Show, Eq)
  -- the analysis only checks *Backward edges and BodyStatements, i.e.,
  -- statements in the sense of LooPo. Other Forward edges are not considered by
  -- CfFada.
  -- In goto programs, the increment operation of a manually programmed loop has
  -- a StatementID, but is only a Forward edge -- once the loop index is
  -- detected by index variable recognition and inserted appropriately in the
  -- symbol table.
  -- As long as a variable is not detected as a loop index, the corresponding
  -- statement is a BodyStatement (and the analysis will make very conservative
  -- approximations).

  -- Similarly, an affine loop condition is not a statement, i.e., the
  -- corresponding edges are of type Forward. However, if the loop bound shall
  -- contain a variable, it must be manually coded (or it is integrated by
  -- normloops) and the statement must be considered by CfFada. Thus, the type
  -- is BodyStatement.

  -- This information comes from the .cfg file:
  --  statment 0 means Forward, statement # 0 means BodyStatement, as long as
  -- there is no additional indication that the statement belongs to static loop
  --  control only ('statement <ID> static' instead of 'statement <ID>' in the
  -- .cfg file).


getNodeType :: CfGraph -> NodeID -> CfgNodeType  
getNodeType graph nodID = test (nodeData (nodeInfo graph nodID))
    where test (CfgNodeData nodeType) = nodeType

getEdgeType :: CfGraph -> EdgeID -> CfgEdgeType  
getEdgeType graph nodID = test (edgeData (edgeInfo graph nodID))
    where test (CfgEdgeData edgeType _ _) = edgeType

getEdgeTypeOfEdge :: CfgEdgeData -> CfgEdgeType
getEdgeTypeOfEdge (CfgEdgeData edgeType _ _) = edgeType

orderOfNode :: [EdgeInfo CfgEdgeData] -> NodeID -> Int
orderOfNode list id = length [ x | x<-list, (edgeSrc x == id || edgeDst x == id)]

getAdvancedCfGraph :: CfGraph -> CfGraph
getAdvancedCfGraph graph = AGraph {graphNodes = (map transformLH) (graphNodes graph), graphEdges = graphEdges graph} where 
              transformLH :: NodeInfo CfgNodeData -> NodeInfo CfgNodeData
	      transformLH x = if (nodeData x == CfgNodeData LoopHead)
	          then
	             if (orderOfNode (graphEdges graph) (nodeID x) == 3) 
	                then NodeInfo {nodeID = nodeID x, nodeData = CfgNodeData DoWhileLoopHead}
	                else NodeInfo {nodeID = nodeID x, nodeData = CfgNodeData ForLoopHead}
		  else x

getSrcOfEdgeFromStatementID :: CfGraph -> StatementID -> NodeID
getSrcOfEdgeFromStatementID (AGraph _ xs) stID = getEdgeID stID xs where
              getMaybeStID (CfgEdgeData _ maybeID _) = maybeID
              getEdgeID st (x:xs) | (getMaybeStID (edgeData x)) == (Just stID) = edgeSrc x 
	      getEdgeID st (x:xs) = getEdgeID st xs


getLoopIDFromLoopHead :: CfGraph -> NodeID -> LoopID
getLoopIDFromLoopHead (AGraph nodes edges) id = getLoopID (head [ x | x <- edges, (edgeDst x) == id, isStructuredBackward (edgeData x) ]) 
                  where

		  	isStructuredBackward (CfgEdgeData x _ _) = isStructuredBackwardHelp x

			isStructuredBackwardHelp (StructuredBackward _)		 	= True
			isStructuredBackwardHelp _					= False
		       
		        getLoopID :: (EdgeInfo CfgEdgeData) -> LoopID
			getLoopID x = getLoopIDHelp (edgeData x)
                        
			getLoopIDHelp :: CfgEdgeData -> LoopID
			getLoopIDHelp (CfgEdgeData (StructuredBackward id) _ _)  = id
			getLoopIDHelp otherwise                                = error "keine Schleife gefunden!"

getIntOfLoopID :: LoopID -> Integer
getIntOfLoopID (LoopID id) = id

getMaybeOfEdgeData :: CfgEdgeData -> (Maybe StatementID)
getMaybeOfEdgeData (CfgEdgeData a b c) = b



-----------------------
-- Reading a CfGraph --
-----------------------

-- |
-- parses tree-file from filepath
readCfGraph :: FilePath -> IO (LPResult CfGraph)
readCfGraph filename = loadAndParse filename parseCfg


-- |
-- parses Tree from given String	    
parseCfg :: Parser (CfGraph)
parseCfg = do nodeList <- many parseNode
	      edgeList <- many parseEdge
	      return (AGraph nodeList edgeList)

-- |
-- parses node of form: node <id> <type> // <origin>
parseNode :: Parser (NodeInfo CfgNodeData)
parseNode = do
	    match "node"
	    spaces
	    id <- parseInteger
	    nodedata <- parseNodeData
	    rest <- many (notFollowedByString "edge"
			  >>notFollowedByString "node"
			  >>anyChar)
	    return (NodeInfo (NodeID id) nodedata)

-- |
-- parses nodedata
-- commented parts are better, require changes on c side
parseNodeData :: Parser (CfgNodeData)
parseNodeData = parseWord>>=(\keyword -> case keyword of
				"mergeloop" -> return (CfgNodeData LoopHead)
				"fork"      -> return (CfgNodeData Fork)
				--"join"      -> return (CfgNodeData Join)
				_           -> parsePossibleJoin) 
				               --return (CfgNodeData Standard))

-- |
-- parses <origin> to detect join-node
-- unnecessary when parseNodeData changed
parsePossibleJoin :: Parser (CfgNodeData)
parsePossibleJoin = (try (match "// IF-Last")>>return (CfgNodeData Join))
		    <|>return (CfgNodeData Standard)

-- |
-- parses edge in form: 
-- edge <ID>: <nodeID source> -> <nodeID destination> <edgetype> // <somedata>
-- <relation>
-- statement <statementID>
parseEdge :: Parser (EdgeInfo CfgEdgeData)
parseEdge = do
	    match "edge"
	    spaces
	    edgeId <- parseInteger
	    match ": "
	    nodeId1 <- parseInteger
	    match "-> "
	    nodeId2 <- parseInteger
	    edgedata <- parseEdgeData
	    rest <- many (notFollowedByString "edge">>anyChar)
	    return (EdgeInfo (EdgeID edgeId) edgedata (NodeID nodeId1)
		                                      (NodeID nodeId2))

-- |
-- parses edgedata, detectes edgeType
parseEdgeData :: Parser (CfgEdgeData)
parseEdgeData = do 
		edgetype <- (manyTill anyChar space)
		manyTill anyChar (match "// ")
		name <- many (notFollowedByString "relation">>anyChar)
		manyTill anyChar space
		relation <- manyTill anyChar (match "\nstatement ")
		id <- parseInteger
		rest <- many (notFollowedByString "edge">>anyChar)
		case edgetype of
		    "Structured"   -> let loopNr = integerToID((read name)::Integer)
				      in return 
					     (CfgEdgeData (StructuredBackward loopNr) Nothing relation)
		    "Unstructured" -> return 
				      (CfgEdgeData UnstructuredBackward Nothing relation)
		    _ -> case id of
			     0 -> return (CfgEdgeData Forward Nothing relation)
			     _ -> return (CfgEdgeData BodyStatement
					           (Just (StatementID id)) relation)


--utilitites, maybe move them to LP module

-- |
-- parses one Word, until closing space is found
-- also consumes that space character (not returned)
parseWord :: Parser String
parseWord = manyTill anyChar space

-- |
-- parses characters until given keyword is found-
-- does NOT consume that word
parseUntilString :: String -> Parser String
parseUntilString st = many (notFollowedByString st>>anyChar)

-- |
-- succeeds if input does not begin with given string
-- never consumes any input
notFollowedByString :: String -> Parser ()
notFollowedByString st = try (do{ c <- try (string st); unexpected (show [c]) }
			      <|> return ())

