module LoopoDS.Dependence
    ( DependenceType(..), GDependence(..), Dependence, GDependenceD(..), DependenceD, Context(..),
      DependenceID(..),
      readDependences, writeDependences
    )
where

import LMath
import ParseBase

import LoopoDS.IntegralID
import LoopoDS.Statement

---
-- The type of a dependence.
--
data DependenceType = UniformTrue
		    | UniformFlow
		    | UniformAnti
		    | UniformOutput
		    | UniformInput
		    | UniformDummy
		    | NonUniformTrue
		    | NonUniformFlow
		    | NonUniformAnti
		    | NonUniformOutput
		    | NonUniformInput
		    | NonUniformDummy
		      deriving (Eq, Show)

---
-- For converting to and from integer values.
--
instance Enum DependenceType where
    ---
    -- Converts an integer to the corresponding type of dependence.
    -- @param intval    -  the integer value of the dependece type.
    -- @return deptype  -  the dependence type.
    --
    toEnum  1 =    UniformTrue
    toEnum  2 =    UniformFlow
    toEnum  3 =    UniformAnti
    toEnum  4 =    UniformOutput
    toEnum  5 =    UniformInput
    toEnum  6 =    UniformDummy
    toEnum  7 = NonUniformTrue
    toEnum  8 = NonUniformFlow
    toEnum  9 = NonUniformAnti
    toEnum 10 = NonUniformOutput
    toEnum 11 = NonUniformInput
    toEnum 12 = NonUniformDummy

    --- 
    -- Converts the dependence type to an integer.
    -- @param deptype  -  the dependence type.
    -- @return intval  -  the corresponding integer value.
    --
    fromEnum    UniformTrue   =  1
    fromEnum    UniformFlow   =  2
    fromEnum    UniformAnti   =  3
    fromEnum    UniformOutput =  4
    fromEnum    UniformInput  =  5
    fromEnum    UniformDummy  =  6
    fromEnum NonUniformTrue   =  7
    fromEnum NonUniformFlow   =  8
    fromEnum NonUniformAnti   =  9
    fromEnum NonUniformOutput = 10
    fromEnum NonUniformInput  = 11
    fromEnum NonUniformDummy  = 12


{-
    number the pseudo-parameter that will be saved right now 
    stride 
    denominator of the stride 
    # of inequalities 
    # of variables 
    # of parameters 
    denominator of coefficient-matrix 
    coefficient-matrix /* REMEMBER the structure described above-*/ 
-}

---
-- Describes the context of a dependence.
--
data Context = Context { parN :: Integer,
			 stride :: Rational,
			 parDescription :: IneqSystem }
	       deriving Show


{-
(dependence 1:) 
    number of dependence 
    type of dependence (see below) 
    source description (rows, denominator, matrix) 
    destination description (rows, denominator, matrix) 
    number of source statement 
    number of target statement 
    h-fransformation (rows, denominator, matrix) 
    context 
-}

newtype DependenceID = DependenceID Integer
    deriving (Show, Eq, Ord)

---
-- Describes a dependence.
--
data GDependence a = Dependence { depID :: DependenceID,
				  depType :: DependenceType,
				  depSrcDomain  :: GIneqSystem a,
				  depDestDomain :: GIneqSystem a,
				  depSrcStmtID  :: StatementID,
				  depDestStmtID :: StatementID,
				  depHTrafo :: GAffineFunction a,
				  depContext :: [Context] }
		     deriving Show
type Dependence = GDependence Rational

data GDependenceD d t = DependenceD { depDID :: DependenceID,
				  depDType :: DependenceType,
				  depDSrcDomain  :: d t,
				  depDDestDomain :: d t,
				  depDSrcStmtID  :: StatementID,
				  depDDestStmtID :: StatementID,
				  depDHTrafo :: GAffineFunction t,
				  depDContext :: [Context] }
		     deriving Show
type DependenceD = GDependenceD Domain (Polynomial Rational)

---
-- Reads the dependeces from a file. The file <code>filename</code>
-- is read from disk, run through a parser for the ".dep" file
-- format, and the result is returned.
-- @param filename - the file to read from.
-- @monadic res    - the parser's result.
--
readDependences :: FilePath -> IO (LPResult [Dependence])
readDependences filename =
    do contents <- readFile filename
       return $ runLP parseDependences contents

parseDependences :: Parser [Dependence]
parseDependences =
    do count  <- parseInteger
       vars   <- parseInteger
       params <- parseInteger
       pseudo <- parseInteger
       parseN count (parseDep vars params pseudo)

writeDependences :: FilePath -> [Dependence] -> IO ()
writeDependences filename deps = writeFile filename (dependencesToString deps)

dependencesToString :: [Dependence] -> String
dependencesToString deps = show (length deps) ++ "\n" ++ (show vars) ++ "\n"
    ++ (show params) ++ "\n" ++ (show pseudo) ++ "\n" ++ (concat . map dependenceToString) deps where
        (vars, params) = getVarsParams deps
        pseudo         = getPseudo deps
        getVarsParams []      = (0, 0)
        getVarsParams (dep:_) = ((nVars . toAffineExp . depSrcDomain) dep, (nParams . toAffineExp . depSrcDomain) dep)
        getPseudo []      = 0
        getPseudo (dep:_) = length (depContext dep)

dependenceToString :: Dependence -> String
dependenceToString dep = (show . integerFromID . depID) dep ++ "\n" ++ (show . fromEnum . depType) dep ++ "\n"
    ++ (ineqSystemToString False True . depSrcDomain) dep
    ++ (ineqSystemToString False True . depDestDomain) dep
    ++ (show . integerFromID . depSrcStmtID) dep ++ "\n"
    ++ (show . integerFromID . depDestStmtID) dep ++ "\n"
    ++ (affineFunctionToString False True . depHTrafo) dep
    ++ (foldl (++) "" . map contextToString . depContext) dep

contextToString :: Context -> String
contextToString context = (show . parN) context ++ "\n"
    ++ (show . numerator . stride) context ++ "\n"
    ++ (show . denominator . stride) context ++ "\n"
    ++ isNew
    where
    -- ineqSystemToString puts the number of inequalities at the wrong
    -- place, so we have to move it
    isOld  = (ineqSystemToString True True . parDescription) context
    (a, b) = splitAt 2 $ lines isOld
    isNew  = unlines $ [head b] ++ a ++ tail b

parseContext :: Parser Context
parseContext =
    do parNr  <- parseInteger
       stride <- parseRational
       ineqs  <- parseInteger
       vars   <- parseInteger
       params <- parseInteger
       matr   <- parseMatrix ineqs (vars+params+1)
       return Context { parN = parNr, stride=stride,
			 parDescription = ineqSystem ((vars, params), matr) }

parseDep :: Integer -> Integer -> Integer -> Parser Dependence
parseDep vars params pseudo =
    do num <- parseInteger
       ty  <- parseInteger
       r   <- parseInteger
       src <- parseMatrix r (vars+params+1)

       r    <- parseInteger
       dest <- parseMatrix r (vars+params+1)
       
       srcNum  <- parseInteger
       destNum <- parseInteger
       
       r   <- parseInteger
       hTr <- parseMatrix r (vars+params+1) 

       ct <- parseN pseudo parseContext

       return Dependence { depID = DependenceID num,
			   depType = toEnum (fromInteger ty),
			   depSrcDomain  = ineqSystem ((vars, params), src),
			   depDestDomain = ineqSystem ((vars, params), dest),
			   depSrcStmtID  = integerToID srcNum,
			   depDestStmtID = integerToID destNum,
			   depHTrafo = affineFunction ((vars, params), hTr)
			               Homogeneous,
			   depContext = ct }


instance HasIntegralID DependenceID where
    integerFromID (DependenceID i) = fromIntegral i
    integerToID = DependenceID . fromIntegral
