-- |
-- Functions to parse and format LooPo's Quasts.
--
module LoopoDS.Quast
    ( parseQuast, parseCommentQuast,
      formatQuast
    )
where

import LMath.Types
import LP

import Control.Monad
import Data.List
import Data.Maybe
import Data.Ratio
import Text.ParserCombinators.Parsec


-- |
-- Parse a quast which begins with a comment:
--
-- >   ( ( some comment ) quast )
--
parseCommentQuast :: Integral a => Parser (GQuast GVector (Ratio a) a)
parseCommentQuast = do char '('
		       spaces
		       char '('
		       skipMany (satisfy (/=')'))
                       char ')'
                       spaces
                       q <- parseQuast
		       spaces
		       char ')'
                       spaces
                       return q
                   
-- |
-- Parse a quast (inner part, without comment).
--
parseQuast :: Integral a => Parser (GQuast GVector (Ratio a) a)
parseQuast =
    do newparams <- parseNewParams
       spaces
       char '('
       spaces
       q <- do (string "list" >> spaces >> qmatrix)
	       <|>
	       (string "if" >> spaces >> qif)
	       <|>
	       return QNoResult
       spaces
       char ')'
       spaces
       return $ newparams q

parseNewParams :: Integral a => Parser (GQuast qvec (Ratio a) a -> GQuast qvec (Ratio a) a)
parseNewParams =
    try (do char '('
            spaces
            string "newparam"
            spaces
            n <- parseInteger
            char '('
            spaces
            string "div"
            spaces
            def <- qvector
            denom <- parseInteger
            char ')'
            spaces
            char ')'
            rec <- parseNewParams
            return (QNewParam (fromIntegral n) (integralVector1 def) (fromIntegral denom) . rec)
        )
    <|>
    return id

-- |
-- Parse a matrix in quast format.
--
qmatrix :: Integral a => Parser (GQuast GVector (Ratio a) a)
qmatrix = do vs <- many (qvector >>= \v -> spaces >> return v)
	     return $ QList vs

-- |
-- Parse an IF in quast format.
--
qif :: Integral a => Parser (GQuast GVector (Ratio a) a)
qif = do v  <- qvector
	 spaces
	 q1 <- parseQuast
	 spaces
	 q2 <- parseQuast
	 return $ QIf (integralVector1 v) q1 q2

-- |
-- Parse a vector in quast format.
--
qvector :: Integral a => Parser (GVector (Ratio a))
qvector = do string "#["
	     spaces
	     cs <- many $ do r <- parseFraction
			     spaces
			     return r
	     char ']'
             spaces
             return $ vector cs


-- |
-- Format a quast in LooPo's format (inner part, without comment).
--
formatQuast :: Integral a => GQuast GVector (Ratio a) a -> String
formatQuast QNoResult = "()"
formatQuast (QList rs) =
    "( list " ++
    concat (intersperse " " (map formatVector rs)) ++
    " )"
formatQuast (QIf c q1 q2) =
    concat ["( if ", formatVector (fmap fromIntegral c), " ",
	    formatQuast q1, " ", formatQuast q2, " )"]
formatQuast (QNewParam n def denom q) =
    "( newparam " ++ show n ++ " ( div " ++ formatVector (fmap fromIntegral def) ++ " " ++
    show denom ++ " )) " ++ formatQuast q

-- |
-- Format a vector in LooPo's format used in quasts.
--
formatVector :: Integral a => GVector (Ratio a) -> String
formatVector v =
    "#[" ++ concat (intersperse " " (map formatRational $ vectorEntries v)) ++ "]"

-- |
-- Format a rational in LooPo's format used in quasts.
--
formatRational :: Integral a => Ratio a -> String
formatRational r
    | denominator r == 1  =  show (numerator r)
    | otherwise           =  show (numerator r) ++ "/" ++ show (denominator r)
