-- |
-- Functions to parse and format LooPo's schedules and placements.
--
module LoopoDS.SchedPlace
    ( Schedules(..), Schedule,
      Placements(..), Placement,
      formatStmtSchedule,
      formatStmtPlacement, formatDataPlacement,
      readSchedule, readPlacement
    )
where

import LoopoDS.IntegralID
import LoopoDS.Access
import LoopoDS.Quast
import LoopoDS.Statement
import LMath.Types
import ParseBase

import Control.Monad
import qualified Data.Map as Map
import Data.Map ( Map )
import Data.Ratio
import Text.ParserCombinators.Parsec

type Schedule  = Quast
type Placement = Quast

newtype          Schedules    = Schedules  (Map StatementID Schedule )
                                deriving Show

newtype Ord a => Placements a = Placements (Map a           Placement)
                                deriving Show

-- |
-- Reads the schedule from a file. The file @filename@
-- is read from disk, run through a parser for the ".sched" file
-- format, and the result is returned.
--
readSchedule :: FilePath             -- ^ filename
	      -> IO (LPResult (Integer, Integer, Schedules))
readSchedule filename =
    do contents <- readFile filename
       return $ runLP parseSchedule contents

-- |
-- Reads the placement from a file. The file @filename@
-- is read from disk, run through a parser for the ".place" file
-- format, and the result is returned.
--
readPlacement :: FilePath             -- ^ filename
	      -> IO (LPResult (Integer, Integer, Placements StatementID))
readPlacement filename =
    do contents <- readFile filename
       return $ runLP parseStmtPlacement contents

-- |
-- Parse a statement placement.
-- Returns the number of variables, the number of parameters
-- and the placement of the statements.
--
parseStmtPlacement :: Parser (Integer, Integer, Placements StatementID)
parseStmtPlacement = liftM (\(nV,nP,theMap) -> (nV,nP,Placements theMap))
		           parseSchedOrStmtPlm

-- |
-- Parse a schedule.
-- Returns the number of variables, the number of parameters
-- and the schedule.
--
parseSchedule :: Parser (Integer, Integer, Schedules)
parseSchedule = liftM (\(nV,nP,theMap) -> (nV,nP,Schedules theMap))
		      parseSchedOrStmtPlm

-- |
-- Parse a schedule or statement placement
-- (they have the same format).
--
parseSchedOrStmtPlm :: (HasIntegralID a, Ord a)
		    => Parser (Integer, Integer, Map a Quast)
parseSchedOrStmtPlm =
    do nStmts <- parseInteger
       nV     <- parseInteger
       nP     <- parseInteger
       pl <- many $ do stmt  <- parseInteger
		       quast <- parseCommentQuast
		       char '#'
		       spaces
		       return (integerToID stmt,quast)
       return (nV, nP, Map.fromList pl)



-- |
-- Format a statement placement in LooPo's file format.
--
formatStmtPlacement :: Integer   -- ^ number of loop indices
		 -> Integer   -- ^ number of structural parameters
		 -> Placements StatementID
		 -> String
formatStmtPlacement nLoopIndices nP (Placements alloc) =
    unlines ([show (length stmtIDs), show nLoopIndices, show nP] ++
	     concatMap fmt stmtIDs)
    where
    fmt stmtID = [show (integerFromID stmtID),
		  formatPlacement alloc stmtID ("FCO placement for " ++
					     show stmtID),
		  "#"
		 ]
    stmtIDs = Map.keys alloc

-- |
-- Format a data placement in LooPo's file format.
--
formatDataPlacement :: Integer   -- ^ total number of arrays
		 -> Integer   -- ^ number of loop indices
		 -> Integer   -- ^ number of structural parameters
		 -> Placements ArrayID
		 -> String
formatDataPlacement nArraysTotal nLoopIndices nP (Placements alloc) =
    unlines ([show nArraysTotal, show (length arrayIDs), show nLoopIndices,
	      show nP] ++
	     concatMap fmt arrayIDs
	    )
    where
    fmt aID = [show (integerFromID aID),
	       formatPlacement alloc aID ("FCO placement for " ++ show aID),
	       "#"
	      ]
    arrayIDs = Map.keys alloc    

-- |
-- Generic placement printer.
--
formatPlacement :: (Ord a, Show a)
	     => Map a Quast  -- ^ placement
	     -> a            -- ^ particular ID to format
	     -> String       -- ^ comment field
	     -> String
formatPlacement alloc pid comment =
    "( ( " ++ comment ++ " )\n" ++
    formatQuast (Map.findWithDefault errmsg pid alloc)
    ++ "\n)"
    where
    errmsg = error ("formatPlacement: looking up " ++ show pid ++
		    " faild (not in placement)")



-- |
-- Format a statement schedule in LooPo's file format.
--
formatStmtSchedule :: Integer   -- ^ number of loop indices
		 -> Integer   -- ^ number of structural parameters
		 -> Schedules
		 -> String
formatStmtSchedule nLoopIndices nP (Schedules alloc) =
    unlines ([show (length stmtIDs), show nLoopIndices, show nP] ++
	     concatMap fmt stmtIDs)
    where
    fmt stmtID = [show (integerFromID stmtID),
		  formatSchedule alloc stmtID ("Schedule for " ++
					     show stmtID),
		  "#"
		 ]
    stmtIDs = Map.keys alloc


-- |
-- Generic schedule printer.
--
formatSchedule :: Map StatementID Quast  -- ^ schedule
	     -> StatementID              -- ^ particular ID to format
	     -> String                   -- ^ comment field
	     -> String
formatSchedule alloc pid comment =
    "( ( " ++ comment ++ " )\n" ++
    formatQuast (Map.findWithDefault errmsg pid alloc)
    ++ "\n)"
    where
    errmsg = error ("formatSchedule: looking up " ++ show pid ++
		    " faild (not in schedule)")



