module LoopoDS.Statement
    ( Statement, StatementID(..), TStatement, TStatementID(..),
      StatementD, TStatementD, GStatementD, GTStatementD,
      readISPC, readTISPC, readISPCDomain, readTISPCDomain,
      writeISPC,
      makeProperTISPCs
    )
where

import Data.Ix

import LMath
import ParseBase
import LoopoDS.IntegralID

type Statement  = (StatementID, IneqSystem)
type TStatement = (TStatementID, StatementID, IneqSystem)

type GStatementD d t = (StatementID, d t)
type GTStatementD d t = (TStatementID, StatementID, d t)

type StatementD  = GStatementD Domain (Polynomial (Ratio Integer))
type TStatementD = GTStatementD Domain (Polynomial (Ratio Integer))

newtype StatementID = StatementID Integer
    deriving (Read, Eq, Ord, Enum, Ix)

instance Show StatementID where
    show (StatementID s) = "S" ++ show(s)

--type TStatementID = Integer
newtype TStatementID = TStatementID Integer
    deriving (Show, Read, Eq, Ord, Enum, Ix)

instance HasIntegralID StatementID where
    integerFromID (StatementID i) = fromIntegral i
    integerToID = StatementID . fromIntegral

instance HasIntegralID TStatementID where
    integerFromID (TStatementID i) = fromIntegral i
    integerToID = TStatementID . fromIntegral

---
-- Reads an index space file.
-- When the given file is parsed correctly, a list of tuples
-- <code>(stmtNr, ispc)</code> representing the number of
-- the statement (<code>stmtNr</code>) and the
-- index space (<code>ispc</code>) is returned.
-- @param filename - the file to read.
-- @monadic r - <code>ParseOk </code> when parsing
--            is successsful, <code>ParseFailed </code>
--            otherwise.
--
readISPC :: FilePath -> IO (LPResult [Statement])
readISPC filename = loadAndParse filename parseISPC
    where
    parseISPC :: Parser [(StatementID, IneqSystem)]
    parseISPC =
	do numberOfStmts <- parseInteger
	   parseN numberOfStmts parseIspcStmt 

    parseIspcStmt :: Parser (StatementID, IneqSystem)
    parseIspcStmt =
	do stmtNr <- parseInteger
	   rows   <- parseInteger
	   vars   <- parseInteger
	   params <- parseInteger
	   matr   <- parseMatrix rows (vars+params+1)
	   return (StatementID stmtNr, ineqSystem ((vars, params), matr))

readISPCDomain :: FilePath -> IO (LPResult [StatementD])
readISPCDomain filename = loadAndParse filename parseISPC
    where
    parseISPC :: Parser [(StatementID, Domain (Polynomial (Ratio Integer)))]
    parseISPC =
	do numberOfStmts <- parseInteger
	   parseN numberOfStmts parseIspcStmt 

    parseIspcStmt :: Parser (StatementID, Domain (Polynomial (Ratio Integer)))
    parseIspcStmt =
	do stmtNr <- parseInteger
	   rows   <- parseInteger
	   vars   <- parseInteger
	   params <- parseInteger
	   matr   <- parseMatrix rows (vars+params+1)
	   let ineq = ineqSystem ((vars, params), matr)
	   return (StatementID stmtNr, ineqToDomain2 Nothing (vars, params) ineq)

---
-- Reads a transformed index space file.
-- When the given file is parsed correctly, a list of tuples
-- <code>(stmtNr, origStmtNr, tispc)</code> representing the number of
-- the target statement (<code>stmtNr</code>), the number of the
-- original statement (<code>origStmtNr</code>), and the
-- transformed index space (<code>tispc</code>) is returned.
-- @param filename - the file to read.
-- @monadic - <code>ParseOk </code> when parsing
--            is successsful, <code>ParseFailed </code>
--            otherwise.
--
readTISPC :: FilePath -> IO (LPResult [TStatement])
readTISPC filename = loadAndParse filename parseTISPC
    where
    parseTISPC :: Parser [TStatement]
    parseTISPC =
	do numberOfStmts <- parseInteger
	   parseN numberOfStmts parseTIspcStmt 

    parseTIspcStmt :: Parser (TStatementID, StatementID, IneqSystem)
    parseTIspcStmt =
	do stmtNr     <- parseInteger
	   origStmtNr <- parseInteger
	   vars       <- parseInteger
	   params     <- parseInteger
	   rows       <- parseInteger
	   matr       <- parseMatrix rows (vars+params+1)
	   return (integerToID stmtNr, integerToID origStmtNr,
		   ineqSystem ((vars, params), matr)
		  )


readTISPCDomain :: FilePath -> IO (LPResult [TStatementD])
readTISPCDomain filename = loadAndParse filename parseTISPC
    where
    parseTISPC :: Parser [TStatementD]
    parseTISPC =
	do numberOfStmts <- parseInteger
	   parseN numberOfStmts parseTIspcStmt

    parseTIspcStmt :: Parser (TStatementID, StatementID, Domain (Polynomial (Ratio Integer)))
    parseTIspcStmt =
	do stmtNr     <- parseInteger
	   origStmtNr <- parseInteger
	   vars       <- parseInteger
	   params     <- parseInteger
	   rows       <- parseInteger
	   matr       <- parseMatrix rows (vars+params+1)
	   let ineq   =  ineqSystem ((vars, params), matr)
	   return (integerToID stmtNr, integerToID origStmtNr,
		   ineqToDomain2 Nothing (vars, params) ineq
		  )

writeISPC :: [Statement] -> FilePath -> IO ()
writeISPC statements filename = writeFile filename (statementsToString statements)

statementsToString :: [Statement] -> String
statementsToString statements = (show . length) statements ++ "\n" ++ (concat . map statementToString) statements

statementToString :: Statement -> String
statementToString (id, ineqSystem) = (show . integerFromID) id ++ "\n" ++ ineqSystemToString' True True ineqSystem



---
-- Use thish function to transform ISPCSs into correct TISPCs
-- Attention: uses TISPC-file *only* for mapping TStatementID -> StatementID
-- 
makeProperTISPCs :: [Statement] 
                 -- ^ trimmed ispcs
                 -> [TStatement]
                 -- ^ tispcs, only used for mapping IDs
                 -> [(TStatementID, AffineFunction)]
                 -- ^ proper trafos
                 -> [TStatement]
makeProperTISPCs trimmedIspcs tispcs properTrafos = concatMap mkTISPC trimmedIspcs
    where
    mkTISPC :: Statement -> [TStatement]
    mkTISPC (stmtID, ispc) =
        [(tStmtID, stmtID, image trafo ispc)
         | (tStmtID, trafo) <- properTrafos,
           (tStmtID', stmtID',_)  <- tispcs, 
           tStmtID' == tStmtID,
           stmtID' == stmtID
        ]
    
