module LoopoDS.SurroundingIfs
    ( SurroundingIfs(..),
      readSIF, writeSIF,
      lookupSurroundingIfs,
      determineSurroundingIfs
    )
where

import Data.List

import LMath
import ParseBase
import LoopoDS.IntegralID
import LoopoDS.Access
import LoopoDS.Dependence
import LoopoDS.Statement
import ScannerParser.ParseTree


-- |
-- Reads the file .sif containing the surrounding ifs.
-- When the given file is parsed correctly, a list of tuples 
-- @(stmtNr, ifss)@ is returned. 
-- Every pair consists of the the number of the statement
-- (@stmtNr@) and the boolean vector -
-- (@slps@) which tells for every loop, 
-- whether it surrounds statement @stmtNr@.
--
readSIF :: FilePath -> IO (LPResult SurroundingIfs)
readSIF filename = loadAndParse filename parseSIF
    where
    parseSIF :: Parser SurroundingIfs
    parseSIF =
	do numberOfStmts <- parseInteger
	   let parseLine :: Parser (StatementID, [StatementID])
               parseLine = do
			   stmt <- parseInteger
			   nsurr <- parseInteger
			   surr <- parseN nsurr parseInteger
			   return $ (integerToID stmt,map integerToID surr)
           
           thens <- parseN numberOfStmts parseLine
           elses <- parseN numberOfStmts parseLine

           let 
               -- | 
               -- assumes that every key appears at most once in each list!
               -- With this restriction it combines the two input maps 
	       merge :: Eq a => [(a,[b])] -> [(a,[c])] -> [(a,([b],[c]))]
	       merge bMap [] =  map (\(a,bs) -> (a,(bs,[]))) bMap
	       merge bMap ((k,vs):xs) = 
		   (case lookup k bMap of Just b -> (k,(b,vs))
					  Nothing -> (k,([],vs))  )
		   -- : merge bMap xs, but without the b just inserted, i.e.
		   : merge (deleteBy (\(k1,_)-> \(k2,_)->k1==k2) (k,[]) bMap) 
		           xs
	       sifs = merge thens elses
           return $ SIF sifs

writeSIF :: SurroundingIfs -> FilePath -> IO ()
writeSIF sifs filename = writeFile filename (surroundingIfsToString sifs)

surroundingIfsToString :: SurroundingIfs -> String
surroundingIfsToString (SIF sifs) = show (length sifs) ++ "\n"
    ++ blockToString ifBlock ++ "\n" ++ blockToString elseBlock where
        ifBlock   = map (\x -> (fst x, fst (snd x))) sifs
        elseBlock = map (\x -> (fst x, snd (snd x))) sifs
        blockToString :: [(StatementID, [StatementID])] -> String
        blockToString block = (concat . map lineToString) block
        lineToString :: (StatementID, [StatementID]) -> String
        lineToString (id, surroundedBy) = (show . integerFromID) id ++ "     "
                                          ++ (show . length) surroundedBy ++ " "
                                          ++ (concat . intersperse " "
                                              . map (show . integerFromID))
                                             surroundedBy
                                          ++ "\n"

-- |
-- Consists of a list of triples, for representing then and else branches
-- surrounding a statement. 
-- The first entry represents any statement @s@,
-- the second and the third entry are lists representing all "then" and "else"
-- branches in which statement @s@ is included, resprectively.
--
newtype SurroundingIfs = SIF [(StatementID, ([StatementID], [StatementID]))]
    deriving (Show)

determineSurroundingIfs :: LambdaTree -> SurroundingIfs
determineSurroundingIfs = SIF . determineSurroundingIfsNodes [] [] where
    
    determineSurroundingIfsNode :: [StatementID] -> [StatementID] -> LambdaNode
        -> [(StatementID, ([StatementID], [StatementID]))]
    determineSurroundingIfsNode ifs elses node@(ForLoop {}) =
        determineSurroundingIfsNodes ifs elses (loopBody node)
    determineSurroundingIfsNode ifs elses node@(WhileLoop {}) =
        determineSurroundingIfsNodes ifs elses (loopBody node)
    determineSurroundingIfsNode ifs elses node@(FuncDecl {}) =
        determineSurroundingIfsNodes ifs elses (funcbody node)
    determineSurroundingIfsNode ifs elses node@(If {}) =
        determineSurroundingIfsNodes (statement node:ifs) elses
            (thenbranch node)
        ++ determineSurroundingIfsNodes ifs (statement node:elses)
            (elsebranch node)
    determineSurroundingIfsNode ifs elses node =
        if hasstatement node then [(statement node, (ifs, elses))] else []
    
    determineSurroundingIfsNodes :: [StatementID] -> [StatementID]
        -> [LambdaNode] -> [(StatementID, ([StatementID], [StatementID]))]
    determineSurroundingIfsNodes ifs elses =
        concat . map (determineSurroundingIfsNode ifs elses)


lookupSurroundingIfs :: SurroundingIfs -> StatementID -> ([StatementID],[StatementID])
lookupSurroundingIfs (SIF sifs) stmtID
    | Just stmts <- lookup stmtID sifs  =  stmts
    | otherwise  =  error ("lookupSIF: statement ID " ++ show stmtID ++
			   "not found")

test :: FilePath -> IO()
test fp = do 
	    sif <- readSIF fp
	    putStrLn $ show sif

