module LoopoDS.SurroundingLoops
    ( SurroundingLoops(..),
      readSLP, writeSLP,
      lookupSurroundingLoops,
      trimToSurroundingLoops,
      trimIndexSpace,
      trimAccess, trimRawAccess,
      trimDependence, trimSchedules, trimPlacements, 
      LoopID(..)
    )
where

import Control.Monad (liftM2)
import Data.List
import Data.Ix

import LMath
import ParseBase
import LoopoDS.IntegralID
import LoopoDS.Access
import LoopoDS.Dependence
import LoopoDS.Statement
import LoopoDS.SchedPlace
import qualified Data.Map as Map
import Data.Map ( Map )
import LoopoDS.Quast


--type LoopID = Integer
newtype LoopID = LoopID Integer
    deriving (Show, Read, Eq, Ord, Enum, Ix)

instance HasIntegralID LoopID where
    integerFromID (LoopID i) = fromIntegral i
    integerToID = LoopID . fromIntegral



-- |
-- Reads the file .slp containing the surrounding loops.
-- When the given file is parsed correctly, a list of tuples 
-- @(stmtNr, slps)@ is returned. 
-- Every pair consists of the the number of the statement
-- (@stmtNr@) and the boolean vector -
-- (@slps@) which tells for every loop, 
-- whether it surrounds statement @stmtNr@.
--
readSLP :: FilePath -> IO (LPResult SurroundingLoops)
readSLP filename = loadAndParse filename parseSLP
    where
    parseSLP :: Parser SurroundingLoops
    parseSLP =
	do numberOfStmts <- parseInteger
           numberOfVars <- parseInteger
           slps <- parseN numberOfStmts $
		          liftM2 (\a b -> (integerToID a, map (/=0) b))
				 parseInteger
			         (parseIntegers numberOfVars)
           return $ SLP slps

newtype SurroundingLoops = SLP [(StatementID, [Bool])]
    deriving (Show)

writeSLP :: SurroundingLoops -> FilePath -> IO ()
writeSLP slps filename = writeFile filename (surroundingLoopsToString slps)

surroundingLoopsToString :: SurroundingLoops -> String
surroundingLoopsToString (SLP slps) = show (length slps) ++ "\n"
    ++ show (getNumberOfLoops slps) ++ "\n"
    ++ (concat . map surroundingLoopsStatementToString) slps where
        getNumberOfLoops []            = 0
        getNumberOfLoops ((_,loops):_) = length loops

surroundingLoopsStatementToString :: (StatementID, [Bool]) -> String
surroundingLoopsStatementToString (id, loops) = show (integerFromID id) ++ " " ++ (concat . intersperse " " . map show . map fromEnum) loops ++ "\n"

lookupSurroundingLoops :: SurroundingLoops -> StatementID -> [Bool]
lookupSurroundingLoops (SLP slps) stmtID
    | Just loops <- lookup stmtID slps  =  loops
    | otherwise  =  error ("lookupSLP: statement ID " ++ show stmtID ++
			   "not found")

-- |
-- Retain only the /columns/ of an 'AffineExpBased' object
-- (i.e., 'GAffineFunction', 'GIneqSystem') which correspond
-- to the dimensions surrounding the given statement.
trimToSurroundingLoops :: AffineExpBased aeb a
		       => SurroundingLoops
		       -> StatementID
		       -> aeb
		       -> aeb
trimToSurroundingLoops slp stmtNr aeb =
    fromAffineExp $ affineExpVPC (keepColumns v loops, p, c)
    where
    (v,p,c) = partsVPC $ toAffineExp aeb
    loops = elemIndices True $ lookupSurroundingLoops slp stmtNr

trimIndexSpace :: SurroundingLoops -> Statement -> Statement
trimIndexSpace slp (stmtNr,ineqs) =
    (stmtNr, trimToSurroundingLoops slp stmtNr ineqs)

trimRawAccess :: SurroundingLoops -> RawAccess -> RawAccess
trimRawAccess slp (arrayId, accType, laNode, stmtNr, indexFn) =
    (arrayId, accType, laNode, stmtNr,
     trimToSurroundingLoops slp (integerToID stmtNr) indexFn
    )

trimAccess :: SurroundingLoops -> Access -> Access
trimAccess slp acc =
    acc { accIndexFct = tr $ accIndexFct acc,
	  accDomain   = tr $ accDomain acc
	}
    where
    tr :: AffineExpBased aeb a => aeb -> aeb
    tr = trimToSurroundingLoops slp (accStmtID acc)

trimSchedules :: SurroundingLoops -> Schedules -> Schedules
trimSchedules slp (Schedules sched) = Schedules (Map.mapWithKey (trimSchedPlace slp) sched) where

trimPlacements :: SurroundingLoops -> Placements StatementID -> Placements StatementID
trimPlacements slp (Placements place) = Placements (Map.mapWithKey (trimSchedPlace slp) place)

-- | Hilfsfunktion fr trimPlacements, trimSchedules
trimSchedPlace :: SurroundingLoops -> StatementID -> GQuast GVector a b -> GQuast GVector a b
trimSchedPlace slp stmtid sch = quastmap (trimInEqVars stmtid) (trimVectVars stmtid) sch where
          trimInEqVars stmtid curm = foldl (\m i -> map (\v -> deleteEntries v i 1) m) curm (delVars stmtid) 

          -- Is that correct? The conditions/new params defs in quasts
          -- are only in the parameters?!
          trimVectVars stmtid curv = foldl (\v i -> deleteEntries v i 1) curv (delVars stmtid) 
          delVars stmt = reverse  $ map fst $ filter (not.snd) $ zip [0..] (lookupSurroundingLoops slp stmt)
          quastmap :: ([GVector a] -> [GVector a]) -> (GVector b -> GVector b) -> GQuast GVector a b -> GQuast GVector a b
          quastmap f1 f2 (QNoResult)     = QNoResult
          quastmap f1 f2 (QList m)       = QList (f1 m)
          quastmap f1 f2 (QIf vec q1 q2) = QIf (f2 vec) (quastmap f1 f2 q1) (quastmap f1 f2 q2)
          quastmap f1 f2 (QNewParam n def denom q) = QNewParam n (f2 def) denom $ quastmap f1 f2 q

trimDependence :: Num a
               => SurroundingLoops -> GDependence a -> GDependence a
trimDependence slp dep =
    dep { depSrcDomain  = ineqSystemVPC (keepColumns sv sloops, sp, sc),
	  depDestDomain = ineqSystemVPC (keepColumns dv dloops, dp, dc),
	  depHTrafo     = affineFunctionVPC (hv', hp', hc') Inhomogeneous
	}
    where
    (sv, sp, sc) = partsVPC $ colFilter "source" sloops $
		   depSrcDomain  dep
    (dv, dp, dc) = partsVPC $ colFilter "destination" dloops $
		   depDestDomain dep
    (hv, hp, hc) = partsVPCih $ depHTrafo dep
    hv' = keepRows (keepColumns hv dloops) sloops
    hp' = keepRows hp sloops 
    hc' = keepRows hc sloops

    sloops = elemIndices True $ lookupSurroundingLoops slp (depSrcStmtID dep)
    dloops = elemIndices True $ lookupSurroundingLoops slp (depDestStmtID dep)

    -- filter out inequalities which have nonzero coefficients
    -- in non-surrounding variables; these inequalities are nonsense since they
    -- restrict loop indices which do not surround the respective statement
    colFilter :: Num a => String -> [Int] -> GIneqSystem a -> GIneqSystem a
    colFilter what loops ineqs = filterIneqs pred ineqs
	where
	pred coeffs
	    | nonsurroundingNull  =  True
	    | surroundingNonnull  =
		error ("trimDependence: don't know how to trim dependence " ++
		       what ++ " space:\n   there are inequalities where both "
		       ++ "surrounding and non-surrounding\n  variables have "
		       ++ "non-null entries:\n" ++ show dep ++
		       "\nwith surrounding loop indices " ++ show loops ++
		       "\n")
	    | otherwise           =  False
	    where
	    v = nVars   ineqs
	    surroundingNonnull = any (/=0) [coeffs!!l | l <- loops]
	    nonsurroundingNull = all (==0) [coeffs!!l | l <- [0..v-1]\\loops]
