--
-- Version: $Id: TrDep.hs 4184 2009-08-13 15:59:24Z groessli $
--
module LoopoDS.TrDep
where

import Data.List  (genericLength, find, genericTake, intersperse, transpose,
                   (\\), elemIndices, genericIndex)
import Data.Maybe (fromMaybe)
import Debug.Trace
import System.IO.Unsafe

import LMath
import LoopoDS.IntegralID
import LoopoDS.Dependence
import LoopoDS.Statement
import LoopoDS.SurroundingLoops
import LoopoDS.Trafo
import LP
import CodeGen

import Control.Exception

filename = "exs/ludec-feaut/ludec.src.out"

newtype TrDependenceID = TrDependenceID Integer
    deriving (Show, Eq, Ord)

instance HasIntegralID TrDependenceID where
    integerFromID (TrDependenceID i) = fromIntegral i
    integerToID = TrDependenceID . fromIntegral

-- Testdaten
{-
(dep, ispc, trafo, nAllSrcIndices) = (d, i, tr, nas)
    where
    ParseOk d  = unsafePerformIO (readDependences (filename++".dep"))
    ParseOk i  = unsafePerformIO (readISPC (filename++".ispc"))
    ParseOk tr = unsafePerformIO (readTrafo (filename++".trafo"))
    nas = nVars (snd (head i))
-}

-- |
-- Describes a dependence in TargetSpace.
--
-- In contrast to 'LoopoDS.Dependence' the h-transformation is replaced by
-- the dependence relation 'trDepRelation'.
--
data TrDependence = TrDependence { trDepID :: TrDependenceID,
				   trDepOrigID :: DependenceID,
				   trDepType :: DependenceType,
				   trDepOrigSrcStmtID  :: StatementID,
				   trDepOrigDestStmtID :: StatementID,
				   trDepSrcStmtID      :: TStatementID,
				   trDepDestStmtID     :: TStatementID,
				   trDepRelation :: (IneqSystem,Integer),
				   trDepContext :: [Context] }
		    deriving Show

-- 
-- Einf"ugen der linear abh"angigen Zeilen
-- 
addLinDepRows :: (TStatementID,StmtTrafo) -> (TStatementID,AffineFunction)
addLinDepRows (tstmtId,strafo) =
    (tstmtId, affineFunction ((nV, nP), extRows) Homogeneous)
    where
    f        = trafoFunc strafo
    usedVect = trafoUsedVector strafo
    lincomb  = trafoLincombs   strafo

    fmWidth = nColumns f
    extRows = case addRows usedVect lincomb (rows' f) of
	          []  -> zeroMatrix 0 fmWidth
		  xss -> matrix xss

    addRows :: [Bool] -> [[Rational]] -> [[Rational]] -> [[Rational]]
    addRows [] [] rs = rs
    addRows (True:uvs)  lcs (r:rs)  = r : addRows uvs lcs rs
    addRows (False:uvs) (lc:lcs) rs = linCo lc : addRows uvs lcs rs
    addRows _ _ _ = error "Kaputte Rekursion"

    fullWidth = nP+nV+1 
    fullHeight = nTargV+nP+1
    nTargV = genericLength usedVect

    nP     = nParams f
    nV     = nVars f

    linCo :: [Rational] -> [Rational]
    linCo xs = map sum $ transpose $ zipWith (\f r -> map (*f) r) xs (rows' f)



properTrafos :: String -> [(TStatementID,AffineFunction)]
properTrafos filename = map addLinDepRows tr
    where ParseOk tr = unsafePerformIO (readTrafo (filename++".trafo"))

makeProperTrafos :: Trafo -> [(TStatementID,AffineFunction)]
makeProperTrafos tr = map addLinDepRows tr


trafoOf :: [(TStatementID,AffineFunction)] -> TStatementID -> AffineFunction
trafoOf sl i = case find (\(nr,_)->nr==i) sl of
	      Just (_,trafo) -> trafo
	      Nothing -> error ("No trafo of statement " ++
					    show i ++ "\n")


addLinDepZeroCols :: (TStatementID,StmtTrafo) -> (TStatementID,AffineFunction)
addLinDepZeroCols (tstmtId,strafo) = 
    (tstmtId, affineFunction ((nV, nP), matr) Homogeneous)
    where
    inv      = trafoInvFunc strafo
    usedVect = trafoUsedVector strafo
    nrows = nRows inv
    nV = length usedVect
    nP = nParams inv
    matr = matrix (transpose(addCols usedVect  (columns' inv)))

    addCols :: [Bool] ->  [[Rational]] -> [[Rational]]
    addCols [] cs = cs  -- for params and const
    addCols (True:uvs)  (c:cs)  =   c  : addCols uvs cs
    addCols (False:uvs) cs =   zeroCol  : addCols uvs  cs
    zeroCol = replicate (fromInteger nrows) 0 


properInvTrafos :: String -> [(TStatementID,AffineFunction)]
properInvTrafos filename = 
    makeProperInvTrafos tr
    where ParseOk tr    = unsafePerformIO (readTrafo (filename++".trafo"))


makeProperInvTrafos :: Trafo
		    -> [(TStatementID,AffineFunction)]
makeProperInvTrafos tr = map addLinDepZeroCols tr



makeProperLinCombEquations :: Trafo -> [(TStatementID, IneqSystem)]
makeProperLinCombEquations tr = 
    map oneProperLinCombEquation tr
    

-- veraltet:
-- ACHTUNG: der anzugebende Parameter w"ahltt ein Statement aus
--          diese Nummer hat eigentlich nichts mit der StmtId zu tun
--          Die L"osungg ist nur so, weil wir keinen Datentypen f"ur
--          eine einzelne Trafo aufmachen wollten.
oneProperLinCombEquation :: (TStatementID,StmtTrafo) -> (TStatementID,IneqSystem)
oneProperLinCombEquation (tstmtId,strafo) =
    (tstmtId, ineqSystem ((nV, nP), eqM `join` (-eqM)))
    where
    trafo    = trafoFunc strafo
    usedVect = trafoUsedVector strafo
    linCombs = trafoLincombs   strafo
    ulength = length usedVect
    nP = nParams trafo
    nV = ulength
	 
    equationLength = nV + nP + 1

    used = elemIndices True usedVect
    notUsed = elemIndices False usedVect
    numNotUsed = genericLength notUsed



    eqM = 
       if (numNotUsed /= length linCombs) 
       then error "failed Assertion: length lincombs != numNotUsed!"
       else matrix $ map genEquations [0..numNotUsed-1]

    genEquations :: Int -> [Rational]
    genEquations nthNotUsed = equationVars ++ equationParsAndConst
	where
	coeffs = genericIndex linCombs nthNotUsed
	currPos = genericIndex notUsed nthNotUsed
	equationParsAndConst = drop (length used) coeffs
	equationVars = 
	    [if i == currPos
	     then -1%1
	     else if elem i used 
	          -- take the n-th coefficient according to the order
	          -- of appearance in the used vector:
	          then coeffs !! (head $ elemIndices i used) 
	          else 0%1
	     | i <- [0..nV-1]]
	

checkTrafoAndInv :: String -> String
checkTrafoAndInv filename = trace (concat ress) $ concat ress
    where trafos = properTrafos filename
	  invs  = properInvTrafos filename
	  combi = zip3 (map fst trafos) (map snd trafos) (map snd invs)
	  mults = map (\(i,a,b) -> (i,(getMatrix b)*(getMatrix a))) combi
	  ress = map show mults

-- extends a statement index space to the correct dimension in the
-- target space (which is the same as the length of the used-vector)
{- DON'T KNOW HOW TO IMPLEMENT THIS CORRECTLY, AG 2004-10-19
padStmnt :: Trafo -> TStatement -> TStatement
padStmnt trafo (tstmtID,stmtID, idxSpc) =
    -- trace (show usedVect) $
    (tstmtID, stmtID, ineqSystem ((nV,nP),newMatr))
    where
    strafo
	| Just st <- lookup stmtID trafo  =  st
	| otherwise  =  error "padStmnt: statement not found"
    usedVect = trafoUsedVector strafo
    linCombs = trafoLincombs   strafo
--    (_, usedVect, linCombs) = uvectAndLincombs trafo !! (integerFromID stmtNr-1)

    -- total (including linearly dependent vars) number of variables
    nV     = genericLength usedVect
    nP     = nParams idxSpc
    origMatr = getMatrix idxSpc

    newMatr1 = multiconc1 $
	       expand (zeroMatrix (nRows origMatr) 1)
		      usedVect
		      (columns origMatr)
    newMatr  = multijoin newMatr1 linDeps

    --
    -- expand 42 [True,False,False,True] [1, 2, 3]
    --   ==
    -- [1, 42, 42, 2, 3]
    --
    expand :: a -> [Bool] -> [a] -> [a]
    expand _ [] cs = cs
    expand toInsert (True:useds) (c:cs) = c        : expand toInsert useds cs
    expand toInsert (False:useds) cs    = toInsert : expand toInsert useds cs

    -- 
    -- return equationss (as two inequalities) for all the
    -- linearly dependent variables
    --
    linDeps :: [Matrix]
    linDeps = zipWith lindep newVars linCombs
	where
	-- the linearly dependent variables
	newVars = [varNr | (False,varNr) <- zip usedVect [0..]]

    --
    -- return the coefficient matrix of two inequalities representing
    -- the equation 
    --      x_{varNr}  =  <linear combination represented by 'coeffs'>
    --
    lindep :: Integer -> [Rational] -> Matrix
    lindep varNr coeffs = m `join` (-m)
	where
	m = setEntry (rowMatrix (expand 0 usedVect coeffs)) 0 varNr (-1)
-}


type GDepRelation a = (GIneqSystem a, Integer)
type DepRelation = GDepRelation Rational

-- depRelation muss aufgerufen werden im Original-Koordinatensystem
-- und liefert die Umkehrrelation nach dem Space-Time-Mapping
-- (und die Dimensionen, die vom IneqSystem f"ur die Beschreibung
--  des src und f"ur die Beschreibung des dest Raums ben"otigt werden)
--
depRelation :: Dependence -> AffineFunction -> AffineFunction -> DepRelation
depRelation  dep srcMap destMap =
      --200203 trace (concat (intersperse ", " (map show [sDim,dDim,nRows hv, nColumns hv,smDim,dmDim]))) $
    case failedAssertions of
    "" -> 
--	 trace ("depRelation" ++ show iesMatrix ++ show mappedIndices ++ loopNest (projectToVars ies mappedIndices)) $
--	(projectToVars ies mappedIndices, [smDim,dmDim])
	(fromMaybe (error "depRelation: infeasible system (according to projectToVars' HsFM)") $ 
	 projectToVars' HsFM ies Nothing mappedIndices, smDim)
    s -> error s
    where
    ies =  ineqSystem((n_vars,n_params),iesMatrix)
    iesMatrix = multijoin1 
	  [ sdv `conc` zerol (dDim+mDim) sdp `conc` sdc,
	    zerol sDim  ddv `conc` zerol mDim ddp `conc` ddc,

            (-esi) `conc`   hv  `conc` zerol mDim   hp  `conc`   hc,
            esi    `conc` (-hv) `conc` zerol mDim (-hp) `conc` (-hc),

            (-smv) `conc` zerol dDim   esm  `conc` zerol dmDim (-smp) `conc` (-smc),
            smv    `conc` zerol dDim (-esm) `conc` zerol dmDim   smp  `conc`   smc ,

            zerol sDim (-dmv) `conc` zerol smDim   edm  `conc` (-dmp) `conc` (-dmc),
            zerol sDim   dmv  `conc` zerol smDim (-edm) `conc`   dmp  `conc`   dmc
	    ]

    -- deconstruct dep
    (h_trafo, srcDepDom, destDepDom) = 
	(depHTrafo dep, depSrcDomain dep, depDestDomain dep)


    -- DOMAINS
    (sdv, sdp, sdc) = partsVPC srcDepDom
    sDim            = nColumns sdv

    (ddv, ddp, ddc) = partsVPC destDepDom
    dDim            = nColumns ddv
		      
    -- H - TRANSFORMATION
    (hv,hp,hc)      = partsVPCih h_trafo
    assertHSrc      = case nRows hv == sDim of
		      True -> ""
		      False -> "dimension of dep's source index space and hTrafo don't match\n"

    assertHDest     = case nColumns hv == dDim of
		      True -> ""
		      False -> "dimension of dep's dest index space and hTrafo don't match\n"

    -- SPACE-TIME-MAPPING
    (smv, smp, smc) = partsVPCih srcMap
    assertSrcMap    = case nColumns smv == sDim of
		      True -> ""
		      False -> "dimension of source index space and STMapping don't match\n"
    smDim           = nRows smv
    
    (dmv, dmp, dmc) = partsVPCih destMap
    assertDestMap   = case nColumns dmv == dDim of
		      True -> ""
		      False -> "dimension of destination index space and STMapping don't match\n"
    dmDim           = nRows dmv
    mDim            = dmDim + smDim

    -- ALL PARAMETERS
    assertParm = case all (== (nColumns sdp)) (map nColumns paramMatrices) of
		 True -> ""
		 False -> "the number of parameters differ: " ++ 
                          concat 
                           (intersperse ", " (map show paramMatrices)) ++
                          "\n"
               where paramMatrices = [ddp,hp,smp,dmp]
    n_params = nColumns sdp

    -- ASSERTS TOGETHER
    failedAssertions = concat [assertHSrc, assertHDest, 
			       assertSrcMap, assertDestMap, assertParm]


    -- REMAINING DEFS    
    esi             = unitMatrix sDim sDim
    esm             = unitMatrix smDim smDim
    edm             = unitMatrix dmDim dmDim

    zerol cols m    = zeroMatrix (nRows m) cols `conc` m

    n_vars          = sDim+dDim+mDim
    mappedIndices   = [(sDim+dDim) .. (n_vars-1)]



invHTrafo :: IneqSystem -> IneqSystem -> Dependence -> 
	     DepRelation
invHTrafo srcIspc destIspc dep = 
    depRelation  dep (selfTrafo srcIspc) (selfTrafo destIspc)



-- selfTrafo can be used as identity for h-trafos or for
-- space-time-mappings:
selfTrafo :: IneqSystem -> AffineFunction
selfTrafo ispc = 
    affineFunction((nV,nP), unitMatrix (nP+nV+1) (nP+nV+1)) Homogeneous
    where nP   = nParams ispc
	  nV   = nVars   ispc

genericDep :: Statement -> AffineFunction -> Dependence
genericDep (stmtid, ispc) hTrafo =
    Dependence {depID = integerToID (-1), depType = UniformDummy,
		depSrcDomain = ispc, depSrcStmtID = stmtid,
		depDestDomain = ispc, depDestStmtID = stmtid,
		depHTrafo = hTrafo,
		depContext = [] }


selfDep :: Statement -> Dependence
selfDep (stmtid, ispc) = genericDep (stmtid, ispc) (selfTrafo ispc)


toDepRelation :: Dependence -> DepRelation
toDepRelation  d = 
    depRelation d  (selfTrafo (depSrcDomain d)) (selfTrafo (depDestDomain d))

genericDepRelation :: Statement -> AffineFunction -> DepRelation
genericDepRelation (stmtid,ispc) hTrafo =
    toDepRelation (genericDep (stmtid,ispc) hTrafo) 


selfDepRelation :: Statement ->  DepRelation
selfDepRelation (stmtid,ispc) = 
    genericDepRelation (stmtid,ispc) (selfTrafo ispc) 


------------------

--
-- konzeptionell falsch
--
wrong_transformDep :: [(TStatementID,AffineFunction)] -> Dependence ->
                      TrDependence
wrong_transformDep trafos d
    = TrDependence { trDepID                = error ("wrong_transformDep: " ++
						     "not implemented here!"),
		     trDepOrigID            = depID d,
                     trDepType              = depType d,
                     trDepOrigSrcStmtID     = depSrcStmtID d,
                     trDepOrigDestStmtID    = depDestStmtID d,
                     trDepSrcStmtID         = integerToID $ integerFromID $ depSrcStmtID d,
                     trDepDestStmtID        = integerToID $ integerFromID $ depDestStmtID d,
                     trDepRelation          = tdr,
                     trDepContext           = depContext d
                   }
    where
          tdr = depRelation  d (trafoOf trafos (integerToID $ integerFromID $ depSrcStmtID d))
                (trafoOf trafos (integerToID $ integerFromID $ depDestStmtID d))


-- |
-- Transform dependences.
-- The original dependences must be trimmed, i.e., 
-- 'trimDependence' must have been applied to them!
-- Note: the trDepID is only guaranteed to be unique for the set
--   of untransformed dependences that are passed to this function.
--   If transfromDeps is called with another set of dependences,
--   the trDepIDs in both sets may contain duplicates!
--   So: it is good practice to always transform ALL dependences at
--       the same time.
--
-- Use like this:
--
-- @
--   lds :: LoopoDS
--   tispc = loopoTISPC lds
--   trafo = loopoTrafo lds
--   deps  = map (trimDependence slp) $ loopoDEP lds
--   slp   = loopoSLP lds
--   proper_trafos = makeProperTrafos trafo
--   trdeps = transformDeps tispc proper_trafos deps
-- @
--
transformDeps :: [TStatement]
	      -> [(TStatementID, AffineFunction)]  -- ^ proper trafos
	      -> [Dependence]                      -- ^ trimmed dependences !
	      -> [TrDependence]
transformDeps tispc properTrs deps =
    numberTrDeps trDeps
    where
    trDeps = concatMap mkTrDep deps
    mkTrDep :: Dependence -> [TrDependence]
    mkTrDep d =
	[TrDependence { trDepID             = 
                            error "transformDep: undefined trDepID",
		        trDepOrigID         = depID d,
			trDepType           = depType d,
		        trDepOrigSrcStmtID  = depSrcStmtID d,
			trDepOrigDestStmtID = depDestStmtID d,
			trDepSrcStmtID      = tstmtidSrc,
			trDepDestStmtID     = tstmtidDest,
			trDepRelation       = tdr,
			trDepContext        = depContext d
		      }
	 | (tstmtidSrc,stmSrc)   <- properTrs,
	   (tstmtidDest,stmDest) <- properTrs,
	   (tstmtidSrc', stmtidSrc,_)  <- tispc, tstmtidSrc'  == tstmtidSrc,
	   (tstmtidDest',stmtidDest,_) <- tispc, tstmtidDest' == tstmtidDest,
	   stmtidSrc == depSrcStmtID d, stmtidDest == depDestStmtID d,
	   let tdr = depRelation d stmSrc stmDest
	]

numberTrDeps :: [TrDependence] -> [TrDependence]
numberTrDeps trDeps = zipWith insertTrDepID trDeps [1..]
    where
    insertTrDepID :: TrDependence -> Integer -> TrDependence
    insertTrDepID trDep i = trDep { trDepID = integerToID i }

