module LoopoDS.Trafo
    ( Trafo, StmtTrafo(..),
      readTrafo
    )
where

import LMath.Types
import LoopoDS.IntegralID
import LoopoDS.Statement
import ParseBase

import Control.Monad
import Data.List
import Data.Ratio


{-
  file               --> num_stat [<func_and_inv>] num_stat [<uvect_and_lincombs>]
                         (where <func_and_inv> and <uvect_and_lincombs> occur
                          num_stat times)
  func_and_inv       --> stat_ID 1 <func> <func>
                         (where the second occurrence of <func> corresponds to
                          the transformation matrix T, while the first one
                          corresponds to its inverse)
  func               --> n_vars n_pars n_rows <matrix>
  matrix             --> denom [entry]
  uvect_and_lincombs --> stat_ID <uvect> <all_lincombs>
  uvect              --> ulength [<is_used>]
                         (with is_used occurring ulength times and
                         the n-th occurrence being 1, iff the n-th outermost
                         loop built a base vector for the transformation)
  is_used            --> 1|0
  all_lincombs       --> nr_lincombs [<lincomb>]
                         (where <lincomb> occurs nr_lincombs times)
  lincomb            --> nr_coef lin_denom [coef]
                         (with coef occurring coef_nr times)
-}

-- |
-- Transformation for a statement.
--
data StmtTrafo =
    StmtTrafo { trafoFunc       :: AffineFunction,
		trafoInvFunc    :: AffineFunction,
		trafoUsedVector :: [Bool],
		trafoLincombs   :: [[Rational]]
	      }
    deriving Show

type Trafo = [(TStatementID, StmtTrafo)]

-- |
-- Reads the "trafo" from a file. The file @filename@
-- is read from disk, run through a parser for the ".trafo" file
-- format, and the result is returned.
--
readTrafo :: FilePath             -- ^ filename
	  -> IO (LPResult Trafo)
readTrafo filename =
    do contents <- readFile filename
       return $ runLP parseTrafo contents

parseTrafo :: Parser Trafo
parseTrafo =
    do num_stat1   <- parseInteger
       funcAndInv <- parseN num_stat1 parseFuncAndInv
       num_stat2  <- parseInteger
       uvectAlc   <- parseN num_stat2 parseUvectAndLincomb
       let stmtNrs1 = sort [n | (n,_) <- funcAndInv]
	   stmtNrs2 = sort [n | (n,_) <- uvectAlc]
       when (stmtNrs1 /= stmtNrs2)
	    (fail "Statement IDs for 'func_and_inv' and 'uvect_and_lincombs' \
		  \do not match")       
       return [(integerToID n,
		StmtTrafo { trafoFunc = g,
			    trafoInvFunc = f,
			    trafoUsedVector = v,
			    trafoLincombs   = lc }
	       )
	      | n <- stmtNrs1, let Just (f,g)  = lookup n funcAndInv,
	                       let Just (v,lc) = lookup n uvectAlc
	      ]


parseFuncAndInv :: Parser (Integer, (AffineFunction, AffineFunction))
parseFuncAndInv =
    do stat_ID <- parseInteger
       one     <- parseInteger
       unless (one == 1) (fail "expected a \'1\' after stat_ID")
       inverse <- parseAffineFunction
       trans   <- parseAffineFunction
       return (stat_ID, (inverse, trans))

parseUvectAndLincomb :: Parser (Integer, ([Bool], [[Rational]]))
parseUvectAndLincomb =
    do stat_ID <- parseInteger
       ulength <- parseInteger
       used_v  <- parseIntegers ulength
       nr_lincombs <- parseInteger
       lincombs    <- parseN nr_lincombs parseLinComb
       return (stat_ID, (map (/=0) used_v, lincombs))

parseLinComb :: Parser [Rational]
parseLinComb =
    do nr_coef   <- parseInteger
       lin_denom <- parseInteger
       coefs     <- parseIntegers nr_coef
       return (map (% lin_denom) coefs)
