module Main ( main )
where

import Control.Exception
import System.Console.GetOpt
import System.Environment
import System.IO
import Text.ParserCombinators.Parsec

import Debug.Trace

import LMath
import LMath.Algo.NLGen


runLGen :: RealFrac a
        => LGenOptions
        -> LGenParams (FracPoly a)
        -> LGenOutputOptions (FracPoly a)
        -> [QfFormula (Polynomial a)]
        -> Integer
        -> String
runLGen lgo lgp lgoo nlCtx level =
    runLGenWithStatements lgo lgfFracPoly lgp lgoo
      (fracPolyDTreeContext nlCtx)
      (showPolyStd (lgooNonlinParamNames lgoo) . snd . ratPoly2intPoly)
      level




data LGenOpt = Level Integer
             | OutputFile FilePath
             | AlgoVariant LGenAlgoVariant
             | Progress
             | Debug
             | SimplifyIFs
             | SimplifyOuterLoops
             | SimplifyAll
             | ConvexUnion
             | Simplex
               deriving (Eq, Show)

options :: [OptDescr LGenOpt]
options = [Option ['l'] ["level"] (ReqArg (Level . read) "N") "pull IFs N levels out"
          ,Option ['o'] ["output"] (ReqArg OutputFile "OUTPUT") "write output to OUTPUT"
--          ,Option []    ["original"] (NoArg (AlgoVariant LGenOriginal)) "use original (slow but correct) algorithm"
--          ,Option []    ["slightly-improved"] (NoArg (AlgoVariant LGenSlightlyImproved)) "use a slightly improved correct algorithm (default)"
--          ,Option []    ["experimental"] (NoArg (AlgoVariant LGenExperimental)) "use experimental algorithm"
          ,Option []    ["progress"] (NoArg Progress) "show progress"
          ,Option []    ["debug"]    (NoArg Debug)    "show debugging output"
          ,Option []    ["simplify-ifs"] (NoArg SimplifyIFs) "simplify if statements"
--          ,Option []    ["simplify-outer-loops"] (NoArg SimplifyOuterLoops) "simplify outer for loops"
--          ,Option []    ["simplify"] (NoArg SimplifyAll) "turn on all --simplify-* options"
--          ,Option []    ["convex"] (NoArg ConvexUnion) "use convex description for outer loops"
          ,Option []    ["simplex"] (NoArg Simplex) "use Simplex to solve LP questions"
          ]

main :: IO ()
main = do args <- getArgs
          mainA args

mainA :: [String] -> IO()
mainA args =
          case getOpt Permute options args of
            (opts,[infile],[]  ) -> run infile opts
            (_,_,errs) ->
                do binary <- getProgName
                   let header = "Usage: " ++ binary ++
                                " [OPTION...] file.cloogInput"
                   hPutStrLn stderr (concat errs ++ usageInfo header options)

run :: FilePath -> [LGenOpt] -> IO ()
run input opts =
    do p <- parseFromFile parseLGenInput input
       case p of
           Right (lgp,lgoo,nlCtx) ->
               let dims = head ([l | Level l <- opts] ++ [nScatterDims])
                   nScatterDims = case concat $ lgpStatements lgp of
                                    [] -> 0
                                    (sys,sc):_ -> nVars sc - nVars sys
                   c  = runLGen lgo lgp lgoo nlCtx dims
               in
               outputFn c
           Left err ->
               hPutStrLn stderr ("Parse error: " ++ show err)
    where
    outputFn = head ([writeFile out | OutputFile out <- opts] ++ [putStrLn])
    algo = head ([a | AlgoVariant a <- opts] ++ [LGenOriginal])
    lgo = lgenOptionsStd { lgoAlgoVariant = algo, 
                           lgoLpVariant = if Simplex `elem` opts then LGenSimplex else LGenQE,
                           lgoProgressToStderr = Progress `elem` opts,
                           lgoDebug = Debug `elem` opts,
                           lgoSimplifyIFs = any (`elem` opts) [SimplifyIFs,SimplifyAll],
                           lgoSimplifyOuterLoops =
                               any (`elem` opts) [SimplifyOuterLoops,SimplifyAll],
                           lgoConvexUnion = ConvexUnion `elem` opts
                         }


handleToolIOError :: IO a -> IO ()
handleToolIOError action =
    catchDyn (action >> return ())
             (\(tio :: ToolIOError) ->
                 do hPutStrLn stderr ("Tool IO error: " ++ show tio)
             )


------------------------------------------


test2 = lgen lgenOptionsStd
             lgfFracPoly
             1
             (ineqSystem ((0,1), zeroMatrix 0 2))
             [[(ineqSystem ((1,1), matrix [[1,0,0], [-1,1,0]])
               ,ineqSystem ((2,1), matrix [[-1, 2, 0, 0]
                                          ,[ 1,-2, 0, 0]
                                          ])
               )
              ]
             ]

noninjTest = lgen lgenOptionsStd
                  lgfFracPoly
                  1
                  (ineqSystem ((0,1), zeroMatrix 0 2))
                  [[(ineqSystem ((2,1), matrix [[1,0,0,0], [-1,0,1,0], [0,1,0,0],[0,-1,0,10]])
                    ,ineqSystem ((3,1), matrix [[-1, 1, 1, 0, 0]
                                               ,[ 1,-1,-1, 0, 0]
                                               ])
                    )
                   ]
                  ]


test10_20 =
    lgen (lgenOptionsStd { lgoAlgoVariant = LGenOriginal, lgoLpVariant = LGenSimplex })
         lgfFracPoly
         1
         (ineqSystem ((0,0), zeroMatrix 0 1))
         [[(ineqSystem ((1,0), matrix [[1,0], [-1,10]])
           ,ineqSystem ((2,0), matrix [[1,-1,0], [-1,1,0]])
           )
          ]
         ,[(ineqSystem ((1,0), matrix [[1,0], [-1,10]])
           ,ineqSystem ((2,0), matrix [[1,-2,0], [-1,2,0]])
           )
          ]
         ]

a,b :: GIneqSystem (FracPoly Rational)
a = ineqSystem ((1,2), matrix [[1,0,0,0], [-1,1,0,0]])
b = ineqSystem ((1,2), matrix [[1,0,0,0], [-1,0,2,0]])

           
testM_N2 :: RealFrac a => DTreeGen (FracPoly a) (Polynomial a) (LCode (FracPoly a))
testM_N2 =
    lgen (lgenOptionsStd { lgoAlgoVariant = LGenOriginal })
         lgfFracPoly
         1
         (ineqSystem ((0,2), zeroMatrix 0 3))
         [[(ineqSystem ((1,2), matrix [[1,0,0,0], [-1,1,0,0]])
           ,ineqSystem ((2,2), matrix [[1,-1,0,0,0], [-1,1,0,0,0]])
           )
          ]
         ,[(ineqSystem ((1,2), matrix [[1,0,0,0], [-1,0,1,0]])
           ,ineqSystem ((2,2), matrix [[1,-2,0,0,0], [-1,2,0,0,0]])
           )
          ]
         ]

codeM_N2 = 
 runLGenWithStatements 
   (lgenOptionsStd { lgoAlgoVariant = LGenOriginal })
   lgfFracPoly
   (LGP { lgpContext = ineqSystem ((0,2), zeroMatrix 0 3)
        , lgpStatements =
            [[(ineqSystem ((1,2), matrix [[1,0,0,0], [-1,1,0,0]])
              ,ineqSystem ((2,2), matrix [[1,-1,0,0,0], [-1,1,0,0,0]])
              )
             ]
            ,[(ineqSystem ((1,2), matrix [[1,0,0,0], [-1,0,1,0]])
              ,ineqSystem ((2,2), matrix [[1,-2,0,0,0], [-1,2,0,0,0]])
              )
             ]
            ]
        }
   )
   (fracPolyLGOO ["i"] ["c"] ["M","N"] [])
   (fracPolyDTreeContext[])
   (showPolyStd [])
   1

testM_N2' :: DTreeGen (FracPoly Rational) (Polynomial Rational) (LCode (FracPoly Rational))
testM_N2' =
    lgen (lgenOptionsStd { lgoAlgoVariant = LGenOriginal, lgoLpVariant = LGenQE })
         lgfFracPoly
         1
         (ineqSystem ((0,0), zeroMatrix 0 1))
         [[(ineqSystem ((1,0), matrix [[1,0], [-1,m%%1]])
           ,ineqSystem ((2,0), matrix [[1,-1,0], [-1,1,0]])
           )
          ]
         ,[(ineqSystem ((1,0), matrix [[1,0], [-1,n%%1]])
           ,ineqSystem ((2,0), matrix [[1,-2,0], [-1,2,0]])
           )
          ]
         ]
    where
      m = indeterminate 2 0 :: Polynomial Rational
      n = indeterminate 2 1

codeM_N2' = 
 runLGenWithStatements 
   (lgenOptionsStd { lgoAlgoVariant = LGenOriginal, lgoLpVariant = LGenQE })
   lgfFracPoly
   (LGP { lgpContext = ineqSystem ((0,0), zeroMatrix 0 1)
        , lgpStatements =
            [[(ineqSystem ((1,0), matrix [[1,0], [-1,m%%1]])
              ,ineqSystem ((2,0), matrix [[1,-1,0], [-1,1,0]])
              )
             ]
            ,[(ineqSystem ((1,0), matrix [[1,0], [-1,n%%1]])
              ,ineqSystem ((2,0), matrix [[1,-2,0], [-1,2,0]])
              )
             ]
            ]
        }
   )
   (fracPolyLGOO ["i"] ["c"] [] ["M","N"])
   (fracPolyDTreeContext [QfAt (AtGt n), QfAt (AtGe (m-1))])
   (showPolyStd ["M","N"] . snd . ratPoly2intPoly)
   1
 where
 m = indeterminate 2 0 :: Polynomial Rational
 n = indeterminate 2 1


codeM_NP' = 
 runLGenWithStatements 
   (lgenOptionsStd { lgoAlgoVariant = LGenOriginal, lgoLpVariant = LGenQE })
   lgfFracPoly
   (LGP { lgpContext = ineqSystem ((0,0), zeroMatrix 0 1)
        , lgpStatements =
            [[(ineqSystem ((1,0), matrix [[1,0], [-1,m%%1]])
              ,ineqSystem ((2,0), matrix [[1,-1,0], [-1,1,0]])
              )
             ]
            ,[(ineqSystem ((1,0), matrix [[1,0], [-1,n%%1]])
              ,ineqSystem ((2,0), matrix [[1,-p%%1,0], [-1,p%%1,0]])
              )
             ]
            ]
        }
   )
   (fracPolyLGOO ["i"] ["c"] [] ["M","N","P"])
   (fracPolyDTreeContext [QfAt (AtGt n), QfAt (AtGe (m-1)), QfAt (AtGt p)])
   (showPolyStd ["M","N","P"] . snd . ratPoly2intPoly)
   1
 where
 m = indeterminate 3 0 :: Polynomial Rational
 n = indeterminate 3 1
 p = indeterminate 3 2




code12 = 
 runLGenWithStatements 
   (lgenOptionsStd { lgoAlgoVariant = LGenOriginal })
   lgfFracPoly
   (LGP { lgpContext = ineqSystem ((0,1), zeroMatrix 0 2)
        , lgpStatements =
            [[(ineqSystem ((1,1), matrix [[1,0,0], [-1,1,0]])
              ,ineqSystem ((3,1), matrix [[1,0,-1,0,0], [-1,0,1,0,0]
                                         ,[0,1, 0,0,0], [0,-1,0,0,0]])
              )
             ]
            ,[(ineqSystem ((2,1), matrix [[1,0,0,0], [-1,0,1,0]
                                         ,[0,1,0,0], [0,-1,1,0]])
              ,ineqSystem ((4,1), matrix [[1,0,-1,0,0,0], [-1,0,1,0,0,0]
                                         ,[0,1,0,-1,0,0], [0,-1,0,1,0,0]])
              )
             ]
            ]
        }
   )
   (fracPolyLGOO ["i","j"] ["c","d"] ["M"] [])
   (fracPolyDTreeContext[])
   (showPolyStd [])
   1
