{-
    This file is part of the LooPo project, see
    <http://www.infosun.fim.uni-passau.de/cl/loopo/>.

    Copyright (C) 2006 Universitaet Passau, Germany
    Author: Armin Groesslinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-}

-- |
-- This module provides som basic parse actions for
-- the data structures of /LooPo/. More complex parse actions
-- (for /LooPo/'s data files, for example) can be composed from
-- the combinators provided by this module.
--
-- Author: Armin Groesslinger
--
-- Version: $Id: ParseBase.hs 4188 2009-08-17 16:42:26Z groessli $
--
module ParseBase
    ( module LP,
      parseMatrix,
      parseAffineFunction,
      parseAffineFunctionI,
      parseIneqSystem
)
where

import Data.Ratio

import LP
import LMath

-- |
-- Parses a matrix.
-- This parse action takes the denominator of the matrix from the input
-- and then parses a matrix of the specified width and height from the
-- input.
--
parseMatrix :: Integer -> Integer -> Parser Matrix
parseMatrix rows cols =
    do denom  <- parseInteger
       list   <- parseIntegers (rows * cols)
       return $ matrix (rows, cols, map (% denom) list)


-- |
-- Parse an affine function in (in)homogenous representation.
-- The number of variables, parameters, and rows (in this order) are
-- taken from the input, and then a matrix of matching dimensions is parsed.
-- The returned affine function is formed from that matrix and the number of
-- variables and parameters read before.
--
parseAffineFunctionHI :: CoordinateSystem -> Parser AffineFunction
parseAffineFunctionHI c =
    do vars   <- parseInteger
       params <- parseInteger
       rows   <- parseInteger
       matr   <- parseMatrix rows (vars+params+1)
       return $ affineFunction ((vars, params), matr) c



-- |
-- Parse an affine function in homogeneous representation.
-- The number of variables, parameters, and rows (in this order) are
-- taken from the input, and then a matrix of matching dimensions is parsed.
-- The returned affine function is formed from that matrix and the number of
-- variables and parameters read before.
--
parseAffineFunction :: Parser AffineFunction
parseAffineFunction = parseAffineFunctionHI Homogeneous

-- |
-- Parse an affine function in inhomogeneous representation.
-- The number of variables, parameters, and rows (in this order) are
-- taken from the input, and then a matrix of matching dimensions is parsed.
-- The returned affine function is formed from that matrix and the number of
-- variables and parameters read before.
--
parseAffineFunctionI :: Parser AffineFunction
parseAffineFunctionI = parseAffineFunctionHI Inhomogeneous



-- |
-- Parse an inequation system.
-- The number of variables, parameters, and rows (in this order) are
-- taken from the input, and then a matrix of matching dimensions is parsed.
-- The returned inequation system is formed from that matrix and the number of
-- variables and parameters read before.
--
parseIneqSystem :: Parser IneqSystem
parseIneqSystem =
    do vars   <- parseInteger
       params <- parseInteger
       rows   <- parseInteger
       matr   <- parseMatrix rows (vars+params+1)
       return $ ineqSystem ((vars, params), matr)
