-- |
-- This module contains functions to convert 'IneqSystem's 
-- to a string that can be used as input for several polylib
-- functions, e.g., barvinok_count, or for the code generator
-- by quillere.
--
-- author: Martin Griebl
-- date: 01 dec 2006
--
module PolylibInterf (showIneqSystemForPolylib, showIneqSystemsForQuillere)
where

import LMath.Types
import List (intersperse)

-- |
-- The polylib format usually requires two parts: one for the variable
-- and one for the context, i.e., the constraints on the parameters.
-- showVarIneqSystemForPolylib creates the part for the variables from
-- a given IneqSystem.
--
-- Note that Polylib requires that each inequality is preceded by
-- a constant 1; a preceding 0 would indicate an equation.
-- showVarIneqSystemForPolylib converts the variables 
showVarIneqSystemForPolylib :: IneqSystem -> String
showVarIneqSystemForPolylib is =
    -- head line for matrix size
    show (nRows is) ++ " " ++ show (1 + nColumns mat) ++ "\n" ++
    -- the extended matrix itself
    canonicShowMat  (conc (columnMatrix [1 | i <- [1..(nRows is)]])  mat) False  ++ "\n"
    where mat = getMatrix is


-- |
-- The polylib format usually requires two parts: one for the variable
-- and one for the context, i.e., the constraints on the parameters.
-- showParamIneqSystemForPolylib creates the part for the parameters. from
-- a given IneqSystem.
-- 
-- Since in the LooPo framework all parameters are implicitly non-negative
-- the context to be written just consists of one inequality per paramteter
-- telling that this parameter is non-zero. Thus, the input is just the
-- number of parameters.
--
-- Note that this function does *not* search for additional constraints 
-- on the parameters that could be hidden in an IneqSystem.
--
showParamIneqSystemForPolylib :: Integer -> String
showParamIneqSystemForPolylib nP =
    -- a matrix for the non-negative parameters
    show (nP+1) ++ " " ++ show (2 + nP) ++ "\n" ++
    intShowMat paramMatrix False ++ "\n"
    where paramMatrix = 
	      conc (columnMatrix [1 | i <- [1..(nP+1)]])
		   (unitMatrix (nP+1) (nP+1))
			

-- |
-- The polylib format usually requires two parts: one for the variable
-- and one for the context, i.e., the constraints on the parameters.
-- showIneqSystemForPolylib returns these two descriptions as a string
-- and is thus the main function of this module.
--
showIneqSystemForPolylib :: IneqSystem -> String
showIneqSystemForPolylib is =
    showVarIneqSystemForPolylib is ++ 
    showParamIneqSystemForPolylib (nParams is)

------------
	
-- |
-- This function converts the given IneqSystems to a String representation
-- that is used by Quillere's code generator.
-- For description of the Integers, see the documnentation of Quillere.
--
showIneqSystemsForQuillere :: Integer -> Integer -> [IneqSystem] -> String
showIneqSystemsForQuillere timedims sep ieqs =
    show (length ieqs) ++ " " ++ show timedims ++ " "  ++
    show sep ++ "\n\n" ++
    "1\n#\n" ++ showParamIneqSystemForPolylib nparams ++
    "1\n#\n" ++ 
    concat (intersperse "1\n#\n" (map showVarIneqSystemForPolylib ieqs))
    where nparams = case alleq of
		    True -> fstparams
		    False -> error ("showIneqSystemForQuillere: " ++
			     "different nubers of parameters is ineqs")
	  fstparams = nParams (head ieqs)
	  alleq = and (map ((== fstparams) . nParams) ieqs)



