-- | Heuristik to get a schedule that allows optimization of memory usage.
--   Implemented by Benjamin Schulte.
--   See my paper for further information.

module QuilMem.BetterSchedule (
        changeSched
       )
where

import QuilMem.QuilMem
import Maybe
import Ratio
import Data.List
import CppInterface
import Foreign.C.String
import LoopoDS
import LoopoDS.Access
import LMath.Types
import LMath.Algo.Sort
import LMath.Algo.Simplex
import qualified Data.Map as Map
import Data.Map ( Map )
import CompSpread.Helpers
import Control.Exception ( evaluate )
import qualified TrDepDirections

-- | aendert die Schedule-Konstante eines Statements sowie aller davon abhaengenden
moveStmtSched :: Schedules
                 -> [TrDependence]
                 -> StatementID
                 -> [Rational]
                 -> Schedules
moveStmtSched scheds deps st toChange = let changestatements = flattenDeps st in
                                        foldl (\sched st -> fst (changeStmtSchedConstant sched st toChange)) scheds changestatements
                 where              
                 flattenDeps curst = let newstats = delete curst $ nub $ map trDepOrigDestStmtID $ filter (\trdep -> trDepOrigSrcStmtID trdep == curst) deps
                                     in if null newstats then [curst] else curst:(nub (concat (map flattenDeps newstats)))

-- | aendert die Schedule-Konstante eines einzelnen Statements
changeStmtSchedConstant :: Schedules                        -- ^ der alte Schedule
                           -> StatementID                   -- ^ das zu aendernde Statement - muss ein QList sein!
                           -> [Rational]                    -- ^ Konstante, um die das Statement verschoben wird
                           -> (Schedules, GMatrix Rational) -- ^ der neue Schedule sowie das neue Statement
changeStmtSchedConstant sched st toChange = 
               let (Schedules scheds) = sched 
                   (QList srcStatementSched') = ((Map.!) scheds st)
                   srcStatementSched = rowVectorsToMatrix1 srcStatementSched'
                   newPos = map (\r -> (entry srcStatementSched r ((nColumns srcStatementSched)-1)) + (toChange!!r)) [0..((nRows srcStatementSched)-1)]
                   newSrcStSched = foldl (\m r -> setEntry m r ((nColumns srcStatementSched)-1) (newPos!!r)) srcStatementSched [0..((nRows srcStatementSched)-1)]
                   newSched = (Schedules (Map.insert st (QList $ rowVectors newSrcStSched ) scheds))
               in (newSched,newSrcStSched)


-- | prueft nach, ob die vom geaenderten Statement abgehenden Lebenszeiten minimal 1 sind.
-- Falls nicht, wird das Statement so verschoben, dass die minimale Lebenszeit sich auf 1 veraendert.
checkSchedConstantsFromDep :: String -> String              -- ^ Pfad- und Dateiname der LooPo-Dateien
               -> ArrayID                                   -- ^ das Array, um das sich die Analyse dreht
               -> TrDependence                              -- ^ die gerade bearbeitete Abhaengigkeit
               -> IO ()                                     
checkSchedConstantsFromDep dir file arr curDep = do
               let ds = unsafeCreateCompSpreadDS dir file
                   deps = filter (\trdep -> trDepSrcStmtID trdep == trDepSrcStmtID curDep) (getArrayTrDeps ds LifeTimeMin arr)
                   (nV, nP, sched) = csSchedule ds
                   minLifeTimes = (map (\trdep -> getDepLifeTimeDTree LifeTimeMin (trdep, (getDepQualVect ds trdep))) deps)
                   (Optimum minLT) = minimumBy dTreeMin (map snd minLifeTimes)
                   dTreeMin (Optimum a) (Optimum b) = compare a b
                   dTreeMin _ _                            = error("BetterSchedule.checkSchedConstants: Unbounded bei LifeTimeMin -> Schedule ungueltig!")
                   curStatement = trDepSrcStmtID curDep
                   toChange = ((head minLT)-1):(replicate ((length minLT)-1) (0%1))
                   (newSched,newSrcStSched) = changeStmtSchedConstant sched (trDepOrigSrcStmtID curDep) toChange

               if ((head minLT) /=1)
                then do
                      putStrLn ("Outgoing dependences from " ++ (show (trDepSrcStmtID curDep)) ++ 
                                " have a minimal lifetime of " ++ (show minLT) ++ 
                                " -> to move.")
                      putStrLn ("Writing new schedule: "   ++ (show newSrcStSched ))
                      writeFile (dir ++ "/" ++ file ++ ".sched") (formatStmtSchedule nV nP newSched)
                      putStr ("Please run TargetLoops and enter OK: ")
                      waitForConfirm
                      return ()
                else do                     
                      putStrLn ("Outgoing dependences from " ++ (show (trDepSrcStmtID curDep)) ++ " have a minimal lifetime of 1 -> ok.")
                      return ()

-- | prueft nach, ob die auf das geaenderte Statement zeigenden Lebenszeiten minimal 1 sind.
-- Falls kleiner, wird das Statement sowie alle abhaengenden entsprechend verschoben.
checkSchedConstantsToDep :: String -> String                             -- ^ Pfad- und Dateiname der LooPo-Dateien
               -> ArrayID                                   -- ^ das Array, um das sich die Analyse dreht
               -> TrDependence                              -- ^ die gerade bearbeitete Abhaengigkeit
               -> IO ()                                     
checkSchedConstantsToDep dir file arr curDep = do
               let ds = unsafeCreateCompSpreadDS dir file
                   deps = filter (\trdep -> trDepDestStmtID trdep == trDepSrcStmtID curDep) (getArrayTrDeps ds LifeTimeMin arr)
                   (nV, nP, scheds) = csSchedule ds
                   minLifeTimes = (map (\trdep -> getDepLifeTimeDTree LifeTimeMin (trdep, (getDepQualVect ds trdep))) deps)
                   (Optimum minLT) = minimumBy dTreeMin (map snd minLifeTimes)
                   dTreeMin (Optimum a) (Optimum b) = compare a b
                   dTreeMin _ _                            = error("BetterSchedule.checkSchedConstants: Unbounded bei LifeTimeMin -> Schedule ungueltig!")
                   curStatement = trDepSrcStmtID curDep
                   realMinLT = if null deps then 1 else head minLT
                   toChange = (1-realMinLT):(replicate ((length minLT)-1) (0%1))
                   newSched = moveStmtSched scheds (csTrDeps ds) (trDepOrigSrcStmtID curDep) toChange

               if (realMinLT/=1)
                then do
                      putStrLn ("To " ++ (show (curStatement)) ++ " directing dependences have a minimal lifetime smaller unequal 1 -> to move.")
                      putStrLn ("Writing new schedule: "++ (show newSched))
                      writeFile (dir ++ "/" ++ file ++ ".sched") (formatStmtSchedule nV nP newSched)
                      putStr ("Please run TargetLoops and enter OK: ")
                      waitForConfirm
                      return ()

                else do                     
                      putStrLn ("To " ++ (show (trDepSrcStmtID curDep)) ++ " directing dependences have a minimal lifetime of 1 -> ok.")
                      return ()

-- | Prueft, ob ein Statement vor Zeitpunkt 0 ausgefuehrt werden kann. Ist dem nicht so, wird verschoben.
checkMinTime :: String -> String -> StatementID -> TStatementID -> IO ()
checkMinTime dir file stmtid tstmtid = do
               let ds = unsafeCreateCompSpreadDS dir file
                   (Just ineq) = lookup stmtid (csISPCs ds)
                   (Just qv)   = lookup tstmtid (csQualVects ds)
                   slp         = csSLPs ds
                   (nV, nP, scheds) = csSchedule ds
                   (Schedules sched) = trimSchedules slp scheds
                   (QList stmtsched)  = ((Map.!) sched stmtid)
                   (_, _, places) = csPlacement ds
                   (Placements place) = trimPlacements slp places
                   (QList stmtplace)  = ((Map.!) place stmtid)
                   schedqv = map fst $ filter (\qv -> snd qv == Sched) $ zip [0..] qv
                   trISPCs = calcTrISPC (stmtid,ineq) (rowVectorsToMatrix1 stmtsched) (rowVectorsToMatrix1 stmtplace) qv
                   mintime = case trISPCs of 
                              (Just tispcs) -> let noParams = moveParamsToVars tispcs 0 (nParams tispcs) (nVars tispcs) 
                                                   onlyTimes = unjust $ projectToVars noParams Nothing schedqv
                                               in case (unLeaf (makeDTree (numOrdDTreeContext ()) (lexmin stdSimplexParams onlyTimes Nothing))) of
                                                   Unbounded -> error("BetterSchedule.checkMinTime: Unbounded MinLaunchTime, probably because of to few ineqs")
                                                   Optimum a -> a
                              Nothing       -> error ("BetterSchedule.checkMinTime: No target index space!")
                   unjust (Just a) = a
                   unjust Nothing = error("BetterSchedule.checkMinTime: Unexpected nothing.")
                   unLeaf (DLeaf a) = a
                   unLeaf x         = error("BetterSchedule.checkMinTime: DTree.Leaf expected, found: "++ (show x))
                   toChange = map (\i -> if i<0 then i*(-1) else 0) mintime
                   newSched = moveStmtSched scheds (csTrDeps ds) stmtid toChange

               if (mintime) < (replicate (length mintime) 0)
                then do
                      putStrLn ("It is possible that " ++ (show tstmtid) ++ " is executed before time 0 -> to move.")
                      putStrLn ("Writing new schedule: "++ (show newSched))
                      writeFile (dir ++ "/" ++ file ++ ".sched") (formatStmtSchedule nV nP newSched)
                      putStr ("Please run TargetLoops and enter OK: ")
                      waitForConfirm
                      return ()

                else do                     
                      putStrLn ("Not possible to execute " ++ (show tstmtid) ++ " before time 0 -> ok.")
                      return ()



changeSched :: String -> String                             -- ^ Pfad- und Dateiname der LooPo-Dateien
               -> ArrayID                                   -- ^ das Array, um das sich die Analyse dreht
               -> [(TrDependenceID, Optimum [Rational])]    -- ^ das letzte Lifetime-Analyseergebnis
               -> [StatementID]                             -- ^ bisher veraenderte Statements
               -> IO [(TrDependenceID, Optimum [Rational])] -- ^ jetziges Lifetime-Analyseergebnis
changeSched dir file arr lts changed = do
               let ds = unsafeCreateCompSpreadDS dir file        
                   unbounddeps  = map (getTrDep ds) $ map (fst) (filter (\dep -> (snd dep) == Unbounded) lts)
                   hasSameStatement = or $ map (\trdep -> trDepOrigSrcStmtID trdep == trDepOrigDestStmtID trdep) unbounddeps
                   -- stats     = map (\trDep -> (trDepOrigSrcStmtID trDep,trDepOrigDestStmtID trDep)) (map (getTrDep ds) unbounddeps)
                   sameTarget   = groupBy (\trDep1 trDep2 -> trDepOrigDestStmtID trDep1 == trDepOrigDestStmtID trDep2) unbounddeps
                   -- Die Reperatur des Schedules beginnt mit dem Statement, dass die letzte Quelle des ersten Ziels ist:
                   curDep       = last $ head sameTarget 
                   curStatement = trDepOrigSrcStmtID curDep
                   curTStatement = trDepSrcStmtID curDep
                   min_lt = getDepLifeTimeDTree LifeTimeMin (curDep, (getDepQualVect ds curDep))
                   (nV, nP, Schedules scheds) = csSchedule ds
                   -- hier kann die Verarbeitung abbrechen, wenn QIf statt QList genutzt ist; damit koennte aber 
                   -- sowieso nicht gearbeitet werden.
                   srcStatementSched = case ((Map.!) scheds curStatement) of
                                        QList a -> a
                                        _       -> error ("BetterSchedule.changeSched: Only unsplitted schedules are accepted by this heuristic.")
                   destStatementSched = case ((Map.!) scheds (trDepOrigDestStmtID curDep)) of
                                        QList a -> a
                                        _       -> error ("BetterSchedule.changeSched: Only unsplitted schedules are accepted by this heuristic.")

                   (SLP slps) = (csSLPs ds)
                   (Just curStmtslp) = lookup curStatement slps 
                   mNewSrcStSched = makeSourceDestSched curStmtslp (rowVectorsToMatrix1 srcStatementSched) (rowVectorsToMatrix1 destStatementSched)
                   (Just newSrcStSched) = mNewSrcStSched
                   newSched = Map.insert curStatement (QList $ rowVectors newSrcStSched) scheds
                   newds = unsafeCreateCompSpreadDS dir file        
                   newlts = getArrayDepLifeTimeTrees newds LifeTimeMax arr 
               
               if (null unbounddeps)
                then do
                      putStrLn ("Found new schedule!")
                      return lts
                else if (elem curStatement changed)
                      then do
                            putStrLn ("Already changed schedule of " ++ (show curStatement) ++ ", no further changes possible.")
                            return lts 
                else if hasSameStatement 
                      then do
                            putStrLn ("There is an Unbounded dependence from a statement directing to itself - no changes possible.")
                            return lts 
                      else do 
                            putStrLn ("Trying to change schedule of " ++ (show curStatement) ++ "...")

                            putStr ("Minimal lifetime of array for " ++ (show (fst min_lt)) ++ " is ")
                            if (snd min_lt) == Unbounded
                             then do
                                   putStrLn "Unbounded. Cannot improve schedule."
                                   return lts
                             else do
                                   putStrLn ("not Unbounded. Trying to adjust schedule to " 
                                             ++ (show (trDepOrigDestStmtID curDep)) ++ ". Schedules at the moment:")
                                   putStrLn ("Source: " ++ (show srcStatementSched))
                                   putStrLn ("Dest: "   ++ (show destStatementSched))

                                   if mNewSrcStSched == Nothing
                                    then do
                                          putStrLn "Schedules cannot be changed."
                                          return lts
                                    else do
                                          putStrLn ("Writing new schedule: "   ++ (show newSrcStSched ))
                                          writeFile (dir ++ "/" ++ file ++ ".sched") (formatStmtSchedule nV nP (Schedules newSched))
                                                             
                                          putStr ("Bitte nun TargetLoops ausfuehren und mit OK bestaetigen: ")
                                          waitForConfirm
                                          checkSchedConstantsFromDep dir file arr curDep 
                                          checkSchedConstantsToDep   dir file arr curDep 

                                          checkMinTime dir file curStatement curTStatement

                                          -- Abschliessendes ueberpruefen, ob Schedule gueltig:
                                          if (TrDepDirections.isSchedule (csTrDeps newds) (0,0))
                                           then do 
                                                 putStrLn "***** New schedule valid *****"
                                                 putStrLn ("Now lifetimes are:")
                                                 putStrLn (showLifeTimeAnalysis newds newlts)

                                                 nlts <- (changeSched dir file arr newlts (curStatement:changed))
                                                 return nlts
                                           else do
                                                 putStrLn "***** New schedule NOT valid *****"
                                                 return lts                                                

makeSourceDestSched :: [Bool] -> GMatrix Rational -> GMatrix Rational -> Maybe (GMatrix Rational)
makeSourceDestSched slps sS dS  = foldl (\sS r -> convertLine r slps sS dS) (Just sS) [0..((nRows sS)-1)] where
                convertLine r slps (Just sS) dS = mSDS (slps++(repeat True)) r sS dS [0..((nColumns sS)-1)]
                convertLine _ _    Nothing   _  = Nothing
                mSDS _          r sS dS [n]     = let i = entry sS r n
                                                      j = entry dS r n
                                                  in case r of
                                                  0 -> Just (setEntry sS r n (j-1)) -- Konstantenwert
                                                  r -> Just (setEntry sS r n  j   )
                mSDS (slp:slps) r sS dS (n:ns)  = let i = entry sS r n
                                                      j = entry dS r n
                                                  in if ((i<j && slp) || (i==j))
                                                      then fmap (\m -> setEntry m r n j) (mSDS slps r sS dS ns)
                                                      else Nothing

waitForConfirm :: IO ()
waitForConfirm = do
                  ok <- getLine
                  if ok == "OK"
                   then return ()
                   else waitForConfirm
                

--- | Calculates a target Index Space for a single target statement
calcTrISPC :: Statement                   -- ^ index space
              -> GMatrix Rational         -- ^ trimmed schedule
              -> GMatrix Rational         -- ^ trimmed placement
              -> [QualEntry]              -- ^ QualVect
              -> Maybe IneqSystem         -- ^ target index space
calcTrISPC (stid, ineq) sched place qv = projectToVars newineq (Just eqsystem) [0..(ts+ps-1)] where
                                   ts      = nRows sched
                                   ps      = nRows place
                                   newineq = newVarsToAffExpBased (ts+ps) ineq
                                   newsched = foldl setSchedPlaceEntries (insertColumns sched 0 (zeroMatrix ts (ts+ps))) schedqvs
                                   schedqvs = zip [0..] $ map (fst) $ filter (\qv->snd qv == Sched) $ zip [0..] qv                                   
                                   newplace = foldl setSchedPlaceEntries (insertColumns place 0 (zeroMatrix ps (ts+ps))) placeqvs
                                   placeqvs = zip [0..] $ map (fst) $ filter (\qv->snd qv == Place) $ zip [0..] qv
                                   schedplace = insertRows newsched 0 newplace
                                   eqsystem = fromAffineExp $ affineExp ((nVars newineq,nParams newineq),schedplace)
                                   setSchedPlaceEntries m (r,c) = setEntry m r c (-1%1)

newVarsToAffExpBased n aeb = newaeb 
                             where
                              ae      = toAffineExp aeb
                              nv      = nVars ae
                              np      = nParams ae
                              m       = getMatrix ae
                              nm      = zeroMatrix (nRows m) n
                              withNewCols = insertColumns m 0 nm
                              newae   = affineExp (((nv+n),np),withNewCols)
                              newaeb = fromAffineExp newae
