module Main
where

{-
**** OPTIMIZING MEMORY USAGE IN THE POLYHEDRAL MODEL ****
Allocation method by Quillere, 
extended for imperative programs and implemented by Benjamin Schulte

What has to be done for each array:
1. Check if accesses are index monotone and return psi.
   Needed: Schedules, Accesses, ISPCs, Statements.
   Done by checkIndexMonotonity 
2. Get the maximum lifetime d for each variable.
   Needed: Usage set, Dependencies, Schedule.
   Done by getDepLifeTimeDTree, maxLT and help functions.

If maximum lifetime is unbounded, stop for this array.

3. Compute z-dach
   Needed: Schedules, Accesses, ISPCs, Statements, psi.
   Done by computeZDach
4. Compute the allocation vectors and matrix R.
   Needed: z-dach
   Done by getAllocationVects 
5f Compute matrices 'Pi', 'Omega', S.
   Needed: Matrix R.
   Done by calculateSOmega, calculatePi 
7. Show memory saving function 'Mem'.
   Needed: Matrices 'Pi', 'Omega', S.
   Done by showMemFunction 
8. Change LooPo's target code array accesses.
   Needed: Matrix 'Mem'
   Done by parseTreeModifier, modifyParseTree 

See my paper for further information.
-}

import QuilMem.QuilMem
import QuilMem.BetterSchedule
import Maybe
import Ratio
import Data.List
import CppInterface
import Foreign.C.String
import LoopoDS
import LoopoDS.Access
import LMath.Types
import LMath.Algo.Sort
import LMath.Algo.Simplex
import qualified Data.Map as Map
import Data.Map ( Map )
import CompSpread.Helpers
import Control.Exception ( evaluate )
import qualified TrDepDirections
import ScannerParser.ScannerParser  -- fuer Arraynamenanzeige
import ScannerParser.SymbolTable
import ScannerParser.ParseTree
import ScannerParser.PrintCode
import ScannerParser.FineCode
import System
import System.IO


optMem :: String -> String -> IO ()
optMem dir file = do 
        let ds = unsafeCreateCompSpreadDS dir file        
            newds = unsafeCreateCompSpreadDS dir file         -- nach Schedule-Aenderung
            curArrays =  map head (group (map (accArrayID) (csAccesses ds)))
        (symtab, _, _, _)    <- openParseFile  (dir ++ "/" ++ file ++ ".parse")

        putStrLn (unlines (zipWith (++) (map ((++ " ").show) [0..]) (map (arrayLexem symtab) curArrays)))
        putStr "Enter number of array to check: " 
        fn <- getLine
        let curArray = curArrays !! read(fn)

        putStrLn "\nLongest lifetime per target dependence: "        

        let lts = getArrayDepLifeTimeTrees ds LifeTimeMax curArray 
        putStrLn (showLifeTimeAnalysis ds lts)
        newlts <- (if (maxLT ds lts) /= Unbounded 
                   then (return lts)
                   else changeSched dir file curArray lts []
                  )

        let maxlt = (maxLT ds newlts)

        case maxlt of
         (Optimum m) -> do 
                         (newst, newpt, _, _) <- openParseFile  (dir ++ "/" ++ file ++ ".parse")
                         let zdach          = computeZDach jim qmss m
                             r              = getAllocationVects zdach
                             pi             = calculatePi r
                             som            = calculateSOmega r
                             im             = checkIndexMonotonity qmss
                             jim            = fromJust im
                             qmss           = getAllQMS newds curArray
                             (modpt, modst) = modifyPTST curArray r (newpt, renameSTInds newst (csQualVects ds))
                         putStrLn ("Maximal lifetime: " ++ (show m)++"\n")
                         putStrLn ("Checking index monotonity...")
                         case im of
                          Nothing -> putStrLn ("No index monotonity, cannot optimize memory usage.")
                          _       -> do
                                      putStrLn ("Index monotonity is " ++ (show im))
                                      putStrLn ("Allocation matrix:\n" ++ (show r))
                                      putStrLn ("Every occurence of the array can be changed the following way: ")
                                      putStrLn (show (showMemFunction (arrayLexem symtab curArray) pi som))
                                      putStrLn ("\n\nNew target code:")
                                      putStrLn (printFineCode modst modpt)
         _           -> putStrLn ("Maximal lifetime " ++ (show maxlt)++", cannot optimize memory usage.")

        return ()

main :: IO ()
main = do         
        hSetBuffering stdout NoBuffering
        args <- getArgs
        if (length args)==0 then showhelp else convert args
        return ()
       where
        showhelp = do 
                    putStrLn "Usage: QuilMem Directory File"
                    putStrLn "Directory: LooPo temp directory"
                    putStrLn "File:      Name of the project out file"
                    putStrLn "Please note: Targgen must be used before running this tool."
                    return ()
        convert args = do
                    let dir  = args!!0
                        file = args!!1
                    optMem dir file
                    return ()                


--- **** TEST FUNCTIONS ****
getFibDS = unsafeCreateCompSpreadDS "/home/cl/schulte/loopofibout" "fib.src.out"
getPasDS = unsafeCreateCompSpreadDS "/home/cl/schulte/loopopasout" "pascal.src.out"
array1 = (integerToID 1)::ArrayID
array2 = (integerToID 2)::ArrayID
tstmt1 = (integerToID 1)::TStatementID
tstmt2 = (integerToID 2)::TStatementID
tstmt3 = (integerToID 3)::TStatementID
tstmt4 = (integerToID 4)::TStatementID
