module QuilMem.QuilMem (
       IndexMonotonity(..), QMStatement, CompSpreadDSMini(..), unsafeCreateCompSpreadDS,          --Datentypen
       getAllQMS, getNoConstQMS,                                                              --Datenkonstruktoren
       checkIndexMonotonity ,                                                                 -- Indexmonotonitaet
       getDepLifeTimeDTree, maxLT,                                                            --Lebenszeitanalyse
       getArrayTrDeps, LifeTimeMinMax(..),
       getDepQualVect, getArrayDepLifeTimeTrees, showLifeTimeAnalysis, getTrDep, getQualVect,
       getKappaDif, computeZDach, getTimeDifISPC, getSmallElement,                            --Z-Dach-Berechnung
       getAllocationVects, 
       calculateSOmega, calculatePi,
       showAllocFunction, showMemFunction, parseTreeModifier, modifyPTST
       )
where

{-
**** OPTIMIZING MEMORY USAGE IN THE POLYHEDRAL MODEL ****
Allocation method by Quillere, 
extended for imperative programs and implemented by Benjamin Schulte

What has to be done for each array:
1. Check if accesses are index monotone and return psi.
   Needed: Schedules, Accesses, ISPCs, Statements.
   Done by checkIndexMonotonity 
2. Get the maximum lifetime d for each variable.
   Needed: Usage set, Dependencies, Schedule.
   Done by getDepLifeTimeDTree, maxLT and help functions.

If maximum lifetime is unbounded, stop for this array.

3. Compute z-dach
   Needed: Schedules, Accesses, ISPCs, Statements, psi.
   Done by computeZDach
4. Compute the allocation vectors and matrix R.
   Needed: z-dach
   Done by getAllocationVects 
5f Compute matrices 'Pi', 'Omega', S.
   Needed: Matrix R.
   Done by calculateSOmega, calculatePi 
7. Show memory saving function 'Mem'.
   Needed: Matrices 'Pi', 'Omega', S.
   Done by showMemFunction 
8. Change LooPo's target code array accesses.
   Needed: Matrix 'Mem'
   Done by parseTreeModifier, modifyParseTree, modifySymbolTable

See my paper for further information.
-}

import Maybe
import Ratio
import Data.List
import CppInterface
import Foreign.C.String
import LoopoDS
import LoopoDS.Access
import LMath.Types
import LMath.Algo.Sort
import LMath.Algo.Simplex
import LMath.Algo.OmegaTest
import qualified Data.Map as Map
import Data.Map ( Map )
import CompSpread.Helpers
import Control.Exception ( evaluate )
import qualified TrDepDirections
import ScannerParser.ParseTree
import ScannerParser.LambdaExpression
import ScannerParser.SymbolTable


-- **** GENERAL FUNCTIONS ****

-- von Michaels Compspread.Data
data CompSpreadDSMini = CompSpreadDSMini
    { csTempDir          :: String
    , csProjFile         :: String
    , csTrDeps           :: [TrDependence]
    , csISPCs            :: [Statement]
    , csTISPCs           :: [TStatement]
    , csTRCs             :: [TRC]
    , csTrafos           :: [(TStatementID, AffineFunction)]
    , csQualVects        :: [(TStatementID, [QualEntry])]
    , csAccesses         :: [Access]
    , csACPs             :: [AccessPair]
    , csSchedule         :: (Integer, Integer, Schedules)
    , csPlacement        :: (Integer, Integer, Placements StatementID)
    , csSLPs             :: SurroundingLoops
    } deriving Show

-- von Michaels Compspread.Data
unsafeCreateCompSpreadDS :: String -> String -> CompSpreadDSMini
unsafeCreateCompSpreadDS tempDir projFile =
    verbose "unsafeCreateCompSpreadDS" $
    CompSpreadDSMini
    { csTempDir = tempDir
    , csProjFile = projFile
    , csTrDeps = trDeps				--ok
    , csTrafos = trafos 				--ok
    , csQualVects = qualVects			--ok
    , csAccesses = accs
    , csTRCs = trcs
    , csACPs = acps
    , csSchedule = loopoSCHED loopoDS
    , csPlacement = loopoPLACE loopoDS
    , csISPCs = ispcs
    , csTISPCs = tispcs
    , csSLPs = slp
    }
    where    
    qualVects = loopoQUAL loopoDS
    loopoFile = tempDir ++ "/" ++ projFile			--ok
    loopoDS = 							--ok
	verbose "reading loopoDS" $
	unsafeReadLoopoDS loopoFile
    deps = map (trimDependence slp) (loopoDEP loopoDS)		--ok
    trDeps = transformDeps tispcs trafos deps			--ok
    trafos = makeProperTrafos (loopoTrafo loopoDS)		--ok
    ispcs = map (trimIndexSpace slp) (loopoISPC loopoDS)
    trcs = loopoTRC loopoDS
    tispcs = loopoTISPC loopoDS					--ok
    slp = loopoSLP loopoDS						--ok
    accs = map ((trimAccess slp). (rawAccessToAccess (loopoISPC loopoDS))) raccs
    raccs = loopoACC loopoDS
    acps = loopoACP loopoDS


data QMStatement = QMS
    { qmTStmtID          :: TStatementID
    , qmSchedule         :: AffineFunction
    , qmISPC             :: IneqSystem
    , qmAccess           :: AffineFunction
    } deriving Show

-- | returns the notconstant components of an affine function
getAFNotConst :: GAffineFunction a -> GMatrix a
getAFNotConst af = let m     = getMatrixIH af
                       nV    = nVars af
                in deleteColumns m nV ((nColumns m)-nV)

-- | returns the notconstant components of the schedule (\Tau_Y in the paper)
getQMSTau :: QMStatement -> GMatrix Rational
getQMSTau qms  = getAFNotConst (qmSchedule qms)

-- | returns the notconstant components of the access function (\mathcal{A}_Y in the paper)
getQMScalA :: QMStatement -> GMatrix Rational
getQMScalA qms = getAFNotConst (qmAccess qms)


-- | Returns for a given DS and ArrayID the list of QMStatements that work not only at a single point of time (paper: S_Y^{\neq 0})
getNoConstQMS ::  CompSpreadDSMini -> ArrayID -> [QMStatement]
getNoConstQMS ds aid = filter isQMSNotConst (getAllQMS ds aid)

-- | Checks whether a QMStatement works not only at a single point of time (paper: \tau_S \neq 0)
isQMSNotConst :: QMStatement -> Bool
isQMSNotConst qms = (sumM 0 (getQMSTau qms) > 0) where
                     sumM r m = (foldl (\sum c -> sum+(if (entry m r c)/=0 then 1 else 0)) 0 [0..((nColumns m)-1)])
                                + (if (r < nRows m-1) then (sumM (r+1) m) else 0)

-- | Returns for a given DS and ArrayID the list of all QMStatements (paper: S_Y)
getAllQMS :: CompSpreadDSMini -> ArrayID -> [QMStatement]
getAllQMS ds aid = concat $ map doublestmt (stmt_ispc_acc)
                where
                writeaccs = filter (\acc -> accType acc == Write && accArrayID acc == aid) (csAccesses ds)
                stmt_ispc_acc = map (\acc -> (accStmtID acc, accDomain acc, accIndexFct acc)) writeaccs
                doublestmt (stmtid, ispc, acc) = map (\(tid, m, vs) -> QMS {qmTStmtID = tid, 
                                                                       qmSchedule = affineFunction ((nVars ispc, nParams ispc),m) Inhomogeneous , 
                                                                       qmISPC = join ispc vs, 
                                                                       qmAccess = acc}) 
                                                     (getStatementSchedules ds stmtid)

-- | returns for a given Statement a list of its schedules and index space restrictions
getStatementSchedules :: CompSpreadDSMini -> StatementID -> [(TStatementID, GMatrix Rational, IneqSystem)]
getStatementSchedules ds stmtid = map getlist $ tstmtlist
                                  where
                                  tstmtlist    = filter (\trc -> trcStmtID trc == stmtid) (csTRCs ds) 
                                  getlist trc  = let tstmtid = (trcTargetStmtID trc) in
                                                 (tstmtid, sched tstmtid, trcConstraints trc)
                                  trafo tstmt  = case lookup tstmt (csTrafos ds) of
                                                 Just a -> a
                                                 _      -> error ((show tstmt) ++ " not found!")
                                  qv    tstmt  = case lookup tstmt (csQualVects ds) of
                                                 Just a -> a
                                                 _      -> error ((show tstmt) ++ " not found!")
                                  sched tstmt  = foldl (\m r -> deleteRows m r 1) (getMatrixIH (trafo tstmt)) (qvnotsch (qv tstmt))
                                  qvnotsch qvs = reverse $ map fst $ filter (\qv -> snd qv /= Sched) $ zip [0..] qvs


{- die folgenden Funktionen sind veraltete Versionen, die bei gesplitteten Schedules ungenau (aber nicht inkorrekt) arbeiten koennen
-- | Returns for a given DS and ArrayID the list of all QMStatements (paper: S_Y)
getAllQMS :: CompSpreadDSMini -> ArrayID -> [QMStatement]
getAllQMS ds aid = concat $ map doublestmt (stmt_ispc_acc)
                where
                writeaccs = filter (\acc -> accType acc == Write && accArrayID acc == aid) (csAccesses ds)
                stmt_ispc_acc = map (\acc -> (accStmtID acc, accDomain acc, accIndexFct acc)) writeaccs
                doublestmt (stmtid, ispc, acc) = map (\(m,vs) -> QMS {qmOrigStmtID = stmtid, 
                                                                      qmSchedule = affineFunction ((nVars ispc, nParams ispc),m) Inhomogeneous , 
                                                                      qmISPC = extendISPCSys ispc vs, 
                                                                      qmAccess = acc}) 
                                                     (getStatementSchedules ds stmtid)

-- | adds the information from a list of condition vectors to the index space
extendISPCSys :: IneqSystem -> [GVector Rational] -> IneqSystem 
extendISPCSys ispc vs = let n = length vs
                            ms = map (rowMatrix . vectorEntries) vs
                        in ineqSystem ((nVars ispc, nParams ispc),(foldl (\a b -> insertRows a 0 b) (getMatrix ispc) ms))

-- | returns for a given Statement a list of its schedules and index space restrictions
getStatementSchedules :: CompSpreadDSMini -> StatementID -> [(GMatrix Rational, [GVector Rational])]
getStatementSchedules ds stmtid = flattenQuast ((Map.!) sched stmtid)
                                  where
                                  (nV,nP, scheds) = csSchedule ds
                                  (Schedules sched) = trimSchedules (csSLPs ds) scheds

-- | flattens a given quast to a list of matrices and a list of condition vects
flattenQuast :: Quast a -> [(GMatrix a, [GVector a])]
flattenQuast (QList a)                     = [(a, [])]
flattenQuast (QIf vect_cond m_then m_else) =    (map (flattencond vect_cond) (flattenQuast m_then)) 
                                             ++                              (flattenQuast m_else)  -- *** negierte vect_cond hinzu!
                                             where
                                             flattencond cond (m, v) = (m, cond:v) -}

-- **** FUNCTIONS TO CHECK INDEX MONOTONITY (STEP 1) **** 

data IndexMonotonity = IMGrowing IMExt | IMFalling IMExt 
  deriving (Show, Eq)
data IMExt           = IMExtended     | IMNormal         
  deriving (Show, Eq)

isIMExtended :: IndexMonotonity -> Bool
isIMExtended im = case im of
                  IMGrowing IMExtended -> True
                  IMFalling IMExtended -> True
                  _                    -> False

roundVect :: Integral a => GVector Rational -> GVector a
roundVect rs = fmap ceiling rs

-- | Hilfsfunktion. Erhaelt eine Matrix mit Ungleichungen und spaltet diese so auf, dass der Vergleich lexikograpgisch ist
-- Die Rueckgabeliste muss mit Oder verknuepft werden.
-- geq: Bool, der angibt, ob der Vergleich >= oder > ist.
getLexiIESParts geq m i = join (join eqentr (negate eqentr)) grentr
                          where
                           eqentr    = subMatrix m 0 0 (i-1) ((nColumns m)-1)
                           grentrm   = subMatrix m i 0 i ((nColumns m)-1)
                           geqconst  = if (geq && i == (nRows m)-1) 
                                        then 0   -- beim letzten lexik. Vergleich ist >= erlaubt
                                        else 1   -- bis dahin muss > sein
                           grentr    = setEntry grentrm 0 ((nColumns grentrm)-1) ((entry grentrm 0 ((nColumns grentrm)-1))-geqconst)


-- | returns D_S^+, the subset of the complete index space with all points which runtime is later then a point in the original index space.
-- every list entry of the return value is a part of D_S^+
-- If IndexMonotonity is extended, then the index space with all possible times (greater equal 0) is used.
getDSPlus :: QMStatement -> [IneqSystem]
getDSPlus qms = map (makeineq.(getLexiIESParts True newschedm)) [0..(ts-1)]
                where
                 sched          = getMatrixIH (qmSchedule qms)
                 ts             = nRows sched
                 t1             = getEarliestTime qms                 
                 nC             = (nColumns sched)-1
                 newschedm      = foldl changem sched [0..((nRows sched)-1)]
                 changem m i    = setEntry m i nC ((entry m i nC)-(vectorEntry t1 i))
                 makeineq m     = ineqSystem (((nColumns m)-1,0),m)


-- | returns the earliest time point the given statement is launched.
getEarliestTime :: QMStatement -> GVector Rational
getEarliestTime qms = case (unZipLexMin $ makeDTree (numOrdDTreeContext ()) (lexmin stdSimplexParams ies Nothing)) of
                       Optimum a -> vector a
                       _         -> error("QuilMem.getEarliestTime: Unbounded earliest time!")
                      where
                       sched       = getMatrixIH (qmSchedule qms)
                       ispc        = qmISPC qms
                       vs          = nVars ispc
                       ps          = nParams ispc
                       ts          = nRows sched
                       schedm      = eqSystem ((vs+ps+ts,0),insertColumns sched 0 (negUnitMatrix ts ts))
                       ispcm       = ineqSystem ((vs+ps+ts,0),insertZeroColumns (getMatrix ispc) 0 ts)
                       ies         = maybe (error ("QuilMem.getEarliestTime: projectToVars gives Nothing"))
                                            id
                                            (projectToVars ispcm (Just schedm) [0..(ts-1)])

-- | gibt das kombinierte Ungleichungssystem fuer Teilsysteme von D_S^+ und D_T^+ zurueck
combineDSPlusDTPlus :: IneqSystem -> IneqSystem -> IneqSystem
combineDSPlusDTPlus dsp dtp = ineqSystem (((nColumns ndsp)-1,0),join ndsp ndtp)
                              where
                               dspm = getMatrix dsp
                               dtpm = getMatrix dtp
                               ndsp = insertZeroColumns dspm ((nColumns dspm)-1) ((nColumns dtpm)-1)
                               ndtp = insertZeroColumns dtpm 0                   ((nColumns dspm)-1)


-- | gibt fuer S und T Liste von Ungleichungssystemen zurueck, die zusammengenommen Ungleichungssystem fuer alle Punkte i_S, i_T
-- ergeben mit \Theta_S (i_S) < \Theta_T (i_T)
getSchedsIneqSyss :: QMStatement -> QMStatement -> [IneqSystem]
getSchedsIneqSyss sqms tqms = getGrEqIneqSyss False ssched tsched
                              where
                              ssched = getMatrixIH (qmSchedule sqms)
                              tsched = getMatrixIH (qmSchedule tqms)

-- | gibt fuer S und T Liste von Ungleichungssystemen zurueck, die zusammengenommen Ungleichungssystem fuer alle Punkte i_S, i_T
-- ergeben mit \Acc_S (i_S) >= \Acc_T (i_T)
getAccsIneqSyss :: IndexMonotonity -> QMStatement -> QMStatement -> [IneqSystem]
getAccsIneqSyss im sqms tqms = getGrEqIneqSyss True (imcs sacc) (imcs tacc)
                            where
                             sacc = getMatrixIH (qmAccess sqms)
                             tacc = getMatrixIH (qmAccess tqms)
                             imcs = case im of
                                    IMFalling _ -> id
                                    IMGrowing _ -> negate


-- | ergibt Ungleichungssytem (1.) < (2.) bzw., wenn geq == True, (1.) <= (2.)
getGrEqIneqSyss :: Bool -> GMatrix Rational -> GMatrix Rational -> [IneqSystem]
getGrEqIneqSyss geq m1 m2 = map (makeineq.(getLexiIESParts geq newm)) [0..(ts-1)]
                              where
                              ts          = nRows m1
                              tsvars      = subMatrix m2 0 0 (ts-1) ((nColumns m2)-2)
                              tsconsts    = subMatrix m2 0 ((nColumns m2)-1) (ts-1) ((nColumns m2)-1)
                              newm        = foldl changem (insertColumns (negate m1) ((nColumns m1)-1) tsvars) [0..(ts-1)]
                              nC          = (nColumns m1)+(nColumns m2)-2
                              changem m i = setEntry m i nC ((entry m i nC)+(entry tsconsts i 0))
                              makeineq m  = ineqSystem ((nC,0),m)


-- | prueft die Indexmonotonitt fr zwei gegebene Statements
-- Rueckgabewert ist negiert!
testQMSIM :: IndexMonotonity -> QMStatement -> QMStatement -> Bool
testQMSIM im sqms tqms = or $ map solvable allcross
                      where
                       dsps        = getDSPlus sqms
                       dtps        = getDSPlus tqms
                       dcross      = cross combineDSPlusDTPlus dsps dtps
                       schieqs     = getSchedsIneqSyss sqms tqms
                       accieqs     = getAccsIneqSyss im sqms tqms
                       sacross     = cross join schieqs accieqs
                       allcross    = if isIMExtended im
                                      then sacross
                                      else cross join dcross sacross
                       solvable m  = omegaTest (Just m) Nothing {-isJust (projectToVars m Nothing [])-}


-- | gibt zurueck, ob alle Statement-Paare indexmonoton wie bergeben sind.
testIM :: IndexMonotonity -> [QMStatement] -> Bool
testIM im qmss = not $ or $ map (\(i,j) -> {- (qmTStmtID i, qmTStmtID j,-} testQMSIM im i j --)
                              ) (cross (\i j -> (i,j)) qmss qmss)

-- | findet heraus, ob IndexMonotonitt besteht
checkIndexMonotonity :: [QMStatement] -> Maybe IndexMonotonity
checkIndexMonotonity qms = case imgn of
                            True  -> case imge of
                                      True  -> Just (IMGrowing IMExtended)
                                      False -> Just (IMGrowing IMNormal)
                            False -> case imfn of
                                      True  -> case imfe of
                                                True  -> Just (IMFalling IMExtended)
                                                False -> Just (IMFalling IMNormal)
                                      False -> Nothing
                           where
                            imgn = testIM (IMGrowing IMNormal)   qms
                            imge = testIM (IMGrowing IMExtended) qms
                            imfn = testIM (IMFalling IMNormal)   qms
                            imfe = testIM (IMFalling IMExtended) qms


cross :: (a -> b -> c) -> [a] -> [b] -> [c]
cross f a b = concat $ map (\i -> map (\j-> f i j) b) a

-- **** FUNCTIONS TO EVALUATE THE MAXIMUM LIFETIME FOR EACH ARRAY (STEP 2) ****

data LifeTimeMinMax = LifeTimeMin | LifeTimeMax deriving (Eq, Show)

-- | gibt alle TargetDependences zu einer bestimmten ArrayID zurueck
-- Fuer die Maximum-Lebenszeitberechnung werden nur Flow/True Dependences herangezogen; 
-- fuer Minumum-Berechnung dagegen auch Output-Dependences.
getArrayTrDeps :: CompSpreadDSMini 
                  -> LifeTimeMinMax 
                  -> ArrayID 
                  -> [TrDependence]
getArrayTrDeps ds max arrayID = filter isNeededDep (csTrDeps ds) where
                            isNeededDep dep = (getTrDepArray ds dep) == (Just arrayID) 
                                              && (case trDepType dep of
                                                  UniformTrue      -> True
                                                  UniformFlow      -> True
                                                  NonUniformTrue   -> True
                                                  NonUniformFlow   -> True
                                                  UniformOutput    -> (max == LifeTimeMin)
                                                  NonUniformOutput -> (max == LifeTimeMin)
                                                  _                -> False 
                                                 )

-- | gibt zurueck, welches Array zu einer TargetDependence gehoert
getTrDepArray :: CompSpreadDSMini -> TrDependence -> Maybe ArrayID
getTrDepArray ds trdep = fmap (accArrayID) (findAcc m_accID)
				  where
				  depID = trDepOrigID trdep
				  acp   = find (\acp -> acpDepID acp == depID) (csACPs ds)
				  m_accID = fmap acpSrcAccID acp -- ginge auch mit acpTrAccID
				  findAcc Nothing      = Nothing
				  findAcc (Just curAccID) = find (\acc -> ((accID acc) == curAccID)) (csAccesses ds)
				  

-- | Setzt die Gleichungen in die Matrix ein, die die Differenz zwischen Source und Target-Positionen definieren
--   und projiziert Ungleichungssystem darauf
-- Ergebnismatrix: ( (tr1-src1), ..., (tr_n-src_n), params, 1) >= 0
-- !!! effizienter waere Erzeugung von EqSystems und uebergabe an projectToVars, lexMax!
-- !!! Additional-Dimensionen geloescht werden
insertSourceTargetDiffEqu :: (TrDependence, ([QualEntry], [QualEntry])) -> Maybe IneqSystem
insertSourceTargetDiffEqu (trDep, (qvs, qvt)) = projectToVars (fromAffineExp newAE) Nothing srcQVectPos  
					where
					 relMatrix = toAffineExp (fst (trDepRelation trDep)) -- :t GAffineExp
					 nv = nVars relMatrix
					 np = nParams relMatrix
					 -- nNewVars = (div nv 2)
					 srcQVectBool = map (\qv -> (qv==Sched {- || qv == Place -})) qvs
					 trgQVectBool = map (\qv -> (qv==Sched {- || qv == Place -})) qvt
					 srcQVectPos  = snd $ unzip $ filter (\x -> fst x) $ zip srcQVectBool [0..]
					 trgQVectPos  = snd $ unzip $ filter (\x -> fst x) $ zip trgQVectBool [(length qvs)..]
					 spVect = snd $ unzip $ filter (\x -> fst x) $ zip srcQVectBool qvs
                                   -- im Weiteren wird sich darauf verlassen, dass die QualVects mit Ausnahme der nun geloeschten
                                   -- Additionals gleich sind; dies wurde schon in getDepQualVect ueberprueft
					 nNewVars =  foldl (\n sched -> if sched then (n+1) else n) 0 srcQVectBool 
					 nm = zeroMatrix (nRows relMatrix) nNewVars
					 withNewCols = insertColumns (getMatrix relMatrix) 0 nm 
						-- GAffineExp muss zerstoert werden da kein Zugriffsfunktionen
					 withNewRows = join (zeroMatrix (nNewVars*2) (nv+np+nNewVars+1)) withNewCols 
						-- buggy:  insertZeroRows withNewCols  0 (nNewVars*2)
					 withEqu = makeEqu withNewRows 0 nNewVars nv srcQVectPos trgQVectPos spVect 
					 newAE = affineExp ((nv+nNewVars+np,0), withEqu) -- und wieder zur GAffineExp machen, aber keine Parameter mehr

-- | Hilfsfunktion fuer insertSourceTargetDiffEqu 
makeEqu :: Integral b => GMatrix Rational -> b -> b -> b-> [b] -> [b] -> [QualEntry] -> GMatrix Rational
makeEqu matr var nNew nVar []      []          _        = matr
makeEqu matr var nNew nVar (sp:spos) (tp:tpos) (qe:qes) = newMatr where 
				   id1     = setEntry matr    (var*2)    var                1
				   diff1_a = setEntry id1     (var*2)    (nNew +sp)         1
				   diff1_b = setEntry diff1_a (var*2)    (nNew+tp)        (-1)
				   id2     = case qe of
				             Sched -> setEntry diff1_b (var*2+1)  var              (-1)
				             Place -> setEntry id1     (var*2+1)  var              (-1) -- Placement wird auf 0 gesetzt
				   diff2_a = setEntry id2     (var*2+1)  (nNew +sp)       (-1)
				   diff2_b = setEntry diff2_a (var*2+1)  (nNew+tp)          1					
				   newMatr = case qe of
				             Sched -> makeEqu  diff2_b (var+1) nNew nVar spos tpos qes
				             Place -> makeEqu  id2     (var+1) nNew nVar spos tpos qes  -- s.o.
				   -- srcVars = (div nVar 2)
makeEqu matr var nNew nVar _       _           _         = error("QualVects not the same!")

-- | gibt Lebenszeit-Entscheidungs"baum" einer TrDependence zurueck
getDepLifeTimeDTree :: LifeTimeMinMax 
                       -> (TrDependence, ([QualEntry], [QualEntry])) 
                       -> (TrDependenceID, Optimum [Rational])
getDepLifeTimeDTree max (trdep, qvs) = case max of 
                                       LifeTimeMax -> (trid, ceilOptVect lt) -- nur fuer Maximumsberechnung anwendbar
                                       LifeTimeMin -> (trDepID trdep, (unLeaf.fromJust) (fmap getDTree  (insertSourceTargetDiffEqu (trdep, qvs))))
                                       where
                                        (trid, lt) = getDepLifeTimeDimDTree max (trdep, qvs)
                                        getDTree ineqsys = makeDTree (numOrdDTreeContext ()) (case max of 
                                                             LifeTimeMax -> (lexmax stdSimplexParams ineqsys Nothing)
                                                             LifeTimeMin -> (lexmin stdSimplexParams ineqsys Nothing)
                                                            )
                                        unLeaf (DLeaf a) = a
                                        unLeaf x         = error("QuilMem.getDepLifeTimeDTree: DTree.Leaf expected, found: "++ (show x))

-- | Hilfsfunktion
ceilOptVect :: [Optimum [Rational]] -> Optimum [Rational]
ceilOptVect (Unbounded:_) = Unbounded
ceilOptVect c             = Optimum (ceilList (map (fmap head) c)) where
				 ceilList []                    = []
				 ceilList [(Optimum a)]         = [a]
				 ceilList ((Optimum a):tb:tcs)  = case tb of 
				                                   (Optimum b) -> (a:(ceilList (tb:tcs)))
				                                   Unbounded   -> ((a+1):(replicate ((length tcs)+1) 0))

-- | gibt Lebenszeit-Entscheidungsbaum jeder Dimension einer TrDependence zurueck
getDepLifeTimeDimDTree :: LifeTimeMinMax 
                       -> (TrDependence, ([QualEntry], [QualEntry])) 
                       -> (TrDependenceID, [Optimum [Rational]])
getDepLifeTimeDimDTree max (trdep, qvs) = (trDepID trdep, map dimlt [0..((nVars ies)-1)])
						where
						dimlt i          = getDimLT (projectToVars ies Nothing [i])
						ies              = fromJust (insertSourceTargetDiffEqu (trdep, qvs))
						getDimLT ineqsys = (unLeaf.fromJust) (fmap getDTree ineqsys)
						getDTree ineqsys = makeDTree (numOrdDTreeContext ()) (case max of 
                                                             LifeTimeMax -> (lexmax stdSimplexParams ineqsys Nothing)
                                                             LifeTimeMin -> (lexmin stdSimplexParams ineqsys Nothing)
                                                            )
						unLeaf (DLeaf a) = a
						unLeaf x         = error("QuilMem.getDepLifeTimeDTree: DTree.Leaf expected, found: "++ (show x))


-- | Lebenszeit-Entscheidugns"baeume" fuer ein spezielles Array
getArrayDepLifeTimeTrees :: CompSpreadDSMini 
                            -> LifeTimeMinMax
                            -> ArrayID 
                            -> [(TrDependenceID, Optimum [Rational])]
getArrayDepLifeTimeTrees ds max arr =  zip trDeps dTrees
                                   where
                                   (trDeps, dTrees) = unzip $ map (getDepLifeTimeDTree max) qualvects
                                   qualvects = (map (\trdep -> (trdep, getDepQualVect ds trdep)) (getArrayTrDeps ds max arr))
                                   noadditionals = filter (\qv -> qv /= Additional) $ fst $ snd $ (head qualvects)

-- | gibt den QualVect zu einem TStatement zurueck
getQualVect :: CompSpreadDSMini -> TStatementID -> Maybe [QualEntry]
getQualVect ds tstmt = lookup tstmt (csQualVects ds)

-- | gibt die QualVects zu einer Target-Dependence zurueck
getDepQualVect :: CompSpreadDSMini -> TrDependence -> ([QualEntry], [QualEntry])
getDepQualVect ds dep = if (rTrimAdditionals m_srcQv == rTrimAdditionals m_destQv) then (m_srcQv, m_destQv)
			   else error("QuilMem.getDepQualVect: Source and destination target dimensions are not the same (" 
			   ++ (show m_srcQv) ++ ", " ++ (show m_destQv) ++ ")")
				where
				m_srcQv  = unjust $ getQualVect ds (trDepSrcStmtID dep)
				m_destQv = unjust $ getQualVect ds (trDepDestStmtID dep)
				unjust (Just a) = a
				unjust _        = error("QuilMem.getDepQualVect: No QualVect found for " ++ show (trDepSrcStmtID dep))
				rTrimAdditionals []     = []
				rTrimAdditionals (x:xs) | (x==Additional) && (notElem Sched xs) && (notElem Place xs) = []
				rTrimAdditionals (x:xs) | otherwise                                                   = x:(rTrimAdditionals xs)

maxLT :: Integral b => CompSpreadDSMini -> [(a1, Optimum [Rational])] -> Optimum (GVector b)
maxLT ds []  = Optimum (vector ((replicate ((getSchedDim ds)-1) 0)++[1]))
maxLT _  lts = fmap roundVect $ fmap vector $ maximumBy myCompareOpt $ map (snd) $ lts
            where
            myCompareOpt :: Optimum [Rational] -> Optimum [Rational] -> Ordering
            myCompareOpt (Optimum x) Unbounded   = if isNotNeg x then LT else negErr x
            myCompareOpt (Optimum x) (Optimum y) = if (isNotNeg x && isNotNeg y) then compare x y else negErr x
            myCompareOpt Unbounded (Optimum x)   = if isNotNeg x then GT else negErr x
            myCompareOpt Unbounded Unbounded     = EQ
            myCompareOpt _         _             = error("Nonexistent value.")
            isNotNeg    x                        = x >= (replicate (length x) (0%1))
            negErr      x                        = error ("BetterSchedule.MaxLT: Negative lifetime!")            

getSmallestNotNullVect :: Integral a => Int -> [a]
getSmallestNotNullVect ndim = (replicate (ndim-1) 0) ++ [1]

getSchedDim :: CompSpreadDSMini -> Int
getSchedDim ds = case (csQualVects ds) of
                 []             -> 1
                 ((stmt,qv):_)  -> length (filter (\qe -> qe == Sched) qv)



-- **** FUNCTIONS TO COMPUTE Z-DACH (STEP 3) ****

-- | returns the smallest cell index a statement writes in
-- (paper: Acc_S(D_S)
getSmallestWrittenCell :: IndexMonotonity -> QMStatement ->  GVector Rational
getSmallestWrittenCell im qms = case (unZipLexMin $ makeDTree (numOrdDTreeContext ()) (case im of 
                                                             IMFalling _ -> (lexmax stdSimplexParams ineqsys Nothing)
                                                             IMGrowing _ -> (lexmin stdSimplexParams ineqsys Nothing)
                                                            )) of
                                 Optimum a -> vector a
                                 _         -> error("QuilMem.getSmallestWrittenCell: No smallest array cell computable; ineqsys: "++(show ineqsys))
                             where
                             (Just ineqsys) = getWrittenCells qms

unZipLexMin ::  DTree Rational (Optimum [Rational]) -> Optimum [Rational]
unZipLexMin = unLeaf where
               unLeaf (DLeaf a) = a
               unLeaf x         = error("QuilMem.unZipLexMin: DTree.Leaf expected, found: "++ (show x))

-- | returns ineqsystem containing the *array* index space of a given statement
-- (paper: Acc_S(D_S)
getWrittenCells :: QMStatement -> Maybe IneqSystem
getWrittenCells qms = projectToVars (ineqSystem ((arraydim + nVars ispc + nParams ispc,0),d))
                       (Just (eqSystem ((arraydim + nVars ispc + nParams ispc,0),insertColumns acc 0 (negUnitMatrix arraydim arraydim))))
                       [0..(arraydim-1)]
                      where
                      ispc = qmISPC qms
                      d = insertZeroColumns (getMatrix ispc) 0 arraydim
                      acc = getMatrixIH (qmAccess qms)
                      arraydim = nRows acc

-- | returns the by all statements smallest written array cell and the by not constant statements written array cell
--  paper: (\kappa_Y^{\neq 0} - \kappa_Y)
getKappaDif :: IndexMonotonity -> [QMStatement] -> GVector Rational
getKappaDif im qms = subVV (kappa  sc_nc) (kappa sc_c)
                   where 
                    sc_ncb = map (\qm -> (getSmallestWrittenCell im qm, isQMSNotConst qm)) qms -- [(Smallest Cell, IsNotConstant)]
                    sc_nc  = map fst (filter snd sc_ncb)  -- smallest cells of not constant statements
                    sc_c   = map fst sc_ncb                 -- smallest cells of all statements
                    kappa  = case im of
                             IMGrowing _ -> minimum
                             IMFalling _ -> negV.maximum


-- | returns all entries of full index space dimension for which notconstant part of schedule is greater or equal than the given vector.
-- given vector must be greater 0!
-- (Paper: \mathcal{I}_S(\Delta t))
getTimeDifISPC :: GVector Rational -> QMStatement -> IneqSystem
getTimeDifISPC dt qms = ineqSystem ((nColumns tau,0),nC)
                        where
                         tau = getQMSTau qms
                         nC  = insertColumns tau (nColumns tau) (columnMatrix (vectorEntries (negV dt)))


-- | returns all array indexes for which the notconstant part of schedule is greater or equal than the given vector.
-- given vector must be greater 0!
-- (Paper: \mathcal{A}_S(\mathcal{I}_S(\Delta t)))
getCellDifSPC :: IndexMonotonity -> GVector Rational -> QMStatement -> Maybe IneqSystem
getCellDifSPC im dt qms = projectToVars dt_ispc_nc (Just eqSysA) [0..(arraydim-1)]
                       where
                        calA       = getQMScalA qms
                        arraydim   = nRows calA
                        dt_ispc    = getTimeDifISPC dt qms
                        dt_ispc_nc = ineqSystem ((arraydim+(nVars dt_ispc), nParams dt_ispc),
                                                 insertZeroColumns (getMatrix dt_ispc) 0 arraydim)
                        calAwConst = insertZeroColumns calA (nColumns calA) 1
                        eqSysA     = eqSystem ((arraydim+(nColumns calA),0),
                                               (insertColumns calAwConst 0 (psiUM arraydim arraydim)))
                        psiUM      = case im of
                                     IMGrowing _ -> negUnitMatrix 
                                     IMFalling _ -> unitMatrix


-- | getCellDifSPC liefert das Polytop zurueck, dass alle Arrayzellen beschreibt, die nach oder zu einer bestimmten Zeitdifferenz
-- geschrieben werden. z-check soll ein moeglichst kleines bzw. grosses Element in der Menge sein. Die folgende Funktion versucht, 
-- solch eines zu finden. lexMin/lexMax ist nicht in jedem Fall hilfreich, da die Menge nicht immer beschraenkt ist und Unbounded 
-- zurueckgeliefert werden kann.
-- 
-- ineqsystem must have no params!!!
getSmallElement :: QMStatement -> IneqSystem -> GVector Rational
getSmallElement qms ies = case (unZipLexMin $ makeDTree (numOrdDTreeContext ()) 
                                (lexmin stdSimplexParams ineqsys Nothing)) of
                                 Optimum a -> vector a
                                 _         -> error("QuilMem.getSmallElement  No small array cell computable; ineqsys: "++(show ineqsys))
                          where
                           indexsteps = getSmallEl 0 ies Nothing
                           ism        = columnMatrix (vectorEntries indexsteps)
                           nIs        = nRows ism
                           calA       = getQMScalA qms
                           zs         = nRows calA
                           calAm      = insertColumns (insertZeroColumns calA (nColumns calA) 1) 0 (negUnitMatrix zs zs)
                           newism     = insertZeroColumns (insertColumns ism 0 (negUnitMatrix nIs nIs)) 0 (zs)
                           togeth     = join newism calAm
                           smallz     = ineqSystem (((nColumns calAm)-1,0),join togeth (negate togeth))
                           ineqsys    = case projectToVars smallz Nothing [0..(zs-1)] of
                                         Just a -> a
                                         _      -> error("QuilMem.getSmallElement  No small array cell computable!")

-- | Hilfsfunktion fuer getSmallElement. Liefert moeglichst kleine Indexschrittmenge
getSmallEl :: Integral a=> a -> IneqSystem -> Maybe EqSystem -> GVector Rational
getSmallEl i ies   _ | i>=(nVars ies) = zeroVector (fromIntegral i)
getSmallEl i ies es = case ies_entry of
                      (Just _) -> setVectorEntry (getSmallEl (i+1) ies (Just neweqsys)) i new_entry
                      Nothing  -> error ("QuilMem.getSmallEl: projevtToVars returned nothing") -- zeroVector (nVars ies)
                      where
                       ies_entry     = projectToVars ies es [i]
                       new_entry     = case ies_entry of
                                       Nothing -> error("QuilMem.getSmallEl: This pattern cannot be reached")
                                       (Just newies) -> let (minel,maxel)=getSmEn newies
                                                        in case minel of
                                                            (Optimum _) -> if ((optvalue minel)<0)
                                                                            then aSmallerNull maxel 
                                                                            else optvalue minel -- >=0
                                                            Unbounded   -> aSmallerNull maxel 
                       aSmallerNull maxel = if (maxel==Unbounded || mx >= 0) 
                                             then 0
                                             else mx 
                                            where
                                             mx = optvalue maxel
                       getSmEn nies = (unZipLexMin $ makeDTree (numOrdDTreeContext ()) 
                                                         (lexmin stdSimplexParams nies Nothing)
                                       ,unZipLexMin $ makeDTree (numOrdDTreeContext ()) 
                                                         (lexmax stdSimplexParams nies Nothing))
                       optvalue v   = case v of
                                      (Optimum a) -> head a
                                      _           -> error("QuilMem.getSmallEl: optvalue with Unbounded!")
                       oldeqsys     = case es of 
                                      (Just esys) -> getMatrix esys
                                      Nothing     -> zeroMatrix 0 ((nVars ies)+1)
                       neweqsys     = eqSystem ((nVars ies,0),
                                        setEntry (setEntry (insertZeroRows oldeqsys 0 1) 
                                           0 i (-1)) 0 ((nColumns oldeqsys)-1) new_entry)


-- | returns \check{Z} as defined in my paper
-- needs indexmonotonity and greatest lifetime vector
computeZDach :: Integral a => IndexMonotonity -> [QMStatement] -> GVector a -> GVector a 
computeZDach im sY dY = roundVect zdachratio
                        where
                         ratiodY     = fmap (fromIntegral) dY
                         sYneqNull   = filter isQMSNotConst sY
                         noctimes    = map (roundVect.getEarliestTime) sYneqNull 
                         ksYneqNull  = filter (not.isQMSNotConst) sY -- Komplementaermenge
                         ctimes      = map (roundVect.getEarliestTime) ksYneqNull 
                         kappadif    = getKappaDif im sY
                         zchecks     = map (\qms -> getSmallElement qms (getTimeDifISPC ratiodY qms)) sYneqNull
                         noconsttog  = and $ map ((==) (head noctimes)) noctimes -- kein nichtkonstantes Statement wird allein ausgefuehrt
                         nocnctog    = and $ cross (<=) (map (addVV dY) ctimes) noctimes -- kein Konstantes und Nichtkonstantes gemeinsam (DA 3.7.2)
                         maxzcheck   = if (isIMExtended im || noconsttog) 
                                        then minimum zchecks -- bei erw. Indexmonotonitaet oder wenn kein nichtkonstantes allein (s. DA 3.7.1)
                                        else maximum zchecks
                         zdachratio  = if ((length sYneqNull)>0) 
                                        then if (length sY) > (length sYneqNull) && (not (isIMExtended im))
                                              then if nocnctog 
                                                    then max   kappadif maxzcheck
                                                    else addVV kappadif maxzcheck 
                                              else maxzcheck -- gibt es keine konstanten Statements, ist die Berechnung von kappa nicht noetig
                                        else error("QuilMem.computeZDach: Not defined for arrays that are only written to on constant time steps.")


-- **** FUNCTIONS TO COMPUTE THE ALLOCATION VECTORS (STEP 4) ****

-- | Berechnet die Allocation Vector Matrix R. 
--   Einheitsmatrix aufstellen, 
--   und nur die letzte Spalte durch die uebergebenes z-dach ersetzen.
getAllocationVects :: GVector Integer -- ^ z-dach
                   -> GMatrix Integer
getAllocationVects d = conc (unitMatrix (vectorLength d) (m-1)) (columnMatrix (vectorEntries d))
                       where
                       m = getLevelOfVector d


-- | Berechnet Grad eines Vektors (die erste Stelle, an der er ungleich 0 ist)
getLevelOfVector:: (Num b, Integral a) => GVector b -> a
getLevelOfVector d = getM 1 (vectorEntries d) where
                       getM i []                 = i-1
                       getM i (x:xs) | x == 0    = getM (i+1) xs
                                     | otherwise = i



-- **** FUNCTIONS TO COMPUTE MATRICES 'Pi', 'Omega', S (STEP 5/6) ****

-- | Berechnet S und Omega wie in Quilleres Paper beschrieben
calculateSOmega :: Integral a => GMatrix a -> ([a]               -- ^ (s_1, s_2, ..., s_my)
                                               , GMatrix a       -- ^ Omega
                                              )
calculateSOmega m = (s, omega)
                    where
                    (nu, _, d) = diagMatrix m
                    ns = map (\i -> entry d i i) [0..((min (nRows d) (nColumns d))-1)]
                    s = map abs ns
                    u = foldl (\m i -> multiplyRow m i (-1)) nu (findIndices (<0) ns)
                    omega = subMatrix u 0 0 ((min (nRows u) (length s))-1) ((nColumns u)-1)

-- | Zu einer gegebenen Matrix A wird ein Tupel (U, V, D) geliefert mit UAV = D
-- Algorithmus aus Martins Buch '***'
diagMatrix :: Integral a => GMatrix a -> (GMatrix a, GMatrix a, GMatrix a)
diagMatrix a = diagStep 0 u v a where
               u = unitMatrix m m
               v = unitMatrix n n
               n = nColumns a
               m = nRows a
               diagStep k u v d | k >= n || k >= m                       = (u,v,d)
               diagStep k u v d | (smallest d 0 0 0 Nothing ) == Nothing = (u,v,d)
               diagStep k u v d | otherwise                              = let (Just (p,q)) = (smallest d 0 0 0 Nothing ) 
                                                                               (nu,nv,nd)   = echolonColumns k (k+1) $ 
                                                                                              echolonRows k (k+1)
                                                                                              (interchangeRows u k p,
                                                                                               interchangeColumns v k q,
                                                                                               interchangeColumns (interchangeRows d k p) k q)
                                                                            in 
                                                                            if ((hasNonZeroInRow nd k (k+1)) || (hasNonZeroInColumn nd k (k+1)))
                                                                            then diagStep k nu nv nd
                                                                            else diagStep (k+1) nu nv nd                                                                           
               hasNonZeroInRow d k i | i<n       = (entry d k i) /= 0 || (hasNonZeroInRow d k (i+1))
               hasNonZeroInRow d k i | otherwise = False
               hasNonZeroInColumn d k j | j<m       = (entry d j k) /= 0 || (hasNonZeroInColumn d k (j+1))
               hasNonZeroInColumn d k j | otherwise = False
               smallest d i j curSm curSmPos | i<m && j<n && (smallerEntry curSm (entry d i j)) = smallest d i (j+1) (entry d i j) (Just (i,j))
               smallest d i j curSm curSmPos | i<m && j<n                                       = smallest d i (j+1) curSm curSmPos
               smallest d i j curSm curSmPos | i<m && j==n                                      = smallest d (i+1) 0 curSm curSmPos
               smallest d i j curSm curSmPos | i==m                                             = curSmPos
               smallerEntry _ 0 = False
               smallerEntry 0 a = True
               smallerEntry a b = (b<a)

-- | Hilfsfunktion zu diagMatrix
echolonRows :: (Integral b, Integral a) => a -> a -> (GMatrix b, GMatrix b, GMatrix b) -> (GMatrix b, GMatrix b, GMatrix b)
echolonRows k i (u,v,d) | i<(nRows d) = let s = signum ((entry d k k) * (entry d i k))
                                            z = div (entry d i k) (entry d k k)
                                        in echolonRows k (i+1) (subtractRow u k i (s*z),v,subtractRow d k i (s*z))
echolonRows k i (u,v,d) | otherwise   = (u,v,d)

-- | Hilfsfunktion zu diagMatrix
echolonColumns :: (Integral b, Integral a) => a -> a -> (GMatrix b, GMatrix b, GMatrix b) -> (GMatrix b, GMatrix b, GMatrix b)
echolonColumns k j (u,v,d) | j<(nColumns d) = let s = signum ((entry d k k) * (entry d k j))
                                                  z = div (entry d k j)  (entry d k k)
                                              in echolonColumns k (j+1) (u,subtractColumn v k j (s*z),subtractColumn d k j (s*z))
echolonColumns k j (u,v,d) | otherwise      = (u,v,d)


-- | Berechnet Matrix 'Pi', so dass 'Pi'*'R'=(0). Besondere Anforderungen:
-- Alle Zeilen von 'Pi' muessen linear unabhaengig und ungleich 0 sein, und anderes
-- (siehe Quilleres Paper).
-- Algorithmus ist eine Eigenkreation, Erklaerung dann in meiner Diplomarbeit.

calculatePi :: GMatrix Integer -> GMatrix Integer
calculatePi r = calcPi 0 (zeroMatrix (n-m) n) where
                n = nRows r
                m = nColumns r
                r_m = (columns' r)!!(m-1)
                calcPi p pi | p < (n-m) && 
                              (r_m!!(p+m-1)) == 0 = calcPi (p+1) (setEntry pi p (m+p-1) (1))
                            | p < (n-m) && 
                              (r_m!!(p+m-1)) /= 0 = newpi
                                                    where
                                                    ls = filter (\i -> i>(p+m-1)) $ findIndices (>0) r_m
                                                    newpi = case ls of
                                                            []    -> foldl (\mr k -> setEntry mr k (m+k) 1) pi [0..(n-m-1)]
                                                            (l:_) -> calcPi (p+1) $ 
                                                                     setEntry (setEntry pi p (m+p-1) j') p l ((-1)*i')
                                                                     where
                                                                     i = r_m!!(m+p-1)
                                                                     j = r_m!!l
                                                                     ggt = gcd i j
                                                                     i' = div i ggt 
                                                                     j' = div j ggt 
                calcPi p pi | otherwise           = pi


-- **** FUNCTIONS TO SHOW ALLOCATION FUNCTION 'Mem' (STEP 7) ****

modifyPTST :: ArrayID -> GMatrix Integer -> (LambdaTree, SymTab) -> (LambdaTree, SymTab)
modifyPTST aid r (pt,st) = let modify = parseTreeModifier aid newpis newoms
                           in (map (modifyAllNodes modify) pt, modifySymbolTable aid newpis newoms st)
                           where
                            pi     = calculatePi r
                            (s,om) = calculateSOmega r
                            newpis = makePiMatrix pi
                            newoms = makeSOmMatrix om s

-- | modifies the number of indices for the 
modifySymbolTable :: ArrayID -> [[Integer]] -> [([Integer], Integer)] -> SymTab -> SymTab
modifySymbolTable aid pis oms st = modifyLNSymbols modLNs st
                        where
                         newdims      = (length pis) + (length oms)
                         modLNs lns   = case sym_type lns of 
                                         LNS_Array -> lns{sym_indices = newdims}
                                         _         -> lns


-- | returns modify functions for parse tree and symbol table
-- Please note: all node ids are zero! All statement and target statement ids are zero!
parseTreeModifier :: ArrayID -> [[Integer]] -> [([Integer], Integer)] -> LambdaNode -> LambdaNode -> LambdaNode
parseTreeModifier aid pis oms  lnorg lnnew = if (hasindices lnnew && name lnnew == integerFromID aid)
                                              then modifyPT pis oms lnnew
                                              else lnnew

-- | modifyer function for modifyMainNodes
-- Please note: all node ids are zero! All statement and target statement ids are zero!
modifyPT :: [[Integer]] -> [([Integer], Integer)] -> LambdaNode -> LambdaNode
modifyPT pis oms ln = ln{indices=projinds ++ pseudoinds}
                          where 
                           les          = map (fromJust.ln2LaExp) (indices ln)
                           projinds     = makeNewIndices les pis
                           pseudoinds   = map (\(ln,sy) -> lnMod 0 0 ln sy) $ 
                                          zip (makeNewIndices les (map fst oms)) (map snd oms)

unusedids :: [Int] -> LambdaTree -> [Int]
unusedids ids lt = [n | n <- ids, not (elem n usednids)] where
                    usednids = allNodeIDs lt

-- | wandelt Pi in von makeNewIndices genutztes Format um
makePiMatrix :: GMatrix Integer -> [[Integer]]
makePiMatrix pi = filter (\r -> not (and (map (== 0) r))) (rows' pi)

-- | wandelt Omega und S in von makeNewIndices genutztes Format um
makeSOmMatrix :: GMatrix Integer -> [Integer] -> [([Integer], Integer)]
makeSOmMatrix om s = filter (\(r,sy) -> not (and (map (== 0) r)) && sy /= 1) (zip (rows' om) s)

-- | erzeugt die neuen Indizes
makeNewIndices :: [LaExp] -> [[Integer]] -> [LambdaNode]
makeNewIndices les mrs = map ((laExp2LN (repeat 0)).(foldl (+) (laExpConstant 0)).(makeNewIndex).(zip les)) mrs

-- | erzeugt einen neuen Indexteil
makeNewIndex :: [(LaExp, Integer)] -> [LaExp]
makeNewIndex = map (\(le,mr) -> multLaExp (fromIntegral mr) le)

-- | fgt den Mod-Teil hinzu
lnMod :: Integral a => Int -> Int -> LambdaNode -> a -> LambdaNode
lnMod id1 id2 ln i = Expression {nodeid=id1, occurence=0, statement=(integerToID 0), targetStatement=(integerToID 0), 
                                 operator=LNOp_Mod, condition=Nothing, firstOperand=Just ln, 
                                 secondOperand=Just constLN }
                     where
                      {- brLN    = case ln of 
                                 ExprIndBlobConst {} -> ln
                                 _ - Expression -  -> Expression {nodeid=id1, occurence=0, statement=(integerToID 0), 
                                                        targetStatement=(integerToID 0), 
                                                        operator=LNOp_Bracket, condition=Nothing, firstOperand=Just ln, 
                                                        secondOperand=Nothing} -}
                      constLN = laExp2LN [id2] (laExpConstant (fromIntegral i))


showAllocFunction :: GMatrix Integer -> String
showAllocFunction r = showMemFunction "A" (calculatePi r) (calculateSOmega r)

-- | Zeigt die Mem-Funktion an. Jede Menge Zuckerguss drin und daher unuebersichtlich,
--   aber die Anzeige ist dafuer leserlich (im Gegensatz zu loopotargout, der musste jetzt sein :D ).
showMemFunction :: Integral a => String -> GMatrix Integer -> ([a], GMatrix a) -> String
showMemFunction name pi (s,om)  = name ++ "[" ++ (showCharList ',' (take ndim buchst)) ++ "] -> " 
                                 ++ name ++ "[" ++ showpseudoproj ++ "]"
                                 where
                                  ndim              = nColumns pi
                                  buchst            = ['i'..]
                                  pir               = rows' pi
                                  omr               = rows' om
                                  showpseudoproj    = if (showproj == "" || showpseudo == "")
                                                       then showproj ++ showpseudo 
                                                       else showproj ++ "," ++ showpseudo 
                                  showproj          = showpir pir         -- zeigt pi an
                                  showpseudo        = showsom (zip s omr) -- zeigt s/omega an
                                  showsom []        = ""
                                  showsom [(s,r)]   = (showmod (showrow r) s)
                                  showsom ((s,r):srs)= let a = showmod (showrow r) s
                                                           b = showsom srs
                                                      in if (a == "" || b == "")
                                                       then a ++ b
                                                       else a ++ "," ++ b
                                  showmod a s       = if (a == "" || s == 1)
                                                       then "" 
                                                       else if (length a==1) 
                                                             then a ++ " mod " ++ (show s)
                                                             else "(" ++ a ++ ") mod " ++ (show s)  -- wenn der ausdruck lang ist, einklammern
                                  showpir []        = ""
                                  showpir [r]       = showrow r
                                  showpir (r:rs)    = (showrow r) ++ "," ++ (showpir rs)
                                  showrow r         = case kvs of
                                                       "" -> "0"
                                                       _  -> kvs
                                                      where
                                                       kvs = showdim (zip r buchst)

-- | zeigt eine Dimension der Zielfunktion an
showdim :: Integral a => [(a, Char)] -> String
showdim []        = ""
showdim [kv]      = showelem kv
showdim (kv:kvs)  = let a = showelem kv 
                        b = showdim kvs
                    in if (a == "" || b == "" || take 2 b == " -")  -- + nur dazwischen, wenn beide Elemente existrieren und das zweite
                        then a ++ b                                 --   positiv, da sonst schon fuehrendes - da.
                        else a ++ " + " ++ b

-- | zeigt ein Element der Zielfunktion an
showelem :: Integral a => (a, Char) -> String
showelem (k,v)    = case k of
                     0 -> ""
                     1 -> [v]                                      -- mit konstanter 1 muss nicht multipliziert werden
                     _ -> if (k>0)
                           then (show k)++"*"++[v]
                           else " - "++(show (-k))++"*"++[v]       -- wenn kleiner null, ein fuehrendes Minus davorsetzen
                                  

showCharList :: Char -> String -> String
showCharList _ []     = []
showCharList _ [c]    = [c]
showCharList d (c:cs) = c:d:(showCharList d cs)



-- *** INTERFACE ***
showLifeTimeAnalysis :: CompSpreadDSMini -> [(TrDependenceID, Optimum [Rational])] -> String
showLifeTimeAnalysis _  []                  = ""
showLifeTimeAnalysis ds ((depID, res):deps) = (showTrDependence ds depID ) ++ ": " ++ (show res) ++ "\n" ++ (showLifeTimeAnalysis ds deps)

showTrDependence :: CompSpreadDSMini -> TrDependenceID -> String
showTrDependence ds depID = (show depID) ++ " (" ++ (show (trDepOrigSrcStmtID trDep)) ++ " --> " ++ (show (trDepOrigDestStmtID trDep)) ++ ")"
				 where
				 trDep = getTrDep ds depID

getTrDep :: CompSpreadDSMini -> TrDependenceID -> TrDependence
getTrDep ds depID = trDep where (Just trDep) = find (\dep -> trDepID dep == depID) (csTrDeps ds)

--- **** TEST FUNCTIONS ****
getFibDS = unsafeCreateCompSpreadDS "/home/cl/schulte/loopofibout" "fib.src.out"
getPasDS = unsafeCreateCompSpreadDS "/home/cl/schulte/loopopasout" "pascal.src.out"
getPas2DS = unsafeCreateCompSpreadDS "/home/cl/schulte/loopopas2" "pascal.src.out"
getExaDS = unsafeCreateCompSpreadDS "/home/cl/schulte/loopoexout" "example_2s.src.out"
array1 = (integerToID 1)::ArrayID
array2 = (integerToID 2)::ArrayID
tstmt1 = (integerToID 1)::TStatementID
tstmt2 = (integerToID 2)::TStatementID
tstmt3 = (integerToID 3)::TStatementID
tstmt4 = (integerToID 4)::TStatementID
rtest  = columnMatrix [1,0]
r2test  = columnMatrix [1,1]
r3test  = columnMatrix [2]
sOtest = calculateSOmega (conc (columnMatrix [(0),(2),(0)]) (columnMatrix [(0),(1),(2)]))
pitest = calculatePi (getAllocationVects (vector [0,0,2,0,0]))
lntest = Assign {nodeid = 39, occurence = 11, name = 1, statement = StatementID 2, targetStatement = (integerToID 2), 
                 indices = [Expression {nodeid = 94, occurence = -1,
                 statement = StatementID 2, targetStatement = (integerToID  2), operator = LNOp_Times, condition = Nothing, 
                 firstOperand = Just (ExprIndBlobConst {nodeid = 93, occurence = -1, name = 1, state = LNS_Constant, 
                 statement = StatementID 2, targetStatement = (integerToID  2)}), 
                 secondOperand = Just (ExprIndBlobConst {nodeid = 92,  occurence = -1, name = 1, state = LNS_Index, 
                 statement = StatementID 2, targetStatement = (integerToID  2)})},ExprIndBlobConst {nodeid = 38, occurence = 10, 
                 name = 0, state = LNS_Constant, statement = StatementID 2, targetStatement = (integerToID  2)}], 
                 firstExpr = ExprIndBlobConst {nodeid = 40, occurence = 8, name = 1, state = LNS_Constant, 
                 statement = StatementID 2, targetStatement = (integerToID  2)}}
