module ScannerParser.LambdaExpression(LaExp(..),
                        laExpConstant, laExpIndex, laExpBlob, 
                        laexpNF,
                        divLaExp, multLaExp, divLaExps, modLaExps,
                        compareLaExps, andLaExps, orLaExps,
                        ln2LaExp, laExp2LN, shrinkLN)

where

import ScannerParser.ParseTree
import ScannerParser.SymbolTable
import LoopoDS.IntegralID ( integerToID )

-- | LamdaExpression
-- 
-- In the first list, snd are the co-efficients, fst are the names of the indices
-- 
-- In the second list, snd are the co-efficients, fst are the names of the blobs
-- Third argument is a constant.
-- The fourth argument is the divisor of the expression.
data LaExp = LE [(Int, Int)] [(Int, Int)] Int Int 
     deriving (Show, Read) 

instance Eq LaExp where
 (==) = eqLaExps
-- (/=) = not (==)

--instance Ord LaExp where
--  (<=) 

instance Num LaExp where
 (+)            = addLaExps
 (-)            = subLaExps
 (*)            = error "ScannerParser.LaExp.(*): Not defined for all cases. Use multLaExp[s] instead" -- multLaExps
 negate         = negateLaExp
 fromInteger i  = laExpConstant (fromInteger i)
 abs            = error "Not defined"
 signum         = error "Not defined"
 

-- | Checks if two LaExps are equal 
eqLaExps :: LaExp -> LaExp -> Bool
eqLaExps le1 le2 = (c1 == c2) && (d1 == d2) && (eqLists inds1 inds2) && (eqLists bls1 bls2)
                   where
                    (LE inds1 bls1 c1 d1) = laexpNF le1
                    (LE inds2 bls2 c2 d2) = laexpNF le2
                    eqLists  xs     ys    = (length xs == length ys) && eqTupels xs ys
                    eqTupels [] _         = True
                    eqTupels (x:xs) ys    = (maybe 0 id (lookup (fst x) ys) == snd x) 
                                             && eqTupels xs ys

-- | Returns a LaExp which represents a constant
laExpConstant :: Int -> LaExp
laExpConstant i = LE [] [] i 1


-- | Returns a LaExp which represents the given index (name in Symtab)
laExpIndex :: Int -> LaExp
laExpIndex n = LE [(n, 1)] [] 0 1


-- | Returns a LaExp which represents the given blob (name in Symtab)
laExpBlob :: Int -> LaExp
laExpBlob n = LE [] [(n, 1)] 0 1

-- | Returns the greatest common divisor of all LaExp elements
laexpGgT :: LaExp -> Int
laexpGgT (LE ind bl c d) = ggt_bl where
                            ggt_c_d = gcd c d
                            ggt_ind = foldl gcd ggt_c_d (map snd ind)                                      
                            ggt_bl  = foldl gcd ggt_ind (map snd bl)

-- | Divides the LaExp with its gcd
laexpNF :: LaExp -> LaExp
laexpNF (LE ind bl c 0)    = error "Division through zero"
laexpNF le@(LE ind bl c d) = LE newind newbl newc newd where
                              leggt  = laexpGgT le
                              ggT    = if d>0                    -- sorgt fr einen positiven Nenner
                                       then leggt
                                       else (-1) * leggt
                              newc   = div c ggT
                              newd   = div d ggT
                              newind = filter ((/=0).snd) $ tupelMap (\ n -> div n ggT) ind
                              newbl  = filter ((/=0).snd) $ tupelMap (\ n -> div n ggT) bl

-- | Multiplies all LaExp elements with the given value (Erweitert die LaExp um den bergebenen Wert)
laexpExt :: Int -> LaExp -> LaExp
laexpExt i (LE ind bl c d) = LE newind newbl newc newd where
                              newc   = i*c
                              newd   = i*d
                              newind = tupelMap (*i) ind
                              newbl  = tupelMap (*i) bl

-- | Helper function. 
tupelMap :: (b -> c) -> [(a,b)] -> [(a,c)]
tupelMap f xs = ys where
                (as,bs) = unzip xs
                cs      = map f bs
                ys      = zip as cs
                            
test1 = LE [(2,2)]        [(2,2), (4,2)] 2 2
test2 = LE [(2,3), (4,5)] [(4,2), (3,4)] 3 1
test3 = LE []             []             3 2

-- | Adds two LaExps
addLaExps :: LaExp -> LaExp -> LaExp
-- Auf Hauptnenner bringen
-- Addieren 
addLaExps le1@(LE _ _ _ d1) le2@(LE _ _ _ d2) = laexpNF (LE newind newbl newc newd)
                               where
                                newd               = d1*d2
                                (LE ind1 bl1 c1 _) = laexpExt d2 le1
                                (LE ind2 bl2 c2 _) = laexpExt d1 le2
                                newind             = tupelAdd ind1 ind2
                                newbl              = tupelAdd bl1  bl2
                                newc               = c1+c2

-- | Negates a LaExp
negateLaExp :: LaExp -> LaExp
negateLaExp = multLaExp (-1) 

-- | Subtracts two LambdaExpressions
subLaExps :: LaExp -> LaExp -> LaExp
subLaExps le1 le2 = addLaExps le1 (negateLaExp le2)

-- | Compares to LaExps, if possible
-- only defined if substraction of the two LaExps returns a constant
compareLaExps :: LaExp -> LaExp -> Maybe Ordering
compareLaExps le1 le2 = if eqLaExps le1 le2 
                         then Just EQ
                         else case (subLaExps le2 le1) of
                               (LE [] [] c d) -> Just (compare 0 ((fromIntegral c) / (fromIntegral d)))
                               _              -> Nothing

-- | SetLessThan (and more) on LaExps. Returns integer constant 0 (False), 1 (True)
sltLaExps :: LambdaNodeOp -> LaExp -> LaExp -> Maybe LaExp
sltLaExps op le1 le2 = let leBool True  = laExpConstant 1
                           leBool False = laExpConstant 0
                           opOrd  LT    = (op==LNOp_NEQ || op==LNOp_LEQ || op==LNOp_LE)
                           opOrd  EQ    = (op==LNOp_GEQ || op==LNOp_EQ  || op==LNOp_LEQ)
                           opOrd  GT    = (op==LNOp_GE  || op==LNOp_GEQ || op==LNOp_NEQ)
                       in fmap (leBool.opOrd) (compareLaExps le1 le2)

-- | "Divides" the given LaExp with the given Int-value
divLaExp :: Int -> LaExp -> LaExp
divLaExp i (LE ind bl c d) = laexpNF (LE ind bl c (d*i))

-- | Divides two LaExps (Divisor must be a constant)
divLaExps :: LaExp -> LaExp -> Maybe LaExp
divLaExps le1 (LE [] [] c2 d2) = multLaExps le1 (LE [] [] d2 c2)
divLaExps _   _ = Nothing -- "Division auf LambdaExpressions nur fuer Konstanten definiert!")

-- | Modulo-Operation zweier LaExps. Nur mglich, wenn beide Operatoren konstant
--   oder zweiter konstant 1
modLaExps :: LaExp -> LaExp -> Maybe LaExp
modLaExps _               (LE [] [] 1  1) = Just (laExpConstant 0)
modLaExps (LE [] [] c1 1) (LE [] [] c2 1) = Just (laExpConstant (mod c1 c2))
modLaExps _               _               = Nothing

-- | "Multiplies" the given LaExp with the given Int-value
multLaExp :: Int -> LaExp -> LaExp
multLaExp i (LE ind bl c d) = laexpNF (LE newind newbl newc d) where
                               newc   = i*c
                               newind = tupelMap (*i) ind
                               newbl  = tupelMap (*i) bl

-- | Multiplies two LaExps (at least one has to be a constant)
multLaExps :: LaExp -> LaExp -> Maybe LaExp
multLaExps le1 (LE [] [] c2 d2)     = Just (divLaExp d2 (multLaExp c2 le1))
multLaExps le1@(LE [] [] c2 d2) le2 = multLaExps le2 le1
multLaExps le1 le2                  = Nothing -- Multiplikation auf LambdaExpressions nur "++
                                              -- fuer eine Konstante definiert!

-- | Helper function. Addiert Listen von Tupeln so, dass gleiche Koeffizienten addiert werden.
tupelAdd :: [(Int, Int)] -> [(Int, Int)] -> [(Int, Int)]
tupelAdd []     ys = ys
tupelAdd (x:xs) ys = tupelAdd xs zs where
                      (na1, ce1) = x
                      zs = case lookup na1 ys of
                           Nothing  -> x:ys
                           Just ce2 -> (na1, ce1+ce2):(filter (\x -> fst x /= na1) ys)

-- | Konjunktion zweier LaExps. Mglich, wenn einer konstant
andLaExps :: LaExp -> LaExp -> Maybe LaExp
andLaExps _               (LE [] [] 0  1) = Just (laExpConstant 0)
andLaExps (LE [] [] 0  1) _               = Just (laExpConstant 0)
andLaExps (LE [] [] 1  1) le2             = Just le2
andLaExps le1             (LE [] [] 1  1) = Just le1
andLaExps _               _               = Nothing


-- | Disjunktion zweier LaExps. Mglich, wenn einer konstant
orLaExps :: LaExp -> LaExp -> Maybe LaExp
orLaExps _               (LE [] [] 1  1) = Just (laExpConstant 1)
orLaExps (LE [] [] 1  1) _               = Just (laExpConstant 1)
orLaExps (LE [] [] 0  1) le2             = Just le2
orLaExps le1             (LE [] [] 0  1) = Just le1
orLaExps _               _               = Nothing


-- | Converts LambdaNode to LambdaExpression, if possible
ln2LaExp :: LambdaNode -> Maybe LaExp
ln2LaExp (ExprIndBlobConst {state=LNS_Constant, name=n}) = Just $ laExpConstant n
ln2LaExp (ExprIndBlobConst {state=LNS_Index,    name=n}) = Just $ laExpIndex    n
ln2LaExp (ExprIndBlobConst {state=LNS_Blob,     name=n}) = Just $ laExpBlob     n
-- ln2LaExp (Expression {operator=LNOp_Bracket, firstOperand=(Just op1)}) = ln2LaExp op1
ln2LaExp (Expression {operator=LNOp_Minus, firstOperand=Nothing, secondOperand=(Just op2)}) 
          = fmap negate (ln2LaExp op2)
ln2LaExp (Expression {operator=op, firstOperand=(Just op1), secondOperand=(Just op2)}) 
          = case (ln2LaExp op1, ln2LaExp op2) of
             (Nothing, Nothing)   -> Nothing
             (Nothing, Just le2)  -> case op of -- erster Operator nicht strikt?
                                      LNOp_Mod   -> if (eqLaExps (laExpConstant 1) le2)
                                                     then Just (laExpConstant 0)
                                                     else Nothing
                                      LNOp_And   -> if (eqLaExps (laExpConstant 0) le2)
                                                     then Just (laExpConstant 0)
                                                     else Nothing
                                      LNOp_Or    -> if (eqLaExps (laExpConstant 1) le2)
                                                     then Just (laExpConstant 1)
                                                     else Nothing
                                      _          -> Nothing
             (Just le1, Nothing)  -> case op of -- zweiter Operator nicht strikt?
                                      LNOp_And   -> if (eqLaExps (laExpConstant 0) le1)
                                                     then Just (laExpConstant 0)
                                                     else Nothing
                                      LNOp_Or    -> if (eqLaExps (laExpConstant 1) le1)
                                                     then Just (laExpConstant 1)
                                                     else Nothing
                                      _          -> Nothing
             (Just le1, Just le2) -> case op of  
                                      LNOp_Plus  -> Just (le1 + le2)
                                      LNOp_Minus -> Just (le1 - le2)
                                      LNOp_Times -> multLaExps le1 le2
                                      LNOp_Div   -> divLaExps le1 le2
                                      LNOp_Mod   -> modLaExps le1 le2
                                      LNOp_GE    -> sltLaExps op le1 le2
                                      LNOp_GEQ   -> sltLaExps op le1 le2
                                      LNOp_NEQ   -> sltLaExps op le1 le2
                                      LNOp_EQ    -> sltLaExps op le1 le2
                                      LNOp_LEQ   -> sltLaExps op le1 le2
                                      LNOp_LE    -> sltLaExps op le1 le2
                                      LNOp_Or    -> orLaExps le1 le2
                                      LNOp_And   -> andLaExps le1 le2
                                      _          -> Nothing
ln2LaExp _ = Nothing
                                      
--lntest1 = ForLoop {nodeid = 59, occurence = 29, name = 1, state = LNS_Do, loopStride = Just (ExprIndBlobConst {nodeid = 58, occurence = 5, name = 1, state = LNS_Constant, statement = 0, targetStatement = 0}), loopUpperBound = Just (Expression {nodeid = 34, occurence = 4, statement = 0, operator = LNOp_Minus, targetStatement = 0, condition = Nothing, firstOperand = Just (ExprIndBlobConst {nodeid = 32, occurence = 2, name = 1, state = LNS_Blob, statement = 0, targetStatement = 0}), secondOperand = Just (ExprIndBlobConst {nodeid = 33, occurence = 3, name = 1, state = LNS_Constant, statement = 0, targetStatement = 0})}), loopLowerBound = Just (ExprIndBlobConst {nodeid = 31, occurence = 1, name = 0, state = LNS_Constant, statement = 0, targetStatement = 0}), loopBody = [ForLoop {nodeid = 52, occurence = 23, name = 2, state = LNS_Do, loopStride = Just (ExprIndBlobConst {nodeid = 51, occurence = 14, name = 1, state = LNS_Constant, statement = 0, targetStatement = 0}), loopUpperBound = Just (ExprFuncCall {nodeid = 42, occurence = 13, name = -1, statement = 0, targetStatement = 0, arguments = [Expression {nodeid = 38, occurence = 9, statement = 0, operator = LNOp_Minus, targetStatement = 0, condition = Nothing, firstOperand = Just (ExprIndBlobConst {nodeid = 36, occurence = 7, name = 1, state = LNS_Blob, statement = 0, targetStatement = 0}), secondOperand = Just (ExprIndBlobConst {nodeid = 37, occurence = 8, name = 1, state = LNS_Constant, statement = 0, targetStatement = 0})},Expression {nodeid = 41, occurence = 12, statement = 0, operator = LNOp_Plus, targetStatement = 0, condition = Nothing, firstOperand = Just (ExprIndBlobConst {nodeid = 39, occurence = 10, name = 1, state = LNS_Index, statement = 0, targetStatement = 0}), secondOperand = Just (ExprIndBlobConst {nodeid = 40, occurence = 11, name = 2, state = LNS_Constant, statement = 0, targetStatement = 0})}]}), loopLowerBound = Just (ExprIndBlobConst {nodeid = 35, occurence = 6, name = 0, state = LNS_Constant, statement = 0, targetStatement = 0}), loopBody = [Assign {nodeid = 44, occurence = 22, name = 1, statement = 1, targetStatement = 0, indices = [ExprIndBlobConst {nodeid = 43, occurence = 21, name = 1, state = LNS_Index, statement = 1, targetStatement = 0}], firstExpr = Expression {nodeid = 50, occurence = 20, statement = 1, operator = LNOp_Plus, targetStatement = 0, condition = Nothing, firstOperand = Just (ExprRead {nodeid = 46, occurence = 16, name = 1, statement = 1, targetStatement = 0, indices = [ExprIndBlobConst {nodeid = 45, occurence = 15, name = 1, state = LNS_Index, statement = 1, targetStatement = 0}]}), secondOperand = Just (ExprRead {nodeid = 49, occurence = 19, name = 2, statement = 1, targetStatement = 0, indices = [ExprIndBlobConst {nodeid = 47, occurence = 17, name = 1, state = LNS_Index, statement = 1, targetStatement = 0},ExprIndBlobConst {nodeid = 48, occurence = 18, name = 2, state = LNS_Index, statement = 1, targetStatement = 0}]})}}]},Assign {nodeid = 53, occurence = 28, name = 3, statement = 2, targetStatement = 0, indices = [], firstExpr = Expression {nodeid = 57, occurence = 27, statement = 2, operator = LNOp_Plus, targetStatement = 0, condition = Nothing, firstOperand = Just (ExprRead {nodeid = 54, occurence = 24, name = 3, statement = 2, targetStatement = 0, indices = []}), secondOperand = Just (ExprRead {nodeid = 56, occurence = 26, name = 1, statement = 2, targetStatement = 0, indices = [ExprIndBlobConst {nodeid = 55, occurence = 25, name = 1, state = LNS_Index, -- statement = 2, targetStatement = 0}]})}}]}
--lntest2 = Expression {nodeid = 34, occurence = 4, statement = 0, operator = LNOp_Minus, targetStatement = 0, condition = Nothing, firstOperand = Just (ExprIndBlobConst {nodeid = 32, occurence = 2, name = 1, state = LNS_Constant, statement = 0, targetStatement = 0}), secondOperand = Just (ExprIndBlobConst {nodeid = 33, occurence = 3, name = 1, state = LNS_Constant, statement = 0, targetStatement = 0})}
--lntest3 = Expression {nodeid = 34, occurence = 4, statement = 0, operator = LNOp_Minus, targetStatement = 0, condition = Nothing, firstOperand = Just (ExprIndBlobConst {nodeid = 32, occurence = 2, name = 1, state = LNS_Constant, statement = 0, targetStatement = 0}), secondOperand = Just (ExprIndBlobConst {nodeid = 33, occurence = 3, name = 2, state = LNS_Constant, statement = 0, targetStatement = 0})}

-- | Takes a LambdaNode and "shrinks" it by converting it into an LaExp and back.
-- By that, operations like "3+4" turn into "7".
-- 
-- But warning: You lose information like occurence, (target)statement aso.
shrinkLN :: LambdaNode -> LambdaNode
shrinkLN = modifyMainNodes p newLN where
               mle ln    = ln2LaExp ln
               newLN ln  = let Just le  = mle ln 
                               ns       = (allNodeIDs [ln]) ++ [1..]
                               Just op1 = firstOperand ln
                               Just op2 = secondOperand ln
                           in {- if (mle ln) /= Nothing
                               then -} laExp2LN ns le
                               {- else if (ln2LaExp op1)==Nothing
                                     then shrinkLN op1
                                     else shrinkLN op2 -}
               p ln      = ((mle ln) /= Nothing && 
                            countNodes (newLN ln) < countNodes ln) -- krzbar per LaExp
                     {-      || (operator ln==LNOp_And   -- krzbar wegen And
                               && ((fmap ln2LaExp (firstOperand ln)) == Just (Just (LE [] [] 1 1))
                                   || (fmap ln2LaExp (secondOperand ln)) == Just (Just (LE [] [] 1 1))
                           )) -}

{- shrinkLN = modifyAllNodes sf where 
           sf orgln@(Expression {}) modln = case (ln2LaExp orgln) of
                                            Nothing   -> modln
                                            -- Rckgabe des neu erzeugten Baums nur, wenn kleiner
                                            (Just le) -> if countNodes newLN < countNodes orgln
                                                         then newLN
                                                         else modln
                                                         where 
                                                          newLN = laExp2LN ns le
                                                          ns = (allNodeIDs [orgln]) ++ [1..]
           -- fr alles andere auer Expressions gar nicht erst prfen
           sf orgln                 modln     = orgln 
   -}                         


-- | Converts a LambdaExpression into a LambdaNode
laExp2LN :: [Int] ->            -- ^ aus dieser Liste werden NodeIDs generiert
            LaExp -> LambdaNode
laExp2LN ns le = fst $ laExp2LNh ns (laexpNF le)

-- | Helper function  for laExp2LN
laExp2LNh :: [Int] ->            -- ^ aus dieser Liste werden NodeIDs generiert
             LaExp -> (LambdaNode, [Int])
laExp2LNh (nid:xs) (LE []   []   0 d)           -- Null
   = (ExprIndBlobConst {nodeid = nid, occurence = 0, name = 0, state = LNS_Constant, 
                       statement = (integerToID 0), targetStatement = (integerToID 0)}, xs)
laExp2LNh (nid:xs) (LE []   []   c 1) | c>0     -- eine einfache Konstante, wenn grergleich Null
   = (ExprIndBlobConst {nodeid = nid, occurence = 0, name = c, state = LNS_Constant, 
                       statement = (integerToID 0), targetStatement = (integerToID 0)}, xs)
                                     | c<0     -- unres Minus
   = (Expression {nodeid = nid, occurence = 0, statement = (integerToID 0), operator = LNOp_Minus, 
                  targetStatement = (integerToID 0), condition = Nothing, firstOperand = Nothing, 
                  secondOperand = Just lnm}, zs)
     where
      (lnm, zs) = laExp2LNh xs (LE []   []   (c*(-1)) 1)
laExp2LNh ns       (LE []   bls  0 1) -- nur noch Blob(s) und keine Konstante (+ 0 vermeiden)
   = getTupelLN (ns) LNS_Blob bls
laExp2LNh (nid:xs) (LE []   bls  c 1) -- Blob(s) und Konstante
   = (Expression {nodeid = nid, occurence = 0, statement = (integerToID 0), operator = LNOp_Plus, 
                  targetStatement = (integerToID 0), condition = Nothing, firstOperand = Just lnb, 
                  secondOperand = Just lnc}, 
      zs)
     where
      (lnb, ys) = laExp2LNh xs (LE [] bls 0 1)
      (lnc, zs) = laExp2LNh ys (LE [] []  c 1)
laExp2LNh ns       (LE inds []   0 1) -- nur noch Indizes (+ 0 vermeiden)
   = getTupelLN (ns) LNS_Index inds
laExp2LNh (nid:xs) (LE inds bls  c 1) -- Indizes und (Konstante und/oder Blobs)
   = (Expression {nodeid = nid, occurence = 0, statement = (integerToID 0), operator = LNOp_Plus, 
                  targetStatement =(integerToID 0), condition = Nothing, firstOperand = Just lni, 
                  secondOperand = Just lnr}, 
      zs)
     where
      (lni, ys) = laExp2LNh xs (LE inds [] 0 1)
      (lnr, zs) = laExp2LNh ys (LE []  bls c 1)      
-- Division kommt als allererstes, da durch alle anderen Patterns ausgeschlossen
laExp2LNh (nid:xs) (LE inds bls  c d) | d /= 1     
   = (Expression {nodeid = nid, occurence = 0, statement = (integerToID 0), operator = LNOp_Div, 
                  targetStatement = (integerToID 0), condition = Nothing, firstOperand = Just ln1, 
                  secondOperand = Just ln2}, zs)
     where
      (ln1, (y:ys)) = laExp2LNh xs (LE inds bls c 1) 
      {- lnb           = Expression {nodeid = y, occurence = 0, statement = (integerToID 0), operator = LNOp_Bracket, 
                      targetStatement = (integerToID 0), condition = Nothing, firstOperand=(Just ln1), 
                      secondOperand = Nothing} -}
      (ln2, zs)     = laExp2LNh ys (LE []   []  d 1) -- Divisor

-- | Helper function for laExp2LNh
getTupelLN :: [Int]              -- ^ aus dieser Liste werden NodeIDs generiert
              -> LambdaNodeState -- ^ Blob oder Index
              -> [(Int, Int)]    -- ^ die zu konvertierenden Tupel
              -> (LambdaNode, [Int])
getTupelLN (nid:ns) st [x]    = case snd x of
                                1 -> (tp, ns)
                                i -> (Expression {nodeid = (head ns), occurence = 0, statement = (integerToID 0), 
                                                  operator = LNOp_Times, targetStatement = (integerToID 0), 
                                                  condition = Nothing, firstOperand = Just lnc, 
                                                  secondOperand = Just tp}, 
                                      ys)
                                where
                                 (lnc, ys) = laExp2LNh (drop 1 ns) (LE [] [] (snd x) 1)
                                 tp = ExprIndBlobConst {nodeid = nid, occurence = 0, 
                                                        name = (fst x), state = st, 
                                                        statement = (integerToID 0), targetStatement = (integerToID 0)}
getTupelLN (nid:ns) st (x:xs) = (Expression {nodeid = nid, occurence = 0, statement = (integerToID 0), 
                                             operator = LNOp_Plus, targetStatement = (integerToID 0), 
                                             condition = Nothing, firstOperand = Just lnx, 
                                             secondOperand = Just lnxs}, 
                                 zs)
                                where
                                 (lnx,  ys) = getTupelLN ns st [x]
                                 (lnxs, zs) = getTupelLN ys st xs
