module ScannerParser.ParseTree (LambdaTree, LambdaNode(..), LambdaNodeOp(..), 
                  findMainNodes, findAllNodes, goThroughTree, getNode, searchLoop, getLevel, 
                  freshNodeIDs, freshStatementIDs, allNodeIDs, countNodes,
                  modifyMainNodes, modifyAllNodes, setNode, hasID, elemUnaryMinus, 
                  parseTreeToXML, parseTreeFromXML,
                  hasstatement, hasloopStride, hasloopUpperBound, hasloopLowerBound,
                  hascondition, hasindices, hasloopBody, hasfuncbody, hasarguments,     
                  hasfirstExpr, hasthenbranch, haselsebranch, hasfirstOperand, hassecondOperand
)

where

import Data.List 
import HXML.HXML -- http://www.flightlab.com/~joe/hxml/
import ScannerParser.XMLHelper -- findXMLSubnodes
import ScannerParser.SymbolTable
import LoopoDS.IntegralID
import LoopoDS.Statement

 
-- | If this data type is changed, you also have to change:
-- - has_ - functions (use DrIFT)
-- - ln2XML (lnArgs, subnodes, showCons)
-- - xml2LN
-- if subnodes were added or deleted: 
-- - modifyAllNodes (add step_x)
-- - goThroughTree (subNodes)
-- - getLevelInNode 
data LambdaNode = 
                -- MC: Schreibzugriff: -> AccessID = nodeid
                Assign {nodeid::Int, occurence::Int, 
                           name::Int,              -- ^ # of variable in Symtab
                           -- state::LambdaNodeState, -- always Write
                           statement::StatementID,
                           targetStatement::TStatementID, 
                           indices::[LambdaNode],  
                           firstExpr::LambdaNode  -- ^ Expression
                           }
                | ForLoop {nodeid::Int, occurence::Int, 
                           name::Int,                        -- ^ # of indexv ariable in Symtab
                           state::LambdaNodeState, 
                           loopStride::Maybe LambdaNode,     -- ^ Expression
                           loopUpperBound::Maybe LambdaNode, 
                           loopLowerBound::Maybe LambdaNode,
                           loopBody::LambdaTree} --children; Block
                | If {nodeid::Int, occurence::Int,
                           statement::StatementID,
                           targetStatement::TStatementID, 
                           condition::Maybe LambdaNode,
                           thenbranch::LambdaTree, --children; Block
                           elsebranch::LambdaTree} --elsechildren; Block
                | Skip {nodeid::Int, occurence::Int,
                           statement::StatementID,
                           targetStatement::TStatementID}
                | WhileLoop {nodeid::Int, occurence::Int, 
                           name::Int,                        -- ^ # of (artificial) index variable
                           state::LambdaNodeState, 
                           loopStride::Maybe LambdaNode,
                           loopLowerBound::Maybe LambdaNode,
                           condition::Maybe LambdaNode,
                           loopBody::LambdaTree} 
                | FuncDecl {nodeid::Int, occurence::Int, 
                           name::Int,                         -- ^ # of procedure name in Symtab
                           state::LambdaNodeState, 
                           funcbody::LambdaTree,              -- ^ can be empty
                           arguments::[LambdaNode]}
                | ProcCall {nodeid::Int, occurence::Int,
                           name::Int,                         -- ^ # of procedure name in Symtab
                           statement::StatementID,
                           targetStatement::TStatementID, 
                           arguments::[LambdaNode]}
                | ExprIndBlobConst {nodeid::Int, occurence::Int, 
                           name::Int,  -- ^ constant: value, other: # in symtab
                           state::LambdaNodeState,  -- ^ Constant | Index | Blob
                           statement::StatementID,
                           targetStatement::TStatementID}
                  -- MC: Lesezugriff: -> AccessID = nodeid
                | ExprRead {nodeid::Int, occurence::Int, 
                           name::Int,              -- ^ # of arrays in Symtab
                           statement::StatementID,
                           targetStatement::TStatementID, 
                           indices::[LambdaNode]} 
                | ExprFuncCall {nodeid::Int, occurence::Int, name::Int, 
                           statement::StatementID,
                           targetStatement::TStatementID, 
                           arguments::[LambdaNode]} 
                | Expression {nodeid::Int, occurence::Int, 
                           statement::StatementID,
                           targetStatement::TStatementID, 
                           operator::LambdaNodeOp, 
                           condition::Maybe LambdaNode, -- ^ for IIf
                           firstOperand::Maybe LambdaNode, --children - state 0 or IIf
                                                           --has to be a Maybe because of unary minus
                           secondOperand::Maybe LambdaNode} --elsechildren - state 0 or IIf
                | Label {nodeid::Int, occurence::Int, 
                           lexem::String}
                | GoTo {nodeid::Int, occurence::Int, 
                           lexem::String}
                {- Unknown {lntype::Int, nodeid::Int, occurence::Int, name::Int, 
                           state::LambdaNodeState, 
                           operator::LambdaNodeOp, 
                           statement::StatementID,
                           targetStatement::TStatementID, 
                           loopStride::Maybe LambdaNode,
                           loopUpperBound::Maybe LambdaNode, 
                           loopLowerBound::Maybe LambdaNode,
                           condition::Maybe LambdaNode,
                           indices::[LambdaNode], 
                           child::[LambdaNode], 
                           elsechild::[LambdaNode],
                           arguments::[LambdaNode]} -}                           
   deriving (Show, Eq) {-!derive : has !-}
   
type LambdaTree = [LambdaNode]

data LambdaNodeOp = {- LNOp_Bracket 
                  | -} LNOp_Plus --279 
                  | LNOp_Minus --280
                  | LNOp_Times 
                  | LNOp_Div 
                  | LNOp_Mod 
                  | LNOp_Not 
                  | LNOp_And 
                  | LNOp_Or
                  | LNOp_GE 
                  | LNOp_GEQ 
                  | LNOp_NEQ 
                  | LNOp_EQ 
                  | LNOp_LEQ 
                  | LNOp_LE 
                  | LNOp_IIf 
                  | LNOp_Parallel 
                  | LNOp_None
   deriving (Eq, Show, Read)

-- | searches for nodes in the tree, where the given function has to be True; 
-- if it is True, child nodes are ignored.
-- 
-- Example 1: see getNodes
-- 
-- Example 2: 
-- finding all constants in syntax tree:
-- 
-- @ findMainNodes (\ ln -> case ln of {ExprIndBlobConst {state=LNS_Constant} -> True; _ -> False}) pt @
findMainNodes :: (LambdaNode -> Bool) -> LambdaTree -> [LambdaNode]
findMainNodes pat pt = goThroughTree findNode pt
                       where
                        findNode xs ln = if (pat ln) 
                                         then [ln]
                                         else xs

-- | like findMainNodes, but also searching in subnodes
findAllNodes :: (LambdaNode -> Bool) -> LambdaTree -> [LambdaNode]
findAllNodes pat pt = goThroughTree findNode pt
                      where
                       findNode xs ln = if (pat ln) 
                                        then ln:xs
                                        else xs
          

-- | converts a tree to a list; see goThrewLN
-- 
-- Example see findMainNodes, findAllNodes, allNodeIDs, countNodes
goThroughTree :: ([a] -> LambdaNode -> [a]) -> LambdaTree -> [a]
goThroughTree f pt = concat $ map (goThroughLN f) pt

-- | converts a node to a list
goThroughLN :: ([a] -> LambdaNode -> [a]) -- ^ how the tree is converted to a list; 
                                        -- snd param is the current node, first is 
                                        -- the list of subnodes
             -> LambdaNode -> [a]
goThroughLN f ln = f sublns ln where 
                    fn = goThroughLN f    -- convert node into list
                    ft = goThroughTree f    -- convert list of nodes into list
                    fm mln = case mln of      -- convert Maybe LambdaNode into list
                             Just ln -> fn ln
                             Nothing -> []     
                    sublns = concat $ map snd $ filter fst subnodes
                    subnodes = [(hasloopStride ln,      fm (loopStride ln)),      -- maybe
                           (hasloopUpperBound ln,       fm (loopUpperBound ln)),  -- maybe
                           (hasloopLowerBound ln,       fm (loopLowerBound ln)),  -- maybe
                           (hascondition ln,            fm (condition ln)),       -- maybe
                           (hasindices ln,              ft (indices ln)),         -- list
                           (hasloopBody ln,             ft (loopBody ln)),        -- list
                           (hasfuncbody ln,             ft (funcbody ln)),        -- list
                           (hasarguments ln,            ft (arguments ln)),       -- list
                           (hasfirstExpr ln,            fn (firstExpr ln)),
                           (hasthenbranch ln,           ft (thenbranch ln)),      -- list
                           (haselsebranch ln,           ft (elsebranch ln)),      -- list
                           (hasfirstOperand ln,         fm (firstOperand ln)), 
                           (hassecondOperand ln,        fm (secondOperand ln))]
                           -- an dieser Stelle um evtl. weitere Subknoten erweitern
                

-- | returns the node with the given NodeID
getNode :: Int -> LambdaTree -> Maybe LambdaNode
getNode i lt = list2maybe $ findMainNodes (hasID i) lt

-- | returns the number of subnodes (includes the node itself)
countNodes :: LambdaNode -> Int
countNodes = sum . (goThroughLN c) where
             c xs ln = [(sum xs)+1]

-- | finds a loop with the given index variable i.
searchLoop :: Int -> LambdaTree -> Maybe LambdaNode                
searchLoop i lt = list2maybe $ findMainNodes (isLoopWithName i) lt 
                  where
                   isLoopWithName i (ForLoop   {name=nm}) = nm == i
                   isLoopWithName i (WhileLoop {name=nm}) = nm == i
                   isLoopWithName _ _                     = False


-- | returns the loop depth of the node with the given NodeID.
getLevel :: Int -> LambdaTree -> Int
getLevel nid lt | getNode nid lt /= Nothing = sum $ map (getLevelInNode nid) lt
getLevel nid _                              = error ("Node ID " ++ (show nid) ++ " not found.")

-- | Hilfsfunktion fr getLevel
getLevelInNode :: Int -> LambdaNode -> Int
-- Abbruchflle:
getLevelInNode nid ln | getNode nid [ln] == Nothing = 0 -- Node nicht vorhanden
getLevelInNode nid ln | nid == nodeid ln            = 0 -- Node gefunden
-- Schleifentiefe erhhen und absteigen:
getLevelInNode nid (ForLoop   {loopBody=body})      = 1 + (getLevel nid body)
getLevelInNode nid (WhileLoop {loopBody=body})      = 1 + (getLevel nid body)                                                      
--    (hier evtl. um weitere Schleifenkonstruktoren erweitern)
-- Schleifentiefe gleich lassen und absteigen:
getLevelInNode nid (FuncDecl  {funcbody=body})      =     getLevel nid body 
getLevelInNode nid (If        {thenbranch=tb, 
                               elsebranch=eb})      =     (getLevel nid tb) + (getLevel nid eb)
--    (hier evtl. um weitere Konstruktoren, in deren Subknoten eine Schleife sein kann, erweitern)
-- Node wurde zwar bisher nicht gefunden und wird in einem Subnode des aktuell untersuchten 
-- Knotens sein; es kann trotzdem abgebrochen werden, da sich weitere Schleifen nur in Subknoten 
-- der oben aufgezhlten Konstruktoren befinden knnen.
getLevelInNode _   _                                = 0 
                                                        
                                                        
                                                        
                                                        

-- | returns list with unused NodeIDs in the given syntax tree.
freshNodeIDs :: LambdaTree -> [Int]
freshNodeIDs ln = [n | n <- [1..], not (elem n usednids)] where
                  usednids = allNodeIDs ln

freshStatementIDs :: LambdaTree -> [StatementID]
freshStatementIDs ln = [id | n <- [1..], let id = integerToID n, not (elem id usednids)] where
                       usednids = allStatementIDs ln

-- | returns _unsorted_ list of all unused NodeIDs in the tree
allNodeIDs :: LambdaTree -> [Int]
allNodeIDs pt = goThroughTree getNodeIDs pt
                where
                getNodeIDs xs ln = (nodeid ln):xs

allStatementIDs :: LambdaTree -> [StatementID]
allStatementIDs pt = goThroughTree getStatementIDs pt
                     where
                     getStatementIDs xs ln = if hasstatement ln then (statement ln):xs else xs

-- | checks if a LambdaNode has a certain ID. Used by getNode, setNode
hasID :: Int -> LambdaNode -> Bool
hasID i ln = nodeid ln == i
                
-- | modifies all nodes in a tree with function  _m_odify.
-- The implementation of modify decides if subnodes get visited.
-- 
-- Example: see modifyMainNodes
modifyAllNodes :: (LambdaNode -> LambdaNode -> LambdaNode) 
                   -- ^ function for modifying nodes
                   -- 
                   -- 1. argument: original node
                   -- 
                   -- 2. Argument: node with subnodes changed; 
                   --              sollte genutzt werden, wenn weiter im Baum abgesteigen werden soll
                  -> LambdaNode  -- ^ Start-Node
                  -> LambdaNode
modifyAllNodes m ln = m ln newln where 
                      modify = modifyAllNodes m
                      modifyMaybe = fmap modify
                      modifyTree = map modify
                      -- hier folgen alle Subknoten, die es geben kann. Evtl erweitern!
                      step0_ln = ln
                      step1_ln = if (hasloopStride ln) 
                                 then step0_ln{loopStride=modifyMaybe (loopStride step0_ln)}
                                 else step0_ln
                      step2_ln = if (hasloopUpperBound ln) 
                                 then step1_ln{loopUpperBound=modifyMaybe (loopUpperBound step1_ln)}
                                 else step1_ln
                      step3_ln = if (hasloopLowerBound ln) 
                                 then step2_ln{loopLowerBound=modifyMaybe (loopLowerBound step2_ln)}
                                 else step2_ln
                      step4_ln = if (hascondition ln) 
                                 then step3_ln{condition=modifyMaybe (condition step3_ln)}
                                 else step3_ln
                      step5_ln = if (hasindices ln) 
                                 then step4_ln{indices=modifyTree (indices step4_ln)}
                                 else step4_ln
                      step6_ln = if (hasindices ln) 
                                 then step5_ln{indices=modifyTree (indices step5_ln)}
                                 else step5_ln
                      step7_ln = if (hasloopBody ln) 
                                 then step6_ln{loopBody=modifyTree (loopBody step6_ln)}
                                 else step6_ln
                      step8_ln = if (hasfuncbody ln) 
                                 then step7_ln{funcbody=modifyTree (funcbody step7_ln)}
                                 else step7_ln
                      step9_ln = if (hasarguments ln) 
                                 then step8_ln{arguments=modifyTree (arguments step8_ln)}
                                 else step8_ln
                      step10_ln = if (hasfirstExpr ln) 
                                 then step9_ln{firstExpr=modify (firstExpr step9_ln)}
                                 else step9_ln
                      step11_ln = if (hasthenbranch ln) 
                                 then step10_ln{thenbranch=modifyTree (thenbranch step10_ln)}
                                 else step10_ln
                      step12_ln = if (haselsebranch ln) 
                                 then step11_ln{elsebranch=modifyTree (elsebranch step11_ln)}
                                 else step11_ln
                      step13_ln = if (hasfirstOperand ln) 
                                 then step12_ln{firstOperand=modifyMaybe (firstOperand step12_ln)}
                                 else step12_ln
                      step14_ln = if (hassecondOperand ln)
                                 then step13_ln{secondOperand=modifyMaybe (secondOperand step13_ln)}
                                 else step13_ln
                      -- an dieser Stelle evtl. erweitern...
                      newln = step14_ln
                        
                
-- | modifies all True _p_attern nodes in a tree with function _m_odify.
-- Wenn das Pattern zutrifft, wird nicht weiter im Baum abgestiegen
-- Examples: elemUnaryMinus, LambdaExpression.shrinkLN
modifyMainNodes :: (LambdaNode -> Bool) -> (LambdaNode -> LambdaNode) -> LambdaNode -> LambdaNode
modifyMainNodes p m ln = modifyAllNodes f ln where 
                         f orgln modln = if (p orgln)
                                         then m orgln
                                         else modln

-- | eleminates unary minuses if possible:
-- x + (-y) -> x - y
elemUnaryMinus :: LambdaNode -> LambdaNode
elemUnaryMinus = modifyMainNodes p m where
                  p ln@(Expression {operator=LNOp_Plus, secondOperand=(Just op2)})
                       = case op2 of
                          Expression {operator=LNOp_Minus, firstOperand=Nothing,
                                      secondOperand=Just ExprIndBlobConst{}} 
                              -> True
                          _   -> False
                  p _  = False
                  m ln = ln{operator=LNOp_Minus, secondOperand=Just newOp} where
                          Just unMinLN = secondOperand ln
                          Just newOp   = secondOperand unMinLN

-- | replaces the node with the given NodeID through the newly given node. Uses modifyNodes.
setNode :: Int            -- ^NodeID of node to replace
           -> LambdaNode  -- ^new node
           -> LambdaTree  -- ^Tree
           -> LambdaTree  -- ^Result
setNode i ln = map (modifyMainNodes (hasID i) makeln) 
               where makeln _ = ln


-- | Converts syntax tree into an XML data structure.
-- To save it into an XML file, @ Tree RTNode [_] @ has to be wrapped around.
parseTreeToXML :: LambdaTree -> XML
parseTreeToXML lt = Tree (ELNode "LambdaTree" []) (map ln2XML lt)

fromXMLtest=Tree RTNode [Tree (ELNode "Parser" []) [Tree (ELNode "Test1" []) [],Tree (ELNode "Syntaxbaum" []) [Tree (ELNode "LambdaTree" []) [Tree (ELNode "ForLoop" [("nodeid","59"),("occurence","29"),("name","1"),("state","LNS_Do")]) [Tree (ELNode "loopStride" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","58"),("occurence","5"),("name","1"),("state","LNS_Constant"),("statement","0"),("targetStatement","0")]) []],Tree (ELNode "loopUpperBound" []) [Tree (ELNode "Expression" [("nodeid","34"),("occurence","4"),("name","0"),("statement","0"),("operator","LNOp_Minus"),("targetStatement","0")]) [Tree (ELNode "condition" []) [],Tree (ELNode "firstOperand" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","32"),("occurence","2"),("name","1"),("state","LNS_Blob"),("statement","0"),("targetStatement","0")]) []],Tree (ELNode "secondOperand" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","33"),("occurence","3"),("name","1"),("state","LNS_Constant"),("statement","0"),("targetStatement","0")]) []]]],Tree (ELNode "loopLowerBound" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","31"),("occurence","1"),("name","0"),("state","LNS_Constant"),("statement","0"),("targetStatement","0")]) []],Tree (ELNode "loopBody" []) [Tree (ELNode "ForLoop" [("nodeid","52"),("occurence","23"),("name","2"),("state","LNS_Do")]) [Tree (ELNode "loopStride" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","51"),("occurence","14"),("name","1"),("state","LNS_Constant"),("statement","0"),("targetStatement","0")]) []],Tree (ELNode "loopUpperBound" []) [Tree (ELNode "ExprFuncCall" [("nodeid","42"),("occurence","13"),("name","-1"),("statement","0"),("targetStatement","0")]) [Tree (ELNode "arguments" []) [Tree (ELNode "Expression" [("nodeid","38"),("occurence","9"),("name","0"),("statement","0"),("operator","LNOp_Minus"),("targetStatement","0")]) [Tree (ELNode "condition" []) [],Tree (ELNode "firstOperand" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","36"),("occurence","7"),("name","1"),("state","LNS_Blob"),("statement","0"),("targetStatement","0")]) []],Tree (ELNode "secondOperand" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","37"),("occurence","8"),("name","1"),("state","LNS_Constant"),("statement","0"),("targetStatement","0")]) []]],Tree (ELNode "Expression" [("nodeid","41"),("occurence","12"),("name","0"),("statement","0"),("operator","LNOp_Plus"),("targetStatement","0")]) [Tree (ELNode "condition" []) [],Tree (ELNode "firstOperand" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","39"),("occurence","10"),("name","1"),("state","LNS_Index"),("statement","0"),("targetStatement","0")]) []],Tree (ELNode "secondOperand" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","40"),("occurence","11"),("name","2"),("state","LNS_Constant"),("statement","0"),("targetStatement","0")]) []]]]]],Tree (ELNode "loopLowerBound" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","35"),("occurence","6"),("name","0"),("state","LNS_Constant"),("statement","0"),("targetStatement","0")]) []],Tree (ELNode "loopBody" []) [Tree (ELNode "Assign" [("nodeid","44"),("occurence","22"),("name","1"),("state","LNS_Write"),("statement","1"),("targetStatement","0")]) [Tree (ELNode "indices" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","43"),("occurence","21"),("name","1"),("state","LNS_Index"),("statement","1"),("targetStatement","0")]) []],Tree (ELNode "firstExpr" []) [Tree (ELNode "Expression" [("nodeid","50"),("occurence","20"),("name","0"),("statement","1"),("operator","LNOp_Plus"),("targetStatement","0")]) [Tree (ELNode "condition" []) [],Tree (ELNode "firstOperand" []) [Tree (ELNode "ExprRead" [("nodeid","46"),("occurence","16"),("name","1"),("statement","1"),("targetStatement","0")]) [Tree (ELNode "indices" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","45"),("occurence","15"),("name","1"),("state","LNS_Index"),("statement","1"),("targetStatement","0")]) []]]],Tree (ELNode "secondOperand" []) [Tree (ELNode "ExprRead" [("nodeid","49"),("occurence","19"),("name","2"),("statement","1"),("targetStatement","0")]) [Tree (ELNode "indices" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","47"),("occurence","17"),("name","1"),("state","LNS_Index"),("statement","1"),("targetStatement","0")]) [],Tree (ELNode "ExprIndBlobConst" [("nodeid","48"),("occurence","18"),("name","2"),("state","LNS_Index"),("statement","1"),("targetStatement","0")]) []]]]]]]]],Tree (ELNode "Assign" [("nodeid","53"),("occurence","28"),("name","3"),("state","LNS_Write"),("statement","2"),("targetStatement","0")]) [Tree (ELNode "indices" []) [],Tree (ELNode "firstExpr" []) [Tree (ELNode "Expression" [("nodeid","57"),("occurence","27"),("name","0"),("statement","2"),("operator","LNOp_Plus"),("targetStatement","0")]) [Tree (ELNode "condition" []) [],Tree (ELNode "firstOperand" []) [Tree (ELNode "ExprRead" [("nodeid","54"),("occurence","24"),("name","3"),("statement","2"),("targetStatement","0")]) [Tree (ELNode "indices" []) []]],Tree (ELNode "secondOperand" []) [Tree (ELNode "ExprRead" [("nodeid","56"),("occurence","26"),("name","1"),("statement","2"),("targetStatement","0")]) [Tree (ELNode "indices" []) [Tree (ELNode "ExprIndBlobConst" [("nodeid","55"),("occurence","25"),("name","1"),("state","LNS_Index"),("statement","2"),("targetStatement","0")]) []]]]]]]]]]]]]

-- | Converts XML data structure into LambdaTree.
parseTreeFromXML :: XML -> Maybe LambdaTree
-- Vorgehen: Suchen des ersten LambdaTree-ELNodes; bergeben an xml2LN.
parseTreeFromXML xml = case xmltrees of 
              [] -> Nothing
              _  -> Just (map xml2LN (onlyELNodes xmltrees))
              where
              xmltrees = findXMLSubnodes p xml -- sucht das erste Vorkommen von LambdaTree
              p (Tree (ELNode "LambdaTree" _) _) = True
              p _                                = False
              

-- | Hilfsfunktion fr fromXML
xml2LN :: XML -> LambdaNode
xml2LN xmlnode@(Tree (ELNode c args) subtrees) 
          = case c of 
            "Assign" 
                    -> Assign {nodeid=nodeid, occurence=occurence, name=name, 
                             statement=(integerToID statmnt), targetStatement=(integerToID targetStmnt),
                             indices=indices, firstExpr=firstExpr}
            "ForLoop" 
                    -> ForLoop {nodeid=nodeid, occurence=occurence, name=name, 
                             state=state,
                             loopStride=loopStride,
                             loopUpperBound=loopUpperBound,
                             loopLowerBound=loopLowerBound,
                             loopBody=loopBody}
            "If"    -> If {nodeid=nodeid, occurence=occurence, 
                             statement=(integerToID statmnt), targetStatement=(integerToID targetStmnt),
                             condition=condition, 
                             thenbranch=thenbranch,
                             elsebranch=elsebranch}
            "Skip"  -> Skip {nodeid=nodeid, occurence=occurence, 
                             statement=(integerToID statmnt), targetStatement=(integerToID targetStmnt)}
            "WhileLoop" 
                    -> WhileLoop {nodeid=nodeid, occurence=occurence, name=name, 
                             state=state,
                             loopStride=loopStride,
                             loopLowerBound=loopLowerBound,
                             condition=condition, 
                             loopBody=loopBody}
            "FuncDecl" 
                    -> FuncDecl {nodeid=nodeid, occurence=occurence, name=name, 
                             state=state,
                             funcbody=funcbody,
                             arguments=arguments}
            "ProcCall" 
                    -> ProcCall {nodeid=nodeid, occurence=occurence, name=name, 
                             statement=(integerToID statmnt), targetStatement=(integerToID targetStmnt),
                             arguments=arguments}
            "ExprIndBlobConst" 
                    -> ExprIndBlobConst {nodeid=nodeid, occurence=occurence, name=name, 
                             statement=(integerToID statmnt), targetStatement=(integerToID targetStmnt),
                             state=state}
            "ExprFuncCall" 
                    -> ExprFuncCall {nodeid=nodeid, occurence=occurence, name=name, 
                             statement=(integerToID statmnt), targetStatement=(integerToID targetStmnt),
                             arguments=arguments}
            "Expression" 
                    -> Expression {nodeid=nodeid, occurence=occurence, 
                             statement=(integerToID statmnt), targetStatement=(integerToID targetStmnt),
                             operator=operator,
                             condition=condition,
                             firstOperand=firstOperand,
                             secondOperand=secondOperand}
            "ExprRead" 
                    -> ExprRead {nodeid=nodeid, occurence=occurence, name=name, 
                             statement=(integerToID statmnt), targetStatement=(integerToID targetStmnt),
                             indices=indices}
            "Label" -> Label {nodeid=nodeid, occurence=occurence, lexem=lexem}
            "GoTo"  -> GoTo  {nodeid=nodeid, occurence=occurence, lexem=lexem}
            -- hier um weitere Konstruktoren erweitern
            _       -> error ("Unknown constructor in XML file: " ++ (show c))
           where
                    getAttVal s = maybe "" id $ attval s xmlnode                                        
                    nodeid              = read $ getAttVal "nodeid"
                    occurence           = read $ getAttVal "occurence"
                    name                = read $ getAttVal "name"
                    state               = read $ getAttVal "state"
                    statmnt             = read $ getAttVal "statement"
                    operator            = read $ getAttVal "operator"
                    targetStmnt         = read $ getAttVal "targetStatement"
                    lexem               = read $ getAttVal "lexem"
                    -- hier um weitere Argumente erweitern
                    
                    consPattern s (Tree (ELNode c _) _) = s == c
                    findSubnodes s = case filter (consPattern s) subtrees of
                                     []   -> [] -- falls der Subknoten nicht existiert
                                     xmls -> onlyELNodes $ treeChildren $ head xmls
                    
                    loopStride     = fmap xml2LN $ list2maybe $ findSubnodes "loopStride"
                    indices        =  map xml2LN $              findSubnodes "indices"
                    firstExpr      =      xml2LN $ head       $ findSubnodes "firstExpr"
                    loopUpperBound = fmap xml2LN $ list2maybe $ findSubnodes "loopUpperBound"
                    loopLowerBound = fmap xml2LN $ list2maybe $ findSubnodes "loopLowerBound"
                    loopBody       =  map xml2LN $              findSubnodes "loopBody" 
                    condition      = fmap xml2LN $ list2maybe $ findSubnodes "condition"
                    thenbranch     =  map xml2LN $              findSubnodes "thenbranch"
                    elsebranch     =  map xml2LN $              findSubnodes "elsebranch"
                    funcbody       =  map xml2LN $              findSubnodes "funcbody"
                    arguments      =  map xml2LN $              findSubnodes "arguments"
                    firstOperand   = fmap xml2LN $ list2maybe $ findSubnodes "firstOperand"
                    secondOperand  = fmap xml2LN $ list2maybe $ findSubnodes "secondOperand"
                    -- hier um weitere Subknoten erweitern
                       

-- | Hilfsfunktion fr toXML
ln2XML :: LambdaNode -> XML
ln2XML ln = Tree (ELNode (showCons ln) thisArgs) thisSubNodes where
            list2XML = map ln2XML
            node2XML n = [ln2XML n]
            maybe2XML Nothing = []
            maybe2XML (Just n) = node2XML n
            thisArgs = map snd $ filter fst lnArgs
            thisSubNodes = map toXMLTree $ map snd $ filter fst subnodes
            toXMLTree (name, xs) = Tree (ELNode name []) xs
            lnArgs = [(hasnodeid ln,     ("nodeid",      show (nodeid ln))), 
                         (hasoccurence ln,  ("occurence",   show (occurence ln))), 
                         (hasname ln,       ("name",        show (name ln))), 
                         (hasstate ln,      ("state",       show (state ln))), 
                         (hasstatement ln,  ("statement",   show (integerFromID (statement ln)))),
                         (hasoperator ln,   ("operator",    show (operator ln))), 
                         (hastargetStatement ln, ("targetStatement", show (integerFromID (targetStatement ln)))),
                         (haslexem ln,      ("lexem",       show (lexem ln)))]
                         -- hier um weitere Argumente erweitern
            subnodes = [(hasloopStride ln,     ("loopStride",     maybe2XML (loopStride ln))),      
                        (hasloopUpperBound ln, ("loopUpperBound", maybe2XML (loopUpperBound ln))),  
                        (hasloopLowerBound ln, ("loopLowerBound", maybe2XML (loopLowerBound ln))),  
                        (hascondition ln,      ("condition",      maybe2XML (condition ln))),       
                        (hasindices ln,        ("indices",        list2XML (indices ln))),
                        (hasloopBody ln,       ("loopBody",       list2XML (loopBody ln))),
                        (hasfuncbody ln,       ("funcbody",       list2XML (funcbody ln))),
                        (hasarguments ln,      ("arguments",      list2XML (arguments ln))),
                        (hasfirstExpr ln,      ("firstExpr",      node2XML (firstExpr ln))),
                        (hasthenbranch ln,     ("thenbranch",     list2XML (thenbranch ln))),
                        (haselsebranch ln,     ("elsebranch",     list2XML (elsebranch ln))),
                        (hasfirstOperand ln,   ("firstOperand",   maybe2XML (firstOperand ln))), 
                        (hassecondOperand ln,  ("secondOperand",  maybe2XML (secondOperand ln)))]
                        -- hier um weitere Sunknoten erweitern
            showCons Assign {}           = "Assign"
            showCons ForLoop {}          = "ForLoop"
            showCons If {}               = "If"
            showCons WhileLoop {}        = "WhileLoop"
            showCons FuncDecl {}         = "FuncDecl"
            showCons ProcCall {}         = "ProcCall"
            showCons ExprFuncCall {}     = "ExprFuncCall"
            showCons ExprRead {}         = "ExprRead"
            showCons Expression {}       = "Expression"
            showCons Skip {}             = "Skip"
            showCons ExprIndBlobConst {} = "ExprIndBlobConst"
            showCons Label {}            = "Label"
            showCons GoTo {}             = "GoTo"
            -- hier um weitere Konstruktoren erweitern
            

-- | Allgemeine Hilfsfunktion
list2maybe :: [a] -> Maybe a
list2maybe []  = Nothing
list2maybe xs  = Just (head xs)


            
{- Die folgenden Funktionen sind fr Auflistungen ber die Argumente 
 der LambdaNodes sehr praktisch. Sie wurden ursprnglich mittels
 DrIFT deriving has generiert. bei kleineren nderungen an den Datentypen
 kann man sie natrlich mit der Hand ndern).-}

hasnodeid    _ = True

hasoccurence _ = True

hasname If{} = False
hasname Skip{} = False
hasname Label{} = False
hasname GoTo{} = False
hasname Expression{} = False
hasname _ = True

hasstate ForLoop{} = True
hasstate WhileLoop{} = True
hasstate FuncDecl{} = True
hasstate ExprIndBlobConst{} = True
hasstate _ = False

hasstatement Assign{} = True
hasstatement ForLoop{} = False
hasstatement If{} = True
hasstatement Skip{} = True
hasstatement WhileLoop{} = False
hasstatement FuncDecl{} = False
hasstatement ProcCall{} = True
hasstatement ExprIndBlobConst{} = True
hasstatement ExprRead{} = True
hasstatement ExprFuncCall{} = True
hasstatement Expression{} = True
hasstatement Label{} = False
hasstatement GoTo{} = False

hastargetStatement = hasstatement

hasindices Assign{}   = True
hasindices ExprRead{} = True
hasindices _          = False

hasfirstExpr Assign{} = True
hasfirstExpr _        = False

hasloopStride ForLoop{}   = True
hasloopStride WhileLoop{} = True
hasloopStride _           = False

hasloopUpperBound ForLoop{} = True
hasloopUpperBound _         = False

hasloopLowerBound ForLoop{}   = True
hasloopLowerBound WhileLoop{} = True
hasloopLowerBound _           = False

hasloopBody ForLoop{}   = True
hasloopBody WhileLoop{} = True
hasloopBody _           = False

hascondition If{}         = True
hascondition Expression{} = True
hascondition WhileLoop{}  = True
hascondition _            = False

hasthenbranch If{} = True
hasthenbranch _    = False

haselsebranch If{} = True
haselsebranch _    = False

hasfuncbody FuncDecl{} = True
hasfuncbody _          = False

hasarguments FuncDecl{}     = True
hasarguments ProcCall{}     = True
hasarguments ExprFuncCall{} = True
hasarguments _              = False

hasoperator Expression{} = True
hasoperator _            = False

hasfirstOperand Expression{} = True
hasfirstOperand _            = False

hassecondOperand Expression{} = True
hassecondOperand _            = False

haslexem Label{} = True
haslexem GoTo{}  = True
haslexem _       = False

--  Imported from other files :-
