module ScannerParser.PrintCode(printCode, TargetLanguage(..),
                               arrayStructurePrefix, 
                               arrayStructureGet, 
                               arrayStructureSet,
                               ehrhartAccessMacro)
where

import ScannerParser.ParseTree
import ScannerParser.SymbolTable
-- import ScannerParser.LambdaExpression
import Data.List
import LoopoDS.IntegralID ( integerToID )

-- | Datatype that determins the target language.
-- | see printCode;
data TargetLanguage = TL_PlainC 
                    | TL_ModAccess 
                    | TL_EhrhartAccess [String]
                    -- ^ ehrhart access with corresponding taskIDs and blobs

-- a few constants for the target code:
arrayStructurePrefix  = "" -- "arrayDataStructure_"
arrayStructureGet     = "get"
arrayStructureSet     = "set"
ehrhartAccessMacro    = "LOOPO_EHRHART_ACCESS_MACRO"
ehrhartAccessFuncName = "ehrhartAccessFunction"

-- | Prints C-code from LambdaTree
printCode :: TargetLanguage -> SymTab -> LambdaTree -> String
printCode tl = showLTree 0 tl
-- Die Funktion könnte man auch mit goThrewTree erstellen, würde aber wenig
-- Code sparen

-- Verbesserungsmöglichkeiten: Andere Sprachen unterstützen!

-- | see printCode; first argument: Einrückung für Körper
showLTree :: Int -> TargetLanguage -> SymTab -> LambdaTree -> String
showLTree 0 tl st [ln] = (showLN 0 tl Nothing st ln) ++ "\n"
showLTree i tl st [ln] = "\n" ++ (space (2*i)) ++ (showLN i tl Nothing st ln)
showLTree 0 tl st lns  = showLNBlock 0 tl st lns
showLTree i tl st lns  = "{\n" ++ (showLNBlock i tl st lns) ++ (space (2*i)) ++ "}"


showLN :: Int                    -- ^ indentation
          -> TargetLanguage      -- ^ target language
          -> Maybe LambdaNodeOp  -- ^ äußerer Operator (für Klammerung)
          -> SymTab              -- ^ Symboltabelle (für Variablenbenennung)
          -> LambdaNode          -- ^ auszu"drucken"der Teilsyntaxbaum
          -> String
showLN i tl _  st (ForLoop {name=var, loopLowerBound=low, loopUpperBound=up, 
                           loopStride=step, loopBody=body, state=par}) 
                 = let vi   = (indexLexem st var)
                       pfor = case par of
                               LNS_DoAll -> "parallel "
                               _         -> ""
                   in pfor ++ "for (" ++ vi ++ "=" ++ 
                    (showLNMaybe 0 tl Nothing st low) ++ "; " ++ vi ++ "<=" ++ 
                    (showLNMaybe 0 tl Nothing st up) ++ "; " ++ vi ++ "+=" ++ 
                    (showLNMaybe 0 tl Nothing st step) ++ ") " ++ (showLTree (i+1) tl st body)
showLN i tl _  st (WhileLoop {condition=cond, loopBody=body}) 
                 = "while (" ++ (showLNMaybe 0 tl Nothing st cond) ++ ") " 
                   ++ (showLTree (i+1) tl st body)
showLN i tl _  st (If {condition=cond, thenbranch=tb, elsebranch=eb})
                 = "if (" ++ (showLNMaybe 0 tl Nothing st cond) ++ ") "
                   ++ (showLTree (i+1) tl st tb) ++
                   case eb of
                   [] -> ""
                   _  -> " else " ++ (showLTree (i+1) tl st eb)
showLN _ _ _  st (ExprIndBlobConst {name=var, state=cib})
                 = case cib of
                   LNS_Constant -> if var >=0 then (show var) else ("(" ++ (show var) ++ ")")
                   LNS_Index -> (indexLexem st var)
                   LNS_Blob  -> (blobLexem st var)
                   LNS_Proto  -> "var" ++ (show var)
                   -- ++ " (- " ++ (show ni) ++ "-)"
showLN _ tl oo st (Expression {operator=op, condition=cond, 
                      firstOperand=op1, secondOperand=op2})
                 = {- if op == LNOp_Bracket then
                      "(" ++ (showLNMaybe 0 tl st op1) ++ ")"
                   else -} 
                   if (op == LNOp_Minus && op1==Nothing) then        -- unäres Minus
                      "(-" ++ (showLNMaybe 0 tl (Just LNOp_Not) st op2) ++ ")" -- Not hat gleiche Präferenz wie unäres Minus
                   else 
                     let outerOp = Just op
                         code = if op == LNOp_IIf 
                                 then (showLNMaybe 0 tl outerOp st cond) ++ " ? " 
                                       ++ (showLNMaybe 0 tl outerOp st op1) ++ " : "  
                                       ++ (showLNMaybe 0 tl outerOp st op2)
                                 else (showLNMaybe 0 tl outerOp st op1)  ++ 
                                      (printLNOpCode op)  ++ 
                                      (showLNMaybe 0 tl outerOp st op2)
                     in if (maybe LNOp_IIf id oo) > op -- ist der äußere Operator stärker, klammern
                         then "(" ++ code ++ ")"
                         else code
showLN _ tl _ st (ExprFuncCall {name=func, arguments=args})
                 = (functionLexem st func) ++ "(" ++ (showLNArgs st tl args) ++ ")"
showLN i tl _ st (FuncDecl {name=name, funcbody=body, arguments=args})
                 = "function " ++ (functionLexem st name) ++ " (" ++ (showLNArgs st tl args) ++ ") "
                   ++ (showLTree (i+1) tl st body)
-- WARNING: might be a bit incomplete procedure call
showLN i tl _ st (ProcCall {name=func, arguments=args})
                 = -- (functionLexem st func) ++ "(" ++ (showLNArgs st tl args) ++ ")"
                   (functionLexem st func) ++ "(" ++ (showLNArgs st TL_PlainC args) ++ ")"
showLN _ TL_PlainC _ st (Assign {name=var, indices=ind, firstExpr=value})
                 = (arrayLexem st (integerToID var)) ++ "[" ++ (showLNArrays st TL_PlainC ind) ++ "] = " ++ 
                   showLN 0 TL_PlainC Nothing st value ++ ";"
showLN _ TL_ModAccess _ st (Assign {name=var, indices=ind, firstExpr=value})
                 = arrayStructurePrefix ++ (arrayLexem st (integerToID var)) ++ 
                   "." ++ arrayStructureSet ++ "(" ++ showLN 0 TL_ModAccess Nothing st value ++ 
                   ", " ++ (showLNArgs st TL_ModAccess ind) ++ ");"
showLN _ (TL_EhrhartAccess params) _ st (Assign {nodeid=accID, name=var, indices=ind, firstExpr=value})
                 = ("ehrhart" ++ arrayLexem st (integerToID var) ++ "[" ++ 
                    ehrhartAccessFuncName ++ show var ++ "(" ++ 
                    ehrhartArguments++ ") "  ++ "] = " ++ 
                    showLN 0 (TL_EhrhartAccess params) Nothing st value ++ ";"
                   )
    where
    ehrhartArguments = 
        ((showLNArgs st (TL_EhrhartAccess params) ind) ++ if null ind then "" else ", " ++
        (concat $ intersperse ", " params))
showLN _ TL_PlainC _ st (ExprRead {name=var, indices=ind})
                 = (arrayLexem st (integerToID var)) ++ "[" ++ (showLNArrays st TL_PlainC ind) ++ "]"
showLN _ TL_ModAccess _ st (ExprRead {name=var, indices=ind})
                 = arrayStructurePrefix ++ (arrayLexem st (integerToID var)) ++ 
                   "." ++ arrayStructureGet ++ "(" ++ (showLNArgs st TL_ModAccess ind) ++ ") "
showLN _ (TL_EhrhartAccess params) _ st (ExprRead {nodeid=accID, name=var, indices=ind})
                 = ("ehrhart" ++ arrayLexem st (integerToID var) ++ 
                    "[" ++ ehrhartAccessFuncName ++ show var ++ "(" ++ 
                    ehrhartArguments ++ ")" ++ "]"
                   )
    where
    ehrhartArguments = 
        ((showLNArgs st (TL_EhrhartAccess params) ind) ++ if null ind then "" else ", " ++
         (concat $ intersperse ", " params))
showLN _ _ _ st (Skip {}) = "skip"
showLN _ _ _ st ln      = error ("printCode function not implemented yet for " ++ (show ln))


-- | help function so that no Just or Nothing is displayed
showLNMaybe :: Int 
            -> TargetLanguage 
            -> Maybe LambdaNodeOp 
            -> SymTab 
            -> Maybe LambdaNode 
            -> String
showLNMaybe i tl op st (Just x) = showLN i tl op st x
showLNMaybe _ _  _ _  Nothing  = " "


-- | Listen von Anweisungen werden pro Einheit je Zeile ausgegeben, mit Einrückung i
showLNBlock :: Int -> TargetLanguage -> SymTab -> [LambdaNode] -> String
showLNBlock i tl st lns = 
  unlines $ zipWith (++) (map (linebegin i) lns) 
                         (map (++ ";") (map (showLN i tl Nothing st) lns))
  where
  linebegin i ln = (space (2*i))


-- problem: doesn't work for multi-dimensional array accesses: 
--          expected: a[i][j]
--          result:   a[i, j] <-- illegal code
showLNArgs :: SymTab -> TargetLanguage -> [LambdaNode] -> String
-- include parents for indices used in the ERHART_ACCESS_MACRO
-- showLNArgs st TL_EhrhartAccess lns = 
--     concat $ intersperse ", " $ 
--            map (\ln -> "(" ++ (showLN 0 TL_EhrhartAccess Nothing st ln) ++ ")") lns
showLNArgs st tl lns = concat $ intersperse ", " $ map (showLN 0 tl Nothing st) lns

-- laNodeID == AccessID... 

showLNArrays :: SymTab -> TargetLanguage -> [LambdaNode] -> String
showLNArrays st tl lns = concat $ intersperse "][" $ map (showLN 0 tl Nothing st) lns

--showLNArrAccs :: SymTab -> TargetLanguage -> [LambdaNode] -> String
--showLNArrAccs st tl lns = -- todo: something like: arrAcc1 = i; arrAcc2 = 2*j-1;
--                          -- d.h. zipWith ueber lns, [1..], jeweils arrAcc i = show ...

-- | prints C expression for operators
printLNOpCode :: LambdaNodeOp -> String
printLNOpCode LNOp_Plus      = "+"
printLNOpCode LNOp_Minus     = "-"   
printLNOpCode LNOp_Times     = "*"
printLNOpCode LNOp_Div       = "/"
printLNOpCode LNOp_Mod       = " % "
printLNOpCode LNOp_Not       = "!"
printLNOpCode LNOp_And       = " && "
printLNOpCode LNOp_Or        = " || "
printLNOpCode LNOp_GE        = " > "
printLNOpCode LNOp_GEQ       = " >= "
printLNOpCode LNOp_NEQ       = " != "
printLNOpCode LNOp_LE        = " < "
printLNOpCode LNOp_LEQ       = " <= "
printLNOpCode LNOp_EQ        = " == "

instance Ord LambdaNodeOp where
  op1 <= op2 = cOperatorPreference op1 <= cOperatorPreference op2

-- | Operator preference in C and C++
cOperatorPreference :: LambdaNodeOp -> Int
cOperatorPreference LNOp_Plus      = 6
cOperatorPreference LNOp_Minus     = 6 -- unäres Minus hätte 8! Ist hier zusammengefasst   
cOperatorPreference LNOp_Times     = 7
cOperatorPreference LNOp_Div       = 7
cOperatorPreference LNOp_Mod       = 7
cOperatorPreference LNOp_Not       = 8
cOperatorPreference LNOp_And       = 3
cOperatorPreference LNOp_Or        = 2
cOperatorPreference LNOp_GE        = 5
cOperatorPreference LNOp_GEQ       = 5
cOperatorPreference LNOp_NEQ       = 4
cOperatorPreference LNOp_LE        = 5
cOperatorPreference LNOp_LEQ       = 5
cOperatorPreference LNOp_EQ        = 4
cOperatorPreference LNOp_IIf       = 1

-- |  Erstellt einen String mit i Leerzeichen
space :: Int -> String
space 0     = []
space i|i>0 = ' ' : space (i-1)                          
                   
