module ScannerParser.SymbolTable
    ( LNSymbol(..), LNSStatus(..), SymTabFrame, SymTab, getLNSStatusClear,
      LNSIntent(..), LNSBasicType(..),
      LambdaNodeState(..), 
      mapSymTab, getSymTabFrame, findSymbols, findAllSymbols, retrieveSymbol,
      findSymbol, findSymbolFrameNumber, findSymbolFrame,
      findSymbolReplacements, isSymbol,
      renameSymbolFrame, deleteSymbolFrame, 
      deleteFrameSymbolsBy, deleteSymbolsBy, deleteSymbol,
      insertSymbol,
      findIndex, indexLexem, getIndexLexems,
      findBlob, blobLexem, getBlobLexems,
      findArray, arrayLexem, getArrayLexems,
      functionLexem,
      modifyLNSymbols, modifySymTabFrames, modifySymbolReplacements, 
      symTabToXML, symTabFromXML
    )
where

import Data.List(sortBy)
import Data.Maybe(isJust, fromJust)
--import ScannerParser.ParseTree(LambdaNodeState(..))
import HXML.HXML
import ScannerParser.XMLHelper
import LoopoDS.Access
import LoopoDS.IntegralID

data LNSymbol = Symbol {sym_lexem::String,
                        sym_token::Int,
                        sym_type::LambdaNodeState,
                        sym_name::Int, -- ^ number
                        sym_indices::Int, -- ^ only for arrays
                        sym_level::Int,
                        sym_intent::Maybe LNSIntent, -- ^ for dummys
                        sym_basictype::Maybe LNSBasicType, -- ^ for functions
                        sym_identify::Int,
                        sym_status::LNSStatus -- ^ Bitvector
                       }
                deriving (Show, Eq)
                       
data LNSStatus    = LNSS {sym_valid::Bool,
                          sym_pure::Bool,
                          sym_no_warn::Bool,
                          sym_dynamic::Bool,
                          sym_local::Bool,
                          sym_pseudo::Bool,
                          sym_normLoops::Bool,
                          sym_assign_idx::Bool}
                deriving (Show, Eq)
                                          
data LambdaNodeState = LNS_ReadAndWrite
                     | LNS_Write
                     | LNS_Read
                     | LNS_Constant
                     | LNS_Index
                     | LNS_Blob
                     | LNS_Array
                     | LNS_Key
                     | LNS_Function
                     | LNS_FunctionCall
                     | LNS_Proto
                     | LNS_Do | LNS_DoAll | LNS_DoAcross -- Loops
                     | LNS_OK                   -- Functions/Expr 0 "OK"
                     | LNS_Various             -- Functions
                     | LNS_TilDo | LNS_TilPar | LNS_TilAss
                     | LNS_Statement
   deriving (Show, Read, Eq)

getLNSStatusClear :: LNSStatus
getLNSStatusClear = LNSS {sym_valid=False, sym_pure=False, sym_no_warn=False,
                          sym_dynamic=False, sym_local=False, 
                          sym_pseudo=False, sym_normLoops=False, 
                          sym_assign_idx=False}

data LNSIntent    = LNSI_In | LNSI_Out | LNSI_InOut     
                deriving (Show, Read, Eq)
                     
data LNSBasicType = LNST_Unknown
                  | LNST_Void
                  | LNST_Int
                  | LNST_Bool
                  | LNST_Float
                  | LNST_Template
                  | LNST_Processor
                  | LNST_UnsignedInt
                deriving (Show, Read, Eq)

-- | Symbol table
--
-- The inner list represents the symbol replacements.
type SymTabFrame = [[LNSymbol]]

type SymTab = [(Int,          -- ^ # of following SymTabFrame
                SymTabFrame)]

mapSymTab :: (a -> b) -> [(Int, a)] -> [(Int, b)]
mapSymTab f = map chSymTab where
              chSymTab (i, xs) = (i, f xs) 
              

-- | Returns a certain frame.
getSymTabFrame :: Int                   -- ^ # of frame
                  -> SymTab -> SymTabFrame
getSymTabFrame i st = case lookup i st of
                      Nothing  -> [[]]
                      Just stf -> stf
                      
-- | Finds symbols with properties defined by f in all frames.
findAllSymbols :: (LNSymbol -> Bool)       -- ^ f
               -> SymTab -> [LNSymbol]                      
findAllSymbols f st = concat $ map (findSymbols f) $ map snd st


-- | Finds symbols with properties defined by f in a certain frame.
findSymbols :: (LNSymbol -> Bool)       -- ^ f
               -> SymTabFrame -> [LNSymbol]                      
findSymbols f stf = concat $ map (filter f) stf                            


-- | Finds symbols with given lexem within a frame.
retrieveSymbol :: Int                   -- ^ # of frame
                  -> String             -- ^ Lexem
                  -> SymTab
                  -> Maybe LNSymbol
retrieveSymbol i s st = let stf = getSymTabFrame i st
                            res = findSymbols (\lns -> sym_lexem lns == s) stf                            
                        in 
                        case res of
                        []     -> Nothing
                        [ln] -> Just (ln)
                        _      -> error "retrieveSymbol: More than one result found!"


-- | Sorts a list of LNSymbols according to its sym_name entry
sortSymbols :: [LNSymbol] -> [LNSymbol]
sortSymbols = sortBy compSymName
    where
    compSymName x y = compare (sym_name x) (sym_name y)


-- | Finds a symbol by name and type
findSymbol :: (Int, LambdaNodeState)    -- ^ Name, Type
              -> SymTab -> Maybe LNSymbol
findSymbol (symname, symtype) st = let isSym = isSymbol (symname, symtype)
                                       filteredst = mapSymTab (findSymbols isSym) st
                                       (_, filtered) = unzip filteredst                                       
                                   in case (concat filtered) of
                                   []       -> Nothing
                                   [ln]     -> Just ln
                                   _        -> error "findSymbol: More than one result found!"

-- | Finds an index
findIndex :: SymTab -> Int -> Maybe LNSymbol
findIndex st indexid = findSymbol (indexid, LNS_Index) st

-- | Returns the index lexem
indexLexem :: SymTab -> Int -> String
indexLexem st indexid = case findIndex st indexid of
                      Nothing -> error ("No index with " ++ (show indexid))
                      Just lns -> sym_lexem lns

-- | Returns all index lexems.
getIndexLexems :: SymTab -> [String]
getIndexLexems st = map sym_lexem $ sortSymbols $ findAllSymbols (\ls -> sym_type ls == LNS_Index) st

-- | Finds an blob
findBlob :: SymTab -> Int -> Maybe LNSymbol
findBlob st blobid = findSymbol (blobid, LNS_Blob) st

-- | Returns the blob lexem
blobLexem :: SymTab -> Int -> String
blobLexem st blobid = case findBlob st blobid of
                      Nothing -> error ("No blob with " ++ (show blobid))
                      Just lns -> sym_lexem lns

-- | Returns all blob lexems.
getBlobLexems :: SymTab -> [String]
getBlobLexems st = map sym_lexem $ sortSymbols $ findAllSymbols (\ls -> sym_type ls == LNS_Blob) st

-- | Finds an array
findArray :: SymTab -> ArrayID -> Maybe LNSymbol
findArray st arrid = findSymbol (integerFromID arrid, LNS_Array) st

-- | Returns the array lexem
arrayLexem :: SymTab -> ArrayID -> String
arrayLexem st arrid = case findArray st arrid of
                      Nothing -> error ("No array with " ++ (show arrid))
                      Just lns -> sym_lexem lns

-- | Returns all array lexems.
getArrayLexems :: SymTab -> [String]
getArrayLexems st = map sym_lexem $ sortSymbols $ findAllSymbols (\ls -> sym_type ls == LNS_Array) st


-- | Finds a function
findFunction :: SymTab -> Int  -> Maybe LNSymbol
findFunction st funcid = findSymbol (funcid, LNS_Function) st

-- | Returns the function lexem
functionLexem :: SymTab -> Int -> String
functionLexem st funcid = case findFunction st funcid of
                           Nothing -> error ("No function with " ++ (show funcid))
                           Just lns -> sym_lexem lns

-- | Finds the number of a frame which holds a certain symbol.
findSymbolFrameNumber :: (Int, LambdaNodeState)    -- ^ Name, Type
                          -> SymTab -> Maybe Int
findSymbolFrameNumber (symname, symtype) st = 
            let isSym = isSymbol (symname, symtype)
                (nums, stfs) = unzip st
                onlyIsSym = map (map (not.null.(filter isSym))) stfs
                hasSym    = map (or) onlyIsSym
                (_, numhS) = unzip $ filter (fst) $ zip hasSym nums
            in case numhS of
               []   -> Nothing
               [i]  -> Just i
               _    -> error "findSymbolFrameNumber: More than one result found!"


-- | Finds the frame which holds a certain symbol.
findSymbolFrame :: (Int, LambdaNodeState)    -- ^ Name, Typ
                   -> SymTab -> Maybe SymTabFrame
findSymbolFrame (symname, symtype) st = 
            let mi = findSymbolFrameNumber (symname, symtype) st
            in case mi of
            Nothing  -> Nothing
            (Just i) -> Just (getSymTabFrame i st)
                          

-- | Finds replacements of a certain symbol (returns also the symbol itself)
findSymbolReplacements :: (Int, LambdaNodeState)    -- ^ Name, Type
                          -> SymTab -> [LNSymbol]
findSymbolReplacements (symname, symtype) st = 
           let isSym    = isSymbol (symname, symtype)
               isIn     = findSymbolFrame (symname, symtype) st
               (Just stf) = isIn
               onlyIsSym = map (not.null.(filter isSym)) stf
           in case isIn of
              Nothing -> []
              _       -> maybe [] id $ lookup True $ zip onlyIsSym stf

-- | identifies a symbol by name and type
isSymbol :: (Int, LambdaNodeState) -> LNSymbol -> Bool
isSymbol (i, st) lns = sym_type lns == st && sym_name lns == i


-- | Renames a frame
renameSymbolFrame :: Int            -- ^ old number
                     -> Int         -- ^ new number
                     -> SymTab -> SymTab
renameSymbolFrame o n st = let ren i = if i == o then n else i
                               (nms, cont) = unzip st
                               newnms      = map ren nms                               
                           in case (elem n nms) of
                              False -> zip newnms cont
                              True  -> error "renameSymbolFrame: New name is already in use"

-- | Deletes a complete symbol frame.
deleteSymbolFrame :: Int -> SymTab -> SymTab
deleteSymbolFrame i st = filter (\x -> fst x /= i) st

-- | Hilfsfunktion; entfernt leere Symbollisten
flushEmptySymRepl :: SymTabFrame -> SymTabFrame
flushEmptySymRepl = filter (not.null)

-- | Hilfsfunktion; entfernt leere Frames
flushEmptySymTabFrames :: SymTab -> SymTab
flushEmptySymTabFrames st = filter (not.null.snd) $ mapSymTab flushEmptySymRepl st

-- | Deletes all symbols in a frame with properties given by f 
deleteFrameSymbolsBy :: (LNSymbol -> Bool) -> SymTabFrame -> SymTabFrame
deleteFrameSymbolsBy f stf = flushEmptySymRepl $ map (filter (not.f)) stf

-- | Deletes all symbols with properties given by f 
deleteSymbolsBy :: (LNSymbol -> Bool) -> SymTab -> SymTab
deleteSymbolsBy f st = flushEmptySymTabFrames $ mapSymTab (deleteFrameSymbolsBy f) st

-- | Deletes a by name and type defined symbol
deleteSymbol :: (Int, LambdaNodeState) -> SymTab -> SymTab
deleteSymbol lns st = deleteSymbolsBy (isSymbol lns) st

-- | Inserts a symbol into a frame. It is  _not_ checked if there is already a
-- symbol with same name and type.
--
-- Used by insertSymbol and insertSymbolReplacement.
insertSymbolToFrame :: LNSymbol                         -- ^ to inserting symbol
                       -> Maybe (Int, LambdaNodeState)  -- ^ Symbol which is a replacement.
                       -> SymTabFrame -> SymTabFrame
insertSymbolToFrame lns repl stf = 
        let rplmts     = case repl of
                         Nothing  -> []
                         Just rpl -> findSymbolReplacements rpl [(0,stf)] -- knstlicher SymTab
            newstf     = deleteFrameSymbolsBy (\ln -> elem ln rplmts) stf -- Replacements entfernt
        in (lns:rplmts):newstf
        

-- | Inserts a symbol into the given frame in the SymTab. 
insertSymbol :: Int                              -- ^ # of frame
                -> LNSymbol                      -- ^ Symbol to insert
                -> Maybe (Int, LambdaNodeState)  -- ^ Symbol which is a replacement.
                -> SymTab -> SymTab
insertSymbol i lns repl st = if findSymbolFrameNumber (sym_name lns,sym_type lns) st == Nothing
                             then map isS st
                             else error "insertSymbol: Symbol with this Type and Name already in table!"
                             where
                             isS (fn,stf) = if fn/=i 
                                            then (fn,stf)
                                            else (fn,insertSymbolToFrame lns repl stf)

-- | Modifies SymTabFrames of a SymTab by using a given function
modifySymTabFrames :: (SymTabFrame -> SymTabFrame) -> SymTab -> SymTab
modifySymTabFrames f = mapSymTab f

-- | Modifies replacements of a SymTab by using a given function
modifySymbolReplacements :: ([LNSymbol] -> [LNSymbol]) -> SymTab -> SymTab
modifySymbolReplacements f = modifySymTabFrames (map f)

-- | Modifies symbols of a SymTab by using a given function
modifyLNSymbols :: (LNSymbol -> LNSymbol) -> SymTab -> SymTab
modifyLNSymbols f = modifySymbolReplacements (map f)

-- | Converts an xml node into LNSymbol structure, if possible.
xml2LNSymbol :: XML -> Maybe LNSymbol
xml2LNSymbol xmlnode@(Tree (ELNode "LNSymbol" args) _) = 
                 let getAttVal s = maybe "" id $ attval s xmlnode
                 in
                 Just (Symbol {sym_lexem    =     (getAttVal "lexem"),
                               sym_token    =read (getAttVal "token"),
                               sym_type     =read (getAttVal "type"),
                               sym_name     =read (getAttVal "name"),
                               sym_indices  =read (getAttVal "indices"),
                               sym_level    =read (getAttVal "level"),
                               sym_intent   =read (getAttVal "intent"),
                               sym_basictype=read (getAttVal "basictype"),
                               sym_identify =read (getAttVal "identify"),
                               sym_status   =LNSS {sym_valid    =read (getAttVal "valid"),
                                                   sym_pure     =read (getAttVal "pure"),
                                                   sym_no_warn  =read (getAttVal "no_warn"),
                                                   sym_dynamic  =read (getAttVal "dynamic"),
                                                   sym_local    =read (getAttVal "local"),
                                                   sym_pseudo   =read (getAttVal "pseudo"),
                                                   sym_normLoops=read (getAttVal "normLoops"),
                                                   sym_assign_idx=read (getAttVal "assign_idx")}
                              })
xml2LNSymbol _ = Nothing

-- | Hilfsfunktion fr xmlToSymReplacements, xmlToSymReplacements
lst2MaybeList :: Eq a => [Maybe a] -> Maybe [a]
lst2MaybeList xs = if (elem Nothing xs) 
                   then Nothing
                   else Just (map (\ (Just b) -> b) xs)

xmlToSymReplacements :: XML -> Maybe [LNSymbol]
xmlToSymReplacements (Tree (ELNode "Replacements" _) xmls)
                        = lst2MaybeList $ map xml2LNSymbol (onlyELNodes xmls)
xmlToSymReplacements _ = Nothing

xmlToSymTabFrame :: XML -> Maybe (Int, SymTabFrame)
xmlToSymTabFrame (Tree (ELNode "Frame" args) xmls) = 
                     let (Just i)   = lookup "Number" args
                         (Just stf) = lst2MaybeList $ map xmlToSymReplacements (onlyELNodes xmls)
                     in Just (read i, stf)
xmlToSymTabFrame _ = Nothing


-- | Converts xml tree into SymTab, if possible.
symTabFromXML :: XML -> Maybe SymTab
symTabFromXML xml = case xmltrees of
              [] -> Nothing
              _  -> lst2MaybeList $ map xmlToSymTabFrame xmltrees
              where
              xmltrees = findXMLSubnodes p xml -- sucht das erste Vorkommen von LambdaTree
              p (Tree (ELNode "SymTab" _) _) = True
              p _                                = False


-- | Converts a single LNSymbol into XML.
lnSymbolToXML :: LNSymbol -> XML
lnSymbolToXML lns = Tree (ELNode "LNSymbol" [("lexem", sym_lexem lns),
                                             ("token", show (sym_token lns)),
                                             ("type", show (sym_type lns)),
                                             ("name", show (sym_name lns)), 
                                             ("indices", show (sym_indices lns)), 
                                             ("level", show (sym_level lns)),
                                             ("intent", show (sym_intent lns)),
                                             ("basictype", show (sym_basictype lns)),
                                             ("identify", show (sym_identify lns)),
                                             ("valid", show (sym_valid (sym_status lns))), 
                                             ("pure", show (sym_pure (sym_status lns))), 
                                             ("no_warn", show (sym_no_warn (sym_status lns))), 
                                             ("dynamic", show (sym_dynamic (sym_status lns))), 
                                             ("local", show (sym_local (sym_status lns))), 
                                             ("pseudo", show (sym_pseudo (sym_status lns))), 
                                             ("normLoops", show (sym_normLoops (sym_status lns))), 
                                             ("assign_idx", show (sym_assign_idx (sym_status lns)))
                                            ]) []

symTabFrameToXML :: SymTabFrame -> [XML]
symTabFrameToXML = let replLstToXML lnss = Tree (ELNode "Replacements" []) 
                                                (map lnSymbolToXML lnss)
                   in map replLstToXML
     
-- | Converts a SymTab into XML.
symTabToXML :: SymTab -> XML
symTabToXML st = let stf2XML (i, stf) = Tree (ELNode "Frame" [("Number", show i)])
                                            (symTabFrameToXML stf)
                in Tree (ELNode "SymTab" []) (map stf2XML st)

testsymbol = Symbol {sym_lexem = "Z", sym_token = 261, sym_type = LNS_Proto, sym_name = 500,sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = True, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}

testtable::SymTab
testtable = [(1,[[Symbol {sym_lexem = "min", sym_token = 263, sym_type = LNS_Function, sym_name = -1, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "max", sym_token = 263, sym_type = LNS_Function, sym_name = -2, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "ceil", sym_token = 263, sym_type = LNS_Function, sym_name = -3, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "IntConv", sym_token = 263, sym_type = LNS_Function, sym_name = -16, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "FloatConv", sym_token = 263, sym_type = LNS_Function, sym_name = -15, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Float, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "FALSE", sym_token = 263, sym_type = LNS_Function, sym_name = -14, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Bool, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "AtomMax", sym_token = 263, sym_type = LNS_Function, sym_name = -8, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Void, sym_identify =0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False,sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "Assign", sym_token = 263, sym_type = LNS_Function, sym_name = -17, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Void, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "A", sym_token = 263, sym_type = LNS_Array, sym_name = 2, sym_indices = 1, sym_level = 1, sym_intent = Nothing, sym_basictype = Nothing, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "AtomMin", sym_token = 263, sym_type = LNS_Function, sym_name = -7, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Void, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "C", sym_token = 0, sym_type = LNS_Array, sym_name = 1, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "GetValid", sym_token = 263, sym_type = LNS_Function, sym_name = -10, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Bool, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False,sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "TRUE", sym_token = 263, sym_type = LNS_Function, sym_name = -13, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Bool, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False,sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "MinInt", sym_token = 263, sym_type = LNS_Function, sym_name = -12, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "MaxInt", sym_token = 263, sym_type = LNS_Function, sym_name = -11, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx= False}}],[Symbol {sym_lexem = "LCM", sym_token = 263, sym_type = LNS_Function, sym_name = -9, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False,sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "ceild", sym_token = 263, sym_type = LNS_Function, sym_name = -18, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "floor", sym_token = 263, sym_type = LNS_Function, sym_name = -4, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "floord", sym_token = 263, sym_type = LNS_Function, sym_name = -19, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "i", sym_token = 260, sym_type = LNS_Index, sym_name = 1, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Nothing, sym_identify = 0, sym_status = LNSS {sym_valid = True, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "loopo_nl_flba_39", sym_token = 261, sym_type = LNS_Array, sym_name = 3, sym_indices= 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Float, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = True, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "loopo_nl_fuba_39", sym_token = 261, sym_type = LNS_Array, sym_name = 4, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Float, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = True, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "loopo_nl_for_ext_ub_39", sym_token = 261, sym_type = LNS_Array, sym_name = 6, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Float, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "loopo_nl_valid_for_ext_ub_39", sym_token = 261, sym_type = LNS_Array, sym_name = 5, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Bool, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "loopo_nl_fubb_o_a_39", sym_token = 261, sym_type = LNS_Blob, sym_name = 1, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = True, sym_assign_idx = False}}],[Symbol {sym_lexem = "sign", sym_token = 263, sym_type = LNS_Function, sym_name = -6, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "return", sym_token = 263, sym_type = LNS_Function, sym_name = -5, sym_indices = -1, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Void, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = False, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}]]),(-17,[[Symbol {sym_lexem = "A", sym_token = 261, sym_type = LNS_Proto, sym_name = 1, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Float, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = True, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}]]),(-16,[[Symbol {sym_lexem = "A", sym_token = 261, sym_type = LNS_Proto, sym_name = 2, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = True, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}]]),(-11,[[Symbol {sym_lexem = "A", sym_token = 261, sym_type = LNS_Proto, sym_name = 3, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = True, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}]]),(-10,[[Symbol {sym_lexem = "A", sym_token = 261, sym_type = LNS_Proto, sym_name = 4, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = True, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "B", sym_token = 261, sym_type = LNS_Proto, sym_name = 5, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = True, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}]]),(-9,[[Symbol {sym_lexem = "A", sym_token = 261, sym_type = LNS_Proto, sym_name = 6, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = True, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "B", sym_token = 261, sym_type = LNS_Proto, sym_name = 7, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = True, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}]]),(-8,[[Symbol {sym_lexem = "A", sym_token = 261, sym_type = LNS_Proto, sym_name = 8, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = True, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "B", sym_token = 261, sym_type = LNS_Proto, sym_name = 9, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = True, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}]]),(-18,[[Symbol {sym_lexem = "A", sym_token = 261, sym_type = LNS_Proto, sym_name = 10, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = True, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}],[Symbol {sym_lexem = "B", sym_token = 261, sym_type = LNS_Proto, sym_name = 11, sym_indices = 0, sym_level = 0, sym_intent = Nothing, sym_basictype = Just LNST_Int, sym_identify = 0, sym_status = LNSS {sym_valid = False, sym_pure = False, sym_no_warn = True, sym_dynamic = False, sym_local = False, sym_pseudo = False, sym_normLoops = False, sym_assign_idx = False}}]])]
