-- |
-- This module is a draft of a new code generation method that
-- tries to overcome the code explosion in ClooG. The idea is
-- to call Cloog several times: 
--
--   * once for all statements, but each first projected to the 
--     outermost dimenstions
--
--   * once for the COMPUTE statements (with the outermost dims as parameters)
--
--   * once for the RECEIVE statement (        \"             )
--
--   * once for the SEND statement (        \"             )
--
-- This avoids unnecessary high dimensionalities as well as the duty
-- to merge all statements simultaneously.
-- Author: Martin Griebl
--
-- Version: $Id: StagedCodeGen.hs 3994 2008-05-21 12:41:18Z classenm $
--
module StagedCodeGen (computeStmtsTileDims, computeStmtsInnerDims)
where

import LoopoDS
import Debug.Trace(trace)
import System.Cmd(system)
-- import TrDepAux
import TrDepDirections(relevantTrDeps)
import LMath
import Tiling
import Maybe
import Ratio
import List
import System(getArgs)

import CompSpread.Data
import CompSpread.CloogInput
import CompSpread.Run
import CompSpread.DomainDescriptions
import CompSpread.Helpers
import CompSpread.Defines

import InsertIfs.Parser(parseSource)
import InsertIfs.ParserDefs(ParserState(..))

import VisualPolylib(showIS2VisualPolylib)

import ScannerParser.SymbolTable
import ScannerParser.ScannerParser
import ScannerParser.ParseTree

import StagedCodeGenInput

main :: IO ()
main = do
       inputStr <- getArgs
       let cl = parseCommandLine inputStr
           csDS = unsafeCreateCompSpreadDS cl

           loopoFile = csTempDir csDS ++ "/" ++ csProjFile csDS           
           outerDimInput = outerTileDimInput csDS
           outerDimRun = RunExternal { runInputExtension = "." ++ outerDimInpExt
                                     , runOutputExtension = "." ++ outerDimOutExt
                                     }
           innerTileInputs = innerTileDimInputs csDS
           innerTileRuns = 
               [ RunExternal { runInputExtension = "." ++ innerTileInpExt ++ show i
                             , runOutputExtension = "." ++ innerTileOutExt ++ show i
                             }
                 | i <- [1..length innerTileInputs]
               ]
           bufferInputs = bufferDimInputs csDS
           bufferRuns = 
               [ RunExternal { runInputExtension = "." ++ bufferInpExt ++ show i
                             , runOutputExtension = "." ++ bufferOutExt ++ show i
                             }
                 -- count starting at 2 (1 is for compute stmts)
                 | i <- [2..length bufferInputs+1]
               ]
       -- putStrLn (showIS2VisualPolylib (snd $ head $ csISPCs csDS) (nVars (snd $ head $ csISPCs csDS)))
       (symtab, pt, _, _) <- openParseFile loopoFile    
       let stmts = findMainNodes (\ln -> hasstatement ln && statement ln /= (integerToID 0)) pt

       -- putStrLn (show $ csTrafos csDS)
       -- putStrLn $ printCode symtab stmts

       writeStagedDefinesFiles csDS
       writeFile (loopoFile ++ "." ++ outerDimInpExt) outerDimInput
       putStrLn "running CLooG for outer tile dims..."
       runCloog csDS outerDimRun 
       runStagedInsertIFs csDS
       putStrLn "outer tile dims done."
       sequence [writeFile (loopoFile ++ "." ++ innerTileInpExt ++ show i) inp |
                 (inp,i) <- zip innerTileInputs [1.. length innerTileInputs]]
       putStrLn "running CLooG for inner tile dims..."
       sequence [runCloog csDS run | run <- innerTileRuns]
       putStrLn "insertIncludes: outerDimOutExt innerTileOutExt"
       insertIncludes csDS finalOutExt innerTileOutExt
       putStrLn "inner tile dims done."
       sequence [writeFile (loopoFile ++ "." ++ bufferInpExt ++ show i) inp |
                 -- count starting at 2 (1 is for compute stmts)
                 (inp,i) <- zip bufferInputs [2.. length innerTileRuns]]
       putStrLn "running CLooG for buffer stmt inner dims..."
       sequence [runCloog csDS run | run <- bufferRuns]
--       sequence [insertInnerIncludes csDS innerTileOutExt bufferOutExt i |
--                 -- count starting at 2 (1 is for compute stmts)
--                 i <- [2 .. genericLength innerTileRuns]]
       putStrLn "substituteInnerDims for sender side"
       substituteInnerDims csDS True

       putStrLn "substituteInnerDims for receiver side"
       substituteInnerDims csDS False

       substituteCompStmts csDS

       substituteBufferStmts csDS True
       substituteBufferStmts csDS False
       
       return ()


{--
insertIncludes :: CompSpreadDS -> String -> String -> IO ()
insertIncludes csDS inpFileExt includeFileExt =
    do
    let inpFile = csTempDir csDS ++ "/" ++ csProjFile csDS ++ inpFileExt
        includeFile = 
            concat $ map (\c -> if c `elem` "/." then "\\" ++ [c] else [c]) $
            csProjFile csDS ++ includeFileExt
        outFile = inpFile ++ ".sed"
        sedCommand =
            ("sed -r " ++
             "-e 's/^[[:space:]]*S([0-9]+).*;/\\n#include \"" ++ includeFile ++ "\\1\\.sed\"/' " ++
            inpFile ++ " > " ++ outFile)
    exitCode <- do trace ("running sed: " ++ show sedCommand) $ system sedCommand
    let (ok, errCode) = checkResult exitCode
    if ok then do putStrLn "sed completed."
          else do putStrLn ("sed exited with error code: " ++ 
			    (show errCode))   
--}

insertIncludes :: CompSpreadDS -> String -> String -> IO ()
insertIncludes csDS inpFileExt includeFileExt =
    do
    let inpFile = csTempDir csDS ++ "/" ++ csProjFile csDS ++ "." ++ inpFileExt
        outFile = inpFile ++ ".sed"
        sedCommand =
            ("sed -r " ++
             "-e 's/^[[:space:]]*S([0-9]+).*;/\\n" ++ includeFileExt ++ "\\1();/' " ++
            inpFile ++ " > " ++ outFile)
    exitCode <- do trace ("running sed: " ++ show sedCommand) $ system sedCommand
    let (ok, errCode) = checkResult exitCode
    if ok then do putStrLn "sed completed."
          else do putStrLn ("\n\n\nsed exited with error code: " ++ 
			    (show errCode) ++ "\n\n\n")   



insertInnerIncludes :: CompSpreadDS -> String -> String -> Integer -> IO ()
insertInnerIncludes csDS inpFileExt includeFileExt stmtNr =
    do
    let inpFile = 
            csTempDir csDS ++ "/" ++ csProjFile csDS ++ "." ++
            inpFileExt ++ (show stmtNr)
--        includeFile = 
--            concat $ map (\c -> if c `elem` "/." then "\\" ++ [c] else [c]) $
--            csProjFile csDS ++ includeFileExt
        outFile = inpFile ++ ".sed"
        sedCommand =
            ("sed -r " ++
             "-e 's/^[[:space:]]*S([0-9]+).*;/\\n" ++ includeFileExt ++ 
             (show stmtNr) ++ "();/' " ++
            inpFile ++ " > " ++ outFile)
    exitCode <- do trace ("running sed: " ++ show sedCommand) $ system sedCommand
    let (ok, errCode) = checkResult exitCode
    if ok then do putStrLn "sed completed."
          else do putStrLn ("\n\n\nsed exited with error code: " ++ 
			    (show errCode) ++ "\n\n\n")   


substituteInnerDims :: CompSpreadDS -> Bool -> IO ()
substituteInnerDims csDS isSender =
    do
    let inpFile = 
            csTempDir csDS ++ "/" ++ csProjFile csDS ++ "." ++
            innerTileOutExt ++ (show extNumber)
        includeFile = 
            concat $ map (\c -> if c `elem` "/." then "\\" ++ [c] else [c]) $
            csProjFile csDS ++ "." ++ stmtExt
        outFile = inpFile ++ ".sed"
        extNumber = if isSender then 2 else 3
        stmtExt = if isSender then sendExt else recvExt
        sedCommand =
            ("sed -r " ++
             (if isSender 
              then ("-e 's/^[[:space:]]*S([0-9]+).*;/\\n" ++ bufferOutExt ++ 
                    (show extNumber) ++ "();" ++ 
                    "\\n#include \"" ++ includeFile ++ "\""
                   ) 
              else ("-e 's/^[[:space:]]*S([0-9]+).*;/\\n#include \"" ++ includeFile ++ "\"" ++ 
                    "\\n" ++ bufferOutExt ++ (show extNumber) ++ "();"
                   )
             ) ++ "/' " ++ inpFile ++ " > " ++ outFile)
    putStrLn "substitute inner dims..."
    putStrLn ("sedCommand: " ++ sedCommand)
    exitCode <- do trace ("running sed: " ++ show sedCommand) $ system sedCommand
    let (ok, errCode) = checkResult exitCode
    if ok then do putStrLn "sed completed."
          else do putStrLn ("\n\n\nsed exited with error code: " ++ 
			    (show errCode) ++ "\n\n\n")   

substituteCompStmts :: CompSpreadDS -> IO ()
substituteCompStmts csDS =
    do
    let inpFile = csTempDir csDS ++ "/" ++ csProjFile csDS ++ "." ++ 
                  innerTileOutExt ++ "1"
        outFile = inpFile ++ ".sed"
        sedCommand =
            ("sed -r " ++
             "-e 's/^[[:space:]]*S([0-9]+)(.+;)/" ++ compStmt_t ++ "\\1\\2/' " ++
            inpFile ++ " > " ++ outFile)
    exitCode <- do trace ("running sed: " ++ show sedCommand) $ system sedCommand
    let (ok, errCode) = checkResult exitCode
    if ok then do putStrLn "sed completed."
          else do putStrLn ("\n\n\nsed exited with error code: " ++ 
			    (show errCode) ++ "\n\n\n")   
    return ()


-- substitute each "Sx" cloog statement by a "writeBuffer" or "unpackBuffer" statement, 
-- depending on the "isBuffer" argument.
substituteBufferStmts :: CompSpreadDS -> Bool -> IO ()
substituteBufferStmts csDS isSender =
    do
    let inpFile = csTempDir csDS ++ "/" ++ csProjFile csDS ++ "." ++ 
                  (bufferOutExt ++ show extNumber)
        outFile = inpFile ++ ".sed"
        extNumber = if isSender then 2 else 3
        stmtName = if isSender then writeBufferStmt_t else unpackBufferStmt_t
--        includeFile = 
--            concat $ map (\c -> if c `elem` "/." then "\\" ++ [c] else [c]) $
--            csProjFile csDS ++ "." ++ sendExt
        sedCommand =
            ("sed -r " ++
             "-e 's/^[[:space:]]*S([0-9]+)(.+;)/" ++ stmtName ++ "\\1\\2" ++ 
             -- " \\n#include \"" ++ includeFile ++ "\"" ++ 
             "/' " ++ inpFile ++ " > " ++ outFile)
    exitCode <- do trace ("running sed: " ++ show sedCommand) $ system sedCommand
    let (ok, errCode) = checkResult exitCode
    if ok then do putStrLn "sed completed."
          else do putStrLn ("\n\n\nsed exited with error code: " ++ 
			    (show errCode) ++ "\n\n\n")   
    return ()



writeStagedDefinesFiles :: CompSpreadDS -> IO ()
writeStagedDefinesFiles csDS =
    do origStmtsCode <- getOrigStatementsCode csDS -- from Compspread/Run.hs
       let loopoFile = csTempDir csDS ++ "/" ++ csProjFile csDS
           defineHeader = "/* created: " ++ getTime ++ " */\n"
           compStmtsDefines = stagedCompStmtsDefines csDS origStmtsCode
           writeStmtsDefines = stagedWriteStmtsDefines csDS
           unpackStmtsDefines = stagedUnpackStmtsDefines csDS
           sendStmtsDefines = stagedSendStmtsDefines csDS
           recvStmtsDefines = stagedRecvStmtsDefines csDS
       writeFile (loopoFile ++ "." ++ compExt) (defineHeader ++ compStmtsDefines)
       writeFile (loopoFile ++ "." ++ writeExt) (defineHeader ++ writeStmtsDefines)
       writeFile (loopoFile ++ "." ++ unpackExt) (defineHeader ++ unpackStmtsDefines)
       -- not really defines, just code:
       writeFile (loopoFile ++ "." ++ sendExt) sendStmtsDefines
       writeFile (loopoFile ++ "." ++ recvExt) recvStmtsDefines
       return ()

stagedCompStmtsDefines :: CompSpreadDS -> [(StatementID, String)] -> String
stagedCompStmtsDefines csDS origStatementsCode = 
    verbose "compStmtsDefines" $
    unlines compStmtsDefines
    where
    allTrStmtIDs = getAllTStatementIDs csDS
    compStmtsDefines = map oneCompStmtDefine allTrStmtIDs

    oneCompStmtDefine :: TStatementID -> String
    oneCompStmtDefine trStmtID = 
        concat 
	[ compDefineHeader csDS trStmtID compStmt_t
	, compStmtIndexMapping csDS trStmtID (innerComputeIteratorNames csDS)
	, fromJust $ lookup (getStmtID csDS trStmtID) origStatementsCode
	]

compDefineHeader :: CompSpreadDS -> TStatementID -> String -> String
compDefineHeader csDS trStmtID stmtName =
    ("#define " ++ stmtName ++ (show $ integerFromID trStmtID) ++ "(" ++
     iteratorNames ++ ") ")
    where
    iteratorNames = concat $ intersperse ", " $ relevantIteratorNames
    allIterators = innerComputeIteratorNames csDS
    isDim = getISPCdim csDS trStmtID
    coordsDim = getCoordsDim csDS trStmtID
    -- tileDim = genericLength $ csTiledDims csDS
    relevantIteratorNames = genericTake (isDim + coordsDim) allIterators

stagedWriteStmtsDefines :: CompSpreadDS -> String
stagedWriteStmtsDefines csDS =
    verbose "writeStmtsDefines" $
    unlines writeStmtsDefines
    where
    --allTrStmtIDs = getAllTStatementIDs csDS
    writeStmtsDefines = map oneWriteStmtDefine (zip allTrDepIDs [1..])
    filteredTrDeps = getCommunicationTrDeps $ csTrDeps csDS
    allTrDepIDs = map trDepID filteredTrDeps

    oneWriteStmtDefine :: (TrDependenceID, Integer) -> String
    oneWriteStmtDefine (trDepID, stmtNr) = 
        concat 
	[ bufferStmtDefineHeader csDS (trDepID,stmtNr) writeBufferStmt_t
        , writeBuffer_t
	]
        where
        arrayName = 
	    genericIndex (csArrayNames csDS) (integerFromID arrayID - 1)
        arrAccNames = arrAcc_t                      
        srcAcc = getSrcAcc csDS trDepID
        accDim = targetDimIH $ accIndexFct srcAcc
        arrayID = accArrayID $ srcAcc
        destProc_t = head $ drop 1 (innerScatterNames csDS)
        writeBuffer_t = 
          -- e.g.: sendVect[destRP].push_back(A[arrAcc1][arrAcc2]);        
          (sendVector_t ++ "[" ++ destProc_t ++ "].push_back(" ++ arrayName ++
	   concat ["[" ++ arrAccNames ++ (show i) ++ "]" | i <- [1..accDim]] ++
	   ");")

stagedUnpackStmtsDefines :: CompSpreadDS -> String
stagedUnpackStmtsDefines csDS =
    verbose "unpackStmtsDefines" $
    unlines unpackStmtsDefines
    where
    --allTrStmtIDs = getAllTStatementIDs csDS
    unpackStmtsDefines = map oneUnpackStmtDefine (zip allTrDepIDs [1..])
    filteredTrDeps = getCommunicationTrDeps $ csTrDeps csDS
    allTrDepIDs = map trDepID filteredTrDeps

    oneUnpackStmtDefine :: (TrDependenceID, Integer) -> String
    oneUnpackStmtDefine (trDepID, stmtNr) = 
        concat 
	[ bufferStmtDefineHeader csDS (trDepID,stmtNr) unpackBufferStmt_t
        , unpackBuffer_t
	]
        where
        arrayName = 
	    genericIndex (csArrayNames csDS) (integerFromID arrayID - 1)
        arrAccNames = arrAcc_t                      
        srcAcc = getSrcAcc csDS trDepID
        accDim = targetDimIH $ accIndexFct srcAcc
        arrayID = accArrayID $ srcAcc
        srcProc_t = head $ drop 1 (outerScatterNames csDS)
        unpackBuffer_t =
            (arrayName ++ 
             concat ["[" ++ arrAccNames ++ (show i) ++ "]" | i <- [1..accDim]] ++
	     " = " ++
             recvVector_t ++ "["++ srcProc_t ++ "]" ++ 
             "[" ++ rdispls_t ++ "[" ++ srcProc_t ++ "]++];")


bufferStmtDefineHeader :: CompSpreadDS -> (TrDependenceID, Integer) -> String -> String
bufferStmtDefineHeader csDS (trDepID,stmtNr) stmtName =
    ("#define " ++ stmtName ++ (show stmtNr) ++ "(" ++
     iteratorNames ++ ") ")
    where
    iteratorNames = concat $ intersperse ", " $ relevantIteratorNames
    allIterators = bufferIteratorNames csDS
    srcTStmtID = trDepSrcStmtID (getTrDep csDS trDepID)
    isDim = getISPCdim csDS srcTStmtID
    coordsDim = getCoordsDim csDS srcTStmtID
    tileDim = genericLength $ csTiledDims csDS
    relevantIteratorNames = genericTake (isDim + coordsDim + tileDim + tileDim) allIterators

stagedSendStmtsDefines :: CompSpreadDS -> String
stagedSendStmtsDefines csDS = send_t
    where
    srcProc_t = head $ drop 1 (outerScatterNames csDS)
    destProc_t = head $ drop 1 (innerScatterNames csDS)
    destRP_t = destProc_t -- for now, only tiling is used for mapping to processors...
    send_t = 
        (mpiSend_t ++ "(&" ++ sendVector_t ++ "[" ++ destProc_t ++ "][0], " ++
         sendVector_t ++ "[" ++ destProc_t ++ "].size(), " ++
         "MPI_DOUBLE, " ++ 
         destRP_t ++ ", " ++
         srcProc_t ++ ", " ++ -- tag=srcProcTile
         mpiComm_t ++ ");\n")


stagedRecvStmtsDefines :: CompSpreadDS -> String
stagedRecvStmtsDefines csDS = recv_t
    where
    srcProc_t = head $ drop 1 (innerScatterNames csDS)
    srcRP_t = srcProc_t -- for now, only tiling is used for mapping to processors...
    destProc_t = head $ drop 1 (outerScatterNames csDS)
    destRP_t = destProc_t -- for now, only tiling is used for mapping to processors...
    recv_t = 
        (mpiProbe_t ++ "(" ++ srcRP_t ++ ", " ++ srcProc_t ++ ", "++
         mpiComm_t ++ ", &" ++ mpiStatus_t ++ "); " ++
         mpiGetCount ++ "(&" ++ mpiStatus_t ++ ", " ++ "MPI_INT, " ++
         "&" ++ mpiCount_t ++ "); " ++
         recvVector_t ++ "[" ++ srcProc_t ++ "].resize(" ++ mpiCount_t ++ ");" ++
         mpiRecv_t ++ "(&" ++ recvVector_t ++ "[" ++ srcProc_t ++ "][0], " ++
         mpiCount_t ++ ", MPI_DOUBLE, " ++
         srcRP_t ++ ", " ++ 
         srcProc_t ++ ", " ++ 
         mpiComm_t ++ 
         ", &" ++ mpiStatus_t ++ "); " ++
         rdispls_t ++ "[" ++ srcProc_t ++ "]" ++ " = 0;\n"
        )


runStagedInsertIFs :: CompSpreadDS -> IO ()
runStagedInsertIFs csDS =
  do
  let baseFile = csTempDir csDS ++ "/" ++ csProjFile csDS
      inpFile = baseFile ++ "." ++ outerDimOutExt
      outputFile = baseFile ++ "." ++ finalOutExt
      rpDim = (genericLength $ csTiledDims csDS) - 1
      globalTimeDim = 1
      outerProc_t = head $ drop 1 (outerScatterNames csDS)

      ifConstraint_t = 
         unlines $
         [ ""
         , (" if(" ++ outerProc_t ++ " == " ++ mpiRank_t ++ ")")
         ]
         
  putStrLn "reading cloogOutput for parsing..."
  input <- readFile inpFile
  putStrLn ("parsing \"" ++ inpFile ++ "\"\n")
  let initState = 
          ParserState { psRPid = outerProc_t
		      , psIfConstraint = ifConstraint_t
                      , psUseCommunication = True
		      , psCloogStmtString = ""
		      }

      firstOutput = 
          trace "first parserPass" $
          parseSource input initState

      output = firstOutput
 
  writeFile outputFile output
  putStrLn ("parsing completed.\n")


{-
mainx :: String 
      -> String
      -> [Vector] 
      -> [Integer] 
 --     -> [IneqSystem] -- temporary for debugging
      -> String
mainx path projName lattice tiledDims =   -- assertedSends 
                     cloogInput
        -- speed: sends2 is double as fast as sends1
-}

{- 24-10-2006

mainx :: CompSpreadDS -> String
mainx csDS = cloogInput
  where
    cloogInput = 
        genGenericCloogInput (map (\x -> [x]) allStmts) iteratorNames scatterNames blobs minBlobSize scatterFcts

    iteratorNames = ["TimeTileIter", "ProcTileIter"]
    scatterNames = ["TimeTileScatter", "ProcTileScatter"]
    blobs = csBlobNames csDS
    minBlobSize = csMinBlobSize csDS
    scatterFcts = 
        Just (map (\is -> identityScatterFct (nVars is) (nParams is)) allStmts)

    allStmts = computes ++ assertedSends ++ recvs

    lds :: LoopoDS
    lds = unsafeReadLoopoDS (csTempDir csDS ++ csProjFile csDS)

    lattice = csLatVects csDS
    tiledDims = csTiledDims csDS
    
    slp   = loopoSLP lds
    deps  = map (trimDependence slp) $ loopoDEP lds
    tispc = loopoTISPC lds  -- has the reduced number of lines
    -- tispc = zipWith properTrafos ispc
    trafo = loopoTrafo lds
    -- ispc  = map snd (loopoISPC lds) 
    -- ACHTUNG: Statement-Namen nicht bercksichtigt!!!
    -- ACHTUNG: die TRCs sind nicht bercksichtigt!!!
    ispcs = map (trimIndexSpace slp) (loopoISPC lds)
    properTrafos = makeProperTrafos trafo
    trdeps1 = transformDeps tispc properTrafos deps
    dimT = fromIntegral (length lattice)
    trdeps2 = relevantTrDeps 0 (dimT-1) lds 

    filteredTrDeps = filterTrDeps trdeps2
                     [ UniformTrue, UniformFlow, UniformOutput
                     , NonUniformTrue, NonUniformFlow, NonUniformOutput]


    sends1, sends2 :: [IneqSystem]
    sends1 = 
        trace "sends1..." $
        map (\trd -> fst ( defaultTile lattice tiledDims (trDepRelation trd)  [SrcTile] )) trdeps1
    sends2 = 
        trace "sends2..." $
        map (\trd -> fst (trace ("trDep#" ++ (show $ integerFromID $ trDepID trd)) $
                          defaultTile lattice tiledDims (trDepRelation trd)  [SrcTile] )) trdeps2

    filteredSends = 
        trace "filteredSends..." $
        map (\trd -> fst (trace ("trDep#" ++ (show $ integerFromID $ trDepID trd)) $
                          defaultTile lattice tiledDims (trDepRelation trd)  [SrcTile] )) 
            filteredTrDeps

    recvs =
        trace "recvs..." $
        map (\trd -> fst ( defaultTile lattice tiledDims (trDepRelation trd)  [DestTile] )) filteredTrDeps

    assertedSends = if map getMatrix sends1 == map getMatrix sends2 
                    then -- sends2
                         filteredSends
		    else error ("descriptions do not match: \n" ++ show sends1 ++ show sends2)

    -- computes = map (fromJust . (\tisp -> projectToVars' HsFM tisp Nothing [0..dimT-1]) . (\(_,_,c) -> c) ) tispc
    computes = 
        trace "computes..." $
        computeStmtsTileDims csDS

24-10-2006 --}

    

-- example run:
-- mainx "/tmp/loopo1434" "cholesky1.src" [vector[5,0], vector[0,7]] [0,1]


