-- |
-- Contains various input data for StagedCodeGen.hs and\/or TaskLoops.hs
-- 
-- Author: Michael Classen
--
module StagedCodeGenInput

where

import LoopoDS
import Debug.Trace(trace)
import System.Cmd(system)
import TrDepDirections(relevantTrDeps)
import LMath
import Tiling
import Maybe
import Ratio
import List

import CompSpread.Data
import CompSpread.CloogInput
import CompSpread.Run
import CompSpread.DomainDescriptions
import CompSpread.Helpers
import CompSpread.Defines

import VisualPolylib(showIS2VisualPolylib)

-- |
-- Determines the types of HOC components that are supported.
--
data HOC_Type = MasterWorkerHOC
              | P2P_HOC
                deriving (Show, Eq)

-- |
-- Determines the type of communication for either communication of soak-data,
-- drain-data or p2p-communication between clients.
data Comm_Type = Comm_Soak
               | Comm_Drain
               | Comm_P2P
                 deriving (Show, Eq)


-- |
-- Determines the buffer type for either communication, execution stmts
-- or soak\/draining.
data Buffer_Type = CommunicationBuffer 
                 | ExecutionBuffer 
                 | SoakDrainBuffer

finalOutExt = "finalOut"

outerDimInpExt = "outerDimInp"
outerDimOutExt = "outerDimOut"

innerTileInpExt = "innerTileInp"
innerTileOutExt = "innerTileOut"

bufferInpExt = "bufferInp"
bufferOutExt = "bufferOut"

compExt = "compDef" 
writeExt = "writeDef"
unpackExt = "unpackDef"
sendExt = "sendDef"
recvExt = "recvDef"

compStmt_t         = "compStmt"
writeBufferStmt_t  = "writeBufferStmt"
unpackBufferStmt_t = "unpackBufferStmt"

arrAcc_t = "arrAcc"     
sendVector_t = "sendVector"     
recvVector_t = "recvVector"     
rdispls_t = "rdispls"
mpiComm_t = "MPI_COMM_WORLD"
                  
mpiSend_t  = "MPI_Bsend"
mpiRecv_t  = "MPI_Recv"
mpiProbe_t = "MPI_Probe"
mpiGetCount_t = "MPI_Get_count"
dmpiStatus_t = "mpiStatus"
mpiCount_t  = "mpiCount"
mpiRank_t = "loopo___rank"


outerIteratorNames csDS = ["outerTimeTileIter"] ++ 
                          ["outerProcTileIter" ++ (show i) | i <- [1..(getTilingDim csDS - 1)]]
outerScatterNames csDS = ["outerTimeTileScatter"] ++
                         ["outerProcTileScatter" ++ (show i) | i <- [1..(getTilingDim csDS - 1)]]

innerIteratorNames csDS = ["innerTimeTileIter"] ++ 
                          ["innerProcTileIter" ++ (show i) | i <- [1..(getTilingDim csDS - 1)]]
innerScatterNames csDS = ["innerTimeTileScatter"] ++
                         ["innerProcTileScatter" ++ (show i) | i <- [1..(getTilingDim csDS - 1)]]

innerComputeIteratorNames csDS = ["compIter" ++ (show i) | 
                                  i <- [1..(getMaxISPCdim csDS + getMaxCoordsDim csDS)]]
innerComputeScatterNames csDS = ["compScatter" ++ (show i) |
                                i <- [1..(getMaxCoordsDim csDS)]]

bufferIteratorNames csDS = ["bufferIter" ++ (show i) |
                           i <- [1..(getMaxISPCdim csDS + getMaxCoordsDim csDS
                                    --  + 2*(genericLength $ csTiledDims csDS)
                                    )]
                           ]
bufferScatterNames csDS = [arrAcc_t ++ (show i) | 
                           i <- [1..(getArrAccessMaxDim csDS)]
                          ]


-- anti and output dependences will be inserted as additional true dependences,
-- with the exception that they will have an access of the form "A0[0]"
getCommunicationTrDeps :: [TrDependence] -> [TrDependence]
getCommunicationTrDeps trDeps =
    filterTrDeps trDeps [ UniformTrue, UniformFlow
                        , NonUniformTrue, NonUniformFlow
                        , UniformOutput, NonUniformOutput
                        ]

-- |
-- Creates an input description for generating the loop nest that
-- enumerates all tasks for the task graph that are not soak or drain tasks.
createAddTaskInput :: CompSpreadDS -> String
createAddTaskInput csDS = 
    verbose "createAddTaskInput" $
    cloogInput
    where
    blobNames = csBlobNames csDS
    minBlobSize = Just (csMinBlobSize csDS)
    scatterFcts = Just [(identityScatterFct (nVars $ head addTaskStmt) 
                         (nParams $ head addTaskStmt))]
    addTaskStmt = addTaskIneqs csDS
    cloogInput = 
        genGenericCloogInput [addTaskStmt] (outerIteratorNames csDS) (outerScatterNames csDS) 
                             blobNames 
                             minBlobSize scatterFcts
    
-- |
-- Creates separate input descriptions for every dependence in the
-- form of a CLooG input format. Note: it's reasonable to 
-- process every dependence seperately, because they just
-- are used for describing dependences between tasks, not for
-- defining any ordering of elements.
-- Extension: now we also add the target dependence for each
--  cloog input description to keep track of the dependence type,
--  which enables optimizations on the task graph.
createDependenceInput :: CompSpreadDS 
                      -> HOC_Type
                      -> [(TrDependence,String)]
createDependenceInput csDS hocType = cloogInputs
    where
    blobNames = csBlobNames csDS
    minBlobSize = Just (csMinBlobSize csDS)
    scatterFcts = Just [(identityScatterFct (nVars $ snd $ head dependenceStmt) 
                         (nParams $ snd $ head dependenceStmt))]
    dependenceStmt = dependenceIneqs csDS hocType                        

    iterators = outerIteratorNames csDS ++ innerIteratorNames csDS
    scatterNames = outerScatterNames csDS ++ innerScatterNames csDS

    cloogInputs = 
        map (\(trDep,domain) ->
               (trDep, 
                genGenericCloogInput [[domain]] iterators scatterNames
                                     blobNames 
                                     minBlobSize scatterFcts)
               )
        dependenceStmt

outerTileDimInput :: CompSpreadDS -> String
outerTileDimInput csDS = cloogInput
    where
    blobNames = csBlobNames csDS
    minBlobSize = Just (csMinBlobSize csDS)
    scatterFcts = Just (map (\dom -> identityScatterFct (nVars $ head dom) 
                                                        (nParams $ head dom)) 
                        allStmts)
    allStmts = outerTileDimIneqs csDS
    cloogInput = 
        genGenericCloogInput allStmts (outerIteratorNames csDS) (outerScatterNames csDS) 
                             blobNames 
                             minBlobSize scatterFcts

-- |
-- Creates a list of inequation systems describing the tasks IDs.
-- This is done by projecting on the outer tile dimensions.
-- Contains inequation systems for all statements except soak and drain 
-- statements.
addTaskIneqs :: CompSpreadDS -> [IneqSystem]
addTaskIneqs csDS = 
    verbose "addTaskIneqs" $
    ineqs
    where
    (soakIDs, drainIDs) = getSoakDrainOrigIDs csDS
    soakDrainStmtIDs = soakIDs ++ drainIDs

    (_,_,ineqs) = unzip3 $ filter (\(trStmtID, stmtID, compIS) -> 
                            not $ stmtID `elem` soakDrainStmtIDs
                           ) (computeStmtsTileDims csDS)

outerTileDimIneqs :: CompSpreadDS -> [[IneqSystem]]
outerTileDimIneqs csDS = [computes, sends, recvs]
    where
    trDeps = csTrDeps csDS
    filteredTrDeps = 
        getCommunicationTrDeps trDeps
    lattice = csLatVects csDS
    tiledDims = csTiledDims csDS
    
    sends = 
        trace "sends..." $
        map (\trd -> fst ( -- trace ("trDep#" ++ (show $ integerFromID $ trDepID trd)) $
                          defaultTile lattice tiledDims 
                                          (tightenedDep $ trDepRelation trd)  [SrcTile] )) 
            filteredTrDeps
    recvs =
        trace "recvs..." $
        map (\trd -> fst ( defaultTile lattice tiledDims 
                           (tightenedDep $ trDepRelation trd)  [DestTile] )) filteredTrDeps

    (_,_,computes) = 
        trace "computes..." $
        (unzip3 $ computeStmtsTileDims csDS)

createCompTaskInput :: CompSpreadDS -> String
createCompTaskInput csDS = 
    verbose "createCompTaskInput" $
    computeStmtInput
    where
    (soakIDs, drainIDs) = getSoakDrainOrigIDs csDS
    soakDrainStmtIDs = soakIDs ++ drainIDs

    blobNames = outerScatterNames csDS ++ csBlobNames csDS
    minBlobSize = Nothing -- no context restriction (given by outer loops)

    innerComputes = filter (\(trStmtID, stmtID, compIS) -> 
                            not $ stmtID `elem` soakDrainStmtIDs
                           ) (computeStmtsInnerDims csDS)
    computeStmtInnerScatterFcts = 
         Just (map (\(trStmtID, stmtID, compIS) -> 
                    getComputeStmtInnerScatterFct csDS trStmtID) innerComputes)
    computeStmtInput = 
        genGenericCloogInput (map (\(_,_,s) -> [s]) innerComputes)
                             (innerComputeIteratorNames csDS) (innerComputeScatterNames csDS)
                             blobNames
                             minBlobSize
                             computeStmtInnerScatterFcts

innerTileDimInputs :: CompSpreadDS -> [String]
innerTileDimInputs csDS = 
    if null failedAssertions
    then (computeStmtInput:sendRecvInputs)
    else trace ("\nin innerTileDimInputs: " ++ failedAssertions ++ "\n")
         (computeStmtInput:sendRecvInputs)
    where
    failedAssertions = 
        if and equivalentIneqs
        then ""
        else ("warning: ineqSystems of sender and receiver side not equivalent! " ++ 
              (show equivalentIneqs))

    equivalentIneqs = 
        zipWith (\send -> \recv -> areIneqsEquivalent send recv Nothing) sendStmts recvStmts

    blobNames = outerScatterNames csDS ++ csBlobNames csDS
    minBlobSize = Nothing -- no context restriction (given by outer loops)

    computeStmtInnerScatterFcts = 
         Just (map (\(trStmtID, _, compIS) -> 
                    getComputeStmtInnerScatterFct csDS trStmtID) innerComputes)
    computeStmtInput = 
        genGenericCloogInput (map (\(_,_,s) -> [s]) innerComputes)
                             (innerComputeIteratorNames csDS) (innerComputeScatterNames csDS)
                             blobNames
                             minBlobSize
                             computeStmtInnerScatterFcts
         
    sendRecvScatterFcts = 
--        (map (\stmts -> Just (map (\is -> identityScatterFct (nVars is) (nParams is)) stmts))
--         sendRecvStmts)
        (map (\stmt -> Just [identityScatterFct (nVars $ head stmt) (nParams $ head stmt)])
             sendRecvStmts)
    innerComputes = computeStmtsInnerDims csDS
    (sendStmts, recvStmts) = (head sendRecvStmts, head $ drop 1 sendRecvStmts)
    sendRecvStmts = [sendRecvStmtsInnerTileDims csDS True, sendRecvStmtsInnerTileDims csDS False]
    sendRecvInputs =
        zipWith (\stmts -> \scatterFcts ->
--                 genGenericCloogInput (map (\s -> [s]) stmts) -- jeweils nur 1 ineqSystem pro stmt
                 genGenericCloogInput [stmts]
                 (innerIteratorNames csDS) (innerScatterNames csDS) blobNames
                 minBlobSize scatterFcts) sendRecvStmts sendRecvScatterFcts


-- |
-- Creates seperate input descriptions for every communication
-- between a sender- and receiver-tile according to a
-- dependence. The result is given in the form of a CLooG input format. 
-- Note: it's reasonable to process every dependence seperately, because they just
-- are used for describing dependences between tasks, not for
-- defining any ordering of elements.
-- The different types of communication for soak-, drain-, and p2p-data are given
-- by the @Comm_Type@ argument.
-- 
genP2PCommInputs :: CompSpreadDS 
                 -> Comm_Type
                 -> [(TrDependenceID, String)]
genP2PCommInputs csDS commType = p2pCommInputPerDep
    where
    blobNames = case commType of
                Comm_Soak  -> (outerScatterNames csDS) ++ csBlobNames csDS
                Comm_Drain -> (outerScatterNames csDS) ++ csBlobNames csDS
                Comm_P2P   -> (outerScatterNames csDS) ++ (innerScatterNames csDS) ++ csBlobNames csDS
    minBlobSize = Nothing -- no context restriction (given by outer loops)
    iteratorNames = bufferIteratorNames csDS
    scatterNames = bufferScatterNames csDS

    (soakIDs, drainIDs) = getSoakDrainOrigIDs csDS
    soakDrainStmtIDs = soakIDs ++ drainIDs
                                                     
    writeBufferStmts = bufferStmtsInnerCoordDims csDS True (Just commType)-- send-recv
    p2pCommInputPerDep = map (\(trDepID,s) -> 
                                (trDepID,
                                 genGenericCloogInput [[s]]
                                    iteratorNames scatterNames blobNames
                                    minBlobSize (Just [writeBufferScatterFct csDS (Just commType) trDepID]))
                              ) writeBufferStmts

bufferDimInputs :: CompSpreadDS -> [String]
bufferDimInputs csDS = [writeBufferInputs, unpackBufferInputs]
    where
    blobNames = (outerScatterNames csDS) ++ (innerScatterNames csDS) ++ csBlobNames csDS
    minBlobSize = Nothing -- no context restriction (given by outer loops)
    writeBufferScatterFcts = 
        Just (map (\(trDepID, is) -> 
              writeBufferScatterFct csDS Nothing trDepID) writeBufferStmts)
    unpackBufferScatterFcts = 
        Just (map (\(trDepID, is) -> 
             unpackBufferScatterFct csDS Nothing trDepID) unpackBufferStmts)

    writeBufferStmts = bufferStmtsInnerCoordDims csDS True Nothing
    unpackBufferStmts = bufferStmtsInnerCoordDims csDS False Nothing

    allStmtTypes = [writeBufferStmts, unpackBufferStmts]
    allScatterFcts = [writeBufferScatterFcts, unpackBufferScatterFcts]

    writeBufferInputs =
        genGenericCloogInput (map (\(_,s) -> [s]) writeBufferStmts) 
                             (bufferIteratorNames csDS)
                             (bufferScatterNames csDS)
                             blobNames
                             minBlobSize
                             writeBufferScatterFcts

    unpackBufferInputs =
        genGenericCloogInput (map (\(_,s) -> [s]) unpackBufferStmts) 
                             (bufferIteratorNames csDS)
                             (bufferScatterNames csDS)
                             blobNames
                             minBlobSize
                             unpackBufferScatterFcts

-- creates a list of inequation systems for each statement
computeStmtsTileDims :: CompSpreadDS -> [(TStatementID, StatementID, IneqSystem)]
computeStmtsTileDims csDS = 
    -- trace (showIS2VisualPolylib firstCompIS (nVars firstCompIS)) $
    verbose "computeStmtsTileDims" $
    -- trace ("compStmtTileISs: " ++ show compStmtTileISs) $
    compStmtTileISs
    where
    (_,_,[firstCompIS]) = unzip3 compStmtTileISs
    compStmtTileISs = map tileTrStmt tispcs
    lattice = csLatVects csDS
    tiledDims = csTiledDims csDS
    infos = [SrcTile]
    tispcs = csTISPCs csDS

    tileTrStmt :: TStatement -> (TStatementID, StatementID, IneqSystem)
    tileTrStmt (trStmtID, stmtID, is) =
--         trace (show (selfDepRelation (stmtID, is))) $
--         trace ("tiled dims: " ++ show tiledDims) $
--         verbose ("tileTrStmt" ++ show stmtID) $
--         trace ("is: " ++ show is) $
--         trace ("lattice: " ++ show lattice) $
        (trStmtID, stmtID,
         fst $ verbose ("tiling for tile dims for compStmt #" ++ show trStmtID) $ 
             defaultTile lattice tiledDims (selfDepRelation (stmtID, is)) 
                         infos)


-- generates a system containing: ISPC | TISPC | Tile
computeStmtsInnerDims :: CompSpreadDS -> [(TStatementID, StatementID, IneqSystem)]
computeStmtsInnerDims csDS = 
--    trace (showIS2VisualPolylib compIS (nVars compIS)) $
--    trace (show compIS) $
    -- trace (show tiledDims) $
    verbose "computeStmtsInnerDims" $
    compISs
    where
    (_, _, compIS) = head compISs
    compISs = 
        -- verbose "compISs" $
        -- trace ("computing " ++ show (genericLength tispcs) ++ " compISs") $
        map tileTrStmt tispcs

    lattice = csLatVects csDS
    tiledDims = csTiledDims csDS
    infos = [SrcCoord,SrcTile]
    tispcs = csTISPCs csDS
    paramsDim = getParamsDim csDS
    tileDim = genericLength tiledDims

    tileTrStmt :: TStatement -> (TStatementID, StatementID, IneqSystem)
    tileTrStmt (trStmtID, stmtID, is) =    
        verbose ("tileTrStmt" ++ show trStmtID) $
--         trace ("tileTrStmt #" ++ show trStmtID ++ 
--                ", (isDim, coordsDim, tileDim, paramsDim, nColumns comp_M) = " ++ 
--                show (isDim, coordsDim, tileDim, paramsDim, nColumns comp_M)) $
        trace (show compIS) $
        compIS
        where
        compIS =
            verbose "compIS" $
            (trStmtID, stmtID, 
             ineqSystem ((isDim + coordsDim, tileDim + paramsDim), comp_M))

        comp_M =
            verbose "comp_M" $                  
            myMultijoin1 $
            [ spaceTimeTrafo_M
            , ispc_M
            , coordsTile_M
            , trc_M
            ]
        
        isDim = getISPCdim csDS trStmtID
        coordsDim = getCoordsDim csDS trStmtID
        paramsDim = getParamsDim csDS
        (is_v, is_p, is_c) = partsVPC $ getISPC csDS trStmtID
        (trafo_v, trafo_p, trafo_c) = partsVPCih trafoFct
        (trc_v, trc_p, trc_c) = partsVPC $ getTRCis csDS trStmtID
        trafoFct = getTrafoFct csDS trStmtID
        emptyIS = 
	      ineqSystem ((coordsDim, paramsDim), 
		          zeroMatrix 0 (coordsDim + paramsDim + 1))

        trafoIneq = 
            -- verbose "trafoIneq" $                  
            trafo_v `conc` ((-(unitMatrix (nRows trafo_v) coordsDim)) `conc` 
                            zeroMatrix (nRows trafo_v) (tileDim ) `conc` trafo_p `conc` 
                            trafo_c)
        spaceTimeTrafo_M = 
            -- verbose "spaceTimeTrafo_M" $                  
            trafoIneq `join` (-trafoIneq)
        ispc_M = 
            -- verbose "ispc_M" $                  
            is_v `conc` zeroMatrix (nRows is_v) (coordsDim + tileDim) `conc`
            is_p `conc` is_c
        coordsTileIS =
            verbose ("coordsTileIS " ++ show trStmtID) $ 
            trace ("tiling for inner dims for compStmt tr#" ++ show trStmtID ++
                   " stmt#" ++ show stmtID) $
            fst $ defaultTile lattice tiledDims (selfDepRelation (stmtID, emptyIS)) infos
        coordsTile_M = 
      --      verbose "coordsTile_M" $                  
            -- trace (show coordsTileIS) $
            zeroMatrix (nRows coordsTileIS) isDim `conc` (getMatrix coordsTileIS)
        trc_M = 
            -- verbose "trc_M" $                              
            trc_v `conc` zeroMatrix (nRows trc_v) (coordsDim + tileDim) `conc`
            trc_p `conc` trc_c


dependenceIneqs :: CompSpreadDS 
                -> HOC_Type 
                -> [(TrDependence,IneqSystem)]
dependenceIneqs csDS hocType =
    verbose "dependenceIneqs" $
    map tileTrDep filteredTrDeps
    where
    lattice = csLatVects csDS
    tiledDims = csTiledDims csDS
    infos = [SrcTile, DestTile] 

    (soakIDs, drainIDs) = getSoakDrainOrigIDs csDS

    paramsDim = getParamsDim csDS
    tileDim = genericLength tiledDims
    trDeps = csTrDeps csDS
    filteredTrDeps = 
        if (hocType == MasterWorkerHOC)
        then getCommunicationTrDeps trDeps
        else -- use all dependence types for p2p-dependence-graph, 
             -- except dependences from/to soak- or drain-statements
             filter (isDepType csDS Comm_P2P) trDeps
    
    tileTrDep :: TrDependence -> (TrDependence,IneqSystem)
    tileTrDep trDep = 
        verbose ("tileTrDep" ++ show trDep) $
        (trDep,
         ineqSystem ((tileDim + tileDim, paramsDim),
                     getMatrix $ fst (defaultTile lattice tiledDims 
                                      (tightenedDep $ trDepRelation trDep) infos)
                    )
        )


sendRecvStmtsInnerTileDims :: CompSpreadDS -> Bool -> [IneqSystem]
sendRecvStmtsInnerTileDims csDS isSender =
    trace ((if isSender then "send" else "recv") ++ "StmtsInnerTileDims...") $
    map tileTrDep filteredTrDeps
    where
    lattice = csLatVects csDS
    tiledDims = csTiledDims csDS
    infos = if isSender 
            then [SrcTile, DestTile] 
            else [DestTile, SrcTile]
    paramsDim = getParamsDim csDS
    tileDim = genericLength tiledDims
    trDeps = csTrDeps csDS
    filteredTrDeps = 
        getCommunicationTrDeps trDeps
    
    tileTrDep :: TrDependence -> IneqSystem
    tileTrDep trDep = 
        ineqSystem ((tileDim, paramsDim + tileDim),
                    getMatrix $ fst (trace ("trDep#" ++ (show $ integerFromID $ trDepID trDep)) $
                                     defaultTile lattice tiledDims 
                                     (tightenedDep $ trDepRelation trDep) infos)
                    )

{--
sendStmtsInnerTileDims :: CompSpreadDS -> [IneqSystem]
sendStmtsInnerTileDims csDS =
    trace "sendStmtsInnerTileDims..." $
    map tileTrDep filteredTrDeps
    where
    lattice = csLatVects csDS
    tiledDims = csTiledDims csDS
    infos = [SrcTile, DestTile]
    paramsDim = getParamsDim csDS
    tileDim = genericLength tiledDims
    trDeps = csTrDeps csDS
    filteredTrDeps = 
        getCommunicationTrDeps trDeps
    
    tileTrDep :: TrDependence -> IneqSystem
    tileTrDep trDep = 
        ineqSystem ((tileDim, paramsDim + tileDim),
                    getMatrix $ fst (trace ("trDep#" ++ (show $ integerFromID $ trDepID trDep)) $
                                     defaultTile lattice tiledDims (trDepRelation trDep) infos)
                    )

recvStmtsInnerTileDims :: CompSpreadDS -> [IneqSystem]
recvStmtsInnerTileDims csDS =
    trace "sendStmtsInnerDims..." $
    map tileTrDep filteredTrDeps
    where
    lattice = csLatVects csDS
    tiledDims = csTiledDims csDS
    infos = [DestTile, SrcTile]
    paramsDim = getParamsDim csDS
    tileDim = genericLength tiledDims
    trDeps = csTrDeps csDS
    filteredTrDeps = 
        getCommunicationTrDeps trDeps
    
    tileTrDep :: TrDependence -> IneqSystem
    tileTrDep trDep = 
        ineqSystem ((tileDim, paramsDim + tileDim),
                    getMatrix $ fst (trace ("trDep#" ++ (show $ integerFromID $ trDepID trDep)) $
                                     defaultTile lattice tiledDims (trDepRelation trDep) infos)
                    )
--}

bufferStmtsInnerCoordDims :: CompSpreadDS 
                          -> Bool 
                          -> Maybe Comm_Type
                          -> [(TrDependenceID, IneqSystem)]
bufferStmtsInnerCoordDims csDS isWrite maybeCommType =
    trace "bufferStmtsInnerCoordDims..." $
    trace (if isWrite 
           then "write: "
           else "unpack: ") $
--    trace (showIS2VisualPolylib (snd $ head ineqs) (nVars $ snd $ head ineqs)) $
    ineqs
    where
    ineqs = map tileTrDep filteredTrDeps

    lattice = csLatVects csDS
    tiledDims = csTiledDims csDS
    infos = 
        if isNothing maybeCommType
        then if isWrite 
             then [SrcCoord, SrcTile, DestTile]
             else [SrcCoord, DestTile, SrcTile]
        else -- in case we are working in the P2P framework:
             case commType of
                  Comm_Soak  -> [SrcCoord, DestTile]
                  Comm_Drain -> [SrcCoord, SrcTile] 
                  Comm_P2P   -> [SrcCoord, SrcTile, DestTile]
    paramsDim = getParamsDim csDS
    tileDim = genericLength tiledDims

    commType = fromJust maybeCommType
    srcTileDim = if (maybeCommType == (Just Comm_Soak))
                 then 0
                 else tileDim
    destTileDim = if (maybeCommType == (Just Comm_Drain))
                  then 0
                  else tileDim
    trDeps = csTrDeps csDS
    trueFlowDeps = filterTrDeps trDeps [ UniformTrue, UniformFlow
                                       , NonUniformTrue, NonUniformFlow
                                       ]

    filteredTrDeps = 
        if isNothing maybeCommType
        then getCommunicationTrDeps trDeps
        else filter (isDepType csDS commType) trueFlowDeps
    
    tileTrDep :: TrDependence -> (TrDependenceID, IneqSystem)
    tileTrDep trDep = 
        verbose ("tileTrDep" ++ show (trDepID trDep)) $
        -- ISPC | TISPC | TILE | TILE | PARAMS | CONST
        (trDepID trDep,
         ineqSystem ((isDim + coordsDim, srcTileDim + destTileDim + paramsDim),
                     is)
        )
        where
        is = myMultijoin1 [ spaceTimeTrafo_M
                          , ispc_M
                          , (zeroMatrix (nRows coordsTile_M) isDim `conc`
                                        coordsTile_M)
                          , trc_M
                          ]

        isDim = getISPCdim csDS trStmtID
        coordsDim = getCoordsDim csDS trStmtID
        paramsDim = getParamsDim csDS
        trStmtID = trDepSrcStmtID trDep
        (is_v, is_p, is_c) = partsVPC $ getISPC csDS trStmtID
        (trafo_v, trafo_p, trafo_c) = partsVPCih trafoFct
        (trc_v, trc_p, trc_c) = partsVPC $ getTRCis csDS trStmtID
        trafoFct = 
            verbose "trafoFct" $
            getTrafoFct csDS trStmtID
        trafoIneq = 
            verbose "trafoIneq" $
            (trafo_v `conc` ((-(unitMatrix coordsDim coordsDim)) `conc` 
             zeroMatrix coordsDim (srcTileDim + destTileDim) `conc` trafo_p `conc` trafo_c))
        ispc_M = 
            verbose "ispc_M" $
            is_v `conc` zeroMatrix (nRows is_v) (coordsDim + srcTileDim + destTileDim) `conc`
            is_p `conc` is_c
        spaceTimeTrafo_M = 
            trafoIneq `join` (-trafoIneq)
        coordsTile_M = getMatrix $ fst (
                                        trace ("tileTrDep" ++ " trDep#" ++ (show $ integerFromID $ trDepID trDep)) $
                                        defaultTile lattice tiledDims 
                                        (tightenedDep $ trDepRelation trDep) infos)
        trc_M = 
            verbose "trc_M" $
            trc_v `conc` zeroMatrix (nRows trc_v) (coordsDim + srcTileDim + destTileDim) `conc`
            trc_p `conc` trc_c



ehrhartAccessZPs :: CompSpreadDS 
                   -> [Access]
                   -- ^ all (trimmed!) accesses
                   -> Buffer_Type
                   -> [(ArrayID, ZDomain)]
ehrhartAccessZPs csDS accesses bufferType = 
    verbose "ehrhartAccessZPs" $
    -- arrayIDsAndZPs
    arrayID2ConvexZPsMap
    where

 -- Optimierungsidee: auf jedes ZPolytop realImage (gibts schon), 
 --   dann convex union, dann ineq2ZP (nimmt Identitaet als Zielfunktion)
 -- Oder: PRO AccessID: convex union auf alle ineqSytem --> gleiche 
 --                     access function --> 1 Z-Polytop

    arrayIDsAndZPs = [ (aID, zps) | 
                         aID <- allArrayIDs,
                         let zps = 
                                 map snd $ 
                                     filter (\(a,i) -> a == aID) separateArrayIDsAndZPs
                       ]
                         
    allArrayIDs = sort $ nub $ map accArrayID accesses

    -- wahrscheinlich unnoetig:
    acc2IneqsMap = [(acc, convexUnionAll ineqs) | 
                   acc <- accesses,
                   let ineqs = [ineq |
                                (trStmtID, stmtID, ineq) <- executionIneqs,
                                accStmtID acc == stmtID
                               ]
                  ]
                      
    arrayID2ConvexZPsMap = [(arrID, [zps]) |
                      arrID <- allArrayIDs,
                      let ineq =
                           (fromJust $
                            trace "===\nUSING CONVEX UNION ALL!\n===" $
                            (convexUnionAll $
                             map (zpRealImage.snd) $ 
                             filter (\(a,i) -> a == arrID) separateArrayIDsAndZPs)
                           ),
                      let zps =
                           trace ("nRows ineq nach convex union all: " ++ 
                                  (show $ nRows ineq)) $
                           trace ("nColumns ineq nach convex union all: " ++ 
                                  (show $ nColumns ineq)) $
                           zpFromIneqs ineq
                     ]                              
                      

    separateArrayIDsAndZPs = 
        -- verbose "separateArrayIDsAndZPs" $
        case bufferType of
--           CommunicationBuffer -> 
--               [(arrayID, ineq) | acc <- accesses,
--                                  trDepID <- srcAccID2TrDepIDs csDS (accID acc),
--                                  (trDepID, ineq) <- communicationIneqs,
--                                  let arrayID = accArrayID acc
--               ]
          ExecutionBuffer -> 
              verbose "executionBuffer" $
              [(arrayID, zpolys) | acc <- accesses,
                                 (trStmtID, stmtID, ineq) <- executionIneqs,
                                 accStmtID acc == stmtID,
                                 let arrayID = accArrayID acc,
                                 let zpolys = 
                                         verbose ("zpolys: array#" ++ show arrayID ++ " accID#" ++ show (accStmtID acc)) $
                                         ZP ineq (getAccFct acc trStmtID)
              ]
--           SoakDrainBuffer -> 
--               error ("ehrhartAccessIndices not implemented yet" ++
--                      " for SoakDrainBuffer!")

-- ...-> [(TrDependenceID, IneqSystem)]
    communicationIneqs = bufferStmtsInnerCoordDims csDS True (Just Comm_P2P)

-- CompSpreadDS -> [(TStatementID, StatementID, IneqSystem)]
    executionIneqs = 
        verbose "executionIneqs" $
        computeStmtsInnerDims csDS 

    acps = csACPs csDS

    getAccFct :: Access -> TStatementID -> GAffineFunction Integer
    getAccFct acc tStmtID = 
        if (assertions == "")
        then affineFunction ((isD+coordsD, tileD+paramsD), acc_M) Inhomogeneous
        else error ("in ehrhartAccessIneqs: something doesn't match: " ++ assertions)
        where
        accFct = accIndexFct acc
        (acc_v, acc_p, acc_c) = partsVPCih (accIndexFct acc)
        accD = targetDimIH (accIndexFct acc)
        isD = getISPCdim csDS tStmtID
        coordsD = getCoordsDim csDS tStmtID
        tileD = getTilingDim csDS
        paramsD = getParamsDim csDS
        acc_M = 
            integralMatrix $ -- ist das korrekt? leider nicht: z.B. fuer A[i/2]
                             -- damit kann LooPo einfach nicht umgehen...
            (acc_v `conc` zeroMatrix accD (coordsD+tileD) `conc`
             acc_p `conc` acc_c)
        assertions = 
            concat $
            [ if accD == nRows acc_v then "" else "accD (" ++ show accD ++ ") != nRows acc_v (" ++ show (nRows acc_v) ++ ")\n"
            , if accD == nRows acc_p then "" else "accD (" ++ show accD ++ ") != nRows acc_p (" ++ show (nRows acc_p) ++ ")\n"
            , if accD == nRows acc_c then "" else "accD (" ++ show accD ++ ") != nRows acc_c (" ++ show (nRows acc_c) ++ ")\n"
            , if nVars accFct == isD then "" else "nVars affFct != isD\n"
            ]
        

-- Fragen: alle Accesses verwenden? oder nur die in AccessPairs?
--         verwendet man SrcAccesses oder DestAccesses? Read/Write?
-- Antwort: - wegen soak/draining: alle wichtigen accesses gehoeren zu
--            Abhaengigkeiten
--          - nur true/flow benoetigt: damit readAcc -> dest, writeAcc -> src

--   AccessID -> [DependenceID]
--   DependenceID -> [TrDependenceID]
--   TrDependenceID -> [ineq]

-- man muss ja alle ungleichungssysteme als union aufzhlen, damit man das 
-- ungleichungssystem kriegt, auf dem man ehrhart anwenden kann
-- --> alle werte <=lex dem aktuellen gibt dann die position im puffer an
-- d.h. alle ungleichungssysteme fr alle dependences vereinigen, und zwar:
-- PRO ARRAY: d.h. fr jedes array (A, B, C...) eine eigene union von ineqs...
-- ABER: man setzt (zur Laufzeit) die TileNr auf die aktuelle taskID 
-- (noch nicht sicher, wie?)
-- also: fuer jedes array: fuer eine feste tileNr: zaehle alle punkte im polytop
--                                                 <=lex aktuelle arrayAccess

    



identityScatterFct :: Integer -> Integer -> Matrix
identityScatterFct isDim params =
    ((unitMatrix isDim isDim) `conc` 
     (-(unitMatrix isDim isDim)) `conc` 
     (zeroMatrix isDim (params + consts))
    )
    where 
    consts = 1

getComputeStmtInnerScatterFct :: CompSpreadDS -> TStatementID -> Matrix
getComputeStmtInnerScatterFct csDS trStmtID = scatterMatrix
    where
    -- coords || ispc | coords| tile | params | const
    scatterMatrix = 
        myMultijoin1 [ coordsM, addM ]

    coordsM =
        unitMatrix coordsD coordsD `conc`
        zeroMatrix coordsD addD `conc`
        zeroMatrix coordsD isDim `conc`
        (-(unitMatrix coordsD coordsD)) `conc`
        zeroMatrix coordsD (tileD + paramsD + 1)

    addM = 
        zeroMatrix addD coordsD `conc`
        unitMatrix addD addD `conc`
        zeroMatrix addD (isDim + coordsD + tileD + paramsD + 1)

    isDim = getISPCdim csDS trStmtID
    maxCoordsD = getMaxCoordsDim csDS
    coordsD = getCoordsDim csDS trStmtID
    addD = maxCoordsD - coordsD
    tileD = genericLength $ csTiledDims csDS
    paramsD = getParamsDim csDS

writeBufferScatterFct :: CompSpreadDS 
                      -> Maybe Comm_Type
                      -> TrDependenceID 
                      -> Matrix
writeBufferScatterFct csDS maybeCommType trDepID = 
    trace "write/sendBufferScatterFct:" $
    scatter_M
    where
    -- accessFct: geht von ISPC -> accDim

    -- ArrAcc(max) || ISPC | TISPC | TILE | TILE | PARAMS | CONST
 
    srcTStmtID = trDepSrcStmtID (getTrDep csDS trDepID)

    tileD = genericLength $ csTiledDims csDS
    
    srcTileD = if (maybeCommType == (Just Comm_Soak))
               then 0
               else tileD
    destTileD = if (maybeCommType == (Just Comm_Drain))
                then 0
                else tileD

    isD = getISPCdim csDS srcTStmtID
    coordsD = getMaxCoordsDim csDS
    srcCoordsD = getCoordsDim csDS srcTStmtID   
    addD = getMaxCoordsDim csDS - getCoordsDim csDS srcTStmtID
    pD = getParamsDim csDS
    accMaxD = getArrAccessMaxDim csDS
    accD = targetDimIH accFct
    accAddD = accMaxD - accD
    accFct = accIndexFct srcAcc
    srcAcc = getSrcAcc csDS  trDepID
    -- arrayNr = integerFromID $ accArrayID srcAcc
    fullAccFct =
        -- trace (show $ getMatrixIH accFct) $        
        verbose "fullAccFct" $
        affineFunction ((isD, pD),
                        -- trace (show $ nColumns (getMatrixIH accFct)) $
                        -- trace "bla" $
                        myMultijoin1 [ getMatrixIH accFct
                        -- trace (show $ isD + pD + 1) $
                                     , zeroMatrix accAddD (isD + pD + 1)
                                     ]
                       ) Inhomogeneous

    (acc_v, acc_p, acc_c) = partsVPCih fullAccFct

    scatter_M = unitMatrix accMaxD accMaxD `conc` 
                 (-acc_v) `conc`
                 zeroMatrix accMaxD (srcCoordsD + srcTileD + destTileD) `conc`
                 (-acc_p) `conc` (-acc_c)
                


unpackBufferScatterFct :: CompSpreadDS -> Maybe Comm_Type -> TrDependenceID -> Matrix
unpackBufferScatterFct = writeBufferScatterFct

compStmtIndexMapping :: CompSpreadDS -> TStatementID -> [String] -> String
compStmtIndexMapping csDS trStmtID iteratorNames = 
    trace (show cloogIndices) $
    concat $
    zipWith (\l c -> l ++ " = " ++ c ++ "; ") 
	    loopoIndices cloogIndices
    where
    origStmtID = getStmtID csDS trStmtID
    tilingDim = getTilingDim csDS
    ispcDim = getISPCdim csDS trStmtID
    (SLP slps) = csSLPs csDS
    slpVector = (fromJust $ lookup origStmtID slps)::[Bool]
    slp = (elemIndices True slpVector) :: [Int]
    loopoIndices = 
	[ genericIndex (csIndexNames csDS) lp | lp <- slp ]
    cloogIndices = 
	genericTake ispcDim iteratorNames

getSoakDrainOrigIDs :: CompSpreadDS 
                    -> ([StatementID], [StatementID])
getSoakDrainOrigIDs csDS = soakDrainStmtIDs
    where
    allOrigStmtIDs = sort $ getAllStatementIDs csDS
    soakDrainStmtIDs = 
        (genericTake (csNumSoakStmts csDS) allOrigStmtIDs,
         genericTake (csNumDrainStmts csDS) (reverse allOrigStmtIDs))

isDepType :: CompSpreadDS 
          -> Comm_Type
          -> TrDependence
          -> Bool
isDepType csDS commType dep | commType == Comm_Soak = trDepOrigSrcStmtID dep `elem` soakIDs
                            | commType == Comm_Drain = trDepOrigDestStmtID dep `elem` drainIDs
                            | otherwise = not ((trDepOrigSrcStmtID dep `elem` soakIDs) ||
                                               (trDepOrigDestStmtID dep `elem` drainIDs))
    where
    (soakIDs, drainIDs) = getSoakDrainOrigIDs csDS
    soakDrainStmtIDs = soakIDs ++ drainIDs

tightenedDep :: (IneqSystem,Integer) -> (IneqSystem,Integer)
tightenedDep (is, x) = (tightenIneqSystem is, x)
                