-- |
-- Generates master and task file for the master/worker framework.
-- The classical master/worker approach is supported as well
-- as the peer-to-peer approach, which adds communication
-- between the workers in order to avoid that the master
-- becomes the bottleneck.
--
-- Authors: Philipp Classen, Michael Classen
--
module TaskLoops
    (-- * generate java files (master-worker framework)
     genMasterJava, genTaskJava, genDataJava,

     -- * generate java files (p2p-framework)
     genP2PMasterJava, genP2PClientJava,
     genP2PDataPacketJava, genP2PDynamicArrayClassJava,

     -- * debug functions
     masterFileTemplate, taskFileTemplate, dataFileTemplate,
     p2pMasterTemplate, p2pClientTemplate,
     p2pDataPacketTemplate, p2pDynamicArrayTemplate
    )
where

import LoopoDS
import Debug.Trace(trace)
import System.Cmd(system)
import System(getArgs,exitWith, exitFailure, ExitCode(ExitSuccess))
import System.Console.GetOpt
import System.Time
import System.IO.Unsafe(unsafePerformIO)
import Monad(when)
import Tiling
import Maybe
import Ratio
import List
import Text.Regex.Posix

import LMath.Types.Vector
import LMath.Types.AffineExp
import LMath.Types.LClasses
import LMath.Types.AffineFunction
import LMath.Types.IneqSystem
import LMath.Types.Matrix
import LMath.Types.Ehrhart
import LMath.Algo.AffineImage
import LMath.Algo.ZPolyhedron

import CompSpread.Data
import CompSpread.Run 
import CompSpread.DomainDescriptions
import CompSpread.Helpers
import CompSpread.Defines

import ScannerParser.PrintCode
import ScannerParser.ScannerParser(openParseFile)

import StagedCodeGenInput
import TaskData (taskDataLoop)

data LoopData = LoopData { ldCompTaskCode :: String
                         -- ^ computation code for a single task

                         , ldArrayReadAccesses_StoreCode :: String
                         -- ^ code that stores the array elements that
                         --   will be accessed by the task, into the buffer
                         , ldArrayReadAccesses_LoadCode :: String
                         -- ^ code that updates the memory by loading
                         --   all array elements that will be accessed
                         --   by the task, from the buffer
                         , ldArrayWriteAccesses_StoreCode :: String
                         -- ^ code that stores the modified array elements
                         --   into the buffer
                         , ldArrayWriteAccesses_LoadCode :: String
                         -- ^ code that updates the memory by writing the
                         --   modified array elements from the buffer into the
                         --   global\/local memory

                         , ldP2P_Communication_writeBuffer :: String
                         -- ^ code executed to prepare the data packet
                         --   that should be sent from one peer to another
                         --   (Worker-Worker communication)
                         , ldP2P_Communication_unpackBuffer :: String
                         -- ^ code executed to write the received
                         --   peer-to-peer data into the local memory
                         --   (Worker-Worker communication)
                         , ldP2P_Soak_writeBuffer :: String
                         -- ^ code executed to write the soak data
                         --   into the buffer (Master-Worker communication)
                         , ldP2P_Soak_unpackBuffer :: String
                         -- ^ code executed to write the received soak data
                         --   into the local memory
                         --   (Master-Worker communication)
                         , ldP2P_Drain_writeBuffer :: String
                         -- ^ code executed to write the drain data
                         --   into the buffer (Worker-Master communication)
                         , ldP2P_Drain_unpackBuffer :: String
                         -- ^ code executed to write the received drain data
                         --   into the local memory
                         --   (Worker-Master communication)
                         , ldP2P_MasterSoak :: String
                         -- ^ code executed to write the soak data
                         --   into the buffer at the master
                         --   (Master-Worker communication)
                         , ldP2P_MasterDrain :: String
                         -- ^ code executed to write the received drain data
                         --   into the local memory at the master
                         --   (Worker-Master communication)

                         , ldAddAllTasksCode :: String
                         -- ^ code executed by the master to
                         --   add all tasks (one for each dependence)
                         , ldCreateDependenceGraphCode :: [String]
                         -- ^ code-pieces executed by the master to
                         --   add dependences between the tasks 
                         --   for each dependence

                         , ldInitDeclarations :: String
                         -- ^ the variable and array declarations
                         --   provided by the user in the, which are
                         --   read from the ".init_decls" file.
                         , ldIndexVariables :: [String]
                         -- ^ names of the loop variables
                         , ldOuterTilingVariables :: [String]
                         -- ^ Names of the tiling variables,
                         --   which are the loop variables of the outer loops.
                         --   (This list is equivalent to the TaskID
                         --    or sourceTaskID.)
                         , ldInnerTilingVariables :: [String]
                         -- ^ Names of the tiling variables,
                         --   which are the loop variables of the inner loops.
                         --   (This list is equivalent to the targetTaskID.)
                         , ldTaskVariables :: [String]
                         -- ^ all integer variables that have to be declared
                         --   in the task class
                         , ldMasterVariables :: [String]
                         -- ^ all integer variables that have to be declared
                         --   in the master class
                         , ldCreateDependenceGraphVariables :: [String]
                         -- ^ all local integer variables that are
                         --   used inside the createDependenceGraph method
                         , ldExecuteTaskVariables :: [String]
                         -- ^ all local integer variables that are
                         --   used inside the execute task method
                         , ldWriteBufferVariables :: [String]
                         -- ^ all local integer variables that are
                         --   used inside the writeBuffer method
                         , ldUnpackBufferVariables :: [String]
                         -- ^ all local integer variables that are
                         --   used inside the unpackBuffer method
                         , ldParameters :: [String]
                         -- ^ names of the parameters (e.g. "n")
                         , ldArrayVariables :: [(String, [String],
                                                 ArrayAccessType)]
                         -- ^ all arrays that are used by the task
                         --   with their dimensions (from the .arrays file)
                         --   and their access type
                         --   (read-only, write-only, read-write)
                         , ldRawTilingVector :: [[Integer]]
                         -- ^ all tiling vectors

                         , ldPackageName :: String
                         -- ^ the name of the project (e.g. userapplication.matmult)
                         , ldAccesses :: [Access]
                         -- ^ all accesses in the source program, 
                         --   trimmed to surrounding loops
                         , ldPreloopCode :: String
                         -- ^ the original user code that is executed
                         --   before the parallel loop
                         , ldPostloopCode :: String
                         -- ^ the original user code that is executed
                         --   after the parallel loop

                         , ldSerialVersionUID :: Int
                         -- ^ can be an arbitrary integer
                         --   (e.g., the current time or a random number
                         --    is just fine).
                         , ldSupportOldFramework :: Bool
                         -- ^ if this flag is set, the generated code supports the
                         --   old framework (not the peer-to-peer branch)
                         , ldHOC_Type :: HOC_Type
                         -- ^ the type of HOC component for which the
                         --   code is generated
                         , ldBufferManagementStrategy :: BufferManagementStrategy
                         -- ^ determines whether the generated code will
                         --   use a dynamic buffer or a fixed-size
                         --   buffer, whose size is precomputed
                         , ldStorageManagementStrategy :: StorageManagementStrategy
                         -- ^ determines the stratege for how to store
                         --   the elements for computation on clients
                         , ldEhrhartAccessCode :: String
                         -- ^ the code for the functions involving ehrhart accesses
                         , ldEhrhartArrayAllocCode :: String
                         -- ^ the code to allocate an ehrhart buffer for each array
                         } deriving Show


-- |
-- The default HOC component (for which the code is generated).
--
defaultHOC_Type :: HOC_Type
defaultHOC_Type = MasterWorkerHOC

-- |
-- Type of array access:
--
--   1) ReadOnlyAccess == IN
--
--   2) WriteOnlyAccess == OUT
--
--   3) ReadWriteAccess == IN-OUT
--
data ArrayAccessType = ReadOnlyAccess | WriteOnlyAccess | ReadWriteAccess
                       deriving (Show, Eq)

-- |
-- Determines the supported buffer management strategies.
--
data BufferManagementStrategy = DynamicBuffer
                              | PrecomputedFixedArray
                                deriving (Show, Eq)

-- |
-- The default management strategy. Can be overridden by
-- the compile line switch '--buffer'.
--
defaultBufferManagementStrategy :: BufferManagementStrategy
defaultBufferManagementStrategy = DynamicBuffer


-- |
-- Determines the choice of code for either master or client.
--
data MasterClientCode = MasterCode | ClientCode
                        deriving (Show, Eq)

-- |
-- Determines the supported storage management strategies for
-- local data needed in computation.
--
data StorageManagementStrategy = LocalArray
                               | LocalDataStructure
                               | EhrhartBuffer
                                 deriving (Show, Eq)

-- |
-- The default storage management strategy for local data. 
-- Can be overridden by the compile line switch '--storage'.
--
defaultStorageManagementStrategy :: StorageManagementStrategy
defaultStorageManagementStrategy = LocalArray
                                   -- trace "warning: using EhrhartBuffer" $
                                   --EhrhartBuffer
                                   --trace "warning: using LocalDataStructure" $
                                   --LocalDataStructure

getTargetLanguage :: CompSpreadDS -> StorageManagementStrategy -> TargetLanguage
getTargetLanguage _ LocalArray = TL_PlainC
getTargetLanguage _ LocalDataStructure = TL_ModAccess
getTargetLanguage csDS EhrhartBuffer = TL_EhrhartAccess params
    where
    params = ([taskIDName ++ "[" ++ show i ++ "]" | 
               i <- [0..getTilingDim csDS-1]] ++ 
              csBlobNames csDS)


getArrayPrefix :: StorageManagementStrategy -> String
getArrayPrefix LocalArray = ""
getArrayPrefix LocalDataStructure = arrayStructurePrefix
getArrayPrefix EhrhartBuffer = "ehrhart"

--
-- Package and classnames of the Master and Task java file.
--
defaultPackageName = "userapplication.project"
masterClassname    = "LoopoAppMaster"
taskClassname      = "LoopoAppTask"
dataClassname      = "LoopoData"

p2pMasterClassname     = "LoopoAppMaster"
p2pClientClassname     = "LoopoAppClient"
p2pDataPacketClassname = "LoopoAppDataPacket"

--
-- Names of the helper variables that are used in the generated java code:
--
varPrefix        = "" -- (all helper variables will use this prefix)
saveArgsName     = varPrefix ++ "args_"
bufferName       = varPrefix ++ "buffer"
dynBufferName    = varPrefix ++ "dynBuffer"
dataPacketName   = varPrefix ++ "dataPacket"
precomputedBufferName = varPrefix ++ "precomputedBuffer"
counterName      = varPrefix ++ "pos"
workerIDName     = varPrefix ++ "workerID"
taskIDName       = varPrefix ++ "taskID"
sourceTaskIDName = varPrefix ++ "sourceTaskID"
targetTaskIDName = varPrefix ++ "targetTaskID"
resultMapName    = varPrefix ++ "map"
taskResultName   = varPrefix ++ "taskResult"
drainDataName    = varPrefix ++ "drainData"
soakDataName     = varPrefix ++ "soakData"
schedDepName     = varPrefix ++ "schedDep"
p2pSchedName     = varPrefix ++ "p2pScheduler"
graphName        = varPrefix ++ "graph"
graphBuilderName = varPrefix ++ "graphBuilder"
numTasksName     = varPrefix ++ "numTasks"
preloopFuncName  = varPrefix ++ "executeOriginalPreloopCode"
postloopFuncName = varPrefix ++ "executeOriginalPostloopCode"
jumpedOutOfOriginalMainFuncFlag =
    varPrefix ++ "jumpedOutOfOriginalMainFunc"

loadFuncName                  = varPrefix ++ "load"
addAllTasksFuncName           = varPrefix ++ "addAllTasks"
createDependenceGraphFuncName = varPrefix ++ "createDependenceGraph"
addDependenceHelperFuncName   = varPrefix ++ "addDependence__"

writeCommBufferFuncName       = varPrefix ++ "writeCommunicationBuffer"
unpackCommBufferFuncName      = varPrefix ++ "unpackCommunicationBuffer"

ehrhartAccessFuncName         = varPrefix ++ "ehrhartAccessFunction"
communicationBufferMapName    = varPrefix ++ "communicationBufferMap"
dependenceIndexName           = "utils.DependenceID"

dynArrayClassname = (if varPrefix == "" then "" else "Class_" ++ varPrefix) ++
                    "DynamicDoubleArray"

addTaskInpExt = "addTaskInp"
addTaskOutExt = "addTaskOut"

compTaskInpExt = "compTaskInp"
compTaskOutExt = "compTaskOut"

dependenceInpExt = "dependenceInp"
dependenceOutExt = "dependenceOut"

p2pCommInpExt = "p2pCommInp"
p2pCommOutExt = "p2pCommOut"

taskDefFile      = "taskDefinitions.loopo"
masterDefFile    = "masterDefinitions.loopo"

p2pClientDefFile = "p2pClientDefinitions.loopo"
p2pClientEhrhartDefFile = "p2pClientEhrhartDef.loopo"


-- |
-- This flags controls whether the generated java code
-- should include annotations to suppress all warnings
-- (only from the auto-generated code).
--
suppressAllWarnings :: Bool
suppressAllWarnings = True

-- |
-- Adds annotations to the generated java code
-- in order to suppress all warnings from the auto-generated code.
--
suppressWarningsCode :: [String]
suppressWarningsCode = ["@SuppressWarnings({ \"unused\", \"cast\" })"]

-- |
-- This flag controls whether the generated java code
-- should eliminate all soak communications by initializing
-- the UserClient with the arrays from the UserMaster.
--
-- (This is only relevant for the P2P-HOC.)
--
eliminateSoakCommunication :: Bool
eliminateSoakCommunication = -- True
                             False -- for Ehrhart buffers

-- |
-- Generates the java code of the master class.
--
genMasterJava :: LoopData
              -> String
genMasterJava ld =
    unlines
    (["#include \"" ++ masterDefFile ++ "\""
     ,genPackageDecl ld
     ,""
     ,"import java.util.Map;"
     ,"import userinterface.UserMaster;"
     ,"import userinterface.UserScheduler;"
     ,"import userinterface.UserTask;"
     ,"import schedulers.IfaceSchTaskDependence;"
     ,"import schedulers.SchTaskDependence;"
     ,""] ++ (if suppressAllWarnings then suppressWarningsCode else []) ++
     ["public class " ++ masterClassname ++
         " extends UserMaster implements IfaceSchTaskDependence {"
     ] ++ declarationCode ++
     [""] ++ masterVarDeclCode ++
     [""] ++ genStartUpCode ld ++
     [""] ++ genEndCode ld ++
     [""] ++ genGetResultsCode ld ++
     [""] ++ joinTaskResultCode ++
     [""] ++ createSchedulerCode ++
     [""] ++ genPreloopFuncCode ld ++
     [""] ++ genPostloopFuncCode ld ++
     [""] ++ getTaskCode ++
     [""] ++ genAddAllTasksCode ld ++
     [""] ++ genCreateDependenceGraphCode ld ++
     [""] ++ helperFunctionCode ++
     ["}"]
    )
    where
    declarationCode =
        let varDecl = ["  // index variable declarations:"] ++
                      ["  private int " ++ x ++ ";"
                       | x <- nub $ ldIndexVariables ld]
            paramDecl = ["  // user declarations:"] ++
                        [if matchVariableDeclaration x then "  static " ++ x else x
                          | x <- lines (ldInitDeclarations ld)] ++
                        [""]
        in varDecl ++ paramDecl

    masterVarDeclCode = 
                        ["  private String[] " ++ saveArgsName ++ ";"
                        ,""
                        ,"  // scheduler:"
                        ,"  private SchTaskDependence " ++ schedDepName ++ ";"
                        ,"  private int " ++ numTasksName ++ ";"
                        ,""
                        ,"  // helper flag to discover whether the user code"
                        ,"  // left its original main function"
                        ,"  private boolean " ++
                           jumpedOutOfOriginalMainFuncFlag ++ ";"
                        ]
    joinTaskResultCode = ["@Override"
                         ,"public long joinTaskResult(" ++
                          (if ldSupportOldFramework ld
                           then ""
                           else ("int " ++ workerIDName ++ ", ")) ++
                          "UserTask task) {"
                         ,"  // update global memory using the task's buffer:"
                         ,"  double[] " ++ bufferName ++ " =  ((" ++ taskClassname ++
                          ") task)." ++ bufferName ++ ";"
                         ,""
                         ,ldArrayWriteAccesses_LoadCode ld
                         ,"  " ++ schedDepName ++ ".joinTask(new int[] { "  ++ 
                               tileVars ++ " });"
                         ,"  " ++ numTasksName ++ "--;"
                         ,"  return " ++ numTasksName ++ ";"
                         ,"}"
                         ]
    tileVars = (concat $ 
                intersperse ", " ["tiledim" ++ show i | 
                                  i <- [0..(length (ldOuterTilingVariables ld) - 1)]
                                 ])

    createSchedulerCode = ["@Override"
                          ,"public UserScheduler createScheduler() {"
                          ,"  " ++ schedDepName ++
                           " = new SchTaskDependence(" ++
                           (if ldSupportOldFramework ld
                            then ""
                            else "this") ++
                           ");"
                          ,"  return " ++ schedDepName ++ ";"
                          ,"}"
                          ]
    getTaskCode = ["@Override"
                  ,"public UserTask getTask(Object index) {"
                  ,"  return new LoopoAppTask((int[]) index);"
                  ,"}"
                  ]

-- |
-- Generates the java code of the task class.
--
genTaskJava :: LoopData
            -> String
genTaskJava ld =
    unlines
    (["#include \"" ++ taskDefFile ++ "\""
     ,genPackageDecl ld
     ,""
     ,"import userinterface.UserTask;"
     ,""] ++ (if suppressAllWarnings then suppressWarningsCode else []) ++
     ["public class " ++ taskClassname ++ " extends UserTask {"
     ] ++ declarationCode ++
     [""
     ,genSerialVersionUIDDecl ld
     ] ++ taskVarDeclCode ++
     [""] ++ constructorCode ++
     [""] ++ loadCode ++
     [""] ++ executeCode ++
     [""] ++ helperFunctionCode ++
     ["}"]
    )
    where
    storageStrategy = ldStorageManagementStrategy ld
    arrayPrefix = getArrayPrefix storageStrategy

    declarationCode =
        let varDecl = ["  // index variable declarations:"] ++
                      ["  private int " ++ x ++ ";"
                       | x <- nub $ ldIndexVariables ld]
            paramDecl = ["  // parameter declarations:"] ++
                        ["  private final int " ++ x ++ ";"
                         | x <- ldParameters ld]
        in varDecl ++ paramDecl

    taskVarDeclCode = [""
                      ,"  final int[] " ++ taskIDName ++ ";"
                      ,""
                      ,"  // communication buffer:"
                      ,"  double[] " ++ bufferName ++ ";"
                      ]

    constructorCode =
      ["public " ++ taskClassname ++ "(int[] " ++ taskIDName ++ ") {"
      ,"  this." ++ taskIDName ++ " = " ++ taskIDName ++ ";"
      ,""
      ,"  // initialize all parameters:"] ++
      ["  " ++ x ++ " = " ++ masterClassname ++ "." ++ x ++ ";"
       | x <- ldParameters ld] ++
      [""
      ,"  // load accessed memory into the buffer"
      ,"  " ++ loadFuncName ++ "();"
      ,"}"
      ]

    loadCode = ["private void " ++ loadFuncName ++ "() {"
               ,"  // take the arrays from the master, not from the task"] ++
               ["  " ++ javaArrayType "double" storageStrategy (genericLength dims) ++ 
                " " ++ arrayPrefix ++ name ++ " = " ++ masterClassname ++ 
                "." ++ arrayPrefix ++ name ++ ";"
                | (name, dims, _) <- ldArrayVariables ld] ++
               [""
               ,ldArrayReadAccesses_StoreCode ld
               ,"}"
               ]
    executeCode = ["@Override"
                  ,"protected " ++
                   (if ldSupportOldFramework ld
                    then "int" else "void") ++ " execute() {"
                  ,"  // allocate local memory"
                  ,allocArraysExecuteCode ld
                  ,""
                  ,"  // update local memory"
                  ,ldArrayReadAccesses_LoadCode ld
                  ,""] ++ genInitTilingVarsCode (ldOuterTilingVariables ld) taskIDName ++
                  [""] ++ genIntegerVariableDefinitionsCode (ldExecuteTaskVariables ld) ++
                  [""
                  ,"  // begin computation code"
                  ,ldCompTaskCode ld
                  ,"  // end loopo computation code"
                  ,""
                  ,"  // write modified local memory into the buffer"
                  ,ldArrayWriteAccesses_StoreCode ld
                  ,""
                  ,"  // free local memory (so it will not be serialized)"] ++
                  ["  " ++ (getArrayPrefix storageStrategy) ++
                   name ++ " = null;" | (name, _, _) <- ldArrayVariables ld] ++
                  ["  return 0;" |  ldSupportOldFramework ld] ++
                  ["}"]

-- |
-- Generates the java code of the data class.
--
genDataJava :: LoopData
            -> String
genDataJava ld =
    unlines 
    ([genPackageDecl ld
     ,""
     ,"// This class contains the arrays that are used by the Task."
     ,"// It also serves as a memory pool, which avoids that the Task"
     ,"// has to allocate its local arrays more than once."] ++
     (if suppressAllWarnings then suppressWarningsCode else []) ++
     ["class " ++ dataClassname ++ " {"
     ,""
     ,"  // array declarations:"
     ] ++ 
     ["  static " ++ javaArrayType "double" storageStrategy (genericLength dims) ++ 
      " " ++ arrayPrefix ++ name ++ ";"
          | (name, dims, _) <- ldArrayVariables ld] ++ 
     [""
     ,bufferDecl
     ,"}"]
    )
    where
    storageStrategy = ldStorageManagementStrategy ld
    arrayPrefix = getArrayPrefix storageStrategy

    bufferDecl = case ldBufferManagementStrategy ld of
                   DynamicBuffer ->
                       unlines $
                       ["  static " ++ dynArrayClassname ++ " " ++ dynBufferName ++
                        " = new " ++ dynArrayClassname ++ "();"
                       ,""] ++ dynDoubleArrayCode True True
                   PrecomputedFixedArray ->
                       "  static double[] " ++ precomputedBufferName ++ ";"

-- |
-- Generates the java code of the peer-to-peer application master.
--
genP2PMasterJava :: LoopData
                 -> String
genP2PMasterJava ld =
    unlines
    ([genPackageDecl ld
     ,""
     ,"import java.io.IOException;"
     ,"import java.io.Serializable;"
     ,"import java.util.List;"
     ,""
     ,"import framework.HOCRuntimeSystem;"
     ,"import framework.MainApplication;"
     ,"import userinterface.UserMaster;"
     ,"import userinterface.UserClient;"
     ,"import userinterface.UserTask;"
     ,"import userinterface.UserScheduler;"
     ,"import schedulers.peer_to_peer.DependenceGraph;"
     ,"import schedulers.peer_to_peer.DependenceGraphBuilder;"
     ,"import schedulers.peer_to_peer.DependenceType;"
     ,"import static schedulers.peer_to_peer.DependenceType.*;"
     ,"import schedulers.peer_to_peer.GenericPeerToPeerTask;"
     ,"import schedulers.peer_to_peer.IfacePeerToPeerScheduler;"
     ,"import schedulers.peer_to_peer.PeerToPeerScheduler;"
     ,""
     ,""] ++ (if suppressAllWarnings then suppressWarningsCode else []) ++
     ["public class " ++ p2pMasterClassname ++
         " extends UserMaster implements IfacePeerToPeerScheduler {"
     ,""] ++ paramDeclarationCode ++
     [""] ++ masterVarDeclCode ++
     [""] ++ schedulerDeclCode ++
     [""] ++ genStartUpCode ld ++
     [""] ++ genEndCode ld ++
     [""] ++ joinTaskResultCode ++
     [""] ++ createSchedulerCode ++
     [""] ++ createUserClientCode ++
     [""] ++ writeSoakDataCode ++
     [""] ++ loadDrainDataCode ++
     [""] ++ genCreateDependenceGraphCode ld ++
     [""] ++ genAddAllTasksCode ld ++
     [""] ++ genPreloopFuncCode ld ++
     [""] ++ genPostloopFuncCode ld ++
-- Master doesn't use communication code anymore, only soak/drain:
--     [""] ++ genP2PCommWriteBufferCode ld ++
--     [""] ++ genP2PCommUnpackBufferCode ld ++
     [""] ++ helperFunctionCode ++
     ["}"]
    )
    where
    paramDeclarationCode =
      ["  // user declarations:"] ++
      [if matchVariableDeclaration x then "  static " ++ x else x
      | x <- lines (ldInitDeclarations ld)]

    masterVarDeclCode =
        (if ldStorageManagementStrategy ld == EhrhartBuffer 
         then [""]
         else ["  // master variable declarations:"] ++
              ["  private int " ++ x ++ ";"
                   | x <- ldMasterVariables ld]) ++                        
        ["  private String[] " ++ saveArgsName ++ ";"
        ,""
        ,"  // helper flag to discover whether the user code"
        ,"  // left its original main function"
        ,"  private boolean " ++ jumpedOutOfOriginalMainFuncFlag ++ ";"
        ,""
        ,"  private " ++ dynArrayClassname ++ " " ++
          dynBufferName ++ " = new " ++
          dynArrayClassname ++ "();"
        ]
    schedulerDeclCode =
      ["  // scheduler:"
      ,"  private PeerToPeerScheduler " ++ p2pSchedName ++ ";"
      ,""
      ,"  // dependence graph:"
      ,"  private DependenceGraph " ++ graphName ++ ";"
      ]

    joinTaskResultCode =
        ["@Override"
        ,"public long joinTaskResult(int " ++ workerIDName ++
         ", UserTask " ++ taskResultName ++ "__raw) {"
	,"  GenericPeerToPeerTask " ++ taskResultName ++
         " = (GenericPeerToPeerTask) " ++ taskResultName ++ "__raw;"
	,"  " ++ p2pSchedName ++ ".joinTask(" ++ taskResultName ++
         ".getTaskID(), " ++ workerIDName ++ ");"
        ,"  return " ++ p2pSchedName ++ ".getNumRemainingTasks();"
	,"}"
        ]
    createSchedulerCode =
        ["@Override"
        ,"public UserScheduler createScheduler() {"
        ,"  if (" ++ graphName ++ " == null) {"
        ,"    // We cannot create the dependence graph at this point"
        ,"    // because it depends on parameters that are not"
        ,"    // available until the startUp method is invoked."
        ,"    return null;"
        ,"  }"
        ,"  if (" ++ p2pSchedName ++ " != null) {"
        ,"    return " ++ p2pSchedName ++ ";"
        ,"  }"
        ,""
        ,"  assert(" ++ p2pSchedName ++ " == null && " ++
         graphName ++ " != null);"
        ,"  " ++ p2pSchedName ++ " = new PeerToPeerScheduler(this, " ++
         graphName ++ ");"
        ,"  return " ++ p2pSchedName ++ ";"
        ,"}"
        ]
    createUserClientCode =
        ["@Override"
        ,"public UserClient createUserClient() {"
        ,"  return new " ++ p2pClientClassname ++ "();"
        ,"}"
        ]
    writeSoakDataCode
        | eliminateSoakCommunication =
          ["@Override"
          ,"public Serializable writeSoakData(int[] " ++ taskIDName ++ ") {"
          ,"  assert(" ++ taskIDName ++ " != null);"
          ,"  return null; // all soak data has been already communicated"
          ,"}"
          ,""]
        | otherwise =
          ["@Override"
          ,"public Serializable writeSoakData(int[] " ++ taskIDName ++ ") {"
          ,"  assert(" ++ taskIDName ++ " != null);"
          ,""] ++
          genInitTilingVarsCode (ldOuterTilingVariables ld) taskIDName ++
          [""] ++ genIntegerVariableDefinitionsCode (ldWriteBufferVariables ld) ++
          [""
          ,"synchronized (" ++ dynBufferName ++ ") {"
          ,"  " ++ dynBufferName ++ ".reset();"
          ,""
          ,"  // begin write soak buffer code"
          ,ldP2P_MasterSoak ld
          ,"  // end write soak buffer code"
          ,""
          ,"  if (" ++ dynBufferName ++ ".isEmpty()) {"
          ,"    return null;"
          ,"  } else {"
          ,"    return " ++ dynBufferName ++ ".getArray();"
          ,"  }"
          ,"} // end synchronized"
          ,""
          ,"}"
          ]
    loadDrainDataCode =
        ["@Override"
        ,"public void loadDrainData(int[] " ++ taskIDName ++
         ", Serializable " ++ drainDataName ++ ") {"
        ,"  assert(" ++ drainDataName ++ " != null);"
        ,"  assert(" ++ taskIDName ++ " != null);"
        ,""] ++
        genInitTilingVarsCode (ldOuterTilingVariables ld) taskIDName ++
        [""] ++ genIntegerVariableDefinitionsCode (ldUnpackBufferVariables ld) ++

        [""
        ,"  final double[] " ++ bufferName ++ " = (double[]) " ++
         drainDataName ++ ";"
        ,"  int " ++ counterName ++ " = 0;"
        ,""
        ,"  // begin unpack drain buffer code"
        ,ldP2P_MasterDrain ld
        ,"  // end unpack drain buffer code"
        ,"}"
        ]

-- |
-- Generates the java code for the peer-to-peer application client.
--
genP2PClientJava :: LoopData
                 -> String
genP2PClientJava ld =
    verbose "genP2PClientJava" $
    unlines
    (["#include \"" ++ p2pClientDefFile ++ "\""
     ,genPackageDecl ld
     ,""
     ,"import java.io.Serializable;"
     ,"import java.util.List;"
     ,"import java.util.HashMap;"
     ,"import userinterface.UserClient;"
     ,"import userinterface.UserDataPacket;"
     ,if ldStorageManagementStrategy ld == EhrhartBuffer 
      then "import schedulers.peer_to_peer.IfaceGenericEhrhartBufferTask;"
      else ""
     ,""] ++ (if suppressAllWarnings then suppressWarningsCode else []) ++
     
     ["public class " ++ p2pClientClassname ++ " implements UserClient" ++ 
      if ldStorageManagementStrategy ld == EhrhartBuffer 
      then ", IfaceGenericEhrhartBufferTask {"
      else " {"
     ,""
     ,genSerialVersionUIDDecl ld
     ,""] ++ declarationCode ++
     [""] ++ clientVarDeclCode ++
     [""] ++ constructorCode ++
     [""] ++ initCode ++
     [""] ++ ehrhartAccessCode ++
     [""] ++ ehrhartArrayAllocCode ++
     [""] ++ executeTaskCode ++
     [""] ++ loadDataFromPeerCode ++
     [""] ++ writeDataForPeerCode ++
     [""] ++ loadSoakDataCode ++
     [""] ++ writeDrainDataCode ++
     [""] ++ genP2PCommWriteBufferCode ld ++
     [""] ++ genP2PCommUnpackBufferCode ld ++
     [""] ++ helperFunctionCode ++
     ["}"]
    )
    where
    storageStrategy = ldStorageManagementStrategy ld
    arrayPrefix = getArrayPrefix storageStrategy

    declarationCode =
        let varDecl = ["  // index variable declarations:"] ++
                      ["  private int " ++ x ++ ";"
                       | x <- nub $ ldIndexVariables ld]
            arrayDecl = [""
                        ,"  // array declarations:"] ++
                        ["  private " ++
                         javaArrayType "double" storageStrategy (genericLength dims) ++
                         " " ++ arrayPrefix ++ name ++ ";"
                         | (name, dims, _) <- ldArrayVariables ld
                        ]
            paramDecl = [""
                        ,"  // parameter declarations:"] ++
                        ["  private int " ++ x ++ ";" | x <- ldParameters ld]
        in varDecl ++ arrayDecl ++ paramDecl

    clientVarDeclCode =
      [""
      ,"  private " ++ dynArrayClassname ++ " " ++ dynBufferName ++ ";"
      ,""
      ,"  // the map holding communication buffers for every destination task:"
      ,"  private " ++ "HashMap<" ++ dependenceIndexName ++ "," ++
       "  double[]> " ++ communicationBufferMapName ++ ";"
      ]

    constructorCode =
      [p2pClientClassname ++ "() {"
      ,"  // initialize parameters"] ++
      ["  this." ++ x ++ " = " ++ p2pMasterClassname ++ "." ++ x ++ ";"
       | x <- ldParameters ld] ++
      initArraysCode ++
      ["}"]
      where
      initArraysCode
        | eliminateSoakCommunication =
            [""
            ,"  // initialize arrays"] ++
            ["  this." ++ arrayPrefix ++ name ++ " = " ++ p2pMasterClassname ++
             "." ++ arrayPrefix ++ name ++ ";"
             | (name, _, access) <- ldArrayVariables ld,
               access /= WriteOnlyAccess]
        | otherwise = ["  // arrays will be initialized in the init method"]

    initCode = ["// This code will be executed on the workers only"
               ,"// (i.e., the allocated arrays will not be serialized"
               ,"//  and sent over the network)."
               ,"@Override"
               ,"public void init() {"
               ] ++
               (if ldStorageManagementStrategy ld == EhrhartBuffer then 
                    ["  " ++ communicationBufferMapName ++ 
                     " = new HashMap<" ++ dependenceIndexName ++ ", double[]>();"
                    ]
                else []) ++
               ["  // init dynamic buffer:"
               ,"  " ++ dynBufferName ++ " = new " ++ dynArrayClassname ++ "();"] ++
               allocationCode ++
               ["}"]
               where
               unitializedArrays
                  | eliminateSoakCommunication =
                     filter (\(_,_,access) -> access == WriteOnlyAccess)
                            (ldArrayVariables ld)
                  | otherwise = ldArrayVariables ld
               allocationCode
                  | null unitializedArrays =
                     [""
                     ,"  // (all arrays have been already initialized)"]
                  | otherwise =
                     [""
                     ,"  // allocate local arrays:"] ++
                     ["  " ++ arrayPrefix ++ name ++ " = " ++
                      getNewArray "double" storageStrategy dims ++ ";"
                     | (name, dims, _) <- unitializedArrays]
    ehrhartAccessCode =
        verbose ("in genP2PClientJava: ehrhartAccessCode***") $
        if ldStorageManagementStrategy ld == EhrhartBuffer
        then [ldEhrhartAccessCode ld]
        else []
    ehrhartArrayAllocCode = 
        verbose ("in genP2PClientJava: ehrhartAllocCode***") $
        if ldStorageManagementStrategy ld == EhrhartBuffer
        then ["public void allocEhrhartArrays(int[] " ++ taskIDName ++ ") {"
             ,"  // code for allocation of ehrhart arrays:"
             ,"  " ++ ldEhrhartArrayAllocCode ld
             ,"}"]
        else []
       
    executeTaskCode =
        ["@Override"
        ,"public void executeTask(int[] " ++ taskIDName ++ ") {"] ++
        genInitTilingVarsCode (ldOuterTilingVariables ld) taskIDName ++
        [""] ++ genIntegerVariableDefinitionsCode (ldExecuteTaskVariables ld) ++
        [""
        ,"  // computation code:"
        ,ldCompTaskCode ld
        ,"}"
        ]
    loadDataFromPeerCode =
        ["@Override"
        ,"public void loadDataFromPeer(UserDataPacket " ++
         dataPacketName ++ "__raw__, int[] " ++ sourceTaskIDName ++
         ", int[] " ++ targetTaskIDName ++ ") {"
        ,"  " ++ p2pDataPacketClassname ++ " " ++ dataPacketName ++
         " = (" ++ p2pDataPacketClassname ++ ") " ++
         dataPacketName ++ "__raw__;"
        ,"  double[] " ++ bufferName ++ " = " ++
         dataPacketName ++ ".getBuffer();"
        ,"  " ++ unpackCommBufferFuncName ++ "(" ++ sourceTaskIDName ++
         ", " ++ targetTaskIDName ++ ", " ++ bufferName ++ ");"
        ,"}"
        ]
    writeDataForPeerCode =
        ["@Override"
        ,"public UserDataPacket writeDataForPeer(int[] " ++
-- TODO: nur entsprechenden communication buffer aus bufferMap rausnehmen (-> remove!)
--       writeCommBuffer wird schon vorher aufgerufen, nach execute()
--       wichtig: ASSERTION: buffer != null
         sourceTaskIDName ++ ", int[] " ++ targetTaskIDName ++ ") {"
        ,"  double[] " ++ bufferName ++ " = " ++
        if ldStorageManagementStrategy ld == EhrhartBuffer
        then (communicationBufferMapName ++ ".remove(new " ++ dependenceIndexName ++
              "(" ++ sourceTaskIDName ++
              ", " ++ targetTaskIDName ++ "));")
        else (writeCommBufferFuncName ++ "(" ++ sourceTaskIDName ++ ", " ++
              targetTaskIDName ++ ");")
        ,"  assert(" ++ bufferName ++ " != null);"
        ,"  return new " ++ p2pDataPacketClassname ++ "(" ++ sourceTaskIDName ++
         ", " ++ targetTaskIDName ++ ", " ++ bufferName ++ ");"
        ,"}"
        ]
    loadSoakDataCode =
        ["@Override"
        ,"public void loadSoakData(Serializable " ++ soakDataName ++
         ", int[] " ++ taskIDName ++ ") {"
        ,"  assert(" ++ soakDataName ++ " != null);"
        ,"  assert(" ++ taskIDName ++ " != null);"
        ,""] ++
        genInitTilingVarsCode (ldOuterTilingVariables ld) taskIDName ++
        [""] ++ genIntegerVariableDefinitionsCode (ldUnpackBufferVariables ld) ++
        [""
        ,"  final double[] " ++ bufferName ++ " = (double[]) " ++
         soakDataName ++ ";"
        ,"  int " ++ counterName ++ " = 0;"
        ,""
        ,"  // begin unpack soak buffer code"
        ,ldP2P_Soak_unpackBuffer ld
        ,"  // end unpack soak buffer code"
        ,"}"
        ]
    writeDrainDataCode =
        ["@Override"
        ,"public Serializable writeDrainData(int[] " ++ taskIDName ++ ") {"
        ,"  assert(" ++ taskIDName ++ " != null);"
        ,""] ++
        genInitTilingVarsCode (ldOuterTilingVariables ld) taskIDName ++
        [""] ++ genIntegerVariableDefinitionsCode (ldWriteBufferVariables ld) ++
        [""
        ,"synchronized (" ++ dynBufferName ++ ") {"
        ,"  " ++ dynBufferName ++ ".reset();"
        ,""
        ,"  // begin write drain buffer code"
        ,ldP2P_Drain_writeBuffer ld
        ,"  // end write drain buffer code"
        ,""
        ,"  if (" ++ dynBufferName ++ ".isEmpty()) {"
        ,"    return null;"
        ,"  } else {"
        ,"    return " ++ dynBufferName ++ ".getArray();"
        ,"  }"
        ,"} // end synchronized"
        ,""
        ,"}"
        ]

-- |
-- Generates the java code for the peer-to-peer data packet.
--
genP2PDataPacketJava :: LoopData
                     -> String
genP2PDataPacketJava ld =
    unlines
    ([genPackageDecl ld
     ,""
     ,"import userinterface.UserDataPacket;"
     ,""
     ,"class " ++ p2pDataPacketClassname ++ " extends UserDataPacket {"
     ,""
     ,genSerialVersionUIDDecl ld
     ,""
     ,"  private final double[] " ++ bufferName ++ ";"
     ,""
     ] ++ constructorCode ++
     [""] ++ accessorCode ++
     ["}"]
    )
    where
    constructorCode =
        [p2pDataPacketClassname ++ "(int[] " ++ sourceTaskIDName ++
         ", int[] " ++ targetTaskIDName ++ ", double[] " ++
         bufferName ++ ") {"
        ,"  super(" ++ sourceTaskIDName ++ ", " ++ targetTaskIDName ++ ");"
        ,"  assert(" ++ bufferName ++ " != null);"
        ,"  this." ++ bufferName ++ " = " ++ bufferName ++ ";"
        ,"}"
        ]
    accessorCode =
        ["  double[] getBuffer() {"
        ,"    return " ++ bufferName ++ ";"
        ,"  }"
        ]

-- |
-- Generates the java code for the dynamic array helper class.
--
genP2PDynamicArrayClassJava :: LoopData
                            -> String
genP2PDynamicArrayClassJava ld =
  unlines ([genPackageDecl ld, ""] ++ dynDoubleArrayCode False False)

-- |
-- Peer-to-peer communication code for preparing
-- the buffer that should be sent.
--
-- Supported HOCs: P2P
--
genP2PCommWriteBufferCode :: LoopData
                          -> [String]
genP2PCommWriteBufferCode ld 
    | ldStorageManagementStrategy ld == EhrhartBuffer =
        (["public void " ++ writeCommBufferFuncName ++
          "(int[] " ++ sourceTaskIDName ++
          ", int[] " ++ targetTaskIDName ++ ") {"] ++
          genInitTilingVarsCode (ldOuterTilingVariables ld) sourceTaskIDName ++
          genInitTilingVarsCode (ldInnerTilingVariables ld) targetTaskIDName ++
          [""] ++ genIntegerVariableDefinitionsCode (ldWriteBufferVariables ld) ++
          [""
          ,"  " ++ dynBufferName ++ ".reset();"
          ,""
          ,"  // begin communication code"
          ,ldP2P_Communication_writeBuffer ld
          ,"  // end communication code"
          ,""
          ,"  " ++ (communicationBufferMapName ++ ".put(new " ++ dependenceIndexName ++
              "(" ++ sourceTaskIDName ++
              ", " ++ targetTaskIDName ++ "), " ++ dynBufferName ++ ".getArray());")
          ,"}"
          ])        
    | otherwise =
        (["public double[] " ++ writeCommBufferFuncName ++
          "(int[] " ++ sourceTaskIDName ++
          ", int[] " ++ targetTaskIDName ++ ") {"] ++
          genInitTilingVarsCode (ldOuterTilingVariables ld) sourceTaskIDName ++
          genInitTilingVarsCode (ldInnerTilingVariables ld) targetTaskIDName ++
          [""] ++ genIntegerVariableDefinitionsCode (ldWriteBufferVariables ld) ++
          [""
          ,"synchronized (" ++ dynBufferName ++ ") {"
          ,"  " ++ dynBufferName ++ ".reset();"
          ,""
          ,"  // begin communication code"
          ,ldP2P_Communication_writeBuffer ld
          ,"  // end communication code"
          ,""
          ,"  return " ++ dynBufferName ++ ".getArray(); // (this function should never return null)"
          ,"} // end synchronized"
          ,"}"
          ])

-- |
-- Peer-to-peer communication code for reading from
-- the buffer that has been received.
--
-- Supported HOCs: P2P
--
genP2PCommUnpackBufferCode :: LoopData
                           -> [String]
genP2PCommUnpackBufferCode ld =
    ["private void " ++ unpackCommBufferFuncName ++
     "(int[] " ++ sourceTaskIDName ++ ", int[] " ++ targetTaskIDName ++
     ", double[] " ++ bufferName ++ ") {"] ++
    genInitTilingVarsCode (ldOuterTilingVariables ld) sourceTaskIDName ++
    genInitTilingVarsCode (ldInnerTilingVariables ld) targetTaskIDName ++
    [""] ++ genIntegerVariableDefinitionsCode (ldUnpackBufferVariables ld) ++
    [""
    ,"  int " ++ counterName ++ " = 0;"
    ,""
    ,ldP2P_Communication_unpackBuffer ld
    ,"}"
    ]

-- |
-- Generates code for the startUp method of the application master.
--
-- Supported HOCs: Master\/Worker, P2P
--
genStartUpCode :: LoopData
               -> [String]
genStartUpCode ld = ["@Override"
                    ,"public long startUp(String[] args" ++
                     (if ldSupportOldFramework ld
                      then ", UserScheduler unused__var__sched"
                      else "")
                     ++ ") {"
                    ,"  // initialization:"
                    ,"  " ++ saveArgsName ++ " = args;"
                    ,"  " ++ preloopFuncName ++ "(args);"] ++
                    ["  DependenceGraphBuilder " ++ graphBuilderName ++
                     " = new DependenceGraphBuilder();"
                    | hocType == P2P_HOC] ++
                    ["  if (" ++ jumpedOutOfOriginalMainFuncFlag ++ ") {"
                    ,"    // loop will not be executed -> do not create tasks"] ++
                    maybePassGraphToScheduler hocType ++
                    ["    return 0;"
                    ,"  }"
                    ,""
                    ,"  // Prepare tasks and the dependence graph"
                    ,"  " ++ addAllTasksFuncName ++ paramlist hocType
                    ,"  " ++ createDependenceGraphFuncName ++ paramlist hocType
                    ,""] ++
                    maybePassGraphToScheduler hocType ++
                    returnCode hocType ++
                    ["}"]
  where
  hocType = ldHOC_Type ld
  paramlist MasterWorkerHOC = "();"
  paramlist P2P_HOC         = "(" ++ graphBuilderName ++ ");"

  maybePassGraphToScheduler P2P_HOC =
    ["  final HOCRuntimeSystem rs = MainApplication.getHOCRuntimeSystem();"
    ,"  if (rs.shouldRemoveTransitiveDependences()) {"
    ,"    " ++ graphBuilderName ++ ".removeTransitiveEdges();"
    ,"  }"
    ,"  " ++ graphName ++ " = " ++ graphBuilderName ++ ".convertToGraph();"
    ,"  if (rs.shouldDrawTaskDependenceGraph()) {"
    ,"    try {"
    ,"      " ++ graphName ++ ".writeDotFile(rs.getTaskDependenceGraphFilename());"
    ,"      if (rs.exitAfterDrawingGraph()) {"
    ,"        // workaround to assure that the task dependence graph is empty"
    ,"        " ++ graphName ++ " = (new DependenceGraphBuilder()).convertToGraph();"
    ,"        createScheduler();"
    ,"        return 0;"
    ,"      }"
    ,"   } catch (IOException e) {"
    ,"      throw new RuntimeException("
    ,"        \"Error: Could not write task dependence graph\" + e);"
    ,"    }"
    ,"  }"
    ,"  createScheduler();"
    ]
  maybePassGraphToScheduler _ = []

  returnCode MasterWorkerHOC =
    ["  " ++ numTasksName ++ " = " ++ schedDepName ++ ".getNumTasks();"
    ,"  return " ++ numTasksName ++ ";"
    ]
  returnCode P2P_HOC = ["  return " ++ p2pSchedName ++ ".getNumTasks();"]

-- |
-- Generates code for the end method of the application master.
--
-- Supported HOCs: Master\/Worker, P2P
--
genEndCode :: LoopData
           -> [String]
genEndCode ld =
  ["@Override"
  ,"public " ++ (if ldSupportOldFramework ld
                 then "int" else "void") ++ " end() {"
  ,"  " ++ postloopFuncName ++ "(this." ++ saveArgsName ++ ");"] ++
  ["  return 0;" | ldSupportOldFramework ld] ++
  ["}"]

genGetResultsCode :: LoopData
                  -> [String]
genGetResultsCode ld =
  ["public Map<String, Object> getResults() {"
  ,"  Map<String, Object> " ++ resultMapName ++
   " = new HashMap<String, Object>();"] ++
  ["  " ++ resultMapName ++ ".put(\"" ++ name ++ "\", " ++ name ++ ");"
   | name <- resultArrays] ++
  ["  return map;"
  ,"}"
  ]
  where
  resultArrays =
    [name | (name, _, access) <- ldArrayVariables ld, access /= ReadOnlyAccess]

-- |
-- Generates code for the private addTasks method and its helper method.
--
-- Supported HOCs: Master\/Worker, P2P
--
genAddAllTasksCode :: LoopData
                   -> [String]
genAddAllTasksCode ld =
  ["private final void " ++ addAllTasksFuncName ++ "(" ++
   maybeGraphBuilder hocType "" ++") {"] ++
  genIntegerVariableDefinitionsCode (ldCreateDependenceGraphVariables ld) ++
  [""
  ,ldAddAllTasksCode ld
  ,"}"
  ,""
  ,"// Only adds a dependence if the source (dep1)"
  ,"// and the target (dep2) are different."
  ,"private final void " ++ addDependenceHelperFuncName ++
   "(int[] dep1, int[] dep2" ++ maybeGraphBuilder hocType ", " ++
   maybeDependenceType hocType ", " ++ ") {"
  ,"  for (int i = 0; i < dep1.length; i++) {"
  ,"    if (dep1[i] != dep2[i]) {"
  ,"      " ++ addDependenceCode hocType
  ,"      return;"
  ,"    }"
  ,"  }"
  ,"}"
  ]
  where
  hocType = ldHOC_Type ld
  addDependenceCode MasterWorkerHOC =
    schedDepName ++ ".addDependence(dep1, dep2);"
  addDependenceCode P2P_HOC =
    graphBuilderName ++ ".addDependence(dep1, dep2, depType);"
  maybeGraphBuilder P2P_HOC prefix =
    prefix ++ "DependenceGraphBuilder " ++ graphBuilderName
  maybeGraphBuilder _ _ = ""
  maybeDependenceType P2P_HOC prefix =
    prefix ++ "DependenceType depType"
  maybeDependenceType _ _ = ""

-- |
-- Match java variable definitions.
--
-- Example:
--
--   1) "int[] a; \/\/ will be matched"
--
--   2) "\/\/ int[] a; \/\/ will be rejected (as it is a commentary)"
--
matchVariableDeclaration :: String
                         -> Bool
matchVariableDeclaration x = x =~ "^[ \t\r]*[a-zA-Z0-9_]+([ \t\r]*\\[[ \t\r]*\\])*[ \t\r]+[a-zA-Z0-9_]+[ \t\r]*([ \t\r]*\\[[ \t\r]*\\])*[ \t\r]*;[ \t\r]*(//.*)?$" -- This regular expression is definitely write-and-forget code.

-- |
-- Generates an unique package name postfix from the tiling vectors.
-- 
getTileNaming :: LoopData 
              -> String
getTileNaming ld =
    let showVect xs = concat $ intersperse "_" (map show xs)
    in concat $ intersperse "__" (map showVect (ldRawTilingVector ld))

-- |
-- Generates code to initialize the tiling variables
-- with the values from the task ID.
--
genInitTilingVarsCode :: [String] -- ^ the name of the tiling variables
                      -> String   -- ^ the name of the task ID variable
                      -> [String]
genInitTilingVarsCode tilingVars taskIDName =
  ["  // initialize tiling variables:"] ++
  ["  final int " ++ var ++ " = " ++ taskIDName ++ "[" ++ show n ++ "];"
   | (var, n) <- zip tilingVars [0..]]

-- |
-- Generates code to define integer variables.
--
genIntegerVariableDefinitionsCode :: [String] -- ^ the name of the variables
                                  -> [String]
genIntegerVariableDefinitionsCode vars =
  ["  // variable definitions:"] ++
  ["  int " ++ x ++ ";" | x <- vars]

-- |
-- Generates the java package declaration code,
-- which is the same for each of the generated java files.
--
genPackageDecl :: LoopData
               -> String
genPackageDecl ld =
    "package " ++ ldPackageName ld ++ getTileNaming ld ++ ";"

genSerialVersionUIDDecl :: LoopData
                        -> String
genSerialVersionUIDDecl ld =
    unlines $
    ["  /** Explicitly set this variable to avoid serialization errors. */"
    ,"  private static final long serialVersionUID = " ++
       show (ldSerialVersionUID ld) ++ "L;"
    ]

-- |
-- Generates the code for the method that creates the dependence graph.
-- It uses the generated loops for each dependence and puts them together.
--
genCreateDependenceGraphCode :: LoopData
                             -> [String]
genCreateDependenceGraphCode ld =
    ["private final void " ++ createDependenceGraphFuncName ++ "(" ++
     maybeGraphBuilder (ldHOC_Type ld) ++ ") {"] ++
    genIntegerVariableDefinitionsCode (ldCreateDependenceGraphVariables ld) ++
    zipWith addDependenceCode (ldCreateDependenceGraphCode ld) [1..] ++
    ["}"]
    where
    addDependenceCode :: String -> Integer -> String
    addDependenceCode rawAddDepCode id =
        unlines $
        [""
        ,"  // " ++ show id ++ ". dependence:"
        ,rawAddDepCode
        ]
    maybeGraphBuilder P2P_HOC = "DependenceGraphBuilder " ++ graphBuilderName
    maybeGraphBuilder _ = ""

-- |
-- Contains the user code that is executed before the loop
-- that should be parallized.
--
genPreloopFuncCode :: LoopData
                   -> [String]
genPreloopFuncCode ld = ["// This method contains the original user code"
                      ,"// that is executed before the parallel loop."
                      ,"private void " ++ preloopFuncName ++
                         "(String[] args) {"
                      ,"  // Set the flag which may be cleared at" ++
                         " the end of the function"
                      ,"  " ++ jumpedOutOfOriginalMainFuncFlag ++ " = true;"
                      ,""
                      ," // Begin of user code"
                      ,ldPreloopCode ld
                      ," // End of user code"
                      ,""
                      ,"  // This line can only be reached, if the user code"
                      ,"  // did not leave the original main function."
                      ,"  " ++ jumpedOutOfOriginalMainFuncFlag ++ " = false;"
                      ,"}"
                      ]

-- |
-- Contains the user code that is executed after the loop
-- that should be parallized.
--
genPostloopFuncCode :: LoopData
                    -> [String]
genPostloopFuncCode ld = ["// This method contains the original user code"
                         ,"// that is executed after the parallel loop."
                         ,"private void " ++ postloopFuncName ++
                          "(String[] args) {"
                         ,"  if (" ++ jumpedOutOfOriginalMainFuncFlag ++ ") {"
                         ,"    return;"
                         ,"  }"
                         ,""
                         ,"  // Begin user code"
                         ,ldPostloopCode ld
                         ,"  // End user code"
                         ,"}"
                         ]

-- |
-- Example: javaArrayType "double" 3 --> "double[][][]"
--
javaArrayType :: String  -- ^ internal stored data type (e.g., "double")
              -> StorageManagementStrategy -- ^ local arrays \/ hashtable, etc.
              -> Integer -- ^ dimension
              -> String
javaArrayType typeStr LocalArray dim = typeStr ++ (concat $ genericReplicate dim "[]")
javaArrayType typeStr LocalDataStructure dim = 
  -- e.g.: doubleArray2D
  typeStr ++ "Array" ++ (show dim) ++ "D"
javaArrayType typeStr EhrhartBuffer dim = typeStr ++ "[]" -- always 1-dim buffer!

-- |
-- Contains the implementation of helper functions
-- that will be used in the master and the task classes:
--
--   1) floor
--
--   2) ceil
--
--   3) floord
--
--   4) ceild
--
--   5) min
--
--   6) max
--
helperFunctionCode :: [String]
helperFunctionCode =
    [floorCode, ""
    ,ceilCode, ""
    ,floordCode, ""
    ,ceildCode, ""
    ,minCode, ""
    ,maxCode
    ]
    where
    floorCode = unlines ["final private static int floor(double x) {"
                        ,"  return (int) Math.floor(x);"
                        ,"}"
                        ]
    ceilCode = unlines ["final private static int ceil(double x) {"
                       ,"  return (int) Math.ceil(x);"
                       ,"}"
                       ]
    floordCode = unlines ["final private static int floord(int x, int y) {"
                         ,"  if (y < 0) {"
                         ,"    x=-x;"
                         ,"    y=-y;"
                         ,"  }"
                         ,"  return (x>=0) ? x/y : (x-y+1)/y;"
                         ,"}"
                         ]
    ceildCode = unlines ["final private static int ceild(int x, int y) {"
                         ,"  if (y < 0) {"
                         ,"     x=-x;"
                         ,"     y=-y;"
                         ,"  }"
                         ,"  return (x>=0) ? (x+y-1)/y : x/y;"
                        ,"}"
                        ]
    minCode = unlines ["final private static int min(int x, int y) {"
                      ,"  return Math.min(x, y);"
                      ,"}"
                      ]
    maxCode = unlines ["final private static int max(int x, int y) {"
                      ,"  return Math.max(x, y);"
                      ,"}"
                      ]

-- |
-- Contains the implementation of a dynamic array class used for buffering.
--
dynDoubleArrayCode :: Bool -- ^ set this flag if the code should be included as an inner class
                   -> Bool -- ^ set this flag if the code should keep references to all allocated arrays
                   -> [String]
dynDoubleArrayCode isInnerClass True = 
  [(if isInnerClass then "  static " else "") ++
   "final class " ++ dynArrayClassname ++ " {"
  ,"    private double[] array;"
  ,"    private int used = 0;"
  ,"    private final double factor;"
  ,"    private java.util.ArrayList<double[]> keepSmallArrays ="
  ,"        new java.util.ArrayList<double[]>();"
  ,""
  ,"    DynamicDoubleArray() {"
  ,"      this(2.0);"
  ,"    }"
  ,""  
  ,"    DynamicDoubleArray(double factor) {"
  ,"      assert(factor > 1.0);"
  ,"      this.factor = factor;"
  ,"      this.array = new double[16];"
  ,"      keepSmallArrays.add(array);"
  ,"    }"
  ,""	  
  ,"    void reset() {"
  ,"      used = 0;"
  ,"    }"
  ,""
  ,"    boolean isEmpty() {"
  ,"      return used == 0;"
  ,"    }"
  ,""
  ,"    void add(double x) {"
  ,"      if (used >= array.length) {"
  ,"        // increase the array"
  ,"        final int newSize = Math.max(array.length + 1, (int) (factor * array.length));"
  ,"        double[] newArray = new double[newSize];"
  ,"        System.arraycopy(array, 0, newArray, 0, array.length);"
  ,"        array = newArray;"
  ,"        keepSmallArrays.add(array);"
  ,"      }"
  ,"      array[used++] = x;"
  ,"    }"
  ,""
  ,"    double[] getArray() {"
  ,"      final int last = keepSmallArrays.size() - 1;"
  ,"      int pos = last;"
  ,"      while (pos > 0 && keepSmallArrays.get(pos-1).length >= used) {"
  ,"        pos--;"
  ,"      }"
  ,"      if (pos == last) {"
  ,"        return array;"
  ,"      } else {"
  ,"        double[] shrinkArray = keepSmallArrays.get(pos);"
  ,"        System.arraycopy(array, 0, shrinkArray, 0, used);"
  ,"        return shrinkArray;"
  ,"      }"
  ,"    }"
  ,"  }"
  ]
dynDoubleArrayCode isInnerClass False =
  [(if isInnerClass then "  static " else "") ++
   "final class " ++ dynArrayClassname ++ " {"
  ,"    private double[] array;"
  ,"    private int used = 0;"
  ,"    private final double factor;"
  ,""
  ,"    DynamicDoubleArray() {"
  ,"      this(2.0);"
  ,"    }"
  ,""  
  ,"    DynamicDoubleArray(double factor) {"
  ,"      assert(factor > 1.0);"
  ,"      this.factor = factor;"
  ,"      this.array = new double[1024];"
  ,"    }"
  ,""	  
  ,"    void reset() {"
  ,"      used = 0;"
  ,"    }"
  ,""
  ,"    boolean isEmpty() {"
  ,"      return used == 0;"
  ,"    }"
  ,""
  ,"    void add(double x) {"
  ,"      if (used >= array.length) {"
  ,"        // increase the array"
  ,"        final int newSize = Math.max(array.length + 1, (int) (factor * array.length));"
  ,"        double[] newArray = new double[newSize];"
  ,"        System.arraycopy(array, 0, newArray, 0, array.length);"
  ,"        array = newArray;"
  ,"      }"
  ,"      array[used++] = x;"
  ,"    }"
  ,""
  ,"    double[] getArray() {"
  ,"      double[] tmp = new double[used];"
  ,"      System.arraycopy(array, 0, tmp, 0, used);"
  ,"      return tmp;"
  ,"    }"
  ,"  }"
  ]

-- |
-- Creates java code to add a task, which is described
-- by its list of tiling indices.
--
genAddSingleTaskCode :: HOC_Type
                     -> [String] -- ^ taskID
                     -> String
genAddSingleTaskCode MasterWorkerHOC taskID =
  schedDepName ++ ".addTask(" ++ genTaskIDCode taskID ++ ");"
genAddSingleTaskCode P2P_HOC taskID =
  graphBuilderName ++ ".registerTask(" ++ genTaskIDCode taskID ++ ");"

-- |
-- Creates java code to add a dependence between two tasks,
-- which are described by their list of tiling indices (task IDs).
-- Extension: now we also pass the dependence type along with
-- the task IDs.
--
genAddDependenceCode :: HOC_Type
                     -> [String] -- ^ taskID 1
                     -> [String] -- ^ taskID 2
                     -> TrDependence
                     -> String
genAddDependenceCode hocType taskID1 taskID2 trDep =
    (addDependenceHelperFuncName ++ "(" ++ 
     genTaskIDCode taskID1 ++ ", " ++
     genTaskIDCode taskID2 ++ 
     maybeGraphBuilder hocType ++ 
     maybeDepType hocType ++
     ");")
    where
    maybeGraphBuilder P2P_HOC = ", " ++ graphBuilderName
    maybeGraphBuilder _       = ""

    maybeDepType P2P_HOC = ", " ++ show (trDepType trDep)
    maybeDepType _       = ""

-- |
-- Creates java code to either write into a buffer or
-- unpack from the buffer. The buffer name and array access names are given.
--
genBufferCode :: (String, ArrayID)
              -- ^ the array name and arrayID
              -> [String]
              -- ^ the array access variable names
              -> Bool
              -- ^ writeMode = True | unpackMode = False
              -> MasterClientCode
              -- ^ the choice of code for either Master or Client
              -> StorageManagementStrategy
              -- ^ storage management strategies for 
              --   local data needed in computation
              -> Maybe ([String])
              -- ^ the parameters for the ehrhart access for the current array
              -> String
genBufferCode (arrayName, _) accNames isWriteMode _ LocalArray Nothing = bufferCode
    where
    bufferCode = 
        if isWriteMode
        then (dynBufferName ++ ".add(" ++ arrayAccess ++ ");")
        else (arrayAccess ++ " = " ++ bufferName ++ "[" ++ counterName ++ "++];")
    arrayAccess = arrayName ++ "[" ++ (concat $ intersperse "][" accNames) ++ "]"
genBufferCode (arrayName, _) accNames isWriteMode _ LocalDataStructure Nothing = bufferCode
    where
    bufferCode = 
        if isWriteMode
        then (dynBufferName ++ ".add(" ++ arrayStructName ++ "." ++ 
              arrayStructureGet ++ "(" ++ arrayArguments ++ ")" ++ ");")
        else (arrayStructName ++ "." ++ arrayStructureSet ++ "(" ++ 
              bufferName ++ "[" ++ counterName ++ "++]" ++ ", " ++ arrayArguments ++ ");")
    arrStructReadAccess = 
        (arrayStructName ++ "." ++ arrayStructureGet ++ "(" ++ arrayArguments ++ ")")
    arrayStructName = arrayStructurePrefix ++ arrayName
    arrayArguments = (concat $ intersperse ", " accNames) 
-- code for master doesn't use ehrhart buffers:
genBufferCode (arrayName, _) accNames isWriteMode MasterCode EhrhartBuffer _ = 
    bufferCode
    where
    bufferCode = 
        if isWriteMode
        then (dynBufferName ++ ".add(" ++ arrayAccess ++ ");")
        else (arrayAccess ++ " = " ++ bufferName ++ "[" ++ counterName ++ "++];")
    arrayAccess = arrayName ++ "[" ++ (concat $ intersperse "][" accNames) ++ "]"
-- code for client uses ehrhart buffers:
genBufferCode (arrayName, arrayID) accNames isWriteMode ClientCode 
              EhrhartBuffer (Just ehrhartParams) = 
    bufferCode
    where
    bufferCode = 
-- TODO:
-- hier EhrhartIndex verwenden!
-- ehrhartIndex fuer dieses Array benutzen, verwendet srcTile als Parameter,
-- sowie arrayAccess als Eingabe, liefert Index im Puffer fuer Array
        if isWriteMode
        then (dynBufferName ++ ".add(" ++ arrayAccess ++ ");")
        else (arrayAccess ++ " = " ++ bufferName ++ "[" ++ counterName ++ "++];")
    arrayAccess = 
        ehrhartArrayName ++ "[" ++ ehrhartAccessCode ++ "]"
    ehrhartArrayName = getArrayPrefix EhrhartBuffer ++ arrayName
    ehrhartAccessCode = ehrhartAccessFuncName ++ arrayNumber ++ 
                        "(" ++ arguments ++ ")"
    arrayNumber = show $ integerFromID arrayID
    arguments = concat $ intersperse ", " $ ehrhartParams

-- |
-- Generate java code to generate the task ID from
-- a list of tiling indices.
--
genTaskIDCode :: [String] -- ^ taskID
              -> String
genTaskIDCode indices =
    "new int[] { " ++ (concat $ intersperse ", " indices) ++ " }"

-- |
-- Replaces varMin and varMax by Math.min and Math.max
-- in the communication loops.
--
eliminateVarMinMax :: String
                   -> String
eliminateVarMinMax str
    | "for" `isPrefixOf` str = rep str
    | otherwise = str
    where
    rep :: String -> String
    rep "" = ""
    rep s@(x:xs)
        | "varMin" `isPrefixOf` s = subst "Math.min" (drop (length "varMin(") s)
        | "varMax" `isPrefixOf` s = subst "Math.max" (drop (length "varMax(") s)
        | otherwise = x : rep xs
    subst :: String -> String -> String
    subst func s = let (args, rem) = parse s
                   in genCode func args ++ rep rem
    parse :: String -> ([String], String)
    parse s = let argstr = findEnd s
                  args   = repspan argstr
              in if null args
                 then error "eliminateVarMinMax should not be reached (1)"
                 else (tail args, drop (1 + length argstr) s)
    findEnd :: String -> String
    findEnd ""          = error "eliminateVarMinMax should not be reached (2)"
    findEnd (')':';':_) = ""
    findEnd (x:xs)      = x : findEnd xs
    repspan :: String -> [String]
    repspan "" = []
    repspan s = let (arg, rem) = span (/= ',') s
                in if null rem
                   then [arg]
                   else arg : repspan (tail rem)
    genCode _ [x] = x
    genCode func (x:y:rem) = genCode func $ (func ++ "(" ++ x ++ ", " ++ y ++ ")") : rem
    
-- |
-- Creates (or resets) the internal buffer.
-- 
createBuffer :: BufferManagementStrategy
             -> [Access]
             -> CompSpreadDS
             -> Bool
             -> Bool
             -> String
createBuffer DynamicBuffer _ _ _ _ =
    unlines $
    ["  " ++ dataClassname ++ "." ++ dynBufferName ++ ".reset();"]
createBuffer PrecomputedFixedArray accesses csDS readMode useMaxBuffer =
    unlines $
    ["  if (" ++ dataClassname ++ "." ++ precomputedBufferName ++ " == null) {"
    ,"    " ++ dataClassname ++ "." ++ precomputedBufferName ++
          " = new double[" ++ bufferSizeCode ++ "];"
    ,"  }"
    ,"  " ++ bufferName ++ " = " ++ dataClassname ++ "." ++
          precomputedBufferName ++ ";"
    ]
    where
    getTileSizes :: [[Integer]] -> [Integer]
    getTileSizes latVect =
        trace "WARNING: getTileSizes assumes that the tiling is rectangular" $
        [(latVect !! i) !! i
         | i <- [0..(length latVect - 1)]
        ]

    tileSizes = getTileSizes (rationalToIntVector $ csLatVects csDS)
    paramNames = csBlobNames csDS

    timeTileSize   = tileSizes !! 0
    spaceTileSize1 = tileSizes !! 1
    spaceTileSize2 = tileSizes !! 2

    mult3Str = if readMode then "3 * " else ""

    -- severe HACK! (works only for matmult example!!!)
    -- (in the future: use Barvinok)
    bufferSizeCode = 
        if useMaxBuffer 
        then error "not used anymore: evalMaxBufferSize accesses csDS readMode"
        else case length tileSizes of
             1 -> error "createBuffer: need space tiling!"
             2 -> mult3Str ++ "(1 + Math.min(" ++ show spaceTileSize1 ++ " * tiledim1 + " ++ 
                  show (spaceTileSize1-1) ++ ", " ++ head paramNames ++" - 1) - Math.max(" ++
                       show spaceTileSize1 ++ " * tiledim1, 0))*(" ++ head paramNames ++ ")" 
             3 -> if allEq tileSizes
                  then (mult3Str ++ show spaceTileSize1 ++ " * " ++ show spaceTileSize1)
                  else error "createBuffer: only cubic tile-shapes supported so far!"
             _ -> error "createBuffer: case fell through: only 3-dim tiling supported so far!"

-- |
-- Generates java code to extract the internal buffer.
-- It depends on the buffer management strategy whether
-- explicit copy code is needed.
--
copyBufferCode :: BufferManagementStrategy
               -> String
copyBufferCode DynamicBuffer =
    "  this." ++ bufferName ++ " = " ++
              dataClassname ++ "." ++ dynBufferName ++ ".getArray();"
copyBufferCode PrecomputedFixedArray = ""

-- |
-- Warning: May return 0 if the tiling has not full dimension.
--
{-- MC not used anymore:
evalMaxBufferSize :: [Access] -- ^ all accesses
                  -> CompSpreadDS
                  -> Bool     -- ^ true: read accesses, false: write accesses
                  -> String
evalMaxBufferSize accs csDS isRead = 
    verbose "evalMaxBufferSize" $
    sumExpression
    where
    theAccType = if isRead then Read else Write

    latVects = csLatVects csDS
    paramNames = csBlobNames csDS

    relevantAccesses = 
	filter (\a -> accType a == theAccType) accs
   
    sumExpression = concat $ intersperse " + " cExpParts
    cExpParts = concatMap (\s -> 
                           map (\acc -> evalIntAccessInTile csDS (getTrStmtID s, 
                                                                  getTrafo (getTrStmtID s),
                                                                  acc)) 
                           (getAccsFromTISPC s)
                          ) (csTISPCs csDS)
                
    srcStmtID :: TStatement -> StatementID
    srcStmtID (_,s,_) = s

    getTrStmtID :: TStatement -> TStatementID
    getTrStmtID (t,_,_) = t

    getTrafo :: TStatementID -> AffineFunction
    getTrafo trStmtID = fromJust $ lookup trStmtID (csTrafos csDS)

    getAccsFromTISPC :: TStatement -> [Access]
    getAccsFromTISPC trStmt = 
        filter (\a -> accStmtID a == (srcStmtID trStmt)) relevantAccesses

--}

-- |
-- Calculates integral points in a given tile under the image of the 
-- access function of an access and returns a valid
-- C/C++/Java expression which evaluates the conditions and
-- computes the value of the respective Ehrhart polynomial.
-- 
{-- MC not used anymore:
evalIntAccessInTile :: CompSpreadDS 
                    -> (TStatementID, AffineFunction, Access)
                    -> String
evalIntAccessInTile csDS (trStmtID,trafo,acc) = 

-- WARNING: incorrect code! acc-function not used!
--          also: may only work for uniform dependences,
--                if acc-function is used only on 1 Tile

    verbose "evalIntAccessInTile" $
    ("(" ++ cExpression ++ ")")
    where
    cevs = countIntegralPoints finalIS Nothing
    cExpression = evaluesWithCondsToCExpression paramNames cevs
    infos = [SrcTile, SrcCoord]
    (tilingIS, _) = defaultTile latVects (csTiledDims csDS) (selfDepRelation (dummyID, dummyIS)) infos

    pD = getParamsDim csDS
    latVects = csLatVects csDS
    paramNames = csBlobNames csDS

    -- dummyIS: unbounded index space, so that each tile contains the full set of iterations
    dummyIS = ineqSystem ((tispcD,pD), zeroMatrix 0 (tispcD + pD + 1))
    dummyID = integerToID 0
    tispcD = getCoordsDim csDS trStmtID
    tileD = genericLength $ csTiledDims csDS
    isD = nVars trafo
    
    (acc_v, acc_p, acc_c) = partsVPCih (accIndexFct acc) -- partsVPCih?
    
    extAccFct :: AffineFunction
    extAccFct = 
        affineFunction 
        ((tispcD + isD, pD),
         myMultijoin1 
         [ acc_v `conc` zeroMatrix (nRows acc_v) isD `conc` acc_p `conc` acc_c
         , zeroMatrix isD tispcD `conc` unitMatrix isD isD `conc` zeroMatrix isD (pD + 1)
         ]
        ) Inhomogeneous

    finalIS = 
        projectToVarsMC accPoly [0 .. (targetDimIH extAccFct)-1]

    accPoly = image extAccFct substituted_M

    substituted_M = 
        -- verbose ("substitutedIS" ++ show (nColumns tile_v)) $
        substituteTileVars (ineqSystem ((tileD+tispcD+isD, 
                                         pD), final_M)) tileD
    final_M = 
        verbose ("final_M" ++ show (nColumns tile_v)) $
        myMultijoin1 $
        [(tile_v `conc` zeroMatrix (nRows tile_v) isD `conc` 
          tile_p `conc` tile_c)
        , tr_M
        ]
          
    tr_M = (trafoIS_M `join` (-trafoIS_M))
    trafoIS_M = 
        -- verbose ("trafoIS_M: " ++ show (nRows $ getMatrixIH trafo) ++ ", " ++ show tispcD) $
        ((zeroMatrix tispcD tileD) `conc` 
         (-(unitMatrix tispcD tispcD)) `conc` getMatrixIH trafo)

    -- tileZeroIS = substituteTileVars tilingIS (genericLength $ csTiledDims csDS)
    (tile_v, tile_p, tile_c) = partsVPC tilingIS

    value = 0%1

    substituteTileVars :: IneqSystem -> Integer -> IneqSystem
    substituteTileVars is 0 = is
    substituteTileVars is tileVarsLeft = substituteTileVars (substVar is 0 value) (tileVarsLeft-1)
        
--}
           
-- |
-- Adds store instructions to the communication loops and indents the loops.
-- Also adds int declaration to the loop variables.
--
addStoreInstructions :: BufferManagementStrategy
                     -> [(String, [String])] -- ^ array with its dimension
                     -> String -- ^ the communication loops (for read\/write accesses)
                     -> String
addStoreInstructions strategy arrayVars code =
    unlines (["{",
              "  int " ++ counterName ++ " = 0;"] ++
             rep 0 (lines code) ++
             ["}"])
    where
    rep :: Integer -> [String] -> [String]
    rep _ [] = []
    rep n (nextLine : remLines)
        | "for (" `isPrefixOf` nextLine =
            (indent n ("for (int " ++ (drop (length "for (") nextLine))) :
            rep (n+1) remLines
        | "ArrayID " `isPrefixOf` nextLine =
            let posStr = drop (length "ArrayID ") nextLine
                pos    = ((read posStr) :: Integer) - 1
                (name, dims) = arrayVars !! (fromIntegral pos)
                dim = genericLength dims
                indexCode = concat ["[i" ++ show i ++ "]" | i <- [0..n-1]]
            in case () of
                 _ | pos < 0 || pos >= genericLength arrayVars ->
                      error "addStoreInstructions: input array array out of bounds"
                 _ | dim /= n ->
                       error "addStoreInstructions: array dimension mismatch"
                 _ -> let addToBufferCode =
                              genAddToBufferStmt strategy (name ++ indexCode)
                          replacedLine =
                              indent n (" { " ++ addToBufferCode ++ " } ")
                      in replacedLine : rep 0 remLines 
        | otherwise = error "addStoreInstructions: Should not be reached (1)"
    indent n str = (genericReplicate (2*(n+1)) ' ') ++ str

-- |
-- Adds load instructions to the communication loops and indents the loops.
-- Also adds int declaration to the loop variables.
--
addLoadInstructions :: [(String, [String])] -- ^ array with its dimension
                    -> String -- ^ the communication loops (for read\/write accesses)
                    -> String
addLoadInstructions arrayVars code =
    unlines (["{",
              "  int " ++ counterName ++ " = 0;"] ++
             rep 0 (lines code) ++
             ["}"])
    where
    rep :: Integer -> [String] -> [String]
    rep _ [] = []
    rep n (nextLine : remLines)
        | "for (" `isPrefixOf` nextLine =
            (indent n ("for (int " ++ (drop (length "for (") nextLine))) :
            rep (n+1) remLines
        | "ArrayID " `isPrefixOf` nextLine =
            let posStr = drop (length "ArrayID ") nextLine
                pos    = ((read posStr) :: Integer) - 1
                (name, dims) = arrayVars !! (fromIntegral pos)
                dim = genericLength dims
                indexCode = concat ["[i" ++ show i ++ "]" | i <- [0..n-1]]
            in case () of
                 _ | pos < 0 || pos >= genericLength arrayVars ->
                      error "addLoadInstructions: input array array out of bounds"
                 _ | dim /= n ->
                      error "addLoadInstructions: array dimension mismatch"
                 _ -> let loadFromBufferCode =
                              genLoadFromBufferStmt (name ++ indexCode)
                          replacedLine = indent n loadFromBufferCode
                      in replacedLine : rep 0 remLines 
        | otherwise = error "addLoadInstructions: Should not be reached (1)"
    indent n str = (genericReplicate (2*(n+1)) ' ') ++ str

-- |
-- Returns Java code for allocating arrays.
--
allocArraysCode :: [(String, [String])] -> String
allocArraysCode arrays = 
    unlines
    ([""
     ,"  // array allocation:"
     ] ++
     arrayAllocations ++
     [""
     ]
    )
    where
    arrayAllocations = map (("  "++).oneArrayAllocLine) arrays

-- |
-- Generates java code, which writes the given value
-- into the internal buffer.
--
genAddToBufferStmt :: BufferManagementStrategy -- ^ the strategy
                   -> String                   -- ^ the value
                   -> String
genAddToBufferStmt DynamicBuffer value =
    dataClassname ++ "." ++ dynBufferName ++ ".add(" ++ value ++ ");"
genAddToBufferStmt PrecomputedFixedArray value =
    -- buffer[count] = A[i0][i1]; count++;
    bufferName ++ "[" ++ counterName ++ "] = " ++ 
    value ++ "; " ++ counterName ++ "++;"

-- |
-- Generates java code, which loads next value from the buffer
-- and write it into the given lvalue (i.e., the target memory cell).
--
genLoadFromBufferStmt :: String -- ^ the lvalue
                      -> String
genLoadFromBufferStmt targetCell =
    -- A[i0][i1] = buffer[count++];
    targetCell ++ " = " ++ bufferName ++ "[" ++ counterName ++ "++];"

-- |
-- Returns one line of array allocation in Java syntax.
-- The array sizes are given for each dimension as a list of identifiers.
--
oneArrayAllocLine :: (String, [String]) -> String
oneArrayAllocLine (arrName, arrSizes) = 
    arrName ++ " = new double" ++ getArraySizes arrSizes ++ ";"

-- |
-- Returns the expression for the dimensions and sizes of an array allocation. 
--
getArraySizes :: [String] -> String
getArraySizes arrSizes = concatMap (\d -> "[" ++ d ++ "]") arrSizes

-- |
-- Returns the expression for allocating a new local array or data structure.
getNewArray :: String   -- ^ the internal data type
            -> StorageManagementStrategy -- ^ e.g. local arrays, hash table, etc.
            -> [String] -- ^ the sizes of the array\/data structure in each dimension
            -> String
getNewArray typeStr LocalArray arrSizes = "new " ++ typeStr ++ getArraySizes arrSizes
getNewArray typeStr LocalDataStructure arrSizes = 
    ("new " ++ typeStr ++ "Array" ++ (show $ length arrSizes) ++ "D(" ++ 
     (concat $ intersperse ", " arrSizes) ++ ")")
getNewArray typeStr EhrhartBuffer _ = "new " ++ typeStr ++ getArraySizes ["0"]

-- |
-- Allocate temporary memory to execute a task on the local worker.
--
-- Supported HOCs: Master\/Worker
--
allocArraysExecuteCode :: LoopData -> String
allocArraysExecuteCode ld =
    unlines
    (["  // allocate local arrays:"
     ] ++
     decls1 ++
     [""
     ,"  if (" ++ dataClassname ++ "." ++ (fst3 $ head arrays) ++ 
                    " == null)"
     ,"  {"
     ,"     // initialize memory pool"
     ] ++ allocateArrays ++
     ["  }"
     , ""
     ] ++ renameArrays
    )
    where
    storageStrategy = ldStorageManagementStrategy ld
    arrayPrefix = getArrayPrefix storageStrategy

    fst3 (x,_,_) = x
    arrays = ldArrayVariables ld
    decls1 = map (\(n,d,_) ->
                   ("  " ++ javaArrayType "double" storageStrategy (genericLength d) ++
                   " " ++ arrayPrefix ++ n ++ ";")
                 ) arrays
    allocateArrays = map (\(n,d,_) ->
                          ("       " ++ dataClassname ++ "." ++ arrayPrefix ++ n ++ 
                           " = " ++ getNewArray "double" storageStrategy d ++ ";")
                         ) arrays
    renameArrays = map (\(n,_,_) -> "  " ++ arrayPrefix ++ n ++ " = " ++ dataClassname ++
                                    "." ++ arrayPrefix ++ n ++ ";") arrays

-- |
-- Sorts the arrays from the ".arrays" file according
-- to the arrays from the symbol table.
--
matchArraysWithLoopoInternals
    :: Show b
    => [(String, b, ArrayType)]       -- ^ arrays from the ".arrays" file
    -> [String]                       -- ^ arrays from the symbol table
    -> [(String, b, ArrayAccessType)] -- ^ sorted arrays
matchArraysWithLoopoInternals arrays order
    | length arrays /= length order =
        error ("matchArraysWithLoopoInternals: array size mismatch\n" ++
               "LooPo internal data and the .arrays file do not match.")
    | otherwise = map (convertArrayType.match) order
    where
    match next = safeHead (filter (\(x,_,_) -> x == next) arrays)
    safeHead (x:_) = x
    safeHead [] =
        error ("matchArraysWithLoopoInternals: Matching error!" ++
               "\narrays from the .arrays file: " ++ show arrays ++
               "\narrays from the symbol table: " ++ show order
              )
    convertArrayType (x, y, ArrayIN)    = (x, y, ReadOnlyAccess)
    convertArrayType (x, y, ArrayOUT)   = (x, y, WriteOnlyAccess)
    convertArrayType (x, y, ArrayINOUT) = (x, y, ReadWriteAccess)

genEhrhartAccessCode :: CompSpreadDS -> [(ArrayID, String, [String])] -> String
genEhrhartAccessCode csDS ehrhartData = ehrhartFunctions
    where
    ehrhartFunctions = concat $ map oneEhrhartFunction ehrhartData

    oneEhrhartFunction :: (ArrayID, String, [String]) -> String
    oneEhrhartFunction (arrayID, ehrhartAccessCode, ehrhartParams) =
        unlines functionLines
        where
        arrayNumber = show $ integerFromID arrayID
        arguments = concat $ intersperse ", " $ map ("int "++) ehrhartParams
        functionLines = 
            ["private int " ++ ehrhartAccessFuncName ++ arrayNumber ++ 
             "(" ++ arguments ++ ") {"
            ,""
            ,"  // ehrhart access code:"
            ,"  " ++ "return " ++ ehrhartAccessCode ++ ";"
            ,"}"
            ,""
            ]

genEhrhartArrayAllocCode :: CompSpreadDS -> [(ArrayID, String)] -> String
genEhrhartArrayAllocCode csDS ehrhartData = ehrhartFunctionCode
    where
    ehrhartFunctionCode = concat $ map oneEhrhartFunction ehrhartData
    tempDir = csTempDir csDS
    projFile = csProjFile csDS
    (symtab,_,_,_) = unsafePerformIO $ openParseFile  (tempDir ++ "/" ++ projFile)

    oneEhrhartFunction :: (ArrayID, String) -> String
    oneEhrhartFunction (arrayID, ehrhartArrayAllocCode) =
        unlines functionLines
        where
        arrayName = 
            verbose "getting array name" $
            getArrayPrefix EhrhartBuffer ++ arrayLexem symtab arrayID
        functionLines = 
            ["// allocate array " ++ arrayName ++ ":"
            , arrayName ++ " = new double[" ++ ehrhartArrayAllocCode ++ "];"
            ]


--
-- Generation of the example files:
--
masterFileTemplate :: String
masterFileTemplate = genMasterJava loopDataTemplate

taskFileTemplate :: String
taskFileTemplate = genTaskJava loopDataTemplate

dataFileTemplate :: String
dataFileTemplate = genDataJava loopDataTemplate

p2pMasterTemplate :: String
p2pMasterTemplate = genP2PMasterJava p2pLoopDataTemplate

p2pClientTemplate :: String
p2pClientTemplate = genP2PClientJava p2pLoopDataTemplate

p2pDataPacketTemplate :: String
p2pDataPacketTemplate = genP2PDataPacketJava p2pLoopDataTemplate

p2pDynamicArrayTemplate :: String
p2pDynamicArrayTemplate = genP2PDynamicArrayClassJava p2pLoopDataTemplate

loopDataTemplate :: LoopData
loopDataTemplate =
    verifyLoopData $
    LoopData
    { ldCompTaskCode                 = "// TODO: task's computation code"
    , ldArrayReadAccesses_StoreCode  = "// TODO: array read accesses store code"
    , ldArrayReadAccesses_LoadCode   = "// TODO: array read accesses load code"
    , ldArrayWriteAccesses_StoreCode = "// TODO: array write accesses store code"
    , ldArrayWriteAccesses_LoadCode  = "// TODO: array write accesses load code"
    , ldP2P_Communication_writeBuffer =
       "// TODO: write into buffer code (P2P communication)"
    , ldP2P_Communication_unpackBuffer =
       "// TODO: unpack buffer code (P2P communication)"
    , ldP2P_Soak_writeBuffer =
       "// TODO: write into buffer code (soak communication)"
    , ldP2P_Soak_unpackBuffer =
       "// TODO: unpack buffer code (soak communication)"
    , ldP2P_Drain_writeBuffer =
       "// TODO: write into buffer code (drain communication)"
    , ldP2P_Drain_unpackBuffer =
       "// TODO: unpack buffer code (drain communication)"
    , ldP2P_MasterSoak =
       "// TODO: master soak communication"
    , ldP2P_MasterDrain =
       "// TODO: master drain communication"
    , ldAddAllTasksCode              = "// TODO: add all tasks code"
    , ldCreateDependenceGraphCode    =
        ["// TODO: create dependence graph code (dependence 1)"
        ,"// TODO: create dependence graph code (dependence 2)"]
    , ldInitDeclarations             =
        unlines ["  double[][] A; // user defined",
                 "  double[] B;   // user defined"]
    , ldIndexVariables               = ["k"]
    , ldOuterTilingVariables         = ["outer_tile_1", "outer_tile_2"]
    , ldInnerTilingVariables         = ["inner_tile_1", "inner_tile_2"]
    , ldTaskVariables                = ["task_var1", "task_var2"]
    , ldMasterVariables              = ["master_var1", "master_var2"]
    , ldCreateDependenceGraphVariables = ["create_graph_1", "create_graph_2"]
    , ldExecuteTaskVariables         = ["execute_helper1", "execute_helper2"]
    , ldWriteBufferVariables         = ["write_buffer_1", "write_buffer_2"]
    , ldUnpackBufferVariables        = ["unpack_buffer_1", "unpack_buffer_2"]
    , ldParameters                   = ["n", "m"]
    , ldArrayVariables               = [("A", ["n","n"], ReadOnlyAccess),
                                        ("B", ["n","n","n"], ReadWriteAccess),
                                        ("C", ["m","n","n"], ReadWriteAccess)]
    , ldRawTilingVector              = [[1,2,3], [4,5,6], [7,8,9]]
    , ldPackageName                  = "userapplication.matmult"
    , ldAccesses                     = error "uninitialized data"
    , ldPreloopCode                  = "// TODO: User preloop code"
    , ldPostloopCode                 = "// TODO: User postloop code"
    , ldSerialVersionUID             = 42
    , ldSupportOldFramework          = False
    , ldHOC_Type                     = defaultHOC_Type
    , ldBufferManagementStrategy     = defaultBufferManagementStrategy
    , ldStorageManagementStrategy    = defaultStorageManagementStrategy
    , ldEhrhartAccessCode            = 
        "// TODO: ehrhartAccess(accIDs, taskIDs.., params...) {...}"
    , ldEhrhartArrayAllocCode        = 
        "// TODO: ehrhartA = new double[ehrhart polynom...]"
    }

p2pLoopDataTemplate :: LoopData
p2pLoopDataTemplate = loopDataTemplate { ldHOC_Type = P2P_HOC }

-- |
-- Contains all command line options.
--
data TaskLoopsCommandLine = TaskLoopsCommandLine
    { tlclInitDeclsFilename     :: String
    , tlclPreloopFilename       :: String
    , tlclPostloopFilename      :: String
    , tlclPackageName           :: String
    , tlclCompSpreadCommandLine :: CommandLine
    , tlclSupportOldFramework   :: Bool
    , tlclNumSoakStmts          :: Integer
    , tlclNumDrainStmts         :: Integer
    , tlclHocType               :: HOC_Type
    , tlclBufferStrategy        :: BufferManagementStrategy
    , tlclStorageStrategy       :: StorageManagementStrategy
    } deriving Show

data TaskLoopsFlag = InitDeclsFilename String
                   | PreloopFilename String | PostloopFilename String
                   | PackageName String | SupportOldFramework
                   | NumSoakStmts String | NumDrainStmts String
                   | HOC_TypeOption String | BufferStrategyOption String
                   | StorageStrategyOption String
                   | HelpCommand
                     deriving Show
    
taskLoopsOptions :: [OptDescr TaskLoopsFlag]
taskLoopsOptions =
     [ Option []    ["init-decls"] (ReqArg InitDeclsFilename "FILE")
                          "set the name of the .init_decls file"
     , Option []    ["pre-file"]  (ReqArg PreloopFilename "FILE")
                          "set the name of the .preloop file"
     , Option []    ["post-file"] (ReqArg PostloopFilename "FILE")
                          "set the name of the .postloop file"
     , Option []    ["package-name"] (ReqArg PackageName "NAME")
                          ("set the name of the package (default: " ++
                           defaultPackageName ++ ")")
     , Option []    ["support-old-framework"] (NoArg SupportOldFramework)
              ("compiles the code for the old framework, which has " ++
               "a slightly modified interface")
     , Option []    ["num-soak-stmts"] (ReqArg NumSoakStmts "INTEGER")
                           "sets the number of soak statements that were inserted"
     , Option []    ["num-drain-stmts"] (ReqArg NumDrainStmts "INTEGER")
                           "sets the number of drain statements that were inserted"
     , Option []    ["hoc-type"] (ReqArg HOC_TypeOption "TYPE")
              ("produce code for the specified HOC component " ++
               "(master-worker, p2p)")
     , Option []    ["buffer"] (ReqArg BufferStrategyOption "TYPE")
                  "override the type of internal buffer (dynamic, precomputed)"
     , Option []    ["storage"] (ReqArg StorageStrategyOption "TYPE")
                  "override the type of storage strategy (array, hashtable)"
     , Option []    ["help"] (NoArg HelpCommand)
                  "show this help screen and exit"
     ]
    
taskLoopsOpts :: [String] -> IO ([TaskLoopsFlag], [String])
taskLoopsOpts argv = 
    case getOpt Permute taskLoopsOptions argv of
      (o,rem,[]) -> return (o,rem)
      (_,_,errs) -> ioError (userError (concat errs ++ usageInfo header taskLoopsOptions))
      where header = "Usage: taskloops [OPTION...]"

-- |
-- Parses the command line options and verifies that no
-- incompatible options have been chosen.
--
parseTaskLoopsCommandLine :: IO TaskLoopsCommandLine
parseTaskLoopsCommandLine =
    do
    argv        <- getArgs
    (opts, rem) <- taskLoopsOpts argv
    when (any isHelpCommand opts)
             (putStrLn usageInfo' >> exitWith ExitSuccess)
    let compSpreadCL = parseCommandLine rem
        initCL = TaskLoopsCommandLine
                 {tlclInitDeclsFilename     = err "init-decls"
                 ,tlclPreloopFilename       = err "pre-file"
                 ,tlclPostloopFilename      = err "post-file"
                 ,tlclPackageName           = defaultPackageName
                 ,tlclCompSpreadCommandLine = compSpreadCL
                 ,tlclSupportOldFramework   = False
                 ,tlclNumSoakStmts          = err "num-soak-stmts"
                 ,tlclNumDrainStmts         = err "num-drain-stmts"
                 ,tlclHocType               = defaultHOC_Type
                 ,tlclBufferStrategy        = defaultBufferManagementStrategy
                 ,tlclStorageStrategy       = defaultStorageManagementStrategy
                 }
        cl = foldl extract initCL opts
        err s = error
              ("Error while parsing command line options.\n" ++
               "Required option '--" ++ s ++ "' is missing.\n\n" ++ usageInfo')
    checkOptions cl
    where
    usageInfo' = usageInfo ("Usage: taskloops [OPTION...]") taskLoopsOptions
    extract cl (InitDeclsFilename str) = cl { tlclInitDeclsFilename   = str }
    extract cl (PreloopFilename   str) = cl { tlclPreloopFilename     = str }
    extract cl (PostloopFilename  str) = cl { tlclPostloopFilename    = str }
    extract cl (PackageName       str) = cl { tlclPackageName         = str }
    extract cl SupportOldFramework     = cl { tlclSupportOldFramework = True }
    extract cl (NumSoakStmts str)      = cl { tlclNumSoakStmts  = (read str) :: Integer }
    extract cl (NumDrainStmts str)     = cl { tlclNumDrainStmts = (read str) :: Integer }
    extract cl (HOC_TypeOption str) =
        cl {tlclHocType = parseHocType str }
    extract cl (BufferStrategyOption str) =
        cl { tlclBufferStrategy = parseBufferStrategy str }
    extract cl (StorageStrategyOption str) =
        cl { tlclStorageStrategy = parseStorageStrategy str }
    isHelpCommand HelpCommand = True
    isHelpCommand _ = False

    -- parse the command line options, which have an argument
    parseHocType "master-worker" = MasterWorkerHOC
    parseHocType "p2p"           = P2P_HOC
    parseHocType str = error ("Could not parse hoc type: '" ++ str ++ "'")
    parseBufferStrategy "dynamic"     = DynamicBuffer
    parseBufferStrategy "precomputed" = PrecomputedFixedArray
    parseBufferStrategy str =
        error ("Could not parse buffer strategy: '" ++ str ++ "'")
    parseStorageStrategy "array"     = LocalArray
    parseStorageStrategy "hashtable" = LocalDataStructure
    parseStorageStrategy str =
        error ("Could not parse storage strategy: '" ++ str ++ "'")

    -- check whether all command line options are compatible
    checkOptions :: TaskLoopsCommandLine -> IO TaskLoopsCommandLine
    checkOptions cl
        | tlclSupportOldFramework cl && tlclHocType cl /= MasterWorkerHOC =
            do
            putStrLn ("Error: Option 'support-old-framework' is incompatible" ++
                     " with the selected HOC type.")
            exitFailure
        | otherwise = return cl

main :: IO ()
main = do
       tlCommandLine <- parseTaskLoopsCommandLine
       serialVersionUID <- genSerialVersionUID -- using the current time

       let cl = tlclCompSpreadCommandLine tlCommandLine
           basicDS = unsafeCreateCompSpreadDS cl
           csDS = basicDS { csNumSoakStmts = numSoakStmts
                          , csNumDrainStmts = numDrainStmts
                          }
           loopoFile = csTempDir csDS ++ "/" ++ csProjFile csDS                      
	   lds = unsafeReadLoopoDS loopoFile
	   accs = loopoAccesses lds    
                   -- (ccAccesses cc) seems to have only such from accpairs!
                   -- (but in this case we need all accesses, because we
                   --  are not working with dependence driven communication)
           trimmedAccs = map (trimAccess (loopoSLP lds)) accs

           addTaskInput = verbose "obtaining createTaskInput" $ createAddTaskInput csDS
           addTaskCodeFile = (loopoFile ++ "." ++ addTaskOutExt ++ ".sed")
           addTaskRun = RunExternal { runInputExtension = "." ++ addTaskInpExt
                                    , runOutputExtension = "." ++ addTaskOutExt
                                    }
           compTaskInput = createCompTaskInput csDS
           compTaskCodeFile = (loopoFile ++ "." ++ compTaskOutExt ++ ".sed")
           compTaskRun = 
               RunExternal { runInputExtension = "." ++ compTaskInpExt
                           , runOutputExtension = "." ++ compTaskOutExt
                           }

           ehrhartZPs =        
               verbose "ehrhartAccessZPs in TaskLoops.hs" $
               ehrhartAccessZPs csDS (csAccesses csDS) ExecutionBuffer
           
           ehrhartParams arrayID countOnly = 
               (accNames ++ 
                taskIDNames ++
                csBlobNames csDS)
               where
               taskIDNames =  
                   [taskIDName ++ 
                    (if countOnly 
                     then "[" ++ show i ++ "]"
                     else show i) | i <- [0..getTilingDim csDS-1]
                   ]
               accDim = getArrayDim csDS arrayID
               accNames = 
                   -- if only used for counting the number of integral points,
                   -- the access variables are no parameters, so they are empty
                   if countOnly 
                   then [] 
                   else genericTake accDim (bufferScatterNames csDS)
                        
           -- the actual accesses in the one-dimensional buffer, expressed as
           -- ehrhart polynomials:
           ehrhartAccesses :: [(ArrayID, String)]
           ehrhartAccesses = 
               verbose "ehrhartAccesses" $
               trace ("length(ehrhartZPs) = " ++ show (length ehrhartZPs)) $
               map (\(arrayID, zDomain) ->
                     trace ("zDomain: (length= " ++ show (length (zdDisjointUnion zDomain)) ++ ")") $
                     -- trace (show (zdDisjointUnion zDomain)) $
                     (arrayID, evaluesWithCondsToCExpression (ehrhartParams arrayID False)
                                 (-- optimizeEPartition $ 
                                  zdLexicographicIndex True Nothing 
                                      (zdDisjointUnion zDomain)))
                                      -- zDomain))
                   ) ehrhartZPs

           -- the basic C-expressions for the size of each array:
           ehrhartArrayAllocExpressions :: [(ArrayID, String)]
           ehrhartArrayAllocExpressions = 
               verbose "ehrhartArrayAllocExpressions" $
               map (\(arrayID, zDomain) ->
                     (arrayID, getOneArrayAllocExpression zDomain 
                                 (ehrhartParams arrayID True))
                   ) ehrhartZPs
           
           getOneArrayAllocExpression :: ZDomain -> [String] -> String
           getOneArrayAllocExpression zD ehrhartPs =
               -- count all zpolytopes in the zdomain,
               -- generate a C-expression for each,
               -- then sum them up.
               concat $
               intersperse " + " $ 
                 map (\zP -> 
                          -- trace (show zP) $
                          "(" ++                             
                          evaluesWithCondsToCExpression ehrhartPs
                             (zpCountIntegralPointsDisj Nothing zP) ++
                          ")") (-- trace "zdDisjointUnion..." $
                                zdDisjointUnion zD)
                                --zD)

           -- just for convenience: a bundle of the arrayIDs, ehrhart access code
           -- and correspoinding ehrhart parameters
           ehrhartAccData = 
               verbose "ehrhartAccData" $
               map (\(arrayID, str) -> 
                     -- trace str $
                     (arrayID, str, ehrhartParams arrayID False)) 
                   ehrhartAccesses
           
           -- the final ehrhart access code for all arrays:
           ehrhartAccessCode = 
               verbose "ehrhartAccessCode" $
               genEhrhartAccessCode csDS ehrhartAccData

           -- the final ehrhart array allocation code for all arrays:
           ehrhartArrayAllocCode = 
               genEhrhartArrayAllocCode csDS ehrhartArrayAllocExpressions

           dependenceInputs = 
               verbose "createDependenceInput" $
               createDependenceInput csDS hocType
           dependenceGraphCodeFile = (loopoFile ++ "." ++ dependenceOutExt ++ ".sed")
           dependenceRun = 
               RunExternal { runInputExtension = "." ++ dependenceInpExt
                           , runOutputExtension = "." ++ dependenceOutExt
                           }

           p2pCommInputs commType = genP2PCommInputs csDS commType
           p2pCommCodeFile = (loopoFile ++ "." ++ p2pCommOutExt ++ ".sed")
           p2pCommRun = 
               RunExternal { runInputExtension = "." ++ p2pCommInpExt
                           , runOutputExtension = "." ++ p2pCommOutExt
                           }

           initDeclsFilename   = tlclInitDeclsFilename   tlCommandLine
           preloopFilename     = tlclPreloopFilename     tlCommandLine
           postloopFilename    = tlclPostloopFilename    tlCommandLine
           packageName         = tlclPackageName         tlCommandLine
           numSoakStmts        = tlclNumSoakStmts        tlCommandLine
           numDrainStmts       = tlclNumDrainStmts       tlCommandLine
           supportOldFramework = tlclSupportOldFramework tlCommandLine
           hocType             = tlclHocType             tlCommandLine
           bufferStrategy      = tlclBufferStrategy      tlCommandLine
           storageStrategy     = tlclStorageStrategy     tlCommandLine

           -- Extension: now, we also pass the dependence to use its type
           --   for later optimizations at runtime
           getDependenceCode :: (TrDependence, String) -> IO String
           getDependenceCode (trDep, depInpDesc) =
               do writeFile (loopoFile ++ "." ++ dependenceInpExt) depInpDesc
                  putStrLn "running CLooG for dependence loops..."
                  runCloog csDS dependenceRun
                  putStrLn "dependence loops CLooG run done."
                  insertDependenceCall csDS hocType trDep
                  strictReadFile dependenceGraphCodeFile

           getP2PCommCode :: (CompSpreadDS 
                              -> Bool 
                              -> Comm_Type
                              -> MasterClientCode
                              -> StorageManagementStrategy 
                              -> [(ArrayID, String)]
                              -> TrDependenceID                                  
                              -> IO ()) 
                          -> Bool
                          -> Comm_Type
                          -> MasterClientCode
                          -> StorageManagementStrategy 
                          -> (TrDependenceID, String) 
                          -> IO String
           getP2PCommCode handleP2Pcode isWriteBuffer commType masterClientCode storageStrategy 
                          (trDepID, p2pCommInpDesc) =
             do writeFile (loopoFile ++ "." ++ p2pCommInpExt) p2pCommInpDesc
                putStrLn "running CLooG for p2pComm loops..."
                runCloog csDS p2pCommRun
                putStrLn "p2pComm loops CLooG run done."
                putStrLn "handleP2Pcode..."
                handleP2Pcode csDS isWriteBuffer commType 
                              masterClientCode storageStrategy ehrhartAccesses trDepID 
                putStrLn "handleP2Pcode done."
                strictReadFile p2pCommCodeFile

       putStrLn ("reading .init_decls file '" ++ initDeclsFilename ++ "'")
       initDecls <- strictReadFile initDeclsFilename
       putStrLn ("reading .preloop file '" ++ preloopFilename ++ "'")
       preloopCode <- strictReadFile preloopFilename
       putStrLn ("reading .postloop file '" ++ postloopFilename ++ "'")
       postloopCode <- strictReadFile postloopFilename

       putStrLn "writing definesFiles..."

       -- Master/Worker HOC:
       when (hocType == MasterWorkerHOC)
          (writeTaskLoopsDefinesFiles csDS storageStrategy >>
           writeMasterLoopsDefinesFiles csDS storageStrategy )

       -- P2P HOC:
       when (hocType == P2P_HOC)
          (writeP2PClientDefinesFiles csDS storageStrategy)
--        when (storageStrategy == EhrhartBuffer)
--           (writeP2PClientEhrhartDefinesFiles csDS ehrhartAccesses)

       putStrLn "definesFiles written."

       putStrLn ("addTaskInput: " ++ addTaskInput)
       writeFile (loopoFile ++ "." ++ addTaskInpExt) addTaskInput
       putStrLn "running CLooG for addTask loops..."
       runCloog csDS addTaskRun 
       putStrLn "addTask loops CLooG run done."
       insertAddTaskCall csDS hocType

       writeFile (loopoFile ++ "." ++ compTaskInpExt) compTaskInput
       putStrLn "running CLooG for compTask loops..."
       runCloog csDS compTaskRun
       putStrLn "compTask loops CLooG run done."
       substituteCompTaskStmts csDS

       addTaskCode  <- strictReadFile addTaskCodeFile
       compTaskCode <- strictReadFile compTaskCodeFile
       -- dependence graph code is a list of code-strings,
       -- one for each dependence:
       putStrLn "Generating code to build the task dependence graph..."
       dependenceGraphCode <- mapM getDependenceCode dependenceInputs
       putStrLn "Generating code for task dependence graph done."

       p2pCommWriteBufferCode <-
           if hocType == P2P_HOC
           then do
                putStrLn "Generating code for peer-to-peer writeBuffer..."
                result <- mapM (getP2PCommCode handleBufferCode True Comm_P2P ClientCode storageStrategy) 
                          (p2pCommInputs Comm_P2P)
                return (Just result)
           else return Nothing

       p2pCommUnpackBufferCode <-
           if hocType == P2P_HOC
           then do
                putStrLn "Generating code for peer-to-peer unpackBuffer..."
                result <- mapM (getP2PCommCode handleBufferCode False Comm_P2P ClientCode storageStrategy) 
                          (p2pCommInputs Comm_P2P)
                return (Just result)
           else return Nothing

       p2pSoakWriteBufferCode <-
           if hocType == P2P_HOC
           then do
                putStrLn "Generating code for soaking/writeBuffer..."
                result <- mapM (getP2PCommCode handleBufferCode True Comm_Soak ClientCode storageStrategy)
                          (p2pCommInputs Comm_Soak)
                return (Just result)
           else return Nothing
       p2pSoakUnpackBufferCode <-
           if hocType == P2P_HOC
           then do
                putStrLn "Generating code for soaking/unpackBuffer..."
                result <- mapM (getP2PCommCode handleBufferCode False Comm_Soak ClientCode storageStrategy)
                          (p2pCommInputs Comm_Soak)
                return (Just result)
           else return Nothing
       p2pDrainWriteBufferCode <-
           if hocType == P2P_HOC
           then do
                putStrLn "Generating code for draining/writeBuffer..."
                result <- mapM (getP2PCommCode handleBufferCode True Comm_Drain ClientCode storageStrategy)
                          (p2pCommInputs Comm_Drain)
                return (Just result)
           else return Nothing
       p2pDrainUnpackBufferCode <-
           if hocType == P2P_HOC
           then do
                putStrLn "Generating code for draining/unpackBuffer..."
                result <- mapM (getP2PCommCode handleBufferCode False Comm_Drain ClientCode storageStrategy)
                          (p2pCommInputs Comm_Drain)
                return (Just result)
           else return Nothing
       p2pDrainUnpackBufferCodeMaster <-
           if hocType == P2P_HOC
           then do
                putStrLn "Generating code for draining/unpackBuffer for Master..."
                result <- mapM (getP2PCommCode handleBufferCode False Comm_Drain MasterCode storageStrategy)
                          (p2pCommInputs Comm_Drain)
                return (Just result)
           else return Nothing
       p2pSoakWriteBufferCodeMaster <-
           if hocType == P2P_HOC
           then do
                putStrLn "Generating code for soaking/writeBuffer for Master..."
                result <- mapM (getP2PCommCode handleBufferCode True Comm_Soak MasterCode storageStrategy)
                          (p2pCommInputs Comm_Soak)
                return (Just result)
           else return Nothing

       let arrayVariables = matchArraysWithLoopoInternals
                            (loopoArrays lds) (csArrayNames csDS)

           outerTilingVariables  = outerScatterNames csDS
           innerTilingVariables  = innerScatterNames csDS
           parameterNames        = csBlobNames csDS
           writeUnpackBufferVars = bufferScatterNames csDS ++
                                   bufferIteratorNames csDS
             
           readAccesses = eliminateVarMinMax $
                          concat $ intersperse "\n" $
                          taskDataLoop (csTempDir csDS) (csProjFile csDS)
                          intLatVects tiledDims tileUsedVects Read
           intLatVects = rationalToIntVector $ csLatVects csDS
           tiledDims = csTiledDims csDS
           tileUsedVects = csTilingUsedVect csDS
           writeAccesses = eliminateVarMinMax $
                           concat $ intersperse "\n" $
                           taskDataLoop (csTempDir csDS) (csProjFile csDS)
                           intLatVects tiledDims tileUsedVects Write
           declTiledimVarsCode local =
               unlines
               (["  // initialize outer tiling variables:"] ++
                ["  final int tiledim" ++ show i ++ " = " ++
                 (if local
                  then "this."
                  else "((" ++ taskClassname ++ ") task).") ++
                    taskIDName ++ "[" ++ show i ++ "]; // alias for " ++ tn
                 | (tn, i) <- zip outerTilingVariables [0..]] ++
                [""]
               )
           declVarsAndParamsCodeTask accessname =
               declTiledimVarsCode True  ++ declParamsCode accessname
           declVarsAndParamsCodeMaster accessname =
               declTiledimVarsCode False ++ declParamsCode accessname
           declParamsCode accessname = unlines
                            (["  // initialize parameters:"] ++
                             ["  final int n" ++ show i ++ " = " ++
                              accessname ++ "." ++ pn ++ ";"
                                  | (pn, i) <- zip parameterNames [0..]] ++
                             [""]
                            )

           (readMode, writeMode) = (True, False)
           useMaxBuffer = True

           -- Master/worker communication code:
           arrayVariables' = [(name, dims) | (name, dims, _) <- arrayVariables]
           readStoreCode  = "  // (readStoreCode)\n" ++
                            declVarsAndParamsCodeTask "this" ++ 
                            createBuffer bufferStrategy trimmedAccs csDS readMode useMaxBuffer ++
                            addStoreInstructions bufferStrategy arrayVariables' readAccesses ++
                            copyBufferCode bufferStrategy
           readLoadCode   = "  // (readLoadCode)\n" ++
                            declVarsAndParamsCodeTask "this" ++
                            addLoadInstructions arrayVariables' readAccesses
           writeStoreCode = "  // (writeStoreCode)\n" ++
                            createBuffer bufferStrategy trimmedAccs csDS writeMode useMaxBuffer ++
                            addStoreInstructions bufferStrategy arrayVariables' writeAccesses ++
                            copyBufferCode bufferStrategy
           writeLoadCode  = "  // (writeLoadCode)\n" ++
                            declVarsAndParamsCodeMaster masterClassname ++
                            addLoadInstructions arrayVariables' writeAccesses

           -- P2P-HOC communication code:
           unlines' (Just xs) = concat (intersperse "\n" xs)
           unlines' Nothing   =
              error "unlines': Internal error (variable not available for the current HOC-type)"
           writeBufferCode  = unlines' p2pCommWriteBufferCode
           unpackBufferCode = unlines' p2pCommUnpackBufferCode
           writeSoakCode    = unlines' p2pSoakWriteBufferCode
           unpackSoakCode   = unlines' p2pSoakUnpackBufferCode
           writeDrainCode   = unlines' p2pDrainWriteBufferCode
           unpackDrainCode  = unlines' p2pDrainUnpackBufferCode
           masterWriteSoakCode = unlines' p2pSoakWriteBufferCodeMaster
           masterUnpackDrainCode = unlines' p2pDrainUnpackBufferCodeMaster

           loopData =
               verifyLoopData $
               LoopData
               { ldCompTaskCode                   = compTaskCode
               , ldArrayReadAccesses_StoreCode    = readStoreCode
               , ldArrayReadAccesses_LoadCode     = readLoadCode
               , ldArrayWriteAccesses_StoreCode   = writeStoreCode
               , ldArrayWriteAccesses_LoadCode    = writeLoadCode
               , ldP2P_Communication_writeBuffer  = writeBufferCode
               , ldP2P_Communication_unpackBuffer = unpackBufferCode
               , ldP2P_Soak_writeBuffer           = writeSoakCode
               , ldP2P_Soak_unpackBuffer          = unpackSoakCode
               , ldP2P_Drain_writeBuffer          = writeDrainCode
               , ldP2P_Drain_unpackBuffer         = unpackDrainCode
               , ldP2P_MasterSoak                 = masterWriteSoakCode
               , ldP2P_MasterDrain                = masterUnpackDrainCode
               , ldAddAllTasksCode                = addTaskCode
               , ldCreateDependenceGraphCode      = dependenceGraphCode
               , ldInitDeclarations               = initDecls
               , ldIndexVariables                 = csIndexNames csDS
               , ldOuterTilingVariables           = outerTilingVariables
               , ldInnerTilingVariables           = innerTilingVariables
               , ldTaskVariables                  = (outerScatterNames csDS ++
                                                     innerComputeIteratorNames csDS ++
                                                     innerComputeScatterNames csDS)
               , ldMasterVariables                = (outerIteratorNames csDS ++
                                                     outerScatterNames csDS ++
                                                     innerIteratorNames csDS ++
                                                     innerScatterNames csDS)
               , ldCreateDependenceGraphVariables = (outerIteratorNames csDS ++
                                                     outerScatterNames csDS ++
                                                     innerIteratorNames csDS ++
                                                     innerScatterNames csDS)
               , ldExecuteTaskVariables           = (innerComputeScatterNames csDS ++
                                                     innerComputeIteratorNames csDS)
               , ldWriteBufferVariables           = writeUnpackBufferVars
               , ldUnpackBufferVariables          = writeUnpackBufferVars
               , ldParameters                     = parameterNames
               , ldArrayVariables                 = arrayVariables
               , ldRawTilingVector                = intLatVects
               , ldPackageName                    = packageName
               , ldAccesses                       = trimmedAccs
               , ldPreloopCode                    = preloopCode
               , ldPostloopCode                   = postloopCode
               , ldSerialVersionUID               = serialVersionUID
               , ldSupportOldFramework            = supportOldFramework
               , ldHOC_Type                       = hocType
               , ldBufferManagementStrategy       = bufferStrategy
               , ldStorageManagementStrategy      = storageStrategy
               , ldEhrhartAccessCode              = ehrhartAccessCode
               , ldEhrhartArrayAllocCode          = ehrhartArrayAllocCode
               }

       case hocType of
         MasterWorkerHOC ->
             do
             putStrLn "*****************************************"
             putStrLn "Generating code for the master/worker-HOC"
             putStrLn "*****************************************"
             writeFile (csTempDir csDS ++ "/" ++ masterClassname ++ ".pre")
                       (genMasterJava loopData)
             writeFile (csTempDir csDS ++ "/" ++ taskClassname ++ ".pre")
                       (genTaskJava loopData)
             writeFile (csTempDir csDS ++ "/" ++ dataClassname ++ ".java")
                       (genDataJava loopData)
             runCPP csDS taskClassname
             runCPP csDS masterClassname
             return ()
         P2P_HOC ->
             do
             putStrLn writeBufferCode
             putStrLn unpackBufferCode
             putStrLn writeSoakCode
             putStrLn unpackSoakCode
             putStrLn writeDrainCode
             putStrLn unpackDrainCode

             putStrLn "*****************************************"
             putStrLn "Generating code for the peer-to-peer-HOC:"
             putStrLn "*****************************************"

             putStrLn "Generating master class..."
             writeFile (csTempDir csDS ++ "/" ++ p2pMasterClassname ++ ".java")
                       (genP2PMasterJava loopData)
             putStrLn "Generating client class..."
             writeFile (csTempDir csDS ++ "/" ++ p2pClientClassname ++ ".pre")
                       (genP2PClientJava loopData)
             putStrLn "Generating data packet class..."
             writeFile (csTempDir csDS ++ "/" ++ p2pDataPacketClassname ++ ".java")
                       (genP2PDataPacketJava loopData)
             putStrLn "Generating dynamic array helper class..."
             writeFile (csTempDir csDS ++ "/" ++ dynArrayClassname ++ ".java")
                       (genP2PDynamicArrayClassJava loopData)
             runCPP csDS p2pClientClassname         
             -- MC: HACK:
--              clientFile <- strictReadFile (csTempDir csDS ++ "/" ++ p2pClientClassname ++ ".java")
--              writeFile (csTempDir csDS ++ "/" ++ p2pClientClassname ++ ".pre") 
--                            ("#include \"" ++ p2pClientEhrhartDefFile ++ "\"" ++ clientFile)
--              putStrLn "HACK! running cpp again..."
--              runCPP csDS p2pClientClassname         
             return ()

rationalToIntVector :: [Vector] -> [[Integer]]
rationalToIntVector latVect =
    if null assertions 
    then intVect 
    else error assertions
    where
    ratVect = map vectorEntries latVect
    allInteger = all (==1) $ concatMap (map denominator) ratVect
    assertions = if allInteger
                 then ""
                 else "rationalToIntVector: non-integral entry!"
    intVect = map (map numerator) ratVect

insertAddTaskCall :: CompSpreadDS
                  -> HOC_Type
                  -> IO ()
insertAddTaskCall csDS hocType =
    do
    let inpFile = 
            csTempDir csDS ++ "/" ++ csProjFile csDS ++ "." ++ addTaskOutExt
        outFile = inpFile ++ ".sed"
        addTaskCall = genAddSingleTaskCode hocType indices
        indices = outerScatterNames csDS
        sedCommand =
            ("sed -r " ++
             "-e 's/^[[:space:]]*S([0-9]+).*;/\\n" ++ addTaskCall ++ "/' " ++
            inpFile ++ " > " ++ outFile)
    exitCode <- do --putStrLn ("running sed: " ++ show sedCommand)
                   system sedCommand
    let (ok, errCode) = checkResult exitCode
    if ok then do --putStrLn "sed completed."
                  return ()
          else do putStrLn ("\n\n\nsed exited with error code: " ++ 
			    (show errCode) ++ "\n\n\n")   

-- |
-- Inserts the call to the addDependence method in the generated
-- loop nest that enumerates all dependences for generating the task graph.
-- Note: the relevant target dependence is passed along to determine the
--   type of the dependence added to the task graph.
insertDependenceCall :: CompSpreadDS
                     -> HOC_Type
                     -> TrDependence
                     -> IO ()
insertDependenceCall csDS hocType trDep =
    do
    let inpFile = 
            csTempDir csDS ++ "/" ++ csProjFile csDS ++ "." ++ dependenceOutExt
        outFile = inpFile ++ ".sed"
        dependenceCall = genAddDependenceCode hocType srcIndices destIndices trDep
        srcIndices = outerScatterNames csDS
        destIndices = innerScatterNames csDS
        sedCommand =
            ("sed -r " ++
             "-e 's/^[[:space:]]*S([0-9]+).*;/\\n" ++ dependenceCall ++ "/' " ++
            inpFile ++ " > " ++ outFile)
    exitCode <- do --putStrLn ("running sed: " ++ show sedCommand)
                   system sedCommand
    let (ok, errCode) = checkResult exitCode
    if ok then do --putStrLn "sed completed."
                  return ()
          else do putStrLn ("\n\n\nsed exited with error code: " ++ 
			    (show errCode) ++ "\n\n\n")   

handleBufferCode :: CompSpreadDS
                 -> Bool
                 -> Comm_Type
                 -> MasterClientCode
                 -> StorageManagementStrategy
                 -> [(ArrayID, String)]
                 -> TrDependenceID
                 -> IO ()
handleBufferCode csDS isWriteMode commType masterClientCode 
                 storageStrategy ehrhartAccs trDepID =
    do
    putStrLn "handleBufferCode starting..."
    let inpFile = 
            csTempDir csDS ++ "/" ++ csProjFile csDS ++ "." ++ p2pCommOutExt
        outFile = inpFile ++ ".sed"
        arrayName = getArrayName csDS trDepID
        arrayID = accArrayID $ getSrcAcc csDS trDepID

        maybeEhrhartAccessCode =
            if storageStrategy == EhrhartBuffer 
            then (Just ehrhartAcc)
            else Nothing

        maybeEhrhartAccessParams =
            if storageStrategy == EhrhartBuffer 
            then (Just ehrhartParams)
            else Nothing

        ehrhartParams = (accNames ++ 
                         taskParams ++
                         csBlobNames csDS)

        taskParams = [correspondingTaskName ++ "[" ++ show i ++ "]"| 
                      i <- [0..getTilingDim csDS-1]]
        correspondingTaskName = 
            if commType /= Comm_P2P
            -- use taskID for soak/drain:
            then taskIDName
            else if isWriteMode 
                 -- use sourceTaskID for writeBuffer:
                 then sourceTaskIDName
                 -- use targetTaskID for unpackBuffer:
                 else targetTaskIDName

        ehrhartAcc = 
            -- verbose ("looking up ehrhartAccess #" ++ show arrayID ++
            --          " in: " ++ (show $ ehrhartAccs)) $
            fromJust $ lookup arrayID ehrhartAccs

        bufferCode = 
            trace ("bufferCode:\n" ++ 
                   show (genBufferCode (arrayName, arrayID) accNames isWriteMode 
                                       masterClientCode storageStrategy 
                                       maybeEhrhartAccessParams)) $
            genBufferCode (arrayName, arrayID) accNames isWriteMode masterClientCode 
                          storageStrategy 
                          maybeEhrhartAccessParams
        accNames = genericTake accDim (bufferScatterNames csDS)
        accDim = targetDimIH $ accIndexFct $ getSrcAcc csDS trDepID
        sedCommand =
            ("sed -r " ++
             "-e 's/^[[:space:]]*S([0-9]+).*;/\\n" ++ bufferCode ++ "/' " ++
            inpFile ++ " > " ++ outFile)
    exitCode <- do --putStrLn ("running sed: " ++ show sedCommand)
                   system sedCommand
    let (ok, errCode) = checkResult exitCode
    if ok then do --putStrLn "sed completed."
                  return ()
          else do putStrLn ("\n\n\nsed exited with error code: " ++ 
			    (show errCode) ++ "\n\n\n")   


substituteCompTaskStmts :: CompSpreadDS -> IO ()
substituteCompTaskStmts csDS =
    do
    let inpFile = csTempDir csDS ++ "/" ++ csProjFile csDS ++ "." ++ 
                  compTaskOutExt
        outFile = inpFile ++ ".sed"
        sedCommand =
            ("sed -r " ++
             "-e 's/^[[:space:]]*S([0-9]+)(.+;)/" ++ compStmt_t ++ "\\1" ++
             "\\2/' " ++
            inpFile ++ " > " ++ outFile)
    exitCode <- do --putStrLn ("running sed: " ++ show sedCommand)
                   system sedCommand
    let (ok, errCode) = checkResult exitCode
    if ok then do --putStrLn "sed completed."
                  return ()
          else do putStrLn ("\n\n\nsed exited with error code: " ++ 
			    (show errCode) ++ "\n\n\n")   
    return ()


writeTaskLoopsDefinesFiles :: CompSpreadDS -> StorageManagementStrategy -> IO ()
writeTaskLoopsDefinesFiles csDS strategy =
    do origStmtsCode <- printOrigStatementsCode (getTargetLanguage csDS strategy) csDS
       let defFile = csTempDir csDS ++ "/" ++ taskDefFile
           defineHeader = "/* created: " ++ getTime ++ " */\n"
           numSoaks = csNumSoakStmts csDS
           numDrains = csNumDrainStmts csDS
           compStmtsDefines = 
               stagedCompStmtsDefines csDS origStmtsCode           
       writeFile defFile (defineHeader ++ compStmtsDefines)
       return ()

writeMasterLoopsDefinesFiles :: CompSpreadDS -> StorageManagementStrategy -> IO ()
writeMasterLoopsDefinesFiles csDS strategy =
    do origStmtsCode <- printOrigStatementsCode (getTargetLanguage csDS strategy) csDS
       let defFile = csTempDir csDS ++ "/" ++ masterDefFile
           defineHeader = "/* created: " ++ getTime ++ " */\n"
           compStmtsDefines = stagedCompStmtsDefines csDS origStmtsCode
           
       writeFile defFile (defineHeader ++ compStmtsDefines)
       return ()

-- VORSICHT:
--     Ich hab nur die Funktion writeTaskLoopsDefinesFiles kopiert
--     und 'taskDefFile' durch 'p2pClientDefFile' ersetzt.
writeP2PClientDefinesFiles :: CompSpreadDS 
                           -> StorageManagementStrategy 
                           -> IO ()
writeP2PClientDefinesFiles csDS strategy =
    verbose "writeP2PClientDefinesFiles" $
    do origStmtsCode <- printOrigStatementsCode (getTargetLanguage csDS strategy) csDS
       let defFile = csTempDir csDS ++ "/" ++ p2pClientDefFile
           defineHeader = "/* created: " ++ getTime ++ " */\n"
           compStmtsDefines = stagedCompStmtsDefines csDS origStmtsCode
           
       putStrLn ("\n\ndefineHeader: " ++ defineHeader)
       putStrLn ("compStmtsDefines: " ++ compStmtsDefines ++ "\n\n")
       writeFile defFile (defineHeader ++ compStmtsDefines)
       putStrLn (defFile ++ " written.\n")
       return ()

-- MC 2008-09-18: wohl nicht mehr benutzt:
writeP2PClientEhrhartDefinesFiles :: CompSpreadDS 
                           -> [(ArrayID, String)]
                           -> IO ()
writeP2PClientEhrhartDefinesFiles csDS ehrhartStrings =
    verbose "writeP2PClientEhrhartDefinesFiles" $
    do let defFile = csTempDir csDS ++ "/" ++ p2pClientEhrhartDefFile
           defineHeader = "/* created: " ++ getTime ++ " */\n"
           ehrhartDefines = 
               unlines $
               map ehrhartDefine ehrhartStrings           
       putStrLn ("ehrhartDefines: " ++ ehrhartDefines ++ "\n")
       writeFile defFile (defineHeader ++ ehrhartDefines)
       putStrLn (defFile ++ " written.\n")
       return ()
    where
      ehrhartDefine :: (ArrayID, String) -> String
      ehrhartDefine (arrID, ehrhartCode) = 
          ("#define " ++ ehrhartAccessMacro ++ (show $ integerFromID arrID) ++ 
           " " ++ ehrhartAccessFuncName ++ (show $ integerFromID arrID) ++ 
           "(" ++ arguments ++ ")")
          where
          arguments = 
              concat $ intersperse ", " $ ehrhartParams
          ehrhartParams = 
              (accNames ++ 
               [taskIDName ++ "[" ++ show i ++ "]"| 
                i <- [0..getTilingDim csDS-1]] ++ 
               csBlobNames csDS)
          accDim = getArrayDim csDS arrID
          accNames = genericTake accDim (bufferScatterNames csDS)

{-- MC 2008-09-18: momentan nicht als macro realisiert, daher veraltet:
          ("#define " ++ ehrhartAccessMacro ++ (show $ integerFromID arrID) ++ 
           "(" ++ (concat $ intersperse ", " macroParams) ++ ") " ++ 
           setAccs ++
           ehrhartCode) 
          where
          accDim = getArrayDim csDS arrID
          accNames = genericTake accDim (bufferScatterNames csDS)
          macroParams = ["x" ++ show i | i <- [1..accDim]]
          setAccs = 
              concat $ 
              -- acc1 = (x1); acc2 = (x2);
              zipWith (\acc -> \x -> acc ++ " = (" ++ x ++ "); ") accNames macroParams
--}

stagedCompStmtsDefines :: CompSpreadDS 
                       -> [(StatementID, String)] 
                       -> String
stagedCompStmtsDefines csDS origStatementsCode = 
    verbose "stagedCompStmtsDefines" $
    unlines compStmtsDefines
    where
    allTrStmtIDs = getAllTStatementIDs csDS
    (soakIDs, drainIDs) = getSoakDrainOrigIDs csDS
    compStmtIDs = filter (\trStmtID-> 
                          not $ (getStmtID csDS trStmtID) `elem` 
                                  (soakIDs ++ drainIDs)
                         ) allTrStmtIDs
    compStmtsDefines = map oneCompStmtDefine (zip compStmtIDs [1..])

    oneCompStmtDefine :: (TStatementID, Integer) -> String
    oneCompStmtDefine (trStmtID, cloogID) = 
        concat 
	[ compDefineHeader csDS trStmtID cloogID compStmt_t
	, compStmtIndexMapping csDS trStmtID (innerComputeIteratorNames csDS)
	, fromJust $ lookup (getStmtID csDS trStmtID) origStatementsCode
	]


-- soak and drain statements not contained in the generated code, but
-- in original code, so statementIDs have to be adjusted
compDefineHeader :: CompSpreadDS -> TStatementID -> Integer -> String -> String
compDefineHeader csDS trStmtID cloogID stmtName =
    ("#define " ++ stmtName ++ (show $ cloogID) ++ "(" ++
     iteratorNames ++ ") ")
    where
    iteratorNames = concat $ intersperse ", " $ relevantIteratorNames
    allIterators = innerComputeIteratorNames csDS
    isDim = getISPCdim csDS trStmtID
    coordsDim = getCoordsDim csDS trStmtID
    -- tileDim = genericLength $ csTiledDims csDS
    relevantIteratorNames = genericTake (isDim + coordsDim) allIterators

runCPP :: CompSpreadDS
       -> String -- ^ class filename
       -> IO ()
runCPP csDS filename =
    do
    let inpFile = (csTempDir csDS ++ "/" ++ filename ++ ".pre")
        outFile = (csTempDir csDS ++ "/" ++ filename ++ ".java")
        cppCall = ("cpp -P -CC " ++ inpFile ++ " "++ outFile)
    exitCode <- do --putStrLn ("running C preprocessor: " ++ show cppCall)
                   system cppCall
    let (ok, errCode) = checkResult exitCode
    if ok then do --putStrLn "CPP completed."
                  return ()
          else do putStrLn ("\n\n\nCPP exited with error code: " ++ 
			    (show errCode) ++ "\n\n\n")   
    return ()

-- |
-- Generates an int value that can be used for the
-- serialVersionUID attribute in the generated java classes.
--
genSerialVersionUID :: IO Int
genSerialVersionUID = do
                      (TOD secs picos) <- getClockTime
                      return $ fromIntegral (secs + picos)

-- |
-- Read the given file, but in contrast to 'readFile',
-- read it immediately and not lazy.
--
strictReadFile :: FilePath  -- ^ the file
               -> IO String
strictReadFile f = do
                   result <- readFile f
                   when (result /= result) exitFailure
                   seq result $ return result

-- |
-- Tests if the LoopData object is consistent.
--
verifyLoopData :: LoopData -> LoopData
verifyLoopData ld
  | otherwise = ld -- all tests passed
  where
  err str = error ("verifyLoopData: " ++ str)
