-- |
-- Functions for tiling.
--
-- Author Armin Groesslinger
--
-- version: $Id: Tiling.hs 4177 2009-07-29 11:00:32Z classenm $
--
module Tiling
    ( TileInformation(..),
      tile, tileP, tileIneqs,
      optimalPartitioning, possiblePartitionings,
      defaultTile, defaultTiledLoops,
      tileIndexSpace, tileCommPartners,
      test_l
    )
where

import LMath
import CodeGen

import Control.Exception ( assert )
import Debug.Trace ( trace )
import List
import Ratio

import LoopoDS

-- Call `possiblePartitionings' and return a partitioning with
-- a minimal number of communication partners.
-- (sec. 4, step 2.,3.)
--
optimalPartitioning :: [Vector] -> ([Vector], [(Vector,Vector)], Integer)
optimalPartitioning vs =
    case possiblePartitionings vs of
	     []    -> ([], [], 0)
	     parts -> head $ parts

-- Calculate all possible partitioning of `vs' in tile spaning
-- and remaining vectors (paired with their coordinates in the
-- coordinate system of the tile spaning vectors), and the
-- number of communication partners caused by this partitioning.
-- The result is sorted by the number of communication partners
-- (in ascending order).
--
possiblePartitionings :: [Vector] -> [([Vector], [(Vector,Vector)], Integer)]
possiblePartitionings [] = []
possiblePartitionings vs
    | or [vectorLength v /= nDims | v <- vs] =
	error ("possiblePartitingings: all vectors must have the same " ++
	       "number\n  of components")
    | otherwise =
	sortBy (\(_,_,p1) (_,_,p2) -> p1 `compare`p2) $
	map countCommPartners linIndeps
    where
    nDims     = vectorLength (head vs)
    linIndeps = findLinIndeps nDims vs ([],zeroMatrix 0 nDims, [])


-- Calculate the number of different communication partners caused
-- by the vectors in `us' when the vectors in `bs' are taken as
-- tile spaning vectors.
-- Return the tile spaning vectors `bs', the vectors `us' paired
-- with their coordinates in the basis `bs' and the number
-- of communication partners (after subtracting 1 for the tile with
-- neighbour coordinates (EQ,...,EQ)). 
-- (sec. 4, steps 1. (a),(b))
--
countCommPartners :: ([Vector], [Vector]) ->
		     ([Vector], [(Vector,Vector)], Integer)
countCommPartners (bs, us) = (bs, zip us ls, genericLength parts - 1)
    where
    -- gaussJordan is only defined if length us > 0 !
    Just (coords, _) = gaussJordan (columnVectorsToMatrix1 bs)
		                   (columnVectorsToMatrix1 us)
    unit = unitMatrix (genericLength bs) (genericLength bs)
    ls :: [Vector]
    ls = if null us then columnVectors unit
	            else columnVectors (coords `conc` unit)
    -- calculate the neightbour coordinates of all communication partner
    -- caused by each of the dependences (concatenate them) and delete
    -- duplicate entries. 
    parts = nub (concat $ map partners ls)


-- Calculate all communication partners for the (short) dependence
-- which has coordinates `linComb' in the coordinate system
-- defined by the tile spaning vectors.
-- Note that [EQ,...,EQ] is always an element of the result list!
-- (sec. 4)
--
partners :: Vector -> [[Ordering]] 
partners linComb = prod [partnersPerDim x | x <- vectorEntries linComb]
    where
    -- return the communication partners in the current dimension
    partnersPerDim x
	| x < 0  =  [LT,EQ]
	| x > 0  =  [EQ,GT]
	| x == 0 =  [EQ]
    -- calculate the neighbour coordinates of all communication 
    -- partners by "multiplying" the individual dimensions, e.g.
    -- prod [[EQ,GT],[EQ],[LT,EQ]] =
    --               [[EQ,EQ,LT],[EQ,EQ,EQ],[GT,EQ,LT],[GT,EQ,EQ]]
    prod :: [[a]] -> [[a]]
    prod []          = [[]]
    prod ([]:ys)     = []
    prod ((x:xs):ys) = [x:zs | zs <- prod ys] ++ prod (xs:ys)
    

-- Partition `vs' in `n' linear indepedent vectors and the
-- remaining vectors; return all such possible partitionings.
-- The accumulator `(bs,m1,us)' contains:
--   `bs' - the linear independent vectors selected so far
--   `m1' - the result of echelon reduction applied to the
--          vectors in `bs' (used as row vectors)
--   `us' - the vectors selected for the "remaining" group so far
-- (sec. 4, step 1.)
--
findLinIndeps :: Integer -> [Vector] -> ([Vector], Matrix, [Vector]) ->
		 [([Vector], [Vector])]
findLinIndeps 0 vs (bs, _, us) = [(bs, us++vs)]
findLinIndeps _ [] _         = []
findLinIndeps n (v:vs) (bs, m1, us)
    | genericLength (v:vs) < n  =   []
    | otherwise =
	(if all (==0) (last (rows' e1))
	    then []                      -- `v' not lin. indep. from `bs'
	    else findLinIndeps (n-1) vs (bs', e1, us)
	) ++ findLinIndeps n vs (bs, m1, v:us)
    where
    bs'    = bs ++ [v]
    (_,e1) = echelonReduction (m1 `join` rowVectorsToMatrix1 [v])



-- Calculate a tile from the lists of short and long dependences.
-- (sec. 4,7.1)
--
optimalTile :: [Vector] -> [Vector] -> IneqSystem
optimalTile shortDeps longDeps = verticesToIneqs vertices
    where
    (basis, optFactors) = optimalBasisAndFactors shortDeps longDeps
    optBasis   = zipWith mulVS basis optFactors
    vertices   = map (foldl addVV zeroVect) (subsets optBasis)
    zeroVect   = zeroVector (genericLength basis)

-- Calculate an optimal basis and optimal scaling factors
-- for that basis from the short and long dependences.
-- (sec. 7.1)
--
optimalBasisAndFactors :: [Vector] -> [Vector] -> ([Vector], [Rational])
optimalBasisAndFactors shortDeps longDeps = (basis, optFactors)
    where
    (basis, _, _) = optimalPartitioning shortDeps
    longCoords    = coordinates basis longDeps 
    optFactors
	| null longDeps = replicate (length basis) 1
	| otherwise     = optimalForm longCoords

-- Take a (non-empty) list of long dependences (as coordinates in the basis)
-- and return a scaling factor for each dimension.
-- In every dimension the returned scaling factor is the least factor
-- with the highest frequency.
--
optimalForm :: [Vector] -> [Rational]
optimalForm coordinates =
    [head factors | factors <- optimalFormFactors coordinates]

-- Take a (non-empty) list of long dependences (as coordinates in the basis!)
-- and return for each dimension the list of scaling factors which
-- occur most often (in the respective dimension).
-- (sec. 7.1)
--
optimalFormFactors :: [Vector] -> [[Rational]]
optimalFormFactors coordinates =
    [snd $ head $ scalingFactorsForDim lambdas | lambdas <- coordsPerDims]
    where
    coordsPerDims = transpose [vectorEntries v | v <- coordinates]

-- Calculate scaling factors for a dimension and their frequency.
-- The list is in descending order of frequencies and the scaling factors
-- in each list (of factors with the same frequency) is in ascending order.
-- `lambdas' is the list of coordinates of the long dependences in
-- the dimension for which the scaling factors are calculated.
-- (sec. 7.1)
--
scalingFactorsForDim :: [Rational] -> [(Integer,[Rational])]
scalingFactorsForDim lambdas
    | null frequenciesAndFactors = [(0,[1 % 1])]
    | otherwise                  = frequenciesAndFactors
    where
    frequenciesAndFactors =
	map (\((frq,f1):ffs) -> (frq, f1:[f | (_,f) <- ffs])) $
	groupBy (\(frq1,_) (frq2,_) -> frq1 == frq2) $
	sortBy (\(frq1,f1) (frq2,f2) -> (frq2,f1) `compare` (frq1,f2)) $
	map (\grp -> (genericLength grp, head grp)) $
	group $ sort $ concat $ map possibleScalingFactors lambdas

-- Return a list of possible scaling factors (in one dimension)
-- resulting from the coordinate `lambda' of a long dependence in
-- the respective dimension (sec. 7.1)
--
possibleScalingFactors :: Rational -> [Rational]
possibleScalingFactors lambda
    | lambda <  0 = [lambda / fromIntegral a | a <- [-1, -2 .. ceiling lambda]]
    | lambda >= 0 = [lambda / fromIntegral a | a <- [ 1,  2 .. floor   lambda]]



-- Calculate the coordinates of the vectors `vs' in the basis `basis'.
--  
coordinates :: [Vector] -> [Vector] -> [Vector]
coordinates _ [] = []
coordinates basis vs
    | length (group (map vectorLength (basis++vs))) /= 1  =
	error "coordinates: all vectors must have the same\n\
	      \  number of components!\n"
    | otherwise  =
	case gauss of
	    Just (coords, _) -> columnVectors coords
	    Nothing -> error "coordinates: `basis' is not a basis for \
			     \the subspace\n  generated by `vs'\n"
    where
    -- gaussJordan is only defined if length vs > 0 !
    gauss = gaussJordan (columnVectorsToMatrix1 basis)
	                (columnVectorsToMatrix1 vs)
    dims = genericLength basis



--------------------------------------------------------
-- Next section: enumerating tiled domains
--------------------------------------------------------

data TileInformation = SrcTile   | DestTile | SrcCoord | DestCoord
		     | SrcOffset | DestOffset
		       deriving (Show, Eq)


-- documentation for the big inequation system below still to come ...
tileIneqs :: Num a =>
	     GIneqSystem a -> [GVector a] -> GAffineFunction a ->  
             (GIneqSystem a,Integer) ->   GIneqSystem a
tileIneqs tileShape latticeVectors offset (dependenceRelation,limitSrc)
    = case failedAssertions of
	 "" ->  --200203 trace ("in tileIneqs'\n" ++ show is) $ 
	       ineqSystem ((4*dimT+dimD+dimS, n_params), is)
	 s  -> error s
    where
    is = --200203 trace (show (zerol dimT tv `conc` zerol (dimS+dimD+2*dimT+n_params)   tc)++ "die lattice-Zeile\n" ++ show(lat  `conc`   e  `conc` (-eSource) `conc` zerol (2*dimT+dimD)   op  `conc`   oc) ++ "\nDie h-Zeile: \n"++show (zerol (2*dimT) hs `conc` zerol (2*dimT) hd `conc` hp `conc` hc)) $
      multijoin1 
     -- columns:    dimT  dimT  simS  dimT  dimT  dimD  n_params  1
     [ zerol dimT tv `conc` zerol (dimS+dimD+2*dimT+n_params)   tc, 

      (-lat) `conc` (-e) `conc` eSource    `conc` zerol (2*dimT+dimD) (-op) `conc` (-oc),
        lat  `conc`   e  `conc` (-eSource) `conc` zerol (2*dimT+dimD)   op  `conc`   oc,

      zerol (3*dimT+dimS) tv `conc` zerol (dimD+n_params) tc,

      zerol (2*dimT+dimS) (-lat) `conc` (-e) `conc` eDest    `conc` (-op) `conc` (-oc),
      zerol (2*dimT+dimS)   lat  `conc`   e  `conc` (-eDest) `conc`   op  `conc`   oc,

      zerol (2*dimT) hs `conc` zerol (2*dimT) hd `conc` hp `conc` hc]


    -- LATTICE

    -- dT ist the dimensionality of the tile space, i.e., the number
    -- of only the commonly surrounding and tiled dimensions
    dimT       = fromIntegral (length latticeVectors)
    assertLat  = case all (==dimT) (map vectorLength latticeVectors) of
		 True -> ""
	         False -> "the lattice vectors do not form a square matrix\n"
    lat        = columnVectorsToMatrix1 latticeVectors 

     -- TILE SHAPE
    (tv,tp,tc) = partsVPC tileShape
    assertTilP = case nColumns tp == 0 of
		 True -> ""
		 False -> "the tile shape contains parameters\n"
    assertTilV = case nColumns tv == dimT of
		 True -> ""
		 False -> "the number of variables of tile shape and lattice differ\n"
    -- the number of rows is arbitrary (e.g., a hexagon as tiles in 2d space)

    -- DEPENDENCE
    -- the dependence is given as dependRelation, where the 
    -- left limitSrc variables correspond to the indices of the 
    -- dependence's source index domain, and the remaining 
    -- variables to the destination's domain.
    -- This form is the result of TrDep.depRelation

    assertDep  = case nRows dependRelationSource == nRows dependRelationDest of
		 True -> ""
		 False -> "the two parts of the dependence relation do not match\n"

    hs' = let m = getMatrix dependRelationSource
          in  subMatrixVer m 0 (limitSrc -1)

    hs         = if limitSrc < dimT
		 then hs' `conc` zeroMatrix (nRows hs') (dimT-limitSrc)
		 else hs'

    (hd',hp,hc) = partsVPC dependRelationDest
    hd = --200203 trace ("dependenceRelation:\n" ++ show dependenceRelation ++ "depRelDest\n" ++ show dependRelationDest ++ "Var-part:\n" ++ show hd') $
	 if nColumns hd' < dimT 
	 then hd' `conc` zeroMatrix (nRows hd') (dimT-(nColumns hd'))
	 else hd'


    -- DEPENDENCE SOURCE DOMAIN -- not given explicitly any more
    dimS       = max dimS' dimT
    dimS'      = limitSrc 

    -- DEPENDENCE DESTINATION DOMAIN -- not given explicitly any more
    dimD       = nColumns hd

    -- GLOBAL TILES' OFFSET
    (ov,op,oc)
	| nParams offset == 0 = --trace ("OV: " ++ show v ++ "\n") 
				(v,zeroMatrix (targetDimIH offset) n_params,c)
	| otherwise           = -- trace ("OV: " ++ show v ++ "\n") 
				(v,p,c)
	where 
	(v,p,c) = partsVPCih offset

    assertOffC = case nColumns ov == dimT of
		 True -> ""
		 False -> "the dimensionality of tile space and offset variables do not match\n"
    assertOffR = case nRows ov == dimT of
		 True -> ""
		 False -> "the dimensionality of tile space and offset dimensions do not match\n"
    assertOffV = case  ov == zeroMatrix dimT dimT of
		 True -> ""
		 False -> "the offset must not contain variables\n"

    -- ALL PARAMETERS
    assertParm = case all (== (nColumns hp)) (map nColumns paramMatrices) of
		 True -> ""
		 False -> "the number of Parameters differ: " ++ 
                          concat 
                           (intersperse ", " (map show paramMatrices)) ++
                          "\n"
               where paramMatrices = [op]
    n_params = nColumns hp

    -- ASSERTS TOGETHER
    failedAssertions = concat [assertLat, assertTilP, assertTilV, assertDep, 
			       assertOffV, assertOffR, 
			       assertOffC, assertParm]

    e          = unitMatrix dimT dimT
    eSource    = e `conc` (zeroMatrix dimT (max(dimS-dimT) 0))
    eDest      = e `conc` (zeroMatrix dimT (max(dimD-dimT) 0))

    zerol cols m = zeroMatrix (nRows m) cols `conc` m
    (dependRelationSource, dependRelationDest) = 
	isplitAt dimS' dependenceRelation



msplitAt :: Integer -> GMatrix a -> (GMatrix a,GMatrix a)
msplitAt i m = (subMatrixVer m 0 (i-1), subMatrixVer m i (nColumns m -1))

isplitAt :: Num a => Integer -> GIneqSystem a -> (GIneqSystem a,GIneqSystem a)
isplitAt i ies = 
    (ineqSystem ((i,0), left), ineqSystem ((nV-i,nP), right))
    where (left',right) = msplitAt i (getMatrix ies)
	  left = left' `conc` zeroMatrix (nRows left') 1
	  nV = nVars ies
	  nP = nParams ies

-- |
-- Calculates inequation systems for tiling.
-- This function takes as input:
-- the shape of the tiles (@tileShape@), 
-- the @lattice@ vectors for enumerating the tiles,
-- the global @offset@, i.e., the position of tile 0 
--    (which can be a constant function or a function in
--     parameters); it usually is the constant zero function,
-- a description of the  @dependence@ relation
--    (details see below), and
-- a list @infos@, which describes what
-- information is \"interesting\" as input or output. The resulting
-- inequation system @tileIS@ has the following structure:
-- the variables are (from left to right) the variables corresponding
-- to the information mentioned in @infos@. I.E.:
-- When @space@ and @tileShape@ are 2-dimensional, then
-- @infos=[SrcTile, SrcOffset, DestTile]@ will prduce
-- an inequation system where the first two variables denote the number
-- of the source tile, the next two variables are the offset within the
-- source tile, and the last two variables denote the number of the
-- destination tile. The result @infoDims@ is a list of dimensions,
-- i.e. it tells how many variables are used for the corresponding information
-- in @infos@. In our case, @infoDims@ would be
-- @[2, 2, 2]@.
--
-- The resulting inequation system is sorted (in the sense that FM or an
-- equivalent algorithm has been applied to it), so it is probably a good
-- idea to put the information you want to use as \"input\" in
-- @infos@ first, followed by the information you need as
-- \"output\"; that will enable you to easily generate the desired loop
-- nest from the resulting inequation system.
--
tile :: IneqSystem              -- ^ the shape of the tiles
     -> [Vector]                -- ^ the tile enumeration vectors
     -> [Integer]               -- ^ the list of tiled dimensions in this order
     -> AffineFunction          -- ^ the offset of the base tile
     -> (IneqSystem,Integer)    -- ^ the dependence relation
     -> [TileInformation]       -- ^ the tile information wanted
     -> (IneqSystem, [Integer]) -- ^ the inequality system which describes the
				--   the information requested in @infos@
				--   and the number of variables used for
				--   each entry in @infos@
tile = tileHelper (\is vars -> simplifyTrivially $ proj is vars)
    where
    proj is vars =
	case projectToVars' HsFM is Nothing vars of
	  Nothing  -> infeasibleIneqSystem (length vars) (nParams is)
	  Just sys -> sys


-- |
-- Comput the tiling of an index space using 
-- 'projectToVars' (which calls the C++ side to sort an inequality system).
--
tile_old :: IneqSystem -> [Vector] -> [Integer] -> AffineFunction ->
	    (IneqSystem,Integer) ->  [TileInformation]  -> (IneqSystem, [Integer])
tile_old = 
    trace "warning: old tiling function!" $
    tileHelper (\is vars -> simplifyTrivially $ proj is vars)
    where
    proj is vars =
	case projectToVars is Nothing vars of
	  Nothing  -> infeasibleIneqSystem (length vars) (nParams is)
	  Just sys -> sys


tileP :: Integral a
      => [QfFormula (Polynomial (Ratio a))]
      -> GIneqSystem (FracPoly (Ratio a))
      -> [GVector (FracPoly (Ratio a))]
      -> [Integer]  -- ^ the list of tiled dimensions in this order
      -> GAffineFunction (FracPoly (Ratio a))
      -> (GIneqSystem (FracPoly (Ratio a)), Integer)
      -> [TileInformation]
      -> (DTree (Polynomial (Ratio a))
                (Maybe (SortedSystem (FracPoly (Ratio a))))
         , [Integer])
tileP ctx = tileHelper (\is vars -> projectToVarsPss' sortIneqsPss ctx is Nothing Nothing vars)

-- |
-- Attention: reordering the tile-dims leads to a corresponding reordering
--            of offset-dims, which cannot be undone like it is done in the 
--            case of Src\/Dest-Coords. The reason is: some offset-dims are 
--            lost (not used) in the tiling process.
tileHelper :: Num a
	   => (GIneqSystem a -> [Integer] -> b)
	   -> GIneqSystem a
	   -> [GVector a]
           -> [Integer] -- tiled dims (for projection)
	   -> GAffineFunction a
	   -> (GIneqSystem a, Integer)
	   -> [TileInformation]
	   -> (b, [Integer])
tileHelper projectionFunc tileShape lattice tiledDims offset (dependence,limitS)  infos
    | nub infos /= infos  =
	error ("tile: a \"tile information\" may not appear twice in " ++
	       "`infos'")
    | otherwise = -- trace ("in tile:\n" ++ show (sD, dD,pD)) $
                  -- trace ("dependence: " ++ show dependence) $
                  (tileIS, infoDims)
    where
    nrsOfVars = map (fromIntegral . length . info2indexes)

    infoDims = nrsOfVars infos
    tileIS = projectionFunc is indexes
    --is  = tileIneqs tileShape lattice offset (dependence,limitS) 
    is  = tileIneqs tileShape lattice offset (reorderedDepIS,limitS) 
    tD = nVars tileShape
    sD = limitS
    dD = nVars dependence - limitS
    pD = nParams dependence
    indexes = 
        -- trace ("in tile:\n" ++ show (info2indexes DestCoord, info2indexesDestCoord)) $
        concat $ map info2indexes infos

    reorderedSrcDims = tiledDims ++ ([0 .. sD-1] \\ tiledDims)
    reorderedDestDims = tiledDims ++ ([0 .. dD-1] \\ tiledDims)
                       
    reorderedDepIS = 
        -- trace "reorderedDepIS:" $
        -- trace (show $ "reorderedSrcDims: " ++ show reorderedSrcDims ++ 
             --         " reorderedDestDims: " ++ show reorderedDestDims) $
        -- trace (show $ "nVars dependence (src,dest): " ++ (show (sD, dD))) $
        reorderVars dependence (reorderedSrcDims ++ (map (+sD) reorderedDestDims))

    restoreCoords coords permutation = 
        -- trace "restoreCoords" $
        -- trace ("coords: " ++ show coords) $
        -- trace ("permutation: " ++ show permutation) $
        -- trace ("inversePermutation: " ++ (show $ inversePermutation permutation)) $
        [ genericIndex coords i | 
          i <- inversePermutation permutation ]


    info2indexesSrcCoord   = [(2*tD)    .. (2*tD+sD-1)] 
    info2indexesDestCoord  = [(4*tD+sD) .. (4*tD+sD+dD-1)]

    info2indexes SrcTile    = [0         .. (tD-1)]
    info2indexes SrcOffset  = [tD        .. (2*tD-1)]
    info2indexes SrcCoord   = restoreCoords [(2*tD)    .. 
                                             (2*tD+sD-1)] reorderedSrcDims
    info2indexes DestTile   = [(2*tD+sD) .. (3*tD+sD-1)]
    info2indexes DestOffset = [(3*tD+sD) .. (4*tD+sD-1)]
    info2indexes DestCoord  = restoreCoords [(4*tD+sD) .. 
                                             (4*tD+sD+dD-1)] reorderedDestDims



-- call `tile' with ins++outs for `infos' and output loops
-- for the out variables
-- tiledLoops generates a loop nest for the tiled space.
-- However: it drops the <code>in</code> dimensions, 
-- i.e., we obtain a description of the <code>out</code> 
-- dimensions that is parameterized in the <code>in</code> 
-- variables. For the other parameters: see <code>tile</code>.
tiledLoops ::  IneqSystem 
           -> [Vector] 
           -> [Integer]
           -> AffineFunction 
           -> (IneqSystem,Integer)  -> [TileInformation] -> 
	       [TileInformation] -> [String]
tiledLoops  tileShape lattice tiledDims offset dependence ins outs =
    drop (fromIntegral (sum (take (length ins) lens))) $
	 lines (loopNest is)
    where
    (is, lens) = tile tileShape lattice tiledDims offset dependence (ins++outs)
		 

-- example 4.1.2 from Martina Wieninger's diploma thesis
test_shape = ineqSystem ((2,0),matrix [[0,1,0],[0,-1,2],[1,-1,0],[-1,1,2 % 1]])
test_l = [vector [0,3], vector [3,3%1]]
test_offset = constantAffineFunction 2 0 (matrix [[0],[0 % 1]])
test_ispc = ineqSystem ((2,0),matrix [[1,0,0],[0,1,0],[-1,-1,15 % 1]])

-- dependence: <(x,y)> -> <(x,y+1)>
test_htrafo = affineFunction ((2,0),matrix [[1,0,0],[0,1,-1 % 1]]) Inhomogeneous
test_deprel = genericDepRelation  (integerToID 1,test_ispc) test_htrafo

-- "enumerate all destination tiles"
--  ex: tile test_ispc test_shape test_l test_offset test_dep [DestTile]
--    tile test_shape test_l test_offset test_deprel [DestTile]
-- should be
--
-- (Inequality system in 2 variables and 0 parameters
-- Matrix with 5 rows and 3 columns
--  1/1  1/1  1/3
--  0/1  1/1  4/3
--  -1/2  -1/1  5/2
--  1/1  0/1  13/3
--  -1/1  0/1  17/3
-- ,[2])

hauptnenner :: Matrix -> Integer 
hauptnenner m = hn (concat (rows' m)) 
  where
     hn :: [Rational] -> Integer 
     hn xs = foldl lcm 1 (map denominator xs)


--integralMatrix  :: Matrix -> Matrix
--integralMatrix m = smul m (fromInteger (hauptnenner m))


matrixGcd :: Matrix -> Integer
-- m is supposed to be integral!
matrixGcd m = mgcd (concat (rows' m)) 
  where
     mgcd :: [Rational] -> Integer 
     mgcd xs = foldl1 gcd (map numerator xs)


smallestIntegralMatrix :: Matrix -> Matrix
smallestIntegralMatrix m = res
  where
     denom = hauptnenner m
     intM = smul m (fromInteger denom)
     mgcd = matrixGcd intM
     res = smul intM (1 % mgcd)



defaultTileShape ::  [Vector] ->  IneqSystem
defaultTileShape lattice = 
       tileShape
  where
       dT = vectorLength (head lattice)
       rawLatticeMatrix = columnVectorsToMatrix1 lattice
 --                       mtranspose  (matrix (map vectorEntries lattice))
       rawHyp = inverseMatrix rawLatticeMatrix
       denom = hauptnenner rawHyp
       intHyp = smul rawHyp (fromInteger denom)
       smat = join  intHyp (-intHyp)
       svect = [0 | i<-[1..dT]] ++ [ fromInteger denom -1 | i<-[1..dT]]
       tileShape = ineqSystem ((dT,0), conc smat (columnMatrix svect))
       

zeroOffset :: Integer -> Integer -> AffineFunction
zeroOffset tileDim nparams = 
    constantAffineFunction tileDim nparams 
	       	       (matrix [[0::Rational] | i<-[1..tileDim]]) 


defaultTile :: [Vector] 
            -> [Integer]
            -> (IneqSystem,Integer) 
            ->  [TileInformation] 
            -> (IneqSystem, [Integer])
defaultTile lattice tiledDims (dependence,limitSrc)  infos = 
    -- trace "defaultTile..." $
    --     trace ("in DefaultTile:\n" ++ show offset ++ show rawHyp ++ show tileShape) $
    tile tileShape lattice tiledDims offset (dependence,limitSrc)  infos
    where
       offset = zeroOffset tileDim (nParams dependence)
       tileShape = defaultTileShape lattice
       tileDim  = vectorLength (head lattice)

-- defaultDepRel :: AffineFunction ->  (IneqSystem, Integer)
-- defaultDepRel htrafo = 
--     ineqSystem((nv,np), matrix)
--     where
--         depSrcVars = nRows htrafo - nparams -1
-- 	depDestVars = nVars htrafo
--         nrows = max depSrcVars depDestVars
-- 	naprams = nparams htrafo
-- 	matrix = (unitMatrix depSrcVars depSrcVars) `join`
		 

-- -- call `tile' with ins++outs for `infos' and output loops
-- -- for the out variables
-- defaultLoops :: IneqSystem -> [Vector] -> AffineFunction ->
-- 	 [TileInformation] -> [TileInformation] -> [String]
-- defaultLoops ispc lattice dependence ins outs =
--     drop (fromIntegral (sum (take (length ins) lens))) $
--     lines (loopNest is)
--     where
--     (is, lens) = defaultTile ispc lattice dependence (ins++outs)


-- noDep :: Integer -> Integer -> AffineFunction
-- noDep nvar nparam = affineFunction ((nvar,nparam),unitMatrix (nvar+nparam+1) (nvar+nparam+1)) Homogeneous
       

-- defaultTileLoops :: IneqSystem -> [Vector] -> [TileInformation] -> [String]
-- defaultTileLoops ispc lattice outs = defaultLoops ispc lattice (noDep dim nparams) [] outs 
--   where dim = nVars ispc
--         nparams = nParams ispc

defaultTiledLoops :: [Vector] 
                  -> [Integer] -- ^ the tiled dims in this order
	          -> (IneqSystem,Integer)  
                  -> [TileInformation] 
                  -> [TileInformation] 
                  -> [String]
defaultTiledLoops lattice tiledDims dependence ins outs =
    drop (fromIntegral (sum (take (length ins) lens))) $
	 lines (loopNest is)
    where
    (is, lens) = defaultTile lattice tiledDims dependence (ins++outs)


-- Armin's test for parametrized offsets 
-- tile (ineqSystem ((1,2),matrix [[1,0,0,0],[-1,0,0,5%1]]))
--   (ineqSystem ((1,2),matrix [[1,0,0,0],[-1,0,0,0%1]])) (matrix [[1%1]])
--   (affineFunction ((0,2),matrix [[0,1,0%1]]))
--   (affineFunction ((1,2),matrix [[1,0,0,0%1]])) [SrcTile]

--check :: IneqSystem -> [[Rational]] -> [[Rational]]
--check is ds = [d | d <- ds, and (is `apply` (d++[1]))]

-- |
-- Tile a given index space.
-- 
-- Return a system which has the original dimensions first,
-- followed by the tile dimensions, followed by the
-- offsets within the tile.
--
tileIndexSpace :: Num a
               => GIneqSystem a     -- ^ index space to tile
               -> [Integer]         -- ^ dimensions to tile
               -> [GVector a]       -- ^ lattice vectors
               -> GIneqSystem a     -- ^ tile shape
               -> GIneqSystem a
tileIndexSpace ispc tiledDims lattice tileShape =
    assert (length lattice == nV) $
    assert (all (==nV) $ map vectorLength lattice) $
    assert (niV >= nV) $
    assert (nParams ispc == nP) $
    assert (all (==0) $ entries tsP) $
    assert (all (\i -> i>=0 && i<fromIntegral niV) tiledDims) $
    ineqSystem ((niV+2*nV,nP),
                multijoin1 $ map multiconc1
                [[ ispcV , zV , zV , ispcP, ispcC],
                 [-tiledV, lat, uV , zP   , zC   ],
                 [ tiledV,-lat,-uV , zP   , zC   ],
                 [ ziV   , zV2, tsV, tsP  , tsC  ]
                ]
               )
    where
    niV = nVars ispc
    nV  = nVars tileShape
    nP  = nParams tileShape
    (ispcV, ispcP, ispcC) = partsVPC ispc
    (tsV, tsP, tsC) = partsVPC tileShape
    zV  = zeroMatrix (nRows ispc) nV
    zV2 = zeroMatrix (nRows tileShape) nV
    uV  = unitMatrix nV nV
    tiledV = multijoin1 [rows (unitMatrix niV niV) !! fromIntegral x | x <- tiledDims]
    ziV = zeroMatrix (nRows tileShape) niV
    lat = columnVectorsToMatrix1 lattice
    zC  = zeroMatrix nV 1
    zP  = zeroMatrix nV nP

-- |
-- Describe communication partners for a given dependence in
-- two tiled index spaces.
-- The column indices of the requested information
-- is returned. The returned system is not projected to
-- the requested dimensions.
--
tileCommPartners :: Num a
  => GDependence a     -- ^ dependence
  -> GIneqSystem a     -- ^ index space of source statement
  -> [Integer]         -- ^ dimensions of source index space to tile
  -> [GVector a]       -- ^ lattice vectors of source tile shape
  -> GIneqSystem a     -- ^ tile shape for source index space
  -> GIneqSystem a     -- ^ index space of target statement
  -> [Integer]         -- ^ dimensions of target index space to tile
  -> [GVector a]       -- ^ lattice vectors of target tile shape
  -> GIneqSystem a     -- ^ tile shape for target index space
  -> [TileInformation]
  -> (GIneqSystem a, [[Integer]])
tileCommPartners dep srcIspc srcDims srcLatt srcShape dstIspc dstDims dstLatt dstShape infos =
    (multijoin1 [insertZeroVars srcSys nVsrc nVdst
                ,insertZeroVars dstSys 0     nVsrc
                ,depEq
                ,fmap negate depEq
                ]
    , map infoToCols infos
    )
    where
    srcSys = tileIndexSpace (srcIspc `join` depSrcDomain dep)
                            srcDims srcLatt srcShape
    dstSys = tileIndexSpace (dstIspc `join` depDestDomain dep)
                            dstDims dstLatt dstShape

    nVsrc, nVdst, nP, nVcoordSrc, nVcoordDst :: Integer
    nVsrc  = nVars srcSys
    nVdst  = nVars dstSys
    nP     = nParams srcSys
    nVcoordSrc = targetDimIH $ depHTrafo dep
    nVcoordDst = nVars $ depHTrafo dep

    depEq = ineqSystem ((nVsrc+nVdst,nP),
                        multiconc1 [unitMatrix nVcoordSrc nVcoordSrc
                                   ,zeroMatrix nVcoordSrc (nVsrc-nVcoordSrc)
                                   ,getMatrixIH $ depHTrafo dep
                                   ,zeroMatrix nVcoordSrc (nVdst-nVcoordDst+nP+1)
                                   ]
                       )

    infoToCols :: TileInformation -> [Integer]
    infoToCols SrcCoord  = [0..nVcoordSrc-1]
    infoToCols SrcTile   = [nVcoordSrc .. nVcoordSrc+genericLength srcLatt-1]
    infoToCols SrcOffset = [nVcoordSrc+genericLength srcLatt-1 .. nVdst-1]
    infoToCols DestCoord = [nVdst .. nVdst+nVcoordDst-1]
    infoToCols DestTile  = [nVdst+nVcoordDst .. nVdst+nVcoordDst+genericLength dstLatt-1]
    infoToCols DestOffset = [nVdst+nVcoordDst+genericLength dstLatt .. nVsrc+nVdst-1]


rectTileIneqs :: [Integer]
              -> (IneqSystem,Integer) 
              ->  IneqSystem
rectTileIneqs tileSizes (dependenceRelation,limitSrc)
    = case failedAssertions of
	 "" -> trace ("in rectTileIneqs'\n" ++ show is) $ 
	       ineqSystem ((2*dimT+dimD+dimS, pD), is)
	 s  -> error ("error in rectTileIneqs: " ++ s)
    where
    is = --200203 trace (show (zerol dimT tv `conc` zerol (dimS+dimD+2*dimT+pD)   tc)++ "die lattice-Zeile\n" ++ show(lat  `conc`   e  `conc` (-eSource) `conc` zerol (2*dimT+dimD)   op  `conc`   oc) ++ "\nDie h-Zeile: \n"++show (zerol (2*dimT) hs `conc` zerol (2*dimT) hd `conc` hp `conc` hc)) $
      multijoin1 
     -- columns:   dimT  simS     dimT     dimD      pD  1
     --         srcTile  srcCoord destTile destCoord params    1
     [(-lat) `conc` eSource    `conc` zeroMatrix dimT (dimT+dimD+pD+1),
        lat  `conc` (-eSource) `conc` zeroMatrix dimT (dimT+dimD+pD) `conc` tileSizeMinus1,

      zerol (dimT+dimS) (-lat) `conc` eDest    `conc` zeroMatrix dimT (pD+1),
      zerol (dimT+dimS)   lat  `conc` (-eDest) `conc` zeroMatrix dimT pD `conc` tileSizeMinus1,

       -- srcCoords/destCoords:
      zerol dimT hs `conc` zerol dimT hd `conc` hp `conc` hc]


    -- dT ist the dimensionality of the tile space, i.e., the number
    -- of only the commonly surrounding and tiled dimensions
    dimT       = genericLength tileSizes

    lat        = setEntries (zeroMatrix dimT dimT) 
                 [(c,c, (genericIndex tileSizes c)%1)| c <- [0..dimT-1]]
                 
    tileSizeMinus1 = setEntries (zeroMatrix dimT 1) 
                     [(c,0, ((genericIndex tileSizes c)-1)%1)| c <- [0..dimT-1]]

    -- DEPENDENCE
    -- the dependence is given as dependRelation, where the 
    -- left limitSrc variables correspond to the indices of the 
    -- dependence's source index domain, and the remaining 
    -- variables to the destination's domain.
    -- This form is the result of TrDep.depRelation

    assertDep  = case nRows dependRelationSource == nRows dependRelationDest of
		 True -> ""
		 False -> "the two parts of the dependence relation do not match\n"

    hs' = let m = getMatrix dependRelationSource
          in  subMatrixVer m 0 (limitSrc -1)

    hs         = if limitSrc < dimT
		 then hs' `conc` zeroMatrix (nRows hs') (dimT-limitSrc)
		 else hs'

    (hd',hp,hc) = partsVPC dependRelationDest
    hd = --200203 trace ("dependenceRelation:\n" ++ show dependenceRelation ++ "depRelDest\n" ++ show dependRelationDest ++ "Var-part:\n" ++ show hd') $
	 if nColumns hd' < dimT 
	 then hd' `conc` zeroMatrix (nRows hd') (dimT-(nColumns hd'))
	 else hd'


    -- DEPENDENCE SOURCE DOMAIN -- not given explicitly any more
    dimS       = max dimS' dimT
    dimS'      = limitSrc 

    -- DEPENDENCE DESTINATION DOMAIN -- not given explicitly any more
    dimD       = nColumns hd

    -- ALL PARAMETERS
    assertParm = case all (== (nColumns hp)) (map nColumns paramMatrices) of
		 True -> ""
		 False -> "the number of Parameters differ: " ++ 
                          concat 
                           (intersperse ", " (map show paramMatrices)) ++
                          "\n"
               where paramMatrices = [hp]
    pD = nParams dependenceRelation

    -- ASSERTS TOGETHER
    failedAssertions = concat [assertDep
			      , assertParm]

    e          = unitMatrix dimT dimT
    eSource    = e `conc` (zeroMatrix dimT (max(dimS-dimT) 0))
    eDest      = e `conc` (zeroMatrix dimT (max(dimD-dimT) 0))

    zerol cols m = zeroMatrix (nRows m) cols `conc` m

    (dependRelationSource, dependRelationDest) = isplitAt dimS' dependenceRelation


-- rectangular tiling version of the defaultTile function:
rectTile :: [Integer]               -- ^ the tile sizes
         -> (IneqSystem,Integer) -- ^ the dependence relation
         -> [TileInformation]       -- ^ the tile information wanted
         -> (IneqSystem, [Integer]) -- ^ the inequality system which describes the
				    --   the information requested in @infos@
				    --   and the number of variables used for
				    --   each entry in @infos@
rectTile = rectTileHelper (\is vars -> simplifyTrivially $ proj is vars)
    where
    proj is vars =
	case projectToVars' HsFM is Nothing vars of
          -- or use: CSidePip
	  Nothing  -> infeasibleIneqSystem (length vars) (nParams is)
	  Just sys -> sys


rectTileHelper :: (Show b) 
               => (IneqSystem -> [Integer] -> b)
	       -> [Integer]    -- tile sizes
               -> (IneqSystem, Integer)
               -> [TileInformation]
	       -> (b, [Integer])
rectTileHelper projectionFunc tileSizes (dependence,limitS)  infos
    | nub infos /= infos  =
	error ("tile: a \"tile information\" may not appear twice in " ++
	       "`infos'")
    | otherwise = trace ("in rectTileHelper:\n" ++ show (tD,sD, dD,pD)) $
                  trace ("tileIS: \n" ++ show tileIS) $
                  if null asserts 
                  then trace ("dependence: " ++ show dependence) $
                           (tileIS, infoDims)
                  else error asserts
    where
    nrsOfVars = map (fromIntegral . length . info2indexes)

    infoDims = nrsOfVars infos
    tileIS = projectionFunc is indexes

    simplifyMethod = (\is -> trace "eliminateRedundantSorted called..." $
                             eliminateRedundantSorted is Nothing DefaultMethod)
    is  = rectTileIneqs tileSizes (dependence,limitS) 
    tD = genericLength tileSizes
    sD = limitS
    dD = nVars dependence - limitS
    pD = nParams dependence
    indexes =        
        concat $ map info2indexes infos


    info2indexes SrcTile    = [0         .. (tD-1)]
    info2indexes SrcCoord   = [tD        .. (tD+sD-1)] 
    info2indexes SrcOffset  = error "SrcOffset not supported in rectTile"

    info2indexes DestTile   = [(tD+sD) .. (2*tD+sD-1)]
    info2indexes DestCoord  = [(2*tD+sD) .. (2*tD+sD+dD-1)]
    info2indexes DestOffset = error "SrcOffset not supported in rectTile"

    asserts = if nVars is ==  2*tD+sD+dD 
               then "" 
               else ("dimension mismatch in rectTileHelper: " ++ 
                     (show (nVars is, 2*tD+sD+dD) ++ "\n"))
