-- |
-- This tool can be used to check the directions of the
-- transformed dependencies, e.g., to test whether a
-- given schedule is correct. It is also possible to
-- verify that dependencies, which were cut by the
-- allocator, do not lead to communication.
--
-- Usage:
-- 
--   1) @make trdep_directions@
--
--   2) @.\/trdep_directions --help@
--
-- Important:
--
-- The schedule (or placement) is taken from the @.trafo@ file
-- (i.e., the @.sched@ and @.place@ files will be ignored).
-- If you want to test a different schedule, you have to
-- modify the @.trafo@ file manually.
--
-- (The @.trafo@ file contains two matrices for each statements.
--  Only the second matrix is important, as it describes the
--  transformation for the current statement. The first matrix
--  is the inverse of the second matrix, but will be
--  ignored in this context.)
--
module TrDepDirections
    (main
    ,trDepsCommunicationDirection
    ,relevantTrDeps, relevantTiledTrDeps, isSchedule)
where

import LoopoDS
import Tiling(TileInformation(..), defaultTile)
import LMath.Algo.Sort ( projectToVars', SortTool(HsFM) )
import LMath.Types
import Debug.Trace(trace)
import Maybe(fromMaybe)
import System (getArgs)

-- |
-- Extracts the rows between @dimFrom@ and @dimTo@ (inclusively)
-- from the given space-time matrix.
--
trafodim :: Integer                   -- ^ @dimFrom@
	 -> Integer                   -- ^ @dimTo@
	 -> (TStatementID, AffineFunction) -- ^ the space-time matrix
	 -> (TStatementID, AffineFunction)
trafodim dimFrom dimTo (stmtNr, affFct) = 
    (stmtNr, affineFunction ((v,p),
     subMatrixHor m dimFrom dimTo) Inhomogeneous)
    where v = nVars affFct
	  p = nParams affFct
	  m = getMatrixIH affFct

interestingDims :: Integer
		-> Integer
		-> Trafo
		-> [(TStatementID, AffineFunction)]
interestingDims from to trafo =
    map (trafodim from to) (makeProperTrafos trafo)

properdeps :: LoopoDS -> [Dependence]
properdeps lds = map (trimDependence (loopoSLP lds)) (loopoDEP lds)

relevantTrDeps :: Integer
	       -> Integer
	       -> LoopoDS
	       -> [TrDependence]
relevantTrDeps from to lds = trdeps
    where
    tispc = loopoTISPC lds
    trafo = loopoTrafo lds
    deps  = map (trimDependence slp) $ loopoDEP lds
    slp   = loopoSLP lds
    properTrafos = interestingDims from to trafo
    trdeps = transformDeps tispc properTrafos deps

-- |
-- liesAtOneSide takes a hyperplane @hyp@ (which must be an ineqSystem
-- with one inequality), and checks whether the inequality system
-- @is@ lies on the one or the other side of the hyperplane
-- [then, @hyp@ or -@hyp@ is redundant in @is@],
-- or (partly) on both sides [no redundancy], 
-- or entirely in the hyperplane [both redundancies].
--
-- The result is
--
-- (@hyp@ redundant for @is@, -@hyp@ redundant for @is@) =
--
-- (@is@ satisfies @hyp@, @is@ satisfies (-@hyp@))
--
liesAtOneSide :: IneqSystem  -- ^ the hyperplane @hyp@
	      -> IneqSystem  -- ^ the inequality system @is@
	      -> (Bool,Bool)
liesAtOneSide hyp is =
    (isIneqRedundant (hyp `join` is) hyprow Nothing PipMethod,
     isIneqRedundant (neghyp `join` is) hyprow Nothing PipMethod)
    where 
    hyprow = case nRows hyp of 
	     1 -> 0
	     x -> error $ "a hyperplane has only one inequality" ++
		          ", but this hyp has " ++ show x ++ " rows"
    neghyp = ineqSystem((nVars hyp, nParams hyp), negate (getMatrix hyp))


liesInHyp :: IneqSystem -> IneqSystem -> Bool
liesInHyp hyp is = left && right 
    where (left,right) = liesAtOneSide hyp is


-- |
-- Creates a hyperplane (containing one inequality) for the dependence @td@,
-- which splits the space at the dimension @pos@.
--
trDepToHyp :: TrDependence  -- ^ @td@
	   -> Integer       -- ^ @pos@
	   -> IneqSystem
trDepToHyp td pos =
    if pos < 0 || pos >= dimSrc || pos + dimSrc >= nv 
    then error $ "trDepToHyp: input variable 'pos' is out of bounds" ++
                 " (pos=" ++ show pos ++ ", dimSrc=" ++ show dimSrc ++
                 ", nv=" ++ show nv ++ ")"
    else ineqSystem ((nv, np), rowMatrix entries)
  where
    nv = nVars tdr
    np = nParams tdr
    (tdr, dimSrc) = trDepRelation td

    entries = let selectRow x = if x == pos then 1 else 0
              in map (negate.selectRow) [0..dimSrc-1]    ++
		 map selectRow          [0..nv-dimSrc-1] ++
		 replicate (np+1) 0


trDepLiesAtOneSide :: TrDependence
		   -> IneqSystem    -- ^ the hyperplane
		   -> (Bool,Bool)
trDepLiesAtOneSide td hyp = liesAtOneSide hyp (fst $ trDepRelation td)

-- |
-- Returns @True@ if the inequality system @is@ lies in the hyperplane @hyp@
-- and also in -@hyp@ (negation of @hyp@).
--
zeroDifferencePossible :: IneqSystem  -- ^ the hyperplane @hyp@
		       -> IneqSystem  -- ^ @is@
		       -> Bool
zeroDifferencePossible hyp is =
    let neghyp = ineqSystem((nVars hyp, nParams hyp), negate (getMatrix hyp))
        is' = is `join` hyp `join` neghyp
    in not (areIneqsInfeasible is')

-- |
-- Tests if all the given dependencies satisfy the schedule conditions.
--
isSchedule :: [TrDependence]     -- ^ the dependences to be tested
	   -> (Integer, Integer) -- ^ describes the interval of the dimensions
	                         --   that contain the (possible) schedule
	   -> Bool
isSchedule tds (dimFrom, dimTo) = all isScheduleImp tds

-- |
-- Helper function of @isSchedule@. Takes only one single dependence
-- instead of the complete list of dependences.
--
isScheduleImp :: TrDependence  -- ^ the dependence to be tested
	      -> Bool
isScheduleImp trDep = recTest dims initialIneq
  where
    (initialIneq, numSrcVars) = trDepRelation trDep
    (src, dest) = (0, numSrcVars)
    dims = [0..numSrcVars-1]

    --
    -- Test each schedule dimension recursively and tighten the inequalities
    -- by adding the constrains defined by the hyperplanes that are
    -- used to isolate each dimension.
    --
    -- The following cases are possible:
    -- 1) containsBackwardCommunication => no Schedule (return False)
    -- 2) isForward && not zeroPossible => Schedule (return True)
    -- 3) isForward && zeroPossible =>
    --      proceed with remaining dimensions (recursive call)
    --
    recTest :: [Integer]   -- the remaining dimensions
	    -> IneqSystem  -- the tightend inequality system
	    -> Bool
    recTest [] _ = False
    recTest (x:xs) ineq =
	let hyp = trDepToHyp trDep x
	    tightendHyp = ineq `join` hyp
	in if containsBackwardCommunication hyp ineq then False    -- case 1)
	   else if not $ zeroDifferencePossible hyp ineq then True -- case 2)
	   else recTest xs tightendHyp -- case 3)


-- |
-- Retuns @True@ if the communication described by @is@ is forward directed
-- according to the hyperplane @hyp@.
--
isForwardCommunication :: IneqSystem  -- ^ the hyperplane @hyp@
 		       -> IneqSystem  -- ^ @is@
 		       -> Bool
isForwardCommunication hyp is = liesAtOneSide hyp is == (True,False)


-- |
-- Retuns @True@ if the communication described by @is@ is backward directed
-- according to the hyperplane @hyp@. Returns also @True@, if the communication
-- is in both directions (backward and forward directed).
--
containsBackwardCommunication :: IneqSystem  -- ^ the hyperplane @hyp@
			      -> IneqSystem  -- ^ @is@
			      -> Bool
containsBackwardCommunication hyp is =
    let direction      = liesAtOneSide hyp is
	backward       = (False, True)
	bothDirections = (False, False)
    in direction == backward || direction == bothDirections


-- |
-- Calculates the communication directions of the dependence @td@
-- for every dimension.
-- 
-- Interpretation of the results:
--
--   * (True, True):   no communication
--
--   * (True, False):  forward communication
--
--   * (False, True):  backward communication
--
--   * (False, False): communication in both directions
--     due to each of the dependences given
--
trDepsCommunicationDirection :: TrDependence -- ^ @td@
			     -> [(Bool,Bool)]
trDepsCommunicationDirection = trDepsFunctionsImp liesAtOneSide


trDepsZeroDifferencePossible :: TrDependence -> [Bool]
trDepsZeroDifferencePossible = trDepsFunctionsImp zeroDifferencePossible


-- |
-- Helper function used by:
--
--   1) trDepsCommunicationDirection
--
--   2) trDepsZeroDifferencePossible
-- 
trDepsFunctionsImp :: (IneqSystem -> IneqSystem -> a) -- ^ the operation @f@
		    -> TrDependence -- ^ the dependence @trDep@
		    -> [a]
trDepsFunctionsImp f trDep = recImp dims initialIneq
  where
    (initialIneq, numSrcVars) = trDepRelation trDep
    (src, dest) = (0, numSrcVars)
    dims = [0..numSrcVars-1]

    --
    -- Apply the operation @f@ recursively at each dimension and tighten the
    -- inequalities by adding the constrains defined by the hyperplanes that
    -- are used to isolate each dimension.
    -- Note: This function is actually a simplified version of the @recTest@
    --       function that is used to implement @isScheduleImp@.
    --
    recImp [] _ = []
    recImp (x:xs) ineq =
	let hyp = trDepToHyp trDep x
	    tightendHyp = ineq `join` hyp
	in f hyp ineq : recImp xs tightendHyp


-- |
-- Interprets the results of the function 'trDepsCommunicationDirection'.
--
showCommunicationDirection :: (Bool,Bool) -> String
showCommunicationDirection (True, True)   = "no communication"
showCommunicationDirection (True, False)  = "forward communication"
showCommunicationDirection (False, True)  = "backward communication"
showCommunicationDirection (False, False) = "backward and forward communication"


-- |
-- Represents the different requests of the user.
--
data RequestType = CallIsSchedule -- (default type)
		 | CallCommunicationDirection
		 | CallZeroDifferencePossible
		 
-- |
-- Calculates the requested function and sends the results to stdout.
--
dealWithUserRequest :: String -- ^ path to the temp-directory
		    -> String -- ^ the name of the project
		    -> (Integer, Integer) -- ^ the rows which contain the
		                          --   schedule dependences
		    -> RequestType -- ^ specifies the function that
		                   --   should be called
		    -> IO ()
dealWithUserRequest path projName dims@(dimFrom, dimTo) request =
    case request of

      -- verify the input
      _ | dimFrom < 0     -> error "invalid input: dimFrom < 0"
      _ | dimTo   < 0     -> error "invalid input: dimTo < 0"
      _ | dimFrom > dimTo -> error "invalid input: dimFrom > dimTo"
      _ | dimTo   > dimToMax ->
	    error ("invalid input: dimTo out of bounds (max=" ++
		   show dimToMax ++ ")")

      CallIsSchedule ->
	  putStrLn $ "isSchedule: " ++ show (isSchedule trDeps dims)

      CallCommunicationDirection ->
	  do putStrLn "Communication directions:"
             forAllDependencies printCommDirections

      CallZeroDifferencePossible ->
	  do putStrLn "zeroDifferencePossible results"
	     forAllDependencies printZeroDiffPossible

  where
    outfile :: String
    outfile = if null path || last path == '/'
	      then path ++ projName ++ extension
	      else path ++ '/':projName ++ extension
    extension = ".out"

    -- this upperbound is dubious!
    dimToMax :: Integer
    dimToMax = let trafo      = loopoTrafo lds
		   stmtTrafos = map snd trafo
		   func       = map trafoFunc stmtTrafos
	       in (maximum $ map (\x -> max (nVars x) (nParams x)) func) - 1

    lds :: LoopoDS
    lds = unsafeReadLoopoDS outfile

    trDeps :: [TrDependence]
    trDeps = relevantTrDeps dimFrom dimTo lds

    forAllDependencies :: (TrDependence -> IO ())
		       -> IO ()
    forAllDependencies f =
	sequence_ $ zipWith (\x y -> do putStrLn $ show x ++ ". dependence"
			                f y)
			    [1..] trDeps

    printCommDirections :: TrDependence -> IO ()
    printCommDirections trDep =
	let directions :: [(Bool,Bool)]
	    directions = trDepsCommunicationDirection trDep
	in sequence_ $ map (putStrLn.showCommunicationDirection) directions
			     
    printZeroDiffPossible :: TrDependence -> IO ()
    printZeroDiffPossible trDep =
	sequence_ $ map (putStrLn.show) (trDepsZeroDifferencePossible trDep)
	

-- |
-- Usage:
--
-- @trdep_directions \<path> \<project> \<dimFrom> \<dimTo> [--type=\<type>]@
-- where @\<type>=isSchedule | communicationDirection | zeroDifferencePossible@
--
-- (if no type is specified @--type=isSchedule@ will be assumed)
--
main :: IO ()
main = do args <- getArgs
	  case args of
	    [] -> error usageInfo
	    [path, projName, dimFrom, dimTo] ->
		-- no explicit type given => take default type
		takeAction path projName (dimFrom, dimTo) defaultType
	    [path, projName, dimFrom, dimTo, typeStr] ->
		let reqType = parseType typeStr
		in takeAction path projName (dimFrom, dimTo) reqType
	    _ -> error $ "Could not parse command line parameters\n" ++
		         usageInfo
  where
    parseType :: String -> RequestType
    parseType "--type=isSchedule" = CallIsSchedule
    parseType "--type=communicationDirection" = CallCommunicationDirection
    parseType "--type=zeroDifferencePossible" = CallZeroDifferencePossible
    parseType _ = error $ "Could not parse request type\n" ++ usageInfo
		  
    defaultType :: RequestType
    defaultType = CallIsSchedule

    takeAction :: String -> String -> (String, String) -> RequestType -> IO()
    takeAction path projName (fromStr, toStr) =
	let dimFrom = (read fromStr)::Integer
	    dimTo   = (read toStr)::Integer
	in dealWithUserRequest path projName (dimFrom, dimTo)

    usageInfo :: String
    usageInfo = "Usage: trdep_directions <path> <project> " ++
		"<dimFrom> <dimTo> [--type=<type>]\n" ++
		"where <type>=isSchedule | communicationDirection | " ++
		"zeroDifferencePossible"


-- |
-- Debugging function. Can be used to display the results of finished functions.
--
verbose :: Show a => String -> a -> a
verbose str x =
    let skipIt = True -- suppress debugging output
    in if skipIt then x
       else seq x $ trace ("Verbose: " ++ str ++ ": result=" ++ show x) x

--------------------------------------------------------------------------------------------
-- MC: 11.10.2005

-- |
-- relevantTiledTrDeps takes all transformed dependences, tiles them (using the
-- given lattice matrix) and returns trDependences with dependence
-- relations that consist of tiled dimensions only.
--
relevantTiledTrDeps :: [TrDependence]   -- ^ the input @trDeps@		       
		    -> Integer          -- ^ @dimFrom@ (of tiled dimensions)
		    -> Integer          -- ^ @dimTo@   (of tiled dimensions)
		    -> [Vector]         -- ^ the lattice matrix used for tiling
		    -> [TrDependence]
relevantTiledTrDeps trDeps from to lattices = 
    map tileTrDeps trDeps
    where
    tileTrDeps :: TrDependence -> TrDependence
    tileTrDeps trDep@TrDependence { trDepRelation = trDepRel} = 
	trDep { trDepRelation = 
		  let (ineq, [srcTileDim,destTileDim]) = 
                        defaultTile lattices [0.. (fromIntegral $ length lattices)] trDepRel 
                        [SrcTile,DestTile]
		      limitSrcDim = snd trDepRel
		      relVars = srcVars ++ destVars
		      srcVars = [from .. to]
		      destVars = [from+srcTileDim .. to+srcTileDim]		
	          in -- trace ("tileTrDeps: relVars=" ++ show relVars) $
		     (fromMaybe (error "problem in projectToVars' HsFM (called from relevantTiledTrDeps))") $
		     projectToVars' HsFM ineq Nothing relVars, fromIntegral $ length srcVars)
	      }
	
