module UniFCO(main) where

import LMath
import Loopo

import Prelude hiding ( catch )

import Control.Exception
import Control.Monad
import Control.Monad.Error
import Control.Monad.State
import Data.Dynamic
import qualified Data.Map as Map
import Data.Map ( Map )
import Data.List
import Data.Maybe
import System.Environment
import System.IO ( openFile, hPrint, hPutStr, hPutStrLn, hClose,
		   stderr, IOMode(..)
		 )
import System.Exit
import System.Console.GetOpt

import Debug.Trace


data ParseError = ParseError String
		  deriving (Typeable)

data FCOParams = FCOParams { fpArrays :: (Maybe [ArrayID]),
			     fpAllocDims :: Integer,
                             fpSchedDims :: Integer,
			     fpStrict :: Bool
			   }
		 deriving (Show)

defaultParams :: FCOParams
defaultParams = FCOParams { fpArrays = Just [],
			    fpAllocDims = 1,
                            fpSchedDims = 0,
			    fpStrict = False
			  }

options :: [OptDescr (ErrorT String (State FCOParams) ())]
options = [ Option ['a'] ["arrays"] (OptArg arrays "A1,...")
	           "arrays to allocate (or all)"
	  , Option ['d'] ["dim"]    (ReqArg dim    "DIM"   )
	           "dimensionality of placement"
	  , Option ['s'] ["strict"] (NoArg strict)
                   "compute strict FCO placement"
          , Option [] ["schedule"]  (ReqArg sdim   "DIM"   )
                   "compute a schedule, too (implies --strict)"
	  ]
    where
    arrays Nothing  = modify (\p -> p { fpArrays = Nothing })
    arrays (Just s) = 
	case reads ('[':s++"]") of
	  (arrs,""):_ -> modify (\p -> p { fpArrays = Just $
					              map integerToID arrs })
	  _           -> throwError "could not parse array list"
    dim s =
	case reads s of
          (d,""):_ -> modify (\p -> p { fpAllocDims = d })
	  _        -> throwError "could not parse placement dimensionality"
    sdim s =
 	case reads s of
          (d,""):_ -> modify (\p -> p { fpSchedDims = d, fpStrict = True })
	  _        -> throwError "could not parse schedule dimensionality"
    strict = modify (\p -> p { fpStrict = True })

parseOptions :: String -> [String] -> Either String FCOParams
parseOptions pn args =
    case getOpt Permute options args of
      (os,[],[]) -> evalState (runErrorT (sequence_ os >> get)) defaultParams
      (_,_,errs) -> Left (concat errs ++ usageInfo (usageInfoHeader pn) options)

usageInfoHeader :: String -> String
usageInfoHeader pn = "USAGE: " ++ pn ++ " TempDir BaseName [Options]"

main :: IO ()
main = do args <- getArgs
	  pn   <- getProgName
	  if length args < 2 
	    then hPutStrLn stderr (usageInfo (usageInfoHeader pn) options)
	    else do let tempdir  = args!!0
			basename = args!!1
	            setTempDir tempdir
	            case parseOptions pn (drop 2 args) of
		      Left errmsg ->
			  appendSessionLog errmsg >>
			  exitWith (ExitFailure 1)
		      Right params ->
			  runIt tempdir basename params >>= exitWith


runIt :: FilePath -> FilePath -> FCOParams -> IO ExitCode
runIt tempdir basename params =
	(do result <- run basename (fpStrict params) (fpArrays params)
	                  (fpAllocDims params) (fpSchedDims params)
	    safeWriteFile (tempdir ++ "/alloc_output") $ fcoMsg result
	    mapM_ (\t -> appendSessionLog (t ++ "\n")) $ fcoLog result
	    appendSessionLog ("AllocatorOutputfile " ++ tempdir ++
			      "/alloc_output")
	    case result of
               FCOResult {} -> 
	          do safeWriteFile (basename++".place")     (fcoStmtPl result)
	             safeWriteFile (basename++".dataplace") (fcoDataPl result)
                     when (fpSchedDims params > 0) $
                          safeWriteFile (basename++".sched") (fcoSched result)
               NoFCO {} -> return ()
	    return ExitSuccess

	 ) `catch` (\ex -> appendSessionLog ("FCO allocator error: " ++
					     show ex ++ "\n") >>
		           return (ExitFailure 1)
		   )


data FCOResult = FCOResult { fcoSched  :: String,
                             fcoStmtPl :: String,
			     fcoDataPl :: String,
			     fcoLog :: [String],
			     fcoMsg :: String
			   }
	       | NoFCO { fcoLog :: [String],
		         fcoMsg :: String
		       }
		 deriving (Show)

--
-- Perform all the steps of the FCO algorithm.
--
run :: FilePath -> Bool -> Maybe [ArrayID] -> Integer -> Integer -> IO FCOResult
run basename strictFCO mbArraysToAllocate nAllocDims nSchedDims =
    do slps   <- loadSurroundingLoops basename
       (stmts',accs') <- loadStatementsAndAccesses basename
       deps'  <- loadDependences  basename
       let stmts = map (trimIndexSpace slps) stmts'
	   accs  = map (trimAccess slps) accs'
	   deps  = map (trimDependence slps) deps'
       acps <- loadAccessPairs basename

       let nLoopIndices = nVars $ snd $ head $ stmts
	   nP = nParams $ snd $ head stmts
	   arrayIDs = nub $ map accArrayID accs
	   arraysToAllocate
	       | Just ata <- mbArraysToAllocate = ata
	       | otherwise                      = arrayIDs
	   nArrays = genericLength arrayIDs
	   relevantDeps = relevantDependences strictFCO arraysToAllocate
			                      deps accs acps
	   relevantAccs = filter ((`elem` arraysToAllocate).accArrayID) accs
	   (totalWidth, nLambdas, paw) =
	       posAndWidth arraysToAllocate stmts relevantAccs relevantDeps
	   sa = statementsAndArrays paw
	   nPis = totalWidth - nLambdas
	   fcoM = fcoMatrix totalWidth nLambdas paw relevantDeps relevantAccs
	   fcoEqs = eqSystem ((totalWidth,0),
			      insertZeroColumns fcoM totalWidth 1)
	   fcoIneqs = ineqSystem ((totalWidth,0),
				  zeroMatrix nLambdas (totalWidth-nLambdas)
				  `conc`
				  unitMatrix nLambdas nLambdas `conc`
				  zeroMatrix nLambdas 1)
	   LRV {lrvLines=lines, lrvRays=rays} = dualOfSystem fcoEqs fcoIneqs
	   goodLines = makeGoodLines paw sa nP lines
           candidates = multidim nP sa (nAllocDims+nSchedDims) $
			map (extract paw) $
			joinMultiples nPis (goodLines++rays)
	   result = sortBy (\(_,(a,_,_)) (_,(b,_,_)) -> compare a b) $
		    [(blowUpStmtPlacements slps c,
		      commCost relevantAccs relevantDeps fcoM paw c)
		    | c <- candidates]
	   result' = map (\(a,b) -> (Map.toAscList a,b)) result
           log1s = [show fcoIneqs, show (map Map.toAscList candidates)]
       if (not $ null result) then
	   do let (bestPlacement,(CommCost costs,cutDeps,cutAccs)) = head result
                  stmtIDs = map fst stmts
                  stmtSchedule  = Schedules $ Map.fromList $
				  [(s,QList $ rowVectors $ mdrop nAllocDims p) | (S s, p) <- Map.toAscList bestPlacement]
		  stmtPlacement = Placements $ Map.fromList $
				  [(s,QList $ rowVectors $ mtake nAllocDims p) | (S s, p) <- Map.toAscList bestPlacement]
		  dataPlacement = Placements $ Map.fromList $
				  [(a,QList $ rowVectors $ mtake nAllocDims p) | (A a, p) <- Map.toAscList bestPlacement]
                  stmtScheduleF  = formatStmtSchedule  nLoopIndices nP stmtSchedule
		  stmtPlacementF = formatStmtPlacement nLoopIndices nP stmtPlacement
		  dataPlacementF = formatDataPlacement nArrays nLoopIndices nP dataPlacement
		  log = (if nSchedDims > 0
                            then ["Schedule:",
                                  "=========",
                                  ""] ++
                                  ["Statement " ++ show (integerFromID s) ++ ": " ++
			                        showCoeff nP "i" (entries $ mdrop nAllocDims p)
		                   | (S s,p) <- Map.toAscList bestPlacement] ++
                                  [""]
                            else []
                        ) ++
                        ["Placements:",
			 "===========",
			 ""] ++
			["Statement " ++ show (integerFromID s) ++ ": " ++
			 showCoeff nP "i" (entries $ mtake nAllocDims p)
		            | (S s,p) <- Map.toAscList bestPlacement] ++
			["Array     " ++ show (integerFromID a) ++ ": " ++
			 showCoeff nP "x" (entries $ mtake nAllocDims p)
		            | (A a,p) <- Map.toAscList bestPlacement] ++
			[""] ++
			["Dependence " ++ show (integerFromID d) ++
			 (if d `elem` cutDeps then "    " else " not") ++
			 " cut"
		            | d <- map depID relevantDeps] ++
			["Access " ++ show (integerFromID a) ++
			 (if a `elem` cutAccs then "    " else " not") ++
			 " cut"
			    | a <- map accID relevantAccs] ++
			[""] ++
			[show n ++ " " ++ show dim ++
			 "-dimensional communication" ++
			 (if n==1 then "" else "s")
			    | (dim,n) <- Map.toAscList costs] ++
			["",
			 "Candidates:",
			 "===========",
			 ""] ++ 
			[show result']

	      return $ FCOResult { fcoMsg = unlines log,
                                   fcoSched  = stmtScheduleF,
		                   fcoStmtPl = stmtPlacementF,
			           fcoDataPl = dataPlacementF,
				   fcoLog = ["FCO allocation computed."]
			         }
         else return $ NoFCO { fcoMsg = "No non-trivial solution for the \
					\following FCO system exists:\n\n" ++
			                unlines log1s,
		               fcoLog = ["No non-trivial FCO allocation exists."] }

    where
    showCoeff :: Integer -> String -> [Rational] -> String
    showCoeff nP prefix coeffs =
	case dropWhile (==' ') res of
	  '+':res' -> res'
	  res'     -> res'
	where
	res = concat [showV v (prefix ++ show n) | v <- vars | n <- [0..]] ++
	      concat [showV v ("n" ++ show n) | v <- init pars | n <- [0..]] ++
	      showV (last pars) ""

	(vars,pars) = splitAt (length coeffs - fromIntegral nP - 1) coeffs

	showV v post
	    | v == 1  && not (null post) = " +" ++ post
	    | v == -1 && not (null post) = " -" ++ post
	    | v /= 0    = ' ' : showR v ++ (if null post then "" else '*':post)
	    | otherwise = ""
	showR r
	    | denominator r == 1  =  showI (numerator r)
	    | otherwise           =  showI (numerator r) ++ '/' : show (denominator r)
	showI i
	    | i >= 0  =  '+' : show i
	    | i <  0  =  show i
	

mtake :: Integral a => a -> GMatrix b -> GMatrix b
mtake n m = subMatrixHor m 0 (n-1)

mdrop :: Integral a => a -> GMatrix b -> GMatrix b
mdrop n m = subMatrixHor m n (nRows m-1)

--
-- Ensure that  hClose  is called when an exception occurs.
--
safeWriteFile :: FilePath -> String -> IO ()
safeWriteFile path contents =
    do h <- openFile path WriteMode
       (hPutStr h contents `finally` hClose h)

--
-- Blow up the computed placements, i.e., add non-surrounding loops
-- to the description of the placements of the statements
--
blowUpStmtPlacements :: SurroundingLoops
		     -> Map ID Matrix
		     -> Map ID Matrix
blowUpStmtPlacements slp alloc = Map.mapWithKey f alloc
    where
    f (S stmtID) a = blowUpStmtPlacement a (lookupSurroundingLoops slp stmtID)
    f _          a = a

blowUpStmtPlacement :: Matrix
		    -> [Bool]
		    -> Matrix
blowUpStmtPlacement alloc slp =
    columnVectorsToMatrix nR $ bu (columnVectors alloc) slp
    where
    bu :: [Vector] -> [Bool] -> [Vector]
    bu xs [] = xs
    bu xs (False:rest) = zero : bu xs rest
    bu [] (True:_) = error "blowUpStmtPlacements: unexpected True \
			   \in surrounding loops list"
    bu (x:xs) (True:rest) = x : bu xs rest

    zero = zeroVector nR
    nR = nRows alloc


data ID = S StatementID
	| A ArrayID
	| D DependenceID
	| C AccessID
	  deriving (Show, Eq, Ord)

statementsAndArrays :: Map ID (Integer,Integer) -> [ID]
statementsAndArrays paw = filter isSA $ Map.keys paw
    where
    isSA (S _) = True
    isSA (A _) = True
    isSA _     = False

lookupFM' :: (Ord key, Show key) => Map key elt -> key -> elt
lookupFM' fm k
    | Just e <- Map.lookup k fm  =  e
    | otherwise                = 
	error ("lookupFM': key " ++ show k ++ " not found")

extract :: Map ID (Integer,Integer) -> Vector -> Map ID Vector
extract paw candidate =
    Map.map (\(pos,width) -> subVector candidate pos width) paw


--
-- join placement candidates which differ only
-- in the lambda values, i.e., the allocation parts
-- of the solutions are the same except for multiples.
--
joinMultiples :: Integer -> [Vector] -> [Vector]
joinMultiples nPis vs = nubBy eq $ map compact vs
    where
    eq v1 v2 = subVector v1 0 nPis == subVector v2 0 nPis
    compact v
	| num == 0  = v
	| otherwise = fmap (/ fromIntegral num) v
	where
	co  = genericTake nPis $ vectorEntries v
	num = foldl gcd 0 $ filter (/=0) $ map numerator co

--
-- Combine the given lines to better lines, e.g.,
-- for the program
--
--    DO i=...
--      A(i) = i
--      B(i) = i
--    END DO
--
-- Polylib computes the lines
--   PI_A(i)=i, PI_B(i)=0
-- and
--   PI_A(i)=0, PI_B(i)=i
-- each of which distributes one of the arrays only. We
-- want to combine the lines to a placement
--   PI_A(i)=i, PI_B(i)=i
--
-- The idea to achieve this is as follows:
-- (1) The lines are written as the rows of a matrix and
--     echelon reduction is performed on the matrix.
-- (2) A subset of the rows representing stairs are selected.
--     Each selected row (except the first one) has at least
--     one more block of zeros at the beginnig
--     (where one "block" refers to the columns corresponding
--     to the coefficients of an array or statement allocation).
-- (3) Then a linear combination of the selected rows which
--     is non-zero in (the columns corresponding to the variables'
--     coefficients of) every block (when possible) can be computed
--     easily.
--
makeGoodLines :: (Ord idtype, Show idtype)
	      => Map idtype (Integer,Integer)
	      -> [idtype]
	      -> Integer
	      -> [Vector]
	      -> [Vector]
makeGoodLines _ _ _ [] = []
makeGoodLines paw (sa::[idtype]) nP lines1 =
    -- (\res -> trace (show (echelon,lines2,res)) res) $
    scanl nondestr goodLine (tail lines2)
    where
    echelon :: Matrix
    (_, echelon) = echelonReduction $ rowVectorsToMatrix1 lines1
    lines2 :: [Vector]
    lines2 = sortBy (\a b -> ordN (firstNZ a) (firstNZ b)) $ rowVectors echelon

    goodLine :: Vector
    goodLine = gl sortedSA (zeroVector $ vectorLength $ head lines1)
	where
	gl [] l = l
	gl (id:sa') l
	    | isZero id l, (h:_) <- lookupFM' leaders id = gl sa' (nondestr l h)
	    | otherwise   = gl sa' l
 
    nondestr :: Vector -> Vector -> Vector
    nondestr l h = head [l' | co <- coeffs, let Just l' = nond (l + co `mulSV` h)]
	where
	nond l'
	    | all (\id -> isZero id l' <= isZero id l) sortedSA = Just l'
	    | otherwise = Nothing

    coeffs = [m*f | f <- [1..], m <- [1,-1]]

    leaders :: Map idtype [Vector]
    leaders = Map.fromList $ lead (sort sa) lines2
	where
	lead [] _ = []
	lead (id:sa') ls = (id,nzs) : lead sa' nz
	    where
	    (nzs,nz) = span (not . isZero id) ls

    isZero id line = all (==0) $ genericTake (width-nP-1) $
		     genericDrop pos $ vectorEntries line
	where
	(pos,width) = lookupFM' paw id

    firstNZ :: Vector -> Maybe idtype
    firstNZ line = listToMaybe [id | id <- sortedSA, not $ isZero id line]

    ordN :: Ord a => Maybe a -> Maybe a -> Ordering
    ordN Nothing  Nothing  = EQ
    ordN (Just _) Nothing  = LT
    ordN Nothing  (Just _) = GT
    ordN (Just a) (Just b) = compare a b

    sortedSA = sort sa

--
-- Compute multidimensional placements from one-dimensional
-- placements. Ensure that the rows of the multidimensional
-- placements are linearly independent (in the variable dimensions).
--
multidim :: Integer
	 -> [ID]
	 -> Integer         -- ^ number of dimensions
	 -> [Map ID Vector]
	 -> [Map ID Matrix]
multidim _  _      _    []   = []
multidim nP totest 1    sols =
    [Map.map (\v -> rowVectorsToMatrix1 [v]) sol
     | sol <- sols, all isOneDim [lookupFM' sol i | i <- totest]
    ]
    where
    isOneDim :: Vector -> Bool
    isOneDim v = vectorLength v == nP + 1 ||
		 (any (/=0) $ genericTake (vectorLength v - nP - 1) $
		  vectorEntries v)
multidim nP totest dims (sol:sols) =
    multidim nP totest dims sols ++
    concatMap app (multidim nP totest (dims-1) sols)
    where
    app :: Map ID Matrix -> [Map ID Matrix]
    app old = [new | all isLinIndep [lookupFM' new i | i <- totest]]
	where
	new = Map.mapWithKey
	            (\k oldM ->
		     insertRows oldM 0 (rowVectorsToMatrix1 [lookupFM' sol k])
		    )
	            old
    
    isLinIndep :: Matrix -> Bool
    isLinIndep m =
	nColumns m - nP < dims ||
	genericLength (kernel (subMatrixVer m 0 (nColumns m - nP - 1)))
          == nColumns m - nP - dims

--
-- Communication cost.
-- Maps the dimensionality (of communications) to the number of
-- communications with the respective dimensionality.
--
data CommCost = CommCost (Map Integer Integer)

instance Show CommCost where
    show (CommCost fm) = "CommCost " ++ show (Map.toAscList fm)

instance Eq CommCost where
    CommCost a == CommCost b  =  a == b

instance Ord CommCost where
    a <= b  =  a == b || a < b
    CommCost a < CommCost b  =
	case Map.elems $ Map.filter (/=0) $ Map.unionWith (-) b a of
	  [] -> False
	  xs -> last xs > 0

mkCommCost :: [Integer] -> CommCost
mkCommCost list =
    CommCost $ Map.fromListWith (+) [(d,1) | d <- list, d > 0]

--
-- Compute the cost of a placement.
-- Return the communication costs and the
--  cut dependences and local accesses.
--
commCost :: [Access]
	 -> [Dependence]
	 -> Matrix
	 -> Map ID (Integer,Integer)
	 -> Map ID Matrix
	 -> (CommCost, [DependenceID], [AccessID])
commCost accs deps matr paw candidate =
    (mkCommCost (map fst costAcc ++ map fst costDep),
     [d | (0,d) <- costDep],
     [a | (0,a) <- costAcc]
    )
    where
    costAcc = [(accCost a, accID a) | a <- accs]
    costDep = [(depCost d, depID d) | d <- deps]

    accCost :: Access -> Integer
    accCost acc
	| commDistZero (C $ accID acc)  = 0
	| otherwise                     = nVars $ accDomain acc

    depCost :: Dependence -> Integer
    depCost dep
	| commDistZero $ D $ depID dep = 0
	| otherwise                    = nVars $ depDestDomain dep
	                               -- depSrcDomain ? oder was anderes ??
	
    commDistZero :: ID -> Bool
    commDistZero id =
	all (==0) $ entries (subMatrixVer matr pos (pos+width-1) * lambdas) 
	where
	(pos,width) = lookupFM' paw id
	lambdas = mtranspose $ lookupFM' candidate id


--
-- Compute positions (= offsets) and column widths for
-- the statemnts, arrays, accesses, and dependences.
-- Return the total width, width of arrays and statements together
-- and the mapping from ID to (position,width)
--
posAndWidth :: [ArrayID]
            -> [Statement]
	    -> [Access]
	    -> [Dependence]
	    -> (Integer, Integer, Map ID (Integer,Integer))
posAndWidth arraysToAllocate stmts accs deps =
    (totalWidth, totalWidth-w2, Map.fromList (pawStmts++pawArrays++pawDeps++pawAccs))
    where
    (w1, pawStmts)  = mapAccumL f  0 stmts
    (w2, pawArrays) = mapAccumL g w1 arraysToAllocate
    (w3, pawDeps)   = mapAccumL h w2 deps
    (totalWidth, pawAccs)   = mapAccumL k w3 accs 

    f :: Integer -> Statement -> (Integer, (ID, (Integer,Integer)))
    f pos (stmtID,dom) = (pos+width, (S stmtID, (pos, width)))
	where
	width = nColumns dom

    g :: Integer -> ArrayID -> (Integer, (ID, (Integer,Integer)))
    g pos aID = (pos+width, (A aID, (pos, width)))
	where
	width = targetDim fct
	Just (Access { accIndexFct=fct }) = find ((==aID).accArrayID) accs
		
    h :: Integer -> Dependence -> (Integer, (ID, (Integer,Integer)))
    h pos dep = (pos+width, (D (depID dep), (pos, width)))
	where
	width = nRows (depDestDomain dep) + 1

    k :: Integer -> Access -> (Integer, (ID, (Integer,Integer)))
    k pos acc = (pos+width, (C (accID acc), (pos, width)))
	where
	width = nRows (accDomain acc) + 1


--
-- retain the dependences which are _not_ caused by arrays listed
-- in 'arraysToAllocate'
--
relevantDependences :: Bool -> [ArrayID] -> [Dependence] -> [Access] -> [AccessPair] -> [Dependence]
relevantDependences strictFCO arraysToAllocate deps accs acps =
    catMaybes [find ((==acpDepID p).depID) depsTU | p <- acps, notAlloc p]
    where
    depsTU
	| strictFCO = filter (strict.depType)    deps
	| otherwise = filter (nonstrict.depType) deps
    nonstrict = (`elem`[UniformTrue,NonUniformTrue,UniformFlow,NonUniformFlow])
    strict    = (`notElem`[UniformInput,NonUniformInput])

    notAlloc :: AccessPair -> Bool
    notAlloc (AccessPair { acpSrcAccID = acc_id }) =
	any (\a -> accID a == acc_id &&
	           accArrayID a `notElem` arraysToAllocate
	    )
	    accs

--
-- Compute the matrix (i.e. equalities) describing
-- FCO allocations (after Farkas' lemma has been applied).
--
fcoMatrix :: Integer
	  -> Integer
	  -> Map ID (Integer,Integer)
	  -> [Dependence]
	  -> [Access]
	  -> Matrix
fcoMatrix totalWidth nLambdas paw deps accs =
--    trace (show (deps,accs,blub))
    multijoin (zeroMatrix 0 totalWidth) $ blub
    where
    blub = (map (multiconc1 . mkDep) deps ++ map (multiconc1 . mkAcc) accs)
    
    mkAcc acc =
	[zeroMatrix sW sPos,
	 condNeg (not isWrite) $ unitMatrix sW sW,
	 zeroMatrix sW (aPos-sPos-sW),
	 condNeg isWrite $ mtranspose $ getMatrix $ accIndexFct acc,
	 zeroMatrix sW (cPos-aPos-aW),
	 mkDom $ getMatrix $ accDomain acc,
	 zeroMatrix sW (totalWidth-cPos-cW)
	]
	where
	condNeg True  = negate
	condNeg False = id
        isWrite = accType acc == Write
	Just (sPos, sW) = Map.lookup (S $ accStmtID  acc) paw
	Just (aPos, aW) = Map.lookup (A $ accArrayID acc) paw
	Just (cPos, cW) = Map.lookup (C $ accID acc) paw
    mkDep dep
	| dPos < sPos =
	    [zeroMatrix dW dPos,
	     negate $ unitMatrix dW dW,
	     zeroMatrix dW (sPos-dPos-dW),
	     mtranspose $ getMatrix $ depHTrafo dep,
	     zeroMatrix dW (hPos-sPos-sW),
	     mkDom $ getMatrix $ depDestDomain dep,
	     zeroMatrix dW (totalWidth-hPos-hW)
	    ]
	| dPos == sPos =
	    [zeroMatrix dW sPos,
	     (mtranspose $ getMatrix $ depHTrafo dep) - unitMatrix dW dW,
	     zeroMatrix dW (hPos-sPos-dW),
	     mkDom $ getMatrix $ depDestDomain dep,
	     zeroMatrix dW (totalWidth-hPos-hW)
	    ]
	| otherwise =
	    [zeroMatrix dW sPos,
	     mtranspose $ getMatrix $ depHTrafo dep,
	     zeroMatrix dW (dPos-sPos-sW),
	     negate $ unitMatrix dW dW,
	     zeroMatrix dW (hPos-dPos-dW),
	     mkDom $ getMatrix $ depDestDomain dep,
	     zeroMatrix dW (totalWidth-hPos-hW)
	    ]
	where
	Just (sPos, sW) = Map.lookup (S $ depSrcStmtID  dep) paw
	Just (dPos, dW) = Map.lookup (S $ depDestStmtID dep) paw
	Just (hPos, hW) = Map.lookup (D $ depID dep) paw
    mkDom m = mtranspose m `conc`
	      columnMatrix (replicate (nColumns m-1) 0 ++ [1])
