-- |
-- This module contains functions to convert 'IneqSystem's or 'Domain's
-- to a visualpolylib input file.
--
module VisualPolylib
    (showIS2VisualPolylib, showDomain2VisualPolylib)
where

import LMath
-- import IOExts(unsafePerformIO,trace) -- just for testing
-- replaced by the following two imports:
import Debug.Trace(trace)
import System.IO.Unsafe (unsafePerformIO)
-- import CodeGen(loopNest, getBoundText)
import IO(openFile, hGetContents, IOMode(..))
import System(getArgs)
import System.Cmd(system)
import System.Time
import GHC.IOBase(ExitCode(..))
import GHC.Base(assert)
import CppInterface
import Foreign.C.String
import Maybe
import Ix
import List
import Ratio

--
-- Default names for the variables, parameters and the constant.
--
varNamesVP, paramNamesVP, constNamesVP :: [String]
varNamesVP   = [[x] | x <- ['a'..'z']]             ++ ['z' : show i | i <- [2..]]
paramNamesVP = ["M", "N", "O", "P", "Q", "R", "S"] ++ ['X' : show i | i <- [1..]]
constNamesVP = ["1"]

-- |
-- Converts the polyhedron @is@ to a string representation
-- which can be displayed by "visualpolylib".
--
-- See also: 'showDomain2VisualPolylib'
--
showIS2VisualPolylib :: IneqSystem -- ^ the IneqSystem @is@
		     -> Int    -- ^ dimension
		     -> String -- ^ the output string (visualpolylib compatible)
showIS2VisualPolylib is dim = 
    contextStr ++ showIneqSystemVP is dim [] nP "S1" ++ initVisuStr
    where
    nV = nVars is
    nP = nParams is
    paramNames = take nP paramNamesVP
    c_t = "C"
    contextStr = 
	(c_t ++ " := {" ++ (concat $ intersperse ", " paramNames) ++ " | };\n\n")
   
    initVisuStr = "initvisu(" ++ show dim ++ "," ++ c_t ++ ");\n"


-- |
-- Converts the domain @d@ to a string representation
-- which can be displayed by "visualpolylib".
--
-- Preconditions:
--
--   (1) @canBeHandledByPolyLib d == True@
--
-- See also: 'showIS2VisualPolylib'
--
showDomain2VisualPolylib :: (BasicDomainElement a, PolyConvertableDomain a)
                         => Domain a -- ^ the domain @d@
                         -> Int      -- ^ dimension
			 -> [Int]    -- ^ list of dimensions which should be set to zero
			 -> String   -- ^ the output string (visualpolylib compatible)
showDomain2VisualPolylib d dim zeroDims =
    if isJust polylibIneqs
    then contextStr ++ ineqSystemStr ++ initVisuStr
    else "domain 'd' cannot be displayed by polylib:\nd=" ++ show d
    where
    polylibIneqs = toPolyLibIneqsD_checked d
    ineqSystemStr = concat $ zipWith showIneqSys (fromJust polylibIneqs) [1..]
    showIneqSys is id = showIneqSystemVP is dim zeroDims nP ('S':show id)

    nV = nVarsD d
    nP = nParamsD d
    paramNames = take nP paramNamesVP
    c_t = "C"
    contextStr = 
	(c_t ++ " := {" ++ (concat $ intersperse ", " paramNames) ++ " | };\n\n")
   
    initVisuStr = "initvisu(" ++ show dim ++ "," ++ c_t ++ ");\n"

-- |
-- Helper function.
-- Converts a single polyhedron @is@ to a visualpolylib compatible format.
--
showIneqSystemVP :: IneqSystem  -- ^ the IneqSystem @is@
	         -> Int    -- ^ dimension
		 -> [Int]  -- ^ list of dimensions which should be set to zero
		 -> Int    -- ^ number of parameters
	         -> String -- ^ name of the polyhedron in the output string
	         -> String -- ^ the output string (visualpolylib compatible)
showIneqSystemVP is dim zeroDims nP polyName =
    polyName ++ " := {" ++
    (concat $ intersperse ", " (varNames ++ paramNames)) ++
    " | \n" ++ ineqLines ++ pRest
    where
    nV = nVars is
    varNames   = take nV varNamesVP
    paramNames = take nP paramNamesVP
    allNames   = varNames ++ paramNames ++ constNamesVP

    rawM = getMatrix $ fmap fromIntegral $ integralIneqSystem is
    allRows = if nP == nParams is
	      then rows' rawM
	      else [init xs ++ replicate (nP - nParams is) 0 ++ [last xs]
		    | xs <- rows' rawM]

    ineqLines = concat $ intersperse ",\n" $
		((map showIneqRow2VP allRows) ++ zeroDimIneqs)
    zeroDimIneqs = let row x = [if x == y then 1 else 0 | y <- [0..nV + nP]]
		       equality x = [showIneqRow2VP (row x),
				     showIneqRow2VP (map negate (row x))
				    ]
                   in concatMap equality zeroDims
    pRest = "\n     };\n\n"

    combineCoeffs :: Integer -> String -> String
    combineCoeffs c n = "(" ++ show c ++ "*" ++ n ++ ")"

    showIneqRow2VP :: [Integer] -> String
    showIneqRow2VP r = 
	(concat $ intersperse "\t+ " $ 
	 zipWith combineCoeffs r allNames) ++ " >= 0"
