-- Calculating the volume and the number of integers in polytopes. See also Rabltest.hs.
--
-- Author: Malte Rosenthal

module VolIntCount where

-- imports --------------------------------------------------------------------------------------------------------------------------------------------
import LoopoDS
import LMath
import List
import Monad
import Time

import PolylibInterf
import NewCommCode hiding (main)
import Barvinok

-- Volume-calculations of Tilmann -------------------------------------------------------------------------------------------------------------

-- Calculate volumes of all indexspaces in the given project.
-- tmpdir       is the directory where the out-files of LooPo are, e.g. /tmp/loopo4711
-- projectname  is the leading part of the out-file with src.out, e.g. adder.src.out
calcVols :: String -> String -> [DTree (Polynomial Rational) (FracPoly Rational)]
calcVols tmpdir projectname = List.map myVol (getLoopoDSIspcs tmpdir projectname)

-- Calculate volumes of all indexspaces with the given type in the given project, where the program is manipulated by cloog.
-- tmpdir 	      	is the directory where the out-files of LooPo are, e.g. /tmp/loopo4711
-- projectname  	is the leading part of the out-file with src.out, e.g. adder.src.out
-- typeOfIneqSyst	is the type of the inequality system that should be handled (see NewCommCode.CloogStatementIneqs)
-- Other parameters: see NewCommCode.CommandLine and NewCommCode.unsafeGetCloogIneqs (TODO: adapt these parameters if they change in NewCommCode)
calcCloogVols :: String -> String -> String -> String -> String -> String -> String -> String -> String -> String -> String -> 
		 String -> String -> (CloogStatementIneqs -> [IneqSystem]) -> [DTree (Polynomial Rational) (FracPoly Rational)]
calcCloogVols 	tileWidth tileWhatever toBeTiled targoutPath cloogPath useCommunication cyclicMapping useMPIWrapper unpackAfterAlltoall
		cloogMinSizeParam cloogArgs
		tmpdir projectname typeOfIneqSyst =
    let commandLine :: [String]
    	commandLine = [	tmpdir, projectname, tileWidth, tileWhatever, toBeTiled, targoutPath, cloogPath, useCommunication,
			cyclicMapping, useMPIWrapper, unpackAfterAlltoall, cloogMinSizeParam, cloogArgs]

    in List.map myVol (getCloogIspcs commandLine typeOfIneqSyst)

-- Calls calcCloogVols with default parameters, where only tiling-parameters and project-parameters can be set
-- tmpdir       	is the directory where the out-files of LooPo are, e.g. /tmp/loopo4711
-- projectname  	is the leading part of the out-file with src.out, e.g. adder.src.out
-- typeOfIneqSyst	is the type of the inequality system that should be handled (see NewCommCode.CloogStatementIneqs)
-- Other parameters: see calcCloogVols
calcCloogVolsTile :: String -> String -> String -> String -> String -> (CloogStatementIneqs -> [IneqSystem]) ->
		     [DTree (Polynomial Rational) (FracPoly Rational)]
calcCloogVolsTile tileWidth tileWhatever toBeTiled tmpdir projectname typeOfIneqSyst = 
	calcCloogVols tileWidth tileWhatever toBeTiled "../loopo/targout/targout" "../cloog/cloog.git/bin/cloog" "--useCommunication=True"
		      "--cyclicMapping=True" "--useMPIWrapper=False" "--unpackAfterAlltoall=True" "100" "" tmpdir projectname typeOfIneqSyst


-- Calls calcCloogVols with default parameters
-- tmpdir       	is the directory where the out-files of LooPo are, e.g. /tmp/loopo4711
-- projectname  	is the leading part of the out-file with src.out, e.g. adder.src.out
-- typeOfIneqSyst	is the type of the inequality system that should be handled (see NewCommCode.CloogStatementIneqs)
calcCloogVolsDef :: String -> String -> (CloogStatementIneqs -> [IneqSystem]) -> [DTree (Polynomial Rational) (FracPoly Rational)]
calcCloogVolsDef tmpdir projectname typeOfIneqSyst =
	calcCloogVols "[[20,0],[0,400]]" "[0,1]" "[True,True]" "../loopo/targout/targout" "../cloog/cloog.git/bin/cloog" "--useCommunication=True"
		      "--cyclicMapping=True" "--useMPIWrapper=False" "--unpackAfterAlltoall=True" "100" "" tmpdir projectname typeOfIneqSyst



-- Barvinok ------------------------------------------------------------------------------------------------------------------------------------

-- Runs the algorithm of Barvinok for counting the number of integers in a linear-parametrized polytope.
-- outputFile	is the file where the output of the algorithm is printed. If it is the empty String "" it is printed to stdout.
-- binPath	is the path to the Barvinok-binary (see www.freshmeat.net or http://193.190.253.81/~skimo/barvinok/).
-- tmpdir	is the directory where the out-files of LooPo are, e.g. /tmp/loopo4711
-- projectname	is the leading part of the out-file with src.out, e.g. adder.src.out
runBarvinok :: String -> String -> String -> String -> IO () 
runBarvinok outputFile binPath tmpdir projectname = 
    let	polylibIneqSystems = List.map showIneqSystemForPolylib (getLoopoDSIspcs tmpdir projectname)
    in	Monad.mapM_ (barvinok outputFile binPath) polylibIneqSystems

-- Calls runBarvinok with "" as outputFile, i.e. all output will be printed on stdout.
runBarvinokStdOut :: String -> String -> String -> IO ()
runBarvinokStdOut binPath tmpdir projectname = runBarvinok "" binPath tmpdir projectname

-- Calculate the number of integers of all indexspaces with the given type in the given project, where the program is manipulated by cloog. This is done by
-- the algorithm of Barvinok. So the polytopes must be linear-parametrized.
-- tmpdir  		is the directory where the out-files of LooPo are, e.g. /tmp/loopo4711
-- projectname  	is the leading part of the out-file with src.out, e.g. adder.src.out
-- typeOfIneqSyst	is the type of the inequality system that should be handled (see NewCommCode.CloogStatementIneqs)
-- Other parameters: see NewCommCode.CommandLine and NewCommCode.unsafeGetCloogIneqs (TODO: adapt these parameters if they change in NewCommCode)
runBarvinokCloog :: String -> String -> String -> String -> String -> String -> String -> String -> String -> String -> String ->
		    String -> String -> String -> String -> (CloogStatementIneqs -> [IneqSystem]) -> IO ()
runBarvinokCloog	tileWidth tileWhatever toBeTiled targoutPath cloogPath useCommunication cyclicMapping useMPIWrapper unpackAfterAlltoall
			cloogMinSizeParam cloogArgs
			outputFile binPath tmpdir projectname typeOfIneqSyst =
    let commandLine :: [String]
    	commandLine = [	tmpdir, projectname, tileWidth, tileWhatever, toBeTiled, targoutPath, cloogPath, useCommunication,
			cyclicMapping, useMPIWrapper, unpackAfterAlltoall, cloogMinSizeParam, cloogArgs]
    
    	polylibIneqSystems = List.map showIneqSystemForPolylib (getCloogIspcs commandLine typeOfIneqSyst)

    in	Monad.mapM_ (barvinok outputFile binPath) polylibIneqSystems

-- Calls calcCloogVols with default parameters, where only tiling-parameters and project-parameters can be set
-- tmpdir       	is the directory where the out-files of LooPo are, e.g. /tmp/loopo4711
-- projectname  	is the leading part of the out-file with src.out, e.g. adder.src.out
-- typeOfIneqSyst	is the type of the inequality system that should be handled (see NewCommCode.CloogStatementIneqs)
-- Other parameters: see calcCloogVols
runBarvinokCloogTile :: String -> String -> String -> String -> String -> String -> (CloogStatementIneqs -> [IneqSystem]) -> String -> IO ()
runBarvinokCloogTile tileWhatever toBeTiled outputFile binPath tmpdir projectname typeOfIneqSyst tileWidth = 
	runBarvinokCloog tileWidth tileWhatever toBeTiled "../loopo/targout/targout" "../cloog/cloog.git/bin/cloog" "--useCommunication=True"
		      	 "--cyclicMapping=True" "--useMPIWrapper=False" "--unpackAfterAlltoall=True" "100" ""
		      	 outputFile binPath tmpdir projectname typeOfIneqSyst

-- Calls calcCloogVols with default parameters
-- tmpdir       	is the directory where the out-files of LooPo are, e.g. /tmp/loopo4711
-- projectname  	is the leading part of the out-file with src.out, e.g. adder.src.out
-- typeOfIneqSyst	is the type of the inequality system that should be handled (see NewCommCode.CloogStatementIneqs)
runBarvinokCloogDef :: String -> String -> String -> String -> (CloogStatementIneqs -> [IneqSystem]) -> IO ()
runBarvinokCloogDef outputFile binPath tmpdir projectname typeOfIneqSyst =
	runBarvinokCloog "[[20,0],[0,400]]" "[0,1]" "[True,True]" "../loopo/targout/targout" "../cloog/cloog.git/bin/cloog" "--useCommunication=True"
	  	      	 "--cyclicMapping=True" "--useMPIWrapper=False" "--unpackAfterAlltoall=True" "100" ""
		      	 outputFile binPath tmpdir projectname typeOfIneqSyst



-- test functions -------------------------------------------------------------------------------------------------------------------------------------

-- Tests calcCloogVols
-- typeOfIneqSyst	is the type of the inequality system that should be handled (see NewCommCode.CloogStatementIneqs)
-- file			output-file where the results should be written in
-- tmpdir		directory where the input-files can be found
-- prog			input-file
-- tileI		list of tile-values in the first dimension
-- tileJ		list of tile-values in the second dimension
testTilmann :: (CloogStatementIneqs -> [IneqSystem]) -> String -> [Int] -> [Int] -> String -> String -> IO ()
testTilmann typeOfIneqSyst file tileI tileJ tmpdir prog = do
    appendFile file ("\n\n" ++ prog ++ ":\n")
    ct1 <- getClockTime
    Monad.mapM_ (appendFile file) ["\n\n(i,j)=(" ++ show i ++ "," ++ show j ++ "):\n" ++
				  (show (calcCloogVolsTile ("[[" ++ show i ++ ",0],[0," ++ show j ++ "]]") "[0,1]" "[True,True]" tmpdir
				  			   (prog ++ ".src.out") typeOfIneqSyst))
    				  | i <- tileI, j <- tileJ]
    ct2 <- getClockTime
    appendFile file ("\n" ++ show (diffClockTimes ct2 ct1))

-- Tests runBarvinokCloog
-- typeOfIneqSyst	is the type of the inequality system that should be handled (see NewCommCode.CloogStatementIneqs)
-- file			output-file where the results should be written in
-- tmpdir		directory where the input-files can be found
-- prog			input-file
-- tileI		list of tile-values in the first dimension
-- tileJ		list of tile-values in the second dimension
testBarvinok :: (CloogStatementIneqs -> [IneqSystem]) -> String -> [Int] -> [Int] -> String -> String -> IO ()
testBarvinok typeOfIneqSyst file tileI tileJ tmpdir prog =
	let binPath = "/public/packages/libraries/barvinok-0.22/bin/barvinok_enumerate"
	in do
    appendFile file ("\n\n" ++ prog ++ ":\n")
    ct1 <- getClockTime
		
    Monad.mapM_ (runBarvinokCloogTile "[0,1]" "[True,True]" file binPath tmpdir (prog ++ ".src.out") typeOfIneqSyst)
    		["[[" ++ show i ++ ",0],[0," ++ show j ++ "]]" | i <- tileI, j <- tileJ]
    ct2 <- getClockTime
    appendFile file ("\n" ++ show (diffClockTimes ct2 ct1))



-- auxiliary functions ---------------------------------------------------------------------------------------------------------------------------------

-- Returns a list of inequality-systems of the given project
-- tmpdir       	is the directory where the out-files of LooPo are, e.g. /tmp/loopo4711
-- projectname  	is the leading part of the out-file with src.out, e.g. adder.src.out
getLoopoDSIspcs :: String -> String -> [IneqSystem]
getLoopoDSIspcs tmpdir projectname = 
    let ds	= LoopoDS.unsafeReadLoopoDS (tmpdir ++ "/" ++ projectname)	-- load loopo-datastructure out of the given project-file...
    in	List.map Prelude.snd (LoopoDS.loopoISPC ds)				-- ... and get the indexspaces out of it

-- Returns a list of inequality-systems of the given type when codegen is called with the given command-line
-- commandLine		list of parameters for codegen (see runBarvinokCloog)
-- typeOfIneqSyst	is the type of the inequality system that should be handled (see NewCommCode.CloogStatementIneqs)
getCloogIspcs :: [String] -> (CloogStatementIneqs -> [IneqSystem]) -> [IneqSystem]
getCloogIspcs commandLine typeOfIneqSyst = typeOfIneqSyst (unsafeGetCloogIneqs (parseCommandLine commandLine))



-- following functions are taken out of Rabltest.hs -----------------------------------------------------------------------------------------------------

myVol :: IneqSystem -> DTree (Polynomial Rational) (FracPoly Rational)
myVol ies = makeDTree (fracPolyDTreeContext () []) (volume (getMatrix (stdIneq2FPIneq ies)))

stdIneq2FPIneq (ies :: IneqSystem ) =
    let mat :: GMatrix Rational
        mat = getMatrix ies
	matFracPoly :: GMatrix (FracPoly Rational)
	matFracPoly = convertMatrixEntries convertToFracPoly mat
	varMat :: GMatrix (FracPoly Rational)
	varMat = subMatrixVer matFracPoly 0 (nVars ies-1)
        paramList :: [FracPoly Rational]
	paramList = [fromPoly (indeterminate p i) | 
		     let p = LMath.nParams ies, i<-[0..p-1]] ++
		    [convertToFracPoly 1]
	paramConv :: [Rational] -> FracPoly Rational
	paramConv paramPartCoeff = 
	    sum (zipWith (*) (map fromRational paramPartCoeff) paramList)
	paramMat :: GMatrix Rational 
	paramMat = subMatrixVer mat (nVars ies) (nColumns mat)  -- -1
	convertedParamMat :: GMatrix (FracPoly Rational)
	convertedParamMat = columnMatrix (map paramConv (rows' paramMat))
	combinedResMat ::  GMatrix (FracPoly Rational)
	combinedResMat = varMat `conc` convertedParamMat
	resIes :: GIneqSystem (FracPoly Rational)
	resIes = ineqSystem((nVars ies,0),combinedResMat)
    in resIes
