#
# Finds specially (user-)marked loops in a .java file
# and calls TaskLoops, which replaces them by the parallel version
# generated by LooPo.
#
# Usage: ruby extract_loops.rb
#
require 'getoptlong'
require 'tempfile'
require 'ftools'

# Controls output behaviour
$verbose = 1 # (where 0=disabled, 1=normal, 2=debug, 3=debug+output)

# Markup commands:
LOOPO_BEGIN_LOOP_MARK = 'loopo\s+begin\s+loop' # "loopo begin loop"
LOOPO_END_LOOP_MARK   = 'loopo\s+end\s+loop'   # "loopo end loop"
LOOPO_END_DECLARATIONS_MARK =
  "loopo\s+end\s+declarations" # "loopo end declarations"
LOOPO_CONSTANTS_DECL     = 'constants'
LOOPO_IN_ARRAYS_DECL     = 'arrays\s*\[\s*IN\s*\]'
LOOPO_OUT_ARRAYS_DECL    = 'arrays\s*\[\s*OUT\s*\]'
LOOPO_IN_OUT_ARRAYS_DECL = 'arrays(\s*\[\s*IN-OUT\s*\])?'

DRAIN_SOAK_VAR_PREFIX = "i_" # "loopo_soak_drain_var_"

# workaround for a former LooPo bug (no longer needed)
# (this is only relevant for the P2P-code)
ENABLE_PUMP_WORKAROUND = false

# helper function to run 'eval' in a (primitive) sandbox
def sandbox_eval(expr)
  result = nil
  t = Thread.start { $SAFE = 4 ; result = eval(expr) }
  abort "sandbox_eval: timeout (hacker attack?)" unless t.join(10)
  result
end

class Extractor

  def initialize(input_filename)
    begin
      puts "Reading input file '#{input_filename}'." if $verbose > 0
      @input = IO.readlines(input_filename)
    rescue IOError
      $stderr << "IOError while reading the input file '#{input_filename}'.\n"
      $stderr << "aborting...\n"
      exit 1
    end

    @num_soak_stmts  = -1
    @num_drain_stmts = -1

    @loop, @init_decls, @preloop, @postloop, @constants, @arrays =
      extractLoop(@input)
  end

  def replaceLoops(output_path, tempfile_prefix)
    if @loop == nil
      puts "Error: No marked loops found!"
      exit 1
    else
      runLoopo(tempfile_prefix)

      if $verbose > 0
	puts("Copying output files from the temporary directory " +
	     "to the output directory '#{output_path}':")
      end

      File.makedirs(output_path)
      for file in Dir[$tempdir_path + "/*.java"]
	short_name = File.basename(file)
	if File.copy(file, output_path + "/" + short_name)
	  puts "  #{short_name} copied." if $verbose > 0
	else
	  puts "Error! Could not copy file #{file}."
	  puts "continuing..."
	end
      end
    end
  end

  private
  def runLoopo(tempfile_prefix)

    if tempfile_prefix && tempfile_prefix != ""
      auto_delete_file = false
    else
      auto_delete_file = true
      tempfile_prefix = "loopo_tmp_file"
    end

    # write .preloop file
    preloop_buffer = @preloop
    preloop_file = writeToTempfile(preloop_buffer, tempfile_prefix,
				   "preloop", auto_delete_file)

    # write .postloop file
    postloop_buffer = @postloop
    postloop_file = writeToTempfile(postloop_buffer, tempfile_prefix,
				    "postloop", auto_delete_file)

    # write .init_delcs file
    init_decls_buffer = @init_decls
    init_decls_file = writeToTempfile(init_decls_buffer, tempfile_prefix,
				      "init_decls", auto_delete_file)

    # write .arrays file
    arrays_buffer = [@arrays["ALL"].length] +
                    @arrays["IN"].map     { |x| "#{x[0]}, [#{x[1].join(", ")}], IN"    } +
                    @arrays["OUT"].map    { |x| "#{x[0]}, [#{x[1].join(", ")}], OUT"   } +
                    @arrays["IN-OUT"].map { |x| "#{x[0]}, [#{x[1].join(", ")}], INOUT" }
    arrays_file = writeToTempfile(arrays_buffer,
                                  $tempdir_path + "/project.src.out",
                                  "arrays", false)

    # write .src file (will not be deleted)
    input_arrays  = @arrays["IN"]  + @arrays["IN-OUT"]
    output_arrays = @arrays["OUT"] + @arrays["IN-OUT"]

    add_drain_soak_stmts = $hoc_type == "p2p"

    if $verbose > 0
      if add_drain_soak_stmts
	puts "Adding soak/drain statements"
      else
	puts "Do not add soak/drain statements"
      end
    end
    src_buffer = createSrcFile(@constants, @loop,
                               input_arrays, output_arrays, add_drain_soak_stmts)
    src_file = writeToTempfile(src_buffer, $tempdir_path + "/project",
			       "src", false)
    unless @num_soak_stmts >= 0 && @num_drain_stmts >= 0
      abort "internal error: num_soak_stmts/num_drain_stmts undefined"
    end

    # Call LooPo to prepare the intermediate LooPo files needed by TaskLoops.
    puts "Starting LooPo." if $verbose > 0

    # Workaround for a bug in LooPo's batchmode
    # (It hangs if the fifo file already exists)
    if File.exists?($tempdir_path + "/fifo")
      if $verbose > 0
	puts "Warning: 'fifo' file does already exists."
	puts "It will now be deleted, but generally it is NOT recommended"
	puts "to reuse existing temporary directories."
      end
      File.delete($tempdir_path + "/fifo")
    end
    loopo_call = "#{$loopo_bin} --batch --config #{$config_file} " +
      "--tempdir #{$tempdir_path} --no-cleanup --project #{src_file.path}"
    puts loopo_call if $verbose >= 2

    unless system_(loopo_call) # prints nothing to stdout if $verbose == 0
      puts "Error! Call to LooPo failed."
      puts "(#{loopo_call})"
      exit 1
    end

    #    "/tmp/loopo23456/example.src"
    # -> "example.src.out"
    project_name = File.basename(src_file.path).to_s + ".out"

    # Call TaskLoops to generate the target java files.
    flags = ENV["TASKLOOP_FLAGS"]
    if $verbose > 0
      puts "Starting taskloops."
      puts "(with TASKLOOP_FLAGS='#{flags}')" if flags
    end
    
    optional_hoc_type = $hoc_type ? " --hoc-type=#{$hoc_type} " : ""
    optional_package_name = $package_name ? " --package #{$package_name} " : ""
    optional_storage_type = $storage_type ? " --storage #{$storage_type} " : ""
    tiling_option1 = (0..$tiling_vectors.length-1).to_a.inspect
    tiling_option2 = "[" + (["True"] * $tiling_vectors.length).join(", ") + "]"
    optional_flags =  flags ? " #{flags} " : ""

    taskloops_call = "#{$taskloops_bin} --pre-file=\"#{preloop_file.path}\" " +
                     " --post-file=\"#{postloop_file.path}\" " +
                     " --init-decls=\"#{init_decls_file.path}\" " +
                     " --num-soak-stmts=#{@num_soak_stmts} " +
                     " --num-drain-stmts=#{@num_drain_stmts} " +
                     optional_hoc_type +
                     optional_package_name +
                     optional_storage_type +
                     optional_flags +
                     " \"#{$tempdir_path}\" \"#{project_name}\" " +
                     " \"#{$tiling_vectors.inspect}\" \"#{tiling_option1}\" " +
                     " \"#{tiling_option2}\" \"#{$targout_bin}\"" +
                     " \"#{$cloog_bin}\" 1 \"\""
    puts taskloops_call if $verbose >= 2
    unless system_(taskloops_call) # prints nothing to stdout if $verbose == 0
      puts "Error! Call to taskloops failed."
      puts "(#{taskloops_call})"
      exit 1
    end
    puts "LooPo finished." if $verbose > 0
  end

  def extractLoop(input)
    puts "Extracting marked loops." if $verbose > 0

    loop       = nil
    init_decls = nil
    pre        = nil
    post       = nil
    constants  = nil
    arrays     = nil
    
    for linecounter in (0..input.length-1)
      nextline = input[linecounter]
      if nextline =~ /\/\/\s*#{LOOPO_BEGIN_LOOP_MARK}/
	# Marked loop found
	# 1) parse declarations
	#    (e.g., "constants: n,m,l; arrays: A{size1,size2}")
	if $verbose >= 2
	  puts "Marked loop detected (in line ##{linecounter + 1})."
	end
	errMsg = nil
	if $' =~ /\([^\(]*\)/
	  # constants or array declarations found
	  constants, arrays = parseLoopDeclarations($&, linecounter+1)
	else
	  puts "No declarations found." if $verbose >= 2
	end

	# 2) find end of the marked loop
	end_of_loop = nil
	for i in (linecounter+1..input.length-1)
	  if input[i] =~ /\/\/\s*#{LOOPO_END_LOOP_MARK}/
	    end_of_loop = i
	    break
	  end
	end
	unless end_of_loop
	  puts("Error! Missing end loop label for marked loop " +
	       "in line #{linecounter+1}.")
	  exit 1
	end
	
	# 3) verify that there is only one marked loop
	for i in (linecounter+1..input.length-1)
	  if input[i] =~ /\/\/\s*#{LOOPO_BEGIN_LOOP_MARK}/
	    puts("Error in line #{i+1}! Second marked loop detected.\n" +
		 "(The first one is in line #{linecounter+1}.)")
	    exit 1
	  end
	end

	unparsed_pre = input[0..linecounter-1]
	loop = input[linecounter+1..end_of_loop-1]
	post = input[end_of_loop+1..-1]

	# parse initial declarations
	init_decls, pre = parseInitialDeclarations(unparsed_pre)

	# error: The LooPo cannot handle (+=, -=, *=, /=, %=)
	if loop.join("\n") =~  /[\+\-\*\/\%]\=/
	  puts "Error! LooPo does not understand the '#{$&}' operator."
	  puts "(Hint: Rewrite it as \"A = A #{$&[0..0]} B\")"
	  exit 1
        end
	if loop.join("\n") =~ /for\s*\(\s*int\s+/
	  if $verbose > 0
	    puts("*"*70)
	    puts("Warning:\nConverting Java style loops to LooPo style " +
		 "(without the integer declaration).")
	    puts "for (int i = ...) // Java"
	    puts "for (i = ...)     // LooPo"
	    puts("*"*70)
	  end
	  loop = loop.join.gsub(/for\s*\(\s*int\s+/, "for (").split("\n")
	end
	return [loop, init_decls, pre, post, constants, arrays]
      end
    end

    # No marked loop found
    return [loop, init_decls, pre, post, constants, arrays]
  end

  # Parse declarations of the following form:
  #   "(constants: n, m, l; \"
  #   " arrays [IN]: A{size1, size2}, B{size3}; \"
  #   " arrays [OUT]: C{size4})"
  #
  # (Note: the parameter 'line' is only used for the error message)
  def parseLoopDeclarations(full_decl, line)
    constants = []
    arrays    = { "IN" => [], "OUT" => [], "IN-OUT" => [], "ALL" => [] }

    # remove '(' and ')' and tokenize the string
    full_decl = full_decl[1..full_decl.length-2].strip
    single_decls = full_decl.split(";").map { |x| x.strip }

    for decl in single_decls
      array_type = nil
      if (decl =~ /#{LOOPO_CONSTANTS_DECL}\s*:/) == 0
	constant = true
      elsif (decl =~ /#{LOOPO_IN_ARRAYS_DECL}\s*:/) == 0
	constant = false
	array_type = "IN"
      elsif (decl =~ /#{LOOPO_OUT_ARRAYS_DECL}\s*:/) == 0
	constant = false
	array_type = "OUT"
      elsif (decl =~ /#{LOOPO_IN_OUT_ARRAYS_DECL}\s*:/) == 0
	constant = false
	array_type = "IN-OUT"
      else
	puts "Error! Could not parse declaration string in line #{line}:"
	puts full_decl
	puts "Stopped at: '#{decl}'"
	exit 1
      end
      var_names = $'.split(/\}\s*,/).map { |x| x.chomp("}") }
      if constant
	constants += var_names.map { |x| x.strip }
      else
	# parse dimension ("A{size1, size2}" -> ["A", ["size1", "size2"]])
	var_names.map! do |x|
	  name, dims, empty = x.split("{")
	  unless name && dims && empty == nil
	    puts "Error! Could not parse declaration string in line #{line}:"
	    puts full_decl
	    puts "Stopped at: '#{x}'"
            exit 1
	  end
	  [name.strip, dims.split(",").map { |x| x.strip }]
	end
	arrays[array_type] += var_names
      end
    end

    if $verbose > 0
      puts "constants:       #{constants.inspect}"
      puts "arrays [IN]:     #{arrays["IN"].inspect}"
      puts "arrays [OUT]:    #{arrays["OUT"].inspect}"
      puts "arrays [IN-OUT]: #{arrays["IN-OUT"].inspect}"
    end
    arrays["ALL"] = arrays["IN"] + arrays["OUT"] + arrays["IN-OUT"]
    return constants, arrays
  end

  #
  # Scan through the input before the loop until
  # the LOOPO_END_DECLARATIONS_MARK is reached.
  #
  def parseInitialDeclarations(unparsed_pre)
    decls = []
    loop {
      nextline = unparsed_pre.shift
      if nextline == nil
	puts "Error! Could not find end-of-declarations label."
	exit 1
      elsif nextline =~ /\/\/\s*#{LOOPO_END_DECLARATIONS_MARK}/
	puts "End of initial declarations found." if $verbose >= 2
	return [decls, unparsed_pre]
      else
	decls << nextline
      end
    }
  end

  #
  # pump_loop_dimensions:
  #   Used as a workaround for a LooPo bug.
  #   LooPo fails when there are loops with different dimensions.
  #
  # Note: This workaround is no longer needed.
  #
  def createDrainSoakCode(arrays, pump_loop_dimension)
    ds_stmts = arrays.map do |name, dims|
      if pump_loop_dimension && pump_loop_dimension > dims.length
	pumped_dims = dims + ([1] * (pump_loop_dimension - dims.length))
      else
	pumped_dims = dims
      end
      begin_loops = ""
      end_loops = ""
      arr_expr = name.clone
      for i in (0..pumped_dims.length-1)
	loop_var = DRAIN_SOAK_VAR_PREFIX + i.to_s # + "_" + name
	begin_loops += "#{"  "*i}for (#{loop_var} = 0; " +
	               "#{loop_var} < #{pumped_dims[i]}; #{loop_var}++) {\n"
	end_loops = "#{"  "*i}}\n" + end_loops
	arr_expr += "[#{loop_var}]" if i < dims.length
      end
      assignment = "#{"  "*dims.length}#{arr_expr} = #{arr_expr};\n"
      begin_loops + assignment + end_loops
    end
    code = ds_stmts.join("\n") + "\n"
    count_stmts = arrays.length
    return [code, count_stmts]
  end

  #
  # Create the .src file and add drain/soak statements (if necessary).
  #
  def createSrcFile(constants, loop, input_arrays,
                    output_arrays, add_soak_drain_statements)
    if add_soak_drain_statements
      if ENABLE_PUMP_WORKAROUND
	pump_workaround =
	  loop.join("\n").split(/for\b*\(/).map { |x|
	    if x =~ /for/ then x else nil end
	  }.compact.length
	puts "Warning: LooPo workaround (pump_workaround=#{pump_workaround})"
      else
	pump_workaround = nil
      end

      soak_stmt_code, @num_soak_stmts =
	createDrainSoakCode(input_arrays,pump_workaround)
      drain_stmt_code, @num_drain_stmts =
	createDrainSoakCode(output_arrays, pump_workaround)

      if $verbose >= 2
	puts "(added #{@num_soak_stmts} soak and #{@num_drain_stmts} drain statements)"
      end
    else
      soak_stmt_code  = nil
      drain_stmt_code = nil
      @num_soak_stmts  = 0
      @num_drain_stmts = 0
    end

    buffer = ""
    buffer += "CONSTANT #{constants.join(", ")};\n\n" unless constants.length == 0
    buffer += "/* soak statements: */\n" + soak_stmt_code if soak_stmt_code
    buffer += "/* user code: */\n" + loop.join("\n")
    buffer += "\n\n/* drain statements: */\n" + drain_stmt_code if drain_stmt_code
    return buffer
  end
  
  def writeToTempfile(text, prefix, suffix, auto_delete_file)
    if $verbose >= 3
      puts "--- begin .#{suffix} file ---"
      puts text
      puts "--- end .#{suffix} file ---"
    end

    begin
      filename = prefix + "." + suffix
      unless auto_delete_file
	f = File.open(filename, "w")
	puts "Creating .#{suffix} file '#{filename}'" if $verbose > 0
      else
	f = Tempfile.open(filename)
	filename = f.path
	puts "Creating temporary .#{suffix} file '#{filename}'" if $verbose > 0
      end
      f.puts(text)
      f.flush
      return f
    rescue IOError
      $stderr << "IOError while writing to the .#{suffix} file '#{filename}'.\n"
      exit 2
    end
  end

  def system_(call)
    if $verbose > 0
      system(call)
    else
      output = `#{call} 2>&1`
      $?.exitstatus
    end
  end

end

def checkFiles(files)
  for f in files
    file, desc = f
    unless File.exists?(file)
      puts "Error! #{desc} not found."
      puts "('#{file}' does not exist)."
      exit 2
    end
  end
end

def allEqual(arr)
  throw "allEqual: empty array" if arr.empty?
  arr = arr.clone
  head = arr.shift
  arr.each { |x| return false if x != head }
  return true
end

def parseTilingVector(vect)
  begin
    i  = '\s*[0-9]+\s*'  # integer
    bo = '\s*\[\s*'      # '['
    bc = '\s*\]\s*'      # ']'
    if vect =~ /^#{bo}#{bo}#{i}(,#{i})*#{bc}(,#{bo}#{i}(,#{i})*#{bc})*#{bc}$/
      vect = sandbox_eval(vect)
      return vect if allEqual(vect.map { |x| x.length })
    end
  rescue
    puts "Error! parseTilingVector: internal error!"
    exit 3
  end
  puts "Error! Could not parse tiling vector."
  exit 2
end

def printHelpScreen
  puts "Usage: #{$0}"
  puts "\t--input <file> --output-dir <path> --tempdir <path>"
  puts "\t[--loopo-dir <path>] [--taskloops-bin <file>]"
  puts "\t[--keep-temp-files <file-prefix>]"
  puts
  puts "Required options:"
  puts "--input <file> or -i <file>:"
  puts "\tsets the name of the input java-file."
  puts "--output-dir <path> or -o <path>:"
  puts "\tsets the name of the output path in which the target java-files will be copied."
  puts "--tempdir <path> or -t <path>:"
  puts "\tsets the name of the temporary directory."
  puts "--tiling \"[vector1,..,vectorN]\" or -l \"[vector1,..,vectorN]\":"
  puts "\tsets the tiling vectors"
  puts "\t(e.g., in the matrix example \"[[50,0,0],[0,50,0],[0,0,50]]\")"
  puts
  puts "Optional options:"
  puts "--hoc-type <type> or -h <type>:"
  puts "\tsets the type of HOC, for which code is generated (master-worker, p2p)"
  puts "--package <name>:"
  puts "\toverrides the package name of the generated java files"
  puts "--storage <type>:"
  puts "\toverrides the type of storage strategy (array, hashtable)"
  puts "--loopo-dir <path>:"
  puts "\tsets the name of the LooPo directory (default: '#{DEFAULT_LOOPO_BIN}')"
  puts "--taskloops-bin <file>:"
  puts "\tsets the name of the TaskLoops tool (default: '#{DEFAULT_TASKLOOPS_BIN}')"
  puts "--cloog-bin <file>:"
  puts "\tsets the name of the CLooG binary (default: '#{DEFAULT_CLOOG_BIN}')"
  puts "--loopo-config <file>:"
  puts "\tsets the name of the LooPo config file (default: '#{DEFAULT_LOOPO_CONFIG_FILE}')"
  puts "--keep-temp-files <file-prefix>:"
  puts "\tsets the prefix of the intermediate files."
  puts "\t(otherwise temporary files will be created, "
  puts "\t which will be automatically deleted at the end.)"
  puts "--check-syntax or -c:"
  puts "\tCheck the syntax of the LooPo tags and exit without generating"
  puts "\tthe target code. If there are no errors, it will print"
  puts "\t\"Syntax OK\" to the standard output."
  puts "--verbose <level>"
  puts "\tsets the verbosity level (default: 1)"
  puts "\t(0: silent, 1: normal, 2: debug, 3: extra debug)"
  puts "--help:"
  puts "\tprints this help screen."
end

HOME_DIR = `echo $HOME`.chomp
DEFAULT_LOOPO_BIN         = "#{HOME_DIR}/loopo"
DEFAULT_CLOOG_BIN         = "/home/cl/classenm/cloog/cloog.git/bin/cloog"
DEFAULT_TASKLOOPS_BIN     = "#{HOME_DIR}/hsloopo/taskloops"
DEFAULT_LOOPO_CONFIG_FILE = "#{HOME_DIR}/.looporc"

if ARGV[0] == nil
  printHelpScreen
  exit 0
end

filename        = nil
output_path     = nil
tempfile_prefix = nil

$loopo_dir      = DEFAULT_LOOPO_BIN
$taskloops_bin  = DEFAULT_TASKLOOPS_BIN
$config_file    = DEFAULT_LOOPO_CONFIG_FILE
$cloog_bin      = DEFAULT_CLOOG_BIN

$hoc_type       = nil
$storage_type   = nil
$package_name   = nil
$tempdir_path   = nil
$loopo_bin      = nil
$targout_bin    = nil
$tiling_vectors = nil

stop_after_syntax_check = false

opts = GetoptLong.new(
  ["--input",            "-i", GetoptLong::REQUIRED_ARGUMENT],
  ["--output-dir",       "-o", GetoptLong::REQUIRED_ARGUMENT],
  ["--tempdir",          "-t", GetoptLong::REQUIRED_ARGUMENT],
  ["--tiling",           "-l", GetoptLong::REQUIRED_ARGUMENT],
  ["--hoc-type",         "-h", GetoptLong::REQUIRED_ARGUMENT],
  ["--package",                GetoptLong::REQUIRED_ARGUMENT],
  ["--storage",                GetoptLong::REQUIRED_ARGUMENT],
  ["--loopo-dir",              GetoptLong::REQUIRED_ARGUMENT],
  ["--taskloops-bin",          GetoptLong::REQUIRED_ARGUMENT],
  ["--cloog-bin",              GetoptLong::REQUIRED_ARGUMENT],
  ["--loopo-config",           GetoptLong::REQUIRED_ARGUMENT],
  ["--keep-temp-files",        GetoptLong::REQUIRED_ARGUMENT],
  ["--check-syntax",     "-c", GetoptLong::NO_ARGUMENT],
  ["--verbose",                GetoptLong::REQUIRED_ARGUMENT],
  ["--help",                   GetoptLong::NO_ARGUMENT]
)

opts.each do |opt, arg|
  case opt
    when "--input", "-i"
      filename = arg
    when "--output-dir", "-o"
      output_path = arg
    when "--tempdir", "-t"
      $tempdir_path = arg
    when "--tiling", "-l"
      $tiling_vectors = parseTilingVector(arg)
    when "--hoc-type"
      $hoc_type = arg
    when "--package"
      $package_name = arg
    when "--storage"
      $storage_type = arg
    when "--loopo-dir"
      $loopo_dir = arg
    when "--taskloops-bin"
      $taskloops_bin = arg
    when "--cloog-bin"
      $cloog_bin = arg
    when "--loopo-config"
      $config_file = arg
    when "--loopo-config"
      $config_file = arg
    when "--keep-temp-files"
      tempfile_prefix = arg.strip
    when "--check-syntax", "-c"
      stop_after_syntax_check = true
    when "--verbose"
      unless arg =~ /^[0-9]+$/
	puts "Error! Could not parse verbosity level ('#{arg}')."
	exit 1
      end
      $verbose = arg.to_i
    when "--help"
      printHelpScreen
      exit 0
  end
end

# Check required command line arguments
unless filename
  puts "Error! Missing input file (--input)."
  exit 1
end

if stop_after_syntax_check
  checkFiles([[filename, "Input file"]])
  ext = Extractor.new(filename)
  puts "Syntax OK"
  exit
end

# Check remaining required command line arguments
unless output_path
  puts "Error! Missing output path (--output-dir)."
  exit 1
end
unless $tempdir_path
  puts "Error! Missing tempdir path (--tempdir)."
  exit 1
end
unless $tiling_vectors
  puts "Error! Tiling vectors missing (--tiling)."
  exit 1
end

$loopo_bin   = "#{$loopo_dir}/loopo"
$targout_bin = "#{$loopo_dir}/targout/targout"

# Checking whether all files exist
checkFiles([[filename, "Input file"] \
	   ,[$loopo_bin, "LooPo binary"] \
	   ,[$targout_bin, "targout binary"] \
           ,[$cloog_bin, "CLooG binary"] \
           ,[$config_file, "LooPo config file"]
           ])

puts "Creating temp directory ('#{$tempdir_path}')." if $verbose >= 2
File.makedirs($tempdir_path)

ext = Extractor.new(filename)
ext.replaceLoops(output_path, tempfile_prefix)
