{-
    This file is part of the LooPo project, see
    <http://www.infosun.fim.uni-passau.de/cl/loopo/>.

    Copyright (C) 2009 Armin Groesslinger
    Author: Armin Groesslinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-}

{-# LANGUAGE NoMonomorphismRestriction #-}
module CADGen(main)
where

import Config
import LMath
import LMath.Types.Polynomial.AlgNum (AlgNum)
import LP

import Debug.Trace
import Control.Exception
import Control.Monad
import Data.Char
import Data.List
import Data.Maybe
import System
import System.Console.GetOpt
import System.IO
import Text.ParserCombinators.Parsec

default ()

version :: String
version = "0.1 (2009-09-30)"

data CInput = CInput { cpScatterNames :: [String]
                     , cpIterNames :: [String]
                     , cpParamNames :: [String]
                     , cpCtx :: QfFormula (Polynomial Rational)
                     , cpDoms :: [QfFormula (Polynomial Rational)]
                     , cpScatter :: [QfFormula (Polynomial Rational)]
                     }
              deriving Show

data COptions = COptions { coCodeSimpl :: Bool
                         , coTestCode  :: Bool
                         , coTestParamVal :: Maybe (Either Integer [Integer])
                         , coInternalProj :: Bool
                         , coAlgImpl :: AlgImpl
                         }
                deriving Show

parseCADGenInput :: Parser CInput
parseCADGenInput =
    do spaces
       string "cadgen"
       spaces
       nP <- parseInt
       paramNames <- parseN nP name
       ctx <- ourParseFormula paramNames
       nI <- parseInt
       iterNames <- parseN nI name
       nDoms <- parseInt
       doms <- parseN nDoms $
                do dim <- parseInt  -- dimensionality of iterator space
                   when (dim > genericLength iterNames) $
                        fail "Not enough iterator names"
                   ourParseFormula (genericTake dim iterNames ++ paramNames)
       let domDim i = unNIndeterminates $
                      indeterminatesInQf (doms `genericIndex` i)
       nS <- parseInt
       scatterNames <- parseN nS name
       nScatter <- parseInt
       when (nScatter /= nDoms) $
            fail "number of scatter relations must match number of domains"
       scatter <- sequence
                  [ourParseFormula names 
                  | i <- [0..nScatter-1]
                  , let names = scatterNames ++
                                genericTake (domDim i - nP) iterNames ++ 
                                paramNames
                  ]
       return $ CInput { cpScatterNames = scatterNames 
                        , cpIterNames  = iterNames
                        , cpParamNames = paramNames
                        , cpCtx        = ctx
                        , cpDoms       = doms
                        , cpScatter    = scatter
                        }

name :: Parser String
name = do spacesOrComment
          x  <- letter <|> char '_'
          xs <- many (char '_' <|> alphaNum)
          return (x:xs)

spacesOrComment :: Parser ()
spacesOrComment =
    do many $ ((many1 space >> return ()) <|>
	       (char '#' >> manyTill anyToken ((newline >> return ())
					       <|> eof
					      ) >>
		return ()
	       )
	      )
       return ()

parseInt :: Parser Integer
parseInt = spacesOrComment >> parseInteger


ourParseFormula :: [String] -> Parser (QfFormula (Polynomial Rational))
ourParseFormula names =
    do spacesOrComment
       qf1 <- parseQfFormula (parsePoly names)
              :: Parser (QfFormula (Polynomial Integer))
       let qf = fmap (fmap' fromIntegral) qf1
       spaces
       char ';'
       return qf
       

handleToolIOError :: IO a -> IO ()
handleToolIOError action =
    catchDyn (action >> return ())
             (\(tio :: ToolIOError) ->
                 do hPutStrLn stderr ("Tool IO error: " ++ show tio)
             )

output :: String -> IO ()
output str =
    handleToolIOError $ putStrLn str


data Method = Internal
            | CADbyGEO
              deriving (Eq, Show)

data AlgImpl = AlgInternal
             | AlgQepcad
               deriving (Eq, Show)

data CADGenOpt = OutputFile FilePath
               | Method Method
               | NoCodeSimpl
               | TestCode (Maybe String)
               | InternalCAD
               | ReducePath FilePath
               | QepcadPath FilePath
                 deriving (Eq, Show)

options :: [OptDescr CADGenOpt]
options = [Option ['o'] ["output"] (ReqArg OutputFile "OUTPUT") "write output to OUTPUT"
          ,Option [] ["no-combine"] (NoArg NoCodeSimpl) "do not combine adjacent cells"
          ,Option [] ["testcode"] (OptArg TestCode "parameter values") "generate test code"
          ,Option [] ["reduce"] (ReqArg ReducePath "REDUCE") "path to Reduce executable"
          ,Option [] ["qepcad"] (ReqArg QepcadPath "QEPCAD") "path to Qepcad executable"
          ]

method :: String -> Method
method "internal" = Internal
method "qep"      = CADbyGEO

main :: IO ()
main = do args <- getArgs
          mainA args

mainA :: [String] -> IO()
mainA args =
          case getOpt Permute options args of
            (opts,[infile],[]  ) -> handleToolIOError $ run infile opts
            (_,_,errs) ->
                do binary <- getProgName
                   let header = "Usage: " ++ binary ++
                                " [OPTION...] file.cadgen"
                   hPutStrLn stderr (concat errs ++ usageInfo header options)

run :: FilePath -> [CADGenOpt] -> IO ()
run input opts =
    do case mbReduce of
         Just reduceCmd -> Config.setExternalToolPath Config.REDUCE reduceCmd
         Nothing        -> return ()
       case mbQepcad of
         Just qepcadCmd -> Config.setExternalToolPath Config.QEPCAD qepcadCmd
         Nothing        -> return ()
       p <- parseFromFile parseInput input
       when (coAlgImpl co == AlgInternal) $
            do hPutStrLn stderr "Note: Qepcad not available, falling back to \
                                \internal algebraic number implementation.\n"
       case p of
         Right cp -> outputFn $ makeCode co cp
         Left err ->
             hPutStrLn stderr ("Parse error: " ++ show err)
    where
    parseInput =
        (lookAhead (spaces >> string "cadgen") >> parseCADGenInput)
        <|>
        parseNLGen
    parseNLGen =
            do (lgp,lgoo,nlCtx) <- parseLGenInput
               let nPlin :: Integer
                   nPlin = nParams $ lgpContext lgp
                   (doms,scatter) = unzip $ lgpStatements lgp
                   domQfs = [disjunctionQf [ineqs2qf d | d <- u]
                            | u <- doms]
                   scatterQfs = [ineqs2qf s | s <- scatter]
                   linCtx = ineqs2qf $ lgpContext lgp
               
               return $ CInput { cpScatterNames = lgooVarNames lgoo
                                , cpIterNames    = lgooIterNames lgoo
                                , cpParamNames   = lgooParamNames lgoo ++
                                                   lgooNonlinParamNames lgoo
                                , cpDoms = domQfs
                                , cpScatter = scatterQfs
                                , cpCtx = fmap (liftPolyN nPlin) $
                                          conjunctionQf (linCtx : nlCtx)
                                }

    outputFn = head ([writeFile out | OutputFile out <- opts] ++ [putStrLn])
    m = head ([m | Method m <- opts] ++ [Internal])

    co = COptions { coCodeSimpl = NoCodeSimpl `notElem` opts
                  , coTestParamVal = do
                      s <- listToMaybe [v | TestCode (Just v) <- opts]
                      case (reads s, reads s) of
                         ((v,""):_, _)  -> return (Left v)
                         (_, (vs,""):_) -> return (Right vs)
                         _              -> Nothing
                  , coTestCode = not $ null [v | TestCode v <- opts]
                  , coInternalProj = True -- InternalCAD `elem` opts
                  , coAlgImpl = if externalToolAvailable QEPCAD
                                then AlgQepcad
                                else AlgInternal
                  }

    mbReduce = listToMaybe [r | ReducePath r <- opts]
    mbQepcad = listToMaybe [q | QepcadPath q <- opts]

ineqs2qf :: RealFrac a => GIneqSystem (FracPoly a) -> QfFormula (Polynomial a)
ineqs2qf ineqs =
    ineqSystemToFormula $ unfracIneqSystem Nothing ineqs


----------------------
-- Test Code Output --
----------------------

makeCode :: COptions -> CInput -> String
makeCode co cp
    | coTestCode co  =  mainProg
    | otherwise      =  loops
    where
    loops :: String
    loops = length code `seq`
            ("/* Code generated by CADGen version " ++ version ++ ". */\n" ++
             code
            )
    code =
        code2text nP (paramNames ++ scatterNames ++ iterNames) $
        cadcodeFn ctx fulltispcs

    paramNames = cpParamNames cp
    iterNames  = cpIterNames cp
    scatterNames = cpScatterNames cp
    ctx        = cpCtx cp
    ispcs      = cpDoms cp
    scatter    = cpScatter cp

    nS, nP :: Integer
    nS = genericLength scatterNames
    nP = genericLength paramNames

    fulltispcs = zipWith mkFullDim ispcs scatter

    mkFullDim d s = fmap (liftPolyN nS) d `qfAnd` s

    cadcodeFn ctx ftispcs
        | coCodeSimpl co  =  codeSimpl code'
        | otherwise       =  code'
        where
          code' = case coAlgImpl co of
                    AlgQepcad   -> code (algDummy :: QAlg)
                    AlgInternal -> code (algDummy :: AlgNum Rational)
          code algImpl = cadgen cadMethod nP ctx nS ftispcs algImpl
                 
          cadMethod = if coInternalProj co then CADInternal else CADQepcadGeo

    -- dummy parameter to select Alg implementation
    -- algDummy :: AlgNum Rational
    algDummy :: Alg alg => alg
    algDummy = error "cadgen should not evaluate the dummy parameter"

    mainProg =
        unlines
        (
         ["#include <stdlib.h>"
         ,"#include <stdio.h>"
         ,"#include <math.h>"
         ,""
         ,"/* CADGen generates positive denominators only. */"
         ,"static inline int ceild(int a, int b) { return (a>=0) ? ((a+b-1)/b) : (a/b); }"
         ,"static inline int floord(int a, int b) { return (a>=0) ? (a/b) : ((a-b+1)/b); }"
         ,""
         ,"static inline int min(int a, int b) { return a<b ? a : b; }"
         ,"static inline int max(int a, int b) { return a>b ? a : b; }"
         ,""
         ,"/* floorRootOf and ceilRootOf are only defined for degree 2 (using"
         ," * floating point); for arbitrary degrees root isolation is needed"
         ," * (can be implemented with exact arithmetic)."
         ," */"
         ,""
         ,"static inline double sq(int root, int a0, int a1, int a2) {"
         ,"  double D=a1*a1 - 4*a2*a0;"
         ,"  double s = (root==1 || root==-2) ? -sqrt(D) : sqrt(D);"
         ,"  return (-a1+s)/(2*a2);"
         ,"}"
         ,""
         ,"static int floorRootOf(int root, int deg, int a0, int a1, int a2, ...) {"
         ,"  if (deg != 2) {"
         ,"    fprintf(stderr, \"Sorry, root isolation not implemented.\\n\");"
         ,"    exit(1);"
         ,"  }"
         ,"  return (int)floor(sq(root,a0,a1,a2));"
         ,"}"
         ,""
         ,"static int ceilRootOf(int root, int deg, int a0, int a1, int a2, ...) {"
         ,"  if (deg != 2) {"
         ,"    fprintf(stderr, \"Sorry, root isolation not implemented.\\n\");"
         ,"    exit(1);"
         ,"  }"
         ,"  return (int)ceil(sq(root,a0,a1,a2));"
         ,"}"
         ,""    
         ,"inline int OneOf(int a, ...) { return a; }"
         ,""
         ] ++
         ["inline void S" ++ show (i+1) ++ "(" ++
          concat (intersperse ", " ["int x"++show x | x <- [0..stmtDims i-1]])
          ++ ") {\n  printf(\"S" ++ show (i+1) ++ "(" ++
          concat (intersperse "," $ genericReplicate (stmtDims i) "%d") ++
          ")\\n\"" ++ concatMap (", " ++) ['x':show x | x <- [0..stmtDims i-1]]
          ++ ");\n}"
         | i <- [0..nStmts-1]] ++
         [""
         ,"void cadgen(" ++ funargs ++ ") {"
         ,varIterDecl
         ,indent 2 loops
         ,"}"
         ,""
         ,"int main(int argc, char *argv[]) {"
         ,paramDecl
         ] ++
         ["  " ++ p ++ " = " ++ getParamVal i ++ ";"
              | (p,i) <- zip paramNames [1..]
         ] ++
         [""
         ,"  cadgen(" ++ callargs ++ ");"
         ,""
         ,"  return 0;"
         ,"}"
         ]
        )

    getParamVal i =
        case coTestParamVal co of
          Nothing -> "atoi(argv[" ++ show i ++ "])"
          Just (Left  v)  -> show v
          Just (Right vs) -> show (vs !! (i-1))

    nStmts = length (cpDoms cp)

    stmtDims n = unNIndeterminates (indeterminatesInQf qf) - nP
        where
          qf = cpDoms cp !! n

    varIterDecl
        | null (scatterNames++iterNames) = ""
        | otherwise  =
            "  int " ++ concat (intersperse ", " (scatterNames++iterNames))
            ++ ";"

    callargs
        | null paramNames = ""
        | otherwise       = concat (intersperse ", " paramNames)
    paramDecl
        | null paramNames = ""
        | otherwise       = "  int " ++ callargs ++ ";"
    funargs   = concat $ intersperse ", " ["int " ++ p | p <- paramNames]
        

indent :: Int -> String -> String
indent n str =
    unlines $
    map (replicate n ' ' ++) $
    lines str