{-
    This file is part of the LooPo project, see
    <http://www.infosun.fim.uni-passau.de/cl/loopo/>.

    Copyright (C) 2009 Armin Groesslinger
    Author: Armin Groesslinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-}

{-# LANGUAGE NoMonomorphismRestriction #-}
module NLGen ( main )
where

import Control.Exception
import Data.Maybe ( listToMaybe )
import System.Console.GetOpt
import System.Environment
import System.IO
import System.IO.Unsafe ( unsafePerformIO )
import Text.ParserCombinators.Parsec

import Data.List
import Debug.Trace

import LMath
import LMath.Algo.NLGen
import LMath.Algo.OmegaTest


default (Integer)

version :: String
version = "1.0 (2009-09-10)"

data RunOpts = RunOpts { roTestCode :: Bool
                       , roTestParamVal :: Maybe (Either Integer [Integer])
                       , roCLooGCode :: Maybe String
                       }
             deriving Show

runLGen :: RealFrac a
        => LGenOptions
        -> LGenParams (FracPoly a)
        -> LGenOutputOptions (FracPoly a)
        -> [QfFormula (Polynomial a)]
        -> Integer
        -> String
runLGen lgo lgp lgoo nlCtx level =
    runLGenWithStatements lgo lgfFracPoly lgp lgoo
      (fracPolyDTreeContext nlCtx)
      (showPolyC (lgooNonlinParamNames lgoo) . snd . ratPoly2intPoly)
      level


makeCode :: RealFrac a
         => RunOpts
         -> LGenOptions
         -> LGenParams (FracPoly a)
         -> LGenOutputOptions (FracPoly a)
         -> [QfFormula (Polynomial a)]
         -> Integer
         -> String
makeCode ro lgo lgp lgoo nlCtx level
    | roTestCode ro  =  mainProg
    | otherwise      =  loops
    where
    loops :: String
    loops = runLGen lgo lgp lgoo nlCtx level

    lgoo' = if roTestCode ro
              then lgoo { lgooStmtGen = testStmtGen showExpr }
              else lgoo
    showExpr e =
--          trace (show e) $
          case (lgooAffineExpToString lgoo) names nP e of
            (num, Left  den) -> "( " ++ num ++ ")/(" ++ den ++ ")"
            (num, Right   1) -> num
            (num, Right den) -> "(" ++ num ++ ")/" ++ show den
          where
          names = varNames ++ iterNames ++ paramNames
 
    varNames   = lgooVarNames lgoo
    iterNames  = lgooIterNames lgoo
    paramNames = lgooParamNames lgoo
    nP         = genericLength paramNames

    mbCloogCode = roCLooGCode ro
    withCloog = mbCloogCode /= Nothing
    Just cloogCode = mbCloogCode

    mainProg =
        unlines
        (
         ["#include <stdlib.h>"
         ,"#include <stdio.h>"
         ,""
         ,"/* We assume that denominators are positive. */"
         ,"inline int ceild(int a, int b) { return (a>=0) ? ((a+b-1)/b) : (a/b); }"
         ,"inline int floord(int a, int b) { return (a>=0) ? (a/b) : ((a-b+1)/b); }"
         ,""
         ,"inline int min(int a, int b) { return a<b ? a : b; }"
         ,"inline int max(int a, int b) { return a>b ? a : b; }"
         ,""
         ] ++
         ["inline void S" ++ show (i+1) ++ "(" ++
          concat (intersperse ", " ["int x"++show x | x <- [0..stmtDims i-1]])
          ++ ") {\n  printf(\"S" ++ show (i+1) ++ "(" ++
          concat (intersperse "," $ replicate (stmtDims i) "%d") ++
          ")\\n\"" ++ concatMap (", " ++) ['x':show x | x <- [0..stmtDims i-1]]
          ++ ");\n}"
         | i <- [0..nStmts-1]] ++
         (if withCloog then
              [""
              ,"void cloog(" ++ funargs ++ ") {"
              ,varIterDecl
              ,indent 2 cloogCode
              ,"}"
              ]
          else
              [""]
         ) ++
         [""
         ,"void nlgen(" ++ funargs ++ ") {"
         ,varIterDecl
         ,indent 2 loops
         ,"}"
         ,""
         ,"int main(int argc, char *argv[]) {"
         ,paramDecl
         ] ++
         ["  " ++ p ++ " = " ++ getParamVal i ++ ";"
              | (p,i) <- zip paramNames [1..]
         ] ++
         (if withCloog then
              ["  if (atoi(argv[1]) == 0)"
              ,"    cloog(" ++ callargs ++ ");"
              ,"  else"
              ,"    nlgen(" ++ callargs ++ ");"
              ]
          else
              ["  nlgen(" ++ callargs ++ ");"]
         ) ++
         [""
         ,"  return 0;"
         ,"}"
         ]
        )

    getParamVal i =
        case roTestParamVal ro of
          Nothing -> "atoi(argv[" ++ show (i+1) ++ "])"
          Just (Left  v)  -> show v
          Just (Right vs) -> show (vs !! (i-1))

    nStmts = length (lgpStatements lgp)

    stmtDims n = nVars is
        where
          (is:_,scatter) = lgpStatements lgp !! n
        

    varIterDecl
        | null (varNames++iterNames) = ""
        | otherwise  =
            "  int " ++ concat (intersperse ", " (varNames++iterNames)) ++ ";"

    callargs
        | null paramNames = ""
        | otherwise       = concat (intersperse ", " paramNames)
    paramDecl
        | null paramNames = ""
        | otherwise       = "  int " ++ callargs ++ ";"
    funargs   = concat $ intersperse ", " ["int " ++ p | p <- paramNames]
        

indent :: Int -> String -> String
indent n str =
    unlines $
    map (replicate n ' ' ++) $
    lines str

data LGenOpt = Level Integer
             | OutputFile FilePath
             | AlgoVariant LGenAlgoVariant
             | Progress
             | Debug
             | SimplifyBounds
             | SimplifyAll
             | ConvexUnion
             | Simplex
             | TestCode (Maybe String)
             | TestCodeWithCLooG FilePath
             | ReducePath FilePath
               deriving (Eq, Show)

options :: [OptDescr LGenOpt]
options = [Option ['l'] ["level"] (ReqArg (Level . read) "N") "pull IFs N levels out"
          ,Option ['o'] ["output"] (ReqArg OutputFile "OUTPUT") "write output to OUTPUT"
--          ,Option []    ["original"] (NoArg (AlgoVariant LGenOriginal)) "use original (slow but correct) algorithm"
--          ,Option []    ["slightly-improved"] (NoArg (AlgoVariant LGenSlightlyImproved)) "use a slightly improved correct algorithm (default)"
--          ,Option []    ["experimental"] (NoArg (AlgoVariant LGenExperimental)) "use experimental algorithm"
--          ,Option []    ["debug"]    (NoArg Debug)    "show debugging output"
          ,Option []    ["simplify"] (NoArg SimplifyBounds) "simplify generated code"
--          ,Option []    ["simplify"] (NoArg SimplifyAll) "turn on all --simplify-*options"
--          ,Option []    ["convex"] (NoArg ConvexUnion) "use convex description for outer loops"
--          ,Option []    ["simplex"] (NoArg Simplex) "use Simplex to solve LP questions"
          ,Option []    ["progress"] (NoArg Progress) "show some progress information"
          ,Option []    ["testcode"] (OptArg TestCode "value for parameters") "generate test code (incl. CLooG)"
          ,Option []    ["testcode-cloog"] (ReqArg TestCodeWithCLooG "CLooG binary") "call CLooG and put CLooG and NLGen code in output"
          ]

main :: IO ()
main = do args <- getArgs
          mainA args

mainA :: [String] -> IO()
mainA args =
          case getOpt Permute options args of
            (opts,[infile],[]  ) -> handleToolIOError $ run infile opts
            (_,_,errs) ->
                do binary <- getProgName
                   let header = "Usage: " ++ binary ++
                                " [OPTION...] file.nlgen"
                   hPutStrLn stderr (concat errs ++ usageInfo header options)

run :: FilePath -> [LGenOpt] -> IO ()
run input opts =
    do p <- parseFromFile parseLGenInput input
       let istest = or [True | TestCode _ <- opts]
       cloog <- if istest
                then case [b | TestCodeWithCLooG b <- opts] of
                       cloogbin:_ ->
                          do res <- runToolIO False cloogbin ["-loopo",input]
                                              "" idReadFn
                             return $ Just res
                       [] -> return Nothing
                else return Nothing
       let runopts = RunOpts { roTestCode = istest
                             , roTestParamVal = do
                                 s <- listToMaybe [v | TestCode (Just v) <- opts]
                                 case (reads s, reads s) of
                                   ((v,""):_, _)  -> return (Left v)
                                   (_, (vs,""):_) -> return (Right vs)
                                   _              -> Nothing
                             , roCLooGCode = cloog
                             }
       case p of
           -- Force the type variable 'a' to Polynomial Rational
           -- (otherwise, we get Polynomial Double and imprecise
           -- arithmetics).
           Right (lgp,lgoo,nlCtx::[QfFormula (Polynomial Rational)]) ->
               let dims = head ([l | Level l <- opts] ++ [nScatterDims])
                   nScatterDims = case lgpStatements lgp of
                                    [] -> 0
                                    (sys:_,sc):_ -> nVars sc - nVars sys
                   c  = makeCode runopts lgo lgp lgoo nlCtx dims
               in
               (length c) `seq` -- print /* ... */ heador only after code
                                -- generation has finished
               outputFn ("/* Code generated by NLGen version " ++ version
                         ++ ". */\n" ++ c
                        )
           Left err ->
               hPutStrLn stderr ("Parse error: " ++ show err)
    where
    outputFn = head ([writeFile out | OutputFile out <- opts] ++ [putStrLn])
    algo = head ([a | AlgoVariant a <- opts] ++ [LGenOriginal])
    lgo = lgenOptionsStd
            { lgoAlgoVariant = algo
            , lgoLpVariant =
                if Simplex `elem` opts then LGenSimplex else LGenQE
            , lgoProgressToStderr = Progress `elem` opts
            , lgoDebug = Debug `elem` opts
            , lgoSimplifyBounds =
                any (`elem` opts) [SimplifyBounds,SimplifyAll]
            , lgoConvexUnion = ConvexUnion `elem` opts
            }

--
-- Read function for runToolIO.
-- Evaluate the length of the string before
-- returning it to create enough strictness such
-- that the string is read before runToolIO terminates
-- the external process.
--
idReadFn :: String -> Either TIOError String
idReadFn str = length str `seq` Right str

handleToolIOError :: IO a -> IO ()
handleToolIOError action =
    catchDyn (action >> return ())
             (\(tio :: ToolIOError) ->
                 do hPutStrLn stderr ("Tool IO error: " ++ show tio)
             )


testStmtGen :: (GVector a -> String) -> Integer -> [GVector a] -> Maybe String
testStmtGen showExpr stmtID exprs =
    Just $
    concat (["printf(\"S", show stmtID, "("] ++
            intersperse "," (replicate n "%d") ++
            [")\", "] ++
            intersperse ", " (map showExpr exprs) ++
            [");"]
           )
    where
    n = length exprs
    
testCode = undefined


------------------------------------------


--
-- LGenFunctions for  RealFrac a => FracPoly a
--

lgfFracPoly :: RealFrac a => LGenFunctions (FracPoly a) (Polynomial a)
lgfFracPoly = LGF { lgfDisjointUnionDifference = disjointUnionDifferenceFP
                  , lgfNonEmpty = hasIntegralPointFP
                  -- check the denominator of the corresponding IFracPoly,
                  -- because in (1%2) %% 1 we want 2 as the denominator
                  , lgfUnitDenom = (\d -> d == 1 || d == -1) .
                                   ifpDenominator .
                                   fracPolyToIFracPoly . 
                                   mapFP toRational
                  , lgfIsConst = fmap toRational . isConstFP
                  , lgfImplies = impliesIneqsFP

                  , lgfProjectToVarsG = \sys mbEqs _ vars ->
                                        chernikovaProjectToVarsG sys mbEqs vars
                  , lgfSortIneqsGss   = chernikovaSortIneqsGss
                  , lgfSimplifyIneqsG = simplifyIneqsFP
                  , lgfSimplifyDomainG = chernikovaSimplifyDomainG
                  }

impliesIneqsFP :: RealFrac a
             => DTreeContext (FracPoly a) (Polynomial a)
             -> GIneqSystem (FracPoly a)
             -> Bool
             -> GIneqSystem (FracPoly a)
             -> QfFormula (FracPoly a)
impliesIneqsFP dtc assump weak cond =
--    trace ("impliesIneqs: " ++ show (assump,cond)) $
    let equiv = fmap (fromPoly . deleteIndeterminates nVP) $
                reduceEliminateQuantifiers (dtcContext dtc) fo in
--    trace ("equiv = " ++ show equiv) $
    equiv
    where
      fo = FoAll [0..nVP-1] $ FoQf (assumpF `qfImpl` condF)
      assumpF = ineqSystemToFormula assump'
      condF   = if weak then
                    ineqSystemToFormula cond'
                else
                    ineqSystemToStrictFormula cond'
      assump' = unfracIneqSystem Nothing assump
      cond'   = unfracIneqSystem Nothing cond
      nVP = nVars cond + nParams cond


hasIntegralPointFP :: RealFrac a
                   => DTreeContext (FracPoly a) (Polynomial a)
                   -> Maybe (GIneqSystem (FracPoly a)) -- context with variables
                   -> GIneqSystem (FracPoly a)
                   -> Bool
hasIntegralPointFP dtc ctx sys =
    or $
    leavesDTree $
    makeDTree dtc $
    fCond (impliesIneqsFP dtc sys True empty) (dLeaf False) (dLeaf True)
    where
      empty = infeasibleIneqSystem (nVars sys) (nParams sys)

simplifyIneqsFP :: RealFrac a
                => Maybe (GIneqSystem (FracPoly a)) -- context with variables
                -> GIneqSystem (FracPoly a)
                -> DTreeGen (FracPoly a)
                            (Polynomial a)
                            (Maybe (GIneqSystem (FracPoly a)))
simplifyIneqsFP = chernikovaSimplifyIneqsG
{-
simplifyIneqsFP _ sys = -- return sys -- naja
    do dtc <- getDTreeContext
       let sel keep [] = multijoin universe keep
           sel keep (x:xs)
               | impl (keep++xs) x  =  sel    keep  xs
               | otherwise          =  sel (x:keep) xs

           impl as b =
               and $
               leavesDTree $
               makeDTree dtc $
               fCond (impliesIneqsFP dtc (multijoin universe as) True b)
                     (dLeaf True)
                     (dLeaf False)
       return $ Just $ sel [] $ unjoin sys
    where
    universe = ineqSystem ((nV,nP), zeroMatrix 0 (nV+nP+1))
    nV = nVars sys
    nP = nParams sys
-}

disjointUnionDifferenceFP :: RealFrac a
                          => Maybe (GIneqSystem (FracPoly a))
                          -> [GIneqSystem (FracPoly a)]
                          -> [GIneqSystem (FracPoly a)]
                          -> [GIneqSystem (FracPoly a)]
disjointUnionDifferenceFP mbExtCtx ins outs =
    reduceDisjointUnionDifference [] ins outs



------------------------------------------


test2 = lgen lgenOptionsStd
             lgfFracPoly
             1
             (ineqSystem ((0,1), zeroMatrix 0 2))
             [([ineqSystem ((1,1), matrix [[1,0,0], [-1,1,0]])]
              ,ineqSystem ((2,1), matrix [[-1, 2, 0, 0]
                                         ,[ 1,-2, 0, 0]
                                         ])
              )
             ]

noninjTest = lgen lgenOptionsStd
                  lgfFracPoly
                  1
                  (ineqSystem ((0,1), zeroMatrix 0 2))
                  [([ineqSystem ((2,1), matrix [[1,0,0,0], [-1,0,1,0], [0,1,0,0],[0,-1,0,10]])]
                   ,ineqSystem ((3,1), matrix [[-1, 1, 1, 0, 0]
                                              ,[ 1,-1,-1, 0, 0]
                                              ])
                   )
                  ]

noninjCode = 
 runLGenWithStatements 
   (lgenOptionsStd { lgoAlgoVariant = LGenOriginal })
   (lgfFracPoly :: LGenFunctions (FracPoly Rational) (Polynomial Rational))
   (LGP { lgpContext = (ineqSystem ((0,1), zeroMatrix 0 2))
        , lgpStatements =
                  [([ineqSystem ((2,1), matrix [[1,0,0,0], [-1,0,1,0], [0,1,0,0],[0,-1,0,10]])]
                   ,ineqSystem ((3,1), matrix [[-1, 1, 1, 0, 0]
                                              ,[ 1,-1,-1, 0, 0]
                                              ])
                   )
                  ]
        }
   )
   (fracPolyLGOO ["i","j"] ["c"] ["M"] [])
   (fracPolyDTreeContext [])
   (showPolyC [])
   1

test10_20 =
    lgen (lgenOptionsStd { lgoAlgoVariant = LGenOriginal, lgoLpVariant = LGenSimplex })
         lgfFracPoly
         1
         (ineqSystem ((0,0), zeroMatrix 0 1))
         [([ineqSystem ((1,0), matrix [[1,0], [-1,10]])]
          ,ineqSystem ((2,0), matrix [[1,-1,0], [-1,1,0]])
          )
         ,([ineqSystem ((1,0), matrix [[1,0], [-1,10]])]
          ,ineqSystem ((2,0), matrix [[1,-2,0], [-1,2,0]])
          )
         ]

a,b :: GIneqSystem (FracPoly Rational)
a = ineqSystem ((1,2), matrix [[1,0,0,0], [-1,1,0,0]])
b = ineqSystem ((1,2), matrix [[1,0,0,0], [-1,0,2,0]])

           
testM_N2 :: RealFrac a => DTreeGen (FracPoly a) (Polynomial a) (LCode (FracPoly a))
testM_N2 =
    lgen (lgenOptionsStd { lgoAlgoVariant = LGenOriginal })
         lgfFracPoly
         1
         (ineqSystem ((0,2), zeroMatrix 0 3))
         [([ineqSystem ((1,2), matrix [[1,0,0,0], [-1,1,0,0]])]
          ,ineqSystem ((2,2), matrix [[1,-1,0,0,0], [-1,1,0,0,0]])
          )
         ,([ineqSystem ((1,2), matrix [[1,0,0,0], [-1,0,1,0]])]
          ,ineqSystem ((2,2), matrix [[1,-2,0,0,0], [-1,2,0,0,0]])
          )
         ]

codeM_N2 = 
 runLGenWithStatements 
   (lgenOptionsStd { lgoAlgoVariant = LGenOriginal })
   (lgfFracPoly :: LGenFunctions (FracPoly Rational) (Polynomial Rational))
   (LGP { lgpContext = ineqSystem ((0,2), zeroMatrix 0 3)
        , lgpStatements =
            [([ineqSystem ((1,2), matrix [[1,0,0,0], [-1,1,0,0]])]
             ,ineqSystem ((2,2), matrix [[1,-1,0,0,0], [-1,1,0,0,0]])
             )
            ,([ineqSystem ((1,2), matrix [[1,0,0,0], [-1,0,1,0]])]
             ,ineqSystem ((2,2), matrix [[1,-2,0,0,0], [-1,2,0,0,0]])
             )
            ]
        }
   )
   (fracPolyLGOO ["i"] ["c"] ["M","N"] [])
   (fracPolyDTreeContext[])
   (showPolyC [])
   1

testM_N2' :: DTreeGen (FracPoly Rational) (Polynomial Rational) (LCode (FracPoly Rational))
testM_N2' =
    lgen (lgenOptionsStd { lgoAlgoVariant = LGenOriginal, lgoLpVariant = LGenQE })
         lgfFracPoly
         1
         (ineqSystem ((0,0), zeroMatrix 0 1))
         [([ineqSystem ((1,0), matrix [[1,0], [-1,m%%1]])]
          ,ineqSystem ((2,0), matrix [[1,-1,0], [-1,1,0]])
          )
         ,([ineqSystem ((1,0), matrix [[1,0], [-1,n%%1]])]
          ,ineqSystem ((2,0), matrix [[1,-2,0], [-1,2,0]])
          )
         ]
    where
      m = indeterminate 2 0 :: Polynomial Rational
      n = indeterminate 2 1

codeM_N2' = 
 runLGenWithStatements 
   (lgenOptionsStd { lgoAlgoVariant = LGenOriginal, lgoLpVariant = LGenQE })
   lgfFracPoly
   (LGP { lgpContext = ineqSystem ((0,0), zeroMatrix 0 1)
        , lgpStatements =
            [([ineqSystem ((1,0), matrix [[1,0], [-1,m%%1]])]
             ,ineqSystem ((2,0), matrix [[1,-1,0], [-1,1,0]])
             )
            ,([ineqSystem ((1,0), matrix [[1,0], [-1,n%%1]])]
             ,ineqSystem ((2,0), matrix [[1,-2,0], [-1,2,0]])
             )
            ]
        }
   )
   (fracPolyLGOO ["i"] ["c"] [] ["M","N"])
   (fracPolyDTreeContext [QfAt (AtGt n), QfAt (AtGe (m-1))])
   (showPolyC ["M","N"] . snd . ratPoly2intPoly)
   1
 where
 m = indeterminate 2 0 :: Polynomial Rational
 n = indeterminate 2 1


codeM_NP' = 
 runLGenWithStatements 
   (lgenOptionsStd { lgoAlgoVariant = LGenOriginal, lgoLpVariant = LGenQE })
   lgfFracPoly
   (LGP { lgpContext = ineqSystem ((0,0), zeroMatrix 0 1)
        , lgpStatements =
            [([ineqSystem ((1,0), matrix [[1,0], [-1,m%%1]])]
             ,ineqSystem ((2,0), matrix [[1,-1,0], [-1,1,0]])
             )
            ,([ineqSystem ((1,0), matrix [[1,0], [-1,n%%1]])]
             ,ineqSystem ((2,0), matrix [[1,-p%%1,0], [-1,p%%1,0]])
             )
            ]
        }
   )
   (fracPolyLGOO ["i"] ["c"] [] ["M","N","P"])
   (fracPolyDTreeContext [QfAt (AtGt n), QfAt (AtGe (m-1)), QfAt (AtGt p)])
   (showPolyC ["M","N","P"] . snd . ratPoly2intPoly)
   1
 where
 m = indeterminate 3 0 :: Polynomial Rational
 n = indeterminate 3 1
 p = indeterminate 3 2




code12 = 
 runLGenWithStatements 
   (lgenOptionsStd { lgoAlgoVariant = LGenOriginal })
   (lgfFracPoly :: LGenFunctions (FracPoly Rational) (Polynomial Rational))
   (LGP { lgpContext = ineqSystem ((0,1), zeroMatrix 0 2)
        , lgpStatements =
            [([ineqSystem ((1,1), matrix [[1,0,0], [-1,1,0]])]
             ,ineqSystem ((3,1), matrix [[1,0,-1,0,0], [-1,0,1,0,0]
                                        ,[0,1, 0,0,0], [0,-1,0,0,0]])
             )
            ,([ineqSystem ((2,1), matrix [[1,0,0,0], [-1,0,1,0]
                                         ,[0,1,0,0], [0,-1,1,0]])]
             ,ineqSystem ((4,1), matrix [[1,0,-1,0,0,0], [-1,0,1,0,0,0]
                                        ,[0,1,0,-1,0,0], [0,-1,0,1,0,0]])
             )
            ]
        }
   )
   (fracPolyLGOO ["i","j"] ["c","d"] ["M"] [])
   (fracPolyDTreeContext[])
   (showPolyC [])
   1

